# -*- coding: utf-8 -*-
# pylint: disable=line-too-long, unused-wildcard-import
"""
The *HydPy-Musk* model family member |musk_mct| realises a modification of the
variable coefficient Muskingum approach of :cite:t:`ref-Cunge1969` (*MC*) proposed by
:cite:t:`ref-Todini2007` (*MCT*).

.. _`Main`: https://de.wikipedia.org/wiki/Main

Opposed to the original Muskingum method of :cite:p:`ref-McCarthy1940`, *MC* does not
rely on predetermined coefficients.  Instead, it calculates the coefficients for each
simulation step based on a water level-discharge relationship.  Essentially, *MC* is a
kinematic wave approach that does not explicitly account for pressure gradients or
other forcings than the channel's bottom slope explicitly.  However, it adjusts its
coefficients so that the numerical error of the underlying finite difference
discretisation (the Muskingum working formula) mimics the physical diffusion of flood
waves due to pressure gradients as good as possible.   Hence, it can simulate flood
waves even in channels with mild bottom slopes without much calibration (provided that
information on the rating curve is available).  Compared to more complete solutions of
the Saint-Vernant equations, it is less computationally demanding.  On the downside, it
cannot deal with cases where, for example, effects or rapid changes in channel geometry
play a significant role.

|musk_mct| implements the Muskingum-Cunge modification of :cite:t:`ref-Todini2007`,
which solves a "mass conservation inconsistency" and a "steady-state inconsistency" of
the original formulation.  It applies the Gauckler-Manning-Strickler equation in
several channel segments with a trapezoidal profile.  |musk_mct| allows to define the
hydraulic parameters (for example, the bottom width or the Strickler coefficient) for
each segment individually.

When applying |musk_mct|, one should be aware that it relies on two nested iterations.
First, it uses a root search algorithm to find the current "reference water level" that
results in the previously determined "reference discharge".  While
:cite:t:`ref-Todini2007` uses the Newton-Raphson approach, |musk_mct| applies the
Pegasus iteration :cite:p:`ref-Dowell1972`.  By default, |musk_mct| strives for very
high accuracies.  There are good chances you can increase the computation speed
significantly in your use case without reducing numerical precision too much (see the
documentation on method |Calc_ReferenceWaterLevel_V1| for further information).  The
second iteration addresses the "reference discharge" itself, which's initial guess can
be quite inaccurate (see |Calc_ReferenceDischarge_V1|).  :cite:t:`ref-Todini2007`
suggests refining it by starting subsequent runs of the whole *MCT* procedure
(including the Newton-Raphson or Pegasus approach described above).  |musk_mct| follows
Todini's suggestion to run everything twice.  You can deviate from this default
behaviour by changing the value of parameter |musk_mct.NmbRuns|.


Integration tests
=================

.. how_to_understand_integration_tests::

All of the following tests are equal to or similar to the examples provided by
:cite:t:`ref-Todini2007`.

The simulation period covers four days; the simulation step is 30 minutes:

>>> from hydpy import pub, Nodes, Element
>>> pub.timegrids = "2000-01-01", "2000-01-05", "30m"

The model shall retrieve its inflow from two |Node| objects (`input1` and `input2`)
and passe its outflow to node `output`.  We define these nodes and connect them to an
|Element| object that handles a |musk_mct| instance:

>>> nodes = Nodes("input1", "input2", "output")
>>> stream = Element("stream", inlets=["input1", "input2"], outlets="output")
>>> from hydpy.models.musk_mct import *
>>> parameterstep("1h")
>>> stream.model = model

The last general preparation step is to define a test function object for controlling
the following test runs:

>>> from hydpy.core.testtools import IntegrationTest
>>> test = IntegrationTest(stream)
>>> test.plotting_options.axis1 = (states.discharge,)
>>> test.dateformat = "%H:%M"

.. _musk_mct_base_example:

Base example
_____________

This example repeats the base example of :cite:t:`ref-Todini2007` dealing with a
trapezoidal profile.  A channel with a total length of 100 km is subdivided into 50
segments:

>>> nmbsegments(50)
>>> length(2.0)

All segments have the same geometry and the same roughness:

>>> bottomwidth(15.0)
>>> sideslope(5.0)
>>> bottomslope(0.00025)
>>> stricklercoefficient(1.0/0.035)

|musk_mct| uses the catchment area for calculating the value of the solver parameter
|ToleranceDischarge|, which determines the accuracy of the iterative estimation of the
reference water level (see methods |ToleranceDischarge.modify_init| and
|Calc_ReferenceWaterLevel_V1|).  We set it 25 000 km², which is approximately the
catchment size of the river `Main`_ at gauge Frankfurt-Osthafen:

>>> catchmentarea(25000.0)

This setting makes up for an error tolerance of 0.025 m³/s, which is 360 times smaller
than the lowest discharge reported for Frankfurt-Ostenhagen (9 m³/s, 1976):

>>> solver.tolerancedischarge.update()
>>> solver.tolerancedischarge
tolerancedischarge(0.025)

Besides "old" discarge values, |musk_mct| requires the Courant number and the cell
Reynolds number as initial conditions. These states must be consistent if you need good
simulation results right from the start.  If there are none available from a previous
simulation run, you can use |Model.prepare_states| to calculate them for definable
discharge values under the assumption of stationarity:

>>> model.prepare_states(discharge=100.0)
>>> test.inits = ((states.discharge, 100.0),
...               (states.courantnumber, states.courantnumber[0]),
...               (states.reynoldsnumber, states.reynoldsnumber[0]))

Node `input1` provides a constant baseflow of 100 m³/s, and node `input2` provides a
synthetic flood wave with a peak discharge of 800 m³/s, making a total peak discharge
of 900 m³/s (which is similar to the average flood discharge at Frankfurt-Ostenhagen)
that occurs 24 hours after the start of the simulation period:

>>> import numpy
>>> q_base, q_peak, t_peak, β = 100.0, 900.0, 24.0, 16.0
>>> ts = pub.timegrids.init.to_timepoints()
>>> nodes.input1.sequences.sim.series = q_base
>>> nodes.input2.sequences.sim.series = (
...     (q_peak - q_base) * ((ts / t_peak) * numpy.exp(1.0 - ts / t_peak)) ** β)

The following table and figure show the discharge calculated for all 51 channel segment
endpoints, including the start and endpoint of the complete channel.  Other properties
like the reference water level, the wetted perimeter, or the Courant number represent
estimated averages for all 50 segments:

.. integration-test::

    >>> conditions = test("musk_mct_base_example", get_conditions="2000-01-01")
    |  date |                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                           referencewaterlevel |                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                             wettedarea |                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                             wettedperimeter |                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                             surfacewidth |                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                           celerity |                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                           correctingfactor |                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                           coefficient1 |                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                            coefficient2 |                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                           coefficient3 |     inflow |                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                             referencedischarge |    outflow |                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                           courantnumber |                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                           reynoldsnumber |                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                          discharge | input1 |     input2 |     output |
    -------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
    | 00:00 | 3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828             3.717828 | 124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652 |  52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556        52.914556 |  52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281     52.178281 | 1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385 | 1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161          1.359161 | 0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158      0.536158 | -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868     -0.201868 | 0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709      0.665709 |      100.0 |      100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0               100.0 |      100.0 |   0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207         0.7207 | 2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116        2.591116 |      100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0 |  100.0 |        0.0 |      100.0 |
    | 00:30 | 3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828             3.717828 | 124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652 |  52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556        52.914556 |  52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281     52.178281 | 1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385 | 1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161          1.359161 | 0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158      0.536158 | -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868     -0.201868 | 0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709      0.665709 |      100.0 |      100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0               100.0 |      100.0 |   0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207         0.7207 | 2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116        2.591116 |      100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0 |  100.0 |        0.0 |      100.0 |
    | 01:00 | 3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828             3.717828 | 124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652 |  52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556        52.914556 |  52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281     52.178281 | 1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385 | 1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161          1.359161 | 0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158      0.536158 | -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868     -0.201868 | 0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709      0.665709 |      100.0 |      100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0               100.0 |      100.0 |   0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207         0.7207 | 2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116        2.591116 |      100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0 |  100.0 |        0.0 |      100.0 |
    | 01:30 | 3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828             3.717828 | 124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652 |  52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556        52.914556 |  52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281     52.178281 | 1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385 | 1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161          1.359161 | 0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158      0.536158 | -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868     -0.201868 | 0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709      0.665709 |      100.0 |      100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0               100.0 |      100.0 |   0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207         0.7207 | 2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116        2.591116 |      100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0 |  100.0 |        0.0 |      100.0 |
    | 02:00 | 3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828             3.717828 | 124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652 |  52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556        52.914556 |  52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281     52.178281 | 1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385 | 1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161          1.359161 | 0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158      0.536158 | -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868     -0.201868 | 0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709      0.665709 |      100.0 |      100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0               100.0 |      100.0 |   0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207         0.7207 | 2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116        2.591116 |      100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0 |  100.0 |        0.0 |      100.0 |
    | 02:30 | 3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828             3.717828 | 124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652 |  52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556        52.914556 |  52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281     52.178281 | 1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385 | 1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161          1.359161 | 0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158      0.536158 | -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868     -0.201868 | 0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709      0.665709 |      100.0 |      100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0               100.0 |      100.0 |   0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207         0.7207 | 2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116        2.591116 |      100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0 |  100.0 |        0.0 |      100.0 |
    | 03:00 | 3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828             3.717828 | 124.878655  124.878654  124.878653  124.878653  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652 |  52.914557   52.914557   52.914557   52.914557   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556        52.914556 |  52.178282   52.178282   52.178282   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281     52.178281 | 1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385 | 1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161          1.359161 | 0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158      0.536158 | -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868     -0.201868 | 0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709      0.665709 | 100.000003 | 100.000003  100.000002  100.000001  100.000001  100.000001       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0               100.0 |      100.0 |   0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207         0.7207 | 2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116        2.591116 | 100.000003  100.000002  100.000001  100.000001  100.000001       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0 |  100.0 |   0.000003 |      100.0 |
    | 03:30 | 3.717829  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828             3.717828 | 124.878674  124.878666  124.878661  124.878658  124.878656  124.878655  124.878654  124.878653  124.878653  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652 |  52.914561   52.914559   52.914558   52.914558   52.914557   52.914557   52.914557   52.914557   52.914557   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556        52.914556 |  52.178286   52.178284   52.178283   52.178282   52.178282   52.178282   52.178282   52.178282   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281     52.178281 | 1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385 | 1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161          1.359161 | 0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158      0.536158 | -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868     -0.201868 | 0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709      0.665709 | 100.000029 | 100.000024  100.000016   100.00001  100.000007  100.000004  100.000003  100.000002  100.000001  100.000001  100.000001       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0               100.0 |      100.0 |   0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207         0.7207 | 2.591117  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116        2.591116 | 100.000029  100.000019  100.000012  100.000008  100.000005  100.000003  100.000002  100.000001  100.000001  100.000001       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0 |  100.0 |   0.000029 |      100.0 |
    | 04:00 | 3.717831   3.71783  3.717829  3.717829  3.717829  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828             3.717828 | 124.878787  124.878741  124.878711  124.878691  124.878678  124.878669  124.878663  124.878659  124.878657  124.878655  124.878654  124.878653  124.878653  124.878653  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652 |  52.914583   52.914574   52.914568   52.914564   52.914561    52.91456   52.914559   52.914558   52.914557   52.914557   52.914557   52.914557   52.914557   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556        52.914556 |  52.178307   52.178298   52.178293   52.178289   52.178286   52.178285   52.178283   52.178283   52.178282   52.178282   52.178282   52.178282   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281     52.178281 | 1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385 | 1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161          1.359161 | 0.536159  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158      0.536158 | -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868     -0.201868 | 0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709      0.665709 | 100.000175 | 100.000147  100.000097  100.000064  100.000042  100.000028  100.000018  100.000012  100.000008  100.000005  100.000003  100.000002  100.000002  100.000001  100.000001       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0               100.0 |      100.0 |   0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207         0.7207 | 2.591118  2.591117  2.591117  2.591117  2.591117  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116        2.591116 | 100.000175  100.000116  100.000076   100.00005  100.000033  100.000022  100.000014   100.00001  100.000006  100.000004  100.000003  100.000002  100.000001  100.000001  100.000001       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0 |  100.0 |   0.000175 |      100.0 |
    | 04:30 |  3.71784  3.717836  3.717834  3.717832  3.717831   3.71783  3.717829  3.717829  3.717829  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828             3.717828 | 124.879291  124.879078  124.878937  124.878842  124.878779  124.878736  124.878708  124.878689  124.878677  124.878669  124.878663  124.878659  124.878657  124.878655  124.878654  124.878653  124.878653  124.878653  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652 |  52.914681    52.91464   52.914612   52.914593   52.914581   52.914573   52.914567   52.914564   52.914561    52.91456   52.914559   52.914558   52.914557   52.914557   52.914557   52.914557   52.914557   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556        52.914556 |  52.178404   52.178363   52.178336   52.178318   52.178306   52.178297   52.178292   52.178288   52.178286   52.178284   52.178283   52.178283   52.178282   52.178282   52.178282   52.178282   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281     52.178281 | 1.088387  1.088386  1.088386  1.088386  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385 |  1.35916   1.35916  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161          1.359161 | 0.536159  0.536159  0.536159  0.536159  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158      0.536158 | -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868     -0.201868 | 0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709      0.665709 | 100.000826 | 100.000695  100.000464   100.00031  100.000207  100.000138  100.000092  100.000061  100.000041  100.000027  100.000018  100.000012  100.000008  100.000005  100.000004  100.000002  100.000002  100.000001  100.000001       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0               100.0 |      100.0 | 0.720701  0.720701    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207         0.7207 | 2.591124  2.591121   2.59112  2.591119  2.591118  2.591117  2.591117  2.591117  2.591117  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116        2.591116 | 100.000826  100.000552  100.000368  100.000246  100.000164  100.000109  100.000073  100.000048  100.000032  100.000021  100.000014  100.000009  100.000006  100.000004  100.000003  100.000002  100.000001  100.000001  100.000001       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0 |  100.0 |   0.000826 |      100.0 |
    | 05:00 | 3.717876   3.71786   3.71785  3.717843  3.717838  3.717835  3.717833  3.717831   3.71783  3.717829  3.717829  3.717829  3.717829  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828             3.717828 | 124.881131  124.880327  124.879783  124.879415  124.879166  124.878998  124.878885  124.878809  124.878758  124.878723    124.8787  124.878684  124.878673  124.878666  124.878662  124.878658  124.878656  124.878655  124.878654  124.878653  124.878653  124.878653  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652 |  52.915041   52.914884   52.914777   52.914705   52.914657   52.914624   52.914602   52.914587   52.914577    52.91457   52.914566   52.914563   52.914561   52.914559   52.914558   52.914558   52.914557   52.914557   52.914557   52.914557   52.914557   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556        52.914556 |  52.178756   52.178602   52.178498   52.178428    52.17838   52.178348   52.178326   52.178311   52.178302   52.178295    52.17829   52.178287   52.178285   52.178284   52.178283   52.178283   52.178282   52.178282   52.178282   52.178282   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281     52.178281 | 1.088392   1.08839  1.088388  1.088387  1.088387  1.088386  1.088386  1.088386  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385 |  1.35916   1.35916   1.35916   1.35916   1.35916   1.35916  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161          1.359161 | 0.536162  0.536161   0.53616  0.536159  0.536159  0.536159  0.536159  0.536159  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158      0.536158 | -0.201869  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868     -0.201868 | 0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709      0.665709 | 100.003194 | 100.002698  100.001823  100.001231   100.00083   100.00056  100.000377  100.000254  100.000171  100.000115  100.000077  100.000052  100.000035  100.000023  100.000016   100.00001  100.000007  100.000005  100.000003  100.000002  100.000001  100.000001  100.000001       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0               100.0 |      100.0 | 0.720705  0.720703  0.720702  0.720701  0.720701    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207         0.7207 | 2.591146  2.591136   2.59113  2.591125  2.591122   2.59112  2.591119  2.591118  2.591118  2.591117  2.591117  2.591117  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116        2.591116 | 100.003194  100.002158  100.001458  100.000984  100.000663  100.000447  100.000301  100.000203  100.000136  100.000092  100.000062  100.000041  100.000028  100.000019  100.000012  100.000008  100.000006  100.000004  100.000003  100.000002  100.000001  100.000001       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0 |  100.0 |   0.003194 |      100.0 |
    | 05:30 | 3.717985  3.717936  3.717902  3.717878  3.717862  3.717851  3.717844  3.717839  3.717835  3.717833  3.717832   3.71783   3.71783  3.717829  3.717829  3.717829  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828             3.717828 | 124.886847  124.884256   124.88248  124.881265  124.880434  124.879866  124.879478  124.879214  124.879034  124.878911  124.878828  124.878771  124.878733  124.878707  124.878689  124.878677  124.878669  124.878663   124.87866  124.878657  124.878655  124.878654  124.878654  124.878653  124.878653  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652 |  52.916158   52.915652   52.915305   52.915067   52.914905   52.914794   52.914718   52.914666   52.914631   52.914607   52.914591    52.91458   52.914572   52.914567   52.914564   52.914561    52.91456   52.914559   52.914558   52.914557   52.914557   52.914557   52.914557   52.914557   52.914557   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556        52.914556 |  52.179852   52.179355   52.179015   52.178782   52.178623   52.178514    52.17844   52.178389   52.178355   52.178331   52.178315   52.178304   52.178297   52.178292   52.178288   52.178286   52.178285   52.178283   52.178283   52.178282   52.178282   52.178282   52.178282   52.178282   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281     52.178281 | 1.088409  1.088402  1.088396  1.088393   1.08839  1.088389  1.088387  1.088387  1.088386  1.088386  1.088386  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385 | 1.359159  1.359159   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161          1.359161 | 0.536171  0.536167  0.536164  0.536162  0.536161   0.53616   0.53616  0.536159  0.536159  0.536159  0.536159  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158      0.536158 | -0.201871   -0.20187  -0.201869  -0.201869  -0.201869  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868     -0.201868 | 0.665708  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709      0.665709 | 100.010515 |  100.00892  100.006099  100.004167  100.002844  100.001939  100.001321    100.0009  100.000612  100.000416  100.000282  100.000192   100.00013  100.000088   100.00006   100.00004  100.000027  100.000018  100.000012  100.000008  100.000006  100.000004  100.000003  100.000002  100.000001  100.000001  100.000001       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0               100.0 |      100.0 | 0.720717  0.720711  0.720708  0.720705  0.720703  0.720702  0.720701  0.720701    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207         0.7207 | 2.591214  2.591183  2.591162  2.591148  2.591138  2.591131  2.591126  2.591123  2.591121  2.591119  2.591118  2.591118  2.591117  2.591117  2.591117  2.591117  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116        2.591116 | 100.010515  100.007193  100.004915  100.003356  100.002289   100.00156  100.001062  100.000723  100.000491  100.000334  100.000227  100.000154  100.000104  100.000071  100.000048  100.000032  100.000022  100.000015   100.00001  100.000007  100.000005  100.000003  100.000002  100.000001  100.000001  100.000001       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0 |  100.0 |   0.010515 |      100.0 |
    | 06:00 | 3.718283  3.718143  3.718046  3.717978  3.717932    3.7179  3.717877  3.717862  3.717851  3.717844  3.717839  3.717836  3.717833  3.717832  3.717831   3.71783  3.717829  3.717829  3.717829  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828             3.717828 | 124.902381  124.895076  124.890007  124.886494  124.884062   124.88238  124.881218  124.880417  124.879865  124.879484  124.879223  124.879043   124.87892  124.878835  124.878777  124.878737   124.87871  124.878692  124.878679   124.87867  124.878665  124.878661  124.878658  124.878656  124.878655  124.878654  124.878653  124.878653  124.878653  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652 |  52.919194   52.917766   52.916776   52.916089   52.915614   52.915285   52.915058   52.914901   52.914793   52.914719   52.914668   52.914633   52.914609   52.914592   52.914581   52.914573   52.914568   52.914564   52.914562    52.91456   52.914559   52.914558   52.914558   52.914557   52.914557   52.914557   52.914557   52.914557   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556        52.914556 |  52.182829   52.181429   52.180458   52.179784   52.179318   52.178996   52.178773    52.17862   52.178514   52.178441   52.178391   52.178356   52.178333   52.178316   52.178305   52.178298   52.178292   52.178289   52.178286   52.178285   52.178284   52.178283   52.178282   52.178282   52.178282   52.178282   52.178282   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281     52.178281 | 1.088456  1.088434  1.088419  1.088408  1.088401  1.088396  1.088393   1.08839  1.088389  1.088388  1.088387  1.088386  1.088386  1.088386  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385 | 1.359156  1.359157  1.359158  1.359159  1.359159   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161          1.359161 | 0.536194  0.536183  0.536175   0.53617  0.536166  0.536164  0.536162  0.536161   0.53616   0.53616  0.536159  0.536159  0.536159  0.536159  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158      0.536158 |  -0.20188  -0.201876  -0.201873  -0.201872   -0.20187   -0.20187  -0.201869  -0.201869  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868     -0.201868 | 0.665707  0.665707  0.665708  0.665708  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709      0.665709 | 100.030315 | 100.025827  100.017877  100.012359  100.008535  100.005888  100.004057  100.002793  100.001921   100.00132  100.000906  100.000621  100.000426  100.000291  100.000199  100.000136  100.000093  100.000064  100.000043   100.00003   100.00002  100.000014  100.000009  100.000006  100.000004  100.000003  100.000002  100.000001  100.000001  100.000001       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0               100.0 |      100.0 | 0.720749  0.720734  0.720723  0.720716  0.720711  0.720707  0.720705  0.720703  0.720702  0.720701  0.720701    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207         0.7207 | 2.591401  2.591313  2.591252   2.59121  2.591181  2.591161  2.591147  2.591137  2.591131  2.591126  2.591123  2.591121  2.591119  2.591118  2.591118  2.591117  2.591117  2.591117  2.591117  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116        2.591116 | 100.030315  100.020992  100.014519  100.010031  100.006922  100.004772  100.003286  100.002261  100.001554  100.001067  100.000732  100.000502  100.000344  100.000235  100.000161   100.00011  100.000075  100.000051  100.000035  100.000024  100.000016  100.000011  100.000007  100.000005  100.000003  100.000002  100.000002  100.000001  100.000001       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0 |  100.0 |   0.030315 |      100.0 |
    | 06:30 | 3.719006  3.718653  3.718406  3.718232   3.71811  3.718024  3.717965  3.717923  3.717894  3.717874   3.71786   3.71785  3.717843  3.717839  3.717835  3.717833  3.717832  3.717831   3.71783  3.717829  3.717829  3.717829  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828             3.717828 | 124.940113  124.921719  124.908788  124.899711  124.893349  124.888896  124.885784  124.883611  124.882097  124.881042  124.880308  124.879799  124.879445    124.8792   124.87903  124.878913  124.878832  124.878776  124.878737  124.878711  124.878692   124.87868  124.878671  124.878665  124.878661  124.878658  124.878656  124.878655  124.878654  124.878653  124.878653  124.878653  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652 |  52.926567   52.922973   52.920446   52.918672   52.917429   52.916559    52.91595   52.915526    52.91523   52.915023    52.91488   52.914781   52.914711   52.914663    52.91463   52.914607   52.914592   52.914581   52.914573   52.914568   52.914564   52.914562    52.91456   52.914559   52.914558   52.914558   52.914557   52.914557   52.914557   52.914557   52.914557   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556        52.914556 |  52.190059   52.186535   52.184057   52.182317   52.181098   52.180245   52.179648   52.179232   52.178941   52.178739   52.178599   52.178501   52.178433   52.178386   52.178354   52.178331   52.178316   52.178305   52.178298   52.178293   52.178289   52.178287   52.178285   52.178284   52.178283   52.178282   52.178282   52.178282   52.178282   52.178282   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281     52.178281 | 1.088568  1.088513  1.088475  1.088448  1.088429  1.088415  1.088406    1.0884  1.088395  1.088392   1.08839  1.088388  1.088387  1.088387  1.088386  1.088386  1.088386  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385 | 1.359148  1.359152  1.359155  1.359156  1.359158  1.359159  1.359159   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161          1.359161 | 0.536251  0.536223  0.536204   0.53619  0.536181  0.536174  0.536169  0.536166  0.536164  0.536162  0.536161   0.53616   0.53616  0.536159  0.536159  0.536159  0.536159  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158      0.536158 | -0.201904  -0.201893  -0.201885   -0.20188  -0.201876  -0.201873  -0.201872   -0.20187   -0.20187  -0.201869  -0.201869  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868     -0.201868 | 0.665703  0.665705  0.665706  0.665707  0.665708  0.665708  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709      0.665709 | 100.078179 | 100.066899  100.046877  100.032801  100.022921  100.015997   100.01115  100.007762  100.005398  100.003749  100.002601  100.001803  100.001248  100.000863  100.000597  100.000412  100.000284  100.000196  100.000135  100.000093  100.000064  100.000044   100.00003  100.000021  100.000014   100.00001  100.000007  100.000005  100.000003  100.000002  100.000001  100.000001  100.000001       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0               100.0 |      100.0 | 0.720827  0.720789  0.720762  0.720743   0.72073  0.720721  0.720714   0.72071  0.720707  0.720705  0.720703  0.720702  0.720701  0.720701    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207         0.7207 | 2.591853  2.591632  2.591477  2.591369  2.591292  2.591239  2.591202  2.591176  2.591158  2.591145  2.591136   2.59113  2.591126  2.591123  2.591121  2.591119  2.591118  2.591118  2.591117  2.591117  2.591117  2.591117  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116        2.591116 | 100.078179  100.054809   100.03837  100.026825   100.01873  100.013061  100.009096  100.006328  100.004397  100.003052  100.002116  100.001466  100.001014  100.000701  100.000484  100.000334   100.00023  100.000159  100.000109  100.000075  100.000052  100.000035  100.000024  100.000017  100.000011  100.000008  100.000005  100.000004  100.000002  100.000002  100.000001  100.000001  100.000001       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0 |  100.0 |   0.078179 |      100.0 |
    | 07:00 | 3.720602  3.719796  3.719222  3.718814  3.718524  3.718319  3.718174  3.718071  3.717999  3.717948  3.717912  3.717887  3.717869  3.717857  3.717848  3.717842  3.717838  3.717835  3.717833  3.717831   3.71783   3.71783  3.717829  3.717829  3.717829  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828             3.717828 | 125.023431   124.98136  124.951394  124.930089  124.914969  124.904256  124.896678  124.891326  124.887551  124.884892  124.883023   124.88171  124.880789  124.880143  124.879692  124.879376  124.879156  124.879002  124.878895  124.878821  124.878769  124.878733  124.878708  124.878691  124.878679   124.87867  124.878665  124.878661  124.878658  124.878656  124.878655  124.878654  124.878653  124.878653  124.878653  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652 |  52.942845   52.934627   52.928772   52.924609   52.921654    52.91956   52.918079   52.917033   52.916296   52.915776   52.915411   52.915154   52.914974   52.914848    52.91476   52.914698   52.914655   52.914625   52.914604   52.914589   52.914579   52.914572   52.914567   52.914564   52.914562    52.91456   52.914559   52.914558   52.914558   52.914557   52.914557   52.914557   52.914557   52.914557   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556        52.914556 |  52.206021   52.197962    52.19222   52.188138   52.185241   52.183188   52.181736    52.18071   52.179987   52.179477   52.179119   52.178867   52.178691   52.178567   52.178481    52.17842   52.178378   52.178348   52.178328   52.178314   52.178304   52.178297   52.178292   52.178289   52.178286   52.178285   52.178284   52.178283   52.178282   52.178282   52.178282   52.178282   52.178282   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281     52.178281 | 1.088815   1.08869  1.088601  1.088538  1.088493  1.088461  1.088439  1.088423  1.088411  1.088404  1.088398  1.088394  1.088391  1.088389  1.088388  1.088387  1.088387  1.088386  1.088386  1.088386  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385 | 1.359132   1.35914  1.359146   1.35915  1.359153  1.359156  1.359157  1.359158  1.359159  1.359159   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161          1.359161 | 0.536377  0.536314  0.536268  0.536236  0.536213  0.536197  0.536186  0.536177  0.536172  0.536168  0.536165  0.536163  0.536162  0.536161   0.53616  0.536159  0.536159  0.536159  0.536159  0.536159  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158      0.536158 | -0.201962  -0.201934  -0.201914    -0.2019   -0.20189  -0.201883  -0.201879  -0.201875  -0.201873  -0.201871   -0.20187  -0.201869  -0.201869  -0.201869  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868     -0.201868 | 0.665696    0.6657  0.665702  0.665704  0.665706  0.665707  0.665708  0.665708  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709      0.665709 | 100.183351 | 100.157608  100.111803   100.07918  100.055988  100.039529  100.027868   100.01962  100.013794  100.009686  100.006792  100.004757  100.003328  100.002325  100.001623  100.001132  100.000788  100.000548  100.000381  100.000265  100.000184  100.000127  100.000088  100.000061  100.000042  100.000029   100.00002  100.000014   100.00001  100.000007  100.000005  100.000003  100.000002  100.000001  100.000001  100.000001       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0               100.0 |      100.0 |    0.721  0.720912   0.72085  0.720806  0.720775  0.720753  0.720737  0.720726  0.720718  0.720713  0.720709  0.720706  0.720704  0.720703  0.720702  0.720701  0.720701    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207         0.7207 | 2.592851  2.592347  2.591988  2.591733  2.591551  2.591423  2.591332  2.591268  2.591223  2.591191  2.591169  2.591153  2.591142  2.591134  2.591129  2.591125  2.591122   2.59112  2.591119  2.591118  2.591118  2.591117  2.591117  2.591117  2.591117  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116        2.591116 | 100.183351  100.130144  100.092223  100.065247  100.046091  100.032511    100.0229  100.016108  100.011315  100.007939  100.005563  100.003893  100.002721    100.0019  100.001325  100.000923  100.000643  100.000447   100.00031  100.000215  100.000149  100.000103  100.000072   100.00005  100.000034  100.000024  100.000016  100.000011  100.000008  100.000005  100.000004  100.000003  100.000002  100.000001  100.000001  100.000001       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0 |  100.0 |   0.183351 |      100.0 |
    | 07:30 | 3.723847  3.722152  3.720928  3.720046  3.719413  3.718958  3.718633    3.7184  3.718234  3.718116  3.718032  3.717972   3.71793    3.7179  3.717879  3.717864  3.717853  3.717846   3.71784  3.717837  3.717834  3.717832  3.717831   3.71783   3.71783  3.717829  3.717829  3.717829  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828             3.717828 | 125.192889  125.104349  125.040441  124.994416  124.961338  124.937613  124.920628   124.90849   124.89983  124.893661  124.889275   124.88616  124.883952  124.882388  124.881283  124.880502  124.879952  124.879564  124.879291    124.8791  124.878965  124.878871  124.878805  124.878758  124.878726  124.878704  124.878688  124.878677  124.878669  124.878664   124.87866  124.878658  124.878656  124.878655  124.878654  124.878653  124.878653  124.878653  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652 |  52.975937    52.95865   52.946168   52.937177   52.930715   52.926079    52.92276   52.920388   52.918695    52.91749   52.916633   52.916024   52.915592   52.915287   52.915071   52.914918    52.91481   52.914735   52.914681   52.914644   52.914618   52.914599   52.914586   52.914577   52.914571   52.914566   52.914563   52.914561    52.91456   52.914559   52.914558   52.914557   52.914557   52.914557   52.914557   52.914557   52.914557   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556        52.914556 |   52.23847   52.221518   52.209279   52.200463   52.194126    52.18958   52.186325   52.183999    52.18234   52.181158   52.180317    52.17972   52.179297   52.178997   52.178786   52.178636    52.17853   52.178456   52.178404   52.178367   52.178341   52.178323   52.178311   52.178302   52.178296   52.178291   52.178288   52.178286   52.178285   52.178284   52.178283   52.178282   52.178282   52.178282   52.178282   52.178282   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281     52.178281 | 1.089318  1.089055  1.088866  1.088729  1.088631   1.08856   1.08851  1.088474  1.088448   1.08843  1.088417  1.088407  1.088401  1.088396  1.088393  1.088391  1.088389  1.088388  1.088387  1.088386  1.088386  1.088386  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385 | 1.359099  1.359116  1.359129  1.359138  1.359144  1.359149  1.359152  1.359155  1.359156  1.359158  1.359158  1.359159   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161          1.359161 | 0.536633  0.536499  0.536403  0.536333  0.536283  0.536247  0.536222  0.536203   0.53619  0.536181  0.536174   0.53617  0.536166  0.536164  0.536162  0.536161   0.53616   0.53616  0.536159  0.536159  0.536159  0.536159  0.536159  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158      0.536158 | -0.202092  -0.202027   -0.20198  -0.201947  -0.201924  -0.201907  -0.201895  -0.201887  -0.201881  -0.201877  -0.201874  -0.201872  -0.201871   -0.20187  -0.201869  -0.201869  -0.201869  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868     -0.201868 | 0.665685  0.665691  0.665696    0.6657  0.665702  0.665704  0.665706  0.665707  0.665707  0.665708  0.665708  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709      0.665709 | 100.396214 |  100.34216  100.245722  100.176129  100.126016  100.090006  100.064178  100.045689  100.032476   100.02305  100.016337  100.011562  100.008172  100.005769  100.004067  100.002863  100.002014  100.001415  100.000993  100.000696  100.000487  100.000341  100.000238  100.000166  100.000116  100.000081  100.000056  100.000039  100.000027  100.000019  100.000013  100.000009  100.000006  100.000004  100.000003  100.000002  100.000001  100.000001  100.000001       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0               100.0 |      100.0 |  0.72135  0.721167  0.721035   0.72094  0.720871  0.720822  0.720787  0.720761  0.720744  0.720731  0.720722  0.720715  0.720711  0.720707  0.720705  0.720703  0.720702  0.720702  0.720701  0.720701    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207         0.7207 | 2.594879   2.59382  2.593054  2.592503  2.592107  2.591823  2.591619  2.591474   2.59137  2.591296  2.591244  2.591206   2.59118  2.591161  2.591148  2.591138  2.591132  2.591127  2.591124  2.591122   2.59112  2.591119  2.591118  2.591118  2.591117  2.591117  2.591117  2.591117  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116        2.591116 | 100.396214  100.284747  100.204241  100.146226  100.104505  100.074562  100.053112  100.037773  100.026824  100.019021  100.013469  100.009524  100.006726  100.004744  100.003342  100.002351  100.001652   100.00116  100.000813   100.00057  100.000399  100.000279  100.000195  100.000136  100.000095  100.000066  100.000046  100.000032  100.000022  100.000015  100.000011  100.000007  100.000005  100.000004  100.000002  100.000002  100.000001  100.000001  100.000001       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0 |  100.0 |   0.396214 |      100.0 |
    | 08:00 | 3.729985  3.726671  3.724246  3.722476  3.721187   3.72025  3.719572  3.719081  3.718727  3.718472  3.718289  3.718157  3.718063  3.717995  3.717947  3.717912  3.717888   3.71787  3.717858  3.717849  3.717843  3.717839  3.717836  3.717833  3.717832  3.717831   3.71783  3.717829  3.717829  3.717829  3.717829  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828             3.717828 | 125.513703   125.34047   125.21373  125.121253  125.053945  125.005068  124.969654  124.944045  124.925565  124.912252   124.90268  124.895809  124.890885  124.887362  124.884845  124.883049   124.88177  124.880861  124.880214  124.879756  124.879431  124.879201  124.879039  124.878924  124.878843  124.878786  124.878746  124.878718  124.878698  124.878684  124.878675  124.878668  124.878663   124.87866  124.878657  124.878656  124.878655  124.878654  124.878653  124.878653  124.878653  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652 |   53.03853   53.004741   52.980006   52.961951   52.948806   52.939258   52.932339   52.927336   52.923724   52.921123   52.919252    52.91791   52.916947   52.916259   52.915767   52.915416   52.915166   52.914988   52.914862   52.914772   52.914709   52.914664   52.914632    52.91461   52.914594   52.914583   52.914575   52.914569   52.914565   52.914563   52.914561   52.914559   52.914559   52.914558   52.914557   52.914557   52.914557   52.914557   52.914557   52.914557   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556        52.914556 |  52.299848   52.266714    52.24246   52.224755   52.211865   52.202503   52.195719   52.190813   52.187271    52.18472   52.182886   52.181569   52.180626   52.179951   52.179468   52.179124   52.178879   52.178705   52.178581   52.178493   52.178431   52.178387   52.178355   52.178333   52.178318   52.178307   52.178299   52.178294    52.17829   52.178287   52.178286   52.178284   52.178283   52.178283   52.178282   52.178282   52.178282   52.178282   52.178282   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281     52.178281 | 1.090269  1.089756   1.08938  1.089106  1.088906  1.088761  1.088655  1.088579  1.088524  1.088485  1.088456  1.088436  1.088421  1.088411  1.088403  1.088398  1.088394  1.088392   1.08839  1.088388  1.088387  1.088387  1.088386  1.088386  1.088386  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385 | 1.359036   1.35907  1.359095  1.359113  1.359126  1.359136  1.359143  1.359148  1.359151  1.359154  1.359156  1.359157  1.359158  1.359159  1.359159   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161          1.359161 | 0.537115  0.536855  0.536664  0.536525  0.536423  0.536349  0.536296  0.536257  0.536229  0.536209  0.536195  0.536184  0.536177  0.536171  0.536168  0.536165  0.536163  0.536162  0.536161   0.53616  0.536159  0.536159  0.536159  0.536159  0.536159  0.536159  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158      0.536158 | -0.202358  -0.202219   -0.20212  -0.202048  -0.201996  -0.201959  -0.201933  -0.201914    -0.2019  -0.201891  -0.201884  -0.201879  -0.201876  -0.201873  -0.201872  -0.201871   -0.20187  -0.201869  -0.201869  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868     -0.201868 | 0.665668  0.665678  0.665686  0.665692  0.665696    0.6657  0.665702  0.665704  0.665706  0.665707  0.665707  0.665708  0.665708  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709      0.665709 | 100.797305 | 100.691784  100.502956  100.364863  100.264133  100.190833  100.137614  100.099058   100.07118  100.051063  100.036572  100.026153  100.018674  100.013315   100.00948   100.00674  100.004786  100.003394  100.002404    100.0017  100.001202  100.000848  100.000598  100.000421  100.000296  100.000208  100.000146  100.000102  100.000072   100.00005  100.000035  100.000025  100.000017  100.000012  100.000008  100.000006  100.000004  100.000003  100.000002  100.000001  100.000001  100.000001       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0               100.0 |      100.0 | 0.722014  0.721656  0.721394  0.721202  0.721063  0.720962  0.720888  0.720835  0.720797  0.720769  0.720749  0.720735  0.720725  0.720718  0.720712  0.720709  0.720706  0.720704  0.720703  0.720702  0.720701  0.720701    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207         0.7207 | 2.598716  2.596645  2.595129  2.594022  2.593216  2.592631  2.592207    2.5919  2.591678  2.591519  2.591404  2.591322  2.591263  2.591221   2.59119  2.591169  2.591154  2.591143  2.591135  2.591129  2.591126  2.591123  2.591121  2.591119  2.591119  2.591118  2.591117  2.591117  2.591117  2.591117  2.591117  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116        2.591116 | 100.797305  100.580172  100.421218  100.305161  100.220635  100.159215  100.114682  100.082459   100.05919  100.042418  100.030351  100.021683  100.015468  100.011019  100.007838  100.005568  100.003951  100.002799  100.001981    100.0014  100.000989  100.000697  100.000491  100.000346  100.000243  100.000171   100.00012  100.000084  100.000059  100.000041  100.000029   100.00002  100.000014   100.00001  100.000007  100.000005  100.000003  100.000002  100.000002  100.000001  100.000001  100.000001       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0 |  100.0 |   0.797305 |      100.0 |
    | 08:30 | 3.740875   3.73481  3.730305  3.726972  3.724513  3.722704  3.721378  3.720407  3.719698  3.719181  3.718806  3.718533  3.718336  3.718193   3.71809  3.718016  3.717963  3.717924  3.717897  3.717877  3.717863  3.717853  3.717846  3.717841  3.717837  3.717834  3.717833  3.717831   3.71783   3.71783  3.717829  3.717829  3.717829  3.717829  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828             3.717828 |  126.08386   125.76616  125.530461  125.356172  125.227676  125.133201  125.063916  125.013225   124.97622  124.949264  124.929667  124.915447  124.905148  124.897702  124.892328  124.888455   124.88567  124.883668  124.882233  124.881205   124.88047  124.879945   124.87957  124.879304  124.879114  124.878979  124.878883  124.878815  124.878767  124.878733  124.878709  124.878692   124.87868  124.878672  124.878666  124.878662  124.878659  124.878657  124.878655  124.878654  124.878654  124.878653  124.878653  124.878653  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652 |   53.14959   53.087734   53.041798   53.007804   52.982728   52.964284   52.950753   52.940852   52.933622   52.928355   52.924526   52.921747   52.919735   52.918279   52.917229   52.916472   52.915928   52.915537   52.915256   52.915055   52.914912   52.914809   52.914736   52.914684   52.914647    52.91462   52.914602   52.914588   52.914579   52.914572   52.914568   52.914564   52.914562    52.91456   52.914559   52.914558   52.914558   52.914557   52.914557   52.914557   52.914557   52.914557   52.914557   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556        52.914556 |  52.408751   52.348096   52.303052   52.269718   52.245129   52.227043   52.213775   52.204066   52.196977   52.191812   52.188057   52.185333   52.183359   52.181932   52.180902    52.18016   52.179626   52.179243   52.178968   52.178771    52.17863   52.178529   52.178457   52.178406    52.17837   52.178344   52.178326   52.178313   52.178303   52.178297   52.178292   52.178289   52.178287   52.178285   52.178284   52.178283   52.178283   52.178282   52.178282   52.178282   52.178282   52.178282   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281     52.178281 | 1.091956  1.091017  1.090319  1.089803  1.089421  1.089141  1.088935  1.088785  1.088675  1.088595  1.088537  1.088494  1.088464  1.088442  1.088426  1.088414  1.088406    1.0884  1.088396  1.088393   1.08839  1.088389  1.088388  1.088387  1.088386  1.088386  1.088386  1.088386  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385 | 1.358925  1.358987  1.359033  1.359067  1.359092   1.35911  1.359124  1.359134  1.359141  1.359147  1.359151  1.359153  1.359155  1.359157  1.359158  1.359159  1.359159   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161          1.359161 | 0.537969  0.537494   0.53714  0.536878  0.536685  0.536543  0.536438  0.536362  0.536306  0.536265  0.536235  0.536214  0.536198  0.536187  0.536179  0.536173  0.536169  0.536166  0.536164  0.536162  0.536161   0.53616   0.53616  0.536159  0.536159  0.536159  0.536159  0.536159  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158      0.536158 | -0.202862   -0.20259  -0.202392  -0.202247  -0.202142  -0.202065   -0.20201   -0.20197  -0.201941   -0.20192  -0.201905  -0.201894  -0.201887  -0.201881  -0.201877  -0.201875  -0.201873  -0.201871   -0.20187  -0.201869  -0.201869  -0.201869  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868     -0.201868 | 0.665645   0.66566  0.665672  0.665681  0.665688  0.665693  0.665697    0.6657  0.665703  0.665704  0.665706  0.665707  0.665707  0.665708  0.665708  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709      0.665709 | 101.507032 | 101.313893  100.967125  100.710055  100.520067  100.380056  100.277146  100.201691  100.146495  100.106207  100.076861  100.055528  100.040049  100.028839  100.020734  100.014885   100.01067  100.007638   100.00546  100.003898  100.002779  100.001979  100.001407     100.001  100.000709  100.000503  100.000356  100.000252  100.000178  100.000125  100.000088  100.000062  100.000044  100.000031  100.000022  100.000015  100.000011  100.000007  100.000005  100.000004  100.000003  100.000002  100.000001  100.000001  100.000001       100.0       100.0       100.0       100.0       100.0               100.0 |      100.0 | 0.723189  0.722535  0.722048  0.721688  0.721422  0.721227  0.721083  0.720978  0.720902  0.720846  0.720805  0.720776  0.720755  0.720739  0.720728   0.72072  0.720714   0.72071  0.720707  0.720705  0.720703  0.720702  0.720702  0.720701  0.720701    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207         0.7207 | 2.605522  2.601732  2.598916  2.596833  2.595296  2.594165  2.593336  2.592728  2.592285  2.591962  2.591728  2.591557  2.591434  2.591345   2.59128  2.591234    2.5912  2.591176  2.591159  2.591147  2.591138  2.591132  2.591127  2.591124  2.591122   2.59112  2.591119  2.591118  2.591118  2.591117  2.591117  2.591117  2.591117  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116        2.591116 | 101.507032  101.110418  100.816029  100.598215  100.437529  100.319308  100.232549  100.169029  100.122627  100.088803  100.064196  100.046329  100.033381  100.024014  100.017249  100.012371   100.00886  100.006337  100.004526  100.003228    100.0023  100.001636  100.001163  100.000825  100.000585  100.000414  100.000293  100.000207  100.000146  100.000103  100.000073  100.000051  100.000036  100.000025  100.000018  100.000012  100.000009  100.000006  100.000004  100.000003  100.000002  100.000001  100.000001  100.000001       100.0       100.0       100.0       100.0       100.0       100.0       100.0 |  100.0 |   1.507032 |      100.0 |
    | 09:00 | 3.759119  3.748656  3.740765  3.734842  3.730413  3.727112  3.724661  3.722846  3.721505  3.720517   3.71979  3.719258  3.718868  3.718583  3.718375  3.718224  3.718114  3.718034  3.717977  3.717935  3.717905  3.717883  3.717868  3.717856  3.717848  3.717843  3.717838  3.717835  3.717833  3.717832  3.717831   3.71783  3.717829  3.717829  3.717829  3.717829  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828             3.717828 | 127.041661  126.491925  126.078102  125.767849  125.536085   125.36352  125.235416   125.14058  125.070552  125.018966  124.981051  124.953245  124.932893  124.918027  124.907189    124.8993   124.89357  124.889413  124.886404  124.884228  124.882657  124.881525   124.88071  124.880125  124.879704  124.879403  124.879187  124.879033  124.878923  124.878845  124.878789  124.878749  124.878721  124.878701  124.878686  124.878676  124.878669  124.878664   124.87866  124.878658  124.878656  124.878655  124.878654  124.878653  124.878653  124.878653  124.878652  124.878652  124.878652  124.878652 |  53.335642   53.228936    53.14847   53.088063   53.042894   53.009238   52.984239   52.965725   52.952049   52.941973   52.934566   52.929133   52.925157   52.922252   52.920133   52.918592   52.917472    52.91666   52.916071   52.915646   52.915339   52.915118   52.914959   52.914844   52.914762   52.914703   52.914661   52.914631   52.914609   52.914594   52.914583   52.914575    52.91457   52.914566   52.914563   52.914561    52.91456   52.914559   52.914558   52.914558   52.914557   52.914557   52.914557   52.914557   52.914557   52.914557   52.914556   52.914556   52.914556        52.914556 |   52.59119   52.486555   52.407653   52.348419   52.304127   52.271124   52.246611   52.228456   52.215046   52.205166   52.197902   52.192575   52.188676   52.185827    52.18375   52.182238    52.18114   52.180344   52.179767    52.17935   52.179049   52.178832   52.178676   52.178564   52.178483   52.178425   52.178384   52.178354   52.178333   52.178318   52.178308     52.1783   52.178294   52.178291   52.178288   52.178286   52.178285   52.178284   52.178283   52.178282   52.178282   52.178282   52.178282   52.178282   52.178281   52.178281   52.178281   52.178281   52.178281     52.178281 | 1.094777  1.093159  1.091939  1.091022  1.090336  1.089824  1.089444  1.089163  1.088955  1.088802  1.088689  1.088607  1.088546  1.088502   1.08847  1.088446  1.088429  1.088417  1.088408  1.088402  1.088397  1.088394  1.088391  1.088389  1.088388  1.088387  1.088387  1.088386  1.088386  1.088386  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385 | 1.358741  1.358846  1.358926  1.358986  1.359032  1.359065   1.35909  1.359109  1.359123  1.359133   1.35914  1.359146   1.35915  1.359153  1.359155  1.359156  1.359158  1.359158  1.359159  1.359159   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161          1.359161 | 0.539391  0.538576   0.53796  0.537496  0.537149   0.53689  0.536697  0.536554  0.536448   0.53637  0.536313  0.536271   0.53624  0.536218  0.536201   0.53619  0.536181  0.536175   0.53617  0.536167  0.536164  0.536163  0.536161  0.536161   0.53616  0.536159  0.536159  0.536159  0.536159  0.536159  0.536159  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158      0.536158 | -0.203758  -0.203261  -0.202891  -0.202617  -0.202415  -0.202267  -0.202158  -0.202078   -0.20202  -0.201978  -0.201947  -0.201925  -0.201909  -0.201897  -0.201889  -0.201883  -0.201879  -0.201876  -0.201873  -0.201872  -0.201871   -0.20187  -0.201869  -0.201869  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868     -0.201868 |  0.66562  0.665638  0.665653  0.665666  0.665675  0.665683  0.665689  0.665694  0.665698  0.665701  0.665703  0.665705  0.665706  0.665707  0.665707  0.665708  0.665708  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709      0.665709 | 102.694853 | 102.361128   101.75973  101.307605  100.968968  100.716187  100.528076  100.388489  100.285182  100.208917  100.152746  100.111466  100.081195   100.05904  100.042858   100.03106  100.022474  100.016237  100.011713  100.008437  100.006069  100.004359  100.003127   100.00224  100.001603  100.001145  100.000818  100.000583  100.000415  100.000295   100.00021  100.000149  100.000106  100.000075  100.000053  100.000037  100.000026  100.000019  100.000013  100.000009  100.000007  100.000005  100.000003  100.000002  100.000002  100.000001  100.000001  100.000001       100.0       100.0               100.0 |      100.0 | 0.725156  0.724028  0.723177  0.722538   0.72206  0.721703  0.721438  0.721242  0.721097   0.72099  0.720912  0.720854  0.720812  0.720781  0.720759  0.720742  0.720731  0.720722  0.720716  0.720711  0.720708  0.720706  0.720704  0.720703  0.720702  0.720701  0.720701    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207         0.7207 | 2.616917  2.610382  2.605453  2.601752  2.598984  2.596921  2.595388  2.594253  2.593415  2.592797  2.592343   2.59201  2.591766  2.591588  2.591458  2.591364  2.591295  2.591245  2.591209  2.591183  2.591164  2.591151  2.591141  2.591134  2.591129  2.591125  2.591123  2.591121  2.591119  2.591119  2.591118  2.591117  2.591117  2.591117  2.591117  2.591117  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116        2.591116 | 102.694853   102.01088  101.495885  101.109626  100.820935  100.605856  100.446087  100.327728  100.240269  100.175797  100.128378  100.093577  100.068089  100.049458  100.035865  100.025966   100.01877  100.013548  100.009764  100.007027   100.00505  100.003624  100.002598   100.00186  100.001329  100.000949  100.000677  100.000482  100.000343  100.000244  100.000173  100.000123  100.000087  100.000062  100.000044  100.000031  100.000022  100.000015  100.000011  100.000008  100.000005  100.000004  100.000003  100.000002  100.000001  100.000001  100.000001       100.0       100.0       100.0       100.0 |  100.0 |   2.694853 |      100.0 |
    | 09:30 |  3.78812  3.771032  3.757931  3.747945   3.74037  3.734649  3.730345  3.727118  3.724707   3.72291  3.721574  3.720584  3.719852  3.719311  3.718913   3.71862  3.718406  3.718249  3.718134   3.71805  3.717989  3.717944  3.717912  3.717889  3.717872   3.71786  3.717851  3.717844   3.71784  3.717836  3.717834  3.717832  3.717831   3.71783   3.71783  3.717829  3.717829  3.717829  3.717829  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828             3.717828 | 128.571088  127.668895  126.979207  126.454614  126.057364  125.757732  125.532531  125.363818  125.237799  125.143926  125.074178  125.022481  124.984249  124.956036   124.93526   124.91999  124.908789  124.900588  124.894593  124.890219  124.887032  124.884715  124.883032  124.881812  124.880929  124.880291   124.87983  124.879497  124.879258  124.879086  124.878962  124.878874   124.87881  124.878765  124.878732  124.878709  124.878693  124.878681  124.878672  124.878666  124.878662  124.878659  124.878657  124.878656  124.878655  124.878654  124.878653  124.878653  124.878653  124.878652 |    53.6314   53.457132    53.32353   53.221686   53.144434   53.086093   53.042201   53.009296   52.984704   52.966378   52.952758    52.94266   52.935191   52.929679   52.925619   52.922635   52.920446   52.918843   52.917672   52.916817   52.916194   52.915741   52.915412   52.915174   52.915001   52.914877   52.914787   52.914722   52.914675   52.914641   52.914617     52.9146   52.914587   52.914578   52.914572   52.914568   52.914564   52.914562    52.91456   52.914559   52.914558   52.914558   52.914557   52.914557   52.914557   52.914557   52.914557   52.914557   52.914556        52.914556 |  52.881204   52.710321   52.579313   52.479446   52.403695   52.346486   52.303447   52.271181   52.247067   52.229096   52.215741   52.205839   52.198515    52.19311   52.189129   52.186203   52.184057   52.182485   52.181336   52.180498   52.179887   52.179443   52.179121   52.178887   52.178718   52.178595   52.178507   52.178443   52.178397   52.178364   52.178341   52.178324   52.178312   52.178303   52.178297   52.178292   52.178289   52.178287   52.178285   52.178284   52.178283   52.178283   52.178282   52.178282   52.178282   52.178282   52.178282   52.178281   52.178281     52.178281 | 1.099251  1.096616  1.094593  1.093049  1.091877  1.090992  1.090325  1.089825  1.089452  1.089173  1.088966  1.088812  1.088699  1.088615  1.088553  1.088508  1.088475   1.08845  1.088432  1.088419   1.08841  1.088403  1.088398  1.088394  1.088392   1.08839  1.088389  1.088388  1.088387  1.088386  1.088386  1.088386  1.088386  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385 | 1.358452  1.358621  1.358753  1.358853   1.35893  1.358988  1.359032  1.359065   1.35909  1.359108  1.359122  1.359132   1.35914  1.359145  1.359149  1.359152  1.359155  1.359156  1.359157  1.359158  1.359159  1.359159   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161          1.359161 |  0.54163  0.540314  0.539298  0.538521  0.537929  0.537481  0.537144   0.53689    0.5367  0.536559  0.536454  0.536376  0.536318  0.536275  0.536244  0.536221  0.536204  0.536191  0.536182  0.536176  0.536171  0.536167  0.536165  0.536163  0.536162  0.536161   0.53616   0.53616  0.536159  0.536159  0.536159  0.536159  0.536159  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158      0.536158 | -0.205257    -0.2044  -0.203753  -0.203266  -0.202902  -0.202631  -0.202429   -0.20228  -0.202169  -0.202088  -0.202028  -0.201985  -0.201953  -0.201929  -0.201912    -0.2019  -0.201891  -0.201885   -0.20188  -0.201876  -0.201874  -0.201872  -0.201871   -0.20187  -0.201869  -0.201869  -0.201869  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868     -0.201868 | 0.665598  0.665617  0.665634  0.665648  0.665661  0.665671  0.665679  0.665686  0.665691  0.665696  0.665699  0.665701  0.665703  0.665705  0.665706  0.665707  0.665707  0.665708  0.665708  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709      0.665709 | 104.586356 | 104.038944   103.04839  102.292759  101.718945  101.284961   100.95793  100.712311  100.528401  100.391084  100.288827  100.212866  100.156573  100.114947  100.084233  100.061616  100.044995  100.032802  100.023875   100.01735  100.012589  100.009121  100.006599  100.004768   100.00344  100.002478  100.001783  100.001282   100.00092  100.000659  100.000472  100.000338  100.000241  100.000172  100.000123  100.000087  100.000062  100.000044  100.000031  100.000022  100.000016  100.000011  100.000008  100.000006  100.000004  100.000003  100.000002  100.000001  100.000001  100.000001               100.0 |      100.0 | 0.728275  0.726438  0.725028  0.723952  0.723135  0.722517  0.722053  0.721704  0.721443  0.721249  0.721105  0.720998  0.720918   0.72086  0.720817  0.720785  0.720762  0.720745  0.720733  0.720724  0.720717  0.720712  0.720709  0.720706  0.720704  0.720703  0.720702  0.720701  0.720701  0.720701    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207         0.7207 | 2.635017  2.624354  2.616175  2.609938  2.605206  2.601631  2.598941  2.596924  2.595417  2.594293  2.593458  2.592839  2.592381  2.592043  2.591795  2.591612  2.591477  2.591379  2.591307  2.591255  2.591217  2.591189  2.591169  2.591154  2.591144  2.591136   2.59113  2.591126  2.591124  2.591121   2.59112  2.591119  2.591118  2.591118  2.591117  2.591117  2.591117  2.591117  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116        2.591116 | 104.586356  103.466536  102.610665  101.959629  101.466496  101.094384  100.814559  100.604796  100.448009  100.331137  100.244239  100.179783  100.132081  100.096856  100.070897  100.051805  100.037791  100.027523  100.020013   100.01453  100.010533  100.007624  100.005511  100.003978  100.002868  100.002065  100.001484  100.001066  100.000765  100.000548  100.000392   100.00028    100.0002  100.000143  100.000102  100.000072  100.000051  100.000036  100.000026  100.000018  100.000013  100.000009  100.000006  100.000005  100.000003  100.000002  100.000002  100.000001  100.000001  100.000001       100.0 |  100.0 |   4.586356 |      100.0 |
    | 10:00 | 3.832038  3.805527  3.784832  3.768792  3.756437  3.746974  3.739758   3.73428  3.730137  3.727014  3.724667  3.722908  3.721593  3.720613  3.719884  3.719343  3.718943  3.718647  3.718428  3.718267  3.718149  3.718062  3.717999  3.717952  3.717918  3.717893  3.717875  3.717862  3.717853  3.717846  3.717841  3.717837  3.717835  3.717833  3.717832  3.717831   3.71783  3.717829  3.717829  3.717829  3.717829  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828             3.717828 | 130.903131  129.493085  128.397221  127.550826  126.900672  126.403657  126.025331  125.738452  125.521666  125.358365  125.235711  125.143836   125.07519  125.024022  124.985969   124.95773  124.936817  124.921361  124.909959  124.901564  124.895393  124.890866  124.887549  124.885124  124.883353  124.882062  124.881122  124.880439  124.879943  124.879583  124.879323  124.879135  124.878999  124.878901  124.878831   124.87878  124.878744  124.878718  124.878699  124.878685  124.878676  124.878669  124.878664  124.878661  124.878658  124.878656  124.878655  124.878654  124.878654  124.878653 |   54.07927   53.808913   53.597859   53.434284   53.308295   53.211783     53.1382   53.082336   53.040083   53.008232   52.984296    52.96636   52.952955   52.942961   52.935527    52.93001   52.925923   52.922903   52.920675   52.919034   52.917828   52.916943   52.916295   52.915821   52.915475   52.915223   52.915039   52.914906   52.914809   52.914738   52.914688   52.914651   52.914624   52.914605   52.914591   52.914581   52.914574   52.914569   52.914566   52.914563   52.914561    52.91456   52.914559   52.914558   52.914558   52.914557   52.914557   52.914557   52.914557        52.914557 |  53.320377    53.05527   52.848315   52.687916   52.564374   52.469736   52.397582   52.342803    52.30137   52.270138   52.246667   52.229079   52.215934   52.206134   52.198845   52.193434   52.189428   52.186466   52.184281   52.182672    52.18149   52.180622   52.179986   52.179522   52.179182   52.178935   52.178755   52.178624   52.178529    52.17846    52.17841   52.178374   52.178348   52.178329   52.178316   52.178306   52.178299   52.178294    52.17829   52.178288   52.178286   52.178285   52.178284   52.178283   52.178282   52.178282   52.178282   52.178282   52.178282     52.178282 | 1.106004  1.101931  1.098744   1.09627  1.094362  1.092899  1.091783  1.090935  1.090293  1.089809  1.089445  1.089173  1.088969  1.088817  1.088704   1.08862  1.088558  1.088512  1.088478  1.088453  1.088435  1.088421  1.088411  1.088404  1.088399  1.088395  1.088392   1.08839  1.088389  1.088388  1.088387  1.088386  1.088386  1.088386  1.088386  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385 | 1.358023  1.358281  1.358484  1.358644  1.358768  1.358863  1.358936  1.358992  1.359034  1.359066   1.35909  1.359108  1.359122  1.359132  1.359139  1.359145  1.359149  1.359152  1.359154  1.359156  1.359157  1.359158  1.359159  1.359159   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161          1.359161 | 0.544976  0.542963  0.541378   0.54014  0.539182  0.538445  0.537881  0.537452  0.537127  0.536882  0.536697  0.536559  0.536455  0.536378   0.53632  0.536278  0.536246  0.536223  0.536206  0.536193  0.536184  0.536177  0.536172  0.536168  0.536165  0.536163  0.536162  0.536161   0.53616   0.53616  0.536159  0.536159  0.536159  0.536159  0.536159  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158      0.536158 | -0.207628  -0.206237  -0.205166  -0.204347  -0.203724  -0.203253  -0.202898  -0.202633  -0.202434  -0.202286  -0.202176  -0.202094  -0.202034  -0.201989  -0.201957  -0.201933  -0.201915  -0.201902  -0.201893  -0.201886  -0.201881  -0.201877  -0.201875  -0.201873  -0.201871   -0.20187   -0.20187  -0.201869  -0.201869  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868     -0.201868 | 0.665591  0.665604  0.665618  0.665632  0.665646  0.665657  0.665668  0.665676  0.665683  0.665689  0.665693  0.665697    0.6657  0.665702  0.665704  0.665705  0.665706  0.665707  0.665708  0.665708  0.665708  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709      0.665709 | 107.466641 | 106.610349  105.053692  103.847863  102.918933  102.206804   101.66325  101.249987  100.936896  100.700465  100.522457   100.38881  100.288728  100.213967  100.158251   100.11682  100.086077  100.063312  100.046487  100.034076  100.024938  100.018222  100.013294  100.009684  100.007044  100.005116  100.003711  100.002688  100.001945  100.001405  100.001014  100.000731  100.000526  100.000378  100.000271  100.000195   100.00014    100.0001  100.000071  100.000051  100.000036  100.000026  100.000018  100.000013  100.000009  100.000007  100.000005  100.000003  100.000002  100.000002          100.000001 | 100.000001 |  0.73298  0.730142  0.727921  0.726197  0.724867  0.723847  0.723069  0.722478   0.72203  0.721693  0.721439  0.721249  0.721107  0.721001  0.720922  0.720864   0.72082  0.720788  0.720765  0.720747  0.720734  0.720725  0.720718  0.720713  0.720709  0.720707  0.720705  0.720703  0.720702  0.720702  0.720701  0.720701    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207         0.7207 | 2.662394  2.645873  2.632966  2.622956  2.615243  2.609332  2.604824  2.601401  2.598811  2.596859  2.595392  2.594292  2.593471  2.592858  2.592402  2.592064  2.591813  2.591628  2.591491  2.591391  2.591317  2.591263  2.591223  2.591194  2.591173  2.591157  2.591146  2.591138  2.591132  2.591127  2.591124  2.591122   2.59112  2.591119  2.591118  2.591118  2.591117  2.591117  2.591117  2.591117  2.591117  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116        2.591116 | 107.466641  105.718149  104.360507  103.312446   102.50751  101.892109  101.423529  101.068054  100.799286  100.596704  100.444443  100.330311  100.244975   100.18132  100.133946  100.098765  100.072693   100.05341  100.039176  100.028688  100.020974   100.01531  100.011159  100.008122  100.005902  100.004283  100.003104  100.002247  100.001624  100.001172  100.000845  100.000609  100.000438  100.000315  100.000226  100.000162  100.000116  100.000083  100.000059  100.000042   100.00003  100.000021  100.000015  100.000011  100.000008  100.000005  100.000004  100.000003  100.000002  100.000001  100.000001 |  100.0 |   7.466641 | 100.000001 |
    | 10:30 | 3.895577  3.856412  3.825223  3.800607   3.78133  3.766336  3.754741  3.745822  3.738991  3.733781  3.729822  3.726822  3.724558  3.722852  3.721571   3.72061  3.719893  3.719357  3.718959  3.718662  3.718443   3.71828   3.71816  3.718072  3.718006  3.717958  3.717923  3.717897  3.717879  3.717865  3.717855  3.717847  3.717842  3.717838  3.717835  3.717833  3.717832  3.717831   3.71783   3.71783  3.717829  3.717829  3.717829  3.717829  3.717828  3.717828  3.717828  3.717828  3.717828             3.717828 | 134.311281  132.205757  130.540002  129.232185  128.212241  127.421463  126.811523  126.343216  125.985123  125.712312  125.505167  125.348364  125.230007  125.140905  125.073994  125.023866  124.986396  124.958446  124.937641  124.922186  124.910725  124.902243  124.895977  124.891355  124.887952  124.885451  124.883615   124.88227  124.881286  124.880567  124.880042   124.87966  124.879382   124.87918  124.879034  124.878927   124.87885  124.878795  124.878755  124.878726  124.878705   124.87869  124.878679  124.878671  124.878666  124.878662  124.878659  124.878657  124.878656  124.878655 |  54.727251   54.327842   54.009774   53.758741   53.562153   53.409239   53.290997   53.200034   53.130374   53.077243   53.036866   53.006281   52.983183   52.965788   52.952722    52.94293    52.93561    52.93015   52.926084   52.923064   52.920825   52.919167   52.917942   52.917039   52.916374   52.915885   52.915526   52.915264   52.915071   52.914931   52.914828   52.914753   52.914699    52.91466   52.914631    52.91461   52.914595   52.914584   52.914576   52.914571   52.914567   52.914564   52.914562    52.91456   52.914559   52.914558   52.914558   52.914557   52.914557        52.914557 |  53.955775   53.564122    53.25223   53.006072   52.813302   52.663358   52.547411   52.458215   52.389908   52.337809   52.298215   52.268224   52.245575   52.228518   52.215705   52.206104   52.198926   52.193572   52.189585   52.186624   52.184428   52.182802   52.181602   52.180716   52.180064   52.179584   52.179233   52.178975   52.178786   52.178648   52.178548   52.178475   52.178421   52.178383   52.178354   52.178334   52.178319   52.178309   52.178301   52.178295   52.178291   52.178289   52.178287   52.178285   52.178284   52.178283   52.178283   52.178282   52.178282     52.178282 | 1.115726   1.10974  1.104958  1.101174  1.098204  1.095891    1.0941  1.092721  1.091664  1.090857  1.090244  1.089779  1.089428  1.089164  1.088965  1.088817  1.088705  1.088622   1.08856  1.088514   1.08848  1.088455  1.088437  1.088423  1.088413  1.088405    1.0884  1.088396  1.088393  1.088391  1.088389  1.088388  1.088387  1.088387  1.088386  1.088386  1.088386  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385 | 1.357422   1.35779  1.358089  1.358329  1.358519  1.358668  1.358785  1.358875  1.358944  1.358997  1.359038  1.359068  1.359091  1.359109  1.359122  1.359132  1.359139  1.359145  1.359149  1.359152  1.359154  1.359156  1.359157  1.359158  1.359159  1.359159   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161          1.359161 |  0.54972  0.546809  0.544461  0.542587  0.541108   0.53995   0.53905  0.538355  0.537821  0.537413  0.537102  0.536867  0.536689  0.536554  0.536453  0.536378  0.536321  0.536279  0.536247  0.536224  0.536207  0.536194  0.536184  0.536178  0.536172  0.536169  0.536166  0.536164  0.536162  0.536161   0.53616   0.53616  0.536159  0.536159  0.536159  0.536159  0.536159  0.536159  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158      0.536158 | -0.211181  -0.209046  -0.207368   -0.20606  -0.205048   -0.20427  -0.203676  -0.203225  -0.202883  -0.202625  -0.202431  -0.202286  -0.202178  -0.202097  -0.202037  -0.201992  -0.201959  -0.201935  -0.201917  -0.201904  -0.201894  -0.201887  -0.201882  -0.201878  -0.201875  -0.201873  -0.201872  -0.201871   -0.20187  -0.201869  -0.201869  -0.201869  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868     -0.201868 |  0.66562  0.665614  0.665617  0.665625  0.665635  0.665646  0.665656  0.665665  0.665674  0.665681  0.665686  0.665691  0.665695  0.665698  0.665701  0.665703  0.665704  0.665705  0.665706  0.665707  0.665708  0.665708  0.665708  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709      0.665709 | 111.678559 |  110.39641  108.053502  106.208913  104.766294  103.644666  102.777139  102.109253  101.597199   101.20609   100.90838  100.682476  100.511559  100.382595  100.285536  100.212666  100.158082  100.117285  100.086857  100.064209  100.047384   100.03491  100.025677  100.018857  100.013826  100.010123    100.0074  100.005402  100.003938  100.002867  100.002084  100.001513  100.001097  100.000795  100.000575  100.000415    100.0003  100.000216  100.000156  100.000112   100.00008  100.000058  100.000041   100.00003  100.000021  100.000015  100.000011  100.000008  100.000005  100.000004          100.000003 | 100.000002 |  0.73975  0.735582  0.732251  0.729614  0.727545  0.725933  0.724684  0.723723  0.722986  0.722424  0.721996  0.721672  0.721427  0.721243  0.721104  0.721001  0.720923  0.720865  0.720822   0.72079  0.720766  0.720749  0.720736  0.720726  0.720719  0.720714   0.72071  0.720707  0.720705  0.720704  0.720703  0.720702  0.720701  0.720701    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207         0.7207 | 2.701933  2.677571  2.658148  2.642805  2.630781  2.621423  2.614183  2.608612  2.604344  2.601089  2.598614  2.596739  2.595324  2.594257  2.593456  2.592856  2.592407  2.592072  2.591823  2.591638  2.591501  2.591399  2.591324  2.591268  2.591228  2.591198  2.591176   2.59116  2.591148  2.591139  2.591133  2.591128  2.591125  2.591123  2.591121   2.59112  2.591119  2.591118  2.591117  2.591117  2.591117  2.591117  2.591117  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116        2.591116 | 111.678559  109.065161   107.00162  105.383639  104.122842   103.14574  102.392167  101.813505  101.370888  101.033531  100.777235  100.583107  100.436477  100.326014  100.243003  100.180768  100.134214  100.099463  100.073578  100.054334  100.040056  100.029481  100.021663  100.015893  100.011642  100.008516   100.00622  100.004537  100.003304  100.002403  100.001746  100.001267  100.000918  100.000664   100.00048  100.000347   100.00025   100.00018  100.000129  100.000093  100.000067  100.000048  100.000034  100.000025  100.000018  100.000013  100.000009  100.000006  100.000005  100.000003  100.000002 |  100.0 |  11.678559 | 100.000002 |
    | 11:00 |  3.98361  3.928411  3.883484  3.847312  3.818466  3.795652   3.77774  3.763766  3.752923  3.744552  3.738116  3.733187  3.729426  3.726564  3.724394  3.722752  3.721513   3.72058   3.71988  3.719355  3.718962  3.718669   3.71845  3.718288  3.718167  3.718078  3.718012  3.717963  3.717927  3.717901  3.717881  3.717867  3.717856  3.717849  3.717843  3.717839  3.717836  3.717834  3.717832  3.717831   3.71783   3.71783  3.717829  3.717829  3.717829  3.717829  3.717828  3.717828  3.717828             3.717828 | 139.099913  136.088203  133.659492  131.718737  130.180387   128.96965    128.0227  127.286137  126.716006  126.276612  125.939292  125.681246  125.484479  125.334882  125.221462  125.135692  125.070989  125.022292  124.985721  124.958315  124.937819   124.92252  124.911123  124.902648  124.896356  124.891694  124.888245  124.885698   124.88382  124.882437   124.88142  124.880674  124.880127  124.879726  124.879434   124.87922  124.879064  124.878951  124.878868  124.878808  124.878765  124.878734  124.878711  124.878694  124.878682  124.878674  124.878668  124.878663   124.87866  124.878658 |  55.625014   55.062084    54.60392    54.23504   53.940862   53.708207    53.52554   53.383028   53.272456   53.187084   53.121452    53.07119   53.032831    53.00365   52.981515    52.96477   52.952135   52.942623   52.935478   52.930124   52.926119    52.92313   52.920902   52.919246   52.918017   52.917105   52.916431   52.915933   52.915566   52.915296   52.915097   52.914952   52.914845   52.914766   52.914709   52.914667   52.914637   52.914615   52.914599   52.914587   52.914578   52.914572   52.914568   52.914565   52.914562   52.914561   52.914559   52.914559   52.914558        52.914557 |  54.836104   54.284105   53.834838   53.473122   53.184657    52.95652   52.777401   52.637655   52.529231   52.445517   52.381159   52.331873   52.294259   52.265645    52.24394    52.22752    52.21513   52.205803   52.198797   52.193547   52.189619   52.186688   52.184504    52.18288   52.181674   52.180781    52.18012   52.179632   52.179272   52.179007   52.178812   52.178669   52.178564   52.178487   52.178431    52.17839    52.17836   52.178339   52.178323   52.178311   52.178303   52.178297   52.178293   52.178289   52.178287   52.178285   52.178284   52.178283   52.178283     52.178282 | 1.129104  1.120728   1.11388  1.108346   1.10392  1.100411  1.097651  1.095494  1.093819  1.092525  1.091529  1.090765  1.090183  1.089739  1.089403  1.089149  1.088956  1.088812  1.088703  1.088622  1.088561  1.088515  1.088482  1.088456  1.088438  1.088424  1.088414  1.088406    1.0884  1.088396  1.088393  1.088391  1.088389  1.088388  1.088387  1.088387  1.088386  1.088386  1.088386  1.088386  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385 | 1.356624   1.35712  1.357535  1.357877  1.358155  1.358378  1.358555  1.358694  1.358803  1.358888  1.358953  1.359003  1.359042  1.359071  1.359093   1.35911  1.359123  1.359132  1.359139  1.359145  1.359149  1.359152  1.359154  1.359156  1.359157  1.359158  1.359159  1.359159   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161          1.359161 | 0.556112  0.552128  0.548826  0.546127  0.543948  0.542208  0.540832  0.539751  0.538909  0.538256  0.537753  0.537367  0.537071  0.536846  0.536676  0.536546  0.536449  0.536375   0.53632  0.536279  0.536248  0.536225  0.536207  0.536195  0.536185  0.536178  0.536173  0.536169  0.536166  0.536164  0.536162  0.536161  0.536161   0.53616  0.536159  0.536159  0.536159  0.536159  0.536159  0.536159  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158      0.536158 | -0.216238  -0.213139  -0.210643  -0.208656  -0.207089  -0.205863  -0.204912  -0.204179  -0.203616  -0.203186  -0.202858  -0.202611  -0.202423  -0.202282  -0.202176  -0.202097  -0.202038  -0.201994  -0.201961  -0.201936  -0.201918  -0.201905  -0.201895  -0.201888  -0.201882  -0.201878  -0.201876  -0.201873  -0.201872  -0.201871   -0.20187  -0.201869  -0.201869  -0.201869  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868     -0.201868 | 0.665713  0.665668  0.665646  0.665637  0.665637  0.665642  0.665649  0.665657  0.665665  0.665672  0.665679  0.665685   0.66569  0.665694  0.665697    0.6657  0.665702  0.665703  0.665705  0.665706  0.665707  0.665707  0.665708  0.665708  0.665708  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709      0.665709 | 117.614836 | 115.771398  112.383429  109.669789  107.513372  105.811738  104.477296  103.436562  102.628863  102.004761  101.524425  101.156061  100.874493  100.659922  100.496866  100.373286  100.279859  100.209393  100.156367   100.11655  100.086714  100.064402  100.047749  100.035343  100.026118   100.01927  100.014195  100.010442  100.007669  100.005625  100.004119  100.003013    100.0022  100.001605  100.001169  100.000851  100.000618  100.000449  100.000325  100.000236   100.00017  100.000123  100.000089  100.000064  100.000046  100.000033  100.000024  100.000017  100.000012  100.000009          100.000006 | 100.000005 | 0.749061  0.743232  0.738465  0.734611  0.731528  0.729083  0.727159  0.725656  0.724488  0.723586  0.722892   0.72236  0.721953  0.721644   0.72141  0.721232  0.721098  0.720997  0.720922  0.720865  0.720822  0.720791  0.720767  0.720749  0.720736  0.720727   0.72072  0.720714   0.72071  0.720707  0.720705  0.720704  0.720703  0.720702  0.720701  0.720701    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207         0.7207 | 2.756584  2.722333  2.694414  2.671906  2.653938  2.639715  2.628541  2.619818  2.613048  2.607819  2.603798  2.600718  2.598367  2.596578  2.595221  2.594195   2.59342  2.592837  2.592399  2.592071  2.591825  2.591642  2.591505  2.591404  2.591328  2.591273  2.591231  2.591201  2.591178  2.591162  2.591149   2.59114  2.591134  2.591129  2.591126  2.591123  2.591121   2.59112  2.591119  2.591118  2.591118  2.591117  2.591117  2.591117  2.591117  2.591117  2.591116  2.591116  2.591116        2.591116 | 117.614836   113.86395  110.849084  108.445876  106.544334  105.049545  103.881289  102.972916  102.269845  101.727911  101.311737  100.993222  100.750211  100.565344  100.425089  100.318953  100.238829  100.178481  100.133129  100.099119  100.073665  100.054653  100.040479  100.029932  100.022097  100.016288  100.011987  100.008809  100.006464  100.004737  100.003466  100.002533  100.001849  100.001347  100.000981  100.000713  100.000518  100.000375  100.000272  100.000197  100.000142  100.000103  100.000074  100.000053  100.000038  100.000028   100.00002  100.000014   100.00001  100.000007  100.000005 |  100.0 |  17.614836 | 100.000005 |
    | 11:30 | 4.100633  4.026293  3.964352  3.913389  3.871928  3.838536  3.811879  3.790764   3.77415  3.761156  3.751045  3.743213  3.737172   3.73253  3.728973  3.726257  3.724189  3.722618  3.721428  3.720528  3.719849  3.719338  3.718955  3.718667  3.718452  3.718291  3.718171  3.718082  3.718016  3.717966   3.71793  3.717903  3.717883  3.717869  3.717858   3.71785  3.717844   3.71784  3.717837  3.717834  3.717833  3.717831   3.71783   3.71783  3.717829  3.717829  3.717829  3.717829  3.717828             3.717828 | 145.585456  141.449571  138.045735  135.273877  133.038053  131.249837  129.830314  128.710901   127.83329  127.148789  126.617349  126.206432  125.889878  125.646837   125.46081  125.318831  125.210757  125.128697  125.066538   125.01956  124.984131  124.957467  124.937441  124.922428  124.911195  124.902805   124.89655  124.891894  124.888435  124.885869  124.883968  124.882562  124.881524  124.880759  124.880196  124.879781  124.879477  124.879254  124.879091  124.878972  124.878884  124.878821  124.878774  124.878741  124.878716  124.878698  124.878686  124.878676  124.878669  124.878665 |  56.818416   56.060293    55.42862   54.908889   54.486073   54.145541   53.873695   53.658356   53.488929   53.356411     53.2533   53.173436   53.111831   53.064484   53.028216   53.000518   52.979425   52.963404   52.951265   52.942089   52.935168   52.929958   52.926045   52.923112   52.920916   52.919277   52.918054   52.917144   52.916468   52.915967   52.915595   52.915321   52.915118   52.914968   52.914858   52.914777   52.914718   52.914674   52.914642   52.914619   52.914602   52.914589    52.91458   52.914574   52.914569   52.914565   52.914563   52.914561    52.91456        52.914559 |  56.006331    55.26293   54.643524   54.133885    53.71928   53.385361   53.118794   52.907637     52.7415   52.611556   52.510446   52.432134   52.371725   52.325297   52.289733   52.262574   52.241891   52.226181   52.214277   52.205279   52.198492   52.193384   52.189547    52.18667   52.184518    52.18291   52.181711   52.180819   52.180156   52.179664     52.1793   52.179031   52.178832   52.178685   52.178577   52.178498   52.178439   52.178397   52.178365   52.178343   52.178326   52.178314   52.178305   52.178298   52.178294    52.17829   52.178288   52.178286   52.178285     52.178284 |  1.14673  1.135553  1.126186  1.118441  1.112114  1.107001  1.102908  1.099658  1.097097  1.095091  1.093529  1.092317  1.091383  1.090664  1.090113  1.089692  1.089371  1.089128  1.088943  1.088804  1.088698  1.088619   1.08856  1.088515  1.088482  1.088457  1.088438  1.088424  1.088414  1.088406  1.088401  1.088397  1.088394  1.088391   1.08839  1.088388  1.088387  1.088387  1.088386  1.088386  1.088386  1.088386  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385 | 1.355622  1.356251  1.356795  1.357258  1.357643  1.357961  1.358219  1.358426   1.35859   1.35872  1.358822  1.358901  1.358963   1.35901  1.359046  1.359074  1.359095  1.359111  1.359124  1.359133   1.35914  1.359145  1.359149  1.359152  1.359154  1.359156  1.359157  1.359158  1.359159  1.359159   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161          1.359161 | 0.564301  0.559139  0.554731   0.55103  0.547968  0.545466  0.543447  0.541833  0.540555  0.539549  0.538762  0.538151  0.537679  0.537315  0.537036  0.536822   0.53666  0.536536  0.536442  0.536371  0.536318  0.536277  0.536247  0.536224  0.536208  0.536195  0.536185  0.536178  0.536173  0.536169  0.536166  0.536164  0.536163  0.536161  0.536161   0.53616   0.53616  0.536159  0.536159  0.536159  0.536159  0.536159  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158      0.536158 |  -0.22308  -0.218819  -0.215296  -0.212422  -0.210106   -0.20826  -0.206801  -0.205657  -0.204766  -0.204077  -0.203545  -0.203138  -0.202827   -0.20259  -0.202411  -0.202275  -0.202172  -0.202095  -0.202037  -0.201994  -0.201961  -0.201937  -0.201919  -0.201906  -0.201896  -0.201888  -0.201883  -0.201879  -0.201876  -0.201874  -0.201872  -0.201871   -0.20187  -0.201869  -0.201869  -0.201869  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868     -0.201868 | 0.665908  0.665799  0.665729  0.665688  0.665665  0.665655  0.665653  0.665656  0.665661  0.665666  0.665673  0.665678  0.665684  0.665688  0.665692  0.665696  0.665698  0.665701  0.665703  0.665704  0.665705  0.665706  0.665707  0.665707  0.665708  0.665708  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709      0.665709 | 125.703736 | 123.151741  118.432011  114.582651  111.471716  108.978126   106.99398   105.42546  104.192662  103.228707  102.478425  101.896861  101.447758  101.102127  100.836961   100.63412  100.479374  100.361624   100.27224  100.204546  100.153392  100.114819  100.085791   100.06399  100.047648  100.035421  100.026289   100.01948  100.014413  100.010648  100.007855  100.005786  100.004256  100.003126  100.002293   100.00168  100.001229  100.000898  100.000656  100.000478  100.000348  100.000253  100.000184  100.000133  100.000097   100.00007  100.000051  100.000037  100.000026  100.000019          100.000014 | 100.000011 | 0.761316  0.753546  0.747031   0.74164  0.737235  0.733674  0.730822  0.728558  0.726773  0.725375  0.724286  0.723442   0.72279  0.722289  0.721904  0.721611  0.721387  0.721218  0.721089  0.720992  0.720918  0.720863  0.720821   0.72079  0.720767   0.72075  0.720737  0.720727   0.72072  0.720715  0.720711  0.720708  0.720706  0.720704  0.720703  0.720702  0.720701  0.720701  0.720701    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207         0.7207 | 2.829006  2.783028  2.744641  2.713002  2.687226  2.666441  2.649833  2.636666    2.6263  2.618189  2.611875  2.606983  2.603208  2.600307  2.598084  2.596386  2.595093  2.594111  2.593367  2.592804   2.59238  2.592061  2.591821  2.591641  2.591506  2.591406  2.591331  2.591275  2.591233  2.591203   2.59118  2.591163  2.591151  2.591141  2.591135   2.59113  2.591126  2.591123  2.591122   2.59112  2.591119  2.591118  2.591118  2.591117  2.591117  2.591117  2.591117  2.591117  2.591116        2.591116 | 125.703736  120.520097  116.275013  112.831663  110.062627  107.852966  106.101732   104.72224  103.641435  102.798713  102.144455  101.638486  101.248576  100.949077  100.719712  100.544549  100.411128  100.309754   100.23291  100.174792  100.130931  100.097898   100.07307  100.054445  100.040499  100.030077  100.022301  100.016509  100.012203  100.009007  100.006638  100.004886  100.003591  100.002635  100.001932  100.001414  100.001034  100.000755   100.00055  100.000401  100.000292  100.000212  100.000154  100.000112  100.000081  100.000058  100.000042   100.00003  100.000022  100.000016  100.000011 |  100.0 |  25.703736 | 100.000011 |
    | 12:00 | 4.250162  4.154331  4.072497  4.003585    3.9463  3.899236  3.860973  3.830156  3.805537  3.786011  3.770622   3.75856  3.749151  3.741844  3.736191  3.731832  3.728483  3.725916  3.723954  3.722458  3.721321  3.720458  3.719805  3.719311  3.718939  3.718658  3.718448   3.71829  3.718172  3.718083  3.718017  3.717968  3.717932  3.717905  3.717885   3.71787  3.717859  3.717851  3.717845   3.71784  3.717837  3.717835  3.717833  3.717832  3.717831   3.71783  3.717829  3.717829  3.717829             3.717829 | 154.071843  148.607303  144.013609  140.197247  137.060919  134.508731  132.450178  130.802799  129.493615  128.459579  127.647291  127.012268  126.517954  126.134656  125.838477  125.610348  125.435151  125.300971  125.198469  125.120356  125.060966  125.015909  124.981799  124.956028  124.936596   124.92197  124.910984  124.902745  124.896577  124.891968  124.888529  124.885968  124.884063  124.882648  124.881599  124.880823  124.880249  124.879825  124.879513  124.879283  124.879114  124.878989  124.878898  124.878832  124.878783  124.878747  124.878721  124.878702  124.878688  124.878678 |  58.343323   57.366031   56.531482   55.828717   55.244521   54.764558   54.374357   54.060077   53.809015   53.609891   53.452952   53.329942   53.233993   53.159474   53.101821   53.057372   53.023211   52.997033   52.977027   52.961776   52.950177   52.941376   52.934712   52.929677    52.92588   52.923022   52.920875   52.919265    52.91806   52.917159   52.916487   52.915986   52.915614   52.915337   52.915132   52.914981   52.914868   52.914786   52.914725    52.91468   52.914647   52.914622   52.914605   52.914591   52.914582   52.914575    52.91457   52.914566   52.914563        52.914562 |  57.501625   56.543311   55.724969   55.035851      54.463   53.992357   53.609734   53.301557    53.05537   52.860113   52.706222     52.5856   52.491514   52.418442   52.361909   52.318323   52.284826   52.259156   52.239539   52.224584    52.21321    52.20458   52.198046   52.193108   52.189385   52.186583   52.184477   52.182898   52.181717   52.180833   52.180174   52.179683   52.179318   52.179047   52.178846   52.178697   52.178587   52.178506   52.178446   52.178402    52.17837   52.178346   52.178329   52.178316   52.178306     52.1783   52.178295   52.178291   52.178288     52.178286 | 1.169001  1.154759  1.142508  1.132125  1.123447  1.116284  1.110438  1.105715  1.101932  1.098926  1.096553   1.09469  1.093236  1.092106  1.091231  1.090556  1.090037  1.089639  1.089335  1.089103  1.088927  1.088793  1.088692  1.088615  1.088557  1.088514  1.088481  1.088457  1.088438  1.088425  1.088414  1.088407  1.088401  1.088397  1.088394  1.088391   1.08839  1.088389  1.088388  1.088387  1.088386  1.088386  1.088386  1.088386  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385 |  1.35443  1.355183  1.355857  1.356448  1.356958  1.357388  1.357747  1.358042  1.358281  1.358473  1.358626  1.358746  1.358841  1.358915  1.358973  1.359017  1.359051  1.359078  1.359098  1.359113  1.359125  1.359134   1.35914  1.359145  1.359149  1.359152  1.359154  1.359156  1.359157  1.359158  1.359159  1.359159  1.359159   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161          1.359161 | 0.574285  0.567946  0.562363  0.557534  0.553428   0.54999   0.54715  0.544834  0.542964  0.541468  0.540282  0.539347  0.538615  0.538044  0.537602   0.53726  0.536997  0.536795  0.536641  0.536523  0.536434  0.536366  0.536314  0.536275  0.536246  0.536224  0.536207  0.536195  0.536185  0.536178  0.536173  0.536169  0.536166  0.536164  0.536163  0.536162  0.536161   0.53616   0.53616  0.536159  0.536159  0.536159  0.536159  0.536159  0.536159  0.536158  0.536158  0.536158  0.536158      0.536158 | -0.231883  -0.226336  -0.221613  -0.217656  -0.214389  -0.211727   -0.20958  -0.207867  -0.206511  -0.205446  -0.204614  -0.203968  -0.203469  -0.203085  -0.202791  -0.202566  -0.202394  -0.202264  -0.202165  -0.202091  -0.202035  -0.201992  -0.201961  -0.201937  -0.201919  -0.201906  -0.201896  -0.201889  -0.201883  -0.201879  -0.201876  -0.201874  -0.201872  -0.201871   -0.20187   -0.20187  -0.201869  -0.201869  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868     -0.201868 | 0.666248  0.666043    0.6659  0.665803   0.66574  0.665702   0.66568   0.66567  0.665666  0.665667   0.66567  0.665674  0.665679  0.665684  0.665688  0.665691  0.665695  0.665698    0.6657  0.665702  0.665703  0.665705  0.665706  0.665706  0.665707  0.665708  0.665708  0.665708  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709      0.665709 | 136.388647 | 132.978491   126.62916  121.352562  117.012066  113.474906  110.616766  108.324831  106.499393  105.054275  103.916384  103.024698  102.328948  101.788184  101.369363  101.046032  100.797167  100.606148  100.459913  100.348238  100.263156  100.198478  100.149417   100.11228  100.084224   100.06307   100.04715  100.035191  100.026223   100.01951  100.014493  100.010751  100.007963  100.005889   100.00435  100.003208  100.002363  100.001738  100.001277  100.000937  100.000687  100.000502  100.000367  100.000268  100.000196  100.000142  100.000104  100.000075  100.000055   100.00004          100.000029 | 100.000024 | 0.776785  0.766895  0.758382  0.751162  0.745124  0.740138  0.736068  0.732778  0.730143  0.728048  0.726394  0.725096  0.724082  0.723294  0.722684  0.722213  0.721851  0.721574  0.721362    0.7212  0.721077  0.720984  0.720913   0.72086   0.72082  0.720789  0.720767   0.72075  0.720737  0.720727   0.72072  0.720715  0.720711  0.720708  0.720706  0.720704  0.720703  0.720702  0.720701  0.720701  0.720701    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207         0.7207 |  2.92119  2.862155  2.811616  2.768963   2.73344  2.704207   2.68041  2.661221  2.645879  2.633702  2.624099  2.616568  2.610692  2.606127  2.602595  2.599871  2.597777  2.596173  2.594946  2.594011    2.5933  2.592761  2.592352  2.592043  2.591811  2.591635  2.591504  2.591405  2.591331  2.591276  2.591235  2.591204  2.591181  2.591164  2.591152  2.591142  2.591135   2.59113  2.591127  2.591124  2.591122   2.59112  2.591119  2.591118  2.591118  2.591117  2.591117  2.591117  2.591117        2.591117 | 136.388647  129.473624  123.700239  118.930984  115.029589  111.866393  109.322164  107.290364  105.678076  104.405896  103.407119  102.626509  102.018873  101.547615  101.183345  100.902642  100.686952  100.521661  100.395312  100.298961  100.225653  100.169999  100.127836  100.095959  100.071906  100.053791  100.040173  100.029954  100.022299  100.016575  100.012302  100.009117  100.006746  100.004985  100.003679  100.002711  100.001995  100.001467  100.001077  100.000789  100.000578  100.000423  100.000309  100.000225  100.000164   100.00012  100.000087  100.000063  100.000046  100.000033  100.000024 |  100.0 |  36.388647 | 100.000024 |
    | 12:30 | 4.434196  4.315692  4.211945  4.122449  4.046325  3.982425   3.92943  3.885959  3.850647   3.82221  3.799483   3.78144  3.767199  3.756017  3.747277  3.740473  3.735195  3.731114  3.727969  3.725551  3.723698   3.72228  3.721198  3.720375  3.719749  3.719274  3.718915  3.718644  3.718439  3.718285  3.718169  3.718083  3.718018  3.717969  3.717933  3.717906  3.717886  3.717871  3.717859  3.717851  3.717845  3.717841  3.717837  3.717835  3.717833  3.717832  3.717831   3.71783   3.71783             3.717829 | 164.823421  157.861374  151.881587  146.809646  142.558632  139.034923  136.143577  133.792785  131.897123  130.379591  129.172585  128.218032  127.466938  126.878591  126.419578  126.062768  125.786317  125.572777  125.408293  125.281927  125.185084   125.11104  125.054553  125.011551  124.978882  124.954112  124.935367  124.921208  124.910533  124.902497  124.896459   124.89193  124.888539  124.886002  124.884109  124.882698  124.881647  124.880867  124.880287  124.879858   124.87954  124.879305  124.879132  124.879004   124.87891  124.878841   124.87879  124.878753  124.878726  124.878706 |  60.220106   59.011597   57.953581   57.040893   56.264585   55.612926   55.072485   54.629164    54.26905   53.979046   53.747273   53.563271   53.418045   53.304011   53.214877   53.145486   53.091661   53.050048   53.017972   52.993317   52.974414   52.959956   52.948924   52.940525   52.934142   52.929303    52.92564   52.922873   52.920787   52.919217   52.918037   52.917152   52.916489   52.915993   52.915623   52.915347   52.915142   52.914989   52.914876   52.914792    52.91473   52.914684    52.91465   52.914625   52.914607   52.914593   52.914583   52.914576   52.914571        52.914567 |  59.341962   58.156921   57.119451   56.224487   55.463255   54.824251   54.294305   53.859593   53.506471   53.222099   52.994827   52.814398   52.671992   52.560173    52.47277   52.404726   52.351947   52.311142   52.279689   52.255512   52.236976     52.2228   52.211982   52.203745   52.197487   52.192741    52.18915   52.186437   52.184391   52.182851   52.181694   52.180826   52.180176    52.17969   52.179327   52.179057   52.178855   52.178706   52.178595   52.178512   52.178452   52.178407   52.178373   52.178349   52.178331   52.178318   52.178308   52.178301   52.178295     52.178292 | 1.196045  1.178676  1.163335  1.149996  1.138572  1.128925  1.120883  1.114258  1.108857  1.104495  1.101001  1.098221  1.096024  1.094297  1.092946  1.091893  1.091076  1.090444  1.089957  1.089582  1.089295  1.089075  1.088908   1.08878  1.088683  1.088609  1.088554  1.088512   1.08848  1.088456  1.088438  1.088425  1.088414  1.088407  1.088401  1.088397  1.088394  1.088392   1.08839  1.088389  1.088388  1.088387  1.088386  1.088386  1.088386  1.088386  1.088385  1.088385  1.088385  1.088385 | 1.353087  1.353937  1.354726  1.355442  1.356079  1.356635  1.357111  1.357512  1.357845  1.358118   1.35834  1.358518   1.35866  1.358772   1.35886  1.358929  1.358983  1.359024  1.359057  1.359081    1.3591  1.359115  1.359126  1.359134  1.359141  1.359146  1.359149  1.359152  1.359154  1.359156  1.359157  1.359158  1.359159  1.359159  1.359159   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916  1.359161  1.359161  1.359161  1.359161  1.359161          1.359161 | 0.585894  0.578501  0.571782   0.56579  0.560543  0.556028  0.552203  0.549009  0.546377  0.544232  0.542501  0.541117  0.540017  0.539149  0.538469  0.537937  0.537524  0.537204  0.536957  0.536767  0.536621  0.536509  0.536424  0.536359   0.53631  0.536272  0.536244  0.536223  0.536207  0.536194  0.536185  0.536178  0.536173  0.536169  0.536167  0.536164  0.536163  0.536162  0.536161   0.53616   0.53616  0.536159  0.536159  0.536159  0.536159  0.536159  0.536159  0.536158  0.536158      0.536158 | -0.242675  -0.235823  -0.229809  -0.224625  -0.220227  -0.216554  -0.213526  -0.211059  -0.209072  -0.207484  -0.206226  -0.205235   -0.20446  -0.203857  -0.203389  -0.203028  -0.202751  -0.202538  -0.202375  -0.202251  -0.202157  -0.202085  -0.202031   -0.20199   -0.20196  -0.201936  -0.201919  -0.201906  -0.201896  -0.201889  -0.201883  -0.201879  -0.201876  -0.201874  -0.201872  -0.201871   -0.20187   -0.20187  -0.201869  -0.201869  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868     -0.201868 | 0.666782  0.666447  0.666197  0.666016  0.665889  0.665803  0.665747  0.665712  0.665691  0.665681  0.665677  0.665676  0.665678  0.665681  0.665684  0.665688  0.665691  0.665694  0.665697  0.665699  0.665701  0.665703  0.665704  0.665705  0.665706  0.665707  0.665707  0.665708  0.665708  0.665708  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709      0.665709 | 150.102726 | 145.693656   137.42686  130.424262  124.557563  119.693089  115.698021  112.445491  109.818287   107.71113  106.031701  104.700669  103.651025  102.826978  102.182639   101.68065   101.29086  100.989117  100.756195  100.576873  100.439163  100.333658  100.253009  100.191495  100.144672  100.109104  100.082139  100.061733  100.046321    100.0347  100.025954  100.019382  100.014452  100.010761     100.008   100.00594  100.004403   100.00326  100.002411   100.00178  100.001313  100.000967  100.000711  100.000522  100.000383  100.000281  100.000206   100.00015   100.00011   100.00008          100.000058 | 100.000049 | 0.795544  0.783499  0.772851  0.763586  0.755645  0.748936   0.74334  0.738728  0.734967  0.731928  0.729494  0.727557  0.726026  0.724822   0.72388  0.723146  0.722576  0.722136  0.721796  0.721535  0.721334  0.721181  0.721064  0.720975  0.720907  0.720856  0.720817  0.720788  0.720766  0.720749  0.720737  0.720727   0.72072  0.720715  0.720711  0.720708  0.720706  0.720704  0.720703  0.720702  0.720701  0.720701  0.720701    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207         0.7207 | 3.034138   2.96147  2.897666  2.842479  2.795427  2.755849  2.722967  2.695953  2.673982  2.656271  2.642104  2.630849  2.621962   2.61498  2.609521   2.60527  2.601972  2.599422  2.597456  2.595945  2.594786    2.5939  2.593223  2.592708  2.592317   2.59202  2.591796  2.591626  2.591498  2.591402   2.59133  2.591275  2.591235  2.591204  2.591182  2.591165  2.591152  2.591143  2.591136  2.591131  2.591127  2.591124  2.591122   2.59112  2.591119  2.591119  2.591118  2.591117  2.591117        2.591117 | 150.102726  141.176813  133.577782  127.181401  121.854721  117.462807  113.874519  110.966949  108.628387  106.759899  105.275774  104.103117   103.18089  102.458652  101.895174   101.45707   101.11752  100.855122  100.652897  100.497444  100.378235   100.28703  100.217404  100.164364  100.124042  100.093447  100.070279  100.052767  100.039554  100.029602   100.02212  100.016504  100.012295  100.009146  100.006794   100.00504  100.003733  100.002762   100.00204  100.001505  100.001109  100.000816    100.0006  100.000441  100.000323  100.000237  100.000173  100.000126  100.000092  100.000067  100.000049 |  100.0 |  50.102726 | 100.000049 |
    | 13:00 | 4.652878  4.511914  4.385471  4.273721  4.176392  4.092813  4.021996  3.962734  3.913701  3.873543  3.840953  3.814715  3.793741  3.777077  3.763909  3.753554  3.745445  3.739119  3.734202   3.73039  3.727444  3.725174  3.723428  3.722089  3.721064  3.720281  3.719684   3.71923  3.718886  3.718624  3.718427  3.718277  3.718165   3.71808  3.718016  3.717969  3.717933  3.717906  3.717886  3.717871   3.71786  3.717852  3.717846  3.717841  3.717838  3.717835  3.717833  3.717832  3.717831              3.71783 | 178.039511  169.465579  161.943847  155.429273  149.857128  145.147791  141.212209  137.957305  135.290772  133.124839  131.378898  129.981002  128.868455  127.987691  127.293694  126.749148  126.323478  125.991864  125.734328  125.534894  125.380866  125.262202  125.170999  125.101058  125.047538  125.006667  124.975518  124.951823  124.933833  124.920198  124.909883  124.902092  124.896218  124.891796  124.888473   124.88598  124.884111  124.882714   124.88167  124.880892  124.880312   124.87988   124.87956  124.879322  124.879146  124.879016   124.87892  124.878849  124.878796  124.878758 |  62.450227    61.01268   59.723205   58.583574   57.591008   56.738668   56.016474   55.412114   54.912072   54.502546   54.170189   53.902617   53.688716   53.518775   53.384492    53.27889   53.196197   53.131686   53.081533   53.042662   53.012622   52.989467   52.971664   52.958007   52.947554   52.939571   52.933485   52.928855    52.92534   52.922676    52.92066   52.919137   52.917989   52.917125   52.916476   52.915988   52.915623    52.91535   52.915146   52.914994   52.914881   52.914796   52.914734   52.914687   52.914653   52.914628   52.914609   52.914595   52.914585        52.914577 |  61.528776   60.119145    58.85471    57.73721   56.763919   55.928131   55.219962   54.627338   54.137006   53.735433   53.409531   53.147155   52.937407   52.770767   52.639091    52.53554   52.454452   52.391195   52.342015   52.303899   52.274442   52.251737    52.23428   52.220888   52.210638    52.20281   52.196842   52.192303   52.188856   52.186243   52.184266   52.182773   52.181648     52.1808   52.180163   52.179686   52.179328    52.17906    52.17886   52.178711   52.178599   52.178517   52.178455    52.17841   52.178376   52.178351   52.178333   52.178319   52.178309     52.178302 |  1.22769  1.207351  1.188923  1.172485  1.158048  1.145557  1.134905  1.125941  1.118488  1.112361  1.107371  1.103344  1.100117  1.097549  1.095516  1.093917  1.092663  1.091684  1.090923  1.090332  1.089876  1.089524  1.089253  1.089046  1.088887  1.088765  1.088673  1.088602  1.088549  1.088509  1.088478  1.088455  1.088437  1.088424  1.088414  1.088407  1.088401  1.088397  1.088394  1.088392   1.08839  1.088389  1.088388  1.088387  1.088387  1.088386  1.088386  1.088386  1.088385  1.088385 | 1.351648  1.352558  1.353431  1.354251  1.355006  1.355687  1.356288   1.35681  1.357255  1.357628  1.357938  1.358191  1.358396  1.358561  1.358693  1.358797  1.358879  1.358943  1.358993  1.359032  1.359062  1.359085  1.359103  1.359117  1.359127  1.359135  1.359141  1.359146   1.35915  1.359152  1.359154  1.359156  1.359157  1.359158  1.359159  1.359159  1.359159   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916  1.359161  1.359161  1.359161          1.359161 | 0.598808  0.590588  0.582889  0.575811  0.569424  0.563764  0.558836  0.554615  0.551053  0.548088  0.545649  0.543663  0.542062   0.54078  0.539762  0.538958  0.538326  0.537831  0.537446  0.537147  0.536916  0.536737    0.5366  0.536494  0.536413  0.536352  0.536305  0.536269  0.536242  0.536221  0.536206  0.536194  0.536185  0.536178  0.536173  0.536169  0.536167  0.536164  0.536163  0.536162  0.536161   0.53616   0.53616  0.536159  0.536159  0.536159  0.536159  0.536159  0.536159      0.536158 | -0.255302  -0.247256  -0.239975  -0.233507  -0.227862  -0.223017  -0.218924  -0.215513  -0.212707  -0.210424  -0.208584  -0.207114  -0.205948  -0.205028  -0.204307  -0.203745  -0.203308   -0.20297  -0.202709  -0.202508  -0.202354  -0.202237  -0.202147  -0.202079  -0.202027  -0.201987  -0.201958  -0.201935  -0.201918  -0.201905  -0.201896  -0.201889  -0.201883  -0.201879  -0.201876  -0.201874  -0.201872  -0.201871   -0.20187   -0.20187  -0.201869  -0.201869  -0.201869  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868     -0.201868 | 0.667548  0.667056  0.666668  0.666371   0.66615   0.66599  0.665878  0.665801  0.665751  0.665719    0.6657   0.66569  0.665685  0.665683  0.665684  0.665686  0.665689  0.665692  0.665694  0.665697  0.665699  0.665701  0.665702  0.665704  0.665705  0.665706  0.665706  0.665707  0.665708  0.665708  0.665708  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709      0.665709 | 167.240349 | 161.711715  151.272237  142.259772  134.567702   128.07447   122.65011  118.162547  114.483072  111.490611  109.074652  107.136875  105.591689  104.365954  103.398136   102.63714  102.041014   101.57563  101.213448  100.932397  100.714887   100.54698  100.417671  100.318315  100.242138  100.183856  100.139355  100.105442  100.079647  100.060063  100.045221  100.033993  100.025513  100.019119  100.014306  100.010689  100.007975  100.005942  100.004421  100.003285  100.002437  100.001806  100.001337  100.000988   100.00073  100.000538  100.000396  100.000291  100.000214  100.000157          100.000115 | 100.000097 | 0.817462  0.803379  0.790606  0.779203  0.769179  0.760501  0.753096   0.74686  0.741673  0.737407  0.733932  0.731126  0.728878  0.727088  0.725672  0.724556  0.723682     0.723  0.722469  0.722057  0.721739  0.721494  0.721305   0.72116   0.72105  0.720965    0.7209  0.720851  0.720814  0.720786  0.720764  0.720748  0.720736  0.720727   0.72072  0.720715  0.720711  0.720708  0.720706  0.720704  0.720703  0.720702  0.720702  0.720701  0.720701    0.7207    0.7207    0.7207    0.7207         0.7207 | 3.167677  3.081677  3.004286   2.93568  2.875759  2.824174  2.780367  2.743637  2.713196  2.688231  2.667947    2.6516  2.638523  2.628127  2.619908  2.613442  2.608377  2.604425  2.601352  2.598969  2.597128  2.595709  2.594617   2.59378  2.593139   2.59265  2.592277  2.591993  2.591777  2.591614   2.59149  2.591397  2.591327  2.591274  2.591234  2.591204  2.591182  2.591165  2.591152  2.591143  2.591136  2.591131  2.591127  2.591124  2.591122  2.591121  2.591119  2.591119  2.591118        2.591118 | 167.240349  156.065508  146.367479  138.048506  130.992611  125.072104  120.154205  116.107075  112.804772  110.130849  107.980591  106.262028  104.895977  103.815405  102.964368  102.296727  101.774815  101.368143  101.052207  100.807442  100.618304  100.472509  100.360384  100.274343   100.20846  100.158114  100.119717   100.09049  100.068284  100.051444  100.038694  100.029059   100.02179  100.016314  100.012196  100.009105  100.006787  100.005053  100.003756  100.002789  100.002068  100.001531  100.001132  100.000836  100.000617  100.000455  100.000335  100.000246  100.000181  100.000132  100.000097 |  100.0 |  67.240349 | 100.000097 |
    | 13:30 | 4.904444  4.742605  4.594088  4.459717  4.339892  4.234571  4.143298  4.065263  3.999388  3.944427  3.899056  3.861959   3.83188  3.807674  3.788322  3.772937  3.760769  3.751188  3.743673    3.7378  3.733225  3.729671  3.726918  3.724791  3.723151   3.72189  3.720922   3.72018  3.719613  3.719181  3.718852  3.718601  3.718411  3.718267  3.718158  3.718076  3.718014  3.717967  3.717932  3.717906  3.717886  3.717871   3.71786  3.717852  3.717846  3.717841  3.717838  3.717835  3.717833             3.717832 | 193.834501  183.600558  174.439545  166.341158  159.271694  153.176519  147.984054  143.610741  139.966339  136.958914  134.499049  132.502995  130.894737  129.607043  128.581727  127.769336  127.128459  126.624861  126.230527  125.922736    125.6832  125.497287  125.353357  125.242195  125.156534  125.090666  125.040121  125.001413  124.971825  124.949251  124.932059   124.91899  124.909071  124.901556  124.895872   124.89158  124.888343  124.885906  124.884075    124.8827   124.88167  124.880899  124.880323  124.879893  124.879572  124.879334  124.879156  124.879025  124.878927  124.878855 |  65.015709   63.365266   61.850689   60.480372   59.258388    58.18432   57.253513   56.457707   55.785914   55.225418   54.762729   54.384404   54.077665   53.830813   53.633452   53.476561    53.35247   53.254759   53.178122   53.118229    53.07157   53.035329   53.007255   52.985562    52.96884   52.955978   52.946106   52.938544   52.932764   52.928353   52.924994    52.92244   52.920501   52.919033   52.917922   52.917083    52.91645   52.915974   52.915616   52.915348   52.915146   52.914996   52.914883   52.914799   52.914736    52.91469   52.914655   52.914629    52.91461        52.914596 |  64.044438   62.426045    60.94088   59.597174    58.39892   57.345709   56.432979   55.652626   54.993879   54.444268   53.990564   53.619585   53.318803   53.076745   52.883216   52.729372   52.607691   52.511877   52.436729   52.377999   52.332246   52.296709    52.26918   52.247908    52.23151   52.218898   52.209218   52.201803   52.196135    52.19181   52.188516   52.186011   52.184111   52.182671   52.181581   52.180759   52.180139   52.179672   52.179321   52.179057    52.17886   52.178712   52.178602   52.178519   52.178458   52.178412   52.178378   52.178353   52.178334      52.17832 |  1.26348  1.240528  1.219233  1.199765  1.182236  1.166691  1.153112  1.141421  1.131491  1.123162  1.116256  1.110589   1.10598  1.102261  1.099282   1.09691  1.095031  1.093551  1.092389   1.09148  1.090771  1.090221  1.089794  1.089465   1.08921  1.089015  1.088865   1.08875  1.088662  1.088595  1.088544  1.088505  1.088475  1.088453  1.088436  1.088423  1.088414  1.088407  1.088401  1.088397  1.088394  1.088392   1.08839  1.088389  1.088388  1.088387  1.088387  1.088386  1.088386  1.088386 | 1.350174  1.351102   1.35202  1.352911  1.353759   1.35455  1.355272  1.355918  1.356485  1.356975   1.35739  1.357738  1.358025   1.35826   1.35845  1.358603  1.358724  1.358821  1.358897  1.358956  1.359003  1.359039  1.359067  1.359089  1.359106  1.359119  1.359129  1.359136  1.359142  1.359147   1.35915  1.359153  1.359155  1.359156  1.359157  1.359158  1.359159  1.359159  1.359159   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916  1.359161          1.359161 | 0.612605  0.603852  0.595424  0.587449  0.580035  0.573268  0.567203  0.561862  0.557236  0.553293  0.549977  0.547224  0.544964  0.543127  0.541646  0.540461  0.539519  0.538774  0.538187  0.537728   0.53737  0.537091  0.536874  0.536707  0.536578  0.536479  0.536402  0.536344  0.536299  0.536265  0.536239  0.536219  0.536204  0.536193  0.536184  0.536178  0.536173  0.536169  0.536167  0.536164  0.536163  0.536162  0.536161   0.53616   0.53616  0.536159  0.536159  0.536159  0.536159      0.536159 | -0.269452  -0.260436  -0.252034  -0.244345  -0.237434  -0.231333  -0.226036  -0.221511  -0.217701  -0.214536  -0.211936  -0.209823   -0.20812   -0.20676   -0.20568  -0.204827  -0.204157  -0.203634  -0.203227   -0.20291  -0.202666  -0.202477  -0.202332  -0.202221  -0.202136  -0.202071  -0.202022  -0.201984  -0.201955  -0.201934  -0.201917  -0.201905  -0.201896  -0.201889  -0.201883  -0.201879  -0.201876  -0.201874  -0.201873  -0.201871   -0.20187   -0.20187  -0.201869  -0.201869  -0.201869  -0.201868  -0.201868  -0.201868  -0.201868     -0.201868 | 0.668576  0.667909  0.667357  0.666914  0.666566  0.666302  0.666107  0.665965  0.665866  0.665798  0.665753  0.665724  0.665706  0.665696  0.665691  0.665689   0.66569  0.665691  0.665693  0.665695  0.665697  0.665699    0.6657  0.665702  0.665703  0.665704  0.665705  0.665706  0.665707  0.665707  0.665708  0.665708  0.665708  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709      0.665709 | 188.127569 | 181.388537  168.574811  157.307237  147.511772   139.09169  131.932884  125.909965  120.892496   116.75072  113.360322  110.605956  108.383484  106.601063  105.179283  104.050638  103.158548  102.456162  101.905076  101.474078  101.137989  100.876624  100.673885     100.517  100.395874  100.302559  100.230821  100.175781  100.133634  100.101421  100.076847  100.058132  100.043906  100.033109   100.02493  100.018743   100.01407  100.010548  100.007896  100.005902  100.004406  100.003285  100.002446  100.001818   100.00135  100.001002  100.000742  100.000549  100.000406  100.000299          100.000221 | 100.000187 | 0.842212  0.826345  0.811608  0.798122  0.785968  0.775181  0.765751  0.757626  0.750721  0.744926  0.740119  0.736173  0.732963  0.730372  0.728296  0.726643  0.725334  0.724301  0.723491  0.722857  0.722364   0.72198  0.721682  0.721452  0.721275  0.721139  0.721034  0.720954  0.720893  0.720846   0.72081  0.720783  0.720763  0.720747  0.720735  0.720726   0.72072  0.720715  0.720711  0.720708  0.720706  0.720704  0.720703  0.720702  0.720702  0.720701  0.720701    0.7207    0.7207         0.7207 | 3.320475  3.222272  3.131845  3.049759  2.976328  2.911596  2.855348  2.807142  2.766363  2.732277  2.704096  2.681023  2.662296  2.647212  2.635143  2.625543  2.617947  2.611964   2.60727    2.6036  2.600741   2.59852  2.596799  2.595469  2.594444  2.593656  2.593051  2.592587  2.592233  2.591962  2.591756    2.5916  2.591481  2.591391  2.591323  2.591271  2.591232  2.591203  2.591181  2.591165  2.591152  2.591143  2.591136  2.591131  2.591127  2.591124  2.591122  2.591121   2.59112        2.591119 | 188.127569  174.526271  162.503058  151.996314  142.919284  135.164843  128.611388  123.129211  118.586652  114.855454   111.81495  109.354927  107.377238  105.796334  104.538991  103.543496  102.758501  102.141753  101.658794   101.28175  100.988218  100.760298  100.583759  100.447336  100.342145   100.26121  100.199063  100.151438   100.11501    100.0872  100.066006  100.049884  100.037641  100.028358  100.021333  100.016024  100.012019  100.009002  100.006733  100.005029  100.003751  100.002794  100.002079  100.001544  100.001146  100.000849  100.000629  100.000465  100.000343  100.000253  100.000187 |  100.0 |  88.127569 | 100.000187 |
    | 14:00 | 5.185446  5.005433  4.836794   4.68085  4.538602  4.410661  4.297215  4.198027   4.11248  4.039649  3.978385  3.927418  3.885439  3.851168  3.823409  3.801079  3.783226  3.769028  3.757789  3.748929  3.741971  3.736524  3.732274  3.728965  3.726397  3.724408  3.722872  3.721687  3.720775  3.720075  3.719538  3.719128  3.718814  3.718575  3.718393  3.718255   3.71815   3.71807   3.71801  3.717965  3.717931  3.717905  3.717886  3.717871   3.71786  3.717852  3.717846  3.717841  3.717838             3.717835 | 212.225932  200.353261  189.524779  179.764544  171.073573  163.429584  156.788497  151.087553  146.249683  142.188529  138.813508  136.034352  133.764756  131.924979  130.443403  129.257218  128.312409  127.563287  126.971733  126.506303  126.141311   125.85594  125.633437  125.460395  125.326142  125.222219  125.141947   125.08007  125.032468  124.995918  124.967905  124.946474  124.930107  124.917629  124.908132  124.900916  124.895441  124.891294  124.888158   124.88579  124.884005   124.88266   124.88165  124.880891  124.880322  124.879896  124.879578   124.87934  124.879163  124.879031 |  67.881379   66.045596   64.325812   62.735493   61.284841   59.980096   58.823164   57.811642   56.939236   56.196493   55.571725   55.051966   54.623857   54.274358   53.991272   53.763557   53.581491   53.436696    53.32208   53.231729   53.160769   53.105222   53.061872   53.028135   53.001945   52.981663   52.965991   52.953908   52.944611   52.937471   52.931998    52.92781   52.924612   52.922174   52.920318   52.918908   52.917838   52.917027   52.916414   52.915952   52.915603    52.91534   52.915142   52.914994   52.914883     52.9148   52.914737   52.914691   52.914656         52.91463 |  66.854459   65.054325   63.367938   61.808502    60.38602   59.106613   57.972148   56.980269   56.124804   55.396485   54.783849   54.274184   53.854388   53.511677   53.234087   53.010795   52.832265   52.690281   52.577891   52.489295   52.419712   52.365244   52.322736   52.289654   52.263973   52.244085   52.228718   52.216869   52.207752   52.200751   52.195384   52.191278   52.188142   52.185751   52.183931   52.182548   52.181499   52.180704   52.180103   52.179649   52.179307    52.17905   52.178856    52.17871   52.178601    52.17852   52.178459   52.178413   52.178379     52.178354 | 1.302742  1.277674  1.253918  1.231701  1.211218  1.192608  1.175953  1.161267  1.148504  1.137567  1.128313  1.120577  1.114178  1.108937  1.104679  1.101246  1.098497  1.096307  1.094571  1.093202  1.092125  1.091282  1.090624  1.090111  1.089714  1.089405  1.089167  1.088983  1.088842  1.088733   1.08865  1.088587  1.088538  1.088501  1.088473  1.088451  1.088435  1.088423  1.088413  1.088406  1.088401  1.088397  1.088394  1.088392   1.08839  1.088389  1.088388  1.088387  1.088387  1.088386 | 1.348723  1.349631  1.350553  1.351475   1.35238  1.353252  1.354075  1.354835  1.355524  1.356136   1.35667  1.357129  1.357517   1.35784  1.358107  1.358324    1.3585  1.358642  1.358754  1.358843  1.358914  1.358969  1.359013  1.359046  1.359073  1.359093  1.359109  1.359121   1.35913  1.359137  1.359143  1.359147   1.35915  1.359153  1.359155  1.359156  1.359157  1.359158  1.359159  1.359159   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916           1.35916 | 0.626833  0.617854  0.608998  0.600395  0.592173  0.584449  0.577322  0.570862  0.565111  0.560076  0.555739  0.552056  0.548971  0.546416  0.544323  0.542623  0.541254  0.540159  0.539287  0.538598  0.538054  0.537628  0.537295  0.537035  0.536833  0.536677  0.536556  0.536463  0.536391  0.536336  0.536293  0.536261  0.536236  0.536217  0.536203  0.536192  0.536184  0.536177  0.536173  0.536169  0.536166  0.536164  0.536163  0.536162  0.536161   0.53616   0.53616  0.536159  0.536159      0.536159 | -0.284696   -0.27501  -0.265738   -0.25701  -0.248935  -0.241596  -0.235043  -0.229292  -0.224326  -0.220101  -0.216556  -0.213619  -0.211211  -0.209256  -0.207681  -0.206423  -0.205423  -0.204633  -0.204012  -0.203526  -0.203146  -0.202851  -0.202623  -0.202446   -0.20231  -0.202205  -0.202125  -0.202063  -0.202016   -0.20198  -0.201953  -0.201932  -0.201916  -0.201904  -0.201895  -0.201888  -0.201883  -0.201879  -0.201876  -0.201874  -0.201873  -0.201871   -0.20187   -0.20187  -0.201869  -0.201869  -0.201869  -0.201868  -0.201868     -0.201868 | 0.669873  0.669028    0.6683  0.667687  0.667185  0.666786  0.666475   0.66624  0.666067  0.665942  0.665854  0.665794  0.665754  0.665728  0.665711  0.665702  0.665697  0.665694  0.665694  0.665695  0.665696  0.665697  0.665699    0.6657  0.665702  0.665703  0.665704  0.665705  0.665706  0.665707  0.665707  0.665708  0.665708  0.665708  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709      0.665709 | 212.993962 | 204.990599  189.671339  175.963815  163.832994  153.216772  144.028948  136.163738   129.50135  123.914021  119.271883  115.448125  112.323077  109.787055  107.742019  106.102187   104.79386  103.754685  102.932592  102.284577  101.775447  101.376631  101.065089  100.822347  100.633667  100.487342  100.374111  100.286671  100.219282  100.167448  100.127651  100.097154  100.073823  100.056007  100.042425  100.032087  100.024233  100.018274   100.01376  100.010347  100.007769  100.005826  100.004363  100.003263  100.002437  100.001818  100.001354  100.001008  100.000749  100.000556          100.000412 | 100.000349 | 0.869317  0.852016  0.835603  0.820238  0.806057  0.793161  0.781609  0.771415  0.762549  0.754946   0.74851  0.743127  0.738673  0.735022  0.732057  0.729665  0.727749  0.726223  0.725013  0.724058  0.723308   0.72272  0.722261  0.721904  0.721626  0.721411  0.721245  0.721117  0.721018  0.720943  0.720885   0.72084  0.720806   0.72078  0.720761  0.720746  0.720734  0.720726  0.720719  0.720714  0.720711  0.720708  0.720706  0.720704  0.720703  0.720702  0.720702  0.720701  0.720701         0.7207 | 3.490217  3.381586  3.279467  3.184709  3.097981  3.019724   2.95012  2.889092  2.836324  2.791295  2.753344  2.721717   2.69563  2.674306  2.657018    2.6431  2.631964  2.623103  2.616087  2.610553  2.606207  2.602803  2.600147  2.598079  2.596474   2.59523   2.59427  2.593529  2.592959  2.592521  2.592186  2.591929  2.591733  2.591583   2.59147  2.591383  2.591317  2.591268   2.59123  2.591202   2.59118  2.591164  2.591152  2.591143  2.591136  2.591131  2.591127  2.591124  2.591122        2.591121 | 212.993962  196.862863  182.355112  169.449781  158.097711  148.223648  139.730091  132.502505  126.415323  121.338079  117.141064  113.700022  110.899629  108.635711  106.816314  105.361848  104.204564  103.287593  102.563763  101.994334  101.547757  101.198523  100.926135  100.714209  100.549707  100.422299  100.323829   100.24788  100.189414  100.144494  100.110044  100.083672  100.063519  100.048144  100.036435  100.027532  100.020774  100.015652  100.011776  100.008847  100.006638  100.004973  100.003721  100.002781  100.002075  100.001547  100.001151  100.000856  100.000636  100.000472  100.000349 |  100.0 | 112.993962 | 100.000349 |
    | 14:30 | 5.491152  5.296408  5.110636  4.935443  4.772257  4.622234  4.486187  4.364523  4.257231  4.163901  4.083776  4.015834  3.958878  3.911625  3.872787   3.84113  3.815513  3.794917   3.77845   3.76535  3.754973  3.746786  3.740347    3.7353  3.731355   3.72828  3.725888  3.724031  3.722594  3.721483  3.720627  3.719967  3.719461  3.719072  3.718774  3.718547  3.718373  3.718241   3.71814  3.718064  3.718006  3.717962  3.717929  3.717904  3.717885  3.717871   3.71786  3.717852  3.717846             3.717841 | 233.131042  219.705796  207.252564  195.824656  185.456014  176.158765  167.922157   160.71313  154.478554  149.148886  144.642794   140.87214  137.746739  135.178436  133.084223  131.388337  130.023386  128.930713  128.060177   127.36955  126.823721  126.393794  126.056198   125.79185  125.585394  125.424541  125.299504  125.202515  125.127438  125.069437  125.024713  124.990289  124.963842  124.943558  124.928027  124.916156  124.907096  124.900193  124.894942  124.890954  124.887929  124.885638  124.883906  124.882598  124.881612  124.880869  124.880311  124.879891  124.879577  124.879341 |  70.998984   69.012974   67.118469   65.331844   63.667659   62.137726   60.750307   59.509572    58.41541   57.463627   56.646511   55.953635   55.372791   54.890906   54.494837   54.171991   53.910749   53.700708   53.532781   53.399186   53.293364   53.209865   53.144208   53.092739   53.052508   53.021142   52.996747   52.977817   52.963158   52.951831   52.943096   52.936371   52.931204    52.92724   52.924206   52.921886   52.920115   52.918766    52.91774   52.916961   52.916369   52.915922   52.915583   52.915328   52.915135    52.91499   52.914881   52.914799   52.914737        52.914691 |  69.911522   67.964078   66.106363   64.354434   62.722566   61.222343   59.861867   58.645227   57.572312   56.639012   55.837764   55.158343   54.588779   54.116252   53.727874   53.411298   53.155129   52.949167   52.784501   52.653499   52.549733   52.467856   52.403473   52.353004   52.313554   52.282797   52.258876   52.240313    52.22594   52.214833   52.206266   52.199672   52.194605   52.190719   52.187743   52.185468   52.183732    52.18241   52.181403   52.180639   52.180059    52.17962   52.179288   52.179038   52.178849   52.178706   52.178599   52.178519   52.178459     52.178413 | 1.344664   1.31805   1.29236  1.267848  1.244753  1.223287  1.203616  1.185852  1.170047  1.156187  1.144202  1.133975  1.125356  1.118172  1.112245  1.107398  1.103466  1.100298   1.09776  1.095739  1.094136   1.09287  1.091874  1.091093  1.090482  1.090005  1.089635  1.089347  1.089124  1.088952  1.088819  1.088717  1.088638  1.088578  1.088532  1.088497   1.08847  1.088449  1.088433  1.088422  1.088413  1.088406  1.088401  1.088397  1.088394  1.088392   1.08839  1.088389  1.088388  1.088387 | 1.347342    1.3482  1.349091  1.350004  1.350926   1.35184  1.352731  1.353581  1.354376  1.355106  1.355762  1.356342  1.356844  1.357274  1.357635  1.357936  1.358183  1.358385  1.358548  1.358678  1.358782  1.358865   1.35893  1.358982  1.359022  1.359053  1.359078  1.359097  1.359112  1.359123  1.359132  1.359139  1.359144  1.359148  1.359151  1.359153  1.359155  1.359156  1.359157  1.359158  1.359159  1.359159   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916           1.35916 | 0.641069  0.632141  0.623159  0.614233  0.605488  0.597052  0.589047  0.581583  0.574744  0.568588  0.563142  0.558401  0.554337  0.550901  0.548032  0.545662  0.543724  0.542152  0.540886  0.539874  0.539068   0.53843  0.537927  0.537532  0.537223  0.536981  0.536793  0.536647  0.536534  0.536446  0.536379  0.536327  0.536287  0.536256  0.536233  0.536215  0.536201  0.536191  0.536183  0.536177  0.536172  0.536169  0.536166  0.536164  0.536163  0.536162  0.536161   0.53616   0.53616      0.536159 | -0.300565  -0.290531  -0.280698  -0.271203  -0.262177  -0.253742  -0.245993  -0.238999  -0.232794  -0.227378  -0.222725  -0.218783  -0.215486  -0.212761  -0.210531  -0.208723  -0.207268  -0.206104  -0.205179  -0.204448  -0.203872  -0.203421  -0.203068  -0.202793   -0.20258  -0.202415  -0.202287  -0.202188  -0.202113  -0.202054   -0.20201  -0.201976   -0.20195   -0.20193  -0.201915  -0.201903  -0.201894  -0.201888  -0.201883  -0.201879  -0.201876  -0.201874  -0.201872  -0.201871   -0.20187   -0.20187  -0.201869  -0.201869  -0.201869     -0.201868 | 0.671426  0.670417  0.669514  0.668723  0.668048  0.667487  0.667031  0.666671  0.666394  0.666185  0.666031  0.665921  0.665843  0.665789  0.665753   0.66573  0.665715  0.665706  0.665701  0.665698  0.665697  0.665698  0.665698  0.665699  0.665701  0.665702  0.665703  0.665704  0.665705  0.665706  0.665706  0.665707  0.665707  0.665708  0.665708  0.665708  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709      0.665709 | 241.948195 | 232.667532  214.792455  198.537301  183.907422  170.880492  159.406865  149.411655  140.798425   133.45415  127.254946  122.071945  117.776757  114.246046  111.364981  109.029474  107.147326  105.638454  104.434451  103.477699  102.720251  102.122599  101.652469  101.283687  100.995154  100.769953  100.594584  100.458314  100.352646  100.270869  100.207703  100.159003  100.121524  100.092731  100.070649  100.053743  100.040821   100.03096  100.023446  100.017731  100.013389  100.010097  100.007604  100.005718  100.004295  100.003221  100.002413  100.001805  100.001349  100.001007           100.00075 | 100.000638 | 0.898211  0.879873  0.862154  0.845229  0.829266  0.814414  0.800791  0.788477  0.777511  0.767887  0.759559  0.752449  0.746453  0.741453  0.737327  0.733951  0.731212  0.729004  0.727236  0.725827  0.724709  0.723827  0.723132  0.722588  0.722162  0.721829  0.721571   0.72137  0.721215  0.721095  0.721002  0.720931  0.720876  0.720834  0.720802  0.720777  0.720759  0.720744  0.720733  0.720725  0.720719  0.720714  0.720711  0.720708  0.720706  0.720704  0.720703  0.720702  0.720702       0.720701 | 3.673884  3.556996  3.445118  3.339247   3.24029  3.149006   3.06595   2.99144  2.925539  2.868057  2.818589  2.776551  2.741245  2.711907  2.687761  2.668057  2.652097  2.639256  2.628984  2.620807  2.614328  2.609214  2.605192  2.602038  2.599573   2.59765  2.596155  2.594995  2.594096  2.593402  2.592866  2.592454  2.592137  2.591894  2.591708  2.591566  2.591457  2.591374  2.591311  2.591264  2.591227    2.5912  2.591179  2.591164  2.591152  2.591143  2.591136  2.591131  2.591127        2.591125 | 241.948195  223.265786  206.194716  190.754186  176.935898  164.703545  153.994023  144.720383  136.776288  130.041536  124.388048  119.685701  115.807446  112.633372  110.053543  107.969635  106.295557  104.957274  103.892079  103.047541  102.380288  101.854776  101.442092  101.118877  100.866363  100.669545  100.516476  100.397683  100.305675  100.234552  100.179676  100.137413  100.104923  100.079988  100.060884  100.046273  100.035115  100.026609  100.020134  100.015213  100.011478  100.008648  100.006508   100.00489   100.00367   100.00275  100.002059  100.001539  100.001149  100.000857  100.000638 |  100.0 | 141.948195 | 100.000638 |
    | 15:00 | 5.816019  5.610333  5.411061  5.219888  5.038455  4.868277  4.710651  4.566567  4.436649  4.321112   4.21976  4.132028  4.057039  3.993695  3.940764  3.896967  3.861042  3.831802  3.808164  3.789169  3.773983  3.761899  3.752322  3.744759  3.738805  3.734132  3.730474  3.727618  3.725393  3.723663  3.722321  3.721281  3.720478  3.719859  3.719381  3.719014  3.718733  3.718517  3.718352  3.718226   3.71813  3.718057  3.718001  3.717959  3.717926  3.717902  3.717884   3.71787   3.71786             3.717852 | 256.370661  241.534161  227.563825  214.534455  202.506968  191.524781  181.610905   172.76619   164.96903  158.176711   152.32829  147.348675  143.153387  139.653406  136.759586   134.38626  132.453864  130.890564  129.633029  128.626518  127.824488  127.187909  126.684424  126.287465  125.975393  125.730703  125.539313  125.389955  125.273649  125.183265  125.113164  125.058896  125.016962  124.984615  124.959707  124.940559  124.925864  124.914603  124.905989  124.899409   124.89439  124.890569  124.887663  124.885456  124.883783  124.882516  124.881559  124.880835   124.88029  124.879879 |  74.311987   72.214393   70.182212   68.232619   66.382361   64.646882   63.039399   61.570031   60.245124   59.066866    58.03328   57.138579   56.373837   55.727854   55.188069   54.741421   54.375058   54.076867   53.835806   53.642088   53.487227   53.363994   53.266325   53.189195    53.12848   53.080827   53.043524   53.014395   52.991701   52.974059   52.960371   52.949773   52.941582   52.935262   52.930396   52.926655   52.923783   52.921582   52.919899   52.918613   52.917632   52.916885   52.916317   52.915886   52.915559   52.915312   52.915124   52.914983   52.914877        52.914796 |  73.160189   71.103328   69.110611   67.198877    65.38455   63.682773   62.106506   60.665672   59.366494   58.211118   57.197603   56.320276   55.570385   54.936947   54.407644   53.969669   53.610421    53.31802   53.081641   52.891685   52.739831   52.618991   52.523218   52.447586   52.388051   52.341323   52.304744   52.276181   52.253928   52.236628   52.223206   52.212814   52.204782   52.198585   52.193813   52.190145   52.187329   52.185171    52.18352   52.182259   52.181297   52.180565   52.180008   52.179585   52.179265   52.179022   52.178838     52.1787   52.178595     52.178517 | 1.388382  1.360798  1.333757  1.307505  1.282295  1.258373  1.235966  1.215262  1.196403  1.179473  1.164495  1.151429  1.140184   1.13063  1.122606  1.115938  1.110449  1.105968  1.102336  1.099412  1.097071  1.095206  1.093726  1.092557  1.091635  1.090912  1.090345  1.089903  1.089558   1.08929  1.089082  1.088921  1.088796    1.0887  1.088626  1.088569  1.088525  1.088492  1.088466  1.088447  1.088432   1.08842  1.088412  1.088405    1.0884  1.088397  1.088394  1.088392   1.08839  1.088389 | 1.346064  1.346853  1.347686  1.348558  1.349458  1.350375  1.351293  1.352197   1.35307  1.353897  1.354665  1.355364  1.355988  1.356535  1.357008  1.357409  1.357746  1.358026  1.358255  1.358442  1.358592  1.358713  1.358809  1.358886  1.358946  1.358994  1.359031   1.35906  1.359083  1.359101  1.359114  1.359125  1.359133   1.35914  1.359145  1.359148  1.359151  1.359153  1.359155  1.359156  1.359157  1.359158  1.359159  1.359159   1.35916   1.35916   1.35916   1.35916   1.35916           1.35916 | 0.654955  0.646302  0.637455  0.628499  0.619537  0.610686  0.602072  0.593819  0.586044  0.578845  0.572296   0.56644   0.56129  0.556832  0.553027  0.549823  0.547156  0.544958  0.543164  0.541711  0.540542  0.539606  0.538862  0.538272  0.537807  0.537441  0.537154  0.536929  0.536754  0.536618  0.536512  0.536431  0.536367  0.536318  0.536281  0.536252   0.53623  0.536213    0.5362   0.53619  0.536182  0.536176  0.536172  0.536169  0.536166  0.536164  0.536163  0.536162  0.536161       0.53616 | -0.316604  -0.306523  -0.296448  -0.286502  -0.276818  -0.267529  -0.258762  -0.250626  -0.243204  -0.236548  -0.230679  -0.225584  -0.221228  -0.217551  -0.214487   -0.21196  -0.209896  -0.208224  -0.206879  -0.205803  -0.204948  -0.204272  -0.203739   -0.20332  -0.202993  -0.202737  -0.202538  -0.202384  -0.202264  -0.202172    -0.2021  -0.202045  -0.202003  -0.201971  -0.201946  -0.201927  -0.201913  -0.201902  -0.201894  -0.201887  -0.201882  -0.201879  -0.201876  -0.201874  -0.201872  -0.201871   -0.20187   -0.20187  -0.201869     -0.201869 | 0.673206  0.672057  0.670996  0.670035  0.669182  0.668444   0.66782  0.667306  0.666893  0.666569  0.666321  0.666135  0.665999  0.665901  0.665832  0.665785  0.665753  0.665731  0.665717  0.665709  0.665704  0.665701    0.6657    0.6657  0.665701  0.665701  0.665702  0.665703  0.665704  0.665705  0.665706  0.665706  0.665707  0.665707  0.665708  0.665708  0.665708  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709      0.665709 |  274.95935 | 264.430599  244.035018  225.211739  208.003529  192.428073  178.476126  166.111112  155.270336  145.867836  137.798666  130.944188  125.177833  120.370742  116.396771  113.136498  110.480072  108.328922  106.596446  105.207928  104.099878   103.21905  102.521267  101.970216  101.536282  101.195468  100.928442  100.719705  100.556886  100.430143  100.331676  100.255323  100.196225  100.150564  100.115346   100.08823  100.067385  100.051388  100.039131  100.029754  100.022592   100.01713   100.01297  100.009807  100.007406  100.005585  100.004206  100.003163  100.002376  100.001783          100.001336 |  100.00114 | 0.928295  0.909319  0.890698  0.872602  0.855206  0.838683  0.823188  0.808858  0.795792  0.784052  0.773656  0.764581  0.756766  0.750122  0.744539  0.739898  0.736075  0.732954  0.730424  0.728387  0.726755  0.725455  0.724424  0.723608  0.722966  0.722462  0.722067  0.721758  0.721517  0.721331  0.721185  0.721073  0.720986  0.720919  0.720868  0.720828  0.720797  0.720774  0.720756  0.720743  0.720732  0.720724  0.720718  0.720714   0.72071  0.720708  0.720706  0.720704  0.720703       0.720702 | 3.868052  3.745231  3.625859  3.510959  3.401542  3.298558  3.202842  3.115054   3.03564  2.964799  2.902478  2.848392  2.802055  2.762834  2.730004  2.702797  2.680453  2.662247  2.647517  2.635671  2.626196  2.618653  2.612672  2.607948  2.604228  2.601308  2.599022  2.597237  2.595846  2.594764  2.593925  2.593275  2.592773  2.592386  2.592087  2.591858  2.591682  2.591547  2.591444  2.591365  2.591305  2.591259  2.591224  2.591198  2.591178  2.591163  2.591151  2.591142  2.591136        2.591131 |  274.95935  253.788045  234.162402  216.137285   199.74369   184.98623  171.841922  160.260553  150.166742    141.4636   134.03767  127.764633  122.515181  118.160473  114.576721  111.648633  109.271667  107.353167  105.812603  104.581124  103.600689  102.822938  102.207976  101.723165  101.341996  101.043061  100.809168  100.626567  100.484308  100.373701  100.287867  100.221382  100.169977  100.130301   100.09973  100.076215  100.058157  100.044312  100.033713  100.025612  100.019431   100.01472  100.011137  100.008414  100.006348  100.004784    100.0036  100.002706  100.002031  100.001522   100.00114 |  100.0 |  174.95935 |  100.00114 |
    | 15:30 |  6.15413  5.941307  5.732419  5.529068   5.33292  5.145634  4.968792  4.803806  4.651831  4.513692  4.389826  4.280264  4.184642  4.102257  4.032133  3.973114  3.923949  3.883367  3.850144  3.823141  3.801332  3.783815  3.769813  3.758669  3.749832  3.742849  3.737346  3.733023  3.729634  3.726984  3.724916  3.723306  3.722054  3.721083  3.720331   3.71975  3.719302  3.718956  3.718691  3.718486   3.71833   3.71821  3.718118  3.718048  3.717995  3.717955  3.717924    3.7179  3.717883             3.717869 | 281.678565  265.615274  250.289427  235.789017  222.193948  209.572244  197.976334  187.439819  177.975134  169.572478   162.20028  155.807248  150.325792  145.676411   141.77248  138.524896  135.846099  133.653219  131.870231  130.429177  129.270621  128.343513  127.604665     127.018  126.553695  126.187314   125.89898  125.672629  125.495345  125.356791  125.248726  125.164604  125.099243   125.04855  125.009302  124.978966  124.955558  124.937524  124.923654  124.913001  124.904833   124.89858  124.893799   124.89015  124.887368   124.88525   124.88364  124.882419  124.881493  124.880791 |  77.760063   75.589684   73.459433   71.385656   69.385322   67.475374   65.671932   63.989396   62.439556   61.030811   59.767622   58.650297   57.675146   56.834976    56.11985   55.517974   55.016582   54.602732   54.263924   53.988546   53.766135   53.587495   53.444704   53.331054   53.240936   53.169717   53.113603    53.06951    53.03495   53.007925   52.986837   52.970415   52.957653   52.947752   52.940085   52.934159   52.929585   52.926061   52.923351   52.921269   52.919673   52.918451   52.917517   52.916803    52.91626   52.915846   52.915531   52.915293   52.915112        52.914975 |  76.541305   74.413073   72.324191   70.290685   68.329196   66.456338   64.687918   63.038055   61.518312   60.136923   58.898265   57.802638   56.846423   56.022569    55.32133   54.731142   54.239487   53.833673   53.501445   53.231415   53.013323   52.838152   52.698134    52.58669   52.498323   52.428487   52.373463   52.330226   52.296337   52.269837   52.249158   52.233055   52.220541   52.210832   52.203314   52.197503   52.193018   52.189563   52.186905   52.184864   52.183299     52.1821   52.181184   52.180485   52.179952   52.179546   52.179237   52.179003   52.178826     52.178691 | 1.433046  1.405028   1.37721  1.349809  1.323064  1.297223  1.272536  1.249238   1.22754  1.207609   1.18956  1.173451  1.159276  1.146973  1.136434  1.127514  1.120049  1.113862   1.10878  1.104638  1.101285  1.098588  1.096428  1.094707  1.093341  1.092261  1.091409   1.09074  1.090215  1.089804  1.089484  1.089234   1.08904   1.08889  1.088773  1.088683  1.088614   1.08856  1.088519  1.088487  1.088463  1.088444   1.08843  1.088419  1.088411  1.088405    1.0884  1.088396  1.088393  1.088391 | 1.344905  1.345615  1.346376  1.347184  1.348034  1.348917  1.349825  1.350742  1.351655  1.352546    1.3534  1.354202  1.354941  1.355608  1.356201  1.356717  1.357161  1.357536   1.35785  1.358109  1.358322  1.358494  1.358634  1.358745  1.358834  1.358905  1.358961  1.359005   1.35904  1.359067  1.359088  1.359104  1.359117  1.359127  1.359135  1.359141  1.359145  1.359149  1.359152  1.359154  1.359155  1.359157  1.359158  1.359158  1.359159  1.359159   1.35916   1.35916   1.35916           1.35916 | 0.668223  0.660004  0.651494  0.642752  0.633852  0.624887  0.615968  0.607213  0.598748  0.590694   0.58316  0.576233  0.569974  0.564412  0.559549  0.555361  0.551803  0.548817  0.546339  0.544303  0.542643    0.5413  0.540219  0.539356  0.538668  0.538123  0.537693  0.537354  0.537088  0.536879  0.536717   0.53659  0.536491  0.536415  0.536356   0.53631  0.536275  0.536247  0.536226   0.53621  0.536198  0.536188  0.536181  0.536176  0.536171  0.536168  0.536166  0.536164  0.536163      0.536162 | -0.332421  -0.322542  -0.312511  -0.302427  -0.292403  -0.282567  -0.273051  -0.263985  -0.255487  -0.247655  -0.240558  -0.234235  -0.228691  -0.223904  -0.219828  -0.216402  -0.213555  -0.211213  -0.209303  -0.207757  -0.206514  -0.205521  -0.204731  -0.204105  -0.203612  -0.203224   -0.20292  -0.202682  -0.202497  -0.202353  -0.202241  -0.202155  -0.202088  -0.202037  -0.201997  -0.201966  -0.201943  -0.201925  -0.201911  -0.201901  -0.201893  -0.201887  -0.201882  -0.201879  -0.201876  -0.201874  -0.201872  -0.201871   -0.20187      -0.20187 | 0.675167  0.673912  0.672721   0.67161  0.670592  0.669678  0.668876  0.668188  0.667613  0.667143  0.666768  0.666477  0.666256  0.666091   0.66597  0.665883  0.665822   0.66578  0.665751  0.665732  0.665719  0.665711  0.665707  0.665704  0.665703  0.665702  0.665702  0.665703  0.665704  0.665704  0.665705  0.665706  0.665706  0.665707  0.665707  0.665708  0.665708  0.665708  0.665708  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709      0.665709 | 311.845554 | 300.138987  277.342974  256.021361   236.24848  218.078227  201.540826  186.640485  173.354306  161.632684  151.401316  142.564733  135.011056  128.617495   123.25605  118.798833  115.122594  112.112173    109.6628  107.681311  106.086471   104.80862  103.788853  102.977958  102.335223  101.827259  101.426875   101.11206  100.865093  100.671768  100.520742  100.402988  100.311349  100.240162  100.184958  100.142223  100.109196  100.083712  100.064081  100.048982  100.037387  100.028497   100.02169  100.016486  100.012514  100.009486  100.007181   100.00543    100.0041  100.003092          100.002329 | 100.001993 | 0.958983  0.939738  0.920611  0.901754  0.883329  0.865509  0.848467  0.832368  0.817358  0.803557  0.791048  0.779874  0.770033  0.761485  0.754159  0.747955  0.742759  0.738452  0.734913  0.732028  0.729692  0.727812  0.726307  0.725107  0.724155  0.723402  0.722809  0.722342  0.721976  0.721689  0.721466  0.721292  0.721157  0.721052   0.72097  0.720908  0.720859  0.720822  0.720793  0.720771  0.720754  0.720741  0.720731  0.720723  0.720718  0.720713   0.72071  0.720707  0.720706       0.720704 |  4.06916  3.942683  3.818178  3.696597  3.578941  3.466224  3.359428  3.259448  3.167039  3.082764  3.006956  2.939702  2.880845  2.830009  2.786643  2.750075  2.719562  2.694341  2.673669  2.656851  2.643257  2.632331  2.623594  2.616636  2.611117  2.606755  2.603317  2.600615  2.598497   2.59684  2.595548  2.594541  2.593758  2.593152  2.592681  2.592318  2.592038  2.591822  2.591655  2.591528   2.59143  2.591355  2.591298  2.591254  2.591221  2.591195  2.591176  2.591161   2.59115        2.591142 | 311.845554  288.329231  266.245927  245.679473  226.695067  209.335647  193.618999   179.53593  167.049869  156.098053  146.594362  138.433579  131.496681  125.656616  120.783968  116.751992  113.440644   110.73946  108.549267  106.782892  105.365075  104.231821  103.329409  102.613226  102.046561  101.599435  101.247523  100.971201   100.75471  100.585447  100.453371  100.350508  100.270543  100.208489  100.160418  100.123241  100.094537   100.07241  100.055381  100.042296  100.032256  100.024565  100.018682  100.014188  100.010761  100.008151  100.006166  100.004658  100.003514  100.002648  100.001993 |  100.0 | 211.845554 | 100.001993 |
    | 16:00 | 6.499555  6.283185    6.0685  5.856924  5.650004  5.449372  5.256691  5.073587  4.901566  4.741927   4.59568  4.463477  4.345576  4.241833  4.151734  4.074454  4.008937  3.953986  3.908342  3.870754  3.840038  3.815106  3.794987  3.778836  3.765927  3.755652    3.7475  3.741054  3.735971  3.731973  3.728835  3.726378  3.724459  3.722961  3.721796   3.72089  3.720188  3.719644  3.719223  3.718898  3.718648  3.718455  3.718307  3.718194  3.718107   3.71804  3.717989   3.71795  3.717921             3.717898 | 308.714417  291.639808  275.160954   259.37168   244.36281  230.218847  217.014346  204.810209  193.650224   183.55827  174.536561  166.565283  159.603778  153.593218  148.460468  144.122668  140.491954  137.479831  135.000792  132.975006  131.330049  130.001763  128.934444  128.080524  127.399945  126.859364  126.431289  126.093239  125.826953  125.617683  125.453576  125.325148  125.224835  125.146628  125.085762  125.038475  125.001798  124.973397  124.951439   124.93449  124.921427  124.911374  124.903649  124.897722   124.89318  124.889706  124.887051  124.885026  124.883482  124.882308 |  81.282718   79.076161   76.886799   74.729143   72.618965   70.572907   68.607937   66.740635   64.986359   63.358358   61.866922   60.518711   59.316351   58.258376   57.339542   56.551437     55.8833   55.322906    54.85742   54.474103   54.160861   53.906601   53.701427   53.536712   53.405072   53.300281   53.217153   53.151415   53.099576   53.058801   53.026805   53.001751   52.982174   52.966905    52.95502   52.945784   52.938619   52.933071    52.92878   52.925469   52.922916   52.920951   52.919442   52.918283   52.917396   52.916717   52.916198   52.915802     52.9155        52.915271 |  79.995552   77.831845   75.684999   73.569244   71.500043   69.493719   67.566907   65.735867   64.015658   62.419271   60.956798   59.634769   58.455757   57.418328   56.517337   55.744537   55.089374   54.539863   54.083416   53.707543   53.400384   53.151061   52.949871   52.788355   52.659272   52.556515   52.475002   52.410541   52.359708   52.319725    52.28835   52.263782   52.244585   52.229614   52.217959   52.208902   52.201877   52.196436   52.192229   52.188982   52.186478   52.184552   52.183072   52.181936   52.181066     52.1804   52.179891   52.179503   52.179207     52.178982 | 1.477863  1.449883  1.421811   1.39383  1.366144   1.33898  1.312583    1.2872  1.263074  1.240432  1.219463  1.200312  1.183071  1.167767  1.154372  1.142802  1.132934  1.124614  1.117672  1.111934  1.107231  1.103404  1.100309   1.09782  1.095828  1.094241  1.092981  1.091983  1.091196  1.090577  1.090091  1.089711  1.089413  1.089181     1.089   1.08886  1.088751  1.088667  1.088601  1.088551  1.088512  1.088482  1.088459  1.088442  1.088428  1.088418   1.08841  1.088404  1.088399  1.088396 | 1.343872  1.344503  1.345183  1.345915  1.346695   1.34752  1.348384  1.349278   1.35019  1.351106  1.352009  1.352885  1.353718  1.354494  1.355204   1.35584  1.356402  1.356888  1.357304  1.357655  1.357947  1.358187  1.358384  1.358544  1.358673  1.358776  1.358858  1.358923  1.358975  1.359016  1.359048  1.359073  1.359092  1.359108   1.35912  1.359129  1.359136  1.359142  1.359146   1.35915  1.359152  1.359154  1.359156  1.359157  1.359158  1.359158  1.359159  1.359159   1.35916           1.35916 | 0.680687  0.673001  0.664968  0.656621  0.648008  0.639193   0.63026  0.621309  0.612453  0.603814  0.595517  0.587677  0.580393  0.573742  0.567771  0.562498  0.557914  0.553984   0.55066   0.54788   0.54558  0.543693  0.542157  0.540916  0.539919  0.539121  0.538486  0.537983  0.537585  0.537271  0.537025  0.536832  0.536681  0.536563  0.536471    0.5364  0.536344  0.536301  0.536268  0.536243  0.536223  0.536208  0.536196  0.536187   0.53618  0.536175  0.536171  0.536168  0.536166      0.536164 | -0.347707  -0.338215  -0.328458  -0.318505  -0.308444  -0.298379  -0.288429  -0.278724  -0.269395  -0.260567  -0.252348  -0.244824  -0.238051  -0.232053  -0.226824  -0.222331  -0.218522  -0.215331  -0.212687  -0.210517   -0.20875  -0.207321  -0.206173  -0.205256  -0.204526  -0.203948  -0.203491  -0.203132   -0.20285   -0.20263  -0.202458  -0.202324  -0.202219  -0.202139  -0.202076  -0.202028   -0.20199  -0.201962  -0.201939  -0.201922  -0.201909  -0.201899  -0.201892  -0.201886  -0.201882  -0.201878  -0.201876  -0.201874  -0.201872     -0.201871 | 0.677259  0.675933  0.674647  0.673417  0.672258  0.671186  0.670212  0.669346  0.668594  0.667954  0.667425  0.666996  0.666657  0.666396  0.666198  0.666051  0.665944  0.665867  0.665813  0.665775   0.66575  0.665732  0.665721  0.665713  0.665709  0.665706  0.665705  0.665704  0.665704  0.665704  0.665705  0.665705  0.665706  0.665706  0.665707  0.665707  0.665708  0.665708  0.665708  0.665708  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709      0.665709 | 352.270855 | 339.494892   314.49817  290.835416  268.605369  247.891828  228.759854   211.25232  195.387177  181.155743  168.522354  157.425533  147.780731  139.484432  132.419269  126.459627  121.477176   117.34583  113.945776  111.166387  108.908014  107.082783  105.614595  104.438555  103.500036  102.753557  102.161599  101.693449  101.324133  101.033456  100.805165  100.626233  100.486258  100.376961  100.291769  100.225481  100.173988  100.134054  100.103133  100.079229  100.060778  100.046559  100.035616  100.027207  100.020756  100.015813  100.012031  100.009141  100.006937  100.005257           100.00398 | 100.003416 | 0.989735  0.970541  0.951268   0.93204  0.912998  0.894296  0.876104  0.858592  0.841931  0.826278  0.811767  0.798502  0.786548  0.775928  0.766626  0.758586  0.751725  0.745936  0.741105   0.73711  0.733834  0.731168  0.729012  0.727277  0.725889  0.724782  0.723904  0.723209   0.72266  0.722228  0.721889  0.721624  0.721417  0.721255  0.721129  0.721031  0.720955  0.720896   0.72085  0.720815  0.720788  0.720767  0.720751  0.720739   0.72073  0.720723  0.720717  0.720713   0.72071       0.720707 | 4.273711  4.145684  4.018315  3.892433   3.76895  3.648839  3.533109  3.422759  3.318731   3.22186  3.132816  3.052059  2.979816  2.916065  2.860552  2.812825  2.772279   2.73821  2.709866  2.686496  2.667377  2.651844    2.6393  2.629224  2.621168  2.614752  2.609661  2.605634  2.602457  2.599959  2.597997  2.596462  2.595262  2.594326  2.593597  2.593031  2.592592  2.592251  2.591988  2.591785  2.591629  2.591508  2.591416  2.591345   2.59129  2.591249  2.591217  2.591193  2.591174         2.59116 | 352.270855  326.628958  302.268626  279.295253  257.801259  237.862181  219.532951  202.844697  187.802414   174.38385  162.539881  152.196482  143.258236  135.613086  129.137864  123.704023  119.183018   115.45088  112.391708  109.900004  107.881911  106.255545  104.950645  103.907759  103.077178  102.417746  101.895671  101.483407  101.158627  100.903332   100.70307  100.546288  100.423776  100.328214  100.253805  100.195965  100.151078    100.1163  100.089397  100.068618  100.052594  100.040255  100.030769  100.023485  100.017902  100.013628   100.01036  100.007866  100.005964  100.004517  100.003416 |  100.0 | 252.270855 | 100.003416 |
    | 16:30 | 6.846601  6.629925   6.41299  6.196998  5.983295  5.773367  5.568801  5.371248  5.182358  5.003708  4.836718  4.682562  4.542097  4.415805  4.303771  4.205687  4.120904  4.048494  3.987339  3.936212   3.89386   3.85906   3.83067  3.807655  3.789098  3.774208   3.76231  3.752837   3.74532  3.739372  3.734677  3.730982  3.728079  3.725803  3.724022  3.722632  3.721547  3.720704  3.720048  3.719539  3.719145  3.718841  3.718605  3.718424  3.718284  3.718177  3.718094  3.718031  3.717983             3.717946 | 337.078735  319.228363   301.82708  284.968858  268.748541  253.259328  238.589746  224.820242  212.019546  200.241107  189.519948  179.870339  171.284669   163.73377  157.168773  151.524327  146.722796  142.678925  139.304424  136.512007  134.218604  132.347611  130.830227  129.606006  128.622814  127.836365  127.209526  126.711482  126.316893  126.005074  125.759242  125.565856  125.414037   125.29508  125.202041  125.129402  125.072785  125.028729  124.994502  124.967952  124.947389  124.931488  124.919209  124.909741  124.902452  124.896847  124.892544  124.889245  124.886719  124.884787 |  84.821903   82.612229   80.399926   78.197223   76.017879   73.877021   71.790851   69.776196    67.84989   66.028012   64.325035   62.752946    61.32048   60.032556   58.890022   57.889761   57.025137   56.286699   55.663035   55.141644   54.709732   54.354843   54.065325   53.830614   53.641374   53.489523   53.368183   53.271578   53.194916   53.134257   53.086387   53.048699   53.019093   52.995883   52.977724   52.963542   52.952485    52.94388   52.937194   52.932007   52.927989   52.924882   52.922482   52.920632   52.919208   52.918112   52.917271   52.916627   52.916133        52.915755 |  83.466009   81.299245   79.129903   76.969976   74.832953   72.733669   70.688011    68.71248   66.823581   65.037083   63.367176   61.825616   60.420968   59.158054   58.037707   57.056871   56.209038   55.484939   54.873386   54.362121   53.938596   53.590598   53.306703    53.07655   52.890985   52.742083   52.623099    52.52837   52.453197   52.393716   52.346775   52.309819   52.280788   52.258029   52.240222   52.226316   52.215474   52.207036   52.200479   52.195393   52.191453   52.188406   52.186053   52.184239   52.182842   52.181768   52.180944   52.180311   52.179827     52.179457 | 1.522125  1.494577  1.466706  1.438651  1.410581  1.382689  1.355188  1.328315  1.302314  1.277433  1.253907  1.231946  1.211724   1.19336  1.176919  1.162405  1.149765  1.138899  1.129668  1.121914  1.115464  1.110145  1.105794  1.102258  1.099402  1.097106  1.095269  1.093806  1.092643  1.091723  1.090996  1.090424  1.089974  1.089621  1.089345   1.08913  1.088962  1.088831  1.088729   1.08865  1.088589  1.088542  1.088506  1.088477  1.088456  1.088439  1.088426  1.088417  1.088409  1.088403 | 1.342963  1.343517  1.344118  1.344769   1.34547  1.346222   1.34702  1.347862  1.348738   1.34964  1.350553  1.351465  1.352357  1.353216  1.354026  1.354775  1.355455   1.35606  1.356591  1.357049  1.357438  1.357765  1.358037   1.35826  1.358442   1.35859  1.358709  1.358804   1.35888   1.35894  1.358988  1.359026  1.359055  1.359079  1.359097  1.359111  1.359122  1.359131  1.359138  1.359143  1.359147   1.35915  1.359153  1.359154  1.359156  1.359157  1.359158  1.359158  1.359159          1.359159 | 0.692235  0.685133  0.677659  0.669827   0.66166  0.653199  0.644498  0.635629  0.626683  0.617766  0.608994  0.600492  0.592379  0.584766  0.577741  0.571369  0.565685  0.560695  0.556379  0.552696  0.549593  0.547007  0.544873  0.543125  0.541706  0.540559  0.539638  0.538902  0.538316  0.537851  0.537483  0.537194  0.536965  0.536787  0.536646  0.536537  0.536452  0.536385  0.536333  0.536293  0.536262  0.536238   0.53622  0.536205  0.536194  0.536186  0.536179  0.536174   0.53617      0.536168 | -0.362235  -0.353255  -0.343938  -0.334327  -0.324479  -0.314473  -0.304403   -0.29438  -0.284527  -0.274974   -0.26585  -0.257273  -0.249341  -0.242128  -0.235676  -0.229996  -0.225071  -0.220859  -0.217302  -0.214333  -0.211879  -0.209869  -0.208235  -0.206915  -0.205855  -0.205008  -0.204334    -0.2038  -0.203378  -0.203045  -0.202784   -0.20258   -0.20242  -0.202295  -0.202198  -0.202123  -0.202064  -0.202019  -0.201984  -0.201957  -0.201936   -0.20192  -0.201908  -0.201898  -0.201891  -0.201885  -0.201881  -0.201878  -0.201875     -0.201874 | 0.679431  0.678067  0.676721  0.675408  0.674142   0.67294  0.671816  0.670785  0.669856  0.669038  0.668333  0.667741  0.667254  0.666864  0.666558  0.666323  0.666146  0.666016  0.665921  0.665853  0.665805  0.665771  0.665748  0.665732  0.665721  0.665715   0.66571  0.665708  0.665706  0.665705  0.665705  0.665705  0.665706  0.665706  0.665706  0.665707  0.665707  0.665707  0.665708  0.665708  0.665708  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709      0.665709 | 395.750601 | 382.047624  355.121347  329.354621  304.863387  281.754038  260.119704  240.036492  221.559904  204.721776  189.528058  175.957758  163.963316  153.472509  144.391838   136.61111  130.008811  124.457693   119.83007  116.002367  112.858648  110.293017  108.210949   106.52972   105.17814  104.095805   103.23208  102.544942  101.999811  101.568435  101.227869  100.959577  100.748648  100.583134  100.453494  100.352129  100.273007  100.211348  100.163376   100.12611  100.097205   100.07482   100.05751  100.044144  100.033839  100.025904  100.019804   100.01512  100.011529   100.00878          100.006678 | 100.005749 | 1.020067  1.001193  0.982083  0.962832  0.943554   0.92438  0.905457  0.886948  0.869022  0.851849  0.835595  0.820408  0.806408  0.793683   0.78228  0.772206  0.763426  0.755872  0.749453  0.744058  0.739567  0.735864  0.732833   0.73037   0.72838   0.72678    0.7255  0.724479  0.723669  0.723027   0.72252  0.722121  0.721808  0.721562  0.721369  0.721219  0.721102  0.721011   0.72094  0.720885  0.720842  0.720809  0.720784  0.720764  0.720749  0.720737  0.720728  0.720722  0.720716       0.720712 | 4.478406  4.350696   4.22253  4.094593  3.967665  3.842614  3.720383  3.601962  3.488357  3.380544   3.27942   3.18575  3.100115  3.022875  2.954148  2.893811  2.841525  2.796769  2.758895  2.727177  2.700865  2.679219  2.661542  2.647199  2.635627  2.626337  2.618909  2.612994  2.608298  2.604582  2.601649  2.599339  2.597525  2.596102  2.594989   2.59412  2.593442  2.592914  2.592504  2.592186   2.59194  2.591749  2.591602  2.591489  2.591401  2.591334  2.591283  2.591243  2.591213         2.59119 | 395.750601  368.269771  341.888623  316.727089   292.89739  270.500404  249.621863  230.328596   212.66511   196.65084  182.278417  169.513261  158.294712  148.538719  140.141936  132.986858  126.947463  121.894814  117.702083   114.24865  111.423066  109.124881   107.26547  105.768038  104.567061  103.607346  102.842909  102.235783  101.754863   101.37483  101.075187  100.839418   100.65427  100.509143  100.395589  100.306891  100.237724  100.183874  100.142015  100.109529  100.084354  100.064875  100.049825  100.038214  100.029269  100.022388  100.017103  100.013048  100.009941  100.007565  100.005749 |  100.0 | 295.750601 | 100.005749 |
    | 17:00 | 7.189981  6.975837  6.759814  6.542882  6.326162  6.110927  5.898592  5.690692  5.488848  5.294716  5.109921  4.935977  4.774203  4.625643  4.490996  4.370575  4.264297  4.171706  4.092032  4.024263  3.967232    3.9197  3.880425   3.84822   3.82199  3.800749  3.783637  3.769912  3.758946  3.750215  3.743284  3.737797  3.733464   3.73005  3.727365  3.725258  3.723608  3.722317  3.721309  3.720524  3.719913  3.719438   3.71907  3.718784  3.718563  3.718393  3.718261   3.71816  3.718082             3.718022 | 366.328806   347.94908  329.872668  312.189771  294.994081  278.381082  262.445835  247.280267  232.969997  219.590851  207.205278  195.858975   185.57811   176.36751  168.210174   161.06827  154.885599  149.591271  145.104132  141.337423  138.203142   135.61573  133.494866  131.767304  130.367863  129.239705  128.334102  127.609865  127.032587  126.573801  126.210151  125.922603  125.695732  125.517101  125.376723  125.266607  125.180378  125.112967  125.060353  125.019352   124.98745  124.962664  124.943436  124.928542   124.91702   124.90812  124.901255  124.895967  124.891899  124.888774 |  88.323702    86.13986   83.936851   81.724568    79.51445   77.319476   75.154074   73.033901   70.975488   68.995724   67.111174   65.337282   63.687507   62.172488   60.799354   59.571297   58.487467   57.543225   56.730706   56.039594   55.457989   54.973252   54.572726   54.244301   53.976798   53.760188   53.585678    53.44571   53.333882   53.244841   53.174159   53.118203   53.074013   53.039193   53.011814   52.990327   52.973495   52.960333   52.950057   52.942049   52.935816   52.930974   52.927217   52.924306   52.922055   52.920315   52.918974    52.91794   52.917145        52.916535 |  86.899805   84.758372   82.598144   80.428822   78.261623   76.109274   73.985922   71.906921   69.888482   67.947163    66.09921   64.359766   62.742029    61.25643   59.909961   58.705753    57.64297   56.717065   55.920324   55.242633   54.672322   54.196998    53.80425   53.482203   53.219895   53.007491    52.83637   52.699121   52.589464   52.502152   52.432843   52.377973   52.334641   52.300497    52.27365    52.25258   52.236075   52.223169   52.213093    52.20524   52.199128    52.19438   52.190696   52.187842   52.185634   52.183929   52.182613     52.1816    52.18082     52.180221 | 1.565216  1.538422  1.511125   1.48343  1.455465  1.427384  1.399366  1.371613  1.344351  1.317817   1.29226  1.267923   1.24503  1.223777  1.204314   1.18674  1.171092   1.15735   1.14544  1.135247  1.126623  1.119402  1.113412  1.108485  1.104461  1.101196   1.09856  1.096443   1.09475    1.0934  1.092328  1.091479  1.090808  1.090279  1.089863  1.089537  1.089281  1.089081  1.088925  1.088803  1.088708  1.088635  1.088578  1.088533  1.088499  1.088473  1.088452  1.088437  1.088424  1.088415 | 1.342171  1.342654   1.34318  1.343752  1.344373  1.345044  1.345766  1.346536  1.347352  1.348208  1.349095  1.350002  1.350915  1.351819  1.352698  1.353537  1.354322  1.355044  1.355693  1.356269   1.35677    1.3572  1.357564  1.357868  1.358121  1.358328  1.358496  1.358633  1.358743   1.35883  1.358901  1.358956     1.359  1.359035  1.359063  1.359084  1.359101  1.359114  1.359125  1.359133  1.359139  1.359144  1.359148  1.359151  1.359153  1.359155  1.359156  1.359157  1.359158          1.359159 | 0.702814   0.69631  0.689432   0.68218  0.674559  0.666589  0.658301  0.649739  0.640966  0.632062  0.623123  0.614261  0.605595  0.597248  0.589336  0.581961  0.575202  0.569111   0.56371  0.558996  0.554939  0.551492  0.548599  0.546195  0.544215  0.542598  0.541286  0.540227  0.539377  0.538698  0.538157  0.537728  0.537388   0.53712  0.536909  0.536744  0.536614  0.536512  0.536433  0.536371  0.536323  0.536285  0.536256  0.536234  0.536216  0.536203  0.536192  0.536184  0.536178      0.536174 | -0.375855  -0.367462  -0.358692  -0.349567  -0.340122  -0.330406  -0.320486  -0.310447   -0.30039  -0.290432  -0.280697  -0.271317  -0.262415  -0.254101  -0.246464  -0.239564   -0.23343   -0.22806  -0.223427  -0.219483  -0.216165  -0.213404  -0.211128  -0.209266  -0.207755  -0.206536  -0.205558  -0.204776  -0.204154  -0.203661  -0.203271  -0.202963  -0.202722  -0.202532  -0.202384  -0.202268  -0.202178  -0.202107  -0.202053   -0.20201  -0.201977  -0.201952  -0.201932  -0.201917  -0.201906  -0.201897   -0.20189  -0.201885  -0.201881     -0.201878 | 0.681635  0.680262  0.678889  0.677528  0.676192  0.674895  0.673654  0.672482  0.671395  0.670405  0.669522  0.668751  0.668093  0.667545  0.667099  0.666745  0.666469  0.666258  0.666101  0.665985    0.6659   0.66584  0.665797  0.665767  0.665746  0.665732  0.665722  0.665716  0.665711  0.665709  0.665707  0.665707  0.665706  0.665706  0.665706  0.665707  0.665707  0.665707  0.665707  0.665708  0.665708  0.665708  0.665708  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709      0.665709 | 441.664962 | 427.206229  398.682633  371.118565  344.640702  319.370874   295.42284  272.898761  251.885504  232.451006  214.640966  198.476193  183.950936  171.032489  159.662273  149.758419  141.219726  133.930626  127.766684  122.600096  118.304674  114.759953  111.854222   109.48645  107.567201  106.018746  104.774573  103.778514  102.983659  102.351191  101.849242   101.45182  101.137843  100.890293  100.695487  100.542464   100.42247  100.328531  100.255108  100.197811  100.153166  100.118432  100.091449  100.070517  100.054303  100.041761  100.032074  100.024602  100.018846  100.014418          100.011016 | 100.009511 | 1.049564  1.031227  1.012532  0.993552  0.974371  0.955096  0.935846  0.916761  0.897995  0.879713  0.862085  0.845281  0.829458  0.814753  0.801275  0.789092  0.778236  0.768695   0.76042  0.753333  0.747334  0.742309  0.738139  0.734708  0.731905   0.72963  0.727793  0.726318  0.725137  0.724197  0.723449  0.722857  0.722389  0.722021  0.721731  0.721503  0.721325  0.721185  0.721076  0.720991  0.720925  0.720874  0.720834  0.720803  0.720779  0.720761  0.720746  0.720736  0.720727       0.720721 | 4.680222  4.554441  4.427289  4.299309  4.171141  4.043515  3.917254  3.793261  3.672503  3.555979  3.444686   3.33957  3.241473  3.151083  3.068891  2.995153  2.929885   2.87287  2.823691  2.781771  2.746427  2.716923  2.692511  2.672472  2.656134  2.642894   2.63222  2.623655  2.616809  2.611357  2.607027  2.603599  2.600891  2.598757  2.597079  2.595762   2.59473  2.593923  2.593293  2.592802   2.59242  2.592123  2.591892  2.591714  2.591576  2.591469  2.591387  2.591324  2.591275        2.591238 | 441.664962  412.688919   384.60814  357.551057  331.642788  307.002115  283.738069  261.946269  241.705181  223.072555  206.082341  190.742431  177.033538  164.909501   154.29913  145.109552  137.230803  130.541223  124.913117  120.218133  116.331894  113.137594  110.528463  108.409135  106.696105  105.317466  104.212176  103.329032   102.62551  102.066594   101.62365  101.273403  100.997032  100.779378  100.608282  100.474021  100.368843  100.286581  100.222345  100.172262  100.133273  100.102966  100.079442   100.06121  100.047098  100.036192  100.027775  100.021288  100.016295  100.012457  100.009511 |  100.0 | 341.664962 | 100.009511 |
    | 17:30 | 7.524907  7.315729  7.103354   6.88854  6.672185  6.455334  6.239182   6.02507  5.814472  5.608966  5.410195  5.219809  5.039391  4.870382  4.713989  4.571116  4.442301  4.327693  4.227049  4.139779  4.065008  4.001655  3.948521  3.904364  3.867965  3.838176  3.813948  3.794351  3.778575  3.765926  3.755822  3.747776  3.741387  3.736327  3.732328  3.729175  3.726693  3.724743  3.723215  3.722018  3.721082  3.720352  3.719783   3.71934  3.718996  3.718729  3.718522  3.718362  3.718239             3.718143 | 395.994739  377.335421  358.838477  340.587982   322.67302  305.186694  288.224703   271.88343  256.257527  241.437016  227.504001  214.529144  202.568183  191.658806  181.818285  173.042229  165.304735  158.560025  152.745457  147.785557  143.596578  140.091058  137.181918  134.785762  132.825262   131.23062   129.94024  128.900778   128.06675  127.399883  126.868314  126.445756  126.110684  125.845582  125.636275  125.471336  125.341598  125.239722  125.159857  125.097346  125.048493  125.010372  124.980669  124.957558  124.939602   124.92567  124.914876  124.906523   124.90007   124.89509 |  91.739296   89.606094   87.440279   85.249596   83.043201   80.831748   78.627423   76.443904   74.296216   72.200458   70.173384   68.231813   66.391908   64.668341   63.073442   61.616416   60.302764   59.133982   58.107612   57.217632   56.455112   55.809037   55.267174   54.816858   54.445662    54.14187   53.894795   53.694943   53.534051    53.40506   53.302017   53.219964    53.15481   53.103205   53.062425   53.030268    53.00496   52.985079   52.969488   52.957282   52.947741   52.940294   52.934491   52.929976   52.926467   52.923745   52.921636   52.920003   52.918742        52.917769 |  90.249071   88.157294   86.033537   83.885396   81.721848    79.55334   77.391822   75.250705   73.144723   71.089664   69.101954   67.198087   65.393912   63.703816   62.139888   60.711157   59.423015   58.276929   57.270491   56.397794   55.650081   55.016554   54.485212   54.043642   53.679654   53.381761   53.139484   52.943513   52.785746    52.65926   52.558218   52.477758   52.413869   52.363266   52.323279   52.291746    52.26693   52.247435   52.232147   52.220177   52.210821    52.20352   52.197829   52.193402   52.189961   52.187292   52.185223   52.183623   52.182386     52.181432 | 1.606618  1.580832  1.554409  1.527424  1.499973  1.472169  1.444155  1.416094  1.388176  1.360614  1.333638  1.307494  1.282426  1.258671  1.236443  1.215919  1.197227  1.180442  1.165576  1.152587  1.141383  1.131834  1.123784  1.117066  1.111508  1.106945  1.103226  1.100211   1.09778  1.095828  1.094267  1.093023  1.092035  1.091252  1.090632  1.090144  1.089759  1.089457   1.08922  1.089035   1.08889  1.088776  1.088688   1.08862  1.088566  1.088525  1.088493  1.088468  1.088449  1.088434 | 1.341486  1.341904  1.342362  1.342861  1.343405  1.343997  1.344637  1.345328  1.346069  1.346858   1.34769  1.348558  1.349454  1.350363  1.351273  1.352168  1.353031  1.353849  1.354608  1.355301   1.35592  1.356465  1.356938  1.357341  1.357681  1.357964  1.358199   1.35839  1.358546  1.358673  1.358774  1.358855   1.35892  1.358971  1.359012  1.359044   1.35907   1.35909  1.359105  1.359118  1.359127  1.359135   1.35914  1.359145  1.359149  1.359151  1.359153  1.359155  1.359156          1.359157 |  0.71241  0.706496  0.700219  0.693571  0.686546  0.679148  0.671389  0.663291  0.654892  0.646243  0.637416  0.628495  0.619584  0.610798  0.602258  0.594085  0.586389  0.579263  0.572775  0.566965  0.561844  0.557397  0.553589  0.550368  0.547672  0.545439  0.543605  0.542109  0.540896  0.539918  0.539134  0.538508  0.538009  0.537613  0.537299  0.537052  0.536857  0.536703  0.536583  0.536489  0.536415  0.536357  0.536312  0.536278   0.53625  0.536229  0.536213    0.5362  0.536191      0.536183 |  -0.38848  -0.380707  -0.372543  -0.363995  -0.355077  -0.345818  -0.336258  -0.326455  -0.316482  -0.306431  -0.296409  -0.286537  -0.276943  -0.267754   -0.25909  -0.251051  -0.243715  -0.237127  -0.231306  -0.226237  -0.221886  -0.218197  -0.215105  -0.212539  -0.210429  -0.208707   -0.20731  -0.206184  -0.205281   -0.20456  -0.203986   -0.20353   -0.20317  -0.202886  -0.202663  -0.202487  -0.202349  -0.202242  -0.202158  -0.202092  -0.202041  -0.202002  -0.201971  -0.201947  -0.201929  -0.201915  -0.201904  -0.201895  -0.201889     -0.201884 | 0.683829  0.682471  0.681098  0.679721  0.678349  0.676995  0.675672  0.674395   0.67318  0.672042  0.670993  0.670047  0.669209  0.668485  0.667873  0.667368   0.66696  0.666638  0.666389  0.666201   0.66606  0.665957  0.665882  0.665828   0.66579  0.665763  0.665744  0.665731  0.665722  0.665716  0.665712   0.66571  0.665708  0.665708  0.665707  0.665707  0.665707  0.665707  0.665707  0.665708  0.665708  0.665708  0.665708  0.665708  0.665709  0.665709  0.665709  0.665709  0.665709      0.665709 | 489.279607 | 474.259572  444.520281   415.52256  387.398474  360.278984    334.2914  309.556408  286.184728  264.273538  243.902833  225.131952  207.996586  192.506573  178.644801  166.367477  155.605888  146.269615  138.250962  131.430204  125.681126  120.876329  116.891859  113.610861  110.926118  108.741538  106.972705  105.546715  104.401513  103.484915  102.753489  102.171392  101.709261  101.343183  101.053785  100.825441  100.645594  100.504184  100.393179  100.306178  100.238095  100.184897  100.143389   100.11105   100.08589  100.066343  100.051177  100.039427  100.030336  100.023312          100.017891 | 100.015487 | 1.077877  1.060246  1.042169  1.023696  1.004891   0.98583   0.96661  0.947341  0.928153  0.909192  0.890616  0.872594  0.855297  0.838888  0.823519  0.809313  0.796363  0.784724  0.774407  0.765386    0.7576  0.750959  0.745359  0.740683  0.736813  0.733635  0.731044  0.728943  0.727249  0.725889  0.724801  0.723934  0.723245  0.722698  0.722267  0.721926  0.721658  0.721447  0.721282  0.721153  0.721052  0.720973  0.720911  0.720863  0.720826  0.720797  0.720775  0.720757  0.720744       0.720734 | 4.876459  4.753967  4.629371  4.503091  4.375628  4.247572  4.119606  3.992505   3.86713  3.744413   3.62534  3.510912  3.402108  3.299834  3.204872   3.11783    3.0391  2.968839  2.906966  2.853177  2.806985  2.767768  2.734818  2.707395  2.684761  2.666217  2.651122  2.638904  2.629062  2.621167  2.614858  2.609833  2.605841   2.60268  2.600181   2.59821  2.596659   2.59544  2.594484  2.593736  2.593151  2.592694  2.592339  2.592062  2.591847   2.59168   2.59155   2.59145  2.591373        2.591313 | 489.279607   459.19787  429.791513  401.192625  373.533373  346.943681  321.548464  297.464444  274.796652  253.634755  234.049421   216.08899  199.776772  185.109304  172.055871  160.559496  150.539475  141.895292   134.51161  128.263823  123.023634   118.66414  115.064027  112.110682  109.702173  107.748206  106.170261  104.901112  103.883957  103.071325  102.423914  101.909432  101.501524  101.178794  100.923954  100.723092   100.56505  100.440905  100.343543  100.267305  100.207697  100.161161  100.124882  100.096641  100.074687  100.057645  100.044433  100.034206  100.026299  100.020194  100.015487 |  100.0 | 389.279607 | 100.015487 |
    | 18:00 |  7.84714   7.64498   7.43857  7.228489  7.015433  6.800229  6.583847  6.367404  6.152162  5.939527  5.731024  5.528264  5.332902  5.146566  4.970789  4.806922   4.65605  4.518931  4.395943  4.287072  4.191929  4.109802  4.039727  3.980571  3.931115  3.890126  3.856413  3.828871  3.806502  3.788426  3.773884   3.76223  3.752921  3.745508   3.73962  3.734955  3.731266  3.728355  3.726062  3.724259  3.722844  3.721734  3.720866  3.720188  3.719659  3.719246  3.718925  3.718676  3.718482             3.718332 | 425.595105    406.9033  388.240203  369.682603  351.312971  333.219023  315.492941    298.2302  281.527941  265.482846  250.188534    235.7325  222.192735  209.634219  198.105583  187.636299   178.23477  169.887673  162.560741  156.201004  150.740262  146.099375  142.192863  138.933293  136.235054  134.017275  132.205807  130.734348  129.544828   128.58727  127.819279  127.205322  126.715911  126.326781  126.018099  125.773749  125.580701   125.42846  125.308605    125.2144  125.140472  125.082544  125.037219  125.001807  124.974178  124.952651  124.935902  124.922888  124.912789  124.904963 |  95.025437   92.963805   90.858831   88.716413   86.543657   84.349003   82.142332   79.935031   77.739991   75.571533   73.445206   71.377456   69.385141   67.484884   65.692305   64.021174   62.482582   61.084238   59.830002   58.719726   57.749453   56.911918   56.197291   55.594019   55.089665   54.671654   54.327851   54.046979   53.818858    53.63452    53.48622   53.367368   53.272438   53.196839   53.136793   53.089213   53.051593   53.021906   52.998523   52.980137   52.965704   52.954391   52.945539   52.938621   52.933223   52.929017   52.925745   52.923201   52.921228        52.919698 |  93.471397     91.4498   89.385704    87.28489   85.154327   83.002292   80.838474   78.674036   76.521623   74.395275   72.310239   70.282644   68.329018   66.465663   64.707895   63.069216   61.560502   60.189313   58.959434   57.870719   56.919287   56.098017   55.397268    54.80571   54.311151   53.901257   53.564131   53.288713   53.065022   52.884264   52.738843     52.6223   52.529213   52.455082   52.396202   52.349546   52.312656   52.283546   52.260617   52.242588   52.228435   52.217343   52.208662   52.201879   52.196586   52.192461   52.189252   52.186758   52.184823     52.183323 | 1.645901  1.621317  1.596003  1.570007  1.543396  1.516253  1.488682   1.46081  1.432788  1.404793  1.377023    1.3497  1.323061  1.297352  1.272816  1.249681  1.228145  1.208368  1.190456  1.174456   1.16036  1.148103  1.137578  1.128644  1.121139  1.114894   1.10974  1.105518  1.102081  1.099298  1.097056  1.095257  1.093819  1.092672  1.091761  1.091039  1.090468  1.090017  1.089662  1.089382  1.089163  1.088991  1.088856  1.088751  1.088669  1.088605  1.088555  1.088517  1.088487  1.088463 | 1.340896  1.341258  1.341655  1.342088  1.342562  1.343078   1.34364  1.344251  1.344911  1.345622  1.346381  1.347187  1.348034  1.348913  1.349814  1.350724  1.351628  1.352511  1.353356   1.35415  1.354883  1.355546  1.356135  1.356651  1.357095  1.357473   1.35779  1.358054  1.358271  1.358449  1.358593   1.35871  1.358803  1.358878  1.358938  1.358985  1.359023  1.359053  1.359076  1.359095  1.359109   1.35912  1.359129  1.359136  1.359142  1.359146  1.359149  1.359152  1.359154          1.359155 | 0.721039  0.715691      0.71  0.703952  0.697536  0.690744  0.683577   0.67604  0.668149  0.659933  0.651436  0.642716  0.633851  0.624933  0.616071  0.607382  0.598988  0.591006  0.583539  0.576671  0.570458  0.564928  0.560082  0.555895  0.552325  0.549318  0.546809  0.544737  0.543038  0.541654  0.540534  0.539632  0.538909  0.538331  0.537871  0.537505  0.537216  0.536987  0.536807  0.536665  0.536554  0.536466  0.536398  0.536344  0.536303   0.53627  0.536245  0.536225   0.53621      0.536198 | -0.400067  -0.392921  -0.385386  -0.377459  -0.369143  -0.360447  -0.351392  -0.342011  -0.332351  -0.322477   -0.31247  -0.302427  -0.292462    -0.2827  -0.273267  -0.264288  -0.255876  -0.248122   -0.24109  -0.234815    -0.2293  -0.224524  -0.220442  -0.216996  -0.214117  -0.211735   -0.20978  -0.208187  -0.206896  -0.205857  -0.205024  -0.204358  -0.203829  -0.203409  -0.203076  -0.202814  -0.202607  -0.202444  -0.202317  -0.202217  -0.202139  -0.202078  -0.202031  -0.201994  -0.201965  -0.201943  -0.201926  -0.201912  -0.201902     -0.201894 | 0.685975  0.684651  0.683303  0.681936  0.680559  0.679181  0.677813  0.676469  0.675162  0.673908  0.672722  0.671619  0.670611  0.669708  0.668917  0.668239  0.667671  0.667206  0.666834  0.666542  0.666318   0.66615  0.666024  0.665932  0.665865  0.665817  0.665783  0.665759  0.665742   0.66573  0.665722  0.665717  0.665713  0.665711  0.665709  0.665708  0.665708  0.665708  0.665708  0.665708  0.665708  0.665708  0.665708  0.665708  0.665708  0.665709  0.665709  0.665709  0.665709      0.665709 |  537.77215 | 522.402417  491.865982  461.841985  432.463717   403.86612  376.183942  349.549433  324.089615  299.923153  277.156923  255.882419  236.172196  218.076622  201.621225  186.804979  173.599801  161.951477  151.782069  142.993686  135.473309  129.098207   123.74142  119.276812    115.5833  112.548036  110.068497  108.053554   106.42371   105.11071   104.05673  103.213335  102.540337  102.004656  101.579239  101.242089  100.975405  100.764835  100.598854  100.468231  100.365593  100.285065  100.221976  100.172621  100.134063  100.103983  100.080548  100.062316  100.048149  100.037156          100.028637 | 100.024851 | 1.104717  1.087923   1.07062  1.052842  1.034632  1.016045  0.997152  0.978039  0.958807  0.939576  0.920483  0.901679  0.883328  0.865599  0.848661  0.832674  0.817777  0.804083  0.791669  0.780571  0.770785  0.762271  0.754954   0.74874  0.743518  0.739171  0.735582  0.732641  0.730246  0.728307  0.726745  0.725491  0.724488  0.723689  0.723054   0.72255  0.722152  0.721838   0.72159  0.721395  0.721242  0.721122  0.721028  0.720955  0.720898  0.720853  0.720818  0.720791   0.72077       0.720754 | 5.064747  4.946676  4.825928  4.702814  4.577718  4.451098    4.3235  4.195556  4.067992  3.941624  3.817345  3.696116   3.57893  3.466786  3.360637   3.26134  3.169609  3.085965  3.010705  2.943887  2.885335  2.834669  2.791344  2.754699  2.724013  2.698544  2.677572  2.660421  2.646481  2.635208  2.626135  2.618859  2.613047  2.608416  2.604738  2.601822  2.599517  2.597697  2.596264  2.595137  2.594252  2.593559  2.593016  2.592592  2.592261  2.592003  2.591802  2.591646  2.591525        2.591432 |  537.77215  507.008032   476.69011  446.950464  417.923662  389.745127  362.549074  336.465985  311.619664  288.123925  266.079034  245.568058  226.653384  209.373664  193.741533  179.742419  167.334709  156.451417  147.003339  138.883475  131.972304  126.143397  121.268818  117.223863  113.890799  111.161493  108.938949  107.137894  105.684644  104.516443  103.580487  102.832804  102.237084  101.763559  101.387972  101.090654  100.855727   100.67042  100.524491  100.409753  100.319676  100.249065  100.193792  100.150587  100.116861  100.090572  100.070107  100.054197  100.041845  100.032267  100.024851 |  100.0 |  437.77215 | 100.024851 |
    | 18:30 | 8.152996  7.959567  7.761059  7.557902  7.350623  7.139858  6.926368  6.711048  6.494936  6.279216  6.065219  5.854405  5.648342  5.448669  5.257035  5.075038  4.904144  4.745601  4.600367  4.469046  4.351857   4.24863  4.158843  4.081675   4.01609  3.960917  3.914929   3.87691  3.845705  3.820254  3.799609  3.782941  3.769541  3.758805  3.750231  3.743402  3.737977  3.733676  3.730274  3.727587   3.72547  3.723803  3.722494  3.721466  3.720662  3.720032   3.71954  3.719156  3.718857             3.718624 | 454.651678  436.167045  417.586072  398.977945  380.417613   361.98574  343.768413  325.856555  308.344976  291.331005  274.912663  259.186338  244.243991  230.169974  217.037585   204.90563  193.815289  183.787665   174.82239  166.897562  159.971146  153.983758  148.862513   144.52549  140.886269  137.858093  135.357294  133.305812  131.632818  130.275508  129.179265  128.297349  127.590291  127.025134  126.574607   126.21632  125.932004  125.706833  125.528829  125.388353  125.277672  125.190602  125.122206  125.068557  125.026535  124.993664  124.967987  124.947955  124.932349  124.920206 |  98.144573   96.171976   94.147583    92.07578   89.961937   87.812552   85.635374    83.43953   81.235607    79.03569   76.853336   74.703447   72.602016   70.565735   68.611444   66.755437    65.01265   63.395825   61.914723   60.575507   59.380407   58.327699   57.412042   56.625083   55.956247    55.39359   54.924601   54.536881   54.218651   53.959098   53.748558   53.578584   53.441922   53.332437   53.244998   53.175359   53.120033   53.076176    53.04148   53.014082   52.992486   52.975491   52.962137    52.95166   52.943452    52.93703   52.932014     52.9281    52.92505        52.922677 |  96.529962   94.595671    92.61059    90.57902   88.506227   86.398581   84.263683   82.110481   79.949356    77.79216   75.652186   73.544046   71.483423   69.486686   67.570346   65.750381   64.041438   62.456011    61.00367   59.690462   58.518569   57.486304   56.588429   55.816752   55.160904   54.609174   54.149292   53.769101   53.457052   53.202539   52.996088   52.829414   52.695406   52.588047   52.502306    52.43402   52.379768   52.336762    52.30274   52.275875   52.254698   52.238032   52.224938   52.214664   52.206615   52.200319     52.1954   52.191562   52.188571     52.186244 | 1.682715  1.659485  1.635458  1.610665  1.585149  1.558968  1.532196  1.504924  1.477269  1.449367   1.42138  1.393494   1.36592  1.338885   1.31263  1.287402  1.263438  1.240956  1.220138  1.201123  1.183993  1.168774  1.155432  1.143886  1.134014  1.125665  1.118676  1.112875    1.1081  1.104195   1.10102  1.098453  1.096386  1.094728  1.093403  1.092347  1.091507  1.090841  1.090314  1.089898   1.08957  1.089311  1.089109  1.088949  1.088824  1.088727  1.088651  1.088591  1.088545  1.088509 | 1.340391  1.340705  1.341047  1.341422  1.341832  1.342281   1.34277  1.343304  1.343885  1.344515  1.345194  1.345924  1.346701  1.347523  1.348383  1.349271  1.350176  1.351084  1.351979  1.352847  1.353672  1.354442  1.355147   1.35578  1.356339  1.356826  1.357243  1.357597  1.357892  1.358137  1.358339  1.358503  1.358636  1.358744   1.35883  1.358899  1.358954  1.358998  1.359033   1.35906  1.359082  1.359099  1.359113  1.359123  1.359131  1.359138  1.359143  1.359147   1.35915          1.359152 | 0.728736  0.723922  0.718788  0.713319    0.7075  0.701318  0.694764  0.687833  0.680527  0.672856  0.664842  0.656519  0.647937  0.639162  0.630277  0.621382  0.612589  0.604018  0.595789  0.588014  0.580788  0.574185  0.568249  0.562997  0.558419  0.554484  0.551143  0.548338  0.546006  0.544084  0.542511  0.541232  0.540198  0.539366  0.538699  0.538166  0.537742  0.537405  0.537138  0.536927   0.53676  0.536629  0.536526  0.536445  0.536382  0.536332  0.536293  0.536263  0.536239      0.536221 | -0.410609  -0.404076   -0.39717  -0.389878  -0.382195   -0.37412  -0.365657  -0.356821  -0.347639  -0.338148  -0.328403  -0.318476  -0.308453   -0.29844  -0.288555  -0.278923  -0.269672  -0.260921  -0.252774  -0.245312  -0.238588  -0.232622  -0.227408  -0.222915  -0.219092  -0.215876  -0.213197  -0.210987  -0.209177  -0.207704  -0.206512  -0.205553  -0.204785  -0.204171  -0.203683  -0.203295  -0.202988  -0.202745  -0.202554  -0.202404  -0.202286  -0.202193  -0.202121  -0.202064   -0.20202  -0.201986  -0.201959  -0.201938  -0.201922      -0.20191 | 0.688045   0.68677  0.685464  0.684129  0.682771  0.681398  0.680018  0.678642  0.677281   0.67595  0.674663  0.673434   0.67228  0.671214  0.670248   0.66939  0.668645  0.668011  0.667486  0.667059   0.66672  0.666457  0.666255  0.666104  0.665992   0.66591  0.665851  0.665808  0.665777  0.665756   0.66574   0.66573  0.665722  0.665717  0.665714  0.665711   0.66571  0.665709  0.665708  0.665708  0.665708  0.665708  0.665708  0.665708  0.665708  0.665708  0.665709  0.665709  0.665709      0.665709 | 586.262607 | 570.765738  539.874884  509.261995   479.05851  449.399391  420.421899   392.26396  365.062114  338.949012  314.050518  290.482445  268.347062  247.729517  228.694416  211.282824  195.510014  181.364263  168.806964   157.77419  148.179706  139.919225  132.875526  126.923951  121.937742  117.792777  114.371376   111.56502  109.276005  107.418151  105.916756  104.708022  103.738141  102.962197  102.343032  101.850124  101.458559  101.148104  100.902402  100.708274   100.55514  100.434527  100.339668   100.26517  100.206745  100.160987  100.125198  100.097243  100.075436  100.058448          100.045229 | 100.039343 | 1.129852  1.113994  1.097584  1.080643  1.063199  1.045289  1.026963  1.008284  0.989327  0.970187  0.950972   0.93181  0.912844   0.89423  0.876136  0.858732  0.842182   0.82664  0.812234  0.799063  0.787188  0.776627  0.767363   0.75934  0.752476  0.746668  0.741804  0.737765  0.734439  0.731719  0.729507  0.727718  0.726278  0.725122  0.724198  0.723462  0.722877  0.722412  0.722045  0.721755  0.721526  0.721346  0.721204  0.721093  0.721006  0.720938  0.720885  0.720843  0.720811       0.720786 | 5.243045  5.130331  5.014494  4.895761  4.774415  4.650805  4.525348  4.398545  4.270981  4.143333  4.016365  3.890932  3.767956  3.648418  3.533316  3.423635  3.320293  3.224093  3.135674  3.055466   2.98367  2.920248  2.864938   2.81729   2.77671   2.74251  2.713959  2.690325  2.670906  2.655052  2.642183  2.631786  2.623423  2.616721  2.611366    2.6071  2.603711  2.601023  2.598897  2.597218  2.595894  2.594852  2.594033  2.593391  2.592888  2.592494  2.592187  2.591947   2.59176        2.591614 | 586.262607  555.260904  524.472222   494.02599  464.055683  434.697801  406.090449  378.371503  351.676333  326.135092  301.869617  278.990007  257.591039  237.748581  219.516302  202.922942  187.970491  174.633574  162.860265  152.574402  143.679314  136.062645  129.601819  124.169617  119.639327  115.889091  112.805176  110.284144  108.233975  106.574336   105.23621  104.161099  103.299971  102.612124   102.06403  101.628257  101.282483  101.008629   100.79211  100.621202  100.486506  100.380507  100.297212   100.23185  100.180631  100.140548  100.109222  100.084772  100.065714  100.050878  100.039343 |  100.0 | 486.262607 | 100.039343 |
    | 19:00 | 8.439345  8.256063  8.067054  7.872631  7.673182  7.469187  7.261224   7.04999  6.836308  6.621137  6.405584  6.190899  5.978472  5.769814   5.56653  5.370268  5.182667  5.005278  4.839486  4.686428  4.546922  4.421419   4.30998   4.21229  4.127703  4.055307  3.994007  3.942607  3.899885  3.864648  3.835783  3.812275  3.793229  3.777865  3.765519  3.755632  3.747736  3.741447  3.736449  3.732486  3.729349  3.726871  3.724916  3.723376  3.722165  3.721214  3.720468  3.719884  3.719427              3.71907 | 482.702924  464.653845  446.392611  427.981025  409.486357  390.981546  372.545245  354.261658  336.220115  318.514329  301.241269  284.499607   268.38768  253.000981  238.429199  224.752928  212.040217  200.343242  189.695437  180.109462  171.576344  164.066026  157.529348  151.901309  147.105205  143.057162  139.670552  136.859868  134.543774  132.647257  131.102886  129.851342  128.841361  128.029303  127.378472  126.858327  126.443665  126.113833   125.85201  125.644559   125.48047  125.350891  125.248719  125.168275  125.105029  125.055372  125.016435  124.985946  124.962101  124.943477 | 101.064773   99.195655   97.268132   95.285395   93.251411   91.171057   89.050247   86.896075   84.716931   82.522613   80.324393   78.135026   75.968686   73.840789   71.767685   69.766205   67.853044   66.044025   64.353271   62.792376   61.369691   60.089805   58.953346   57.957102   57.094476   56.356176   55.731037   55.206862   54.771176   54.411834   54.117462   53.877732   53.683496   53.526816   53.400914   53.300079   53.219558   53.155423   53.104457    53.06404    53.03205   53.006774   52.986835   52.971132   52.958783   52.949084   52.941479   52.935522   52.930864        52.927225 |  99.393453   97.560632    95.67054   93.726306   91.731822   89.691866   87.612242   85.499902   83.363075    81.21137   79.055837   76.908986   74.784715    72.69814   70.665295   68.702682   66.826674   65.052785   63.394864   61.864281   60.469223   59.214192   58.099802   57.122904    56.27703   55.553067   54.940068   54.426072   53.998847   53.646483   53.357827   53.122753   52.932289   52.778652   52.655194   52.556318    52.47736    52.41447   52.364494   52.324862   52.293493   52.268708   52.249157   52.233758   52.221649   52.212139   52.204681    52.19884   52.194272     52.190704 | 1.716787  1.695022  1.672416  1.648986  1.624759  1.599772  1.574074   1.54773  1.520823  1.493454  1.465749  1.437855  1.409944  1.382214  1.354881  1.328181  1.302357  1.277653  1.254299    1.2325  1.212422   1.19418  1.177834  1.163386  1.150782  1.139924  1.130677  1.122886  1.116383     1.111  1.106578  1.102968  1.100038   1.09767  1.095765  1.094238  1.093017  1.092044  1.091271  1.090657  1.090171  1.089787  1.089484  1.089245  1.089058   1.08891  1.088794  1.088704  1.088633  1.088578 | 1.339961  1.340232  1.340528  1.340852  1.341206  1.341594  1.342019  1.342482  1.342989   1.34354  1.344139  1.344788  1.345487  1.346235   1.34703  1.347866  1.348737  1.349631  1.350538  1.351441  1.352326  1.353177   1.35398  1.354723  1.355399  1.356002  1.356532  1.356991  1.357382  1.357712  1.357987  1.358215  1.358401  1.358553  1.358677  1.358776  1.358855  1.358919   1.35897   1.35901  1.359042  1.359068  1.359088  1.359104  1.359116  1.359126  1.359133  1.359139  1.359144          1.359148 | 0.735545  0.731229  0.726619  0.721698   0.71645  0.710859  0.704912  0.698597  0.691906  0.684838  0.677397    0.6696  0.661471  0.653052  0.644398  0.635584  0.626698  0.617846  0.609143   0.60071  0.592664  0.585111  0.578137  0.571805  0.566147   0.56117  0.556854  0.553161  0.550038  0.547425  0.545259  0.543477  0.542022  0.540841  0.539887   0.53912  0.538505  0.538013  0.537622  0.537312  0.537065  0.536871  0.536717  0.536596    0.5365  0.536425  0.536366   0.53632  0.536284      0.536256 | -0.420119  -0.414178  -0.407881  -0.401217  -0.394171  -0.386737  -0.378909  -0.370689  -0.362086   -0.35312   -0.34382  -0.334232  -0.324418  -0.314455  -0.304438  -0.294476  -0.284693  -0.275213  -0.266163  -0.257655  -0.249785  -0.242622  -0.236205  -0.230546  -0.225625  -0.221405  -0.217829  -0.214831  -0.212342  -0.210292  -0.208617  -0.207256  -0.206156  -0.205272  -0.204563  -0.203997  -0.203546  -0.203189  -0.202906  -0.202682  -0.202505  -0.202366  -0.202257  -0.202171  -0.202104  -0.202052   -0.20201  -0.201978  -0.201954     -0.201934 | 0.690015  0.688801  0.687548  0.686261  0.684942  0.683597  0.682233  0.680855  0.679476  0.678104  0.676754  0.675439  0.674174  0.672974  0.671854  0.670827  0.669904  0.669091  0.668391  0.667802  0.667316  0.666926  0.666618   0.66638  0.666199  0.666064  0.665964  0.665891  0.665838    0.6658  0.665772  0.665753  0.665739  0.665729  0.665722  0.665717  0.665714  0.665712   0.66571  0.665709  0.665709  0.665709  0.665708  0.665708  0.665708  0.665708  0.665709  0.665709  0.665709      0.665709 | 633.846018 | 618.449354   587.65833   556.91044  526.333106  496.058405  466.222368  436.963933   408.42353  380.741244  354.054558  328.495681  304.188479  281.245124  259.762549  239.818943  221.470492  204.748696  189.658539  176.177839  164.257969  153.826042  144.788476  137.035644  130.447204   124.89757  120.261038  116.416157  113.249088  110.655882  108.543728  106.831334  105.448651  104.336147  103.443808  102.730026  102.160464  101.706975  101.346622    101.0608  100.834476  100.655551  100.514311  100.402981  100.315348  100.246463  100.192387  100.149991  100.116795  100.090836          100.070562 | 100.061517 | 1.153099  1.138251  1.122822  1.106824  1.090275  1.073197  1.055624  1.037598  1.019175  1.000423  0.981426  0.962285  0.943116  0.924053  0.905246  0.886856  0.869051  0.852001  0.835866  0.820791  0.806891  0.794251  0.782915  0.772887  0.764132  0.756585  0.750155  0.744734  0.740207   0.73646   0.73338  0.730865  0.728823  0.727173  0.725845   0.72478  0.723929  0.723251  0.722712  0.722284  0.721945  0.721677  0.721466  0.721299  0.721169  0.721066  0.720985  0.720922  0.720873       0.720834 | 5.409633  5.303042  5.192984  5.079622  4.963159  4.843851  4.722013  4.598025  4.472345  4.345511  4.218148  4.090975  3.964795  3.840495  3.719024  3.601374  3.488543  3.381493    3.2811  3.188104  3.103062  3.026313  2.957962  2.897878  2.845723  2.800984  2.763028  2.731148  2.704611  2.682697  2.664726  2.650079  2.638204  2.628619  2.620913  2.614739  2.609808  2.605879  2.602756   2.60028  2.598319   2.59677  2.595547  2.594585  2.593828  2.593233  2.592767  2.592402  2.592116        2.591893 | 633.846018  603.060736  572.255878  541.556394   511.09217   480.99749   451.41017  422.470342  394.318834  367.095134  340.934918  315.967162  292.310913  270.071787  249.338384   230.17882  212.637666  196.733588  182.458027  169.775178  158.623442  148.918354  140.556806  133.422181   127.38992  122.332956  118.126565   114.65228  111.800703  109.473235  107.582826  106.053972  104.822145  103.832891   103.04073  102.408015  101.903807  101.502832  101.184557  100.932363  100.732857  100.575274  100.450987    100.3531  100.276111   100.21564  100.168205  100.131043  100.101968  100.079249  100.061517 |  100.0 | 533.846018 | 100.061517 |
    | 19:30 | 8.703587  8.531614  8.353409  8.169195  7.979253  7.783936  7.583673  7.378993   7.17053  6.959035  6.745396  6.530637  6.315928  6.102583  5.892046  5.685873  5.485693  5.293158  5.109878  4.937346  4.776852  4.629412    4.4957   4.37601  4.270253  4.177976  4.098424  4.030607  3.973388  3.925558  3.885907  3.853276  3.826591   3.80489  3.787326  3.773169  3.761798  3.752694  3.745424  3.739634  3.735031  3.731379  3.728488  3.726202  3.724398  3.722975  3.721856  3.720976  3.720286             3.719744 | 509.315914  491.916425  474.198355  456.216656  438.031232  419.707295  401.315609  382.932623  364.640409   346.52639  328.682772  311.205642  294.193668   277.74634  261.961749  246.933887  232.749552  219.484984   207.20245  195.947087  185.744353   176.59846  168.492081  161.387488   155.22909  149.947073  145.461752  141.688087    138.5399  135.933425   133.78999  132.037797   130.61286  129.459291  128.529078  127.781536   127.18259  126.703964   126.32238  126.018801   125.77774  125.586656  125.435433   125.31594  125.221655  125.147365  125.088906  125.042967  125.006911  124.978648 | 103.759517  102.005735  100.188392   98.309769   96.372738   94.380878   92.338597   90.251263    88.12534   85.968513   83.789811    81.59969   79.410082   77.234379   75.087319   72.984759   70.943314   68.979833   67.110737   65.351242   63.714523   62.210925   60.847322   59.626724   58.548205   57.607165   56.795887   56.104291   55.520769   55.032999   54.628635   54.295855   54.023725   53.802417   53.623298   53.478921   53.362963   53.270118   53.195983   53.136929    53.08999   53.052753   53.023266   52.999954   52.981553   52.967049   52.955634   52.946661   52.939618        52.934097 | 102.035868  100.316143   98.534091    96.69195   94.792535   92.839355   90.836734   88.789934   86.705295   84.590353   82.453959   80.306369   78.159282    76.02583   73.920464   71.858735   69.856933   67.931581   66.098782   64.373455    62.76852   61.294121   59.956998   58.760104   57.702528   56.779763   55.984239   55.306073   54.733883   54.255585   53.859074   53.532756    53.26591     53.0489   52.873259   52.731686    52.61798   52.526939   52.454243   52.396336   52.350308   52.313795    52.28488   52.262021   52.243977   52.229755   52.218561   52.209763   52.202856     52.197442 | 1.747905  1.727687  1.706601  1.684653  1.661857  1.638237  1.613822  1.588655  1.562793  1.536309  1.509293  1.481858  1.454137  1.426289  1.398497  1.370966  1.343922  1.317603  1.292254  1.268115  1.245407  1.224319  1.204998  1.187537  1.171972  1.158284  1.146399  1.136204  1.127556  1.120294   1.11425  1.109259  1.105168  1.101833  1.099129  1.096946   1.09519  1.093784  1.092659  1.091763  1.091051  1.090485  1.090038  1.089683  1.089404  1.089183   1.08901  1.088873  1.088766  1.088682 | 1.339597   1.33983  1.340086  1.340366  1.340672  1.341007  1.341373  1.341775  1.342213  1.342693  1.343216  1.343786  1.344404  1.345071  1.345789  1.346555  1.347365  1.348215  1.349095  1.349994  1.350899  1.351795  1.352666  1.353498  1.354277  1.354994   1.35564  1.356214  1.356715  1.357146  1.357512   1.35782  1.358076  1.358287   1.35846    1.3586  1.358714  1.358805  1.358879  1.358938  1.358984  1.359022  1.359051  1.359075  1.359093  1.359108  1.359119  1.359128  1.359135          1.359141 | 0.741516  0.737663   0.73354  0.729131   0.72442   0.71939  0.714026  0.708312  0.702235  0.695787  0.688961  0.681759   0.67419  0.666272  0.658038  0.649535  0.640825  0.631988  0.623121  0.614332   0.60574  0.597465  0.589619    0.5823  0.575587  0.569529  0.564149  0.559442  0.555381   0.55192  0.549005  0.546574  0.544564  0.542915  0.541569  0.540479  0.539598  0.538891  0.538324  0.537872  0.537511  0.537225  0.536998  0.536818  0.536676  0.536564  0.536476  0.536407  0.536352      0.536309 | -0.428627  -0.423248  -0.417536  -0.411476  -0.405055  -0.398258  -0.391075    -0.3835   -0.37553   -0.36717  -0.358434  -0.349344  -0.339939  -0.330268    -0.3204   -0.31042  -0.300428  -0.290541   -0.28088  -0.271574  -0.262742  -0.254492  -0.246909  -0.240049   -0.23394  -0.228582  -0.223947  -0.219989  -0.216649  -0.213857  -0.211546  -0.209647  -0.208098  -0.206841  -0.205826   -0.20501  -0.204357  -0.203835   -0.20342   -0.20309  -0.202829  -0.202622  -0.202459  -0.202331   -0.20223   -0.20215  -0.202088  -0.202039  -0.202001     -0.201971 | 0.691868   0.69072  0.689532  0.688304  0.687039   0.68574  0.684412  0.683059  0.681689   0.68031  0.678933  0.677569  0.676232  0.674936  0.673696  0.672527  0.671444  0.670458  0.669579  0.668811  0.668156  0.667609  0.667162  0.666805  0.666526  0.666311  0.666149  0.666028  0.665939  0.665874  0.665826  0.665792  0.665767   0.66575  0.665737  0.665728  0.665722  0.665717  0.665714  0.665712  0.665711   0.66571  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709      0.665709 | 679.625318 | 664.555012  634.317386   603.89196  573.400675    542.9707  512.734128  482.827412  453.390487  424.565545  396.495419  369.321557  343.181577  318.206421  294.517159  272.221552  251.410512  232.154695  214.501462  198.472537  184.062662  171.239499  159.944968  150.098021  141.598685  134.333038  128.178639  123.009901  118.702933   115.13951  112.209993  109.815171  107.867144  106.289423  105.016452  103.992763   103.17193  102.515441  101.991588   101.57443  101.242856  100.979758  100.771329  100.606456  100.476224  100.373497  100.292572  100.228905  100.178879   100.13962           100.10885 |   100.0951 | 1.174319  1.160534  1.146151  1.131174  1.115614  1.099482    1.0828  1.065596  1.047906   1.02978  1.011277  0.992474   0.97346  0.954344  0.935249  0.916316    0.8977  0.879565  0.862081  0.845414  0.829719  0.815129  0.801748  0.789645  0.778847  0.769343  0.761086  0.753999  0.747984   0.74293  0.738722  0.735247  0.732397  0.730074  0.728189  0.726668  0.725444  0.724464   0.72368  0.723055  0.722559  0.722164  0.721852  0.721605   0.72141  0.721256  0.721135   0.72104  0.720965       0.720907 | 5.563093  5.463247  5.359671  5.252477   5.14181  5.027852  4.910833  4.791037  4.668808  4.544561  4.418792  4.292076   4.16508  4.038559  3.913355  3.790383  3.670612  3.555042   3.44466  3.340398  3.243082   3.15338  3.071766  2.998486  2.933547  2.876736   2.82764  2.785699  2.750245  2.720562  2.695921  2.675619  2.659001  2.645476  2.634522  2.625688   2.61859  2.612905  2.608364  2.604746   2.60187  2.599588  2.597781  2.596352  2.595224  2.594335  2.593635  2.593085  2.592653        2.592314 | 679.625318  649.507798  619.142613  588.649035  558.151652    527.7802  497.669136  467.956936  438.785049  410.296493  382.634041  355.937985  330.343483   305.97752  282.955558   261.37799  241.326599  222.861243  206.017072  190.802589  177.198852  165.160069  154.615665  145.473774  137.625879  130.952166  125.327086  120.624586  116.722591  113.506479  110.871442  108.723801  106.981433  105.573524   104.43985  103.529788    102.8012   102.21929  101.755522  101.386625   101.09371  100.861509  100.677721  100.532462  100.417816  100.327454  100.256324  100.200404  100.156498  100.122067    100.0951 |  100.0 | 579.625318 |   100.0951 |
    | 20:00 | 8.943624  8.783908  8.617563  8.444748  8.265663  8.080562  7.889758  7.693642   7.49269  7.287477   7.07869  6.867142  6.653779   6.43969  6.226109  6.014405  5.806074  5.602706  5.405946  5.217437  5.038748  4.871299  4.716276  4.574562  4.446683   4.33278  4.232615  4.145614  4.070923  4.007492  3.954152  3.909694  3.872928  3.842732  3.818079  3.798056  3.781866  3.768825  3.758355  3.749975  3.743283  3.737952  3.733714  3.730351  3.727686  3.725579  3.723914  3.722601  3.721567             3.720753 | 534.096426  517.543773  500.575387  483.240058   465.59091  447.685822  429.587807  411.365304  393.092374  374.848733  356.719603  338.795306  321.170562  303.943418  287.213765  271.081389  255.643554  240.992147  227.210459  214.369795  202.526143  191.717248  181.960431  173.251534  165.565198  158.856584  153.064384  148.114787  143.925935  140.412345  137.488877  135.073932  133.091764  131.473903   130.15982  129.096989  128.240526  127.552564  127.001494  126.561164  126.210089  125.930727  125.708827  125.532859  125.393529  125.283368   125.19639  125.127807  125.073797  125.031317 | 106.207428  104.578632  102.882242  101.119869   99.293558   97.405885   95.460063   93.460066   91.410746   89.317972   87.188756    85.03138   82.855498   80.672214   78.494098   76.335134   74.210565   72.136612   70.130049   68.207625   66.385351   64.677696   63.096766   61.651565   60.347448   59.185855   58.164375   57.277133   56.515435   55.868561   55.324598   54.871209   54.496268   54.188326   53.936918   53.732724   53.567614    53.43462   53.327853   53.242387   53.174147   53.119785   53.076564   53.042265   53.015092   52.993598   52.976621    52.96323   52.952683        52.944386 | 104.436241  102.839076  101.175628   99.447479   97.656634   95.805618   93.897583   91.936424   89.926901   87.874767   85.786899   83.671418    81.53779   79.396904   77.261085   75.144047   73.060736   71.027058   69.059461   67.174369   65.387483   63.712989   62.162759   60.745623   59.466831   58.327795   57.326152    56.45614   55.709233   55.074921   54.541521   54.096938   53.729278   53.427316    53.18079    52.98056   52.818657   52.688246   52.583552   52.499746   52.432831   52.379524   52.337143    52.30351   52.276865   52.255788   52.239141    52.22601   52.215668     52.207531 | 1.775914  1.757304  1.737807  1.717426  1.696166  1.674037  1.651057  1.627254  1.602662  1.577331  1.551324  1.524722  1.497624  1.470152  1.442451  1.414688  1.387056  1.359769  1.333058  1.307166  1.282336  1.258801   1.23677  1.216416  1.197866   1.18119  1.166401  1.153458  1.142272  1.132716  1.124639  1.117878  1.112267  1.107644   1.10386  1.100781  1.098287  1.096275  1.094659  1.093363  1.092328  1.091503  1.090847  1.090326  1.089913  1.089587  1.089329  1.089125  1.088965  1.088839 |  1.33929  1.339491  1.339712  1.339953  1.340217  1.340506  1.340823  1.341169  1.341548  1.341963  1.342417  1.342913  1.343454  1.344041  1.344678  1.345364    1.3461  1.346883  1.347709   1.34857  1.349457  1.350358  1.351259  1.352145  1.353001  1.353811  1.354565  1.355253   1.35587  1.356414  1.356887  1.357292  1.357634  1.357921  1.358158  1.358354  1.358514  1.358644  1.358748  1.358833  1.358901  1.358955  1.358998  1.359032  1.359059  1.359081  1.359098  1.359112  1.359122           1.35913 | 0.746698  0.743274  0.739604   0.73567  0.731459  0.726954  0.722139  0.716998  0.711515  0.705677  0.699472  0.692891  0.685932    0.6786  0.670906  0.662876  0.654547  0.645973  0.637222  0.628381  0.619552  0.610847  0.602386  0.594287  0.586656  0.579585   0.57314  0.567359  0.562254  0.557811  0.553996  0.550759  0.548042  0.545782  0.543919  0.542392  0.541149  0.540143  0.539331  0.538679  0.538157   0.53774  0.537408  0.537144  0.536935  0.536769  0.536638  0.536535  0.536453      0.536389 |  -0.43617  -0.431322  -0.426164  -0.420681  -0.414858  -0.408678   -0.40213  -0.395201  -0.387881  -0.380167  -0.372059  -0.363565  -0.354702  -0.345499  -0.335999  -0.326259  -0.316354  -0.306376  -0.296432  -0.286639  -0.277124  -0.268011  -0.259415  -0.251436  -0.244147  -0.237593  -0.231791  -0.226728  -0.222369  -0.218663  -0.215547  -0.212951  -0.210807  -0.209049  -0.207616  -0.206456   -0.20552  -0.204768  -0.204166  -0.203685  -0.203302  -0.202998  -0.202757  -0.202567  -0.202416  -0.202297  -0.202204   -0.20213  -0.202073     -0.202028 |  0.69359  0.692515  0.691396  0.690236  0.689035  0.687795  0.686518  0.685209  0.683872  0.682513   0.68114  0.679763  0.678391  0.677037  0.675716  0.674442   0.67323  0.672095   0.67105  0.670107  0.669272   0.66855  0.667938  0.667432  0.667022  0.666696  0.666444   0.66625  0.666105  0.665996  0.665917  0.665858  0.665816  0.665785  0.665763  0.665747  0.665735  0.665727  0.665721  0.665717  0.665714  0.665712  0.665711   0.66571   0.66571  0.665709  0.665709  0.665709  0.665709      0.665709 | 722.742733 | 708.218133  678.975348  649.321351  619.371669  589.247215  559.074282   528.98434  499.113599  469.602278  440.593564  412.232203  384.662695  358.027083  332.462327  308.097307  285.049516  263.421573   243.29772  224.740554  207.788261   192.45266  178.718363  166.543256  155.860439  146.581532  138.601145   131.80207  126.060736  121.252399  117.255646  113.955949  111.248142   109.03786  107.242092  105.789032  104.617444  103.675727  102.920835  102.317154  101.835425  101.451752  101.146711  100.904577  100.712668  100.560781  100.440733  100.345973   100.27127  100.212451          100.166194 | 100.145491 | 1.193411  1.180727  1.167435  1.153535  1.139031  1.123928  1.108239  1.091979  1.075172  1.057851  1.040058  1.021845  1.003281  0.984447   0.96544  0.946375  0.927383   0.90861  0.890216  0.872368  0.855235  0.838978  0.823745  0.809657  0.796806  0.785243   0.77498  0.765991  0.758217  0.751573  0.745954  0.741248  0.737341  0.734122  0.731486  0.729341  0.727603  0.726201  0.725074  0.724171  0.723449  0.722874  0.722416  0.722053  0.721765  0.721538  0.721358  0.721216  0.721104       0.721016 | 5.702291  5.609692  5.513161  5.412773  5.308628  5.200854  5.089615  4.975114  4.857607  4.737405  4.614885  4.490497  4.364771  4.238322  4.111856  3.986164  3.862122  3.740669   3.62279  3.509484  3.401719   3.30039  3.206263  3.119933  3.041782  2.971962  2.910392  2.856777  2.810643  2.771384  2.738313  2.710707  2.687848  2.669054  2.653697  2.641214  2.631115  2.622976   2.61644  2.611206  2.607026  2.603696  2.601047  2.598945   2.59728  2.595962  2.594921    2.5941  2.593454        2.592945 | 722.742733  693.730479  664.250376  634.415266  604.343331  574.158186  543.988788  513.969122  484.237632  454.936354  426.209697  398.202844  371.059756  344.920747   319.91966  296.180692  273.814953  252.916911   233.56093  215.798158  199.654072  185.126992  172.187844  160.781358  150.828734  142.231632  134.877144  128.643276  123.404418  119.036292  115.420046  112.445268  110.011919  108.031272  106.426048   105.12997  104.086923   103.24991  102.579925  102.044835  101.618335  101.279006   101.00948   100.79573  100.626459  100.492596  100.386876  100.303488  100.237797   100.18611  100.145491 |  100.0 | 622.742733 | 100.145491 |
    | 20:30 | 9.157834  9.011135  8.857496  8.697029  8.529875  8.356212  8.176267  7.990321  7.798724  7.601905  7.400384  7.194787  6.985854  6.774457  6.561603  6.348442  6.136267  5.926503  5.720688  5.520438  5.327404  5.143206   4.96936  4.807198  4.657792  4.521886  4.399854  4.291685  4.197007  4.115131  4.045121  3.985879  3.936219   3.89494  3.860881   3.83296  3.810199  3.791734  3.776816  3.764808  3.755171  3.747458  3.741301  3.736395  3.732493  3.729396  3.726941  3.724998  3.723463             3.722251 | 556.697087  541.169798  525.138651  508.647015  491.741933  474.474578  456.900685  439.080946  421.081337  402.973367  384.834192  366.746571  348.798601  331.083191  313.697208  296.740236  280.312905  264.514758  249.441653  235.182755  221.817245  209.410935  198.013084  187.653748  178.342033  170.065568  162.791368  156.468089  151.029457  146.398466  142.491854  139.224359  136.512396  134.276887  132.445211  130.952303  129.741064  128.762242  127.973953  127.341004  126.834111  126.429103  126.106165  125.849141  125.644922  125.482911   125.35457  125.253042  125.172829  125.109537 | 108.391945  106.895907  105.329093  103.692642  101.987995  100.216976   98.381887   96.485603   94.531691   92.524523   90.469407   88.372715   86.242013   84.086177   81.915483   79.741663   77.577895   75.438712   73.339796   71.297643   69.329077   67.450616   65.677726   64.023995   62.500348   61.114372    59.86988   58.766773   57.801244   56.966264   56.252305   55.648153   55.141717    54.72075   54.373412   54.088674   53.856558   53.668251    53.51612   53.393657    53.29538   53.216728   53.153931   53.103898   53.064111   53.032526   53.007492   52.987679   52.972021        52.959663 | 106.578336   105.11135  103.574963  101.970291  100.298747    98.56212   96.762667   94.903208    92.98724   91.019049   89.003842   86.947866   84.858541    82.74457   80.616029   78.484423   76.362675   74.265033   72.206877   70.204381   68.274043   66.432061   64.693599   63.071982   61.577923   60.218862   58.998537   57.916852   56.970072   56.151307   55.451213   54.858793   54.362192     53.9494   53.608807   53.329598   53.101989    52.91734   52.768163   52.648078    52.55171   52.474585   52.413007   52.363946   52.324931    52.29396   52.269412   52.249984    52.23463     52.222512 |  1.80071  1.783749  1.765892  1.747136  1.727482  1.706933  1.685498   1.66319  1.640032  1.616053  1.591295  1.565815  1.539682  1.512984  1.485831  1.458354  1.430707  1.403067  1.375638  1.348639  1.322307  1.296886  1.272615   1.24972  1.228395  1.208797  1.191028  1.175137  1.161115  1.148901  1.138391  1.129447  1.121915  1.115628  1.110424  1.106145  1.102649  1.099808  1.097508  1.095655  1.094167  1.092974  1.092021  1.091262  1.090658  1.090178  1.089798  1.089497  1.089259  1.089071 | 1.339034  1.339207  1.339397  1.339605  1.339833  1.340082  1.340355  1.340653  1.340981  1.341339  1.341731  1.342161   1.34263  1.343143  1.343701  1.344307  1.344962  1.345667  1.346421   1.34722  1.348059  1.348929  1.349822  1.350723  1.351618  1.352491  1.353328  1.354116  1.354843  1.355502  1.356089  1.356604  1.357049  1.357428  1.357748  1.358015  1.358235  1.358416  1.358564  1.358684   1.35878  1.358858  1.358921  1.358971   1.35901  1.359042  1.359067  1.359087  1.359103          1.359115 |  0.75114  0.748116  0.744864  0.741371  0.737623  0.733606  0.729303  0.724699  0.719778  0.714523   0.70892  0.702956  0.696621  0.689909   0.68282  0.675361   0.66755  0.659416  0.651003   0.64237  0.633595  0.624768  0.615997  0.607397  0.599088  0.591182  0.583781  0.576967  0.570795  0.565292  0.560459  0.556274  0.552697  0.549673  0.547144  0.545046  0.543319  0.541908   0.54076  0.539832  0.539084  0.538483  0.538002  0.537618  0.537312  0.537069  0.536876  0.536723  0.536602      0.536507 |  -0.44279  -0.438443  -0.433807  -0.428869  -0.423613  -0.418023  -0.412085  -0.405785  -0.399109  -0.392047  -0.384592  -0.376741  -0.368498  -0.359875  -0.350893  -0.341586  -0.332004   -0.32221  -0.312287  -0.302331  -0.292454  -0.282779   -0.27343  -0.264529  -0.256186  -0.248488  -0.241499  -0.235253  -0.229753  -0.224979  -0.220888  -0.217423  -0.214519  -0.212107   -0.21012  -0.208493   -0.20717  -0.206099  -0.205236  -0.204543  -0.203988  -0.203545  -0.203193  -0.202913  -0.202691  -0.202515  -0.202376  -0.202266   -0.20218     -0.202112 | 0.695172  0.694172  0.693128  0.692041   0.69091  0.689738  0.688525  0.687274  0.685987  0.684669  0.683325  0.681962  0.680588  0.679213   0.67785  0.676509  0.675207  0.673957  0.672776  0.671677  0.670673  0.669773  0.668984  0.668306  0.667737  0.667269  0.666894  0.666598   0.66637  0.666196  0.666065  0.665968  0.665897  0.665844  0.665806  0.665779  0.665759  0.665744  0.665734  0.665726  0.665721  0.665717  0.665715  0.665713  0.665711   0.66571   0.66571  0.665709  0.665709      0.665709 | 762.408181 | 748.636706  720.807656  692.354598  663.386083  634.015926  604.363404  574.553318  544.715878  514.986376  485.504598   456.41395  427.860239  399.990087  372.948944  346.878696  321.914872  298.183514  275.797784  254.854454   235.43049  217.579962  201.331596  186.687252  173.621605  162.083224   151.99707  143.268304   135.78708  129.433887  124.084928   119.61706  115.911928  112.859083  110.358034  108.319312  106.664733  105.327035  104.249121  103.383056  102.688972  102.133966   101.69106  101.338248  101.057669  100.834872  100.658212  100.518321  100.407691  100.320309          100.251373 | 100.220473 | 1.210305  1.198749   1.18658  1.173795  1.160394  1.146378  1.131751  1.116524  1.100708  1.084325  1.067401  1.049974  1.032089  1.013806  0.995198  0.976354  0.957377   0.93839   0.91953  0.900948  0.882808  0.865277  0.848522  0.832701   0.81795   0.80438  0.792066  0.781043   0.77131  0.762825  0.755519  0.749299  0.744058  0.739682  0.736058  0.733078  0.730642  0.728662   0.72706  0.725768  0.724731  0.723899  0.723235  0.722706  0.722285   0.72195  0.721685  0.721475  0.721309       0.721178 | 5.826356  5.741407  5.652367  5.559288  5.462236  5.361301  5.256595  5.148262  5.036486  4.921494  4.803566  4.683042  4.560331  4.435916  4.310365  4.184332  4.058558  3.933871  3.811174  3.691428  3.575627   3.46476  3.359772  3.261508   3.17067   3.08777  3.013101  2.946722  2.888464   2.83796  2.794682   2.75799  2.727182  2.701537  2.680352  2.662968  2.648785  2.637271  2.627964  2.620469  2.614452  2.609635  2.605788  2.602722  2.600284  2.598348  2.596814  2.595599  2.594639        2.593882 | 762.408181  734.914669  706.743967  678.002042  648.799994  619.254318  589.487046  559.625729  529.803227  500.157268  470.829726  441.965589  413.711568  386.214316  359.618245  334.062922  309.680096    286.5904  264.899855  244.696334  226.046231  208.991592  193.548041  179.703778  167.419923  156.632313  147.254718  139.183269  132.301667  126.486706   121.61356  117.560407  114.212091  111.462701  109.217079  107.391433  105.913218  104.720524  103.761148  102.991514  102.375552   101.88361  101.491456  101.179382  100.931425  100.734699  100.578835  100.455506  100.358044  100.281118  100.220473 |  100.0 | 662.408181 | 100.220473 |
    | 21:00 | 9.345031  9.211955  9.071689  8.924314   8.76993  8.608662  8.440671  8.266155  8.085367   7.89862  7.706298   7.50887  7.306903   7.10107  6.892169  6.681128  6.469015   6.25704  6.046551  5.839022  5.636027  5.439201  5.250186  5.070568  4.901795  4.745101  4.601433  4.471388   4.35519  4.252684  4.163368  4.086454   4.02094   3.96569  3.919514  3.881228  3.849705  3.823908  3.802907  3.785887  3.772148  3.761094  3.752227  3.745131  3.739466  3.734952  3.731361   3.72851  3.726249             3.724458 | 576.823462  562.479881  547.553048  532.081624  516.107307  499.675278  482.834654   465.63892  448.146339  430.420303  412.529628  394.548731  376.557676  358.642036  340.892515  323.404291  306.276002  289.608333  273.502162  258.056237  243.364414   229.51253  216.575062  204.611806  193.664888  183.756454  174.887398  167.037383  160.166269  154.216869  149.118706  144.792364  141.153894  138.118862  135.605681  133.538084   131.84672  130.469962  129.354089  128.453017  127.727734  127.145576   126.67944  126.307009  126.010018  125.773597  125.585691  125.436565  125.318377   125.22483 | 110.300988  108.943875  107.513439  106.010504  104.436088  102.791475  101.078288   99.298572   97.454892   95.550437   93.589129   91.575754   89.516079    87.41699    85.28661   83.134406   80.971268   78.809536   76.662963   74.546577   72.476424   70.469181   68.541603   66.709848   64.988695   63.390729   61.925589   60.599391   59.414401   58.369038   57.458193   56.673821   56.005703   55.442266   54.971361   54.580917   54.259442   53.996359   53.782189   53.608625   53.468515   53.355788   53.265357   53.192995    53.13522   53.089183   53.052565   53.023487    53.00043        52.982173 | 108.450308  107.119548  105.716891  104.243141    102.6993  101.086624   99.406705    97.66155   95.853674   93.986201   92.062982   90.088704   88.069027   86.010701   83.921691   81.811282    79.69015   77.570398    75.46551   73.390222   71.360271   69.392007   67.501861   65.705678   64.017949   62.451014   61.014326   59.713882   58.551903    57.52684   56.633684   55.864544     55.2094   54.656905   54.195144   53.812282    53.49705   53.239076   53.029065   52.858872   52.721482   52.610945   52.522269   52.451312    52.39466   52.349517    52.31361   52.285096   52.262487     52.244584 | 1.822229  1.806945   1.79076   1.77367   1.75567  1.736761  1.716944  1.696224  1.674613  1.652129  1.628796   1.60465  1.579738  1.554123  1.527882  1.501113  1.473932  1.446481  1.418923  1.391447  1.364262  1.337595  1.311686  1.286779  1.263107  1.240884  1.220292  1.201464  1.184483  1.169374  1.156107  1.144604  1.134746  1.126389  1.119374  1.113535  1.108713  1.104756  1.101528  1.098907  1.096788  1.095082  1.093711  1.092614  1.091738  1.091039  1.090483  1.090041  1.089691  1.089413 | 1.338823  1.338971  1.339135  1.339314   1.33951  1.339724  1.339959  1.340217  1.340499  1.340807  1.341146  1.341517  1.341923  1.342367  1.342852  1.343382  1.343958  1.344582  1.345256   1.34598   1.34675  1.347564  1.348415  1.349293  1.350189  1.351087  1.351973  1.352831  1.353648  1.354411   1.35511   1.35574  1.356297  1.356783  1.357201  1.357556  1.357854  1.358102  1.358306  1.358474   1.35861  1.358721   1.35881  1.358882  1.358939  1.358985  1.359022  1.359051  1.359074          1.359092 | 0.754891  0.752237  0.749374  0.746289   0.74297  0.739404  0.735576  0.731471  0.727073  0.722367  0.717336  0.711965  0.706241   0.70015  0.693685  0.686843  0.679626  0.672045  0.664123  0.655894  0.647409  0.638734   0.62995  0.621157  0.612465   0.60399  0.595851  0.588155  0.580998  0.574449  0.568553  0.563327  0.558762  0.554828  0.551479  0.548659  0.546306  0.544361  0.542763  0.541459    0.5404  0.539544  0.538855  0.538301  0.537858  0.537505  0.537223  0.536999  0.536822      0.536681 | -0.448531  -0.444655  -0.440511  -0.436086  -0.431364  -0.426332  -0.420974  -0.415275  -0.409221  -0.402797  -0.395993  -0.388799  -0.381209  -0.373224   -0.36485  -0.356103   -0.34701  -0.337611   -0.32796  -0.318128  -0.308204  -0.298291  -0.288504  -0.278967  -0.269805  -0.261134  -0.253056   -0.24565  -0.238966  -0.233027  -0.227825  -0.223332  -0.219498  -0.216263  -0.213561  -0.211322  -0.209482  -0.207978  -0.206757  -0.205769  -0.204973  -0.204335  -0.203824  -0.203416  -0.203091  -0.202833  -0.202629  -0.202466  -0.202338     -0.202237 | 0.696608  0.695684  0.694717  0.693706  0.692651  0.691552  0.690411  0.689227  0.688003  0.686741  0.685445  0.684119   0.68277  0.681405  0.680033  0.678664  0.677311  0.675986  0.674706  0.673484  0.672336  0.671275  0.670312  0.669457  0.668713  0.668079  0.667551  0.667121  0.666778   0.66651  0.666303  0.666147   0.66603  0.665943  0.665879  0.665832  0.665798  0.665773  0.665755  0.665742  0.665733  0.665726  0.665721  0.665717  0.665715  0.665713  0.665711  0.665711   0.66571       0.66571 | 797.923464 | 785.096097  759.067949  732.216245  704.642249   676.45219  647.757589  618.675504  589.328644  559.845334  530.359289  501.009154  471.937783  443.291202  415.217224  387.863686  361.376285  335.896008  311.556199  288.479303  266.773417  246.528805  227.814591  210.675915  195.131832  181.174263  168.768232  157.853513  148.347671  140.150297  133.148051  127.220052  122.243112  118.096367   114.66501  111.842972  109.534572  107.655242  106.131525  104.900552  103.909171  103.112916  102.474905  101.964764  101.557634  101.233266  100.975238  100.770277  100.607689  100.478879          100.376955 | 100.331203 | 1.224961  1.214552  1.203527  1.191881  1.179613  1.166721  1.153206  1.139071  1.124322   1.10897  1.093032  1.076531  1.059498  1.041974   1.02401  1.005672  0.987039  0.968206  0.949285  0.930402  0.911702  0.893342  0.875485  0.858302  0.841954  0.826591  0.812341    0.7993  0.787527  0.777044  0.767832  0.759838  0.752985  0.747172   0.74229  0.738225  0.734866   0.73211  0.729861  0.728035  0.726558  0.725369  0.724413  0.723648  0.723037   0.72255  0.722162  0.721854   0.72161       0.721417 | 5.934663  5.857679   5.77648    5.6911  5.601584  5.507993  5.410403  5.308914  5.203654  5.094784  4.982508  4.867076  4.748793   4.62803  4.505227  4.380902   4.25566  4.130191  4.005273  3.881764  3.760594   3.64274  3.529195  3.420936   3.31887   3.22379  3.136324  3.056899  2.985716  2.922742  2.867729  2.820244  2.779713  2.745471  2.716808  2.693011  2.673396  2.657329  2.644239  2.633624  2.625051  2.618151  2.612613  2.608181  2.604641   2.60182  2.599576  2.597794  2.596381        2.595262 | 797.923464  772.329174  745.861719  718.619959  690.707528  662.233208  633.311228  604.061444  574.609394  545.086176  515.628118    486.3762  457.475187  429.072432  401.316309  374.354257  348.330412  323.382844  299.640442  277.219523  256.220306  236.723444   218.78686  202.443177  187.698061  174.529743  162.889945  152.706259  143.885867  136.320318  129.890901    124.4741  119.946643   116.18975  113.092359  110.553262  108.482246  106.800392  105.439756  104.342615  103.460477  102.752989  102.186826  101.734649  101.374146  101.087194  100.859125  100.678107   100.53462  100.421025  100.331203 |  100.0 | 697.923464 | 100.331203 |
    | 21:30 | 9.504436  9.385452  9.259074  9.125364  8.984394  8.836255  8.681057  8.518942  8.350087  8.174715  7.993104    7.8056  7.612622  7.414681  7.212389  7.006472  6.797781  6.587303  6.376164  6.165632  5.957108  5.752108  5.552233  5.359126  5.174411  4.999624  4.836133  4.685063  4.547224  4.423068  4.312668   4.21573  4.131637  4.059513  3.998303   3.94685  3.903966  3.868493  3.839343  3.815526   3.79616  3.780482  3.767834  3.757663  3.749506  3.742979  3.737769  3.733616  3.730312             3.727688 | 594.238098  581.215344  567.538362  553.241771  538.362623  522.940837  507.019634  490.645983  473.871029  456.750512  439.345139  421.720912  403.949357  386.107648  368.278571  350.550287  333.015836  315.772336  298.919809  282.559596  266.792311  251.715344  237.419952  223.988037  211.488801   199.97554  189.482908  180.024998  171.594579  164.163678  157.685547  152.097846  147.326669   143.29094    139.9067  137.090871  134.764245  132.853592  131.292929   130.02406  128.996572  128.167433  127.500363  126.965087  126.536549   126.19416  125.921095  125.703675   125.53082  125.393585 | 111.926613  110.713207  109.424398  108.060815  106.623204  105.112473  103.529761  101.876502  100.154509   98.366058   96.513989   94.601809   92.633812   90.615202    88.55222    86.45227   84.324035   82.177571   80.024371   77.877359   75.750822   73.660221   71.621889   69.652575   67.768844   65.986357   64.319073   62.778451   61.372766   60.106621   58.980757   57.992179   57.134594   56.399075   55.774854    55.25013   54.812798   54.451044   54.153772   53.910879   53.713391     53.5535   53.424516   53.320791   53.237605   53.171049   53.117909    53.07556   53.041868        53.015103 | 110.044364  108.854522   107.59074  106.253637  104.843943   103.36255  101.810573  100.189419   98.500866   96.747146   94.931042   93.055995   91.126215   89.146806   87.123885   85.064715   82.977809   80.873028   78.761642   76.656323   74.571082   72.521079    70.52233   68.591258   66.744108   64.996237    63.36133   61.850626   60.472238   59.230681    58.12668     57.1573   56.316369   55.595133   54.983034   54.468499    54.03966   53.684931   53.393432   53.155256   52.961603   52.804817   52.678338   52.576627   52.495057   52.429793   52.377685   52.336159    52.30312     52.276875 | 1.840446  1.826858  1.812365  1.796963  1.780648  1.763415  1.745263  1.726192  1.706206  1.685313  1.663526  1.640866  1.617363  1.593059  1.568005  1.542271  1.515943  1.489125  1.461944  1.434551   1.40712  1.379846  1.352947  1.326656  1.301213  1.276861  1.253824  1.232305  1.212465  1.194421   1.17823  1.163897   1.15137  1.140557  1.131327   1.12353  1.117005  1.111588  1.107124  1.103468  1.100489  1.098074  1.096122  1.094552  1.093291  1.092281  1.091475  1.090832   1.09032  1.089913 | 1.338651  1.338778  1.338918  1.339072   1.33924  1.339424  1.339627  1.339848  1.340091  1.340357  1.340649  1.340969  1.341319  1.341703  1.342123  1.342582  1.343084  1.343631  1.344225  1.344868  1.345561  1.346301  1.347088  1.347916  1.348777  1.349661  1.350557  1.351449  1.352324  1.353165   1.35396  1.354696  1.355367  1.355967  1.356495  1.356953  1.357345  1.357676  1.357953  1.358183  1.358373  1.358527  1.358653  1.358755  1.358838  1.358904  1.358957  1.358999  1.359033          1.359059 | 0.757992  0.755685  0.753184  0.750478  0.747556  0.744407  0.741018  0.737374  0.733462  0.729266   0.72477  0.719958  0.714815  0.709325  0.703477  0.697259  0.690665  0.683694  0.676352  0.668655  0.660629  0.652316  0.643772  0.635069  0.626296  0.617556  0.608963  0.600633  0.592682  0.585212  0.578308  0.572031  0.566414  0.561462  0.557159  0.553468  0.550338  0.547712  0.545527  0.543725  0.542247  0.541043  0.540066  0.539277  0.538643  0.538133  0.537726    0.5374  0.537141      0.536935 | -0.453435  -0.450005  -0.446324   -0.44238  -0.438161  -0.433653  -0.428841  -0.423712  -0.418249  -0.412438  -0.406265  -0.399717  -0.392784  -0.385457  -0.377734  -0.369618  -0.361117  -0.352254  -0.343058  -0.333577  -0.323872  -0.314019  -0.304113  -0.294263  -0.284588  -0.275211  -0.266255  -0.257831  -0.250031  -0.242924  -0.236547  -0.230913  -0.226004  -0.221783  -0.218196   -0.21518  -0.212668  -0.210593   -0.20889  -0.207501  -0.206374  -0.205463   -0.20473  -0.204142  -0.203672  -0.203296  -0.202997   -0.20276  -0.202571     -0.202422 | 0.697894  0.697048  0.696158  0.695225  0.694248  0.693227  0.692161  0.691051  0.689898  0.688703  0.687468  0.686197  0.684893  0.683562   0.68221  0.680844  0.679476  0.678115  0.676775  0.675469  0.674213  0.673022  0.671909  0.670888  0.669969  0.669159  0.668459  0.667868  0.667381  0.666986  0.666674   0.66643  0.666244  0.666104  0.665999  0.665921  0.665863  0.665821  0.665791  0.665768  0.665752   0.66574  0.665731  0.665725   0.66572  0.665717  0.665715  0.665713  0.665712      0.665711 | 828.701418 |   816.9889  793.109331  768.222123  742.421108  715.804675  688.476186  660.544327  632.123381  603.333393  574.300207  545.155324  516.035568  487.082493  458.441524  430.260755  402.689407  375.875879  349.965419  325.097392  301.402208  278.997978  257.987032  238.452498  220.455154  204.030864  189.188885  175.911311  164.153873   153.84815  144.905101  137.219651  130.675902  125.152492  120.527603  116.683242  113.508554  110.902082  108.773029  107.041683  105.639198  104.506921  103.595458  102.863613  102.277302  101.808513  101.434352  101.136197  100.898958  100.710445          100.560843 | 100.493595 | 1.237366  1.228113  1.218243  1.207752  1.196636  1.184892  1.172518  1.159514  1.145882  1.131625  1.116753  1.101278  1.085221  1.068607  1.051472  1.033862  1.015832  0.997456  0.978817  0.960017  0.941175  0.922424  0.903914  0.885805  0.868262  0.851454  0.835538  0.820655  0.806921  0.794418   0.78319  0.773241  0.764541  0.757025  0.750607  0.745182  0.740641  0.736869   0.73376  0.731213  0.729137  0.727454  0.726094  0.724999   0.72412  0.723416  0.722854  0.722406  0.722049       0.721765 | 6.026813  5.958037  5.884943  5.807559  5.725915   5.64005  5.550018  5.455885  5.357739  5.255691  5.149885    5.0405   4.92776  4.811939  4.693369  4.572451  4.449656   4.32554  4.200741  4.075985  3.952086  3.829929  3.710466  3.594683  3.483568  3.378074  3.279066  3.187273  3.103246  3.027323  2.959613  2.899997  2.848151  2.803586   2.76569  2.733781  2.707147  2.685089  2.666944  2.652105  2.640032  2.630252  2.622358  2.616008  2.610913  2.606836  2.603581  2.600986  2.598921         2.59728 | 828.701418  805.346262  780.937487  755.566733  729.329986  702.328005  674.666717  646.457533  617.817575  588.869785  559.742886  530.571153   501.49396  472.655072  444.201619  416.282743  389.047856  362.644506   337.21584  312.897692  289.815339   268.08005  247.785566  229.004739  211.786581  196.154045   182.10282  169.601406  158.592616  148.996508  140.714546   133.63465  127.636628  122.597499  118.396232  114.917588  112.054906  109.711828  107.803089  106.254555  105.002723   103.99387  103.183019  102.532843  102.012583  101.597049  101.265714  101.001916  100.792184  100.625656  100.493595 |  100.0 | 728.701418 | 100.493595 |
    | 22:00 | 9.635642  9.531101  9.418994  9.299375    9.1723  9.037835   8.89606  8.747073  8.590997  8.427989  8.258248  8.082018  7.899607  7.711391  7.517828  7.319472  7.116981  6.911133  6.702834  6.493129  6.283201  6.074372   5.86809  5.665909  5.469448  5.280347  5.100197  4.930473   4.77245  4.627132  4.495188  4.376921  4.272257  4.180774   4.10175  4.034236  3.977137  3.929285  3.889506  3.856673  3.829741  3.807768  3.789922  3.775486  3.763847  3.754492  3.746991   3.74099  3.736199              3.73238 | 608.762608   597.17591   584.87214  571.882458  558.239913  543.979849  529.140312  513.762479   497.89109  481.574877  464.866981   447.82534  430.513042    412.9986  395.356136  377.665433  360.011811   342.48579  325.182463  308.200552  291.641066  275.605534  260.193781  245.501268   231.61605  218.615499     206.563  195.504915  185.468155  176.458703  168.461381   161.44099  155.344775   150.10597  145.647998  141.888846  138.745153  136.135666  133.983867  132.219735  130.780705  129.611981  128.666363  127.903745  127.290445  126.798435  126.404575  126.089885  125.838881  125.638981 | 113.264653  112.198537  111.055267  109.835385  108.539471  107.168196  105.722367  104.202987  102.611321  100.948966   99.217931   97.420734   95.560499   93.641064   91.667106   89.644261    87.57925   85.480004   83.355767   81.217182   79.076326    76.94668   74.843015   72.781162   70.777648    68.84918   67.012008   65.281156   63.669633   62.187668     60.8421   59.636009   58.568647   57.635697   56.829803   56.141297   55.558999      55.071   54.665333   54.330503    54.05585   53.831762    53.64977    53.50255   53.383862   53.288457   53.211961   53.150763   53.101899        53.062953 |  111.35642  110.311007  109.189939  107.993746  106.722998  105.378352  103.960599  102.470725  100.909969   99.279895   97.582476   95.820179   93.996068   92.113908   90.178283    88.19472    86.16981    84.11133   82.028344    79.93129   77.832007   75.743717   73.680904   71.659091   69.694483   67.803466    66.00197    64.30473   62.724502   61.271315   59.951878   58.769208   57.722574   56.807741   56.017497   55.342361   54.771371   54.292848   53.895059   53.566731   53.297412   53.077675   52.899218   52.754857   52.638474   52.544921    52.46991   52.409901   52.361986     52.323796 | 1.855369  1.843484  1.830694  1.816993  1.802377  1.786842  1.770383  1.752997  1.734683  1.715443  1.695282  1.674212  1.652248  1.629416  1.605749  1.581295  1.556112  1.530275  1.503879  1.477037  1.449885  1.422583  1.395314  1.368283  1.341713  1.315841  1.290907  1.267148   1.24478  1.223991  1.204923   1.18767  1.172268    1.1587  1.146897  1.136751  1.128124  1.120861  1.114799   1.10978  1.105651  1.102275  1.099528  1.097303  1.095507  1.094062  1.092902  1.091973  1.091232   1.09064 | 1.338516  1.338624  1.338742  1.338873  1.339017  1.339175  1.339349  1.339539  1.339748  1.339977  1.340229  1.340504  1.340806  1.341137  1.341499  1.341897  1.342332  1.342807  1.343326   1.34389  1.344503  1.345164  1.345875  1.346633  1.347434  1.348274  1.349143  1.350031  1.350925  1.351809   1.35267  1.353492  1.354262  1.354971  1.355613  1.356183  1.356682  1.357112  1.357479  1.357788  1.358046  1.358259  1.358434  1.358577  1.358693  1.358787  1.358863  1.358924  1.358973          1.359011 | 0.760483  0.758502   0.75634  0.753987  0.751435  0.748672  0.745688  0.742471  0.739007  0.735282  0.731282   0.72699  0.722392  0.717472  0.712214  0.706604   0.70063  0.694285  0.687562  0.680464  0.673001  0.665194  0.657073  0.648687  0.640097  0.631383  0.622639  0.613973  0.605499  0.597334  0.589588  0.582357  0.575717  0.569716  0.564377  0.559697  0.555649  0.552192  0.549272  0.546829  0.544802  0.543134  0.541769  0.540658  0.539757  0.539031  0.538446  0.537978  0.537603      0.537303 | -0.457545  -0.454535  -0.451291  -0.447802  -0.444055  -0.440039  -0.435741  -0.431146  -0.426241   -0.42101  -0.415439  -0.409513   -0.40322  -0.396546  -0.389483  -0.382024  -0.374169  -0.365924  -0.357302  -0.348329  -0.339041  -0.329491  -0.319746   -0.30989  -0.300022  -0.290257  -0.280714  -0.271519  -0.262788  -0.254625  -0.247114   -0.24031  -0.234242  -0.228909  -0.224286  -0.220328  -0.216977   -0.21417  -0.211838  -0.209916  -0.208341  -0.207059  -0.206019   -0.20518  -0.204505  -0.203964  -0.203531  -0.203185   -0.20291     -0.202691 | 0.699028  0.698259  0.697448  0.696594  0.695696  0.694753  0.693766  0.692734  0.691657  0.690537  0.689373  0.688168  0.686923  0.685644  0.684333  0.682998  0.681644   0.68028  0.678918  0.677568  0.676243  0.674959  0.673731  0.672572  0.671497  0.670518  0.669643  0.668878  0.668223  0.667674  0.667224  0.666863  0.666579  0.666359  0.666191  0.666065  0.665971  0.665901  0.665849  0.665812  0.665784  0.665764  0.665749  0.665738   0.66573  0.665724   0.66572  0.665717  0.665714      0.665713 |  854.27953 | 843.829311  822.400281  799.796835  776.105316  751.416181  725.824452  699.430125  672.338547  644.660711  616.513465  588.019611  559.307841  530.512507  501.773158  473.233832  445.042035  417.347394  390.299939  364.047993  338.735666  314.499984  291.467691  269.751847  249.448346  230.632596  213.356582  197.646629  183.502146  170.895602   159.77388  150.061027  141.662222  134.468634   128.36272   123.22346  118.931092  115.371004  112.436621  110.031251  108.069009  106.474962  105.184725  104.143685   103.30601   102.63358  102.094932  101.664251  101.320471  101.046472          100.828393 |  100.73023 | 1.247525  1.239434  1.230726  1.221396  1.211441  1.200857  1.189641  1.177791  1.165304  1.152182  1.138428  1.124048  1.109052  1.093456  1.077283  1.060563  1.043334  1.025648  1.007567  0.989168  0.970542  0.951798  0.933061   0.91447  0.896179   0.87835  0.861151  0.844746  0.829285  0.814902  0.801697  0.789737  0.779053  0.769633  0.761432  0.754379  0.748379  0.743324  0.739105   0.73561  0.732734  0.730382  0.728468  0.726917  0.725665  0.724658  0.723849  0.723202  0.722685       0.722272 | 6.102608   6.04222  5.977429  5.908257  5.834729  5.756873  5.674723  5.588324  5.497736  5.403032  5.304313  5.201702   5.09536  4.985483  4.872317   4.75616  4.637373  4.516385  4.393702  4.269913  4.145693  4.021803  3.899086  3.778455  3.660876  3.547338  3.438819  3.336238  3.240408   3.15199  3.071453  2.999044   2.93478   2.87846  2.829695  2.787945   2.75257  2.722876  2.698158  2.677733  2.660963  2.647269  2.636141  2.627134  2.619869  2.614028  2.609343  2.605594    2.6026        2.600213 |  854.27953  833.456799  811.417315  788.245466  764.029518  738.862085  712.840558  686.067507  658.651018  630.704956  602.349127  573.709301  544.917079   516.10956  487.428761  459.020771  431.034579  403.620547  376.928504  351.105441  326.292811  302.623478  280.218376  259.183022  239.604039  221.545946  205.048476  190.124743  176.760534   164.91494  154.522434  145.496305  137.733195  131.118328  125.530925  120.849308   116.95529  113.737587  111.094177  108.933631  107.175593  105.750594  104.599401  103.672099  102.927027  102.329695  101.851742  101.469975  101.165515  100.923057   100.73023 |  100.0 |  754.27953 |  100.73023 |
    | 22:30 | 9.738576  9.648725   9.55116  9.445932  9.333092  9.212692  9.084791  8.949459  8.806782  8.656868  8.499855  8.335915  8.165267  7.988183  7.804996  7.616119  7.422047  7.223377  7.020817  6.815195  6.607475  6.398755  6.190277  5.983412  5.779651  5.580576  5.387817  5.202995  5.027661  4.863211  4.710817  4.571351   4.44534  4.332935  4.233924  4.147764  4.073642  4.010552  3.957369  3.912925  3.876069   3.84571  3.820848   3.80059  3.784155   3.77087  3.760165  3.751563  3.744668             3.739152 | 620.277911  610.220312  599.390663  587.817172  575.529452  562.558888  548.939021  534.705947  519.898728  504.559812  488.735454  472.476138  455.836971  438.878051  421.664779  404.268095    386.7646  369.236538  351.771575  334.462352  317.405727  300.701673  284.451768   268.75725  253.716616  239.422813  225.960094  213.400725  201.801769   191.20228  181.621233  173.056532  165.485322  158.865661  153.139439  148.236203  144.077452  140.580907  137.664355  135.248768  133.260572  131.633068  130.307127  129.231291  128.361463  127.660338  127.096686  126.644593  126.282712   125.99356 | 114.314374  113.398071    112.4031  111.329986   110.17924  108.951396  107.647055   106.26693  104.811902  103.283075  101.681848  100.009988   98.269716   96.463798   94.595657   92.669475   90.690323   88.664281   86.598561   84.501626   82.383284   80.254756   78.128682   76.019066    73.94111   71.910936   69.945165   68.060348    66.27228   64.595217   63.041095   61.618819   60.333748   59.187442   58.177725   57.299061   56.543165   55.899764   55.357398   54.904159     54.5283   54.218699   53.965159   53.758569   53.590959   53.455477   53.346311    53.25859    53.18827        53.132016 | 112.385756  111.487247  110.511598  109.459323  108.330924  107.126924  105.847912  104.494588  103.067815  101.568677   99.998545   98.359152   96.652674   94.881827   93.049963   91.161186   89.220469   87.233771   85.208166   83.151951   81.074747   78.987553   76.902766   74.834117   72.796513   70.805764   68.878167   67.029952   65.276607   63.632111   62.108169   60.713513   59.453397   58.329351   57.339243   56.477642   55.736425   55.105518   54.573685   54.129247   53.760687   53.457098   53.208482   53.005904   52.841549   52.708697   52.601651   52.515635    52.44668     52.391518 | 1.867031  1.856853  1.845768  1.833772  1.820861   1.80703  1.792275  1.776592  1.759976  1.742425  1.723939  1.704523  1.684183  1.662933  1.640793  1.617791  1.593967  1.569372  1.544072   1.51815  1.491707  1.464866  1.437773  1.410597  1.383528   1.35678  1.330581   1.30517  1.280786  1.257657  1.235989  1.215953   1.19767  1.181213  1.166595  1.153779   1.14268  1.133178  1.125127   1.11837  1.112747    1.1081  1.104286  1.101171   1.09864  1.096591  1.094938  1.093609  1.092543  1.091689 | 1.338413  1.338503  1.338603  1.338713  1.338836  1.338971  1.339119  1.339283  1.339462  1.339659  1.339875  1.340112  1.340372  1.340657   1.34097  1.341312  1.341688  1.342099  1.342549  1.343041  1.343577  1.344159   1.34479   1.34547  1.346198  1.346972  1.347788  1.348639  1.349514  1.350404  1.351292  1.352166   1.35301   1.35381  1.354555  1.355236  1.355847  1.356388  1.356858  1.357262  1.357605  1.357892  1.358132  1.358329  1.358491  1.358623   1.35873  1.358817  1.358887          1.358943 |   0.7624  0.760728  0.758885  0.756863  0.754655  0.752252  0.749644  0.746821  0.743771  0.740481  0.736938  0.733128  0.729036  0.724645  0.719942   0.71491  0.709534  0.703801  0.697701  0.691227  0.684377  0.677155  0.669576  0.661665  0.653459  0.645011  0.636391  0.627684  0.618988  0.610416  0.602081  0.594099  0.586574  0.579595  0.573225  0.567504  0.562442  0.558028  0.554228  0.550996  0.548276  0.546006  0.544129  0.542586  0.541326  0.540301  0.539472  0.538803  0.538265      0.537834 | -0.460898   -0.45829   -0.45546  -0.452399  -0.449097  -0.445543  -0.441726  -0.437633  -0.433251  -0.428566  -0.423563  -0.418227  -0.412545  -0.406502  -0.400085  -0.393284   -0.38609    -0.3785  -0.370516  -0.362146  -0.353409  -0.344334  -0.334964  -0.325358  -0.315589  -0.305748  -0.295939   -0.28628  -0.276892  -0.267897  -0.259408   -0.25152  -0.244306  -0.237811   -0.23205  -0.227013  -0.222668  -0.218963  -0.215838  -0.213228  -0.211066  -0.209287  -0.207832  -0.206649  -0.205691  -0.204919  -0.204297  -0.203799  -0.203401     -0.203083 |  0.70001  0.699319  0.698585   0.69781  0.696991  0.696128  0.695221  0.694269  0.693272  0.692229  0.691142   0.69001  0.688836  0.687621  0.686368  0.685082  0.683766  0.682427  0.681073  0.679714   0.67836  0.677023  0.675717  0.674457  0.673259  0.672135    0.6711  0.670164  0.669334  0.668614  0.668002  0.667494  0.667081  0.666752  0.666494  0.666295  0.666144   0.66603  0.665946  0.665883  0.665837  0.665803  0.665778  0.665759  0.665746  0.665736  0.665729  0.665723  0.665719      0.665716 | 874.327871 | 865.261866  846.534985  826.485696  805.192852  782.738975  759.210705  734.699246  709.300787  683.116876  656.254748  628.827567  600.954576  572.761106  544.378439  515.943457  487.598075  459.488388  431.763511   404.57407  378.070319  352.399862  327.704993  304.119687  281.766312   260.75218  241.166117  223.075281  206.522483  191.524323   178.07041  166.123874  155.623277  146.485857  138.611865  131.889621  126.200804  121.425504  117.446617  114.153342  111.443631   109.22566  107.418427   105.95167  104.765308  103.808574  103.039003   102.42137  101.926654  101.531088          101.215299 |  101.07295 | 1.255463  1.248535  1.240989  1.232821   1.22403   1.21461  1.204559  1.193872  1.182548  1.170583  1.157978  1.144733  1.130854  1.116348  1.101228  1.085513  1.069228  1.052407  1.035094  1.017344  0.999225  0.980821  0.962229  0.943564  0.924957  0.906553   0.88851  0.870992  0.854164  0.838188  0.823205  0.809336   0.79667  0.785259  0.775115  0.766214  0.758501  0.751894  0.746293  0.741591  0.737676   0.73444  0.731783  0.729613  0.727849  0.726421  0.725269  0.724342  0.723598       0.723003 | 6.162039  6.110163  6.053809  5.993001  5.927759  5.858106  5.784067  5.705671  5.622959  5.535978  5.444796  5.349497   5.25019  5.147015  5.040148  4.929804  4.816252  4.699815  4.580882  4.459913  4.337447  4.214108  4.090606  3.967734  3.846363   3.72743  3.611909  3.500787   3.39502  3.295487  3.202943  3.117974   3.04096  2.972058  2.911198  2.858103  2.812324  2.773279  2.740308  2.712714  2.689802  2.670908  2.655422  2.642795  2.632543  2.624253   2.61757  2.612199  2.607891        2.604445 | 874.327871  856.279772  836.870575  816.177403  794.280814  771.265259  747.219535  722.237221  696.417077  669.863405  642.686338  615.002042  586.932806  558.606987  530.158777  501.727758    473.4582  445.498069  417.997705  391.108131   364.97898  339.756022  315.578317  292.575039  270.862066  250.538486  231.683209  214.351951  198.574871  184.355165   171.66888  160.466115  150.673642  142.198793  134.934292  128.763583  123.566132  119.222261  115.617143  112.643788  110.204974   108.21423  106.596031  105.285427  104.227278  103.375276  102.690885  102.142271  101.703305  101.352648   101.07295 |  100.0 | 774.327871 |  101.07295 |
    | 23:00 | 9.813468  9.738463   9.65561  9.564967  9.466583  9.360505  9.246782  9.125465  8.996616   8.86031  8.716641  8.565729  8.407724  8.242818   8.07125  7.893317  7.709386  7.519901  7.325401  7.126524  6.924029  6.718798  6.511846  6.304331  6.097545  5.892908  5.691946  5.496261  5.307481  5.127202  4.956914  4.797928  4.651297  4.517758  4.397688  4.291094   4.19763  4.116645  4.047249   3.98839  3.938927  3.897702  3.863592  3.835548  3.812617  3.793954  3.778826  3.766606  3.756765             3.748859 | 628.722832  620.265221  610.988222  600.917511  590.079743   578.50289  566.216586  553.252497  539.644702  525.430088  510.648762  495.344458  479.564961  463.362497  446.794124  429.922059  412.813961  395.543112  378.188481  360.834616  343.571339  326.493166  309.698414   293.28794  277.363458  262.025414  247.370426  233.488335  220.459002  208.349035   197.20872  187.069492  177.942286   169.81705  162.663604  156.433822  151.064924  146.483515  142.609884   139.36211  136.659615   134.42593  132.590614  131.090367  129.869472  128.879726  128.080021  127.435712  126.917886  126.502613 | 115.078133  114.313223  113.468292   112.54391  111.540586    110.4588  109.299041  108.061845  106.747838  105.357784  103.892645  102.353634  100.742296   99.060578   97.310922   95.496359    93.62062   91.688249   89.704721   87.676574    85.61152    83.51856   81.408064   79.291817   77.183003   75.096102   73.046686   71.051083   69.125901   67.287407   65.550807   63.929458    62.43411   61.072271   59.847791   58.760739   57.807592   56.981708   56.274008   55.673754   55.169328   54.748916   54.401065   54.115069   53.881212   53.690887   53.536615   53.411998   53.311635        53.231012 | 113.134684  112.384627  111.556104  110.649673  109.665833  108.605054  107.467817  106.254647  104.966157  103.603097  102.166409  100.657286   99.077239   97.428178   95.712499   93.933174   92.093861   90.199015   88.254006   86.265244   84.240292   82.187976   80.118464   78.043314   75.975451   73.929076   71.919459   69.962609   68.074812   66.272021   64.569144   62.979281   61.512972   60.177579   58.976878   57.910935   56.976298   56.166452   55.472495   54.883898   54.389266   53.977019   53.635923   53.355481   53.126165   52.939536    52.78826   52.666063   52.567649     52.488592 | 1.875492  1.867018  1.857634  1.847339  1.836129  1.824002  1.810952  1.796975  1.782065  1.766219  1.749434  1.731708  1.713043  1.693443  1.672919  1.651488  1.629172  1.606004  1.582029  1.557304  1.531901  1.505911  1.479444  1.452631  1.425628  1.398611  1.371782  1.345358  1.319572  1.294663  1.270867  1.248403  1.227463  1.208198  1.190711   1.17505  1.161208  1.149128  1.138711  1.129827  1.122326   1.11605  1.110839  1.106542  1.103021  1.100149  1.097818  1.095933  1.094413  1.093191 |  1.33834  1.338414  1.338496  1.338588  1.338691  1.338806  1.338932  1.339071  1.339225  1.339394  1.339579  1.339783  1.340007  1.340252  1.340521  1.340817   1.34114  1.341495  1.341884   1.34231  1.342776  1.343284  1.343838  1.344439  1.345088  1.345786  1.346531  1.347321  1.348149  1.349009  1.349888  1.350776  1.351658  1.352519  1.353344   1.35412  1.354838   1.35549  1.356071  1.356582  1.357024  1.357403  1.357722   1.35799  1.358212  1.358394  1.358544  1.358666  1.358764          1.358844 | 0.763774  0.762398  0.760857  0.759148  0.757263  0.755195  0.752937   0.75048  0.747812  0.744924  0.741803  0.738437  0.734811  0.730912  0.726723  0.722231  0.717418  0.712271  0.706775  0.700918  0.694691  0.688088  0.681112   0.67377   0.66608  0.658073  0.649792  0.641296  0.632662  0.623979  0.615352  0.606893  0.598717  0.590936  0.583647  0.576929  0.570836  0.565395  0.560608  0.556453  0.552894  0.549877  0.547346  0.545241  0.543503  0.542078  0.540915  0.539971  0.539208      0.538592 | -0.463533  -0.461308  -0.458872  -0.456217  -0.453335  -0.450216  -0.446852  -0.443229  -0.439337  -0.435163  -0.430693  -0.425913  -0.420809  -0.415365  -0.409569  -0.403405  -0.396864  -0.389934  -0.382609  -0.374889  -0.366777  -0.358286   -0.34944  -0.340272  -0.330833  -0.321185  -0.311409  -0.301602  -0.291873  -0.282341  -0.273129  -0.264354  -0.256122  -0.248521  -0.241611  -0.235425  -0.229969  -0.225223  -0.221146  -0.217685  -0.214775  -0.212352  -0.210348  -0.208704  -0.207361   -0.20627  -0.205388  -0.204677  -0.204105     -0.203647 | 0.700842  0.700227  0.699571  0.698874  0.698134  0.697351  0.696524  0.695652  0.694736  0.693774  0.692766  0.691713  0.690615  0.689473  0.688289  0.687065  0.685804  0.684511  0.683191  0.681851    0.6805  0.679146  0.677803  0.676482  0.675198  0.673965  0.672798  0.671712  0.670718  0.669827  0.669042  0.668367  0.667798  0.667329   0.66695   0.66665  0.666417  0.666237  0.666101     0.666  0.665924  0.665867  0.665826  0.665795  0.665772  0.665756  0.665743  0.665734  0.665728      0.665723 | 888.651531 | 881.064673   865.23817  847.961189  829.305117  809.344583  788.157896  765.827492  742.440374   718.08852  692.869281  666.885714  640.246865  613.067953   585.47045  557.582018   529.53627   501.47232   473.53409   445.86932  418.628264  391.962017  366.020471  340.949876  316.890034  293.971174  272.310583  252.009166  233.148091  215.785792  199.955603    185.6643  172.891828  161.592358  151.696733  143.116152  135.746811  129.475069   124.18265  119.751444  116.067537  113.024282  110.524342  108.480802  106.817479  105.468648  104.378352  103.499484  102.792753  102.225641          101.771412 | 101.566343 | 1.261221  1.255454  1.249067  1.242058  1.234427  1.226169  1.217281   1.20776  1.197602  1.186804  1.175362  1.163276  1.150545  1.137173  1.123166  1.108533  1.093289  1.077457  1.061065   1.04415  1.026761   1.00896  0.990818  0.972427  0.953889  0.935327  0.916877  0.898689  0.880922  0.863743  0.847315  0.831791  0.817305  0.803966  0.791846  0.780983  0.771374  0.762983  0.755742  0.749564  0.744345  0.739976  0.736347  0.733355  0.730901  0.728901  0.727276  0.725962  0.724902        0.72405 | 6.205263  6.161974  6.114139  6.061786  6.004937  5.943612  5.877831  5.807617  5.732995  5.653998  5.570668  5.483061  5.391252  5.295334  5.195429  5.091691  4.984312   4.87353  4.759635  4.642976  4.523972  4.403113  4.280974  4.158211  4.035567  3.913868   3.79401  3.676945  3.563653   3.45511  3.352242   3.25588  3.166714  3.085248  3.011774  2.946358  2.888848  2.838895  2.795999  2.759546  2.728863  2.703254   2.68204   2.66458  2.650292  2.638656  2.629218  2.621592  2.615447         2.61051 | 888.651531  873.566327  856.995594  839.009202  819.680035  799.084435   777.30265  754.419275  730.523682  705.710426  680.079614  653.737222   626.79533  599.372267  571.592626  543.587121   515.49226  487.449778  459.605812  432.109768  405.112835  378.766141  353.218508  328.613828  305.088073  282.766022  261.757805  242.155441  224.029582  207.426737  192.367277  178.844493  166.824954  156.250266  147.040205  139.096994  132.310344  126.562795  121.734834  117.709397  114.375458  111.630575   109.38243  107.549485   106.06094  104.856202  103.884029   103.10153    102.4731  101.969389  101.566343 |  100.0 | 788.651531 | 101.566343 |
    | 23:30 | 9.860821  9.800732  9.732675  9.656713  9.572903  9.481292  9.381925  9.274844  9.160097  9.037736  8.907827   8.77045  8.625707  8.473731  8.314688  8.148786   7.97629  7.797523  7.612883  7.422852  7.228009   7.02904  6.826751  6.622079  6.416095  6.210008  6.005161  5.803016   5.60513  5.413113  5.228583  5.053094  4.888065  4.734707  4.593947  4.466379  4.352232  4.251372   4.16333  4.087359  4.022504   3.96768  3.921742  3.883551  3.852018  3.826136  3.805001   3.78782  3.773904              3.76267 | 634.091275  627.282738  619.614911   611.11128  601.795943  591.693914  580.831441  569.236337   556.93833  543.969429  530.364305  516.160686  501.399758  486.126575   470.39046  454.245392   437.75036  420.969674  403.973193   386.83646  369.640693  352.472595  335.423935  318.590846  302.072785  285.971109   270.38721   255.42023  241.164339  227.705686  215.119161  203.465197    192.7869  183.107859  174.430952  166.738387    159.9931  154.141386  149.116514  144.842881  141.240244  138.227606  135.726442  133.663118  131.970465  130.588608  129.465196  128.555182  127.820316  127.228477 | 115.561038  114.948249  114.254197  113.479541  112.624842   111.69059  110.677236  109.585223  108.415026  107.167187  105.842367  104.441388  102.965301  101.415442   99.793508   98.101641   96.342517   94.519444   92.636479   90.698537   88.711516   86.682422   84.619476   82.532221   80.431585   78.329903   76.240867   74.179386   72.161331   70.203142   68.321296   66.531649   64.848682   63.284727   61.849251   60.548307   59.384232   58.355655   57.457798   56.683043   56.021651   55.462552    54.99408   54.604607   54.283026   54.019081   53.803552   53.628332    53.48642        53.371855 | 113.608211  113.007322  112.326748  111.567135  110.729033  109.812924  108.819248  107.748442  106.600969  105.377363  104.078269  102.704497  101.257075   99.737312   98.146876   96.487864     94.7629    92.97523   91.128831   89.228522   87.280089   85.290397   83.267513   81.220791   79.160948   77.100079    75.05161   73.030162   71.051297   69.131134   67.285832   65.530939   63.880654    62.34707    60.93947   59.663789   58.522321   57.513718   56.633297   55.873586   55.225039   54.676797   54.217422   53.835512   53.520177   53.261357   53.050013   52.878196    52.73904       52.6267 | 1.880831  1.874054  1.866364  1.857759  1.848241  1.837807  1.826454  1.814177  1.800971  1.786831  1.771752  1.755731  1.738765  1.720853     1.702  1.682212    1.6615  1.639886  1.617395  1.594066  1.569948  1.545103  1.519613  1.493575  1.467107   1.44035  1.413468  1.386648  1.360096  1.334037  1.308706   1.28434  1.261169  1.239402  1.219213  1.200735  1.184049   1.16918  1.156101  1.144739  1.134982  1.126691  1.119713   1.11389  1.109067  1.105098   1.10185  1.099205  1.097059  1.095325 | 1.338295  1.338353  1.338419  1.338495   1.33858  1.338676  1.338782    1.3389  1.339031  1.339175  1.339334  1.339509  1.339701  1.339912  1.340144  1.340398  1.340677  1.340983  1.341318  1.341686  1.342089   1.34253  1.343012  1.343538  1.344109  1.344728  1.345396  1.346112  1.346873  1.347677  1.348517  1.349383  1.350265  1.351149  1.352021  1.352866   1.35367  1.354421  1.355111  1.355732  1.356284  1.356766  1.357181  1.357534  1.357832   1.35808  1.358286  1.358455  1.358593          1.358705 | 0.764635  0.763542  0.762291  0.760878  0.759298  0.757547  0.755616  0.753499  0.751186   0.74867  0.745939  0.742981  0.739786  0.736339  0.732626  0.728633  0.724345  0.719746  0.714822  0.709556  0.703938  0.697954  0.691599  0.684869  0.677769   0.67031  0.662516  0.654422  0.646077  0.637549  0.628917  0.620277  0.611738  0.603414  0.595416  0.587852  0.580812  0.574363   0.56855  0.563389  0.558872  0.554971  0.551642  0.548831   0.54648   0.54453  0.542923  0.541607  0.540536      0.539666 | -0.465482  -0.463626  -0.461567  -0.459299  -0.456815  -0.454108   -0.45117  -0.447989  -0.444557  -0.440861   -0.43689   -0.43263  -0.428068  -0.423189  -0.417979  -0.412424   -0.40651  -0.400224  -0.393555  -0.386495   -0.37904  -0.371191  -0.362955  -0.354349    -0.3454   -0.33615  -0.326652  -0.316978  -0.307214  -0.297461  -0.287833   -0.27845  -0.269433  -0.260896  -0.252936   -0.24563  -0.239029  -0.233153  -0.227998  -0.223535  -0.219718  -0.216489  -0.213783  -0.211536  -0.209682  -0.208163  -0.206925   -0.20592  -0.205107     -0.204453 | 0.701525  0.700986  0.700407  0.699787  0.699125  0.698421  0.697673  0.696882  0.696047  0.695166   0.69424  0.693267  0.692249  0.691186  0.690077  0.688925  0.687732  0.686499  0.685231  0.683933  0.682611  0.681271  0.679924  0.678579  0.677248  0.675946  0.674686  0.673482  0.672351  0.671304  0.670353  0.669507  0.668768  0.668137   0.66761  0.667178  0.666832  0.666559  0.666347  0.666186  0.666064  0.665972  0.665904  0.665853  0.665816  0.665788  0.665767  0.665752  0.665741      0.665733 | 897.187983 | 891.147552  878.364804  864.024233  848.189792  830.928283   812.30978  792.408058   771.30102   749.07113  725.805825  701.597906  676.545894  650.754331  624.334007  597.402092  570.082143  542.503957  514.803237  487.121036   459.60293  432.397898   405.65686  379.530854  354.168824  329.715033  306.306122  284.067881  263.111847  243.531884   225.40096  208.768384  193.657785  180.066112  167.963875  157.296756  147.988545  139.945216  133.059789  127.217517  122.300938  118.194363  114.787514   111.97817  109.673825  107.792462  106.262621  105.022957  104.021456  103.214471          102.565699 |  102.27229 | 1.264854  1.260242  1.255008  1.249152  1.242673  1.235569  1.227839  1.219478  1.210482  1.200849  1.190574  1.179655  1.168088  1.155873  1.143012  1.129508   1.11537  1.100609  1.085243  1.069296  1.052801    1.0358  1.018346  1.000506  0.982358  0.963998  0.945537  0.927102  0.908835  0.890891   0.87343  0.856618  0.840615  0.825565  0.811594  0.798794  0.787226  0.776909  0.767828  0.759933  0.753149  0.747382  0.742526  0.738472  0.735113  0.732348  0.730085  0.728242  0.726747       0.725538 | 6.232586  6.197914  6.158633  6.114776  6.066371  6.013438  5.955997  5.894067  5.827666  5.756815  5.681543  5.601886  5.517889  5.429617  5.337149  5.240593  5.140082  5.035785  4.927913  4.816724  4.702532  4.585714  4.466718  4.346066  4.224366  4.102308  3.980668  3.860299  3.742118  3.627091  3.516194  3.410384  3.310552  3.217471  3.131759  3.053835    2.9839  2.921934  2.867705  2.820803  2.780681  2.746705  2.718192  2.694456  2.674835  2.658717  2.645545   2.63483  2.626147        2.619134 | 897.187983  885.198679  871.620137   856.51498  839.948468  821.988922  802.708144   782.18185  760.490105  737.717738  713.954757  689.296724  663.845085  637.707442  610.997743  583.836352  556.349997  528.671536  500.939521  473.297522  445.893166  418.876855  392.400133  366.613672  341.664874  317.695093   294.83653  273.208876  252.915851  234.041803  216.648639  200.773334  186.426344  173.591166  162.225244  152.262274    143.6158  136.183802  129.853876  124.508486  120.029846  116.304048  113.224219  110.692652  108.621979  106.935533  105.567101  104.460269  103.567502  102.849128   102.27229 |  100.0 | 797.187983 |  102.27229 |
    | 00:00 | 9.881373  9.836196  9.782936  9.721669  9.652461  9.575365  9.490426  9.397686  9.297183   9.18896   9.07306  8.949539  8.818466  8.679929  8.534039  8.380942  8.220819  8.053901  7.880475  7.700894  7.515589  7.325079  7.129988  6.931047  6.729114  6.525179  6.320366  6.115937   5.91328  5.713895  5.519358  5.331283  5.151261  4.980793  4.821214  4.673623  4.538813  4.417233  4.308965  4.213739  4.130975  4.059843  3.999338  3.948352  3.905748   3.87041  3.841288  3.817423   3.79796             3.782153 | 636.428225  631.296671  625.273185  618.379328  610.636963  602.068524  592.697309  582.547773  571.645851  560.019286   547.69798  534.714355  521.103741  506.904758   492.15972  476.915035  461.221589  445.135119  428.716543  412.032229  395.154187  378.160137  361.133426  344.162745  327.341597   310.76746  294.540594  278.762444  263.533613  248.951402   235.10696   222.08216  209.946366  198.753362  188.538741  179.318102   171.08634  163.818235  157.470346  151.984053  147.289399  143.309288  139.963576  137.172696  134.860536  132.956506  131.396797  130.124958  129.091926  128.255698 | 115.770625  115.309908   114.76676  114.141965  113.436177  112.649944  111.783734  110.837966   109.81304  108.709369  107.527419  106.267751  104.931064  103.518253  102.030466   100.46917   98.836231   97.133996    95.36539   93.534015   91.644266   89.701446   87.711891   85.683085   83.623769   81.544026   79.455336   77.370559   75.303862   73.270524   71.286632   69.368636   67.532761   65.794317    64.16693   62.661789   61.286997   60.047115   58.942989   57.971872   57.127845    56.40244   55.785402   55.265447   54.830967    54.47059   54.173607   53.930233   53.731749        53.570543 | 113.813727  113.361958  112.829357  112.216695  111.524613  110.753648  109.904259  108.976857  107.971834  106.889596  105.730599  104.495393  103.184664  101.799289  100.340393   98.809416   97.208188    95.53901   93.804749   92.008938   90.155886   88.250795   86.299876   84.310467   82.291141   80.251786   78.203656   76.159365   74.132802    72.13895   70.193584   68.312833    66.51261   64.807926   63.212141    61.73623   60.388135   59.172331   58.089646   57.137388    56.30975   55.598433   54.993377    54.48352   54.057476   53.704098   53.412882   53.174234   52.979605     52.821529 | 1.883146  1.878055  1.872045  1.865118  1.857277  1.848521  1.838849  1.828257  1.816742  1.804297  1.790919  1.776601   1.76134  1.745131  1.727973  1.709868  1.690819  1.670836  1.649935  1.628137  1.605475  1.581989  1.557736  1.532785  1.507223  1.481157  1.454713  1.428041  1.401315  1.374727  1.348492  1.322839  1.298004   1.27422  1.251709  1.230666  1.211248  1.193569  1.177685  1.163601  1.151271  1.140606  1.131483  1.123758  1.117277  1.111882  1.107423   1.10376  1.100766  1.098331 | 1.338276  1.338319   1.33837   1.33843  1.338499  1.338578  1.338666  1.338765  1.338875  1.338998  1.339133  1.339283  1.339447  1.339628  1.339827  1.340046  1.340286  1.340549  1.340839  1.341156  1.341504  1.341885  1.342303  1.342759  1.343258  1.343801   1.34439  1.345028  1.345714  1.346447  1.347224  1.348041   1.34889  1.349761  1.350642   1.35152  1.352379  1.353206  1.353987  1.354712  1.355372  1.355964  1.356486  1.356939  1.357328  1.357658  1.357935  1.358165  1.358355          1.358511 | 0.765006  0.764188  0.763216  0.762087  0.760798  0.759345  0.757722  0.755924  0.753943  0.751772  0.749402  0.746823  0.744023  0.740993  0.737718  0.734186  0.730382  0.726292    0.7219  0.717192  0.712152  0.706766  0.701022  0.694911  0.688428  0.681571   0.67435   0.66678  0.658888  0.650717  0.642322  0.633776  0.625164  0.616588  0.608157  0.599986  0.592185  0.584854  0.578072    0.5719  0.566369  0.561485  0.557233  0.553577  0.550469  0.547854  0.545674  0.543869  0.542385      0.541172 | -0.466776  -0.465278  -0.463582  -0.461685  -0.459581  -0.457265  -0.454729  -0.451965  -0.448965  -0.445718  -0.442213   -0.43844  -0.434384  -0.430034  -0.425375  -0.420393  -0.415074  -0.409404  -0.403369  -0.396958  -0.390161   -0.38297  -0.375385  -0.367408  -0.359052  -0.350336  -0.341295  -0.331972  -0.322431  -0.312746  -0.303012  -0.293334  -0.283828  -0.274615  -0.265813   -0.25753  -0.249854  -0.242852  -0.236562  -0.230995  -0.226135  -0.221946  -0.218379  -0.215372   -0.21286  -0.210778  -0.209064  -0.207662  -0.206521     -0.205596 | 0.702062  0.701599  0.701095  0.700551  0.699967   0.69934  0.698672   0.69796  0.697205  0.696405   0.69556   0.69467  0.693733  0.692751  0.691722  0.690648  0.689529  0.688368  0.687165  0.685925  0.684651  0.683349  0.682026  0.680688  0.679347  0.678012  0.676698  0.675416  0.674183  0.673012  0.671917  0.670912  0.670005  0.669204   0.66851  0.667923  0.667435  0.667039  0.666724  0.666476  0.666285   0.66614   0.66603  0.665947  0.665886   0.66584  0.665807  0.665782  0.665763      0.665749 |      900.0 | 895.545692  885.895215  874.600785   861.71894  847.308733   831.43212  814.154359  795.544424   775.67542  754.625001  732.475779  709.315717  685.238496  660.343831  634.737734   608.53269  581.847725  554.808343  527.546294  500.199137   472.90957  445.824476  419.093664  392.868254  367.298692   342.53239  318.710989  295.967304  274.422016   254.18026  235.328269  217.930328  202.026287  187.629935  174.728482  163.283334  153.232232   144.49264  136.966151  130.543474  125.109582  120.548529  116.747591  113.600496  111.009653  108.887441  107.156694   105.75055   104.61187          103.692391 | 103.275704 | 1.266429  1.262965  1.258875  1.254161  1.248824  1.242863  1.236278  1.229067  1.221224  1.212748  1.203635  1.193879  1.183478  1.172428  1.160729  1.148379  1.135383  1.121744  1.107472  1.092583  1.077095  1.061037  1.044446  1.027367  1.009859  0.991993  0.973855  0.955547  0.937185  0.918903  0.900847  0.883174  0.866048  0.849631  0.834076  0.819521  0.806078  0.793827  0.782811  0.773036  0.764472   0.75706  0.750716  0.745341   0.74083  0.737073  0.733968  0.731416  0.729331       0.727634 | 6.244442  6.218378  6.187643   6.15228  6.112321  6.067792  6.018716   5.96511  5.906989  5.844371  5.777273  5.705718  5.629735  5.549363  5.464655  5.375681   5.28253   5.18532  5.084198   4.97935  4.871006  4.759446  4.645009    4.5281  4.409195  4.288851  4.167707  4.046488  3.925999  3.807118  3.690781  3.577958  3.469616  3.366687  3.270014  3.180309   3.09811  3.023749  2.957338   2.89877  2.847743   2.80379  2.766331  2.734713  2.708254  2.686281  2.668155  2.653288  2.641155        2.631294 |      900.0  891.184497  880.697261  868.593659  854.931385  839.770832  823.175493   805.21237  785.952381  765.470789  743.847609  721.168016  697.522729  673.008353   647.72768  621.789909  595.310781  568.412585  541.224023  513.879886  486.520511  459.290986  432.340046  405.818655  379.878208  354.668363  330.334486  307.014742  284.836883  263.914844  244.345295  226.204369  209.544799  194.393783  180.751832  168.592854  157.865619  148.496568  140.393819  133.451993  127.557435  122.593317  118.444212  114.999822   112.15771  109.825026  107.919343  106.368767  105.111519  104.095172  103.275704 |  100.0 |      800.0 | 103.275704 |
    | 00:30 |  9.87607  9.845729  9.807196  9.760563  9.705905  9.643286  9.572757  9.494362  9.408139  9.314122  9.212347  9.102852  8.985682  8.860895  8.728565  8.588787  8.441685  8.287417  8.126182  7.958234  7.783885  7.603518  7.417599  7.226688   7.03145  6.832668  6.631249  6.428235  6.224806   6.02227  5.822059  5.625701  5.434787  5.250918  5.075649  4.910414  4.756453  4.614742  4.485934  4.370329  4.267869  4.178157  4.100512  4.034035  3.977682  3.930338  3.890877  3.858217   3.83135             3.809365 | 635.824817  632.377864  628.013435  622.751367  616.611531  609.614087  601.779735  593.129994  583.687479  573.476201  562.521887  550.852303  538.497611  525.490731  511.867716  497.668145  482.935512  467.717618  452.066944  436.040998   419.70262  403.120214  386.367882  369.525434   352.67822  335.916747   319.33602  303.034571  287.113102  271.672723  256.812751    242.6281  229.206323  216.624441  204.945768  194.217022  184.466024  175.700323  167.907012  161.053834  155.091552  149.957322  145.578687  141.877742  138.775027  136.192841  134.057782  132.302461  130.866477   129.69676 | 115.716545  115.407131   115.01417    114.5386  113.981198  113.342604  112.623349  111.823876  110.944568  109.985782  108.947874  107.831235  106.636333  105.363754  104.014247  102.588788  101.088634   99.515398   97.871126   96.158386   94.380365   92.540975   90.644967   88.698048   86.707004   84.679811   82.625732   80.555394    78.48081   76.415344   74.373586   72.371122   70.424165   68.549062   66.761662   65.076591   63.506494   62.061315   60.747727   59.568789   58.523892   57.609006   56.817184   56.139251   55.564561   55.081738   54.679317    54.34625   54.072259        53.848048 | 113.760698  113.457293  113.071962  112.605627   112.05905  111.432858   110.72757  109.943621   109.08139  108.141223   107.12347  106.028515  104.856818   103.60895  102.285651  100.887873   99.416851   97.874166   96.261825   94.582345   92.838852   91.035182   89.175992   87.266882   85.314503   83.326676   81.312486   79.282352   77.248055   75.222699   73.220591   71.257014   69.347866   67.509176   65.756485   64.104138    62.56453   61.147416   59.859337   58.703294   57.678688   56.781568   56.005122   55.340354   54.776825   54.303378   53.908771   53.582173   53.313502     53.093646 | 1.882548   1.87913  1.874784  1.869517  1.863334  1.856236  1.848225  1.839298  1.829453  1.818685   1.80699  1.794363  1.780797  1.766288   1.75083  1.734423  1.717064  1.698756  1.679506  1.659324  1.638231   1.61625  1.593419  1.569784  1.545406  1.520362  1.494747  1.468674  1.442281  1.415725  1.389187   1.36287  1.336993  1.311787  1.287487  1.264322  1.242502  1.222208  1.203579  1.186704   1.17162   1.15831  1.146712  1.136721  1.128206  1.121021  1.115008  1.110016  1.105898  1.102521 | 1.338281   1.33831  1.338347  1.338392  1.338446  1.338508   1.33858  1.338662  1.338754  1.338857  1.338971  1.339098  1.339238  1.339393  1.339564  1.339751  1.339958  1.340184  1.340433  1.340707  1.341007  1.341336  1.341697  1.342092  1.342525  1.342998  1.343513  1.344074  1.344682  1.345338  1.346042  1.346791  1.347583  1.348411  1.349268  1.350141  1.351019  1.351888  1.352732  1.353539  1.354295  1.354992  1.355623  1.356184  1.356677  1.357102  1.357466  1.357773   1.35803          1.358243 | 0.764911  0.764361   0.76366  0.762805  0.761795  0.760626  0.759296  0.757798  0.756128   0.75428  0.752245  0.750016  0.747583  0.744937  0.742065  0.738957  0.735599  0.731978  0.728079  0.723888  0.719389  0.714567  0.709408  0.703899  0.698028  0.691789  0.685178  0.678197  0.670858  0.663182  0.655202  0.646965  0.638534  0.629985  0.621412  0.612919  0.604618   0.59662  0.589032  0.581946  0.575433  0.569542  0.564292  0.559683  0.555688  0.552269  0.549373  0.546944  0.544924      0.543256 | -0.467444  -0.466294   -0.46495  -0.463411  -0.461671  -0.459728  -0.457575  -0.455206  -0.452612  -0.449787   -0.44672  -0.443402   -0.43982  -0.435964   -0.43182  -0.427375  -0.422615  -0.417526  -0.412094  -0.406306  -0.400147  -0.393609  -0.386681   -0.37936  -0.371646  -0.363545  -0.355072  -0.346254  -0.337128  -0.327746  -0.318174  -0.308498  -0.298813   -0.28923  -0.279868  -0.270846  -0.262277  -0.254263  -0.246882  -0.240189  -0.234211  -0.228948  -0.224376  -0.220454  -0.217126  -0.214329     -0.212  -0.210074  -0.208492     -0.207199 | 0.702458  0.702069   0.70164  0.701172  0.700663  0.700113  0.699522  0.698889  0.698213  0.697493  0.696729   0.69592  0.695065  0.694165  0.693218  0.692225  0.691185  0.690101  0.688972  0.687801   0.68659  0.685343  0.684064   0.68276  0.681436  0.680103   0.67877  0.677448  0.676152  0.674894   0.67369  0.672553  0.671498  0.670535  0.669674  0.668919   0.66827  0.667724  0.667275  0.666913  0.666626  0.666401  0.666229  0.666098  0.665999  0.665925   0.66587  0.665829  0.665799      0.665776 | 897.264896 | 894.409558  887.926349  879.734443  869.882983  858.423344  845.409487  830.898325  814.950103  797.628793  779.002493  759.143834  738.130384  716.045037  692.976384  669.019052  644.273985  618.848661  592.857217  566.420453  539.665689  512.726446  485.741908  458.856128  432.216951  405.974601  380.279922  355.282237  331.126833  307.952101   285.88637  265.044559   245.52478  227.405094  210.740681  195.561682  181.872003  169.649301  158.846299  149.393422  141.202592  134.171869  128.190509  123.143973  118.918469  115.404706  112.500711  110.113659  108.160827  106.569807          105.278185 | 104.691406 | 1.266023  1.263697  1.260739  1.257155  1.252946  1.248115  1.242661  1.236584  1.229881  1.222548  1.214583  1.205981  1.196738   1.18685  1.176314  1.165127  1.153288  1.140798  1.127661  1.113884  1.099478   1.08446  1.068853  1.052689  1.036007  1.018859  1.001309  0.983433  0.965323  0.947087  0.928848  0.910745  0.892927  0.875555  0.858791  0.842793   0.82771  0.813668  0.800765  0.789067  0.778603  0.769362  0.761304  0.754358  0.748436  0.743436  0.739251  0.735774  0.732906       0.730553 | 6.241383  6.223878  6.201644  6.174731  6.143179  6.107022  6.066286  6.020991  5.971153  5.916787  5.857906  5.794524   5.72666  5.654337  5.577586  5.496452  5.410993  5.321285  5.227429  5.129554  5.027822  4.922437  4.813648  4.701757  4.587131  4.470201  4.351477  4.231547  4.111083   3.99084  3.871653  3.754421  3.640092  3.529635  3.424003  3.324091  3.230688  3.144438  3.065795  2.995002  2.932081  2.876847  2.828931  2.787821  2.752908   2.72353  2.699011  2.678694  2.661965        2.648265 | 897.264896  891.647486  884.297236  875.262253  864.592703  852.341154  838.562941  823.316534  806.663931  788.671052  769.408146  748.950197  727.377324  704.775162  681.235226  656.855218  631.739291  605.998219  579.749473   553.11716  526.231802   499.22992  472.253377  445.448457  418.964629  392.952972  367.564228  342.946476  319.242434  296.586425  275.101089  254.893953  236.054048   218.64879  202.721393  188.289118  175.342597   163.84646  153.741335  144.947132  137.367384  130.894233  125.413605  120.810097  116.971197  113.790597  111.170503  109.022985  107.270507  105.845821  104.691406 |  100.0 | 797.264896 | 104.691406 |
    | 01:00 | 9.846037  9.830391  9.806448  9.774316  9.734086  9.685832  9.629616  9.565486  9.493484  9.413644  9.325995  9.230568  9.127395  9.016511  8.897965  8.771816  8.638145  8.497057  8.348686  8.193206  8.030837  7.861854  7.686596  7.505481  7.319008  7.127781  6.932509  6.734024  6.533286  6.331389  6.129562  5.929162  5.731655  5.538593  5.351567  5.172159  5.001874  4.842065  4.693865  4.558118  4.435335  4.325667   4.22892  4.144582  4.071883  4.009869  3.957469  3.913569  3.877067             3.846916 | 632.412743  630.638826  627.928835  624.301018  619.773443  614.364221  608.091743  600.974926  593.033463   584.28809  574.760871  564.475492  553.457576  541.735021   529.33834  516.301032  502.659958  488.455719   473.73305   458.54119  442.934248  426.971521  410.717768  394.243398   377.62455  360.943032  344.286057  327.745759  311.418402  295.403258  279.801092  264.712226  250.234182   236.45893  223.469848  211.338552  200.121826  189.858964  180.569837  172.253991   164.89098  158.441989  152.852616  148.056506  143.979421  140.543286  137.669849  135.283657  133.314246   131.69754 | 115.410266  115.250713   115.00654  114.678857  114.268589  113.776496  113.203198  112.549202  111.814921  111.000705  110.106864  109.133697  108.081526  106.950731   105.74179  104.455323  103.092144  101.653316  100.140223   98.554632   96.898787   95.175491    93.38821   91.541183   89.639534    87.68939   85.697999    83.67384   81.626703   79.567751   77.509514   75.465824   73.451643   71.482784   69.575489   67.745883   66.009306   64.379571   62.868222   61.483869   60.231715   59.113322    58.12669   57.266605   56.525225   55.892801   55.358425   54.910731    54.53848        54.230997 | 113.460367  113.303912   113.06448  112.743161   112.34086  111.858323  111.296158  110.654862  109.934841  109.136437  108.259953  107.305684  106.273946   105.16511  103.979646  102.718161  101.381454   99.970568   98.486857   96.932058   95.308368   93.618537   91.865964   90.054805   88.190085   86.277811   84.325092    82.34024   80.332858   78.313889   76.295621   74.291618   72.316552   70.385926    68.51567   66.721594    65.01874   63.420653   61.938653   60.581184   59.353345   58.256671   57.289199   56.445816   55.718834   55.098691   54.574692   54.135692    53.77067     53.469157 | 1.879165  1.877401    1.8747  1.871071  1.866523   1.86106  1.854685  1.847398  1.839198  1.830082  1.820047  1.809087  1.797197  1.784372  1.770605  1.755891  1.740226  1.723609   1.70604  1.687522  1.668065  1.647682  1.626395  1.604233  1.581238  1.557461  1.532969  1.507848  1.482198  1.456143  1.429828   1.40342  1.377107    1.3511   1.32562  1.300901  1.277176  1.254664  1.233565  1.214041  1.196211  1.180144  1.165853  1.153304  1.142416  1.133075  1.125142  1.118468  1.112899  1.108285 | 1.338309  1.338324  1.338347  1.338378  1.338418  1.338466  1.338522  1.338588  1.338663  1.338748  1.338844   1.33895  1.339069  1.339201  1.339346  1.339507  1.339684  1.339879  1.340093  1.340328  1.340587  1.340871  1.341182  1.341523  1.341898  1.342308  1.342756  1.343245  1.343779  1.344358  1.344984  1.345658  1.346379  1.347144   1.34795  1.348788  1.349649  1.350523  1.351395  1.352252   1.35308  1.353864  1.354594  1.355262  1.355862  1.356394  1.356857  1.357256  1.357595          1.357881 | 0.764367  0.764083  0.763646  0.763058  0.762317  0.761422   0.76037  0.759158  0.757781  0.756236  0.754515  0.752612  0.750519  0.748228  0.745729  0.743011  0.740064  0.736874  0.733429  0.729715  0.725717   0.72142   0.71681  0.711871   0.70659  0.700956  0.694957  0.688589   0.68185  0.674748  0.667296  0.659521  0.651462  0.643172  0.634719  0.626186  0.617672  0.609282  0.601129  0.593323  0.585964  0.579135  0.572898  0.567289  0.562321   0.55798  0.554236  0.551043   0.54835      0.546097 | -0.467511  -0.466703  -0.465703   -0.46451  -0.463122  -0.461537   -0.45975  -0.457755  -0.455548   -0.45312  -0.450465  -0.447573  -0.444435  -0.441041  -0.437377  -0.433434  -0.429197  -0.424653  -0.419788  -0.414588   -0.40904  -0.403129  -0.396845  -0.390177  -0.383119  -0.375667  -0.367825  -0.359602  -0.351019  -0.342106  -0.332906  -0.323477  -0.313892  -0.304241  -0.294626  -0.285161  -0.275963  -0.267151  -0.258833    -0.2511  -0.244022  -0.237642  -0.231975  -0.227012  -0.222721  -0.219054  -0.215955  -0.213358  -0.211201     -0.209421 | 0.702716    0.7024  0.702045  0.701651  0.701217  0.700743  0.700228  0.699672  0.699073  0.698432  0.697748  0.697019  0.696246  0.695427  0.694562  0.693651  0.692693  0.691689  0.690639  0.689544  0.688406  0.687226  0.686007  0.684754  0.683471  0.682166  0.680844  0.679517  0.678194  0.676888  0.675612  0.674381  0.673208  0.672108  0.671094  0.670176   0.66936   0.66865  0.668046  0.667541  0.667129  0.666798  0.666537  0.666334  0.666179  0.666061  0.665972  0.665906  0.665856      0.665819 | 889.260928 | 887.991891  884.659952  879.575838  872.781306  864.320124  854.238381   842.58482  829.411179  814.772556  798.727778  781.339791   762.67605  742.808912   721.81603   699.78072  676.792311   652.94645  628.345352  603.097965   577.32004   551.13407  524.669066  498.060143  471.447873  444.977376  418.797102  393.057275  367.907981  343.496877  319.966538  297.451469   276.07487  255.945266   237.15316  219.767955  203.835375  189.375691  176.382976  164.825625  154.648185  145.774459   138.11164  131.555118  125.993514  121.313497  117.403988  114.159523   111.48265   109.28539          107.489875 | 106.671706 |  1.26372   1.26252  1.260681  1.258212  1.255117  1.251399  1.247059  1.242099  1.236516  1.230309  1.223475  1.216011  1.207912  1.199174  1.189792  1.179763  1.169084  1.157752  1.145768  1.133133  1.119852  1.105934  1.091392  1.076247  1.060523  1.044257  1.027493  1.010287  0.992707  0.974837  0.956774  0.938632  0.920541  0.902642  0.885091  0.868047  0.851672  0.836119  0.821527  0.808012  0.795659  0.784517    0.7746  0.765884  0.758318  0.751822  0.746304  0.741659  0.737782       0.734569 | 6.224056  6.215028  6.201212  6.182669  6.159448   6.13159  6.099128  6.062086  6.020483  5.974336  5.923654   5.86845  5.808734   5.74452  5.675827  5.602678   5.52511   5.44317  5.356924  5.266456  5.171878  5.073333  4.970997  4.865092  4.755888  4.643714   4.52896  4.412089   4.29364  4.174235  4.054578  3.935454  3.817722    3.7023  3.590143  3.482212  3.379435  3.282664   3.19263  3.109897  3.034835  2.967596  2.908117   2.85614  2.811236  2.772856  2.740371  2.713114  2.690423        2.671659 | 889.260928  886.814967  882.596333  876.645852  869.006203  859.722228   848.84125  836.413413  822.492033  807.133966  790.399988  772.355188  753.069357  732.617391  711.079665  688.542402  665.098005  640.845331  615.889917  590.344095   564.32701  537.964485  511.388712  484.737735  458.154683  431.786714  405.783647  380.296225  355.474026  331.462981  308.402541  286.422546   265.63987  246.155016  228.048823  211.379558  196.180648  182.459351   170.19659  159.348123  149.847052  141.607533  134.529376  128.503107  123.415021  119.151786  115.604288   112.67051  110.257443  108.282085  106.671706 |  100.0 | 789.260928 | 106.671706 |
    | 01:30 | 9.792548  9.791394  9.781841  9.764017  9.738027  9.703961  9.661889  9.611871  9.553953  9.488173  9.414561  9.333143  9.243942  9.146984  9.042299  8.929922  8.809903  8.682308  8.547223  8.404762  8.255075  8.098353  7.934835  7.764821   7.58868  7.406862  7.219907  7.028459  6.833276  6.635238  6.435357   6.23478  6.034784  5.836767  5.642225  5.452725  5.269849  5.095147  4.930059  4.775846  4.633524  4.503798  4.387032  4.283236  4.192085  4.112963  4.045027  3.987276  3.938629              3.89798 |  626.35823  626.227895  625.149687  623.140349  620.216264  616.393665  611.688851  606.118405   599.69942  592.449737  584.388191  575.534882  565.911453  555.541386  544.450321  532.666385  520.220548  507.146978  493.483422   479.27158  464.557487  449.391875  433.830512  417.934504  401.770523  385.410954  368.933911  352.423092  335.967429  319.660476  303.599489  287.884146  272.614873  257.890731  243.806918  230.451895    217.9043  206.229816  195.478266  185.681237  176.850568  168.977934  162.035714  155.979082  150.749147  146.276765  142.486601  139.301012  136.643414  134.440916 | 114.864789  114.853018  114.755597  114.573822   114.30878  113.961369  113.532323  113.022237  112.431589  111.760759  111.010059  110.179751  109.270082  108.281304  107.213716  106.067695  104.843739  103.542516  102.164911  100.712094   99.185582   97.587317   95.919752   94.185944   92.389655   90.535468   88.628896   86.676503   84.686014   82.666415   80.628026   78.582535   76.542966   74.523575   72.539635   70.607098   68.742129   66.960507    65.27693   63.704268   62.252856   60.929905   59.739121   58.680606   57.751044   56.944157   56.251339   55.662401    55.16629        54.751748 | 112.925483   112.91394  112.818411  112.640166  112.380271  112.039606  111.618892  111.118712  110.539533  109.881731  109.145608  108.331425   107.43942  106.469844  105.422988  104.299222  103.099035   101.82308  100.472227   99.047623   97.550755   95.983527   94.348345   92.648206     90.8868    89.06862   87.199072   85.284593   83.332758   81.352379   79.353574   77.347805   75.347843   73.367667   71.422254   69.527246   67.698493    65.95147   64.300586   62.758464   61.335237   60.037977   58.870317   57.832358   56.920848   56.129629   55.450266   54.872764   54.386288     53.979795 |  1.87313     1.873  1.871921  1.869907  1.866969  1.863114  1.858346   1.85267  1.846086  1.838592  1.830187  1.820866  1.810625  1.799458  1.787359  1.774322   1.76034   1.74541  1.729527  1.712692  1.694904  1.676171  1.656502  1.635915  1.614435  1.592095   1.56894  1.545031  1.520439  1.495256  1.469593  1.443581  1.417374  1.391147  1.365096  1.339437  1.314396  1.290204   1.26709  1.245264   1.22491  1.206173  1.189151   1.17389  1.160383  1.148577  1.138377  1.129659  1.122281  1.116092 | 1.338361  1.338362  1.338371  1.338388  1.338414  1.338448   1.33849   1.33854    1.3386  1.338669  1.338747  1.338836  1.338935  1.339046   1.33917  1.339307  1.339458  1.339625  1.339809  1.340011  1.340233  1.340478  1.340746  1.341041  1.341364  1.341718  1.342107  1.342532  1.342996  1.343503  1.344054  1.344651  1.345296  1.345988  1.346726  1.347506  1.348323  1.349169  1.350034  1.350905  1.351769  1.352612  1.353419  1.354179  1.354882   1.35552   1.35609  1.356592  1.357027            1.3574 | 0.763392  0.763371  0.763196  0.762869   0.76239  0.761759  0.760975  0.760036  0.758938  0.757679  0.756254  0.754656   0.75288  0.750919  0.748765  0.746408  0.743838  0.741045  0.738018  0.734742  0.731206  0.727394  0.723293  0.718887  0.714162  0.709104  0.703699  0.697936  0.691808  0.685311  0.678447  0.671226  0.663668  0.655802  0.647675  0.639345  0.630886  0.622388  0.613951  0.605685  0.597701  0.590103  0.582986  0.576424  0.570468  0.565144  0.560452  0.556374  0.552872      0.549897 | -0.467001   -0.46653  -0.465867  -0.465013  -0.463967  -0.462727  -0.461291  -0.459655  -0.457815  -0.455764  -0.453498  -0.451008  -0.448286  -0.445323  -0.442109  -0.438634  -0.434884  -0.430849  -0.426515  -0.421867  -0.416893  -0.411579   -0.40591  -0.399875  -0.393462  -0.386663  -0.379473  -0.371891  -0.363924  -0.355584  -0.346897  -0.337898  -0.328635  -0.319173  -0.309592  -0.299986  -0.290462  -0.281135  -0.272123   -0.26354  -0.255487  -0.248047  -0.241278  -0.235212  -0.229854  -0.225184  -0.221165  -0.217744  -0.214862     -0.212454 |  0.70284  0.702597  0.702316  0.701995  0.701635  0.701234  0.700794  0.700313  0.699791  0.699227   0.69862   0.69797  0.697277  0.696539  0.695755  0.694926  0.694051  0.693129   0.69216  0.691145  0.690084  0.688979  0.687831  0.686642  0.685417  0.684159  0.682874  0.681568  0.680251  0.678931  0.677621  0.676333   0.67508  0.673877  0.672739  0.671678  0.670706  0.669833  0.669064  0.668399  0.667837  0.667372  0.666994  0.666693  0.666457  0.666274  0.666134  0.666028  0.665948      0.665888 | 876.351729 | 876.632637  876.388511  874.369597  870.610273  865.146762  858.017406  849.262953  838.926863  827.055634  813.699137  798.910976  782.748853  765.274949  746.556309  726.665221  705.679597  683.683319  660.766568  637.026091  612.565417  587.494972  561.932086  536.000868  509.831889  483.561679  457.331963  431.288628  405.580367  380.356977  355.767299  331.956774  309.064667  287.220975  266.543138  247.132674  229.071944  212.421261   197.21663  183.468374  171.160892  160.253698  150.683757     142.369  135.212729  129.108514  123.945122  119.611079  115.998511  113.006098          110.541067 | 109.413639 | 1.259614  1.259525  1.258791   1.25742   1.25542  1.252796  1.249551  1.245688  1.241205  1.236104  1.230381  1.224033  1.217059  1.209452  1.201209  1.192325  1.182796  1.172618  1.161789  1.150306   1.13817  1.125385  1.111957  1.097896  1.083219  1.067948  1.052112   1.03575  1.018912  1.001658  0.984063  0.966216   0.94822  0.930196  0.912277  0.894611  0.877354  0.860666  0.844705   0.82962  0.815538  0.802563  0.790765  0.780178  0.770801    0.7626  0.755509  0.749447  0.744313       0.740005 | 6.193191  6.192525  6.187011  6.176724  6.161723  6.142056  6.117765  6.088879  6.055423  6.017414  5.974866  5.927788  5.876188  5.820075  5.759458  5.694349  5.624769  5.550744  5.472313   5.38953  5.302467  5.211218  5.115908   5.01669  4.913761  4.807359  4.697779  4.585373  4.470559  4.353831  4.235759  4.116997  3.998279   3.88042    3.7643  3.650849  3.541025  3.435772  3.335987  3.242471  3.155886  3.076716  3.005243  2.941529  2.885431  2.836621  2.794623  2.758856  2.728678        2.703426 | 876.351729  877.003313  875.863265  872.965108   868.34407  862.037358  854.084437  844.527324  833.410909  820.783278  806.696069  791.204828  774.369388  756.254253  736.928977  716.468554  694.953778  672.471588  649.115369  624.985202  600.188028   574.83773  549.055068   522.96748  496.708675  470.418014  444.239624  418.321211  392.812544  367.863578  343.622202  320.231624  297.827415  276.534289   256.46273  237.705627   220.33513  204.399995   189.92368  176.903478  165.310872  155.093245  146.176866  138.470954  131.872466  126.271146  121.554388  117.611511  114.337198  111.633956  109.413639 |  100.0 | 776.351729 | 109.413639 |
    | 02:00 | 9.717003  9.730077  9.734656  9.730887  9.718892  9.698775  9.670619  9.634495  9.590457  9.538548  9.478803  9.411248  9.335904  9.252789  9.161923  9.063326  8.957027  8.843064  8.721488  8.592371  8.455807  8.311924  8.160884  8.002897  7.838224  7.667194  7.490205  7.307742  7.120385  6.928824  6.733862  6.536432  6.337597  6.138553  5.940622   5.74524  5.553927  5.368251  5.189779  5.020012  4.860315  4.711846   4.57549  4.451812   4.34103  4.243015  4.157323  4.083245  4.019877             3.966188 | 617.855812  619.323171  619.837509  619.414091  618.067687  615.812765   612.66369  608.634916  603.741188  597.997759  591.420601  584.026646  575.834028  566.862355   557.13298   546.66931  535.497117  523.644871  511.144096  498.029726   484.34048  470.119234  455.413386  440.275213  424.762181  408.937219  392.868908  376.631571  360.305224  343.975343  327.732411   311.67118  295.889609  280.487426  265.564285  251.217499  237.539395  224.614363  212.515749  201.302814  191.018048  181.685135  173.307886  165.870346  159.338161  153.661106    148.7765  144.613126  141.095209  138.146078 | 114.094379  114.227707  114.274405  114.235963   114.11364  113.908482  113.621354  113.252958  112.803855  112.274489  111.665205  110.976273  110.207909  109.360303  108.433645  107.428155  106.344116  105.181912  103.942076  102.625331  101.232651   99.765322   98.225012   96.613851   94.934519   93.190343     91.3854   89.524635   87.613968   85.660414   83.672187   81.658787   79.631058   77.601199   75.582696    73.59018   71.639159   69.745634   67.925574   66.194283   64.565684   63.051586   61.661024   60.399755   59.269994   58.270432   57.396539   56.641092   55.994862        55.447344 | 112.170033  112.300772  112.346563  112.308868   112.18892  111.987746  111.706194  111.344952  110.904571  110.385484  109.788032  109.112478  108.359035   107.52789  106.619227  105.633263  104.570275  103.430641  102.214881  100.923706   99.558072   98.119237   96.608839   95.028965   93.382245   91.671939   89.902048   88.077417   86.203854   84.288237   82.338619   80.364318   78.375967   76.385526   74.406221   72.452398   70.539265   68.682511   66.897795   65.200125   63.603152   62.118457   60.754899   59.518123   58.410301    57.43015   56.573227    55.83245    55.19877     54.661884 |  1.86459   1.86607  1.866588  1.866161  1.864804  1.862526  1.859336  1.855239  1.850237  1.844332  1.837523  1.829808  1.821183  1.811642  1.801181  1.789793  1.777471  1.764209  1.750002  1.734844  1.718734  1.701671  1.683659  1.664704  1.644821  1.624029  1.602356  1.579842  1.556537  1.532505  1.507827  1.482602  1.456948  1.431006  1.404938  1.378926  1.353176  1.327905  1.303342  1.279716  1.257248  1.236137   1.21655  1.198614  1.182403  1.167942  1.155206  1.144122  1.134585  1.126465 | 1.338435  1.338422  1.338417  1.338421  1.338433  1.338453  1.338481  1.338517  1.338562  1.338616  1.338678  1.338751  1.338833  1.338925  1.339029  1.339145  1.339273  1.339416  1.339573  1.339746  1.339938  1.340148  1.340379  1.340633  1.340912  1.341218  1.341553  1.341921  1.342324  1.342765  1.343246   1.34377  1.344339  1.344955  1.345618  1.346327  1.347081  1.347875  1.348702  1.349554   1.35042  1.351286  1.352139  1.352966  1.353751  1.354485  1.355159  1.355767  1.356307          1.356779 | 0.762001  0.762243  0.762327  0.762258  0.762036  0.761662  0.761138  0.760461  0.759631  0.758645  0.757499  0.756189  0.754711  0.753058  0.751224  0.749201   0.74698  0.744554   0.74191  0.739038  0.735926  0.732561  0.728929  0.725016  0.720808  0.716289  0.711446  0.706265  0.700733  0.694841  0.688583  0.681958  0.674971  0.667636  0.659977   0.65203  0.643846  0.635491  0.627044  0.618599  0.610261  0.602139  0.594341  0.586968  0.580106  0.573819  0.568147  0.563105  0.558687      0.554864 | -0.465934  -0.465799  -0.465469  -0.464947  -0.464235  -0.463331  -0.462234  -0.460943  -0.459454  -0.457763  -0.455865  -0.453754  -0.451424  -0.448867  -0.446074  -0.443035  -0.439741   -0.43618   -0.43234  -0.428209  -0.423772  -0.419018  -0.413931  -0.408497  -0.402706  -0.396543  -0.389999  -0.383068  -0.375746  -0.368036  -0.359946  -0.351494   -0.34271  -0.333634  -0.324322  -0.314843  -0.305283  -0.295742   -0.28633  -0.277163  -0.268356   -0.26002  -0.252246  -0.245108  -0.238653    -0.2329  -0.227846  -0.223462  -0.219705     -0.216519 | 0.702835  0.702666  0.702456  0.702208   0.70192  0.701592  0.701225  0.700818   0.70037  0.699881  0.699351  0.698778  0.698163  0.697504    0.6968  0.696052  0.695258  0.694418  0.693532  0.692599  0.691619  0.690593  0.689521  0.688406  0.687249  0.686052  0.684821  0.683558  0.682272  0.680968  0.679656  0.678347  0.677052  0.675784  0.674557  0.673385  0.672283  0.671263  0.670336  0.669508  0.668785  0.668165  0.667645  0.667217  0.666872  0.666598  0.666384  0.666219  0.666093      0.665998 | 858.969595 | 860.742648  863.479776  864.439706  863.649445  861.137735  856.935289  851.075036   843.59239  834.525527  823.915692  811.807514  798.249345  783.293608  766.997169  749.421704  730.634078  710.706714  689.717958  667.752403  644.901195  621.262264  596.940497  572.047804  546.703062   521.03191  495.166352  469.244144   443.40792  417.804026  392.581025  367.887851  343.871594  320.674931  298.433216  277.271317  257.300298  238.614104  221.286461  205.368235  190.885527  177.838742  166.202858  155.928985   146.94717   139.17027  132.498548  126.824575  122.037996  118.029777          114.695666 |  113.16387 | 1.253801  1.254808  1.255161  1.254871  1.253947  1.252397  1.250225  1.247436  1.244031  1.240012  1.235376  1.230123  1.224249  1.217751  1.210626  1.202867  1.194471  1.185433  1.175749  1.165415  1.154428  1.142788  1.130496  1.117557  1.103979  1.089775  1.074963  1.059569  1.043625  1.027175  1.010273  0.992984  0.975389  0.957583  0.939675  0.921792  0.904072  0.886666   0.86973  0.853426  0.837904  0.823307   0.80975  0.797324  0.786084   0.77605  0.767205  0.759504  0.752873       0.747224 | 6.149586  6.157133  6.159777  6.157601  6.150677  6.139062  6.122806  6.101946   6.07651  6.046523  6.011999   5.97295  5.929385  5.881307  5.828722  5.771636  5.710057  5.643998   5.57348  5.498533    5.4192  5.335542  5.247637  5.155593  5.059543  4.959659  4.856152  4.749284  4.639372  4.526792  4.411993  4.295498  4.177911  4.059914  3.942275   3.82583   3.71148  3.600162  3.492827  3.390398  3.293731  3.203569  3.120499  3.044921  2.977026  2.916792     2.864   2.81826  2.779055         2.74578 | 858.969595  862.602048  864.444126  864.522021  862.863549  859.498374  854.458248  847.777269  839.492152  829.642523  818.271228  805.424661  791.153108   775.51111   758.55783   740.35743  720.979447   700.49916  678.997944  656.563582  633.290544  609.280192  584.640906  559.488099  533.944095  508.137839    482.2044  456.284243  430.522216  405.066234  380.065612  355.669044  332.022203  309.264992  287.528484  266.931644  247.577959  229.552159  212.917269  197.712251  183.950505  171.619495  160.681639  151.076529  142.724352  135.530239  129.389155  124.190858  119.824507  116.182576   113.16387 |  100.0 | 758.969595 |  113.16387 |
    | 02:30 |   9.6209  9.647882   9.66628  9.676258  9.677958  9.671498  9.656975   9.63447  9.604048   9.56576  9.519647  9.465737  9.404054  9.334613  9.257429  9.172512  9.079879  8.979547  8.871545  8.755912  8.632704     8.502  8.363905  8.218557  8.066137  7.906873  7.741053   7.56903  7.391238  7.208198  7.020533  6.828975   6.63438   6.43773  6.240138  6.042849  5.847227  5.654736  5.466914  5.285323  5.111498  4.946881  4.792747  4.650141  4.519813  4.402183  4.297323  4.204975  4.124585             4.055367 | 607.122099  610.126405  612.179035  613.293761  613.483763  612.761807  611.140426  608.632095  605.249411  601.005276  595.913095  589.986977  583.241955  575.694215  567.361347  558.262604  548.419192  537.854566  526.594747  514.668665    502.1085  488.950049   475.23309  461.001752  446.304868  431.196307  415.735271  399.986525  384.020546  367.913553  351.747379  335.609151  319.590718  303.787785  288.298702  273.222862  258.658696  244.701256  231.439449  218.953031  207.309531   196.56135  186.743344  177.871181  169.940753  162.928804  156.794774  151.483692  146.929806   143.06052 | 113.114315  113.389481    113.5771  113.678861  113.696197  113.630311  113.482205    113.2527  112.942456  112.551999  112.081731  111.531959  110.902908   110.19475  109.407618  108.541639   107.59696  106.573775  105.472364  104.293133  103.036657  101.703731  100.295428   98.813165   97.258778     95.6346   93.943557   92.189263   90.376131   88.509483   86.595665   84.642157    82.65767   80.652221   78.637174   76.625211   74.630245   72.667221   70.751802   68.899929   67.127256   65.448482   63.876624   62.422321   61.093231   59.893631   58.824268   57.882498   57.062683        56.356792 | 111.209001  111.478824  111.662799  111.762584  111.779583  111.714977  111.569747  111.344699   111.04048  110.657605   110.19647  109.657373  109.040539  108.346132  107.574286  106.725124  105.798789  104.795474  103.715452  102.559121  101.327045  100.020003   98.639048    97.18557   95.661368   94.068731   92.410527     90.6903   88.912378   87.081979   85.205326   83.289753   81.343803     79.3773   77.401383   75.428491   73.472266   71.547363   69.669139   67.853229    66.11498   64.468806   62.927473   61.501411   60.198132   59.021827   57.973231   57.049749   56.245854     55.553671 | 1.853695  1.856758  1.858844  1.859975  1.860168  1.859436  1.857789  1.855236  1.851781  1.847429   1.84218  1.836033  1.828986  1.821035  1.812176  1.802402  1.791707  1.780085  1.767528  1.754031  1.739587  1.724193  1.707846  1.690549  1.672306  1.653126  1.633026  1.612028  1.590167  1.567484  1.544036  1.519895  1.495147  1.469899   1.44428  1.418436  1.392539   1.36678  1.341368  1.316525   1.29248  1.269457  1.247667  1.227297  1.208496  1.191369  1.175969  1.162299  1.150316  1.139933 | 1.338531  1.338504  1.338485  1.338475  1.338474   1.33848  1.338495  1.338517  1.338548  1.338588  1.338635  1.338692  1.338758  1.338834   1.33892  1.339017  1.339125  1.339246   1.33938  1.339528  1.339692  1.339872  1.340071  1.340289   1.34053  1.340793  1.341083  1.341401   1.34175  1.342132   1.34255  1.343007  1.343505  1.344047  1.344635  1.345269   1.34595  1.346676  1.347445  1.348251  1.349087  1.349943  1.350806  1.351665  1.352505  1.353312  1.354074   1.35478  1.355424          1.356002 | 0.760206  0.760713  0.761057  0.761243  0.761275  0.761154  0.760883  0.760461  0.759888  0.759163  0.758283  0.757246  0.756049  0.754685  0.753151  0.751439  0.749543  0.747454  0.745165  0.742664  0.739942  0.736987  0.733787  0.730328  0.726597  0.722579   0.71826  0.713625   0.70866  0.703353  0.697693   0.69167  0.685283  0.678531  0.671424   0.66398  0.656228   0.64821  0.639983  0.631618  0.623201   0.61483   0.60661  0.598651  0.591059  0.583925  0.577328  0.571322  0.565935      0.561174 | -0.464331  -0.464531  -0.464532  -0.464339  -0.463953  -0.463377   -0.46261  -0.461652  -0.460501  -0.459154  -0.457608  -0.455858  -0.453898  -0.451723  -0.449325  -0.446696  -0.443827  -0.440709   -0.43733  -0.433679  -0.429744  -0.425512   -0.42097  -0.416103  -0.410899  -0.405344  -0.399425  -0.393133  -0.386457  -0.379393  -0.371939  -0.364102  -0.355894  -0.347336  -0.338464  -0.329323  -0.319974  -0.310495  -0.300976  -0.291521  -0.282242  -0.273255  -0.264673  -0.256599  -0.249116  -0.242287  -0.236147  -0.230706  -0.225949     -0.221843 | 0.702707  0.702609  0.702472  0.702294  0.702078  0.701822  0.701526  0.701191  0.700816  0.700401  0.699945  0.699447  0.698908  0.698326  0.697701  0.697032  0.696319  0.695559  0.694755  0.693903  0.693005  0.692061  0.691069  0.690032   0.68895  0.687825  0.686659  0.685455  0.684218  0.682953  0.681667  0.680367  0.679063  0.677766  0.676488  0.675243  0.674044  0.672906  0.671842  0.670865  0.669982  0.669201  0.668523  0.667947  0.667467  0.667075   0.66676  0.666512  0.666319       0.66617 |   837.5984 | 840.786902   846.36061  850.174014   852.24676  852.600178  851.257475  848.243937  843.587158  837.317279  829.467248  820.073097  809.174248  796.813824  783.038991  767.901302  751.457063  733.767698  714.900119  694.927086  673.927548  651.986955  629.197527  605.658459  581.476039  556.763664  531.641717  506.237282  480.683656  455.119626  429.688478  404.536696  379.812328  355.662992   332.23352  309.663256  288.083049  267.612027  248.354288  230.395668  213.800831  198.610916  184.842027  172.484778  161.505081  151.846207  143.432015  136.171114  129.961575  124.695761          120.264864 | 118.218296 | 1.246386   1.24847   1.24989   1.25066  1.250791  1.250293  1.249172  1.247434  1.245083   1.24212  1.238546  1.234361  1.229563  1.224148  1.218115  1.211457  1.204172  1.196253  1.187696  1.178496  1.168648  1.158151  1.147001  1.135198  1.122747  1.109651  1.095922  1.081575   1.06663  1.051115   1.03507  1.018539  1.001583  0.984273  0.966695   0.94895  0.931153  0.913436  0.895941  0.878822  0.862236   0.84634  0.831282   0.81719  0.804172  0.792302   0.78162  0.772132  0.763808       0.756592 | 6.094094  6.109677    6.1203  6.126062  6.127044  6.123313  6.114927  6.101931  6.084361  6.062244  6.035602  6.004448  5.968791  5.928638  5.883991  5.834852  5.781222  5.723106  5.660511  5.593452  5.521951  5.446042  5.365774  5.281213  5.192449  5.099598  5.002809  4.902269  4.798208   4.69091  4.580715  4.468027  4.353325  4.237162  4.120173  4.003072  3.886654  3.771777  3.659356  3.550331  3.445637  3.346169  3.252734   3.16601  3.086504  3.014528  2.950186  2.893372  2.843798        2.801021 |   837.5984  844.057384  848.746556  851.684769  852.892462  852.391851  850.207126  846.364667  840.893272  833.824412  825.192498  815.035173  803.393617  790.312878  775.842215  760.035453  742.951351  724.653971  705.213044  684.704325  663.209924  640.818598  617.625996  593.734826   569.25493  544.303236  519.003568  493.486267  467.887597  442.348901  417.015461  392.035048  367.556106  343.725584  320.686397  298.574558  277.516032  257.623422  238.992643  221.699772  205.798339  191.317313   178.26005  166.604425   156.30426   147.29202  139.482612  132.777941  127.071818  122.254762  118.218296 |  100.0 |   737.5984 | 118.218296 |
    | 03:00 | 9.505814   9.54633   9.57818  9.601548  9.616593  9.623447  9.622222  9.613012  9.595892  9.570921  9.538147  9.497606  9.449323  9.393317  9.329598  9.258174  9.179054  9.092242  8.997752    8.8956  8.785815   8.66844  8.543536   8.41119  8.271519  8.124676  7.970858  7.810316  7.643364  7.470384  7.291846  7.108309  6.920439  6.729017  6.534946  6.339259  6.143119  5.947816   5.75475  5.565411  5.381341  5.204087  5.035139  4.875867  4.727448  4.590798  4.466531  4.354918  4.255893             4.169077 | 594.389717  598.857028  602.380334  604.971855  606.643163  607.405341  607.269152  606.245189  604.344042  601.576452  597.953489  593.486723  588.188416  582.071725  575.150912  567.441581  558.960925  549.727991  539.763967  529.092485  517.739938  505.735822  493.113084   479.90848  466.162938  451.921917  437.235748  422.159949  406.755487  391.088974  375.232763  359.264912  343.268978  327.333589  311.551758  296.019874  280.836338  266.099799   251.90699  238.350171  225.514279  213.473889  202.290218  192.008426  182.655516  174.239124  166.747431  160.150299  154.401547  149.442154 | 111.940663  112.353846  112.678651  112.916963  113.070387  113.140286  113.127799  113.033873  112.859276  112.604625  112.270399   111.85696  111.364569  110.793409    110.1436  109.415224  108.608346  107.723042  106.759424  105.717675  104.598083  103.401086  102.127314  100.777648   99.353276    97.85576    96.28712   94.649911    92.94732   91.183271   89.362528   87.490812   85.574909   83.622776   81.643631   79.648006   77.647767   75.656057   73.687165    71.75628   69.879127   68.071479   66.348545   64.724286   63.210695   61.817141   60.549852   59.411619   58.401767        57.516407 | 110.058141    110.4633  110.781798  111.015481  111.165927  111.234468  111.222224  111.130121  110.958915  110.709209  110.381474  109.976063  109.493234  108.933165  108.295975  107.581744  106.790536  105.922424  104.977518  103.955999  102.858149  101.684396  100.435361   99.111904   97.715192   96.246758   94.708579   93.103163   91.433636   89.703843   87.918458   86.083089   84.204392   82.290168   80.349456   78.392586    76.43119   74.478158     72.5475   70.654111   68.813411   67.040866   65.351391   63.758674   62.274476   60.907984   59.665305   58.549176   57.558934     56.690767 | 1.840603  1.845218  1.848841  1.851497  1.853206  1.853985  1.853846    1.8528  1.850855  1.848016  1.844287  1.839668  1.834159  1.827757   1.82046  1.812261  1.803156  1.793137  1.782197  1.770329  1.757527  1.743783  1.729093  1.713453  1.696863  1.679325  1.660846  1.641438   1.62112  1.599919  1.577872  1.555028  1.531448  1.507211  1.482411  1.457164  1.431604   1.40589  1.380199   1.35473  1.329696  1.305321  1.281831  1.259446  1.238365  1.218759  1.200757  1.184443  1.169849  1.156958 |  1.33865  1.338608  1.338575  1.338551  1.338535  1.338529   1.33853  1.338539  1.338557  1.338582  1.338616  1.338659   1.33871   1.33877   1.33884  1.338919  1.339009   1.33911  1.339224  1.339349  1.339489  1.339643  1.339814  1.340002  1.340208  1.340436  1.340686   1.34096  1.341262  1.341592  1.341954  1.342351  1.342785  1.343258  1.343774  1.344334   1.34494  1.345593  1.346292  1.347034  1.347817  1.348634  1.349476  1.350333  1.351192  1.352041  1.352864   1.35365  1.354387          1.355065 | 0.758018  0.758793  0.759398  0.759841  0.760125  0.760254  0.760231  0.760057  0.759734  0.759261  0.758637   0.75786  0.756929  0.755839  0.754586  0.753166  0.751572  0.749798  0.747836  0.745678  0.743316  0.740738  0.737934  0.734892  0.731599  0.728042  0.724208  0.720081  0.715647  0.710893  0.705804  0.700369  0.694578  0.688424  0.681907  0.675031  0.667808  0.660262  0.652426  0.644349  0.636094  0.627737  0.619368   0.61109   0.60301  0.595233  0.587861   0.58098  0.574657      0.568935 | -0.462209  -0.462745  -0.463077   -0.46321  -0.463148  -0.462894  -0.462449  -0.461815  -0.460991  -0.459975  -0.458765  -0.457359  -0.455751  -0.453938  -0.451912  -0.449668  -0.447198  -0.444493  -0.441545  -0.438342  -0.434874  -0.431129  -0.427094  -0.422758  -0.418105  -0.413123  -0.407798  -0.402118   -0.39607  -0.389645  -0.382835  -0.375638  -0.368054  -0.360093  -0.351771  -0.343114  -0.334163  -0.324969    -0.3156  -0.306138   -0.29668  -0.287331  -0.278207  -0.269421  -0.261081  -0.253283  -0.246101  -0.239586  -0.233762     -0.228629 |  0.70246  0.702434  0.702368  0.702261  0.702114  0.701929  0.701703  0.701439  0.701135  0.700791  0.700408  0.699984  0.699519  0.699012  0.698463  0.697871  0.697236  0.696556  0.695831  0.695061  0.694244  0.693381  0.692472  0.691515  0.690512  0.689464  0.688372  0.687237  0.686063  0.684853  0.683612  0.682345   0.68106  0.679765  0.678471  0.677188   0.67593   0.67471  0.673542   0.67244  0.671417  0.670483  0.669647  0.668912  0.668279  0.667745  0.667303  0.666945  0.666659      0.666434 | 812.756821 | 817.267945  825.500862  832.008558  836.803345  839.899215  841.311994  841.059509  839.161764  835.641143   830.52263  823.834044  815.606302  805.873695  794.674195  782.049766  768.046705  752.715988  736.113632  718.301055  699.345434  679.320059  658.304651  636.385655  613.656475  590.217641  566.176882  541.649079  516.756075  491.626304  466.394203  441.199385  416.185515   391.49888  367.286612  343.694546  320.864733  298.932614  278.023937  258.251496   239.71186  222.482273  206.617972  192.150172  179.084986  167.403468  157.062916  147.999401  140.131377   133.36405          127.594129 | 124.912751 | 1.237473  1.240615  1.243081  1.244889  1.246053  1.246582  1.246488  1.245776  1.244452  1.242519   1.23998  1.236836  1.233085  1.228726  1.223757  1.218173  1.211971  1.205146  1.197692  1.189605  1.180879  1.171509  1.161492  1.150825  1.139507  1.127538  1.114923  1.101669  1.087788  1.073297  1.058222  1.042593  1.026452   1.00985  0.992853  0.975537  0.957993   0.94033  0.922667  0.905142    0.8879  0.871096  0.854886  0.839424  0.824848  0.811279   0.79881  0.787499  0.777373       0.768423 | 6.027609  6.051019  6.069419  6.082917  6.091606  6.095565  6.094858  6.089538   6.07965  6.065225  6.046291  6.022866  5.994961  5.962584  5.925738  5.884423  5.838638  5.788381  5.733653  5.674456  5.610798  5.542695  5.470171  5.393267  5.312035  5.226551  5.136914  5.043253  4.945731  4.844551  4.739966   4.63228  4.521859  4.409137   4.29462  4.178894  4.062625  3.946556  3.831506  3.718354  3.608022  3.501444  3.399538  3.303159  3.213057  3.129838  3.053927  2.985548  2.924716        2.871249 | 812.756821   821.85586  829.223841  834.872272  838.814268  841.064701  841.640349  840.560079  837.845029  833.518823  827.607792  820.141232  811.151667  800.675142  788.751533  775.424874    760.7437  744.761405    727.5366  709.133478  689.622173  669.079096  647.587244  625.236475  602.123709  578.353055  554.035834  529.290471  504.242216  479.022686  453.769162  428.623626  403.731495  379.240024  355.296352  332.045186  309.626134  288.170729  267.799222  248.617267  230.712667  214.152411  198.980238   185.21502  172.850181  161.854356  152.173327  143.733166  136.444336  130.206375  124.912751 |  100.0 | 712.756821 | 124.912751 |
    | 03:30 | 9.373376  9.426998  9.471884  9.508236  9.536228  9.556009  9.567705  9.571419  9.567239  9.555235  9.535461  9.507961  9.472764  9.429892  9.379359  9.321171  9.255331   9.18184  9.100699   9.01191  8.915483  8.811436  8.699799  8.580619  8.453965  8.319933  8.178652  8.030292   7.87507  7.713259  7.545201  7.371312  7.192096  7.008154  6.820197  6.629053  6.435675  6.241146  6.046678    5.8536  5.663345  5.477418  5.297359  5.124684  4.960829  4.807079  4.664497   4.53387  4.415666             4.310013 |  579.90155  585.746449  590.661157  594.656249  597.741629  599.926681  601.220415  601.631603  601.168922  599.841094  597.657031  594.625992   590.75774  586.062723  580.552251  574.238704  567.135742  559.258544  550.624055  541.251253  531.161442  520.378557  508.929483  496.844395  484.157112  470.905446  457.131565  442.882337  428.209656  413.170735  397.828333  382.250912  366.512672  350.693445  334.878395  319.157489  303.624672  288.376723  273.511729  259.127158  245.317549  232.171834  219.770423  208.182184  197.461582  187.646221  178.755115  170.787953   163.72551  157.531264 | 110.590056  111.136896  111.594639  111.965357  112.250824  112.452553  112.571824  112.609705  112.567079  112.444658  112.243007  111.962554  111.603616  111.166409  110.651068  110.057665  109.386228  108.636764   107.80928  106.903809  105.920445  104.859368  103.720888  102.505487  101.213867   99.847004   98.406217   96.893232   95.310268   93.660119   91.946257   90.172931   88.345277   86.469431   84.552636   82.603338   80.631263   78.647453   76.664256    74.69524   72.755012   70.858926    69.02267   67.261726   65.590732   64.022778   62.568722   61.236586   60.031132        58.953682 | 108.733762  109.269982  109.718837  110.082355  110.362279  110.560091  110.677045  110.714191  110.672392  110.552349  110.354613  110.079607  109.727639  109.298923  108.793589  108.211709  107.553312  106.818401  106.006986  105.119099  104.154831   103.11436  101.997988   100.80619   99.539652   98.199333   96.786524   95.302921   93.750697   92.132593   90.452013   88.713123   86.920961   85.081543   83.201971   81.290527   79.356748   77.411462   75.466778   73.535999   71.633449   69.774184   67.973586   66.246839   64.608294   63.070789    61.64497   60.338703   59.156658     58.100132 | 1.825475  1.831609  1.836734  1.840879  1.844068   1.84632   1.84765  1.848073  1.847597  1.846231  1.843981  1.840848  1.836834  1.831939   1.82616  1.819494  1.811935  1.803477  1.794114  1.783838  1.772643  1.760519  1.747461  1.733461  1.718516  1.702624  1.685783  1.667999  1.649281  1.629643  1.609108  1.587706   1.56548  1.542483  1.518783  1.494466  1.469634  1.444411   1.41894  1.393387  1.367938  1.342797  1.318181  1.294313  1.271417  1.249703  1.229357  1.210533   1.19334  1.177839 | 1.338792  1.338734  1.338686  1.338647  1.338618  1.338598  1.338586  1.338582  1.338586  1.338598  1.338619  1.338648  1.338685  1.338731  1.338785  1.338849  1.338922  1.339006    1.3391  1.339206  1.339325  1.339456  1.339602  1.339763   1.33994  1.340136  1.340351  1.340588  1.340848  1.341133  1.341447   1.34179  1.342167  1.342579  1.343029  1.343519  1.344053  1.344631  1.345256  1.345927  1.346643  1.347401  1.348196  1.349021  1.349867  1.350723  1.351576  1.352412  1.353217          1.353979 | 0.755448  0.756496  0.757365  0.758065    0.7586  0.758977    0.7592   0.75927  0.759191  0.758963  0.758586  0.758059  0.757382  0.756552  0.755566  0.754419  0.753109  0.751628  0.749972  0.748132  0.746101  0.743871  0.741432  0.738773  0.735883   0.73275  0.729361  0.725703  0.721761  0.717522   0.71297  0.708092  0.702876  0.697311  0.691388  0.685104  0.678459  0.671461  0.664128  0.656487  0.648577  0.640454  0.632186  0.623855  0.615555  0.607391  0.599469  0.591893  0.584757      0.578139 | -0.459585   -0.46046  -0.461123  -0.461582   -0.46184  -0.461904  -0.461777  -0.461459  -0.460952  -0.460257  -0.459372  -0.458295  -0.457023  -0.455554  -0.453881     -0.452  -0.449905  -0.447588  -0.445042  -0.442257  -0.439225  -0.435933  -0.432372  -0.428528   -0.42439  -0.419944  -0.415177  -0.410076  -0.404627  -0.398818  -0.392638  -0.386079  -0.379135  -0.371805  -0.364093  -0.356011  -0.347581  -0.338834  -0.329815  -0.320582  -0.311209  -0.301783  -0.292405  -0.283185  -0.274236   -0.26567  -0.257589   -0.25008  -0.243206     -0.237008 | 0.702101  0.702146  0.702149  0.702112  0.702035  0.701918  0.701762  0.701567  0.701332  0.701059  0.700746  0.700393       0.7  0.699566  0.699091  0.698574  0.698015  0.697412  0.696766  0.696075  0.695338  0.694556  0.693728  0.692853  0.691931  0.690963  0.689948  0.688889  0.687786  0.686643  0.685461  0.684245  0.683001  0.681734  0.680452  0.679164  0.677882  0.676616   0.67538  0.674188  0.673054   0.67199  0.671008  0.670119  0.669329   0.66864  0.668051  0.667558  0.667153      0.666827 | 784.982427 | 790.710094  801.397827  810.412313  817.758568  823.443338   827.47521  829.864731  830.624555  829.769591  827.317189  823.287334  817.702865  810.589714   801.97717  791.898162  780.389558  767.492491  753.252693  737.720844  720.952927  703.010587  683.961481  663.879611  642.845633  620.947116  598.278751   574.94247  551.047472  526.710112  502.053635  477.207715  452.307768  427.494007  402.910189  378.702053  355.015392  331.993778  309.775944  288.492859  268.264578  249.196984  231.378593  214.877627  199.739607  185.985719  173.612191  162.590847  152.870913  144.381975          137.037896 |  133.60186 | 1.227172  1.231349  1.234838  1.237661  1.239832  1.241365   1.24227  1.242558  1.242234  1.241305  1.239772  1.237639  1.234907  1.231574  1.227639  1.223099   1.21795   1.21219  1.205811   1.19881  1.191181  1.182918  1.174017  1.164471  1.154279  1.143437  1.131946  1.119807  1.107026  1.093612  1.079579  1.064947  1.049744  1.034006  1.017778  1.001117  0.984091  0.966785  0.949296  0.931736  0.914233  0.896925  0.879963  0.863502  0.847695  0.832689  0.818616  0.805583  0.793668       0.782918 | 5.951054  5.982056     6.008  6.029008  6.045182  6.056611  6.063367  6.065513  6.063098  6.056163  6.044739  6.028849  6.008509  5.983729  5.954513  5.920864  5.882778  5.840251  5.793278  5.741855  5.685981  5.625658  5.560894  5.491709   5.41813    5.3402  5.257983  5.171561  5.081044  4.986574  4.888331  4.786536  4.681463  4.573439  4.462858  4.350181  4.235947   4.12077  4.005348  3.890451  3.776922  3.665653  3.557568  3.453591   3.35461  3.261435  3.174752  3.095091  3.022789        2.957982 | 784.982427  796.508674    806.3595  814.539083   821.05328  825.909728  829.117963  830.689545  830.638212  828.980045  825.733654   820.92039  814.564569  806.693727  797.338891  786.534872   774.32058  760.739347   745.83928  729.673605  712.301035  693.786118  674.199585  653.618678  632.127432  609.816917  586.785411  563.138476  538.988922  514.456634  489.668212  464.756412  439.859339  415.119362  390.681709  366.692737  343.297825  320.638928  298.851788  278.062868  258.386114  239.919664  222.742723  206.912815  192.463681  179.404087  167.717744  157.364491  148.282714  140.392879   133.60186 |  100.0 | 684.982427 |  133.60186 |
    | 04:00 | 9.225254  9.291504   9.34896  9.397843  9.438343  9.470621  9.494817  9.511046  9.519406  9.519977  9.512821   9.49799  9.475519  9.445436  9.407756  9.362488  9.309634  9.249192  9.181157  9.105523  9.022287  8.931448  8.833016  8.727009   8.61346  8.492422  8.363974  8.228221  8.085306  7.935418  7.778794  7.615735  7.446611  7.271871  7.092057  6.907815  6.719901   6.52919  6.336688  6.143524  5.950956   5.76035  5.573164   5.39091  5.215113   5.04725  4.888687   4.74061  4.603964             4.479397 | 563.905408  571.032753  577.249664  582.564939  586.986729  590.522666  593.179993  594.965682  595.886559   595.94942  595.161158  593.528894  591.060115   587.76282   583.64568  578.718208  572.990948  566.475672  559.185599  551.135631   542.34261  532.825583  522.606101  511.708523  500.160346  487.992537   475.23989  461.941373  448.140478  433.885559  419.230135  404.233157  388.959199  373.478563  357.867258  342.206803  326.583841  311.089485  295.818363  280.867319   266.33373  252.313427  238.898237  226.173219  214.213714  203.082394  192.826577  183.476074   175.04188  167.515914 | 109.079504  109.755117  110.341059  110.839572  111.252589  111.581766  111.828513  111.994019  112.079275  112.085093  112.012122   111.86087  111.631715  111.324923  110.940661  110.479016  109.940012  109.323624    108.6298  107.858483  107.009634   106.08326  105.079442  103.998374  102.840399  101.606056  100.296131   98.911714   97.454269   95.925702    94.32845   92.665567   90.940825    89.15882   87.325079   85.446172    83.52981   81.584938   79.621787   77.651901   75.688082   73.744276   71.835343   69.976713   68.183926    66.47205   64.855016   63.344927   61.951404        60.681061 | 107.252544  107.915036    108.4896  108.978433  109.383429  109.706214  109.948169  110.110461  110.194061  110.199766  110.128212  109.979898  109.755193  109.454358  109.077558  108.624878  108.096341  107.491923  106.811572  106.055234  105.222869  104.314484   103.33016  102.270086  101.134598   99.924225   98.639737   97.282205   95.853062   94.354179   92.787945   91.157354   89.466105   87.718705   85.920575   84.078154   82.199007   80.291903   78.366876   76.435243    74.50956   72.603502   70.731639   68.909102    67.15113   65.472497   63.886865   62.406101    61.03964     59.793965 | 1.808476   1.81609  1.822679  1.828275  1.832905   1.83659   1.83935    1.8412  1.842152  1.842217  1.841402  1.839711  1.837149  1.833715  1.829409  1.824229   1.81817  1.811229  1.803398  1.794672  1.785041  1.774499  1.763037  1.750648  1.737325  1.723062  1.707855  1.691702  1.674606  1.656572  1.637612  1.617744  1.596993  1.575395  1.552996  1.529857  1.506051  1.481672   1.45683  1.431657  1.406306  1.380948  1.355778  1.331004  1.306845  1.283524  1.261257  1.240244  1.220656  1.202629 | 1.338956  1.338882  1.338818  1.338765  1.338721  1.338687  1.338662  1.338644  1.338636  1.338635  1.338643  1.338658  1.338682  1.338714  1.338754  1.338803  1.338862  1.338929  1.339007  1.339095  1.339194  1.339305  1.339428  1.339566  1.339718  1.339886  1.340071  1.340274  1.340499  1.340745  1.341016  1.341313  1.341639  1.341996  1.342387  1.342815  1.343282   1.34379  1.344342  1.344939  1.345582  1.346271  1.347003  1.347775  1.348581  1.349413  1.350261  1.351113  1.351957          1.352777 | 0.752505  0.753831  0.754968  0.755927  0.756716  0.757341  0.757807  0.758118  0.758279   0.75829  0.758153  0.757868  0.757435  0.756854  0.756121  0.755234  0.754191  0.752986  0.751614  0.750071  0.748348   0.74644  0.744337  0.742031  0.739512  0.736768  0.733788  0.730561  0.727072  0.723308  0.719255  0.714899  0.710226  0.705223  0.699877   0.69418  0.688125  0.681709  0.674938  0.667823  0.660386  0.652659  0.644688  0.636533  0.628269  0.619982  0.611771  0.603741  0.595997      0.588639 | -0.456474  -0.457693   -0.45869  -0.459474  -0.460053  -0.460432  -0.460617   -0.46061  -0.460414  -0.460031  -0.459459    -0.4587  -0.457751  -0.456609  -0.455272  -0.453736  -0.451995  -0.450043  -0.447875  -0.445481  -0.442855  -0.439986  -0.436865  -0.433481  -0.429822  -0.425876   -0.42163  -0.417071  -0.412186  -0.406961  -0.401385  -0.395444   -0.38913  -0.382436  -0.375357  -0.367895  -0.360057   -0.35186  -0.343328    -0.3345  -0.325425  -0.316168  -0.306809  -0.297439  -0.288163  -0.279092  -0.270339  -0.262012  -0.254204     -0.246994 | 0.701634  0.701749  0.701822  0.701854  0.701845  0.701797  0.701708  0.701581  0.701414  0.701209  0.700964  0.700681  0.700357  0.699994  0.699591  0.699147  0.698661  0.698134  0.697563  0.696949  0.696291  0.695588   0.69484  0.694046  0.693205  0.692317  0.691383  0.690403  0.689377  0.688307  0.687195  0.686042  0.684854  0.683633  0.682387  0.681121  0.679844  0.678566  0.677297  0.676051   0.67484  0.673678  0.672579  0.671555  0.670617  0.669774   0.66903  0.668386   0.66784      0.667387 | 754.817086 | 761.644889  774.561774   785.87282  795.575811  803.670355  810.157951   815.04207  818.328251  820.024218  820.140019  818.688181  815.683888  811.145179  805.093176  797.552323  788.550656  778.120087  766.296713   753.12114   738.63882    722.9004  705.962076  687.885943  668.740336  648.600149   627.54712  605.670068   583.06506  559.835491  536.092047  511.952536  487.541532  462.989827  438.433635  414.013527  389.873048   366.15701  343.009432  320.571127  298.976965  278.352866  258.812614   240.45463  223.358896  207.584248  193.166283  180.116147  168.420393  158.042048          148.922886 |  144.62564 | 1.215595   1.22078  1.225268  1.229079  1.232231   1.23474  1.236619  1.237879  1.238527  1.238572  1.238017  1.236866  1.235121  1.232783  1.229851  1.226323  1.222197   1.21747  1.212136  1.206191   1.19963  1.192446  1.184635   1.17619  1.167106  1.157379  1.147006  1.135985  1.124317  1.112005  1.099056  1.085481  1.071297  1.056527  1.041202  1.025362  1.009056  0.992346  0.975308   0.95803  0.940615  0.923183  0.905863  0.888801  0.872147  0.856055  0.840675  0.826148  0.812593       0.800107 | 5.865375  5.903704  5.936935  5.965201  5.988614  6.007271  6.021254  6.030632  6.035463  6.035792  6.031658  6.023087  6.010101  5.992714  5.970932  5.944758  5.914191  5.879226  5.839855  5.796071  5.747866  5.695234  5.638172  5.576683  5.510778  5.440477  5.365814  5.286838  5.203618  5.116248   5.02485   4.92958  4.830634  4.728255  4.622736  4.514432  4.403763  4.291221  4.177372  4.062865  3.948424  3.834848  3.722994  3.613766  3.508084  3.406854  3.310928  3.221059  3.137867        3.061797 | 754.817086  768.537161  780.652864  791.161084  800.060505  807.351664  813.037025  817.121066  819.610386  820.513834  819.842651  817.610643  813.834363  808.533329  801.730254    793.4513  783.726359  772.589346  760.078519  746.236809  731.112163  714.757902  697.233068  678.602777  658.938549  638.318619  616.828197  594.559676  571.612766  548.094521  524.119243  499.808229  475.289332    450.6963  426.167852  401.846468  377.876851  354.404049  331.571214  309.517025   288.37279  268.259319  249.283673  231.535949  215.086315   199.98253  186.248218  173.882136  162.858604  153.129203   144.62564 |  100.0 | 654.817086 |  144.62564 |
    | 04:30 | 9.063136  9.141484     9.211  9.271919  9.324445  9.368752   9.40499  9.433287   9.45375  9.466468  9.471513  9.468943    9.4588  9.441118  9.415916  9.383206  9.342992  9.295272  9.240038  9.177279  9.106984  9.029142  8.943745  8.850794  8.750296  8.642272  8.526762  8.403823  8.273544  8.136045  7.991483  7.840066  7.682056  7.517779  7.347637  7.172117  6.991801  6.807377  6.619648   6.42954  6.238107  6.046527  5.856097  5.668217  5.484359  5.306036  5.134746  4.971911  4.818818             4.676544 | 546.649201  554.955869   562.37758    568.9212  574.593005  579.398798  583.344013  586.433814  588.673202   590.06711  590.620509  590.338518  589.226517  587.290274  584.536074  580.970869  576.602428  571.439516  565.492077  558.771437  551.290527   543.06412   534.10909  524.444689  514.092839  503.078449  491.429741  479.178587  466.360858  453.016772  439.191234  424.934153  410.300731  395.351694  380.153443  364.778104  349.303434  333.812544    318.3934  303.138047  288.141515  273.500362  259.310835  245.666639  232.656366  220.360654  208.849254  198.178191  188.387311  179.498488 | 107.426214  108.225207  108.934135  109.555392   110.09105  110.542894  110.912453  111.201028  111.409712  111.539411  111.590861  111.564647  111.461214  111.280886  111.023875  110.690301  110.280202  109.793549  109.230268   108.59025  107.873376  107.079539  106.208665  105.260743   104.23586  103.134231  101.956248  100.702519   99.373929     97.9717   96.497454   94.953298   93.341904   91.666604   89.931494   88.141533   86.302661   84.421894   82.507426   80.568703   78.616461    76.66272   74.720708   72.804693   70.929709   69.111166   67.364337   65.703747   64.142495         62.69158 | 105.631359  106.414836  107.109998   107.71919  108.244446  108.687515  109.049898  109.332869  109.537501  109.664681  109.715132  109.689427  109.588003  109.411176  109.159157   108.83206  108.429925  107.952723   107.40038   106.77279  106.069838  105.291417  104.437454  103.507941   102.50296  101.422724  100.267616   99.038234   97.735445   96.360445   94.914829   93.400659   91.820557    90.17779   88.476375   86.721174   84.918011   83.073768   81.196478   79.295403   77.381072   75.465272   73.560973   71.682165   69.843592   68.060364   66.347457   64.719115   63.188181     61.765441 | 1.789771  1.798823  1.806835  1.813841  1.819869  1.824946  1.829093  1.832327  1.834664  1.836116  1.836692  1.836398  1.835241  1.833222  1.830342  1.826601  1.821995  1.816522  1.810176  1.802951   1.79484  1.785835  1.775928   1.76511  1.753374  1.740711  1.727115   1.71258  1.697104  1.680686   1.66333  1.645044  1.625841  1.605743   1.58478  1.562991  1.540429  1.517159  1.493264  1.468843  1.444015   1.41892   1.39372  1.368593  1.343741  1.319373  1.295711  1.272974  1.251369  1.231084 | 1.339145  1.339053  1.338973  1.338904  1.338845  1.338797  1.338757  1.338727  1.338705  1.338691  1.338686  1.338689    1.3387  1.338718  1.338746  1.338781  1.338825  1.338878   1.33894  1.339011  1.339093  1.339186   1.33929  1.339406  1.339535  1.339679  1.339837  1.340012  1.340205  1.340418  1.340652  1.340908   1.34119    1.3415  1.341839   1.34221  1.342616   1.34306  1.343544   1.34407  1.344641  1.345257  1.345918  1.346624  1.347371  1.348156  1.348971  1.349808  1.350656          1.351502 | 0.749197  0.750807  0.752218   0.75344  0.754484  0.755357  0.756067  0.756618  0.757015  0.757261  0.757358  0.757308  0.757112   0.75677   0.75628  0.755641   0.75485  0.753905  0.752802  0.751536  0.750101  0.748491    0.7467   0.74472  0.742541  0.740156  0.737552   0.73472  0.731648  0.728321  0.724729  0.720856  0.716688  0.712212  0.707415  0.702283  0.696806  0.690975  0.684787  0.678243  0.671349  0.664122  0.656588  0.648785  0.640765  0.632594  0.624351  0.616129  0.608028      0.600152 | -0.452891  -0.454458  -0.455793  -0.456905  -0.457804  -0.458498  -0.458992  -0.459292  -0.459401  -0.459322  -0.459056  -0.458604  -0.457966   -0.45714  -0.456124  -0.454915   -0.45351  -0.451903   -0.45009  -0.448064  -0.445818  -0.443344  -0.440634  -0.437677  -0.434464  -0.430983  -0.427222  -0.423169  -0.418812  -0.414136  -0.409129  -0.403778   -0.39807  -0.391996  -0.385546  -0.378715  -0.371502  -0.363909   -0.35595  -0.347643  -0.339018   -0.33012  -0.321003  -0.311738   -0.30241  -0.293117  -0.283964  -0.275064  -0.266527     -0.258454 | 0.701065  0.701251  0.701393  0.701493  0.701551  0.701569  0.701547  0.701486  0.701386  0.701247   0.70107  0.700853  0.700598  0.700304   0.69997  0.699596  0.699182  0.698727   0.69823  0.697691  0.697108  0.696483  0.695812  0.695097  0.694336  0.693529  0.692675  0.691775  0.690829  0.689836  0.688799  0.687719  0.686597  0.685436  0.684242  0.683018  0.681771  0.680507  0.679237   0.67797  0.676718  0.675494   0.67431   0.67318  0.672119  0.671136  0.670243  0.669446  0.668749       0.66815 |  722.79402 | 730.598093   745.50288  758.883087  770.729355  781.034286  789.792466    797.0005  802.657069  806.763004  809.321385  810.337655  809.819761  807.778315  804.226773  799.181646  792.662723  784.693326  775.300579  764.515703  752.374329  738.916821   724.18862   708.24059  691.129368   672.91771  653.674818  633.476642  612.406145  590.553507  568.016252  544.899284  521.314786  497.381976  473.226674  448.980649  424.780714  400.767533  377.084109  353.873938  331.278804  309.436249  288.476729  268.520541  249.674633  232.029451  215.656013  200.603473  186.897388  174.538955          163.505363 | 158.267707 | 1.202852   1.20902  1.214477  1.219249  1.223354  1.226812  1.229635  1.231838  1.233429  1.234418   1.23481   1.23461  1.233822  1.232447  1.230486  1.227938  1.224802  1.221075  1.216753  1.211832  1.206306  1.200171   1.19342  1.186048  1.178048  1.169415  1.160143   1.15023  1.139671  1.128467  1.116619  1.104132  1.091014  1.077279  1.062946  1.048042    1.0326  1.016666  1.000293  0.983548  0.966513  0.949282  0.931964  0.914683  0.897575  0.880785  0.864466   0.84877  0.833841       0.819811 | 5.771526  5.816891  5.857126  5.892375  5.922757   5.94838  5.969333  5.985691  5.997519   6.00487  6.007786    6.0063  6.000438  5.990218  5.975649  5.956738  5.933484  5.905884  5.873929  5.837611  5.796917  5.751837   5.70236   5.64848  5.590194  5.527506  5.460427  5.388984  5.313213  5.233172  5.148939  5.060617  4.968343  4.872288  4.772665  4.669739  4.563826  4.455308  4.344631  4.232318  4.118968  4.005258   3.89194  3.779833  3.669812  3.562784  3.459659  3.361315   3.26856        3.182087 |  722.79402  738.459753   752.60598  765.222366  776.300536  785.834079  793.818582  800.251671  805.133083  808.464744  810.250884  810.498158  809.215799  806.415788   802.11305  796.325669  789.075128  780.386571  770.289086  758.816008   746.00524  731.899585  716.547095  700.001418   682.32215  663.575174  643.832978  623.174943   601.68758  579.464704  556.607523  533.224609   509.43174  485.351569   461.11309  436.850871  412.704015  388.814821  365.327102  342.384173  320.126472  298.688862  278.197649  258.767411  240.497774  223.470306  207.745763   193.36192  180.332264  168.645741  158.267707 |  100.0 |  622.79402 | 158.267707 |
    | 05:00 | 8.888712  8.978581  9.059601   9.13202  9.196051  9.251882  9.299671  9.339557  9.371654  9.396061  9.412857  9.422108  9.423863  9.418161  9.405026  9.384475  9.356515  9.321144  9.278354  9.228134  9.170467  9.105337  9.032725  8.952617  8.865005  8.769886  8.667271  8.557184   8.43967  8.314796  8.182661  8.043399  7.897186   7.74425  7.584878  7.419427  7.248331  7.072114  6.891402  6.706926  6.519539  6.330217   6.14006  5.950292  5.762247  5.577349  5.397081  5.222943  5.056392             4.898788 | 528.376671  537.753274  546.275878  553.949227  560.777585  566.764821  571.914493  576.229927  579.714299  582.370711   584.20228  585.212221  585.403945  584.781156  583.347964  581.109006  578.069572  574.235755  569.614608  564.214313  558.044375  551.115828  543.441459   535.03605  525.916644  516.102824  505.617011  494.484781  482.735195  470.401138   457.51966  444.132322   430.28552   416.03079  401.425067  386.530879  371.416457  356.155719  340.828099  325.518177  310.315066  295.311502  280.602605  266.284267  252.451156   239.19435  226.598666  214.739786  203.681377   193.47242 |  105.64743  106.563917  107.390166  108.128694  108.781692  109.351053   109.83841  110.245162  110.572493  110.821395  110.992687  111.087026  111.104926  111.046769  110.912823  110.703245  110.418103  110.057387  109.621019  109.108872  108.520785  107.856579   107.11608  106.299141  105.405667  104.435643  103.389169    102.2665  101.068084   99.794618     98.4471   97.026897   95.535811   93.976164   92.350885   90.663605    88.91876   87.121698   85.278782   83.397494   81.486518   79.555801   77.616574   75.681312    73.76362   71.878024   70.039648   68.263772   66.565282        64.958028 | 103.887119  104.785808  105.596011  106.320198  106.960515  107.518819  107.996712  108.395565   108.71654  108.960609  109.128574   109.22108  109.238633  109.181606   109.05026  108.844752  108.565148  108.211437  107.783543  107.281342  106.704674  106.053367  105.327248  104.526174   103.65005  102.698863  101.672711  100.571843     99.3967   98.147964   96.826614    95.43399    93.97186     92.4425   90.848783   89.194269   87.483308   85.721143   83.914015   82.069261   80.195395   78.302171   76.400603   74.502922    72.62247   70.773491   68.970815   67.229426   65.563919     63.987877 | 1.769528  1.779973  1.789362  1.797731  1.805114  1.811538  1.817027  1.821602  1.825278  1.828071  1.829992   1.83105   1.83125  1.830598  1.829097  1.826746  1.823545  1.819491   1.81458  1.808807  1.802166   1.79465   1.78625  1.776959  1.766767  1.755665  1.743646  1.730701  1.716825  1.702013  1.686262  1.669575  1.651955  1.633414  1.613969  1.593644  1.572473  1.550501  1.527786    1.5044  1.480432  1.455991  1.431204  1.406218  1.381202  1.356344  1.331847  1.307927  1.284801  1.262683 | 1.339358  1.339247  1.339149  1.339064   1.33899  1.338926  1.338873  1.338829  1.338793  1.338767  1.338749  1.338739  1.338737  1.338743  1.338757   1.33878   1.33881  1.338849  1.338897  1.338953  1.339019  1.339095  1.339181  1.339279  1.339388   1.33951  1.339645  1.339795  1.339961  1.340143  1.340345  1.340566   1.34081  1.341078  1.341371  1.341693  1.342046  1.342432  1.342854  1.343315  1.343817  1.344361   1.34495  1.345584  1.346263  1.346986  1.347748  1.348544  1.349366          1.350205 | 0.745532  0.747434  0.749124  0.750614  0.751916   0.75304  0.753993  0.754783  0.755414  0.755892   0.75622    0.7564  0.756435  0.756324  0.756068  0.755666  0.755117  0.754419  0.753569  0.752563  0.751397  0.750067  0.748566  0.746887  0.745025  0.742969  0.740712  0.738243  0.735553  0.732629  0.729459   0.72603   0.72233  0.718344  0.714059   0.70946  0.704534  0.699271  0.693661  0.687697  0.681377  0.674705  0.667693   0.66036  0.652738  0.644871  0.636816  0.628646  0.620444      0.612306 |  -0.44885  -0.450773  -0.452449  -0.453892  -0.455112   -0.45612  -0.456922  -0.457526  -0.457936  -0.458156  -0.458189  -0.458036  -0.457698  -0.457176  -0.456469  -0.455573  -0.454488  -0.453209  -0.451732  -0.450053  -0.448164   -0.44606  -0.443734  -0.441176  -0.438377  -0.435329  -0.432019  -0.428437  -0.424571  -0.420407  -0.415934  -0.411138  -0.406005  -0.400524  -0.394684  -0.388473  -0.381886  -0.374918  -0.367571  -0.359851  -0.351774  -0.343364  -0.334656  -0.325699  -0.316554  -0.307299  -0.298023  -0.288827   -0.27982     -0.271111 | 0.700401  0.700656  0.700866  0.701033  0.701158  0.701242  0.701286  0.701289  0.701254   0.70118  0.701068  0.700917  0.700727    0.7005  0.700233  0.699927  0.699582  0.699197  0.698771  0.698304  0.697795  0.697244   0.69665  0.696012  0.695329    0.6946  0.693826  0.693006   0.69214  0.691227  0.690268  0.689263  0.688215  0.687124  0.685993  0.684826  0.683626    0.6824  0.681153  0.679894  0.678633   0.67738  0.676146  0.674946  0.673792  0.672697  0.671675  0.670736  0.669888      0.669138 | 689.426711 | 698.078496  714.719825  729.930003  743.692649   755.99352  766.820486  776.163519  784.014708  790.368294  795.220728  798.570747  800.419477  800.770553  799.630266  797.007726  792.915058  787.367611  780.384196  771.987344  762.203585  751.063749  738.603279  724.862567  709.887291  693.728767   676.44429  658.097473   638.75856  618.504713  597.420248  575.596806  553.133439  530.136588  506.719926  483.004028  459.115856  435.187993   411.35763  387.765241  364.552933  341.862468  319.832923  298.598042  278.283306  259.002826  240.856177  223.925348  208.272029  193.935476          180.931185 | 174.713131 | 1.189058  1.196177  1.202574  1.208276  1.213305   1.21768  1.221419  1.224534  1.227038   1.22894  1.230248  1.230968  1.231105  1.230661  1.229638  1.228037  1.225857  1.223097  1.219752   1.21582  1.211297  1.206176  1.200454  1.194122  1.187177   1.17961  1.171416  1.162589  1.153125  1.143021  1.132273  1.120882  1.108852  1.096188  1.082901  1.069007  1.054528  1.039494  1.023944  1.007924  0.991496  0.974732  0.957718  0.940555  0.923357  0.906253  0.889382  0.872893  0.856936        0.84166 | 5.670463  5.722546  5.769479  5.811412  5.848476  5.880782  5.908429  5.931497  5.950058   5.96417  5.973881  5.979229  5.980244  5.976947  5.969354  5.957472  5.941304  5.920848  5.896097  5.867041  5.833668  5.795963  5.753912  5.707502   5.65672  5.601559  5.542016  5.478099  5.409821  5.337212  5.260317  5.179199  5.093947  5.004677  4.911537  4.814717  4.714451  4.611022  4.504774  4.396114  4.285519  4.173541  4.060809  3.948029  3.835979  3.725498   3.61747  3.512798  3.412376        3.317048 | 689.426711  706.780669  722.712105  737.203565  750.239752  761.807488  771.895698  780.495407   787.59977  793.204117  797.306019  799.905379  801.004542  800.608429  798.724692  795.363894  790.539709  784.269151  776.572813  767.475147  757.004745  745.194651  732.082687  717.711788   702.13035  685.392578  667.558827  648.695934  628.877519  608.184252  586.704063  564.532272  541.771636  518.532261  494.931371  471.092896  447.146839    423.2284  399.476816   376.03388  353.042138  330.642719  308.972844  288.163013  268.333959  249.593466  232.033202  215.725772  200.722212   187.05019  174.713131 |  100.0 | 589.426711 | 174.713131 |
    | 05:30 | 8.703663  8.804429  8.896355  8.979698   9.05468  9.121495  9.180312  9.231276   9.27451   9.31012  9.338191  9.358797  9.371992   9.37782  9.376313  9.367491  9.351364  9.327934  9.297194  9.259132  9.213732  9.160971  9.100827  9.033277    8.9583  8.875878  8.786001  8.688668  8.583893  8.471704  8.352153  8.225318  8.091311   7.95028  7.802421  7.647987   7.48729  7.320718  7.148742  6.971926  6.790937  6.606555  6.419681  6.231339  6.042677  5.854959  5.669552  5.487898  5.311479              5.14177 | 509.323711  519.656267  529.170993  537.870307  545.756297  552.830777  559.095344  564.551434  569.200376  573.043456  576.081974  578.317315   579.75102  580.384868  580.220961  579.261826  577.510517  574.970739  571.646978  567.544643  562.670234   557.03151  550.637688  543.499653  535.630186  527.044213  517.759076  507.794811   497.17446  485.924384  474.074593  461.659089  448.716201  435.288921  421.425214  407.178304  392.606901  377.775357   362.75372  347.617653  332.448176  317.331195  302.356764  287.618041  273.209897  259.227145  245.762403  232.903603  220.731246  209.315541 | 103.760297  104.787909  105.725376  106.575306  107.339975  108.021362  108.621182  109.140914  109.581819  109.944966  110.231241  110.441373  110.575937  110.635377   110.62001  110.530041  110.365576  110.126632  109.813146  109.424993  108.961995  108.423939  107.810591  107.121714  106.357092  105.516549   104.59998  103.607379  102.538873  101.394766  100.175582   98.882118   97.515502   96.077261   94.569397   92.994467   91.355675   89.656971   87.903153   86.099974    84.25424   82.373909   80.468161   78.547442   76.623458   74.709105   72.818316     70.9658   69.166672        67.435971 | 102.036632  103.044288  103.963551  104.796976  105.546795   106.21495  106.803122  107.312761  107.745104  108.101199  108.381915  108.587966  108.719917  108.778203  108.763134  108.674912  108.513641  108.279337  107.971939  107.591323  107.137317   106.60971  106.008272  105.332773  104.582999  103.758779  102.860009  101.886683  100.838927   99.717038    98.52153   97.253184   95.913107   94.502796   93.024213   91.479867   89.872899   88.207183   86.487423    84.71926   82.909369   81.065553   79.196814   77.313394   75.426772   73.549595   71.695523   69.878982   68.114792       66.4177 | 1.747914  1.759701  1.770417  1.780103  1.788792  1.796516  1.803301  1.809169  1.814138  1.818226  1.821445  1.823806  1.825317  1.825984  1.825812  1.824802  1.822955  1.820269  1.816743  1.812371   1.80715  1.801071  1.794129  1.786314  1.777619  1.768033  1.757548  1.746156  1.733847  1.720614  1.706451  1.691356  1.675326  1.658365   1.64048  1.621685  1.601998  1.581449  1.560077  1.537931  1.515075  1.491589   1.46757  1.443133  1.418413  1.393568  1.368773  1.344221  1.320121  1.296687 | 1.339597  1.339465  1.339348  1.339246  1.339155  1.339076  1.339008   1.33895  1.338901  1.338861   1.33883  1.338807  1.338793  1.338787  1.338788  1.338798  1.338816  1.338841  1.338875  1.338918  1.338969   1.33903    1.3391  1.339181  1.339272  1.339374  1.339489  1.339616  1.339758  1.339915  1.340088  1.340279  1.340489   1.34072  1.340974  1.341253  1.341559  1.341894  1.342261  1.342663  1.343101  1.343579  1.344099  1.344662  1.345269  1.345922  1.346618  1.347356  1.348131          1.348937 | 0.741517   0.74372  0.745695  0.747457  0.749022  0.750398  0.751597  0.752626  0.753492    0.7542  0.754756  0.755162  0.755421  0.755535  0.755506  0.755333  0.755015  0.754553  0.753944  0.753185  0.752273  0.751204  0.749974  0.748577  0.747007  0.745257   0.74332  0.741186  0.738847  0.736292  0.733511  0.730491   0.72722  0.723686  0.719875  0.715772  0.711365   0.70664  0.701585  0.696189  0.690444  0.684346  0.677895  0.671099  0.663973  0.656542  0.648842  0.640923  0.632849      0.624697 | -0.444366   -0.44665  -0.448673   -0.45045  -0.451994  -0.453317  -0.454427  -0.455332   -0.45604  -0.456555   -0.45688   -0.45702  -0.456976  -0.456748  -0.456338  -0.455744  -0.454964  -0.453998   -0.45284  -0.451488  -0.449937  -0.448181  -0.446215   -0.44403  -0.441619  -0.438973  -0.436083  -0.432939  -0.429529  -0.425842  -0.421866  -0.417588  -0.412995  -0.408075  -0.402813    -0.3972  -0.391223  -0.384875   -0.37815  -0.371046  -0.363567  -0.355723  -0.347534  -0.339029  -0.330248  -0.321245   -0.31209  -0.302863  -0.293659     -0.284582 | 0.699646  0.699971  0.700249  0.700482  0.700672  0.700821  0.700928  0.700996  0.701024  0.701013  0.700964  0.700877  0.700752  0.700589  0.700387  0.700147  0.699869  0.699551  0.699194  0.698797  0.698359   0.69788  0.697358  0.696795  0.696187  0.695536   0.69484  0.694098  0.693311  0.692478  0.691598  0.690672    0.6897  0.688684  0.687625  0.686524  0.685385  0.684211  0.683007  0.681779  0.680535  0.679283  0.678032  0.676795  0.675583  0.674409  0.673287  0.672229  0.671248      0.670354 | 655.199767 | 664.568615  682.690174  699.484436  714.928153   729.00043  741.682645   752.95839  762.813448  771.235784  778.215567  783.745208  787.819423  790.435322  791.592513  791.293233  789.542505   786.34831  781.721789  775.677465   768.23349  759.411907  749.238941  737.745305  724.966519  710.943249  695.721645  679.353692  661.897543  643.417852  623.986071  603.680713  582.587563   560.79981  538.418087  515.550391  492.311857  468.824349   445.21584  421.619548  398.172796  375.015554  352.288664   330.13171  308.680569   288.06465  268.403904  249.805702  232.361732  216.145102          201.207876 |  194.01565 | 1.174325  1.182361  1.189665  1.196265  1.202186  1.207448   1.21207  1.216067  1.219451  1.222235  1.224428  1.226035  1.227064  1.227519  1.227401  1.226713  1.225456  1.223627  1.221225  1.218248  1.214691  1.210551  1.205822  1.200497  1.194572   1.18804  1.180893  1.173127  1.164734   1.15571  1.146049  1.135749  1.124809  1.113229  1.101015  1.088174  1.074719  1.060668  1.046047   1.03089  1.015238  0.999145  0.982675  0.965908  0.948934   0.93186  0.914807  0.897906  0.881301       0.865139 | 5.563137  5.621594  5.674894  5.723193  5.766629  5.805319  5.839366   5.86886  5.893874  5.914472  5.930708  5.942624  5.950254  5.953624  5.952753  5.947651  5.938326  5.924775  5.906995  5.884977  5.858707  5.828171  5.793352  5.754232  5.710794  5.663023  5.610906  5.554435   5.49361  5.428438   5.35894  5.285148  5.207116  5.124915  5.038644  4.948433  4.854446   4.75689  4.656019  4.552141  4.445624  4.336904  4.226487  4.114956   4.00297  3.891261  3.780631  3.671933  3.566056        3.463894 | 655.199767  673.980446  691.445944  707.571823  722.336041  735.718861  747.702783  758.272502  767.414892  775.119014  781.376146  786.179833  789.525963  791.412862  791.841415  790.815209  788.340693   784.42737  779.088009  772.338875  764.199988  754.695399  743.853487  731.707277  718.294762  703.659252  687.849713  670.921114  652.934764  633.958628  614.067612  593.343804  571.876652  549.763058  527.107362  504.021202    480.6232  457.038466  433.397859     409.837  386.494979  363.512747  341.031167  319.188719  298.118883  277.947245  258.788398  240.742789  223.893646  208.304235   194.01565 |  100.0 | 555.199767 |  194.01565 |
    | 06:00 | 8.509648  8.620641  8.722833  8.816487   8.90183  8.979061  9.048354   9.10986  9.163708   9.21001  9.248857  9.280326  9.304481  9.321368  9.331026  9.333477  9.328737  9.316811  9.297694  9.271377  9.237841  9.197066  9.149024  9.093688  9.031029   8.96102  8.883635  8.798856  8.706673  8.607087  8.500115  8.385794  8.264182  8.135369  7.999481  7.856685  7.707198  7.551295  7.389321  7.221696  7.048928  6.871622   6.69049  6.506361  6.320184   6.13303  5.946092  5.760672  5.578162             5.400013 | 489.715297  500.886822  511.281613  520.899531   529.74031  537.803574  545.088861  551.595647  557.323383  562.271527  566.439585  569.827158  572.433997  574.260058  575.305575  575.571131  575.057746  573.766973  571.701006  568.862798  565.256199  560.886099  555.758595  549.881172    543.2629   535.91465  527.849331  519.082141  509.630842  499.516049   488.76154  477.394571  465.446214  452.951689  439.950696  426.487743  412.612437  398.379743  383.850181  369.089934  354.170843  339.170249  324.170646  309.259095  294.526361  280.065723  265.971432  252.336785  239.251853  226.800892 | 101.781726  102.913628  103.955795   104.91088  105.781207  106.568812  107.275467  107.902708  108.451856  108.924037    109.3202  109.641128  109.887455  110.059678  110.158163  110.183164  110.134826    110.0132  109.818249  109.549863  109.207868  108.792037  108.302107  107.737789   107.09879  106.384827  105.595652  104.731074  103.790988   102.77541  101.684509   100.51865   99.278447   97.964812   96.579019   95.122777   93.598302   92.008405   90.356589   88.647141   86.885242   85.077066    83.22988   81.352125   79.453479   77.544878   75.638478   73.747557   71.886309        70.069543 | 100.096483  101.206405  102.228334  103.164871  104.018297  104.790608   105.48354    106.0986  106.637084  107.100096  107.488565  107.803261  108.044805  108.213683  108.310256  108.334771  108.287372  108.168107  107.976943  107.713769  107.378415  106.970659  106.490243  105.936884  105.310294  104.610196  103.836345  102.988557  102.066727  101.070871  100.001154   98.857936   97.641816   96.353691    94.99481   93.566847   92.071976   90.512954   88.893215   87.216963   85.489279   83.716217   81.904902   80.063612   78.201836   76.330298    74.46092   72.606719   70.781615     69.000129 | 1.725096  1.738169  1.750159  1.761109  1.771054  1.780029   1.78806  1.795172  1.801387  1.806721   1.81119  1.814806  1.817579  1.819516  1.820624  1.820905  1.820361  1.818994    1.8168  1.813778  1.809924  1.805231  1.799693  1.793304  1.786054  1.777934  1.768936   1.75905  1.748266  1.736576   1.72397  1.710443  1.695989  1.680606  1.664293  1.647056  1.628905  1.609855   1.58993  1.569163  1.547597  1.525288  1.502306  1.478739  1.454689  1.430282  1.405662  1.380991  1.356454  1.332248 | 1.339861  1.339708  1.339571   1.33945  1.339342  1.339246  1.339163   1.33909  1.339027  1.338974   1.33893  1.338894  1.338867  1.338849  1.338838  1.338835  1.338841  1.338854  1.338875  1.338904  1.338942  1.338989  1.339044  1.339109  1.339183  1.339268  1.339364  1.339472  1.339593  1.339726  1.339875  1.340039   1.34022  1.340419  1.340638   1.34088  1.341144  1.341435  1.341754  1.342103  1.342485  1.342902  1.343358  1.343853  1.344391  1.344973  1.345599  1.346269  1.346982          1.347735 | 0.737162  0.739672  0.741939  0.743981  0.745811  0.747444   0.74889   0.75016   0.75126  0.752198  0.752979  0.753608  0.754088  0.754423  0.754614  0.754663  0.754569  0.754333  0.753954  0.753429  0.752758  0.751936  0.750961  0.749827   0.74853  0.747065  0.745423  0.743599  0.741584  0.739369  0.736944    0.7343  0.731424  0.728305   0.72493  0.721286   0.71736  0.713138  0.708606  0.703752  0.698564  0.693033  0.687154  0.680922  0.674343  0.667427  0.660194  0.652672  0.644906       0.63695 | -0.439453  -0.442106  -0.444481  -0.446596  -0.448466  -0.450105  -0.451523   -0.45273  -0.453733  -0.454539  -0.455154   -0.45558  -0.455822  -0.455881  -0.455758  -0.455455   -0.45497  -0.454302   -0.45345  -0.452409  -0.451178   -0.44975  -0.448122  -0.446287  -0.444239  -0.441969  -0.439471  -0.436734  -0.433749  -0.430505  -0.426991  -0.423196  -0.419107   -0.41471  -0.409994  -0.404946  -0.399552  -0.393803  -0.387689  -0.381201  -0.374338  -0.367098   -0.35949  -0.351527  -0.343233  -0.334642    -0.3258  -0.316767  -0.307618     -0.298439 | 0.698809  0.699201  0.699546  0.699845  0.700099  0.700311  0.700482  0.700611  0.700702  0.700753  0.700766   0.70074  0.700677  0.700576  0.700438  0.700262  0.700047  0.699795  0.699504  0.699174  0.698804  0.698395  0.697944  0.697452  0.696918  0.696341  0.695721  0.695056  0.694346  0.693591   0.69279  0.691943   0.69105  0.690111  0.689127  0.688099  0.687028  0.685918  0.684771  0.683592  0.682386  0.681158  0.679918  0.678674  0.677436  0.676217  0.675028  0.673884  0.672796      0.671778 | 620.561767 | 630.517326  649.862603   667.99308  684.878838  700.492554  714.809364  727.806763  739.464531  749.764688  758.691476  766.231363   772.37307  777.107622  780.428423  782.331348  782.814865  781.880172  779.531366  775.775625  770.623421  764.088754    756.1894  746.947194  736.388322  724.543633  711.448967  697.145493  681.680051  665.105496  647.481029  628.872514  609.352759  589.001759  567.906866   546.16289  523.872085  501.144003  478.095198  454.848725  431.533424   408.28294  385.234455  362.527104  340.300057  318.690267  297.829895  277.843473  258.844865  240.934172          224.194722 | 216.081698 | 1.158766  1.167682  1.175856   1.18332  1.190099  1.196214  1.201687  1.206533  1.210766    1.2144  1.217444  1.219906  1.221795  1.223114  1.223868   1.22406   1.22369  1.222758  1.221264  1.219206  1.216581  1.213384  1.209612  1.205259   1.20032  1.194787  1.188655  1.181917  1.174566  1.166595  1.157999  1.148772  1.138911  1.128412  1.117277  1.105506  1.093107  1.080089  1.066468  1.052264  1.037507  1.022233   1.00649  0.990335  0.973839  0.957086  0.940173  0.923212  0.906328       0.889658 | 5.450486  5.514947  5.574261  5.628587  5.678067  5.722824  5.762965  5.798582  5.829756  5.856554  5.879032  5.897238   5.91121  5.920978  5.926564  5.927981   5.92524  5.918342  5.907285  5.892061  5.872658  5.849063  5.821256  5.789219   5.75293   5.71237   5.66752  5.618362  5.564883  5.507078  5.444947  5.378501  5.307765  5.232779  5.153601  5.070316  4.983033  4.891896  4.797086  4.698829    4.5974  4.493133  4.386423  4.277732  4.167599  4.056636   3.94553  3.835039  3.725984        3.619229 | 620.561767  640.508365  659.255677  676.772458  693.030133  708.002645  721.666336  733.999861  744.984132  754.602285   762.83967  769.683875  775.124755  779.154505  781.767735  782.961582  782.735837    781.0931  778.038954  773.582165  767.734898  760.512968  751.936102  742.028223  730.817758  718.337956  704.627224  689.729465  673.694428  656.578042  638.442748    619.3578  599.399529  578.651558  557.204941  535.158216  512.617329  489.695428  466.512468  443.194616  419.873416  396.684667  373.767012  351.260191  329.302953  308.030637  287.572454  268.048513  249.566718   232.21966  216.081698 |  100.0 | 520.561767 | 216.081698 |
    | 06:30 | 8.308292  8.428797  8.540578  8.643895  8.738977  8.826026  8.905217  8.976705  9.040621  9.097082  9.146185  9.188009  9.222623  9.250079  9.270417  9.283668  9.289848  9.288967  9.281023  9.266009  9.243909    9.2147  9.178357  9.134848   9.08414    9.0262  8.960993  8.888488   8.80866  8.721488  8.626964  8.525094  8.415898   8.29942   8.17573  8.044931  7.907164  7.762616  7.611527  7.454199  7.291008  7.122408  6.948948  6.771275  6.590149  6.406445  6.221161  6.035416  5.850442             5.667574 | 469.762998  481.655022  492.816014  503.243043  512.933298  521.884066  530.092716  537.556688  544.273506  550.240779  555.456224  559.917693  563.623199  566.570961   568.75945  570.187443  570.854094  570.759001  569.902299   568.28475   565.90786  562.773995  558.886522  554.249959  548.870146  542.754429  535.911863  528.353437  520.092312  511.144079  501.527044  491.262516  480.375122  468.893125   456.84876  444.278565  431.223702  417.730273  403.849591   389.63841  375.159082  360.479616  345.673601  330.819973  316.002561  301.309388  286.831665  272.662463  258.895008  245.620618 |   99.72829  100.957198  102.097146   103.15078  104.120433  105.008158  105.815752  106.544783  107.196611  107.772402  108.273147  108.699675  109.052667  109.332664  109.540078  109.675205  109.738232  109.729244  109.648236  109.495122  109.269741  108.971872  108.601241  108.157536  107.640419   107.04954  106.384555   105.64515  104.831054  103.942074   102.97812  101.939238  100.825652   99.637805   98.376414   97.042523    95.63757   94.163461   92.622646   91.018214   89.353981   87.634597   85.865642   84.053729   82.206598   80.333179   78.443647   76.549407   74.663039        72.798146 |  98.082924   99.287967  100.405778  101.438951  102.389775   103.26026  104.052171  104.767045  105.406215  105.970824  106.461845  106.880091  107.226228  107.500787  107.704174  107.836677   107.89848  107.889666  107.810231   107.66009  107.439086  107.147001  106.783568   106.34848  105.841405     105.262  104.609929  103.884882  103.086596  102.214879  101.269644  100.250937   99.158976   97.994196   96.757301   95.449313   94.071643    92.62616   91.115267   89.541991   87.910077   86.224082   84.489479   82.712753   80.901491   79.064453   77.211614   75.354159   73.504423     71.675745 |  1.70124  1.715538  1.728744  1.740902  1.752049  1.762222  1.771448  1.779755  1.787165  1.793696  1.799366  1.804188  1.808173  1.811331  1.813668   1.81519    1.8159  1.815798  1.814886  1.813162  1.810621  1.807261  1.803076  1.798058    1.7922  1.785494  1.777931  1.769501  1.760195  1.750002  1.738912  1.726918  1.714011  1.700184  1.685434  1.669759  1.653161  1.635647   1.61723  1.597928  1.577768   1.55679   1.53504   1.51258  1.489489   1.46586  1.441805  1.417457  1.392967  1.368507 | 1.340153  1.339976  1.339818  1.339676   1.33955  1.339437  1.339337  1.339249  1.339172  1.339105  1.339047  1.338999  1.338959  1.338928  1.338905  1.338891  1.338884  1.338885  1.338894   1.33891  1.338935  1.338968   1.33901   1.33906   1.33912  1.339189  1.339268  1.339358   1.33946  1.339573  1.339699   1.33984  1.339995  1.340166  1.340356  1.340564  1.340793  1.341045  1.341321  1.341624  1.341956   1.34232  1.342717  1.343151  1.343623  1.344137  1.344693  1.345294  1.345938          1.346626 | 0.732474  0.735301  0.737867  0.740192  0.742294  0.744187  0.745883  0.747395   0.74873  0.749897  0.750903  0.751753  0.752452  0.753003   0.75341  0.753675  0.753798   0.75378  0.753622  0.753322   0.75288  0.752292  0.751557  0.750672  0.749631   0.74843  0.747064  0.745527  0.743811   0.74191  0.739814  0.737514  0.735001  0.732264   0.72929  0.726069  0.722586  0.718829  0.714785  0.710439   0.70578  0.700794  0.695472  0.689806  0.683791  0.677428  0.670723  0.663692  0.656358      0.648758 | -0.434126  -0.437154  -0.439886  -0.442344  -0.444544  -0.446501  -0.448228  -0.449736  -0.451033  -0.452129  -0.453028  -0.453737  -0.454259  -0.454598  -0.454756  -0.454735  -0.454534  -0.454154  -0.453593   -0.45285  -0.451922  -0.450806  -0.449497  -0.447992  -0.446284  -0.444367  -0.442234  -0.439877  -0.437287  -0.434455  -0.431372  -0.428025  -0.424404  -0.420496   -0.41629  -0.411771  -0.406929  -0.401749  -0.396221  -0.390334   -0.38408  -0.377452  -0.370451  -0.363078  -0.355344  -0.347267  -0.338876  -0.330209   -0.32132     -0.312273 | 0.697894  0.698354  0.698764  0.699127  0.699445  0.699719  0.699951  0.700142  0.700293  0.700404  0.700477  0.700512  0.700509  0.700469  0.700392  0.700277  0.700125  0.699935  0.699708  0.699443  0.699139  0.698796  0.698413   0.69799  0.697527  0.697022  0.696474  0.695884  0.695249  0.694571  0.693847  0.693078  0.692263  0.691402  0.690496  0.689544  0.688547  0.687508  0.686427  0.685308  0.684154  0.682971  0.681762  0.680537  0.679303  0.678069  0.676847  0.675648  0.674486      0.673372 | 585.920022 | 596.334382  616.650944  635.872055  653.961361  670.885374  686.613272  701.116758  714.369932  726.349217  737.033294  746.403078  754.441709  761.134568  766.469321  770.435981   773.02699  774.237331  774.064655  772.509435  769.575144  765.268447  759.599429  752.581833  744.233328  734.575792  723.635613  711.444012  698.037373  683.457581  667.752368  650.975646  633.187838  614.456172  594.854951  574.465762  553.377615  531.686995  509.497796  486.921109   464.07484   441.08312  418.075484  395.185768  372.550718  350.308269  328.595497  307.546238  287.288418  267.941142          249.611646 | 240.673729 | 1.142493  1.152248  1.161255  1.169545  1.177145  1.184079  1.190367  1.196028  1.201077  1.205527  1.209389  1.212674  1.215388  1.217539  1.219131  1.220168  1.220651  1.220582  1.219961  1.218786  1.217056  1.214767  1.211916  1.208498  1.204508  1.199939  1.194785  1.189041  1.182697  1.175749  1.168188  1.160009  1.151206  1.141773  1.131708  1.121008  1.109675  1.097713  1.085129  1.071936  1.058151  1.043798   1.02891  1.013529  0.997705  0.981503  0.964997  0.948277  0.931447       0.914624 | 5.333429  5.403501  5.468453  5.528448  5.583628  5.634119  5.680031  5.721459  5.758486  5.791184  5.819612  5.843821  5.863853  5.879739  5.891506  5.899171  5.902746  5.902236  5.897641  5.888955  5.876169  5.859268  5.838235  5.813049   5.78369  5.750133  5.712355  5.670333  5.624047   5.57348  5.518619  5.459458  5.396003  5.328267  5.256281  5.180092  5.099768  5.015402  4.927119  4.835077  4.739475  4.640559  4.538628  4.434041  4.327219  4.218657  4.108922  3.998654   3.88857        3.779449 | 585.920022  606.776721  626.556768  645.222337   662.73857   679.07336  694.197189  708.082987  720.706036  732.043904  742.076395  750.785535  758.155579  764.173038  768.826732  772.107863  774.010111  774.529756  773.665814  771.420207  767.797947  762.807344  756.460241  748.772268  739.763114  729.456825  717.882113  705.072684  691.067572  675.911485  659.655149  642.355634  624.076679  604.888973  584.870397  564.106211  542.689149  520.719422  498.304588  475.559262  452.604643  429.567816  406.580802  383.779318  361.301231  339.284672  317.865821  297.176373  277.340721  258.472947  240.673729 |  100.0 | 485.920022 | 240.673729 |
    | 07:00 | 8.101178  8.230437  8.351091  8.463391  8.567563  8.663805  8.752292  8.833178  8.906598  8.972669  9.031488  9.083142  9.127699  9.165216  9.195738  9.219296  9.235913  9.245601  9.248362   9.24419  9.233072  9.214988   9.18991  9.157809  9.118648  9.072391  9.018996  8.958426  8.890643  8.815614  8.733311  8.643717  8.546825  8.442643  8.331201  8.212552  8.086776  7.953991  7.814356  7.668079  7.515425  7.356725  7.192388  7.022906  6.848868  6.670966  6.490007  6.306915  6.122735             5.938629 | 449.663064  462.157038  473.969952  485.095832  495.529128  505.264639  514.297445  522.622871  530.236455  537.133929  543.311219  548.764438  553.489907  557.484171  560.744024  563.266552  565.049177   566.08971  566.386419  565.938104  564.744184  562.804798  560.120914   556.69446  552.528463  547.627206  541.996402  535.643388   528.57733  520.809456  512.353298  503.224963  493.443413  483.030764  472.012603  460.418306  448.281373  435.639752  422.536153  409.018343  395.139399    380.9579  366.538041  351.949632  337.267956  322.573431  307.951061    293.4896  279.280418   265.41601 |  97.616126    98.93432   100.16475  101.309995  102.372342  103.353819  104.256214  105.081097  105.829838  106.503624  107.103469  107.630233  108.084628  108.467231  108.778492  109.018742  109.188205     109.287  109.315156  109.272611   109.15923  108.974803  108.719064  108.391693  107.992332  107.520593  106.976075   106.35838  105.667126  104.901975  104.062649  103.148962   102.16085  101.098404   99.961917   98.751923   97.469258   96.115113   94.691109   93.199367   91.642593    90.02417   88.348253   86.619869    84.84502   83.030772   81.185345   79.318164   77.439887        75.562371 |  96.011777   97.304372   98.510908   99.633913   100.67563  101.638048  102.522919  103.331783  104.065984  104.726685  105.314882  105.831417  106.276988  106.652161  106.957377  107.192962  107.359133  107.456011  107.483619  107.441901  107.330721  107.149876  106.899103   106.57809  106.186483  105.723905  105.189962  104.584262  103.906432  103.156139  102.333113  101.437169  100.468245   99.426432   98.312014   97.125517    95.86776   94.539912   93.143562   91.680788   90.154245   88.567251   86.923879    85.22906   83.488677   81.709661    79.90007   78.069149   76.227347      74.38629 | 1.676509  1.691967  1.706325  1.719631  1.731924  1.743239  1.753607  1.763056  1.771609  1.779287  1.786107  1.792085  1.797233  1.801561  1.805078   1.80779  1.809702  1.810816  1.811133  1.810654  1.809375  1.807294  1.804407  1.800707  1.796188  1.790841   1.78466  1.777633  1.769752  1.761007  1.751386  1.740881  1.729481  1.717177  1.703963  1.689832  1.674782  1.658813  1.641928  1.624137  1.605454  1.585904  1.565516  1.544334  1.522412  1.499817  1.476635  1.452967  1.428933  1.404674 | 1.340473  1.340271   1.34009  1.339927  1.339781   1.33965  1.339533  1.339428  1.339336  1.339254  1.339183  1.339121  1.339069  1.339025   1.33899  1.338963  1.338944  1.338933   1.33893  1.338935  1.338948  1.338968  1.338997  1.339034  1.339079  1.339134  1.339198  1.339272  1.339356  1.339451  1.339557  1.339677  1.339809  1.339956  1.340119  1.340299  1.340496  1.340714  1.340953  1.341216  1.341504   1.34182  1.342166  1.342545  1.342958  1.343409  1.343899  1.344431  1.345006          1.345625 | 0.727464  0.730614  0.733486  0.736101  0.738478  0.740635  0.742585  0.744341  0.745913   0.74731   0.74854   0.74961  0.750525  0.751291   0.75191  0.752385  0.752719  0.752914  0.752969  0.752885  0.752662  0.752298  0.751792   0.75114   0.75034  0.749388   0.74828   0.74701  0.745573  0.743962  0.742169  0.740188  0.738009  0.735622  0.733017  0.730183  0.727108   0.72378  0.720186  0.716313  0.712147  0.707675  0.702885  0.697766  0.692308  0.686505  0.680356  0.673863  0.667038      0.659898 | -0.428399  -0.431809  -0.434905  -0.437709  -0.440242   -0.44252  -0.444557  -0.446366  -0.447958  -0.449341  -0.450523  -0.451511  -0.452309  -0.452923  -0.453354  -0.453607  -0.453681  -0.453579  -0.453299  -0.452841  -0.452203  -0.451383  -0.450378  -0.449184  -0.447797  -0.446211   -0.44442  -0.442419  -0.440198   -0.43775  -0.435066  -0.432136   -0.42895  -0.425497  -0.421764  -0.417741  -0.413413   -0.40877  -0.403797  -0.398483  -0.392818  -0.386791  -0.380396  -0.373628   -0.36649  -0.358986  -0.351131  -0.342948  -0.334468     -0.325739 | 0.696909  0.697435   0.69791  0.698336  0.698715   0.69905  0.699342  0.699593  0.699803  0.699973  0.700105  0.700198  0.700254  0.700273  0.700254  0.700199  0.700107  0.699978  0.699812  0.699609  0.699368  0.699089  0.698772  0.698416   0.69802  0.697584  0.697107  0.696588  0.696026  0.695422  0.694774  0.694081  0.693343   0.69256   0.69173  0.690856  0.689935   0.68897  0.687962  0.686911   0.68582  0.684693  0.683534  0.682347  0.681139  0.679917   0.67869  0.677469  0.676265      0.675089 |  551.63713 | 562.386699  583.429946  603.502172   622.56087  640.566681  657.483147  673.276505  687.915528  701.371399   713.61762   724.62995  734.386361   742.86703  750.054345  755.932941  760.489752  763.714088  765.597733  766.135068  765.323211  763.162184    759.6551  754.808373  748.631951   741.13957  732.349025  722.282469  710.966719  698.433582  684.720194  669.869353  653.929868  636.956886  619.012207  600.164568  580.489887  560.071442  538.999971  517.373675  495.298091  472.885805  450.255989  427.533712  404.849001  382.335626   360.12957  338.367179  317.182984  296.707208          277.063006 | 267.430129 | 1.125616  1.136166  1.145963  1.155039  1.163423  1.171138  1.178207  1.184648  1.190477  1.195709  1.200356  1.204429  1.207936  1.210884   1.21328  1.215128   1.21643  1.217189  1.217405  1.217078  1.216207   1.21479  1.212823  1.210303  1.207224  1.203582   1.19937  1.194582  1.189211  1.183251  1.176693   1.16953  1.161757  1.153365  1.144351  1.134709  1.124437  1.113535  1.102003  1.089849  1.077081  1.063714  1.049769  1.035273  1.020263  1.004784  0.988892  0.972657  0.956159       0.939494 | 5.212864  5.288128  5.358322  5.423608  5.484127  5.540005  5.591352  5.638266  5.680831   5.71912  5.753196  5.783112  5.808911  5.830629  5.848294  5.861928  5.871543  5.877148  5.878746  5.876332  5.869899  5.859434  5.844921  5.826341  5.803671  5.776886   5.74596  5.710868  5.671583  5.628081   5.58034  5.528344  5.472081  5.411549  5.346754  5.277717  5.204474  5.127079  5.045611  4.960175   4.87091   4.77799  4.681634  4.582109  4.479739  4.374908  4.268068  4.159741  4.050524        3.941088 |  551.63713  573.156838   593.72748  613.304912  631.848282  649.319747  665.684246  680.909326  694.964997  707.823625  719.459855  729.850561   738.97482  746.813909   753.35133   758.57285  762.466567  765.022999  766.235193  766.098857  764.612512  761.777671  757.599036  752.084721  745.246493  737.100039  727.665247  716.966511  705.033049  691.899234  677.604934  662.195854  645.723874  628.247376  609.831545  590.548639  570.478202  549.707214  528.330147  506.448909  484.172648  461.617388  438.905458  416.164693  393.527363  371.128812  349.105769  327.594337  306.727653  286.633239  267.430129 |  100.0 |  451.63713 | 267.430129 |
    | 07:30 | 7.889835  8.027052  8.155828    8.2764  8.388984  8.493771  8.590931  8.680615  8.762955  8.838068  8.906052  8.966995  9.020966  9.068026   9.10822  9.141583  9.168142   9.18791  9.200892  9.207087  9.206482   9.19906  9.184796  9.163658  9.135612  9.100618  9.058633  9.009613  8.953515  8.890294  8.819912  8.742333  8.657531  8.565489  8.466206  8.359695  8.245993  8.125164  7.997301  7.862539  7.721056  7.573083  7.418912  7.258905  7.093507  6.923247  6.748758  6.570779  6.390163             6.207883 | 429.595039  442.573638  454.925053  466.640014  477.710036  488.127275   497.88441  506.974558  515.391216  523.128203  530.179638   536.53991  542.203673  547.165847  551.421623  554.966486   557.79624  559.907048  561.295475  561.958549  561.893823   561.09946  559.574316  557.318049  554.331234  550.615493  546.173642  541.009856  535.129845  528.541057  521.252888   513.27692  504.627177  495.320391  485.376291  474.817913  463.671906  451.968865   439.74365  427.035708  413.889376  400.354153  386.484924   372.34212  357.991775  343.505464  328.960075  314.437381  300.023366  285.807265 |  95.460849   96.860194    98.17345   99.403053  100.551188  101.619806  102.610646  103.525246  104.364956  105.130958   105.82427  106.445764  106.996168  107.476081   107.88598  108.226224  108.497067   108.69866  108.831058  108.894231  108.888066  108.812376  108.666907  108.451346   108.16533  107.808456  107.380292  106.880388  106.308292  105.663566  104.945804  104.154653  103.289839  102.351196  101.338697  100.252494   99.092959   97.860735    96.55679   95.182482   93.739631   92.230593    90.65835   89.026601   87.339857   85.603545   83.824101   82.009062   80.167132        78.308228 |  93.898353   95.270524   96.558278   97.764003   98.889841   99.937708  100.909307  101.806145  102.629549  103.380676  104.060524  104.669949  105.209664  105.680258  106.082197  106.415834  106.681417  106.879095  107.008923  107.070869  107.064824  106.990603  106.847959  106.636584  106.356122  106.006178  105.586329  105.096133  104.535147  103.902941  103.199117  102.423329   101.57531  100.654895   99.662058   98.596948   97.459931   96.251635   94.973012   93.625393   92.210561   90.730827   89.189116   87.589054   85.935066   84.232472   82.487584   80.707792   78.901631     77.078825 | 1.651067   1.66761  1.683054  1.697444  1.710821  1.723221  1.734675  1.745211  1.754855  1.763626  1.771546  1.778628  1.784888  1.790336  1.794983  1.798835  1.801898  1.804176  1.805672  1.806385  1.806315  1.805461  1.803818  1.801381  1.798146  1.794105   1.78925  1.783572  1.777063  1.769712  1.761508  1.752442  1.742503   1.73168  1.719964  1.707347  1.693822  1.679384  1.664031  1.647765  1.630592  1.612525   1.59358  1.573786  1.553177  1.531802  1.509721  1.487008  1.463755  1.440073 | 1.340823  1.340593  1.340387  1.340201  1.340034  1.339884  1.339748  1.339627  1.339519  1.339422  1.339336  1.339261  1.339195  1.339139  1.339092  1.339053  1.339022  1.338999  1.338984  1.338977  1.338978  1.338986  1.339003  1.339027   1.33906  1.339101   1.33915  1.339209  1.339278  1.339356  1.339445  1.339546  1.339658  1.339783  1.339923  1.340077  1.340247  1.340435  1.340642  1.340869  1.341119  1.341394  1.341694  1.342024  1.342384  1.342778  1.343208  1.343676  1.344184          1.344735 | 0.722141  0.725622  0.728805  0.731716  0.734374  0.736799  0.739005  0.741008  0.742818  0.744446  0.745901   0.74719  0.748321  0.749298  0.750126  0.750809   0.75135  0.751751  0.752014  0.752139  0.752127  0.751977  0.751688  0.751259  0.750687   0.74997  0.749104  0.748084  0.746906  0.745565  0.744055  0.742367  0.740496  0.738431  0.736165  0.733688  0.730988  0.728054  0.724875  0.721438  0.717729  0.713736  0.709445  0.704844  0.699921  0.694666  0.689071  0.683133   0.67685      0.670231 |  -0.42229  -0.426087  -0.429551  -0.432707  -0.435576  -0.438177  -0.440526  -0.442638  -0.444523  -0.446192  -0.447655  -0.448919   -0.44999  -0.450873  -0.451574  -0.452094  -0.452436  -0.452603  -0.452594   -0.45241   -0.45205  -0.451514  -0.450798  -0.449899  -0.448816  -0.447542  -0.446074  -0.444405  -0.442529  -0.440439  -0.438128  -0.435586  -0.432805  -0.429774  -0.426482  -0.422918  -0.419071  -0.414928  -0.410476  -0.405703  -0.400597  -0.395147  -0.389342  -0.383173  -0.376637   -0.36973  -0.362456  -0.354825  -0.346855     -0.338573 | 0.695861  0.696451  0.696988  0.697476  0.697916  0.698311  0.698662   0.69897  0.699238  0.699465  0.699654  0.699804  0.699917  0.699992  0.700031  0.700033  0.699999  0.699929  0.699822  0.699679  0.699499  0.699282  0.699027  0.698735  0.698404  0.698034  0.697624  0.697174  0.696683   0.69615  0.695574  0.694955  0.694293  0.693585  0.692833  0.692035  0.691192  0.690303  0.689369  0.688391  0.687371  0.686309  0.685209  0.684075  0.682911  0.681722  0.680515  0.679299  0.678082      0.676876 | 518.029154 | 528.996258   550.53263  571.225767  591.027329  609.892427  627.779312  644.649117  660.465653  675.195245  688.806608  701.270744  712.560879  722.652419   731.52293  739.152144  745.521984  750.616613  754.422502  756.928522  758.126059   758.00914  756.574598  753.822245  749.755072  744.379473  737.705487  729.747064  720.522345  710.053969  698.369388  685.501197   671.48747  656.372108  640.205166  623.043186   604.94949  585.994454  566.255716  545.818328  524.774812  503.225103  481.276363  459.042615  436.644183  414.206905  391.861078  369.740113  347.978876   326.71171          306.070129 | 295.895863 | 1.108245  1.119541  1.130083  1.139904   1.14903  1.157488  1.165299  1.172483  1.179057  1.185036  1.190434   1.19526  1.199526  1.203238  1.206403  1.209028  1.211114  1.212666  1.213685   1.21417  1.214123  1.213541  1.212422  1.210762  1.208558  1.205805  1.202497  1.198629  1.194193  1.189184  1.183593  1.177413  1.170636  1.163256  1.155266   1.14666  1.137431  1.127578  1.117097   1.10599   1.09426  1.081914  1.068964  1.055427  1.041326  1.026694   1.01157  0.996004  0.980059       0.963807 | 5.089659  5.169673  5.244693  5.314875  5.380357  5.441261  5.497697  5.549761  5.597538  5.641101  5.680515  5.715833  5.747102  5.774358  5.797633  5.816949  5.832322  5.843764  5.851277  5.854862  5.854512  5.850217  5.841961  5.829727  5.813492  5.793231  5.768918  5.740525  5.708022   5.67138  5.630573  5.585574  5.536363  5.482922  5.425243  5.363326  5.297181  5.226835  5.152331  5.073732  4.991129  4.904639  4.814416  4.720652  4.623587  4.523511  4.420772  4.315783  4.209024        4.101047 | 518.029154  539.976699  561.105958  581.366789  600.712676   619.10038  636.489657  652.843032  668.125613  682.304953  695.350933  707.235686  717.933536  727.420971  735.676638  742.681356  748.418151   752.87232  756.031506  757.885805  758.427884  757.653126  755.559803  752.149257  747.426117  741.398528  734.078407  725.481718  715.628762  704.544493  692.258837  678.807032  664.229964  648.574514  631.893885  614.247921  595.703394  576.334248  556.221787   535.45478  514.129474  492.349479  470.225497  447.874872  425.420927  402.992052   380.72051  358.740947  337.188571  316.197013  295.895863 |  100.0 | 418.029154 | 295.895863 |
    | 08:00 | 7.675738  7.820077   7.95619  8.084295  8.204589  8.317251  8.422441  8.520304  8.610967  8.694542  8.771127  8.840807  8.903654  8.959726  9.009072  9.051729  9.087723  9.117073  9.139785   9.15586  9.165288  9.168054  9.164134    9.1535  9.136116  9.111942  9.080935  9.043049  8.998235  8.946443  8.887626  8.821737  8.748736  8.668587  8.581265  8.486757  8.385066  8.276212  8.160243  8.037233  7.907291   7.77057  7.627268  7.477642  7.322013  7.160777  6.994414  6.823495  6.648697             6.470806 | 409.720867  423.069176  435.847684  448.043584  459.645265  470.642085  481.024207   490.78246  499.908235  508.393405  516.230264  523.411479  529.930062  535.779344  540.952971  545.444904  549.249427  552.361168  554.775131   556.48673  557.491842  557.786866  557.368792  556.235284  554.384773  551.816567  548.530972  544.529428  539.814662  534.390856  528.263834  521.441268  513.932897  505.750773  496.909516  487.426594  477.322617  466.621638   455.35148  443.544047  431.235655  418.467333  405.285116  391.740292  377.889594  363.795313  349.525304  335.152847  320.756333  306.418728 |  93.277479    94.74945    96.13754   97.443959   98.670722    99.81965  100.892385  101.890393  102.814974  103.667276  104.448297  105.158898  105.799809  106.371636  106.874868  107.309884  107.676957  107.976264  108.207885  108.371815  108.467965  108.496171  108.456198  108.347748  108.170463   107.92394  107.607734  107.221368  106.764349  106.236175  105.636353  104.964418  104.219946  103.402585  102.512075  101.548281  100.511227   99.401136   98.218477   96.964014   95.638866   94.244575   92.783178   91.257286   89.670177   88.025886   86.329302   84.586269   82.803674         80.98953 |  91.757383    93.20077   94.561904   95.842953   97.045893    98.17251   99.224413   100.20304  101.109667  101.945417  102.711271  103.408073  104.036538   104.59726   105.09072  105.517288  105.877233  106.170727  106.397851  106.558597   106.65288  106.680539  106.641342  106.534997  106.361156   106.11942  105.809354  105.430492  104.982347   104.46443  103.876257  103.217369  102.487355  101.685867   100.81265   99.867572   98.850657   97.762123   96.602431   95.372328   94.072914   92.705699   91.272681   89.776421   88.220133   86.607772   84.944135    83.23495   81.486972     79.708058 | 1.625071  1.642622  1.659078  1.674485  1.688882  1.702305  1.714786  1.726353  1.737032  1.746844   1.75581  1.763946  1.771267  1.777784  1.783509  1.788451  1.792614  1.796006  1.798627  1.800482  1.801569  1.801888  1.801436   1.80021  1.798204  1.795413   1.79183  1.787446  1.782253  1.776241  1.769401  1.761721  1.753191    1.7438  1.733537  1.722393  1.710357  1.697422  1.683582  1.668834  1.653176  1.636614  1.619154  1.600812   1.58161  1.561578  1.540757  1.519201  1.496975  1.474163 | 1.341202  1.340943   1.34071    1.3405  1.340311   1.34014  1.339985  1.339846  1.339721  1.339609  1.339508  1.339419  1.339339   1.33927   1.33921  1.339159  1.339116  1.339081  1.339055  1.339036  1.339025  1.339022  1.339026  1.339039  1.339059  1.339087  1.339124  1.339169  1.339223  1.339286  1.339359  1.339443  1.339537  1.339643  1.339762  1.339894   1.34004  1.340201   1.34038  1.340576  1.340793  1.341031  1.341292  1.341578  1.341892  1.342235   1.34261   1.34302  1.343467          1.343953 | 0.716519  0.720336  0.723836  0.727047   0.72999  0.732687  0.735153  0.737405  0.739456  0.741316  0.742996  0.744505   0.74585  0.747037  0.748073   0.74896  0.749704  0.750308  0.750772    0.7511  0.751292  0.751348  0.751269  0.751052  0.750697  0.750202  0.749565   0.74878  0.747846  0.746757  0.745508  0.744094  0.742507  0.740741  0.738788  0.736638  0.734283  0.731711  0.728913  0.725877  0.722589  0.719039  0.715212  0.711096  0.706677  0.701944  0.696886  0.691494  0.685762      0.679688 | -0.415816  -0.420006  -0.423842  -0.427353  -0.430561  -0.433488   -0.43615  -0.438564  -0.440743  -0.442699  -0.444441  -0.445979   -0.44732  -0.448469  -0.449433  -0.450215   -0.45082  -0.451248  -0.451502  -0.451583  -0.451491  -0.451226  -0.450786   -0.45017  -0.449374  -0.448396  -0.447232  -0.445877  -0.444325  -0.442571  -0.440607  -0.438427  -0.436022  -0.433384  -0.430501  -0.427365  -0.423964  -0.420287  -0.416321  -0.412055  -0.407475  -0.402571  -0.397329  -0.391739  -0.385792  -0.379481  -0.372803  -0.365758  -0.358352     -0.350599 | 0.694755  0.695408  0.696006  0.696554  0.697053  0.697506  0.697914  0.698279  0.698603  0.698886   0.69913  0.699335  0.699503  0.699634  0.699728  0.699786  0.699808  0.699794  0.699744  0.699659  0.699537  0.699379  0.699185  0.698954  0.698685  0.698378  0.698033  0.697649  0.697225  0.696761  0.696256  0.695708  0.695118  0.694484  0.693807  0.693084  0.692317  0.691505  0.690647  0.689744  0.688797  0.687806  0.686774  0.685702  0.684594  0.683454  0.682286  0.681096  0.679893      0.678684 | 485.365238 | 496.439435  518.249068   539.34494   559.67325  579.183957    597.8304  615.568991  632.358972  648.162204  662.943015  676.668063   689.30625  700.828641  711.208428  720.420904  728.443467  735.255637  740.839103  745.177786   748.25792  750.068164   750.59972  749.846486  747.805223  744.475748  739.861142  733.967987   726.80662  718.391407  708.741035  697.878824  685.833052  672.637285  658.330718  642.958516  626.572145  609.229686  590.996117  571.943563  552.151477  531.706752  510.703726  489.244075  467.436543  445.396503  423.245297  401.109338  379.118934  357.406819          336.106362 | 325.557379 | 1.090488  1.102477  1.113716  1.124234  1.134061   1.14322  1.151734  1.159624  1.166906  1.173596  1.179709  1.185254  1.190243  1.194685  1.198586  1.201953   1.20479    1.2071  1.208886  1.210149   1.21089  1.211107  1.210799  1.209964  1.208598  1.206696  1.204255  1.201268   1.19773  1.193634  1.188972  1.183738  1.177923  1.171521  1.164523  1.156923  1.148713  1.139888  1.130444  1.120377  1.109686  1.098373  1.086444  1.073908  1.060778  1.047074  1.032825  1.018064  1.002836       0.987197 | 4.964652  5.048951  5.128362  5.203029  5.273083  5.338641  5.399807  5.456676  5.509331  5.557844  5.602279   5.64269  5.679125  5.711621  5.740211   5.76492  5.785765  5.802759  5.815908  5.825213   5.83067  5.832271  5.830002  5.823847  5.813783  5.799788  5.781834  5.759892  5.733933  5.703924  5.669833  5.631631  5.589289   5.54278  5.492084  5.437185  5.378078  5.314765  5.247264  5.175606  5.099842  5.020047  4.936322  4.848799  4.757649  4.663083  4.565362    4.4648  4.361772        4.256718 | 485.365238  507.519985  528.988731  549.715706  569.649104  588.740668  606.945351  624.221043  640.528345  655.830394  670.092714  683.283109  695.371577   706.33026  716.133402  724.757347  732.180546  738.383589  743.349261  747.062616  749.511067  750.684508  750.575447  749.179163  746.493889  742.521012  737.265297  730.735127  722.942776  713.904684  703.641765  692.179723  679.549383  665.787026  650.934734  635.040725   618.15968  600.353046  581.689304  562.244189  542.100842  521.349878  500.089341   478.42452  456.467611  434.337172  412.157354  390.056877  368.167711  346.623453  325.557379 |  100.0 | 385.365238 | 325.557379 |
    | 08:30 | 7.460295  7.610884  7.753521  7.888393  8.015673   8.13552  8.248082   8.35349  8.451864   8.54331  8.627921  8.705779  8.776953  8.841501  8.899472  8.950903  8.995823  9.034249  9.066192  9.091653  9.110625  9.123096  9.129043  9.128439  9.121251  9.107439   9.08696  9.059767  9.025809  8.985032  8.937385  8.882813  8.821265  8.752693  8.677057  8.594322  8.504464  8.407476  8.303366  8.192164  8.073926  7.948742  7.816737  7.678082  7.533001  7.381776   7.22476  7.062384  6.895165             6.723719 | 390.184427  403.791006  416.888256  429.459602  441.490121  452.966239  463.875493   474.20634  483.948013    493.0904  501.623956  509.539625  516.828791  523.483234  529.495101  534.856892  539.561453  543.601978  546.972024  549.665532  551.676862  553.000832  553.632773  553.568591  552.804842  551.338817  549.168644  546.293399  542.713235  538.429522  533.445009  527.763995  521.392529  514.338614  506.612441  498.226636  489.196525  479.540418  469.279905  458.440167  447.050291  435.143591  422.757922  409.935973  396.725543  383.179753  369.357211  355.322067  341.143956  326.897784 |  91.080379   92.616089    94.07071   95.446138   96.744142   97.966352    99.11426  100.189218  101.192442  102.125011  102.987879  103.781874  104.507707  105.165974  105.757164   106.28166  106.739749  107.131621  107.457376  107.717029  107.910514  108.037688  108.098337  108.092178  108.018872  107.878019  107.669176  107.391858  107.045549  106.629711  106.143799   105.58727  104.959602  104.260309  103.488965  102.645226  101.728858   100.73977   99.678049   98.544003   97.338212   96.061577   94.715387   93.301383   91.821839   90.279643   88.678387   87.022464   85.317159         83.56875 |  89.602949   91.108837    92.53521   93.883929   95.156726   96.355201   97.480818   98.534901   99.518643  100.433102  101.279214   102.05779  102.769528  103.415012  103.994721  104.509032  104.958225  105.342487  105.661916  105.916527  106.106254  106.230959   106.29043  106.284391  106.212508  106.074391  105.869603   105.59767  105.258086  104.850324  104.373848  103.828127  103.212647  102.526934   101.77057  100.943215  100.044643   99.074761   98.033658   96.921635    95.73926   94.487416   93.167368   91.780823   90.330011   88.817763   87.247603   85.623836   83.951647     82.237192 | 1.598678  1.617151  1.634542  1.650892  1.666241  1.680624  1.694071   1.70661  1.718268  1.729066  1.739025  1.748161  1.756491  1.764027   1.77078   1.77676  1.781973  1.786427  1.790124  1.793069  1.795261  1.796701  1.797388  1.797318  1.796488  1.794893  1.792526  1.789381  1.785449  1.780722  1.775189  1.768841  1.761666  1.753654  1.744794  1.735074  1.724484  1.713013  1.700654  1.687398  1.673241   1.65818  1.642217  1.625357  1.607612  1.588999  1.569544  1.549283  1.528261  1.506537 | 1.341612  1.341322  1.341061  1.340825  1.340612  1.340419  1.340244  1.340086  1.339943  1.339814  1.339698  1.339594    1.3395  1.339418  1.339345  1.339281  1.339226  1.339179  1.339141  1.339111  1.339089  1.339074  1.339067  1.339068  1.339076  1.339093  1.339117  1.339149   1.33919  1.339239  1.339298  1.339365  1.339443  1.339532  1.339632  1.339744  1.339869  1.340007  1.340161   1.34033  1.340517  1.340723  1.340949  1.341198   1.34147  1.341769  1.342096  1.342454  1.342845          1.343272 |  0.71061  0.714767  0.718589  0.722104  0.725337  0.728309  0.731038  0.733542  0.735835  0.737929  0.739835  0.741564  0.743123   0.74452  0.745761  0.746851  0.747796  0.748597   0.74926  0.749785  0.750175  0.750431  0.750553   0.75054  0.750393   0.75011  0.749689  0.749127  0.748422  0.747569  0.746566  0.745406  0.744084  0.742594  0.740929  0.739082  0.737044  0.734805  0.732358  0.729689   0.72679  0.723647  0.720248  0.716582  0.712633  0.708391  0.703842  0.698975   0.69378       0.68825 | -0.408997  -0.413583  -0.417794  -0.421663  -0.425212  -0.428466  -0.431444  -0.434162  -0.436635  -0.438876  -0.440897  -0.442707  -0.444315  -0.445728  -0.446951  -0.447991  -0.448851  -0.449535  -0.450045  -0.450383   -0.45055  -0.450547  -0.450372  -0.450025  -0.449504  -0.448807  -0.447931  -0.446871  -0.445625  -0.444185  -0.442548  -0.440706  -0.438653  -0.436379  -0.433878  -0.431139  -0.428153  -0.424909  -0.421395  -0.417601  -0.413513   -0.40912   -0.40441   -0.39937   -0.39399  -0.388258  -0.382169  -0.375715  -0.368895     -0.361714 |   0.6936  0.694313  0.694971  0.695577  0.696133  0.696642  0.697105  0.697525  0.697903  0.698241  0.698538  0.698797  0.699018  0.699203   0.69935  0.699462  0.699538  0.699579  0.699584  0.699554  0.699488  0.699387  0.699251  0.699078  0.698869  0.698624   0.69834   0.69802   0.69766  0.697262  0.696823  0.696344  0.695824  0.695262  0.694656  0.694008  0.693315  0.692578  0.691796  0.690968  0.690096  0.689179  0.688219  0.687216  0.686172  0.685091  0.683976   0.68283  0.681659      0.680471 |  453.86844 | 464.947556   486.82638  508.121027  528.772653  548.726477  567.931339  586.339343  603.905565  620.587817  636.346449  651.144195  664.946051  677.719174  689.432825   700.05831   709.56897  717.940172  725.149328   731.17594  736.001651  739.610331  741.988171  743.123803  743.008443  741.636049  739.003504  735.110817  729.961351  723.562066  715.923782   707.06147  696.994545   685.74719  673.348683  659.833732  645.242819   629.62253  613.025887  595.512643   577.14955  558.010582  538.177084  517.737835  496.789013  475.434009  453.783092  431.952872  410.065544  388.247872          366.629891 | 355.876475 | 1.072449  1.085076  1.096958  1.108126  1.118607  1.128425  1.137601  1.146157   1.15411  1.161474  1.168265  1.174494  1.180173  1.185309  1.189912  1.193987   1.19754  1.200574  1.203093  1.205099  1.206593  1.207574  1.208042  1.207994  1.207429  1.206342   1.20473  1.202587  1.199908  1.196687  1.192916   1.18859    1.1837  1.178239  1.172199  1.165571  1.158349  1.150525  1.142093  1.133048  1.123385  1.113102  1.102201  1.090683  1.078556  1.065831  1.052525   1.03866  1.024269       1.009389 | 4.838645  4.926743  5.010091  5.088817   5.16304  5.232867  5.298397  5.359718  5.416909  5.470041  5.519175  5.564365  5.605658  5.643092  5.676701  5.706509  5.732536  5.754795  5.773296   5.78804  5.799026  5.806246  5.809689  5.809339  5.805177  5.797181  5.785323  5.769575  5.749906  5.726284  5.698674  5.667043  5.631357  5.591585  5.547696  5.499666  5.447474  5.391107  5.330563  5.265849  5.196987  5.124018  5.047001  4.966021  4.881194  4.792666  4.700626  4.605306  4.506988        4.406014 |  453.86844  476.026352  497.630449  518.619751  538.937579   558.53107  577.350786   595.35039  612.486387  628.717904  644.006524  658.316141  671.612857  683.864895  695.042549  705.118148  714.066042  721.862615  728.486311  733.917684  738.139466  741.136658  742.896637  743.409286  742.667144  740.665579  737.402978  732.880961  727.104619  720.082764   711.82821  702.358061  691.694025  679.862735  666.896088  652.831578  637.712641  621.588988  604.516917  586.559602  567.787345   548.27776  528.115891  507.394231   486.21262  464.677993  442.903953  421.010133  399.121322  377.366315  355.876475 |  100.0 |  353.86844 | 355.876475 |
    | 09:00 | 7.244844  7.400779  7.549098  7.689948  7.823469  7.949797  8.069056  8.181365  8.286829  8.385545  8.477599  8.563067  8.642014  8.714497  8.780561  8.840243  8.893572  8.940566  8.981236  9.015586   9.04361  9.065297  9.080627  9.089574  9.092106  9.088187  9.077773  9.060816  9.037266  9.007069  8.970167  8.926503   8.87602  8.818661  8.754373  8.683107  8.604823  8.519489  8.427084  8.327607  8.221072  8.107521  7.987021  7.859677  7.725632   7.58508  7.438269   7.28551  7.127189             6.963774 |  371.11146  384.869291  398.180901   411.02574  423.385415  435.243291  446.584181  457.394101  467.660073   477.36997  486.512391  495.076561  503.052249   510.42971  517.199637  523.353125  528.881651  533.777061  538.031566  541.637752  544.588598  546.877499  548.498305  549.445366  549.713591   549.29851  548.196362  546.404178  543.919892  540.742454  536.871971  532.309848  527.058959  521.123827  514.510824   507.22839  499.287272  490.700773  481.485028  471.659291  461.246235  450.272264  438.767826  426.767734  414.311462  401.443433  388.213262  374.675943  360.891966  346.927321 |  88.883199   90.473428   91.985999   93.422392   94.784047   96.072337    97.28855   98.433875   99.509402  100.516112  101.454884  102.326489  103.131597  103.870777    104.5445  105.153143   105.69699  106.176238  106.590998  106.941299  107.227091  107.448252  107.604585  107.695828  107.721656  107.681685  107.575479  107.402556  107.162393  106.854437  106.478112  106.032828  105.517999  104.933049  104.277437  103.550669  102.752324  101.882079  100.939736   99.925261   98.838816   97.680812   96.451951   95.153289     93.7863   92.352946   90.855757   89.297917   87.683353        86.016835 |  87.448437   89.007785   90.490983   91.899482   93.234695   94.497967   95.690562   96.813646   97.868286   98.855447   99.775988  100.630667  101.420141  102.144967  102.805607   103.40243  103.935716  104.405657  104.812362  105.155861  105.436104   105.65297  105.806267  105.895738  105.921064  105.881869  105.777726  105.608161  105.372662  105.070686  104.701669  104.265032    103.7602   103.18661   102.54373  101.831075  101.048233  100.194887   99.270844   98.276069   97.210723   96.075206   94.870209   93.596766   92.256324   90.850804    89.38269   87.855101   86.271892     84.637736 |  1.57204  1.591344  1.609586  1.626803  1.643033  1.658307  1.672656  1.686107  1.698686  1.710414  1.721311  1.731394  1.740681  1.749183  1.756913   1.76388  1.770093  1.775559  1.780281  1.784265  1.787511  1.790021  1.791794  1.792828  1.793121  1.792668  1.791464  1.789502  1.786776  1.783277  1.778996  1.773924   1.76805  1.761362  1.753851  1.745504  1.736309  1.726257  1.715336  1.703536  1.690849   1.67727  1.662793  1.647419  1.631149  1.613994  1.595966  1.577087  1.557387  1.536905 | 1.342053  1.341731  1.341439  1.341176  1.340937  1.340721  1.340525  1.340347  1.340185  1.340039  1.339906  1.339787  1.339679  1.339582  1.339496  1.339419  1.339352  1.339294  1.339244  1.339202  1.339168  1.339142  1.339124  1.339114  1.339111  1.339115  1.339128  1.339148  1.339176  1.339212  1.339257  1.339311  1.339374  1.339447   1.33953  1.339624  1.339729  1.339847  1.339979  1.340124  1.340285  1.340463  1.340659  1.340874  1.341111  1.341371  1.341656  1.341968  1.342309          1.342682 | 0.704432  0.708931  0.713077  0.716899  0.720424  0.723674  0.726669  0.729427  0.731964  0.734294  0.736428  0.738377   0.74015  0.741756  0.743202  0.744493  0.745635  0.746633   0.74749  0.748209  0.748792  0.749241  0.749558  0.749743  0.749795  0.749714  0.749499  0.749149   0.74866  0.748031  0.747257  0.746335   0.74526  0.744028  0.742631  0.741063  0.739318  0.737387  0.735261  0.732932  0.730388   0.72762  0.724616  0.721364  0.717851  0.714064  0.709991   0.70562  0.700938      0.695935 | -0.401855  -0.406837  -0.411425  -0.415653  -0.419546  -0.423128  -0.426422  -0.429444  -0.432212  -0.434739  -0.437037  -0.439118  -0.440991  -0.442664  -0.444145  -0.445438   -0.44655  -0.447484  -0.448244  -0.448831  -0.449249  -0.449498  -0.449579  -0.449491  -0.449233  -0.448804  -0.448201  -0.447423  -0.446464  -0.445322  -0.443991  -0.442466  -0.440741  -0.438809  -0.436663  -0.434294  -0.431693  -0.428851  -0.425757  -0.422402  -0.418772  -0.414857  -0.410644  -0.406122  -0.401278    -0.3961  -0.390579  -0.384704   -0.37847     -0.371873 | 0.692403  0.693173  0.693888  0.694549  0.695161  0.695724  0.696241  0.696714  0.697145  0.697534  0.697884  0.698195  0.698468  0.698704  0.698903  0.699067  0.699195  0.699288  0.699346  0.699369  0.699358  0.699312  0.699231  0.699114  0.698963  0.698775  0.698551  0.698291  0.697993  0.697658  0.697284   0.69687  0.696417  0.695923  0.695388   0.69481   0.69419  0.693526  0.692818  0.692065  0.691268  0.690426  0.689539  0.688609  0.687635   0.68662  0.685566  0.684476  0.683354      0.682204 | 423.717566 | 434.708461  456.469778  477.775127  498.561091  518.768599  538.342433  557.230822  575.385107  592.759469  609.310703  624.998038  639.782987  653.629231  666.502527  678.370644  689.203318  698.972231  707.651007  715.215231  721.642484  726.912398  731.006731  733.909462    735.6069  736.087822  735.343621  733.368485  730.159588  725.717307  720.045455  713.151544  705.047051  695.747719  685.273864  673.650698  660.908662  647.083768  632.217931  616.359309  599.562609  581.889377  563.408245  544.195119  524.333294  503.913472  483.033663  461.798935  440.320992  418.717547          397.111453 | 386.319679 | 1.054232  1.067435  1.079905  1.091671  1.102758  1.113189  1.122986  1.132167  1.140751  1.148752  1.156185  1.163062  1.169394  1.175191   1.18046  1.185209  1.189444  1.193168  1.196387  1.199101  1.201313  1.203023  1.204231  1.204935  1.205135  1.204826  1.204006  1.202669  1.200812  1.198428  1.195511  1.192054  1.188051  1.183493  1.178373  1.172682  1.166413  1.159558  1.152109   1.14406  1.135403  1.126135  1.116252  1.105754  1.094641  1.082918  1.070595  1.057684  1.044207       1.030188 | 4.712406  4.803796  4.890611  4.972955  5.050931  5.124633   5.19415  5.259562  5.320943  5.378357  5.431864  5.481516  5.527356  5.569424  5.607751  5.642363   5.67328  5.700518  5.724085  5.743985  5.760218  5.772778  5.781655  5.786836  5.788303  5.786033  5.780002  5.770183  5.756543  5.739051  5.717671  5.692368  5.663105  5.629848  5.592559  5.551208  5.505764  5.456202  5.402506  5.344664  5.282677  5.216559  5.146337  5.072062  4.993802  4.911655   4.82575  4.736252  4.643366        4.547347 | 423.717566  445.692732  467.244659  488.307632  508.820556   528.72641  547.971801  566.506595  584.283621  601.258423  617.389061   632.63594  646.961688  660.331044  672.710787   684.06968  694.378435  703.609704  711.738085  718.740149  724.594485  729.281766  732.784831  735.088793  736.181153  736.051953  734.693933  732.102716  728.277015  723.218858  716.933832  709.431351  700.724941   690.83254  679.776819  667.585507  654.291731  639.934358  624.558326  608.214976  590.962351  572.865469  553.996545  534.435145  514.268265  493.590292  472.502842  451.114431   429.53996  407.899975  386.319679 |  100.0 | 323.717566 | 386.319679 |
    | 09:30 | 7.030648  7.190997  7.344133  7.490152  7.629154   7.76124  7.886511  8.005064  8.116988  8.222366  8.321274   8.41378  8.499943  8.579815  8.653437  8.720845  8.782066  8.837117  8.886011   8.92875  8.965331  8.995743  9.019969  9.037983  9.049757  9.055254  9.054433  9.047247  9.033646  9.013575  8.986975  8.953788  8.913951  8.867403  8.814083  8.753932  8.686896  8.612928  8.531987  8.444044  8.349086  8.247115  8.138156   8.02226  7.899512  7.770029  7.633978  7.491574  7.343091             7.188874 | 352.609816  366.417127  379.843479  392.864181  405.457231  417.602827  429.282979    440.4812  451.182264  461.372001  471.037147  480.165204  488.744347  496.763331  504.211433  511.078393  517.354381  523.029969  528.096111  532.544141  536.365772  539.553114  542.098689  543.995465  545.236898  545.816981  545.730306  544.972137  543.538494   541.42625  538.633243    535.1584  531.001877  526.165213  520.651505  514.465595  507.614282  500.106541  491.953771  483.170054  473.772435  463.781209  453.220231  442.117221  430.504083  418.417208  405.897769   392.99198  379.751321  366.232683 |  86.698827   88.334066   89.895759   91.384864   92.802407   94.149428   95.426951   96.635954   97.777358   98.852009    99.86068   100.80406  101.682755  102.497286   103.24809  103.935521  104.559849  105.121267  105.619886  106.055742  106.428796  106.738939  106.985991  107.169707  107.289778  107.345838  107.337464  107.264182  107.125475  106.920785  106.649524  106.311078   105.90482  105.430121  104.886359  104.272939  103.589308  102.834974  102.009532   101.11269  100.144301   99.104396    97.99323   96.811325   95.559527   94.239061   92.851605    91.39936   89.885132        88.312421 |  85.306485   86.909968   88.441334   89.901522   91.291536   92.612399   93.865114   95.050639   96.169877    97.22366   98.212743   99.137803   99.999435  100.798148  101.534372  102.208453  102.820658  103.371173  103.860109  104.287501  104.653311  104.957431  105.199685  105.379833  105.497573  105.552544  105.544333  105.472474   105.33646  105.135746  104.869752  104.537878   104.13951  103.674029  103.140827  102.539319  101.868963  101.129278  100.319865    99.44044   98.490856   97.471145   96.381557   95.222605   93.995115   92.700292    91.33978   89.915736   88.430913     86.888743 | 1.545305  1.565343  1.584347   1.60235  1.619384   1.63548  1.650665  1.664965  1.678404  1.691004  1.702783   1.71376   1.72395  1.733367  1.742022  1.749926  1.757089  1.763516  1.769213  1.774185  1.778435  1.781964  1.784772  1.786859  1.788222  1.788859  1.788764  1.787932  1.786357  1.784031  1.780947  1.777095  1.772464  1.767046  1.760828  1.753799  1.745948  1.737262  1.727731  1.717343  1.706087  1.693955  1.680939  1.667034  1.652236  1.636548  1.619974  1.602525  1.584218  1.565079 | 1.342527  1.342169  1.341846  1.341553  1.341288  1.341047  1.340828  1.340629  1.340448  1.340283  1.340134  1.339998  1.339875  1.339764  1.339664  1.339574  1.339494  1.339423  1.339361  1.339308  1.339263  1.339226  1.339197  1.339175  1.339161  1.339154  1.339155  1.339164   1.33918  1.339204  1.339237  1.339277  1.339327  1.339385  1.339453   1.33953  1.339619  1.339718   1.33983  1.339954  1.340093  1.340246  1.340415  1.340601  1.340806  1.341032  1.341279  1.341551  1.341848          1.342174 | 0.698004  0.702844  0.707314  0.711445  0.715263  0.718793  0.722056   0.72507  0.727853   0.73042  0.732782  0.734952   0.73694  0.738755  0.740405  0.741896  0.743234  0.744426  0.745474  0.746383  0.747155  0.747794    0.7483  0.748675   0.74892  0.749034  0.749017  0.748868  0.748585  0.748167   0.74761  0.746912  0.746069  0.745076  0.743929  0.742621  0.741147    0.7395  0.737671  0.735654  0.733438  0.731015  0.728373  0.725502   0.72239  0.719025  0.715394  0.711484  0.707284      0.702781 | -0.394414   -0.39979  -0.404755  -0.409341  -0.413577   -0.41749  -0.421099  -0.424427  -0.427489    -0.4303  -0.432876  -0.435227  -0.437363  -0.439295  -0.441029  -0.442573  -0.443932  -0.445111  -0.446115  -0.446947  -0.447609  -0.448103   -0.44843  -0.448591  -0.448586  -0.448413  -0.448072  -0.447561  -0.446876  -0.446014  -0.444973  -0.443746   -0.44233  -0.440717  -0.438902  -0.436878  -0.434636  -0.432168  -0.429464  -0.426516  -0.423312  -0.419842  -0.416093  -0.412054  -0.407713  -0.403058  -0.398077  -0.392759  -0.387094     -0.381075 |  0.69117  0.691995  0.692764  0.693479  0.694143  0.694758  0.695327  0.695852  0.696333  0.696773  0.697173  0.697534  0.697857  0.698142  0.698392  0.698606  0.698784  0.698928  0.699037  0.699111  0.699152  0.699158   0.69913  0.699068  0.698971  0.698839  0.698672   0.69847  0.698231  0.697956  0.697643  0.697293  0.696904  0.696475  0.696007  0.695498  0.694947  0.694354  0.693718  0.693039  0.692315  0.691547  0.690734  0.689877  0.688976  0.688032  0.687046  0.686019  0.684956      0.683858 | 395.049792 | 405.868877  427.344437  448.489494  469.236568  489.522794  509.289365  528.481087  547.046007   564.93511  582.102064  598.503016  614.096419  628.842897  642.705134  655.647791  667.637441  678.642533  688.633365  697.582088  705.462717  712.251169  717.925312  722.465038  725.852352  728.071478  729.108985  728.953937  727.598057  725.035912   721.26512  716.286581  710.104719   702.72775  694.167971  684.442056  673.571376  661.582328  648.506672  634.381865  619.251403  603.165143    586.1796  568.358223   549.77161   530.49767   510.62169  490.236308  469.441342  448.343473          427.055729 | 416.381526 | 1.035938  1.049651   1.06265  1.074959  1.086602  1.097599  1.107971  1.117735   1.12691  1.135509  1.143546  1.151035  1.157985  1.164407  1.170309  1.175698   1.18058  1.184961  1.188844  1.192233  1.195129  1.197533  1.199447  1.200869  1.201797  1.202231  1.202166    1.2016  1.200526  1.198942   1.19684  1.194215   1.19106  1.187367  1.183129  1.178338  1.172985  1.167063  1.160564  1.153478    1.1458  1.137523   1.12864  1.119148  1.109044  1.098328  1.087004  1.075078  1.062561        1.04947 | 4.586659  4.680818  4.770612  4.856121  4.937424  5.014599   5.08772  5.156856  5.222073   5.28343  5.340981  5.394773  5.444847  5.491242  5.533987  5.573107  5.608624   5.64055  5.668898   5.69367  5.714869  5.732489  5.746523  5.756958  5.763778  5.766962  5.766486  5.762324  5.754446  5.742819   5.72741   5.70818  5.685093   5.65811  5.627193  5.592303  5.553407  5.510469  5.463463  5.412363  5.357156  5.297834  5.234402  5.166879  5.095304  5.019731  4.940244  4.856953  4.770001        4.679572 | 395.049792  416.675449  438.005432  458.969823  479.503612  499.546099  519.040394  537.933015  556.173553  573.714397  590.510507  606.519224  621.700121   636.01488  649.427193  661.902691  673.408895  683.915181  693.392773  701.814747  709.156058  715.393582  720.506178  724.474767  727.282431  728.914532  729.358841  728.605705  726.648213  723.482398  719.107454   713.52597  706.744189  698.772285  689.624655   679.32023  667.882798   655.34134  641.730366  627.090255  611.467587  594.915462  577.493783   559.26951  540.316851  520.717384  500.560084  479.941229  458.964169  437.738914  416.381526 |  100.0 | 295.049792 | 416.381526 |
    | 10:00 | 6.818894  6.982699  7.139767  7.290128  7.433834  7.570947  7.701535  7.825667  7.943414  8.054841  8.160011   8.25898  8.351799  8.438512  8.519157  8.593765   8.66236  8.724958   8.78157    8.8322  8.876843   8.91549  8.948124  8.974723  8.995257  9.009693   9.01799  9.020103  9.015983  9.005576  8.988823  8.965664  8.936035  8.899871  8.857107  8.807678   8.75152  8.688574  8.618785  8.542108  8.458503  8.367947  8.270432  8.165968  8.054591  7.936364  7.811386  7.679796  7.541779             7.397577 | 334.769979  348.530953  361.977874  375.081784  387.816976  400.160402  412.091209  423.590365  434.640363  445.224978  455.329076  464.938448  474.039687  482.620078  490.667517  498.170436  505.117754  511.498835  517.303455  522.521784  527.144376  531.162168  534.566488  537.349073  539.502096  541.018198   541.89054  542.112854  541.679512  540.585601  538.827019  536.400571  533.304092  529.536574  525.098316  519.991082  514.218285  507.785183  500.699091  492.969616  484.608908  475.631925  466.056714  455.904713  445.201049  433.974853  422.259566  410.093241  397.518812  384.584345 |  84.539346   86.209842   87.811623   89.345014   90.810532   92.208814   93.540552   94.806462   96.007248   97.143588   98.216112     99.2254  100.171969  101.056272  101.878694  102.639548   103.33908  103.977463  104.554798   105.07112  105.526394  105.920516  106.253319  106.524572  106.733982  106.881197  106.965811  106.987364  106.945348  106.839212  106.668366  106.432187   106.13003  105.761231  105.325122  104.821039   104.24834  103.606415   102.89471  102.112746  101.260144  100.336654   99.342189   98.276863   97.141034   95.935353   94.660821   93.318856   91.911354        90.440775 |  83.188939   84.826995   86.397671   87.901284   89.338343   90.709471   92.015348   93.256674   94.434142   95.548415   96.600111   97.589799   98.517987   99.385117  100.191568  100.937648  101.623595   102.24958  102.815705     103.322  103.768432  104.154901  104.481241  104.747227   104.95257  105.096927  105.179897  105.201032  105.159832  105.055757  104.888228  104.656636  104.360346  103.998709  103.571069  103.076775  102.515197  101.885738  101.187854  100.421075    99.58503   98.679474   97.704321   96.659683    95.54591   94.363643   93.113862   91.797956   90.417787     88.975766 | 1.518618  1.539285  1.558957  1.577659   1.59542  1.612263  1.628215  1.643299  1.657537  1.670949  1.683554  1.695369  1.706409  1.716688  1.726218  1.735008  1.743069  1.750408  1.757031  1.762942  1.768146  1.772643  1.776437  1.779525  1.781908  1.783581  1.784543  1.784788  1.784311  1.783104  1.781161  1.778474  1.775032  1.770826  1.765846   1.76008  1.753517  1.746145  1.737951  1.728924  1.719053  1.708326  1.696734  1.684267  1.670919  1.656687  1.641568  1.625566  1.608688  1.590949 | 1.343032  1.342638  1.342281  1.341958  1.341664  1.341398  1.341155  1.340933  1.340732  1.340548   1.34038  1.340227  1.340088  1.339962  1.339848  1.339745  1.339652  1.339568  1.339494  1.339429  1.339373  1.339325  1.339284  1.339252  1.339227  1.339209  1.339199  1.339197  1.339201  1.339214  1.339234  1.339263  1.339299  1.339344  1.339398  1.339461  1.339534  1.339617   1.33971  1.339816  1.339934  1.340065   1.34021  1.340371  1.340548  1.340744  1.340958  1.341194  1.341453          1.341737 | 0.691346  0.696523  0.701316  0.705754  0.709866  0.713677  0.717209  0.720481  0.723512  0.726316  0.728907  0.731299  0.733502  0.735526  0.737379  0.739069  0.740603  0.741986  0.743224   0.74432  0.745278  0.746102  0.746793  0.747353  0.747784  0.748086  0.748259  0.748303  0.748217     0.748  0.747649  0.747162  0.746537  0.745769  0.744855   0.74379  0.742568  0.741184  0.739631  0.737902  0.735988  0.733882  0.731573  0.729053  0.726309  0.723332  0.720109  0.716627  0.712876       0.70884 | -0.386701  -0.392466  -0.397803  -0.402746  -0.407325  -0.411566  -0.415492  -0.419123  -0.422479  -0.425576  -0.428427  -0.431047  -0.433445  -0.435633  -0.437619   -0.43941  -0.441013  -0.442434  -0.443677  -0.444747  -0.445647   -0.44638  -0.446946  -0.447348  -0.447586  -0.447661   -0.44757  -0.447313  -0.446889  -0.446294  -0.445526  -0.444581  -0.443454  -0.442142  -0.440637  -0.438935  -0.437028  -0.434909  -0.432568  -0.429999  -0.427191  -0.424133  -0.420816  -0.417227  -0.413356   -0.40919  -0.404718  -0.399927  -0.394806     -0.389345 |  0.68991  0.690786  0.691606  0.692371  0.693085   0.69375  0.694369  0.694942  0.695473  0.695961  0.696409  0.696819   0.69719  0.697523  0.697821  0.698083   0.69831  0.698502   0.69866  0.698784  0.698875  0.698932  0.698955  0.698944    0.6989  0.698821  0.698709  0.698562  0.698379  0.698162  0.697908  0.697617   0.69729  0.696924   0.69652  0.696076  0.695592  0.695067  0.694501  0.693892  0.693241  0.692545  0.691806  0.691022  0.690195  0.689323  0.688408   0.68745  0.686451      0.685414 | 367.963873 | 378.537392  399.578015    420.4096  440.961174  461.166598  480.963975  500.295162  519.105367  537.342819   554.95849  571.905868  588.140766  603.621169  618.307106  632.160548  645.145333  657.227108   668.37329  678.553052  687.737313  695.898764  703.011891  709.053036  714.000452  717.834398  720.537238  722.093559  722.490317   721.71699  719.765759   716.63171  712.313046  706.811329  700.131738  692.283344  683.279405  673.137676  661.880731  649.536297   636.13759  621.723655  606.339697   590.03739   572.87517  554.918481  536.239965  516.919583  497.044644  476.709711          456.016366 | 445.601377 | 1.017665  1.031817  1.045281  1.058076  1.070221  1.081735  1.092636   1.10294  1.112664  1.121821  1.130425  1.138488   1.14602  1.153032  1.159532  1.165526  1.171023  1.176026  1.180541   1.18457  1.188116  1.191182  1.193767  1.195871  1.197495  1.198635  1.199291  1.199457  1.199132   1.19831  1.196986  1.195155   1.19281  1.189943  1.186549  1.182619  1.178145  1.173119  1.167533  1.161378  1.154645  1.147328  1.139419  1.130911    1.1218  1.112083  1.101757  1.090826  1.079292       1.067164 |  4.46209  4.558475   4.65075  4.738959  4.823154   4.90339  4.979725  5.052214  5.120911  5.185868  5.247129  5.304739  5.358734  5.409148   5.45601  5.499343  5.539166  5.575494  5.608336  5.637699  5.663583  5.685985  5.704898  5.720311  5.732208  5.740571  5.745377  5.746601  5.744215  5.738186   5.72848  5.715061  5.697892  5.676932   5.65214  5.623478  5.590904   5.55438   5.51387  5.469342  5.420767  5.368124  5.311402  5.250598  5.185721  5.116799  5.043877  4.967022  4.886329        4.801921 | 367.963873  389.092965  410.049659   430.76042  451.156867  471.175126  490.755297  509.841011  528.379068  546.319137  563.613505   580.21687  596.086168  611.180438  625.460704  638.889891  651.432753   663.05583  673.727421  683.417571  692.098077  699.742517   706.32629  711.826673  716.222897  719.496241  721.630148  722.610346  722.425006  721.064908  718.523625  714.797739  709.887063  703.794894  696.528275  688.098286  678.520344   667.81452  656.005866  643.124756  629.207223  614.295294  598.437318  581.688267  564.110008  545.771529  526.749097  507.126342  486.994228  466.450901  445.601377 |  100.0 | 267.963873 | 445.601377 |
    | 10:30 | 6.610682   6.77697  6.937067  7.090931  7.238555  7.379952  7.515153  7.644195  7.767124  7.883984  7.994819  8.099674  8.198588  8.291595  8.378728  8.460011  8.535464  8.605101   8.66893  8.726953  8.779166  8.825559  8.866117  8.900816  8.929631  8.952528  8.969468  8.980408  8.985299  8.984089   8.97672  8.963131  8.943259  8.917037  8.884396  8.845269  8.799585  8.747278  8.688283   8.62254  8.549995  8.470604  8.384333  8.291163  8.191093  8.084142  7.970358  7.849818  7.722638             7.588977 |  317.66581  331.291187  344.670508  357.770495  370.561682  383.017729  395.114886  406.831554  418.147931  429.045733  439.507959  449.518704     459.063   468.12669  476.696324  484.759072  492.302655  499.315289  505.785641  511.702797  517.056243  521.835847  526.031857   529.63491  532.636038  535.026699  536.798802  537.944752  538.457499  538.330598  537.558286   536.13556  534.058276  531.323257   527.92842  523.872908  519.157247  513.783515   507.75553  501.079053   493.76201  485.814733  477.250213  468.084378  458.336376  448.028875   437.18837  425.845486  414.035285  401.797551 |  82.415993   84.111808   85.744481   87.313593   88.819062   90.261038    91.63982   92.955804   94.209433   95.401172   96.531478   97.600792   98.609516   99.558012  100.446594   101.27552  102.044992  102.755152  103.406082  103.997803  104.530275  105.003395  105.417002  105.770873  106.064728  106.298229  106.470985  106.582551  106.632432  106.620089  106.544939  106.406361  106.203703  105.936287  105.603419  105.204394   104.73851  104.205081  103.603447  102.932997  102.193184  101.383551  100.503757   99.553605   98.533082   97.442394   96.282018   95.052751   93.755767        92.392688 |   81.10682   82.769703   84.370671   85.909312   87.385546   88.799519   90.151526   91.441955   92.671239   93.839835   94.948192    95.99674   96.985875   97.915953   98.787279   99.600108  100.354637  101.051006  101.689295  102.269526  102.791658   103.25559  103.661165  104.008164  104.296312  104.525279   104.69468  104.804079  104.852992  104.840889  104.767198  104.631311  104.432588  104.170366  103.843962  103.452686  102.995849  102.472778  101.882828  101.225397  100.499951   99.706041   98.843332   97.911631   96.910926   95.841419   94.703576    93.49818   92.226383     90.889774 | 1.492117  1.513303  1.533544  1.552855  1.571258  1.588773  1.605421  1.621222  1.636195  1.650359  1.663732  1.676329  1.688165  1.699253  1.709605  1.719231  1.728141  1.736342   1.74384  1.750642   1.75675  1.762167  1.766896  1.770936  1.774288  1.776948  1.778915  1.780185  1.780753  1.780612  1.779757   1.77818  1.775871  1.772823  1.769025  1.764466  1.759135  1.753021  1.746111  1.738392  1.729854  1.720484   1.71027  1.699202   1.68727  1.674466  1.660786  1.646225  1.630785  1.614471 | 1.343568  1.343136  1.342745   1.34239  1.342067  1.341773  1.341505   1.34126  1.341037  1.340833  1.340646  1.340476   1.34032  1.340178  1.340049  1.339931  1.339825  1.339729  1.339643  1.339566  1.339498  1.339438  1.339386  1.339343  1.339307  1.339279  1.339258  1.339245  1.339239   1.33924  1.339249  1.339266   1.33929  1.339323  1.339363  1.339413  1.339471  1.339539  1.339617  1.339705  1.339805  1.339916  1.340041  1.340179  1.340332  1.340501  1.340687  1.340891  1.341116          1.341363 | 0.684485  0.689991    0.6951  0.699843  0.704248  0.708339  0.712139   0.71567  0.718948  0.721991  0.724813  0.727427  0.729845  0.732078  0.734134  0.736023  0.737751  0.739324  0.740749   0.74203  0.743171  0.744176  0.745048   0.74579  0.746402  0.746886  0.747242  0.747472  0.747575   0.74755  0.747395  0.747109   0.74669  0.746134  0.745439  0.744601  0.743615  0.742475  0.741178  0.739715  0.738081  0.736267  0.734265  0.732067  0.729663  0.727043  0.724195  0.721108  0.717771       0.71417 | -0.378745   -0.38489  -0.390594   -0.39589  -0.400808  -0.405375  -0.409615   -0.41355  -0.417199  -0.420579  -0.423705  -0.426592  -0.429251  -0.431693  -0.433928  -0.435964  -0.437808  -0.439467  -0.440945  -0.442249  -0.443382  -0.444347  -0.445146  -0.445782  -0.446255  -0.446567  -0.446717  -0.446705  -0.446529  -0.446188   -0.44568  -0.445002  -0.444149  -0.443119  -0.441907  -0.440507  -0.438913  -0.437119  -0.435117    -0.4329  -0.430459  -0.427786  -0.424869  -0.421699  -0.418265  -0.414555  -0.410559  -0.406262  -0.401655     -0.396726 | 0.688631  0.689553   0.69042  0.691233  0.691994  0.692706  0.693372  0.693992  0.694569  0.695104  0.695599  0.696054  0.696472  0.696852  0.697196  0.697504  0.697778  0.698017  0.698222  0.698394  0.698532  0.698637  0.698709  0.698748  0.698754  0.698727  0.698666  0.698572  0.698444  0.698281  0.698083   0.69785  0.697581  0.697275  0.696932  0.696551  0.696131  0.695672  0.695173  0.694632   0.69405  0.693425  0.692757  0.692046  0.691291  0.690492  0.689649  0.688762  0.687833      0.686862 | 342.523746 | 352.787857   373.26364  393.646724   413.86326  433.844397  453.525667  472.846465  491.749627  510.181072  528.089508  545.426186  562.144694  578.200786  593.552243  608.158756  621.981836  634.984744  647.132439  658.391545  668.730334  678.118725  686.528304  693.932352  700.305894  705.625766  709.870694  713.021392  715.060678  715.973612   715.74764  714.372771  711.841769  708.150356   703.29745  697.285407  690.120295  681.812174  672.375403  661.828954  650.196741  637.507957  623.797408  609.105852  593.480319  576.974417  559.648597  541.570377  522.814505  503.463032          483.605293 | 473.574309 | 0.999507  1.014024  1.027886  1.041106  1.053698  1.065677  1.077058  1.087857  1.098087  1.107762  1.116893  1.125493  1.133571  1.141138    1.1482  1.154767  1.160843  1.166436  1.171548  1.176185  1.180349  1.184042  1.187265  1.190018  1.192302  1.194115  1.195456  1.196321  1.196708  1.196612  1.196029  1.194954  1.193382  1.191304  1.188716  1.185609  1.181975  1.177807  1.173096  1.167834  1.162012  1.155621  1.148654  1.141103   1.13296  1.124222  1.114882  1.104938  1.094392       1.083244 | 4.339339  4.437396  4.531641  4.622075  4.708717  4.791598  4.870751  4.946217  5.018035  5.086245  5.150884  5.211988  5.269589  5.323716  5.374394  5.421643  5.465483  5.505925  5.542979  5.576651  5.606942  5.633848  5.657364   5.67748  5.694181   5.70745  5.717266  5.723605  5.726439  5.725737  5.721468  5.713594  5.702078  5.686881  5.667961  5.645277  5.618785  5.588443  5.554211   5.51605  5.473923  5.427799  5.377652  5.323464  5.265225  5.202939  5.136622  5.066309  4.992053        4.913934 | 342.523746  363.029072  383.479832  403.799517   423.91696  443.765653  463.283184  482.410773  501.092886  519.276907  536.912879  553.953272  570.352797  586.068258  601.058419  615.283903  628.707112  641.292169  653.004873  663.812673  673.684663  682.591591  690.505877  697.401661  703.254854  708.043211  711.746424  714.346226  715.826515    716.1735   715.37586  713.424925  710.314878  706.042971   700.60977  694.019409  686.279868  677.403271   667.40619  656.309973  644.141076    630.9314   616.71863  601.546566  585.465438  568.532198  550.810766  532.372229  513.294965  493.664677  473.574309 |  100.0 | 242.523746 | 473.574309 |
    | 11:00 | 6.407028  6.574813  6.737027  6.893544   7.04429  7.189224  7.328329  7.461609  7.589075  7.710748  7.826654  7.936818  8.041266  8.140023   8.23311  8.320546  8.402343  8.478513   8.54906  8.613983  8.673278  8.726933  8.774932  8.817255  8.853874  8.884757  8.909867  8.929162  8.942595  8.950114  8.951663  8.947183   8.93661  8.919876  8.896914  8.867651  8.832016  8.789936  8.741339  8.686158  8.624327  8.555787  8.480484  8.398379  8.309442   8.21366   8.11104  8.001613   7.88544             7.762612 |  301.35547  314.763031   327.99305  341.007944   353.77449  366.263059  378.446998  390.302138  401.806392  412.939423  423.682382  434.017683  443.928826  453.400244  462.417179  470.965583  479.032027  486.603637  493.668036  500.213299  506.227922  511.700794  516.621189  520.978752  524.763505  527.965861  530.576636  532.587083  533.988927  534.774408  534.936343   534.46819  533.364121  531.619122  529.229086   526.19093  522.502732  518.163865  513.175168  507.539114  501.260012  494.344212  486.800341  478.639543  469.875744  460.525934   450.61045  440.153284  429.182383  417.729951 |  80.339123     82.0502   83.704462   85.300636   86.837948   88.315986   89.734589   91.093775    92.39368   93.634512   94.816524   95.939981   97.005149   98.012275    98.96158   99.853249  100.687426  101.464209  102.183648  102.845739  103.450427  103.997603  104.487104   104.91871   105.29215  105.607096   105.86317  106.059941   106.19693  106.273611  106.289413  106.243724  106.135896  105.965248  105.731074   105.43265   105.06924  104.640106  104.144521  103.581782  102.951226  102.252245  101.484311  100.646998   99.740012   98.763223   97.716703   96.600766   95.416021        94.163425 |  79.070281   80.748131   82.370268   83.935445   85.442904   86.892239   88.283294   89.616085   90.890747   92.107483   93.266541   94.368181   95.412664   96.400233   97.331103   98.205456   99.023434   99.785133    100.4906  101.139834   101.73278   102.26933  102.749325   103.17255  103.538737  103.847567  104.098668  104.291618  104.425948   104.50114  104.516634  104.471833  104.366098  104.198764  103.969138  103.676509  103.320156  102.899355  102.413394  101.861584  101.243273  100.557865   99.804844   98.983791   98.094418   97.136598     96.1104   95.016134   93.854396     92.626125 | 1.465935  1.487525   1.50823  1.528056  1.547016  1.565122  1.582391  1.598839  1.614483  1.629337  1.643419  1.656741  1.669319  1.681163  1.692285  1.702696  1.712405  1.721419  1.729744  1.737387  1.744351  1.750639  1.756255  1.761198  1.765469  1.769067  1.771989  1.774233  1.775794  1.776668  1.776848  1.776327  1.775099  1.773154  1.770482  1.767075  1.762921  1.758008  1.752326  1.745861  1.738603  1.730537  1.721652  1.711935  1.701376  1.689965  1.677691   1.66455  1.650535  1.635647 | 1.344135  1.343665  1.343238  1.342849  1.342495  1.342173  1.341878  1.341609  1.341363  1.341138  1.340932  1.340743   1.34057  1.340412  1.340267  1.340135  1.340015  1.339905  1.339806  1.339717  1.339637  1.339566  1.339503  1.339449  1.339402  1.339363  1.339332  1.339308  1.339291  1.339282   1.33928  1.339285  1.339298  1.339319  1.339348  1.339385   1.33943  1.339484  1.339547   1.33962  1.339703  1.339797  1.339902   1.34002  1.340151  1.340297  1.340458  1.340635   1.34083          1.341045 | 0.677448   0.68327  0.688687  0.693729  0.698422  0.702791  0.706859  0.710647  0.714173  0.717455  0.720507  0.723344  0.725977  0.728419  0.730678  0.732765  0.734686  0.736449  0.738059  0.739523  0.740845  0.742029  0.743079  0.743997  0.744786  0.745447  0.745982  0.746392  0.746676  0.746835  0.746867  0.746773  0.746549  0.746195  0.745706  0.745081  0.744316  0.743405  0.742345  0.741131  0.739755  0.738212  0.736494  0.734593  0.732502   0.73021  0.727707  0.724984  0.722028      0.718829 | -0.370581  -0.377092  -0.383152  -0.388793  -0.394045  -0.398935  -0.403486  -0.407722  -0.411662  -0.415324  -0.418724  -0.421876  -0.424794  -0.427489   -0.42997  -0.432248   -0.43433  -0.436224  -0.437935  -0.439468  -0.440829  -0.442021  -0.443047   -0.44391  -0.444612  -0.445153  -0.445535  -0.445758  -0.445821  -0.445723  -0.445462  -0.445037  -0.444445  -0.443682  -0.442744  -0.441628  -0.440328  -0.438839  -0.437153  -0.435265  -0.433167   -0.43085  -0.428306  -0.425525  -0.422497  -0.419212  -0.415658  -0.411824  -0.407698     -0.403269 | 0.687339  0.688304  0.689213   0.69007  0.690875  0.691632  0.692341  0.693006  0.693627  0.694207  0.694746  0.695245  0.695707  0.696132   0.69652  0.696873  0.697192  0.697476  0.697726  0.697943  0.698128  0.698279  0.698399  0.698485   0.69854  0.698561   0.69855  0.698507   0.69843  0.698319  0.698175  0.697997  0.697784  0.697535   0.69725  0.696929  0.696571  0.696174  0.695738  0.695263  0.694748  0.694191  0.693593  0.692952  0.692268  0.691541   0.69077  0.689955  0.689097      0.688196 | 318.762374 | 328.663042  348.463205  368.280889  388.040029   407.66966  427.103282  446.278334  465.135749  483.619587  501.676722  519.256585  536.310946  552.793729  568.660861  583.870145  598.381162  612.155183  625.155108  637.345424  648.692171   659.16293  668.726827  677.354547  685.018365  691.692196  697.351655  701.974137  705.538905  708.027211  709.422415  709.710136  708.878413  706.917891  703.822021  699.587281   694.21342   687.70371  680.065232  671.309163  661.451093   650.51134  638.515287  625.493717  611.483152  596.526179  580.671766  563.975548  546.500078  528.315018          509.497263 | 499.956953 | 0.981554  0.996359  1.010548  1.024129  1.037109    1.0495  1.061312  1.072559  1.083252  1.093403  1.103022   1.11212  1.120707  1.128792  1.136383  1.143487   1.15011  1.156259  1.161936  1.167148  1.171896  1.176184  1.180012  1.183381  1.186292  1.188744  1.190736  1.192265  1.193329  1.193924  1.194047  1.193692  1.192855  1.191529  1.189709  1.187387  1.184555  1.181207  1.177333  1.172926  1.167977  1.162477  1.156417   1.14979  1.142586  1.134799  1.126423  1.117452  1.107882       1.097713 | 4.219002  4.318165  4.413858  4.506035  4.594676  4.679777  4.761351  4.839414  4.913991  4.985108   5.05279  5.117065  5.177957   5.23549  5.289684  5.340557  5.388124  5.432396   5.47338  5.511083  5.545503   5.57664  5.604486  5.629033  5.650266  5.668171  5.682726  5.693909  5.701694  5.706051  5.706949  5.704353  5.698225  5.688527  5.675218  5.658254  5.637592  5.613188  5.584998  5.552978  5.517088  5.477287   5.43354  5.385819  5.334098  5.278362  5.218609  5.154844  5.087094          5.0154 | 318.762374  338.536372  358.367162   378.17597  397.889464  417.439067  436.760375  455.792683  474.478578  492.763602  510.595974  527.926347  544.707613  560.894734  576.444609  591.315952   605.46921  618.866484  631.471477   643.24946  654.167243   664.19318  673.297167  681.450676  688.626788  694.800247  699.947535  704.046954  707.078729  709.025129  709.870599  709.601922  708.208387  705.681982   702.01761  697.213312  691.270525  684.194345  675.993815   666.68223  656.277448  644.802225  632.284546  618.757966  604.261941   588.84216  572.550841  555.447004  537.596702  519.073178  499.956953 |  100.0 | 218.762374 | 499.956953 |
    | 11:30 | 6.208858   6.37715  6.540563   6.69888   6.85195  6.999667  7.141967  7.278807  7.410166  7.536036  7.656416  7.771315   7.88074  7.984703  8.083214  8.176284  8.263917  8.346119  8.422889  8.494225  8.560117  8.620555   8.67552  8.724991  8.768942  8.807342  8.840153  8.867336  8.888844  8.904627  8.914631  8.918797  8.917062   8.90936  8.895623  8.875777  8.849751  8.817467  8.778852   8.73383  8.682328  8.624278  8.559612  8.488275  8.410216  8.325398  8.233797   8.13541  8.030252             7.918365 | 285.882459  298.997426  312.003262  324.858191  337.525307  349.971734  362.167954  374.087258  385.705293  396.999698  407.949802  418.536369  428.741399   438.54795  447.939998  456.902319  465.420385  473.480286  481.068659  488.172635  494.779794  500.878129  506.456024  511.502236  516.005887  519.956466  523.343835  526.158245  528.390364  530.031307  531.072678  531.506623  531.325893  530.523912  529.094866  527.033792  524.336694  521.000659  517.023996  512.406391  507.149068  501.254979  494.729004  487.578171  479.811888  471.442204  462.484067  452.955618  442.878473  432.278034 |  78.318176    80.03442   81.700916   83.315443   84.876449   86.382882   87.834057   89.229557   90.569162   91.852785   93.080433   94.252169   95.368092   96.428312   97.432937   98.382059   99.275749  100.114046  100.896954  101.624435   102.29641  102.912752   103.47329  103.977802  104.426017  104.817615  105.152227  105.429435  105.648775  105.809733  105.911753  105.954237  105.936545  105.858004  105.717908  105.515525  105.250103  104.920876  104.527076   104.06794  103.542725  102.950719  102.291261  101.563758  100.767709   99.902731   98.968588   97.965227   96.892819        95.751799 |   77.08858   78.771496   80.405629   81.988803   83.519496   84.996675   86.419668    87.78807    89.10166   90.360356   91.564163   92.713146   93.807398   94.847029   95.832145   96.762836   97.639171   98.461189   99.228893   99.942247  100.601172  101.205546  101.755199  102.249913  102.689424  103.073417  103.401531  103.673357  103.888437  104.046269  104.146308  104.187967  104.170619  104.093603  103.956228  103.757775  103.497507  103.174673   102.78852  102.338301  101.823285  101.242775  100.596124   99.882748   99.102158   98.253977   97.337975   96.354099   95.302515     94.183654 |   1.4402  1.462072  1.483132  1.503375  1.522801  1.541417  1.559231  1.576256  1.592502  1.607984  1.622714  1.636704  1.649967  1.662514  1.674355    1.6855  1.695958  1.705735  1.714839  1.723275  1.731047  1.738159  1.744614  1.750412  1.755555  1.760041   1.76387  1.767038  1.769543   1.77138  1.772544  1.773028  1.772826   1.77193  1.770332  1.768021  1.764989  1.761223  1.756713  1.751447  1.745412  1.738597  1.730987  1.722572  1.713338  1.703273  1.692367   1.68061  1.667995  1.654514 | 1.344732  1.344222  1.343759  1.343336   1.34295  1.342598  1.342276  1.341982  1.341712  1.341464  1.341237  1.341029  1.340838  1.340663  1.340502  1.340355   1.34022  1.340097  1.339985  1.339883  1.339791  1.339708  1.339634  1.339568  1.339511  1.339461  1.339419  1.339385  1.339358  1.339338  1.339326  1.339321  1.339323  1.339332  1.339349  1.339374  1.339407  1.339448  1.339498  1.339557  1.339625  1.339703  1.339792  1.339891  1.340003  1.340128  1.340266  1.340419  1.340588          1.340774 | 0.670267  0.676387    0.6821  0.687431  0.692406  0.697049  0.701382  0.705425  0.709198  0.712717  0.715999  0.719058  0.721907  0.724558   0.72702  0.729304  0.731417  0.733369  0.735164  0.736809   0.73831  0.739671  0.740895  0.741987  0.742949  0.743783  0.744491  0.745075  0.745534  0.745871  0.746083  0.746172  0.746135  0.745971  0.745679  0.745255  0.744697  0.744002  0.743164  0.742181  0.741046  0.739754  0.738298  0.736673  0.734869   0.73288  0.730695  0.728306  0.725702      0.722873 | -0.362243  -0.369102  -0.375504   -0.38148  -0.387058  -0.392264  -0.397122  -0.401656  -0.405884  -0.409825  -0.413497  -0.416913  -0.420087  -0.423033  -0.425759  -0.428277  -0.430594  -0.432719  -0.434659  -0.436418  -0.438003  -0.439417  -0.440665   -0.44175  -0.442673  -0.443438  -0.444044  -0.444493  -0.444785   -0.44492  -0.444897  -0.444713  -0.444368  -0.443858  -0.443181  -0.442333  -0.441309  -0.440106  -0.438716  -0.437136  -0.435357  -0.433373  -0.431175  -0.428756  -0.426106  -0.423215  -0.420073  -0.416669  -0.412993     -0.409031 | 0.686043  0.687045  0.687993  0.688889  0.689734  0.690532  0.691283  0.691989  0.692653  0.693274  0.693855  0.694397  0.694901  0.695368  0.695799  0.696195  0.696556  0.696884  0.697178  0.697439  0.697667  0.697864  0.698028   0.69816   0.69826  0.698329  0.698366   0.69837  0.698343  0.698282  0.698189  0.698063  0.697903  0.697709   0.69748  0.697216  0.696916  0.696579  0.696204  0.695792   0.69534  0.694849  0.694318  0.693745   0.69313  0.692473  0.691773  0.691029  0.690242      0.689412 |  296.68568 | 306.178412  325.210834  344.364012  363.560377  382.727474  401.797335  420.705953  439.392838  457.800636  475.874816    493.5634  510.816735  527.587304  543.829562  559.499806  574.556057  588.957976  602.666787  615.645222  627.857483  639.269217  649.847501  659.560853  668.379243  676.274129  683.218498  689.186931  694.155675  698.102735  701.007978  702.853257  703.622553  703.302128  701.880705  699.349658   695.70323  690.938759  685.056933  678.062059  669.962346   660.77021   650.50259  639.181274  626.833235  613.490967  599.192818  583.983316  567.913464  551.041014          533.430692 | 524.469322 | 0.963895  0.978904  0.993347  1.007222   1.02053  1.033277  1.045469  1.057116  1.068226   1.07881  1.088877  1.098435  1.107494  1.116062  1.124146  1.131753  1.138889   1.14556  1.151771  1.157524  1.162825  1.167675  1.172076  1.176029  1.179534  1.182593  1.185202  1.187362  1.189069  1.190321  1.191114  1.191444  1.191306  1.190696  1.189607  1.188032  1.185965  1.183398  1.180324  1.176734   1.17262  1.167973  1.162784  1.157045  1.150747   1.14388  1.136439  1.128415  1.119804       1.110599 | 4.101626  4.201323  4.297938   4.39137  4.481553   4.56845  4.652042  4.732322  4.809294  4.882969   4.95336  5.020482  5.084352  5.144987  5.202399  5.256604  5.307612  5.355431  5.400068  5.441525  5.479803  5.514898  5.546804  5.575513  5.601011  5.623283  5.642311  5.658071   5.67054  5.679689  5.685487  5.687902  5.686896  5.682432  5.674469  5.662965  5.647875  5.629155  5.606759  5.580641  5.550756  5.517059  5.479509  5.438067    5.3927  5.343379  5.290083  5.232802  5.171537        5.106302 |  296.68568  315.639889  334.754887  353.950392  373.151572  392.288367  411.294916   430.10908  448.672029  466.927898    484.8235  502.308074  519.333083  535.852034  551.820338  567.195181  581.935425  596.001531   609.35549  621.960779  633.782325  644.786492  654.941076  664.215314  672.579909   680.00707  686.470563   691.94578   696.40982  699.841586  702.221905  703.533651  703.761901  702.894097  700.920231  697.833048  693.628274  688.304849  681.865191  674.315476   665.66593  655.931138  645.130367  633.287903   620.43338  606.602122  591.835474  576.181113   559.69334  542.433328  524.469322 |  100.0 |  196.68568 | 524.469322 |
    | 12:00 | 6.017004  6.184817  6.348515  6.507779  6.662372  6.812122  6.956905  7.096631  7.231239  7.360689  7.484952  7.604012   7.71786   7.82649  7.929901   8.02809  8.121055  8.208794  8.291299  8.368564  8.440578  8.507324  8.568786  8.624939  8.675757  8.721209  8.761259  8.795867  8.824989  8.848575  8.866573  8.878925   8.88557  8.886443  8.881476  8.870596   8.85373  8.830803  8.801735  8.766449  8.724868  8.676914  8.622513  8.561597  8.494101  8.419969  8.339156  8.251629  8.157373              8.05639 |  271.27676  284.032089  296.745946  309.372585  321.871563  334.206863  346.346177  358.260315   369.92271  381.309019  392.396786   403.16517  413.594708  423.667125  433.365175  442.672502   451.57353  460.053364  468.097714  475.692826  482.825433  489.482703  495.652213  501.321919   506.48014  511.115549  515.217171  518.774389  521.776959  524.215029  526.079172  527.360424  528.050328  528.140995  527.625171   526.49631  524.748672  522.377415   519.37872  515.749911  511.489604  506.597865  501.076384  494.928667  488.160249  480.778912  472.794938  464.221359  455.074237  445.372947 |  76.361644   78.073009   79.742405   81.366579   82.943127   84.470286   85.946785   87.371719   88.744462    90.06459    91.33183   92.546011   93.707037   94.814856   95.869441   96.870776   97.818839   98.713597   99.554994  100.342947  101.077342  101.758026   102.38481  102.957462  103.475707  103.939228   104.34766  104.700595  104.997579  105.238114  105.421657  105.547624   105.61539  105.624293  105.573634  105.462685  105.290689   105.05687  104.760437  104.400593  103.976542  103.487505  102.932728    102.3115  101.623171  100.867171  100.043037   99.150436   98.189205        97.159379 |  75.170042   76.848174   78.485151   80.077785   81.623717    83.12122   84.569046   85.966309   87.312394   88.606887   89.849517    91.04012   92.178599   93.264905   94.299011   95.280901   96.210553   97.087936   97.912993   98.685645   99.405778  100.073243  100.687856  101.249387  101.757569  102.212088  102.612589   102.95867  103.249887  103.485751  103.665729   103.78925  103.855701   103.86443  103.814755   103.70596  103.537305  103.308026   103.01735  102.664493  102.248678  101.769137  101.225134   100.61597   99.941007   99.199689   98.391558   97.516292   96.573727       95.5639 |  1.41503   1.43706  1.458364  1.478921  1.498721   1.51776  1.536041  1.553568  1.570349  1.586394  1.601711  1.616311  1.630203  1.643399  1.655907  1.667735  1.678891  1.689384  1.699218  1.708399  1.716933  1.724821  1.732068  1.738674  1.744641  1.749969  1.754656  1.758701  1.762101  1.764852  1.766949  1.768388  1.769162  1.769263  1.768685  1.767418  1.765453  1.762779  1.759386  1.755263  1.750398  1.744777  1.738389  1.731221   1.72326  1.714493  1.704908  1.694493  1.683239  1.671135 | 1.345356  1.344808  1.344307  1.343849  1.343431  1.343049  1.342698  1.342377  1.342082  1.341812  1.341563  1.341335  1.341125  1.340932  1.340754  1.340591  1.340442  1.340304  1.340179  1.340064  1.339959  1.339864  1.339779  1.339702  1.339634  1.339573  1.339521  1.339476  1.339439  1.339409  1.339386   1.33937  1.339362  1.339361  1.339367  1.339381  1.339402  1.339431  1.339469  1.339514  1.339569  1.339632  1.339705  1.339789  1.339883  1.339989  1.340107  1.340239  1.340384          1.340545 | 0.662977  0.669373  0.675363  0.680971  0.686219  0.691128   0.69572  0.700016  0.704033  0.707789    0.7113   0.71458  0.717644  0.720503  0.723167  0.725648  0.727954  0.730092  0.732071  0.733896  0.735574  0.737109  0.738506  0.739769    0.7409  0.741904  0.742781  0.743534  0.744164  0.744672  0.745058  0.745323  0.745465  0.745483  0.745377  0.745144  0.744783   0.74429  0.743661  0.742894  0.741984  0.740926  0.739714  0.738343  0.736806  0.735096  0.733204  0.731123  0.728843      0.726354 | -0.353772  -0.360954   -0.36768  -0.373977   -0.37987  -0.385384  -0.390543  -0.395368   -0.39988  -0.404098  -0.408038  -0.411715  -0.415144  -0.418337  -0.421307  -0.424062  -0.426612  -0.428966  -0.431131  -0.433112  -0.434917   -0.43655  -0.438015  -0.439316  -0.440456  -0.441437  -0.442261  -0.442929  -0.443443  -0.443801  -0.444005  -0.444053  -0.443943  -0.443675  -0.443244   -0.44265  -0.441887  -0.440953  -0.439842  -0.438549   -0.43707  -0.435396  -0.433522  -0.431439   -0.42914  -0.426616  -0.423857  -0.420853  -0.417594     -0.414069 | 0.684751  0.685784  0.686765  0.687696  0.688578  0.689413  0.690202  0.690947   0.69165  0.692311  0.692932  0.693514  0.694058  0.694566  0.695038  0.695474  0.695876  0.696245  0.696581  0.696884  0.697155  0.697394  0.697601  0.697777  0.697921  0.698035  0.698117  0.698168  0.698187  0.698175  0.698131  0.698054  0.697945  0.697803  0.697627  0.697417  0.697172  0.696892  0.696576  0.696224  0.695833  0.695405  0.694937   0.69443  0.693882  0.693292  0.692661  0.691987  0.691271      0.690511 | 276.276442 |   285.3259  303.516397  321.923161  340.467883  359.077271  377.682468  396.218543  414.624057  432.840691  450.812929  468.487783  485.814565  502.744691  519.231512  535.230176  550.697508  565.591913  579.873299  593.503006  606.443766  618.659664  630.116117   640.77987  650.618998  659.602927  667.702464   674.88984  681.138772  686.424534   690.72404  694.015952  696.280793  697.501084  697.661494  696.749006  694.753109  691.665997  687.482797  682.201809  675.824766  668.357114  659.808304  650.192103  639.526915  627.836113  615.148375  601.498014  586.925317  571.476844          555.205728 |  546.89363 |  0.94661  0.961739   0.97636   0.99046  1.004033  1.017077  1.029596  1.041594  1.053076  1.064049  1.074522  1.084501  1.093994  1.103008   1.11155  1.119626  1.127242  1.134403  1.141114  1.147378  1.153199  1.158579  1.163521  1.168026  1.172095  1.175727  1.178922  1.181679  1.183996  1.185871  1.187301  1.188282  1.188809  1.188878  1.188484   1.18762  1.186281  1.184459  1.182146  1.179336  1.176019  1.172187  1.167832  1.162944  1.157514  1.151535  1.144996   1.13789  1.130209       1.121948 | 3.987709  4.087367  4.184374   4.27857  4.369839  4.458102  4.543308  4.625422  4.704427  4.780312  4.853078  4.922726  4.989262  5.052694   5.11303  5.170278  5.224443   5.27553  5.323544  5.368484  5.410349  5.449136  5.484837  5.517443  5.546942  5.573318  5.596554  5.616629  5.633517  5.647194  5.657629   5.66479  5.668642  5.669148  5.666268  5.659961  5.650183  5.636889  5.620032  5.599565  5.575441  5.547613  5.516035  5.480662  5.441452   5.39837  5.351381  5.300461  5.245592        5.186769 | 276.276442   294.34072  312.661655   331.15828  349.754957  368.380755  386.968903  405.456322  423.783225  441.892772  459.730777  477.245459  494.387231  511.108517    527.3636  543.108498  558.300848  572.899826  586.866072  600.161634  612.749925  624.595703  635.665048  645.925367  655.345405  663.895266  671.546458  678.271936  684.046174  688.845241  692.646894  695.430693   697.17812  697.872725  697.500287  696.048988  693.509609  689.875746  685.144044  679.314442  672.390451  664.379433  655.292908  645.146869  633.962108  621.764552    608.5856  594.462456  579.438454  563.563357   546.89363 |  100.0 | 176.276442 |  546.89363 |
    | 12:30 | 5.832203  5.998568  6.161644  6.321006  6.476328  6.627362   6.77392  6.915861  7.053079  7.185495  7.313051  7.435706  7.553427  7.666191  7.773981  7.876782  7.974581  8.067367  8.155126  8.237846   8.31551    8.3881  8.455594  8.517969  8.575197  8.627245  8.674079  8.715657  8.751938  8.782873   8.80841  8.828494  8.843064  8.852057  8.855406   8.85304  8.844886  8.830868  8.810906  8.784922  8.752835  8.714563  8.670028  8.619151  8.561858  8.498081  8.427758  8.350836  8.267274             8.177046 | 257.556023  269.892613  282.253963  294.590694  306.859064  319.020085  331.038784  342.883582  354.525775    365.9391  377.099376  387.984202  398.572703   408.84532  418.783629  428.370193  437.588435   446.42253  454.857312  462.878205  470.471153  477.622568  484.319292  490.548559  496.297971   501.55548  506.309376   510.54829  514.261191  517.437402  520.066617  522.138932  523.644873  524.575448  524.922192  524.677235  523.833374  522.384152  520.323961  517.648143  514.353119  510.436517  505.897327  500.736068  494.954969  488.558168  481.551927  473.944872  465.748233  456.976116 |  74.477036   76.173631   77.836689   79.461869   81.045849   82.586099   84.080703    85.52822   86.927572   88.277956   89.578784   90.829619   92.030143   93.180119   94.279363   95.327729   96.325089   97.271321   98.166296   99.009873   99.801892  100.542167  101.230481  101.866585  102.450193  102.980984  103.458592  103.882615  104.252608  104.568083  104.828512  105.033325  105.181913  105.273625  105.307778  105.283652  105.200497  105.057535  104.853968  104.588981  104.261752  103.871457  103.417282  102.898437  102.314166  101.663767  100.946608  100.162154   99.309986        98.389836 |  73.322033    74.98568   76.616442   78.210063   79.763283   81.273622   82.739203    84.15861   85.530787   86.854948   88.130514   89.357059    90.53427   91.661914   92.739811   93.767819   94.745811   95.673667   96.551262   97.378458   98.155097   98.880996   99.555943  100.179694   100.75197  101.272452  101.740786  102.156575  102.519383  102.828732  103.084103  103.284939  103.430641  103.520573  103.554062  103.530405  103.448864  103.308678  103.109065  102.849224  102.528349  102.145633  101.700278  101.191508  100.618584   99.980815   99.277583   98.508362   97.672743     96.770462 | 1.390539  1.412598   1.43403  1.454796  1.474874   1.49425  1.512916  1.530871  1.548117  1.564658    1.5805   1.59565  1.610116  1.623906  1.637028  1.649488  1.661295  1.672453   1.68297   1.69285  1.702098  1.710717  1.718709  1.726078  1.732823  1.738945  1.744445  1.749319  1.753566  1.757183  1.760166   1.76251  1.764209  1.765258  1.765648  1.765372  1.764422  1.762787  1.760457  1.757422  1.753671  1.749191  1.743969  1.737994  1.731252   1.72373  1.715415  1.706295  1.696358  1.685591 | 1.346005  1.345418  1.344881  1.344389  1.343937  1.343524  1.343144  1.342796  1.342475  1.342181   1.34191  1.341661  1.341431  1.341219  1.341024  1.340845   1.34068  1.340528  1.340388   1.34026  1.340142  1.340035  1.339938   1.33985   1.33977  1.339699  1.339636  1.339581  1.339533  1.339493   1.33946  1.339434  1.339416  1.339404    1.3394  1.339403  1.339413  1.339431  1.339457   1.33949  1.339532  1.339582  1.339641   1.33971  1.339788  1.339878  1.339978   1.34009  1.340215          1.340354 | 0.655616  0.662258  0.668505  0.674373  0.679881  0.685047  0.689892  0.694434  0.698691  0.702681  0.706418  0.709919  0.713196  0.716262  0.719129  0.721805  0.724302  0.726627  0.728788  0.730792  0.732646  0.734353  0.735921  0.737352  0.738651   0.73982  0.740863  0.741782  0.742577  0.743252  0.743806   0.74424  0.744554  0.744747  0.744819  0.744768  0.744593  0.744291   0.74386  0.743296  0.742597  0.741757  0.740773  0.739639  0.738349  0.736898  0.735277   0.73348  0.731498      0.729322 | -0.345211  -0.352686  -0.359712  -0.366311  -0.372505  -0.378316  -0.383766  -0.388877  -0.393668  -0.398157  -0.402361  -0.406297  -0.409977  -0.413416  -0.416626  -0.419615  -0.422396  -0.424976  -0.427363  -0.429564  -0.431586  -0.433433  -0.435111  -0.436624  -0.437975  -0.439167  -0.440202  -0.441083  -0.441811  -0.442385  -0.442808  -0.443077  -0.443193  -0.443155   -0.44296  -0.442606  -0.442091  -0.441411  -0.440562   -0.43954  -0.438341  -0.436958  -0.435386  -0.433617  -0.431645  -0.429462  -0.427059  -0.424427  -0.421557     -0.418437 | 0.683471  0.684528  0.685537  0.686498  0.687412  0.688281  0.689105  0.689885  0.690624  0.691322   0.69198    0.6926  0.693183  0.693729  0.694239  0.694714  0.695156  0.695564   0.69594  0.696283  0.696594  0.696874  0.697122   0.69734  0.697527  0.697683  0.697809  0.697904  0.697968  0.698002  0.698004  0.697975  0.697914  0.697822  0.697696  0.697538  0.697347  0.697121   0.69686  0.696564  0.696232  0.695863  0.695457  0.695012  0.694528  0.694004   0.69344  0.692834  0.692186      0.691496 | 257.498082 |  266.07758  283.368962  300.963801  318.783844  336.755645  354.810047  372.881723  390.908765  408.832328  426.596318   444.14713  461.433413  478.405874  495.017112  511.221468  526.974912  542.234932  556.960451  571.111763  584.650468  597.539441  609.742793  621.225863   631.95521  641.898618  651.025118   659.30502  666.709956  673.212934  678.788411  683.412375  687.062443  689.717975    691.3602  691.972363  691.539885  690.050546  687.494675  683.865372  679.158737  673.374126  666.514417  658.586298  649.600567   639.57245  628.521929  616.474068  603.459356  589.514034          574.680419 | 567.070969 | 0.929778  0.944939  0.959659  0.973912  0.987685  1.000968  1.013759  1.026056  1.037863  1.049182  1.060019  1.070379  1.080268  1.089692  1.098656  1.107167  1.115229  1.122847  1.130026  1.136769  1.143079  1.148959  1.154411  1.159436  1.164036  1.168211   1.17196  1.175283  1.178179  1.180644  1.182678  1.184275  1.185434  1.186148  1.186414  1.186226  1.185578  1.184464  1.182876  1.180808   1.17825  1.175196  1.171636  1.167562  1.162965  1.157835  1.152164  1.145942  1.139162       1.131815 | 3.877699  3.976747  4.073618  4.168086  4.259982  4.349183  4.435599  4.519166  4.599839  4.677589  4.752397   4.82425  4.893143  4.959073  5.022039  5.082043  5.139085  5.193166  5.244285   5.29244  5.337628  5.379843  5.419077   5.45532   5.48856  5.518782  5.545968  5.570097  5.591147  5.609092  5.623903   5.63555  5.643999  5.649213  5.651155  5.649783  5.645055  5.636926  5.625351   5.61028  5.591667  5.569462  5.543616  5.514082  5.480813  5.443765  5.402897  5.358174  5.309565        5.257048 | 257.498082  274.619595   292.08488  309.813152  327.728697  345.760343  363.840961  381.907033  399.898269  417.757278  435.429274  452.861836   470.00469  486.809531  503.229865  519.220877   534.73932  549.743422  564.192808  578.048437  591.272556  603.828666  615.681494  626.796988  637.142317  646.685881  655.397342  663.247656  670.209126  676.255465   681.36187  685.505115  688.663656   690.81775  691.949592  692.043469  691.085928  689.065968   685.97524  681.808278  676.562737  670.239655  662.843733  654.383626  644.872256  634.327132  622.770679  610.230578  596.740097  582.338421  567.070969 |  100.0 | 157.498082 | 567.070969 |
    | 13:00 | 5.655093  5.819066  5.980632  6.139258  6.294522  6.446098   6.59373  6.737219  6.876411  7.011186  7.141453   7.26714   7.38819  7.504561  7.616216  7.723127   7.82527  7.922621   8.01516  8.102866  8.185719  8.263694  8.336768  8.404914  8.468101  8.526298  8.579468  8.627571  8.670564  8.708401   8.74103  8.768397  8.790444  8.807109  8.818325  8.824024  8.824133  8.818577  8.807277  8.790155  8.767126  8.738109  8.703021  8.661778    8.6143  8.560508  8.500331  8.433699  8.360554             8.280848 | 244.726788  256.593605  268.549283  280.541284  292.522899  304.452406  316.292346  328.008885  339.571286  350.951453  362.123551  373.063683   383.74962  394.160569  404.276984  414.080396   423.55328  432.678929  441.441359  449.825224  457.815739  465.398622  472.560047  479.286595  485.565229  491.383263  496.728347  501.588454  505.951879  509.807239  513.143485  515.949917  518.216209  519.932443  521.089147  521.677344   521.68861  521.115144  519.949845    518.1864   515.81939  512.844399  509.258149  505.058635   500.24529  494.819154  488.783066   482.14187  474.902639  467.074913 |   72.67086   74.343058    75.99072   77.608389   79.191785   80.737564   82.243118   83.706422   85.125905   86.500351   87.828817   89.110573   90.345051   91.531803   92.670469   93.760755   94.802406   95.795198   96.738916   97.633349   98.478279   99.273476  100.018688   100.71364  101.358029   101.95152  102.493747  102.984304  103.422752  103.808612  104.141366  104.420458  104.645293  104.815238  104.929622   104.98774  104.988853  104.932191  104.816959  104.642339  104.407494  104.111579  103.753744  103.333148  102.848964    102.3004  101.686704  101.007191  100.261257        99.448408 |  71.550931   73.190656   74.806321   76.392576   77.945224   79.460985   80.937302    82.37219   83.764108   85.111862    86.41453   87.671396   88.881901   90.045607   91.162161   92.231274   93.252698    94.22621   95.151601   96.028665   96.857187   97.636942   98.367682   99.049139   99.681014  100.262981  100.794677  101.275708  101.705642  102.084008  102.410301  102.683973  102.904442  103.071086  103.183249  103.240239   103.24133  103.185769  103.072775  102.901545  102.671261  102.381092  102.030206  101.617777  101.142997  100.605085  100.003307   99.336989   98.605541     97.808477 | 1.366828  1.388788   1.41023  1.431098  1.451357  1.470979  1.489948  1.508254  1.525893  1.542863  1.559167  1.574808  1.589791   1.60412  1.617803  1.630845  1.643251  1.655028   1.66618  1.676712  1.686628  1.695931  1.704624   1.71271  1.720188   1.72706  1.733326  1.738984  1.744032  1.748469   1.75229  1.755491  1.758068  1.760014  1.761323  1.761988  1.762001  1.761353  1.760034  1.758034  1.755342  1.751948  1.747838  1.743001  1.737424  1.731093  1.723996  1.716119  1.707449  1.697974 | 1.346675  1.346053   1.34548  1.344953  1.344468  1.344023  1.343614  1.343237  1.342891  1.342572  1.342277  1.342006  1.341756  1.341525  1.341312  1.341116  1.340934  1.340767  1.340613  1.340471   1.34034   1.34022  1.340111  1.340011   1.33992  1.339838  1.339764  1.339698  1.339641   1.33959  1.339547  1.339512  1.339483  1.339462  1.339447   1.33944   1.33944  1.339447  1.339461  1.339483  1.339513  1.339551  1.339597  1.339652  1.339717   1.33979  1.339874  1.339969  1.340076          1.340194 | 0.648225   0.65508  0.661556  0.667662  0.673413  0.678824  0.683912  0.688693  0.693186  0.697405  0.701366  0.705085  0.708574  0.711846  0.714912  0.717784  0.720471  0.722981  0.725324  0.727506  0.729533  0.731412  0.733148  0.734746  0.736209  0.737542  0.738747  0.739828  0.740785  0.741622  0.742339  0.742937  0.743416  0.743778   0.74402  0.744143  0.744146  0.744026  0.743781   0.74341  0.742909  0.742275  0.741503   0.74059   0.73953  0.738319  0.736949  0.735415  0.733708      0.731822 | -0.336605  -0.344337  -0.351635  -0.358514   -0.36499  -0.371084  -0.376815  -0.382202  -0.387264  -0.392019  -0.396483  -0.400672  -0.404601  -0.408282  -0.411728   -0.41495  -0.417958  -0.420762  -0.423368  -0.425785   -0.42802  -0.430079  -0.431966  -0.433686  -0.435244  -0.436642  -0.437883  -0.438971  -0.439906   -0.44069  -0.441323  -0.441806  -0.442139   -0.44232   -0.44235  -0.442225  -0.441945  -0.441506  -0.440905  -0.440139  -0.439204  -0.438094  -0.436804   -0.43533  -0.433663  -0.431798  -0.429726   -0.42744   -0.42493     -0.422188 | 0.682209  0.683284  0.684315    0.6853  0.686242   0.68714  0.687995  0.688808   0.68958  0.690312  0.691005   0.69166  0.692279  0.692861  0.693408   0.69392  0.694399  0.694845  0.695258   0.69564   0.69599  0.696308  0.696597  0.696854  0.697082  0.697279  0.697446  0.697583  0.697691  0.697768  0.697814  0.697831  0.697816  0.697771  0.697694  0.697585  0.697443  0.697269  0.697061   0.69682  0.696543  0.696231  0.695883  0.695498  0.695075  0.694614  0.694113  0.693573  0.692991      0.692369 | 240.298231 | 248.389159  264.740134  281.472955  298.510263  315.779159   333.21079  350.739961  368.304769  385.846273  403.308207  420.636719  437.780153  454.688854  471.314995  487.612442  503.536623  519.044426  534.094111  548.645232  562.658584   576.09615  588.921067  601.097604  612.591148  623.368202  633.396393  642.644495   651.08246  658.681456  665.413928  671.253663  676.175869  680.157269  683.176214  685.212797  686.249002  686.268852  685.258583  683.206835  680.104855  675.946727  670.729612  664.454011  657.124042  648.747732  639.337331  628.909627  617.486279  605.094151          591.765644 | 584.896578 | 0.913469  0.928574  0.943312  0.957646  0.971552  0.985013   0.99802  1.010565  1.022647  1.034267  1.045425  1.056126  1.066372  1.076169  1.085521  1.094432  1.102907   1.11095  1.118565  1.125754  1.132522  1.138871  1.144802  1.150318  1.155419  1.160106  1.164379  1.168237  1.171679  1.174704  1.177309  1.179491  1.181247  1.182574  1.183466   1.18392  1.183928  1.183486  1.182587  1.181224   1.17939  1.177076  1.174274  1.170976  1.167173  1.162856  1.158016  1.152643  1.146729       1.140265 |  3.77199  3.869867  3.966079  4.060332    4.1524   4.24211  4.329333  4.413971  4.495952  4.575221   4.65174  4.725481  4.796423  4.864553  4.929861  4.992339  5.051982  5.108784  5.162741  5.213848  5.262097  5.307482  5.349993  5.389618  5.426345  5.460159   5.49104  5.518971  5.543928  5.565886  5.584819  5.600695  5.613483  5.623148  5.629653  5.632958  5.633021  5.629799  5.623246  5.613315  5.599958  5.583124  5.562764  5.538828  5.511266  5.480029  5.445072  5.406351  5.363825        5.317462 | 240.298231  256.440293   273.00399  289.909628   307.08218  324.450874  341.948785  359.512452  377.081536  394.598503  412.008357    429.2584  446.298023  463.078526  479.552965  495.676016  511.403864  526.694104  541.505661  555.798725  569.534694  582.676137  595.186762  607.031399  618.175992  628.587603  638.234426  647.085812  655.112309  662.285707  668.579103  673.966969  678.425245  681.931435  684.464727  686.006118  686.538566  686.047152   684.51926  681.944771  678.316289  673.629365  667.882754  661.078684  653.223145  644.326185   634.40223  623.470409  611.554887  598.685197  584.896578 |  100.0 | 140.298231 | 584.896578 |
    | 13:30 | 5.486211  5.646885   5.80608  5.963154   6.11759  6.268978   6.41699  6.561369  6.701907  6.838441  6.970842  7.099005  7.222848  7.342304  7.457319  7.567847   7.67385  7.775293  7.872146   7.96438  8.051966  8.134878  8.213086   8.28656  8.355267  8.419174  8.478242  8.532432    8.5817  8.626001  8.665284  8.699495  8.728578  8.752472  8.771113  8.784433  8.792362  8.794824  8.791744   8.78304  8.768632  8.748436  8.722366  8.690337  8.652264  8.608063  8.557653  8.500956    8.4379              8.36842 | 232.785701  244.139814  255.644051  267.243315  278.888362  290.535058  302.143686  313.678339  325.106382  336.397996  347.525779  358.464417  369.190393  379.681743  389.917854   399.87928  409.547599  418.905278  427.935565  436.622398  444.950319  452.904411  460.470238  467.633797  474.381477  480.700031  486.576547   491.99843  496.953396  501.429462  505.414948  508.898488  511.869043  514.315924  516.228822  517.597841  518.413549  518.667029  518.349941  517.454601  515.974056  513.902192  511.233829  507.964852  504.092343  499.614728  494.531944  488.845623  482.559284  475.678551 |  70.948591   72.587152   74.210634   75.812474   77.387418   78.931277   80.440719   81.913096   83.346308   84.738689   86.088914   87.395928   88.658887   89.877108   91.050034   92.177203    93.25822   94.292739   95.280448   96.221053   97.114269   97.959808   98.757376   99.506664  100.207343  100.859062  101.461441  102.014072  102.516514  102.968293    103.3689  103.717789  104.014379  104.258052  104.448154  104.583994  104.664848  104.689961  104.658545  104.569789  104.422855  104.216891  103.951028  103.624397  103.236128  102.785366  102.271282  101.693083  101.050034       100.341472 |  69.862107   71.468848   73.060804   74.631537   76.175897   77.689775   79.169904   80.613689   82.019069    83.38441   84.708415   85.990048   87.228481   88.423045   89.573194   90.678474   91.738498   92.752927   93.721456   94.643795   95.519665   96.348784   97.130864   97.865601   98.552674   99.191737   99.782418  100.324317  100.817002  101.260008  101.652835   101.99495   102.28578  102.524721  102.711131  102.844333  102.923617  102.948242  102.917437  102.830404  102.686324  102.484359   102.22366  101.903371  101.522642  101.080634  100.576532  100.009562      99.379     98.684198 | 1.343992  1.365724  1.387057  1.407919  1.428258  1.448035  1.467222  1.485801  1.503761  1.521093  1.537794  1.553865  1.569306  1.584121  1.598312  1.611884  1.624841  1.637187  1.648927  1.660065  1.670604  1.680547  1.689896  1.698654  1.706821  1.714399  1.721387  1.727784  1.733589  1.738799  1.743413  1.747425  1.750832  1.753628  1.755808  1.757365  1.758292  1.758579  1.758219  1.757203  1.755518  1.753156  1.750105  1.746351  1.741884   1.73669  1.730757   1.72407  1.716616  1.708382 | 1.347363  1.346707    1.3461   1.34554  1.345023  1.344546  1.344107  1.343702  1.343328  1.342983  1.342665  1.342372    1.3421   1.34185  1.341618  1.341403  1.341205  1.341022  1.340853  1.340697  1.340552   1.34042  1.340298  1.340186  1.340083   1.33999  1.339906  1.339829  1.339761  1.339701  1.339648  1.339602  1.339564  1.339532  1.339508  1.339491  1.339481  1.339477  1.339481  1.339493  1.339511  1.339538  1.339572  1.339614  1.339665  1.339725  1.339794  1.339873  1.339963          1.340064 | 0.640848  0.647874  0.654547  0.660868  0.666842  0.672482    0.6778  0.682812  0.687531  0.691974  0.696155  0.700088  0.703786  0.707261  0.710527  0.713592  0.716468  0.719163  0.721686  0.724044  0.726244  0.728293  0.730196  0.731958  0.733584  0.735078  0.736442  0.737682  0.738798  0.739792  0.740668  0.741425  0.742066  0.742589  0.742996  0.743286  0.743458  0.743511  0.743445  0.743256  0.742942  0.742501  0.741929  0.741223  0.740378  0.739391  0.738254  0.736963  0.735512      0.733893 | -0.328002   -0.33595  -0.343485  -0.350616  -0.357354  -0.363713   -0.36971  -0.375362  -0.380686    -0.3857  -0.390417  -0.394856  -0.399028  -0.402948  -0.406628  -0.410078  -0.413311  -0.416334  -0.419157  -0.421788  -0.424233  -0.426499  -0.428591  -0.430515  -0.432275  -0.433875  -0.435318  -0.436607  -0.437743   -0.43873  -0.439567  -0.440256  -0.440798  -0.441191  -0.441435   -0.44153  -0.441474  -0.441264  -0.440898  -0.440374  -0.439687  -0.438834  -0.437811  -0.436611   -0.43523  -0.433662    -0.4319  -0.429936  -0.427762      -0.42537 | 0.680974  0.682059  0.683105   0.68411  0.685074  0.685997  0.686879  0.687721  0.688523  0.689286  0.690011  0.690699  0.691351  0.691967  0.692548  0.693095   0.69361  0.694091   0.69454  0.694958  0.695345  0.695701  0.696027  0.696323  0.696589  0.696825  0.697032   0.69721  0.697358  0.697477  0.697566  0.697625  0.697655  0.697654  0.697623  0.697561  0.697468  0.697343  0.697185  0.696995  0.696771  0.696514  0.696221  0.695893  0.695529  0.695127  0.694688   0.69421  0.693692      0.693135 | 224.612057 | 232.203236  247.587184  263.422219  279.632737  296.147101  312.897375  329.819025  346.850621  363.933548  381.011753  398.031503  414.941174  431.691072  448.233263  464.521437  480.510783  496.157882  511.420622  526.258117  540.630644  554.499595  567.827434  580.577671  592.714837   604.20448  615.013162  625.108474  634.459052  643.034611  650.805987   657.74519  663.825469  669.021386  673.308913   676.66553  679.070345  680.504228  680.949957  680.392386  678.818624  676.218237  672.583462  667.909444  662.194487  655.440324  647.652403  638.840189  629.017478  618.202719          606.419352 | 600.314285 | 0.897748  0.912708  0.927383  0.941724  0.955696  0.969272  0.982437  0.995177  1.007486   1.01936  1.030796  1.041797  1.052362  1.062495  1.072198  1.081476   1.09033  1.098765  1.106784   1.11439  1.121585  1.128372  1.134753  1.140729  1.146301   1.15147  1.156237    1.1606  1.164558  1.168111  1.171257  1.173992  1.176315  1.178221  1.179707  1.180769    1.1814  1.181596  1.181351  1.180658   1.17951  1.177899  1.175819   1.17326  1.170215  1.166673  1.162627  1.158066  1.152983       1.147366 | 3.670921  3.767084  3.862125  3.955681  4.047469  4.137265  4.224895  4.310226  4.393154    4.4736  4.551503  4.626816  4.699504   4.76954  4.836904  4.901577  4.963548  5.022805  5.079338  5.133137  5.184192  5.232493  5.278029  5.320786  5.360751  5.397908  5.432238  5.463722  5.492337   5.51806  5.540863  5.560718  5.577594  5.591457   5.60227  5.609997  5.614595  5.616024  5.614237  5.609189  5.600831  5.589115   5.57399  5.555404  5.533306  5.507645  5.478371  5.445436  5.408792        5.368399 | 224.612057  239.752837  255.383497  271.426379  287.807919  304.458403  321.311673  338.304819  355.377892  372.473622  389.537172  406.515917  423.359244  440.018378  456.446231  472.597274  488.427417  503.893917  518.955293  533.571256  547.702653  561.311421  574.360554  586.814076  598.637028  609.795463  620.256454  629.988107  638.959588   647.14116   654.50423  661.021407  666.666575  671.414976  675.243303  678.129818  680.054471  680.999043  680.947302  679.885177  677.800949  674.685456  670.532322  665.338198  659.103025  651.830312  643.527426  634.205909  623.881789  612.575913  600.314285 |  100.0 | 124.612057 | 600.314285 |
    | 14:00 |  5.32599   5.48251   5.63851  5.793242  5.946098  6.096583  6.244297  6.388918  6.530184  6.667885  6.801851  6.931943  7.058051   7.18008  7.297957  7.411616  7.521004  7.626073  7.726782  7.823094  7.914971  8.002379  8.085285  8.163654  8.237451  8.306637  8.371176  8.431025  8.486141  8.536478  8.581985  8.622611  8.658298  8.688989  8.714621  8.735126  8.750437   8.76048  8.765179  8.764457  8.758231  8.746419  8.728935  8.705693  8.676604  8.641582   8.60054  8.553393   8.50006             8.440466 | 221.720713  232.527259  243.541639   254.70693  265.971896  277.290381  288.620711  299.925123  311.169258  322.321706  333.353612  344.238333  354.951143  365.468975   375.77021  385.834486  395.642536  405.176055  414.417576  423.350372  431.958368  440.226064  448.138472  455.681062  462.839715  469.600688  475.950578  481.876303   487.36508  492.404416  496.982103  501.086219  504.705131  507.827513  510.442364  512.539033  514.107259  515.137205  515.619515  515.545369  514.906556  513.695547  511.905593  509.530818  506.566341    503.0084  498.854494   494.10354  488.756048  482.814304 |  69.314656   70.910855   72.501747   74.079713   75.638542   77.173193   78.679589   80.154433   81.595069   82.999349   84.365537   85.692228   86.978275    88.22274   89.424847   90.583948   91.699489   92.770992   93.798029   94.780214    95.71718   96.608577   97.454055   98.253264   99.005842   99.711413   100.36958  100.979924  101.541999  102.055332  102.519418  102.933719  103.297665  103.610652   103.87204  104.081157  104.237295  104.339713  104.387639  104.380273  104.316785  104.196326  104.018023  103.780994  103.484346  103.127188  102.708638  102.227834  101.683948         101.0762 |  68.259902   69.825104   71.385102   72.932425   74.460983   75.965832   77.442974   78.889178   80.301838   81.678847   83.018505   84.319432   85.580505   86.800804   87.979567   89.116159   90.210037   91.260731   92.267825   93.230936   94.149707   95.023793   95.852853   96.636542   97.374505   98.066374    98.71176   99.310252   99.861412  100.364776   100.81985  101.226105  101.582984  101.889893  102.146205  102.351261  102.504367  102.604796  102.651791  102.644568  102.582314  102.464193  102.289354  102.056927   101.76604  101.415817  101.005395  100.533929  100.000605     99.404658 | 1.322113  1.343489  1.364596  1.385343  1.405663  1.425502  1.444822  1.463594    1.4818  1.499424  1.516459  1.532898   1.54874  1.563983   1.57863  1.592681  1.606139  1.619008  1.631289  1.642987  1.654104  1.664642  1.674604   1.68399  1.692803  1.701043  1.708709  1.715802   1.72232  1.728261  1.733622  1.738401  1.742593  1.746193  1.749197  1.751599   1.75339  1.754565  1.755115   1.75503  1.754302   1.75292  1.750874  1.748151  1.744741   1.74063  1.735806  1.730255  1.723964  1.716919 | 1.348065  1.347379  1.346741  1.346148  1.345599  1.345091  1.344622  1.344188  1.343787  1.343417  1.343074  1.342757  1.342464  1.342193  1.341942  1.341709  1.341493  1.341294  1.341109  1.340938   1.34078  1.340634  1.340499  1.340374   1.34026  1.340156   1.34006  1.339973  1.339894  1.339824  1.339761  1.339705  1.339657  1.339616  1.339582  1.339555  1.339535  1.339522  1.339516  1.339517  1.339525   1.33954  1.339563  1.339594  1.339633   1.33968  1.339735    1.3398  1.339875           1.33996 | 0.633528  0.640682  0.647516   0.65402  0.660194  0.666043  0.671577  0.676806  0.681744  0.686403  0.690797  0.694939  0.698843   0.70252  0.705981  0.709239  0.712302   0.71518  0.717881  0.720414  0.722786  0.725003  0.727071  0.728996  0.730783  0.732435  0.733957  0.735353  0.736624  0.737774  0.738804  0.739717  0.740513  0.741193  0.741759  0.742209  0.742545  0.742764  0.742867  0.742851  0.742715  0.742457  0.742073  0.741562  0.740919   0.74014  0.739222  0.738158  0.736943      0.735571 | -0.319453  -0.327568  -0.335302  -0.342654  -0.349627   -0.35623  -0.362477   -0.36838  -0.373955  -0.379217  -0.384182  -0.388862  -0.393273  -0.397426  -0.401336  -0.405012  -0.408465  -0.411706  -0.414742  -0.417583  -0.420235  -0.422705     -0.425  -0.427124  -0.429082  -0.430879  -0.432519  -0.434004  -0.435337   -0.43652  -0.437556  -0.438444  -0.439187  -0.439784  -0.440235   -0.44054  -0.440698  -0.440706  -0.440564  -0.440269  -0.439818  -0.439208  -0.438435  -0.437494  -0.436381  -0.435091  -0.433618  -0.431954  -0.430094     -0.428028 | 0.679771  0.680859  0.681914  0.682933  0.683914  0.684857  0.685762  0.686628  0.687457  0.688248  0.689002   0.68972  0.690403   0.69105  0.691664  0.692244  0.692791  0.693307   0.69379  0.694243  0.694665  0.695056  0.695418   0.69575  0.696053  0.696327  0.696572  0.696788  0.696975  0.697134  0.697263  0.697364  0.697435  0.697477   0.69749  0.697473  0.697425  0.697347  0.697237  0.697096  0.696922  0.696716  0.696476  0.696202  0.695894  0.695549  0.695168   0.69475  0.694294      0.693799 | 210.365288 | 217.452286  231.855953  246.770575  262.123174  277.844105   293.86696  310.128392  326.567903  343.127618  359.752069  376.387984  392.984098  409.490982  425.860887  442.047614  458.006387  473.693759  489.067512  504.086589  518.711023  532.901889  546.621254  559.832147  572.498537  584.585311  596.058275  606.884153  617.030599   626.46622  635.160608  643.084378  650.209222   656.50797  661.954668   666.52466  670.194689  672.943012  674.749525  675.595905  675.465772   674.34486  672.221207  669.085369  664.930642  659.753306  653.552889  646.332441  638.098829  628.863048          618.640532 |  613.31055 | 0.882674  0.897402  0.911932  0.926205  0.940173  0.953803  0.967067  0.979948  0.992434  1.004515  1.016186  1.027444  1.038289   1.04872   1.05874  1.068349  1.077549  1.086344  1.094736  1.102727  1.110319  1.117515  1.124315  1.130722  1.136736  1.142359  1.147589  1.152428  1.156873  1.160925  1.164581   1.16784  1.170698  1.173153  1.175201  1.176838  1.178059   1.17886  1.179234  1.179177  1.178681  1.177739  1.176344  1.174488  1.172162  1.169359   1.16607  1.162285  1.157994        1.15319 | 3.574776  3.668704  3.762078  3.854469  3.945533  4.034995  4.122637  4.208287  4.291807   4.37309  4.452052  4.528627   4.60276  4.674412  4.743548  4.810143  4.874175  4.935624  4.994474  5.050712  5.104322  5.155291  5.203606  5.249251   5.29221  5.332467  5.370002  5.404797  5.436828  5.466072  5.492502  5.516091  5.536809  5.554622  5.569495  5.581393  5.590276  5.596102  5.598828  5.598409  5.594798  5.587945  5.577801  5.564315  5.547433  5.527105  5.503276  5.475897  5.444915        5.410284 | 210.365288  224.496437  239.175863  254.328899  269.884234  285.773875   301.93299  318.299715  334.814918  351.421983  368.066592  384.696522  401.261468  417.712881  434.003818  450.088822  465.923806   481.46596  496.673665  511.506425  525.924806   539.89039  553.365734   566.31434   578.70064  590.489982  601.648626   612.14376  621.943505  631.016953  639.334193  646.866363  653.585708  659.465643  664.480836  668.607304  671.822514  674.105504  675.437021  675.799668   675.17807  673.559059  670.931873  667.288371  662.623272  656.934404  650.222976  642.493862  633.755904  624.022224   613.31055 |  100.0 | 110.365288 |  613.31055 |
    | 14:30 | 5.174762  5.326336  5.478362  5.629999   5.78055  5.929437  6.076189  6.220418  6.361805  6.500091  6.635064  6.766548  6.894399  7.018499  7.138748  7.255061  7.367369  7.475609  7.579727  7.679674  7.775406  7.866881  7.954059    8.0369  8.115365  8.189414  8.259005  8.324096  8.384643  8.440597  8.491911  8.538531  8.580402  8.617467  8.649663  8.676927  8.699191  8.716384  8.728432  8.735257   8.73678  8.732919  8.723589  8.708702  8.688171  8.661906  8.629819   8.59182  8.547822             8.497741 | 211.512255  221.744314  232.237671  242.934433  253.782009  264.732701  275.743231  286.774253  297.789891  308.757309  319.646325  330.429075  341.079712  351.574145  361.889816  372.005507  381.901167   391.55777  400.957188  410.082082  418.915808  427.442335  435.646179  443.512337  451.026239  458.173705  464.940907  471.314341  477.280801  482.827366  487.941386  492.610473   496.82251  500.565645  503.828312  506.599246  508.867506  510.622509  511.854067  512.552437   512.70837  512.313184  511.358831  509.837986  507.744141  505.071714  501.816168  497.974149  493.543632  488.524085 |  67.772427   69.318182   70.868548    72.41495    73.95027   75.468634   76.965217   78.436063   79.877936   81.288185   82.664641    84.00552   85.309355   86.574929   87.801229     88.9874   90.132717   91.236551   92.298349   93.317616   94.293897   95.226763   96.115806   96.960623   97.760811   98.515962   99.225657   99.889459  100.506913   101.07754  101.600836  102.076267  102.503273   102.88126  103.209603  103.487643  103.714691  103.890024  104.012886  104.082493  104.098028  104.058652  103.963498  103.811683  103.602307   103.33446  103.007231  102.619715  102.171022        101.66029 |  66.747622    68.26336   69.783619   71.299991   72.805496   74.294374   75.761894   77.204178    78.61805   80.000914   81.350639    82.66548   83.943995   85.184992   86.387478   87.550615    88.67369   89.756088   90.797267   91.796741   92.754063   93.668814   94.540592   95.369003   96.153652   96.894139   97.590051   98.240963   98.846426   99.405972   99.919106  100.385305  100.804019  101.174665  101.496632  101.769273  101.991912   102.16384  102.284316  102.352571  102.367805  102.329193  102.235887   102.08702   101.88171  101.619065   101.29819  100.918199   100.47822     99.977406 | 1.301262  1.322161  1.342925  1.363449  1.383648  1.403456  1.422822  1.441708  1.460085  1.477932  1.495234   1.51198  1.528164  1.543781   1.55883  1.573309  1.587219  1.600562  1.613338  1.625551  1.637201  1.648291  1.658821  1.668794   1.67821  1.687069  1.695372  1.703119  1.710307  1.716935  1.723001  1.728503  1.733436  1.737796  1.741579  1.744779  1.747389  1.749404  1.750815  1.751614  1.751792   1.75134  1.750248  1.748504  1.746097  1.743016  1.739248   1.73478  1.729598   1.72369 | 1.348775  1.348064  1.347397  1.346774  1.346195  1.345657  1.345158  1.344696  1.344268  1.343871  1.343503  1.343163  1.342847  1.342555  1.342283  1.342032  1.341798  1.341582  1.341381  1.341195  1.341022  1.340862  1.340714  1.340577  1.340451  1.340334  1.340227   1.34013   1.34004  1.339959  1.339886  1.339821  1.339763  1.339712  1.339669  1.339632  1.339602   1.33958  1.339564  1.339555  1.339553  1.339558   1.33957   1.33959  1.339617  1.339652  1.339695  1.339747  1.339808          1.339878 | 0.626313  0.633545  0.640497   0.64715  0.653496  0.659532  0.665263  0.670696  0.675839  0.680705  0.685306  0.689652  0.693756   0.69763  0.701285  0.704732   0.70798  0.711039  0.713918  0.716624  0.719166   0.72155  0.723782  0.725868  0.727814  0.729623    0.7313  0.732849  0.734273  0.735574  0.736756   0.73782  0.738768  0.739601  0.740321  0.740926  0.741419  0.741798  0.742062  0.742212  0.742246  0.742161  0.741956  0.741628  0.741175  0.740593  0.739878  0.739025  0.738031       0.73689 | -0.311008  -0.319237  -0.327126  -0.334662   -0.34184  -0.348664  -0.355139  -0.361278  -0.367091  -0.372591  -0.377792  -0.382707  -0.387349  -0.391732  -0.395866  -0.399763  -0.403434  -0.406888  -0.410134  -0.413182  -0.416038  -0.418709  -0.421202  -0.423523  -0.425676  -0.427667  -0.429499  -0.431175    -0.4327  -0.434075  -0.435303  -0.436385  -0.437323  -0.438116  -0.438766  -0.439273  -0.439636  -0.439853  -0.439925  -0.439848  -0.439621   -0.43924  -0.438704  -0.438007  -0.437147  -0.436117  -0.434915  -0.433532  -0.431964     -0.430203 | 0.678608  0.679692  0.680748  0.681774  0.682766  0.683725  0.684648  0.685535  0.686387  0.687202  0.687983  0.688728  0.689439  0.690115  0.690759   0.69137  0.691948  0.692495  0.693011  0.693496  0.693952  0.694377  0.694773   0.69514  0.695478  0.695787  0.696068  0.696321  0.696545  0.696742   0.69691   0.69705  0.697161  0.697244  0.697298  0.697323  0.697319  0.697285  0.697221  0.697127  0.697001  0.696844  0.696654  0.696432  0.696176  0.695886   0.69556  0.695199  0.694802      0.694367 |  197.47692 | 204.061346  217.483488  231.466974  245.942327  260.842616  276.103572  291.663553  307.463447  323.446515  339.558233  355.746116  371.959562  388.149695  404.269225  420.272323  436.114512  451.752559  467.144396  482.249039  497.026527  511.437862  525.444968  539.010654  552.098583  564.673253   576.69999  588.144937  598.975068  609.158193  618.662984  627.459006  635.516756  642.807715  649.304406  654.980466  659.810731  663.771329  666.839789  668.995163  670.218164  670.491313  669.799113  668.128226  665.467678  661.809076  657.146842  651.478466  644.804776  637.130224          628.463183 | 623.908461 | 0.868296  0.882706  0.897013  0.911143  0.925039  0.938657  0.951962   0.96493  0.977541  0.989782  1.001643  1.013118  1.024203  1.034895  1.045194    1.0551  1.064614  1.073737   1.08247  1.090815  1.098775  1.106349   1.11354  1.120349  1.126777  1.132824   1.13849  1.143775  1.148679    1.1532  1.157338   1.16109  1.164454  1.167427  1.170007  1.172188  1.173968  1.175341  1.176303  1.176848   1.17697  1.176662  1.175917  1.174728  1.173087  1.170986  1.168417   1.16537  1.161837       1.157808 | 3.483779  3.574984  3.666217  3.756989  3.846898  3.935617  4.022882   4.10848  4.192241  4.274027  4.353728  4.431255  4.506538  4.579519  4.650152  4.718398  4.784226  4.847609  4.908525  4.966952  5.022871  5.076266  5.127119  5.175412  5.221128  5.264249  5.304754  5.342622  5.377833   5.41036   5.44018  5.467264  5.491583  5.513105  5.531796  5.547621  5.560542  5.570519  5.577509  5.581469  5.582353  5.580113  5.574699  5.566061  5.554147  5.538903  5.520276  5.498213  5.472661        5.443567 |  197.47692  210.602151  224.324105  238.572059  253.277744  268.375546  283.802524   299.49833  315.405079  331.467181  347.631174  363.845556  380.060624   396.22833  412.302146  428.236946  443.988899  459.515376  474.774871  489.726926  504.332075  518.551792  532.348452  545.685296  558.526411   570.83671  582.581927  593.728618   604.24417  614.096819  623.255671  631.690748  639.373022  646.274478  652.368174  657.628324  662.030381  665.551144  668.168866  669.863392  670.616294  670.411038  669.233151   667.07042    663.9131  659.754137  654.589413  648.418012   641.24249  633.069172  623.908461 |  100.0 |   97.47692 | 623.908461 |
    | 15:00 | 5.032754  5.178664  5.325995  5.473826  5.621379  5.768001  5.913147  6.056366  6.197282  6.335585  6.471017  6.603365  6.732451  6.858127  6.980268  7.098769   7.21354  7.324503  7.431592  7.534744  7.633906  7.729028  7.820061  7.906962  7.989685  8.068187  8.142424  8.212351  8.277921  8.339087  8.395798  8.448004  8.495648  8.538674  8.577022  8.610628  8.639426  8.663347  8.682319  8.696267  8.705112  8.708773  8.707167  8.700208  8.687808  8.669877  8.646324  8.617059  8.581991              8.54103 | 202.134363  211.772752  221.721006  231.921217  242.320169  252.869176  263.523777  274.243363  284.990785  295.731972  306.435569  317.072612  327.616234  338.041405  348.324706  358.444126  368.378889  378.109307  387.616642  396.882992  405.891197  414.624745  423.067703  431.204648   439.02061  446.501028  453.631705  460.398778  466.788684  472.788141  478.384133  483.563893  488.314902  492.624887  496.481825  499.873952  502.789782  505.218129  507.148135  508.569308  509.471562  509.845275  509.681342  508.971249  507.707154  505.881973  503.489486  500.524448   496.98272  492.861408 |   66.32422   67.812216     69.3147   70.822287   72.327039   73.822297   75.302509    76.76306   78.200128   79.610544   80.991685   82.341373   83.657796   84.939443   86.185043   87.393521   88.563962   89.695572   90.787661   91.839613   92.850872   93.820924   94.749289   95.635505    96.47912   97.279688   98.036758   98.749873    99.41856  100.042332   100.62068  101.153072  101.638951  102.077732  102.468801  102.811516  103.105202  103.349153  103.542631  103.684871  103.775073  103.812411  103.796034  103.725064  103.598605  103.415742  103.175553  102.877107   102.51948       102.101756 |  65.327538   66.786638   68.259945   69.738256   71.213786   72.680008   74.131475   75.563664   76.972824   78.355851   79.710171   81.033649   82.324508   83.581266   84.802678   85.987688   87.135399   88.245035   89.315916    90.34744    91.33906   92.290275   93.200612   94.069618    94.89685   95.681872    96.42424   97.123507   97.779209   98.390868   98.957984   99.480037   99.956481  100.386741  100.770216  101.106276  101.394258  101.633472  101.823193   101.96267  102.051121  102.087734  102.071675  102.002083  101.878079  101.698768  101.463243  101.170593  100.819911    100.410299 | 1.281498  1.301803  1.322114  1.342308  1.362288  1.381971  1.401297  1.420215  1.438688  1.456687   1.47419  1.491181  1.507648  1.523582  1.538979  1.553835  1.568149  1.581918  1.595143  1.607825  1.619965  1.631563   1.64262  1.653137  1.663114  1.672552   1.68145  1.689808  1.697625    1.7049  1.711629  1.717811  1.723443   1.72852  1.733038  1.736992  1.740377  1.743185  1.745411  1.747047  1.748083  1.748512  1.748324  1.747509  1.746055  1.743952  1.741187  1.737748  1.733623  1.728797 | 1.349488  1.348756  1.348065  1.347416  1.346809  1.346242  1.345714  1.345224  1.344768  1.344345  1.343952  1.343588  1.343249  1.342935  1.342643  1.342372   1.34212  1.341886  1.341669  1.341467  1.341279  1.341105  1.340943  1.340793  1.340655  1.340526  1.340408  1.340299  1.340199  1.340107  1.340024  1.339949  1.339881  1.339821  1.339768  1.339721  1.339682   1.33965  1.339625  1.339606  1.339595   1.33959  1.339592  1.339601  1.339618  1.339642  1.339673  1.339713  1.339761          1.339817 | 0.619247  0.626503  0.633529  0.640293  0.646779  0.652976  0.658883  0.664501  0.669837  0.674899  0.679696  0.684238  0.688537  0.692603  0.696448   0.70008  0.703511  0.706749  0.709803  0.712681  0.715392  0.717941  0.720335  0.722581  0.724683  0.726647  0.728478  0.730178  0.731752  0.733203  0.734533  0.735746  0.736842  0.737824  0.738692  0.739448  0.740092  0.740625  0.741046  0.741354  0.741549   0.74163  0.741595  0.741441  0.741167   0.74077  0.740246  0.739592  0.738804      0.737877 | -0.302719  -0.311004  -0.318999  -0.326678  -0.334028  -0.341043  -0.347725  -0.354079  -0.360114  -0.365839  -0.371266  -0.376407  -0.381274  -0.385878  -0.390231  -0.394344  -0.398228  -0.401892  -0.405345  -0.408595  -0.411651  -0.414521  -0.417209  -0.419723  -0.422067  -0.424248  -0.426268  -0.428133  -0.429845  -0.431407  -0.432822  -0.434092  -0.435219  -0.436203  -0.437045  -0.437746  -0.438306  -0.438724  -0.438999  -0.439131  -0.439117  -0.438955  -0.438642  -0.438176  -0.437554   -0.43677  -0.435822  -0.434703  -0.433408     -0.431932 |  0.67749  0.678561  0.679612  0.680639  0.681638  0.682606  0.683543  0.684447  0.685317  0.686154  0.686957  0.687727  0.688463  0.689166  0.689837  0.690476  0.691084  0.691661  0.692207  0.692723  0.693209  0.693667  0.694095  0.694495  0.694866  0.695209  0.695525  0.695812  0.696072  0.696305   0.69651  0.696687  0.696837  0.696958  0.697052  0.697118  0.697155  0.697163  0.697143  0.697092  0.697012  0.696901  0.696759  0.696586   0.69638  0.696142  0.695869  0.695563  0.695221      0.694844 | 185.861607 |  191.95039  204.400392   217.45267  231.042094   245.10516  259.580335  274.408195   289.53145  304.894883  320.445254  336.131179  351.903002  367.712665  383.513591  399.260567  414.909643  430.418036  445.744051  460.847004  475.687161  490.225681  504.424572  518.246651  531.655514  544.615513  557.091741  569.050022  580.456913  591.279708  601.486453  611.045969  619.927882  628.102661  635.541671  642.217224  648.102655  653.172396  657.402072  660.768599  663.250307  664.827064  665.480428  665.193801  663.952609  661.744496  658.559527  654.390423    649.2328  643.085429          635.950517 | 632.161886 | 0.854656  0.868669  0.882674  0.896588  0.910344  0.923886  0.937173  0.950172  0.962857   0.97521  0.987216  0.998865   1.01015  1.021065  1.031608  1.041777   1.05157  1.060989  1.070032  1.078702  1.086998  1.094923  1.102476  1.109659  1.116471  1.122915  1.128989  1.134693  1.140027   1.14499  1.149581  1.153798  1.157639  1.161102  1.164183  1.166879  1.169187  1.171102  1.172619  1.173734  1.174441  1.174734  1.174605  1.174049  1.173058  1.171624  1.169739  1.167395  1.164582       1.161291 | 3.398096   3.48613  3.574778  3.663498  3.751835  3.839412  3.925919  4.011104   4.09476  4.176719  4.256843   4.33502  4.411161   4.48519  4.557045  4.626678  4.694044  4.759109  4.821841  4.882213  4.940202  4.995786  5.048942   5.09965  5.147891  5.193644  5.236888  5.277601   5.31576  5.351341  5.384319  5.414665  5.442352  5.467348   5.48962  5.509134  5.525853  5.539739  5.550751  5.558845  5.563978  5.566103  5.565171  5.561132  5.553936  5.543529  5.529858  5.512868  5.492506        5.468716 | 185.861607  197.995248  210.764143   224.10243   237.94566  252.231246  266.898683  281.889616  297.147809   312.61906   328.25109  343.993408   359.79719   375.61515  391.401419  407.111441  422.701874  438.130497   453.35614  468.338609   483.03863  497.417797  511.438532  525.064048  538.258325  550.986088  563.212798  574.904643  586.028546  596.552173   606.44395  615.673091  624.209632  632.024476  639.089443  645.377338  650.862021  655.518493  659.322996  662.253119  664.287923   665.40808  665.596023  664.836116  663.114838  660.420984  656.745881  652.083627  646.431341  639.789432  632.161886 |  100.0 |  85.861607 | 632.161886 |
    | 15:30 |  4.90009  5.039705  5.181686  5.325052  5.468955  5.612673  5.755596  5.897208  6.037078  6.174842  6.310199  6.442894  6.572716  6.699484  6.823049   6.94328  7.060068  7.173319  7.282949  7.388885  7.491062   7.58942  7.683904  7.774462  7.861044  7.943602  8.022088  8.096455  8.166653  8.232635  8.294348  8.351741  8.404759  8.453345  8.497441  8.536982  8.571907  8.602146  8.627629  8.648282   8.66403  8.674793   8.68049  8.681034  8.676341   8.66632  8.650881  8.629932  8.603381             8.571134 | 193.555751   202.58874   211.97467  221.656652  231.581653   241.70062  251.968389  262.343457  272.787688  283.265995  293.746026  304.197858  314.593717   324.90773  335.115694  345.194879   355.12385  364.882309  374.450962  383.811398  392.945987  401.837785  410.470457  418.828205  426.895711  434.658082  442.100806  449.209712  455.970938  462.370905  468.396295  474.034035  479.271284   484.09543  488.494084  492.455088  495.966524  499.016726  501.594304  503.688168  505.287562  506.382108   506.96185  507.017313  506.539567  505.520304  503.951921  501.827617  499.141502  495.888717 |  64.971308   66.395113   67.843041   69.305085   70.772614   72.238263   73.695796   75.139961   76.566352   77.971279   79.351654   80.704889   82.028814   83.321604   84.581716    85.80784   86.998852   88.153787   89.271798   90.352139   91.394144   92.397202   93.360752   94.284264   95.167232   96.009162   96.809567   97.567961    98.28385   98.956731   99.586088  100.171384  100.712063  101.207547   101.65723  102.060481  102.416639  102.725016  102.984893   103.19552  103.356119  103.465882  103.523972  103.529528  103.481663  103.379469  103.222022  103.008384  102.737612       102.408761 |  64.000899   65.397055   66.816865   68.250517   69.689548   71.126735   72.555963   73.972084   75.370775    76.74842   78.101988   79.428944    80.72716   81.994845   83.230486   84.432799   85.600683    86.73319    87.82949   88.888852   89.910621   90.894201   91.839039   92.744618   93.610439   94.436019   95.220881   95.964547   96.666534   97.326348   97.943483   98.517413   99.047593   99.533455   99.974405  100.369825  100.719067  101.021456  101.276286  101.482823  101.640303  101.747934  101.804897  101.810345  101.763409  101.663199   101.50881  101.299321  101.033806    100.711342 | 1.262866   1.28247  1.302221  1.321984  1.341646  1.361114  1.380312  1.399182  1.417676  1.435756  1.453393  1.470566  1.487256  1.503452  1.519144  1.534326  1.548993  1.563141  1.576769  1.589876  1.602462  1.614525  1.626067  1.637086  1.647584   1.65756  1.667013  1.675943  1.684349  1.692229  1.699581  1.706403  1.712691  1.718443  1.723654   1.72832  1.732436  1.735994  1.738991  1.741417  1.743266  1.744528  1.745197  1.745261   1.74471  1.743534  1.741722  1.739261   1.73614  1.732345 | 1.350198  1.349452  1.348742  1.348069  1.347437  1.346843  1.346289  1.345771  1.345288  1.344839  1.344421  1.344032  1.343671  1.343334  1.343021   1.34273   1.34246  1.342208  1.341973  1.341755  1.341552  1.341363  1.341187  1.341024  1.340872  1.340731  1.340601  1.340481   1.34037  1.340268  1.340174  1.340089  1.340011  1.339941  1.339878  1.339823  1.339775  1.339733  1.339698   1.33967  1.339649  1.339635  1.339627  1.339627  1.339633  1.339646  1.339667  1.339695  1.339731          1.339776 | 0.612375    0.6196   0.62665  0.633485  0.640075  0.646404  0.652461  0.658245  0.663756     0.669  0.673982  0.678712  0.683198  0.687451   0.69148  0.695295  0.698904  0.702318  0.705545  0.708593   0.71147  0.714183  0.716738  0.719141  0.721399  0.723516  0.725498  0.727347  0.729069  0.730667  0.732143  0.733501  0.734742  0.735869  0.736883  0.737785  0.738576  0.739258  0.739829   0.74029   0.74064  0.740879  0.741005  0.741017  0.740913  0.740691  0.740348   0.73988  0.739285      0.738559 | -0.294634  -0.302916  -0.310965  -0.318743  -0.326226  -0.333401  -0.340263   -0.34681  -0.353048  -0.358983  -0.364623  -0.369979   -0.37506  -0.379879  -0.384445  -0.388768   -0.39286  -0.396729  -0.400384  -0.403834  -0.407087  -0.410151  -0.413031  -0.415735  -0.418267  -0.420634  -0.422839  -0.424888  -0.426783  -0.428528  -0.430125  -0.431578  -0.432888  -0.434057  -0.435085  -0.435974  -0.436724  -0.437334  -0.437806  -0.438136  -0.438326  -0.438372  -0.438272  -0.438025  -0.437628  -0.437077  -0.436367  -0.435496  -0.434459     -0.433249 | 0.676423  0.677474  0.678512  0.679533  0.680532  0.681506  0.682451  0.683367  0.684253  0.685107  0.685929   0.68672  0.687479  0.688206  0.688903  0.689568  0.690202  0.690806  0.691381  0.691926  0.692442  0.692929  0.693388  0.693819  0.694222  0.694597  0.694945  0.695266   0.69556  0.695827  0.696067   0.69628  0.696466  0.696624  0.696756   0.69686  0.696937  0.696985  0.697006  0.696997   0.69696  0.696894  0.696797   0.69667  0.696512  0.696323  0.696101  0.695846  0.695557      0.695234 | 175.431719 | 181.036389  192.532895  204.663291  217.367591  230.586404  244.261514  258.336212  272.755454  287.465913  302.415961  317.555605  332.836401  348.211361  363.634849   379.06249  394.451076  409.758483  424.943592  439.966223   454.78707  469.367654   483.67027  497.657953  511.294442   524.54416  537.372191  549.744271  561.626778   572.98674  583.791837  594.010419  603.611527  612.564925  620.841136  628.411491  635.248184  641.324338   646.61408   651.09263  654.736402  657.523109  659.431895  660.443471  660.540265  659.706599  657.928868   655.19574  651.498382  646.830686          641.189529 | 638.149957 | 0.841787  0.855327  0.868957  0.882585  0.896132  0.909536  0.922745   0.93572  0.948427  0.960844  0.972949   0.98473  0.996174  1.007275  1.018025  1.028422  1.038462  1.048144  1.057467  1.066431  1.075036  1.083281  1.091168  1.098696  1.105867  1.112679  1.119134   1.12523  1.130967  1.136345  1.141361  1.146016  1.150306  1.154229  1.157783  1.160965  1.163772  1.166199  1.168242  1.169896  1.171156  1.172018  1.172473  1.172517  1.172141   1.17134  1.170104  1.168426  1.166298        1.16371 | 3.317836  3.402296  3.487951  3.574212  3.660579  3.746629   3.83201  3.916428  3.999641  4.081449  4.161685  4.240215  4.316927  4.391726  4.464537  4.535295  4.603946  4.670444   4.73475   4.79683  4.856654  4.914193  4.969423   5.02232   5.07286  5.121021  5.166779  5.210113  5.250997  5.289407  5.325318  5.358701  5.389528   5.41777  5.443393  5.466365  5.486649  5.504209  5.519005  5.530995  5.540136  5.546383  5.549689  5.550005  5.547281  5.541464  5.532503  5.520342  5.504926        5.486201 | 175.431719  186.597275  198.426867  210.860358  223.837853  237.300423  251.190536  265.452283  280.031471  294.875633  309.933973  325.157297  340.497911  355.909529   371.34717  386.767064  402.126565  417.384072  432.498951  447.431474  462.142762  476.594734  490.750063  504.572145  518.025066  531.073584  543.683111  555.819704  567.450066  578.541547  589.062157  598.980585  608.266224  616.889206  624.820445  632.031686  638.495569  644.185696  649.076717  653.144419  656.365834  658.719354  660.184866  660.743897  660.379773  659.077798  656.825446  653.612574  649.431641  644.277961  638.149957 |  100.0 |  75.431719 | 638.149957 |
    | 16:00 | 4.776796  4.909583  5.045636  5.183935  5.323582  5.463796  5.603905  5.743337  5.881603  6.018291  6.153053  6.285593  6.415663  6.543053  6.667581  6.789096  6.907466  7.022577  7.134331   7.24264  7.347427  7.448622  7.546161  7.639985  7.730037  7.816264  7.898614  7.977036  8.051479  8.121892  8.188222  8.250417  8.308421  8.362177  8.411628  8.456712  8.497365  8.533521  8.565112  8.592066  8.614308  8.631762  8.644346  8.651979  8.654574  8.652044    8.6443  8.631249  8.612799             8.588856 | 185.740833   194.16377  202.976735  212.124926   221.55635  231.222274  241.077365  251.079639  261.190296   271.37349  281.596065  291.827297  302.038637   312.20347  322.296902  332.295557  342.177405    351.9216  361.508347  370.918776  380.134833  389.139187  397.915149  406.446591  414.717891   422.71387  430.419746  437.821091  444.903797   451.65404   458.05826  464.103139  469.775581  475.062708  479.951849  484.430537  488.486517   492.10775   495.28243  497.998999  500.246174   502.01298  503.288784  504.063341   504.32685  504.070013  503.284105  501.961056   500.09354  497.675078 |  63.713951   65.068119    66.45559    67.86597   69.290096   70.720005   72.148847   73.570773   74.980814   76.374765    77.74907   79.100723   80.427186   81.726305   82.996252   84.235465   85.442606   86.616515   87.756185   88.860726   89.929348   90.961339   91.956045   92.912861   93.831216   94.710565   95.550376   96.350128   97.109301   97.827373   98.503811   99.138072   99.729598  100.277811  100.782111  101.241876  101.656458  102.025183   102.34735  102.622226  102.849053  103.027043  103.155378  103.233215  103.259682  103.233885  103.154908  103.021817  102.833663       102.589489 |  62.767959    64.09583   65.456357   66.839349   68.235819    69.63796   71.039055   72.433368   73.816028   75.182909   76.530525   77.855931   79.156634   80.430525   81.675811   82.890959   84.074658   85.225771   86.343309   87.426401   88.474271   89.486221   90.461611   91.399846   92.300367    93.16264   93.986142   94.770364   95.514794   96.218921   96.882223   97.504168   98.084207   98.621773    99.11628   99.567117   99.973648  100.335213  100.651123  100.920662  101.143084  101.317617   101.44346  101.519785  101.545738  101.520443  101.442999  101.312492  101.127992     100.88856 | 1.245399  1.264205  1.283298  1.302533  1.321783  1.340944  1.359931  1.378672   1.39711    1.4152  1.432905  1.450196  1.467051  1.483452  1.499385  1.514842  1.529813  1.544293  1.558278  1.571766  1.584754  1.597241  1.609226  1.620707  1.631685  1.642159  1.652128   1.66159  1.670546  1.678992  1.686927  1.694349  1.701255  1.707641  1.713505  1.718841  1.723645  1.727912  1.731635  1.734809  1.737425  1.739476  1.740955  1.741851  1.742156  1.741859  1.740949  1.739416  1.737247  1.734431 |   1.3509  1.350146  1.349422  1.348731  1.348076  1.347459  1.346879  1.346335  1.345826  1.345352  1.344909  1.344495  1.344111  1.343752  1.343418  1.343106  1.342816  1.342545  1.342293  1.342058  1.341839  1.341635  1.341445  1.341268  1.341103   1.34095  1.340808  1.340676  1.340553   1.34044  1.340336   1.34024  1.340153  1.340073  1.340001  1.339936  1.339878  1.339828  1.339784  1.339747  1.339716  1.339693  1.339676  1.339666  1.339662  1.339665  1.339676  1.339693  1.339719          1.339751 | 0.605737  0.612876    0.6199   0.62676  0.633416  0.639843  0.646024  0.651951  0.657618  0.663027  0.668182  0.673089  0.677754  0.682185  0.686393  0.690384  0.694169  0.697756  0.701152  0.704368  0.707409  0.710283  0.712996  0.715556  0.717968  0.720236  0.722367  0.724364  0.726232  0.727974  0.729594  0.731094  0.732478  0.733746  0.734902  0.735947  0.736881  0.737706  0.738423  0.739031  0.739531  0.739921  0.740202  0.740372   0.74043  0.740374  0.740201   0.73991  0.739497      0.738959 | -0.286802  -0.295019  -0.303066  -0.310894   -0.31847   -0.32577  -0.332781  -0.339497  -0.345917  -0.352043  -0.357882   -0.36344  -0.368726   -0.37375  -0.378521  -0.383048  -0.387342  -0.391411  -0.395264   -0.39891  -0.402356   -0.40561  -0.408679  -0.411569  -0.414285  -0.416835  -0.419222   -0.42145  -0.423524  -0.425448  -0.427224  -0.428855  -0.430344  -0.431691    -0.4329  -0.433971  -0.434904  -0.435701   -0.43636  -0.436882  -0.437266  -0.437511  -0.437615  -0.437576  -0.437392   -0.43706  -0.436577   -0.43594  -0.435143     -0.434184 | 0.675409  0.676433  0.677453  0.678462  0.679455  0.680428  0.681377  0.682301  0.683197  0.684065  0.684904  0.685712  0.686491   0.68724  0.687958  0.688647  0.689306  0.689936  0.690536  0.691108  0.691652  0.692167  0.692655  0.693115  0.693548  0.693954  0.694333  0.694686  0.695012  0.695311  0.695584  0.695831  0.696052  0.696246  0.696413  0.696554  0.696668  0.696755  0.696814  0.696846   0.69685  0.696825  0.696772  0.696689  0.696576  0.696433  0.696259  0.696053  0.695815      0.695544 | 166.099094 | 171.235081  181.804644   193.03065  204.858981  217.235001  230.104368  243.413557  257.110181  271.143157  285.462777  300.020714  314.769986  329.664899  344.660975  359.714876  374.784332  389.828063   404.80571  419.677779  434.405575  448.951158  463.277294  477.347421  491.125612  504.576552  517.665515  530.358352  542.621479  554.421875  565.727086  576.505233  586.725028  596.355796  605.367505  613.730805  621.417068  628.398448   634.64794  640.139455  644.847902  648.749286  651.820809  654.040994  655.389817  655.848853  655.401437  654.032844  651.730481  648.484096          644.286008 | 641.971983 | 0.829713  0.842712  0.855899  0.869172  0.882446  0.895649  0.908722  0.921617  0.934295  0.946727  0.958886  0.970756  0.982319  0.993566  1.004488  1.015078  1.025331  1.035245  1.044817  1.054045  1.062928  1.071466  1.079659  1.087506  1.095007  1.102162   1.10897  1.115431  1.121545  1.127311  1.132726  1.137791  1.142503  1.146861  1.150861  1.154501  1.157777  1.160687  1.163226   1.16539  1.167174  1.168573  1.169581  1.170192    1.1704  1.170197  1.169577  1.168532  1.167053       1.165132 | 3.243046  3.323586  3.405878  3.489305  3.573329  3.657486  3.741385  3.824693  3.907134  3.988474  4.068519   4.14711   4.22411  4.299409  4.372911  4.444539  4.514226  4.581916  4.647559  4.711113  4.772542  4.831812  4.888892  4.943756  4.996375  5.046725  5.094781  5.140517  5.183908   5.22493  5.263555  5.299756  5.333504   5.36477  5.393521  5.419726  5.443349  5.464354  5.482704  5.498356  5.511271  5.521404   5.52871   5.53314  5.534647  5.533178  5.528683  5.521106  5.510395        5.496492 | 166.099094  176.327832  187.239944   198.78178  210.898691     223.536  236.639659  250.156643  264.035183   278.22487  292.676688  307.342992  322.177459  337.135022  352.171793  367.244985  382.312841  397.334559  412.270227  427.080765  441.727868   456.17396  470.382151  484.316205  497.940507  511.220042  524.120374  536.607638  548.648533  560.210319  571.260826  581.768466  591.702251  601.031817  609.727463  617.760186  625.101735  631.724663  637.602405  642.709346  647.020915  650.513685  653.165484  654.955525  655.864539  655.874935  654.970968  653.138924  650.367318   646.64712  641.971983 |  100.0 |  66.099094 | 641.971983 |
    | 16:30 | 4.662803  4.788331  4.917961  5.050661  5.185499   5.32165   5.45839  5.595094  5.731223  5.866316  5.999978  6.131875   6.26172   6.38927  6.514316  6.636679  6.756205  6.872761  6.986231  7.096513  7.203516  7.307159  7.407368  7.504077  7.597221  7.686742  7.772582  7.854686  7.933001  8.007472  8.078045  8.144667  8.207281   8.26583  8.320256  8.370499  8.416496  8.458182  8.495489  8.528349  8.556687   8.58043  8.599498  8.613811  8.623285  8.627835  8.627372  8.621805  8.611043             8.594991 | 178.650675  186.465533  194.701133  203.305811   212.22951  221.424536  230.845967  240.451799  250.202927  260.063026  269.998363  279.977589   289.97152  299.952919  309.896293  319.777704  329.574595  339.265636  348.830584  358.250159  367.505934  376.580237  385.456062  394.116998  402.547158  410.731121  418.653877  426.300788  433.657542   440.71012  447.444773  453.847988  459.906478   465.60716  470.937147  475.883742  480.434433  484.576901  488.299021  491.588879  494.434786  496.825303  498.749268  500.195836  501.154514  501.615215  501.568316  501.004724  499.915948  498.294189 |  62.551442   63.831586   65.153561    66.50684   67.881925   69.270393   70.664876   72.058988   73.447235   74.824915   76.188012   77.533102   78.857269   80.158027   81.433247   82.681109   83.900041   85.088685   86.245857   87.370514   88.461735   89.518691   90.540632   91.526868   92.476757   93.389692   94.265094   95.102398   95.901054   96.660512   97.380223    98.05963   98.698167   99.295255   99.850296  100.362674  100.831751  101.256866   101.63733  101.972432   102.26143  102.503556  102.698013  102.843978  102.940599  102.986999  102.982276  102.925508  102.815752       102.652051 |  61.628025    62.88331   64.179612   65.506612   66.854994   68.216499   69.583901   70.950941    72.31223   73.663156   74.999782   76.318751   77.617204   78.892702   80.143159   81.366787   82.562049    83.72761    84.86231   85.965128   87.035158   88.071589   89.073684   90.040768   90.972211   91.867418   92.725819   93.546864    94.33001    95.07472   95.780454   96.446668   97.072805   97.658298   98.202561   98.704989   99.164957   99.581816   99.954892  100.283486  100.566872  100.804296  100.994977  101.138107  101.232852  101.278351   101.27372  101.218054  101.110429    100.949907 | 1.229114  1.247039  1.265386     1.284  1.302749  1.321517  1.340208  1.358741  1.377049  1.395079  1.412784  1.430131  1.447091   1.46364  1.479761   1.49544  1.510667  1.525432  1.539729  1.553553  1.566901   1.57977  1.592157  1.604061   1.61548  1.626413  1.636858  1.646814  1.656281  1.665255  1.673735  1.681719  1.689203  1.696185  1.702662  1.708629  1.714082  1.719015  1.723424  1.727302  1.730643  1.733439  1.735683  1.737366   1.73848  1.739015   1.73896  1.738306  1.737041  1.735153 | 1.351587  1.350832    1.3501  1.349396  1.348723  1.348085  1.347482  1.346914  1.346381  1.345881  1.345414  1.344977  1.344568  1.344187  1.343831  1.343499  1.343189    1.3429  1.342629  1.342377  1.342142  1.341922  1.341717  1.341526  1.341348  1.341182  1.341027  1.340883  1.340749  1.340625   1.34051  1.340404  1.340307  1.340217  1.340135  1.340061  1.339994  1.339934  1.339881  1.339835  1.339796  1.339763  1.339737  1.339717  1.339704  1.339698  1.339699  1.339706  1.339721          1.339743 | 0.599372  0.606369  0.613316  0.620154  0.626836  0.633326    0.6396  0.645642  0.651444  0.657001  0.662313  0.667384  0.672216  0.676819  0.681197   0.68536  0.689315   0.69307  0.696633  0.700012  0.703215  0.706248  0.709118  0.711832  0.714395  0.716814  0.719092  0.721235  0.723246  0.725131  0.726892  0.728533  0.730055  0.731463  0.732758  0.733941  0.735015  0.735981  0.736838  0.737589  0.738232  0.738769  0.739198  0.739519  0.739732  0.739833  0.739823  0.739699  0.739457      0.739097 | -0.279266  -0.287355  -0.295344  -0.303173  -0.310796  -0.318183  -0.325311  -0.332167  -0.338746  -0.345044  -0.351064   -0.35681  -0.362289  -0.367507  -0.372474  -0.377198  -0.381687   -0.38595  -0.389996  -0.393832  -0.397467  -0.400908  -0.404162  -0.407234  -0.410132  -0.412861  -0.415425   -0.41783   -0.42008  -0.422178  -0.424129  -0.425934  -0.427596  -0.429119  -0.430503   -0.43175   -0.43286  -0.433836  -0.434677  -0.435383  -0.435954  -0.436389  -0.436687  -0.436847  -0.436866  -0.436743  -0.436474  -0.436057  -0.435489     -0.434765 | 0.674454  0.675445  0.676439   0.67743  0.678411  0.679377  0.680325  0.681252  0.682155  0.683033  0.683884  0.684707  0.685502  0.686269  0.687008  0.687718  0.688399  0.689052  0.689677  0.690274  0.690843  0.691385    0.6919  0.692387  0.692849  0.693283  0.693692  0.694074  0.694431  0.694761  0.695066  0.695345  0.695598  0.695825  0.696027  0.696203  0.696352  0.696475  0.696572  0.696642  0.696685    0.6967  0.696688  0.696647  0.696577  0.696478   0.69635   0.69619     0.696      0.695778 | 157.776495 | 162.462441  172.138219  182.484301  193.453069  204.995216  217.060787  229.599895  242.563196  255.902187  269.569372  283.518346  297.703822   312.08161  326.608583  341.242623  355.942571  370.668156  385.379948  400.039296  414.608276  429.049649  443.326814  457.403773  471.245102  484.815916  498.081857  511.009071  523.564198  535.714365  547.427187  558.670771  569.413722  579.625165  589.274762  598.332745  606.769949   614.55786  621.668659  628.075291  633.751528  638.672055  642.812554  646.149811  648.661826  650.327943   651.12899  651.047432  650.067544  648.175593          645.360045 | 643.742834 | 0.818447  0.830847  0.843528  0.856384  0.869322  0.882263  0.895142  0.907903  0.920501  0.932899  0.945067  0.956982  0.968624   0.97998  0.991036  1.001784  1.012218  1.022332  1.032121  1.041583  1.050717   1.05952   1.06799  1.076129  1.083933  1.091404   1.09854  1.105341  1.111806  1.117933  1.123722  1.129172   1.13428  1.139045  1.143464  1.147534  1.151254  1.154619  1.157626  1.160271  1.162549  1.164456  1.165986  1.167134  1.167894  1.168258  1.168221  1.167775  1.166912       1.165625 |  3.17372  3.250052   3.32866  3.408914  3.490248  3.572167  3.654245  3.736114  3.817463  3.898027  3.977585   4.05595  4.132964  4.208495  4.282432  4.354681  4.425161  4.493804  4.560551  4.625353  4.688163  4.748943  4.807656   4.86427  4.918755  4.971082  5.021222   5.06915  5.114838   5.15826  5.199388  5.238194   5.27465  5.308725  5.340389  5.369609  5.396351   5.42058  5.442259   5.46135   5.47781  5.491599  5.502672  5.510982  5.516483  5.519124  5.518856  5.515624  5.509375        5.500054 | 157.776495  167.106073  177.129349  187.799808  199.068651  210.886007   223.20179  235.966278  249.130483  262.646368  276.466966  290.546429  304.840021  319.304095  333.896041  348.574236  363.297979  378.027434  392.723573  407.348121   421.86351  436.232828  450.419786  464.388679   478.10436  491.532212  504.638133  517.388517   529.75025  541.690702  553.177732  564.179692  574.665443  584.604371  593.966415  602.722097  610.842564  618.299636  625.065858  631.114568  636.419971  640.957226  644.702536  647.633262   649.72804  650.966916  651.331491  650.805086  649.372919  647.022299  643.742834 |  100.0 |  57.776495 | 643.742834 |
    | 17:00 | 4.557952  4.675901  4.798705  4.925347  5.054883  5.186458  5.319312  5.452773  5.586256  5.719256  5.851333  5.982114  6.111275  6.238538  6.363666  6.486453  6.606721  6.724317  6.839106  6.950971  7.059806  7.165519  7.268026  7.367249  7.463118  7.555567  7.644534  7.729959  7.811784  7.889952  7.964406  8.035092  8.101951  8.164927  8.223961  8.278992  8.329959  8.376798  8.419443  8.457825  8.491874  8.521516  8.546674  8.567271  8.583225  8.594451  8.600863  8.602373   8.59889             8.590319 | 172.243889  179.458735   187.11841   195.17539  203.582431  212.293627   221.26508  230.455258  239.825147  249.338254  258.960515  268.660157  278.407515  288.174858  297.936199  307.667125  317.344628  326.946956  336.453475  345.844548  355.101418  364.206115  373.141363  381.890504  390.437426  398.766504  406.862544  414.710731  422.296594  429.605964  436.624942  443.339875  449.737329  455.804075   461.52707  466.893447  471.890513  476.505742  480.726776  484.541439  487.937737  490.903885  493.428319  495.499729  497.107088  498.239694  498.887214  499.039741  498.687855  497.822693 |  61.482169   62.685016   63.937379   65.228876    66.54989   67.891705   69.246551   70.607591    71.96886   73.325191   74.672127   76.005833   77.323018   78.620855   79.896912   81.149097   82.375599   83.574848   84.745474   85.886273    86.99618   88.074244   89.119608    90.13149   91.109169   92.051971    92.95926   93.830425   94.664877   95.462035   96.221327   96.942179   97.624014   98.266245   98.868273   99.429486   99.949252  100.426918  100.861811  101.253234  101.600463   101.90275  102.159318  102.369364  102.532059  102.646546  102.711941   102.72734  102.691812       102.604409 |  60.579516   61.759005   62.987048   64.253465   65.548826   66.864584    68.19312   69.527729   70.862564   72.192556   73.513334   74.821141   76.112747   77.385381   78.636658   79.864526    81.06721   82.243171   83.391064   84.509709   85.598063   86.655192   87.680256   88.672488   89.631181   90.555674   91.445344   92.299592   93.117839   93.899517   94.644064   95.350918   96.019512   96.649271   97.239608   97.789923   98.299594   98.767985   99.194433   99.578255   99.918741  100.215157  100.466743   100.67271  100.832246  100.944509  101.008635  101.023734  100.988896    100.903191 | 1.214017  1.230992  1.248514  1.266426   1.28459  1.302882  1.321196  1.339444  1.357548  1.375446  1.393086  1.410425   1.42743  1.444071  1.460326  1.476178   1.49161  1.506613  1.521177  1.535294   1.54896  1.562169  1.574918  1.587204  1.599025  1.610379  1.621263  1.631676  1.641616  1.651081  1.660068  1.668576  1.676602  1.684142  1.691194  1.697753  1.703815  1.709376  1.714431  1.718973  1.722997  1.726496  1.729463   1.73189  1.733768  1.735089  1.735844  1.736021  1.735611  1.734603 | 1.352254  1.351506  1.350772   1.35006  1.349374  1.348719  1.348096  1.347506   1.34695  1.346426  1.345935  1.345475  1.345043   1.34464  1.344262  1.343909  1.343579   1.34327  1.342982  1.342712   1.34246  1.342225  1.342004  1.341799  1.341606  1.341427  1.341259  1.341103  1.340958  1.340822  1.340697   1.34058  1.340472  1.340372  1.340281  1.340197  1.340121  1.340052   1.33999  1.339935  1.339886  1.339845  1.339809  1.339781  1.339759  1.339744  1.339735  1.339733  1.339738          1.339749 | 0.593313  0.600115  0.606935  0.613704  0.620368  0.626882  0.633216  0.639347  0.645258  0.650943  0.656394  0.661614  0.666603  0.671365  0.675906  0.680233  0.684351   0.68827  0.691995  0.695535  0.698896  0.702086   0.70511  0.707976  0.710689  0.713255  0.715679  0.717965  0.720119  0.722144  0.724044  0.725823  0.727483  0.729027  0.730458  0.731778  0.732987  0.734089  0.735084  0.735972  0.736755  0.737433  0.738005  0.738472  0.738832  0.739085  0.739229  0.739263  0.739185      0.738992 | -0.272066  -0.279968  -0.287842  -0.295618  -0.303243  -0.310675  -0.317884   -0.32485   -0.33156  -0.338008   -0.34419  -0.350108  -0.355765  -0.361167   -0.36632  -0.371231  -0.375908  -0.380359  -0.384591  -0.388614  -0.392433  -0.396056   -0.39949  -0.402742  -0.405817  -0.408721   -0.41146  -0.414038   -0.41646  -0.418729   -0.42085  -0.422825  -0.424657   -0.42635  -0.427904  -0.429322  -0.430605  -0.431755  -0.432771  -0.433654  -0.434405  -0.435023  -0.435507  -0.435856  -0.436069  -0.436145   -0.43608  -0.435872  -0.435519     -0.435017 |  0.67356   0.67451  0.675473  0.676439  0.677403  0.678358    0.6793  0.680225   0.68113  0.682014  0.682873  0.683708  0.684517  0.685299  0.686055  0.686783  0.687484  0.688158  0.688805  0.689425  0.690018  0.690584  0.691124  0.691638  0.692126  0.692588  0.693024  0.693434  0.693819  0.694179  0.694514  0.694824  0.695108  0.695367  0.695601   0.69581  0.695994  0.696151  0.696283  0.696389  0.696469  0.696522  0.696549  0.696548  0.696519  0.696462  0.696376  0.696261  0.696117      0.695941 | 150.378786 | 154.635896  163.456402  172.952867  183.084668   193.80833  205.078748  216.850087  229.076409  241.712108  254.712175  268.032372  281.629317  295.460517  309.484373  323.660153   337.94796  352.308686  366.703968  381.096142  395.448198  409.723735  423.886927  437.902486  451.735632  465.352065  478.717943  491.799868  504.564868   516.98039  529.014297  540.634866  551.810795  562.511215  572.705701  582.364303  591.457565  599.956565  607.832955  615.059014  621.607699  627.452718  632.568605  636.930804  640.515765   643.30106  645.265495   646.38925  646.654028  646.043213          644.542053 | 643.588839 | 0.807996  0.819747  0.831867  0.844247  0.856791  0.869413  0.882042  0.894615  0.907081  0.919397  0.931529  0.943446  0.955127  0.966552  0.977706  0.988579  0.999159  1.009441  1.019417  1.029085   1.03844  1.047479  1.056201  1.064604  1.072686  1.080447  1.087886     1.095   1.10179  1.108255  1.114392  1.120201   1.12568  1.130826  1.135638  1.140114   1.14425  1.148044  1.151492   1.15459  1.157335  1.159721  1.161745    1.1634  1.164681  1.165581  1.166096  1.166217  1.165938        1.16525 | 3.109794  3.181694   3.25635  3.333132  3.411463  3.490825  3.570762  3.650877  3.730827  3.810318    3.8891  3.966961  4.043719  4.119223  4.193342  4.265967  4.337001  4.406366  4.473992  4.539818  4.603792  4.665866     4.726  4.784156  4.840298  4.894395  4.946415   4.99633   5.04411  5.089727  5.133153  5.174358  5.213315  5.249992  5.284359  5.316383  5.346033  5.373272  5.398064  5.420373  5.440159   5.45738  5.471995  5.483957  5.493222  5.499741  5.503465  5.504341  5.502318        5.497342 | 150.378786  158.851944  168.020668  177.846069  188.285711  199.295012  210.828294  222.839525  235.282835  248.112856  261.284927  274.755219  288.480788   302.41959  316.530464  330.773106  345.108023  359.496492  373.900511  388.282752  402.606522  416.835715  430.934784    444.8687  458.602931  472.103411  485.336524  498.269087   510.86834  523.101938   534.93795  546.344862  557.291584  567.747465   577.68231  587.066407  595.870558  604.066115   611.62503  618.519902  624.724048  630.211563  634.957414  638.937518  642.128857  644.509583  646.059152  646.758464  646.590013  645.538065  643.588839 |  100.0 |  50.378786 | 643.588839 |
    | 17:30 | 4.462004  4.572163  4.687834  4.808038  4.931844  5.058387   5.18688  5.316617  5.446976  5.577407  5.707435  5.836644  5.964676   6.09122  6.216007  6.338806  6.459414  6.577656  6.693379  6.806447  6.916742  7.024157  7.128597  7.229976  7.328214  7.423238   7.51498  7.603374  7.688359  7.769873  7.847859  7.922257  7.993009  8.060058  8.123343  8.182805  8.238381  8.290009  8.337624  8.381158  8.420543  8.455706  8.486573  8.513068  8.535111  8.552621  8.565513  8.573701  8.577094             8.575602 | 166.477454  173.105845  180.196427  187.706708  195.593095  203.812213  212.321816  221.081366  230.052361  239.198482  248.485605  257.881746  267.356937  276.883089  286.433839   295.98439  305.511361  314.992646  324.407281  333.735318  342.957722  352.056266  361.013443  369.812386  378.436797  386.870882  395.099296  403.107089  410.879665  418.402742  425.662313  432.644625  439.336145  445.723541  451.793665   457.53354  462.930347   467.97142  472.644244  476.936455  480.835847  484.330381    487.4082  490.057648  492.267297  494.025974    495.3228  496.147234  496.489122  496.338757 |   60.50369   61.627101   62.806711   64.032558   65.295139    66.58563   67.896004   69.219072   70.548471   71.878619   73.204646   74.522326   75.827998   77.118497   78.391085    79.64339   80.873357   82.079194   83.259338   84.412413   85.537204   86.632628   87.697712   88.731575   89.733411   90.702471   91.638056   92.539505   93.406181    94.23747    95.03277   95.791485   96.513023   97.196786    97.84217   98.448562   99.015334   99.541839  100.027416   100.47138  100.873024  101.231617  101.546402  101.816599  102.041398  102.219966   102.35144  102.434935   102.46954       102.454322 |  59.620039   60.721635   61.878337   63.080379   64.318441   65.583872   66.868799   68.166174   69.469758   70.774075   72.074351   73.366443    74.64676   75.912198   77.160072   78.388059   79.594141   80.776562   81.933788   83.064471   84.167419    85.24157   86.285971   87.299758   88.282138    89.23238   90.149797    91.03374   91.883586   92.698732   93.478587   94.222569   94.930095    95.60058   96.233431   96.828048   97.383813   97.900094   98.376241   98.811584   99.205428   99.557057    99.86573  100.130679  100.351113  100.526213  100.655134  100.737008   100.77094    100.756018 | 1.200098   1.21607  1.232701  1.249839  1.267341  1.285079   1.30294  1.320827  1.338654  1.356352  1.373861   1.39113  1.408121  1.424797  1.441133  1.457105  1.472695  1.487889  1.502674  1.517041  1.530982  1.544491  1.557562  1.570192  1.582377  1.594114    1.6054  1.616232   1.62661  1.636529  1.645988  1.654984  1.663514  1.671576  1.679165   1.68628  1.692914  1.699064  1.704726  1.709893  1.714561  1.718722  1.722371  1.725499    1.7281  1.730164  1.731683  1.732647  1.733046  1.732871 | 1.352896  1.352161  1.351433  1.350718  1.350024  1.349356  1.348717  1.348108  1.347531  1.346986  1.346472  1.345989  1.345535  1.345109   1.34471  1.344336  1.343985  1.343657   1.34335  1.343063  1.342794  1.342542  1.342306  1.342085  1.341879  1.341686  1.341505  1.341336  1.341179  1.341032  1.340895  1.340767  1.340649  1.340539  1.340438  1.340345  1.340259  1.340181  1.340109  1.340045  1.339988  1.339938  1.339894  1.339856  1.339826  1.339801  1.339783  1.339772  1.339767          1.339769 | 0.587587  0.594146  0.600789  0.607443  0.614045  0.620544  0.626903   0.63309  0.639085  0.644873  0.650445  0.655797  0.660928  0.665838  0.670532  0.675015  0.679291  0.683367  0.687249  0.690945  0.694461  0.697804   0.70098  0.703995  0.706856  0.709567  0.712135  0.714563  0.716857   0.71902  0.721058  0.722972  0.724767  0.726445   0.72801  0.729462  0.730805   0.73204  0.733168  0.734191  0.735109  0.735923  0.736634   0.73724  0.737743   0.73814  0.738432  0.738617  0.738694       0.73866 | -0.265236  -0.272893  -0.280596  -0.288268  -0.295846   -0.30328  -0.310532  -0.317575  -0.324388   -0.33096  -0.337283  -0.343354  -0.349174  -0.354746  -0.360073  -0.365162  -0.370018  -0.374649  -0.379062  -0.383264  -0.387262  -0.391063  -0.394674  -0.398101  -0.401349  -0.404426  -0.407335  -0.410083  -0.412673   -0.41511  -0.417397  -0.419538  -0.421537  -0.423395  -0.425115    -0.4267   -0.42815  -0.429468  -0.430654  -0.431709  -0.432633  -0.433427  -0.434089   -0.43462  -0.435019  -0.435283  -0.435412  -0.435404  -0.435255     -0.434963 | 0.672727  0.673633  0.674558  0.675495  0.676436  0.677374  0.678305  0.679223  0.680127  0.681012  0.681876  0.682719  0.683538  0.684332  0.685102  0.685847  0.686565  0.687258  0.687924  0.688565   0.68918  0.689769  0.690332   0.69087  0.691383   0.69187  0.692332  0.692769  0.693182  0.693569  0.693932  0.694271  0.694585  0.694874  0.695139   0.69538  0.695595  0.695786  0.695951  0.696092  0.696207  0.696296  0.696359  0.696396  0.696406  0.696389  0.696344  0.696271  0.696169      0.696038 | 143.823868 | 147.675291  155.683213  164.365128  173.687757  183.613853  194.103564  205.115476  216.607397  228.536908  240.861749  253.540072  266.530599  279.792713    293.2865  306.972761     320.813  334.769402  348.804798  362.882636  376.966937  391.022264  405.013687  418.906752  432.667448   446.26219  459.657788  472.821434  485.720689  498.323468  510.598036  522.513006  534.037338   545.14035  555.791725  565.961529  575.620233  584.738738  593.288414  601.241133  608.569322  615.246015  621.244919  626.540486  631.107993  634.923637   637.96464  640.209359  641.637422  642.229864          641.969284 | 641.644175 | 0.798353  0.809417   0.82093  0.832783  0.844879  0.857128  0.869453  0.881787  0.894071  0.906258  0.918307  0.930184  0.941862  0.953319  0.964535  0.975496  0.986191  0.996608  1.006742  1.016585  1.026132  1.035381  1.044327  1.052968  1.061303  1.069328  1.077044  1.084448  1.091539  1.098316  1.104776   1.11092  1.116745  1.122248  1.127429  1.132285  1.136812  1.141009  1.144872  1.148397  1.151581  1.154419  1.156908  1.159042  1.160815  1.162223  1.163258  1.163916  1.164188       1.164069 | 3.051157  3.118468  3.188958  3.262016  3.337066   3.41358  3.491079  3.569143  3.647401  3.725532   3.80326  3.880346  3.956587   4.03181  4.105865  4.178626  4.249983  4.319843  4.388126   4.45476  4.519684  4.582844  4.644193  4.703685  4.761283  4.816949   4.87065  4.922353  4.972026  5.019641  5.065166  5.108572  5.149829  5.188907  5.225775  5.260401  5.292752  5.322793  5.350491  5.375807  5.398704  5.419141  5.437079  5.452473  5.465278  5.475449  5.482936  5.487691  5.489662        5.488795 | 143.823868   151.48719  159.840164  168.851837  178.486528  188.705356  199.467438  210.730788  222.452973  234.591565   247.10445  259.950031  273.087341  286.476104  300.076763  313.850474  327.759086  341.765117  355.831713  369.922618  384.002134  398.035085  411.986785  425.823011  439.509969  453.014277   466.30294   479.34334  492.103214  504.550654  516.654099  528.382336    539.7045  550.590087  561.008969  570.931406  580.328077  589.170107  597.429104  605.077206  612.087126  618.432217  624.086538  629.024928  633.223099   636.65773  639.306578  641.148599  642.164082  642.334801  641.644175 |  100.0 |  43.823868 | 641.644175 |
    | 18:00 | 4.374648  4.476919  4.585244  4.698716  4.816435  4.937546  5.061251  5.186824  5.313609  5.441026  5.568561  5.695762  5.822237  5.947643  6.071682  6.194093   6.31465  6.433156  6.549437  6.663341  6.774733  6.883493  6.989514    7.0927  7.192961  7.290217  7.384393  7.475416  7.563222  7.647744  7.728922  7.806693  7.880998  7.951777   8.01897  8.082515  8.142352  8.198417  8.250647  8.298975  8.343333  8.383652  8.419858  8.451878  8.479633  8.503045  8.522031  8.536507  8.546385             8.551576 | 161.307468  167.367785  173.900995  180.870377  188.236746  195.959966   204.00007  212.318049  220.876361  229.639219  238.572738  247.644955   256.82579  266.086951  275.401819  284.745315  294.093771  303.424797  312.717156  321.950648  331.106002  340.164777  349.109274  357.922456   366.58787   375.08959  383.412151    391.5405  399.459952  407.156144  414.614999  421.822697  428.765645  435.430452  441.803908  447.872972  453.624751  459.046496  464.125593  468.849561  473.206054  477.182863  480.767928  483.949348  486.715403  489.054575  490.955576  492.407381  493.399275  493.920892 |  59.612835   60.655791   61.760501   62.917685   64.118191   65.353283   66.614835   67.895429   69.188395   70.487797   71.788397   73.085603   74.375403   75.654298   76.919247   78.167599   79.397047   80.605573   81.791411   82.953007   84.088988   85.198131   86.279341    87.33163   88.354099   89.345919   90.306323   91.234588   92.130032   92.991996   93.819847   94.612963   95.370728   96.092533   96.777765   97.425804   98.036022    98.60778   99.140421   99.633272  100.085639  100.496808  100.866042  101.192579  101.475631  101.714388   101.90801  102.055633   102.15637       102.209305 |  58.746484   59.769187   60.852444   61.987156   63.164348   64.375456   65.612509   66.868236   68.136093   69.410261   70.685605    71.95762   73.222372   74.476433   75.716817   76.940927     78.1465   79.331557   80.494367   81.633406   82.747326    83.83493   84.895144   85.926999   86.929612   87.902172   88.843925   89.754164   90.632219   91.477445   92.289219   93.066933   93.809983   94.517771   95.189696    95.82515   96.423519   96.984174   97.506471   97.989751   98.433333   98.836518   99.198581   99.518777   99.796333  100.030453  100.220315  100.365072  100.463852    100.515759 | 1.187337  1.202268  1.217958  1.234259  1.251031  1.268143  1.285479  1.302932  1.320413  1.337843  1.355156  1.372294  1.389211  1.405867   1.42223  1.438272  1.453971  1.469309  1.484271  1.498844  1.513019  1.526787  1.540142  1.553077  1.565588   1.57767  1.589322  1.600538  1.611317  1.621655   1.63155  1.640999  1.649998  1.658546  1.666638  1.674271  1.681442  1.688145  1.694376  1.700131  1.705404  1.710189   1.71448   1.71827  1.721551  1.724316  1.726557  1.728264  1.729429  1.730041 | 1.353508  1.352794  1.352077  1.351366   1.35067  1.349993  1.349341  1.348717  1.348122  1.347556  1.347022  1.346517  1.346041  1.345594  1.345173  1.344778  1.344408   1.34406  1.343734  1.343429  1.343142  1.342874  1.342622  1.342386  1.342165  1.341958  1.341764  1.341582  1.341412  1.341253  1.341105  1.340967  1.340838  1.340718  1.340606  1.340503  1.340408   1.34032   1.34024  1.340167  1.340101  1.340042  1.339989  1.339943  1.339904   1.33987  1.339844  1.339824   1.33981          1.339803 | 0.582215  0.588489   0.59491  0.601402  0.607898  0.614343  0.620689    0.6269  0.632949  0.638816  0.644487  0.649953  0.655209  0.660255   0.66509  0.669718  0.674143   0.67837  0.682404  0.686251  0.689918   0.69341  0.696735  0.699897  0.702902  0.705757  0.708465  0.711033  0.713465  0.715766  0.717938  0.719986  0.721914  0.723724   0.72542  0.727003  0.728476  0.729841    0.7311  0.732253  0.733303  0.734249  0.735093  0.735835  0.736474  0.737011  0.737445  0.737774  0.737999      0.738116 | -0.258801  -0.266163  -0.273643  -0.281159  -0.288642  -0.296033  -0.303288  -0.310372  -0.317257  -0.323926  -0.330366   -0.33657  -0.342536  -0.348262  -0.353751  -0.359006  -0.364032  -0.368835  -0.373421  -0.377796  -0.381968  -0.385941  -0.389724  -0.393321  -0.396739  -0.399984   -0.40306  -0.405974  -0.408729  -0.411329   -0.41378  -0.416083  -0.418244  -0.420264  -0.422146  -0.423893  -0.425506  -0.426987  -0.428338  -0.429558   -0.43065  -0.431614  -0.432448  -0.433154   -0.43373  -0.434176   -0.43449  -0.434671  -0.434716     -0.434624 | 0.671956  0.672813  0.673697  0.674599  0.675512  0.676429  0.677343  0.678251  0.679147   0.68003  0.680896  0.681742  0.682568  0.683372  0.684154  0.684911  0.685645  0.686354  0.687038  0.687697  0.688332  0.688941  0.689526  0.690086  0.690622  0.691133   0.69162  0.692082   0.69252  0.692934  0.693324   0.69369  0.694032   0.69435  0.694644  0.694914   0.69516  0.695382  0.695579  0.695752  0.695901  0.696024  0.696122  0.696195  0.696242  0.696262  0.696256  0.696223  0.696162      0.696073 | 138.033385 | 141.503641  148.744745  156.650926  165.196435  174.350538  184.078952  194.345012  205.110567   216.33666  227.984017  240.013389  252.385788  265.062632  278.005841  291.177878  304.541773  318.061115  331.700041  345.423215  359.195797  372.983416  386.752143  400.468462  414.099245  427.611726  440.973484  454.152422  467.116753  479.834988  492.275931  504.408669  516.202576   527.62731  538.652828  549.249391  559.387582  569.038328  578.172925  586.763073   594.78091  602.199062  608.990693  615.129566  620.590112   625.34751  629.377773  632.657849  635.165728   636.88057          637.782834 |  638.04775 | 0.789506  0.799857  0.810725  0.822007  0.833607  0.845433  0.857404  0.869448  0.881502  0.893513  0.905435  0.917229  0.928865  0.940314  0.951555  0.962571  0.973346  0.983868  0.994128  1.004117   1.01383   1.02326  1.032403  1.041257  1.049818  1.058084  1.066052  1.073721  1.081088  1.088153  1.094914  1.101369  1.107515  1.113352  1.118877  1.124088  1.128983  1.133558   1.13781  1.141737  1.145334  1.148599  1.151526  1.154111  1.156349  1.158235  1.159763  1.160927  1.161721       1.162139 | 2.997649  3.060281  3.126449   3.19558  3.267112  3.340518  3.415309  3.491045  3.567335  3.643833  3.720238  3.796288  3.871758  3.946452  4.020204  4.092868  4.164322  4.234456  4.303179   4.37041  4.436078   4.50012  4.562482  4.623113   4.68197  4.739011  4.794199  4.847497  4.898872  4.948291  4.995724  5.041138  5.084503  5.125788  5.164962  5.201992  5.236846   5.26949   5.29989  5.328009  5.353811  5.377256  5.398306  5.416917  5.433046   5.44665   5.45768  5.466089  5.471827        5.474842 | 138.033385  144.936151  152.515642   160.74896  169.607432  179.058193  189.065485  199.591686  210.598081   222.04544  233.894421  246.105852  258.640915  271.461258  284.529062  297.807069  311.258591  324.847495  338.538185  352.295581  366.085084  379.872552  393.624267  407.306915  420.887554  434.333595  447.612783   460.69318   473.54315  486.131353  498.426734  510.398524  522.016237  533.249678  544.068953  554.444479  564.347004  573.747632  582.617852   590.92957  598.655157  605.767492   612.24002   618.04682  623.162673   627.56315  631.224704  634.124771  636.241892  637.555838   638.04775 |  100.0 |  38.033385 |  638.04775 |
    | 18:30 | 4.295512  4.389899  4.490767  4.597297  4.708646  4.823987  4.942531   5.06354  5.186341  5.310326  5.434948  5.559727  5.684236  5.808102  5.930997  6.052634  6.172762  6.291161  6.407637  6.522019  6.634157  6.743916  6.851177  6.955831  7.057781  7.156936  7.253215   7.34654  7.436838  7.524041  7.608083  7.688899  7.766427  7.840606  7.911373  7.978668  8.042429  8.102592  8.159094   8.21187  8.260852  8.305973  8.347161  8.384344  8.417446   8.44639  8.471097  8.491483  8.507464             8.518952 | 156.689822  162.204565  168.196467  174.635128  181.486413  188.714069  196.281015  204.150306  212.285812  220.652668  229.217533  237.948726  246.816257  255.791797  264.848605  273.961432  283.106412  292.260945  301.403585  310.513932  319.572528  328.560762  337.460778  346.255404  354.928067  363.462738  371.843866  380.056324  388.085365  395.916579  403.535847  410.929318   418.08337  424.984587   431.61974  437.975758  444.039723  449.798849  455.240475   460.35206   465.12118  469.535525  473.582908   477.25127  480.528694  483.403418  485.863862  487.898649  489.496643  490.646979 |  58.805803   59.768364   60.797021    61.88341   63.018954    64.19521   65.404122   66.638182   67.890513   69.154907   70.425814   71.698313   72.968064   74.231254   75.484542   76.725002   77.950072   79.157505   80.345328   81.511803   82.655389   83.774719   84.868569   85.935838   86.975525   87.986717    88.96857   89.920299   90.841164   91.730464   92.587524   93.411691   94.202328   94.958804   95.680494    96.36677   97.017001   97.630548   98.206758   98.744967   99.244493   99.704636  100.124675  100.503866  100.841445   101.13662  101.388576  101.596474  101.759446       101.876603 |  57.955124   58.898992   59.907673   60.972966   62.086458   63.239872   64.425308   65.635403   66.863415   68.103255   69.349482    70.59727   71.842363   73.081023   74.309973   75.526344   76.727624   77.911609   79.076366   80.220188   81.341567   82.439161   83.511769   84.558312   85.577809   86.569364    87.53215   88.465397    89.36838    90.24041   91.080826   91.888989   92.664272   93.406058   94.113733   94.786682   95.424287   96.025918   96.590939   97.118696   97.608522   98.059729   98.471611   98.843439   99.174462   99.463905   99.710968   99.914829  100.074636    100.189518 | 1.175702  1.189571  1.204281  1.219696  1.235679  1.252102  1.268845  1.285798  1.302866  1.319962  1.337015  1.353961  1.370746  1.387326  1.403663  1.419724  1.435484   1.45092  1.466014  1.480751  1.495118  1.509105  1.522704  1.535906  1.548706  1.561099  1.573079  1.584644   1.59579  1.606512  1.616808  1.626676   1.63611  1.645109  1.653669  1.661787  1.669458  1.676679  1.683445  1.689751  1.695592  1.700964  1.705859  1.710271  1.714194   1.71762  1.720542  1.722951  1.724838  1.726194 | 1.354087  1.353399    1.3527  1.351999  1.351306  1.350626  1.349966   1.34933  1.348719  1.348137  1.347583  1.347058  1.346561  1.346093  1.345652  1.345236  1.344846  1.344479  1.344134   1.34381  1.343506   1.34322  1.342952  1.342701  1.342465  1.342243  1.342036  1.341841  1.341658  1.341487  1.341327  1.341178  1.341038  1.340907  1.340786  1.340673  1.340568  1.340471  1.340382    1.3403  1.340225  1.340157  1.340095  1.340041  1.339993  1.339951  1.339916  1.339887  1.339864          1.339848 | 0.577212  0.583164  0.589322  0.595611  0.601959  0.608307  0.614603  0.620804  0.626877  0.632795  0.638541    0.6441  0.649465   0.65463  0.659594  0.664357  0.668922  0.673291   0.67747  0.681463  0.685275  0.688913  0.692382  0.695687  0.698835   0.70183  0.704678  0.707383  0.709951  0.712386  0.714691  0.716871   0.71893   0.72087  0.722694  0.724405  0.726006  0.727499  0.728885  0.730166  0.731344  0.732419  0.733393  0.734266  0.735037  0.735708  0.736278  0.736746  0.737112      0.737374 | -0.252783  -0.259804  -0.267012  -0.274324  -0.281663  -0.288968  -0.296184  -0.303271  -0.310195  -0.316931  -0.323463  -0.329778   -0.33587  -0.341734   -0.34737  -0.352779  -0.357964  -0.362929   -0.36768  -0.372221  -0.376559    -0.3807  -0.384649  -0.388412  -0.391995  -0.395404  -0.398644  -0.401719  -0.404636  -0.407397  -0.410006  -0.412469  -0.414788  -0.416966  -0.419006  -0.420911  -0.422683  -0.424323  -0.425833  -0.427215  -0.428469  -0.429596  -0.430596   -0.43147  -0.432217  -0.432837  -0.433328   -0.43369  -0.433921     -0.434018 | 0.671248  0.672053  0.672891  0.673754  0.674634  0.675524  0.676418   0.67731  0.678196  0.679072  0.679935  0.680782  0.681612  0.682422  0.683212   0.68398  0.684726  0.685449  0.686148  0.686824  0.687476  0.688105  0.688709   0.68929  0.689847   0.69038  0.690889  0.691375  0.691837  0.692275  0.692691  0.693082  0.693451  0.693796  0.694118  0.694416  0.694691  0.694943   0.69517  0.695374  0.695554   0.69571  0.695842  0.695948   0.69603  0.696087  0.696117  0.696122    0.6961      0.696051 | 132.933235 | 136.047689  142.569805  149.741881  157.545711  165.957223  174.947941  184.486203  194.538162  205.068565  216.041342  227.420038  239.168124  251.249205  263.627163  276.266245   289.13111  302.186855  315.399018  328.733567   342.15689  355.635773  369.137374  382.629205  396.079111  409.455241  422.726038  435.860215  448.826743  461.594841  474.133961  486.413789  498.404235  510.075438  521.397768  532.341832  542.878488  552.978861   562.61436  571.756708  580.377969  588.450587  595.947431   602.84184  609.107685  614.719435  619.652228  623.881958  627.385368  630.140156          632.125091 |  632.94054 | 0.781435  0.791056  0.801252  0.811928   0.82299  0.834348  0.845918  0.857625  0.869402  0.881191  0.892942  0.904612  0.916164  0.927569  0.938799  0.949834  0.960657  0.971252  0.981608  0.991715  1.001564  1.011148  1.020463  1.029504  1.038266  1.046747  1.054943  1.062853  1.070474  1.077804  1.084841  1.091584   1.09803  1.104176  1.110023  1.115565  1.120803  1.125732   1.13035  1.134654   1.13864  1.142305  1.145645  1.148655  1.151331  1.153668  1.155661  1.157304   1.15859       1.159515 | 2.949073  3.006999   3.06875    3.1338  3.201621  3.271695  3.343535  3.416691  3.490755  3.565361  3.640188  3.714951  3.789403   3.86333  3.936545  4.008887  4.080215  4.150408  4.219362  4.286983  4.353192  4.417918  4.481098  4.542676  4.602602  4.660829  4.717315  4.772022  4.824913  4.875952  4.925106  4.972342  5.017628  5.060933  5.102224  5.141468  5.178634  5.213688  5.246595  5.277321  5.305828   5.33208  5.356037  5.377659  5.396904  5.413728  5.428086  5.439932  5.449217        5.455891 | 132.933235  139.126367  145.977135  153.470612  161.585236  170.294367  179.567629  189.372026  199.672819  210.434207  221.619827  233.193113  245.117553  257.356859  269.875076  282.636646  295.606441  308.749777  322.032411  335.420525  348.880708  362.379939   375.88556  389.365253   402.78702  416.119162  429.330261  442.389162  455.264963  467.927001  480.344847  492.488298   504.32738  515.832343  526.973676  537.722106  548.048619  557.924478   567.32124  576.210789   584.56537  592.357626  599.560646   606.14802  612.093898  617.373061  621.961003  625.834015  628.969286  631.345017   632.94054 |  100.0 |  32.933235 |  632.94054 |
    | 19:00 | 4.224173  4.310781  4.404173  4.503637   4.60841  4.717712  4.830775   4.94687  5.065314  5.185479  5.306799  5.428763  5.550917  5.672858   5.79423  5.914721  6.034054  6.151987  6.268305  6.382819  6.495362  6.605785  6.713954  6.819749  6.923063  7.023797  7.121859  7.217166   7.30964  7.399207  7.485796   7.56934  7.649774  7.727033  7.801055  7.871777  7.939136   8.00307  8.063514  8.120403  8.173673  8.223255  8.269081  8.311078  8.349174  8.383293  8.413357  8.439288  8.461001             8.478411 | 152.580804  157.575855  163.046271  168.968315   175.31339  182.049686  189.143564  196.560659  204.266715  212.228176  220.412581  228.788798  237.327149  245.999448  254.778977  263.640436  272.559855  281.514505  290.482799  299.444192  308.379088  317.268744  326.095192  334.841152  343.489968  352.025535  360.432242  368.694919  376.798785  384.729405  392.472651  400.014665  407.341828  414.440731  421.298153  427.901034  434.236463   440.29166  446.053961  451.510813  456.649766  461.458468  465.924671  470.036226  473.781098  477.147373  480.123274  482.697181  484.857658  486.593477 |  58.078285   58.961509   59.913924   60.928271   61.996749   63.111407   64.264433   65.448372   66.656266   67.881722   69.118946   70.362739   71.608468   72.852026   74.089787   75.318555   76.535519   77.738201   78.924417   80.092239   81.239958   82.366052   83.469164   84.548071   85.601669   86.628951   87.628994   88.600941   89.543994   90.457399    91.34044   92.192428   93.012695   93.800589   94.555467   95.276691   95.963621   96.615616   97.232026   97.812191   98.355438   98.861079   99.328406   99.756695  100.145198  100.493146  100.799747  101.064183  101.285614       101.463171 |  57.241734   58.107806   59.041726   60.036375   61.084104   62.177116   63.307751   64.468699   65.653136   66.854794   68.067993   69.287632    70.50917   71.728578   72.942303    74.14721   75.340541   76.519867   77.683048   78.828192   79.953623   81.057849   82.139539   83.197494   84.230632   85.237965   86.218588    87.17166     88.0964   88.992068    89.85796   90.693403   91.497741   92.270335   93.010553   93.717771   94.391362   95.030696   95.635136   96.204035   96.736732   97.232553   97.690805   98.110777   98.491735   98.832927   99.133574   99.392875   99.610005     99.784115 | 1.165149  1.177952   1.19166   1.20615  1.221297  1.236975  1.253065  1.269455  1.286046  1.302746  1.319478  1.336172  1.352769  1.369217  1.385475  1.401506  1.417278  1.432766  1.447947  1.462805  1.477324  1.491491  1.505294  1.518726  1.531779  1.544446  1.556721    1.5686  1.580077   1.59115  1.601814  1.612066  1.621903   1.63132  1.640315  1.648883  1.657021  1.664725  1.671991  1.678813  1.685188   1.69111  1.696573  1.701571  1.706098  1.710147   1.71371   1.71678  1.719348  1.721407 | 1.354631  1.353974  1.353298  1.352613  1.351928  1.351251  1.350588  1.349943  1.349321  1.348723  1.348153  1.347609  1.347094  1.346606  1.346144  1.345709  1.345298  1.344912  1.344548  1.344206  1.343884  1.343581  1.343297   1.34303  1.342778  1.342542  1.342321  1.342113  1.341917  1.341734  1.341562  1.341401   1.34125  1.341108  1.340977  1.340853  1.340739  1.340632  1.340534  1.340443  1.340359  1.340282  1.340212  1.340149  1.340092  1.340042  1.339999  1.339961   1.33993          1.339905 | 0.572586  0.578188  0.584048  0.590094  0.596254  0.602466  0.608674  0.614829  0.620893  0.626835   0.63263   0.63826  0.643714  0.648982   0.65406  0.658946  0.663639  0.668142  0.672458  0.676589  0.680542   0.68432  0.687929  0.691374   0.69466  0.697793  0.700777  0.703618   0.70632  0.708887  0.711323  0.713633   0.71582  0.717888  0.719839  0.721676  0.723403   0.72502  0.726531  0.727937   0.72924  0.730441  0.731541  0.732541   0.73344  0.734241  0.734942  0.735544  0.736046      0.736446 | -0.247193  -0.253837  -0.260731  -0.267791  -0.274942  -0.282115  -0.289251  -0.296302   -0.30323  -0.310003  -0.316599  -0.323001  -0.329198  -0.335181  -0.340948  -0.346497  -0.351828  -0.356946  -0.361852  -0.366551  -0.371049   -0.37535   -0.37946  -0.383384  -0.387127  -0.390696  -0.394095  -0.397329  -0.400403   -0.40332  -0.406086  -0.408704  -0.411178  -0.413511  -0.415705  -0.417764   -0.41969  -0.421485  -0.423151  -0.424688    -0.4261  -0.427386  -0.428546  -0.429582  -0.430494  -0.431281  -0.431942  -0.432477  -0.432884     -0.433163 | 0.670601  0.671352  0.672141  0.672961  0.673805  0.674664  0.675533  0.676405  0.677276  0.678141  0.678998  0.679842  0.680672  0.681485   0.68228  0.683056  0.683811  0.684546  0.685258  0.685948  0.686616  0.687261  0.687884  0.688483  0.689059  0.689613  0.690143  0.690651  0.691135  0.691597  0.692036  0.692452  0.692846  0.693216  0.693564  0.693889  0.694192  0.694471  0.694728  0.694961  0.695171  0.695358  0.695521   0.69566  0.695775  0.695865  0.695931  0.695971  0.695986      0.695975 | 128.453911 | 131.238302  137.090401  143.571957  150.672128   158.37353  166.653616  175.485921  184.841113  194.687854  204.993475  215.724492  226.846991  238.326899  250.130184  262.222981  274.571676  287.142959  299.903848  312.821697  325.864195  338.999355  352.195499  365.421242  378.645474  391.837339  404.966223  418.001734   430.91369  443.672104   456.24718  468.609299   480.72902  492.577072  504.124362  515.341973  526.201177  536.673442   546.73045  556.344119   565.48662  574.130414  582.248286  589.813383  596.799262  603.179949  608.929997   614.02456  618.439471  622.151331          625.137612 | 626.463357 | 0.774111  0.782997  0.792504  0.802546  0.813036  0.823886  0.835013  0.846339  0.857795   0.86932  0.880857  0.892361  0.903791  0.915112  0.926296  0.937316  0.948154  0.958791  0.969212  0.979407  0.989365  0.999077  1.008537  1.017739  1.026678   1.03535  1.043751  1.051879   1.05973  1.067302  1.074593  1.081601  1.088323  1.094757  1.100902  1.106754  1.112311  1.117571  1.122532  1.127189   1.13154  1.135581  1.139309  1.142719  1.145808   1.14857     1.151  1.153095  1.154847        1.15625 | 2.905195  2.958453  3.015747  3.076617  3.140577  3.207135  3.275814  3.346162  3.417762  3.490236  3.563242  3.636478  3.709677  3.782605  3.855058  3.926856  3.997844  4.067887  4.136866  4.204678  4.271232  4.336449  4.400257  4.462594  4.523403  4.582632  4.640236  4.696171  4.750397  4.802876  4.853572   4.90245  4.949478  4.994621  5.037847  5.079123  5.118416  5.155694  5.190921  5.224063  5.255084  5.283948  5.310616   5.33505  5.357207  5.377048  5.394527  5.409599  5.422219        5.432336 | 128.453911  133.989029  140.157436  146.951895  154.357903  162.355133  170.918765  180.020637  189.630205    199.7153  210.242724  221.178687   232.48913  244.139955  256.097183  268.327055  280.796099  293.471161  306.319426  319.308416  332.405985  345.580303  358.799843  372.033362  385.249881   398.41867  411.509229  424.491273  437.334721  450.009685  462.486458   474.73551  486.727485  498.433202  509.823653  520.870013  531.543648   541.81613  551.659252   561.04505  569.945833  578.334214  586.183145  593.465967  600.156453  606.228877  611.658071  616.419506  620.489372  623.844679  626.463357 |  100.0 |  28.453911 | 626.463357 |
    | 19:30 |  4.16017   4.23919  4.325176   4.41754  4.515608  4.618666   4.72599  4.836869  4.950625  5.066622  5.184278   5.30306   5.42249  5.542139  5.661626  5.780613  5.898799  6.015919  6.131739   6.24605  6.358669   6.46943  6.578186  6.684806  6.789169  6.891169  6.990706  7.087688  7.182031  7.273656  7.362487  7.448453  7.531486  7.611518  7.688485  7.762323  7.832968  7.900356  7.964423  8.025106  8.082338  8.136053  8.186183  8.232658  8.275407  8.314356   8.34943   8.38055  8.407637             8.430607 | 148.937618  153.441496  158.413403  163.836371  169.687671  175.940374   182.56477  189.529556  196.802798  204.352646  212.147853  220.158124  228.354328  236.708614  245.194446  253.786596  262.461099  271.195181  279.967182  288.756473  297.543368  306.309038  315.035435  323.705212  332.301652  340.808606  349.210432  357.491937  365.638333  373.635187  381.468381  389.124078  396.588687  403.848832  410.891327  417.703155  424.271443  430.583449  436.626546  442.388207  447.856001  453.017585    457.8607   462.37317  466.542908   470.35792  473.806316  476.876325  479.556309  481.834793 |  57.425575   58.231423   59.108318   60.050241   61.050342   62.101337   63.195832   64.326581   65.486664   66.669613   67.869468   69.080812   70.298763   71.518951   72.737484   73.950913   75.156178   76.350575   77.531711   78.697464   79.845951   80.975495   82.084596   83.171908   84.236215   85.276413   86.291491    87.28052   88.242635   89.177026    90.08293   90.959617   91.806387   92.622561   93.407475   94.160475   94.880912   95.568137     96.2215   96.840343   97.423998   97.971786   98.483013   98.956969   99.392923   99.790127  100.147809  100.465174  100.741406       100.975661 |  56.601699   57.391898   58.251764   59.175395   60.156076   61.186661   62.259902   63.368692   64.506247   65.666224   66.842779     68.0306   69.224898   70.421391   71.616261   72.806126   73.987985   75.159189   76.317388   77.460503   78.586687   79.694296   80.781859   81.848056   82.891695   83.911692   84.907059   85.876881   86.820313   87.736559    88.62487   89.484533   90.314859   91.115183   91.884855   92.623232   93.329678   94.003558   94.644233   95.251058   95.823379    96.36053   96.861829   97.326581    97.75407    98.14356   98.494296   98.805498   99.076365     99.306072 |  1.15563  1.167376  1.180072  1.193613  1.207886  1.222773  1.238157  1.253928  1.269983  1.286228   1.30258  1.318964  1.335316   1.35158  1.367707  1.383657  1.399393  1.414887  1.430113   1.44505  1.459679  1.473986  1.487957  1.501582  1.514851  1.527756  1.540291   1.55245  1.564226  1.575617  1.586616   1.59722  1.607426  1.617229  1.626625  1.635612  1.644184  1.652338  1.660071  1.667376   1.67425  1.680687  1.686683  1.692231  1.697326  1.701961  1.706128  1.709821  1.713032  1.715753 | 1.355136  1.354515  1.353868  1.353204  1.352533  1.351863  1.351201  1.350553  1.349923  1.349314  1.348729   1.34817  1.347637   1.34713   1.34665  1.346195  1.345765   1.34536  1.344977  1.344616  1.344277  1.343957  1.343656  1.343373  1.343106  1.342855  1.342619  1.342397  1.342189  1.341992  1.341808  1.341635  1.341473  1.341321  1.341178  1.341045  1.340921  1.340805  1.340697  1.340596  1.340503  1.340418  1.340339  1.340268  1.340203  1.340144  1.340092  1.340046  1.340007          1.339974 | 0.568338  0.573569  0.579103  0.584872  0.590808  0.596846  0.602928  0.609002  0.615025  0.620959  0.626776  0.632454  0.637975  0.643328  0.648504  0.653498  0.658309  0.662935  0.667378  0.671641  0.675727   0.67964  0.683385  0.686965  0.690387  0.693654  0.696772  0.699745  0.702578  0.705275   0.70784  0.710278  0.712591  0.714785   0.71686  0.718822  0.720671  0.722412  0.724045  0.725573  0.726998  0.728322  0.729544  0.730668  0.731692  0.732618  0.733447  0.734177  0.734809      0.735343 | -0.242038  -0.248276  -0.254819  -0.261588  -0.268507  -0.275505  -0.282519  -0.289496  -0.296392  -0.303169    -0.3098  -0.316262   -0.32254  -0.328623  -0.334503  -0.340177  -0.345642  -0.350899   -0.35595  -0.360799  -0.365448  -0.369902  -0.374167  -0.378246  -0.382145  -0.385869  -0.389423  -0.392811  -0.396038  -0.399109  -0.402027  -0.404797  -0.407422  -0.409906  -0.412251   -0.41446  -0.416537  -0.418483  -0.420299  -0.421989  -0.423553  -0.424993  -0.426309  -0.427502  -0.428572  -0.429519  -0.430344  -0.431045  -0.431622     -0.432074 | 0.670014  0.670709  0.671447  0.672222  0.673025  0.673849  0.674689  0.675537  0.676389   0.67724  0.678086  0.678924   0.67975  0.680563  0.681361  0.682142  0.682904  0.683647   0.68437  0.685073  0.685754  0.686414  0.687052  0.687668  0.688262  0.688834  0.689384  0.689912  0.690417  0.690901  0.691362  0.691801  0.692218  0.692613  0.692985  0.693336  0.693664   0.69397  0.694254  0.694515  0.694754   0.69497  0.695162  0.695332  0.695478  0.695601    0.6957  0.695774  0.695824      0.695848 | 124.530701 | 127.010719    132.2421  138.077884  144.514263  151.540408   159.13973  167.291068  175.969762  185.148563  194.798373  204.888839  215.388805  226.266647  237.490526  249.028563  260.848956  272.920065  285.210457  297.688936  310.324554  323.086613  335.944657  348.868464  361.828031  374.793562   387.73545  400.624267  413.430747   426.12578  438.680395  451.065757   463.25316  475.214025  486.919894  498.342437  509.453456   520.22489  530.628829  540.637528  550.223426  559.359168  568.017637  576.171984   583.79567  590.862506  597.346713  603.222974  608.466506  613.053132          616.959373 | 618.754992 |   0.7675  0.775656  0.784467  0.793858  0.803749  0.814058  0.824704   0.83561  0.846704  0.857922  0.869205  0.880503  0.891772  0.902973  0.914073  0.925045  0.935865  0.946512  0.956969  0.967224  0.977262  0.987076  0.996655  1.005993  1.015084  1.023923  1.032506  1.040828  1.048887  1.056679  1.064201  1.071452  1.078429  1.085129  1.091549  1.097689  1.103544  1.109114  1.114393  1.119381  1.124074  1.128468   1.13256  1.136346  1.139823  1.142985  1.145828  1.148348  1.150538       1.152394 | 2.865755  2.914438  2.967293  3.023936  3.083933  3.146829   3.21217  3.279511  3.348434  3.418552  3.489512  3.560994  3.632715  3.704422  3.775894  3.846935  3.917375  3.987063  4.055869  4.123676  4.190384  4.255904  4.320156   4.38307  4.444583  4.504637  4.563182   4.62017  4.675557  4.729302  4.781366  4.831713  4.880307  4.927114    4.9721  5.015232  5.056475  5.095797  5.133163  5.168538  5.201889  5.233178  5.262368  5.289421  5.314297  5.336957  5.357356  5.375453  5.391201        5.404554 | 124.530701  129.459284  134.992485  141.130297  147.865063  155.182738  163.064133  171.486067  180.422359  189.844652  199.723079  210.026775   220.72427  231.783775  243.173387  254.861238  266.815589  279.004891  291.397824  303.963313  316.670535  329.488913  342.388112  355.338022  368.308747  381.270588  394.194032  407.049735   419.80851   432.44132  444.919264  457.213572  469.295602  481.136834  492.708876   503.98346  514.932455  525.527872  535.741877  545.546814  554.915216  563.819841  572.233695  580.130072  587.482596  594.265268  600.452522  606.019289  610.941069  615.194008  618.754992 |  100.0 |  24.530701 | 618.754992 |
    | 20:00 | 4.103014  4.174715  4.253449  4.338754  4.430065  4.526748  4.628135   4.73355  4.842332   4.95385  5.067512  5.182773  5.299134  5.416145  5.533401  5.650541  5.767241  5.883215  5.998209  6.111993  6.224368  6.335153  6.444186  6.551323  6.656433  6.759399  6.860111  6.958471  7.054387  7.147774  7.238552  7.326645   7.41198  7.494489  7.574104  7.650758  7.724386  7.794925   7.86231  7.926475  7.987357  8.044889  8.099003  8.149633  8.196707  8.240153  8.279899  8.315867  8.347982             8.376161 | 145.718828   149.76196  154.260864  159.205235  164.578348  170.358465  176.520199  183.035739  189.875891  197.010897  204.411071  212.047246  219.891086  227.915279   236.09364  244.401154  252.813967  261.309353  269.865657  278.462225  287.079333  295.698113  304.300471  312.869027  321.387034  329.838326  338.207251  346.478621   354.63766  362.669958  370.561433  378.298287   385.86698   393.25419  400.446795   407.43184  414.196521  420.728162  427.014201  433.042174  438.799707  444.274503  449.454341  454.327065  458.880593  463.102912  466.982088  470.506273  473.663719  476.442793 |  56.842697   57.573908   58.376837   59.246781   60.177975   61.163954   62.197902   63.272929   64.382293   65.519555   66.678685   67.854117   69.040771   70.234054   71.429838   72.624433    73.81455   74.997261   76.169965   77.330348   78.476352   79.606138   80.718062    81.81065   82.882568   83.932611   84.959676   85.962754   86.940911   87.893276   88.819033   89.717409   90.587666   91.429095   92.241006   93.022727   93.773594    94.49295    95.18014   95.834503   96.455378   97.042089   97.593954   98.110273   98.590334   99.033403    99.43873   99.805541  100.133043       100.420418 |   56.03014   56.747151   57.534488   58.387539   59.300649   60.267481   61.281351   62.335502   63.423322     64.5385    65.67512   66.827726   67.991336   69.161446   70.334009   71.505406   72.672411   73.832155   74.982085   76.119935   77.243684    78.35153   79.441862   80.513232   81.564335   82.593986   83.601107   84.584706   85.543867   86.477738   87.385517   88.266447   89.119805   89.944893   90.741038   91.507578   92.243864   92.949251   93.623096   94.264752   94.873569   95.448887   95.990035   96.496328   96.967066   97.401531   97.798987   98.158675   98.479817     98.761611 | 1.147087  1.157798  1.169487  1.182068  1.195442  1.209501  1.224135  1.239236  1.254702  1.270436  1.286352  1.302372  1.318425   1.33445  1.350396  1.366216   1.38187  1.397324  1.412551  1.427524  1.442224  1.456631  1.470732  1.484513  1.497963  1.511072  1.523833  1.536238  1.548281  1.559956  1.571258  1.582183  1.592726  1.602883   1.61265  1.622023  1.630998  1.639571  1.647737  1.655493  1.662833  1.669754  1.676249  1.682313   1.68794  1.693125  1.697861   1.70214  1.705956  1.709301 | 1.355602   1.35502  1.354405  1.353768  1.353116  1.352459  1.351803  1.351156  1.350522  1.349905  1.349309  1.348736  1.348188  1.347664  1.347166  1.346693  1.346245  1.345821   1.34542  1.345041  1.344684  1.344347  1.344029  1.343729  1.343447  1.343181   1.34293  1.342695  1.342472  1.342263  1.342067  1.341882  1.341708  1.341545  1.341392  1.341248  1.341113  1.340987   1.34087   1.34076  1.340658  1.340564  1.340477  1.340396  1.340323  1.340256  1.340196  1.340142  1.340094          1.340053 | 0.564464  0.569312  0.574498  0.579963  0.585641  0.591471  0.597391  0.603349  0.609296  0.615192  0.621004  0.626703  0.632269  0.637687  0.642943  0.648031  0.652945  0.657683  0.662244   0.66663  0.670842  0.674883  0.678757  0.682468  0.686021  0.689419  0.692667  0.695769  0.698731  0.701556  0.704248  0.706811  0.709249  0.711565  0.713764  0.715847  0.717817  0.719678  0.721432   0.72308  0.724625  0.726068   0.72741  0.728654    0.7298  0.730848  0.731799  0.732654  0.733412      0.734074 | -0.237315  -0.243127  -0.249291  -0.255734  -0.262382  -0.269165  -0.276017  -0.282881  -0.289707  -0.296455   -0.30309  -0.309585   -0.31592  -0.322081  -0.328055  -0.333836   -0.33942  -0.344804   -0.34999  -0.354977  -0.359769  -0.364369  -0.368781  -0.373009  -0.377058  -0.380932  -0.384636  -0.388174  -0.391551  -0.394771  -0.397838  -0.400756  -0.403529   -0.40616  -0.408652  -0.411009  -0.413232  -0.415325  -0.417289  -0.419126  -0.420839  -0.422428  -0.423894  -0.425239  -0.426462  -0.427565  -0.428547  -0.429408  -0.430148     -0.430765 | 0.669483  0.670123  0.670809  0.671535  0.672295  0.673082  0.673889  0.674709  0.675539  0.676371  0.677203  0.678031  0.678851   0.67966  0.680458   0.68124  0.682007  0.682756  0.683488    0.6842  0.684893  0.685565  0.686217  0.686848  0.687458  0.688047  0.688615  0.689161  0.689686  0.690189  0.690671  0.691132  0.691571  0.691988  0.692384  0.692758  0.693111  0.693442  0.693752  0.694039  0.694305  0.694548   0.69477  0.694968  0.695145  0.695298  0.695428  0.695534  0.695617      0.695675 | 121.103761 | 123.304688  127.964252  133.199479  139.013098  145.400579  152.351196  159.849121  167.874458  176.404176  185.412893  194.873524  204.757799   215.03666  225.680563  236.659699  247.944157  259.504032  271.309501  283.330872  295.538612  307.903358  320.395927  332.987307  345.648651   358.35127  371.066619  383.766289   396.42199  409.005544  421.488875  433.844001  446.043026  458.058138  469.861604  481.425772  492.723069  503.726013  514.407215  524.739391  534.695379  544.248154  553.370854  562.036801  570.219541  577.892872  585.030896   591.60806  597.599219  602.979695          607.725353 | 609.950719 | 0.761564  0.769006  0.777122  0.785852  0.795126  0.804868  0.815001  0.825451  0.836145  0.847017  0.858007  0.869061  0.880131  0.891175  0.902158  0.913047  0.923816  0.934442  0.944906  0.955192  0.965284  0.975171  0.984844  0.994294  1.003513  1.012496  1.021237  1.029731  1.037975  1.045965  1.053697   1.06117  1.068379  1.075323  1.081999  1.088405  1.094537  1.100393  1.105971  1.111268   1.11628  1.121005  1.125438  1.129577  1.133418  1.136956  1.140188  1.143108  1.145711       1.147993 | 2.830476  2.874724  2.923211  2.975628  3.031607  3.090739  3.152601  3.216766  3.282824  3.350386  3.419089  3.488605  3.558633  3.628908   3.69919  3.769269  3.838958  3.908094  3.976533  4.044146  4.110823  4.176463  4.240979  4.304293  4.366335  4.427042  4.486357  4.544228  4.600608   4.65545  4.708715  4.760363  4.810357   4.85866  4.905236  4.950053  4.993074  5.034268  5.073599  5.111032  5.146534  5.180068  5.211598  5.241086  5.268494  5.293783  5.316911  5.337836  5.356514        5.372901 | 121.103761  125.476426  130.421653  135.946046  142.048432  148.720886  155.949836   163.71715   172.00113  180.777372  190.019483  199.699661  209.789144   220.25856   231.07818  242.218105  253.648401  265.339186  277.260692  289.383299  301.677556  314.114189  326.664101  339.298364  351.988213  364.705033  377.420351  390.105817  402.733202  415.274381  427.701326    439.9861  452.100849  464.017801  475.709263  487.147622  498.305349  509.155006  519.669252   529.82086  539.582732  548.927916  557.829636  566.261317  574.196619   581.60948  588.474161  594.765296  600.457955  605.527713  609.950719 |  100.0 |  21.103761 | 609.950719 |
    | 20:30 | 4.052202  4.116916  4.188621  4.266988  4.351564  4.441808  4.537123  4.636881  4.740453   4.84722  4.956594  5.068023  5.180995  5.295043  5.409742  5.524709    5.6396  5.754108  5.867958  5.980905  6.092728  6.203232  6.312242    6.4196  6.525163  6.628802  6.730401  6.829851  6.927054  7.021919  7.114359  7.204294  7.291648  7.376347  7.458322  7.537503  7.613826  7.687223  7.757629   7.82498  7.889211  7.950255  8.008047  8.062519  8.113602  8.161226  8.205319  8.245807  8.282615             8.315665 |  142.88472  146.498743  150.552065  155.040731  159.953977  165.275413  170.984263  177.056566  183.466242  190.185998  197.188041  204.444632  211.928488  219.613051  227.472666  235.482675  243.619461  251.860441  260.184045  268.569661  276.997584  285.448945  293.905653  302.350319  310.766201  319.137138  327.447495  335.682107  343.826232  351.865504  359.785892   367.57366  375.215336  382.697674  390.007633  397.132343   404.05909   410.77529  417.268471  423.526262  429.536374  435.286592  440.764768  445.958811  450.856687  455.446417  459.716079  463.653813   467.24783  470.486421 |  56.324511   56.984473   57.715724   58.514906   59.377414   60.297732   61.269756   62.287097    63.34332   64.432137   65.547541   66.683897   67.835992   68.999056   70.168759   71.341197   72.512864   73.680622   74.841667   75.993498   77.133877   78.260806   79.372494   80.467331   81.543866   82.600785   83.636891   84.651089   85.642371   86.609802   87.552509    88.46967   89.360507   90.224274   91.060255   91.867754   92.646091   93.394595   94.112603   94.799452   95.454477    96.07701   96.666374   97.221879   97.742827   98.228499   98.678163   99.091066   99.466436        99.803479 |  55.522017   56.169163   56.886214   57.669876   58.515635    59.41808   60.371229   61.368814   62.404526   63.472198   64.565942    65.68023   66.809953   67.950431   69.097419   70.247089   71.396003   72.541084   73.679583   74.809045   75.927279   77.032324   78.122424      79.196   80.251629   81.288024   82.304009   83.298512   84.270544   85.219188   86.143588   87.042939   87.916476   88.763469   89.583216   90.375035   91.138257   91.872225    92.57629     93.2498   93.892105   94.502549   95.080468   95.625186   96.136017   96.612258   97.053189   97.458074   97.826155     98.156652 | 1.139457  1.149168  1.159868   1.17149   1.18395  1.197155  1.211004  1.225393  1.240221  1.255394  1.270822  1.286424  1.302125  1.317862  1.333576  1.349218  1.364743  1.380113  1.395297  1.410266  1.424995  1.439465  1.453659  1.467559  1.481155  1.494434  1.507387  1.520005  1.532282   1.54421  1.555784  1.566998  1.577848  1.588328  1.598435  1.608164  1.617511  1.626471  1.635041  1.643216  1.650991  1.658362  1.665324  1.671871  1.677998    1.6837  1.688969  1.693799  1.698184  1.702116 | 1.356029  1.355488  1.354909  1.354302  1.353675  1.353035   1.35239  1.351748  1.351115  1.350494   1.34989  1.349307  1.348745  1.348207  1.347692  1.347202  1.346736  1.346294  1.345875  1.345479  1.345104   1.34475  1.344415  1.344099  1.343801   1.34352  1.343255  1.343005  1.342769  1.342547  1.342337   1.34214  1.341955   1.34178  1.341616  1.341462  1.341317  1.341181  1.341054  1.340935  1.340824   1.34072  1.340624  1.340535  1.340453  1.340378   1.34031  1.340248  1.340192          1.340142 | 0.560953  0.565413  0.570238  0.575376   0.58077  0.586359  0.592085  0.597893  0.603732  0.609559  0.615335  0.621029  0.626617  0.632077  0.637395  0.642559  0.647562  0.652399  0.657068  0.661566  0.665896  0.670058  0.674056  0.677892  0.681571  0.685095   0.68847  0.691699  0.694786  0.697735  0.700551  0.703238  0.705798  0.708236  0.710555  0.712758  0.714847  0.716826  0.718698  0.720463  0.722125  0.723685  0.725145  0.726507   0.72777  0.728937  0.730008  0.730984  0.731864      0.732649 | -0.233017  -0.238392  -0.244156  -0.250244  -0.256588  -0.263118   -0.26977  -0.276483  -0.283204  -0.289887  -0.296494  -0.302994  -0.309361  -0.315575  -0.321623  -0.327494   -0.33318  -0.338677  -0.343984  -0.349099  -0.354024  -0.358762  -0.363314  -0.367684  -0.371876  -0.375895  -0.379744  -0.383427   -0.38695  -0.390315  -0.393527  -0.396589  -0.399506  -0.402281  -0.404917  -0.407417  -0.409783   -0.41202  -0.414127  -0.416109  -0.417966    -0.4197  -0.421312  -0.422804  -0.424176  -0.425429  -0.426563  -0.427578  -0.428474      -0.42925 | 0.669006  0.669591  0.670225  0.670902  0.671617  0.672363  0.673133  0.673923  0.674726  0.675537  0.676351  0.677166  0.677976  0.678779  0.679572  0.680354  0.681122  0.681875  0.682613  0.683332  0.684034  0.684717  0.685381  0.686025   0.68665  0.687254  0.687837  0.688401  0.688943  0.689465  0.689966  0.690447  0.690906  0.691345  0.691763  0.692159  0.692535   0.69289  0.693224  0.693536  0.693827  0.694097  0.694345  0.694572  0.694776  0.694958  0.695118  0.695255  0.695369       0.69546 | 118.118088 | 120.064494  124.200134  128.879856  134.112308  139.898871   146.23448  153.108572  160.506046  168.408168  176.793378  185.637977  194.916697  204.603146  214.670169  225.090108   235.83501  246.876766  258.187213  269.738211  281.501682   293.44964  305.554208  317.787622   330.12223   342.53049  354.984957  367.458283  379.923198  392.352508  404.719081  416.995844  429.155771  441.171882  453.017237  464.664936  476.088117  487.259961  498.153695  508.742602  519.000028    528.8994  538.414241  547.518191  556.185033  564.388724  572.103429  579.303561  585.963831  592.059297          597.565427 | 600.181103 |  0.75626  0.763011  0.770444  0.778512  0.787158  0.796313  0.805909  0.815872  0.826132  0.836623  0.847284  0.858056  0.868891  0.879743  0.890573  0.901347  0.912033  0.922608  0.933049  0.943336  0.953455   0.96339  0.973132  0.982668  0.991992  1.001095  1.009971  1.018615  1.027022  1.035189   1.04311  1.050783  1.058205  1.065372  1.072283  1.078933  1.085321  1.091444  1.097299  1.102883  1.108194  1.113227   1.11798   1.12245  1.126633  1.130524   1.13412  1.137416  1.140408       1.143091 | 2.799062  2.839062  2.883296  2.931538  2.983489  3.038797  3.097076   3.15793  3.220962  3.285789  3.352046  3.419398  3.487534  3.556174  3.625066  3.693984  3.762729  3.831122  3.899005  3.966241  4.032705  4.098289  4.162896  4.226438  4.288841  4.350033  4.409952  4.468542  4.525751  4.581529  4.635833  4.688619   4.73985  4.789486   4.83749  4.883828  4.928463  4.971362   5.01249  5.051813  5.089295  5.124901  5.158595  5.190341  5.220101  5.247837  5.273508  5.297073  5.318491        5.337718 | 118.118088  121.983988  126.387922  131.342357  136.852113  142.915105  149.523254  156.663472  164.318606  172.468304  181.089764  190.158364  199.648163  209.532302   219.78331  230.373335  241.274309  252.458072  263.896458  275.561345  287.424695  299.458573  311.635156  323.926738   336.30572  348.744613  361.216022  373.692639  386.147235  398.552649   410.88178   423.10758  435.203046  447.141221  458.895183  470.438051  481.742983  492.783183  503.531902  513.962452  524.048215   533.76266  543.079364  551.972029  560.414519  568.380883  575.845398  582.782614  589.167399     594.975  600.181103 |  100.0 |  18.118088 | 600.181103 |
    | 21:00 | 4.007224  4.065334  4.130296  4.201911  4.279843  4.363653  4.452823  4.546789  4.644962  4.746752  4.851582  4.958901  5.068191  5.178974  5.290809  5.403295  5.516069  5.628804  5.741207  5.853014  5.963989  6.073921  6.182617  6.289909  6.395641  6.499673  6.601879  6.702143  6.800357  6.896424  6.990251  7.081753   7.17085  7.257465  7.341523  7.422955  7.501692  7.577665   7.65081   7.72106   7.78835  7.852615  7.913788  7.971803  8.026592  8.078086  8.126215  8.170906  8.212087             8.249681 |  140.39759  143.614704  147.251179   151.30892  155.782911  160.662111  165.930524  171.568306  177.552816  183.859555  190.462949  197.336984  204.455687  211.793481  219.325428  227.027386  234.876101  242.849242  250.925407  259.084101  267.305691  275.571361  283.863056  292.163418  300.455734  308.723877  316.952248  325.125727  333.229626  341.249638  349.171803  356.982462  364.668226  372.215944   379.61267  386.845642  393.902252  400.770026  407.436608  413.889739  420.117242  426.107013  431.847008  437.325234  442.529743  447.448633  452.070041  456.382148  460.373182  464.031426 |  55.865828   56.458433   57.120922   57.851248   58.646003     59.5007   60.410065   61.368335   62.369508   63.407563   64.476619   65.571062   66.685613   67.815381   68.955876    70.10301   71.253084   72.402764   73.549055   74.689269   75.820997   76.942079   78.050574   79.144738   80.222997   81.283924   82.326224   83.348713   84.350305   85.329996   86.286852   87.219998   88.128611   89.011909   89.869141   90.699586   91.502543   92.277325   93.023258   93.739671   94.425899   95.081272   95.705117   96.296755   96.855495   97.380634   97.871454   98.327222   98.747185        99.130572 |  55.072242   55.653338   56.302962   57.019106   57.798427   58.636526   59.528233   60.467893   61.449624    62.46752   63.515817   64.589006   65.681913   66.789742   67.908089   69.032947   70.160687   71.288041   72.412072   73.530144   74.639894   75.739205   76.826175   77.899091    78.95641   79.996735   81.018794   82.021427   83.003569   83.964235   84.902509   85.817535   86.708503   87.574647   88.415233   89.229551   90.016915   90.776652   91.508099     92.2106   92.883501   93.526148   94.137878   94.718027   95.265917   95.780858   96.262146   96.709063   97.120871     97.496813 | 1.132675  1.141432   1.15117  1.161844  1.173389  1.185725  1.198761  1.212403  1.226554   1.24112  1.256012  1.271146  1.286447  1.301846   1.31728  1.332696  1.348046  1.363288  1.378387  1.393309   1.40803  1.422524  1.436772  1.450757  1.464464  1.477879  1.490991  1.503791  1.516269  1.528419  1.540234  1.551707  1.562833  1.573607  1.584024  1.594079  1.603768  1.613086  1.622029  1.630593  1.638773  1.646564  1.653961   1.66096  1.667555   1.67374  1.679509  1.684857  1.689777  1.694261 | 1.356417  1.355918  1.355378  1.354805  1.354205  1.353587  1.352959  1.352327  1.351698  1.351077  1.350469  1.349878  1.349306  1.348755  1.348226   1.34772  1.347238  1.346779  1.346343  1.345929  1.345537  1.345166  1.344815  1.344482  1.344169  1.343872  1.343592  1.343327  1.343078  1.342842   1.34262  1.342411  1.342213  1.342027  1.341851  1.341686  1.341531  1.341385  1.341248  1.341119  1.340999  1.340887  1.340782  1.340684  1.340594   1.34051  1.340433  1.340363  1.340299          1.340242 | 0.557792  0.561867  0.566323  0.571119  0.576206  0.581528   0.58703  0.592656  0.598356  0.604082  0.609793  0.615455  0.621038  0.626518  0.631877    0.6371  0.642177  0.647099  0.651862  0.656463  0.660901  0.665176   0.66929  0.673246  0.677045  0.680691  0.684188  0.687539  0.690749   0.69382  0.696757  0.699565  0.702245  0.704802  0.707239  0.709559  0.711766  0.713861  0.715848  0.717729  0.719506  0.721181  0.722756  0.724232   0.72561  0.726893   0.72808  0.729173  0.730172      0.731076 |  -0.22913  -0.234066  -0.239417  -0.245129  -0.251139  -0.257385  -0.263801  -0.270326  -0.276906  -0.283491  -0.290039  -0.296512  -0.302883  -0.309128  -0.315227  -0.321167  -0.326938  -0.332533  -0.337948  -0.343179  -0.348227  -0.353092  -0.357775  -0.362281   -0.36661  -0.370767  -0.374755  -0.378579  -0.382242  -0.385748  -0.389101  -0.392304  -0.395362  -0.398277  -0.401053  -0.403693  -0.406199  -0.408575  -0.410823  -0.412944  -0.414942  -0.416817  -0.418571  -0.420206  -0.421722   -0.42312  -0.424401  -0.425565  -0.426611     -0.427541 |  0.66858  0.669111  0.669693  0.670321  0.670989  0.671692  0.672424   0.67318  0.673953  0.674739  0.675533   0.67633  0.677127   0.67792  0.678707  0.679485  0.680252  0.681007  0.681747  0.682472  0.683181  0.683873  0.684546  0.685202  0.685838  0.686456  0.687054  0.687633  0.688191   0.68873  0.689249  0.689748  0.690227  0.690685  0.691123  0.691541  0.691939  0.692316  0.692672  0.693008  0.693324  0.693618  0.693892  0.694144  0.694375  0.694585  0.694773  0.694939  0.695082      0.695204 | 115.523427 | 117.238921  120.897004  125.065566  129.758459  134.982473  140.737897  147.019279  153.816262  161.114439  168.896141  177.141147  185.827282  194.930918  204.427368  214.291202  224.496486  235.016959  245.826169  256.897561  268.204547  279.720551   291.41903  303.273496  315.257519   327.34473  339.508817  351.723522  363.962632   376.19997  388.409389  400.564767  412.639996  424.608979  436.445623  448.123843  459.617551  470.900663  481.947103  492.730802   503.22571  513.405806  523.245108  532.717694  541.797717  550.459434  558.677231  566.425659  573.679469  580.413658          586.603522 | 589.571078 | 0.751544  0.757633  0.764401  0.771816   0.77983  0.788388  0.797426  0.806878  0.816675  0.826753   0.83705  0.847507  0.858073  0.868698  0.879342  0.889967   0.90054  0.911033   0.92142  0.931682    0.9418  0.951758  0.961542  0.971141  0.980545  0.989745  0.998735  1.007507  1.016056  1.024377  1.032467   1.04032  1.047934  1.055304  1.062429  1.069305  1.075928  1.082297  1.088409   1.09426  1.099848   1.10517  1.110222  1.115001  1.119503  1.123726  1.127664  1.131314  1.134671       1.137731 | 2.771217  2.807185  2.847323  2.891484  2.939442  2.990906  3.045539  3.102979  3.162854  3.224791  3.288434  3.353442    3.4195  3.486316  3.553628  3.621197   3.68881  3.756274  3.823422  3.890101  3.956178  4.021534  4.086061  4.149666  4.212265   4.27378  4.334144  4.393293  4.451173  4.507729  4.562915  4.616684  4.668995  4.719807  4.769083  4.816784  4.862874  4.907319  4.950083  4.991131  5.030429  5.067941  5.103632  5.137465  5.169404  5.199411  5.227448  5.253473  5.277447        5.299328 | 115.523427  118.929757  122.837984  127.265606  132.222833  137.713023  143.733372  150.275745  157.327532   164.87249  172.891495  181.363198  190.264575  199.571369  209.258448  219.300071  229.670099  240.342141  251.289673  262.486108  273.904853  285.519341  297.303054   309.22953  321.272371  333.405235   345.60184  357.835951  370.081374  382.311951  394.501547   406.62405  418.653356  430.563374  442.328014  453.921189  465.316814  476.488806  487.411092   498.05761  508.402322  518.419221  528.082351  537.365821  546.243831  554.690693  562.680865  570.188987  577.189922  583.658801  589.571078 |  100.0 |  15.523427 | 589.571078 |
    | 21:30 | 3.967576  4.019499  4.078054  4.143163  4.214606  4.292049  4.375065   4.46316    4.5558   4.65243  4.752498  4.855462  4.960808  5.068049  5.176735  5.286449  5.396814  5.507485  5.618151  5.728531  5.838372  5.947447  6.055552  6.162501  6.268127  6.372281  6.474825  6.575634  6.674593  6.771596  6.866546  6.959351  7.049926  7.138191  7.224067  7.307482  7.388363  7.466643  7.542253  7.615127  7.685199  7.752403  7.816673  7.877944  7.936147  7.991216   8.04308  8.091671  8.136916             8.178741 | 138.221951  141.074338  144.323446  147.976417  152.033607  156.489185  161.331972  166.546411   172.11355  178.011974  184.218627  190.709501  197.460197  204.446345  211.643922  219.029468  226.580229  234.274239  242.090361  250.008286  258.008521  266.072351  274.181798  282.319572   290.46902   298.61407  306.739183  314.829305  322.869816  330.846487   338.74544   346.55311  354.256208  361.841688  369.296718  376.608653  383.765009  390.753441  397.561724   404.17773  410.589418  416.784815  422.752006  428.479124  433.954341  439.165863  444.101927  448.750797  453.100769  457.140171 |  55.461498   55.991006   56.588156   57.252133   57.980716   58.770488   59.617089   60.515484   61.460223   62.445663   63.466155   64.516193   65.590513   66.684163   67.792541   68.911416   70.036922   71.165548   72.294122    73.41978   74.539946   75.652299   76.754751   77.845421   78.922608   79.984774    81.03052    82.05857   83.067755   84.056998   85.025302   85.971735   86.895425   87.795547   88.671316   89.521982   90.346817   91.145118   91.916194   92.659366   93.373961   94.059309   94.714739   95.339576   95.933136   96.494728   97.023649    97.51918   97.980589        98.407124 |  54.675763   55.194988   55.780542   56.431625    57.14606   57.920495   58.750655   59.631604   60.557997     61.5243   62.524975   63.554622    64.60808   65.680491   66.767346   67.864493   68.968142   70.074851   71.181509   72.285308    73.38372   74.474472   75.555516   76.625005   77.681274   78.722814   79.748252   80.756338   81.745925   82.715958   83.665458   84.593512   85.499264   86.381906   87.240669   88.074815   88.883633   89.666431   90.422533   91.151273   91.851992   92.524031   93.166733   93.779435   94.361469   94.912155   95.430805   95.916714   96.369162     96.787414 | 1.126675  1.134528  1.143343  1.153092   1.16373  1.175191  1.187398  1.200266  1.213706  1.227626  1.241939  1.256561  1.271414  1.286427  1.301535   1.31668  1.331811  1.346881  1.361852  1.376689   1.39136  1.405841  1.420108  1.434142  1.447924  1.461442  1.474681   1.48763   1.50028  1.512621  1.524646  1.536349  1.547722   1.55876  1.569458   1.57981  1.589812  1.599459  1.608746   1.61767  1.626225  1.634406  1.642209  1.649629   1.65666  1.663298  1.669536  1.675369  1.680791  1.685794 | 1.356767   1.35631   1.35581  1.355273  1.354705  1.354113  1.353505  1.352888  1.352268  1.351651  1.351043  1.350447  1.349868  1.349307  1.348766  1.348246  1.347749  1.347274  1.346822  1.346391  1.345982  1.345594  1.345227  1.344878  1.344549  1.344237  1.343942  1.343663  1.343399   1.34315  1.342915  1.342692  1.342483  1.342285  1.342098  1.341922  1.341756  1.341599  1.341452  1.341314  1.341184  1.341063  1.340949  1.340843  1.340744  1.340652  1.340567  1.340488  1.340416          1.340351 | 0.554963   0.55866  0.562747  0.567193  0.571957   0.57699  0.582241  0.587657  0.593187  0.598782  0.604399  0.610001  0.615554  0.621031  0.626409  0.631671  0.636804  0.641796   0.64664  0.651332  0.655868  0.660247   0.66447  0.668537  0.672451  0.676214  0.679828  0.683298  0.686626  0.689816  0.692872  0.695797  0.698595  0.701268  0.703822  0.706257  0.708578  0.710788  0.712889  0.714883  0.716772   0.71856  0.720247  0.721835  0.723327  0.724722  0.726023  0.727229  0.728343      0.729364 | -0.225636  -0.230137   -0.23507  -0.240392  -0.246048  -0.251981  -0.258129  -0.264434  -0.270839  -0.277292  -0.283747  -0.290165  -0.296511   -0.30276  -0.308887  -0.314876  -0.320714   -0.32639  -0.331897  -0.337231   -0.34239  -0.347372  -0.352178   -0.35681  -0.361269  -0.365558   -0.36968  -0.373639  -0.377437   -0.38108  -0.384569  -0.387909  -0.391103  -0.394155  -0.397067  -0.399844  -0.402487  -0.404999  -0.407383  -0.409642  -0.411776  -0.413789  -0.415681  -0.417454   -0.41911  -0.420649  -0.422072  -0.423379  -0.424572     -0.425649 | 0.668201  0.668681  0.669212   0.66979   0.67041  0.671069  0.671761   0.67248  0.673221  0.673979  0.674749  0.675526  0.676307  0.677088  0.677865  0.678637    0.6794  0.680153  0.680894  0.681622  0.682335  0.683033  0.683715   0.68438  0.685027  0.685656  0.686267  0.686859  0.687433  0.687987  0.688522  0.689038  0.689534  0.690011  0.690468  0.690906  0.691324  0.691722    0.6921  0.692458  0.692796  0.693114  0.693412  0.693689  0.693945  0.694181  0.694395  0.694589  0.694761      0.694911 | 113.274094 | 114.781134  118.006087  121.706665  125.901139  130.601119  135.811847  141.532732  147.758047  154.477675  161.677869   169.34194  177.450888  185.983921    194.9189  204.232687  213.901426  223.900755  234.205968  244.792134  255.634184  266.706971  277.985311  289.444009  301.057879  312.801744  324.650445  336.578836  348.561782  360.574151  372.590812  384.586622  396.536428  408.415052  420.197297  431.857934  443.371703  454.713316  465.857451  476.778761  487.451873  497.851401  507.951953   517.72814  527.154599  536.206006  544.857099   553.08271  560.857789  568.157447          574.956992 | 578.239266 | 0.747371  0.752833  0.758962  0.765737  0.773125  0.781081  0.789549   0.79847   0.80778  0.817417   0.82732   0.83743  0.847693  0.858059  0.868484  0.878929  0.889357  0.899738  0.910044  0.920253  0.930342  0.940296  0.950098  0.959735  0.969197  0.978472  0.987552  0.996431  1.005101  1.013557  1.021794  1.029807  1.037592  1.045146  1.052466  1.059547  1.066387  1.072983  1.079331   1.08543  1.091276  1.096865  1.102195  1.107263  1.112065  1.116597  1.120856  1.124838  1.128539       1.131953 |  2.74664   2.77882  2.815051  2.855263  2.899304  2.946945  2.997905  3.051864   3.10848  3.167402  3.228283  3.290787  3.354596  3.419414  3.484967  3.551007  3.617308  3.683668  3.749904  3.815855  3.881375  3.946335   4.01062  4.074126  4.136761  4.198442  4.259094  4.318649  4.377047  4.434229  4.490145  4.544746  4.597987  4.649825  4.700219  4.749132  4.796525  4.842361  4.886606  4.929224   4.97018  5.009438  5.046964  5.082721  5.116673  5.148783  5.179014  5.207326   5.23368        5.258035 | 113.274094  116.265714  119.722278  123.665436  128.110098  133.064577  138.531029  144.506095  150.981647  157.945553  165.382414  173.274228  181.600965  190.341054  199.471768  208.969543  218.810217  228.969219  239.421701  250.142641  261.106918  272.289355  283.664758  295.207932  306.893693  318.696872  330.592317   342.55489  354.559456  366.580887  378.594046  390.573788  402.494947  414.332337  426.060745  437.654929  449.089613  460.339492   471.37923  482.183466  492.726818  502.983894    512.9293  522.537656  531.783611  540.641867    549.0872  557.094494  564.638767  571.695218  578.239266 |  100.0 |  13.274094 | 578.239266 |
    | 22:00 | 3.932765  3.978941  4.031463  4.090361  4.155527  4.226731  4.303642  4.385843  4.472865  4.564201  4.659331  4.757733  4.858901  4.962351  5.067625  5.174299   5.28198   5.39031  5.498962  5.607639  5.716073   5.82402  5.931263  6.037602   6.14286  6.246873  6.349495   6.45059  6.550037  6.647721  6.743538  6.837391   6.92919  7.018848  7.106286  7.191426  7.274195  7.354521  7.432334  7.507568  7.580154  7.650028  7.717122  7.781372   7.84271  7.901069  7.956383  8.008581  8.057593             8.103349 | 136.324682  138.843997  141.735441  145.010668   148.67491  152.727264  157.161284  161.965759  167.125596  172.622696  178.436782  184.546122  190.928138  197.559899  204.418497  211.481328  218.726286  226.131889  233.677364  241.342673  249.108526  256.956366  264.868338  272.827255  280.816551  288.820239  296.822861  304.809445  312.765457  320.676761  328.529578  336.310445  344.006185  351.603871  359.090796  366.454446  373.682474  380.762676  387.682971  394.431381  400.996015  407.365049  413.526721   419.46931  425.181133  430.650534   435.86588  440.815555  445.487961  449.871517 |  55.106492     55.5774   56.113022   56.713659   57.378223   58.104372   58.888706   59.727002   60.614454   61.545905   62.516038   63.519548   64.551266   65.606249   66.679839   67.767702   68.865841   69.970596   71.078632   72.186923   73.292733   74.393586    75.48725   76.571706   77.645127   78.705859   79.752398   80.783373    81.79753   82.793715   83.770862    84.72798   85.664146   86.578488   87.470182   88.338445   89.182523   90.001688   90.795234   91.562469    92.30271   93.015283   93.699515   94.354733   94.980261   95.575415   96.139502   96.671821   97.171652        97.638265 |  54.327651   54.789414   55.314635   55.903608   56.555267   57.267314   58.036417   58.858433   59.728652   60.642014   61.593308   62.577332   63.589014   64.623509   65.676251   66.742989   67.819803   68.903104   69.989623   71.076392   72.160727   73.240203   74.312629   75.376025   76.428601   77.468734    78.49495   79.505905   80.500367   81.477207   82.435378    83.37391   84.291896   85.188482    86.06286   86.914262   87.741948   88.545206   89.323342   90.075677   90.801543   91.500279   92.171223   92.813718   93.427098   94.010695   94.563828   95.085809   95.575934     96.033486 |  1.12139  1.128397  1.136333   1.14519  1.154938  1.165529    1.1769  1.188977  1.201679   1.21492  1.228616  1.242685  1.257048  1.271631  1.286368  1.301198  1.316065  1.330922  1.345724  1.360434  1.375019  1.389449  1.403698  1.417745   1.43157  1.445157  1.458491  1.471558  1.484348  1.496851  1.509057   1.52096  1.532551  1.543825  1.554775  1.565396  1.575683  1.585631  1.595235   1.60449  1.613391  1.621934  1.630114  1.637925  1.645364  1.652425  1.659101  1.665388   1.67128   1.67677 |  1.35708  1.356666  1.356207  1.355707  1.355173  1.354611  1.354027  1.353428  1.352821  1.352213  1.351608  1.351012  1.350428   1.34986  1.349309  1.348778  1.348267  1.347778   1.34731  1.346864  1.346439  1.346035  1.345651  1.345287  1.344941  1.344614  1.344304   1.34401  1.343733   1.34347  1.343221  1.342986  1.342764  1.342554  1.342355  1.342168  1.341991  1.341825  1.341667  1.341519   1.34138  1.341249  1.341126  1.341011  1.340904  1.340803   1.34071  1.340623  1.340543           1.34047 | 0.552446  0.555778  0.559502  0.563595  0.568026  0.572754  0.577732  0.582912  0.588244   0.59368  0.599175  0.604688  0.610185  0.615634  0.621009  0.626291   0.63146  0.636506  0.641417  0.646186  0.650809  0.655282  0.659605  0.663777  0.667798  0.671672  0.675398  0.678982  0.682424   0.68573    0.6889   0.69194  0.694853  0.697641  0.700308  0.702857   0.70529  0.707612  0.709824  0.711929  0.713929  0.715827  0.717624  0.719323  0.720924   0.72243  0.723841  0.725159  0.726384      0.727518 | -0.222513  -0.226592  -0.231109  -0.236032  -0.241319  -0.246917  -0.252772  -0.258826  -0.265023  -0.271312  -0.277643  -0.283974  -0.290268  -0.296493  -0.302624   -0.30864  -0.314524  -0.320263  -0.325847   -0.33127  -0.336527  -0.341615  -0.346534  -0.351283  -0.355864  -0.360278  -0.364527  -0.368615  -0.372544  -0.376318  -0.379939  -0.383412  -0.386738  -0.389923  -0.392969  -0.395878  -0.398654  -0.401299  -0.403816  -0.406208  -0.408476  -0.410622  -0.412649  -0.414557  -0.416348  -0.418024  -0.419585  -0.421031  -0.422364     -0.423584 | 0.667865  0.668297  0.668778  0.669307  0.669881  0.670495  0.671145  0.671825  0.672532  0.673259  0.674002  0.674756  0.675518  0.676283  0.677048   0.67781  0.678567  0.679316  0.680055  0.680784    0.6815  0.682202   0.68289  0.683562  0.684217  0.684856  0.685478  0.686083  0.686669  0.687237  0.687787  0.688319  0.688831  0.689325    0.6898  0.690256  0.690693  0.691111  0.691509  0.691888  0.692247  0.692587  0.692908  0.693208  0.693488  0.693748  0.693988  0.694208  0.694406      0.694583 | 111.328777 | 112.648524  115.482516  118.756727  122.493033  126.707214  131.408976  136.602269  142.285793  148.453634   155.09594  162.199586  169.748792  177.725672  186.110694  194.883075  204.021091  213.502323  223.303848  233.402389  243.774418  254.396239  265.244042  276.293941  287.522001  298.904253    310.4167  322.035324  333.736083  345.494907  357.287695  369.090312  380.878579  392.628271  404.315112  415.914768  427.402849  438.754902  449.946414   460.95281   471.74946  482.311678  492.614734  502.633859   512.34426  521.721133  530.739679   539.37513  547.602771  555.397967          562.736206 | 566.297499 | 0.743693  0.748569  0.754088  0.760246  0.767019  0.774374  0.782267  0.790644  0.799448  0.808621  0.818103  0.827836  0.837766  0.847842  0.858018  0.868252  0.878505  0.888744  0.898941  0.909068  0.919104  0.929028  0.938823  0.948475   0.95797  0.967297  0.976447  0.985411  0.994181  1.002751  1.011115  1.019269  1.027207  1.034925   1.04242  1.049687  1.056724  1.063528  1.070095  1.076422  1.082506  1.088344  1.093933   1.09927  1.104351  1.109172  1.113732  1.118024  1.122046       1.125794 | 2.725037  2.753688  2.786228  2.822657  2.862891  2.906769  2.954067  3.004513  3.057801  3.113609  3.171605  3.231465  3.292872   3.35553  3.419158  3.483499  3.548319  3.613404  3.678562  3.743618  3.808417  3.872821  3.936703  3.999953   4.06247  4.124164  4.184954  4.244766  4.303534  4.361195  4.417695  4.472982  4.527007  4.579724  4.631092   4.68107  4.729618  4.776698  4.822276  4.866313  4.908775  4.949626  4.988831  5.026355  5.062161  5.096213  5.128474  5.158906  5.187471        5.214129 | 111.328777  113.947931  116.994969    120.4948  124.466298  128.922161  133.869105  139.308297  145.235948  151.643993  158.520765  165.851653  173.619684  181.806032  190.390444   199.35159  208.667336  218.314967  228.271346  238.513045  249.016433  259.757744  270.713119  281.858643  293.170359  304.624282  316.196406  327.862697  339.599103  351.381537  363.185884  374.987987  386.763646  398.488611  410.138578  421.689185  433.116011  444.394573  455.500327  466.408669  477.094939  487.534428  497.702382  507.574018  517.124532  526.329119  535.162988   543.60139  551.619643  559.193162  566.297499 |  100.0 |  11.328777 | 566.297499 |
    | 22:30 | 3.902316  3.943198  3.990087  4.043107  4.102255  4.167402  4.238311  4.314652  4.396025   4.48198  4.572038  4.665707  4.762499  4.861935  4.963561  5.066945  5.171687  5.277417  5.383792  5.490504  5.597267  5.703827  5.809949  5.915422  6.020056  6.123676  6.226125  6.327258  6.426943  6.525061  6.621499  6.716155  6.808931  6.899738  6.988491  7.075109  7.159516  7.241638  7.321403  7.398742  7.473587  7.545871  7.615528  7.682492  7.746696  7.808075   7.86656  7.922086  7.974582             8.023978 | 134.675101  136.892049  139.455283  142.380198  145.676306  149.347235  153.391061   157.80088  162.565534  167.670407  173.098214  178.829735  184.844453    191.1211  197.638088  204.373844  211.307061  218.416876  225.682983  233.085704  240.606025  248.225604  255.926757  263.692439  271.506206  279.352178  287.215003  295.079808  302.932161  310.758034  318.543755  326.275979  333.941652  341.527973  349.022372  356.412476  363.686083  370.831143  377.835731   384.68803  391.376312  397.888919  404.214255  410.340766  416.256932   421.95126   427.41227  432.628496  437.588478  442.280763 |  54.795973   55.212892   55.691064   56.231767   56.834957    57.49933    58.22246   59.000988   59.830831   60.707404    61.62582   62.581065   63.568148   64.582205   65.618586   66.672904    67.74107     68.8193   69.904124    70.99237   72.081152   73.167848   74.250085   75.325707   76.392765   77.449487   78.494261    79.52562    80.54222   81.542828   82.526308   83.491606   84.437742   85.363795   86.268901   87.152238   88.013023   88.850503   89.663951   90.452658   91.215931   91.953088   92.663452    93.34635   94.001108   94.627049    95.22349   95.789739   96.325094        96.828838 |  54.023162   54.431985   54.900871   55.431074    56.02255   56.674021   57.383109   58.146518   58.960247   59.819797   60.720378   61.657073   62.624988   63.619352   64.635608   65.669452   66.716874   67.774166   68.837923   69.905036   70.972674   72.038268   73.099488   74.154223   75.200559    76.23676   77.261245   78.272576   79.269434   80.250612   81.214993   82.161546   83.089308   83.997378   84.884907    85.75109    86.59516   87.416376   88.214027   88.987418   89.735869   90.458711    91.15528   91.824917    92.46696   93.080746   93.665604   94.220857   94.745816     95.239778 | 1.116753  1.122976  1.130084  1.138088  1.146973  1.156708  1.167245  1.178522  1.190468  1.203004  1.216052  1.229531  1.243364  1.257477  1.271801  1.286273  1.300836  1.315437  1.330031  1.344575  1.359035  1.373376  1.387573  1.401599  1.415433  1.429056  1.442453  1.455607  1.468507  1.481142    1.4935  1.505574  1.517356  1.528838  1.540013  1.550876   1.56142  1.571642  1.581534  1.591093  1.600313   1.60919  1.617719  1.625894  1.633712  1.641166  1.648252  1.654963  1.661295   1.66724 |  1.35736  1.356986  1.356567  1.356107  1.355609  1.355078  1.354522  1.353945  1.353356  1.352759  1.352162  1.351569  1.350984  1.350411  1.349853  1.349312   1.34879  1.348288  1.347806  1.347345  1.346905  1.346486  1.346086  1.345706  1.345346  1.345003  1.344678   1.34437  1.344078  1.343801  1.343539  1.343291  1.343057  1.342834  1.342624  1.342426  1.342238   1.34206  1.341893  1.341735  1.341586  1.341445  1.341313  1.341189  1.341073  1.340964  1.340862  1.340768   1.34068          1.340598 | 0.550217  0.553203  0.556574  0.560318  0.564412  0.568823  0.573512  0.578435  0.583544  0.588794  0.594139  0.599537  0.604951  0.610347  0.615697  0.620975  0.626163  0.631244  0.636206  0.641039  0.645737  0.650293  0.654706  0.658974  0.663096  0.667073  0.670907  0.674599  0.678151  0.681567   0.68485  0.688001  0.691025  0.693925  0.696703  0.699363  0.701907  0.704338   0.70666  0.708874  0.710982  0.712988  0.714893    0.7167  0.718409  0.720022  0.721542  0.722968  0.724302      0.725545 | -0.219738   -0.22341   -0.22752  -0.232045  -0.236953    -0.2422  -0.247739  -0.253516  -0.259477   -0.26557  -0.271747  -0.277961  -0.284173  -0.290349  -0.296459  -0.302479  -0.308388   -0.31417  -0.319815  -0.325311  -0.330652  -0.335835  -0.340855  -0.345712  -0.350406  -0.354937  -0.359306  -0.363517  -0.367571   -0.37147  -0.375219   -0.37882  -0.382275  -0.385589  -0.388763  -0.391802  -0.394707  -0.397482  -0.400129   -0.40265  -0.405048  -0.407325  -0.409482  -0.411522  -0.413445  -0.415253  -0.416948  -0.418529  -0.419998     -0.421356 |  0.66757  0.667955   0.66839  0.668871  0.669398  0.669967  0.670574  0.671215  0.671885  0.672579  0.673293  0.674021  0.674761  0.675507  0.676257  0.677007  0.677755  0.678498  0.679233   0.67996  0.680676   0.68138  0.682072  0.682749  0.683412  0.684059   0.68469  0.685305  0.685903  0.686483  0.687047  0.687592   0.68812  0.688629   0.68912  0.689593  0.690048  0.690484  0.690901    0.6913  0.691679   0.69204  0.692381  0.692704  0.693007   0.69329  0.693554  0.693798  0.694021      0.694225 | 109.650285 | 110.802511  113.285216  116.172809  119.489955  123.255912  127.484306  132.183224  137.355532  142.999362  149.108686  155.673928  162.682555  170.119627  177.968277  186.210128  194.825636  203.794365  213.095215  222.706585  232.606512  242.772767  253.182922  263.814411  274.644558  285.650606  296.809728  308.099039  319.495595  330.976398  342.518391  354.098455  365.693405  377.279985  388.834861  400.334623  411.755776  423.074737  434.267839  445.311323  456.181346  466.853978  477.305211  487.510959  497.447076  507.089356  516.413557   525.39541  534.010643  542.235002          550.044282 | 553.850514 | 0.740465  0.744796  0.749742  0.755308  0.761485  0.768249  0.775566  0.783392  0.791677  0.800367  0.809405  0.818736  0.828305  0.838063   0.84796  0.857953  0.868002  0.878072   0.88813   0.89815  0.908105  0.917974  0.927738   0.93738  0.946886  0.956244  0.965441  0.974469  0.983318  0.991983  1.000455  1.008729    1.0168  1.024664  1.032315  1.039751  1.046967   1.05396  1.060726  1.067263  1.073567  1.079635  1.085464   1.09105  1.096391  1.101483  1.106323  1.110906  1.115229       1.119289 | 2.706121  2.731514  2.760598  2.793435  2.830007  2.870215  2.913897   2.96083  3.010754  3.063376  3.118391  3.175488  3.234361  3.294712  3.356262  3.418747  3.481926  3.545575  3.609492  3.673493  3.737414  3.801104  3.864431  3.927273  3.989523  4.051082  4.111864  4.171788  4.230782  4.288781  4.345723  4.401554  4.456222  4.509679   4.56188  4.612781  4.662343  4.710525  4.757291  4.802603  4.846426  4.888723  4.929459  4.968598  5.006105  5.041944  5.076079  5.108472  5.139085        5.167881 | 109.650285  111.936415  114.613875  117.709968  121.246765   125.24073  129.702665  134.637936  140.046896  145.925432  152.265578  159.056123  166.283189  173.930748  181.981076  190.415126   199.21284  208.353395    217.8154  227.577041  237.616196  247.910522  258.437507  269.174523  280.098847  291.187683  302.418173  313.767402  325.212398  336.730135  348.297525  359.891417  371.488592  383.065755  394.599536  406.066482  417.443056  428.705632  439.830497   450.79385  461.571803  472.140386  482.475549  492.553175  502.349084  511.839048  520.998806  529.804082  538.230606  546.254142  553.850514 |  100.0 |   9.650285 | 553.850514 |
    | 23:00 | 3.875779   3.91182  3.953491  4.000997  4.054424  4.113741  4.178804  4.249368  4.325111  4.405645  4.490543  4.579348  4.671598   4.76683  4.864596  4.964465  5.066033  5.168918  5.272771  5.377268  5.482112  5.587034  5.691787  5.796148  5.899912  6.002896  6.104929   6.20586  6.305546   6.40386  6.500682  6.595902  6.689419  6.781138  6.870969  6.958827  7.044634  7.128311  7.209786  7.288986  7.365844   7.44029  7.512258  7.581682  7.648495  7.712632  7.774026   7.83261  7.888316             7.941077 | 133.244993  135.188985  137.452813  140.054822  143.008149  146.320466  149.994058  154.026174  158.409581  163.133233  168.183001  173.542371  179.193106  185.115812   191.29042  197.696571  204.313916  211.122343  218.102129   225.23405  232.499448  239.880256  247.359017  254.918867  262.543516  270.217224  277.924762  285.651379   293.38276  301.104995  308.804535  316.468161  324.082949  331.636238  339.115599  346.508807  353.803819  360.988742  368.051818  374.981401  381.765935  388.393941  394.854001  401.134741  407.224823   413.11293  418.787761  424.238019   429.45241  434.419636 |  54.525344   54.892894   55.317854    55.80232   56.347178   56.952096   57.615603   58.335223   59.107649   59.928943   60.794728   61.700371   62.641138   63.612322   64.609341   65.627813   66.663598   67.712829   68.771923   69.837585   70.906793   71.976791   73.045068   74.109342   75.167537   76.217765   77.258309   78.287601   79.304207   80.306812   81.294205   82.265268    83.21896   84.154311   85.070409   85.966394   86.841449   87.694792   88.525674   89.333367   90.117163   90.876369   91.610304   92.318289   92.999653   93.653721   94.279817   94.877259   95.445357        95.983409 |  53.757789   54.118201   54.534909   55.009967   55.544243   56.137415   56.788037   57.493682   58.251108   59.056453   59.905426   60.793482   61.715979   62.668303   63.645961   64.644655   65.660325   66.689181   67.727709   68.772676   69.821121    70.87034   71.917872   72.961478   73.999124   75.028958   76.049295   77.058598   78.055462   79.038598   80.006817   80.959022   81.894194   82.811381   83.709689   84.588274   85.446336   86.283109   87.097855   87.889863   88.658438   89.402902   90.122583    90.81682   91.484952   92.126319   92.740257   93.326097   93.883162     94.410766 | 1.112702  1.118202  1.124538  1.131734  1.139791  1.148693  1.158407  1.168883  1.180062  1.191875  1.204249  1.217107  1.230375   1.24398  1.257853  1.271928  1.286146  1.300452  1.314798  1.329139  1.343435  1.357653   1.37176  1.385731  1.399541  1.413169  1.426597  1.439809  1.452789  1.465526  1.478008  1.490226   1.50217  1.513832  1.525205  1.536283  1.547059  1.557527  1.567682  1.577518  1.587031  1.596215  1.605066  1.613578  1.621747  1.629567  1.637033  1.644141  1.650883  1.657255 | 1.357607  1.357272  1.356893  1.356471   1.35601  1.355514  1.354987  1.354437  1.353868  1.353288  1.352702  1.352115  1.351532  1.350958  1.350396  1.349848  1.349317  1.348804   1.34831  1.347835  1.347381  1.346947  1.346532  1.346137  1.345761  1.345404  1.345064  1.344741  1.344435  1.344145  1.343869  1.343608  1.343361  1.343126  1.342904  1.342694  1.342495  1.342306  1.342128   1.34196  1.341801  1.341652   1.34151  1.341377  1.341252  1.341135  1.341024  1.340921  1.340825          1.340736 | 0.548254  0.550915  0.553948  0.557351  0.561108  0.565197  0.569585  0.574233  0.579099  0.584139  0.589309  0.594566  0.599872   0.60519  0.610489  0.615743  0.620929  0.626028  0.631024  0.635906  0.640664  0.645292  0.649785  0.654139  0.658354  0.662427  0.666361  0.670156  0.673814  0.677336  0.680726  0.683986  0.687118  0.690126  0.693013   0.69578  0.698433  0.700972    0.7034  0.705721  0.707936  0.710048  0.712059  0.713971  0.715786  0.717505   0.71913  0.720662  0.722102      0.723452 | -0.217285  -0.220572  -0.224287  -0.228419  -0.232946  -0.237833  -0.243039  -0.248517  -0.254216  -0.260087  -0.266079  -0.272148   -0.27825  -0.284349  -0.290411  -0.296411  -0.302323   -0.30813  -0.313816  -0.319369  -0.324781  -0.330044  -0.335154  -0.340108  -0.344906  -0.349545  -0.354028  -0.358354  -0.362526  -0.366547  -0.370417  -0.374141  -0.377721  -0.381159  -0.384459  -0.387623  -0.390654  -0.393555  -0.396328  -0.398976  -0.401501  -0.403904  -0.406189  -0.408356  -0.410408  -0.412345   -0.41417  -0.415882  -0.417483     -0.418973 |  0.66731  0.667653  0.668043  0.668479   0.66896  0.669484  0.670048  0.670648   0.67128   0.67194  0.672622  0.673323  0.674038  0.674763  0.675495  0.676231  0.676967    0.6777  0.678429  0.679152  0.679866   0.68057  0.681264  0.681944  0.682612  0.683266  0.683904  0.684528  0.685136  0.685727  0.686302   0.68686  0.687401  0.687925  0.688431   0.68892  0.689391  0.689844  0.690278  0.690695  0.691094  0.691474  0.691836  0.692179  0.692503  0.692808  0.693094  0.693361  0.693609      0.693837 | 108.205287 |  109.20832  111.376768  113.915382  116.850836  120.205156  123.995303  128.233044   132.92508  138.073371  143.675592  149.725659  156.214282  163.129487   170.45711  178.181225   186.28451  194.748556  203.554111  212.681282  222.109688  231.818577  241.786918  251.993466  262.416809  273.035401  283.827587  294.771613  305.845638  317.027737  328.295898  339.628025  351.001934  362.395348  373.785891  385.151089  396.468363  407.715024  418.868277   429.90521  440.802803  451.537924   462.08733  472.427674  482.535513  492.387308  501.959443  511.228231  520.169934   528.76078          536.976984 | 540.995797 | 0.737645  0.741474  0.745884   0.75089  0.756493  0.762681  0.769429  0.776703   0.78446  0.792653  0.801229  0.810136   0.81932  0.828732  0.838323  0.848047  0.857865  0.867737  0.877631  0.887516  0.897364  0.907154  0.916862  0.926472  0.935966  0.945331  0.954555  0.963626  0.972535  0.981273  0.989834   0.99821  1.006396  1.014386  1.022176  1.029762  1.037138  1.044303  1.051251  1.057979  1.064485  1.070765  1.076815  1.082634  1.088216  1.093559   1.09866  1.103515   1.10812       1.112471 | 2.689621  2.712027  2.737902  2.767359  2.800437  2.837101  2.877245  2.920701  2.967253  3.016649  3.068612  3.122852  3.179075  3.236992  3.296325  3.356809  3.418196  3.480257  3.542781  3.605575  3.668465  3.731292  3.793914  3.856201  3.918038   3.97932  4.039951  4.099847  4.158929  4.217127  4.274376  4.330615  4.385791  4.439851  4.492749  4.544438  4.594877  4.644024  4.691841   4.73829  4.783333  4.826934  4.869058  4.909668   4.94873  4.986208  5.022065  5.056266  5.088773        5.119548 | 108.205287  110.194932  112.540372  115.270486   118.40979  121.977856  125.989051  130.452548   135.37257  140.748798  146.576882  152.848985  159.554335  166.679739  174.210041  182.128525  190.417249  199.057323  208.029124  217.312478   226.88679  236.731146  246.824393   257.14519   267.67205  278.383368  289.257434   300.27245  311.406531  322.637709  333.943931  345.303056  356.692855  368.091002  379.475074  390.822544  402.110779  413.317038  424.418466    435.3921   446.21486   456.86356  467.314903  477.545492  487.531834  497.250347  506.677378  515.789208  524.562078  532.972203  540.995797 |  100.0 |   8.205287 | 540.995797 |
    | 23:30 |  3.85273  3.884375  3.921249  3.963622  4.011658  4.065411  4.124827  4.189747  4.259927  4.335048  4.414739  4.498589  4.586169  4.677038  4.770761  4.866913  4.965091  5.064908  5.166007  5.268055  5.370743   5.47379  5.576937  5.679948  5.782608  5.884721  5.986107  6.086603   6.18606   6.28434  6.381317  6.476874  6.570905  6.663307  6.753987  6.842857  6.929833  7.014835  7.097787  7.178615  7.257249  7.333619  7.407657  7.479297  7.548473  7.615117  7.679166  7.740551  7.799207             7.855066 | 132.008587  133.707479  135.699711  138.005812  140.641862  143.619021  146.943384  150.616106  154.633773  158.988922  163.670671  168.665372  173.957247  179.528973  185.362192  191.437937  197.736981  204.240099  210.928274  217.782838   224.78557  231.918751  239.165203  246.508289  253.931913  261.420501  268.958974   276.53272  284.127557  291.729708  299.325755  306.902617  314.447507   321.94791  329.391546  336.766347   344.06043  351.262071  358.359683  365.341797   372.19704  378.914118  385.481802  391.888909  398.124293   404.17683   410.03541  415.688927   421.12627  426.336323 |   54.29029   54.613009   54.989052   55.421169   55.911043   56.459224   57.065145   57.727204   58.442898   59.208988   60.021678    60.87679   61.769927   62.696613   63.652403   64.632974   65.634187   66.652132   67.683145    68.72383    69.77105   70.821926   71.873824   72.924335   73.971264   75.012612   76.046551   77.071414   78.085677    79.08794   80.076916   81.051417   82.010342   82.952665   83.877428   84.783727   85.670708   86.537561   87.383506   88.207797   89.009707   89.788532   90.543579   91.274167   91.979619   92.659265   93.312431   93.938443   94.536619        95.106272 |  53.527299   53.843752   54.212491   54.636217   55.116579   55.654114   56.248268    56.89747   57.599266   58.350479   59.147387   59.985894   60.861687   61.770377   62.707606   63.669135   64.650906   65.649082   66.660074   67.680549   68.707433   69.737902   70.769372   71.799483   72.826082   73.847207   74.861068   75.866029   76.860596   77.843395   78.813166   79.768743   80.709046    81.63307   82.539875   83.428574   84.298331   85.148349   85.977867   86.786151   87.572489   88.336189   89.076574   89.792974   90.484727   91.151174   91.791657   92.405511   92.992072     93.550662 | 1.109176  1.114016  1.119638  1.126075  1.133345  1.141443  1.150352  1.160035  1.170445  1.181524  1.193204  1.205417  1.218091  1.231155   1.24454  1.258181  1.272016  1.285989  1.300049  1.314148  1.328246  1.342304  1.356288   1.37017  1.383923  1.397524  1.410953  1.424191  1.437222  1.450033  1.462611  1.474945  1.487024   1.49884  1.510385  1.521651  1.532632  1.543321  1.553713  1.563801   1.57358  1.583046  1.592193  1.601015  1.609509  1.617669  1.625489  1.632965   1.64009   1.64686 | 1.357825  1.357527  1.357186  1.356802  1.356378  1.355917  1.355423    1.3549  1.354356  1.353795  1.353224  1.352647  1.352071  1.351499  1.350935  1.350383  1.349845  1.349323  1.348818  1.348332  1.347864  1.347416  1.346988  1.346578  1.346187  1.345815  1.345461  1.345124  1.344804  1.344499   1.34421  1.343936  1.343676  1.343429  1.343195  1.342973  1.342762  1.342563  1.342374  1.342196  1.342027  1.341868  1.341717  1.341575   1.34144  1.341314  1.341196  1.341084   1.34098          1.340882 | 0.546533  0.548892  0.551606  0.554679  0.558106  0.561872  0.565951  0.570313  0.574919  0.579728    0.5847  0.589792  0.594964   0.60018  0.605406  0.610613  0.615776  0.620873  0.625886  0.630801  0.635606  0.640292  0.644853  0.649283  0.653581  0.657744   0.66177  0.665662  0.669419  0.673043  0.676536    0.6799  0.683137   0.68625  0.689242  0.692116  0.694873  0.697517  0.700051  0.702476  0.704796  0.707012  0.709126  0.711142   0.71306  0.714882  0.716611  0.718246   0.71979      0.721244 | -0.215126  -0.218052  -0.221391  -0.225141   -0.22929  -0.233812  -0.238675  -0.243837  -0.249253  -0.254876  -0.260658  -0.266553  -0.272517  -0.278512  -0.284501  -0.290456  -0.296349  -0.302159  -0.307868  -0.313461  -0.318927  -0.324257  -0.329444  -0.334485  -0.339375  -0.344114  -0.348701  -0.353136   -0.35742  -0.361555  -0.365542  -0.369384  -0.373083  -0.376642  -0.380063  -0.383348  -0.386502  -0.389525  -0.392421  -0.395192   -0.39784  -0.400367  -0.402776  -0.405067  -0.407244  -0.409307  -0.411257  -0.413097  -0.414826     -0.416445 | 0.667083  0.667387  0.667735  0.668128  0.668565  0.669045  0.669566  0.670125  0.670718  0.671341   0.67199  0.672661   0.67335  0.674052  0.674763  0.675482  0.676204  0.676926  0.677646  0.678362  0.679072  0.679774  0.680467  0.681149   0.68182  0.682478  0.683123  0.683753  0.684369   0.68497  0.685555  0.686125  0.686678  0.687214  0.687734  0.688237  0.688723  0.689192  0.689643  0.690077  0.690493  0.690892  0.691272  0.691634  0.691979  0.692305  0.692612  0.692901  0.693171      0.693422 | 106.964036 |  107.83474  109.723235  111.948232  114.537679  117.515684   120.90193  124.711364  128.954138  133.635769  138.757455   144.31651  150.306838  156.719436  163.542862  170.763671  178.366805  186.335911  194.653621  203.301771  212.261581  221.513792  231.038773  240.816606   250.82714  261.050041   271.46482  282.050852  292.787392  303.653585  314.628463  325.690953   336.81987  347.993919  359.191689   370.39165  381.572149  392.711407  403.787515  414.778434  425.661989  436.415876  447.017654  457.444754  467.674479  477.684011  487.450417  496.950659  506.161606  515.060048          523.622716 | 527.823549 | 0.735189  0.738559  0.742473  0.746953   0.75201  0.757642  0.763833   0.77056  0.777787  0.785474  0.793574  0.802039  0.810817   0.81986  0.829119  0.838549  0.848108  0.857756  0.867459  0.877183    0.8869  0.896585  0.906214  0.915768  0.925228   0.93458  0.943809  0.952902   0.96185  0.970644  0.979274  0.987733  0.996015  1.004114  1.012025  1.019742  1.027262   1.03458  1.041692  1.048596  1.055286   1.06176  1.068015  1.074047  1.079853  1.085429  1.090773  1.095881  1.100748       1.105372 | 2.675279  2.694968  2.717885  2.744187  2.773964  2.807233  2.843946   2.88399  2.927196  2.973353   3.02222  3.073532  3.127013  3.182386   3.23938   3.29773  3.357187  3.417517  3.478503  3.539944  3.601657  3.663477   3.72525  3.786841  3.848126  3.908991  3.969335  4.029067  4.088103  4.146367  4.203789  4.260305  4.315857   4.37039  4.423854    4.4762  4.527385  4.577365  4.626101  4.673552  4.719681  4.764451  4.807825  4.849768  4.890245  4.929219  4.966655  5.002516  5.036768        5.069372 | 106.964036  108.690799  110.739252   113.13911  115.916605  119.093761  122.687947  126.711701  131.172791  136.074474  141.415879  147.192483  153.396602  160.017887  167.043784  174.459944  182.250579   190.39876  198.886666  207.695783  216.807055  226.201013  235.857861   245.75755  255.879828  266.204275  276.710329  287.377302  298.184392  309.110686  320.135164  331.236696  342.394044  353.585855  364.790658  375.986865  387.152758  398.266495  409.306102  420.249473  431.074368  441.758413  452.279104  462.613806  472.739759  482.634086  492.273799  501.635809  510.696943  519.433955  527.823549 |  100.0 |   6.964036 | 527.823549 |
    | 00:00 | 3.832776  3.860455  3.892951  3.930581  3.973575  4.022062  4.076068  4.135519  4.200248  4.270009  4.344491  4.423336  4.506153  4.592532  4.682059  4.774319  4.868914  4.965459  5.063591   5.16297  5.263279  5.364226   5.46554  5.566975  5.668305  5.769321  5.869836  5.969675   6.06868  6.166705  6.263617  6.359292  6.453616  6.546483  6.637794  6.727455  6.815379  6.901484  6.985689   7.06792  7.148104  7.226169  7.302047  7.375669   7.44697  7.515883  7.582343  7.646284  7.707639             7.766343 |  130.94249  132.422396  134.169578  136.206034  138.550104  141.215847  144.212698  147.545395  151.214147  155.215006  159.540368  164.179546   169.11937  174.344757  179.839237  185.585411  191.565328  197.760801  204.153643  210.725846  217.459718  224.337963   231.34374  238.460694   245.67296   252.96516  260.322393  267.730206  275.174567  282.641844  290.118764  297.592387  305.050077  312.479466  319.868433  327.205068  334.477656  341.674644  348.784623  355.796306  362.698508  369.480127  376.130128   382.63753  388.991385  395.180775  401.194791  407.022531  412.653086  418.075536 |  54.086797   54.369072   54.700462   55.084215   55.522671   56.017146   56.567905   57.174188   57.834295   58.545716   59.305288   60.109353   60.953924   61.834825   62.747815   63.688694   64.653375   65.637945     66.6387    67.65217   68.675125   69.704582   70.737791   71.772231   72.805592   73.835764   74.860814   75.878976   76.888633     77.8883   78.876613   79.852312   80.814233   81.761292   82.692479   83.606847   84.503503     85.3816   86.240333   87.078929   87.896643   88.692753   89.466557   90.217364   90.944493   91.647273   92.325032     92.9771   93.602804        94.201466 |  53.327758   53.604551   53.929505   54.305807   54.735748   55.220621   55.760685   56.355194   57.002482   57.700088   58.444909    59.23336    60.06153   60.925324   61.820585   62.743193    63.68914    64.65459   65.635911     66.6297   67.632791   68.642256     69.6554   70.669752   71.683047   72.693213   73.698357   74.696748   75.686798   76.667052   77.636172   78.592924   79.536165   80.464833   81.377937   82.274549   83.153792   84.014837   84.856894   85.679205   86.481039    87.26169   88.020467   88.756693   89.469703   90.158835   90.823432   91.462837   92.076391     92.663427 | 1.106117  1.110359  1.115325  1.121058  1.127584  1.134915  1.143045   1.15195  1.161597  1.171936  1.182912   1.19446  1.206515  1.219009  1.231874  1.245046  1.258464  1.272068  1.285805  1.299628  1.313491  1.327354  1.341181  1.354941  1.368605  1.382148  1.395547  1.408781  1.421835  1.434692  1.447337   1.45976  1.471948  1.483892  1.495583  1.507012  1.518173  1.529058  1.539661  1.549976  1.559997  1.569719  1.579137  1.588244  1.597038  1.605511  1.613659  1.621477  1.628959    1.6361 | 1.358016  1.357752  1.357447    1.3571  1.356713  1.356288  1.355827  1.355335  1.354818  1.354279  1.353726  1.353163  1.352596   1.35203  1.351468  1.350914  1.350372  1.349843  1.349329  1.348833  1.348354  1.347893  1.347451  1.347028  1.346623  1.346237  1.345869  1.345517  1.345183  1.344865  1.344563  1.344275  1.344002  1.343742  1.343496  1.343262  1.343041   1.34283  1.342631  1.342442  1.342263  1.342093  1.341933  1.341781  1.341638  1.341503  1.341376  1.341256  1.341144          1.341038 | 0.545032  0.547112  0.549526  0.552286  0.555394  0.558841   0.56261  0.566677  0.571009  0.575571  0.580324  0.585228  0.590244  0.595334  0.600463  0.605601   0.61072  0.615795  0.620806  0.625737  0.630574  0.635305  0.639921  0.644418  0.648789  0.653031  0.657143  0.661125  0.664975  0.668695  0.672286   0.67575  0.679088  0.682303  0.685397  0.688373  0.691233   0.69398  0.696616  0.699144  0.701565  0.703883    0.7061  0.708217  0.710236   0.71216  0.713989  0.715726  0.717372      0.718927 | -0.213236  -0.215826   -0.21881  -0.222194  -0.225973  -0.230132  -0.234645  -0.239479  -0.244596   -0.24995  -0.255497  -0.261193  -0.266993  -0.272857  -0.278748  -0.284634  -0.290484  -0.296276  -0.301988  -0.307602  -0.313106  -0.318487  -0.323739  -0.328853  -0.333826  -0.338655  -0.343337  -0.347873  -0.352261  -0.356504  -0.360602  -0.364556   -0.36837  -0.372044  -0.375582  -0.378985  -0.382257  -0.385399  -0.388414  -0.391305  -0.394073   -0.39672   -0.39925  -0.401662   -0.40396  -0.406145  -0.408218  -0.410181  -0.412035      -0.41378 | 0.666885  0.667153  0.667462  0.667815   0.66821  0.668648  0.669127  0.669645  0.670198  0.670784  0.671398  0.672037  0.672697  0.673373  0.674063  0.674762  0.675467  0.676175  0.676885  0.677592  0.678296  0.678993  0.679684  0.680366  0.681038  0.681698  0.682347  0.682983  0.683606  0.684214  0.684808  0.685387  0.685951  0.686499  0.687032  0.687548  0.688047  0.688531  0.688997  0.689447  0.689879  0.690295  0.690693  0.691073  0.691437  0.691782   0.69211  0.692419   0.69271      0.692983 | 105.900084 | 106.653874  108.293973  110.238325  112.515493  115.151005   118.16666  121.580057  125.404373  129.648357  134.316507  139.409387  144.924024  150.854347  157.191638  163.924958  171.041533  178.527104  186.366211  194.542446  203.038646  211.837056  220.919456  230.267252  239.861558  249.683244   259.71298  269.931265  280.318441  290.854713  301.520152  312.294698  323.158165  334.090235  345.070459  356.078253  367.092894  378.093518  389.059114  399.968526  410.800445  421.533411  432.145811  442.615882  452.921708   463.04123  472.952243  482.632411  492.059267  501.210232          510.062621 | 514.416746 | 0.733058  0.736013  0.739471  0.743462  0.748003  0.753102  0.758755  0.764944  0.771644  0.778822  0.786437  0.794445  0.802799  0.811452  0.820357  0.829469  0.838745  0.848144   0.85763  0.867168   0.87673  0.886286  0.895812  0.905287  0.914691  0.924007   0.93322  0.942317  0.951284  0.960113  0.968794  0.977318  0.985678  0.993868  1.001882  1.009714   1.01736  1.024815  1.032074  1.039135  1.045993  1.052645  1.059086  1.065315  1.071327   1.07712  1.082689  1.088032  1.093144       1.098023 | 2.662853  2.680087    2.7003   2.72368   2.75036  2.780407  2.813823  2.850547   2.89046  2.933396  2.979149  3.027486  3.078155  3.130894  3.185443  3.241542  3.298943   3.35741  3.416722  3.476672  3.537071  3.597744  3.658531  3.719289  3.779885  3.840199  3.900124  3.959561   4.01842  4.076621  4.134088  4.190754  4.246555  4.301435  4.355338  4.408216   4.46002  4.510707  4.560233  4.608558  4.655644  4.701452  4.745946  4.789089  4.830845  4.871179  4.910055  4.947438  4.983291        5.017579 | 105.900084  107.394663   109.17857  111.281706  113.731343  116.551312  119.761411  123.377062  127.409206    131.8644  136.745077  142.049917  147.774276  153.910639  160.449065  167.377598  174.682635   182.34924  190.361421  198.702342  207.354504   216.29989  225.520069  234.996282  244.709508  254.640507  264.769857  275.077974  285.545128  296.151455  306.876959  317.701518  328.604879   339.56666  350.566344  361.583278  372.596671  383.585586  394.528939  405.405497  416.193878  426.872544  437.419807  447.813826  458.032614  468.054037  477.855822  487.415565   496.71074   505.71871  514.416746 |  100.0 |   5.900084 | 514.416746 |
    | 00:30 | 3.815554  3.839677  3.868202  3.901482  3.939796  3.983337  4.032204  4.086398  4.145829  4.210323  4.279637  4.353466  4.431466  4.513264   4.59847   4.68669  4.777533  4.870622  4.965592  5.062098  5.159819  5.258453   5.35772  5.457363  5.557146  5.656851  5.756279  5.855246  5.953586  6.051144  6.147779   6.24336  6.337767  6.430887  6.522616  6.612857  6.701517  6.788511  6.873757  6.957174   7.03869  7.118231  7.195726  7.271107  7.344307  7.415259  7.483898  7.550158  7.613973             7.675278 | 130.025594  131.310769   132.83798  134.630044   136.70691  139.084936  141.776388  144.789189  148.126911   151.78897  155.770999  160.065315  164.661453  169.546711  174.706665  180.125642  185.787129  191.674118  197.769379  204.055682  210.515955  217.133406  223.891601  230.774514  237.766559  244.852591  252.017911  259.248249  266.529742  273.848914  281.192649  288.548159  295.902961  303.244846   310.56185  317.842231  325.074441  332.247104  339.348991  346.369002  353.296141  360.119504  366.828257   373.41162  379.858857  386.159257  392.302126  398.276773  404.072502  409.678605 |  53.911173   54.157179   54.448078   54.787467   55.178195   55.622228    56.12057   56.673242   57.279323   57.937043   58.643902   59.396815   60.192265   61.026441   61.895374   62.795044   63.721472   64.670792   65.639298   66.623478   67.620037   68.625905   69.638235   70.654403   71.671993   72.688789   73.702756   74.712031   75.714905   76.709807   77.695294   78.670034   79.632796   80.582437   81.517894   82.438172   83.342336   84.229504   85.098837   85.949537   86.780836   87.591995   88.382295   89.151036   89.897532   90.621106   91.321085   91.996803   92.647592        93.272782 |  53.155544   53.396773   53.682023   54.014821   54.397961   54.833372   55.322037   55.863976   56.458287   57.103235   57.796367   58.534659   59.314661   60.132639   60.984697   61.866896   62.775334   63.706219   64.655917   65.620985   66.598191   67.584526   68.577197   69.573632   70.571461   71.568511   72.562788   73.552464   74.535863   75.511445   76.477794   77.433605    78.37767    79.30887   80.226162   81.128568   82.015174   82.885114   83.737565   84.571745   85.386901   86.182307    86.95726   87.711073   88.443073   89.152595   89.838981   90.501577   91.139728     91.752777 | 1.103472  1.107176  1.111544  1.116626  1.122459  1.129063  1.136445  1.144595   1.15349  1.163094  1.173359   1.18423  1.195647  1.207547  1.219865  1.232538  1.245504  1.258705  1.272086  1.285597  1.299191  1.312826  1.326463  1.340069  1.353612  1.367065  1.380404  1.393606  1.406654  1.419528  1.432215  1.444699   1.45697  1.469016  1.480828  1.492395  1.503711  1.514767  1.525558  1.536075  1.546314  1.556268  1.565932    1.5753  1.584368   1.59313  1.601581  1.609716  1.617529  1.625015 | 1.358183   1.35795  1.357679  1.357368  1.357017  1.356627    1.3562  1.355741  1.355252  1.354739  1.354207  1.353661  1.353107  1.352549  1.351992   1.35144  1.350896  1.350362  1.349842  1.349337  1.348848  1.348376  1.347922  1.347486  1.347068  1.346668  1.346286  1.345921  1.345573  1.345241  1.344925  1.344625  1.344339  1.344067  1.343808  1.343562  1.343329  1.343108  1.342897  1.342698  1.342508  1.342329  1.342159  1.341998  1.341846  1.341702  1.341566  1.341437  1.341316          1.341203 | 0.543727  0.545552   0.54769  0.550155  0.552957  0.556093  0.559554  0.563323  0.567373  0.571675  0.576193  0.580889  0.585727  0.590668  0.595679  0.600725  0.605778  0.610811  0.615802  0.620731  0.625583  0.630344  0.635004  0.639554  0.643987    0.6483   0.65249  0.656553   0.66049    0.6643  0.667984  0.671542  0.674977   0.67829  0.681483  0.684559  0.687518  0.690365  0.693101  0.695729   0.69825  0.700668  0.702984    0.7052  0.707319  0.709342   0.71127  0.713106  0.714851      0.716506 | -0.211587   -0.21387  -0.216523  -0.219558   -0.22298  -0.226782  -0.230945  -0.235445  -0.240249  -0.245318   -0.25061  -0.256083  -0.261694  -0.267403   -0.27317  -0.278962  -0.284747  -0.290498  -0.296192  -0.301809  -0.307333  -0.312751  -0.318051  -0.323227   -0.32827  -0.333178  -0.337946  -0.342574  -0.347059  -0.351403  -0.355605  -0.359666  -0.363588  -0.367374  -0.371024  -0.374541  -0.377926  -0.381184  -0.384314  -0.387321  -0.390206   -0.39297  -0.395617  -0.398147  -0.400564  -0.402867   -0.40506  -0.407143  -0.409117     -0.410983 | 0.666714  0.666949  0.667222  0.667537  0.667892  0.668289  0.668727  0.669205  0.669719  0.670267  0.670846  0.671452  0.672081   0.67273  0.673395  0.674072  0.674758  0.675451  0.676147  0.676843  0.677538   0.67823  0.678916  0.679595  0.680267  0.680928  0.681579  0.682219  0.682847  0.683462  0.684063   0.68465  0.685223  0.685782  0.686325  0.686853  0.687365  0.687861  0.688342  0.688806  0.689254  0.689685    0.6901  0.690497  0.690878  0.691242  0.691588  0.691917  0.692228      0.692522 |  104.99001 | 105.640887  107.061431  108.755669  110.752196  113.077356  115.754471  118.803258  122.239462  126.074701  130.316503  134.968493  140.030701   145.49993  151.370162  157.632964  164.277868  171.292729  178.664025  186.377122  194.416497  202.765909  211.408555  220.327171  229.504129  238.921501  248.561109  258.404561  268.433279  278.628517  288.971371  299.442785  310.023561   320.69435  331.435658  342.227843   353.05111  363.885508   374.71093  385.507104  396.253595  406.929802  417.514955  427.988114  438.328171  448.513853  458.523717  468.336166  477.929445  487.281654          496.370756 | 500.851295 | 0.731216  0.733796  0.736838  0.740377  0.744437  0.749032  0.754166  0.759832  0.766014  0.772684  0.779809  0.787351  0.795268  0.803514  0.812045  0.820816  0.829785  0.838912  0.848157  0.857486  0.866867  0.876271  0.885672  0.895046  0.904372  0.913631  0.922808  0.931887  0.940854    0.9497  0.958412  0.966983  0.975404  0.983667  0.991767  0.999697  1.007452  1.015027  1.022418  1.029619  1.036628  1.043441  1.050054  1.056462  1.062664  1.068656  1.074434  1.079994  1.085334       1.090449 | 2.652123  2.667152  2.684908  2.705603  2.729402  2.756414  2.786686  2.820208  2.856908  2.896666  2.939315  2.984656  3.032465    3.0825  3.134515  3.188261  3.243494  3.299978   3.35749   3.41582  3.474773  3.534167  3.593837  3.653629  3.713406   3.77304  3.832417  3.891431  3.949988  4.008001   4.06539  4.122082  4.178011  4.233114  4.287336  4.340623  4.392925  4.444195  4.494389  4.543466  4.591385  4.638106  4.683592  4.727807  4.770714  4.812277  4.852461   4.89123  4.928549        4.964383 |  104.99001  106.280272  107.829468   109.66713  111.820966     114.316  117.173889  120.412438  124.045348  128.082162  132.528389  137.385765  142.652598  148.324169  154.393146  160.849977  167.683265  174.880098  182.426333  190.306835   198.50568  207.006308  215.791659  224.844266  234.146334  243.679796   253.42636  263.367534  273.484652  283.758886  294.171253  304.702627  315.333732  326.045149  336.817309  347.630495  358.464833  369.300297  380.116696  390.893678  401.610727  412.247155  422.782109  433.194565  443.463329  453.567045  463.484191  473.193088  482.671908  491.898677  500.851295 |  100.0 |    4.99001 | 500.851295 |
    | 01:00 | 3.800735  3.821687  3.846634  3.875949   3.90995   3.94888  3.992899   4.04208  4.096402  4.155765  4.219989  4.288831  4.361998  4.439156   4.51995  4.604008  4.690958  4.780428  4.872059  4.965508  5.060446  5.156567  5.253584  5.351231  5.449261  5.547448  5.645583  5.743473  5.840942  5.937829  6.033982  6.129266  6.223551  6.316721  6.408666  6.499282  6.588475  6.676153  6.762232   6.84663  6.929269  7.010074  7.088975    7.1659  7.240782  7.313553  7.384148  7.452501  7.518546              7.58222 | 129.238958  130.351743  131.682449  133.254147  135.087795  137.201463  139.609718   142.32323  145.348601  148.688398  152.341378  156.302843  160.565085  165.117882  169.948986  175.044598  180.389788  185.968868    191.7657  197.763942  203.947247   210.29941  216.804476   223.44681  230.211149  237.082621  244.046762  251.089506  258.197178  265.356477  272.554451  279.778477  287.016229  294.255662  301.484975  308.692595  315.867147   322.99743    330.0724  337.081143  344.012859   350.85684  357.602455  364.239134  370.756352  377.143612  383.390438   389.48636  395.420904  401.183582 |  53.760044   53.973709   54.228119   54.527081   54.873823   55.270832   55.719744   56.221285   56.775272   57.380655   58.035614    58.73767   59.483823   60.270688   61.094626   61.951858    62.83857    63.75099   64.685451    65.63844   66.606624    67.58687   68.576253    69.57206   70.571777   71.573092   72.573873   73.572162   74.566159   75.554209   76.534788   77.506491    78.46802   79.418171   80.355824   81.279936   82.189526   83.083674   83.961507   84.822198   85.664953   86.489013   87.293643    88.07813   88.841778   89.583902   90.303831   91.000895   91.674431        92.323773 |   53.00735   53.216866   53.466335   53.759492     54.0995   54.488799   54.928994   55.420796   55.964024   56.557652   57.199891   57.888314   58.619977   59.391562   60.199499   61.040085   61.909577   62.804278   63.720593   64.655076   65.604458   66.565668   67.535839   68.512307   69.492611   70.474481   71.455827    72.43473   73.409424   74.378287   75.339824   76.292657   77.235514   78.167213   79.086658   79.992824   80.884751   81.761535   82.622321   83.466298   84.292688   85.100745    85.88975   86.659002    87.40782   88.135533   88.841481   89.525009   90.185465     90.822198 | 1.101193  1.104415  1.108242  1.112728  1.117917  1.123838   1.13051  1.137933  1.146096  1.154973  1.164529  1.174716  1.185482  1.196769  1.208516  1.220663  1.233149  1.245916  1.258908  1.272074  1.285366   1.29874  1.312155  1.325574  1.338965  1.352299  1.365548   1.37869  1.391703  1.404568  1.417268  1.429789  1.442117   1.45424  1.466147  1.477828  1.489275  1.500479  1.511432  1.522129  1.532561  1.542724  1.552611  1.562216  1.571535  1.580562  1.589292  1.597718  1.605838  1.613644 | 1.358328  1.358124  1.357883  1.357606  1.357289  1.356934  1.356542  1.356115  1.355657  1.355172  1.354663  1.354138  1.353599  1.353053  1.352504  1.351957  1.351414  1.350879  1.350354  1.349843  1.349346  1.348864  1.348399  1.347951  1.347521  1.347108  1.346713  1.346334  1.345973  1.345628  1.345299  1.344985  1.344686  1.344402  1.344131  1.343873  1.343628  1.343395  1.343174  1.342963  1.342764  1.342574  1.342394  1.342224  1.342062  1.341909  1.341764  1.341627  1.341498          1.341376 | 0.542597  0.544192  0.546076  0.548267  0.550778  0.553615  0.556775  0.560246   0.56401  0.568042  0.572311  0.576784  0.581424  0.586197  0.591067     0.596  0.600965  0.605936  0.610887  0.615797  0.620648  0.625424  0.630112  0.634703  0.639188  0.643562  0.647819  0.651956  0.655972  0.659866  0.663637  0.667285  0.670812  0.674218  0.677506  0.680677  0.683734  0.686678  0.689511  0.692236  0.694855   0.69737  0.699784  0.702097  0.704313  0.706433  0.708458  0.710392  0.712234      0.713986 | -0.210155  -0.212158  -0.214504  -0.217213  -0.220295  -0.223749  -0.227567  -0.231731  -0.236214  -0.240985  -0.246005  -0.251236  -0.256635  -0.262165  -0.267784  -0.273458  -0.279154  -0.284843  -0.290498  -0.296098  -0.301625  -0.307062  -0.312396  -0.317619  -0.322721  -0.327696   -0.33254   -0.33725  -0.341824  -0.346261   -0.35056  -0.354722  -0.358748  -0.362638  -0.366396  -0.370021  -0.373517  -0.376886  -0.380128  -0.383247  -0.386245  -0.389123  -0.391884  -0.394529   -0.39706   -0.39948  -0.401788  -0.403988  -0.406079     -0.408064 | 0.666565  0.666771  0.667012  0.667291  0.667609  0.667968  0.668366  0.668804  0.669279  0.669789  0.670332  0.670904  0.671501  0.672121   0.67276  0.673413  0.674079  0.674753  0.675434  0.676117  0.676802  0.677485  0.678165   0.67884  0.679508  0.680169  0.680821  0.681463  0.682094  0.682714  0.683321  0.683915  0.684496  0.685063  0.685616  0.686154  0.686678  0.687186  0.687679  0.688157  0.688618  0.689064  0.689494  0.689908  0.690305  0.690686   0.69105  0.691397  0.691727       0.69204 | 104.213141 | 104.773748  106.000939  107.473146  109.218505  111.263631  113.632815  116.347365  119.425123  122.880184   126.72279  130.959404    135.5929   140.62286  146.045917  151.856126  158.045333   164.60352  171.519124  178.779309  186.370205  194.277104  202.484623  210.976835  219.737368  228.749488   237.99616  247.460093  257.123774  266.969492  276.979354  287.135296  297.419091  307.812348  318.296518  328.852889  339.462585  350.106567  360.765621  371.420364  382.051236  392.638498   403.16223  413.602329  423.938508  434.150296  444.217042  454.117912  463.831897  473.337815          482.614323 | 487.196237 | 0.729628  0.731872  0.734539  0.737663  0.741275  0.745397  0.750038  0.755201  0.760875  0.767044   0.77368  0.780751   0.78822  0.796045  0.804186  0.812597  0.821239   0.83007  0.839052  0.848148  0.857327  0.866556  0.875808  0.885059  0.894286  0.903468  0.912588  0.921629  0.930578  0.939421  0.948147  0.956747  0.965211  0.973531  0.981699  0.989711  0.997558  1.005237  1.012742  1.020069  1.027213  1.034171  1.040938  1.047511  1.053887  1.060061  1.066031  1.071793  1.077343       1.082679 | 2.642885  2.655945  2.671483  2.689727  2.710865   2.73504  2.762341  2.792799  2.826391  2.863038  2.902618  2.944967   2.98989  3.037173  3.086586  3.137893  3.190859  3.245252  3.300849  3.357439  3.414823  3.472813  3.531238  3.589939  3.648771    3.7076  3.766305  3.824774  3.882907  3.940611  3.997801  4.054401  4.110338  4.165548   4.21997  4.273549  4.326231  4.377967  4.428712  4.478421  4.527053  4.574567  4.620926   4.66609  4.710024  4.752691  4.794055  4.834083  4.872737        4.909983 | 104.213141  105.324238  106.665989  108.267098   110.15518  112.355895  114.892203   117.78379   121.04668  124.693048  128.731211  133.165771   137.99787  143.225514  148.843948  154.846023  161.222564  167.962697  175.054154  182.483521   190.23646  198.297887  206.652113  215.282964  224.173867  233.307923  242.667958  252.236562  261.996115  271.928813  282.016675  292.241552  302.585135  313.028954  323.554379  334.142617  344.774709  355.431531  366.093785  376.741999  387.356525  397.917534  408.405015  418.798772  429.078427  439.223419  449.213002  459.026256  468.642083  478.039218  487.196237 |  100.0 |   4.213141 | 487.196237 |
    | 01:30 | 3.788018  3.806157  3.827897  3.853624  3.883676  3.918337   3.95782  4.002254  4.051689  4.106088   4.16534  4.229261  4.297613   4.37011  4.446431  4.526237  4.609174  4.694887  4.783025  4.873246  4.965224  5.058647  5.153224  5.248681  5.344763  5.441235  5.537879  5.634495  5.730896  5.826914   5.92239   6.01718  6.111149  6.204173  6.296137  6.386934  6.476462  6.564628  6.651343  6.736521  6.820083  6.901952  6.982054  7.060317  7.136673  7.211052   7.28339   7.35362  7.421677             7.487495 | 128.565676  129.526487  130.682455  132.056426  133.669834  135.541901  137.688971  140.123997  142.856232  145.891119  149.230363  152.872172  156.811592  161.040936  165.550232  170.327677   175.36006  180.633149  186.132023  191.841348  197.745607  203.829274  210.076945  216.473443   223.00388    229.6537  236.408707  243.255067  250.179316  257.168341   264.20937  271.289949  278.397922   285.52141   292.64878   299.76863  306.869757  313.941143  320.971924  327.951376  334.868892  341.713966  348.476169  355.145141  361.710569  368.162177   374.48971  380.682921  386.731563  392.625377 |  53.630356   53.815333   54.037047   54.299404   54.605879   54.959358   55.361998   55.815143   56.319279   56.874047   57.478296   58.130171   58.827227   59.566549   60.344881   61.158744    62.00454   62.878645   63.777477   64.697557   65.635548   66.588283   67.552779   68.526249   69.506098   70.489925   71.475507   72.460796   73.443905   74.423096   75.396765   76.363433   77.321733   78.270399   79.208253   80.134201   81.047216   81.946335   82.830651   83.699302   84.551471   85.386373   86.203255   87.001389   87.780066   88.538595   89.276297   89.992502   90.686547        91.357771 |  52.880181   53.061566   53.278974   53.536236    53.83676   54.183374   54.578196    55.02254   55.516886   56.060881   56.653396   57.292612   57.976132   58.701096   59.464314   60.262372   61.091744   61.948874   62.830251   63.732464    64.65224   65.586473    66.53224   67.486805   68.447627   69.412348    70.37879   71.344946   72.308964    73.26914     74.2239   75.171796   76.111487    77.04173   77.961372   78.869339   79.764623   80.646282   81.513425   82.365208   83.200828   84.019517   84.820536   85.603171   86.366726   87.110525   87.833901   88.536198   89.216766     89.874955 | 1.099235  1.102028  1.105368  1.109313  1.113909  1.119195  1.125195  1.131924  1.139379  1.147547  1.156401  1.165904  1.176011  1.186672  1.197829  1.209427  1.221407  1.233711  1.246285  1.259076  1.272035  1.285115  1.298276  1.311479  1.324688  1.337872  1.351003  1.364055  1.377006  1.389834  1.402522  1.415054  1.427413  1.439588  1.451567  1.463338  1.474891  1.486219  1.497313  1.508165  1.518769  1.529117  1.539204  1.549024  1.558571  1.567839  1.576824   1.58552  1.593921  1.602023 | 1.358453  1.358275  1.358063  1.357817  1.357533  1.357212  1.356854   1.35646  1.356033  1.355577  1.355095  1.354591  1.354072  1.353541  1.353003  1.352463  1.351924   1.35139  1.350863  1.350348  1.349844  1.349355   1.34888  1.348422   1.34798  1.347556  1.347148  1.346757  1.346382  1.346024  1.345682  1.345355  1.345044  1.344747  1.344464  1.344194  1.343937  1.343693   1.34346  1.343239  1.343029  1.342829  1.342639  1.342459  1.342288  1.342126  1.341972  1.341826  1.341689          1.341558 | 0.541623  0.543011  0.544663    0.5466   0.54884  0.551393  0.554261   0.55744  0.560917  0.564674  0.568685   0.57292  0.577347  0.581932  0.586641   0.59144  0.596298  0.601187  0.606079  0.610951  0.615783  0.620557   0.62526  0.629878  0.634403  0.638826  0.643141  0.647343  0.651431    0.6554  0.659252  0.662984  0.666598  0.670093  0.673472  0.676736  0.679885  0.682923  0.685851  0.688671  0.691385  0.693995  0.696503  0.698912  0.701223  0.703438  0.705558  0.707586  0.709523      0.711371 | -0.208917  -0.210666  -0.212732  -0.215137  -0.217897  -0.221018  -0.224499  -0.228329  -0.232489  -0.236953  -0.241688  -0.246659  -0.251828  -0.257156  -0.262605  -0.268139  -0.273723  -0.279327  -0.284922  -0.290486  -0.295996  -0.301435  -0.306788  -0.312043  -0.317189   -0.32222   -0.32713  -0.331912  -0.336566  -0.341087  -0.345476  -0.349732  -0.353855  -0.357846  -0.361705  -0.365435  -0.369037  -0.372512  -0.375862   -0.37909  -0.382197  -0.385186  -0.388057  -0.390814  -0.393457  -0.395988   -0.39841  -0.400723  -0.402928     -0.405028 | 0.666437  0.666616  0.666828  0.667074  0.667358   0.66768  0.668041  0.668441  0.668878   0.66935  0.669857  0.670394  0.670959  0.671548  0.672158  0.672787   0.67343  0.674084  0.674747  0.675415  0.676087   0.67676  0.677432  0.678101  0.678765  0.679423  0.680074  0.680716  0.681349  0.681972  0.682583  0.683183   0.68377  0.684345  0.684906  0.685453  0.685987  0.686506  0.687011    0.6875  0.687975  0.688435  0.688879  0.689307  0.689719  0.690116  0.690497  0.690861  0.691209       0.69154 | 103.551298 | 104.032989  105.090494  106.366341  107.887805  109.681299  111.771576  114.181029  116.929129  120.032032  123.502357  127.349135  131.577895  136.190867  141.187257  146.563571  152.313953  158.430524   164.90369  171.722431   178.87455  186.346882  194.125479  202.195749  210.542578  219.150424  228.003384  237.085257  246.379581  255.869665   265.53861  275.369328  285.344546  295.446817   305.65852   315.96186  326.338869  336.771404  347.241142  357.729579  368.218024  378.687601   389.11924  399.493678  409.791458  419.992925  430.078227  440.027317  449.819954  459.435706          468.853953 | 473.514016 | 0.728264  0.730209  0.732537  0.735284  0.738485  0.742165  0.746341  0.751022  0.756207  0.761884  0.768036  0.774635   0.78165  0.789045  0.796781  0.804817  0.813113  0.821628  0.830326  0.839168  0.848121  0.857153  0.866236  0.875342  0.884448  0.893533  0.902576   0.91156  0.920471  0.929293  0.938015  0.946626  0.955115  0.963475  0.971696  0.979772  0.987697  0.995464  1.003068  1.010504  1.017768  1.024855  1.031762  1.038483  1.045017  1.051358  1.057505  1.063452  1.069197       1.074736 | 2.634953  2.646265  2.659814  2.675835  2.694533  2.716076  2.740588  2.768138  2.798745  2.832374  2.868943  2.908327  2.950363  2.994866  3.041627  3.090427  3.141043   3.19325  3.246829  3.301569  3.357268  3.413736  3.470798  3.528288  3.586054  3.643958  3.701871  3.759677  3.817267  3.874545  3.931421  3.987813  4.043644  4.098847  4.153356  4.207113  4.260061  4.312151  4.363332   4.41356   4.46279  4.510982  4.558095  4.604092  4.648934  4.692585  4.735009  4.776171  4.816035        4.854566 | 103.551298  104.505809  105.664878  107.056027  108.706329  110.641577  112.885522  115.459234  118.380613  121.664087  125.320477  129.357026  133.777546  138.582675  143.770175  149.335282  155.271037  161.568622  168.217657   175.20647  182.522326  190.151625  198.080059  206.292748  214.774339  223.509093  232.480944   241.67355  251.070326  260.654474  270.408996  280.316708  290.360248  300.522083  310.784504   321.12963  331.539406  341.995599  352.479797  362.973402  373.457633  383.913517  394.321887  404.663384  414.918452  425.067337  435.090091  444.966567  454.676429  464.199148  473.514016 |  100.0 |   3.551298 | 473.514016 |
    | 02:00 | 3.777134  3.792789  3.811673  3.834168  3.860629  3.891365   3.92663  3.966605  4.011396  4.061031  4.115461  4.174563  4.238156  4.306001  4.377824  4.453315   4.53215  4.613991  4.698501  4.785346  4.874202  4.964757  5.056716    5.1498  5.243749  5.338319  5.433285  5.528437  5.623582  5.718542  5.813152  5.907259  6.000723  6.093414  6.185209  6.275998  6.365673  6.454137  6.541297  6.627065  6.711358  6.794096  6.875204  6.954608  7.032238  7.108025  7.181901  7.253801  7.323659              7.39141 | 127.990726  128.818092  129.819351   131.01674  132.431699  134.084104  135.991564  138.168844   140.62744  143.375346  146.416991  149.753342  153.382147  157.298263  161.494056  165.959813  170.684159  175.654442  180.857082  186.277873   191.90223  197.715399  203.702616  209.849224  216.140771  222.563062  229.102207  235.744638  242.477119  249.286748  256.160944  263.087437  270.054245  277.049661  284.062224  291.080705  298.094077  305.091499  312.062294  318.995924  325.881978  332.710145  339.470205  346.152004  352.745445  359.240469  365.627045  371.895153  378.034775  384.035887 |  53.519362   53.679013    53.87159   54.100995   54.370842   54.684296   55.043926   55.451591   55.908375   56.414556   56.969628   57.572359   58.220876   58.912771   59.645217   60.415083   61.219041   62.053662   62.915498   63.801145   64.707298   65.630783   66.568586   67.517865    68.47596    69.44039   70.408853   71.379217   72.349511   73.317915   74.282749   75.242459    76.19561   77.140871   78.077008    79.00287   79.917384   80.819544   81.708403   82.583067   83.442688   84.286457   85.113598   85.923366    86.71504   87.487916   88.241309   88.974545   89.686959        90.377891 |  52.771342   52.927893    53.11673    53.34168   53.606287   53.913654     54.2663   54.666049   55.113962   55.610313   56.154606   56.745633   57.381556   58.060014   58.778237   59.533153   60.321498   61.139912   61.985011    62.85346   63.742016   64.647567   65.567159   66.498004   67.437493   68.383194    69.33285   70.284371   71.235822   72.185421   73.131518   74.072591   75.007232   75.934137   76.852095   77.759977   78.656732   79.541373   80.412971    81.27065   82.113577    82.94096   83.752039   84.546083   85.322382    86.08025   86.819012   87.538009   88.236588     88.914103 | 1.097557   1.09997  1.102876  1.106331  1.110385  1.115083  1.120457  1.126528  1.133305  1.140785  1.148951  1.157775  1.167222  1.177248  1.187802  1.198833  1.210284    1.2221  1.234228  1.246615  1.259211  1.271969  1.284846  1.297801  1.310799  1.323804  1.336788  1.349723  1.362585   1.37535     1.388  1.400516  1.412883  1.425085  1.437111  1.448948  1.460586  1.472015  1.483226  1.494212  1.504964  1.515476   1.52574  1.535752  1.545505  1.554993   1.56421  1.573152  1.581813  1.590188 | 1.358561  1.358406  1.358221  1.358003   1.35775  1.357462  1.357136  1.356775   1.35638  1.355954    1.3555  1.355021  1.354523  1.354009  1.353485  1.352955  1.352423  1.351893  1.351368   1.35085  1.350342  1.349847  1.349365  1.348897  1.348445  1.348009   1.34759  1.347187    1.3468  1.346429  1.346075  1.345735  1.345411  1.345102  1.344806  1.344525  1.344256     1.344  1.343757  1.343525  1.343304  1.343094  1.342894  1.342704  1.342523  1.342352  1.342189  1.342034  1.341888           1.34175 | 0.540785  0.541989  0.543432  0.545137  0.547125  0.549409  0.551998  0.554893  0.558088  0.561568  0.565314  0.569302  0.573502  0.577883  0.582413   0.58706  0.591791  0.596577   0.60139  0.606206  0.611002  0.615759   0.62046  0.625092  0.629643  0.634103  0.638466  0.642724  0.646874  0.650913  0.654838  0.658648  0.662343  0.665922  0.669387  0.672739  0.675978  0.679106  0.682125  0.685037  0.687843  0.690546  0.693147  0.695649  0.698052   0.70036  0.702574  0.704695  0.706725      0.708665 | -0.207849   -0.20937  -0.211181  -0.213307  -0.215766  -0.218572  -0.221728  -0.225231  -0.229069  -0.233221  -0.237662  -0.242361  -0.247282   -0.25239  -0.257648  -0.263019  -0.268469  -0.273966  -0.279481  -0.284987  -0.290462  -0.295885   -0.30124  -0.306512   -0.31169  -0.316764  -0.321726  -0.326571  -0.331294  -0.335892  -0.340363  -0.344705  -0.348919  -0.353004   -0.35696  -0.360789  -0.364491  -0.368069  -0.371523  -0.374856  -0.378069  -0.381164  -0.384143  -0.387007  -0.389759    -0.3924  -0.394931  -0.397354  -0.399671     -0.401882 | 0.666326  0.666482  0.666667  0.666885  0.667137  0.667425   0.66775  0.668113  0.668513  0.668948  0.669418  0.669921  0.670452   0.67101  0.671591  0.672193  0.672811  0.673444  0.674087  0.674739  0.675396  0.676057  0.676719  0.677379  0.678037  0.678691  0.679339   0.67998  0.680613  0.681238  0.681852  0.682456  0.683048  0.683629  0.684197  0.684752  0.685294  0.685823  0.686338  0.686839  0.687325  0.687797  0.688254  0.688696  0.689123  0.689534   0.68993  0.690311  0.690675      0.691023 | 102.988541 | 103.401463  104.310548  105.413364  106.736011  108.304302  110.143003  112.275122  114.721301  117.499338  120.623867  124.106192   127.95427  132.172816  136.763501   141.72522  147.054386  152.745247  158.790188  165.180018  171.904227  178.951208  186.308457  193.962725  201.900157  210.106396  218.566669  227.265851  236.188518  245.318983  254.641325  264.139412  273.796909  283.597291  293.523849  303.559689  313.687733  323.890721  334.151207  344.451554  354.773934  365.100322  375.412497  385.692033  395.920303   406.07847  416.147493   426.10812  435.940893  445.626147          455.144013 |  459.86077 | 0.727094  0.728776    0.7308  0.733207  0.736031  0.739302  0.743043  0.747268  0.751983  0.757184  0.762859   0.76899   0.77555  0.782508  0.789829  0.797476   0.80541  0.813593  0.821986  0.830554  0.839261  0.848076  0.856967  0.865908  0.874873  0.883839  0.892786  0.901695  0.910548  0.919332  0.928032  0.936636  0.945134  0.953517  0.961774  0.969899  0.977885  0.985724  0.993412  1.000942  1.008311  1.015512  1.022543  1.029398  1.036075  1.042568  1.048876  1.054993  1.060917       1.066644 | 2.628163  2.637929  2.649704   2.66372  2.680195  2.699314  2.721227  2.746038  2.773802  2.804524  2.838163  2.874631  2.913801  2.955517  2.999596   3.04584  3.094039  3.143979  3.195449  3.248239  3.302148  3.356985   3.41257  3.468734   3.52532  3.582183   3.63919  3.696217  3.753152  3.809892  3.866341  3.922413  3.978028  4.033112  4.087599  4.141424   4.19453  4.246863  4.298371  4.349008  4.398726  4.447485  4.495241  4.541957  4.587593  4.632113   4.67548  4.717659  4.758613        4.798309 | 102.988541  103.806634   104.80539  106.010881  107.449272  109.146061  111.125319  113.409027  116.016517  118.964077  122.264701   125.92801  129.960299  134.364697  139.141409  144.288007  149.799735  155.669829  161.889812  168.449767  175.338579  182.544143  190.053542  197.853192  205.928961  214.266261  222.850129  231.665276  240.696139  249.926906  259.341546  268.923822  278.657302  288.525368  298.511216  308.597862  318.768136  329.004687  339.289972   349.60626  359.935626  370.259948  380.560901  390.819962  401.018397  411.137268  421.157429  431.059525  440.823992  450.431064   459.86077 |  100.0 |   2.988541 |  459.86077 |
    | 02:30 | 3.767843  3.781315  3.797665  3.817268  3.840481  3.867632  3.899002  3.934816  3.975228  4.020323   4.07011  4.124527  4.183448  4.246687  4.314015  4.385161   4.45983   4.53771  4.618481   4.70182   4.78741  4.874942   4.96412  5.054663  5.146305  5.238796  5.331904  5.425413   5.51912  5.612841  5.706403  5.799647  5.892425  5.984601  6.076047  6.166646  6.256287  6.344867   6.43229  6.518463  6.603302  6.686724  6.768651  6.849007   6.92772   7.00472   7.07994  7.153311   7.22477              7.29425 | 127.500825  128.211452  129.076298  130.116687  131.353681  132.807361  134.496132  136.436105  138.640602  141.119809    143.8806  146.926514  150.257887  153.872086  157.763833  161.925571  166.347852  171.019714  175.929033  181.062841  186.407597  191.949415  197.674252  203.568051  209.616852  215.806876  222.124579  228.556692  235.090238  241.712543  248.411234  255.174236   261.98975  268.846247  275.732443  282.637282  289.549915  296.459681  303.356085  310.228778  317.067539  323.862258  330.602914  337.279564  343.882322  350.401349  356.826836   363.14899  369.358024  375.444148 |  53.424605      53.562   53.728741   53.928647   54.165374   54.442261   54.762178   55.127403   55.539529   55.999407   56.507136   57.062086   57.662963   58.307885    58.99449   59.720038   60.481518   61.275744   62.099447   62.949341    63.82219   64.714847   65.624292   66.547653   67.482219   68.425449   69.374968   70.328568   71.284203   72.239973   73.194122   74.145027   75.091182   76.031193   76.963765   77.887694   78.801856     79.7052    80.59674   81.475544   82.340734   83.191473   84.026963   84.846438   85.649159   86.434412   87.201503   87.949749   88.678483        89.387044 |  52.678425   52.813152   52.976655   53.172679   53.404809   53.676319   53.990023   54.348156   54.752279   55.203226   55.701095   56.245269   56.834477   57.466875   58.140147   58.851605   59.598297   60.377101   61.184807   62.018198   62.874096   63.749418   64.641202   65.546632    66.46305   67.387963   68.319043   69.254125   70.191201   71.128411   72.064032   72.996471   73.924252   74.846008    75.76047   76.666457   77.562867   78.448669   79.322895   80.184634   81.033023   81.867241   82.686506   83.490067     84.2772   85.047204   85.799398   86.533114   87.247696     87.942498 | 1.096124  1.098202  1.100721  1.103736  1.107299  1.111457  1.116248  1.121702  1.127835  1.134653  1.142149  1.150307  1.159098  1.168486  1.178429  1.188877  1.199781  1.211089  1.222746  1.234703  1.246908  1.259316   1.27188  1.284559  1.297316  1.310115  1.322924  1.335715   1.34846  1.361136  1.373722  1.386198  1.398547  1.410754  1.422804  1.434684  1.446383  1.457891  1.469197  1.480294  1.491173  1.501826  1.512247  1.522429  1.532366  1.542052   1.55148  1.560647  1.569545   1.57817 | 1.358653  1.358519  1.358358  1.358166  1.357942  1.357684  1.357391  1.357062  1.356699  1.356303  1.355877  1.355425   1.35495  1.354457   1.35395  1.353433  1.352911  1.352387  1.351865  1.351347  1.350838  1.350338   1.34985  1.349375  1.348914  1.348469  1.348038  1.347624  1.347225  1.346843  1.346476  1.346124  1.345788  1.345466  1.345159  1.344865  1.344585  1.344318  1.344063   1.34382  1.343588  1.343368  1.343157  1.342958  1.342767  1.342587  1.342415  1.342251  1.342096          1.341949 | 0.540067  0.541107  0.542362  0.543857  0.545613  0.547647  0.549973  0.552595  0.555512  0.558718  0.562198  0.565931  0.569894  0.574058  0.578394   0.58287  0.587455   0.59212  0.596836  0.601576  0.606318  0.611041  0.615726  0.620356   0.62492  0.629406  0.633804  0.638108  0.642312  0.646411  0.650402  0.654283  0.658053  0.661711  0.665258  0.668693  0.672017  0.675232  0.678339   0.68134  0.684236  0.687029   0.68972  0.692312  0.694807  0.697205  0.699509  0.701721  0.703842      0.705874 | -0.206931  -0.208249   -0.20983    -0.2117  -0.213881   -0.21639  -0.219237  -0.222423  -0.225944  -0.229785  -0.233927  -0.238343  -0.243004  -0.247875  -0.252923  -0.258112  -0.263407  -0.268776  -0.274189  -0.279618  -0.285038  -0.290427  -0.295767   -0.30104  -0.306234  -0.311338  -0.316341  -0.321237   -0.32602  -0.330685  -0.335229  -0.339651  -0.343948   -0.34812  -0.352167  -0.356089  -0.359888  -0.363563  -0.367117  -0.370551  -0.373866  -0.377064  -0.380147  -0.383116  -0.385973   -0.38872  -0.391357  -0.393888  -0.396312     -0.398632 | 0.666232  0.666367  0.666528  0.666719  0.666941  0.667198   0.66749  0.667818  0.668182  0.668582  0.669016  0.669484  0.669981  0.670507  0.671058  0.671632  0.672225  0.672834  0.673456  0.674089   0.67473  0.675377  0.676026  0.676677  0.677327  0.677975  0.678618  0.679257  0.679888  0.680513  0.681128  0.681735  0.682331  0.682916   0.68349  0.684052  0.684602  0.685139  0.685663  0.686173   0.68667  0.687153  0.687622  0.688077  0.688517  0.688942  0.689352  0.689748  0.690127      0.690492 |  102.51094 | 102.864117  103.643795  104.594665  105.741418  107.108944  108.721635   110.60269  112.773489   115.25306  118.057676  121.200599  124.691957  128.538756  132.744996   137.31187  142.238009  147.519766    153.1515  159.125856  165.434021  172.065963  179.010628  186.256121  193.789848  201.598638  209.668838  217.986393  226.536902  235.305668  244.277731  253.437894  262.770741  272.260653  281.891807  291.648193  301.513605  311.471648  321.505734   331.59908  341.734706   351.89543  362.063866   372.22242  382.353289  392.438454  402.459683  412.398525  422.236311  431.954153           441.53295 | 446.286663 | 0.726095  0.727544  0.729299  0.731399  0.733882  0.736777  0.740114   0.74391  0.748177   0.75292  0.758132  0.763802  0.769909  0.776427  0.783327  0.790574  0.798133  0.805968   0.81404  0.822315  0.830757  0.839333  0.848014  0.856769  0.865573  0.874402  0.883233  0.892047  0.900825  0.909551  0.918212  0.926793  0.935283  0.943672   0.95195  0.960108  0.968139  0.976035  0.983792  0.991401  0.998859  1.006161  1.013301  1.020275   1.02708  1.033711  1.040165  1.046437  1.052525       1.058425 | 2.622363  2.630772  2.640971  2.653191  2.667652  2.684553  2.704062   2.72631  2.751386   2.77933  2.810139  2.843761  2.880107  2.919051  2.960439  3.004094  3.049826  3.097435  3.146716  3.197468  3.249492  3.302596    3.3566   3.41133  3.466627  3.522339  3.578329  3.634468  3.690637  3.746729  3.802643  3.858288  3.913579  3.968439  4.022797  4.076585  4.129743  4.182214  4.233944  4.284883  4.334983  4.384201  4.432494  4.479821  4.526142   4.57142  4.615619  4.658701  4.700631        4.741375 |  102.51094  103.210545  104.069086  105.110997  106.361256  107.844699  109.585308  111.605535  113.925717    116.5636  119.534005  122.848643  126.516066  130.541734  134.928183  139.675251  144.780354   150.23877  156.043926  162.187673  168.660526  175.451891  182.550249  189.943321  197.618192  205.561428  213.759155  222.197129  230.860792  239.735306  248.805586  258.056322  267.471992  277.036873  286.735048  296.550408  306.466653  316.467292   326.53564  336.654815  346.807738  356.977124  367.145487  377.295126  387.408133  397.466381  407.451529  417.345019  427.128071   436.78169  446.286663 |  100.0 |    2.51094 | 446.286663 |
    | 03:00 | 3.759929  3.771492  3.785606  3.802633  3.822925   3.84682  3.874618  3.906574  3.942884  3.983682  4.029032  4.078926  4.133294  4.192004  4.254872  4.321671   4.39214  4.465995  4.542939  4.622663  4.704862  4.789233  4.875483  4.963328    5.0525  5.142746  5.233827   5.32552  5.417618  5.509926  5.602267  5.694473  5.786391  5.877878  5.968801  6.059035  6.148467  6.236989  6.324499  6.410902   6.49611  6.580037  6.662602  6.743729  6.823342  6.901371  6.977747    7.0524  7.125267             7.196281 | 127.084277  127.693138  128.438169  129.339559  130.417679  131.692434    133.1826  134.905197  136.874952  139.103868  141.600953  144.372082  147.420022  150.744561  154.342757  158.209237  162.336547  166.715501  171.335536  176.185028  181.251583  186.522282   191.98389  197.623025   203.42629  209.380376  215.472132  221.688624  228.017164  234.445331  240.960978  247.552236  254.207503  260.915431  267.664919  274.445088  281.245267  288.054974  294.863897  301.661872   308.43887  315.184975  321.890368   328.54531  335.140127  341.665196  348.110929  354.467759   360.72613  366.876485 |  53.343905   53.461822   53.605761   53.779395   53.986343   54.230023   54.513507   54.839391   55.209689   55.625749   56.088222   56.597048   57.151496   57.750223    58.39135   59.072564   59.791211   60.544395   61.329065   62.142101    62.98037   63.840789   64.720361    65.61621   66.525594   67.445925   68.374772   69.309862   70.249075   71.190443   72.132137   73.072462   74.009846   74.942831   75.870062    76.79028    77.70231   78.605055   79.497484   80.378631   81.247583   82.103472   82.945477   83.772808    84.58471   85.380454   86.159331   86.920654   87.663748        88.387951 |  52.599292    52.71492   52.856063   53.026325   53.229255   53.468203   53.746181   54.065737   54.428844   54.836825   55.290316   55.789261   56.332943   56.920042    57.54872   58.216705   58.921396   59.659953   60.429386   61.226633   62.048623   62.892334   63.754826   64.633277   65.525001   66.427461    67.33827   68.255201   69.176176   70.099263   71.022669   71.944734   72.863915   73.778782   74.688007   75.590355   76.484674   77.369887   78.244987   79.109023   79.961099   80.800368   81.626021   82.437286   83.233422   84.013713   84.777465   85.524003   86.252667     86.962806 | 1.094902  1.096687  1.098864  1.101485  1.104605  1.108271  1.112525  1.117402  1.122928  1.129115  1.135966  1.143474  1.151618  1.160371  1.169698  1.179556  1.189899  1.200679  1.211845  1.223348  1.235138  1.247168  1.259392  1.271768  1.284257  1.296822  1.309429  1.322049  1.334651  1.347213   1.35971  1.372121  1.384429  1.396615  1.408666  1.420567  1.432305  1.443869  1.455249  1.466436   1.47742  1.488194   1.49875  1.509081  1.519181  1.529044  1.538662  1.548032  1.557147  1.566001 | 1.358733  1.358617  1.358477  1.358309  1.358112  1.357882  1.357618  1.357321  1.356989  1.356624  1.356228  1.355803  1.355354  1.354883  1.354395  1.353893  1.353383  1.352868  1.352352  1.351838  1.351329  1.350827  1.350335  1.349854  1.349386  1.348932  1.348492  1.348067  1.347658  1.347264  1.346885  1.346522  1.346173   1.34584   1.34552  1.345215  1.344923  1.344644  1.344378  1.344124  1.343882  1.343651  1.343431  1.343221  1.343021   1.34283  1.342649  1.342477  1.342313          1.342158 | 0.539453  0.540349  0.541437  0.542742  0.544286   0.54609  0.548168   0.55053  0.553181  0.556117  0.559331  0.562807  0.566526  0.570463  0.574591  0.578881  0.583303  0.587829  0.592429  0.597077  0.601747  0.606418   0.61107  0.615684  0.620247  0.624745  0.629168  0.633506  0.637754  0.641904  0.645953  0.649898  0.653737  0.657468   0.66109  0.664604   0.66801  0.671308  0.674499  0.677585  0.680568  0.683447  0.686227  0.688907  0.691489  0.693976   0.69637   0.69867   0.70088      0.703001 | -0.206145  -0.207283  -0.208658  -0.210296  -0.212222  -0.214455  -0.217009  -0.219892  -0.223104  -0.226637  -0.230479  -0.234607  -0.238997  -0.243619  -0.248441  -0.253429   -0.25855  -0.263771  -0.269062  -0.274393  -0.279739  -0.285076  -0.290383  -0.295641  -0.300837  -0.305956  -0.310987  -0.315922  -0.320754  -0.325476  -0.330085  -0.334577   -0.33895  -0.343203  -0.347335  -0.351345  -0.355234  -0.359002  -0.362651  -0.366182  -0.369595  -0.372892  -0.376075  -0.379146  -0.382105  -0.384954  -0.387695   -0.39033  -0.392859     -0.395284 | 0.666151  0.666267  0.666407  0.666574   0.66677  0.666998  0.667258  0.667554  0.667884  0.668249  0.668649  0.669082  0.669546  0.670039  0.670559  0.671104   0.67167  0.672254  0.672854  0.673466  0.674089   0.67472  0.675356  0.675995  0.676635  0.677275  0.677913  0.678547  0.679176  0.679798  0.680414  0.681021   0.68162  0.682208  0.682787  0.683354   0.68391  0.684454  0.684986  0.685506  0.686012  0.686505  0.686985  0.687451  0.687903  0.688341  0.688765  0.689174  0.689568      0.689947 | 102.106358 | 102.407781  103.074971  103.892852  104.884543  106.073766  107.484211  109.138899  111.059547  113.266015  115.775846  118.603925  121.762274  125.259968  129.103171  133.295256     137.837   142.72682  147.961031   153.53411  159.438951  165.667103  172.208979  179.054043  186.190968  193.607768   201.29191  209.230398  217.409845  225.816529  234.436434   243.25528   252.25855  261.431506  270.759196  280.226468  289.817969  299.518148  309.311257  319.181348  329.112271  339.087672  349.090988  359.105444  369.114051  379.099602  389.044667  398.931595  408.742506  418.459298          428.063638 | 432.836227 | 0.725243  0.726488  0.728005  0.729832  0.732005  0.734558  0.737521  0.740917  0.744763  0.749068  0.753833  0.759053  0.764713  0.770793  0.777268  0.784109  0.791283  0.798756  0.806492  0.814457  0.822616  0.830936  0.839386  0.847937  0.856561  0.865233  0.873929  0.882629  0.891314  0.899966  0.908569   0.91711  0.925576  0.933955  0.942237  0.950413  0.958474  0.966413  0.974223  0.981898  0.989431  0.996817  1.004053  1.011132  1.018051  1.024805  1.031391  1.037805  1.044042       1.050101 | 2.617423  2.624641  2.633449  2.644068  2.656717    2.6716  2.688899  2.708767  2.731319  2.756628  2.784723   2.81559  2.849173   2.88538  2.924086   2.96514  3.008372  3.053599  3.100628  3.149266  3.199319  3.250599  3.302924   3.35612  3.410024  3.464481  3.519348  3.574493  3.629791   3.68513  3.740404  3.795518  3.850382  3.904915  3.959041   4.01269  4.065798  4.118305  4.170154  4.221293  4.271674  4.321249  4.369974  4.417808   4.46471   4.51064  4.555563   4.59944  4.642237        4.683918 | 102.106358  102.703346  103.439639  104.337919  105.421771  106.715078  108.241374  110.023193  112.081471   114.43502  117.100133  120.090304  123.416092  127.085094  131.102037  135.468938  140.185323  145.248489  150.653759  156.394752  162.463629  168.851315  175.547702  182.541819  189.821975  197.375885  205.190762  213.253396  221.550221  230.067355  238.790645  247.705685  256.797844  266.052274  275.453922  284.987534  294.637657  304.388642  314.224641  324.129606  334.087285  344.081221  354.094746  364.110979  374.112822  384.082957  394.003845   403.85772  413.626592  423.292243  432.836227 |  100.0 |   2.106358 | 432.836227 |
    | 03:30 | 3.753205  3.763103  3.775252  3.789996  3.807677  3.828631  3.853172  3.881574  3.914068  3.950828  3.991964  4.037523  4.087485  4.141771  4.200244  4.262723  4.328986  4.398779  4.471831  4.547855  4.626558  4.707647  4.790833  4.875838  4.962391  5.050236  5.139131  5.228847   5.31917    5.4099  5.500853  5.591856  5.682747  5.773377  5.863608  5.953312  6.042367  6.130661  6.218089  6.304553   6.38996   6.47422  6.557252  6.638975  6.719313  6.798193  6.875545    6.9513   7.02539             7.097752 | 126.730822  127.251264  127.891444  128.670265  129.607167  130.721562  132.032226  133.556689  135.310674  137.307626   139.55835  142.070789  144.849936  147.897874   151.21393  154.794906  158.635373  162.727991   167.06384  171.632739  176.423538  181.424379  186.622928  192.006558  197.562503  203.277983  209.140295  215.136878  221.255367  227.483621  233.809741   240.22208  246.709239  253.260069  259.863649  266.509283  273.186481  279.884938  286.594524  293.305261  300.007306  306.690936  313.346531  319.964553  326.535537  333.050073  339.498789  345.872345  352.161411  358.356665 |  53.275333   53.376271   53.500172   53.650523   53.830837   54.044532   54.294793   54.584442   54.915821   55.290697   55.710206   56.174814    56.68433   57.237938   57.834256    58.47142   59.147165   59.858924   60.603909   61.379201   62.181815   63.008763   63.857107   64.723987    65.60666   66.502508   67.409057   68.323981   69.245099   70.170376   71.097918   72.025962   72.952871   73.877124   74.797308   75.712106   76.620292    77.52072   78.412319   79.294081   80.165057    81.02435    81.87111   82.704524   83.523817   84.328242   85.117079   85.889627   86.645206        87.383148 |  52.532052    52.63103   52.752525   52.899956   53.076768   53.286314   53.531715   53.815739   54.140682   54.508279   54.919641   55.375227   55.874849   56.417705   57.002444   57.627234   58.289857   58.987794   59.718312   60.478548   61.265576   62.076466   62.908335   63.758381   64.623912   65.502364   66.391309   67.288465   68.191696   69.099004   70.008534   70.918556   71.827465   72.733771   73.636085   74.533118   75.423667    76.30661   77.180895   78.045533   78.899595   79.742202   80.572518   81.389748   82.193131   82.981934   83.755452   84.512998   85.253904     85.977516 | 1.093863  1.095392  1.097267   1.09954  1.102261  1.105481  1.109244  1.113588  1.118544  1.124134  1.130368  1.137246  1.144758  1.152884  1.161596  1.170859  1.180632  1.190871  1.201528  1.212557  1.223908  1.235536  1.247395  1.259442  1.271637  1.283941   1.29632  1.308742  1.321177  1.333598  1.345981  1.358304  1.370546   1.38269  1.394719  1.406617  1.418373  1.429972  1.441405   1.45266  1.463729  1.474603  1.485273  1.495734  1.505976  1.515995  1.525783  1.535336  1.544646  1.553708 |   1.3588  1.358701  1.358579  1.358433   1.35826  1.358056  1.357821  1.357553  1.357251  1.356917  1.356551  1.356155  1.355731  1.355285  1.354818  1.354335   1.35384  1.353336  1.352829   1.35232  1.351813  1.351312  1.350818  1.350333  1.349859  1.349398   1.34895  1.348516  1.348096  1.347692  1.347302  1.346927  1.346567  1.346222  1.345891  1.345574  1.345271  1.344981  1.344703  1.344438  1.344185  1.343943  1.343713  1.343493  1.343283  1.343083  1.342893  1.342711  1.342539          1.342375 | 0.538931    0.5397   0.54064  0.541774  0.543127  0.544719  0.546566  0.548684   0.55108  0.553756  0.556708  0.559927  0.563397  0.567099  0.571009    0.5751  0.579345  0.583715  0.588182  0.592719  0.597301  0.601903  0.606506  0.611089  0.615636  0.620133  0.624567  0.628929   0.63321  0.637402  0.641501  0.645502  0.649402  0.653199  0.656892  0.660479  0.663961  0.667338   0.67061  0.673778  0.676843  0.679807  0.682671  0.685437  0.688105  0.690679  0.693158  0.695545  0.697842       0.70005 | -0.205473  -0.206453  -0.207645  -0.209074  -0.210767  -0.212744  -0.215025   -0.21762  -0.220535  -0.223769  -0.227312  -0.231151  -0.235263  -0.239625  -0.244207  -0.248979  -0.253908  -0.258963  -0.264113  -0.269327   -0.27458  -0.279846  -0.285102  -0.290329   -0.29551   -0.30063  -0.305676  -0.310638  -0.315507  -0.320276   -0.32494  -0.329494  -0.333935  -0.338262  -0.342471  -0.346563  -0.350537  -0.354393  -0.358132  -0.361755  -0.365262  -0.368654  -0.371934  -0.375102   -0.37816  -0.381108   -0.38395  -0.386685  -0.389316     -0.391843 | 0.666082  0.666182  0.666303  0.666448  0.666621  0.666822  0.667054  0.667318  0.667616  0.667948  0.668314  0.668713  0.669144  0.669605  0.670094  0.670609  0.671147  0.671705  0.672281  0.672872  0.673475  0.674088  0.674709  0.675335  0.675964  0.676594  0.677224  0.677852  0.678476  0.679096   0.67971  0.680317  0.680916  0.681507  0.682088   0.68266  0.683221  0.683771   0.68431  0.684837  0.685352  0.685854  0.686343   0.68682  0.687283  0.687732  0.688168   0.68859  0.688998      0.689391 | 101.764251 | 102.020965  102.590658   103.29251  104.147973  105.179415  106.409571   107.86096  109.555283  111.512861   113.75214  116.289296  119.137947  122.308996  125.810576  129.648103  133.824399  138.339882  143.192792  148.379425  153.894382  159.730799  165.880566  172.334514  179.082591  186.114001  193.417326  200.980626  208.791519  216.837244  225.104713  233.580546  242.251106  251.102514  260.120667   269.29125  278.599738  288.031404  297.571318  307.204346  316.915148  326.688181  336.507686  346.357694  356.222015  366.084242  375.927739  385.735646  395.490871  405.176089          414.773742 |  419.54871 | 0.724519  0.725585  0.726892  0.728476  0.730372  0.732615  0.735236  0.738261  0.741712  0.745603   0.74994  0.754723  0.759946  0.765593  0.771644  0.778074  0.784856  0.791957  0.799344  0.806984  0.814844  0.822891  0.831093  0.839421  0.847846  0.856343  0.864886  0.873455  0.882028  0.890588  0.899118  0.907602  0.916027   0.92438  0.932651  0.940829  0.948906  0.956872  0.964722  0.972446  0.980041  0.987499  0.994815  1.001985  1.009004  1.015868  1.022572  1.029113  1.035487       1.041689 | 2.613224  2.619404  2.626988  2.636187  2.647213  2.660272  2.675554  2.693226  2.713424  2.736249  2.761761  2.789979   2.82088  2.854404  2.890457  2.928916  2.969632  3.012442  3.057169   3.10363  3.151639  3.201013  3.251571   3.30314  3.355554  3.408657  3.462301  3.516351  3.570677  3.625161  3.679695  3.734176  3.788513  3.842619  3.896414  3.949825  4.002784  4.055229  4.107099  4.158341  4.208903  4.258736  4.307795  4.356035  4.403416  4.449898  4.495442  4.540011  4.583569         4.62608 | 101.764251   102.27261  102.902644  103.675228   104.61241  105.736898  107.071486  108.638447  110.458941  112.552477  114.936462  117.625858  120.632956  123.967277   127.63557  131.641912  135.987867    140.6727  145.693613  151.045988  156.723634  162.719007  169.023417  175.627212  182.519931  189.690436  197.127025  204.817515  212.749321  220.909504  229.284821  237.861758   246.62655  255.565203  264.663506  273.907036  283.281168  292.771071  302.361714  312.037862  321.784072  331.584695  341.423866  351.285509  361.153325  371.010793  380.841166  390.627469  400.352494  409.998799   419.54871 |  100.0 |   1.764251 |  419.54871 |
    | 04:00 | 3.747504  3.755956  3.766386  3.779114  3.794471  3.812785  3.834372  3.859521  3.888487  3.921476   3.95864   4.00007  4.045797   4.09579  4.149964  4.208179  4.270256  4.335976  4.405097  4.477355  4.552477  4.630181  4.710189  4.792225  4.876022  4.961322  5.047881  5.135466  5.223859  5.312855  5.402261    5.4919  5.581603  5.671217  5.760596  5.849607  5.938124  6.026029  6.113214  6.199575  6.285016  6.369446  6.452778  6.534931  6.615826  6.695388  6.773544  6.850226  6.925365             6.998895 | 126.431497  126.875356    127.4241  128.095249  128.907141  129.878445  131.027622  132.372353  133.928982  135.712021  137.733732  140.003829  142.529297  145.314342  148.360453   151.66656   155.22926  159.043095  163.100856  167.393884  171.912364  176.645593  181.582223  186.710463  192.018254   197.49341  203.123725  208.897063  214.801417  220.824955   226.95605  233.183295  239.495513  245.881753  252.331291  258.833615  265.378414  271.955565  278.555119  285.167278  291.782387  298.390913  304.983427  311.550592  318.083146  324.571889  331.007664  337.381349  343.683842  349.906047 |  53.217193   53.303384    53.40975   53.539557   53.696168   53.882935   54.103075   54.359551   54.654946   54.991367   55.370361   55.792866   56.259191   56.769029   57.321493   57.915178   58.548235   59.218455   59.923352   60.660244   61.426334   62.218766   63.034689   63.871295   64.725858   65.595755   66.478487   67.371685   68.273121   69.180701   70.092471   71.006606   71.921406    72.83529   73.746786    74.65452   75.557216   76.453679   77.342792   78.223508   79.094841    79.95586   80.805687   81.643483    82.46845   83.279824   84.076868   84.858869   85.625138           86.375 |  52.475041   52.559558   52.663859   52.791145   52.944715   53.127854    53.34372   53.595215   53.884874   54.214762   54.586396   55.000696   55.457965   55.957902   56.499638   57.081794   57.702558   58.359763   59.050971   59.773553   60.524766    61.30181   62.101888   62.922248   63.760216    64.61322    65.47881   66.354663   67.238593   68.128548   69.022612   69.918995   70.816031   71.712168   72.605963    73.49607   74.381236    75.26029   76.132138    76.99575   77.850162   78.694461   79.527785   80.349311   81.158259   81.953876   82.735441   83.502257   84.253646     84.988946 | 1.092981  1.094288  1.095899  1.097863  1.100229  1.103047  1.106362  1.110216  1.114644  1.119672   1.12532  1.131594  1.138492  1.146004  1.154108  1.162775  1.171973   1.18166  1.191795  1.202332  1.213225  1.224429    1.2359  1.247593  1.259468  1.271486  1.283612  1.295811  1.308053   1.32031  1.332555  1.344765  1.356919  1.368997  1.380981  1.392856  1.404607   1.41622  1.427684  1.438988  1.450121  1.461075   1.47184  1.482409  1.492775   1.50293  1.512868  1.522583  1.532069   1.54132 | 1.358858  1.358773  1.358668  1.358541  1.358389   1.35821  1.358001  1.357761  1.357488  1.357184  1.356847  1.356479  1.356084  1.355662  1.355218  1.354755  1.354278  1.353788  1.353292  1.352791   1.35229  1.351791  1.351297   1.35081  1.350332  1.349865   1.34941  1.348968  1.348539  1.348125  1.347725   1.34734  1.346969  1.346612   1.34627  1.345942  1.345627  1.345326  1.345037  1.344761  1.344497  1.344245  1.344004  1.343774  1.343555  1.343345  1.343145  1.342955  1.342773            1.3426 | 0.538486  0.539145  0.539954  0.540938  0.542118  0.543516  0.545152  0.547042  0.549196  0.551622   0.55432  0.557285  0.560506  0.563968  0.567652  0.571533  0.575587  0.579787  0.584105  0.588515  0.592991  0.597509  0.602046  0.606582  0.611098   0.61558  0.620014  0.624387   0.62869  0.632914  0.637053  0.641102  0.645056  0.648912  0.652669  0.656324  0.659877  0.663328  0.666676  0.669922  0.673068  0.676112  0.679058  0.681907  0.684659  0.687316  0.689879  0.692351  0.694733      0.697025 | -0.204901  -0.205742  -0.206771  -0.208014  -0.209496   -0.21124  -0.213267  -0.215591  -0.218224  -0.221167  -0.224418  -0.227968  -0.231801  -0.235896  -0.240228   -0.24477  -0.249492  -0.254363  -0.259354  -0.264433  -0.269574   -0.27475  -0.279939  -0.285117  -0.290268  -0.295373   -0.30042  -0.305396   -0.31029  -0.315095  -0.319804  -0.324411  -0.328912  -0.333303  -0.337584  -0.341751  -0.345804  -0.349742  -0.353566  -0.357276  -0.360873  -0.364357  -0.367729  -0.370991  -0.374144  -0.377188  -0.380127   -0.38296  -0.385689     -0.388316 | 0.666023  0.666109  0.666213   0.66634   0.66649  0.666667  0.666873  0.667109  0.667377  0.667678  0.668011  0.668377  0.668776  0.669204  0.669662  0.670147  0.670656  0.671187  0.671738  0.672306  0.672888  0.673482  0.674086  0.674697  0.675313  0.675932  0.676553  0.677173  0.677792  0.678407  0.679018  0.679623  0.680222  0.680813  0.681397  0.681971  0.682536  0.683091  0.683635  0.684169  0.684691  0.685201  0.685699  0.686185  0.686658  0.687118  0.687565  0.687998  0.688418      0.688825 |  101.47548 | 101.693674  102.179097  102.780025    103.5162  104.408511  105.478542  106.748051  108.238409  109.970062  111.962013  114.231396  116.793126   119.65967   122.84092  126.344162  130.174145  134.333209  138.821464     143.637  148.776116  154.233537  160.002633  166.075616  172.443713  179.097322  186.026145  193.219292  200.665381  208.352601  216.268777  224.401416  232.737738  241.264705   249.96904  258.837242  267.855593  277.010165  286.286822  295.671223  305.148819  314.704851   324.32435   333.99213  343.692788  353.410698  363.130006  372.834631  382.508258  392.134337          401.696078 |  406.45844 | 0.723904  0.724815  0.725938  0.727307  0.728956  0.730919  0.733229  0.735913  0.738997  0.742497  0.746428  0.750792   0.75559  0.760812  0.766442  0.772463  0.778847  0.785569  0.792597  0.799901  0.807447  0.815205  0.823143   0.83123  0.839439  0.847742  0.856115  0.864535   0.87298  0.881431   0.88987   0.89828  0.906648  0.914961  0.923205  0.931371  0.939448  0.947427  0.955301  0.963063  0.970704   0.97822  0.985604  0.992852  0.999958  1.006917  1.013726  1.020381  1.026877        1.03321 | 2.609663  2.614942  2.621454  2.629398  2.638979  2.650397  2.663847  2.679506  2.697525  2.718026  2.741096  2.766784  2.795099  2.826012   2.85946  2.895345  2.933548  2.973923  3.016313  3.060548  3.106451  3.153847  3.202559  3.252416  3.303251  3.354907  3.407234  3.460092  3.513348  3.566881  3.620577   3.67433  3.728042  3.781625  3.834993  3.888071  3.940786  3.993074   4.04487   4.09612  4.146768  4.196763   4.24606  4.294611  4.342375   4.38931  4.435377  4.480538  4.524756        4.567996 |  101.47548  101.907495  102.445435  103.108366  103.916715  104.891855  106.055606  107.429686  109.035153  110.891864  113.017997  115.429659  118.140593     121.162  124.502466  128.167983  132.162053  136.485846  141.138404  146.116859  151.416664   157.03181  162.955037  169.178017  175.691526   182.48558  189.549557  196.872299   204.44219  212.247225  220.275058  228.513042  236.948262  245.567558  254.357535  263.304583  272.394877  281.614384  290.948865  300.383876  309.904764  319.496664  329.144501   338.83298  348.546587  358.269584  367.986005  377.679652  387.334093  396.932659   406.45844 |  100.0 |    1.47548 |  406.45844 |
    | 04:30 | 3.742681   3.74988  3.758811   3.76977  3.783068  3.799023  3.817946  3.840135  3.865858  3.895348  3.928789  3.966314  4.007997  4.053855  4.103847  4.157882   4.21582  4.277482  4.342657  4.411109  4.482584  4.556821   4.63355  4.712504  4.793421  4.876045   4.96013  5.045442  5.131758  5.218869  5.306578    5.3947  5.483063  5.571506  5.659879  5.748042  5.835865  5.923226  6.010011  6.096113  6.181432  6.265874  6.349348   6.43177   6.51306   6.59314  6.671937  6.749379  6.825397             6.899926 | 126.178495  126.556222  127.025488  127.602394  128.304056  129.148213  130.152753  131.335198  132.712175  134.298907  136.108774  138.152949  140.440151  142.976503  145.765515  148.808154  152.103011   155.64652  159.433221  163.456044  167.706584   172.17538  176.852154  181.726034  186.785741  192.019749  197.416411  202.964062  208.651095  214.466023  220.397516  226.434429   232.56582  238.780955  245.069308  251.420559   257.82458  264.271431  270.751341  277.254694  283.772017  290.293963  296.811295  303.314869  309.795624  316.244562  322.652739  329.011247  335.311204  341.543743 |  53.168002   53.241427   53.332506   53.444264   53.579879   53.742583   53.935563   54.161845   54.424173   54.724912   55.065948   55.448626   55.873709   56.341367   56.851194   57.402245     57.9931   58.621933   59.286589   59.984661   60.713571   61.470635    62.25312     63.0583   63.883495   64.726097   65.583601   66.453615   67.333872   68.222233   69.116689   70.015359   70.916487   71.818431   72.719663   73.618756    74.51438   75.405291   76.290329   77.168402   78.038489   78.899623   79.750895    80.59144   81.420436   82.237097    83.04067   83.830428   84.605668         85.36571 |  52.426805   52.498804   52.588114   52.697703   52.830683   52.990228   53.179461   53.401348   53.658583   53.953481   54.287894   54.663141   55.079969   55.538546   56.038472   56.578822   57.158203   57.774825   58.426573   59.111089   59.825845   60.568206   61.335496   62.125041    62.93421    63.76045   64.601302   65.454421   66.317584   67.188693   68.065779   68.946998   69.830627   70.715056   71.598786   72.480419   73.358651   74.232261   75.100112   75.961134   76.814324   77.658736   78.493477   79.317699   80.130596   80.931398   81.719366   82.493787   83.253973     83.999255 | 1.092235  1.093349  1.094729  1.096421  1.098472   1.10093   1.10384  1.107246  1.111185  1.115691  1.120785  1.126484  1.132792  1.139705  1.147211  1.155289   1.16391   1.17304  1.182642  1.192674  1.203092  1.213853  1.224914  1.236231  1.247763  1.259471  1.271319  1.283271  1.295296  1.307364  1.319448  1.331522  1.343564  1.355554  1.367472  1.379302  1.391027  1.402633  1.414108   1.42544  1.436617  1.447631  1.458472   1.46913    1.4796  1.489872  1.499941  1.509799  1.519442  1.528861 | 1.358907  1.358834  1.358744  1.358634  1.358502  1.358345   1.35816  1.357946  1.357701  1.357425  1.357117  1.356778   1.35641  1.356015  1.355595  1.355155  1.354696  1.354223   1.35374  1.353249  1.352755  1.352261  1.351769  1.351283  1.350803  1.350332  1.349872  1.349423  1.348986  1.348563  1.348154  1.347758  1.347377   1.34701  1.346656  1.346317  1.345991  1.345679   1.34538  1.345093  1.344818  1.344556  1.344304  1.344064  1.343835  1.343615  1.343406  1.343206  1.343016          1.342834 |  0.53811  0.538672  0.539367  0.540216  0.541242  0.542466  0.543909  0.545587  0.547515  0.549703  0.552156  0.554872  0.557847  0.561068  0.564521  0.568184  0.572037  0.576054  0.580209  0.584476   0.58883  0.593247  0.597702  0.602175  0.606647    0.6111  0.615519   0.61989  0.624204   0.62845  0.632619  0.636707  0.640707  0.644616   0.64843  0.652147  0.655766  0.659285  0.662705  0.666025  0.669246  0.672369  0.675393  0.678321  0.681154  0.683892  0.686538  0.689092  0.691555      0.693931 | -0.204415  -0.205135  -0.206021  -0.207097  -0.208389  -0.209922  -0.211715  -0.213788  -0.216153  -0.218819  -0.221787  -0.225052  -0.228605   -0.23243  -0.236505  -0.240807  -0.245309  -0.249981  -0.254796  -0.259722  -0.264734  -0.269803  -0.274906  -0.280019  -0.285122  -0.290199  -0.295231  -0.300207  -0.305115  -0.309944  -0.314687  -0.319337  -0.323889  -0.328338  -0.332681  -0.336916  -0.341042  -0.345057  -0.348961  -0.352753  -0.356434  -0.360005  -0.363466  -0.366818  -0.370062    -0.3732  -0.376232   -0.37916  -0.381985     -0.384707 | 0.665973  0.666046  0.666137  0.666246  0.666377  0.666532  0.666714  0.666924  0.667163  0.667435  0.667737  0.668072  0.668438  0.668836  0.669262  0.669716  0.670196    0.6707  0.671225  0.671769  0.672329  0.672903  0.673488  0.674083  0.674684  0.675291  0.675901  0.676512  0.677123  0.677732  0.678338  0.678941  0.679538  0.680129  0.680712  0.681288  0.681856  0.682414  0.682963  0.683502   0.68403  0.684547  0.685053  0.685547  0.686029  0.686498  0.686956    0.6874  0.687831       0.68825 | 101.232151 | 101.417242  101.830021  102.343419  102.975465  103.745505   104.67383  105.781227  107.088478  108.615852  110.382592  112.406469  114.703394  117.287128  120.169085  123.358241  126.861123   130.68189  134.822452  139.282654  144.060468  149.152205  154.552719  160.255604  166.253377  172.537637  179.099205  185.928248  193.014372  200.346714     207.914  215.704606  223.706593  231.907745  240.295591  248.857419  257.580296  266.451068  275.456373  284.582636  293.816075  303.142697  312.548296  322.018452  331.538526  341.093657  350.668756  360.248507  369.817359  379.359522          388.858968 | 393.595174 | 0.723384  0.724161  0.725123  0.726302  0.727732  0.729444  0.731472  0.733845  0.736589  0.739726  0.743272  0.747237  0.751626  0.756433  0.761651  0.767263   0.77325  0.779588   0.78625  0.793207  0.800428  0.807882   0.81554  0.823372  0.831348  0.839441  0.847627   0.85588   0.86418  0.872505  0.880836  0.889158  0.897453  0.905709  0.913912   0.92205  0.930113  0.938091  0.945976   0.95376  0.961435  0.968995  0.976434  0.983746  0.990925  0.997968  1.004869  1.011624  1.018229        1.02468 |  2.60665  2.611147  2.616725  2.623567  2.631865  2.641817  2.653614  2.667437  2.683449   2.70179  2.722568  2.745857  2.771696  2.800085   2.83099  2.864344  2.900051  2.937991  2.978024  3.019996  3.063746  3.109103    3.1559  3.203968  3.253142  3.303265  3.354186  3.405761  3.457856  3.510344  3.563108  3.616039  3.669034     3.722  3.774849  3.827501  3.879881  3.931919  3.983551  4.034716  4.085358  4.135425  4.184867  4.233636   4.28169  4.328984   4.37548  4.421138   4.46592         4.50979 | 101.232151   101.59857  102.056914  102.624476   103.32003  104.163508  105.175578  106.377169  107.788947  109.430802  111.321362  113.477561  115.914305  118.644224  121.677527  125.021959  128.682833  132.663144  136.963719   141.58341  146.519301  151.766919  157.320437  163.172864  169.316216  175.741668  182.439682  189.400122  196.612337  204.065241  211.747374  219.646943  227.751863  236.049784  244.528111  253.174018  261.974459  270.916176  279.985699   289.16935   298.45324  307.823273  317.265134  326.764296  336.306009    345.8753  355.456967  365.035576   374.59546  384.120709  393.595174 |  100.0 |   1.232151 | 393.595174 |
    | 05:00 | 3.738608  3.744727  3.752356  3.761766  3.773248  3.787105  3.803639  3.823147  3.845908  3.872171  3.902147  3.936001  3.973847  4.015745  4.061699  4.111661  4.165534  4.223178  4.284416  4.349043   4.41683  4.487534  4.560903  4.636679  4.714606  4.794435   4.87592  4.958825  5.042927  5.128013   5.21388  5.300341  5.387216  5.474341  5.561559  5.648727  5.735707  5.822374   5.90861  5.994303  6.079348  6.163649  6.247112   6.32965   6.41118  6.491623  6.570902  6.648945  6.725682             6.801045 | 125.965045  126.285823  126.686221  127.180918  127.785744  128.517371  129.392917  130.429485   131.64368  133.051117  134.665961  136.500537  138.565017  140.867215  143.412479  146.203696  149.241381  152.523835  156.047366  159.806531  163.794404  168.002835  172.422698  177.044116  181.856669  186.849557  192.011754   197.33212  202.799493  208.402768  214.130942  219.973159  225.918728  231.957146  238.078097  244.271457   250.52729  256.835835  263.187503  269.572858  275.982607  282.407585  288.838742  295.267125  301.683865  308.080168  314.447291  320.776538  327.059243  333.286759 |  53.126466   53.188875   53.266674   53.362639   53.479734    53.62104   53.789656   53.988605   54.220723   54.488553   54.794247   55.139492   55.525449   55.952724   56.421365   56.930879   57.480279   58.068135   58.692646   59.351712   60.043008    60.76405   61.512262   62.285029   63.079741   63.893835   64.724818   65.570294    66.42797   67.295675   68.171355    69.05308    69.93904   70.827541   71.716999   72.605934   73.492965   74.376802   75.256234   76.130132   76.997432   77.857134   78.708295   79.550021   80.381466   81.201822   82.010314   82.806202    83.58877        84.357325 |  52.386075   52.447273   52.523561   52.617662   52.732484   52.871045   53.036387   53.231473   53.459083   53.721712    54.02147    54.36001   54.738472    55.15745    55.61699    56.11661   56.655341   57.231781   57.844164   58.490432   59.168303   59.875343   60.609025   61.366785   62.146065    62.94435   63.759196   64.588253   65.429274   66.280128   67.138803   68.003406   68.872161   69.743408   70.615593   71.487266   72.357071   73.223744   74.086099   74.943026   75.793483   76.636491   77.471123   78.296504   79.111803   79.916227   80.709019   81.489452   82.256823     83.010452 | 1.091605  1.092552  1.093731  1.095185  1.096958  1.099094   1.10164  1.104639  1.108131  1.112151  1.116728  1.121882  1.127625  1.133962  1.140885  1.148382   1.15643  1.165002  1.174064   1.18358   1.19351  1.203811  1.214443  1.225363  1.236531  1.247907  1.259454  1.271136   1.28292  1.294776  1.306675   1.31859  1.330499  1.342379  1.354209  1.365972   1.37765  1.389229  1.400695  1.412035  1.423238  1.434292  1.445188  1.455918  1.466472  1.476843  1.487024  1.497007  1.506787  1.516357 | 1.358948  1.358886  1.358809  1.358714  1.358599  1.358462  1.358299  1.358109   1.35789  1.357641  1.357361  1.357051  1.356711  1.356343  1.355948  1.355531  1.355093  1.354639  1.354171  1.353693  1.353209  1.352722  1.352235   1.35175   1.35127  1.350797  1.350333  1.349879  1.349436  1.349005  1.348587  1.348182  1.347791  1.347414   1.34705    1.3467  1.346364  1.346041  1.345731  1.345433  1.345148  1.344875  1.344613  1.344363  1.344123  1.343895  1.343676  1.343467  1.343267          1.343076 | 0.537791   0.53827  0.538865  0.539596  0.540485  0.541552  0.542819  0.544303  0.546021  0.547986  0.550204  0.552681  0.555413  0.558395  0.561614  0.565055  0.568698  0.572521    0.5765  0.580611  0.584829  0.589128  0.593487  0.597881  0.602293  0.606702  0.611093  0.615451  0.619763  0.624019  0.628209  0.632326  0.636364  0.640316  0.644181  0.647953  0.651632  0.655215  0.658702  0.662091  0.665384   0.66858   0.67168  0.674685  0.677595  0.680412  0.683137  0.685771  0.688315      0.690771 | -0.204003  -0.204617  -0.205377  -0.206307   -0.20743  -0.208771  -0.210351  -0.212191  -0.214307  -0.216709  -0.219405  -0.222394  -0.225671  -0.229224  -0.233038  -0.237093  -0.241363  -0.245824  -0.250448  -0.255206  -0.260071  -0.265016  -0.270015  -0.275046  -0.280087  -0.285118  -0.290122  -0.295085  -0.299992  -0.304834    -0.3096  -0.314282  -0.318875  -0.323373  -0.327772  -0.332068   -0.33626  -0.340345  -0.344322  -0.348192  -0.351953  -0.355606  -0.359152   -0.36259  -0.365922  -0.369148  -0.372271   -0.37529  -0.378207     -0.381023 |  0.66593  0.665993  0.666071  0.666165  0.666279  0.666415  0.666574   0.66676  0.666974  0.667217  0.667491  0.667796  0.668131  0.668498  0.668893  0.669317  0.669768  0.670244  0.670742  0.671261  0.671797   0.67235  0.672915  0.673492  0.674078  0.674671  0.675268  0.675869  0.676471  0.677073  0.677673  0.678271  0.678865  0.679454  0.680037  0.680613  0.681182  0.681743  0.682295  0.682838  0.683371  0.683894  0.684406  0.684908  0.685397  0.685876  0.686342  0.686796  0.687238      0.687668 | 101.027458 |  101.18417  101.534486  101.972181  102.513608  103.176544  103.979894   104.94332  106.086806  107.430182  108.992644  110.792303  112.845768   115.16782  117.771155  120.666228  123.861185  127.361875  131.171923  135.292867  139.724322  144.464165   149.50873     154.853  160.490785  166.414891  172.617268  179.089135  185.821093  192.803214  200.025119  207.476041  215.144868  223.020186  231.090309  239.343296  247.766973  256.348939  265.076579  273.937061  282.917346  292.004182  301.184106  310.443437  319.768282  329.144523  338.557818  347.993595  357.437052  366.873147          376.286599 | 380.984449 | 0.722945  0.723605  0.724427  0.725441  0.726676  0.728165  0.729939  0.732029  0.734461  0.737261  0.740447  0.744035  0.748032  0.752439  0.757254  0.762464  0.768056  0.774008  0.780299  0.786901  0.793786  0.800926  0.808291  0.815851  0.823579  0.831447  0.839429    0.8475  0.855638  0.863821  0.872029  0.880245  0.888453  0.896637  0.904783  0.912879  0.920914  0.928877  0.936759  0.944552  0.952248  0.959839  0.967319  0.974682  0.981922  0.989035  0.996015  1.002858   1.00956       1.016116 | 2.604105  2.607928  2.612694   2.61857  2.625738  2.634383  2.644695  2.656855  2.671032  2.687377  2.706016  2.727046  2.750529  2.776495  2.804937  2.835816  2.869063  2.904582  2.942253  2.981943  3.023501  3.066772  3.111595  3.157807  3.205246  3.253758  3.303189  3.353396  3.404242  3.455597   3.50734  3.559359  3.611547  3.663807  3.716047  3.768184  3.820139  3.871839  3.923217   3.97421  4.024758  4.074807  4.124306  4.173204  4.221458  4.269022  4.315855  4.361918  4.407171        4.451578 | 101.027458  101.337652  101.727386  102.212243  102.809354   103.53715  104.415028  105.462934  106.700909  108.148599  109.824777  111.746909  113.930772  116.390164  119.136693  122.179669  125.526076  129.180625  133.145856  137.422299  142.008645  146.901948  152.097816  157.590596  163.373556  169.439035  175.778583  182.383084  189.242851  196.347713   203.68708      211.25  219.025203  227.001129  235.165961  243.507637  252.013867  260.672141  269.469734  278.393711  287.430926   296.56802  305.791425  315.087354    324.4418  333.840536  343.269106  352.712823  362.156767  371.585776  380.984449 |  100.0 |   1.027458 | 380.984449 |
    | 05:30 | 3.735175  3.740366  3.746867  3.754927  3.764814  3.776812  3.791214  3.808309  3.828378  3.851682  3.878451  3.908876  3.943105  3.981236  4.023314  4.069333  4.119237  4.172928  4.230263   4.29107  4.355148  4.422276   4.49222  4.564737  4.639582  4.716509  4.795278  4.875656  4.957416  5.040344  5.124235  5.208896  5.294144   5.37981  5.465732  5.551761  5.637756  5.723585  5.809127  5.894266  5.978893  6.062908  6.146213   6.22872  6.310341  6.390996  6.470606  6.549098  6.626399              6.70244 | 125.785285  126.057154  126.398055  126.821272  127.341328  127.973743  128.734711  129.640714  130.708075  131.952509  133.388661  135.029703  136.886977  138.969736  141.284969  143.837327  146.629146  149.660542  152.929574  156.432451  160.163766  164.116744  168.283478  172.655165  177.222313  181.974921  186.902646  191.994927    197.2411  202.630479  208.152426  213.796396  219.551978  225.408911  231.357104  237.386638  243.487771  249.650929  255.866704  262.125837  268.419216  274.737855  281.072885   287.41554  293.757142  300.089087  306.402836  312.689896  318.941811  325.150148 |   53.09146   53.144394   53.210694   53.292889   53.393719    53.51608   53.662945   53.837282   54.041951   54.279605   54.552592   54.862871   55.211942   55.600799    56.02991   56.499212   57.008144   57.555679   58.140388   58.760498   59.413965    60.09854   60.811831   61.551365   62.314634   63.099143   63.902435   64.722127   65.555921   66.401622   67.257145    68.12052   68.989889    69.86351   70.739747   71.617071   72.494051   73.369347   74.241706   75.109953   75.972986   76.829768   77.679323   78.520726   79.353105   80.175626   80.987497   81.787957   82.576277         83.35175 |   52.35175   52.403655   52.468668   52.549267   52.648139   52.768123   52.912137   53.083088   53.283783   53.516821   53.784507   54.088761   54.431053   54.812359   55.233137   55.693326   56.192374   56.729277   57.302631   57.910699   58.551476   59.222756   59.922196   60.647369   61.395816    62.16509   62.952783   63.756557   64.574159   65.403437   66.242347   67.088955   67.941442   68.798098   69.657319   70.517606   71.377555   72.235854   73.091272   73.942659   74.788932   75.629076   76.462133   77.287197   78.103411    78.90996   79.706065    80.49098   81.263991     82.024405 | 1.091073  1.091877  1.092883  1.094129  1.095656  1.097508  1.099727  1.102359  1.105442  1.109015  1.113111  1.117753  1.122961  1.128745  1.135104  1.142033  1.149516  1.157532  1.166053  1.175046  1.184477  1.194305  1.204492  1.214997  1.225781  1.236803  1.248027  1.259416  1.270938  1.282559  1.294251  1.305986  1.317739  1.329486  1.341207  1.352883  1.364494  1.376026  1.387463  1.398792      1.41  1.421076   1.43201  1.442791  1.453412  1.463863  1.474137  1.484228  1.494127  1.503829 | 1.358983   1.35893  1.358864  1.358783  1.358684  1.358564  1.358421  1.358254  1.358059  1.357835  1.357582  1.357299  1.356987  1.356645  1.356277  1.355884  1.355469  1.355034  1.354584  1.354121  1.353649  1.353171   1.35269   1.35221  1.351732  1.351258  1.350792  1.350334  1.349886  1.349449  1.349024  1.348611  1.348211  1.347824  1.347451   1.34709  1.346744   1.34641  1.346089  1.345781  1.345486  1.345202   1.34493   1.34467  1.344421  1.344182  1.343954  1.343735  1.343526          1.343327 | 0.537522  0.537929  0.538437  0.539065  0.539832   0.54076  0.541868  0.543175  0.544699  0.546455  0.548452  0.550699  0.553197  0.555942  0.558929  0.562144  0.565571  0.569192  0.572985  0.576928  0.580995  0.585163   0.58941  0.593711  0.598048  0.602399  0.606748  0.611079  0.615378  0.619632  0.623832  0.627969  0.632034  0.636023   0.63993  0.643751  0.647483  0.651124  0.654673  0.658127  0.661487  0.664753  0.667925  0.671003  0.673987   0.67688  0.679681  0.682392  0.685015      0.687549 | -0.203654  -0.204177  -0.204828  -0.205628  -0.206601   -0.20777  -0.209157  -0.210783  -0.212667  -0.214822  -0.217259  -0.219981  -0.222988  -0.226273  -0.229825  -0.233627  -0.237659  -0.241898  -0.246318  -0.250893  -0.255596    -0.2604   -0.26528  -0.270212  -0.275174  -0.280144  -0.285105   -0.29004  -0.294934  -0.299775  -0.304553  -0.309257  -0.313881  -0.318419  -0.322864  -0.327214  -0.331464  -0.335613  -0.339659    -0.3436  -0.347436  -0.351166  -0.354791  -0.358312  -0.361727  -0.365039  -0.368248  -0.371355  -0.374361     -0.377268 | 0.665895  0.665948  0.666014  0.666096  0.666194  0.666312  0.666452  0.666616  0.666807  0.667024  0.667271  0.667547  0.667853  0.668189  0.668555  0.668949  0.669371  0.669818  0.670289  0.670782  0.671294  0.671824  0.672369  0.672927  0.673495  0.674072  0.674656  0.675245  0.675837   0.67643  0.677023  0.677615  0.678205   0.67879  0.679371  0.679947  0.680516  0.681078  0.681633  0.682178  0.682715  0.683243  0.683761  0.684268  0.684765  0.685251  0.685726  0.686189   0.68664      0.687079 | 100.855551 |  100.98799   101.28473  101.657124  102.119917  102.689325  103.382823  104.218841  105.216394  106.394661  107.772542  109.368208  111.198691  113.279519   115.62442  118.245112  121.151176  124.350007  127.846838  131.644824  135.745173  140.147301   144.84901  149.846668  155.135384  160.709178  166.561131  172.683521  179.067942  185.705402  192.586411  199.701047  207.039012   214.58968   222.34213  230.285172   238.40737  246.697052  255.142326  263.731081  272.450993  281.289529   290.23394  299.271268  308.388336   317.57175  326.807892  336.082917  345.382749  354.693078          363.999354 | 368.647924 | 0.722574  0.723134  0.723836  0.724704  0.725769   0.72706  0.728607   0.73044  0.732588  0.735077  0.737929  0.741162  0.744786   0.74881  0.753234  0.758051  0.763252  0.768821  0.774738   0.78098  0.787523  0.794338  0.801398  0.808675   0.81614  0.823767   0.83153  0.839403  0.847363  0.855388  0.863458  0.871554  0.879658  0.887755   0.89583   0.90387  0.911862  0.919797  0.927663  0.935451  0.943154  0.950763  0.958272  0.965673  0.972962  0.980133   0.98718  0.994098  1.000884       1.007533 |  2.60196  2.605203  2.609265  2.614299  2.620473  2.627962  2.636947  2.647607  2.660114  2.674626  2.691283  2.710198  2.731456  2.755111  2.781182  2.809658  2.840495  2.873622  2.908943  2.946342  2.985688  3.026836  3.069637  3.113935  3.159575  3.206403   3.25427  3.303029  3.352543  3.402681  3.453318  3.504339  3.555633  3.607101  3.658646  3.710182  3.761626  3.812903  3.863941  3.914675  3.965044  4.014991   4.06446  4.113402  4.161769  4.209516    4.2566  4.302979  4.348615        4.393469 | 100.855551  101.117665  101.448408  101.861742  102.373194  102.999679  103.759241  104.670712   105.75331  107.026198  108.508033  110.216526  112.168049  114.377302  116.857057  119.617989   122.66859   126.01516  129.661863  133.610844  137.862367  142.414993  147.265758  152.410353    157.8433  163.558108  169.547423  175.803151  182.316567   189.07841  196.078955  203.308079   210.75531  218.409865  226.260681  234.296441  242.505588  250.876336  259.396682   268.05441  276.837088  285.732077  294.726524  303.807362  312.961305  322.174849  331.434264   340.72559  350.034635   359.34697  368.647924 |  100.0 |   0.855551 | 368.647924 |
    | 06:00 | 3.732287  3.736681  3.742209  3.749096  3.757587  3.767948  3.780454  3.795387  3.813023  3.833629  3.857447  3.884688  3.915529  3.950098  3.988478  4.030703  4.076759  4.126582  4.180071  4.237086  4.297455  4.360983  4.427459  4.496654  4.568337  4.642272  4.718224  4.795962  4.875263   4.95591  5.037697  5.120429  5.203918   5.28799   5.37248  5.457234  5.542106  5.626961   5.71167  5.796116  5.880185  5.963773  6.046781  6.129114  6.210684  6.291407  6.371204  6.449998  6.527715             6.604285 | 125.634162  125.864133  126.153775  126.515033  126.961131  127.506391  128.165985  128.955606  129.891091  130.987996  132.261172  133.724345  135.389749  137.267817  139.366954    141.6934   144.25118   147.04214  150.066047  153.320757  156.802405  160.505636  164.423823  168.549303   172.87358  177.387522  182.081532   186.94569  191.969876  197.143873  202.457442  207.900384  213.462586  219.134054  224.904935  230.765528   236.70629  242.717839  248.790949  254.916542  261.085682   267.28956  273.519486  279.766874  286.023233  292.280147   298.52927  304.762308  310.971009   317.14715 |  53.062013   53.106817   53.163193   53.233425   53.320024   53.425683    53.55322   53.705503   53.885362   54.095499   54.338391   54.616204   54.930713   55.283249   55.674654    56.10527   56.574943   57.083047    57.62853   58.209964   58.825611    59.47348   60.151395   60.857054   61.588081    62.34207    63.11663   63.909408    64.71812   65.540564   66.374636   67.218332   68.069759    68.92713   69.788765   70.653086   71.518614   72.383964   73.247838   74.109017    74.96636   75.818793   76.665304    77.50494   78.336797   79.160018   79.973788   80.777327   81.569888        82.350753 |  52.322875   52.366809    52.42209   52.490958   52.575875   52.679482   52.804542   52.953868   53.130234   53.336291   53.574466   53.846884   54.155286   54.500976    54.88478   55.307034   55.767586   56.265823   56.800713   57.370856   57.974547   58.609835   59.274585   59.966541   60.683372   61.422719   62.182237    62.95962   63.752628   64.559101   65.376975   66.204287    67.03918   67.879902   68.724804    69.57234   70.421061   71.269606   72.116704    72.96116   73.801854   74.637733   75.467806   76.291136   77.106839   77.914074   78.712041   79.499976   80.277146     81.042847 | 1.090626  1.091306  1.092162  1.093227   1.09454   1.09614  1.098069   1.10037  1.103083  1.106248  1.109898  1.114064  1.118767  1.124023  1.129841  1.136218  1.143148  1.150614  1.158595  1.167064  1.175988  1.185333  1.195062  1.205136  1.215517  1.226167  1.237048  1.248124  1.259361  1.270726  1.282189  1.293722  1.305298  1.316892  1.328484  1.340051  1.351575   1.36304  1.374429  1.385727  1.396922  1.408001  1.418954  1.429769  1.440438  1.450952  1.461302  1.471482  1.481483  1.491299 | 1.359012  1.358968  1.358911  1.358842  1.358756  1.358652  1.358528   1.35838  1.358208  1.358008  1.357781  1.357524  1.357238  1.356924  1.356581  1.356213  1.355821  1.355408  1.354977  1.354531  1.354073  1.353607  1.353135   1.35266  1.352186  1.351715  1.351248  1.350788  1.350336  1.349894  1.349463  1.349043  1.348635  1.348239  1.347857  1.347487   1.34713  1.346786  1.346456  1.346137  1.345832  1.345538  1.345256  1.344986  1.344726  1.344478   1.34424  1.344012  1.343794          1.343585 | 0.537296   0.53764  0.538073  0.538611  0.539272  0.540075  0.541041  0.542188  0.543534  0.545096  0.546887  0.548915  0.551187  0.553703   0.55646  0.559449  0.562658  0.566071  0.569669  0.573432  0.577337  0.581361  0.585481  0.589676  0.593922  0.598202  0.602495  0.606785  0.611058  0.615299  0.619498  0.623644  0.627728  0.631744  0.635686  0.639548  0.643327  0.647019  0.650624  0.654138  0.657561  0.660892  0.664132  0.667279  0.670335    0.6733  0.676175  0.678961  0.681659      0.684269 |  -0.20336  -0.203804   -0.20436  -0.205047  -0.205887  -0.206902  -0.208115  -0.209547  -0.211217  -0.213142  -0.215334  -0.217801  -0.220547  -0.223569   -0.22686  -0.230408  -0.234196  -0.238205  -0.242411   -0.24679  -0.251317  -0.255966  -0.260711  -0.265528  -0.270395  -0.275289  -0.280191  -0.285084  -0.289951  -0.294779  -0.299556  -0.304272  -0.308917  -0.313484  -0.317967  -0.322362  -0.326664  -0.330869  -0.334976  -0.338983  -0.342888  -0.346691  -0.350392  -0.353989  -0.357485  -0.360878   -0.36417  -0.367362  -0.370454     -0.373447 | 0.665864   0.66591  0.665966  0.666036  0.666121  0.666224  0.666346  0.666491  0.666659  0.666853  0.667074  0.667323  0.667601  0.667908  0.668245   0.66861  0.669003  0.669422  0.669865  0.670332   0.67082  0.671326  0.671849  0.672387  0.672937  0.673497  0.674066  0.674641  0.675221  0.675805  0.676389  0.676974  0.677558  0.678139  0.678717   0.67929  0.679859  0.680421  0.680976  0.681524  0.682064  0.682595  0.683117   0.68363  0.684133  0.684625  0.685107  0.685578  0.686038      0.686487 | 100.711411 | 100.823137  101.074029  101.390242  101.784989  102.272971  102.870219  103.593865  104.461837  105.492491  106.704212  108.114991  109.742022  111.601318  113.707391  116.072996  118.708954  121.624043  124.824979   128.31644  132.101163  136.180062  140.552385  145.215874  150.166939  155.400815  160.911721  166.693001  172.737243  179.036393  185.581843  192.364512  199.374906  206.603172  214.039139   221.67235  229.492084  237.487378  245.647039  253.959651  262.413585  270.996997  279.697833  288.503827    297.4025  306.381154  315.426873  324.526517  333.666717  342.833873          352.014148 | 356.603706 | 0.722262  0.722737  0.723333  0.724076  0.724991  0.726106  0.727451  0.729054  0.730945   0.73315  0.735692  0.738593  0.741867  0.745525  0.749573  0.754009  0.758827  0.764016   0.76956   0.77544  0.781634  0.788117  0.794862  0.801844  0.809035  0.816408  0.823937  0.831597  0.839365  0.847217  0.855133  0.863093  0.871079  0.879075  0.887064  0.895033   0.90297  0.910861  0.918698  0.926469  0.934166   0.94178  0.949305  0.956733  0.964058  0.971274  0.978376  0.985358  0.992216       0.998946 | 2.600155  2.602901  2.606355  2.610657  2.615961  2.622429  2.630234  2.639549  2.650546  2.663385  2.678214  2.695162  2.714331  2.735796    2.7596  2.785757   2.81425   2.84503  2.878026  2.913142  2.950266  2.989268  3.030011  3.072348  3.116133  3.161214  3.207446  3.254685  3.302791  3.351632  3.401083  3.451023  3.501341  3.551933    3.6027  3.653552  3.704402  3.755172  3.805788  3.856182  3.906289   3.95605  4.005408   4.05431  4.102708  4.150554  4.197804  4.244415  4.290348        4.335563 | 100.711411  100.932502  101.212649  101.564295  102.001427   102.53946  103.195052  103.985822  104.930019  106.046127  107.352452  108.866696   110.60556  112.584389  114.816879  117.314857  120.088144  123.144484  126.489561  130.127051  134.058743  138.284675  142.803295  147.611634  152.705472  158.079495  163.727448  169.642264  175.816185  182.240854  188.907409  195.806546  202.928579  210.263485  217.800943  225.530358  233.440882  241.521436  249.760711  258.147185   266.66912  275.314568  284.071367  292.927146  301.869317  310.885075  319.961391  329.085013  338.242458  347.420008  356.603706 |  100.0 |   0.711411 | 356.603706 |
    | 06:30 | 3.729863  3.733574  3.738265  3.744136  3.751411  3.760334  3.771163  3.784166  3.799615  3.817773   3.83889  3.863193  3.890877  3.922099  3.956974  3.995572  4.037915  4.083981  4.133702  4.186973  4.243656  4.303582  4.366563  4.432392  4.500851  4.571716  4.644762  4.719762  4.796496   4.87475  4.954315  5.034995  5.116599  5.198951   5.28188  5.365228  5.448845  5.532592  5.616337  5.699956  5.783334   5.86636  5.948934  6.030957  6.112339  6.192994  6.272837  6.351792  6.429783             6.506739 | 125.507328  125.701496  125.947093  126.254803  126.636581  127.105543  127.675774  128.362068  129.179601  130.143563  131.268758  132.569204  134.057755  135.745766  137.642825   139.75656   142.09253  144.654195  147.442964  150.458305  153.697899  157.157834  160.832812  164.716363  168.801054  173.078687  177.540472  182.177191  186.979324  191.937167  197.040922  202.280771  207.646929   213.12969  218.719455  224.406754  230.182255  236.036773  241.961268  247.946843  253.984738   260.06632  266.183074  272.326592  278.488561   284.66075     290.835   297.00321  303.157325  309.289324 |  53.037287   53.075136   53.122971   53.182843   53.257034   53.348028   53.458463   53.591076   53.748623   53.933801   54.149155   54.396995   54.679312   54.997715   55.353375      55.747   56.178818   56.648593   57.155649   57.698913   58.276966     58.8881   59.530381   60.201705   60.899853    61.62254    62.36746   63.132316   63.914856   64.712887     65.5243   66.347071   67.179279     68.0191   68.864814   69.714801   70.567537   71.421592   72.275625   73.128375   73.978661   74.825371   75.667459   76.503938   77.333876    78.15639    78.97064   79.775826   80.571182        81.355975 |  52.298629   52.335742   52.382649   52.441358   52.514109   52.603335   52.711626   52.841663   52.996151   53.177733   53.388905   53.631932   53.908767   54.220986    54.56974    54.95572   55.379153   55.839806   56.337015   56.869729   57.436556   58.035822   58.665631   59.323918   60.008508   60.717162   61.447615   62.197619   62.964962   63.747497   64.543152   65.349946   66.165993   66.989505   67.818796   68.652277   69.488453   70.325923   71.163371   71.999562   72.833336   73.663603   74.489338   75.309573   76.123395   76.929936   77.728373   78.517923   79.297834     80.067387 | 1.090251  1.090825  1.091552   1.09246  1.093585  1.094964  1.096636  1.098642  1.101021  1.103813  1.107055  1.110778  1.115008  1.119767  1.125067  1.130914  1.137305  1.144232  1.151679  1.159623  1.168037  1.176891  1.186152  1.195782  1.205745  1.216005  1.226525  1.237268    1.2482  1.259288  1.270502  1.281811  1.293189  1.304611  1.316052  1.327491  1.338909  1.350287  1.361608  1.372857   1.38402  1.395084  1.406038   1.41687   1.42757  1.438128  1.448537  1.458788  1.468874  1.478787 | 1.359037  1.358999  1.358952  1.358892  1.358818  1.358729   1.35862  1.358491  1.358339  1.358161  1.357958  1.357726  1.357466  1.357178  1.356862  1.356519  1.356151  1.355761   1.35535  1.354922   1.35448  1.354027  1.353566    1.3531  1.352632  1.352164  1.351699  1.351239  1.350785  1.350339  1.349903  1.349477  1.349062  1.348658  1.348267  1.347889  1.347523   1.34717  1.346829  1.346501  1.346185  1.345881  1.345589  1.345309   1.34504  1.344782  1.344534  1.344297   1.34407          1.343852 | 0.537106  0.537397  0.537764  0.538224  0.538791  0.539485  0.540324  0.541327  0.542511  0.543895  0.545493  0.547316  0.549373  0.551668    0.5542  0.556965  0.559955  0.563156  0.566553  0.570128  0.573861  0.577729   0.58171  0.585783  0.589927   0.59412  0.598344  0.602581  0.606815  0.611031  0.615217  0.619361  0.623454  0.627488  0.631456  0.635351  0.639169  0.642907  0.646562   0.65013  0.653611  0.657003  0.660306  0.663519  0.666643  0.669677  0.672623  0.675481  0.678251      0.680935 | -0.203112  -0.203489  -0.203962   -0.20455  -0.205273  -0.206152  -0.207209  -0.208465   -0.20994  -0.211652  -0.213615  -0.215841  -0.218337  -0.221103  -0.224138  -0.227432  -0.230974  -0.234747  -0.238731  -0.242904  -0.247242  -0.251722  -0.256317  -0.261005  -0.265761  -0.270563  -0.275392  -0.280228  -0.285054  -0.289857  -0.294621  -0.299335  -0.303991  -0.308578   -0.31309  -0.317521  -0.321866  -0.326121  -0.330283  -0.334349  -0.338318  -0.342188  -0.345958  -0.349629    -0.3532  -0.356671  -0.360042  -0.363315  -0.366489     -0.369565 | 0.665839  0.665877  0.665926  0.665985  0.666059  0.666148  0.666254  0.666381  0.666529  0.666701  0.666898  0.667122  0.667374  0.667654  0.667963  0.668299  0.668664  0.669055  0.669471  0.669911  0.670373  0.670855  0.671356  0.671872  0.672403  0.672945  0.673498  0.674059  0.674626  0.675198  0.675772  0.676349  0.676925  0.677501  0.678074  0.678645  0.679211  0.679772  0.680327  0.680876  0.681418  0.681951  0.682477  0.682994  0.683502     0.684  0.684488  0.684966  0.685434       0.68589 | 100.590745 | 100.684832   100.89658  101.164575  101.500596   101.91789  102.431066  103.055929  103.809235  104.708389  105.771082  107.014919  108.457013  110.113622  111.999798  114.129107    116.5134  119.162671  122.084969  125.286389  128.771112  132.541498  136.598205  140.940334  145.565584  150.470411  155.650177  161.099292  166.811347  172.779224  178.995201  185.451032  192.138018  199.047071  206.168755  213.493328  221.010767  228.710797  236.582902  244.616338  252.800145  261.123147  269.573957  278.140977    286.8124  295.576202  304.420144  313.331767  322.298386  331.307091          340.344737 | 344.866663 |    0.722  0.722401  0.722908  0.723541  0.724325  0.725287  0.726452   0.72785  0.729508  0.731454  0.733712  0.736305   0.73925  0.742563  0.746252  0.750319  0.754765   0.75958  0.764755  0.770273  0.776116  0.782261  0.788684  0.795361  0.802266  0.809373  0.816655  0.824089   0.83165  0.839315  0.847062  0.854872  0.862726  0.870605  0.878495   0.88638  0.894247  0.902082  0.909876  0.917617  0.925295  0.932903  0.940431  0.947873  0.955223  0.962472  0.969617  0.976651   0.98357       0.990368 |  2.59864   2.60096  2.603891  2.607559  2.612103  2.617675  2.624436   2.63255  2.642186  2.653506  2.666662  2.681791   2.69901  2.718413  2.740063  2.763997  2.790222  2.818714  2.849424   2.88228  2.917186  2.954031   2.99269  3.033031  3.074914  3.118195  3.162731  3.208382  3.255009   3.30248  3.350667   3.39945  3.448713   3.49835  3.548259  3.598345  3.648522  3.698705   3.74882  3.798793  3.848558  3.898053   3.94722  3.996003  4.044351  4.092216  4.139552  4.186314  4.232462        4.277955 | 100.590745  100.776908  101.013762  101.312339  101.685166  102.146207  102.710724  103.395073  104.216425   105.19243  106.340845  107.679135  109.224088  110.991447  112.995593  115.249286  117.763485  120.547222  123.607567  126.949641  130.576686  134.490173  138.689941  143.174349  147.940431  152.984055  158.300069  163.882443  169.724382  175.818441  182.156615  188.730413  195.530928  202.548885  209.774686  217.198441  224.809997  232.598953  240.554676  248.666313  256.922794  265.312837   273.82495  282.447431  291.168363  299.975617  308.856845  317.799476  326.790713  335.817529  344.866663 |  100.0 |   0.590745 | 344.866663 |
    | 07:00 |  3.72783   3.73096  3.734932  3.739926  3.746144  3.753808  3.763159  3.774451  3.787941   3.80389   3.82255  3.844154  3.868914  3.897009  3.928581  3.963732  4.002517  4.044951  4.091003  4.140601  4.193641  4.249983  4.309463  4.371898  4.437088  4.504822  4.574886  4.647062  4.721134   4.79689  4.874125  4.952638   5.03224   5.11275  5.193995  5.275813  5.358051  5.440564  5.523216   5.60588  5.688437  5.770773  5.852783  5.934366  6.015428   6.09588  6.175638  6.254619  6.332749             6.409952 | 125.401056  125.564697  125.772547  126.034103  126.360112    126.7625  127.254238  127.849141  128.561599  129.406261  130.397675  131.549909  132.876185  134.388528   136.09747  138.011818  140.138484  142.482409  145.046545  147.831914  150.837721  154.061503  157.499319  161.145937  164.995044  169.039433  173.271186  177.681843  182.262541  187.004138  191.897323  196.932693  202.100824  207.392324  212.797869  218.308235  223.914314  229.607123  235.377817   241.21768   247.11813  253.070707  259.067071  265.098985  271.158313  277.237002  283.327074  289.420615   295.50976  301.586685 |   53.01656   53.048473   53.088979   53.139908   53.203318   53.281482   53.376847   53.491994    53.62957   53.792222   53.982511   54.202832   54.455335   54.741849   55.063825    55.42229   55.817828   56.250568   56.720203   57.226014   57.766911   58.341489   58.948075   59.584788   60.249592    60.94035   61.654864   62.390918   63.146309   63.918875   64.706513   65.507197   66.318984   67.140026   67.968567   68.802951   69.641613   70.483081   71.325973   72.168987   73.010902   73.850569   74.686907   75.518896   76.345573   77.166027   77.979395   78.784853   79.581618        80.368938 |  52.278304   52.309597   52.349316   52.399256   52.461436   52.538082   52.631595   52.744505    52.87941   53.038903   53.225497    53.44154   53.689139   53.970089   54.285812   54.637317   55.025173    55.44951   55.910025   56.406013   56.936407   57.499827   58.094633   58.718981   59.370876   60.048219   60.748858   61.470618    62.21134   62.968903   63.741246   64.526381   65.322404   66.127502   66.939954   67.758134   68.580509   69.405637    70.23216   71.058804    71.88437   72.707731   73.527827   74.343659   75.154283   75.958805   76.756377   77.546195   78.327487     79.099518 | 1.089936   1.09042  1.091036  1.091809  1.092771  1.093956  1.095401  1.097143  1.099223  1.101679  1.104547  1.107862  1.111653  1.115944  1.120753  1.126092  1.131964  1.138365  1.145286   1.15271  1.160615  1.168974  1.177758  1.186934  1.196467  1.206322  1.216463  1.226855  1.237464  1.248256    1.2592  1.270266  1.281426  1.292654  1.303925  1.315217  1.326508   1.33778  1.349015  1.360197  1.371311  1.382342  1.393279  1.404109  1.414823  1.425409   1.43586  1.446166   1.45632  1.466313 | 1.359058  1.359026  1.358985  1.358935  1.358872  1.358794    1.3587  1.358587  1.358454  1.358297  1.358115  1.357907  1.357672  1.357409  1.357119  1.356801  1.356458  1.356091  1.355702  1.355294   1.35487  1.354432  1.353984  1.353528  1.353068  1.352605  1.352144  1.351685  1.351231  1.350783  1.350343  1.349912  1.349491  1.349081  1.348682  1.348296  1.347921  1.347558  1.347208  1.346871  1.346545  1.346232   1.34593   1.34564  1.345361  1.345094  1.344837   1.34459  1.344354          1.344127 | 0.536946  0.537192  0.537503  0.537894   0.53838  0.538978  0.539704  0.540578  0.541617  0.542838  0.544258  0.545889  0.547743  0.549826  0.552141  0.554687  0.557459  0.560447   0.56364   0.56702  0.570571  0.574273  0.578104  0.582044  0.586071  0.590164  0.594306  0.598476  0.602657  0.606836  0.610997  0.615129  0.619221  0.623263  0.627248  0.631168  0.635019  0.638795  0.642493   0.64611  0.649643  0.653091  0.656453  0.659728  0.662916  0.666016   0.66903  0.671956  0.674797      0.677552 | -0.202904  -0.203223  -0.203624  -0.204127  -0.204747  -0.205506  -0.206424  -0.207522  -0.208819  -0.210335  -0.212086  -0.214086  -0.216343  -0.218864  -0.221649  -0.224693  -0.227989  -0.231523  -0.235279  -0.239237  -0.243377  -0.247675  -0.252108  -0.256651  -0.261282  -0.265978  -0.270719  -0.275484  -0.280256  -0.285018  -0.289757  -0.294459  -0.299113   -0.30371  -0.308241    -0.3127   -0.31708  -0.321376  -0.325586  -0.329705  -0.333731  -0.337663  -0.341498  -0.345237  -0.348879  -0.352423  -0.355869  -0.359219  -0.362472     -0.365628 | 0.665817   0.66585  0.665891  0.665942  0.666005  0.666082  0.666174  0.666285  0.666415  0.666567  0.666743  0.666944  0.667171  0.667425  0.667706  0.668015  0.668352  0.668715  0.669104  0.669518  0.669954  0.670412  0.670889  0.671384  0.671894  0.672418  0.672953  0.673498  0.674051   0.67461  0.675174   0.67574  0.676309  0.676877  0.677445  0.678011  0.678574  0.679133  0.679687  0.680235  0.680778  0.681313  0.681841  0.682361  0.682873  0.683376  0.683869  0.684354  0.684828      0.685291 | 100.489885 | 100.568984  100.747384  100.974092  101.259562   101.61566  102.055618  102.593916  103.246094  104.028507  104.958021  106.051668  107.326283  108.798136  110.482592    112.3938  114.544439  116.945526  119.606291  122.534118  125.734536  129.211281  132.966373  137.000244  141.311872  145.898923  150.757905  155.884302  161.272711  166.916955  172.810198  178.945028  185.313538  191.907394  198.717885  205.735971   212.95231  220.357296  227.941069  235.693536  243.604382  251.663075   259.85887  268.180814  276.617744  285.158287  293.790856  302.503648  311.284641  320.121589          329.002018 | 333.448727 | 0.721781  0.722119  0.722548  0.723087  0.723758  0.724584  0.725591  0.726806  0.728255  0.729966  0.731965  0.734274  0.736914  0.739902   0.74325  0.746965   0.75105  0.755501  0.760313  0.765471  0.770962  0.776766  0.782862  0.789227  0.795836  0.802665   0.80969  0.816884  0.824225  0.831688  0.839254  0.846899  0.854606  0.862356  0.870133   0.87792  0.885703   0.89347  0.901207  0.908905  0.916553  0.924141  0.931661  0.939106  0.946467  0.953739  0.960915   0.96799  0.974957       0.981813 |  2.59737  2.599325  2.601808  2.604928  2.608813    2.6136   2.61944  2.626488  2.634905  2.644852  2.656482  2.669939   2.68535  2.702823  2.722439  2.744255  2.768301  2.794576  2.823054  2.853683  2.886389  2.921078  2.957643  2.995962  3.035907  3.077341  3.120129  3.164132  3.209216  3.255248  3.302101  3.349652  3.397786  3.446391  3.495365   3.54461  3.594035  3.643555   3.69309  3.742566  3.791912  3.841065  3.889963   3.93855  3.986771  4.034577   4.08192  4.128755  4.175039        4.220731 | 100.489885  100.646374  100.846266  101.099299  101.416638  101.810851  102.295835  102.886659  103.599345  104.450594  105.457452  106.636954  108.005753  109.579759  111.373807  113.401373  115.674353  118.202896  120.995318  124.058071  127.395767  131.011257  134.905732  139.078858  143.528916  148.252955  153.246934  158.505857  164.023904  169.794543  175.810624   182.06447  188.547943  195.252507  202.169277  209.289051  216.602351  224.099437   231.77033  239.604823   247.59249  255.722692   263.98458  272.367096  280.858972  289.448728  298.124671  306.874889  315.687249  324.549392  333.448727 |  100.0 |   0.489885 | 333.448727 |
    | 07:30 |  3.72613  3.728763   3.73212  3.736359  3.741662  3.748229  3.756284   3.76606  3.777806  3.791771  3.808204  3.827345  3.849412  3.874602  3.903079  3.934972  3.970368  4.009314  4.051813  4.097829  4.147288  4.200082  4.256076  4.315108     4.377  4.441557  4.508579  4.577857  4.649183   4.72235  4.797153  4.873395  4.950885   5.02944  5.108885  5.189055  5.269792   5.35095  5.432387  5.513974  5.595586  5.677107  5.758428  5.839445  5.920061  6.000184  6.079727  6.158606  6.236743             6.314062 | 125.312153  125.449827  125.625402  125.847286   126.12508  126.469561  126.892584  127.406944  128.026157  128.764194   129.63517  130.652997  131.831032  133.181732  134.716339  136.444618  138.374647  140.512687  142.863113  145.428421  148.209289  151.204693  154.412055  157.827416  161.445622  165.260512  169.265099  173.451736  177.812271  182.338181  187.020683  191.850831  196.819597  201.917931  207.136809   212.46727  217.900442   223.42756  229.039975  234.729161  240.486714  246.304348  252.173892  258.087283  264.036554  270.013826  276.011298  282.021234  288.035955  294.047826 |  52.999215   53.026073   53.060306   53.103536   53.157611    53.22459   53.306726   53.406428   53.526208   53.668627   53.836217   54.031409   54.256452   54.513342   54.803757   55.129001   55.489971   55.887138   56.320543   56.789815   57.294201   57.832604   58.403629   59.005642   59.636813   60.295174   60.978663   61.685166   62.412552   63.158707   63.921554   64.699074    65.48932   66.290425   67.100609   67.918183   68.741548   69.569194   70.399699   71.231723   72.064006   72.895362   73.724674   74.550891   75.373017   76.190114   77.001292   77.805705   78.602548        79.391051 |  52.261296   52.287633   52.321201   52.363591   52.416616   52.482294   52.562836   52.660601   52.778055   52.917709   53.082044   53.273445   53.494118    53.74602   54.030795   54.349723   54.703683   55.093137   55.518126   55.978285   56.472876   57.000823    57.56076   58.151082   58.769996   59.415572   60.085788   60.778571   61.491832   62.223497   62.971531   63.733952   64.508852     65.2944   66.088851   66.890548   67.697923   68.509497   69.323874   70.139741   70.955861   71.771073   72.584281   73.394453   74.200614   75.001843   75.797269   76.586061   77.367429      78.14062 | 1.089672   1.09008    1.0906  1.091257  1.092077  1.093093  1.094338  1.095849  1.097661  1.099813  1.102342  1.105283  1.108668  1.112522   1.11687  1.121725  1.127098  1.132991  1.139398  1.146309  1.153708  1.161572  1.169876   1.17859  1.187682  1.197119  1.206867  1.216892   1.22716  1.237637  1.248293  1.259097   1.27002  1.281035  1.292116  1.303241  1.314388  1.325536  1.336666  1.347762  1.358808  1.369789  1.380691  1.391503  1.402214  1.412812  1.423287  1.433632  1.443837  1.453895 | 1.359075  1.359048  1.359014  1.358971  1.358917  1.358851  1.358769  1.358671  1.358554  1.358416  1.358255  1.358069  1.357857  1.357618  1.357353   1.35706  1.356742  1.356399  1.356032  1.355645   1.35524  1.354819  1.354385  1.353942  1.353491  1.353037   1.35258  1.352125  1.351672  1.351223  1.350781  1.350347  1.349921  1.349505    1.3491  1.348706  1.348324  1.347953  1.347594  1.347247  1.346912  1.346589  1.346278  1.345978   1.34569  1.345413  1.345147  1.344891  1.344645           1.34441 | 0.536812  0.537019  0.537283  0.537615   0.53803  0.538543   0.53917  0.539929  0.540837  0.541911  0.543167  0.544621  0.546284  0.548166  0.550273  0.552606  0.555164   0.55794  0.560926  0.564108  0.567472  0.570998  0.574669  0.578464  0.582362  0.586344  0.590389  0.594479  0.598597  0.602725   0.60685  0.610959  0.615038  0.619079  0.623071  0.627008  0.630883   0.63469  0.638425  0.642084  0.645664  0.649163  0.652579  0.655911  0.659159  0.662321  0.665399  0.668391  0.671299      0.674122 | -0.202729  -0.202998  -0.203339  -0.203767  -0.204298  -0.204951  -0.205746  -0.206702  -0.207839  -0.209176  -0.210732   -0.21252  -0.214553  -0.216839  -0.219383  -0.222183  -0.225235   -0.22853  -0.232055  -0.235793  -0.239726  -0.243832   -0.24809  -0.252476  -0.256968  -0.261544  -0.266182  -0.270862  -0.275566  -0.280275  -0.284976  -0.289653  -0.294295   -0.29889   -0.30343  -0.307907  -0.312313  -0.316643  -0.320893  -0.325058  -0.329135  -0.333122  -0.337017  -0.340819  -0.344527   -0.34814  -0.351657   -0.35508  -0.358408     -0.361641 | 0.665799  0.665827  0.665862  0.665905  0.665959  0.666025  0.666105  0.666202  0.666316   0.66645  0.666606  0.666785  0.666989  0.667218  0.667474  0.667757  0.668067  0.668403  0.668765  0.669152  0.669563  0.669996  0.670449  0.670921   0.67141  0.671914  0.672431  0.672959  0.673497  0.674042  0.674593  0.675149  0.675708  0.676269   0.67683   0.67739  0.677948  0.678504  0.679056  0.679603  0.680145  0.680681   0.68121  0.681733  0.682248  0.682754  0.683252  0.683741  0.684221      0.684692 | 100.405708 | 100.472098  100.622146  100.813583  101.055643  101.358904   101.73528   102.19795  102.761225  103.440352  104.251265  105.210282  106.333779  107.637843  109.137933  110.848565  112.783037  114.953202  117.369304   120.03987  122.971671   126.16973  129.637372  133.376319  137.386801  141.667691  146.216634  151.030193  156.103973  161.432747  167.010565   172.83085  178.886487  185.169889  191.673065   198.38766  205.305005   212.41614  219.711847  227.182662  234.818892  242.610626  250.547739  258.619897  266.816559  275.126973  283.540183  292.045016  300.630088  309.283795          317.994307 | 322.359175 | 0.721597  0.721882  0.722244  0.722702  0.723274  0.723983  0.724851  0.725903  0.727166  0.728666  0.730429  0.732478  0.734835   0.73752  0.740546  0.743926  0.747665  0.751764   0.75622  0.761024  0.766165  0.771627  0.777392  0.783439  0.789745  0.796288  0.803043  0.809987  0.817095  0.824345  0.831714  0.839182  0.846729  0.854336  0.861985  0.869661  0.877348  0.885033  0.892702  0.900344  0.907949  0.915505  0.923006  0.930441  0.937803  0.945086  0.952282  0.959385  0.966391       0.973294 | 2.596306  2.597953  2.600051    2.6027  2.606013  2.610116  2.615146  2.621251  2.628581  2.637294  2.647542   2.65947  2.673213  2.688889  2.706596  2.726407  2.748372  2.772511  2.798821  2.827272  2.857808  2.890358  2.924827   2.96111  2.999091  3.038643  3.079637  3.121942  3.165425  3.209956  3.255408  3.301659  3.348591  3.396093  3.444059  3.492389  3.540989   3.58977  3.638651  3.687555  3.736408  3.785144  3.833699  3.882015  3.930035  3.977707  4.024983  4.071815  4.118159        4.163973 | 100.405708  100.537037  100.705441  100.919482  101.189059  101.525429  101.941164  102.450052  103.066934  103.807476   104.68789  105.724617  106.933984  108.331856  109.933304  111.752305  113.801491  116.091948  118.633083  121.432552  124.496242  127.828309  131.431248  135.306002  139.452085  143.867719  148.549972  153.494893  158.697638  164.152591  169.853463  175.793387  181.964994  188.360479  194.971657  201.790006  208.806704  216.012656  223.398516  230.954702  238.671409   246.53862  254.546101  262.683417   270.93992  279.304757  287.766865  296.314966  304.937565  313.622948  322.359175 |  100.0 |   0.405708 | 322.359175 |
    | 08:00 | 3.724709  3.726921  3.729753  3.733344  3.737856  3.743471   3.75039  3.758833   3.76903  3.781221  3.795649   3.81255   3.83215  3.854657  3.880251  3.909083  3.941268  3.976882  4.015962  4.058502  4.104463  4.153767  4.206306  4.261945  4.320527  4.381878  4.445812  4.512132  4.580641  4.651136  4.723418  4.797293  4.872568  4.949062  5.026598  5.105007  5.184131  5.263817  5.343923  5.424314  5.504863   5.58545  5.665962  5.746292  5.826341  5.906012  5.985217  6.063869  6.141889             6.219197 | 125.237898  125.353528  125.501574   125.68944  125.925665  126.219926   126.58299  127.026614  127.563383  128.206492  128.969484  129.865934  130.909133  132.111746  133.485503  135.040908  136.787004  138.731201  140.879158  143.234737  145.800023   148.57539  151.559616  154.750026  158.142661  161.732447  165.513376  169.478674  173.620952  177.932353  182.404674   187.02947  191.798143  196.702018  201.732397  206.880604  212.138022  217.496112  222.946434  228.480656  234.090555  239.768021  245.505055   251.29376  257.126338  262.995078   268.89235  274.810593  280.742306  286.680035 |  52.984723   53.007288   53.036165   53.072786   53.118799   53.176061   53.246626   53.332724   53.436714   53.561042   53.708175   53.880533   54.080418    54.30994   54.570951   54.864983   55.193208   55.556401   55.954932   56.388764   56.857473   57.360275   57.896068   58.463477   59.060901   59.686563    60.33856   61.014901   61.713552   62.432464   63.169604   63.922977   64.690643    65.47073   66.261441   67.061063   67.867968   68.680613   69.497539   70.317371   71.138812   71.960639   72.781701   73.600913   74.417249   75.229742   76.037474   76.839576   77.635219        78.423617 |  52.247086   52.269212   52.297528   52.333439   52.378558   52.434707   52.503903   52.588328   52.690299   52.812213   52.956488   53.125499   53.321503   53.546568    53.80251   54.090832   54.412683   54.768824   55.159615   55.585023   56.044629   56.537667   57.063056   57.619446   58.205268    58.81878   59.458116   60.121323   60.806406   61.511357   62.234183   62.972926   63.725684   64.490622   65.265978   66.050073   66.841308   67.638171   68.439234   69.243145   70.048634   70.854502    71.65962   72.462923   73.263407   74.060121   74.852168   75.638693   76.418886     77.191973 | 1.089452  1.089795  1.090233   1.09079  1.091488  1.092357  1.093428  1.094732  1.096307  1.098188   1.10041  1.103011  1.106021  1.109471  1.113386  1.117785  1.122682  1.128085  1.133994  1.140404  1.147304  1.154675  1.162497  1.170744  1.179387  1.188396  1.197739  1.207382  1.217294   1.22744   1.23779  1.248313   1.25898  1.269763  1.280637  1.291577   1.30256  1.313565  1.324573  1.335565  1.346526  1.357439   1.36829  1.379067  1.389758  1.400351  1.410835  1.421202  1.431443  1.441549 |  1.35909  1.359067  1.359038  1.359002  1.358956  1.358899  1.358829  1.358744  1.358642   1.35852  1.358378  1.358212  1.358022  1.357807  1.357565  1.357297  1.357003  1.356684  1.356341  1.355976   1.35559  1.355188   1.35477  1.354341  1.353902  1.353457  1.353007  1.352556  1.352107  1.351659  1.351217   1.35078  1.350351  1.349931   1.34952   1.34912   1.34873  1.348351  1.347984  1.347629  1.347285  1.346953  1.346633  1.346324  1.346026   1.34574  1.345464  1.345199  1.344944            1.3447 |   0.5367  0.536874  0.537097  0.537379  0.537732  0.538172  0.538711  0.539368  0.540159    0.5411  0.542208  0.543498  0.544985  0.546678  0.548586  0.550714  0.553063  0.555631   0.55841  0.561392  0.564563  0.567908   0.57141   0.57505  0.578808  0.582666  0.586603  0.590601  0.594642  0.598708  0.602785  0.606858  0.610915  0.614943  0.618934  0.622879  0.626769  0.630599  0.634364  0.638058  0.641679  0.645223  0.648689  0.652074  0.655377  0.658598  0.661736   0.66479  0.667762      0.670651 | -0.202583  -0.202809  -0.203098  -0.203461  -0.203915  -0.204476  -0.205162  -0.205992  -0.206985  -0.208161  -0.209537  -0.211129  -0.212953  -0.215017   -0.21733  -0.219894  -0.222707  -0.225765  -0.229057  -0.232571  -0.236291  -0.240197  -0.244269  -0.248487  -0.252828   -0.25727  -0.261792  -0.266373  -0.270994  -0.275638  -0.280287  -0.284927  -0.289544  -0.294128  -0.298666  -0.303151  -0.307575   -0.31193  -0.316212  -0.320415  -0.324537  -0.328573  -0.332522  -0.336381   -0.34015  -0.343827  -0.347412  -0.350903  -0.354303     -0.357609 | 0.665784  0.665807  0.665837  0.665874   0.66592  0.665977  0.666046   0.66613  0.666229  0.666347  0.666485  0.666644  0.666827  0.667033  0.667265  0.667523  0.667807  0.668117  0.668453  0.668814  0.669199  0.669606  0.670036  0.670485  0.670951  0.671435  0.671933  0.672443  0.672964  0.673495  0.674033  0.674577  0.675125  0.675676  0.676229  0.676782  0.677335  0.677886  0.678435   0.67898   0.67952  0.680056  0.680586   0.68111  0.681626  0.682136  0.682638  0.683131  0.683616      0.684091 |  100.33556 | 100.391192  100.517185  100.678558  100.883429  101.141185  101.462497  101.859289  102.344651  102.932694  103.638348  104.477107  105.464742  106.616977  107.949175  109.476017  111.211215  113.167266  115.355251  117.784697  120.463485  123.397829  126.592282  130.049806  133.771851  137.758473  142.008451  146.519422  151.288006  156.309926  161.580124  167.092864  172.841819   178.82015  185.020576  191.435424  198.056681  204.876025  211.884857  219.074326  226.435342  233.958592  241.634548   249.45347  257.405415  265.480231  273.667563  281.956847   290.33731  298.797965          307.327606 | 311.604896 | 0.721444  0.721683  0.721989  0.722376  0.722863  0.723469  0.724215  0.725125  0.726223  0.727533  0.729083  0.730894  0.732991  0.735395  0.738121  0.741184  0.744592  0.748352  0.752462  0.756919  0.761715  0.766837  0.772269  0.777994  0.783992  0.790241  0.796718    0.8034  0.810265  0.817289   0.82445  0.831728  0.839102  0.846552  0.854062  0.861613  0.869191  0.876781  0.884369  0.891943  0.899493  0.907006  0.914475  0.921889  0.929241  0.936522  0.943728   0.95085  0.957883       0.964821 | 2.595418  2.596801  2.598571  2.600816  2.603635  2.607143  2.611466  2.616738  2.623104  2.630713  2.639713   2.65025  2.662464  2.676478  2.692403  2.710327  2.730316  2.752411  2.776629   2.80296   2.83137  2.861806  2.894191  2.928437  2.964438  3.002082  3.041247  3.081809   3.12364  3.166614  3.210605  3.255493  3.301158  3.347488  3.394376   3.44172  3.489423  3.537394   3.58555  3.633809  3.682097  3.730344  3.778485  3.826457  3.874204   3.92167  3.968806  4.015562  4.061893        4.107756 |  100.33556  100.445596  100.587233  100.767965  100.996534  101.282967  101.638577  102.075902  102.608588  103.251219  104.019081  104.927891  105.993485  107.231496  108.657031  110.284361  112.126654  114.195747  116.501977  119.054065  121.859061  124.922338  128.247635  131.837129  135.691542  139.810256  144.191446  148.832204   153.72867  158.876148  164.269213  169.901811   175.76734  181.858722  188.168467  194.688719  201.411303   208.32775  215.429331   222.70707  230.151763  237.753987  245.504106  253.392279  261.408458  269.542389  277.783614  286.121465  294.545062  303.043311  311.604896 |  100.0 |    0.33556 | 311.604896 |
    | 08:30 | 3.723523  3.725379  3.727763  3.730799  3.734631   3.73942  3.745351  3.752623  3.761452  3.772064  3.784692  3.799568   3.81692   3.83696  3.859883  3.885855  3.915016  3.947467  3.983273  4.022462  4.065023   4.11091  4.160045  4.212318  4.267597  4.325729  4.386543  4.449861  4.515492  4.583246  4.652929  4.724348  4.797317  4.871651  4.947175  5.023718   5.10112  5.179225  5.257888  5.336971  5.416343  5.495881  5.575468  5.654996   5.73436  5.813462  5.892211  5.970517  6.048299             6.125476 | 125.175971  125.272927  125.397541  125.556311   125.75679  126.007623  126.318527  126.700232  127.164363  127.723275  128.389824  129.177111   130.09818   131.16571  132.391702  133.787191  135.361986  137.124459  139.081398   141.23791  143.597393  146.161562   148.93052  151.902874  155.075872  158.445565  162.006965  165.754215  169.680745  173.779415  178.042648  182.462543  187.030976   191.73968   196.58031  201.544502   206.62391  211.810241  217.095273  222.470875  227.929012  233.461753  239.061268  244.719829  250.429802  256.183642  261.973885  267.793137  273.634071  279.489407 |  52.972635    52.99156   53.015875   53.046838   53.085909   53.134754   53.195234   53.269394    53.35943   53.467652   53.596432   53.748144   53.925097   54.129468   54.363232   54.628105   54.925485   55.256418    55.62157    56.02122   56.455261   56.923222   57.424296   57.957381    58.52112   59.113948   59.734141   60.379854   61.049168   61.740123   62.450747   63.179086   63.923221   64.681287   65.451481   66.232075   67.021418    67.81794    68.62015    69.42664   70.236079   71.047209   71.858845   72.669869   73.479224   74.285914   75.088993   75.887568   76.680787        77.467841 |  52.235232    52.25379   52.277632   52.307994   52.346307   52.394203   52.453508   52.526228   52.614516   52.720636   52.846916   52.995681   53.169198     53.3696   53.598825   53.858554   54.150159   54.474665   54.832727   55.224616   55.650228   56.109101   56.600445   57.123178   57.675969   58.257285   58.865434   59.498608   60.154924   60.832461   61.529285   62.243481   62.973165    63.71651   64.471747   65.237183   66.011198   66.792251   67.578883   68.369712   69.163431    69.95881   70.754684   71.549959   72.343597   73.134621   73.922106   74.705172   75.482988     76.254758 | 1.089268  1.089556  1.089925  1.090396  1.090989   1.09173  1.092648  1.093773  1.095137  1.096775  1.098723  1.101014  1.103682  1.106759  1.110271  1.114242  1.118689  1.123624  1.129053  1.134975  1.141385  1.148269  1.155611   1.16339   1.17158  1.180153   1.18908  1.198329   1.20787   1.21767  1.227697  1.237923  1.248316   1.25885  1.269498  1.280234  1.291036   1.30188  1.312748  1.323619  1.334478  1.345306   1.35609  1.366815   1.37747  1.388041  1.398519  1.408893  1.419154  1.429292 | 1.359102  1.359083  1.359059  1.359028  1.358989   1.35894   1.35888  1.358806  1.358717  1.358611  1.358486  1.358339   1.35817  1.357976  1.357757  1.357513  1.357243  1.356947  1.356627  1.356284   1.35592  1.355537  1.355137  1.354723  1.354298  1.353864  1.353423  1.352979  1.352534   1.35209  1.351648  1.351211   1.35078  1.350357  1.349941  1.349535  1.349139  1.348754  1.348379  1.348016  1.347663  1.347323  1.346994  1.346676  1.346369  1.346073  1.345789  1.345515  1.345251          1.344997 | 0.536607  0.536753  0.536941  0.537179   0.53748  0.537855  0.538318  0.538885  0.539572  0.540393  0.541367  0.542508  0.543831  0.545348  0.547069  0.549002  0.551149  0.553513  0.556088  0.558869  0.561845  0.565004   0.56833  0.571806  0.575415  0.579138  0.582956   0.58685  0.590801  0.594793   0.59881  0.602837  0.606859  0.610866  0.614845  0.618788  0.622685   0.62653  0.630317   0.63404  0.637696  0.641279  0.644789  0.648221  0.651575   0.65485  0.658045  0.661158   0.66419      0.667141 | -0.202461  -0.202651  -0.202894  -0.203203  -0.203589  -0.204069   -0.20466  -0.205378  -0.206243  -0.207273  -0.208486  -0.209899  -0.211527  -0.213383  -0.215477  -0.217814  -0.220396   -0.22322  -0.226282   -0.22957  -0.233072  -0.236772  -0.240651   -0.24469  -0.248868  -0.253164  -0.257556  -0.262025  -0.266551  -0.271115    -0.2757  -0.280291  -0.284872  -0.289432  -0.293958  -0.298441  -0.302873  -0.307244   -0.31155  -0.315785  -0.319943  -0.324023  -0.328019   -0.33193  -0.335755  -0.339491  -0.343138  -0.346695  -0.350161     -0.353537 | 0.665771  0.665791  0.665816  0.665847  0.665887  0.665935  0.665995  0.666067  0.666154  0.666258  0.666379   0.66652  0.666683  0.666868  0.667077  0.667311  0.667571  0.667855  0.668166  0.668501  0.668861  0.669243  0.669648  0.670074  0.670518   0.67098  0.671458   0.67195  0.672454  0.672968  0.673492  0.674023  0.674559    0.6751  0.675644   0.67619  0.676736  0.677281  0.677825  0.678367  0.678905  0.679439  0.679969  0.680493  0.681011  0.681522  0.682026  0.682523  0.683012      0.683492 | 100.277186 | 100.323731  100.429356  100.565154  100.738239  100.956902  101.230652  101.570219  101.987504  102.495478  103.108023  103.839731  104.705647  105.720991  106.900853  108.259892  109.812043  111.570258  113.546281  115.750474  118.191697   120.87724    123.8128   127.00251  130.448997  134.153471  138.115831  142.334782  146.807957  151.532034  156.502852  161.715519  167.164498    172.8437  178.746549   184.86605  191.194836  197.725209   204.44918   211.35849  218.444636  225.698882  233.112274  240.675646  248.379627  256.214642  264.170914  272.238462  280.407098  288.666429          297.005849 | 301.190638 | 0.721315  0.721516  0.721774  0.722102  0.722515  0.723032  0.723672  0.724456  0.725407  0.726549  0.727906  0.729503  0.731362  0.733505  0.735952  0.738717  0.741813  0.745247  0.749025  0.753144  0.757601  0.762386  0.767487  0.772889  0.778575  0.784523  0.790715  0.797127  0.803738  0.810525  0.817467  0.824542   0.83173  0.839012  0.846369  0.853783   0.86124  0.868722  0.876217  0.883712  0.891194  0.898653  0.906078   0.91346   0.92079   0.92806  0.935264  0.942393  0.949442       0.956406 | 2.594677  2.595837  2.597328  2.599225   2.60162  2.604613  2.608318   2.61286  2.618373  2.624998  2.632878  2.642157  2.652974  2.665459  2.679731  2.695888  2.714013  2.734163  2.756373  2.780655  2.806993  2.835353  2.865678  2.897894  2.931913  2.967632  3.004942  3.043725  3.083862   3.12523  3.167706  3.211171  3.255507  3.300602  3.346346  3.392636  3.439374  3.486467  3.533828  3.581373  3.629027  3.676715  3.724371  3.771931  3.819335  3.866526  3.913451  3.960062   4.00631        4.052151 | 100.277186  100.369238  100.488165  100.640507  100.833948  101.077382  101.380932  101.755933  102.214854  102.771173  103.439187   104.23379  105.170195  106.263642  107.529087  108.980905  110.632609  112.496602  114.583979  116.904379  119.465889  122.275002  125.336623  128.654112  132.229361  136.062893  140.153977   144.50075  149.100336  153.948963  159.042076  164.374437  169.940212   175.73305   181.74615  187.972317  194.404009  201.033374  207.852282  214.852345   222.02494   229.36122  236.852123   244.48838  252.260516  260.158856   268.17352  276.294425  284.511282  292.813591  301.190638 |  100.0 |   0.277186 | 301.190638 |
    | 09:00 | 3.722536   3.72409  3.726094  3.728656  3.731903   3.73598  3.741051  3.747299  3.754924  3.764135  3.775156   3.78821  3.803522  3.821307  3.841767  3.865083   3.89141  3.920874  3.953565  3.989539  4.028815  4.071377  4.117175  4.166127  4.218126  4.273041  4.330722  4.391005  4.453714  4.518669  4.585683   4.65457  4.725147  4.797232  4.870649  4.945228  5.020805  5.097225  5.174339  5.252006  5.330093  5.408472  5.487024  5.565637  5.644204  5.722625  5.800803  5.878649  5.956078             6.033007 | 125.124403   125.20557  125.310284  125.444222   125.61404   125.82742  126.093082  126.420756  126.821106  127.305603  127.886345  128.575838  129.386737  130.331558  131.422386  132.670582  134.086522  135.679359  137.456841   139.42518  141.588978  143.951209  146.513254  149.274979  152.234842  155.390033  158.736618  162.269702  165.983576  169.871866  173.927671  178.143679   182.51228   187.02565  191.675833  196.454802  201.354506  206.366912  211.484032  216.697948  222.000818  227.384894   232.84252  238.366134  243.948266  249.581532  255.258631  260.972334  266.715476  272.480951 |  52.962566   52.978413    52.99885    53.02498   53.058092   53.099667   53.151385   53.215106   53.292857   53.386799   53.499186   53.632316   53.788468   53.969841   54.178492   54.416268   54.684754   54.985223   55.318607   55.685472   56.086013   56.520062   56.987108   57.486326   58.016617   58.576643   59.164876   59.779641   60.419153    61.08156   61.764971    62.46749   63.187234   63.922358   64.671067   65.431625   66.202368   66.981703   67.768115   68.560166   69.356491   70.155804   70.956885   71.758584   72.559814   73.359549   74.156817   74.950696   75.740313        76.524838 |  52.225358   52.240898   52.260938   52.286561   52.319029   52.359798   52.410511   52.472994   52.549235   52.641353   52.751558   52.882102   53.035222   53.213073   53.417672   53.650831   53.914102   54.208737   54.535647   54.895388    55.28815    55.71377   56.171746   56.661271   57.181263   57.730414   58.307224    58.91005   59.537144   60.186687   60.856827   61.545703   62.251471   62.972319   63.706488   64.452277   65.208053   65.972254   66.743394   67.520063   68.300925   69.084715    69.87024    70.65637   71.442042   72.226246   73.008031   73.786494   74.560777     75.330067 | 1.089115  1.089356  1.089667  1.090064  1.090567  1.091198  1.091983   1.09295  1.094128  1.095551  1.097252  1.099265  1.101622  1.104356  1.107496  1.111067   1.11509  1.119581   1.12455  1.130001  1.135934   1.14234  1.149207  1.156518  1.164252  1.172384  1.180888  1.189733  1.198891   1.20833  1.218021  1.227933  1.238037  1.248304  1.258709  1.269224  1.279827  1.290493  1.301204  1.311937  1.322676  1.333403  1.344103   1.35476  1.365363  1.375897  1.386353  1.396718  1.406983   1.41714 | 1.359112  1.359096  1.359076   1.35905  1.359016  1.358975  1.358923   1.35886  1.358783   1.35869   1.35858  1.358451    1.3583  1.358127   1.35793  1.357708  1.357461  1.357189  1.356892  1.356572   1.35623  1.355867  1.355485  1.355088  1.354678  1.354257  1.353827  1.353391  1.352953  1.352513  1.352074  1.351638  1.351206  1.350781  1.350362  1.349952   1.34955  1.349159  1.348777  1.348407  1.348047  1.347698   1.34736  1.347034  1.346718  1.346414   1.34612  1.345837  1.345564          1.345302 | 0.536529  0.536652  0.536809  0.537011  0.537266  0.537585  0.537982   0.53847  0.539064   0.53978  0.540632  0.541637   0.54281  0.544164   0.54571  0.547457  0.549412  0.551578  0.553954  0.556535  0.559316  0.562285  0.565431  0.568737  0.572189  0.575767  0.579455  0.583233  0.587084   0.59099  0.594935  0.598904  0.602881  0.606855  0.610812  0.614744  0.618639  0.622491  0.626292  0.630037   0.63372  0.637337  0.640884  0.644359   0.64776  0.651084  0.654331  0.657499  0.660588      0.663599 | -0.202359  -0.202518  -0.202723  -0.202984  -0.203313  -0.203723   -0.20423   -0.20485  -0.205601  -0.206499  -0.207565  -0.208814  -0.210262  -0.211925  -0.213812  -0.215932  -0.218291  -0.220889  -0.223722  -0.226786  -0.230068  -0.233557  -0.237236  -0.241088  -0.245094  -0.249232  -0.253484  -0.257829  -0.262246  -0.266718  -0.271226  -0.275755  -0.280288  -0.284812  -0.289315  -0.293787  -0.298216  -0.302595  -0.306916  -0.311174  -0.315362  -0.319477  -0.323515  -0.327473  -0.331347  -0.335137  -0.338842  -0.342459  -0.345988     -0.349429 | 0.665761  0.665777  0.665798  0.665825  0.665858    0.6659  0.665951  0.666014  0.666089  0.666179  0.666286  0.666411  0.666555  0.666721  0.666909  0.667121  0.667357  0.667617  0.667903  0.668213  0.668548  0.668906  0.669286  0.669688   0.67011   0.67055  0.671007   0.67148  0.671966  0.672464  0.672972  0.673488  0.674012  0.674542  0.675075  0.675612   0.67615  0.676689  0.677228  0.677765    0.6783  0.678832   0.67936  0.679883  0.680401  0.680914   0.68142  0.681919  0.682411      0.682895 | 100.228681 | 100.267562  100.355973  100.470061  100.616036  100.801192  101.033965  101.323961  101.681935  102.119736  102.650184  103.286914  104.044159  104.936513  105.978646  107.185027  108.569636  110.145698  111.925444  113.919909  116.138787  118.590321   121.28125  124.216804  127.400729  130.835351  134.521663  138.459427  142.647285  147.082872  151.762928  156.683406  161.839567  167.226071   172.83705   178.66618  184.706734  190.951633  197.393482  204.024603   210.83706  217.822678  224.973057  232.279584  239.733438  247.325596  255.046836  262.887735  270.838668   278.88981          287.031127 | 291.119246 | 0.721209  0.721377  0.721593   0.72187  0.722221  0.722661  0.723208  0.723882  0.724704  0.725696  0.726882  0.728284  0.729927  0.731832  0.734019  0.736506  0.739307  0.742433  0.745892  0.749685  0.753811  0.758265  0.763038  0.768117  0.773488  0.779133  0.785033  0.791168  0.797516  0.804057  0.810768   0.81763  0.824621  0.831722  0.838914   0.84618  0.853502  0.860865  0.868255  0.875658  0.883062  0.890454  0.897824  0.905163   0.91246  0.919708  0.926899  0.934025  0.941081       0.948059 |  2.59406  2.595031  2.596284  2.597886  2.599915  2.602463  2.605632  2.609535  2.614297  2.620049  2.626928  2.635073  2.644622  2.655708  2.668453  2.682967  2.699342  2.717652  2.737947  2.760258  2.784588  2.810922  2.839221  2.869429  2.901471   2.93526  2.970698  3.007677  3.046084  3.085803  3.126716  3.168706  3.211657  3.255456  3.299994  3.345167  3.390876  3.437024  3.483523  3.530289   3.57724  3.624304  3.671408  3.718488  3.765481  3.812329  3.858977  3.905374  3.951471        3.997221 | 100.228681  100.305569  100.405268  100.533462  100.696881  100.903377  101.161969  101.482845  101.877332  102.357801  102.937529  103.630515  104.451242   105.41442    106.5347  107.826382   109.30314  110.977764  112.861935  114.966056  117.299117  119.868619  122.680544  125.739368  129.048106  132.608393  136.420576  140.483828  144.796255  149.355015  154.156427  159.196072  164.468888  169.969251  175.691049   181.62774  187.772412  194.117816  200.656408  207.380376  214.281658  221.351964  228.582785  235.965403  243.490893  251.150134  258.933802  266.832374  274.836126  282.935129  291.119246 |  100.0 |   0.228681 | 291.119246 |
    | 09:30 | 3.721715  3.723014  3.724695  3.726854    3.7296  3.733063   3.73739  3.742746  3.749314  3.757289   3.76688  3.778301  3.791772  3.807505  3.825706  3.846566  3.870253  3.896911  3.926654  3.959561   3.99568  4.035023  4.077568  4.123261   4.17202  4.223738  4.278286  4.335517  4.395272  4.457382   4.52167  4.587959  4.656069  4.725823  4.797045  4.869568  4.943226  5.017862  5.093327  5.169475  5.246172  5.323287  5.400698   5.47829  5.555952  5.633582  5.711082  5.788357  5.865322             5.941891 | 125.081522  125.149366  125.237213  125.350008  125.493588  125.674751  125.901281   126.18195  126.526476  126.945429    127.4501  128.052319  128.764229  129.598037  130.565739  131.678843  132.948098  134.383253  135.992845   137.78404  139.762517  141.932415  144.296322  146.855322  149.609067  152.555889  155.692932  159.016291   162.52116  166.201977   170.05256  174.066234  178.235941  182.554343  187.013906  191.606968  196.325799  201.162648  206.109776  211.159489   216.30415  221.536199  226.848156  232.232629  237.682309  243.189976  248.748484  254.350764  259.989813  265.658686 |  52.954192    52.96744    52.98459   53.006601   53.034608   53.069924   53.114051   53.168674   53.235648   53.316978   53.414784   53.531263   53.668634   53.829083   54.014702   54.227431   54.468993   54.740853   55.044166   55.379759   55.748105   56.149325   56.583197   57.049177   57.546425   58.073846   58.630125   59.213771   59.823155   60.456551   61.112169   61.788187   62.482776   63.194125   63.920457   64.660042    65.41121   66.172356   66.941943   67.718509   68.500662   69.287084   70.076528   70.867811   71.659818   72.451491   73.241832   74.029894   74.814779        75.595633 |  52.217147   52.230138   52.246955   52.268539   52.296001   52.330632   52.373902   52.427464   52.493138   52.572888   52.668795   52.783012   52.917715   53.075048   53.257063   53.465661   53.702532   53.969112   54.266536   54.595612   54.956804   55.350233    55.77568    56.23261   56.720202   57.237381   57.782858    58.35517    58.95272   59.573816   60.216702   60.879592   61.560692   62.258227   62.970454   63.695678   64.432259   65.178623   65.933266   66.694751   67.461715   68.232866   69.006979   69.782896   70.559522   71.335822   72.110815   72.883573   73.653216     74.418907 | 1.088988  1.089189   1.08945  1.089784   1.09021  1.090746  1.091416  1.092245  1.093261  1.094494  1.095975  1.097737  1.099813  1.102235  1.105032  1.108232  1.111858  1.115929   1.12046   1.12546   1.13093   1.13687   1.14327  1.150118  1.157396  1.165085  1.173159  1.181593  1.190358  1.199425  1.208765  1.218349  1.228148  1.238133  1.248278  1.258556  1.268943  1.279415  1.289951   1.30053  1.311133  1.321742  1.332341  1.342915   1.35345  1.363932   1.37435  1.384691  1.394946  1.405106 | 1.359121  1.359107   1.35909  1.359068   1.35904  1.359005   1.35896  1.358906   1.35884  1.358759  1.358663  1.358549  1.358416  1.358261  1.358085  1.357884  1.357659   1.35741  1.357136  1.356838  1.356518  1.356176  1.355815  1.355436  1.355041  1.354634  1.354217  1.353792  1.353361  1.352928  1.352493  1.352059  1.351629  1.351202  1.350782  1.350368  1.349963  1.349566  1.349178  1.348801  1.348434  1.348078  1.347732  1.347397  1.347073   1.34676  1.346458  1.346166  1.345885          1.345614 | 0.536465  0.536567  0.536699  0.536869  0.537085  0.537357  0.537696  0.538115  0.538628  0.539248  0.539992  0.540875  0.541911  0.543114  0.544497  0.546071  0.547843  0.549819     0.552  0.554386  0.556973  0.559752  0.562714  0.565845  0.569132  0.572558  0.576106  0.579758  0.583498  0.587307  0.591169  0.595068  0.598989  0.602919  0.606845  0.610755  0.614639  0.618489  0.622297  0.626055  0.629759  0.633402  0.636981  0.640493  0.643935  0.647304  0.650599  0.653818  0.656961      0.660027 | -0.202274  -0.202408   -0.20258    -0.2028  -0.203079  -0.203428  -0.203862  -0.204396  -0.205046  -0.205828  -0.206761  -0.207861  -0.209144  -0.210627  -0.212321  -0.214237  -0.216383  -0.218761  -0.221373  -0.224214  -0.227277  -0.230553  -0.234028  -0.237686   -0.24151  -0.245482  -0.249582   -0.25379  -0.258087  -0.262454  -0.266873  -0.271327  -0.275801  -0.280278  -0.284747  -0.289196  -0.293613   -0.29799  -0.302318   -0.30659  -0.310801  -0.314945  -0.319017  -0.323014  -0.326933  -0.330772  -0.334529  -0.338202   -0.34179     -0.345292 | 0.665752  0.665766  0.665783  0.665806  0.665834   0.66587  0.665914  0.665968  0.666033  0.666112  0.666205  0.666315  0.666443   0.66659  0.666759   0.66695  0.667163  0.667401  0.667663  0.667949  0.668259  0.668593  0.668949  0.669328  0.669727  0.670145  0.670581  0.671033    0.6715   0.67198  0.672472  0.672974  0.673484  0.674001  0.674523   0.67505   0.67558  0.676111  0.676644  0.677175  0.677706  0.678234   0.67876  0.679281  0.679799  0.680311  0.680818  0.681319  0.681813        0.6823 | 100.188431 | 100.220863  100.294751  100.390446  100.513349  100.669852  100.867406   101.11457  101.421015  101.797498  102.255784   102.80852   103.46907  104.251303  105.169351  106.237345  107.469147  108.878086  110.476705  112.276556  114.288011  116.520136  118.980604  121.675649  124.610069  127.787256  131.209265  134.876896  138.789789  142.946532   147.34477  151.981306  156.852202  161.952868  167.278149   172.82239  178.579503  184.543018  190.706126  197.061719  203.602413  210.320578  217.208351   224.25765  231.460185  238.807463  246.290792  253.901286  261.629858  269.467226          277.403908 | 281.391872 |  0.72112   0.72126  0.721442  0.721675  0.721972  0.722346  0.722813  0.723391  0.724099  0.724959  0.725991   0.72722  0.728666  0.730354  0.732302  0.734531  0.737057  0.739892  0.743046  0.746525  0.750331  0.754462  0.758911   0.76367  0.768727  0.774066  0.779671  0.785522  0.791601  0.797886  0.804358  0.810996  0.817779  0.824688  0.831704  0.838808  0.845985  0.853218  0.860491   0.86779  0.875104  0.882418  0.889722  0.897006   0.90426  0.911476  0.918643  0.925756  0.932808       0.939791 | 2.593546  2.594359   2.59541  2.596759  2.598476   2.60064  2.603344  2.606691  2.610793  2.615774  2.621762  2.628891  2.637295  2.647105  2.658449  2.671441  2.686184  2.702762  2.721242  2.741668  2.764064  2.788432   2.81475  2.842979  2.873063  2.904926  2.938485  2.973641  3.010292  3.048328  3.087637  3.128105  3.169619  3.212067  3.255342  3.299339  3.343956  3.389097  3.434671  3.480592  3.526778  3.573151  3.619639  3.666173  3.712691  3.759131  3.805437  3.851555  3.897434        3.943028 | 100.188431  100.252558  100.336007  100.443703  100.581519  100.756361  100.976224  101.250223  101.588587  102.002603  102.504516  103.107384  103.824886   104.67109  105.660204  106.806298   108.12304  109.623431  111.319572  113.222471  115.341879  117.686182  120.262341  123.075868  126.130844  129.429975  132.974661  136.765095  140.800363  145.078551  149.596856  154.351683  159.338746  164.553152  169.989475  175.641829  181.503922  187.569104  193.830406  200.280575    206.9121  213.717228  220.687985  227.816181  235.093425  242.511122  250.060483  257.732522  265.518055    273.4077  281.391872 |  100.0 |   0.188431 | 281.391872 |
    | 10:00 | 3.721033  3.722117  3.723526  3.725341   3.72766  3.730595   3.73428  3.738861  3.744504  3.751391  3.759715   3.76968  3.781495   3.79537   3.81151  3.830111  3.851351  3.875388  3.902354  3.932353  3.965455  4.001698  4.041088  4.083599  4.129174   4.17773   4.22916  4.283338   4.34012  4.399353  4.460871  4.524505  4.590084  4.657433  4.726382  4.796763  4.868413  4.941174  5.014893  5.089426  5.164633  5.240384  5.316553   5.39302  5.469675   5.54641  5.623126  5.699725  5.776119             5.852221 | 125.045916  125.102538  125.176118  125.270945  125.392125  125.545644  125.738417  125.978312  126.274127   126.63554  127.073008  127.597624   128.22093  128.954703  129.810703  130.800415  131.934789  133.223988  134.677168  136.302287  138.105967  140.093397  142.268288  144.632882   147.18799  149.933079  152.866372   155.98498  159.285031  162.761815  166.409911  170.223321  174.195587  178.319894  182.589165  186.996136  191.533425  196.193586  200.969149  205.852658  210.836693  215.913889   221.07695  226.318654  231.631857  237.009497  242.444587  247.930211  253.459524  259.025737 |  52.947237   52.958296   52.972663   52.991173   53.014818   53.044758    53.08233   53.129048   53.186601   53.256832    53.34172   53.443341    53.56383   53.705329    53.86993   54.059621   54.276228   54.521359   54.796363   55.102288    55.43986   55.809471   56.211174   56.644701   57.109476   57.604652   58.129139   58.681646   59.260717   59.864771   60.492136   61.141082   61.809851   62.496682   63.199828   63.917579   64.648267   65.390281   66.142074   66.902162   67.669133   68.441642   69.218411   69.998233    70.77996    71.56251   72.344854    73.12602   73.905085        74.681174 |  52.210328   52.221171    52.23526    52.25341   52.276596   52.305954   52.342797   52.388608   52.445043   52.513911    52.59715   52.696798   52.814947   52.953697   53.115102    53.30111    53.51351   53.753881   54.023544   54.323528   54.654545   55.016979   55.410881   55.835989   56.291738   56.777298     57.2916   57.833378   58.401204   58.993528   59.608709   60.245053   60.900835   61.574328    62.26382   62.967632   63.684131   64.411736   65.148929   65.894257   66.646334   67.403841   68.165526   68.930204   69.696751   70.464104   71.231255   71.997252   72.761188     73.522206 | 1.088882   1.08905  1.089269   1.08955  1.089909  1.090364  1.090935  1.091644  1.092517  1.093582  1.094869  1.096407   1.09823  1.100368  1.102851  1.105708  1.108965  1.112643  1.116759  1.121327  1.126353   1.13184  1.137783  1.144175  1.151002  1.158247  1.165889  1.173905  1.182269  1.190955  1.199933  1.209176  1.218656  1.228344  1.238213  1.248238  1.258393  1.268654  1.278999  1.289407  1.299858  1.310334  1.320818  1.331292  1.341744  1.352158  1.362523  1.372826  1.383056  1.393204 | 1.359128  1.359116  1.359102  1.359083   1.35906   1.35903  1.358992  1.358945  1.358888  1.358819  1.358735  1.358635  1.358517   1.35838  1.358222  1.358042  1.357839  1.357611   1.35736  1.357084  1.356786  1.356465  1.356124  1.355764  1.355387  1.354996  1.354592  1.354179  1.353758  1.353332  1.352904  1.352474  1.352046   1.35162  1.351199  1.350783  1.350375  1.349974  1.349581  1.349198  1.348825  1.348461  1.348108  1.347766  1.347434  1.347112  1.346802  1.346502  1.346212          1.345932 | 0.536411  0.536496  0.536607   0.53675  0.536933  0.537163  0.537452  0.537811  0.538252  0.538789  0.539437  0.540209  0.541121  0.542187  0.543419   0.54483   0.54643  0.548225   0.55022  0.552415  0.554811    0.5574  0.560177   0.56313  0.566247  0.569513  0.572913  0.576431  0.580049   0.58375  0.587518  0.591337  0.595192  0.599067   0.60295  0.606829  0.610693  0.614532  0.618338  0.622102  0.625819  0.629482  0.633087   0.63663  0.640107  0.643516  0.646854   0.65012  0.653313       0.65643 | -0.202203  -0.202315   -0.20246  -0.202645  -0.202881  -0.203178  -0.203549  -0.204007  -0.204568  -0.205247   -0.20606  -0.207026  -0.208159  -0.209477  -0.210992  -0.212716  -0.214659  -0.216828  -0.219224  -0.221847  -0.224694  -0.227757  -0.231024  -0.234484  -0.238121  -0.241917  -0.245855  -0.249917  -0.254082  -0.258333  -0.262651  -0.267018   -0.27142  -0.275839  -0.280262  -0.284678  -0.289073  -0.293438  -0.297764  -0.302042  -0.306267  -0.310432  -0.314532  -0.318562   -0.32252  -0.326402  -0.330205  -0.333929  -0.337571      -0.34113 | 0.665745  0.665756  0.665771   0.66579  0.665814  0.665844  0.665882  0.665928  0.665985  0.666053  0.666134  0.666231  0.666344  0.666475  0.666625  0.666796   0.66699  0.667206  0.667445  0.667708  0.667994  0.668304  0.668637  0.668992  0.669368  0.669763  0.670178   0.67061  0.671058  0.671519  0.671994   0.67248  0.672975  0.673479  0.673989  0.674505  0.675025  0.675548  0.676073  0.676598  0.677123  0.677647  0.678169  0.678689  0.679204  0.679716  0.680223  0.680724   0.68122      0.681709 | 100.155078 |  100.18209  100.243749   100.32389  100.427197  100.559251  100.726608  100.936857  101.198653  101.521708  101.916752  102.395443  102.970234  103.654204  104.460843  105.403825   106.49675   107.75289  109.184946  110.804817  112.623408  114.650472   116.89449  119.362604  122.060582  124.992825  128.162405   131.57113  135.219625  139.107423  143.233071  147.594225  152.187751   157.00982  162.055984  167.321265  172.800208  178.486951  184.375265  190.458601  196.730117  203.182714  209.809048  216.601552  223.552446  230.653747  237.897275  245.274652  252.777311  260.396487          268.123224 | 272.008178 | 0.721046  0.721163  0.721316  0.721512  0.721762   0.72208  0.722477  0.722972  0.723581  0.724323   0.72522  0.726293  0.727563  0.729053  0.730783  0.732773  0.735042  0.737603   0.74047  0.743649  0.747147  0.750964  0.755097   0.75954  0.764285  0.769318  0.774624  0.780188  0.785991  0.792015  0.798239  0.804643  0.811208  0.817914  0.824743  0.831676  0.838696  0.845786  0.852931  0.860116  0.867327  0.874553  0.881781     0.889    0.8962  0.903371  0.910506  0.917595  0.924632        0.93161 |  2.59312  2.593798  2.594679  2.595813  2.597263  2.599098    2.6014  2.604263  2.607789  2.612091  2.617289   2.62351  2.630883  2.639539  2.649602  2.661193   2.67442  2.689378  2.706145  2.724781  2.745324  2.767793  2.792185  2.818478   2.84663  2.876583  2.908264   2.94159  2.976467  3.012793  3.050463  3.089369    3.1294  3.170449  3.212408  3.255171  3.298639  3.342714  3.387302  3.432317  3.477674  3.523295  3.569104  3.615031   3.66101  3.706979  3.752879  3.798655  3.844254        3.889628 | 100.155078  100.208484  100.278224  100.368554  100.484583  100.632359   100.81894  101.052445   101.34206  101.698026  102.131564  102.654771  103.280465  104.021993  104.893004  105.907204  107.078101  108.418747  109.941501  111.657816  113.578058  115.711372  118.065585  120.647155  123.461161  126.511328  129.800074  133.328591  137.096933  141.104113  145.348203   149.82644  154.535318  159.470674  164.627779  170.001396  175.585855  181.375095  187.362719  193.542023  199.906029   206.44751  213.159007  220.032844  227.061137  234.235804  241.548563  248.990943  256.554278  264.229708  272.008178 |  100.0 |   0.155078 | 272.008178 |
    | 10:30 | 3.720467  3.721371  3.722549  3.724073  3.726027  3.728511  3.731642  3.735552   3.74039  3.746323  3.753528  3.762198  3.772532  3.784731     3.799  3.815533  3.834516  3.856117  3.880484  3.907737  3.937969  3.971244  4.007592  4.047011  4.089473  4.134918  4.183261  4.234399  4.288205  4.344541  4.403255  4.464191  4.527182  4.592063  4.658668  4.726831  4.796391   4.86719  4.939075    5.0119  5.085525  5.159816  5.234644   5.30989  5.385437  5.461178  5.537008   5.61283   5.68855             5.764082 | 125.016392  125.063577  125.125111  125.204704  125.306801  125.436656  125.600384  125.805005  126.058443  126.369501  126.747797  127.203652  127.747943  128.391918  129.146979  130.024444    131.0353  132.189957  133.498018  134.968078  136.607553  138.422557  140.417825   142.59668  144.961043  147.511484  150.247305  153.166638   156.26657   159.54327  162.992121  166.607843  170.384615  174.316187  178.395976  182.617149  186.972704  191.455522  196.058427  200.774216  205.595699  210.515719  215.527169  220.623004  225.796246  231.039993  236.347411  241.711741  247.126287  252.584415 |   52.94147   52.950687   52.962704   52.978244   52.998171   53.023504    53.05543   53.095301   53.144644   53.205144   53.278628   53.367045   53.472424   53.596836   53.742345   53.910952   54.104543   54.324834   54.573323   54.851251   55.159565   55.498902   55.869575   56.271581   56.704605   57.168052   57.661064   58.182562   58.731279   59.305794    59.90457   60.525989   61.168379   61.830043   62.509281   63.204412   63.913784    64.63579   65.368878   66.111553   66.862383   67.620001   68.383107   69.150465   69.920902   70.693306   71.466624   72.239858   73.012059         73.78233 |  52.204673    52.21371   52.225494   52.240732   52.260272   52.285114   52.316419   52.355516   52.403901   52.463225   52.535283   52.621983   52.725315   52.847312   52.989995   53.155328   53.345159   53.561172   53.804836   54.077366   54.379693    54.71244   55.075916   55.470114    55.89473   56.349176   56.832615   57.343986   57.882047   58.445405   59.032554   59.641905    60.27182   60.920635   61.586683   62.268314   62.963911   63.671897   64.390749   65.119001    65.85525   66.598156   67.346443   68.098899   68.854375    69.61178    70.37008   71.128298   71.885504     72.640817 | 1.088794  1.088934  1.089117  1.089353  1.089656  1.090041  1.090526  1.091132  1.091881  1.092798  1.093913  1.095252  1.096847  1.098729  1.100926  1.103469  1.106384  1.109695  1.113421   1.11758  1.122181  1.127231  1.132731  1.138675  1.145056  1.151861   1.15907  1.166666  1.174623  1.182919  1.191525  1.200416  1.209564  1.218941  1.228521  1.238277  1.248185   1.25822  1.268359   1.27858  1.288864   1.29919  1.309542  1.319903  1.330256  1.340588  1.350885  1.361135  1.371326  1.381447 | 1.359133  1.359124  1.359112  1.359096  1.359076  1.359051  1.359019  1.358979   1.35893   1.35887  1.358797   1.35871  1.358607  1.358485  1.358345  1.358183     1.358  1.357793  1.357563   1.35731  1.357033  1.356734  1.356414  1.356073  1.355715   1.35534  1.354952  1.354552  1.354142  1.353726  1.353305  1.352881  1.352456  1.352033  1.351612  1.351196  1.350786  1.350382  1.349985  1.349597  1.349218  1.348848  1.348488  1.348139  1.347799   1.34747  1.347151  1.346843  1.346545          1.346257 | 0.536366  0.536438   0.53653   0.53665  0.536804  0.536999  0.537245  0.537552  0.537931  0.538394  0.538956   0.53963  0.540429   0.54137  0.542464  0.543725  0.545163  0.546787  0.548603  0.550615  0.552824  0.555227  0.557818  0.560591  0.563534  0.566636  0.569882  0.573256  0.576744  0.580328  0.583992   0.58772  0.591496  0.595307  0.599137  0.602975  0.606809  0.610628  0.614423  0.618185  0.621907  0.625583  0.629207  0.632775  0.636282  0.639725  0.643102   0.64641  0.649648      0.652814 | -0.202144  -0.202238  -0.202359  -0.202515  -0.202714  -0.202967  -0.203283  -0.203675  -0.204157  -0.204745  -0.205453  -0.206297  -0.207295  -0.208461  -0.209811  -0.211356  -0.213109  -0.215078  -0.217267  -0.219679  -0.222313  -0.225164  -0.228224  -0.231482  -0.234926  -0.238541   -0.24231  -0.246215  -0.250238  -0.254361  -0.258566  -0.262836  -0.267154  -0.271503   -0.27587  -0.280241  -0.284604  -0.288947  -0.293261  -0.297537  -0.301768  -0.305946  -0.310066  -0.314123  -0.318112  -0.322031  -0.325877  -0.329646  -0.333337     -0.336949 | 0.665738  0.665748  0.665761  0.665777  0.665797  0.665823  0.665855  0.665895  0.665943  0.666002  0.666073  0.666157  0.666257  0.666373  0.666507   0.66666  0.666834  0.667029  0.667247  0.667488  0.667752  0.668038  0.668348  0.668679  0.669032  0.669406  0.669799   0.67021  0.670638  0.671081  0.671537  0.672007  0.672487  0.672976  0.673473  0.673977  0.674487     0.675  0.675516  0.676034  0.676553  0.677072   0.67759  0.678105  0.678619  0.679128  0.679634  0.680136  0.680632      0.681122 | 100.127477 | 100.149943  100.201322  100.268334   100.35503  100.466266  100.607788  100.786299  101.009507  101.286137   101.62592  102.039534  102.538507   103.13508  103.842031  104.672472  105.639616  106.756544  108.035956  109.489951  111.129814  112.965843  115.007204  117.261837   119.73639  122.436201  125.365307  128.526487  131.921325  135.550287  139.412816  143.507422  147.831779  152.382816  157.156807  162.149445  167.355916   172.77096  178.388925  184.203814   190.20932   196.39886  202.765599  209.302466  216.002176  222.857236  229.859955  237.002447  244.276637   251.67426           259.18686 | 262.966516 | 0.720985  0.721083   0.72121  0.721375  0.721586  0.721854  0.722193  0.722615  0.723137  0.723777  0.724554  0.725487  0.726599  0.727911  0.729442  0.731214  0.733244   0.73555  0.738145  0.741041  0.744243  0.747757  0.751583  0.755717  0.760153  0.764881   0.76989  0.775164  0.780687  0.786442  0.792411  0.798573  0.804911  0.811406  0.818037  0.824787  0.831639  0.838576  0.845582  0.852641  0.859741  0.866867  0.874007   0.88115  0.888285  0.895403  0.902494   0.90955  0.916563       0.923526 | 2.592766  2.593331  2.594068  2.595021  2.596242  2.597795  2.599752  2.602195  2.605219  2.608925  2.613426   2.61884   2.62529  2.632903  2.641802   2.65211  2.663937  2.677387  2.692547   2.70949  2.728268  2.748915  2.771445  2.795851  2.822108  2.850175  2.879992  2.911489  2.944582  2.979179  3.015184  3.052494  3.091004  3.130608  3.171202  3.212681  3.254945  3.297897  3.341443  3.385494  3.429963  3.474771   3.51984  3.565099  3.610479  3.655916  3.701349  3.746722   3.79198        3.837073 | 100.127477   100.17189  100.230087  100.305734  100.403259  100.527942   100.68599    100.8846  101.131985  101.437384  101.811021  102.264029  102.808337  103.456505  104.221539  105.116666  106.155099  107.349795  108.713216  110.257105  111.992301  113.928568  116.074484  118.437353  121.023172   123.83662  126.881092  130.158749  133.670594  137.416556  141.395584  145.605747   150.04432  154.707884  159.592399  164.693287  170.005496  175.523557  181.241633  187.153562  193.252893  199.532909  205.986652  212.606942  219.386387  226.317391  233.392167  240.602733  247.940918  255.398363  262.966516 |  100.0 |   0.127477 | 262.966516 |
    | 11:00 | 3.719999  3.720751  3.721735  3.723012  3.724656  3.726754  3.729409  3.732739  3.736878  3.741977  3.748199  3.755722  3.764735   3.77543  3.788006  3.802656  3.819567  3.838916   3.86086  3.885536  3.913055  3.943501  3.976929  4.013362  4.052795  4.095194  4.140497  4.188621   4.23946  4.292894  4.348785  4.406988  4.467348  4.529708  4.593906  4.659782  4.727177  4.795934  4.865902  4.936934  5.008887  5.081626  5.155023  5.228952  5.303298  5.377947  5.452796  5.527742  5.602691             5.677553 | 124.991944  125.031209  125.082591   125.14929  125.235166  125.344808  125.483607  125.657793  125.874467  126.141592  126.467954   126.86309  127.337166  127.900831   128.56503  129.340787  130.238986  131.270127    132.4441  133.769976  135.255818  136.908533  138.733763  140.735824  142.917681  145.280966  147.826031  150.552031  153.457023  156.538081  159.791418   163.21251  166.796215  170.536884  174.428465  178.464594  182.638676  186.943951  191.373556  195.920567  200.578036  205.339028  210.196632  215.143989  220.174291  225.280796  230.456827  235.695776  240.991096  246.336302 |  52.936694   52.944365   52.954401   52.967425    52.98419   53.005587   53.032661   53.066619    53.10883   53.160823   53.224278   53.301004   53.392913   53.501987   53.630232   53.779631   53.952097   54.149416   54.373201   54.624845   54.905487    55.21598   55.556878   55.928426   56.330565   56.762946    57.22495   57.715718   58.234182   58.779096   59.349077   59.942631    60.55819   61.194137   61.848835   62.520642   63.207937   63.909126   64.622661   65.347041   66.080823   66.822624   67.571122   68.325059   69.083236   69.844517   70.607821   71.372127   72.136461        72.899902 |  52.199989   52.207511   52.217352   52.230124   52.246563   52.267544   52.294093   52.327391   52.368782   52.419766   52.481988   52.557224   52.647349   52.754304   52.880059   53.026557   53.195674   53.389161     53.6086   53.855357   54.130549   54.435013   54.769291   55.133624   55.527954   55.951938    56.40497   56.886208   57.394603   57.928936   58.487848   59.069876   59.673481   60.297079   60.939062   61.597824   62.271771   62.959344   63.659022   64.369335   65.088868   65.816264   66.550227   67.289522   68.032976   68.779473   69.527955   70.277418    71.02691     71.775525 | 1.088722  1.088838  1.088991  1.089189  1.089444  1.089769   1.09018  1.090696  1.091337  1.092126  1.093089  1.094252  1.095644  1.097295  1.099233  1.101489  1.104089  1.107059  1.110421  1.114193   1.11839  1.123022  1.128092  1.133602  1.139546  1.145914  1.152694  1.159868  1.167416  1.175315  1.183542  1.192071  1.200876   1.20993  1.219207  1.228681  1.238327   1.24812  1.258038  1.268057  1.278158   1.28832  1.298526  1.308757  1.318997  1.329232  1.339447  1.349629  1.359767  1.369848 | 1.359138  1.359131   1.35912  1.359107   1.35909  1.359069  1.359042  1.359008  1.358966  1.358914  1.358851  1.358775  1.358684  1.358578  1.358453  1.358309  1.358144  1.357957  1.357748  1.357516  1.357261  1.356983  1.356684  1.356363  1.356024  1.355667  1.355295  1.354909  1.354513  1.354107  1.353695  1.353278  1.352859   1.35244  1.352021  1.351606  1.351194  1.350788  1.350389  1.349997  1.349613  1.349238  1.348872  1.348515  1.348169  1.347832  1.347506   1.34719  1.346884          1.346587 |  0.53633  0.536389  0.536466  0.536567  0.536696  0.536861   0.53707  0.537331  0.537656  0.538055  0.538541  0.539127  0.539826  0.540653  0.541622  0.542744  0.544032  0.545495  0.547142  0.548978  0.551006  0.553225  0.555634  0.558227  0.560995  0.563927  0.567013  0.570238  0.573587  0.577045  0.580595  0.584222  0.587911  0.591646  0.595414  0.599201  0.602994  0.606784  0.610559  0.614311  0.618031  0.621712  0.625348  0.628934  0.632465  0.635937  0.639348  0.642693  0.645972      0.649182 | -0.202095  -0.202173  -0.202275  -0.202406  -0.202574  -0.202788  -0.203056  -0.203392  -0.203806  -0.204313  -0.204927  -0.205663  -0.206538  -0.207567  -0.208765  -0.210145  -0.211721  -0.213501  -0.215493  -0.217701  -0.220128   -0.22277  -0.225623  -0.228679  -0.231927  -0.235355  -0.238947  -0.242688   -0.24656  -0.250546  -0.254627  -0.258788  -0.263011  -0.267279  -0.271578  -0.275894  -0.280214  -0.284526  -0.288818  -0.293083   -0.29731  -0.301494  -0.305627  -0.309703  -0.313718  -0.317668  -0.321549  -0.325359  -0.329094     -0.332754 | 0.665733  0.665741  0.665752  0.665765  0.665783  0.665804  0.665832  0.665866  0.665908  0.665959   0.66602  0.666094  0.666181  0.666283  0.666402  0.666539  0.666695  0.666871  0.667069  0.667288   0.66753  0.667794  0.668081   0.66839  0.668721  0.669072  0.669443  0.669833   0.67024  0.670664  0.671102  0.671554  0.672018  0.672492  0.672976  0.673467  0.673965  0.674468  0.674975  0.675485  0.675996  0.676509  0.677021  0.677533  0.678042   0.67855  0.679054  0.679554   0.68005      0.680541 | 100.104666 | 100.123325  100.166076  100.222027  100.294669  100.388215  100.507683   100.65897  100.848911  101.085306  101.376936  101.733524  102.165674  102.684765  103.302809  104.032278  104.885896  105.876424  107.016433  108.318071  109.792861  111.451505  113.303724  115.358141  117.622184  120.102048   122.80267  125.727754   128.87981  132.260218  135.869306  139.706433  143.770082  148.057948  152.567024  157.293682   162.23375  167.382572  172.735072    178.2858  184.028978  189.958529  196.068114  202.351148  208.800823  215.410119  222.171815  229.078499  236.122567  243.296229           250.59151 | 254.264101 | 0.720934  0.721016  0.721122   0.72126  0.721438  0.721665  0.721951  0.722311  0.722758  0.723308  0.723979   0.72479  0.725761  0.726911  0.728262  0.729834  0.731646  0.733714  0.736056  0.738683  0.741605  0.744828  0.748357  0.752189  0.756322  0.760749   0.76546  0.770443  0.775684  0.781167  0.786874  0.792789  0.798892  0.805165   0.81159  0.818148  0.824822  0.831595   0.83845  0.845374   0.85235  0.859366  0.866408  0.873465  0.880526   0.88758  0.894617   0.90163  0.908609       0.915547 | 2.592474  2.592944  2.593559  2.594358  2.595385  2.596697  2.598356  2.600438  2.603024   2.60621  2.610097  2.614796  2.620423  2.627099  2.634946  2.644082  2.654624  2.666678  2.680339  2.695689  2.712794  2.731702   2.75244  2.775019  2.799429  2.825644  2.853618  2.883295  2.914604  2.947463  2.981784  3.017471  3.054426  3.092547  3.131732   3.17188  3.212891  3.254668  3.297117  3.340147  3.383672   3.42761  3.471882  3.516414  3.561136  3.605983   3.65089  3.695801  3.740658         3.78541 | 100.104666  100.141549  100.190043  100.253296  100.335137  100.440159    100.5738  100.742413  100.953308  101.214779  101.536089  101.927424  102.399809  102.964984   103.63524  104.423236  105.341775  106.403588  107.621094  109.006186  110.570024  112.322862    114.2739  116.431184  118.801534  121.390517  124.202445  127.240412  130.506342  134.001068   137.72441  141.675263  145.851692  150.251019  154.869905   159.70443  164.750164  170.002226  175.455341  181.103882  186.941914  192.963223   199.16134  205.529564  212.060977  218.748459  225.584688  232.562157  239.673168  246.909839  254.264101 |  100.0 |   0.104666 | 254.264101 |
    | 11:30 | 3.719612  3.720237  3.721057  3.722126  3.723507  3.725276  3.727523  3.730353  3.733887  3.738258  3.743618  3.750131   3.75797  3.767321  3.778372  3.791314  3.806333  3.823609  3.843307  3.865575  3.890541  3.918307  3.948947   3.98251  4.019011  4.058442  4.100765  4.145916  4.193813  4.244351  4.297411   4.35286  4.410557  4.470351   4.53209  4.595619  4.660781  4.727425  4.795398  4.864555  4.934753  5.005856  5.077732  5.150255  5.223308  5.296775  5.370548  5.444525  5.518608             5.592705 | 124.971726  125.004354  125.047196  125.103004  125.175117  125.267535  125.384986  125.532979  125.717845  125.946748   126.22767  126.569361  126.981256  127.473355  128.056065   128.74002  129.535872  130.454071  131.504649  132.696996  134.039674  135.540245  137.205137  139.039557   141.04743  143.231393  145.592816  148.131856  150.847539  153.737857   156.79988  160.029869  163.423395  166.975448  170.680545  174.532825  178.526134  182.654101  186.910199  191.287805  195.780235  200.380786  205.082762   209.87949  214.764343  219.730742  224.772168  229.882162  235.054326   240.28232 |  52.932744   52.939119   52.947487   52.958387   52.972468   52.990508   53.013425   53.042289   53.078321   53.122904   53.177567   53.243979   53.323927   53.419287   53.531987   53.663968   53.817137   53.993315   54.194195   54.421289   54.675891   54.959046   55.271519   55.613788   55.986034   56.388151   56.819757   57.280217    57.76867    58.28406   58.825167   59.390639   59.979028   60.588813    61.21843   61.866297    62.53083   63.210463   63.903659   64.608922   65.324804   66.049911   66.782905   67.522506   68.267497   69.016717   69.769061   70.523482   71.278985        72.034624 |  52.196116   52.202367   52.210573   52.221261   52.235068   52.252758    52.27523   52.303533   52.338866   52.382583   52.436184   52.501307   52.579703    52.67321   52.783722    52.91314   53.063334   53.236091    53.43307   53.655754   53.905412   54.183068   54.489474   54.825096   55.190113   55.584421   56.007645   56.459163   56.938131   57.443513   57.974111   58.528603   59.105566   59.703509   60.320899   60.956185   61.607813   62.274248   62.953983    63.64555   64.347531   65.058556   65.777316   66.502555   67.233079   67.967748   68.705483   69.445253   70.186085      70.92705 | 1.088662  1.088759  1.088886  1.089051  1.089266   1.08954  1.089888  1.090326  1.090874  1.091551   1.09238  1.093387  1.094599  1.096043  1.097748  1.099743  1.102055   1.10471  1.107732  1.111142  1.114957   1.11919  1.123849  1.128937  1.134455  1.140395  1.146749  1.153503   1.16064   1.16814  1.175982  1.184141  1.192593  1.201313  1.210275  1.219454  1.228824  1.238362  1.248044  1.257847  1.267751  1.277734  1.287778  1.297864  1.307977    1.3181  1.328219   1.33832  1.348391  1.358419 | 1.359142  1.359136  1.359127  1.359116  1.359102  1.359084  1.359061  1.359032  1.358996  1.358952  1.358897  1.358831  1.358752  1.358659  1.358548   1.35842  1.358273  1.358105  1.357915  1.357704  1.357469  1.357213  1.356934  1.356634  1.356314  1.355976  1.355621  1.355251  1.354868  1.354475  1.354073  1.353665  1.353253  1.352839  1.352424   1.35201  1.351599  1.351193  1.350791  1.350396  1.350009  1.349629  1.349257  1.348895  1.348542  1.348199  1.347865  1.347541  1.347228          1.346924 | 0.536299  0.536348  0.536413  0.536497  0.536606  0.536745  0.536922  0.537144  0.537421  0.537764  0.538183  0.538691  0.539301  0.540026   0.54088  0.541876  0.543025  0.544338  0.545826  0.547494  0.549348   0.55139   0.55362  0.556034  0.558626  0.561388   0.56431  0.567379  0.570583  0.573906  0.577334  0.580851  0.584442  0.588092  0.591787  0.595513  0.599257  0.603008  0.606755  0.610487  0.614197  0.617876  0.621517  0.625114  0.628663  0.632158  0.635597  0.638974  0.642289      0.645539 | -0.202055   -0.20212  -0.202205  -0.202315  -0.202456  -0.202637  -0.202865   -0.20315  -0.203505  -0.203941  -0.204472  -0.205113  -0.205878  -0.206783  -0.207842  -0.209071  -0.210481  -0.212084  -0.213889  -0.215903  -0.218129  -0.220568  -0.223217  -0.226072  -0.229123   -0.23236  -0.235771   -0.23934  -0.243053  -0.246892  -0.250841  -0.254882  -0.258999  -0.263175  -0.267396  -0.271646  -0.275912  -0.280182  -0.284444  -0.288687  -0.292903  -0.297084  -0.301222   -0.30531  -0.309344  -0.313318  -0.317229  -0.321073  -0.324848      -0.32855 | 0.665729  0.665736  0.665745  0.665756  0.665771  0.665789  0.665812  0.665841  0.665877  0.665921  0.665974  0.666038  0.666115  0.666205   0.66631  0.666431  0.666571  0.666729  0.666908  0.667107  0.667328  0.667571  0.667836  0.668123  0.668431  0.668761   0.66911  0.669479  0.669866   0.67027  0.670689  0.671123   0.67157  0.672029  0.672497  0.672975   0.67346  0.673952  0.674449  0.674949  0.675453  0.675958  0.676465  0.676971  0.677476   0.67798  0.678482   0.67898  0.679475      0.679966 | 100.085838 | 100.101314  100.136836  100.183484  100.244257    100.3228  100.423481  100.551469  100.712798  100.914415  101.164198  101.470956  101.844386  102.294999  102.834007  103.473183  104.224679  105.100835  106.113966  107.276143  108.598985  110.093461  111.769719  113.636942  115.703235  117.975556  120.459672  123.160155  126.080401  129.222675  132.588171   136.17709  139.988722  144.021528  148.273225   152.74087  157.420933  162.309368  167.401673  172.692946  178.177929   183.85105  189.706453  195.738029  201.939433  208.304104  214.825275  221.495985  228.309084  235.257239          242.332932 | 245.897158 | 0.720893   0.72096  0.721049  0.721164  0.721314  0.721505  0.721748  0.722053  0.722435  0.722907  0.723485  0.724187  0.725032  0.726039  0.727227  0.728617  0.730228  0.732078  0.734183  0.736558  0.739215  0.742161  0.745404  0.748945  0.752782  0.756913   0.76133  0.766023   0.77098  0.776187  0.781629   0.78729  0.793151  0.799195  0.805404   0.81176  0.818247  0.824846  0.831542  0.838319  0.845162  0.852057  0.858991  0.865952  0.872927  0.879908  0.886882  0.893841  0.900777       0.907681 | 2.592231  2.592622  2.593135  2.593804  2.594667  2.595773  2.597177  2.598946  2.601155  2.603887  2.607236  2.611304    2.6162  2.622038  2.628935  2.637009  2.646375  2.657142  2.669412  2.683273  2.698802  2.716057  2.735082    2.7559  2.778518  2.802923  2.829086  2.856962  2.886495  2.917614  2.950239  2.984285  3.019658  3.056263  3.094002  3.132776  3.172488  3.213042  3.254342    3.2963  3.338827   3.38184  3.425259  3.469008  3.513016  3.557215   3.60154  3.645931  3.690331        3.734684 | 100.085838  100.116426  100.156776  100.209588  100.278161  100.366477  100.479285  100.622171  100.801615  101.025025  101.300751  101.638061  102.047082  102.538711  103.124485  103.816416  104.626808  105.568048  106.652388   107.89173   109.29742  110.880062  112.649356  114.613972  116.781459  119.158186  121.749321   124.55884  127.589558  130.843186  134.320401  138.020926  141.943612  146.086526  150.447033  155.021877  159.807252  164.798868  169.992009  175.381584  180.962167  186.728037  192.673204  198.791435  205.076275  211.521054  218.118908   224.86278  231.745427  238.759422  245.897158 |  100.0 |   0.085838 | 245.897158 |
    | 12:00 | 3.719292   3.71981  3.720494  3.721387  3.722545  3.724033  3.725932  3.728333  3.731343  3.735084  3.739691  3.745314  3.752116  3.760269  3.769953  3.781353  3.794651  3.810028  3.827653  3.847684   3.87026  3.895497   3.92349  3.954306  3.987986   4.02454  4.063954  4.106189   4.15118  4.198844  4.249077  4.301764  4.356774  4.413969  4.473206  4.534335  4.597207  4.661671   4.72758  4.794787  4.863152  4.932537  5.002809  5.073842  5.145515  5.217711  5.290321  5.363239  5.436365             5.509605 | 124.955028  124.982104  125.017773  125.064398  125.124859  125.202628  125.301839  125.427344  125.584762  125.780505  126.021784  126.316581  126.673592  127.102135  127.612027  128.213427  128.916655  129.731988  130.669459  131.738639  132.948441  134.306943  135.821234  137.497296  139.339925   141.35269  143.537926  145.896763  148.429183    151.1341  154.009454  157.052316  160.258999  163.625167   167.14594  170.815995  174.629653   178.58096  182.663759  186.871747  191.198526  195.637644  200.182628   204.82701  209.564343  214.388222  219.292286   224.27023  229.315801  234.422803 |  52.929482   52.934772    52.94174   52.950847   52.962655   52.977839   52.997202   53.021688   53.052384   53.090529    53.13751   53.194856   53.264221   53.347367   53.446129   53.562382   53.698001   53.854815    54.03456   54.238835   54.469058   54.726431   55.011907   55.326172   55.669632   56.042412   56.444362   56.875073   57.333897   57.819972   58.332257   58.869555    59.43055   60.013832   60.617928   61.241325   61.882494   62.539905   63.212045    63.89743   64.594615   65.302202   66.018842   66.743241   67.474162   68.210423   68.950899   69.694518   70.440263        71.187166 |  52.192917   52.198104   52.204937   52.213868   52.225446   52.240335   52.259322   52.283333   52.313433   52.350837   52.396905   52.453137   52.521156   52.602687   52.699531   52.813526   52.946512    53.10028   53.276535   53.476843   53.702596   53.954971   54.234903   54.543065   54.879855   55.245396   55.639541   56.061888   56.511801   56.988437   57.490774   58.017638   58.567738   59.139693   59.732058    60.34335   60.972068   61.616712   62.275799   62.947875   63.631522   64.325367   65.028091   65.738422   66.455149   67.177112   67.903209   68.632387    69.36365     70.096049 | 1.088612  1.088692  1.088798  1.088937  1.089116  1.089347  1.089641  1.090014   1.09048  1.091059  1.091772  1.092642  1.093694  1.094954   1.09645  1.098208  1.100257  1.102623  1.105331  1.108403  1.111859  1.115713  1.119979  1.124662  1.129766  1.135289  1.141224  1.147562  1.154289  1.161388   1.16884  1.176623  1.184715  1.193092  1.201728    1.2106  1.219683  1.228952  1.238384  1.247957  1.257649  1.267439  1.277307  1.287235  1.297206  1.307204  1.317213  1.327219  1.337208  1.347169 | 1.359146   1.35914  1.359133  1.359124  1.359112  1.359097  1.359077  1.359053  1.359022  1.358984  1.358937   1.35888  1.358811  1.358729  1.358632  1.358519  1.358388  1.358237  1.358066  1.357873  1.357659  1.357423  1.357165  1.356886  1.356586  1.356266  1.355929  1.355576  1.355209  1.354829  1.354439  1.354041  1.353637  1.353229  1.352819  1.352409     1.352  1.351594  1.351192  1.350795  1.350404  1.350021  1.349645  1.349277  1.348918  1.348569  1.348228  1.347898  1.347577          1.347265 | 0.536274  0.536315  0.536369  0.536439   0.53653  0.536647  0.536797  0.536985  0.537222  0.537515  0.537876  0.538316  0.538846   0.53948   0.54023   0.54111  0.542131  0.543306  0.544645  0.546155  0.547843  0.549714  0.551769  0.554007  0.556425  0.559015  0.561771  0.564681  0.567734  0.570916  0.574214  0.577612  0.581097  0.584652  0.588265   0.59192  0.595605  0.599308  0.603017  0.606721  0.610412  0.614081  0.617719  0.621322  0.624881  0.628394  0.631854  0.635259  0.638605       0.64189 | -0.202022  -0.202076  -0.202146  -0.202238  -0.202357  -0.202509  -0.202703  -0.202946  -0.203249  -0.203623  -0.204081  -0.204637  -0.205304  -0.206097  -0.207031   -0.20812  -0.209378  -0.210816  -0.212446  -0.214275  -0.216309  -0.218551  -0.221001  -0.223656   -0.22651  -0.229555  -0.232781  -0.236174   -0.23972  -0.243405  -0.247211  -0.251123  -0.255125  -0.259199   -0.26333  -0.267504  -0.271706  -0.275924  -0.280145  -0.284358  -0.288554  -0.292723  -0.296857   -0.30095  -0.304996  -0.308988  -0.312922  -0.316795  -0.320603     -0.324343 | 0.665726  0.665731  0.665739  0.665748   0.66576  0.665776  0.665796   0.66582  0.665851  0.665889  0.665935   0.66599  0.666057  0.666136  0.666228  0.666336  0.666461  0.666603  0.666764  0.666944  0.667146  0.667368  0.667612  0.667877  0.668164  0.668471  0.668799  0.669147  0.669513  0.669897  0.670297  0.670713  0.671143  0.671585  0.672038  0.672502  0.672974  0.673453  0.673939   0.67443  0.674924  0.675422  0.675921  0.676421  0.676921  0.677421  0.677919  0.678415  0.678908      0.679398 | 100.070317 | 100.083136  100.112612  100.151447  100.202216  100.268059  100.352769  100.460859  100.597638  100.769263  100.982775  101.246112  101.568093  101.958367  102.427334  102.986027  103.645964  104.418977  105.317023  106.351974   107.53542  108.878467  110.391557  112.084312  113.965404  116.042458  118.321988  120.809369  123.508831  126.423487  129.555378  132.905531  136.474033  140.260111  144.262211  148.478075  152.904823  157.539021  162.376743  167.413635  172.644958  178.065639  183.670302  189.453301  195.408745  201.530517  207.812289  214.247538  220.829549  227.551423          234.406082 | 237.861066 | 0.720858  0.720914  0.720988  0.721084   0.72121  0.721371  0.721576  0.721835   0.72216  0.722564  0.723062  0.723668  0.724401   0.72528  0.726322  0.727548  0.728975  0.730624  0.732511  0.734651  0.737057  0.739741  0.742711   0.74597  0.749521  0.753362  0.757489  0.761895  0.766569  0.771499  0.776673  0.782075  0.787688  0.793496  0.799483  0.805629  0.811919  0.818335  0.824861  0.831482  0.838182  0.844946  0.851762  0.858617  0.865498  0.872394  0.879296  0.886192  0.893075       0.899936 | 2.592031  2.592356  2.592783  2.593341  2.594065  2.594996  2.596183  2.597684  2.599565  2.601903  2.604782  2.608295  2.612543  2.617635  2.623681  2.630795  2.639091  2.648678  2.659662  2.672137  2.686188  2.701883  2.719277  2.738408  2.759295  2.781941  2.806332  2.832436   2.86021  2.889594  2.920521  2.952913  2.986686  3.021749  3.058009  3.095373  3.133745   3.17303  3.213136  3.253972  3.295449  3.337485  3.379998  3.422911   3.46615  3.509647  3.553334  3.597151  3.641037        3.684937 | 100.070317  100.095651  100.129179  100.173209  100.230576  100.304725  100.399789   100.52066  100.673058  100.863574  101.099698  101.389815  101.743178  102.169836  102.680543  103.286617  103.999783  104.831992  105.795214  106.901236  108.161459  109.586705  111.187047  112.971662  114.948722  117.125312  119.507379  122.099724   124.90601  127.928798  131.169605  134.628967  138.306519  142.201076   146.31071  150.632833  155.164268  159.901319   164.83983  169.975244  175.302642  180.816791  186.512172  192.383011  198.423298  204.626807  210.987111  217.497587  224.151427  230.941643  237.861066 |  100.0 |   0.070317 | 237.861066 |
    | 12:30 | 3.719028  3.719458  3.720026  3.720771   3.72174  3.722991  3.724593  3.726626  3.729185  3.732379  3.736329  3.741174  3.747061  3.754151  3.762615  3.772627  3.784368  3.798013  3.813735  3.831696  3.852044  3.874909  3.900401  3.928604  3.959578  3.993357  4.029949  4.069334   4.11147  4.156293  4.203718  4.253644  4.305958  4.360532  4.417232  4.475919  4.536449  4.598677  4.662458  4.727648  4.794106  4.861697  4.930288  4.999749   5.06996  5.140802  5.212162  5.283935  5.356017             5.428312 | 124.941255  124.963693  124.993349  125.032245  125.082858  125.148195  125.231855  125.338097  125.471887  125.638938  125.845731  126.099504  126.408223  126.780513  127.225562  127.752992  128.372706  129.094708  129.928907  130.884923  131.971885   133.19825  134.571635  136.098681  137.784953  139.634868  141.651661  143.837393  146.192976  148.718235  151.411984  154.272122  157.295728  160.479168  163.818199  167.308069  170.943607  174.719309  178.629414  182.667968  186.828878  191.105965  195.492994  199.983713   204.57187  209.251236  214.015615  218.858855  223.774853  228.757556 |   52.92679   52.931175   52.936969   52.944567   52.954453   52.967211   52.983544   53.004277   53.030376   53.062944   53.103234   53.152634    53.21267   53.284978    53.37129     53.4734   53.593129   53.732285    53.89262   54.075788   54.283297   54.516476   54.776438   55.064052    55.37993   55.724413   56.097574   56.499225   56.928931   57.386035   57.869678   58.378832   58.912325   59.468874   60.047109   60.645602   61.262888   61.897488   62.547924   63.212734   63.890484   64.579779   65.279266    65.98764    66.70365   67.426096   68.153835   68.885773   69.620872        70.358141 |  52.190278   52.194577   52.200258   52.207709   52.217403   52.229914   52.245929    52.26626   52.291852   52.323788   52.363295   52.411736   52.470606    52.54151   52.626146   52.726273   52.843676    52.98013   53.137352   53.316962   53.520442   53.749093   54.004006   54.286035   54.595779   54.933572   55.299487   55.693338     56.1147   56.562927   57.037178   57.536445   58.059578   58.605319   59.172325   59.759195   60.364494   60.986771   61.624575   62.276475   62.941064   63.616973   64.302876   64.997494     65.6996   66.408017   67.121623   67.839348   68.560171     69.283123 | 1.088571  1.088638  1.088726  1.088841  1.088992  1.089186  1.089434  1.089749  1.090146   1.09064  1.091252  1.092002  1.092913  1.094009  1.095317  1.096862  1.098673  1.100774  1.103193  1.105951  1.109071  1.112569  1.116461  1.120757  1.125462  1.130579  1.136105  1.142033  1.148353  1.155052  1.162113  1.169516  1.177242  1.185267  1.193568  1.202123  1.210906  1.219895  1.229065  1.238394   1.24786  1.257443  1.267122  1.276878  1.286694  1.296552  1.306437  1.316334   1.32623   1.33611 | 1.359148  1.359144  1.359138   1.35913   1.35912  1.359108  1.359091   1.35907  1.359044  1.359011  1.358971  1.358922  1.358862  1.358791  1.358706  1.358606  1.358489  1.358354  1.358201  1.358027  1.357832  1.357616  1.357378  1.357118  1.356838  1.356538   1.35622  1.355884  1.355532  1.355167   1.35479  1.354404   1.35401   1.35361  1.353206  1.352801  1.352395  1.351991  1.351589  1.351191  1.350799  1.350412  1.350033  1.349661  1.349297  1.348942  1.348595  1.348258   1.34793          1.347611 | 0.536253  0.536287  0.536332   0.53639  0.536467  0.536565  0.536691  0.536851  0.537052  0.537303  0.537613  0.537992  0.538452  0.539004  0.539662  0.540437  0.541342  0.542389  0.543588   0.54495  0.546482  0.548189  0.550076  0.552142  0.554388  0.556807  0.559396  0.562144  0.565042  0.568078  0.571239  0.574511   0.57788  0.581333  0.584853  0.588429  0.592045  0.595691  0.599353  0.603021  0.606684  0.610334  0.613963  0.617563  0.621127   0.62465  0.628126  0.631553  0.634925       0.63824 | -0.201994  -0.202039  -0.202098  -0.202175  -0.202274  -0.202403  -0.202566  -0.202772  -0.203031  -0.203351  -0.203746  -0.204226  -0.204806  -0.205499   -0.20632  -0.207282    -0.2084  -0.209686  -0.211151  -0.212806  -0.214657   -0.21671  -0.218966  -0.221426  -0.224085  -0.226938  -0.229976  -0.233189  -0.236564  -0.240088  -0.243744  -0.247519  -0.251395  -0.255356  -0.259389  -0.263476  -0.267604   -0.27176   -0.27593  -0.280104   -0.28427  -0.288418  -0.292542  -0.296631  -0.300681  -0.304684  -0.308635  -0.312531  -0.316366     -0.320138 | 0.665723  0.665727  0.665734  0.665741  0.665752  0.665765  0.665782  0.665803  0.665829  0.665861  0.665901  0.665949  0.666007  0.666076  0.666157  0.666253  0.666363   0.66649  0.666635  0.666798   0.66698  0.667183  0.667407  0.667652  0.667917  0.668204   0.66851  0.668837  0.669183  0.669546  0.669927  0.670324  0.670736  0.671161  0.671598  0.672047  0.672505  0.672972  0.673445  0.673925   0.67441  0.674899   0.67539  0.675884  0.676378  0.676872  0.677366  0.677859  0.678349      0.678837 | 100.057539 | 100.068142  100.092568  100.124855  100.167204  100.222318  100.293475  100.384608  100.500369  100.646193  100.828346  101.053946  101.330973  101.668237  102.075323  102.562504  103.140617  103.820923  104.614928  105.534207  106.590201  107.794031  109.156309  110.686977  112.395156  114.289038  116.375795   118.66153   121.15125  123.848872  126.757248  129.878211  133.212638  136.760514  140.521011  144.492563  148.672943  153.059331  157.648383  162.436294  167.418846  172.591461  177.949237  183.486988  189.199266  195.080387  201.124452  207.325356  213.676804  220.172317          226.805236 | 230.150482 | 0.720829  0.720876  0.720937  0.721018  0.721123  0.721258  0.721431  0.721651  0.721927  0.722272  0.722699  0.723222  0.723856  0.724621  0.725532   0.72661  0.727871  0.729336  0.731021  0.732943  0.735116  0.737552  0.740262  0.743252  0.746527  0.750086   0.75393  0.758051  0.762444  0.767098  0.772002  0.777142  0.782504  0.788071  0.793827  0.799756  0.805841  0.812066  0.818413  0.824868  0.831415  0.838039  0.844727  0.851466  0.858243  0.865046  0.871865   0.87869  0.885511       0.892319 | 2.591866  2.592135   2.59249  2.592956  2.593562  2.594344  2.595346  2.596617  2.598216  2.600212  2.602681  2.605708  2.609386  2.613815    2.6191   2.62535  2.632676  2.641187  2.650989  2.662181  2.674852   2.68908  2.704931  2.722452  2.741678  2.762624   2.78529  2.809659  2.835698  2.863364  2.892597  2.923331   2.95549  2.988991  3.023747  3.059669  3.096665  3.134642  3.173509  3.213177  3.253558  3.294567  3.336124  3.378148  3.420567  3.463309  3.506306  3.549494  3.592814        3.636207 | 100.057539  100.078493  100.106314   100.14297  100.190892  100.253051   100.33303  100.435103  100.564298  100.726455  100.928261  101.177268   101.48188  101.851314  102.295527  102.825112  103.451159  104.185103  105.038536   106.02302  107.149888  108.430057  109.873847  111.490826  113.289678  115.278107   117.46276  119.849195  122.441872  125.244163  128.258399  131.485916  134.927125  138.581585  142.448077  146.524686  150.808867  155.297522  159.987061  164.873458  169.952303  175.218847  180.668036  186.294546  192.092805  198.057018  204.181178  210.459084  216.884347  223.450395  230.150482 |  100.0 |   0.057539 | 230.150482 |
    | 13:00 |  3.71881  3.719166  3.719638  3.720259  3.721069  3.722119  3.723467  3.725186  3.727357  3.730079  3.733459  3.737623  3.742706  3.748857  3.756234  3.765005   3.77534  3.787413  3.801395  3.817451  3.835733  3.856383  3.879521   3.90525  3.933645  3.964761  3.998625   4.03524  4.074584  4.116612  4.161258   4.20844  4.258058  4.309999  4.364141  4.420352  4.478498  4.538439  4.600035  4.663146  4.727632  4.793359  4.860194  4.928009  4.996679  5.066086  5.136116  5.206661  5.277616             5.348882 | 124.929909  124.948478  124.973101  125.005502  125.047808  125.102614  125.173045  125.262823  125.376322  125.518609  125.695482  125.913471  126.179829  126.502489  126.889988  127.351373  127.896062  128.533697  129.273963   130.12641  131.100253  132.204196  133.446257  134.833617  136.372497   138.06807  139.924397  141.944411  144.129915  146.481627  148.999229  151.681452  154.526158   157.53044  160.690721  164.002849  167.462195  171.063737  174.802138  178.671819  182.667022  186.781856  191.010347  195.346469  199.784179  204.317433  208.940206  213.646505  218.430377  223.285911 |  52.924573   52.928202   52.933013   52.939343   52.947607   52.958311   52.972063   52.989588   53.011735   53.039487   53.073964   53.116425   53.168262   53.230988   53.306223   53.395665   53.501065   53.624189   53.766778   53.930511   54.116959   54.327545   54.563512   54.825888   55.115467   55.432788   55.778136   56.151535   56.552763   56.981365   57.436675   57.917838   58.423841   58.953538   59.505675   60.078924   60.671898    61.28318   61.911337   62.554941   63.212579   63.882865   64.564449   65.256023   65.956326   66.664144   67.378316   68.097732   68.821331        69.548106 |  52.188104   52.191662   52.196379   52.202587    52.21069   52.221186   52.234671   52.251856   52.273573   52.300786   52.334593    52.37623    52.42706   52.488568   52.562342   52.650047     52.7534   52.874133   53.013954   53.174507   53.357334    53.56383   53.795215   54.052496   54.336451   54.647611   54.986252     55.3524   55.745837   56.166116   56.612583   57.084403    57.58058    58.09999   58.641405   59.203522   59.784981   60.384391    61.00035   61.631456   62.276323   62.933593   63.601941   64.280085   64.966788   65.660861   66.361164   67.066609   67.776158     68.488818 | 1.088537  1.088593  1.088666  1.088762  1.088888   1.08905  1.089259  1.089526  1.089862  1.090284  1.090808  1.091452  1.092239   1.09319  1.094331  1.095686  1.097281  1.099142  1.101295  1.103764  1.106571  1.109735  1.113274  1.117201  1.121524  1.126248  1.131376  1.136902  1.142822  1.149123  1.155792  1.162814  1.170169  1.177837  1.185796  1.194024  1.202498  1.211194   1.22009  1.229163  1.238392  1.247754   1.25723  1.266801  1.276448  1.286153  1.295902  1.305677  1.315465  1.325252 |  1.35915  1.359147  1.359142  1.359136  1.359127  1.359116  1.359103  1.359085  1.359063  1.359035     1.359  1.358958  1.358906  1.358844   1.35877  1.358682  1.358579  1.358459  1.358321  1.358165  1.357988  1.357791  1.357572  1.357333  1.357072  1.356792  1.356492  1.356174   1.35584   1.35549  1.355127  1.354753   1.35437   1.35398  1.353584  1.353184  1.352783  1.352382  1.351982  1.351585  1.351191  1.350803  1.350421  1.350045  1.349677  1.349317  1.348965  1.348622  1.348287          1.347962 | 0.536236  0.536264  0.536301   0.53635  0.536414  0.536497  0.536603  0.536738  0.536909  0.537123  0.537388  0.537714  0.538112  0.538592  0.539166  0.539847  0.540646  0.541576  0.542648  0.543871  0.545255  0.546807  0.548532  0.550433   0.55251  0.554761  0.557182  0.559767  0.562507  0.565392  0.568411  0.571551  0.574798  0.578138  0.581559  0.585045  0.588584  0.592163  0.595769  0.599392   0.60302  0.606643  0.610254  0.613844  0.617405  0.620932  0.624419  0.627861  0.631254      0.634594 | -0.201971  -0.202008  -0.202057  -0.202122  -0.202205  -0.202313  -0.202451  -0.202626  -0.202845  -0.203119  -0.203458  -0.203872  -0.204375  -0.204979  -0.205698  -0.206546  -0.207536  -0.208681  -0.209994  -0.211486  -0.213164  -0.215036  -0.217106  -0.219375  -0.221843  -0.224505  -0.227356  -0.230387  -0.233587  -0.236943  -0.240443  -0.244072  -0.247814  -0.251655  -0.255578  -0.259569  -0.263613  -0.267697  -0.271807  -0.275931  -0.280058  -0.284178  -0.288281  -0.292359  -0.296405  -0.300412  -0.304374  -0.308286  -0.312143     -0.315942 |  0.66572  0.665724  0.665729  0.665736  0.665745  0.665756   0.66577  0.665788   0.66581  0.665838  0.665872  0.665913  0.665964  0.666024  0.666095  0.666179  0.666277   0.66639  0.666519  0.666666  0.666832  0.667016  0.667221  0.667445  0.667691  0.667956  0.668242  0.668548  0.668873  0.669217  0.669578  0.669956   0.67035  0.670757  0.671178  0.671611  0.672055  0.672508  0.672969  0.673437  0.673912  0.674391  0.674873  0.675359  0.675847  0.676335  0.676824  0.677312  0.677799      0.678284 | 100.047032 | 100.055791  100.076005  100.102811  100.138087  100.184151  100.243832  100.320543  100.418347  100.542026  100.697131   100.89002  101.127875  101.418699  101.771275  102.195109  102.700331  103.297576  103.997837  104.812296  105.752148  106.828413  108.051758  109.432323  110.979568  112.702144  114.607787  116.703248  118.994242  121.485437  124.180458  127.081917  130.191459  133.509821  137.036897  140.771814  144.712999  148.858251  153.204814  157.749434  162.488415  167.417675  172.532785  177.829013  183.301346  188.944527  194.753072  200.721285  206.843275  213.112967          219.524103 | 222.759457 | 0.720806  0.720844  0.720895  0.720962   0.72105  0.721164  0.721309  0.721495   0.72173  0.722024  0.722389  0.722838  0.723387   0.72405  0.724845   0.72579  0.726901  0.728199  0.729699  0.731419  0.733374  0.735579  0.738043  0.740777  0.743786  0.747074  0.750641  0.754484    0.7586  0.762979  0.767613  0.772489  0.777595  0.782917  0.788438  0.794143  0.800016  0.806041  0.812201  0.818481  0.824867  0.831341  0.837892  0.844506  0.851169   0.85787  0.864597  0.871341  0.878091       0.884837 |  2.59173  2.591953  2.592248  2.592636  2.593143  2.593799  2.594642  2.595716  2.597074  2.598775  2.600888   2.60349  2.606666  2.610508  2.615116  2.620592  2.627043  2.634576  2.643296  2.653305  2.664695  2.677552  2.691949  2.707944  2.725583  2.744894  2.765889  2.788566  2.812905  2.838874  2.866426  2.895506  2.926046  2.957972  2.991205  3.025658  3.061246   3.09788   3.13547  3.173928  3.213168  3.253104  3.293656  3.334744  3.376292  3.418228  3.460484  3.502993  3.545694        3.588528 | 100.047032   100.06434  100.087395   100.11787  100.157844  100.209872  100.277054  100.363108  100.472441  100.610206  100.782348  100.995636  101.257667  101.576846  101.962339  102.423992  102.972227  103.617899  104.372143  105.246199  106.251226  107.398116  108.697318  110.158672  111.791267  113.603322  115.602096  117.793832  120.183724  122.775913  125.573507  128.578621  131.792428  135.215227  138.846506  142.685023  146.728873   150.97556   155.42206  160.064886  164.900134   169.92354  175.130511  180.516167  186.075368  191.802736  197.692678  203.739398  209.936909  216.279043  222.759457 |  100.0 |   0.047032 | 222.759457 |
    | 13:30 | 3.718631  3.718926  3.719317  3.719833   3.72051  3.721389  3.722522  3.723972  3.725812  3.728126  3.731013  3.734584  3.738963  3.744286    3.7507  3.758362  3.767436  3.778088  3.790486  3.804794  3.821171  3.839761  3.860698  3.884093  3.910042  3.938614  3.969856   4.00379  4.040415  4.079706  4.121617  4.166081  4.213016  4.262323  4.313893  4.367606  4.423334  4.480947   4.54031  4.601286   4.66374  4.727539   4.79255  4.858646  4.925702  4.993599  5.062222   5.13146  5.201207             5.271363 | 124.920573  124.935921  124.956338  124.983291  125.018602  125.064503  125.123701  125.199438  125.295551  125.416519  125.567501  125.754362   125.98367  126.262677  126.599267   127.00188  127.479406  128.041055  128.696201  129.454217  130.324292  131.315253  132.435394  133.692312  135.092775  136.642609  138.346618  140.208536  142.231011  144.415616  146.762888  149.272383  151.942758  154.771848  157.756764  160.893987  164.179456   167.60866   171.17672  174.878462   178.70848  182.661202   186.73093  190.911885  195.198243  199.584156  204.063781  208.631287  213.280876  218.006781 |  52.922749   52.925748   52.929738   52.935004   52.941902   52.950868   52.962429   52.977216   52.995975   53.019577   53.049019   53.085436   53.130091   53.184375   53.249788   53.327927   53.420459   53.529087   53.655522   53.801442   53.968449   54.158035   54.371545   54.610137   54.874762   55.166139   55.484745   55.830809   56.204313   56.605005   57.032411   57.485859   57.964501   58.467339    58.99325   59.541014   60.109337   60.696877   61.302258   61.924095   62.561007   63.211626   63.874612    64.54866   65.232503    65.92492   66.624736   67.330826   68.042112        68.757565 |  52.186315   52.189256   52.193168   52.198332   52.205096   52.213888   52.225224   52.239724   52.258119   52.281262   52.310133   52.345843   52.389631    52.44286   52.507003   52.583625   52.674359   52.780878   52.904858   53.047944   53.211708   53.397613   53.606976   53.840935   54.100421    54.38614   54.698559   55.037903   55.404153   55.797064    56.21617   56.660812    57.13016   57.623233   58.138931   58.676058   59.233345   59.809474   60.403099   61.012861   61.637404   62.275389     62.9255   63.586458   64.257022   64.935993   65.622219   66.314597    67.01207      67.71363 |  1.08851  1.088555  1.088616  1.088696  1.088801  1.088937  1.089113  1.089338  1.089623  1.089981  1.090429  1.090982   1.09166  1.092483  1.093475   1.09466  1.096061  1.097704  1.099615  1.101818  1.104335  1.107188  1.110396  1.113973  1.117931  1.122279  1.127021  1.132156  1.137682  1.143591  1.149872  1.156511  1.163494    1.1708  1.178411  1.186305   1.19446  1.202854  1.211465  1.220271  1.229249  1.238378  1.247639  1.257011  1.266476  1.276016  1.285614  1.295255  1.304923  1.314604 | 1.359152  1.359149  1.359145   1.35914  1.359133  1.359124  1.359112  1.359097  1.359079  1.359055  1.359025  1.358989  1.358944   1.35889  1.358826  1.358748  1.358657  1.358551  1.358429  1.358288  1.358129  1.357949   1.35775  1.357529  1.357289  1.357027  1.356746  1.356447   1.35613  1.355797  1.355449  1.355089  1.354718  1.354338  1.353951  1.353559  1.353164  1.352767   1.35237  1.351974  1.351581  1.351192  1.350808   1.35043  1.350058  1.349693  1.349336  1.348988  1.348648          1.348316 | 0.536222  0.536245  0.536276  0.536316   0.53637  0.536439  0.536528  0.536643  0.536787  0.536969  0.537196  0.537476  0.537819  0.538235  0.538736  0.539332  0.540035  0.540858  0.541812  0.542907  0.544153  0.545559   0.54713  0.548871  0.550785  0.552871  0.555127  0.557549   0.56013  0.562861  0.565733  0.568734  0.571852  0.575074  0.578386  0.581775  0.585228  0.588732  0.592274  0.595842  0.599426  0.603015  0.606599  0.610171  0.613723  0.617247  0.620738  0.624189  0.627597      0.630958 | -0.201953  -0.201983  -0.202024  -0.202077  -0.202147  -0.202238  -0.202354  -0.202502  -0.202688  -0.202922  -0.203212  -0.203569  -0.204003  -0.204528  -0.205156    -0.2059  -0.206775  -0.207792  -0.208964  -0.210303  -0.211819   -0.21352  -0.215411  -0.217497  -0.219778  -0.222253  -0.224917  -0.227764  -0.230787  -0.233973  -0.237311  -0.240787  -0.244388  -0.248099  -0.251904  -0.255789   -0.25974  -0.263742  -0.267782  -0.271847  -0.275926  -0.280008  -0.284083  -0.288142  -0.292177   -0.29618  -0.300145  -0.304066  -0.307939     -0.311759 | 0.665718  0.665722  0.665726  0.665731  0.665739  0.665748   0.66576  0.665775  0.665794  0.665818  0.665847  0.665883  0.665926  0.665979  0.666041  0.666115  0.666201  0.666301  0.666417  0.666549  0.666698  0.666865  0.667051  0.667257  0.667483  0.667729  0.667994   0.66828  0.668585  0.668909   0.66925  0.669609  0.669984  0.670374  0.670778  0.671195  0.671623  0.672062   0.67251  0.672966  0.673429  0.673897  0.674371  0.674848  0.675328   0.67581  0.676293  0.676776  0.677259       0.67774 | 100.038402 | 100.045629  100.062336  100.084562  100.113905  100.152349  100.202329  100.266798  100.349293  100.454007  100.585838  100.750441  100.954252  101.204502  101.509199  101.877086  102.317574   102.84064  103.456705  104.176492  105.010856   105.97062  107.066393  108.308403  109.706336  111.269196  113.005184  114.921611  117.024828  119.320192  121.812054  124.503768  127.397726  130.495399  133.797398  137.303536  141.012899   144.92391  149.034402  153.341675  157.842563  162.533479  167.410467  172.469244  177.705234  183.113595   188.68925  194.426903  200.321054  206.366017          212.555921 | 215.681533 | 0.720787  0.720818  0.720861  0.720916   0.72099  0.721085  0.721207  0.721364  0.721563  0.721813  0.722125   0.72251  0.722983  0.723557  0.724249  0.725074  0.726051  0.727197  0.728528  0.730063  0.731817  0.733805  0.736039  0.738529  0.741285  0.744311  0.747611  0.751184  0.755026  0.759134  0.763499  0.768112  0.772961  0.778033  0.783315   0.78879  0.794445  0.800263  0.806228  0.812326  0.818541  0.824857  0.831262  0.837741  0.844281  0.850871  0.857498  0.864151  0.870821       0.877497 | 2.591619  2.591802  2.592047   2.59237  2.592793  2.593343  2.594051  2.594958  2.596108  2.597554  2.599359  2.601591  2.604327  2.607653  2.611659  2.616445  2.622109  2.628758  2.636493  2.645416  2.655623  2.667204  2.680238  2.694792  2.710922  2.728668  2.748054  2.769089  2.791769  2.816072  2.841964  2.869401  2.898324  2.928669  2.960364   2.99333  3.027485  3.062744  3.099022  3.136233   3.17429  3.213111  3.252613  3.292717  3.333347   3.37443  3.415895  3.457676  3.499709        3.541933 | 100.038402  100.052681  100.071761  100.097063  100.130361  100.173844  100.230189  100.302621  100.394988   100.51182  100.658383  100.840716  101.065654  101.340821   101.67461  102.076118  102.555069  103.121701  103.786627  104.560682  105.454755  106.479609  107.645709  108.963055  110.441026  112.088256  113.912524  115.920674  118.118569  120.511062  123.102001  125.894247  128.889719  132.089442  135.493608  139.101649  142.912301  146.923673  151.133316  155.538281  160.135177  164.920221  169.889284  175.037926  180.361431  185.854833   191.51294  197.330352  203.301471  209.420517  215.681533 |  100.0 |   0.038402 | 215.681533 |
    | 14:00 | 3.718484  3.718727  3.719051   3.71948  3.720044  3.720779  3.721731  3.722952  3.724508  3.726472  3.728933  3.731989  3.735753  3.740349  3.745912  3.752589  3.760533  3.769905  3.780867  3.793582  3.808206  3.824892  3.843777  3.864985  3.888623  3.914776  3.943509  3.974862  4.008853  4.045476  4.084705   4.12649  4.170766   4.21745  4.266445  4.317646  4.370933  4.426185  4.483273  4.542067  4.602435  4.664247  4.727371  4.791682  4.857056  4.923371  4.990513  5.058369  5.126832               5.1958 | 124.912901  124.925571  124.942477  124.964868  124.994298  125.032683  125.082362   125.14615    125.2274  125.330054  125.458688  125.618543  125.815538  126.056265  126.347959  126.698445   127.11605  127.609502  128.187792  128.860029  129.635272   130.52236  131.529742  132.665316   133.93628   135.34901  136.908957  138.620583  140.487317  142.511544  144.694624   147.03693  149.537909  152.196149  155.009468     157.975  161.089282  164.348343  167.747789  171.282877  174.948588  178.739685  182.650771  186.676333   190.81078  195.048475  199.383764  203.810989  208.324508  212.918706 |   52.92125   52.923726   52.927029   52.931404   52.937154   52.944653   52.954356   52.966812   52.982674   53.002708   53.027802   53.058969   53.097353   53.144221   53.200956   53.269047   53.350065    53.44564   53.557431   53.687093   53.836237   54.006395   54.198984   54.415267   54.656327   54.923041    55.21606     55.5358   55.882441   56.255926   56.655976   57.082103    57.53363   58.009715   58.509378   59.031518   59.574947   60.138407   60.720595   61.320178   61.935816    62.56617   63.209918   63.865763   64.532443    65.20873   65.893441    66.58544   67.283632        67.986974 |  52.184845   52.187272   52.190512   52.194802    52.20044   52.207793   52.217308   52.229522   52.245076   52.264721   52.289327    52.31989   52.357528   52.403486   52.459119   52.525888   52.605332   52.699052   52.808672   52.935816   53.082063   53.248917   53.437766   53.649849   53.886228   54.147763   54.435091   54.748622   55.088532   55.454764   55.847045   56.264897   56.707655   57.174496   57.664455   58.176456   58.709332    59.26185   59.832732   60.420672   61.024354   61.642467   62.273714   62.916823   63.570556    64.23371   64.905125   65.583685   66.268319     66.958003 | 1.088487  1.088524  1.088575  1.088641  1.088729  1.088843   1.08899   1.08918  1.089421  1.089725  1.090106   1.09058  1.091163  1.091874  1.092735  1.093767  1.094995  1.096442  1.098133  1.100092  1.102343  1.104907  1.107804  1.111052  1.114664  1.118652  1.123023  1.127779  1.132921  1.138444  1.144341  1.150601   1.15721  1.164152  1.171409  1.178963  1.186792  1.194876  1.203192  1.211719  1.220436  1.229321  1.238354  1.247516  1.256786  1.266148  1.275583  1.285076  1.294612  1.304175 | 1.359154  1.359151  1.359148  1.359144  1.359138   1.35913   1.35912  1.359108  1.359092  1.359072  1.359047  1.359015  1.358977   1.35893  1.358874  1.358807  1.358727  1.358633  1.358524  1.358398  1.358255  1.358092  1.357911  1.357709  1.357487  1.357245  1.356983  1.356702  1.356403  1.356086  1.355755  1.355409  1.355051  1.354683  1.354306  1.353923  1.353535  1.353144  1.352751  1.352358  1.351967  1.351578  1.351193  1.350813  1.350439  1.350071   1.34971  1.349356  1.349011          1.348674 |  0.53621  0.536229  0.536255  0.536289  0.536333  0.536391  0.536466  0.536562  0.536685  0.536839  0.537033  0.537273  0.537568  0.537928  0.538362  0.538883    0.5395  0.540226  0.541073  0.542049  0.543167  0.544435  0.545861   0.54745  0.549206  0.551132  0.553226  0.555486  0.557908  0.560484  0.563206  0.566064  0.569048  0.572144  0.575341  0.578625  0.581983  0.585403  0.588872  0.592377  0.595909  0.599454  0.603005  0.606552  0.610086  0.613601  0.617088  0.620543  0.623961      0.627335 | -0.201937  -0.201962  -0.201996  -0.202041  -0.202099  -0.202175  -0.202272  -0.202397  -0.202555  -0.202754  -0.203002  -0.203308  -0.203683  -0.204138  -0.204685  -0.205337  -0.206106  -0.207006  -0.208049  -0.209248  -0.210612  -0.212151  -0.213873  -0.215783  -0.217883  -0.220174  -0.222654   -0.22532  -0.228163  -0.231176  -0.234348  -0.237667   -0.24112  -0.244693  -0.248372  -0.252143  -0.255991  -0.259902  -0.263863   -0.26786  -0.271882  -0.275917  -0.279955  -0.283986  -0.288001  -0.291993  -0.295955  -0.299879  -0.303761     -0.307596 | 0.665717  0.665719  0.665723  0.665728  0.665734  0.665741  0.665751  0.665764   0.66578  0.665801  0.665826  0.665856  0.665894  0.665939  0.665994  0.666058  0.666135  0.666223  0.666326  0.666444  0.666578  0.666729  0.666898  0.667086  0.667293   0.66752  0.667766  0.668032  0.668317  0.668621  0.668943  0.669282  0.669639  0.670011  0.670397  0.670798   0.67121  0.671634  0.672068  0.672511  0.672962   0.67342  0.673883  0.674351  0.674823  0.675297  0.675774  0.676251  0.676729      0.677206 | 100.031322 | 100.037278  100.051069  100.069473  100.093848  100.125888  100.167682  100.221778  100.291248  100.379754  100.491604  100.631805  100.806103     101.021  101.283759  101.602379   101.98555  102.442571  102.983258  103.617812  104.356679  105.210394  106.189412   107.30394  108.563785  109.978196  111.555741  113.304201   115.23048  117.340559  119.639457   122.13123  124.818983  127.704903  130.790302  134.075677  137.560769  141.244626  145.125675  149.201777  153.470295  157.928142  162.571839   167.39755  172.401131  177.578154  182.923942  188.433588  194.101976  199.923794          205.893546 | 208.909841 | 0.720771  0.720797  0.720832  0.720878  0.720939  0.721019  0.721122  0.721254  0.721422  0.721634    0.7219   0.72223  0.722636  0.723133  0.723733  0.724452  0.725308  0.726317  0.727495  0.728861  0.730429  0.732215  0.734234  0.736495  0.739011  0.741787  0.744829  0.748139  0.751716  0.755556  0.759656  0.764006  0.768597  0.773418  0.778456  0.783698  0.789129  0.794733  0.800497  0.806404  0.812441  0.818591  0.824841  0.831177  0.837585  0.844055  0.850572  0.857126  0.863707       0.870305 | 2.591527  2.591678  2.591881  2.592149  2.592502  2.592962  2.593556   2.59432  2.595292   2.59652  2.598059  2.599969  2.602321  2.605193  2.608668  2.612839    2.6178  2.623651  2.630492  2.638424  2.647543  2.657942  2.669704  2.682906  2.697609  2.713864  2.731707  2.751159  2.772226  2.794901  2.819161  2.844973  2.872289  2.901054  2.931204  2.962669  2.995371  3.029231  3.064167  3.100095  3.136934  3.174599  3.213009  3.252086  3.291753  3.331936  3.372563  3.413568  3.454885        3.496452 | 100.031322  100.043088  100.058858  100.079837  100.107535  100.143825  100.191009  100.251879  100.329783  100.428691  100.553248  100.708821   100.90153  101.138259   101.42665  101.775063  102.192515  102.688595  103.273349  103.957141  104.750507  105.663988   106.70796  107.892474  109.227096  110.720769  112.381694  114.217232  116.233839   118.43702  120.831313  123.420291   126.20659  129.191946  132.377248  135.762595  139.347367  143.130282  147.109468   151.28252  155.646563  160.198299  164.934061  169.849849  174.941367  180.204058  185.633122  191.223543  196.970104  202.867398  208.909841 |  100.0 |   0.031322 | 208.909841 |
    | 14:30 | 3.718364  3.718564  3.718832  3.719188  3.719657  3.720271  3.721069  3.722096  3.723409  3.725073  3.727166  3.729777  3.733005  3.736964  3.741778  3.747582   3.75452  3.762744  3.772409  3.783675  3.796697  3.811628   3.82861  3.847776  3.869242  3.893107  3.919451   3.94833   3.97978  4.013815  4.050425  4.089581  4.131233  4.175315  4.221746   4.27043  4.321262  4.374128  4.428909  4.485481  4.543716  4.603488  4.664669  4.727134   4.79076  4.855426  4.921017   4.98742  5.054527             5.122234 | 124.906603  124.917049  124.931031  124.949606  124.974101  125.006155  125.047782  125.101419  125.169989  125.256949   125.36634  125.502825  125.671713  125.878966  126.131188  126.435584  126.799906   127.23236  127.741502  128.336101  129.024999  129.816944  130.720435  131.743553  132.893816  134.178042  135.602234  137.171494  138.889961  140.760779  142.786093  144.967066  147.303925  149.796014  152.441866  155.239287  158.185435  161.276907  164.509819   167.87989   171.38251  175.012808  178.765706  182.635979  186.618286  190.707218  194.897318  199.183113  203.559125  208.019892 |  52.920019    52.92206   52.924793   52.928422   52.933208   52.939471   52.947602   52.958078   52.971467   52.988442   53.009788   53.036409   53.069332   53.109706   53.158799   53.217988   53.288742   53.372608   53.471178   53.586064   53.718865   53.871128   54.044317   54.239772   54.458684   54.702062   54.970712   55.265222   55.585955   55.933042   56.306393   56.705704   57.130475   57.580027   58.053527   58.550007   59.068395   59.607529   60.166188   60.743106   61.336992   61.946547   62.570474   63.207495   63.856354   64.515827   65.184727   65.861907   66.546264        67.236739 |  52.183638    52.18564   52.188319   52.191878   52.196571   52.202712   52.210685   52.220957   52.234086   52.250732   52.271663   52.297768   52.330051   52.369641   52.417781    52.47582   52.545201   52.627438   52.724093   52.836749   52.966971   53.116277   53.286102   53.477762   53.692423   53.931075   54.194508   54.483299   54.797803    55.13815   55.504251   55.895808    56.31233   56.753152   57.217456   57.704296   58.212616   58.741281   59.289092   59.854806   60.437159   61.034877   61.646688   62.271338   62.907597   63.554263   64.210173   64.874203   65.545271     66.222336 | 1.088468  1.088499  1.088541  1.088596  1.088669  1.088764  1.088888  1.089047   1.08925  1.089508  1.089833  1.090237  1.090737   1.09135  1.092095  1.092993  1.094066  1.095336  1.096829  1.098566  1.100572  1.102869  1.105478  1.108417  1.111703  1.115349  1.119364  1.123755  1.128524   1.13367  1.139189  1.145073   1.15131  1.157887  1.164789  1.171998  1.179496  1.187261  1.195274  1.203513  1.211958  1.220588  1.229382  1.238321  1.247385  1.256556  1.265816   1.27515   1.28454  1.293973 | 1.359155  1.359153   1.35915  1.359147  1.359142  1.359135  1.359127  1.359117  1.359103  1.359086  1.359065  1.359038  1.359005  1.358965  1.358916  1.358857  1.358787  1.358704  1.358608  1.358496  1.358367  1.358221  1.358057  1.357873  1.357669  1.357445  1.357202  1.356939  1.356658  1.356359  1.356044  1.355714   1.35537  1.355015   1.35465  1.354276  1.353896  1.353512  1.353125  1.352736  1.352347   1.35196  1.351575  1.351194  1.350818  1.350448  1.350083  1.349726  1.349376          1.349034 | 0.536201  0.536216  0.536237  0.536266  0.536303  0.536351  0.536414  0.536495  0.536598  0.536729  0.536894  0.537099  0.537352  0.537663  0.538039  0.538493  0.539033  0.539672   0.54042  0.541289  0.542288  0.543428  0.544717  0.546162  0.547767  0.549538  0.551474  0.553575  0.555838  0.558259  0.560829  0.563542  0.566386  0.569352  0.572427  0.575599  0.578855  0.582183   0.58557  0.589005  0.592475  0.595969  0.599478  0.602992  0.606502  0.609999  0.613477  0.616929   0.62035      0.623733 | -0.201924  -0.201945  -0.201973   -0.20201  -0.202059  -0.202122  -0.202204  -0.202309  -0.202443  -0.202612  -0.202824  -0.203086  -0.203408  -0.203802  -0.204277  -0.204846  -0.205521  -0.206315   -0.20724  -0.208309  -0.209532  -0.210921  -0.212482  -0.214224   -0.21615  -0.218263  -0.220563  -0.223049  -0.225714  -0.228552  -0.231555  -0.234712  -0.238012  -0.241441  -0.244987  -0.248636  -0.252372  -0.256184  -0.260056  -0.263976  -0.267932  -0.271911  -0.275903  -0.279898  -0.283886  -0.287859  -0.291809   -0.29573  -0.299615     -0.303458 | 0.665715  0.665718  0.665721  0.665724  0.665729  0.665736  0.665744  0.665755  0.665769  0.665786  0.665807  0.665834  0.665866  0.665906  0.665953  0.666009  0.666076  0.666155  0.666246  0.666351  0.666471  0.666607   0.66676  0.666931  0.667121  0.667329  0.667556  0.667803  0.668068  0.668353  0.668655  0.668976  0.669314  0.669667  0.670036   0.67042  0.670816  0.671225  0.671645  0.672074  0.672512  0.672958  0.673411  0.673869  0.674331  0.674798  0.675267  0.675738   0.67621      0.676682 | 100.025521 | 100.030423  100.041793  100.057012  100.077234  100.103899  100.138798  100.184122  100.242532  100.317214  100.411947  100.531147  100.679922  100.864093  101.090216  101.365574  101.698142   102.09654  102.569951  103.128014  103.780707  104.538201  105.410708  106.408323  107.540867  108.817735  110.247762  111.839102  113.599132   115.53438  117.650473  119.952119  122.443098  125.126283  128.003673  131.076434  134.344959  137.808921  141.467343  145.318652  149.360743  153.591034  158.006523  162.603828  167.379231   172.32872   177.44801  182.732579  188.177683  193.778378          199.529529 | 202.437175 | 0.720758  0.720779  0.720808  0.720847  0.720897  0.720964   0.72105  0.721161  0.721303  0.721483  0.721709  0.721991   0.72234  0.722767  0.723287  0.723913  0.724661  0.725546  0.726586  0.727797  0.729195  0.730795  0.732613   0.73466  0.736949  0.739488  0.742283  0.745339  0.748657  0.752237  0.756074  0.760164  0.764499  0.769068  0.773861  0.778865  0.784067  0.789453  0.795009  0.800719   0.80657  0.812546  0.818633  0.824817  0.831086  0.837426  0.843825  0.850272  0.856756       0.863266 | 2.591451  2.591576  2.591744  2.591966   2.59226  2.592644  2.593142  2.593785  2.594605  2.595646  2.596954  2.598586  2.600604  2.603078  2.606086  2.609712  2.614045   2.61918  2.625214  2.632244  2.640367  2.649676  2.660258  2.672195  2.685555  2.700397  2.716768  2.734699  2.754208    2.7753  2.797963  2.822175  2.847901  2.875094    2.9037  2.933654  2.964889   2.99733  3.030899  3.065517  3.101102  3.137575  3.174856  3.212865  3.251526  3.290765   3.33051  3.370693  3.411248        3.452111 | 100.025521  100.035204  100.048223  100.065594  100.088602  100.118847  100.158302  100.209376  100.274977  100.358569  100.464236  100.596729  100.761505  100.964753  101.213394  101.515069  101.878094  102.311393   102.82441   103.42699  104.129247  104.941415  105.873693  106.936082   108.13823  109.489288   110.99778  112.671496  114.517406  116.541598  118.749243  121.144584  123.730941  126.510739  129.485547  132.656129  136.022502  139.583997  143.339322  147.286621  151.423537  155.747264  160.254597  164.941976  169.805527  174.841093  180.044265  185.410403   190.93466  196.611992  202.437175 |  100.0 |   0.025521 | 202.437175 |
    | 15:00 | 3.718265   3.71843  3.718651  3.718946  3.719336  3.719848  3.720515  3.721378  3.722484  3.723892  3.725669  3.727895  3.730658  3.734061  3.738217   3.74325  3.749294  3.756492  3.764991  3.774946  3.786508  3.799829  3.815055  3.832323  3.851757  3.873467  3.897546  3.924064  3.953076   3.98461  4.018677  4.055263  4.094338   4.13585  4.179734  4.225909   4.27428  4.324746  4.377196  4.431512  4.487575  4.545261  4.604447  4.665011   4.72683  4.789785  4.853761  4.918643  4.984324             5.050698 |  124.90144  124.910041   124.92159   124.93698  124.957339  124.984069  125.018897  125.063929  125.121703  125.195241    125.2881  125.404411  125.548915  125.726979  125.944596  126.208367  126.525458  126.903534  127.350673  127.875251  128.485814  129.190935  129.999058  130.918347  131.956529  133.120758  134.417489  135.852374  137.430183  139.154755   141.02897  143.054748  145.233075  147.564041  150.046902  152.680143  155.461563  158.388347  161.457151  164.664179  168.005259  171.475907  175.071395  178.786801  182.617058  186.556996  190.601375  194.744912  198.982306  203.308252 |   52.91901   52.920691   52.922948   52.925955   52.929933   52.935156    52.94196   52.950756   52.962039   52.976397   52.994521   53.017215   53.045396   53.080101   53.122485   53.173812    53.23545   53.308851   53.395529   53.497042   53.614953   53.750804   53.906082    54.08218   54.280371   54.501771   54.747321   55.017762   55.313622   55.635212   55.982623   56.355733   56.754216   57.177562   57.625092    58.09598   58.589277   59.103932   59.638813   60.192733   60.764463    61.35275   61.956335   62.573964   63.204397   63.846418   64.498841   65.160517   65.830333         66.50722 |  52.182648   52.184297    52.18651   52.189459    52.19336   52.198481   52.205152   52.213778   52.224841   52.238921   52.256693   52.278946    52.30658   52.340611   52.382172   52.432503   52.492944   52.564919   52.649914   52.749455   52.865076    52.99829   53.150552   53.323231   53.517573   53.734674   53.975455   54.240644   54.530759   54.846104   55.186768   55.552632   55.943378   56.358503   56.797342   57.259085   57.742803   58.247463   58.771958   59.315121   59.875748   60.452611   61.044475    61.65011     62.2683   62.897853   63.537607   64.186433   64.843243     65.506985 | 1.088453  1.088478  1.088513  1.088558  1.088619  1.088698  1.088802  1.088935  1.089107  1.089325  1.089601  1.089946  1.090374  1.090901  1.091544  1.092323  1.093258  1.094371  1.095684   1.09722  1.099002  1.101054  1.103396  1.106047  1.109027  1.112349  1.116026  1.120066  1.124476  1.129255  1.134404  1.139917  1.145786     1.152  1.158545  1.165407  1.172567  1.180008   1.18771  1.195653  1.203817  1.212182  1.220726  1.229431  1.238277  1.247246   1.25632  1.265482  1.274716  1.284006 | 1.359156  1.359154  1.359152  1.359149  1.359145   1.35914  1.359133  1.359124  1.359113  1.359098   1.35908  1.359057  1.359029  1.358994  1.358952  1.358901   1.35884  1.358767  1.358682  1.358583  1.358468  1.358337  1.358188  1.358021  1.357835  1.357629  1.357404   1.35716  1.356897  1.356616  1.356317  1.356003  1.355674  1.355333   1.35498  1.354617  1.354247  1.353871   1.35349  1.353106  1.352722  1.352337  1.351954  1.351573  1.351196  1.350824  1.350457  1.350096  1.349742          1.349396 | 0.536193  0.536206  0.536223  0.536246  0.536277  0.536318   0.53637  0.536438  0.536525  0.536636  0.536776  0.536951  0.537168  0.537435  0.537761  0.538154  0.538626  0.539186  0.539846  0.540616  0.541507  0.542528  0.543689  0.544998   0.54646  0.548082  0.549865  0.551811  0.553918  0.556184  0.558602  0.561166  0.563869  0.566699  0.569647    0.5727  0.575847  0.579076  0.582374   0.58573  0.589131  0.592566  0.596025  0.599498  0.602975  0.606448   0.60991  0.613352   0.61677      0.620156 | -0.201914  -0.201931  -0.201954  -0.201985  -0.202026  -0.202078  -0.202147  -0.202236  -0.202349  -0.202492  -0.202672  -0.202896  -0.203173  -0.203512  -0.203924  -0.204419   -0.20501  -0.205708  -0.206526  -0.207476   -0.20857  -0.209817  -0.211228  -0.212811  -0.214571  -0.216513  -0.218638  -0.220946  -0.223435  -0.226099  -0.228932  -0.231924  -0.235067  -0.238347  -0.241753  -0.245271  -0.248889  -0.252592  -0.256368  -0.260202  -0.264082  -0.267997  -0.271935  -0.275885  -0.279838  -0.283784  -0.287716  -0.291625  -0.295506     -0.299352 | 0.665714  0.665716  0.665719  0.665722  0.665726  0.665731  0.665739  0.665748  0.665759  0.665774  0.665792  0.665815  0.665842  0.665876  0.665918  0.665967  0.666025  0.666094  0.666175  0.666268  0.666376  0.666498  0.666636  0.666791  0.666964  0.667154  0.667364  0.667592  0.667838  0.668104  0.668387  0.668689  0.669008  0.669344  0.669695  0.670061  0.670441  0.670834  0.671239  0.671654  0.672079  0.672513  0.672953  0.673401  0.673854  0.674311  0.674772  0.675236  0.675702      0.676169 | 100.020773 | 100.024803  100.034165  100.046736  100.063489  100.085652  100.114751   100.15267  100.201705  100.264622  100.344722  100.445888  100.572641  100.730175  100.924379  101.161848   101.44987  101.796385  102.209933  102.699564  103.274741   103.94521  104.720867  105.611604  106.627167  107.776999  109.070102  110.514912  112.119189   113.88993  115.833305  117.954613  120.258267  122.747793  125.425849  128.294257  131.354051   134.60553  138.048308  141.681382  145.503185  149.511645  153.704238  158.078039  162.629761  167.355802  172.252271  177.315024  182.539685  187.921667          193.456187 | 196.256068 | 0.720747  0.720765  0.720789  0.720821  0.720863  0.720918   0.72099  0.721083  0.721203  0.721355  0.721547  0.721788  0.722086  0.722454  0.722903  0.723446  0.724097  0.724873  0.725788  0.726859  0.728101  0.729531  0.731163   0.73301  0.735085  0.737399  0.739959  0.742771   0.74584  0.749166  0.752747  0.756581   0.76066  0.764978  0.769525   0.77429   0.77926  0.784423  0.789765  0.795272   0.80093  0.806724  0.812641  0.818667  0.824787   0.83099  0.837263  0.843594  0.849973       0.856387 | 2.591389  2.591492  2.591631  2.591815  2.592059  2.592379  2.592796  2.593336  2.594027  2.594908  2.596019   2.59741  2.599137  2.601264  2.603861  2.607006  2.610781  2.615276  2.620583  2.626797  2.634011   2.64232  2.651812  2.662571  2.674673  2.688183  2.703156  2.719633  2.737644  2.757203  2.778311  2.800956  2.825115  2.850751  2.877818  2.906262  2.936022  2.967029  2.999211  3.032493  3.066797  3.102046   3.13816  3.175064   3.21268  3.250934  3.289755  3.329073  3.368821        3.408936 | 100.020773  100.028733  100.039466  100.053832   100.07292  100.098091  100.131036  100.173827   100.22898  100.299511  100.388999  100.501632  100.642259  100.816415  101.030345  101.290994  101.605989  101.983589  102.432615  102.962355  103.582448  104.302755  105.133212  106.083681  107.163794  108.382813  109.749489  111.271949  112.957593  114.813018  116.843968  119.055303  121.450988  124.034109  126.806895  129.770763  132.926362  136.273633  139.811866  143.539759  147.455479  151.556714  155.840725  160.304395  164.944271  169.756596  174.737346  179.882254  185.186834  190.646398  196.256068 |  100.0 |   0.020773 | 196.256068 |
    | 15:30 | 3.718184  3.718319  3.718502  3.718746   3.71907  3.719496  3.720054  3.720777  3.721708  3.722896  3.724403  3.726296  3.728657  3.731576  3.735156  3.739511  3.744764  3.751047  3.758502  3.767274  3.777511  3.789362  3.802974  3.818486  3.836027  3.855717  3.877658  3.901935  3.928616  3.957747  3.989353   4.02344  4.059993  4.098978  4.140345  4.184026  4.229943  4.278002  4.328104  4.380141  4.433999   4.48956  4.546707  4.605319  4.665277  4.726463  4.788762  4.852061  4.916251             4.981226 | 124.897211  124.904286  124.913813  124.926548  124.943447  124.965705  124.994803  125.032553  125.081154  125.143239  125.221926  125.320863  125.444266  125.596943  125.784308  126.012373  126.287727  126.617482  127.009212  127.470854    128.0106   128.63677  129.357671  130.181448  131.115938  132.168528  133.346022  134.654529  136.099368  137.684999   139.41498  141.291943  143.317603  145.492781  147.817446  150.290772  152.911207  155.676545  158.584001  161.630291  164.811705  168.124175  171.563341  175.124612  178.803212  182.594231  186.492658  190.493416  194.591388  198.781438 |  52.918184   52.919566   52.921428   52.923917   52.927219   52.931568   52.937253   52.944627    52.95412   52.966244   52.981606   53.000915   53.024989   53.054759    53.09127   53.135678   53.189245   53.253326   53.329349   53.418803   53.523202   53.644065   53.782878   53.941066   54.119956   54.320753   54.544506   54.792089   55.064181   55.361256   55.683577   56.031197   56.403966   56.801539   57.223398   57.668863   58.137119   58.627233   59.138177   59.668849   60.218092   60.784713   61.367498   61.965225   62.576679   63.200658   63.835985   64.481511    65.13612        65.798735 |  52.181838   52.183194   52.185019    52.18746   52.190698   52.194962   52.200537   52.207768   52.217077   52.228965   52.244029   52.262963   52.286569   52.315761   52.351563   52.395109   52.447636   52.510472   52.585019   52.672736   52.775108   52.893623   53.029741   53.184857   53.360273   53.557171   53.776579   54.019354   54.286162   54.577468    54.89353   55.234399   55.599929   55.989781   56.403448   56.840263   57.299425   57.780022   58.281043    58.80141   59.339987   59.895605   60.467072   61.053192   61.652772   62.264634   62.887623   63.520613   64.162511     64.812258 |  1.08844  1.088461   1.08849  1.088527  1.088578  1.088644   1.08873  1.088842  1.088987  1.089171  1.089404  1.089698  1.090064  1.090516   1.09107  1.091744  1.092557  1.093529  1.094681  1.096036  1.097616  1.099442  1.101538  1.103923  1.106616  1.109633   1.11299  1.116695  1.120759  1.125184  1.129973  1.135123  1.140629  1.146482  1.152671  1.159184  1.166005  1.173117  1.180502  1.188142  1.196016  1.204106  1.212391  1.220852  1.229469  1.238225  1.247101   1.25608  1.265145  1.274281 | 1.359157  1.359156  1.359154  1.359151  1.359148  1.359143  1.359138   1.35913  1.359121  1.359108  1.359093  1.359074   1.35905   1.35902  1.358983  1.358939  1.358886  1.358822  1.358747  1.358659  1.358557   1.35844  1.358306  1.358155  1.357985  1.357797   1.35759  1.357363  1.357118  1.356855  1.356574  1.356276  1.355963  1.355636  1.355296  1.354946  1.354586  1.354219  1.353846  1.353469  1.353089  1.352708  1.352327  1.351948  1.351572  1.351198   1.35083  1.350467   1.35011          1.349759 | 0.536186  0.536197  0.536211  0.536231  0.536256   0.53629  0.536334  0.536391  0.536464  0.536558  0.536676  0.536825  0.537011   0.53724  0.537521  0.537862  0.538273  0.538763  0.539343  0.540023  0.540814  0.541726  0.542768   0.54395  0.545277  0.546757  0.548394  0.550189  0.552143  0.554256  0.556522  0.558938  0.561496  0.564187  0.567003  0.569932  0.572964  0.576087  0.579289  0.582558  0.585882   0.58925  0.592651  0.596075  0.599513  0.602955  0.606392  0.609819  0.613227       0.61661 | -0.201905   -0.20192  -0.201939  -0.201964  -0.201998  -0.202042    -0.2021  -0.202174  -0.202269   -0.20239  -0.202543  -0.202734  -0.202972  -0.203263  -0.203619  -0.204049  -0.204565  -0.205177  -0.205899   -0.20674  -0.207714  -0.208832  -0.210102  -0.211535  -0.213138  -0.214915  -0.216872  -0.219008  -0.221323  -0.223814  -0.226476  -0.229303  -0.232284  -0.235411  -0.238672  -0.242054  -0.245546  -0.249133  -0.252803  -0.256543   -0.26034  -0.264182  -0.268057  -0.271954  -0.275863  -0.279775   -0.28368  -0.287571  -0.291441     -0.295282 | 0.665713  0.665715  0.665717   0.66572  0.665723  0.665728  0.665734  0.665741  0.665751  0.665763  0.665779  0.665798  0.665822  0.665851  0.665887   0.66593  0.665981  0.666042  0.666113  0.666195  0.666291    0.6664  0.666525  0.666665  0.666822  0.666996  0.667188  0.667398  0.667626  0.667873  0.668138  0.668421  0.668722  0.669039  0.669373  0.669722  0.670085  0.670462  0.670851  0.671252  0.671663  0.672083  0.672512  0.672948  0.673391  0.673839  0.674291  0.674747  0.675206      0.675666 | 100.016892 |   100.0202  100.027901   100.03827  100.052133  100.070529  100.094759  100.126437   100.16754  100.220462  100.288078  100.373791  100.481589  100.616084  100.782547  100.986924  101.235837  101.536566  101.897005    102.3256  102.831265  103.423276  104.111145  104.904493  105.812902   106.84577  108.012177  109.320747  110.779534  112.395925  114.176556  116.127259  118.253027  120.557995  123.045453  125.717857  128.576871  131.623405   134.85767  138.279231  141.887061  145.679602  149.654815  153.810233  158.143005  162.649938  167.327536  172.172029  177.179406  182.345429           187.66566 | 190.358854 | 0.720738  0.720753  0.720773  0.720799  0.720834   0.72088   0.72094  0.721019  0.721119  0.721248   0.72141  0.721615   0.72187  0.722186  0.722572  0.723042  0.723609  0.724286  0.725089  0.726034  0.727135  0.728408  0.729868   0.73153  0.733406  0.735508  0.737845  0.740425  0.743254  0.746333  0.749665  0.753247  0.757075  0.761144  0.765445  0.769969  0.774705  0.779642  0.784766  0.790064  0.795524   0.80113  0.806869  0.812728  0.818693  0.824751   0.83089  0.837097  0.843362       0.849672 | 2.591339  2.591423  2.591538   2.59169  2.591893  2.592159  2.592508   2.59296  2.593542  2.594285  2.595227  2.596411  2.597886  2.599711  2.601948  2.604669  2.607951  2.611876  2.616532  2.622008  2.628398  2.635792  2.644281   2.65395  2.664878  2.677138   2.69079  2.705885   2.72246  2.740542  2.760143   2.78126  2.803882  2.827982  2.853524  2.880464  2.908746   2.93831  2.969091  3.001017  3.034015  3.068011  3.102929  3.138692  3.175226  3.212457  3.250313  3.288725  3.327624        3.366947 | 100.016892  100.023427  100.032265  100.044131  100.059945  100.080866  100.108336  100.144135  100.190433   100.24985  100.325508   100.42109  100.540886  100.689829  100.873524  101.098258  101.370989  101.699317  102.091433  102.556043  103.102273  103.739552  104.477486  105.325715  106.293769  107.390926  108.626071  110.007575  111.543183  113.239926   115.10405  117.140976   119.35527   121.75064  124.329953  127.095258  130.047828  133.188211  136.516278  140.031283   143.73192  147.616376  151.682382   155.92727  160.348005  164.941234  169.703316  174.630354  179.718216  184.962561  190.358854 |  100.0 |   0.016892 | 190.358854 |
    | 16:00 | 3.718118  3.718229  3.718379  3.718581  3.718849  3.719204  3.719669  3.720275  3.721056  3.722059  3.723333  3.724941  3.726954  3.729453  3.732531   3.73629  3.740844  3.746317  3.752838  3.760548  3.769588  3.780102  3.792236  3.806129  3.821916   3.83972  3.859653  3.881812  3.906276  3.933105  3.962342  3.994009  4.028106  4.064616  4.103505   4.14472  4.188195  4.233852    4.2816   4.33134  4.382969  4.436374  4.491442  4.548059  4.606107  4.665471  4.726037  4.787694  4.850331             4.913842 | 124.893752  124.899564  124.907413  124.917938  124.931947  124.950457  124.974733  125.006332  125.047153  125.099484  125.166049  125.250058  125.355244  125.485898  125.646887   125.84366   126.08224  126.369189   126.71156  127.116822  127.592767    128.1474  128.788813  129.525049  130.363958  131.313063  132.379419  133.569497  134.889077  136.343164  137.935926  139.670655  141.549756  143.574754  145.746317  148.064305  150.527821  153.135279  155.884473  158.772653  161.796592  164.952665  168.236904  171.645071  175.172704   178.81517  182.567706  186.425455  190.383496   194.43687 |  52.917508   52.918643   52.920177   52.922234   52.924972   52.928589   52.933332   52.939505   52.947479     52.9577   52.970697   52.987097   53.007623   53.033108   53.064494    53.10283   53.149275   53.205083   53.271593   53.350214   53.442401   53.549631   53.673372   53.815057    53.97605   54.157617   54.360895   54.586869   54.836351   55.109961   55.408123   55.731056   56.078778   56.451112     56.8477   57.268014   57.711378   58.176985    58.66392   59.171178   59.697684   60.242312   60.803904    61.38128   61.973257   62.578657   63.196313   63.825086   64.463861        65.111556 |  52.181175   52.182289   52.183793    52.18581   52.188494   52.192041   52.196692   52.202746   52.210565   52.220587   52.233332   52.249413   52.269541   52.294531   52.325307   52.362899   52.408442   52.463166   52.528385   52.605479   52.695876   52.801023   52.922361   53.061295   53.219162   53.397203   53.596533   53.818119   54.062756   54.331053   54.623425   54.940086   55.281056    55.64616   56.035046   56.447197   56.881952   57.338517   57.815997   58.313404   58.829685   59.363737   59.914423   60.480587   61.061069   61.654711   62.260374   62.876936   63.503306     64.138424 |  1.08843  1.088447  1.088471  1.088502  1.088543  1.088598  1.088671  1.088764  1.088886  1.089041  1.089239  1.089488    1.0898  1.090187  1.090664  1.091246  1.091951  1.092798  1.093806  1.094997  1.096393  1.098015  1.099885  1.102023   1.10445  1.107182  1.110236  1.113624  1.117357  1.121442  1.125881  1.130677  1.135827  1.141324   1.14716  1.153325  1.159805  1.166585  1.173648  1.180978  1.188556  1.196363  1.204379  1.212586  1.220965  1.229497  1.238164  1.246949  1.255835  1.264805 | 1.359158  1.359156  1.359155  1.359153   1.35915  1.359146  1.359142  1.359135  1.359127  1.359117  1.359104  1.359088  1.359067  1.359041   1.35901  1.358972  1.358925   1.35887  1.358804  1.358726  1.358636  1.358531  1.358411  1.358275  1.358121   1.35795   1.35776  1.357551  1.357323  1.357077  1.356814  1.356533  1.356236  1.355924  1.355598  1.355261  1.354913  1.354556  1.354192  1.353822  1.353449  1.353073  1.352696  1.352319  1.351943   1.35157  1.351201  1.350836  1.350477          1.350123 | 0.536181   0.53619  0.536202  0.536218  0.536239  0.536267  0.536304  0.536351  0.536413  0.536492  0.536592  0.536719  0.536877  0.537073  0.537315   0.53761  0.537966  0.538394  0.538902  0.539501  0.540202  0.541014  0.541946  0.543009   0.54421  0.545556  0.547052  0.548702  0.550508   0.55247  0.554587  0.556854  0.559266  0.561817  0.564497  0.567298   0.57021   0.57322  0.576319  0.579494  0.582734  0.586027  0.589363  0.592731  0.596121  0.599524  0.602931  0.606334  0.609726        0.6131 | -0.201898   -0.20191  -0.201926  -0.201947  -0.201975  -0.202012   -0.20206  -0.202122  -0.202202  -0.202305  -0.202434  -0.202597    -0.2028   -0.20305  -0.203357   -0.20373  -0.204179  -0.204714  -0.205348  -0.206092  -0.206957  -0.207954  -0.209095  -0.210387  -0.211841  -0.213463  -0.215257  -0.217226  -0.219372  -0.221693  -0.224186  -0.226845  -0.229664  -0.232634  -0.235745  -0.238986  -0.242346   -0.24581  -0.249368  -0.253005   -0.25671  -0.260471  -0.264275  -0.268111  -0.271968  -0.275837  -0.279708  -0.283574  -0.287425     -0.291256 | 0.665713  0.665714  0.665716  0.665718  0.665721  0.665725  0.665729  0.665736  0.665744  0.665755  0.665768  0.665784  0.665805   0.66583  0.665861  0.665898  0.665942  0.665995  0.666058  0.666131  0.666216  0.666314  0.666425  0.666552  0.666694  0.666852  0.667028  0.667221  0.667432  0.667661  0.667907  0.668172  0.668454  0.668754  0.669069  0.669401  0.669747  0.670108  0.670481  0.670867  0.671264  0.671671  0.672087  0.672512  0.672943  0.673381  0.673824  0.674271  0.674722      0.675176 | 100.013722 | 100.016435  100.022761  100.031305  100.042761   100.05801   100.07816  100.104587   100.13899  100.183437  100.240424  100.312923  100.404439  100.519056  100.661468  100.837015  101.051687  101.312121  101.625579  101.999895  102.443416  102.964909  103.573458  104.278342  105.088907  106.014427  107.063968  108.246253  109.569544  111.041533  112.669247  114.458983  116.416257  118.545769  120.851399  123.336214  126.002486  128.851728  131.884739  135.101651   138.50198  142.084685  145.848217   149.79057  153.909331  158.201722  162.664641  167.294693  172.088225   177.04135          182.149965 | 184.737719 | 0.720731  0.720743  0.720759  0.720781   0.72081  0.720848  0.720899  0.720964  0.721049  0.721157  0.721295  0.721469  0.721686  0.721956  0.722289  0.722694  0.723186  0.723776  0.724479   0.72531  0.726283  0.727413  0.728716  0.730206  0.731897    0.7338  0.735927  0.738287  0.740886  0.743729  0.746818  0.750155  0.753736  0.757558  0.761615  0.765899  0.770399  0.775107   0.78001  0.785096  0.790352  0.795764   0.80132  0.807005  0.812807  0.818712  0.824709  0.830785  0.836928       0.843127 | 2.591297  2.591367  2.591461  2.591587  2.591755  2.591977  2.592267  2.592646  2.593135  2.593761  2.594558  2.595564  2.596822  2.598384  2.600307  2.602657  2.605502  2.608921  2.612995  2.617809  2.623453  2.630015  2.637584  2.646248  2.656088  2.667178  2.679588  2.693373  2.708582  2.725248  2.743393  2.763028  2.784149  2.806741  2.830778  2.856224  2.883033  2.911152  2.940522  2.971077   3.00275  3.035469  3.069162  3.103754  3.139172  3.175344  3.212198  3.249664  3.287675        3.326165 | 100.013722  100.019081   100.02635  100.036139  100.049225   100.06659  100.089464   100.11937  100.158177  100.208152  100.272014  100.352988  100.454856  100.581998  100.739425    100.9328  101.168441  101.453305  101.794957  102.201508  102.681541  103.244011  103.898136  104.653266  105.518748  106.503789   107.61732  108.867866  110.263432  111.811399  113.518448  115.390496  117.432656  119.649221  122.043659  124.618632  127.376022  130.316973  133.441936  136.750721  140.242551  143.916116  147.769626  151.800858  156.007205  160.385718  164.933138  169.645935  174.520331  179.552327  184.737719 |  100.0 |   0.013722 | 184.737719 |
    | 16:30 | 3.718063  3.718155  3.718279  3.718445  3.718667  3.718962  3.719349  3.719855  3.720511  3.721355  3.722431  3.723795  3.725508  3.727643  3.730283  3.733521  3.737461  3.742216  3.747908  3.754666  3.762628  3.771931  3.782717  3.795126  3.809292  3.825344  3.843399  3.863564  3.885928  3.910565  3.937531  3.966863  3.998578  4.032675  4.069135   4.10792  4.148978  4.192244   4.23764  4.285077  4.334459  4.385683  4.438641  4.493225   4.54932  4.606814  4.665596  4.725555  4.786582             4.848571 | 124.890926  124.895694  124.902154  124.910842   124.92244  124.937812  124.958037   124.98445  125.018685  125.062724  125.118942  125.190152  125.279649  125.391244  125.529295  125.698719  125.904991  126.154133  126.452673  126.807592  127.226248  127.716283   128.28551  128.941795  129.692917  130.546442  131.509583  132.589081  133.791085  135.121063  136.583722  138.182955  139.921808  141.802474  143.826299  145.993815  148.304781  150.758237   153.35257  156.085579  158.954547  161.956307  165.087314  168.343702  171.721345  175.215908  178.822894  182.537682  186.355561   190.27176 |  52.916955   52.917887    52.91915   52.920847   52.923114   52.926118    52.93007    52.93523   52.941918    52.95052   52.961499   52.975403   52.992872   53.014646    53.04157   53.074595   53.114774   53.163263   53.221308   53.290234   53.371423     53.4663   53.576299   53.702843    53.84731   54.011007   54.195137   54.400774   54.628842   54.880094   55.155096   55.454224   55.777656   56.125379   56.497193   56.892725   57.311443   57.752673   58.215619   58.699382    59.20298   59.725364   60.265439   60.822079    61.39414   61.980473   62.579934   63.191395   63.813749        64.445915 |  52.180633   52.181547   52.182785    52.18445   52.186673   52.189618   52.193493   52.198554   52.205112   52.213547   52.224313   52.237946   52.255076   52.276428   52.302829   52.335212   52.374611   52.422158   52.479076   52.546663   52.626276    52.71931   52.827173    52.95126   53.092922    53.25344   53.433994   53.635638   53.859277   54.105649   54.375311    54.66863   54.985781   55.326752   55.691346   56.079197   56.489783   56.922445   57.376401    57.85077   58.344588   58.856828   59.386415   59.932245   60.493197   61.068144   61.655964   62.255551   62.865819     63.485709 | 1.088422  1.088436  1.088455  1.088481  1.088515  1.088561  1.088621  1.088699  1.088801  1.088932  1.089099   1.08931  1.089576  1.089907  1.090316  1.090817  1.091427  1.092163  1.093044  1.094089  1.095319  1.096755  1.098418   1.10033   1.10251  1.104976  1.107746  1.110834  1.114253  1.118011  1.122114  1.126567  1.131368  1.136516  1.142003  1.147821   1.15396  1.160407  1.167146  1.174162  1.181437  1.188954  1.196694  1.204638  1.212769  1.221067  1.229515  1.238095  1.246791  1.255586 | 1.359158  1.359157  1.359156  1.359154  1.359152  1.359149  1.359145   1.35914  1.359133  1.359124  1.359113  1.359099  1.359082   1.35906  1.359033     1.359   1.35896  1.358911  1.358854  1.358786  1.358706  1.358613  1.358505  1.358383  1.358244  1.358088  1.357914  1.357723  1.357512  1.357284  1.357037  1.356773  1.356493  1.356196  1.355886  1.355562  1.355226  1.354881  1.354527  1.354166    1.3538  1.353429  1.353057  1.352683   1.35231  1.351939   1.35157  1.351204  1.350843          1.350486 | 0.536177  0.536184  0.536194  0.536207  0.536224  0.536248  0.536278  0.536318   0.53637  0.536436  0.536521  0.536629  0.536763  0.536931  0.537139  0.537393  0.537702  0.538074  0.538518  0.539044  0.539663  0.540383  0.541215  0.542168   0.54325   0.54447  0.545833  0.547344  0.549007  0.550823  0.552792  0.554912  0.557179  0.559587   0.56213  0.564799  0.567585  0.570479  0.573468  0.576543  0.579691  0.582902  0.586166   0.58947  0.592805  0.596162  0.599531  0.602904  0.606273      0.609631 | -0.201893  -0.201902  -0.201915  -0.201933  -0.201956  -0.201986  -0.202027  -0.202079  -0.202146  -0.202232  -0.202342  -0.202481  -0.202654  -0.202868  -0.203132  -0.203454  -0.203844  -0.204311  -0.204867  -0.205522  -0.206287  -0.207175  -0.208195  -0.209358  -0.210672  -0.212146  -0.213785  -0.215594  -0.217576   -0.21973  -0.222056   -0.22455  -0.227206  -0.230017  -0.232976  -0.236071  -0.239292  -0.242628  -0.246066  -0.249594  -0.253199   -0.25687  -0.260595  -0.264361  -0.268159  -0.271977  -0.275807  -0.279639  -0.283465     -0.287278 | 0.665712  0.665713  0.665714  0.665716  0.665719  0.665722  0.665726  0.665731  0.665738  0.665747  0.665758  0.665772   0.66579  0.665811  0.665838   0.66587  0.665909  0.665955   0.66601  0.666075   0.66615  0.666237  0.666337   0.66645  0.666578  0.666722  0.666882  0.667059  0.667253  0.667465  0.667694  0.667941  0.668205  0.668487  0.668785  0.669099  0.669428  0.669772   0.67013    0.6705  0.670882  0.671275  0.671678   0.67209   0.67251  0.672937   0.67337  0.673808  0.674251      0.674697 | 100.011136 | 100.013359  100.018549  100.025581  100.035037  100.047661  100.064395  100.086412  100.115166   100.15244  100.200393  100.261615  100.339178   100.43668   100.55829  100.708782  100.893551   101.11862  101.390633   101.71682  102.104949  102.563255  103.100354  103.725133  104.446639  105.273944   106.21602  107.281608  108.479091  109.816381   111.30082    112.9391  114.737194  116.700318    118.8329   121.13858  123.620217   126.27991  129.119037   132.13829  135.337731  138.716836  142.274548  146.009329  149.919209  154.001826  158.254475  162.674137  167.257518  172.001076          176.901041 | 179.384754 | 0.720725  0.720735  0.720748  0.720766   0.72079  0.720822  0.720864  0.720919   0.72099  0.721081  0.721197  0.721345   0.72153  0.721761  0.722046  0.722396  0.722821  0.723334  0.723948  0.724676  0.725534  0.726535  0.727694  0.729026  0.730545  0.732264  0.734193  0.736344  0.738724  0.741341  0.744197  0.747295  0.750636  0.754215  0.758031  0.762075   0.76634  0.770818  0.775497  0.780367  0.785414  0.790627  0.795993  0.801499  0.807131  0.812877  0.818725  0.824661  0.830675       0.836755 | 2.591263   2.59132  2.591398  2.591502  2.591641  2.591825  2.592067  2.592384  2.592794  2.593321  2.593994  2.594847  2.595918  2.597252  2.598902  2.600926  2.603388  2.606359  2.609915  2.614136  2.619108  2.624915  2.631646  2.639387   2.64822  2.658223  2.669469  2.682021  2.695933  2.711247  2.727996  2.746198   2.76586  2.786978  2.809535  2.833505  2.858851  2.885528  2.913484  2.942659  2.972992  3.004414  3.036858  3.070251  3.104524  3.139604   3.17542  3.211905  3.248989        3.286608 | 100.011136  100.015526  100.021498  100.029564  100.040378  100.054773  100.073794  100.098743  100.131224  100.173194  100.227012  100.295497   100.38197  100.490305  100.624964  100.791028  100.994206  101.240837  101.537868  101.892814  102.313698  102.808973   103.38742  104.058041   104.82993  105.712146  106.713582  107.842831  109.108072  110.516958  112.076526  113.793124  115.672356  117.719048  119.937236  122.330163  124.900302   127.64938   130.57842  133.687787  136.977233  140.445956  144.092645  147.915532   151.91244  156.080824  160.417813  164.920242  169.584685  174.407481  179.384754 |  100.0 |   0.011136 | 179.384754 |
    | 17:00 | 3.718019  3.718094  3.718196  3.718333  3.718517  3.718761  3.719084  3.719506  3.720055  3.720764  3.721672  3.722826  3.724281  3.726102  3.728362  3.731146  3.734547  3.738669  3.743624  3.749535  3.756529  3.764739  3.774301  3.785353  3.798029   3.81246  3.828766  3.847062  3.867446  3.890004  3.914802  3.941893  3.971309  4.003062   4.03715  4.073551  4.112226  4.153124  4.196178  4.241312  4.288438  4.337464  4.388288  4.440806  4.494911  4.550494  4.607445  4.665656   4.72502              4.78543 | 124.888618  124.892527  124.897837  124.904999   124.91459   124.92734  124.944167  124.966214  124.994883  125.031888  125.079291   125.13955  125.215564  125.310705   125.42886  125.574442  125.752409  125.968255  126.227989  126.538096  126.905477  127.337374  127.841272  128.424792  129.095571  129.861133  130.728762  131.705374  132.797404  134.010697  135.350423  136.821008  138.426085  140.168471  142.050161  144.072338  146.235406  148.539034  150.982207  153.563289  156.280088   159.12992  162.109679  165.215899   168.44481  171.792397  175.254446   178.82659  182.504344  186.283138 |  52.916504   52.917268   52.918306   52.919706    52.92158   52.924071    52.92736   52.931667   52.937268   52.944497   52.953756   52.965524   52.980364   52.998933   53.021984   53.050372   53.085056   53.127091   53.177629   53.237906   53.309228   53.392953   53.490472    53.60318   53.732451   53.879611    54.04591   54.232491   54.440369   54.670409   54.923307   55.199581   55.499559   55.823385   56.171015   56.542229   56.936642   57.353717   57.792785    58.25306   58.733661   59.233625   59.751933   60.287516   60.839281   61.406117   61.986908   62.580545   63.185934           63.802 |  52.180191    52.18094   52.181958   52.183331   52.185168   52.187612   52.190836    52.19506   52.200552   52.207641    52.21672   52.228259   52.242811   52.261019   52.283623    52.31146    52.34547   52.386688   52.436245   52.495351   52.565288   52.647388   52.743013   52.853532   52.980293   53.124596   53.287665   53.470623   53.674464   53.900037   54.148024   54.418932   54.713085   55.030623   55.371502   55.735507    56.12226   56.531236   56.961778   57.413115   57.884383   58.374638    58.88288   59.408063   59.949113   60.504941   61.074454   61.656563   62.250196     62.854298 | 1.088415  1.088426  1.088442  1.088463  1.088492   1.08853   1.08858  1.088645   1.08873   1.08884  1.088981   1.08916  1.089386  1.089668  1.090018  1.090449  1.090976  1.091614  1.092381  1.093295  1.094376  1.095645   1.09712  1.098825  1.100777  1.102997  1.105502  1.108308  1.111428  1.114875  1.118656  1.122777  1.127241  1.132046   1.13719  1.142666  1.148466  1.154579  1.160992   1.16769  1.174658  1.181879  1.189335  1.197009  1.204883  1.212938  1.221158  1.229523  1.238019  1.246627 | 1.359159  1.359158  1.359157  1.359155  1.359153  1.359151  1.359148  1.359143  1.359138   1.35913  1.359121  1.359109  1.359094  1.359076  1.359053  1.359024  1.358989  1.358947  1.358897  1.358837  1.358767  1.358684  1.358589  1.358479  1.358354  1.358213  1.358055  1.357879  1.357686  1.357474  1.357245  1.356998  1.356734  1.356453  1.356158  1.355848  1.355526  1.355193   1.35485  1.354498  1.354141  1.353778  1.353411  1.353042  1.352672  1.352303  1.351935  1.351569  1.351207          1.350849 | 0.536173  0.536179  0.536187  0.536198  0.536213  0.536232  0.536257  0.536291  0.536334   0.53639  0.536461  0.536552  0.536667   0.53681  0.536988  0.537206  0.537473  0.537796  0.538184  0.538645  0.539189  0.539826  0.540566  0.541418   0.54239  0.543491  0.544729  0.546108  0.547634  0.549308  0.551134  0.553109  0.555231  0.557497  0.559901  0.562436  0.565093  0.567865   0.57074  0.573708  0.576759  0.579881  0.583064  0.586298  0.589571  0.592874  0.596199  0.599534  0.602874       0.60621 | -0.201888  -0.201896  -0.201907  -0.201921   -0.20194  -0.201966  -0.201999  -0.202043  -0.202099  -0.202172  -0.202265  -0.202382  -0.202529  -0.202713  -0.202939  -0.203217  -0.203554  -0.203961  -0.204447  -0.205022  -0.205698  -0.206485  -0.207395  -0.208437  -0.209621  -0.210956  -0.212449  -0.214105  -0.215928  -0.217921  -0.220083  -0.222413  -0.224906  -0.227558  -0.230362  -0.233308  -0.236387  -0.239588  -0.242901  -0.246312  -0.249811  -0.253385  -0.257022  -0.260712  -0.264442  -0.268202  -0.271983  -0.275774  -0.279568     -0.283356 | 0.665712  0.665712  0.665714  0.665715  0.665717   0.66572  0.665723  0.665728  0.665734  0.665741   0.66575  0.665762  0.665777  0.665796  0.665819  0.665846   0.66588   0.66592  0.665968  0.666025  0.666092  0.666169  0.666258  0.666359  0.666475  0.666605  0.666751  0.666912   0.66709  0.667285  0.667497  0.667727  0.667974  0.668238  0.668518  0.668815  0.669127  0.669454  0.669796  0.670151  0.670518  0.670897  0.671286  0.671685  0.672093  0.672509  0.672931  0.673359  0.673793      0.674231 | 100.009028 | 100.010847  100.015102  100.020881  100.028677  100.039116  100.052995  100.071313  100.095313  100.126525  100.166816  100.218433   100.28406   100.36686   100.47052   100.59929   100.75801  100.952121  101.187674  101.471304  101.810202  102.212059  102.684992   103.23746  103.878152  104.615878  105.459445  106.417528  107.498551  108.710566  110.061154  111.557322  113.205442   115.01118   116.97947  119.114485   121.41964  123.897599  126.550303  129.378999  132.384288  135.566164  138.924067   142.45693  146.163227   150.04102  154.088002  158.301534  162.678681  167.216244          171.910786 | 174.291992 |  0.72072  0.720728  0.720739  0.720754  0.720774  0.720801  0.720835  0.720881   0.72094  0.721017  0.721115   0.72124  0.721397  0.721594  0.721838  0.722139  0.722506  0.722951  0.723486  0.724123  0.724877  0.725761   0.72679  0.727977  0.729338  0.730885   0.73263  0.734584  0.736758  0.739158   0.74179  0.744658  0.747764  0.751107  0.754684  0.758492  0.762523   0.76677  0.771224  0.775875  0.780711  0.785721  0.790892  0.796212  0.801669  0.807249   0.81294  0.818731  0.824608       0.830562 | 2.591236  2.591283  2.591346  2.591432  2.591547    2.5917  2.591901  2.592165  2.592509  2.592952   2.59352  2.594241  2.595151  2.596289  2.597702  2.599442  2.601567  2.604143  2.607239  2.610932    2.6153  2.620426  2.626395  2.633291  2.641197  2.650194  2.660356   2.67175  2.684437  2.698467   2.71388  2.730704  2.748955  2.768638  2.789748  2.812266  2.836165  2.861409  2.887952  2.915743  2.944725  2.974836  3.006011  3.038183  3.071283  3.105241  3.139988  3.175457  3.211579        3.248289 | 100.009028  100.012621  100.017521  100.024159  100.033086  100.045004    100.0608  100.081585  100.108732  100.143927  100.189212   100.24704   100.32032  100.412466  100.527439  100.669775  100.844613    101.0577  101.315386  101.624596  101.992787  102.427887  102.938208  103.532355  104.219111  105.007319  105.905754  106.922999  108.067324  109.346573  110.768064  112.338504  114.063926  115.949638  118.000198  120.219399  122.610275  125.175121  127.915522  130.832388  133.926005  137.196077  140.641775  144.261792  148.054383  152.017415  156.148403  160.444553  164.902789  169.519786  174.291992 |  100.0 |   0.009028 | 174.291992 |
    | 17:30 | 3.717983  3.718044  3.718128  3.718241  3.718393  3.718595  3.718863  3.719215  3.719675  3.720269  3.721034  3.722009  3.723243  3.724793  3.726724  3.729112  3.732042  3.735607  3.739912  3.745069  3.751197  3.758424  3.766879  3.776696  3.788007  3.800943  3.815629  3.832181  3.850707  3.871299  3.894038  3.918987  3.946191  3.975679  4.007462  4.041532  4.077866  4.116426  4.157159  4.199998   4.24487  4.291687   4.34036  4.390788  4.442872  4.496507  4.551586  4.608003  4.665654             4.724434 | 124.886737  124.889937  124.894297  124.900195  124.908115  124.918677  124.932659  124.951036  124.975011  125.006058  125.045966  125.096876  125.161327  125.242296  125.343232  125.468084  125.621318  125.807923    126.0334  126.303739  126.625373  127.005116  127.450086  127.967609  128.565112  129.250006  130.029561  130.910789  131.900318  133.004289  134.228255  135.577105  137.054997  138.665323  140.410684  142.292889  144.312972   146.47122  148.767221  151.199911  153.767636  156.468213  159.298998  162.256941  165.338653  168.540459   171.85845  175.288529  178.826454   182.46787 |  52.916137   52.916762   52.917614   52.918767   52.920315   52.922379   52.925111   52.928702   52.933386   52.939452   52.947247    52.95719   52.969775   52.985582   53.005279   53.029634    53.05951    53.09587   53.139771   53.192359   53.254858   53.328555   53.414782   53.514894   53.630247   53.762169   53.911934   54.080735   54.269659   54.479663   54.711556   54.965983   55.243413   55.544133   55.868253   56.215702   56.586241   56.979476   57.394866   57.831748   58.289346   58.766795   59.263156   59.777432   60.308585    60.85555   61.417249   61.992599   62.580523        63.179958 |  52.179831   52.180444    52.18128    52.18241   52.183928   52.185952   52.188631   52.192152   52.196745   52.202693   52.210337   52.220087   52.232428   52.247927   52.267243   52.291124    52.32042   52.356074   52.399122   52.450689   52.511974    52.58424   52.668793   52.766961   52.880074   53.009434   53.156291   53.321813   53.507068   53.712994   53.940385   54.189871   54.461913   54.756794   55.074619   55.415321   55.778665   56.164263   56.571587   56.999984   57.448696   57.916874   58.403595   58.907884   59.428723   59.965066   60.515857   61.080034   61.656541     62.244336 | 1.088409  1.088419  1.088432  1.088449  1.088473  1.088504  1.088546    1.0886  1.088671  1.088764  1.088882  1.089033  1.089225  1.089465  1.089764  1.090134  1.090588   1.09114  1.091807  1.092605  1.093552  1.094669  1.095975   1.09749  1.099234  1.101225  1.103484  1.106026  1.108866  1.112018  1.115491  1.119293   1.12343  1.127903  1.132711   1.13785  1.143315  1.149095  1.155181   1.16156  1.168217  1.175137  1.182304  1.189701  1.197311  1.205115  1.213096  1.221238  1.229523  1.237935 | 1.359159  1.359158  1.359157  1.359156  1.359155  1.359153   1.35915  1.359146  1.359142  1.359135  1.359128  1.359118  1.359105  1.359089  1.359069  1.359045  1.359015  1.358979  1.358935  1.358883  1.358821  1.358748  1.358663  1.358565  1.358453  1.358326  1.358182  1.358022  1.357845   1.35765  1.357437  1.357206  1.356959  1.356695  1.356415   1.35612  1.355812  1.355492   1.35516   1.35482  1.354471  1.354116  1.353756  1.353393  1.353028  1.352661  1.352295  1.351931  1.351569          1.351211 | 0.536171  0.536175  0.536182  0.536191  0.536203  0.536219   0.53624  0.536268  0.536304  0.536351  0.536411  0.536488  0.536585  0.536707  0.536859  0.537047  0.537277  0.537556  0.537893  0.538296  0.538774  0.539337  0.539992  0.540751  0.541622  0.542613  0.543732  0.544987  0.546382  0.547921  0.549607   0.55144   0.55342  0.555545  0.557809  0.560208  0.562734   0.56538  0.568136  0.570993   0.57394  0.576967  0.580064   0.58322  0.586424  0.589667  0.592939  0.596231  0.599534      0.602842 | -0.201884  -0.201891  -0.201899  -0.201911  -0.201927  -0.201948  -0.201976  -0.202013   -0.20206  -0.202121  -0.202199  -0.202298  -0.202424   -0.20258  -0.202774  -0.203013  -0.203305  -0.203658  -0.204081  -0.204585   -0.20518  -0.205877  -0.206685  -0.207617   -0.20868  -0.209885  -0.211239   -0.21275  -0.214422  -0.216258  -0.218261   -0.22043  -0.222763  -0.225256  -0.227903  -0.230698  -0.233631  -0.236694  -0.239875  -0.243165   -0.24655   -0.25002  -0.253563  -0.257168  -0.260822  -0.264517   -0.26824  -0.271984  -0.275737     -0.279493 | 0.665711  0.665712  0.665713  0.665714  0.665716  0.665718  0.665721  0.665725  0.665729  0.665736  0.665744  0.665754  0.665767  0.665782  0.665802  0.665826  0.665855   0.66589  0.665932  0.665982   0.66604  0.666109  0.666188  0.666279  0.666382    0.6665  0.666632  0.666779  0.666942  0.667121  0.667317  0.667529  0.667759  0.668006  0.668269  0.668549  0.668844  0.669155   0.66948  0.669819  0.670171  0.670535   0.67091  0.671296  0.671691  0.672095  0.672507  0.672925  0.673348      0.673777 | 100.007313 |   100.0088  100.012283  100.017028  100.023447  100.032069  100.043565  100.058786   100.07879   100.10489  100.138692  100.182145  100.237583  100.307779  100.395983  100.505965  100.642048  100.809129   101.01269  101.258794  101.554061  101.905633  102.321116  102.808504  103.376091  104.032368  104.785912  105.645263  106.618808  107.714664  108.940566  110.303771  111.810974  113.468237  115.280939  117.253747  119.390594  121.694683  124.168499  126.813833  129.631815  132.622957  135.787194  139.123934  142.632101  146.310184  150.156278  154.168126  158.343156  162.678514          167.171089 | 169.451444 | 0.720716  0.720723  0.720732  0.720744  0.720761  0.720783  0.720812   0.72085  0.720899  0.720964  0.721046  0.721152  0.721285  0.721453  0.721661  0.721919  0.722236  0.722621  0.723085  0.723642  0.724302  0.725081  0.725991  0.727047  0.728262   0.72965  0.731224  0.732995  0.734973  0.737168  0.739586  0.742234  0.745113  0.748225  0.751569  0.755143  0.758942   0.76296  0.767188  0.771618   0.77624  0.781043  0.786016  0.791146  0.796421  0.801829  0.807358  0.812996   0.81873       0.824551 | 2.591213  2.591251  2.591304  2.591374  2.591469  2.591596  2.591763  2.591984  2.592271  2.592643  2.593121   2.59373  2.594502  2.595471  2.596678  2.598171  2.600002   2.60223   2.60492  2.608142   2.61197  2.616483  2.621762  2.627889  2.634947  2.643015   2.65217  2.662483  2.674019  2.686835  2.700976   2.71648  2.733371  2.751665  2.771364   2.79246  2.814934  2.838759  2.863898  2.890306  2.917932  2.946722  2.976613  3.007544  3.039448  3.072258  3.105907  3.140329  3.175456        3.211223 | 100.007313  100.010249  100.014266  100.019723  100.027082  100.036937  100.050038  100.067331  100.089989   100.11946  100.157507  100.206259  100.268257  100.346499  100.444487  100.566257  100.716416  100.900152  101.123239  101.392026   101.71341  102.094779  102.543957  103.069113  103.678672  104.381198  105.185282  106.099423  107.131902  108.290676  109.583266   111.01667  112.597281  114.330835  116.222359  118.276158  120.495797  122.884116  125.443244  128.174631  131.079089  134.156827  137.407504  140.830274  144.423831  148.186455  152.116055  156.210207  160.466189  164.881011  169.451444 |  100.0 |   0.007313 | 169.451444 |
    | 18:00 | 3.717954  3.718004  3.718072  3.718165  3.718291  3.718458   3.71868  3.718973  3.719357  3.719855  3.720498  3.721321  3.722365  3.723682   3.72533  3.727375  3.729893   3.73297  3.736702   3.74119  3.746548  3.752893   3.76035  3.769047  3.779113  3.790678  3.803866  3.818798  3.835586  3.854331  3.875121  3.898031  3.923118  3.950425  3.979976  4.011778  4.045822  4.082084  4.120523  4.161086   4.20371  4.248318  4.294828   4.34315  4.393188  4.444843  4.498014  4.552598  4.608492             4.665593 | 124.885205  124.887822  124.891398  124.896248  124.902782   124.91152  124.923123   124.93842  124.958441  124.984453  125.017999  125.060942  125.115498  125.184286  125.270357  125.377228  125.508906    125.6699  125.865222  126.100373  126.381315   126.71442  127.106409  127.564272  128.095173  128.706343  129.404969  130.198074  131.092403  132.094306  133.209643  134.443689  135.801062  137.285671   138.90068  140.648486  142.530729  144.548303  146.701388  148.989497  151.411525  153.965804  156.650165  159.461999  162.398315    165.4558   168.63087  171.919717  175.318359  178.822672 |  52.915837   52.916349   52.917047   52.917995   52.919272    52.92098   52.923247   52.926237   52.930149   52.935231   52.941784   52.950172   52.960827   52.974258   52.991059   53.011912   53.037595   53.068979   53.107029   53.152804    53.20744   53.272148   53.348196   53.436887   53.539542   53.657477   53.791973   53.944254   54.115458   54.306619   54.518638   54.752271   55.008114    55.28659    55.58795   55.912269   56.259455   56.629251   57.021253   57.434922   57.869596   58.324514   58.798825   59.291611   59.801901   60.328684   60.870924   61.427573   61.997579        62.579897 |  52.179537   52.180039   52.180724   52.181654   52.182906    52.18458   52.186804   52.189735   52.193571   52.198554    52.20498   52.213206   52.223653   52.236823   52.253298   52.273746    52.29893   52.329705   52.367017   52.411902   52.465477   52.528929     52.6035   52.690468    52.79113   52.906775   53.038659   53.187982   53.355862   53.543311   53.751213   53.980309   54.231183   54.504251   54.799759   55.117781   55.458224   55.820839   56.205229   56.610864   57.037098   57.483181   57.948281   58.431498   58.931878   59.448431   59.980142   60.525981   61.084918     61.655928 | 1.088405  1.088412  1.088423  1.088437  1.088457  1.088483  1.088517  1.088563  1.088622  1.088699  1.088799  1.088927  1.089089  1.089293  1.089548  1.089865  1.090255  1.090732   1.09131  1.092004  1.092833  1.093814  1.094967   1.09631  1.097863  1.099645  1.101675  1.103971  1.106548  1.109421  1.112602    1.1161  1.119922  1.124073  1.128554  1.133363  1.138496  1.143948  1.149708  1.155767  1.162111  1.168728  1.175601  1.182714  1.190053  1.197598  1.205334  1.213243  1.221308  1.229514 | 1.359159  1.359159  1.359158  1.359157  1.359156  1.359154  1.359152  1.359149  1.359145   1.35914  1.359133  1.359125  1.359114    1.3591  1.359084  1.359063  1.359037  1.359005  1.358967  1.358922  1.358868  1.358803  1.358728  1.358641  1.358541  1.358427  1.358297  1.358152  1.357989   1.35781  1.357614    1.3574  1.357169  1.356921  1.356657  1.356377  1.356084  1.355777  1.355458  1.355129  1.354791  1.354445  1.354093  1.353736  1.353376  1.353014  1.352651  1.352289  1.351928           1.35157 | 0.536168  0.536172  0.536178  0.536185  0.536195  0.536208  0.536226  0.536249  0.536279  0.536318  0.536369  0.536434  0.536516   0.53662  0.536749   0.53691  0.537108   0.53735  0.537642  0.537993  0.538412  0.538906  0.539486   0.54016  0.540937  0.541827  0.542836  0.543973  0.545244  0.546653  0.548205  0.549901  0.551742  0.553727  0.555852  0.558115  0.560508  0.563025  0.565659    0.5684  0.571238  0.574164  0.577169   0.58024  0.583368  0.586544  0.589757  0.592998  0.596259      0.599531 | -0.201881  -0.201886  -0.201894  -0.201903  -0.201916  -0.201934  -0.201957  -0.201987  -0.202027  -0.202078  -0.202144  -0.202228  -0.202334  -0.202468  -0.202634  -0.202839   -0.20309  -0.203396  -0.203764  -0.204204  -0.204727  -0.205341  -0.206058  -0.206887  -0.207839  -0.208924  -0.210149  -0.211522  -0.213049  -0.214736  -0.216585  -0.218597  -0.220772  -0.223107  -0.225598   -0.22824  -0.231026  -0.233946  -0.236993  -0.240154   -0.24342   -0.24678  -0.250222  -0.253734  -0.257306  -0.260927  -0.264586  -0.268274  -0.271981     -0.275698 | 0.665711  0.665711  0.665712  0.665713  0.665715  0.665716  0.665719  0.665722  0.665726  0.665731  0.665738  0.665747  0.665757  0.665771  0.665788  0.665808  0.665833  0.665864    0.6659  0.665944  0.665995  0.666056  0.666126  0.666207    0.6663  0.666405  0.666524  0.666658  0.666806  0.666971  0.667151  0.667348  0.667561  0.667791  0.668037    0.6683  0.668578  0.668872  0.669181  0.669504  0.669841   0.67019  0.670551  0.670923  0.671306  0.671697  0.672097  0.672504  0.672918      0.673337 | 100.005918 | 100.007132  100.009981  100.013872  100.019152  100.026264  100.035775  100.048405  100.065057  100.086851  100.115169  100.151693  100.198452  100.257865  100.332788  100.426556  100.543014  100.686552  100.862115  101.075216  101.331923   101.63883  102.003023  102.432015  102.933669  103.516115  104.187643  104.956597  105.831256  106.819726  107.929825  109.168984  110.544152  112.061719  113.727459  115.546476  117.523187  119.661299  121.963816  124.433055  127.070668  129.877677  132.854512  136.001055  139.316684  142.800319  146.450463  150.265243  154.242454  158.379587          162.673863 | 164.855127 | 0.720713  0.720719  0.720726  0.720736   0.72075  0.720768  0.720792  0.720824  0.720865  0.720919  0.720988  0.721077   0.72119  0.721333  0.721511  0.721732  0.722004  0.722336  0.722739  0.723223  0.723801  0.724485  0.725288  0.726225  0.727307  0.728549  0.729964  0.731563  0.733359   0.73536  0.737575   0.74001  0.742671   0.74556  0.748678  0.752023  0.755593  0.759382  0.763386  0.767595  0.772001  0.776595  0.781365    0.7863  0.791389   0.79662  0.801981  0.807459  0.813045       0.818724 | 2.591195  2.591226  2.591269  2.591327  2.591405   2.59151  2.591649  2.591832  2.592072  2.592384  2.592786    2.5933  2.593953  2.594776  2.595806  2.597085  2.598659  2.600582  2.602914  2.605719  2.609066  2.613029  2.617686  2.623115  2.629397  2.636612  2.644837  2.654145  2.664604  2.676275  2.689212  2.703458  2.719046  2.735999  2.754329  2.774038  2.795115  2.817541  2.841289   2.86632  2.892592  2.920054  2.948652  2.978326  3.009015  3.040654   3.07318  3.106525  3.140626        3.175418 | 100.005918  100.008315  100.011604  100.016084  100.022145  100.030284  100.041137  100.055505   100.07439  100.099033  100.130952  100.171989  100.224358  100.290683  100.374049  100.478037  100.606756  100.764871  100.957612  101.190772  101.470691  101.804225  102.198688  102.661793  103.201563  103.826236   104.54416  105.363678  106.293015  107.340162  108.512772  109.818056  111.262706  112.852817   114.59384  116.490541  118.546983   120.76652  123.151808  125.704823  128.426891  131.318728  134.380478   137.61176  141.011708  144.579024  148.312011  152.208622  156.266488  160.482958  164.855127 |  100.0 |   0.005918 | 164.855127 |
    | 18:30 |  3.71793  3.717971  3.718027  3.718103  3.718206  3.718345  3.718529  3.718773  3.719093   3.71951  3.720049  3.720742  3.721625  3.722742  3.724145  3.725892  3.728053  3.730704  3.733931  3.737829  3.742502   3.74806   3.75462  3.762305  3.771239   3.78155  3.793361  3.806794  3.821965  3.838979  3.857933   3.87891   3.90198  3.927196  3.954595  3.984199  4.016012  4.050023  4.086205  4.124518   4.16491  4.207315   4.25166  4.297864  4.345838   4.39549  4.446723  4.499438  4.553535             4.608914 | 124.883958  124.886096  124.889025   124.89301  124.898393  124.905614  124.915231  124.927948  124.944644  124.966406  124.994563  125.030729  125.076836  125.135177  125.208442  125.299753   125.41269  125.551311   125.72016  125.924265  126.169119  126.460646  126.805152  127.209259  127.679823  128.223845   128.84836  129.560336  130.366551  131.273491   132.28724  133.413385  134.656936  136.022258  137.513021  139.132174  140.881926  142.763758  144.778433  146.926037  149.206014  151.617221  154.157982  156.826142  159.619131  162.534015  165.567554  168.716252  171.976401  175.344126 |  52.915593   52.916011   52.916584   52.917363   52.918415   52.919826   52.921705    52.92419   52.927453   52.931705   52.937206   52.944271   52.953277   52.964669   52.978974   52.996795    53.01883   53.045863   53.078772   53.118526   53.166178   53.222858    53.28976   53.368131   53.459246   53.564393   53.684845   53.821838   53.976547   54.150059   54.343353    54.55728   54.792544   55.049693   55.329112   55.631013   55.955444   56.302289   56.671278      57.062   57.473912   57.906362   58.358598   58.829786   59.319029   59.825379   60.347852   60.885439   61.437123        62.001881 |  52.179298   52.179708   52.180269   52.181033   52.182065   52.183448   52.185291   52.187728   52.190927   52.195097   52.200491   52.207419    52.21625   52.227421   52.241448   52.258923    52.28053   52.307038   52.339308    52.37829   52.425017   52.480596   52.546199   52.623048   52.712394   52.815499   52.933611   53.067944   53.219649   53.389791   53.579332   53.789104   54.019799   54.271956   54.545948   54.841986   55.160117   55.500227    55.86205   56.245184   56.649098    57.07315   57.516603   57.978641   58.458384   58.954901   59.467227   59.994375   60.535345     61.089136 | 1.088401  1.088407  1.088416  1.088428  1.088444  1.088465  1.088494  1.088532  1.088581  1.088646  1.088729  1.088837  1.088974  1.089147  1.089364  1.089635   1.08997  1.090381  1.090881  1.091484  1.092207  1.093067  1.094081  1.095269  1.096648  1.098238  1.100058  1.102126  1.104457  1.107069  1.109973  1.113181  1.116702  1.120543  1.124706  1.129193  1.134002  1.139129  1.144566  1.150306  1.156337  1.162647  1.169222  1.176048  1.183109  1.190389  1.197872   1.20554  1.213378  1.221369 |  1.35916  1.359159  1.359159  1.359158  1.359157  1.359155  1.359153  1.359151  1.359148  1.359143  1.359138  1.359131  1.359122   1.35911  1.359096  1.359078  1.359056  1.359029  1.358996  1.358956  1.358909  1.358852  1.358786  1.358709  1.358619  1.358517    1.3584  1.358268  1.358121  1.357957  1.357776  1.357578  1.357363  1.357131  1.356883  1.356619   1.35634  1.356048  1.355742  1.355425  1.355098  1.354762  1.354419   1.35407  1.353716   1.35336  1.353001  1.352642  1.352283          1.351925 | 0.536166   0.53617  0.536174   0.53618  0.536188  0.536199  0.536214  0.536233  0.536258  0.536291  0.536333  0.536388  0.536458  0.536546  0.536656  0.536794  0.536963  0.537172  0.537425   0.53773  0.538096   0.53853  0.539041  0.539638   0.54033  0.541125  0.542033   0.54306  0.544214    0.5455  0.546923  0.548487  0.550192   0.55204  0.554028  0.556154  0.558414  0.560801  0.563309   0.56593  0.568656  0.571476  0.574382  0.577363  0.580409  0.583511  0.586659  0.589842  0.593053      0.596283 | -0.201879  -0.201883  -0.201889  -0.201897  -0.201908  -0.201922  -0.201941  -0.201967     -0.202  -0.202043  -0.202098  -0.202169  -0.202259  -0.202372  -0.202514   -0.20269  -0.202906   -0.20317  -0.203489  -0.203873   -0.20433  -0.204871  -0.205504  -0.206241  -0.207091  -0.208063  -0.209168  -0.210412  -0.211803  -0.213347  -0.215047  -0.216907  -0.218928  -0.221108  -0.223444  -0.225934   -0.22857  -0.231346  -0.234253  -0.237283  -0.240425  -0.243668  -0.247002  -0.250416  -0.253898  -0.257438  -0.261026  -0.264651  -0.268303     -0.271974 | 0.665711  0.665711  0.665712  0.665713  0.665714  0.665715  0.665717   0.66572  0.665723  0.665728  0.665733  0.665741   0.66575  0.665761  0.665776  0.665793  0.665815  0.665841  0.665873  0.665911  0.665956  0.666009  0.666071  0.666143  0.666226  0.666321  0.666428  0.666549  0.666684  0.666834  0.666999  0.667181  0.667378  0.667592  0.667822  0.668068   0.66833  0.668607    0.6689  0.669207  0.669528  0.669862  0.670209  0.670567  0.670936  0.671314  0.671702  0.672098  0.672501      0.672911 | 100.004785 | 100.005775  100.008102   100.01129  100.015628  100.021487  100.029346  100.039814  100.053657  100.071831  100.095522  100.126177  100.165554   100.21576  100.279297  100.359102  100.458586  100.581665  100.732787   100.91694  101.139657     101.407  101.725535  102.102279  102.544647  103.060368  103.657404  104.343842  105.127796  106.017291  107.020157  108.143924  109.395728  110.782223  112.309511  113.983089  115.807805  117.787835  119.926675  122.227146  124.691404  127.320973  130.116774  133.079164  136.207974  139.502559  142.961833  146.584315  150.368167   154.31123          158.411057 | 160.495087 | 0.720711  0.720715  0.720721  0.720729  0.720741  0.720756  0.720775  0.720802  0.720836  0.720881   0.72094  0.721015   0.72111  0.721231  0.721383  0.721571  0.721805  0.722091   0.72244  0.722861  0.723365  0.723964  0.724671  0.725499   0.72646  0.727569  0.728837  0.730278  0.731902  0.733721  0.735744  0.737978   0.74043  0.743103     0.746  0.749122  0.752467  0.756032  0.759812  0.763801  0.767991  0.772373  0.776938  0.781676  0.786574  0.791623   0.79681  0.802124  0.807553       0.813087 |  2.59118  2.591205  2.591241  2.591288  2.591353  2.591439  2.591555  2.591707  2.591907  2.592168  2.592505  2.592938   2.59349  2.594189  2.595066  2.596158  2.597509  2.599166  2.601182  2.603618  2.606538   2.61001  2.614107  2.618906  2.624484  2.630918  2.638286  2.646663  2.656118  2.666717  2.678517  2.691569  2.705912  2.721578  2.738587  2.756947   2.77666  2.797714  2.820089  2.843756  2.868678  2.894813   2.92211  2.950517  2.979976  3.010426  3.041805   3.07405  3.107097        3.140884 | 100.004785   100.00674   100.00943  100.013104  100.018089  100.024803  100.033782  100.045705  100.061426  100.082003  100.108742  100.143234    100.1874  100.243531  100.314337  100.402981  100.513119  100.648926  100.815119  101.016958  101.260249  101.551313  101.896957  102.304416  102.781283  103.335432  103.974913  104.707859  105.542369    106.4864   107.54766  108.733507  110.050859  111.506111  113.105077  114.852938   116.75421  118.812733   121.03166  123.413474   125.96001  128.672478  131.551507   134.59718  137.809077  141.186323  144.727622  148.431304  152.295361  156.317484  160.495087 |  100.0 |   0.004785 | 160.495087 |
    | 19:00 |  3.71791  3.717944   3.71799  3.718052  3.718137  3.718252  3.718404  3.718606  3.718873  3.719221  3.719673  3.720256  3.721001  3.721947  3.723139   3.72463   3.72648  3.728759  3.731544  3.734923  3.738989  3.743846  3.749603  3.756377  3.764287  3.773455  3.784004  3.796056  3.809727  3.825127  3.842358  3.861511  3.882666  3.905885  3.931218    3.9587  3.988348  4.020164  4.054135  4.090232  4.128415  4.168632  4.210817    4.2549  4.300799  4.348429  4.397699  4.448514   4.50078             4.554398 | 124.882944  124.884689  124.887086  124.890357  124.894787  124.900746  124.908707  124.919265   124.93317  124.951351  124.974952  125.005364  125.044269  125.093668  125.155926  125.233803  125.330486  125.449611  125.595281  125.772069  125.985009  126.239579   126.54166  126.897483  127.313566  127.796631   128.35351  128.991047  129.715989  130.534879   131.45395  132.479028   133.61544  134.867944  136.240662  137.737041  139.359826  141.111053   142.99205  145.003466  147.145293  149.416918  151.817166  154.344351   156.99634  159.770597  162.664246   165.67412  168.796807  172.028695 |  52.915395   52.915736   52.916205   52.916844    52.91771   52.918874    52.92043   52.922494   52.925211   52.928763   52.933374   52.939316   52.946916   52.956564   52.968721   52.983924   53.002792   53.026031   53.054435   53.088886   53.130352   53.179883   53.238598   53.307677   53.388341    53.48184   53.589424   53.712327   53.851742   54.008793   54.184518   54.379845   54.595574   54.832364   55.090718   55.370979    55.67333   55.997789   56.344222   56.712345    57.10174   57.511867   57.942078   58.391633   58.859715   59.345447   59.847902   60.366124   60.899131        61.445932 |  52.179104   52.179438   52.179898   52.180524   52.181373   52.182516   52.184041   52.186064   52.188729   52.192212   52.196734    52.20256   52.210012   52.219473   52.231394   52.246302   52.264804   52.287592   52.315443   52.349225   52.389886   52.438455    52.49603   52.563768   52.642866   52.734549   52.840044    52.96056   53.097267   53.251268   53.423581   53.615115   53.826655   54.058846   54.312183   54.587002   54.883481    55.20164   55.541345   55.902319   56.284153   56.686316   57.108172   57.548997   58.007989   58.484288   58.976986   59.485144   60.007801     60.543983 | 1.088398  1.088403   1.08841   1.08842  1.088433  1.088451  1.088474  1.088506  1.088547  1.088601  1.088671  1.088762  1.088877  1.089024  1.089209   1.08944  1.089726   1.09008  1.090511  1.091034  1.091664  1.092415  1.093306  1.094353  1.095575   1.09699  1.098617  1.100473  1.102577  1.104943  1.107587   1.11052  1.113755  1.117297  1.121155  1.125329  1.129821  1.134629  1.139747   1.14517  1.150889  1.156892  1.163167  1.169702  1.176481   1.18349  1.190713  1.198133  1.205735  1.213503 |  1.35916  1.359159  1.359159  1.359158  1.359157  1.359156  1.359155  1.359153   1.35915  1.359146  1.359142  1.359136  1.359128  1.359118  1.359106  1.359091  1.359072  1.359048   1.35902  1.358986  1.358944  1.358895  1.358837  1.358768  1.358689  1.358597  1.358493  1.358374   1.35824   1.35809  1.357924  1.357742  1.357543  1.357327  1.357095  1.356846  1.356583  1.356304  1.356013  1.355709  1.355393  1.355068  1.354735  1.354394  1.354048  1.353698  1.353344  1.352989  1.352633          1.352277 | 0.536165  0.536167  0.536171  0.536176  0.536183  0.536192  0.536204   0.53622  0.536241  0.536268  0.536304   0.53635  0.536409  0.536483  0.536577  0.536694   0.53684  0.537019  0.537238  0.537503  0.537821  0.538201   0.53865  0.539177  0.539791  0.540501  0.541314  0.542239  0.543283  0.544453  0.545754  0.547191  0.548765  0.550479  0.552333  0.554324   0.55645  0.558707  0.561088  0.563586  0.566195  0.568905  0.571707  0.574592  0.577551  0.580572  0.583648  0.586768  0.589923      0.593104 | -0.201877   -0.20188  -0.201885  -0.201892    -0.2019  -0.201912  -0.201928  -0.201949  -0.201977  -0.202013  -0.202059  -0.202119  -0.202195  -0.202291  -0.202412  -0.202562  -0.202748  -0.202975  -0.203252  -0.203586  -0.203985  -0.204459  -0.205018   -0.20567  -0.206426  -0.207296  -0.208288  -0.209412  -0.210675  -0.212083  -0.213642  -0.215355  -0.217226  -0.219254  -0.221438  -0.223776  -0.226262  -0.228892  -0.231658  -0.234552  -0.237565  -0.240687  -0.243907  -0.247216  -0.250602  -0.254055  -0.257564  -0.261119   -0.26471     -0.268328 |  0.66571  0.665711  0.665711  0.665712  0.665713  0.665714  0.665716  0.665718  0.665721  0.665725  0.665729  0.665735  0.665743  0.665753  0.665765   0.66578  0.665799  0.665822  0.665849  0.665882  0.665922  0.665968  0.666023  0.666087  0.666161  0.666246  0.666342  0.666451  0.666574   0.66671  0.666861  0.667028   0.66721  0.667408  0.667622  0.667852  0.668098  0.668359  0.668635  0.668927  0.669232  0.669551  0.669883  0.670227  0.670582  0.670947  0.671323  0.671707  0.672099      0.672498 | 100.003865 | 100.004672  100.006571   100.00918  100.012739  100.017561  100.024048  100.032713  100.044206  100.059342  100.079133  100.104825  100.137936  100.180296   100.23409  100.301896   100.38673  100.492075   100.62191  100.780735   100.97357  101.205964  101.483965  101.814098   102.20331  102.658908  103.188486  103.799832  104.500836  105.299381  106.203243  107.219984  108.356857  109.620711  111.017919   112.55431  114.235116  116.064941  118.047737  120.186802  122.484779  124.943681  127.564909  130.349292  133.297117  136.408172  139.681789   143.11688  146.711981  150.465286          154.374685 | 156.363423 | 0.720709  0.720712  0.720717  0.720724  0.720733  0.720745  0.720762  0.720784  0.720813   0.72085  0.720899  0.720962  0.721043  0.721145  0.721274  0.721435  0.721635  0.721881  0.722182  0.722547  0.722986   0.72351  0.724131  0.724861  0.725712  0.726699  0.727832  0.729126  0.730592   0.73224  0.734082  0.736125  0.738378  0.740844  0.743529  0.746434  0.749559  0.752903  0.756462  0.760232  0.764206  0.768376  0.772735  0.777271  0.781976  0.786838  0.791847  0.796991  0.802259        0.80764 | 2.591168  2.591189  2.591217  2.591257   2.59131  2.591381  2.591476  2.591603  2.591769  2.591987   2.59227  2.592634    2.5931  2.593692  2.594437  2.595369  2.596526   2.59795  2.599691  2.601802  2.604343  2.607377  2.610974  2.615205  2.620143  2.625866  2.632449  2.639967  2.648491  2.658088  2.668821  2.680744  2.693904  2.708339  2.724076  2.741134   2.75952  2.779232  2.800258  2.822577  2.846161  2.870973   2.89697  2.924103   2.95232  2.981565  3.011779  3.042901   3.07487        3.107624 | 100.003865  100.005458  100.007655  100.010666  100.014761  100.020293  100.027713  100.037595  100.050663  100.067823  100.090191  100.119139   100.15633   100.20376  100.263798  100.339232  100.433301  100.549731  100.692755  100.867133  101.078151  101.331615  101.633822  101.991522  102.411862  102.902315  103.470598  104.124576  104.872167  105.721227  106.679454  107.754282  108.952786  110.281598  111.746833  113.354035   115.10813    117.0134   119.07347  121.291308  123.669236  126.208954  128.911566   131.77762  134.807143  137.999684  141.354353  144.869865  148.544573  152.376512  156.363423 |  100.0 |   0.003865 | 156.363423 |
    | 19:30 | 3.717895  3.717922  3.717959  3.718011  3.718081  3.718175  3.718301  3.718469   3.71869  3.718981  3.719359  3.719849  3.720477  3.721276  3.722288  3.723557  3.725139  3.727094  3.729493  3.732414  3.735945   3.74018  3.745221  3.751177  3.758162  3.766293  3.775691  3.786474   3.79876  3.812661  3.828282  3.845721  3.865064  3.886385  3.909744  3.935186  3.962742  3.992425  4.024236   4.05816  4.094167  4.132216  4.172255  4.214219  4.258039  4.303636  4.350925  4.399817  4.450222             4.502045 | 124.882121  124.883544  124.885503  124.888184  124.891826  124.896739   124.90332  124.912076  124.923641   124.93881  124.958563    124.9841  125.016877  125.058637  125.111454  125.177758  125.260376  125.362555  125.487982  125.640796  125.825593  126.047409  126.311695  126.624284  126.991328  127.419236   127.91459  128.484058  129.134289  129.871816   130.70295  131.633681  132.669585  133.815742  135.076665   136.45625  137.957731  139.583664  141.335918  143.215685  145.223502  147.359282  149.622356   152.01152  154.525088  157.160943  159.916588  162.789204  165.775693  168.872725 |  52.915234   52.915512   52.915895   52.916419   52.917131   52.918091   52.919377   52.921089   52.923349   52.926313   52.930172   52.935162   52.941565   52.949722   52.960037   52.972983   52.989111   53.009049   53.033514   53.063307   53.099312   53.142497   53.193906   53.254646   53.325881   53.408808   53.504647   53.614614   53.739903   53.881662    54.04097   54.218816   54.416079   54.633511   54.871725   55.131183   55.412194   55.714906   56.039316   56.385269   56.752471   57.140499   57.548814   57.976775   58.423652   58.888646   59.370898   59.869506   60.383535        60.912031 |  52.178946   52.179219   52.179594   52.180108   52.180806   52.181748   52.183009   52.184687   52.186903   52.189809   52.193594   52.198487   52.204765   52.212764   52.222879   52.235574   52.251388   52.270939   52.294929   52.324143   52.359449   52.401795   52.452206   52.511767   52.581618   52.662935   52.756912   52.864744     52.9876   53.126607   53.282821   53.457213   53.650645   53.863855   54.097443   54.351863   54.627416    54.92425    55.24236   55.581595   55.941666   56.322159   56.722545   57.142195   57.580394   58.036358   58.509245    58.99817   59.502217     60.020451 | 1.088395    1.0884  1.088405  1.088413  1.088424  1.088439  1.088458  1.088484  1.088519  1.088564  1.088623  1.088698  1.088796   1.08892  1.089077  1.089273  1.089519  1.089821  1.090193  1.090646  1.091192  1.091848  1.092628  1.093549  1.094629  1.095885  1.097335  1.098997  1.100889  1.103028  1.105427  1.108102  1.111064  1.114323  1.117886  1.121758  1.125942  1.130438  1.135243  1.140353   1.14576  1.151457  1.157431  1.163672  1.170166  1.176899  1.183857  1.191023  1.198382  1.205919 |  1.35916   1.35916  1.359159  1.359159  1.359158  1.359157  1.359156  1.359154  1.359152  1.359149  1.359145   1.35914  1.359133  1.359125  1.359115  1.359102  1.359086  1.359065  1.359041  1.359011  1.358975  1.358932  1.358881  1.358821   1.35875  1.358669  1.358575  1.358468  1.358347  1.358211   1.35806  1.357892  1.357708  1.357508  1.357291  1.357058   1.35681  1.356547  1.356269  1.355978  1.355676  1.355362  1.355039  1.354708   1.35437  1.354027  1.353679  1.353329  1.352977          1.352624 | 0.536164  0.536166  0.536169  0.536173  0.536178  0.536186  0.536196  0.536209  0.536226  0.536249  0.536279  0.536318  0.536367   0.53643   0.53651   0.53661  0.536734  0.536888  0.537077  0.537306  0.537583  0.537914  0.538308  0.538773  0.539316  0.539947  0.540673  0.541504  0.542446  0.543507  0.544692  0.546007  0.547456  0.549041  0.550763  0.552622  0.554615  0.556741  0.558994  0.561368  0.563857  0.566453  0.569147  0.571932  0.574796  0.577732  0.580729  0.583779  0.586872      0.589998 | -0.201875  -0.201878  -0.201882  -0.201887  -0.201894  -0.201904  -0.201917  -0.201935  -0.201958  -0.201988  -0.202027  -0.202077  -0.202141  -0.202223  -0.202325  -0.202454  -0.202613  -0.202808  -0.203047  -0.203337  -0.203685    -0.2041  -0.204591  -0.205167  -0.205838  -0.206613  -0.207502  -0.208514  -0.209656  -0.210937  -0.212361  -0.213934   -0.21566  -0.217541  -0.219575  -0.221763  -0.224101  -0.226584  -0.229207  -0.231963  -0.234843  -0.237839  -0.240941  -0.244139  -0.247423  -0.250782  -0.254205  -0.257683  -0.261206     -0.264764 |  0.66571  0.665711  0.665711  0.665712  0.665712  0.665713  0.665715  0.665717  0.665719  0.665722  0.665726  0.665731  0.665738  0.665746  0.665756  0.665769  0.665785  0.665805  0.665829  0.665857  0.665892  0.665933  0.665981  0.666037  0.666103  0.666179  0.666265  0.666363  0.666474  0.666598  0.666736  0.666889  0.667056  0.667239  0.667438  0.667652  0.667882  0.668127  0.668387  0.668663  0.668953  0.669256  0.669573  0.669903  0.670244  0.670596  0.670958   0.67133  0.671711      0.672099 | 100.003119 | 100.003776  100.005324  100.007457  100.010375  100.014338  100.019686   100.02685   100.03638  100.048969  100.065481  100.086983  100.114785  100.150471  100.195943   100.25346  100.325677  100.415681  100.527023   100.66374  100.830372  101.031972  101.274089  101.562754  101.904442  102.306023  102.774694  103.317907   103.94328  104.658501  105.471229  106.388994  107.419099  108.568527  109.843855  111.251185  112.796083  114.483533  116.317907  118.302946  120.441758  122.736824   125.19002  127.802635  130.575409  133.508568   136.60186  139.854596  143.265689  146.833692          150.556831 | 152.452299 | 0.720707   0.72071  0.720714  0.720719  0.720727  0.720737  0.720751  0.720769  0.720793  0.720824  0.720865  0.720918  0.720986  0.721073  0.721182  0.721319   0.72149  0.721701   0.72196  0.722276  0.722657  0.723114  0.723658    0.7243  0.725053  0.725928  0.726939  0.728098  0.729416  0.730906  0.732578  0.734441  0.736504  0.738773  0.741254  0.743949  0.746861  0.749988   0.75333  0.756883  0.760642  0.764601  0.768751  0.773085  0.777594  0.782266  0.787092  0.792062  0.797163       0.802386 | 2.591158  2.591175  2.591198   2.59123  2.591274  2.591333  2.591412  2.591517  2.591655  2.591837  2.592074   2.59238  2.592772  2.593272  2.593905  2.594698  2.595687  2.596909  2.598409  2.600235  2.602441  2.605087  2.608237  2.611957  2.616319  2.621396  2.627262   2.63399  2.641653  2.650319  2.660054  2.670915  2.682955  2.696218  2.710737   2.72654  2.743641  2.762047  2.781753  2.802748  2.825009  2.848507  2.873207  2.899065  2.926034  2.954063  2.983097  3.013077  3.043946        3.075644 | 100.003119  100.004416  100.006209  100.008673  100.012034  100.016586   100.02271  100.030891  100.041741  100.056031  100.074717  100.098979   100.13025  100.170264   100.22109  100.285175  100.365379  100.465015  100.587868  100.738227  100.920886  101.141149  101.404814  101.718147  102.087834  102.520929  103.024779  103.606943  104.275102  105.036957   105.90013  106.872064  107.959928  109.170521  110.510205  111.984826  113.599671  115.359425  117.268147  119.329262   121.54556  123.919216  126.451802  129.144325  131.997257  135.010575  138.183798  141.516026  145.005984  148.652051  152.452299 |  100.0 |   0.003119 | 152.452299 |
    | 20:00 | 3.717882  3.717904  3.717935  3.717977  3.718034  3.718112  3.718216  3.718355  3.718539  3.718781  3.719097  3.719508  3.720036  3.720711  3.721568  3.722647  3.723996  3.725671  3.727733  3.730254  3.733313  3.736997  3.741401  3.746626  3.752779  3.759973  3.768324  3.777947  3.788958  3.801471  3.815594  3.831429  3.849067  3.868591  3.890069  3.913558  3.939099  3.966719   3.99643  4.028229  4.062099  4.098011  4.135923  4.175781  4.217525  4.261083  4.306378  4.353329  4.401848             4.451848 | 124.881454  124.882612  124.884212  124.886407  124.889397  124.893443  124.898878  124.906129  124.915737  124.928375  124.944885  124.966298  124.993871   125.02912  125.073853  125.130209  125.200684  125.288166  125.395957  125.527788  125.687833  125.880698  126.111415  126.385405  126.708445  127.086602  127.526173    128.0336  128.615382  129.277979  130.027712  130.870666  131.812595  132.858835  134.014229   135.28306  136.669004  138.175097  139.803716  141.556577  143.434741  145.438645  147.568126  149.822468  152.200443  154.700363  157.320132  160.057292  162.909078   165.87246 |  52.915104    52.91533   52.915643   52.916072   52.916656   52.917447   52.918509   52.919926   52.921804   52.924274     52.9275   52.931684   52.937071   52.943957   52.952694     52.9637   52.977459   52.994534   53.015566   53.041277   53.072473   53.110043   53.154952   53.208235   53.270988   53.344356   53.429512   53.527647   53.639941   53.767551   53.911579   54.073061   54.252937    54.45204   54.671077   54.910619   55.171088   55.452758    55.75575   56.080036   56.425447   56.791678   57.178302   57.584781   58.010481   58.454687   58.916612   59.395417   59.890223        60.400117 |  52.178818    52.17904   52.179347   52.179768   52.180341   52.181116   52.182158   52.183547   52.185388    52.18781   52.190973   52.195076   52.200358   52.207111   52.215678    52.22647   52.239963   52.256706   52.277329   52.302541   52.333131   52.369972   52.414009   52.466257   52.527792   52.599734   52.683237   52.779466    52.88958   53.014711   53.155943   53.314288   53.490671   53.685908   53.900692   54.135582   54.390993   54.667193     54.9643   55.282289   55.620993   55.980112   56.359228   56.757813   57.175247   57.610826   58.063781   58.533288   59.018485     59.518478 | 1.088393  1.088397  1.088402  1.088408  1.088417  1.088429  1.088445  1.088467  1.088495  1.088533  1.088582  1.088645  1.088727  1.088832  1.088965  1.089132  1.089341  1.089601  1.089921  1.090311  1.090785  1.091355  1.092037  1.092845  1.093797  1.094909  1.096198  1.097683   1.09938  1.101307  1.103479   1.10591  1.108615  1.111603  1.114885  1.118467  1.122353  1.126545  1.131044  1.135845  1.140945  1.146337  1.152011  1.157957  1.164163  1.170616  1.177303   1.18421   1.19132  1.198619 |  1.35916   1.35916  1.359159  1.359159  1.359158  1.359158  1.359157  1.359155  1.359153  1.359151  1.359148  1.359143  1.359138  1.359131  1.359122  1.359111  1.359097   1.35908  1.359059  1.359033  1.359002  1.358964   1.35892  1.358867  1.358805  1.358732  1.358649  1.358553  1.358444  1.358321  1.358183  1.358029   1.35786  1.357675  1.357474  1.357256  1.357023  1.356774  1.356511  1.356235  1.355945  1.355644  1.355332  1.355011  1.354683  1.354347  1.354007  1.353662  1.353315          1.352966 | 0.536163  0.536164  0.536167   0.53617  0.536175  0.536181  0.536189    0.5362  0.536214  0.536233  0.536258  0.536291  0.536332  0.536386  0.536453  0.536538  0.536644  0.536776  0.536938  0.537136  0.537376  0.537665   0.53801  0.538418  0.538898  0.539457  0.540104  0.540847  0.541695  0.542653   0.54373   0.54493  0.546259  0.547719  0.549313  0.551043  0.552906  0.554902  0.557026  0.559275  0.561642  0.564121  0.566704  0.569383  0.572149  0.574994  0.577907  0.580881  0.583905       0.58697 | -0.201873  -0.201876  -0.201879  -0.201884   -0.20189  -0.201898  -0.201909  -0.201923  -0.201942  -0.201967     -0.202  -0.202042  -0.202096  -0.202165  -0.202252  -0.202361  -0.202497  -0.202665  -0.202871  -0.203122  -0.203424  -0.203787  -0.204217  -0.204724  -0.205318  -0.206008  -0.206802   -0.20771   -0.20874    -0.2099  -0.211198  -0.212638  -0.214225  -0.215962  -0.217851  -0.219892  -0.222082  -0.224419  -0.226899  -0.229515   -0.23226  -0.235127  -0.238106  -0.241188  -0.244364  -0.247623  -0.250955  -0.254349  -0.257797     -0.261289 |  0.66571   0.66571  0.665711  0.665711  0.665712  0.665713  0.665714  0.665715  0.665717   0.66572  0.665723  0.665728  0.665733   0.66574  0.665749   0.66576  0.665774  0.665791  0.665811  0.665836  0.665866  0.665902  0.665944  0.665994  0.666052  0.666119  0.666196  0.666284  0.666384  0.666497  0.666622  0.666762  0.666915  0.667084  0.667268  0.667467  0.667681  0.667911  0.668156  0.668415   0.66869  0.668978   0.66928  0.669595  0.669922   0.67026   0.67061  0.670969  0.671337      0.671714 | 100.002515 | 100.003049   100.00431  100.006052  100.008441  100.011695  100.016099  100.022015  100.029907  100.040365  100.054122  100.072094  100.095404  100.125423  100.163801  100.212512  100.273886  100.350651   100.44596  100.563427  100.707139  100.881676  101.092106   101.34398    101.6433  101.996487  102.410326  102.891902  103.448522  104.087631  104.816718  105.643222  106.574433  107.617401  108.778848  110.065091  111.481971  113.034806  114.728342  116.566731  118.553515  120.691625  122.983389  125.430554  128.034304    130.7953  133.713709  136.789242  140.021194   143.40848          146.949668 | 148.753956 | 0.720705  0.720708  0.720711  0.720716  0.720722   0.72073  0.720742  0.720757  0.720777  0.720803  0.720837  0.720881  0.720938  0.721011  0.721104  0.721221  0.721367  0.721548   0.72177  0.722043  0.722373  0.722771  0.723246   0.72381  0.724473  0.725248  0.726147  0.727182  0.728364  0.729707   0.73122  0.732914  0.734798   0.73688  0.739165  0.741658  0.744363   0.74728   0.75041  0.753749  0.757295  0.761043  0.764986  0.769116  0.773426  0.777906  0.782546  0.787337  0.792268       0.797328 |  2.59115  2.591164  2.591183  2.591209  2.591245  2.591293  2.591359  2.591446  2.591561  2.591712   2.59191  2.592166  2.592497  2.592919  2.593454  2.594129  2.594973  2.596019  2.597309  2.598884  2.600796  2.603099   2.60585  2.609114  2.612958   2.61745  2.622664  2.628669   2.63554  2.643344  2.652148  2.662014  2.672998  2.685149  2.698508  2.713107  2.728969  2.746108  2.764529  2.784225  2.805185  2.827384  2.850795  2.875381    2.9011  2.927906  2.955748  2.984572  3.014322        3.044941 | 100.002515  100.003569  100.005032  100.007046  100.009801  100.013543  100.018592  100.025355  100.034352  100.046238  100.061828  100.082132  100.108387  100.142094  100.185054  100.239408  100.307675  100.392788  100.498122  100.627523  100.785323  100.976346  101.205906  101.479788  101.804216  102.185808  102.631519  103.148568  103.744354  104.426372  105.202111  106.078963  107.064123    108.1645  109.386633   110.73662  112.220053  113.841973  115.606836  117.518491  119.580174  121.794511  124.163533  126.688698   129.37092  132.210602  135.207674  138.361629   141.67156  145.136201  148.753956 |  100.0 |   0.002515 | 148.753956 |
    | 20:30 | 3.717871   3.71789  3.717915  3.717949  3.717996   3.71806  3.718146  3.718261  3.718413  3.718615  3.718879  3.719222  3.719666  3.720235   3.72096  3.721876  3.723025  3.724456  3.726226  3.728397  3.731042   3.73424  3.738079  3.742652   3.74806  3.754409   3.76181  3.770375  3.780219  3.791453  3.804187  3.818526  3.834565  3.852394  3.872089  3.893716  3.917326  3.942957  3.970633  4.000364  4.032144  4.065955  4.101767  4.139538  4.179215  4.220736  4.264032  4.309029  4.355645             4.403796 | 124.880912  124.881855   124.88316  124.884956  124.887408  124.890736  124.895219  124.901217  124.909188  124.919705  124.933486  124.951415  124.974577  125.004284  125.042112  125.089933  125.149946  125.224712  125.317174  125.430684  125.569015  125.736363  125.937345  126.176978  126.460648  126.794069  127.183222  127.634286  128.153561   128.74738  129.422013  130.183575  131.037932  131.990608  133.046709  134.210849  135.487093  136.878911  138.389149  140.020018  141.773086  143.649298  145.648995  147.771947  150.017393  152.384086   154.87034  157.474079  160.192886  163.024047 |  52.914998   52.915182   52.915437   52.915788   52.916268   52.916918   52.917794   52.918966   52.920524   52.922579   52.925272   52.928776   52.933301   52.939105   52.946494   52.955834   52.967553   52.982149   53.000195    53.02234   53.049314   53.081929   53.121073   53.167707   53.222858   53.287609   53.363085   53.450436    53.55082   53.665387    53.79525   53.941474   54.105046   54.286864   54.487716   54.708265   54.949041   55.210431   55.492674   55.795868   56.119962   56.464772   56.829985   57.215171   57.619794   58.043228   58.484768   58.943644   59.419037        59.910085 |  52.178714   52.178895   52.179145   52.179489   52.179959   52.180597   52.181456   52.182606   52.184133   52.186149   52.188789   52.192225   52.196662   52.202353   52.209599   52.218758   52.230249   52.244562   52.262257   52.283972   52.310423   52.342404   52.380788   52.426516   52.480596    52.54409     52.6181   52.703754    52.80219   52.914531   53.041873   53.185257   53.345652    53.52394   53.720891   53.937158   54.173258   54.429571   54.706334   55.003639    55.32144   55.659554   56.017675    56.39538   56.792146   57.207357   57.640323   58.090288   58.556449     59.037962 | 1.088392  1.088395  1.088398  1.088404  1.088411  1.088421  1.088434  1.088452  1.088476  1.088507  1.088548  1.088601   1.08867  1.088758  1.088871  1.089013  1.089191  1.089413  1.089687  1.090023  1.090433  1.090929  1.091523  1.092231  1.093067  1.094049  1.095192  1.096515  1.098033  1.099764  1.101725  1.103929  1.106392  1.109124  1.112138  1.115442   1.11904  1.122939  1.127138  1.131638  1.136436  1.141525    1.1469  1.152551  1.158468  1.164639  1.171053  1.177694   1.18455  1.191605 |  1.35916   1.35916   1.35916  1.359159  1.359159  1.359158  1.359157  1.359156  1.359155  1.359152   1.35915  1.359146  1.359142  1.359136  1.359128  1.359119  1.359107  1.359093  1.359074  1.359052  1.359025  1.358993  1.358953  1.358907  1.358852  1.358788  1.358714  1.358628   1.35853  1.358419  1.358294  1.358154  1.357999  1.357829  1.357642   1.35744  1.357222  1.356988   1.35674  1.356477  1.356201  1.355912  1.355613  1.355303  1.354984  1.354658  1.354325  1.353987  1.353645          1.353301 | 0.536162  0.536163  0.536165  0.536168  0.536172  0.536177  0.536183  0.536192  0.536204   0.53622  0.536241  0.536268  0.536303  0.536348  0.536405  0.536477  0.536568  0.536681   0.53682  0.536991  0.537198  0.537449   0.53775  0.538108   0.53853  0.539024  0.539599  0.540263  0.541023  0.541886  0.542861  0.543953  0.545167  0.546508  0.547979  0.549583  0.551319  0.553186  0.555183  0.557306   0.55955   0.56191  0.564378  0.566948  0.569612   0.57236  0.575185  0.578076  0.581026      0.584025 | -0.201872  -0.201874  -0.201877  -0.201881  -0.201886  -0.201892  -0.201901  -0.201913  -0.201929   -0.20195  -0.201977  -0.202013  -0.202058  -0.202116   -0.20219  -0.202283  -0.202399  -0.202543   -0.20272  -0.202937  -0.203199  -0.203514  -0.203891  -0.204337  -0.204861  -0.205472  -0.206179  -0.206992  -0.207918  -0.208966  -0.210144  -0.211458  -0.212913  -0.214513   -0.21626  -0.218157  -0.220203  -0.222396  -0.224732  -0.227208  -0.229816   -0.23255  -0.235403  -0.238365  -0.241428  -0.244581  -0.247816  -0.251121  -0.254487     -0.257906 |  0.66571   0.66571   0.66571  0.665711  0.665711  0.665712  0.665713  0.665714  0.665716  0.665718  0.665721  0.665725  0.665729  0.665735  0.665743  0.665752  0.665764  0.665778  0.665796  0.665817  0.665844  0.665875  0.665912  0.665956  0.666007  0.666067  0.666135  0.666214  0.666304  0.666405  0.666519  0.666646  0.666787  0.666942  0.667111  0.667296  0.667495   0.66771  0.667939  0.668183  0.668442  0.668716  0.669003  0.669303  0.669615   0.66994  0.670276  0.670622  0.670979      0.671344 | 100.002026 |  100.00246  100.003486  100.004907  100.006861   100.00953  100.013152  100.018032  100.024561  100.033236  100.044684  100.059685  100.079202  100.104417   100.13676  100.177948  100.230022  100.295383  100.376826  100.477569  100.601279  100.752092  100.934616  101.153934  101.415584  101.725538  102.090154   102.51613  103.010433  103.580223  104.232772  104.975375   105.81525  106.759455  107.814795   108.98774  110.284355  111.710236  113.270458  114.969546  116.811442    118.7995  120.936485  123.224581  125.665413  128.260069  131.009135  133.912725  136.970516   140.18179          143.545462 | 145.260727 | 0.720704  0.720706  0.720709  0.720713  0.720718  0.720725  0.720734  0.720746  0.720763  0.720785  0.720813   0.72085  0.720898   0.72096  0.721038  0.721137  0.721262  0.721416  0.721607  0.721842  0.722128  0.722473  0.722888  0.723381  0.723964  0.724649  0.725446  0.726367  0.727426  0.728632  0.729998  0.731534   0.73325  0.735153  0.737252  0.739552  0.742057  0.744771  0.747693  0.750823   0.75416  0.757698  0.761434  0.765361  0.769471  0.773757  0.778209  0.782817  0.787573       0.792465 | 2.591143  2.591155   2.59117  2.591192  2.591221  2.591261  2.591315  2.591387  2.591482  2.591608  2.591773  2.591988  2.592266  2.592621  2.593074  2.593647  2.594365   2.59526  2.596366  2.597724  2.599377  2.601376  2.603774  2.606632   2.61001  2.613976  2.618597  2.623944  2.630087  2.637096  2.645037  2.653975  2.663968  2.675069  2.687326  2.700775  2.715448  2.731364  2.748536  2.766966  2.786649  2.807569  2.829705  2.853026  2.877498  2.903078   2.92972  2.957376  2.985993        3.015516 | 100.002026  100.002883  100.004074  100.005718  100.007975  100.011048  100.015205   100.02079  100.028242  100.038115  100.051104  100.068073  100.090086  100.118439  100.154694  100.200721  100.258731  100.331312  100.421464  100.532622  100.668684  100.834021  101.033479  101.272376  101.556476  101.891955  102.285358  102.743537  103.273577   103.88272  104.578272  105.367517  106.257618  107.255529  108.367911  109.601049   110.96079  112.452481  114.080932   115.85038  117.764476  119.826278  122.038255  124.402308  126.919784  129.591513  132.417833  135.398632  138.533379  141.821163  145.260727 |  100.0 |   0.002026 | 145.260727 |
    | 21:00 | 3.717863  3.717878  3.717898  3.717926  3.717965  3.718017  3.718088  3.718183  3.718309  3.718477  3.718697  3.718984  3.719357  3.719836  3.720447  3.721223    3.7222  3.723422  3.724937  3.726804  3.729087  3.731857  3.735196  3.739189  3.743931  3.749521  3.756065   3.76367  3.772447  3.782507  3.793958  3.806907  3.821453   3.83769  3.855701  3.875558  3.897324  3.921047   3.94676  3.974484  4.004228  4.035982   4.06973  4.105438  4.143064  4.182557  4.223855  4.266891  4.311591             4.357876 | 124.880474   124.88124  124.882304  124.883771  124.885781  124.888514  124.892208  124.897164  124.903769  124.912509  124.923997  124.938989  124.958417  124.983417  125.015356   125.05587  125.106889  125.170676  125.249846  125.347399  125.466729  125.611644   125.78636  125.995495  126.244047  126.537362  126.881085  127.281105  127.743483  128.274378  128.879953  129.566294  130.339312   131.20466  132.167642  133.233142  134.405556  135.688738  137.085964  138.599902  140.232604  141.985504  143.859435  145.854652  147.970862  150.207265  152.562598  155.035179  157.622953  160.323542 |  52.914912   52.915062    52.91527   52.915557    52.91595   52.916484   52.917206   52.918174   52.919465   52.921173   52.923418   52.926348   52.930144   52.935028   52.941268   52.949182   52.959146   52.971601   52.987055   53.006092    53.02937   53.057625    53.09167   53.132393   53.180752   53.237763   53.304495   53.382053   53.471561   53.574149   53.690931   53.822983   53.971327    54.13691   54.320585   54.523096   54.745067   54.986987   55.249211   55.531948   55.835269   56.159106   56.503261   56.867413   57.251129   57.653879   58.075041   58.513924   58.969774        59.441787 |  52.178631   52.178777   52.178981   52.179262   52.179647   52.180171   52.180879   52.181829   52.183095    52.18477   52.186971   52.189844   52.193566   52.198356   52.204474   52.212234   52.222005   52.234218   52.249373    52.26804   52.290865   52.318571   52.351955   52.391888   52.439307   52.495212   52.560648   52.636699   52.724469   52.825066   52.939579   53.069067    53.21453   53.376898   53.557006   53.755584   53.973244   54.210467   54.467598   54.744845   55.042275   55.359824   55.697295   56.054376   56.430641   56.825569   57.238553   57.668913   58.115911     58.578758 |  1.08839  1.088393  1.088396    1.0884  1.088406  1.088414  1.088425   1.08844   1.08846  1.088486   1.08852  1.088564  1.088622  1.088696  1.088791  1.088912  1.089063  1.089252  1.089487  1.089777   1.09013  1.090559  1.091076  1.091695  1.092428  1.093293  1.094305   1.09548  1.096834  1.098386   1.10015  1.102143  1.104379  1.106871  1.109631  1.112669  1.115992  1.119607  1.123517  1.127722  1.132222  1.137014  1.142092   1.14745  1.153078  1.158966  1.165102  1.171475  1.178072  1.184877 |  1.35916   1.35916   1.35916   1.35916  1.359159  1.359159  1.359158  1.359157  1.359156  1.359154  1.359152  1.359149  1.359145   1.35914  1.359134  1.359126  1.359116  1.359103  1.359088  1.359068  1.359045  1.359017  1.358983  1.358942  1.358894  1.358837  1.358772  1.358695  1.358607  1.358507  1.358394  1.358267  1.358126  1.357969  1.357797  1.357609  1.357406  1.357187  1.356954  1.356705  1.356443  1.356168  1.355881  1.355582  1.355274  1.354957  1.354633  1.354303  1.353968          1.353629 | 0.536161  0.536162  0.536164  0.536166  0.536169  0.536173  0.536179  0.536186  0.536196  0.536209  0.536227   0.53625  0.536279  0.536317  0.536365  0.536426  0.536503  0.536599  0.536718  0.536865  0.537045  0.537262  0.537524  0.537837  0.538208  0.538644  0.539153  0.539744  0.540423  0.541199  0.542078  0.543068  0.544175  0.545403  0.546756  0.548238  0.549849  0.551591  0.553461  0.555459   0.55758  0.559819  0.562171   0.56463  0.567187  0.569835  0.572565   0.57537   0.57824      0.581166 | -0.201871  -0.201873  -0.201875  -0.201878  -0.201882  -0.201888  -0.201895  -0.201905  -0.201918  -0.201936  -0.201958  -0.201988  -0.202026  -0.202075  -0.202138  -0.202217  -0.202316  -0.202439  -0.202591  -0.202777  -0.203004  -0.203278  -0.203607  -0.203998  -0.204459  -0.204999  -0.205627  -0.206353  -0.207183  -0.208127  -0.209193  -0.210388  -0.211717  -0.213186  -0.214798  -0.216555   -0.21846   -0.22051  -0.222704  -0.225039   -0.22751   -0.23011  -0.232834  -0.235672  -0.238618  -0.241661  -0.244792  -0.248002  -0.251281      -0.25462 |  0.66571   0.66571   0.66571  0.665711  0.665711  0.665712  0.665712  0.665713  0.665715  0.665717  0.665719  0.665722  0.665726  0.665731  0.665737  0.665745  0.665755  0.665768  0.665783  0.665802  0.665824  0.665851  0.665884  0.665922  0.665967   0.66602  0.666081  0.666152  0.666232  0.666324  0.666427  0.666542   0.66667  0.666812  0.666968  0.667138  0.667323  0.667523  0.667738  0.667967  0.668211  0.668469  0.668741  0.669026  0.669325  0.669635  0.669958  0.670291  0.670635      0.670988 | 100.001631 | 100.001983  100.002817  100.003975  100.005571  100.007759  100.010734  100.014754  100.020149  100.027338  100.036852  100.049356  100.065676  100.086825  100.114042  100.148815  100.192929  100.248489  100.317963  100.404209  100.510506  100.640571  100.798579  100.989163  101.217411  101.488846    101.8094  102.185365  102.623345  103.130187  103.712905  104.378597  105.134362  105.987208  106.943963  108.011194   109.19513  110.501593  111.935942  113.503027  115.207156  117.052076  119.040962  121.176423  123.460507  125.894726  128.480079   131.21708  134.105795  137.145873          140.336581 | 141.965038 | 0.720703  0.720705  0.720707   0.72071  0.720714   0.72072  0.720728  0.720738  0.720752   0.72077  0.720794  0.720825  0.720865  0.720917  0.720983  0.721067  0.721172  0.721304  0.721468   0.72167  0.721917  0.722216  0.722576  0.723007  0.723519  0.724122  0.724827  0.725646   0.72659  0.727672  0.728901   0.73029  0.731847  0.733583  0.735506  0.737621  0.739935  0.742452  0.745173  0.748099  0.751229  0.754562  0.758093  0.761816  0.765727  0.769817  0.774078  0.778502  0.783079         0.7878 | 2.591138  2.591147   2.59116  2.591178  2.591202  2.591234  2.591279  2.591338  2.591417  2.591522   2.59166  2.591839  2.592072  2.592371  2.592754  2.593239   2.59385  2.594614  2.595561  2.596728  2.598155  2.599886  2.601973  2.604468  2.607431  2.610923   2.61501  2.619758  2.625237  2.631515  2.638658  2.646733  2.655799  2.665914  2.677128  2.689484  2.703019  2.717759  2.733724  2.750924   2.76936  2.789025  2.809903  2.831972  2.855202  2.879558  2.904998  2.931479   2.95895        2.987362 | 100.001631  100.002326  100.003295  100.004637  100.006482  100.009003  100.012423   100.01703  100.023194  100.031385  100.042193  100.056357  100.074787  100.098601  100.129152  100.168068  100.217282  100.279073  100.356095  100.451409  100.568509  100.711336  100.884296   101.09225  101.340512  101.634817  101.981292    102.3864  102.856888  103.399705  104.021933  104.730694  105.533067  106.435992   107.44619   108.57008  109.813705  111.182668  112.682087  114.316547   116.09008  118.006149  120.067644  122.276889  124.635658    127.1452  129.806261  132.619125  135.583636  138.699246  141.965038 |  100.0 |   0.001631 | 141.965038 |
    | 21:30 | 3.717856  3.717868  3.717885  3.717908  3.717939  3.717982   3.71804  3.718119  3.718224  3.718363  3.718546  3.718786  3.719098    3.7195  3.720016  3.720672  3.721502  3.722542  3.723838   3.72544  3.727405  3.729801  3.732698  3.736178  3.740326  3.745237  3.751009  3.757745  3.765552  3.774537  3.784808  3.796471  3.809628  3.824375  3.840801  3.858986  3.878998  3.900894  3.924721  3.950507  3.978273  4.008022  4.039745  4.073423  4.109023  4.146503  4.185811  4.226886  4.269663             4.314068 |  124.88012  124.880742  124.881608  124.882805   124.88445  124.886694  124.889733  124.893823   124.89929  124.906545  124.916109  124.928628  124.944902   124.96591  124.992837  125.027106  125.070409  125.124736  125.192407  125.276091  125.378837   125.50408  125.655656  125.837793  126.055107  126.312571  126.615483  126.969418  127.380169  127.853678  128.395956  129.013005  129.710726  130.494834   131.37077  132.343628  133.418078  134.598306   135.88797  137.290159  138.807372  140.441513  142.193891  144.065234  146.055716  148.164988  150.392217  152.736124  155.195034  157.766915 |  52.914843   52.914965   52.915134   52.915368    52.91569   52.916128   52.916722   52.917521    52.91859   52.920008   52.921877   52.924323   52.927503   52.931608   52.936869   52.943563   52.952021   52.962631   52.975843   52.992178   53.012226   53.036654   53.066203   53.101688   53.143995   53.194076   53.252937   53.321631   53.401242   53.492871   53.597616   53.716556   53.850731   54.001123   54.168637   54.354085   54.558169   54.781474   55.024454   55.287429   55.570583   55.873961    56.19748   56.540929   56.903981   57.286201   57.687061    58.10595   58.542186        58.995031 |  52.178563   52.178682   52.178848   52.179077   52.179392   52.179822   52.180405   52.181189   52.182236   52.183627   52.185459   52.187858   52.190977   52.195002    52.20016   52.206725   52.215019   52.225422   52.238378   52.254395   52.274054   52.298008   52.326983   52.361779   52.403265   52.452373   52.510091   52.577451   52.655516   52.745365   52.848076   52.964706   53.096276   53.243748   53.408009   53.589855   53.789976   54.008945   54.247206   54.505075   54.782729   55.080217   55.397453   55.734233   56.090234   56.465031   56.858107   57.268861   57.696626     58.140677 | 1.088389  1.088391  1.088394  1.088397  1.088402  1.088409  1.088418   1.08843  1.088446  1.088468  1.088496  1.088534  1.088582  1.088644  1.088724  1.088826  1.088955  1.089116  1.089317  1.089565   1.08987  1.090241   1.09069  1.091229  1.091871  1.092631  1.093523  1.094564   1.09577  1.097157   1.09874  1.100537  1.102561  1.104827  1.107348  1.110134  1.113194  1.116537  1.120166  1.124086  1.128296  1.132796  1.137581  1.142647  1.147987  1.153591   1.15945  1.165552  1.171885  1.178436 |  1.35916   1.35916   1.35916   1.35916  1.359159  1.359159  1.359158  1.359158  1.359156  1.359155  1.359153  1.359151  1.359148  1.359143  1.359138  1.359131  1.359123  1.359112  1.359099  1.359082  1.359062  1.359038  1.359008  1.358973  1.358931  1.358881  1.358822  1.358755  1.358676  1.358587  1.358485   1.35837  1.358241  1.358098  1.357939  1.357766  1.357577  1.357373  1.357154   1.35692  1.356672   1.35641  1.356136  1.355849  1.355553  1.355246  1.354932   1.35461  1.354282           1.35395 | 0.536161  0.536161  0.536163  0.536165  0.536167   0.53617  0.536175  0.536181   0.53619    0.5362  0.536215  0.536234  0.536258   0.53629  0.536331  0.536383  0.536448   0.53653  0.536632  0.536758  0.536913  0.537101  0.537328  0.537601  0.537926   0.53831   0.53876  0.539284  0.539889  0.540584  0.541376  0.542271  0.543276  0.544396  0.545637  0.547002  0.548493  0.550112  0.551859  0.553733   0.55573  0.557849  0.560083  0.562427  0.564875  0.567419  0.570051  0.572764  0.575549      0.578397 | -0.201871  -0.201872  -0.201874  -0.201876   -0.20188  -0.201884   -0.20189  -0.201898  -0.201909  -0.201924  -0.201943  -0.201968     -0.202  -0.202041  -0.202094   -0.20216  -0.202245   -0.20235   -0.20248   -0.20264  -0.202836  -0.203074   -0.20336  -0.203702  -0.204107  -0.204583   -0.20514  -0.205785  -0.206527  -0.207375  -0.208337   -0.20942   -0.21063  -0.211974  -0.213457   -0.21508  -0.216847  -0.218758  -0.220812  -0.223007   -0.22534  -0.227805  -0.230398   -0.23311  -0.235934  -0.238863  -0.241886  -0.244996  -0.248182     -0.251435 |  0.66571   0.66571   0.66571   0.66571  0.665711  0.665711  0.665712  0.665713  0.665714  0.665715  0.665717   0.66572  0.665723  0.665727  0.665733   0.66574  0.665748  0.665759  0.665772  0.665788  0.665807  0.665831  0.665859  0.665893  0.665933  0.665979  0.666033  0.666096  0.666168   0.66625  0.666343  0.666448  0.666565  0.666694  0.666837  0.666994  0.667165  0.667351  0.667551  0.667765  0.667994  0.668238  0.668495  0.668766   0.66905  0.669346  0.669655  0.669975  0.670306      0.670647 | 100.001312 | 100.001597  100.002274  100.003217   100.00452   100.00631  100.008752  100.012061  100.016513  100.022463   100.03036   100.04077  100.054397  100.072113  100.094982  100.124297  100.161608  100.208761  100.267926  100.341634  100.432804  100.544768  100.681291  100.846579  101.045284  101.282494  101.563711  101.894818  102.282041  102.731884  103.251071  103.846467  104.525001  105.293579  106.158998  107.127865  108.206518  109.400951  110.716755   112.15906  113.732502  115.441185  117.288669  119.277961  121.411522  123.691273   126.11862  128.694477  131.419296  134.293095          137.315496 | 138.859416 | 0.720703  0.720704  0.720706  0.720708  0.720712  0.720716  0.720723  0.720731  0.720742  0.720757  0.720777  0.720803  0.720837   0.72088  0.720936  0.721007  0.721097  0.721209  0.721349  0.721523  0.721735  0.721994  0.722307  0.722682   0.72313   0.72366  0.724282  0.725008  0.725848  0.726815  0.727919  0.729171  0.730581   0.73216  0.733916  0.735856  0.737987  0.740315  0.742841  0.745569  0.748498  0.751628  0.754956  0.758478   0.76219  0.766084  0.770153   0.77439  0.778786       0.783332 | 2.591134  2.591141  2.591152  2.591166  2.591186  2.591213  2.591249  2.591298  2.591364   2.59145  2.591565  2.591715   2.59191  2.592162  2.592484  2.592895  2.593413  2.594064  2.594874  2.595875  2.597104  2.598601  2.600412  2.602587  2.605179  2.608247  2.611852  2.616059  2.620933  2.626541   2.63295  2.640225  2.648429  2.657619  2.667851  2.679172  2.691623  2.705238  2.720041   2.73605  2.753274  2.771711  2.791354  2.812187  2.834187  2.857324  2.881563  2.906864  2.933183        2.960471 | 100.001312  100.001875  100.002663  100.003756  100.005265   100.00733   100.01014  100.013935  100.019028  100.025816  100.034798  100.046605  100.062015  100.081989  100.107696  100.140548  100.182234  100.234753  100.300445   100.38203  100.482626  100.605776  100.755465  100.936122   101.15262  101.410262   101.71475  102.072152  102.488851  102.971479  103.526852   104.16189  104.883534  105.698658  106.613989  107.636019  108.770933   110.02454  111.402214  112.908848  114.548818  116.325961  118.243559  120.304346  122.510507  124.863701  127.365081  130.015319  132.814644  135.762866  138.859416 |  100.0 |   0.001312 | 138.859416 |
    | 22:00 | 3.717851   3.71786  3.717874  3.717893  3.717918  3.717954  3.718002  3.718066  3.718153  3.718268   3.71842   3.71862  3.718881  3.719219  3.719653  3.720208  3.720911  3.721796  3.722901  3.724273  3.725963   3.72803   3.73054  3.733565  3.737187  3.741491   3.74657  3.752522  3.759449  3.767454  3.776643   3.78712  3.798989  3.812348  3.827289  3.843897  3.862248  3.882406  3.904425  3.928347    3.9542  3.981999  4.011747  4.043434  4.077038  4.112527  4.149857  4.188978  4.229831             4.272349 | 124.879833  124.880338  124.881042  124.882018  124.883363  124.885203  124.887701  124.891073  124.895592  124.901607  124.909559      124.92  124.933613  124.951242  124.973909  125.002852  125.039548  125.085743  125.143484  125.215145  125.303448  125.411483  125.542723  125.701026  125.890629  126.116138  126.382495  126.694949  127.058998  127.480336  127.964783  128.518208  129.146448  129.855223  130.650055  131.536185  132.518499  133.601463  134.789065  136.084772  137.491497  139.011581  140.646788  142.398308  144.266773  146.252285  148.354441  150.572377  152.904804  155.350054 |  52.914787   52.914886   52.915023   52.915214   52.915477   52.915837   52.916325   52.916984   52.917867   52.919043   52.920597   52.922637   52.925297   52.928742   52.933171   52.938825   52.945994   52.955016   52.966292   52.980282   52.997516   53.018594   53.044189   53.075044   53.111977   53.155871   53.207669   53.268368   53.339003   53.420639   53.514348   53.621201   53.742243   53.878477   54.030845   54.200213   54.387352   54.592926   54.817482   55.061439   55.325087   55.608584   55.911955   56.235099   56.577794   56.939709   57.320407   57.719365   58.135979         58.56958 |  52.178508   52.178604   52.178739   52.178926   52.179184   52.179537   52.180016   52.180662   52.181528    52.18268   52.184204   52.186205   52.188814   52.192191   52.196534   52.202079   52.209108   52.217955   52.229012   52.242731    52.25963   52.280299   52.305396   52.335652   52.371868    52.41491   52.465702   52.525222   52.594486   52.674536   52.766426   52.871204   52.989895   53.123483   53.272893   53.438972   53.622476   53.824058   54.044253   54.283473   54.542002   54.819993   55.117472   55.434341   55.770382   56.125268   56.498574   56.889784   57.298308     57.723488 | 1.088389   1.08839  1.088392  1.088395  1.088399  1.088405  1.088412  1.088422  1.088435  1.088453  1.088477  1.088508  1.088548  1.088601  1.088668  1.088754  1.088863     1.089  1.089172  1.089384  1.089646  1.089967  1.090355  1.090824  1.091385  1.092051  1.092837  1.093757  1.094827  1.096064  1.097482  1.099097  1.100925   1.10298  1.105275  1.107823  1.110633  1.113715  1.117075  1.120718  1.124646   1.12886  1.133358  1.138137   1.14319  1.148511  1.154092  1.159921  1.165989  1.172282 |  1.35916   1.35916   1.35916   1.35916   1.35916  1.359159  1.359159  1.359158  1.359157  1.359156  1.359154  1.359152   1.35915  1.359146  1.359142  1.359136  1.359129   1.35912  1.359108  1.359094  1.359077  1.359056   1.35903  1.358999  1.358963  1.358919  1.358867  1.358807  1.358737  1.358657  1.358566  1.358462  1.358345  1.358214  1.358069   1.35791  1.357735  1.357545   1.35734  1.357121  1.356887  1.356639  1.356377  1.356104  1.355819  1.355524  1.355219  1.354907  1.354587          1.354262 |  0.53616  0.536161  0.536162  0.536163  0.536165  0.536168  0.536172  0.536177  0.536184  0.536193  0.536205  0.536221  0.536241  0.536268  0.536302  0.536346  0.536401  0.536471  0.536558  0.536666  0.536799  0.536962  0.537159  0.537396   0.53768  0.538017  0.538414  0.538878  0.539416  0.540037  0.540747  0.541554  0.542464  0.543483  0.544617   0.54587  0.547246  0.548746  0.550372  0.552124     0.554  0.555997  0.558112  0.560341  0.562677  0.565114  0.567645  0.570262  0.572957      0.575723 |  -0.20187  -0.201871  -0.201873  -0.201875  -0.201877  -0.201881  -0.201886  -0.201893  -0.201902  -0.201914   -0.20193   -0.20195  -0.201977  -0.202012  -0.202056  -0.202113  -0.202184  -0.202274  -0.202385  -0.202523  -0.202692  -0.202898  -0.203146  -0.203444  -0.203799  -0.204218   -0.20471  -0.205282  -0.205944  -0.206704  -0.207569  -0.208547  -0.209647  -0.210873  -0.212231  -0.213725   -0.21536  -0.217135  -0.219052  -0.221109  -0.223305  -0.225635  -0.228095  -0.230679  -0.233379   -0.23619  -0.239101  -0.242105  -0.245193     -0.248355 |  0.66571   0.66571   0.66571   0.66571   0.66571  0.665711  0.665711  0.665712  0.665713  0.665714  0.665716  0.665718  0.665721  0.665724  0.665729  0.665735  0.665742  0.665751  0.665762  0.665776  0.665793  0.665813  0.665838  0.665867  0.665902  0.665943  0.665991  0.666047  0.666111  0.666185  0.666268  0.666363  0.666469  0.666587  0.666718  0.666862   0.66702  0.667192  0.667378  0.667578  0.667792  0.668021  0.668264   0.66852   0.66879  0.669072  0.669367  0.669674  0.669991       0.67032 | 100.001054 | 100.001285  100.001835  100.002601  100.003664  100.005128   100.00713  100.009849  100.013519  100.018438  100.024985   100.03364  100.045005  100.059824  100.079014  100.103691  100.135201  100.175156  100.225459  100.288345  100.366404  100.462611  100.580349  100.723423  100.896067  101.102945  101.349138   101.64012  101.981725  102.380103  102.841658  103.372989  103.980813  104.671886  105.452926  106.330526  107.311077  108.400692  109.605143  110.929796  112.379566  113.958881  115.671651  117.521262   119.51056  121.641866  123.916983  126.337219  128.903406  131.615938          134.474791 | 135.936495 | 0.720702  0.720703  0.720705  0.720707  0.720709  0.720713  0.720718  0.720725  0.720735  0.720747  0.720764  0.720785  0.720814   0.72085  0.720897  0.720957  0.721033  0.721129  0.721248  0.721396  0.721579  0.721802  0.722074    0.7224  0.722791  0.723256  0.723804  0.724445  0.725191  0.726053  0.727041  0.728167  0.729441  0.730873  0.732472  0.734246  0.736204   0.73835  0.740689  0.743225  0.745959  0.748891   0.75202  0.755343  0.758856  0.762554  0.766431   0.77048  0.774693       0.779062 |  2.59113  2.591136  2.591145  2.591157  2.591173  2.591195  2.591225  2.591265  2.591319  2.591391  2.591487  2.591612  2.591775  2.591986  2.592258  2.592604  2.593044  2.593597  2.594288  2.595146  2.596202  2.597494  2.599063  2.600954  2.603217  2.605906   2.60908  2.612797  2.617123   2.62212  2.627856  2.634393  2.641796  2.650125  2.659435  2.669779  2.681202  2.693743  2.707432  2.722293  2.738342  2.755584  2.774019  2.793637  2.814422  2.836351  2.859394  2.883516  2.908677        2.934834 | 100.001054  100.001511   100.00215  100.003041  100.004272  100.005962  100.008268  100.011392  100.015595  100.021213  100.028669  100.038497  100.051365  100.068095  100.089695  100.117388  100.152643  100.197209  100.253146  100.322859  100.409122  100.515111  100.644413  100.801049  100.989469  101.214549  101.481574  101.796211  102.164464  102.592626   103.08722  103.654922  104.302493  105.036692  105.864196   106.79152  107.824935  108.970402  110.233504  111.619394  113.132752  114.777754  116.558051  118.476758  120.536457  122.739205  125.086552  127.579561  130.218836  133.004554  135.936495 |  100.0 |   0.001054 | 135.936495 |
    | 22:30 | 3.717846  3.717854  3.717865   3.71788  3.717901   3.71793   3.71797  3.718023  3.718094   3.71819  3.718316  3.718483  3.718701  3.718984  3.719349  3.719817  3.720411  3.721163  3.722105  3.723278  3.724728  3.726508  3.728677  3.731303  3.734458  3.738222  3.742681  3.747928  3.754059  3.761174  3.769374  3.778763  3.789443  3.801513  3.815067  3.830195  3.846977  3.865486  3.885782  3.907916  3.931927  3.957838  3.985664  4.015405   4.04705  4.080576   4.11595  4.153129  4.192062             4.232692 | 124.879602  124.880011  124.880583  124.881378  124.882477  124.883983  124.886035  124.888811  124.892543  124.897523  124.904127  124.912823  124.924196  124.938969  124.958023  124.982432   125.01348  125.052697  125.101884  125.163142  125.238896  125.331914  125.445329  125.582642   125.74773  125.944834  126.178543  126.453767  126.775696  127.149751  127.581526  128.076716  128.641047  129.280193  129.999699  130.804897  131.700833  132.692195  133.783252  134.977801  136.279129  137.689982  139.212552   140.84847  142.598817  144.464134  146.444455   148.53933  150.747869  153.068773 |  52.914742   52.914822   52.914934   52.915089   52.915304   52.915598   52.915999   52.916542   52.917271   52.918245   52.919535   52.921235   52.923457   52.926344   52.930067   52.934836   52.940901   52.948562   52.958168    52.97013   52.984918   53.003071   53.025196   53.051971   53.084144   53.122531   53.168011   53.221521   53.284043   53.356598   53.440226   53.535977    53.64489   53.767977   53.906205   54.060479   54.231625   54.420376   54.627359   54.853084    55.09794   55.362187   55.645958   55.949259   56.271974   56.613872   56.974616    57.35377   57.750815        58.165156 |  52.178463   52.178542   52.178651   52.178804   52.179014   52.179303   52.179696   52.180228   52.180943   52.181898   52.183163    52.18483   52.187009    52.18984   52.193491   52.198167   52.204115   52.211626   52.221046   52.232776   52.247277   52.265077   52.286772   52.313027   52.344576   52.382217   52.426814   52.479285   52.540593   52.611738   52.693743   52.787634   52.894432   53.015129   53.150672   53.301951   53.469773   53.654859   53.857823   54.079164   54.319265   54.578381   54.856641   55.154052     55.4705   55.805759   56.159497   56.531289   56.920623     57.326917 | 1.088388  1.088389  1.088391  1.088393  1.088396  1.088401  1.088407  1.088415  1.088426  1.088441  1.088461  1.088487   1.08852  1.088564  1.088621  1.088693  1.088786  1.088902  1.089048   1.08923  1.089455  1.089731  1.090067  1.090474  1.090962  1.091545  1.092235  1.093047  1.093995  1.095094   1.09636  1.097809  1.099455  1.101313  1.103398  1.105721  1.108295  1.111128  1.114231  1.117607  1.121262  1.125198  1.129415   1.13391  1.138681  1.143721  1.149024   1.15458   1.16038  1.166413 |  1.35916   1.35916   1.35916   1.35916   1.35916   1.35916  1.359159  1.359159  1.359158  1.359157  1.359156  1.359154  1.359152  1.359149  1.359145   1.35914  1.359134  1.359126  1.359117  1.359105   1.35909  1.359071  1.359049  1.359022   1.35899  1.358952  1.358907  1.358854  1.358792   1.35872  1.358638  1.358545  1.358439   1.35832  1.358188  1.358041   1.35788  1.357704  1.357514  1.357308  1.357088  1.356854  1.356606  1.356346  1.356073  1.355789  1.355496  1.355193  1.354882          1.354565 |  0.53616   0.53616  0.536161  0.536162  0.536164  0.536166  0.536169  0.536174  0.536179  0.536187  0.536197   0.53621  0.536227  0.536249  0.536278  0.536315  0.536362  0.536421  0.536495  0.536588  0.536702  0.536842  0.537013  0.537219  0.537466  0.537761   0.53811   0.53852  0.538997   0.53955  0.540185   0.54091  0.541732  0.542657   0.54369  0.544837  0.546102  0.547487  0.548996  0.550629  0.552384  0.554262  0.556259  0.558371  0.560594  0.562921  0.565347  0.567865  0.570467      0.573144 |  -0.20187  -0.201871  -0.201872  -0.201873  -0.201876  -0.201879  -0.201883  -0.201888  -0.201896  -0.201906  -0.201919  -0.201936  -0.201959  -0.201988  -0.202025  -0.202073  -0.202134   -0.20221  -0.202305  -0.202423  -0.202568  -0.202746  -0.202961   -0.20322   -0.20353  -0.203898  -0.204332  -0.204838  -0.205427  -0.206105  -0.206881  -0.207763  -0.208758  -0.209873  -0.211114  -0.212486  -0.213992  -0.215636   -0.21742  -0.219342  -0.221402  -0.223597  -0.225924  -0.228378  -0.230953  -0.233642  -0.236438  -0.239333  -0.242318     -0.245384 |  0.66571   0.66571   0.66571   0.66571   0.66571  0.665711  0.665711  0.665712  0.665712  0.665713  0.665715  0.665717  0.665719  0.665722  0.665726  0.665731  0.665737  0.665744  0.665754  0.665766   0.66578  0.665798  0.665819  0.665845  0.665876  0.665912  0.665954  0.666003  0.666061  0.666126  0.666201  0.666287  0.666382   0.66649  0.666609  0.666741  0.666887  0.667045  0.667218  0.667404  0.667604  0.667819  0.668047  0.668289  0.668545  0.668813  0.669094  0.669387  0.669692      0.670007 | 100.000846 | 100.001034  100.001479  100.002102  100.002967  100.004163  100.005802  100.008036  100.011058  100.015119   100.02054  100.027728  100.037193  100.049573  100.065653  100.086396  100.112969  100.146772  100.189474  100.243038  100.309757  100.392278  100.493631  100.617242  100.766951  100.947017  101.162108  101.417294  101.718016  102.070053  102.479475  102.952584  103.495852  104.115847  104.819156  105.612311  106.501704  107.493518  108.593649  109.807652   111.14068   112.59744  114.182162  115.898576  117.749897  119.738822   121.86754  124.137743  126.550644  129.107004          131.807159 | 133.189015 | 0.720702  0.720702  0.720704  0.720705  0.720708  0.720711  0.720715  0.720721  0.720728  0.720739  0.720752   0.72077  0.720794  0.720825  0.720864  0.720915  0.720979   0.72106  0.721162  0.721289  0.721446  0.721638  0.721872  0.722156  0.722497  0.722903  0.723384   0.72395  0.724611  0.725377   0.72626  0.727269  0.728417  0.729711  0.731164  0.732782  0.734575  0.736549  0.738709   0.74106  0.743604  0.746343  0.749277  0.752404  0.755721  0.759225   0.76291  0.766771  0.770799       0.774988 | 2.591128  2.591133  2.591139  2.591149  2.591162   2.59118  2.591205  2.591238  2.591283  2.591342  2.591422  2.591526  2.591662  2.591839  2.592067   2.59236  2.592732  2.593201   2.59379  2.594523   2.59543  2.596543  2.597899   2.59954  2.601512  2.603864   2.60665  2.609928  2.613757    2.6182  2.623319  2.629179  2.635842   2.64337  2.651819  2.661246  2.671697  2.683217  2.695843  2.709602  2.724516  2.740599  2.757856  2.776284  2.795875  2.816609  2.838465  2.861412  2.885416        2.910438 | 100.000846  100.001216  100.001735  100.002459  100.003463  100.004845  100.006736  100.009304  100.012769  100.017413  100.023594  100.031766  100.042497  100.056491  100.074615  100.097925  100.127696  100.165455  100.213008  100.272476  100.346321  100.437374   100.54886  100.684408  100.848068  101.044305  101.277994  101.554398  101.879138  102.258154  102.697646  103.204023  103.783822  104.443646  105.190073  106.029588  106.968499  108.012863  109.168424  110.440548  111.834177  113.353787  115.003364  116.786382  118.705798  120.764051  122.963077  125.304324   127.78877  130.416957  133.189015 |  100.0 |   0.000846 | 133.189015 |
    | 23:00 | 3.717843  3.717849  3.717858   3.71787  3.717888  3.717911  3.717943  3.717987  3.718046  3.718125   3.71823  3.718369  3.718551  3.718787  3.719094  3.719488   3.71999  3.720627  3.721428  3.722429  3.723671  3.725202  3.727074  3.729347   3.73209  3.735375  3.739282  3.743897  3.749311  3.755619   3.76292  3.771312  3.780898  3.791775  3.804039  3.817783  3.833091   3.85004  3.868699  3.889126  3.911367  3.935458  3.961421  3.989268  4.018996  4.050594  4.084038  4.119294   4.15632             4.195065 | 124.879415  124.879746  124.880211  124.880858  124.881754  124.882987   124.88467  124.886953  124.890031   124.89415  124.899628  124.906862  124.916351  124.928713  124.944708  124.965263  124.991493  125.024734  125.066566  125.118843  125.183717  125.263661  125.361487  125.480364  125.623819  125.795742  126.000372  126.242279  126.526332  126.857662  127.241609  127.683663  128.189396  128.764389  129.414159  130.144075  130.959287   131.86465  132.864663  133.963403  135.164488  136.471032  137.885623  139.410311  141.046605   142.79548  144.657396  146.632322  148.719767  150.918815 |  52.914705    52.91477   52.914861   52.914988   52.915163   52.915404   52.915733   52.916179    52.91678   52.917585   52.918656    52.92007   52.921924    52.92434   52.927465   52.931481   52.936606     52.9431   52.951271    52.96148   52.974147   52.989752   53.008841   53.032029   53.059998   53.093497   53.133342   53.180408    53.23562    53.29995     53.3744   53.459989   53.557741   53.668665    53.79374   53.933899   54.090009    54.26286   54.453147    54.66146   54.888278   55.133957   55.398731    55.68271   55.985882   56.308118   56.649178    57.00872   57.386311        57.781435 |  52.178428   52.178491    52.17858   52.178704   52.178876   52.179112   52.179435   52.179872   52.180462   52.181252   52.182301   52.183687   52.185506   52.187875    52.19094   52.194878   52.199903    52.20627   52.214283   52.224294   52.236715   52.252016   52.270735   52.293473   52.320898   52.353747   52.392819    52.43897    52.49311   52.556191   52.629195   52.713122   52.808976   52.917745   53.040392   53.177829   53.330908   53.500402   53.686993   53.891262   54.113674   54.354582   54.614215   54.892679   55.189964   55.505942   55.840379   56.192939   56.563198     56.950648 | 1.088387  1.088388   1.08839  1.088392  1.088394  1.088398  1.088403   1.08841  1.088419  1.088431  1.088447  1.088469  1.088497  1.088534  1.088581  1.088642   1.08872  1.088819  1.088943  1.089099  1.089291  1.089528  1.089818  1.090171  1.090595  1.091104  1.091709  1.092423  1.093261  1.094236  1.095364  1.096659  1.098138  1.099814  1.101702  1.103815  1.106165  1.108764   1.11162  1.114741  1.118133  1.121799  1.125742   1.12996  1.134453  1.139215  1.144241  1.149524  1.155056  1.160826 |  1.35916   1.35916   1.35916   1.35916   1.35916   1.35916  1.359159  1.359159  1.359158  1.359158  1.359156  1.359155  1.359153  1.359151  1.359148  1.359143  1.359138  1.359132  1.359124  1.359113  1.359101  1.359085  1.359066  1.359042  1.359014  1.358981  1.358941  1.358894   1.35884  1.358776  1.358703  1.358619  1.358523  1.358416  1.358295  1.358161  1.358013  1.357851  1.357674  1.357482  1.357276  1.357056  1.356822  1.356574  1.356314  1.356043   1.35576  1.355468  1.355167          1.354859 | 0.536159   0.53616  0.536161  0.536162  0.536163  0.536165  0.536167  0.536171  0.536176  0.536182   0.53619  0.536201  0.536215  0.536234  0.536258  0.536289  0.536329  0.536379  0.536442  0.536521  0.536619  0.536739  0.536886  0.537065   0.53728  0.537538  0.537844  0.538205  0.538627  0.539119  0.539685  0.540335  0.541075  0.541911   0.54285  0.543896  0.545056  0.546332  0.547727  0.549244  0.550882  0.552641   0.55452  0.556516  0.558625  0.560841   0.56316  0.565575  0.568079      0.570666 | -0.201869   -0.20187  -0.201871  -0.201872  -0.201874  -0.201877   -0.20188  -0.201885  -0.201891  -0.201899   -0.20191  -0.201924  -0.201943  -0.201967  -0.201999  -0.202039   -0.20209  -0.202155  -0.202236  -0.202337  -0.202462  -0.202615  -0.202801  -0.203026  -0.203297  -0.203619     -0.204  -0.204447  -0.204969  -0.205573  -0.206267   -0.20706  -0.207958  -0.208969  -0.210099  -0.211354  -0.212739  -0.214257   -0.21591    -0.2177  -0.219627   -0.22169  -0.223885  -0.226208  -0.228656  -0.231222  -0.233899  -0.236681  -0.239558     -0.242524 |  0.66571   0.66571   0.66571   0.66571   0.66571   0.66571  0.665711  0.665711  0.665712  0.665713  0.665714  0.665715  0.665717   0.66572  0.665723  0.665727  0.665732  0.665739  0.665747  0.665757   0.66577  0.665785  0.665804  0.665826  0.665852  0.665884  0.665921  0.665965  0.666016  0.666074  0.666142  0.666218  0.666305  0.666402  0.666511  0.666631  0.666765  0.666911   0.66707  0.667243   0.66743  0.667631  0.667845  0.668073  0.668314  0.668569  0.668836  0.669115  0.669407      0.669709 | 100.000679 |  100.00083  100.001191  100.001696  100.002401  100.003377  100.004718   100.00655  100.009035  100.012385  100.016869  100.022831  100.030704  100.041033   100.05449  100.071902  100.094278  100.122834  100.159025  100.204576  100.261507  100.332168   100.41926  100.525859  100.655435  100.811857    100.9994  101.222735  101.486914  101.797341  102.159734  102.580081  103.064579  103.619572  104.251479  104.966721  105.771646  106.672451  107.675114  108.785325  110.008429  111.349372  112.812665  114.402351  116.121983  117.974618  119.962812  122.088629  124.353655  126.759017          129.305406 | 130.609826 | 0.720701  0.720702  0.720703  0.720704  0.720706  0.720709  0.720712  0.720717  0.720723  0.720732  0.720743  0.720758  0.720778  0.720803  0.720837  0.720879  0.720933  0.721002  0.721089  0.721197  0.721331  0.721497  0.721699  0.721945  0.722241  0.722596  0.723017  0.723515  0.724099  0.724779  0.725565  0.726468  0.727499  0.728667  0.729982  0.731455  0.733092  0.734902  0.736891  0.739064  0.741426  0.743978  0.746721  0.749656  0.752781  0.756092  0.759586  0.763258  0.767101       0.771109 | 2.591125  2.591129  2.591135  2.591143  2.591153  2.591168  2.591188  2.591216  2.591253  2.591302  2.591368  2.591454  2.591568  2.591716  2.591908  2.592154  2.592468  2.592866  2.593367  2.593993   2.59477  2.595726  2.596896  2.598318  2.600032  2.602085  2.604526   2.60741  2.610792  2.614731   2.61929  2.624529  2.630511  2.637296  2.644945  2.653512  2.663049  2.673604  2.685217  2.697922  2.711746  2.726709  2.742822   2.76009  2.778509  2.798068  2.818749   2.84053   2.86338        2.887266 | 100.000679  100.000978  100.001398  100.001987  100.002805  100.003934  100.005482  100.007591  100.010444  100.014279  100.019397  100.026184  100.035121  100.046812  100.061998  100.081591  100.106695  100.138637  100.178999  100.229644  100.292751  100.370837  100.466786  100.583867  100.725747  100.896499  101.100598  101.342913  101.628678  101.963467  102.353149   102.80383  103.321801  103.913463  104.585257  105.343588  106.194749  107.144848  108.199735  109.364943  110.645632  112.046539  113.571948  115.225663   117.01099  118.930734  120.987203  123.182218  125.517129  127.992838  130.609826 |  100.0 |   0.000679 | 130.609826 |
    | 23:30 |  3.71784  3.717845  3.717852  3.717862  3.717876  3.717896  3.717922  3.717958  3.718007  3.718072  3.718159  3.718274  3.718425  3.718623   3.71888  3.719212  3.719636  3.720174  3.720855  3.721708   3.72277  3.724083  3.725695   3.72766   3.73004    3.7329  3.736316  3.740367  3.745137  3.750717    3.7572  3.764684  3.773266  3.783044  3.794113  3.806567  3.820494  3.835975  3.853085  3.871887  3.892437  3.914778  3.938942  3.964951  3.992811  4.022522  4.054068  4.087426  4.122561             4.159432 | 124.879264  124.879532  124.879909  124.880435  124.881166  124.882173  124.883552  124.885429  124.887964  124.891367  124.895905  124.901916  124.909822  124.920154  124.933563  124.950848  124.972975  125.001107  125.036628  125.081167  125.136629  125.205215  125.289441  125.392162  125.516577  125.666235  125.845035  126.057207  126.307298  126.600135  126.940785  127.334502  127.786673  128.302744  128.888156  129.548266  130.288275  131.113154  132.027575   133.03585  134.141879  135.349104  136.660474   138.07843  139.604887  141.241238  142.988361  144.846638  146.815981  148.895859 |  52.914676   52.914728   52.914802   52.914905   52.915048   52.915245   52.915514   52.915881   52.916376   52.917042   52.917928   52.919103   52.920648   52.922667   52.925287   52.928665   52.932988   52.938484   52.945423   52.954122   52.964953   52.978344   52.994783   53.014825   53.039091   53.068264   53.103098   53.144404   53.193051   53.249956   53.316077   53.392397   53.479914   53.579625   53.692511   53.819519   53.961546   54.119424   54.293907   54.485655   54.695224   54.923059   55.169489   55.434722   55.718847   56.021835   56.343544   56.683728   57.042042        57.418051 |  52.178399    52.17845   52.178522   52.178623   52.178763   52.178956    52.17922    52.17958   52.180066   52.180718   52.181588    52.18274   52.184255   52.186235   52.188804   52.192116   52.196355   52.201745   52.208549   52.217079   52.227699    52.24083    52.25695   52.276603   52.300397   52.329004   52.363162   52.403665   52.451368   52.507168   52.572005   52.646843    52.73266   52.830435   52.941129    53.06567   53.204939   53.359751   53.530846    53.71887   53.924369    54.14778   54.389424   54.649507   54.928114   55.225219   55.540681   55.874259   56.225614     56.594321 | 1.088387  1.088388  1.088389   1.08839  1.088393  1.088396    1.0884  1.088405  1.088413  1.088423  1.088436  1.088454  1.088478  1.088508  1.088548    1.0886  1.088665  1.088749  1.088854  1.088987  1.089151  1.089355  1.089605  1.089909  1.090278  1.090721   1.09125  1.091877  1.092615  1.093478   1.09448  1.095636  1.096961  1.098469  1.100174  1.102091  1.104231  1.106608   1.10923  1.112107  1.115246  1.118652  1.122329  1.126277  1.130496  1.134985  1.139737  1.144749  1.150013   1.15552 |  1.35916   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916  1.359159  1.359159  1.359158  1.359157  1.359156  1.359154  1.359152   1.35915  1.359146  1.359142  1.359136  1.359129  1.359121   1.35911  1.359096   1.35908   1.35906  1.359035  1.359006  1.358971   1.35893  1.358882  1.358825   1.35876  1.358685  1.358599  1.358502  1.358393  1.358271  1.358135  1.357986  1.357822  1.357644  1.357452  1.357245  1.357024   1.35679  1.356543  1.356284  1.356013  1.355732  1.355441          1.355142 | 0.536159   0.53616   0.53616  0.536161  0.536162  0.536164  0.536166  0.536169  0.536172  0.536178  0.536184  0.536193  0.536205  0.536221  0.536241  0.536267  0.536301  0.536343  0.536397  0.536464  0.536548  0.536651  0.536778  0.536933   0.53712  0.537344  0.537612  0.537929  0.538302  0.538737  0.539241  0.539822  0.540486   0.54124   0.54209  0.543042  0.544102  0.545273   0.54656  0.547964  0.549488  0.551132  0.552895  0.554774  0.556769  0.558873  0.561083  0.563393  0.565797      0.568289 | -0.201869   -0.20187   -0.20187  -0.201871  -0.201873  -0.201875  -0.201878  -0.201881  -0.201887  -0.201893  -0.201902  -0.201914   -0.20193  -0.201951  -0.201977  -0.202011  -0.202054  -0.202109  -0.202178  -0.202264  -0.202371  -0.202503  -0.202663  -0.202859  -0.203094  -0.203375  -0.203709  -0.204104  -0.204565  -0.205102  -0.205721  -0.206431  -0.207239  -0.208153   -0.20918  -0.210325  -0.211593   -0.21299  -0.214519  -0.216181  -0.217978  -0.219909  -0.221973  -0.224167  -0.226487  -0.228928  -0.231484   -0.23415  -0.236917     -0.239777 |  0.66571   0.66571   0.66571   0.66571   0.66571   0.66571  0.665711  0.665711  0.665711  0.665712  0.665713  0.665714  0.665716  0.665718  0.665721  0.665724  0.665729  0.665734  0.665741   0.66575  0.665761  0.665774   0.66579  0.665809  0.665832   0.66586  0.665893  0.665931  0.665976  0.666028  0.666088  0.666157  0.666235  0.666323  0.666422  0.666532  0.666653  0.666788  0.666935  0.667095  0.667269  0.667456  0.667656   0.66787  0.668098  0.668338  0.668592  0.668858  0.669136      0.669426 | 100.000544 | 100.000667  100.000958  100.001368  100.001941  100.002736  100.003832  100.005334  100.007376  100.010135   100.01384  100.018779  100.025321  100.033927  100.045173  100.059769  100.078585  100.102673  100.133301  100.171976  100.220477  100.280879  100.355586   100.44735  100.559291  100.694915  100.858118  101.053187  101.284787   101.55795  101.878036    102.2507  102.681846  103.177562  103.744063  104.387622  105.114494  105.930847  106.842688  107.855798  108.975664  110.207429  111.555845  113.025233  114.619455    116.3419  118.195473  120.182596  122.305217  124.564821          126.962454 | 128.191889 | 0.720701  0.720701  0.720702  0.720703  0.720705  0.720707   0.72071  0.720714  0.720719  0.720726  0.720735  0.720748  0.720764  0.720786  0.720813  0.720849  0.720895  0.720953  0.721027  0.721119  0.721234  0.721376   0.72155  0.721763   0.72202  0.722329  0.722697  0.723134  0.723649  0.724251  0.724949  0.725755  0.726678  0.727729  0.728918  0.730253  0.731745    0.7334  0.735227  0.737231  0.739416  0.741787  0.744346  0.747094  0.750029  0.753151  0.756456   0.75994  0.763598       0.767424 | 2.591124  2.591127  2.591131  2.591138  2.591146  2.591158  2.591175  2.591197  2.591228  2.591269  2.591323  2.591395   2.59149  2.591614  2.591774  2.591981  2.592246  2.592583  2.593009  2.593542  2.594206  2.595027  2.596035  2.597263   2.59875  2.600538  2.602673  2.605204  2.608184   2.61167  2.615719  2.620392  2.625748   2.63185  2.638755  2.646521  2.655201  2.664846  2.675499    2.6872   2.69998  2.713865  2.728872  2.745011  2.762286  2.780692  2.800217  2.820843  2.842547          2.8653 | 100.000544  100.000786  100.001126  100.001604   100.00227  100.003191  100.004458  100.006188  100.008535  100.011697  100.015931   100.02156  100.028995  100.038748  100.051456  100.067901  100.089039  100.116021  100.150226  100.193289  100.247129   100.31398  100.396412  100.497357  100.620124  100.768413  100.946316  101.158314   101.40926  101.704361  102.049136  102.449378    102.9111  103.440471  104.043756  104.727238   105.49715  106.359597  107.320492  108.385486  109.559908  110.848718   112.25646  113.787233  115.444667  117.231909  119.151623  121.205989  123.396721  125.725078  128.191889 |  100.0 |   0.000544 | 128.191889 |
    | 00:00 | 3.717838  3.717842  3.717848  3.717856  3.717867  3.717883  3.717905  3.717934  3.717974  3.718028    3.7181  3.718195  3.718322  3.718487  3.718702   3.71898  3.719337  3.719793  3.720369  3.721095  3.722002  3.723126  3.724512  3.726208  3.728268  3.730754  3.733734  3.737281  3.741475    3.7464  3.752145  3.758802  3.766467  3.775234    3.7852  3.796457  3.809095  3.823199  3.838847   3.85611  3.875048  3.895714  3.918148  3.942379  3.968426  3.996295  4.025983  4.057473  4.090741             4.125754 | 124.879143   124.87936  124.879665  124.880093  124.880688   124.88151  124.882639  124.884179  124.886265  124.889074  124.892829  124.897817  124.904397  124.913021  124.924246  124.938761    124.9574  124.981173  125.011288  125.049174  125.096511  125.155251  125.227639  125.316236  125.423932  125.553954  125.709869  125.895578  126.115301  126.373556  126.675123  127.025002  127.428362  127.890482  128.416685   129.01227  129.682439  130.432228  131.266434   132.18955  133.205713  134.318647  135.531629  136.847453  138.268417  139.796311  141.432417  143.177523  145.031939  146.995523 |  52.914652   52.914695   52.914754   52.914838   52.914954   52.915115   52.915336   52.915637   52.916044   52.916593   52.917327   52.918302   52.919588   52.921273   52.923467   52.926303   52.929945    52.93459   52.940473   52.947874   52.957119   52.968589   52.982721   53.000012   53.021023   53.046378   53.076767   53.112941   53.155708   53.205932   53.264519   53.332411   53.410575   53.499986   53.601615   53.716414   53.845298   53.989131   54.148711   54.324756   54.517892   54.728643   54.957424   55.204536   55.470164   55.754376   56.057127   56.378267   56.717539        57.074597 |  52.178375   52.178417   52.178475   52.178557   52.178671   52.178829   52.179045   52.179341    52.17974   52.180279   52.180998   52.181954   52.183215   52.184868   52.187019     52.1898   52.193371   52.197926   52.203695   52.210952   52.220018   52.231265   52.245122   52.262077    52.28268   52.307543   52.337342   52.372813    52.41475   52.463998   52.521448   52.588022   52.664668   52.752342   52.851998   52.964567   53.090948   53.231988   53.388469   53.561096   53.750481    53.95714   54.181478   54.423791   54.684261   54.962953   55.259826   55.574729   55.907413     56.257537 | 1.088387  1.088387  1.088388  1.088389  1.088391  1.088394  1.088397  1.088401  1.088408  1.088416  1.088427  1.088442  1.088462  1.088487  1.088521  1.088564  1.088619   1.08869  1.088779  1.088892  1.089032  1.089207  1.089421  1.089684  1.090003  1.090389   1.09085  1.091399  1.092048   1.09281  1.093699  1.094728  1.095911  1.097264  1.098801  1.100535  1.102479  1.104647  1.107048  1.109693  1.112591  1.115746  1.119166  1.122851  1.126804  1.131023  1.135507   1.14025  1.145247  1.150491 |  1.35916   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916  1.359159  1.359159  1.359159  1.359158  1.359157  1.359155  1.359154  1.359152  1.359149  1.359145   1.35914  1.359134  1.359127  1.359118  1.359106  1.359092  1.359075  1.359053  1.359028  1.358998  1.358962  1.358919  1.358869  1.358811  1.358744  1.358667   1.35858  1.358481   1.35837  1.358246  1.358109  1.357958  1.357793  1.357614  1.357421  1.357214  1.356993  1.356759  1.356513  1.356254  1.355984  1.355704          1.355415 | 0.536159  0.536159   0.53616   0.53616  0.536161  0.536163  0.536164  0.536167   0.53617  0.536174   0.53618  0.536187  0.536197   0.53621  0.536227  0.536249  0.536277  0.536313  0.536359  0.536416  0.536487  0.536576  0.536685  0.536818   0.53698  0.537176  0.537409  0.537687  0.538016    0.5384  0.538848  0.539366   0.53996  0.540638  0.541406   0.54227  0.543235  0.544307   0.54549  0.546787    0.5482   0.54973  0.551379  0.553144  0.555024  0.557016  0.559117   0.56132  0.563622      0.566014 | -0.201869  -0.201869   -0.20187  -0.201871  -0.201872  -0.201874  -0.201876  -0.201879  -0.201883  -0.201889  -0.201896  -0.201906  -0.201919  -0.201936  -0.201959  -0.201987  -0.202024   -0.20207  -0.202129  -0.202202  -0.202294  -0.202406  -0.202545  -0.202714  -0.202918  -0.203163  -0.203456  -0.203802  -0.204209  -0.204685  -0.205236   -0.20587  -0.206596   -0.20742  -0.208349   -0.20939  -0.210549  -0.211831   -0.21324  -0.214779  -0.216449  -0.218252  -0.220186  -0.222251  -0.224444  -0.226759  -0.229194  -0.231741  -0.234394     -0.237147 |  0.66571   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571  0.665711  0.665711  0.665712  0.665712  0.665713  0.665715  0.665717  0.665719  0.665722  0.665726   0.66573  0.665736  0.665744  0.665753  0.665764  0.665778  0.665795  0.665815  0.665839  0.665868  0.665902  0.665941  0.665988  0.666041  0.666102  0.666172  0.666252  0.666341  0.666441  0.666552  0.666675  0.666811  0.666959  0.667119  0.667293  0.667481  0.667681  0.667895  0.668122  0.668362  0.668615   0.66888      0.669156 | 100.000436 | 100.000535   100.00077  100.001103  100.001568  100.002215   100.00311  100.004339  100.006015  100.008286  100.011343   100.01543  100.020859  100.028022  100.037408  100.049628  100.065427  100.085718  100.111599  100.144386  100.185638  100.237187  100.301161  100.380015  100.476547  100.593919  100.735669  100.905713  101.108346  101.348224  101.630351  101.960042  102.342886  102.784695  103.291452  103.869243  104.524191  105.262392  106.089835  107.012343  108.035504  109.164611  110.404614  111.760076  113.235134  114.833484  116.558354  118.412508   120.39824  122.517387          124.771341 | 125.928277 | 0.720701  0.720701  0.720702  0.720703  0.720704  0.720706  0.720708  0.720711  0.720715  0.720721  0.720729  0.720739  0.720753  0.720771  0.720794  0.720824  0.720863  0.720912  0.720974  0.721053  0.721151  0.721273  0.721422  0.721606  0.721828  0.722097  0.722419  0.722801  0.723254  0.723785  0.724405  0.725122  0.725947   0.72689  0.727961  0.729169  0.730524  0.732034  0.733707  0.735549  0.737567  0.739764  0.742145   0.74471   0.74746  0.750395  0.753514  0.756812  0.760285       0.763929 | 2.591122  2.591125  2.591128  2.591134  2.591141   2.59115  2.591164  2.591182  2.591207  2.591241  2.591286  2.591346  2.591425  2.591528  2.591663  2.591836   2.59206  2.592345  2.592705  2.593159  2.593726  2.594429  2.595295  2.596355  2.597643  2.599197  2.601059  2.603276  2.605896  2.608973  2.612562  2.616719  2.621504  2.626977  2.633195  2.640217  2.648097  2.656887  2.666634  2.677382  2.689167  2.702017  2.715958  2.731005  2.747167  2.764445  2.782835  2.802323  2.822892        2.844518 | 100.000436  100.000631  100.000906  100.001294  100.001835  100.002586  100.003621  100.005039  100.006968  100.009573  100.013071  100.017734  100.023911  100.032038  100.042658  100.056444  100.074217  100.096975  100.125919  100.162477  100.208338  100.265474   100.33617  100.423047  100.529081  100.657619  100.812387  100.997492  101.217415  101.476992  101.781393  102.136081  102.546771  103.019378  103.559952  104.174619  104.869507  105.650678  106.524056  107.495364  108.570057  109.753272  111.049775  112.463925  113.999643  115.660396  117.449179  119.368521  121.420484  123.606678  125.928277 |  100.0 |   0.000436 | 125.928277 |
    | 00:30 | 3.717836  3.717839  3.717844   3.71785   3.71786  3.717873   3.71789  3.717914  3.717947  3.717992  3.718051   3.71813  3.718235  3.718373  3.718553  3.718786  3.719086  3.719471  3.719959  3.720576  3.721348   3.72231  3.723499  3.724959   3.72674  3.728896  3.731491  3.734591   3.73827  3.742607  3.747685  3.753594  3.760423  3.768266  3.777216  3.787365  3.798805  3.811621  3.825896  3.841705  3.859114  3.878182  3.898957  3.921477  3.945768  3.971848   3.99972   4.02938   4.06081             4.093986 | 124.879046  124.879221  124.879468  124.879815  124.880299  124.880969  124.881893  124.883155  124.884871  124.887186  124.890289  124.894424  124.899893  124.907082  124.916468   124.92864  124.944319  124.964379  124.989872  125.022049  125.062385  125.112608  125.174713  125.250992  125.344043  125.456789   125.59248  125.754697   125.94734  126.174615  126.441005  126.751242  127.110254  127.523122   127.99502  128.531145  129.136658  129.816608  130.575867  131.419065  132.350525   133.37421  134.493679   135.71205   137.03197  138.455601  139.984614  141.620189  143.363029  145.213376 |  52.914633   52.914667   52.914716   52.914784   52.914878   52.915009    52.91519   52.915437   52.915772   52.916224   52.916831   52.917639   52.918708   52.920113   52.921947   52.924325   52.927389   52.931309   52.936289   52.942575   52.950454   52.960262   52.972389   52.987279   53.005438   53.027431   53.053889   53.085501   53.123019   53.167247   53.219041   53.279297   53.348941   53.428922   53.520192   53.623696   53.740359   53.871063   54.016642   54.177858   54.355397   54.549852   54.761715   54.991373   55.239101   55.505061   55.789303   56.091769   56.412297        56.750625 |  52.178357    52.17839   52.178438   52.178504   52.178597   52.178725   52.178902   52.179144   52.179473   52.179917   52.180512   52.181304   52.182352    52.18373   52.185528   52.187861   52.190865   52.194708   52.199592   52.205756   52.213482     52.2231   52.234991   52.249592   52.267398   52.288964   52.314908   52.345907   52.382696   52.426065   52.476853   52.535938   52.604231   52.682658   52.772155    52.87365   52.988047   53.116214   53.258965   53.417051   53.591142    53.78182   53.989569   54.214767   54.457684    54.71848   54.997202   55.293795   55.608098     55.939856 | 1.088386  1.088387  1.088387  1.088388   1.08839  1.088392  1.088395  1.088398  1.088404   1.08841   1.08842  1.088432  1.088448   1.08847  1.088497  1.088534   1.08858   1.08864  1.088716  1.088811  1.088931   1.08908  1.089264  1.089491  1.089767  1.090101  1.090503  1.090983  1.091552  1.092224  1.093009  1.093923  1.094978  1.096189   1.09757  1.099135  1.100896  1.102868  1.105061  1.107486  1.110153   1.11307  1.116241  1.119672  1.123366  1.127323  1.131541  1.136019  1.140752  1.145733 |  1.35916   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916  1.359159  1.359159  1.359158  1.359157  1.359156  1.359155  1.359153  1.359151  1.359148  1.359144  1.359139  1.359132  1.359124  1.359115  1.359102  1.359087  1.359069  1.359047  1.359021  1.358989  1.358951  1.358907  1.358856  1.358796  1.358728  1.358649   1.35856  1.358459  1.358347  1.358221  1.358083  1.357931  1.357765  1.357585  1.357391  1.357184  1.356963  1.356729  1.356483  1.356225  1.355956          1.355677 | 0.536159  0.536159   0.53616   0.53616  0.536161  0.536162  0.536163  0.536165  0.536168  0.536171  0.536176  0.536182   0.53619  0.536201  0.536215  0.536234  0.536258  0.536288  0.536326  0.536375  0.536436  0.536512  0.536605   0.53672   0.53686   0.53703  0.537233  0.537477  0.537765  0.538104  0.538501  0.538961  0.539492    0.5401  0.540791  0.541572  0.542449  0.543428  0.544512  0.545705  0.547011  0.548432  0.549969  0.551622   0.55339   0.55527   0.55726  0.559356  0.561552      0.563844 | -0.201869  -0.201869  -0.201869   -0.20187  -0.201871  -0.201872  -0.201874  -0.201877   -0.20188  -0.201885  -0.201891  -0.201899   -0.20191  -0.201925  -0.201943  -0.201967  -0.201998  -0.202037  -0.202087  -0.202149  -0.202227  -0.202324  -0.202443  -0.202589  -0.202766  -0.202979  -0.203235  -0.203538  -0.203897  -0.204317  -0.204806  -0.205372  -0.206021  -0.206762  -0.207601  -0.208545  -0.209601  -0.210774  -0.212068  -0.213488  -0.215036  -0.216714  -0.218522   -0.22046  -0.222525  -0.224716  -0.227027  -0.229454  -0.231991     -0.234633 |  0.66571   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571  0.665711  0.665711  0.665712  0.665713  0.665714  0.665715  0.665717   0.66572  0.665723  0.665727  0.665732  0.665738  0.665746  0.665756  0.665768  0.665782    0.6658  0.665821  0.665846  0.665876  0.665911  0.665952  0.665999  0.666054  0.666117  0.666188  0.666269  0.666359  0.666461  0.666573  0.666697  0.666833  0.666982  0.667144  0.667318  0.667506  0.667706   0.66792  0.668146  0.668386  0.668637      0.668901 | 100.000349 | 100.000429  100.000619  100.000888  100.001265  100.001792  100.002522  100.003527  100.004901  100.006768  100.009288  100.012666  100.017166  100.023119  100.030944  100.041161   100.05441  100.071478  100.093316  100.121069  100.156102  100.200024  100.254716   100.32236  100.405457  100.506848  100.629734   100.77768  100.954617  101.164843  101.413003  101.704067  102.043301  102.436224  102.888558  103.406175  103.995031  104.661108  105.410336  106.248536  107.181348  108.214174   109.35212  110.599949  111.962043  113.442367  115.044452  116.771379  118.625774  120.609811          122.725222 |  123.81218 |   0.7207  0.720701  0.720701  0.720702  0.720703  0.720704  0.720706  0.720709  0.720713  0.720717  0.720724  0.720732  0.720744  0.720759  0.720778  0.720803  0.720836  0.720877   0.72093  0.720997   0.72108  0.721184  0.721313  0.721471  0.721663  0.721896  0.722176  0.722511  0.722908  0.723376  0.723924  0.724561  0.725296   0.72614  0.727103  0.728193  0.729421  0.730795  0.732323  0.734012   0.73587  0.737901  0.740109  0.742497  0.745068  0.747821  0.750756   0.75387  0.757161       0.760623 | 2.591121  2.591123  2.591126   2.59113  2.591136  2.591144  2.591155   2.59117  2.591191  2.591219  2.591256  2.591305  2.591371  2.591457  2.591569  2.591715  2.591903  2.592143  2.592449  2.592834  2.593317  2.593918  2.594662  2.595575  2.596688  2.598036  2.599657  2.601595  2.603894  2.606603  2.609776  2.613467  2.617731  2.622627  2.628213  2.634546  2.641681  2.649672  2.658568  2.668414  2.679252  2.691116  2.704033  2.718026  2.733109  2.749289  2.766568  2.784938  2.804387        2.824897 | 100.000349  100.000506  100.000729  100.001043  100.001482  100.002094  100.002939    100.0041  100.005683  100.007827  100.010714  100.014573  100.019699  100.026462  100.035326  100.046866   100.06179  100.080958  100.105414  100.136403  100.175405  100.224159  100.284687  100.359327  100.450743  100.561954   100.69634  100.857649  101.049999  101.277865  101.546062  101.859718  102.224239  102.645258  103.128589  103.680165  104.305971  105.011982  105.804094  106.688053  107.669397  108.753395  109.944994  111.248775  112.668921  114.209184  115.872872   117.66284  119.581487  121.630762   123.81218 |  100.0 |   0.000349 |  123.81218 |
    | 01:00 | 3.717834  3.717837  3.717841  3.717846  3.717854  3.717864  3.717879  3.717898  3.717925  3.717962  3.718011  3.718077  3.718164  3.718278  3.718429  3.718624  3.718876    3.7192  3.719613  3.720136  3.720793  3.721614  3.722633  3.723888  3.725424  3.727291  3.729545  3.732249  3.735469   3.73928   3.74376  3.748992  3.755063  3.762062   3.77008  3.779209  3.789538  3.801156  3.814145  3.828586  3.844548  3.862098  3.881288  3.902165  3.924765  3.949111  3.975217  4.003087  4.032714              4.06408 | 124.878967  124.879108  124.879308   124.87959  124.879983   124.88053  124.881284  124.882318  124.883727  124.885633  124.888195  124.891618  124.896159  124.902145  124.909982  124.920176  124.933347   124.95025  124.971799  124.999084  125.033401  125.076271  125.129464  125.195021  125.275274  125.372856  125.490721  125.632138  125.800696  126.000291  126.235109  126.509599  126.828436  127.196479  127.618718  128.100218  128.646055  129.261249  129.950703  130.719129  131.570992   132.51045  133.541303  134.666949  135.890355  137.214027  138.639999   140.16983  141.804603  143.544942 |  52.914618   52.914646   52.914685    52.91474   52.914817   52.914923   52.915071   52.915273   52.915548   52.915921   52.916422    52.91709   52.917978   52.919148   52.920679   52.922671   52.925245   52.928548   52.932758   52.938089   52.944793   52.953166   52.963554   52.976354   52.992018   53.011059   53.034049   53.061619   53.094462   53.133327   53.179013   53.232369   53.294279   53.365654   53.447423   53.540517   53.645855   53.764332   53.896803   54.044067   54.206855   54.385821   54.581527   54.794434   55.024903   55.273183   55.539417   55.823637   56.125772        56.445648 |  52.178342   52.178369   52.178407   52.178461   52.178536   52.178641   52.178786   52.178984   52.179254   52.179619    52.18011   52.180766   52.181636   52.182784   52.184285   52.186239   52.188763   52.192001    52.19613   52.201357   52.207931   52.216141   52.226327   52.238878   52.254239    52.27291   52.295453   52.322488   52.354693   52.392803   52.437603   52.489923    52.55063   52.620619     52.7008   52.792086   52.895379   53.011555   53.141453   53.285857   53.445485   53.620975    53.81288   54.021653   54.247646   54.491105   54.752169    55.03087   55.327137     55.640802 | 1.088386  1.088386  1.088387  1.088388  1.088389  1.088391  1.088393  1.088396    1.0884  1.088406  1.088413  1.088424  1.088437  1.088455  1.088478  1.088508  1.088548  1.088598  1.088662  1.088743  1.088845  1.088972   1.08913  1.089325  1.089563  1.089852  1.090201   1.09062  1.091119  1.091709  1.092402  1.093211   1.09415  1.095231  1.096469  1.097877  1.099469  1.101258  1.103256  1.105474  1.107922   1.11061  1.113544   1.11673  1.120173  1.123874  1.127833   1.13205  1.136521  1.141243 | 1.359161   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916  1.359159  1.359159  1.359158  1.359157  1.359156  1.359154  1.359152   1.35915  1.359146  1.359142  1.359137   1.35913  1.359122  1.359111  1.359098  1.359083  1.359063   1.35904  1.359013   1.35898  1.358941  1.358896  1.358843  1.358782  1.358711  1.358631   1.35854  1.358438  1.358324  1.358197  1.358057  1.357903  1.357736  1.357556  1.357361  1.357153  1.356932  1.356699  1.356453  1.356196          1.355928 | 0.536159  0.536159  0.536159   0.53616   0.53616  0.536161  0.536162  0.536164  0.536166  0.536169  0.536173  0.536178  0.536185  0.536194  0.536206  0.536221  0.536241  0.536267  0.536299   0.53634  0.536392  0.536457  0.536537  0.536636  0.536757  0.536904  0.537081  0.537293  0.537545  0.537844  0.538194  0.538603  0.539075  0.539619   0.54024  0.540945  0.541739  0.542629   0.54362  0.544715  0.545919  0.547234  0.548663  0.550206  0.551862  0.553632  0.555512  0.557499   0.55959      0.561779 | -0.201868  -0.201869  -0.201869   -0.20187   -0.20187  -0.201872  -0.201873  -0.201875  -0.201878  -0.201882  -0.201887  -0.201894  -0.201903  -0.201915   -0.20193   -0.20195  -0.201976  -0.202009  -0.202052  -0.202105  -0.202171  -0.202254  -0.202356  -0.202482  -0.202635   -0.20282  -0.203042  -0.203308  -0.203623  -0.203993  -0.204426  -0.204929  -0.205509  -0.206173  -0.206929  -0.207783  -0.208741  -0.209811  -0.210997  -0.212303  -0.213734  -0.215291  -0.216976  -0.218789  -0.220729  -0.222795  -0.224983  -0.227289  -0.229709     -0.232236 |  0.66571   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571  0.665711  0.665711  0.665711  0.665712  0.665713  0.665714  0.665716  0.665718  0.665721  0.665724  0.665728  0.665734   0.66574  0.665749  0.665759  0.665772  0.665787  0.665805  0.665827  0.665853  0.665884   0.66592  0.665962  0.666011  0.666067  0.666131  0.666203  0.666285  0.666377   0.66648  0.666594  0.666719  0.666856  0.667005  0.667167  0.667342   0.66753   0.66773  0.667944   0.66817  0.668408      0.668659 | 100.000279 | 100.000343  100.000497  100.000714   100.00102  100.001449  100.002043  100.002864   100.00399  100.005523  100.007598  100.010387  100.014112  100.019055   100.02557  100.034101  100.045197  100.059534  100.077935  100.101393  100.131099  100.168463  100.215145  100.273075  100.344482  100.431913   100.53825  100.666726  100.820929  101.004804  101.222646  101.479082  101.779047  102.127755   102.53065  102.993362  103.521654  104.121352  104.798294  105.558252  106.406879  107.349639  108.391753  109.538147  110.793404  112.161732  113.646931  115.252377  116.981007  118.835319          120.817374 | 121.836899 |   0.7207  0.720701  0.720701  0.720702  0.720702  0.720704  0.720705  0.720707   0.72071  0.720714  0.720719  0.720727  0.720736  0.720748  0.720765  0.720786  0.720813  0.720848  0.720893  0.720949   0.72102  0.721109  0.721219  0.721355  0.721521  0.721723  0.721966  0.722258  0.722606  0.723017  0.723501  0.724065  0.724719  0.725473  0.726336  0.727317  0.728427  0.729673  0.731065   0.73261  0.734316  0.736188  0.738231  0.740449  0.742846  0.745421  0.748176   0.75111  0.754219       0.757502 |  2.59112  2.591122  2.591124  2.591127  2.591132  2.591139  2.591148   2.59116  2.591177    2.5912  2.591231  2.591272  2.591326  2.591398  2.591492  2.591614  2.591772  2.591974  2.592232  2.592559   2.59297  2.593483   2.59412  2.594905  2.595865  2.597032  2.598441  2.600131  2.602144  2.604525  2.607324  2.610593  2.614384  2.618754   2.62376  2.629457  2.635901  2.643147  2.651245  2.660243  2.670185  2.681109  2.693048  2.706027  2.720068  2.735183  2.751379  2.768654  2.787002         2.80641 | 100.000279  100.000406  100.000585   100.00084  100.001197  100.001694  100.002384  100.003333  100.004631  100.006394  100.008773  100.011963  100.016211  100.021833  100.029222   100.03887  100.051384  100.067507  100.088141   100.11437  100.147488  100.189022   100.24076  100.304776  100.383452  100.479498  100.595969  100.736273  100.904178  101.103806  101.339625  101.616422  101.939283  102.313549  102.744771  103.238662  103.801032  104.437734  105.154588  105.957325  106.851522  107.842534   108.93545  110.135036  111.445695   112.87144  114.415863  116.082121  117.872934  119.790579  121.836899 |  100.0 |   0.000279 | 121.836899 |
    | 01:30 | 3.717833  3.717835  3.717838  3.717843  3.717849  3.717857  3.717869  3.717885  3.717907  3.717937  3.717978  3.718032  3.718104    3.7182  3.718325  3.718488    3.7187  3.718973  3.719321  3.719764  3.720322  3.721022  3.721893   3.72297  3.724293  3.725906  3.727861  3.730214  3.733028   3.73637  3.740313  3.744936   3.75032  3.756551  3.763718  3.771908  3.781212  3.791717  3.803508  3.816666  3.831265  3.847376  3.865059  3.884366  3.905339  3.928011  3.952406  3.978533  4.006396             4.035986 | 124.878904  124.879018   124.87918  124.879407  124.879727  124.880172  124.880787  124.881634  124.882789  124.884357  124.886471  124.889301  124.893067  124.898044   124.90458  124.913107  124.924156  124.938379  124.956568  124.979671  125.008822  125.045357  125.090841  125.147087  125.216179  125.300484  125.402671  125.525717  125.672909  125.847839  126.054397  126.296744   126.57929  126.906653  127.283618  127.715085   128.20601  128.761346  129.385978  130.084661  130.861954  131.722162  132.669284   133.70696  134.838437  136.066536   137.39363  138.821633  140.351995   141.98571 |  52.914606   52.914628   52.914659   52.914704   52.914766   52.914853   52.914974   52.915139   52.915365   52.915671   52.916084   52.916638   52.917374   52.918346   52.919624    52.92129   52.923449   52.926229   52.929782   52.934296   52.939991   52.947128   52.956012   52.966995   52.980484   52.996938   53.016876   53.040873   53.069565   53.103644   53.143857   53.190999   53.245908   53.309456    53.38254   53.466068   53.560949   53.668078   53.788322   53.922504   54.071394   54.235693   54.416021   54.612911   54.826797   55.058013   55.306786   55.573237   55.857385        56.159144 |   52.17833   52.178351   52.178382   52.178426   52.178487   52.178573   52.178691   52.178853   52.179074   52.179375    52.17978   52.180322   52.181044   52.181998    52.18325   52.184884   52.187001   52.189727   52.193212   52.197638   52.203222   52.210221   52.218932   52.229702   52.242928   52.259063   52.278613   52.302145    52.33028   52.363697   52.403129   52.449355   52.503198   52.565512   52.637177   52.719083   52.812122    52.91717   53.035079   53.166655   53.312654   53.473762   53.650589   53.843655   54.053388   54.280114   54.524055   54.785333   55.063962     55.359861 | 1.088386  1.088386  1.088387  1.088387  1.088388   1.08839  1.088391  1.088394  1.088397  1.088402  1.088408  1.088417  1.088428  1.088443  1.088462  1.088487   1.08852  1.088563  1.088617  1.088685  1.088772   1.08888  1.089015  1.089182  1.089387  1.089637   1.08994  1.090305  1.090741  1.091258  1.091869  1.092584  1.093417   1.09438  1.095487  1.096751  1.098186  1.099805   1.10162  1.103643  1.105885  1.108356  1.111063  1.114014  1.117214  1.120667  1.124374  1.128335   1.13255  1.137015 | 1.359161   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916  1.359159  1.359159  1.359158  1.359158  1.359157  1.359155  1.359154  1.359152  1.359149  1.359145  1.359141  1.359135  1.359128  1.359119  1.359108  1.359094  1.359078  1.359058  1.359034  1.359005  1.358971  1.358931  1.358884  1.358829  1.358767  1.358695  1.358613   1.35852  1.358416  1.358301  1.358172  1.358031  1.357876  1.357708  1.357527  1.357332  1.357124  1.356903  1.356669  1.356424          1.356168 | 0.536159  0.536159  0.536159  0.536159   0.53616  0.536161  0.536162  0.536163  0.536165  0.536167   0.53617  0.536174   0.53618  0.536188  0.536198   0.53621  0.536227  0.536249  0.536276  0.536311  0.536355   0.53641  0.536479  0.536564  0.536668  0.536795  0.536948  0.537133  0.537354  0.537616  0.537925  0.538286  0.538706  0.539191  0.539747  0.540381  0.541099  0.541907  0.542809  0.543811  0.544918  0.546132  0.547456  0.548891  0.550439  0.552099   0.55387  0.555749  0.557734      0.559819 | -0.201868  -0.201869  -0.201869  -0.201869   -0.20187  -0.201871  -0.201872  -0.201874  -0.201876  -0.201879  -0.201884  -0.201889  -0.201897  -0.201907   -0.20192  -0.201936  -0.201958  -0.201986  -0.202022  -0.202067  -0.202123  -0.202194  -0.202282   -0.20239  -0.202522  -0.202682  -0.202876  -0.203107  -0.203383  -0.203709  -0.204092  -0.204538  -0.205054  -0.205648  -0.206326  -0.207096  -0.207965  -0.208938  -0.210021  -0.211219  -0.212537  -0.213978  -0.215543  -0.217234  -0.219052  -0.220994  -0.223059  -0.225245  -0.227546     -0.229958 |  0.66571   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571  0.665711  0.665711  0.665712  0.665712  0.665714  0.665715  0.665717  0.665719  0.665722  0.665725   0.66573  0.665736  0.665743  0.665751  0.665762  0.665775  0.665791   0.66581  0.665833   0.66586  0.665892  0.665929  0.665973  0.666023   0.66608  0.666145  0.666219  0.666302  0.666396  0.666499  0.666614   0.66674  0.666878  0.667028  0.667191  0.667366  0.667554  0.667754  0.667967  0.668193       0.66843 | 100.000223 | 100.000275  100.000398  100.000574  100.000822   100.00117  100.001654  100.002324  100.003245  100.004503  100.006209   100.00851   100.01159  100.015689  100.021107  100.028221  100.037502  100.049529  100.065012  100.084812  100.109963  100.141701  100.181481  100.231011  100.292269   100.36753  100.459382  100.570745  100.704881  100.865397  101.056247  101.281719  101.546418  101.855241  102.213345  102.626099   103.09904  103.637817  104.248131  104.935675  105.706069  106.564798  107.517157  108.568192  109.722653  110.984952  112.359129  113.848828  115.457275  117.187273          119.041196 | 119.995856 |   0.7207    0.7207  0.720701  0.720701  0.720702  0.720703  0.720704  0.720706  0.720708  0.720711  0.720716  0.720722   0.72073   0.72074  0.720753  0.720771  0.720794  0.720823  0.720861  0.720909  0.720969  0.721045  0.721139  0.721256  0.721399  0.721573  0.721784  0.722038  0.722342  0.722703  0.723129  0.723627  0.724208  0.724879  0.725651  0.726532  0.727532  0.728661  0.729925  0.731335  0.732897  0.734618  0.736503  0.738558  0.740786  0.743189  0.745769  0.748526  0.751457       0.754562 | 2.591119  2.591121  2.591123  2.591125  2.591129  2.591134  2.591142  2.591152  2.591166  2.591185   2.59121  2.591244  2.591289  2.591349  2.591427  2.591529  2.591661  2.591832   2.59205  2.592327  2.592676  2.593113  2.593658  2.594331  2.595158  2.596167  2.597389   2.59886  2.600618  2.602706   2.60517  2.608058  2.611422  2.615313  2.619788  2.624901  2.630707  2.637261  2.644613  2.652815  2.661912  2.671946  2.682952  2.694962     2.708  2.722085  2.737228  2.753435  2.770704        2.789027 | 100.000223  100.000325   100.00047  100.000676  100.000965   100.00137  100.001931  100.002707   100.00377  100.005218  100.007177  100.009811  100.013328  100.017995  100.024147  100.032203  100.042683  100.056226   100.07361  100.095778  100.123857  100.159186  100.203339  100.258152  100.325747  100.408549  100.509309  100.631115  100.777401  100.951949  101.158883  101.402656  101.688027  102.020033  102.403949  102.845243  103.349523  103.922482  104.569833   105.29725  106.110303  107.014398  108.014719  109.116177  110.323364  111.640515  113.071478   114.61969   116.28817  118.079504  119.995856 |  100.0 |   0.000223 | 119.995856 |
    | 02:00 | 3.717832  3.717834  3.717836   3.71784  3.717845  3.717852  3.717861  3.717875  3.717893  3.717917  3.717951  3.717996  3.718055  3.718135  3.718239  3.718375  3.718553  3.718782  3.719076   3.71945  3.719924  3.720519  3.721263  3.722185  3.723322  3.724714  3.726406  3.728451  3.730904  3.733828  3.737291  3.741367  3.746132  3.751667  3.758057  3.765389  3.773749  3.783225  3.793901  3.805861  3.819181  3.833934  3.850187  3.867997  3.887414  3.908477  3.931217  3.955654  3.981798             4.009649 | 124.878854  124.878945  124.879076   124.87926  124.879519  124.879881  124.880383  124.881075  124.882022   124.88331  124.885051  124.887389  124.890509  124.894643  124.900087   124.90721  124.916467  124.928419  124.943749  124.963283  124.988008  125.019096  125.057925  125.106102  125.165483  125.238187   125.32662   125.43348  125.561764  125.714773  125.896102  126.109625  126.359479   126.65003  126.985839  127.371614  127.812162  128.312331  128.876953   129.51078  130.218421  131.004286  131.872526  132.826984   133.87115  135.008124  136.240588  137.570789  139.000524  140.531144 |  52.914596   52.914614   52.914639   52.914675   52.914726   52.914797   52.914895    52.91503   52.915215   52.915467   52.915807   52.916264   52.916874   52.917682   52.918746   52.920138   52.921947   52.924282   52.927278   52.931095   52.935925   52.941998   52.949583   52.958992   52.970587    52.98478   53.002038   53.022885   53.047901   53.077723   53.113043   53.154604   53.203195   53.259646   53.324816   53.399585   53.484843   53.581476   53.690353   53.812314   53.948155   54.098614   54.264361   54.445989      54.644   54.858802   55.090704   55.339911   55.606527        55.890554 |   52.17832   52.178338   52.178363   52.178398   52.178448   52.178517   52.178613   52.178746   52.178927   52.179174   52.179508   52.179956   52.180554   52.181346   52.182389   52.183754   52.185528   52.187818   52.190756   52.194498   52.199235    52.20519   52.212628   52.221854   52.233224   52.247141   52.264064   52.284506   52.309036   52.338279   52.372913   52.413667   52.461315    52.51667   52.580574   52.653891   52.737494    52.83225   52.939013   53.058605   53.191808   53.339345   53.501874   53.679975    53.87414   54.084771    54.31217   54.556537   54.817976     55.096487 | 1.088386  1.088386  1.088386  1.088387  1.088388  1.088389   1.08839  1.088392  1.088395  1.088399  1.088404  1.088411   1.08842  1.088433  1.088449   1.08847  1.088497  1.088533  1.088579  1.088637   1.08871  1.088802  1.088918  1.089061  1.089237  1.089453  1.089715  1.090032  1.090412  1.090865  1.091401  1.092032  1.092769  1.093625  1.094613  1.095745  1.097035  1.098497  1.100141  1.101982   1.10403  1.106295  1.108786  1.111513   1.11448  1.117693  1.121154  1.124867   1.12883  1.133041 | 1.359161  1.359161   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916  1.359159  1.359159  1.359158  1.359157  1.359156  1.359155  1.359153  1.359151  1.359148  1.359144  1.359139  1.359133  1.359125  1.359116  1.359104   1.35909  1.359073  1.359052  1.359027  1.358997  1.358961   1.35892  1.358872  1.358816  1.358751  1.358678  1.358594    1.3585  1.358395  1.358278  1.358148  1.358005   1.35785  1.357681  1.357498  1.357303  1.357095  1.356874   1.35664          1.356396 | 0.536159  0.536159  0.536159  0.536159   0.53616   0.53616  0.536161  0.536162  0.536163  0.536165  0.536168  0.536172  0.536176  0.536182  0.536191  0.536201  0.536215  0.536234  0.536257  0.536286  0.536324  0.536371  0.536429  0.536502  0.536591  0.536701  0.536834  0.536995  0.537187  0.537417  0.537688  0.538007  0.538379  0.538811  0.539308  0.539877  0.540524  0.541254  0.542074  0.542989  0.544002  0.545119  0.546343  0.547675  0.549117   0.55067  0.552333  0.554105  0.555983      0.557964 | -0.201868  -0.201868  -0.201869  -0.201869   -0.20187   -0.20187  -0.201871  -0.201873  -0.201875  -0.201877  -0.201881  -0.201885  -0.201892    -0.2019  -0.201911  -0.201925  -0.201943  -0.201966  -0.201997  -0.202035  -0.202083  -0.202143  -0.202218  -0.202311  -0.202424  -0.202563  -0.202731  -0.202933  -0.203174   -0.20346  -0.203797  -0.204192   -0.20465   -0.20518  -0.205788  -0.206481  -0.207265  -0.208147  -0.209134   -0.21023  -0.211441   -0.21277  -0.214219  -0.215793   -0.21749  -0.219311  -0.221255   -0.22332  -0.225502     -0.227798 | 0.665709   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571  0.665711  0.665711  0.665712  0.665713  0.665714  0.665715  0.665717   0.66572  0.665723  0.665727  0.665731  0.665738  0.665745  0.665754  0.665766  0.665779  0.665796  0.665816   0.66584  0.665868    0.6659  0.665939  0.665983  0.666034  0.666093   0.66616  0.666235  0.666319  0.666414  0.666519  0.666634  0.666761    0.6669  0.667051  0.667214   0.66739  0.667577  0.667778  0.667991      0.668215 | 100.000178 |  100.00022  100.000319  100.000461  100.000662  100.000944  100.001338  100.001884  100.002637  100.003668   100.00507  100.006965   100.00951  100.012905  100.017405  100.023331  100.031083  100.041159  100.054169  100.070858  100.092122   100.11904  100.152887  100.195167  100.247632  100.312306  100.391507  100.487862  100.604326  100.744185  100.911064  101.108918  101.342025  101.614967  101.932598  102.300015  102.722509  103.205523  103.754594  104.375297  105.073182  105.853718   106.72223  107.683846  108.743443  109.905602   111.17457  112.554227  114.048065  115.659169          117.390214 | 118.282595 |   0.7207    0.7207  0.720701  0.720701  0.720701  0.720702  0.720703  0.720705  0.720707  0.720709  0.720713  0.720718  0.720724  0.720733  0.720744  0.720759  0.720778  0.720803  0.720835  0.720875  0.720926  0.720991  0.721071  0.721171  0.721294  0.721444  0.721627  0.721848  0.722113  0.722429  0.722803  0.723242  0.723756  0.724353  0.725042  0.725831   0.72673  0.727749  0.728895  0.730178  0.731604  0.733182  0.734918  0.736816  0.738882  0.741119  0.743529  0.746112  0.748869       0.751799 | 2.591119   2.59112  2.591121  2.591124  2.591127  2.591131  2.591137  2.591145  2.591157  2.591172  2.591193  2.591221  2.591258  2.591308  2.591373  2.591458  2.591569  2.591713  2.591896   2.59213  2.592426  2.592799  2.593264  2.593841  2.594551  2.595421  2.596479  2.597757   2.59929  2.601118  2.603282  2.605829  2.608806  2.612263  2.616254  2.620832   2.62605  2.631963  2.638623   2.64608  2.654383  2.663575  2.673696   2.68478  2.696857   2.70995  2.724075  2.739244  2.755459        2.772719 | 100.000178   100.00026  100.000377  100.000543  100.000778  100.001106  100.001564  100.002197  100.003066  100.004254  100.005866  100.008038  100.010947  100.014817  100.019933  100.026651  100.035416  100.046776  100.061403  100.080112  100.103884  100.133888  100.171508  100.218366  100.276342  100.347603  100.434617   100.54017  100.667382  100.819708   101.00094  101.215198  101.466919  101.760829  102.101913  102.495379  102.946608  103.461104  104.044441  104.702197    105.4399   106.26296  107.176622  108.185899  109.295534   110.50995  111.833219  113.269031   114.82068  116.491046  118.282595 |  100.0 |   0.000178 | 118.282595 |
    | 02:30 | 3.717831  3.717833  3.717835  3.717838  3.717842  3.717847  3.717855  3.717866  3.717881  3.717901  3.717928  3.717965  3.718015  3.718081  3.718167  3.718281   3.71843  3.718622  3.718869  3.719185  3.719586  3.720092  3.720726  3.721515  3.722491  3.723689  3.725152  3.726924  3.729059  3.731613  3.734648  3.738234  3.742441  3.747347  3.753033  3.759581  3.767075  3.775602  3.785246  3.796089  3.808212   3.82169  3.836592  3.852981  3.870913  3.890433   3.91158  3.934381  3.958855              3.98501 | 124.878813  124.878887  124.878992  124.879141  124.879351  124.879645  124.880054  124.880619  124.881394  124.882452  124.883885  124.885814  124.888395  124.891825  124.896354  124.902297  124.910042  124.920072  124.932976  124.949468  124.970408   124.99682  125.029917  125.071115  125.122063  125.184654  125.261047  125.353679  125.465273  125.598848  125.757711  125.945455  126.165941  126.423273  126.721774  127.065943  127.460409  127.909888  128.419119  128.992814  129.635593  130.351925  131.146072  132.022038  132.983515  134.033848  135.175996  136.412509  137.745512  139.176696 |  52.914588   52.914602   52.914623   52.914652   52.914693   52.914751    52.91483   52.914941   52.915092   52.915299   52.915579   52.915956   52.916461   52.917131   52.918016   52.919177   52.920691   52.922651   52.925173   52.928395   52.932487   52.937647   52.944112   52.952159   52.962109    52.97433   52.989242   53.007318   53.029086    53.05513   53.086088   53.122652    53.16556   53.215595   53.273576    53.34035   53.416781   53.503738   53.602085   53.712668   53.836298   53.973744   54.125715   54.292853   54.475719   54.674789   54.890446   55.122976   55.372562        55.639292 |  52.178312   52.178326   52.178346   52.178375   52.178415   52.178472    52.17855   52.178658   52.178807    52.17901   52.179284   52.179654   52.180148   52.180806   52.181674   52.182813   52.184297   52.186219   52.188692   52.191852   52.195863   52.200923   52.207263   52.215154    52.22491   52.236894   52.251516   52.269241   52.290587   52.316125   52.346482   52.382336    52.42441   52.473474   52.530329   52.595806   52.670753   52.756021   52.852459   52.960894   53.082124     53.2169    53.36592   53.529812   53.709127   53.904332   54.115801   54.343815   54.588554     54.850104 | 1.088386  1.088386  1.088386  1.088386  1.088387  1.088388  1.088389  1.088391  1.088393  1.088396  1.088401  1.088406  1.088414  1.088424  1.088438  1.088455  1.088478  1.088508  1.088546  1.088595  1.088658  1.088736  1.088834  1.088957  1.089108  1.089294   1.08952  1.089795  1.090126  1.090522  1.090992  1.091547  1.092198  1.092957  1.093836  1.094848  1.096005  1.097321  1.098808  1.100478  1.102344  1.104415  1.106702  1.109214  1.111958  1.114941  1.118165  1.121636  1.125353  1.129316 | 1.359161  1.359161   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916  1.359159  1.359159  1.359158  1.359157  1.359156  1.359154  1.359152   1.35915  1.359147  1.359142  1.359137  1.359131  1.359123  1.359113    1.3591  1.359085  1.359067  1.359045  1.359019  1.358988  1.358952  1.358909  1.358859  1.358802  1.358736  1.358661  1.358576   1.35848  1.358373  1.358255  1.358124   1.35798  1.357823  1.357653   1.35747  1.357274  1.357066  1.356845          1.356612 | 0.536159  0.536159  0.536159  0.536159  0.536159   0.53616   0.53616  0.536161  0.536162  0.536164  0.536166  0.536169  0.536173  0.536178  0.536185  0.536194  0.536206  0.536221   0.53624  0.536265  0.536297  0.536337  0.536387  0.536449  0.536526   0.53662  0.536735  0.536875  0.537042  0.537243  0.537481  0.537762  0.538091  0.538474  0.538917  0.539426  0.540007  0.540667   0.54141  0.542242  0.543168  0.544193   0.54532  0.546552  0.547892   0.54934  0.550898  0.552563  0.554335      0.556212 | -0.201868  -0.201868  -0.201868  -0.201869  -0.201869   -0.20187  -0.201871  -0.201872  -0.201873  -0.201875  -0.201878  -0.201882  -0.201887  -0.201894  -0.201903  -0.201915   -0.20193   -0.20195  -0.201975  -0.202008  -0.202048    -0.2021  -0.202164  -0.202243  -0.202341  -0.202461  -0.202606  -0.202782  -0.202992  -0.203242  -0.203539  -0.203887  -0.204294  -0.204765  -0.205308  -0.205929  -0.206636  -0.207434   -0.20833   -0.20933  -0.210439  -0.211661     -0.213  -0.214459   -0.21604  -0.217742  -0.219567  -0.221512  -0.223576     -0.225755 | 0.665709   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571  0.665711  0.665711  0.665712  0.665712  0.665713  0.665714  0.665716  0.665718  0.665721  0.665724  0.665728  0.665733   0.66574  0.665748  0.665757  0.665769  0.665784  0.665801  0.665822  0.665846  0.665875  0.665909  0.665948  0.665994  0.666046  0.666106  0.666174   0.66625  0.666336  0.666432  0.666538  0.666654  0.666782  0.666922  0.667073  0.667237  0.667413  0.667601  0.667801      0.668013 | 100.000142 | 100.000176  100.000256   100.00037  100.000533  100.000761  100.001081  100.001526  100.002141  100.002985  100.004136  100.005695  100.007795  100.010604  100.014337  100.019267  100.025735  100.034166  100.045084   100.05913  100.077083  100.099879  100.128634  100.164669   100.20953  100.265014  100.333189  100.416413   100.51735  100.638983  100.784624  100.957905  101.162786  101.403529  101.684686  102.011067  102.387707  102.819817  103.312746  103.871918  104.502781  105.210747  106.001137  106.879118  107.849656  108.917465  110.086964  111.362239  112.747019  114.244649          115.858081 |  116.69078 |   0.7207    0.7207    0.7207  0.720701  0.720701  0.720702  0.720703  0.720704  0.720705  0.720708   0.72071  0.720714   0.72072  0.720727  0.720736  0.720749  0.720765  0.720785  0.720812  0.720846   0.72089  0.720945  0.721013  0.721098  0.721204  0.721333  0.721491  0.721683  0.721914  0.722189  0.722517  0.722904  0.723358  0.723887    0.7245  0.725206  0.726012   0.72693  0.727966   0.72913   0.73043  0.731873  0.733466  0.735216  0.737127  0.739203  0.741448  0.743864   0.74645       0.749208 | 2.591118  2.591119   2.59112  2.591122  2.591125  2.591128  2.591133   2.59114  2.591149  2.591162  2.591179  2.591202  2.591233  2.591274  2.591328    2.5914  2.591492  2.591613  2.591767  2.591965  2.592216  2.592532  2.592928  2.593422  2.594032  2.594781  2.595695  2.596803  2.598137  2.599733  2.601631  2.603871    2.6065  2.609565  2.613116  2.617205  2.621884  2.627206  2.633224  2.639988  2.647547  2.655947   2.66523  2.675435  2.686594  2.698734  2.711878   2.72604   2.74123        2.757451 | 100.000142  100.000208  100.000302  100.000436  100.000626  100.000893  100.001265  100.001781  100.002492  100.003465   100.00479  100.006579  100.008982  100.012188  100.016437  100.022033  100.029354  100.038871  100.051162  100.066929  100.087025   100.11247  100.144475  100.184467  100.234112  100.295337  100.370349  100.461655  100.572076  100.704759  100.863174  101.051122  101.272717  101.532373   101.83478   102.18487  102.587779  103.048802  103.573338  104.166842  104.834759   105.58247  106.415237  107.338139  108.356029  109.473484  110.694767  112.023792  113.464102  115.018847   116.69078 |  100.0 |   0.000142 |  116.69078 |
    | 03:00 | 3.717831  3.717832  3.717833  3.717836  3.717839  3.717844   3.71785  3.717859  3.717871  3.717887   3.71791  3.717941  3.717981  3.718036  3.718108  3.718203  3.718327  3.718488  3.718696  3.718962  3.719302  3.719731   3.72027  3.720944   3.72178   3.72281  3.724071  3.725605  3.727459  3.729685  3.732341  3.735489  3.739196  3.743535  3.748582  3.754417   3.76112  3.768775  3.777465  3.787274   3.79828  3.810562  3.824192  3.839237  3.855757  3.873804  3.893422  3.914648  3.937505             3.962011 | 124.878781   124.87884  124.878925  124.879045  124.879215  124.879454  124.879787  124.880248  124.880882  124.881749  124.882927  124.884517   124.88665  124.889492  124.893256  124.898208  124.904681  124.913087  124.923934  124.937838  124.955547  124.977954   125.00612  125.041295  125.084936  125.138729  125.204606  125.284759  125.381654  125.498041  125.636951  125.801699  125.995872   126.22331  126.488086  126.794475  127.146911  127.549948  128.008204  128.526314  129.108869   129.76036  130.485118  131.287264  132.170656  133.138844  134.195029  135.342039  136.582297  137.917814 |  52.914582   52.914593    52.91461   52.914633   52.914666   52.914713   52.914778   52.914868   52.914992   52.915162   52.915392   52.915703    52.91612   52.916675   52.917411   52.918378   52.919643   52.921286   52.923406   52.926123   52.929583   52.933961   52.939464   52.946335   52.954858   52.965363   52.978225   52.993869   53.012776   53.035476   53.062557   53.094658   53.132466   53.176719    53.22819   53.287688   53.356047   53.434114   53.522739   53.622765   53.735011   53.860263   53.999262    54.15269   54.321159   54.505204   54.705274   54.921728   55.154829        55.404741 |  52.178306   52.178317   52.178334   52.178357   52.178389   52.178435   52.178499   52.178587   52.178709   52.178875   52.179101   52.179405   52.179814   52.180359    52.18108   52.182029   52.183269    52.18488   52.186959   52.189623   52.193016   52.197309   52.202705   52.209443   52.217801   52.228101   52.240713   52.256054   52.274593   52.296853   52.323408   52.354885    52.39196   52.435352   52.485824   52.544167   52.611199    52.68775   52.774654   52.872737   52.982803   53.105623   53.241923   53.392371   53.557568   53.738039   53.934225   54.146475   54.375049     54.620109 | 1.088385  1.088386  1.088386  1.088386  1.088387  1.088387  1.088388   1.08839  1.088392  1.088394  1.088398  1.088402  1.088409  1.088417  1.088428  1.088443  1.088462  1.088487   1.08852  1.088561  1.088614   1.08868  1.088764  1.088868  1.088998  1.089158  1.089353  1.089591  1.089878  1.090223  1.090634  1.091122  1.091696  1.092367  1.093148   1.09405  1.095086  1.096268  1.097609  1.099121  1.100816  1.102705  1.104799  1.107108  1.109639    1.1124  1.115397  1.118633   1.12211  1.125831 | 1.359161  1.359161  1.359161   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916  1.359159  1.359159  1.359158  1.359158  1.359157  1.359155  1.359154  1.359152  1.359149  1.359145  1.359141  1.359135  1.359129   1.35912  1.359109  1.359096  1.359081  1.359062  1.359039  1.359012   1.35898  1.358942  1.358898  1.358847  1.358788  1.358721  1.358644  1.358557   1.35846  1.358352  1.358232  1.358099  1.357954  1.357797  1.357626  1.357442  1.357246  1.357037          1.356817 | 0.536158  0.536159  0.536159  0.536159  0.536159   0.53616   0.53616  0.536161  0.536162  0.536163  0.536165  0.536167   0.53617  0.536175   0.53618  0.536188  0.536198   0.53621  0.536227  0.536248  0.536275  0.536308  0.536351  0.536404   0.53647  0.536551   0.53665  0.536771  0.536917  0.537092    0.5373  0.537547  0.537837  0.538177  0.538571  0.539025  0.539546  0.540139   0.54081  0.541565  0.542409  0.543347  0.544382  0.545519   0.54676  0.548107  0.549561  0.551122   0.55279      0.554563 | -0.201868  -0.201868  -0.201868  -0.201869  -0.201869  -0.201869   -0.20187  -0.201871  -0.201872  -0.201874  -0.201876   -0.20188  -0.201884   -0.20189  -0.201897  -0.201907   -0.20192  -0.201936  -0.201958  -0.201985  -0.202019  -0.202063  -0.202118  -0.202185  -0.202269  -0.202373  -0.202498   -0.20265  -0.202834  -0.203053  -0.203313  -0.203619  -0.203979  -0.204397  -0.204881  -0.205437  -0.206072  -0.206792  -0.207604  -0.208513  -0.209526  -0.210647   -0.21188  -0.213229  -0.214697  -0.216284  -0.217992  -0.219819  -0.221765     -0.223828 | 0.665709  0.665709   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571  0.665711  0.665711  0.665712  0.665713  0.665714  0.665715  0.665717  0.665719  0.665722  0.665725  0.665729  0.665735  0.665742   0.66575   0.66576  0.665773  0.665788  0.665806  0.665828  0.665853  0.665883  0.665918  0.665958  0.666005  0.666059  0.666119  0.666189  0.666266  0.666353   0.66645  0.666557  0.666674  0.666803  0.666944  0.667096   0.66726  0.667435  0.667623      0.667823 | 100.000113 |  100.00014  100.000205  100.000297  100.000428  100.000613  100.000873  100.001235  100.001737  100.002427  100.003371  100.004653  100.006384  100.008705  100.011799  100.015895  100.021285   100.02833  100.037481  100.049287  100.064423  100.083701  100.108094  100.138759  100.177058   100.22458  100.283166  100.354922  100.442247  100.547838  100.674706  100.826179  101.005899  101.217822  101.466192   101.75553  102.090598  102.476365  102.917964  103.420645  103.989722  104.630517  105.348306  106.148264  107.035405  108.014538  109.090219  110.266709  111.547942    112.9375          114.438591 | 115.214209 |   0.7207    0.7207    0.7207    0.7207  0.720701  0.720701  0.720702  0.720703  0.720704  0.720706  0.720709  0.720712  0.720716  0.720722   0.72073   0.72074  0.720754  0.720771  0.720793  0.720822  0.720859  0.720905  0.720964  0.721037  0.721127  0.721238  0.721375   0.72154  0.721741  0.721981  0.722268  0.722608  0.723008  0.723476  0.724021  0.724649  0.725371  0.726195   0.72713  0.728184  0.729365  0.730681   0.73214  0.733749  0.735512  0.737435  0.739521  0.741774  0.744194       0.746783 | 2.591118  2.591118   2.59112  2.591121  2.591123  2.591126   2.59113  2.591135  2.591143  2.591153  2.591167  2.591187  2.591212  2.591246  2.591291  2.591351  2.591428  2.591529  2.591659  2.591825  2.592038  2.592306  2.592643  2.593065  2.593587  2.594231   2.59502  2.595979  2.597138  2.598529  2.600189  2.602156  2.604472  2.607184  2.610337  2.613981  2.618166  2.622945  2.628369  2.634489  2.641354  2.649011  2.657506  2.666877  2.677163  2.688393  2.700593  2.713784  2.727979        2.743188 | 100.000113  100.000166  100.000242   100.00035  100.000504   100.00072  100.001022  100.001443  100.002023   100.00282  100.003907   100.00538  100.007363  100.010015   100.01354  100.018196  100.024304  100.032267   100.04258   100.05585  100.072815  100.094361  100.121548  100.155628  100.198072  100.250586  100.315141  100.393984   100.48966  100.605019   100.74323   100.90778  101.102471  101.331406  101.598977  101.909833  102.268849  102.681091  103.151762   103.68616  104.289618  104.967451    105.7249  106.567073  107.498896  108.525063  109.649994  110.877794  112.212227  113.656693  115.214209 |  100.0 |   0.000113 | 115.214209 |
    | 03:30 |  3.71783  3.717831  3.717832  3.717834  3.717837  3.717841  3.717846  3.717853  3.717863  3.717876  3.717895   3.71792  3.717954  3.717999  3.718059  3.718138  3.718241  3.718376  3.718551  3.718775  3.719062  3.719425  3.719884  3.720458  3.721173  3.722057  3.723143  3.724468  3.726075  3.728012   3.73033  3.733088  3.736348  3.740178  3.744649  3.749836  3.755817  3.762674  3.770487  3.779338  3.789307  3.800473  3.812909  3.826687  3.841869  3.858514  3.876671  3.896382  3.917679             3.940587 | 124.878755  124.878802  124.878871  124.878968  124.879106  124.879299   124.87957  124.879945  124.880463  124.881174  124.882141   124.88345  124.885211  124.887564  124.890688  124.894809  124.900212  124.907248  124.916354  124.928061  124.943016  124.961997  124.985931  125.015916  125.053238  125.099393  125.156104  125.225338  125.309319  125.410541  125.531772  125.676058  125.846716  126.047322  126.281697  126.553877  126.868087  127.228695  127.640175  128.107054  128.633858  129.225061  129.885025   130.61795  131.427816  132.318342  133.292939  134.354674  135.506245  136.749956 |  52.914576   52.914586   52.914599   52.914618   52.914645   52.914683   52.914736   52.914809    52.91491   52.915049   52.915238   52.915494   52.915838   52.916298   52.916909   52.917714    52.91877   52.920145   52.921925   52.924212   52.927135   52.930843    52.93552   52.941377   52.948668   52.957682   52.968756   52.982272   52.998662   53.018411   53.042053   53.070179   53.103425    53.14248   53.188073   53.240971   53.301973   53.371898   53.451575   53.541837   53.643504    53.75737   53.884198   54.024699   54.179529   54.349273   54.534439   54.735452   54.952646        55.186264 |  52.178301    52.17831   52.178323   52.178342   52.178368   52.178405   52.178457   52.178529   52.178628   52.178765    52.17895   52.179201   52.179538   52.179989   52.180588   52.181378   52.182413   52.183762   52.185506    52.18775   52.190615   52.194252   52.198837   52.204581    52.21173   52.220569   52.231428   52.244682   52.260754   52.280119   52.303302   52.330882   52.363483   52.401779   52.446486   52.498358   52.558175   52.626741   52.704872   52.793381   52.893073   53.004728   53.129093   53.266866   53.418689   53.585136   53.766707   53.963816   54.176793     54.405874 | 1.088385  1.088385  1.088386  1.088386  1.088386  1.088387  1.088388  1.088389   1.08839  1.088393  1.088395  1.088399  1.088405  1.088412  1.088421  1.088433  1.088449   1.08847  1.088497  1.088532  1.088576  1.088633  1.088704  1.088793  1.088904  1.089041  1.089209  1.089415  1.089664  1.089964  1.090323   1.09075  1.091255  1.091848   1.09254  1.093342  1.094266  1.095326  1.096532  1.097897  1.099434  1.101153  1.103066  1.105182  1.107512  1.110061  1.112839  1.115848  1.119094  1.122579 | 1.359161  1.359161  1.359161  1.359161   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916  1.359159  1.359159  1.359158  1.359157  1.359156  1.359155  1.359153  1.359151  1.359148  1.359144  1.359139  1.359134  1.359126  1.359117  1.359106  1.359092  1.359076  1.359056  1.359032  1.359004  1.358971  1.358932  1.358887  1.358834  1.358774  1.358705  1.358627  1.358539   1.35844   1.35833  1.358209  1.358075  1.357929   1.35777  1.357599  1.357415  1.357218          1.357009 | 0.536158  0.536159  0.536159  0.536159  0.536159  0.536159   0.53616   0.53616  0.536161  0.536162  0.536164  0.536166  0.536168  0.536172  0.536177  0.536183  0.536191  0.536202  0.536215  0.536233  0.536256  0.536284   0.53632  0.536366  0.536422  0.536492  0.536577  0.536682  0.536808   0.53696  0.537142  0.537359  0.537614  0.537914  0.538264  0.538668  0.539134  0.539666  0.540271  0.540955  0.541722  0.542577  0.543526  0.544571  0.545717  0.546966   0.54832   0.54978  0.551344      0.553014 | -0.201868  -0.201868  -0.201868  -0.201868  -0.201869  -0.201869   -0.20187   -0.20187  -0.201871  -0.201873  -0.201875  -0.201877  -0.201881  -0.201886  -0.201892    -0.2019  -0.201911  -0.201925  -0.201943  -0.201966  -0.201995  -0.202032  -0.202078  -0.202136  -0.202208  -0.202297  -0.202405  -0.202537  -0.202696  -0.202887  -0.203115  -0.203385  -0.203702  -0.204072  -0.204502  -0.204999  -0.205567  -0.206215  -0.206949  -0.207774  -0.208697  -0.209722  -0.210854  -0.212098  -0.213457  -0.214932  -0.216526  -0.218238  -0.220068     -0.222015 | 0.665709  0.665709   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571  0.665711  0.665711  0.665712  0.665713  0.665714  0.665715  0.665717   0.66572  0.665723  0.665726  0.665731  0.665737  0.665744  0.665753  0.665764  0.665777  0.665792  0.665811  0.665834   0.66586  0.665891  0.665927  0.665968  0.666016  0.666071  0.666133  0.666203  0.666282   0.66637  0.666468  0.666576  0.666694  0.666824  0.666965  0.667118  0.667282  0.667458      0.667646 |  100.00009 | 100.000112  100.000164  100.000238  100.000344  100.000494  100.000704  100.000999  100.001408  100.001972  100.002745  100.003798  100.005223  100.007139    100.0097    100.0131  100.017586  100.023466  100.031125  100.041036   100.05378   100.07006  100.090722  100.116779  100.149425  100.190063  100.240325  100.302091  100.377508  100.469009  100.579322  100.711483  100.868833  101.055021  101.273994  101.529978  101.827456  102.171141  102.565935   103.01689  103.529159  104.107943  104.758442  105.485798  106.295042   107.19104  108.178448  109.261669  110.444813  111.731668          113.125672 |  113.84681 |   0.7207    0.7207    0.7207    0.7207  0.720701  0.720701  0.720702  0.720702  0.720703  0.720705  0.720707   0.72071  0.720713  0.720718  0.720725  0.720733  0.720744  0.720759  0.720778  0.720802  0.720833  0.720872  0.720922  0.720984  0.721061  0.721157  0.721274  0.721418  0.721591    0.7218  0.722051  0.722349  0.722701  0.723114  0.723596  0.724156    0.7248  0.725539  0.726379  0.727331  0.728402    0.7296  0.730933  0.732407   0.73403  0.735806   0.73774  0.739835  0.742095        0.74452 | 2.591117  2.591118  2.591119   2.59112  2.591122  2.591124  2.591127  2.591132  2.591138  2.591146  2.591158  2.591174  2.591195  2.591223   2.59126   2.59131  2.591375  2.591459  2.591568  2.591708  2.591887  2.592115  2.592402  2.592761  2.593208   2.59376  2.594439  2.595268  2.596272  2.597483  2.598932  2.600656  2.602693  2.605086  2.607879  2.611119  2.614855  2.619137  2.624014  2.629538  2.635757  2.642721  2.650474   2.65906  2.668516  2.678878  2.690176  2.702433  2.715667        2.729893 |  100.00009  100.000133  100.000194  100.000281  100.000405   100.00058  100.000825  100.001168  100.001641  100.002293  100.003185  100.004395   100.00603  100.008222  100.011143  100.015012  100.020102  100.026756  100.035398  100.046552  100.060854  100.079073  100.102133  100.131131   100.16736  100.212331  100.267794  100.335759  100.418511  100.518627  100.638988  100.782781  100.953504  101.154957  101.391232  101.666691  101.985941    102.3538  102.775257  103.255428  103.799506  104.412706  105.100213  105.867128  106.718414  107.658845  108.692963  109.825032  111.059009  112.398515   113.84681 |  100.0 |    0.00009 |  113.84681 |
    | 04:00 |  3.71783   3.71783  3.717831  3.717833  3.717835  3.717838  3.717842  3.717848  3.717856  3.717867  3.717883  3.717903  3.717931  3.717968  3.718018  3.718084   3.71817  3.718283  3.718429  3.718618  3.718859  3.719167  3.719556  3.720045  3.720655  3.721413  3.722346  3.723489   3.72488  3.726561  3.728582  3.730993  3.733854  3.737227  3.741179  3.745781  3.751107  3.757234  3.764242  3.772211  3.781219  3.791346  3.802666  3.815252  3.829172  3.844487  3.861251  3.879513   3.89931             3.920675 | 124.878734  124.878772  124.878827  124.878905  124.879017  124.879174  124.879393  124.879699  124.880122  124.880703  124.881497  124.882574  124.884026  124.885972  124.888562  124.891988  124.896492  124.902374  124.910008  124.919852  124.932464   124.94852  124.968828   124.99435  125.026218  125.065755  125.114494  125.174193  125.246852  125.334725   125.44033  125.566454  125.716151  125.892737  126.099777  126.341067  126.620605  126.942563  127.311243  127.731039  128.206382  128.741695  129.341333  130.009536  130.750371  131.567685   132.46506  133.445774  134.512765  135.668607 |  52.914572    52.91458   52.914591   52.914606   52.914628   52.914658   52.914701   52.914761   52.914844   52.914957   52.915112   52.915323   52.915607   52.915987   52.916493   52.917163   52.918043   52.919193   52.920684   52.922608   52.925073    52.92821   52.932178   52.937164    52.94339   52.951112   52.960631   52.972287   52.986471   53.003619   53.024221   53.048815   53.077991   53.112387   53.152688   53.199616   53.253932   53.316422   53.387891   53.469154   53.561021   53.664291   53.779736   53.908092   54.050045   54.206224   54.377188    54.56342    54.76532          54.9832 |  52.178297   52.178304   52.178315    52.17833   52.178351   52.178381   52.178423   52.178482   52.178563   52.178674   52.178827   52.179033   52.179311   52.179684    52.18018   52.180837     52.1817   52.182827    52.18429   52.186177   52.188593    52.19167   52.195561    52.20045   52.206555   52.214127   52.223461   52.234891   52.248799   52.265615   52.285816   52.309933   52.338542   52.372271   52.411788   52.457805   52.511066   52.572343   52.642425   52.722109   52.812192   52.913457    53.02666   53.152523    53.29172   53.444866    53.61251   53.795125   53.993104     54.206754 | 1.088385  1.088385  1.088386  1.088386  1.088386  1.088387  1.088387  1.088388  1.088389  1.088391  1.088393  1.088397  1.088401  1.088407  1.088414  1.088425  1.088438  1.088456  1.088478  1.088507  1.088545  1.088593  1.088653  1.088729  1.088823  1.088941  1.089086  1.089263  1.089478  1.089739  1.090052  1.090426  1.090869  1.091391  1.092003  1.092715  1.093538  1.094485  1.095568  1.096798  1.098187  1.099748  1.101491  1.103426  1.105564  1.107913   1.11048  1.113273  1.116295   1.11955 | 1.359161  1.359161  1.359161  1.359161   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916  1.359159  1.359159  1.359158  1.359157  1.359156  1.359154  1.359152   1.35915  1.359147  1.359143  1.359138  1.359132  1.359124  1.359114  1.359102  1.359088  1.359071   1.35905  1.359026  1.358996  1.358962  1.358922  1.358875  1.358821   1.35876  1.358689   1.35861   1.35852   1.35842  1.358309  1.358186  1.358051  1.357904  1.357744  1.357572  1.357388          1.357191 | 0.536158  0.536158  0.536159  0.536159  0.536159  0.536159  0.536159   0.53616  0.536161  0.536161  0.536163  0.536164  0.536166  0.536169  0.536173  0.536178  0.536185  0.536194  0.536206  0.536221   0.53624  0.536264  0.536295  0.536333  0.536381  0.536441  0.536514  0.536604  0.536714  0.536846  0.537005  0.537194  0.537419  0.537683  0.537992  0.538352  0.538767  0.539244  0.539788  0.540405    0.5411  0.541878  0.542745  0.543704  0.544759  0.545914  0.547171  0.548531  0.549995      0.551564 | -0.201868  -0.201868  -0.201868  -0.201868  -0.201869  -0.201869  -0.201869   -0.20187  -0.201871  -0.201872  -0.201874  -0.201876  -0.201879  -0.201882  -0.201888  -0.201894  -0.201903  -0.201915   -0.20193  -0.201949  -0.201974  -0.202005  -0.202045  -0.202094  -0.202156  -0.202232  -0.202325  -0.202439  -0.202578  -0.202744  -0.202943  -0.203179  -0.203458  -0.203785  -0.204167  -0.204609  -0.205117  -0.205699   -0.20636  -0.207107  -0.207945   -0.20888  -0.209917  -0.211061  -0.212315  -0.213682  -0.215165  -0.216765  -0.218481     -0.220313 | 0.665709  0.665709  0.665709   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571  0.665711  0.665711  0.665712  0.665712  0.665713  0.665714  0.665716  0.665718   0.66572  0.665724  0.665728  0.665733  0.665739  0.665746  0.665756  0.665767  0.665781  0.665797  0.665817   0.66584  0.665867  0.665899  0.665936  0.665978  0.666027  0.666083  0.666146  0.666218  0.666298  0.666387  0.666486  0.666595  0.666714  0.666845  0.666986  0.667139  0.667304       0.66748 | 100.000072 | 100.000089  100.000131  100.000191  100.000276  100.000397  100.000568  100.000807   100.00114    100.0016  100.002233  100.003097  100.004269  100.005849  100.007967  100.010786  100.014515  100.019417   100.02582  100.034129  100.044844  100.058573  100.076051  100.098159  100.125944  100.160642  100.203694  100.256771  100.321794  100.400946  100.496694  100.611793  100.749299  100.912566  101.105245  101.331271  101.594847  101.900419  102.252646  102.656363  103.116537  103.638226   104.22652  104.886496  105.623165  106.441417  107.345974  108.341345  109.431784  110.621255          111.913403 | 112.582651 |   0.7207    0.7207    0.7207    0.7207    0.7207  0.720701  0.720701  0.720702  0.720703  0.720704  0.720706  0.720708  0.720711  0.720715   0.72072  0.720727  0.720737  0.720749  0.720765  0.720785  0.720811  0.720844  0.720887  0.720939  0.721005  0.721087  0.721188  0.721312  0.721462  0.721644  0.721862  0.722123  0.722432  0.722796  0.723222  0.723718  0.724293  0.724953  0.725708  0.726565  0.727533  0.728621  0.729835  0.731184  0.732673  0.734309  0.736098  0.738042  0.740146       0.742412 | 2.591117  2.591118  2.591118  2.591119  2.591121  2.591122  2.591125  2.591129  2.591134  2.591141   2.59115  2.591163  2.591181  2.591204  2.591235  2.591276   2.59133  2.591401  2.591492   2.59161  2.591761  2.591953  2.592197  2.592502  2.592884  2.593358  2.593941  2.594656  2.595525  2.596576  2.597839  2.599346  2.601134  2.603242  2.605711  2.608586  2.611913   2.61574  2.620116   2.62509  2.630712  2.637029  2.644089  2.651935  2.660608  2.670146  2.680582  2.691943  2.704253        2.717528 | 100.000072  100.000106  100.000155  100.000225  100.000325  100.000467  100.000666  100.000944   100.00133  100.001863  100.002593  100.003588  100.004933  100.006743  100.009161  100.012372  100.016608  100.022162  100.029396  100.038759  100.050799  100.066183  100.085715  100.110351  100.141228  100.179678  100.227253  100.285742  100.357193  100.443927  100.548553  100.673974  100.823396  101.000324  101.208553  101.452159  101.735473  102.063056  102.439668  102.870221  103.359741  103.913315  104.536043  105.232983  106.009098  106.869209  107.817942   108.85969  109.998572  111.238398  112.582651 |  100.0 |   0.000072 | 112.582651 |
    | 04:30 | 3.717829   3.71783  3.717831  3.717832  3.717834  3.717836   3.71784  3.717844  3.717851   3.71786  3.717873   3.71789  3.717912  3.717943  3.717984  3.718039  3.718111  3.718205  3.718327  3.718486  3.718689  3.718949  3.719279  3.719694  3.720215  3.720863  3.721664  3.722648  3.723849  3.725307  3.727064  3.729168  3.731674  3.734639  3.738124  3.742198   3.74693  3.752394  3.758666  3.765824  3.773945  3.783108  3.793388  3.804859   3.81759  3.831648  3.847089  3.863968  3.882329             3.902209 | 124.878717  124.878748  124.878792  124.878855  124.878945  124.879072  124.879251  124.879499  124.879844   124.88032   124.88097  124.881855  124.883051  124.884658  124.886803  124.889648  124.893398   124.89831  124.904702  124.912968   124.92359  124.937153  124.954359   124.97605  125.003219  125.037034  125.078854  125.130244  125.192998  125.269147  125.360972  125.471013  125.602073   125.75721  125.939738  126.153207  126.401384  126.688229  127.017858  127.394507  127.822486  128.306135  128.849771  129.457635  130.133843  130.882336   131.70683  132.610777  133.597325  134.669287 |  52.914569   52.914575   52.914584   52.914596   52.914614   52.914638   52.914673   52.914722   52.914789   52.914882   52.915009   52.915182   52.915416    52.91573   52.916149   52.916705   52.917438   52.918398   52.919647   52.921263   52.923339   52.925989   52.929351   52.933589   52.938897   52.945503   52.953671   52.963706   52.975959   52.990822   53.008741   53.030205   53.055759   53.085991   53.121539   53.163083   53.211341   53.267063   53.331026   53.404019    53.48684    53.58028   53.685117   53.802099   53.931937   54.075293   54.232769   54.404899   54.592142        54.794875 |  52.178294     52.1783   52.178308    52.17832   52.178337   52.178362   52.178396   52.178444    52.17851   52.178601   52.178726   52.178895   52.179124   52.179432   52.179843   52.180389   52.181107   52.182049   52.183274   52.184858   52.186893   52.189492   52.192789   52.196944   52.202149   52.208627   52.216636   52.226477   52.238491   52.253066   52.270636   52.291685   52.316742   52.346387   52.381244   52.421981   52.469302   52.523943   52.586663   52.658239   52.739451   52.831077   52.933878   53.048588   53.175905   53.316477   53.470895   53.639683   53.823289     54.022086 | 1.088385  1.088385  1.088385  1.088386  1.088386  1.088386  1.088387  1.088388  1.088389   1.08839  1.088392  1.088395  1.088398  1.088403  1.088409  1.088418  1.088429  1.088443  1.088462  1.088487  1.088519  1.088559   1.08861  1.088674  1.088755  1.088856   1.08898  1.089132  1.089319  1.089545  1.089817  1.090143  1.090531   1.09099   1.09153   1.09216  1.092892  1.093737  1.094707  1.095812  1.097065  1.098478  1.100062  1.101828  1.103785  1.105944  1.108312  1.110896  1.113703  1.116737 | 1.359161  1.359161  1.359161  1.359161  1.359161   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916  1.359159  1.359159  1.359158  1.359158  1.359157  1.359155  1.359154  1.359152  1.359149  1.359146  1.359141  1.359136  1.359129  1.359121  1.359111  1.359099  1.359084  1.359066  1.359044  1.359019  1.358988  1.358953  1.358912  1.358864  1.358808  1.358745  1.358674  1.358593  1.358502    1.3584  1.358287  1.358163  1.358027  1.357879  1.357719  1.357546          1.357361 | 0.536158  0.536158  0.536159  0.536159  0.536159  0.536159  0.536159   0.53616   0.53616  0.536161  0.536162  0.536163  0.536165  0.536167  0.536171  0.536175  0.536181  0.536188  0.536198   0.53621  0.536226  0.536247  0.536273  0.536306  0.536347  0.536398  0.536461  0.536538  0.536633  0.536748  0.536886  0.537051  0.537248   0.53748  0.537753  0.538072  0.538442  0.538868  0.539355  0.539911  0.540539  0.541245  0.542035  0.542912  0.543882  0.544947   0.54611  0.547374   0.54874      0.550209 | -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201869  -0.201869  -0.201869   -0.20187  -0.201871  -0.201872  -0.201874  -0.201877   -0.20188  -0.201884   -0.20189  -0.201897  -0.201907   -0.20192  -0.201936  -0.201957  -0.201983  -0.202017  -0.202059  -0.202111  -0.202177  -0.202257  -0.202355  -0.202475  -0.202619  -0.202793  -0.202999  -0.203245  -0.203533  -0.203871  -0.204263  -0.204717  -0.205238  -0.205832  -0.206505  -0.207265  -0.208115  -0.209063  -0.210112  -0.211266   -0.21253  -0.213906  -0.215396  -0.217001     -0.218721 | 0.665709  0.665709  0.665709  0.665709   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571  0.665711  0.665711  0.665712  0.665713  0.665714  0.665715  0.665716  0.665719  0.665721  0.665725  0.665729  0.665734  0.665741  0.665749  0.665759   0.66577  0.665785  0.665802  0.665822  0.665846  0.665874  0.665907  0.665945  0.665988  0.666038  0.666095   0.66616  0.666232  0.666313  0.666404  0.666503  0.666613  0.666734  0.666865  0.667007   0.66716      0.667325 | 100.000057 | 100.000071  100.000104  100.000152  100.000221  100.000319  100.000457  100.000651  100.000922  100.001298  100.001815  100.002523  100.003486  100.004788  100.006537  100.008872  100.011968   100.01605  100.021396  100.028353  100.037351  100.048913  100.063677  100.082408  100.106021  100.135601  100.172419  100.217957  100.273925  100.342278  100.425237  100.525298  100.645241   100.78814  100.957359  101.156545  101.389621  101.660762  101.974376  102.335064  102.747594   103.21685  103.747789  104.345394  105.014621  105.760352  106.587339  107.500164  108.503192  109.600534          110.796014 | 111.415947 |   0.7207    0.7207    0.7207    0.7207    0.7207  0.720701  0.720701  0.720701  0.720702  0.720703  0.720704  0.720706  0.720709  0.720712  0.720717  0.720722   0.72073   0.72074  0.720754  0.720771  0.720793  0.720821  0.720857  0.720901  0.720958  0.721028  0.721114  0.721221  0.721351  0.721508  0.721698  0.721925  0.722196  0.722516  0.722892  0.723332  0.723842  0.724431  0.725107  0.725878  0.726751  0.727736   0.72884   0.73007  0.731434  0.732938  0.734587  0.736387  0.738342       0.740454 | 2.591117  2.591117  2.591118  2.591119   2.59112  2.591121  2.591123  2.591126  2.591131  2.591136  2.591144  2.591155  2.591169  2.591188  2.591214  2.591248  2.591293  2.591352  2.591428  2.591527  2.591655  2.591817  2.592023  2.592283  2.592609  2.593014  2.593514   2.59413  2.594881  2.595792   2.59689  2.598206  2.599772  2.601625  2.603803  2.606348  2.609304  2.612717  2.616634  2.621103  2.626172   2.63189  2.638303  2.645456  2.653392   2.66215  2.671767  2.682273  2.693695        2.706054 | 100.000057  100.000084  100.000124   100.00018  100.000261  100.000376  100.000537  100.000763  100.001077  100.001512   100.00211  100.002926  100.004033  100.005525  100.007525  100.010187  100.013708  100.018338  100.024385  100.032234  100.042357   100.05533  100.071849   100.09275  100.119027   100.15185  100.192593  100.242845  100.304435  100.379445  100.470231  100.579429  100.709965  100.865058  101.048216  101.263229  101.514151  101.805281  102.141134  102.526403  102.965928  103.464642  104.027529  104.659572  105.365705  106.150757  107.019408  107.976144  109.025212  110.170589  111.415947 |  100.0 |   0.000057 | 111.415947 |
    | 05:00 | 3.717829   3.71783   3.71783  3.717831  3.717833  3.717835  3.717837  3.717841  3.717847  3.717854  3.717864  3.717878  3.717897  3.717922  3.717956  3.718002  3.718061   3.71814  3.718242  3.718375  3.718546  3.718766  3.719045  3.719397   3.71984  3.720394   3.72108  3.721926  3.722962  3.724223  3.725748  3.727582  3.729772  3.732373  3.735441   3.73904  3.743235  3.748097  3.753698  3.760113  3.767418  3.775689  3.785003  3.795433   3.80705  3.819923  3.834113  3.849677  3.866665              3.88512 | 124.878704  124.878729  124.878764  124.878815  124.878887   124.87899  124.879135  124.879337  124.879618  124.880006  124.880539  124.881265   124.88225  124.883576   124.88535   124.88771  124.890828  124.894925   124.90027  124.907203  124.916137  124.927578  124.942136  124.960544  124.983673  125.012549  125.048373  125.092539  125.146647  125.212521  125.292221  125.388055  125.502582  125.638614  125.799216  125.987694   126.20758  126.462613  126.756709  127.093928  127.478438  127.914467  128.406262  128.958035  129.573915  130.257899  131.013802  131.845215  132.755463  133.747568 |  52.914567   52.914571   52.914578   52.914588   52.914602   52.914622   52.914651    52.91469   52.914745   52.914821   52.914925   52.915067    52.91526   52.915519   52.915865   52.916327   52.916936   52.917737   52.918781   52.920136   52.921882   52.924118   52.926963   52.930559   52.935078    52.94072   52.947717   52.956343   52.966909    52.97977   52.995326   53.014024   53.036362   53.062881   53.094174   53.130874   53.173659    53.22324   53.280358   53.345776   53.420271   53.504623   53.599604    53.70597   53.824448   53.955723   54.100433   54.259155     54.4324        54.620602 |  52.178291   52.178296   52.178303   52.178313   52.178326   52.178346   52.178374   52.178413   52.178466   52.178541   52.178643   52.178782   52.178971   52.179225   52.179565   52.180017   52.180615     52.1814   52.182424   52.183753   52.185465   52.187657   52.190447   52.193974   52.198405   52.203936   52.210798   52.219257   52.229617   52.242228   52.257482   52.275818   52.297721   52.323726   52.354411   52.390399   52.432353   52.480971   52.536979   52.601127   52.674175   52.756889   52.850026   52.954327   53.070503   53.199229   53.341129   53.496769   53.666649     53.851197 | 1.088385  1.088385  1.088385  1.088386  1.088386  1.088386  1.088386  1.088387  1.088388  1.088389  1.088391  1.088393  1.088396    1.0884  1.088405  1.088412  1.088421  1.088433  1.088449   1.08847  1.088496   1.08853  1.088574  1.088628  1.088697  1.088783  1.088889   1.08902  1.089181  1.089377  1.089613  1.089897  1.090236  1.090639  1.091114  1.091672  1.092321  1.093073  1.093939   1.09493  1.096058  1.097335  1.098771  1.100377  1.102165  1.104144  1.106322  1.108708  1.111309  1.114129 | 1.359161  1.359161  1.359161  1.359161  1.359161  1.359161   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916  1.359159  1.359159  1.359158  1.359157  1.359156  1.359155  1.359153  1.359151  1.359148  1.359144   1.35914  1.359134  1.359127  1.359118  1.359108  1.359095  1.359079   1.35906  1.359038  1.359012   1.35898  1.358944  1.358901  1.358852  1.358795  1.358731  1.358658  1.358575  1.358483   1.35838  1.358266  1.358141  1.358004  1.357854  1.357693           1.35752 | 0.536158  0.536158  0.536158  0.536159  0.536159  0.536159  0.536159  0.536159   0.53616   0.53616  0.536161  0.536162  0.536164  0.536166  0.536168  0.536172  0.536177  0.536183  0.536191  0.536201  0.536215  0.536232  0.536254  0.536282  0.536317  0.536361  0.536415  0.536481  0.536563  0.536662  0.536782  0.536926  0.537099  0.537303  0.537543  0.537825  0.538153  0.538533  0.538969  0.539468  0.540034  0.540673  0.541391  0.542191  0.543079  0.544059  0.545133  0.546304  0.547575      0.548947 | -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201869  -0.201869   -0.20187  -0.201871  -0.201872  -0.201873  -0.201875  -0.201878  -0.201881  -0.201886  -0.201892    -0.2019  -0.201911  -0.201924  -0.201942  -0.201964  -0.201993  -0.202029  -0.202073  -0.202129  -0.202198  -0.202283  -0.202386  -0.202511  -0.202662  -0.202843  -0.203058  -0.203312   -0.20361  -0.203958  -0.204361  -0.204826  -0.205359  -0.205966  -0.206652  -0.207423  -0.208286  -0.209246  -0.210306  -0.211471  -0.212744  -0.214128  -0.215625     -0.217235 | 0.665709  0.665709  0.665709  0.665709   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571  0.665711  0.665711  0.665711  0.665712  0.665713  0.665714  0.665715  0.665717  0.665719  0.665722  0.665726   0.66573  0.665736  0.665743  0.665751  0.665762  0.665774  0.665789  0.665807  0.665828  0.665852  0.665881  0.665915  0.665954  0.665999   0.66605  0.666108  0.666173  0.666247  0.666329   0.66642  0.666521  0.666632  0.666753  0.666885  0.667028      0.667182 | 100.000046 | 100.000057  100.000083  100.000122  100.000177  100.000256  100.000368  100.000526  100.000746  100.001052  100.001474  100.002054  100.002844  100.003916  100.005359   100.00729  100.009858  100.013253  100.017711   100.02353  100.031076  100.040801  100.053254  100.069102  100.089141   100.11432  100.145759  100.184766  100.232861   100.29179  100.363545  100.450379  100.554815  100.679656  100.827992  101.003189  101.208894   101.44901  101.727685  102.049282  102.418348  102.839578  103.317772   103.85779  104.464508  105.142762  105.897305  106.732758  107.653565  108.663953          109.767894 | 110.341062 |   0.7207    0.7207    0.7207    0.7207    0.7207    0.7207  0.720701  0.720701  0.720702  0.720702  0.720704  0.720705  0.720707   0.72071  0.720714  0.720718  0.720725  0.720733  0.720744  0.720759  0.720777  0.720801  0.720831  0.720869  0.720917  0.720977  0.721051  0.721143  0.721255  0.721391  0.721556  0.721754  0.721991  0.722272  0.722603  0.722991  0.723444  0.723968  0.724572  0.725263  0.726049  0.726939   0.72794  0.729059  0.730305  0.731684  0.733201  0.734863  0.736674       0.738639 | 2.591117  2.591117  2.591118  2.591118  2.591119   2.59112  2.591122  2.591124  2.591128  2.591132  2.591139  2.591148  2.591159  2.591175  2.591197  2.591225  2.591262  2.591311  2.591375  2.591458  2.591565  2.591702  2.591877  2.592097  2.592374   2.59272  2.593149  2.593678  2.594326  2.595114  2.596068  2.597214  2.598583  2.600208  2.602126  2.604375  2.606996  2.610033  2.613532  2.617537  2.622098  2.627261  2.633072  2.639578  2.646822  2.654845  2.663686  2.673378  2.683951         2.69543 | 100.000046  100.000067  100.000099  100.000144  100.000209  100.000302  100.000433  100.000616  100.000872  100.001227  100.001715  100.002384  100.003293  100.004523  100.006175  100.008379  100.011303  100.015158  100.020207  100.026779  100.035279  100.046204  100.060156  100.077863  100.100191  100.128169  100.163007  100.206112  100.259113  100.323875  100.402515  100.497418  100.611248  100.746948  100.907748  101.097158  101.318956  101.577174  101.876076  102.220126  102.613955  103.062324  103.570075   104.14209  104.783237  105.498325  106.292054  107.168968  108.133413  109.189498  110.341062 |  100.0 |   0.000046 | 110.341062 |
    | 05:30 | 3.717829  3.717829   3.71783  3.717831  3.717832  3.717833  3.717836  3.717839  3.717843  3.717849  3.717858  3.717869  3.717884  3.717905  3.717934  3.717971  3.718021  3.718086  3.718172  3.718283  3.718427  3.718611  3.718847  3.719146  3.719523  3.719994  3.720581  3.721307  3.722199  3.723288   3.72461  3.726205  3.728116  3.730392  3.733088  3.736261  3.739973   3.74429   3.74928  3.755017  3.761573  3.769022  3.777441  3.786903  3.797479   3.80924  3.822249  3.836567  3.852248             3.869341 | 124.878693  124.878713  124.878741  124.878782  124.878841  124.878924  124.879041  124.879205  124.879434  124.879751  124.880187  124.880782  124.881592  124.882684  124.884151  124.886106  124.888697  124.892109  124.896574  124.902381  124.909885  124.919524  124.931824  124.947424  124.967084  124.991705  125.022346  125.060242  125.106818  125.163706  125.232762  125.316073  125.415968  125.535023  125.676062  125.842148  126.036579  126.262867  126.524719  126.826005   127.17073   127.56299  128.006933  128.506712  129.066437  129.690125  130.381658  131.144729  131.982805   132.89909 |  52.914564   52.914568   52.914574   52.914582   52.914593    52.91461   52.914632   52.914665   52.914709   52.914771   52.914856   52.914973   52.915131   52.915344   52.915631   52.916013    52.91652   52.917186   52.918059   52.919194    52.92066   52.922544   52.924948   52.927996   52.931837   52.936648   52.942633   52.950036   52.959132    52.97024   52.983721    52.99998   53.019469   53.042687   53.070179   53.102536   53.140389   53.184411   53.235307   53.293807   53.360664   53.436638   53.522494   53.618984   53.726842   53.846774   53.979442   54.125458   54.285378        54.459686 |  52.178289   52.178293   52.178298   52.178306   52.178318   52.178333   52.178356   52.178387   52.178431   52.178492   52.178575    52.17869   52.178845   52.179054   52.179335    52.17971   52.180206    52.18086   52.181716   52.182829   52.184267   52.186114   52.188471    52.19146   52.195227   52.199943   52.205813   52.213072   52.221991   52.232884   52.246103   52.262046   52.281157   52.303924   52.330882    52.36261   52.399729   52.442896   52.492803   52.550167   52.615726   52.690225   52.774413   52.869029   52.974793   53.092396   53.222487   53.365669   53.522482     53.693406 | 1.088385  1.088385  1.088385  1.088385  1.088386  1.088386  1.088386  1.088387  1.088387  1.088388   1.08839  1.088391  1.088394  1.088397  1.088401  1.088407  1.088415  1.088425  1.088438  1.088456  1.088478  1.088507  1.088543  1.088589  1.088648  1.088721  1.088812  1.088925  1.089063  1.089232  1.089437  1.089684   1.08998  1.090333   1.09075  1.091241  1.091816  1.092484  1.093256  1.094143  1.095156  1.096306  1.097605  1.099063  1.100692  1.102502  1.104501  1.106699  1.109102  1.111718 | 1.359161  1.359161  1.359161  1.359161  1.359161  1.359161   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916  1.359159  1.359159  1.359159  1.359158  1.359157  1.359156  1.359154  1.359153   1.35915  1.359147  1.359143  1.359138  1.359132  1.359125  1.359116  1.359104  1.359091  1.359075  1.359055  1.359032  1.359004  1.358972  1.358934   1.35889   1.35884  1.358782  1.358716  1.358642  1.358558  1.358464   1.35836  1.358245  1.358118   1.35798   1.35783          1.357668 | 0.536158  0.536158  0.536158  0.536158  0.536159  0.536159  0.536159  0.536159  0.536159   0.53616  0.536161  0.536162  0.536163  0.536164  0.536167   0.53617  0.536173  0.536179  0.536185  0.536194  0.536206   0.53622  0.536239  0.536262  0.536292  0.536329  0.536375  0.536433  0.536503  0.536589  0.536693  0.536818  0.536968  0.537147  0.537359  0.537608  0.537898  0.538236  0.538625  0.539072  0.539581  0.540158  0.540809  0.541537  0.542348  0.543246  0.544235  0.545318  0.546497      0.547775 | -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201869  -0.201869  -0.201869   -0.20187  -0.201871  -0.201872  -0.201874  -0.201876  -0.201879  -0.201883  -0.201888  -0.201895  -0.201904  -0.201915   -0.20193  -0.201949  -0.201973  -0.202003  -0.202041  -0.202089  -0.202148  -0.202221   -0.20231  -0.202418  -0.202549  -0.202707  -0.202895  -0.203118   -0.20338  -0.203688  -0.204046  -0.204461  -0.204937  -0.205482    -0.2061  -0.206799  -0.207583  -0.208457  -0.209428  -0.210499  -0.211674  -0.212956  -0.214348     -0.215851 | 0.665709  0.665709  0.665709  0.665709  0.665709   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571  0.665711  0.665711  0.665712  0.665712  0.665713  0.665714  0.665716  0.665718   0.66572  0.665723  0.665727  0.665732  0.665738  0.665745  0.665754  0.665765  0.665778  0.665793  0.665812  0.665833  0.665859  0.665889  0.665923  0.665963  0.666009  0.666061   0.66612  0.666187  0.666262  0.666345  0.666437  0.666539  0.666651  0.666772  0.666905      0.667048 | 100.000036 | 100.000045  100.000066  100.000097  100.000142  100.000206  100.000296  100.000424  100.000602  100.000851  100.001196   100.00167  100.002319    100.0032  100.004389  100.005985  100.008113  100.010933  100.014646  100.019506  100.025827  100.033996  100.044487  100.057876  100.074858   100.09626  100.123065  100.156426   100.19769  100.248411  100.310371  100.385596  100.476369  100.585238  100.715027  100.868836  101.050036  101.262264  101.509406  101.795577  102.125095  102.502449  102.932262   103.41925  103.968177  104.583808  105.270865  106.033976  106.877631  107.806141          108.823597 | 109.352519 |   0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207  0.720701  0.720701  0.720702  0.720703  0.720704  0.720706  0.720708  0.720711  0.720715   0.72072  0.720728  0.720737  0.720749  0.720764  0.720784   0.72081  0.720842  0.720883  0.720934  0.720997  0.721076  0.721172   0.72129  0.721433  0.721605  0.721812  0.722058  0.722349  0.722691  0.723092  0.723558  0.724096  0.724714   0.72542  0.726222  0.727127  0.728144  0.729279   0.73054  0.731933  0.733463  0.735138       0.736959 | 2.591117  2.591117  2.591117  2.591118  2.591119  2.591119  2.591121  2.591123  2.591126  2.591129  2.591135  2.591142  2.591151  2.591165  2.591182  2.591206  2.591237  2.591277  2.591331  2.591401  2.591491  2.591606  2.591753   2.59194  2.592176  2.592471  2.592838  2.593292  2.593849   2.59453  2.595357  2.596353  2.597548  2.598971  2.600656  2.602639  2.604958  2.607655  2.610772  2.614355  2.618449    2.6231  2.628354  2.634258  2.640854  2.648187  2.656295  2.665214  2.674979        2.685616 | 100.000036  100.000054  100.000079  100.000115  100.000168  100.000242  100.000348  100.000497  100.000705  100.000994  100.001393  100.001941  100.002687  100.003699  100.005062  100.006886  100.009311  100.012517  100.016727  100.022223  100.029351   100.03854  100.050308  100.065287  100.084233  100.108046  100.137789  100.174706  100.220243  100.276063  100.344066  100.426402  100.525485     100.644  100.784909  100.951447  101.147123  101.375702  101.641191  101.947816  102.299988  102.702274  103.159357  103.675987  104.256943  104.906982   105.63079   106.43294  107.317844  108.289712  109.352519 |  100.0 |   0.000036 | 109.352519 |
    | 06:00 | 3.717829  3.717829   3.71783   3.71783  3.717831  3.717832  3.717834  3.717837   3.71784  3.717845  3.717852  3.717861  3.717874  3.717891  3.717915  3.717946  3.717987  3.718041  3.718113  3.718206  3.718326  3.718482  3.718681  3.718934  3.719253  3.719655  3.720156  3.720778  3.721545  3.722484  3.723628  3.725011  3.726676  3.728665  3.731029  3.733821  3.737098  3.740923  3.745361  3.750479   3.75635  3.763045  3.770638  3.779201  3.788808  3.799527  3.811426  3.824567  3.839009             3.854803 | 124.878685  124.878701  124.878723  124.878756  124.878804  124.878871  124.878965  124.879099  124.879285  124.879543  124.879899  124.880387  124.881052  124.881951  124.883162   124.88478   124.88693  124.889769  124.893494  124.898352  124.904648  124.912757  124.923136  124.936337  124.953025  124.973987  125.000155   125.03262  125.072647  125.121694  125.181425  125.253722  125.340698  125.444703  125.568327    125.7144  125.885985  126.086368  126.319037  126.587665  126.896077  127.248219  127.648116  128.099836  128.607437  129.174929  129.806221  130.505079  131.275079  132.119569 |  52.914563   52.914566    52.91457   52.914577   52.914586   52.914599   52.914618   52.914644    52.91468   52.914731     52.9148   52.914895   52.915025   52.915201   52.915438   52.915754   52.916174   52.916729   52.917457   52.918406   52.919637   52.921222    52.92325    52.92583    52.92909   52.933186   52.938298    52.94464   52.952458   52.962037   52.973699   52.987812   53.004785   53.025074    53.04918    53.07765   53.111073   53.150078   53.195333   53.247534   53.307404   53.375681   53.453112   53.540443   53.638409   53.747724   53.869069   54.003085   54.150362         54.31143 |  52.178288   52.178291   52.178295   52.178301    52.17831   52.178323   52.178341   52.178367   52.178403   52.178452    52.17852   52.178614   52.178741   52.178914   52.179146   52.179456   52.179868   52.180412   52.181126   52.182057   52.183263   52.184817   52.186806   52.189336   52.192533   52.196549   52.201562   52.207781   52.215447    52.22484   52.236276   52.250114   52.266758   52.286653   52.310291   52.338208   52.370982    52.40923   52.453606   52.504793     52.5635   52.630451   52.706378   52.792014   52.888077   52.995269   53.114258   53.245672   53.390089     53.548029 | 1.088385  1.088385  1.088385  1.088385  1.088385  1.088386  1.088386  1.088386  1.088387  1.088388  1.088389   1.08839  1.088392  1.088395  1.088398  1.088403   1.08841  1.088418  1.088429  1.088444  1.088462  1.088486  1.088517  1.088556  1.088606  1.088668  1.088746  1.088843  1.088961  1.089107  1.089284  1.089499  1.089757  1.090065  1.090431  1.090864  1.091371  1.091963   1.09265  1.093441  1.094349  1.095383  1.096555  1.097876  1.099357  1.101007  1.102838  1.104857  1.107073  1.109493 | 1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916  1.359159  1.359159  1.359158  1.359158  1.359157  1.359155  1.359154  1.359152  1.359149  1.359146  1.359142  1.359137   1.35913  1.359122  1.359113  1.359101  1.359087   1.35907  1.359049  1.359025  1.358997  1.358963  1.358925   1.35888  1.358828  1.358769  1.358701  1.358625   1.35854  1.358445   1.35834  1.358223  1.358096  1.357957          1.357806 | 0.536158  0.536158  0.536158  0.536158  0.536159  0.536159  0.536159  0.536159  0.536159   0.53616   0.53616  0.536161  0.536162  0.536163  0.536165  0.536168  0.536171  0.536175  0.536181  0.536188  0.536198   0.53621  0.536226  0.536246  0.536271  0.536302  0.536342  0.536391  0.536451  0.536525  0.536615  0.536724  0.536855  0.537012  0.537197  0.537416  0.537673  0.537972  0.538319  0.538718  0.539175  0.539695  0.540283  0.540944  0.541683  0.542505  0.543413  0.544411  0.545502      0.546689 | -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201869  -0.201869   -0.20187   -0.20187  -0.201871  -0.201873  -0.201874  -0.201877   -0.20188  -0.201884   -0.20189  -0.201897  -0.201907  -0.201919  -0.201935  -0.201956  -0.201981  -0.202014  -0.202054  -0.202105  -0.202168  -0.202244  -0.202338  -0.202452  -0.202589  -0.202753  -0.202948  -0.203179   -0.20345  -0.203768  -0.204136  -0.204561  -0.205049  -0.205605  -0.206236  -0.206946  -0.207742  -0.208629   -0.20961  -0.210692  -0.211877  -0.213167     -0.214566 | 0.665709  0.665709  0.665709  0.665709  0.665709   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571  0.665711  0.665711  0.665712  0.665713  0.665714  0.665715  0.665716  0.665718  0.665721  0.665724  0.665728  0.665734   0.66574  0.665747  0.665757  0.665768  0.665781  0.665798  0.665817  0.665839  0.665866  0.665896  0.665932  0.665973  0.666019  0.666073  0.666133  0.666201  0.666276  0.666361  0.666454  0.666556  0.666669  0.666792      0.666925 | 100.000029 | 100.000036  100.000053  100.000078  100.000114  100.000165  100.000238  100.000341  100.000486  100.000689   100.00097  100.001357  100.001888  100.002612  100.003591  100.004908   100.00667   100.00901  100.012099  100.016154  100.021442  100.028295  100.037121  100.048419  100.062789  100.080955  100.103775  100.132264  100.167612  100.211198  100.264612  100.329671  100.408431  100.503203   100.61656  100.751342  100.910656  101.097877  101.316628  101.570775  101.864399  102.201772  102.587321  103.025598  103.521234  104.078895  104.703243  105.398881  106.170317  107.021914          107.957852 | 108.445009 |   0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207  0.720701  0.720701  0.720701  0.720702  0.720703  0.720705  0.720706  0.720709  0.720712  0.720717  0.720723   0.72073   0.72074  0.720754   0.72077  0.720792  0.720819  0.720854  0.720897  0.720951  0.721019  0.721102  0.721203  0.721327  0.721476  0.721656  0.721871  0.722126  0.722428  0.722782  0.723194  0.723673  0.724225  0.724858  0.725579  0.726396  0.727316  0.728348  0.729499  0.730774  0.732181  0.733724        0.73541 | 2.591117  2.591117  2.591117  2.591117  2.591118  2.591119   2.59112  2.591122  2.591124  2.591127  2.591131  2.591137  2.591145  2.591156   2.59117   2.59119  2.591215  2.591249  2.591294  2.591352  2.591428  2.591525  2.591649  2.591807  2.592007  2.592258  2.592572  2.592961   2.59344  2.594027  2.594742  2.595607  2.596648  2.597891  2.599369  2.601114  2.603162  2.605552  2.608324  2.611521  2.615188  2.619368  2.624108  2.629452  2.635446  2.642131   2.64955  2.657739  2.666735        2.676569 | 100.000029  100.000043  100.000063  100.000092  100.000134  100.000195   100.00028  100.000401   100.00057  100.000805   100.00113  100.001578  100.002191  100.003023  100.004146  100.005653  100.007662  100.010325  100.013833  100.018423  100.024394  100.032111  100.042024  100.054679  100.070732  100.090971  100.116326  100.147894  100.186956  100.234991  100.293697  100.365008  100.451104  100.554425  100.677677  100.823832  100.996135  101.198087  101.433438  101.706167  102.020459  102.380674  102.791312  103.256975  103.782324  104.372036  105.030756  105.763053  106.573372  107.465997  108.445009 |  100.0 |   0.000029 | 108.445009 |
    | 06:30 | 3.717829  3.717829  3.717829   3.71783   3.71783  3.717831  3.717833  3.717835  3.717838  3.717842  3.717848  3.717855  3.717866   3.71788  3.717899  3.717925  3.717959  3.718004  3.718063  3.718141  3.718242  3.718373  3.718541  3.718754  3.719025  3.719367  3.719794  3.720326  3.720984  3.721793   3.72278  3.723979  3.725426  3.727162   3.72923  3.731682   3.73457  3.737952   3.74189  3.746448  3.751693  3.757697  3.764529  3.772263  3.780968  3.790716  3.801575  3.813608  3.826878             3.841438 | 124.878678  124.878691  124.878709  124.878735  124.878773  124.878828  124.878904  124.879012  124.879163  124.879374  124.879665  124.880064  124.880609  124.881349  124.882347  124.883685  124.885467  124.887826  124.890931   124.89499  124.900265  124.907079  124.915825  124.926982  124.941126  124.958947  124.981262  125.009031  125.043376  125.085593  125.137171  125.199803  125.275398  125.366091  125.474252  125.602481  125.753613  125.930706  126.137033  126.376058  126.651417  126.966887  127.326354  127.733773  128.193129  128.708391  129.283466  129.922158  130.628121  131.404816 |  52.914561   52.914564   52.914568   52.914573    52.91458   52.914591   52.914606   52.914627   52.914656   52.914697   52.914754   52.914832   52.914939   52.915084   52.915279    52.91554   52.915888   52.916349   52.916956   52.917749    52.91878   52.920112   52.921821   52.924001   52.926765   52.930247   52.934607   52.940032   52.946741   52.954987   52.965059   52.977287   52.992042   53.009739   53.030837   53.055838    53.08529    53.11978   53.159936   53.206418   53.259915    53.32114   53.390819   53.469683   53.558462   53.657872   53.768606   53.891325   54.026646        54.175138 |  52.178286   52.178289   52.178292   52.178297   52.178305   52.178315    52.17833    52.17835   52.178379    52.17842   52.178475   52.178552   52.178656   52.178798    52.17899   52.179246   52.179587    52.18004   52.180634   52.181412   52.182423   52.183729   52.185405   52.187543   52.190253   52.193668   52.197943   52.203263   52.209841   52.217927   52.227803   52.239794   52.254263   52.271616   52.292304    52.31682   52.345699    52.37952   52.418896   52.464475   52.516934    52.57697   52.645295   52.722628   52.809683   52.907162   53.015746   53.136082   53.268775     53.414383 | 1.088385  1.088385  1.088385  1.088385  1.088385  1.088386  1.088386  1.088386  1.088387  1.088387  1.088388  1.088389  1.088391  1.088393  1.088396    1.0884  1.088405  1.088412  1.088422  1.088434  1.088449   1.08847  1.088496  1.088529  1.088571  1.088624   1.08869  1.088772  1.088874     1.089  1.089153  1.089339  1.089563  1.089832  1.090153  1.090532   1.09098  1.091503  1.092113  1.092818  1.093629  1.094557  1.095612  1.096806  1.098149  1.099651  1.101323  1.103173  1.105212  1.107446 | 1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916  1.359159  1.359159  1.359158  1.359157  1.359156  1.359155  1.359153  1.359151  1.359148  1.359145   1.35914  1.359135  1.359128   1.35912   1.35911  1.359097  1.359083  1.359065  1.359044  1.359019  1.358989  1.358955  1.358915  1.358869  1.358816  1.358755  1.358687  1.358609  1.358523  1.358426   1.35832  1.358202  1.358073          1.357933 | 0.536158  0.536158  0.536158  0.536158  0.536158  0.536159  0.536159  0.536159  0.536159  0.536159   0.53616   0.53616  0.536161  0.536162  0.536164  0.536166  0.536169  0.536172  0.536177  0.536183  0.536191  0.536201  0.536215  0.536231  0.536253   0.53628  0.536313  0.536355  0.536407  0.536471  0.536549  0.536643  0.536757  0.536893  0.537056  0.537248  0.537475   0.53774  0.538048  0.538404  0.538813   0.53928   0.53981  0.540409   0.54108   0.54183  0.542661  0.543579  0.544586      0.545685 | -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201869  -0.201869   -0.20187  -0.201871  -0.201872  -0.201873  -0.201875  -0.201878  -0.201881  -0.201886  -0.201892    -0.2019  -0.201911  -0.201924  -0.201941  -0.201963  -0.201991  -0.202025  -0.202068  -0.202122  -0.202188  -0.202269  -0.202367  -0.202486  -0.202629    -0.2028  -0.203002  -0.203242  -0.203522  -0.203849  -0.204227  -0.204663  -0.205162   -0.20573  -0.206372  -0.207094  -0.207902    -0.2088  -0.209792  -0.210884  -0.212078     -0.213377 | 0.665709  0.665709  0.665709  0.665709  0.665709  0.665709   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571  0.665711  0.665711  0.665711  0.665712  0.665713  0.665714  0.665715  0.665717  0.665719  0.665722  0.665726   0.66573  0.665735  0.665742   0.66575   0.66576  0.665771  0.665785  0.665802  0.665822  0.665845  0.665872  0.665904   0.66594  0.665982   0.66603  0.666084  0.666145  0.666214  0.666291  0.666376   0.66647  0.666574  0.666687       0.66681 | 100.000023 | 100.000028  100.000042  100.000062  100.000091  100.000132  100.000191  100.000275  100.000392  100.000557  100.000786  100.001102  100.001537   100.00213  100.002936  100.004022  100.005478  100.007418  100.009985  100.013364  100.017782  100.023525  100.030941  100.040461  100.052605  100.068002  100.087402  100.111695  100.141928  100.179324  100.225297  100.281469  100.349691  100.432048  100.530877  100.648771  100.788585  100.953434  101.146687  101.371957  101.633085  101.934114  102.279269  102.672918  103.119537  103.623672  104.189895  104.822761  105.526761  106.306283          107.165567 |  107.61339 |   0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207  0.720701  0.720701  0.720702  0.720703  0.720704  0.720705  0.720707   0.72071  0.720714  0.720719  0.720725  0.720734  0.720744  0.720759  0.720777    0.7208  0.720829  0.720866  0.720912   0.72097  0.721041  0.721128  0.721235  0.721365  0.721521  0.721709  0.721932  0.722197  0.722509  0.722874  0.723299   0.72379  0.724356  0.725003  0.725738   0.72657  0.727506  0.728553  0.729718  0.731008  0.732428       0.733984 | 2.591117  2.591117  2.591117  2.591117  2.591118  2.591118  2.591119  2.591121  2.591122  2.591125  2.591128  2.591133   2.59114  2.591149  2.591161  2.591177  2.591198  2.591226  2.591263  2.591312  2.591375  2.591457  2.591562  2.591695  2.591865  2.592078  2.592346  2.592678   2.59309  2.593595  2.594213  2.594962  2.595867  2.596951  2.598245  2.599777  2.601582  2.603695  2.606155  2.609003   2.61228  2.616029  2.620295  2.625122  2.630555  2.636636  2.643408   2.65091  2.659179        2.668248 | 100.000023  100.000034   100.00005  100.000074  100.000108  100.000156  100.000225  100.000323   100.00046  100.000651  100.000917  100.001283  100.001784  100.002468  100.003392  100.004637    100.0063  100.008509  100.011427  100.015257  100.020252  100.026726  100.035065  100.045742  100.059326  100.076501  100.098084  100.125039  100.158494  100.199763  100.250361   100.31202  100.386703  100.476619  100.584232  100.712265  100.863703  101.041791  101.250023  101.492131  101.772065  102.093967  102.462141   102.88102  103.355129  103.889036  104.487318  105.154511  105.895065  106.713305   107.61339 |  100.0 |   0.000023 |  107.61339 |
    | 07:00 | 3.717829  3.717829  3.717829  3.717829   3.71783  3.717831  3.717832  3.717834  3.717836  3.717839  3.717844   3.71785  3.717859   3.71787  3.717886  3.717907  3.717936  3.717973  3.718023  3.718088  3.718172  3.718282  3.718423  3.718603  3.718833  3.719123  3.719487  3.719941  3.720505    3.7212  3.722051  3.723088  3.724344  3.725855  3.727662  3.729811  3.732351  3.735335  3.738822  3.742872   3.74755  3.752922  3.759057  3.766025  3.773896  3.782741  3.792627  3.803621  3.815786             3.829179 | 124.878673  124.878683  124.878697  124.878719  124.878749  124.878793  124.878855  124.878942  124.879065  124.879236  124.879473    124.8798  124.880247  124.880855  124.881677  124.882782  124.884258  124.886216    124.8888  124.892188  124.896603  124.902321  124.909681  124.919097   124.93107  124.946198  124.965199  124.988916   125.01834  125.054621  125.099087  125.153254  125.218843  125.297789  125.392248  125.504605   125.63747  125.793681  125.976289  126.188549    126.4339  126.715939  127.038395  127.405093  127.819917  128.286769  128.809529  129.392005  130.037896  130.750747 |   52.91456   52.914562   52.914565   52.914569   52.914575   52.914584   52.914596   52.914613   52.914637   52.914671   52.914717   52.914781   52.914868   52.914987   52.915148   52.915364   52.915652   52.916035    52.91654   52.917202   52.918065   52.919182   52.920621   52.922461     52.9248   52.927756   52.931469   52.936103   52.941851   52.948938   52.957622   52.968199   52.981004   52.996412   53.014842   53.036756   53.062658   53.093096   53.128654   53.169958    53.21766   53.272443   53.335008   53.406069   53.486343   53.576542   53.677363    53.78948   53.913533        54.050117 |  52.178285   52.178287    52.17829   52.178294     52.1783   52.178308    52.17832   52.178337    52.17836   52.178393   52.178439   52.178501   52.178587   52.178704   52.178861   52.179073   52.179356   52.179731   52.180226   52.180875   52.181721   52.182817   52.184228   52.186032   52.188326   52.191225   52.194865   52.199409   52.205046   52.211995   52.220511   52.230882   52.243438   52.258547    52.27662   52.298108   52.323507   52.353353   52.388222   52.428723   52.475499   52.529218   52.590568   52.660249   52.738964   52.827411   52.926275   53.036215   53.157859     53.291791 | 1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088386  1.088386  1.088386  1.088387  1.088387  1.088388   1.08839  1.088392  1.088394  1.088397  1.088402  1.088408  1.088415  1.088425  1.088438  1.088455  1.088477  1.088505  1.088541  1.088586  1.088642  1.088713    1.0888  1.088908   1.08904  1.089201  1.089395  1.089629   1.08991  1.090242  1.090636  1.091098  1.091638  1.092265  1.092988  1.093819  1.094767  1.095843  1.097058  1.098422  1.099945  1.101638  1.103508  1.105565 | 1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916  1.359159  1.359159  1.359159  1.359158  1.359157  1.359156  1.359154  1.359153   1.35915  1.359147  1.359144  1.359139  1.359133  1.359126  1.359117  1.359107  1.359094  1.359078   1.35906  1.359038  1.359012  1.358981  1.358946  1.358905  1.358857  1.358803  1.358741  1.358672  1.358593  1.358505  1.358407  1.358299  1.358181          1.358051 | 0.536158  0.536158  0.536158  0.536158  0.536158  0.536159  0.536159  0.536159  0.536159  0.536159   0.53616   0.53616  0.536161  0.536162  0.536163  0.536165  0.536167   0.53617  0.536174  0.536179  0.536185  0.536194  0.536205  0.536219  0.536238   0.53626  0.536289  0.536325  0.536369  0.536424  0.536491  0.536573  0.536672  0.536791  0.536933  0.537102  0.537301  0.537535  0.537808  0.538125   0.53849  0.538909  0.539386  0.539926  0.540535  0.541217  0.541976  0.542818  0.543744       0.54476 | -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201869  -0.201869  -0.201869   -0.20187  -0.201871  -0.201872  -0.201874  -0.201876  -0.201879  -0.201883  -0.201888  -0.201895  -0.201903  -0.201915  -0.201929  -0.201948  -0.201971     -0.202  -0.202037  -0.202083   -0.20214  -0.202209  -0.202294  -0.202397  -0.202522  -0.202671  -0.202848  -0.203058  -0.203306  -0.203595  -0.203931   -0.20432  -0.204766  -0.205276  -0.205856  -0.206509  -0.207243  -0.208062  -0.208971  -0.209974  -0.211075     -0.212278 | 0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571  0.665711  0.665711  0.665712  0.665712  0.665713  0.665714  0.665716  0.665718   0.66572  0.665723  0.665727  0.665731  0.665737  0.665744  0.665752  0.665763  0.665775  0.665789  0.665807  0.665827  0.665851  0.665879  0.665912  0.665949  0.665992  0.666041  0.666096  0.666158  0.666228  0.666306  0.666392  0.666487  0.666591      0.666705 | 100.000018 | 100.000023  100.000034   100.00005  100.000073  100.000106  100.000153  100.000221  100.000316   100.00045  100.000636  100.000894   100.00125  100.001736  100.002398  100.003293  100.004495  100.006101  100.008233  100.011045  100.014733  100.019538  100.025762  100.033773  100.044023  100.057055  100.073524  100.094208  100.120029  100.152064  100.191569  100.239991  100.298986  100.370432  100.456445  100.559384  100.681862  100.826744   100.99715  101.196444  101.428222    101.6963  102.004684  102.357547  102.759194   103.21403  103.726515  104.301126  104.942314   105.65446          106.441833 | 106.852701 |   0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207  0.720701  0.720701  0.720701  0.720702  0.720703  0.720704  0.720706  0.720708  0.720711  0.720715  0.720721  0.720728  0.720737  0.720749  0.720764  0.720783  0.720808   0.72084  0.720879  0.720928  0.720989  0.721064  0.721156  0.721268  0.721404  0.721567  0.721763  0.721995  0.722269  0.722591  0.722968  0.723405  0.723909  0.724488  0.725149  0.725899  0.726746  0.727697  0.728758  0.729938  0.731241       0.732674 | 2.591116  2.591117  2.591117  2.591117  2.591117  2.591118  2.591119   2.59112  2.591121  2.591123  2.591126   2.59113  2.591135  2.591143  2.591152  2.591166  2.591183  2.591207  2.591238  2.591278  2.591331    2.5914  2.591488  2.591601  2.591744  2.591926  2.592153  2.592437   2.59279  2.593224  2.593757  2.594405   2.59519  2.596135  2.597264  2.598607  2.600195   2.60206  2.604239  2.606769  2.609692  2.613047  2.616878  2.621229  2.626142  2.631661  2.637828  2.644684  2.652267        2.660613 | 100.000018  100.000027   100.00004  100.000059  100.000086  100.000125  100.000181   100.00026  100.000371  100.000526  100.000743  100.001042  100.001452  100.002013  100.002774  100.003799  100.005174  100.007006  100.009431  100.012622  100.016795   100.02222  100.029227  100.038223  100.049701  100.064257  100.082603  100.105584  100.134194  100.169595  100.213135   100.26636  100.331034   100.40915  100.502943  100.614899  100.747755  100.904505  101.088394  101.302906  101.551752   101.83885  102.168298  102.544344  102.971353   103.45377  103.996073   104.60274  105.278198  106.026783  106.852701 |  100.0 |   0.000018 | 106.852701 |
    | 07:30 | 3.717828  3.717829  3.717829  3.717829   3.71783   3.71783  3.717831  3.717833  3.717835  3.717837  3.717841  3.717846  3.717853  3.717863  3.717876  3.717893  3.717916  3.717948  3.717989  3.718043  3.718114  3.718205  3.718324  3.718476   3.71867  3.718916  3.719225  3.719613  3.720095  3.720691  3.721424   3.72232  3.723408  3.724721  3.726297  3.728177  3.730406  3.733035  3.736117  3.739708   3.74387  3.748667  3.754164  3.760429  3.767531  3.775538  3.784519  3.794541  3.805667             3.817958 | 124.878668  124.878676  124.878688  124.878705   124.87873  124.878765  124.878815  124.878886  124.878985  124.879125  124.879318  124.879585  124.879951   124.88045  124.881126  124.882038  124.883258  124.884882  124.887031  124.889855  124.893545  124.898338  124.904524  124.912461  124.922581  124.935407  124.951563  124.971789  124.996958  125.028089  125.066361  125.113131  125.169943  125.238544  125.320891  125.419162  125.535752  125.673282  125.834586   126.02271  126.240888  126.492531  126.781196  127.110563  127.484395  127.906504  128.380712  128.910807  129.500503  130.153395 |   52.91456   52.914561   52.914563   52.914567   52.914572   52.914578   52.914588   52.914602   52.914622   52.914649   52.914686   52.914739    52.91481   52.914908    52.91504   52.915218   52.915457   52.915774   52.916194   52.916746   52.917467   52.918404   52.919613   52.921164   52.923141   52.925648   52.928805   52.932757   52.937674   52.943755   52.951231   52.960365   52.971458   52.984849    53.00092   53.020092   53.042829   53.069638   53.101063    53.13769   53.180137   53.229053   53.285111   53.349001   53.421424   53.503083   53.594674   53.696874   53.810339        53.935686 |  52.178284   52.178286   52.178288   52.178292   52.178296   52.178303   52.178313   52.178326   52.178345   52.178372   52.178409    52.17846    52.17853   52.178626   52.178756    52.17893   52.179164   52.179475   52.179887   52.180428   52.181136   52.182054   52.183239    52.18476     52.1867   52.189157   52.192253   52.196128    52.20095   52.206913   52.214244     52.2232   52.234078   52.247209   52.262968   52.281768   52.304063   52.330351   52.361166   52.397082   52.438705   52.486671    52.54164   52.604289   52.675306    52.75538   52.845191   52.945407   53.056668     53.179582 | 1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088386  1.088386  1.088386  1.088386  1.088387  1.088388  1.088389   1.08839  1.088392  1.088395  1.088399  1.088404   1.08841  1.088418  1.088429  1.088444  1.088462  1.088486  1.088516  1.088554  1.088602  1.088662  1.088737  1.088829  1.088943  1.089082   1.08925  1.089454  1.089698  1.089989  1.090335  1.090742  1.091219  1.091775  1.092419  1.093161  1.094011  1.094979  1.096076  1.097311  1.098696   1.10024  1.101952  1.103842 | 1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916  1.359159  1.359159  1.359158  1.359158  1.359157  1.359155  1.359154  1.359152  1.359149  1.359146  1.359142  1.359137  1.359131  1.359124  1.359114  1.359103   1.35909  1.359074  1.359054  1.359032  1.359005  1.358973  1.358937  1.358895  1.358846  1.358791  1.358728  1.358657  1.358577  1.358488  1.358389  1.358279           1.35816 | 0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536159  0.536159  0.536159  0.536159  0.536159   0.53616   0.53616  0.536161  0.536162  0.536163  0.536165  0.536168  0.536171  0.536175  0.536181  0.536188  0.536197  0.536209  0.536225  0.536244  0.536269  0.536299  0.536337  0.536384  0.536442  0.536512  0.536598  0.536701  0.536825  0.536973  0.537148  0.537355  0.537596  0.537877  0.538203  0.538577  0.539005  0.539492  0.540043  0.540662  0.541354  0.542123  0.542974      0.543909 | -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201869  -0.201869   -0.20187   -0.20187  -0.201871  -0.201873  -0.201875  -0.201877   -0.20188  -0.201885   -0.20189  -0.201897  -0.201907  -0.201919  -0.201935  -0.201954  -0.201979  -0.202011   -0.20205  -0.202098  -0.202158  -0.202231  -0.202321  -0.202429  -0.202558  -0.202714  -0.202898  -0.203116  -0.203371  -0.203669  -0.204015  -0.204414  -0.204871  -0.205392  -0.205982  -0.206647  -0.207392  -0.208222  -0.209141  -0.210155     -0.211265 | 0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571  0.665711  0.665711  0.665712  0.665713  0.665714  0.665715  0.665716  0.665718  0.665721  0.665724  0.665728  0.665733  0.665739  0.665746  0.665755  0.665766  0.665778  0.665794  0.665812  0.665833  0.665857  0.665886   0.66592  0.665958  0.666002  0.666051  0.666108  0.666171  0.666242   0.66632  0.666407  0.666503      0.666609 | 100.000014 | 100.000018  100.000027  100.000039  100.000058  100.000085  100.000123  100.000177  100.000254  100.000363  100.000514  100.000725  100.001015  100.001414  100.001957  100.002693  100.003685  100.005014  100.006781  100.009119  100.012193   100.01621  100.021426   100.02816  100.036799  100.047815  100.061776  100.079363  100.101382  100.128784  100.162679  100.204353  100.255286  100.317164  100.391895  100.481619  100.588719  100.715822  100.865804  101.041785  101.247122  101.485395  101.760388  102.076071  102.436562  102.846107  103.309033  103.829718  104.412543  105.061857          105.781933 | 106.158164 |   0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207  0.720701  0.720701  0.720702  0.720702  0.720703  0.720705  0.720707  0.720709  0.720713  0.720717  0.720723  0.720731   0.72074  0.720753   0.72077  0.720791  0.720817  0.720851  0.720893  0.720945  0.721009  0.721089  0.721185  0.721303  0.721445  0.721615  0.721818  0.722059  0.722343  0.722676  0.723063  0.723512   0.72403  0.724622  0.725297  0.726061  0.726923  0.727888  0.728964  0.730157       0.731473 | 2.591116  2.591117  2.591117  2.591117  2.591117  2.591118  2.591118  2.591119   2.59112  2.591122  2.591124  2.591127  2.591132  2.591138  2.591146  2.591157  2.591171  2.591191  2.591217   2.59125  2.591295  2.591352  2.591426  2.591521  2.591643  2.591796   2.59199  2.592232  2.592534  2.592906  2.593365  2.593925  2.594605  2.595426  2.596411  2.597586   2.59898  2.600623  2.602548  2.604793  2.607393  2.610389  2.613823  2.617735  2.622169  2.627166   2.63277  2.639022   2.64596        2.653621 | 100.000014  100.000021  100.000032  100.000047  100.000069    100.0001  100.000145  100.000209  100.000299  100.000425  100.000601  100.000845  100.001181   100.00164  100.002265  100.003111  100.004246  100.005763  100.007776  100.010432  100.013915  100.018454  100.024334  100.031904  100.041591   100.05391  100.069482  100.089047  100.113477  100.143799  100.181206  100.227077   100.28299   100.35074  100.432347  100.530071  100.646415  100.784132   100.94622  101.135922  101.356709  101.612269  101.906486  102.243413   102.62724  103.062265  103.552851  104.103389  104.718255  105.401772  106.158164 |  100.0 |   0.000014 | 106.158164 |
    | 08:00 | 3.717828  3.717829  3.717829  3.717829  3.717829   3.71783  3.717831  3.717832  3.717833  3.717835  3.717838  3.717843  3.717848  3.717856  3.717867  3.717881  3.717901  3.717926  3.717961  3.718006  3.718065  3.718142  3.718241  3.718369  3.718533  3.718741  3.719004  3.719334  3.719746  3.720257  3.720887  3.721659    3.7226  3.723739  3.725111  3.726752  3.728706  3.731017  3.733735  3.736913   3.74061  3.744884  3.749798  3.755419  3.761813  3.769046  3.777187  3.786301  3.796455              3.80771 | 124.878665  124.878671  124.878681  124.878694  124.878714  124.878742  124.878783   124.87884  124.878921  124.879034  124.879191  124.879409  124.879708  124.880118  124.880674  124.881425  124.882434  124.883779  124.885563  124.887915  124.890996  124.895008  124.900201  124.906882  124.915426  124.926284  124.940002  124.957226  124.978724  125.005395  125.038285  125.078602   125.12773   125.18724  125.258904  125.344702  125.446825  125.567683  125.709901  125.876311  126.069947  126.294023   126.55192  126.847155  127.183355  127.564221  127.993494  128.474916  129.012186  129.608921 |  52.914559    52.91456   52.914562   52.914565   52.914569   52.914574   52.914582   52.914593   52.914609   52.914631   52.914662   52.914704   52.914763   52.914843   52.914952   52.915098   52.915295   52.915558   52.915907   52.916367   52.916969   52.917753   52.918768   52.920074   52.921743   52.923865   52.926546   52.929911   52.934112   52.939322   52.945747   52.953621   52.963215   52.974835   52.988823   53.005566   53.025488   53.049055   53.076773   53.109189   53.146883    53.19047   53.240591   53.297912   53.363111   53.436877   53.519897    53.61285   53.716398        53.831174 |  52.178284   52.178285   52.178287   52.178289   52.178293   52.178299   52.178306   52.178317   52.178333   52.178354   52.178385   52.178426   52.178484   52.178562   52.178669   52.178813   52.179006   52.179264   52.179606   52.180056   52.180647   52.181416   52.182411   52.183691   52.185329   52.187409   52.190038   52.193338   52.197457   52.202566   52.208866   52.216588   52.225995   52.237389   52.251106   52.267524   52.287059   52.310168   52.337348   52.369134   52.406096   52.448837   52.497985   52.554192   52.618125   52.690458   52.771866   52.863015   52.964551     53.077099 | 1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088386  1.088386  1.088386  1.088387  1.088387  1.088388  1.088389  1.088391  1.088393  1.088396    1.0884  1.088406  1.088413  1.088422  1.088434  1.088449  1.088469  1.088494  1.088527  1.088567  1.088619  1.088682  1.088762  1.088859  1.088979  1.089125  1.089302  1.089514  1.089769  1.090071  1.090429   1.09085  1.091342  1.091915  1.092576  1.093336  1.094205  1.095193  1.096309  1.097566  1.098971  1.100535  1.102266 | 1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916  1.359159  1.359159  1.359158  1.359157  1.359156  1.359155  1.359153  1.359151  1.359148  1.359145  1.359141  1.359136  1.359129  1.359121  1.359112    1.3591  1.359086  1.359069  1.359049  1.359025  1.358998  1.358965  1.358928  1.358884  1.358835  1.358778  1.358714  1.358641   1.35856   1.35847   1.35837           1.35826 | 0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536159  0.536159  0.536159  0.536159  0.536159  0.536159   0.53616  0.536161  0.536161  0.536162  0.536164  0.536166  0.536169  0.536172  0.536177  0.536183  0.536191  0.536201  0.536214   0.53623  0.536251  0.536277   0.53631   0.53635  0.536399   0.53646  0.536534  0.536624  0.536732  0.536861  0.537015  0.537196   0.53741  0.537659  0.537948  0.538282  0.538665  0.539103    0.5396   0.54016  0.540789  0.541491   0.54227       0.54313 | -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201869  -0.201869  -0.201869   -0.20187  -0.201871  -0.201872  -0.201873  -0.201875  -0.201878  -0.201882  -0.201886  -0.201892    -0.2019  -0.201911  -0.201924   -0.20194  -0.201962  -0.201988  -0.202022  -0.202063  -0.202114  -0.202178  -0.202255  -0.202348  -0.202461  -0.202596  -0.202758  -0.202949  -0.203174  -0.203438  -0.203745    -0.2041  -0.204509  -0.204976  -0.205508  -0.206109  -0.206785  -0.207541  -0.208382  -0.209312     -0.210335 | 0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571  0.665711  0.665711  0.665711  0.665712  0.665713  0.665714  0.665715  0.665717  0.665719  0.665722  0.665725  0.665729  0.665734  0.665741  0.665748  0.665758  0.665769  0.665782  0.665798  0.665816  0.665838  0.665864  0.665893  0.665928  0.665967  0.666011  0.666062  0.666119  0.666184  0.666255  0.666335  0.666423       0.66652 | 100.000011 | 100.000014  100.000021  100.000031  100.000046  100.000068  100.000099  100.000142  100.000205  100.000293  100.000416  100.000587  100.000824   100.00115  100.001596  100.002201  100.003019  100.004116   100.00558  100.007522  100.010082  100.013435  100.017802  100.023455  100.030727  100.040026  100.051846  100.066778  100.085529  100.108933   100.13797  100.173781  100.217683  100.271186  100.336006  100.414078  100.507566  100.618873  100.750639  100.905748  101.087318  101.298697  101.543445  101.825316  102.148237  102.516276  102.933612    103.4045  103.933232  104.524099          105.181346 | 105.525191 |   0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207  0.720701  0.720701  0.720702  0.720703  0.720704  0.720705  0.720707   0.72071  0.720714  0.720719  0.720725  0.720734  0.720744  0.720758  0.720776  0.720798  0.720827  0.720862  0.720907  0.720962   0.72103  0.721114  0.721216  0.721339  0.721487  0.721664  0.721875  0.722125  0.722419  0.722762  0.723161  0.723622  0.724152  0.724757  0.725446  0.726224    0.7271  0.728079  0.729169       0.730376 | 2.591116  2.591116  2.591117  2.591117  2.591117  2.591117  2.591118  2.591118  2.591119  2.591121  2.591123  2.591125  2.591129  2.591134   2.59114  2.591149  2.591162  2.591178  2.591199  2.591227  2.591264  2.591312  2.591374  2.591455  2.591557  2.591687  2.591851  2.592058  2.592315  2.592635  2.593029  2.593511    2.5941  2.594812  2.595669  2.596696  2.597917  2.599361   2.60106  2.603046  2.605356  2.608026  2.611096  2.614607  2.618599  2.623114  2.628195  2.633882  2.640216        2.647233 | 100.000011  100.000017  100.000025  100.000037  100.000055   100.00008  100.000116  100.000168  100.000241  100.000343  100.000486  100.000685  100.000959  100.001336  100.001849  100.002544  100.003481  100.004736  100.006405  100.008613  100.011517  100.015311  100.020239  100.026601  100.034765  100.045177  100.058377  100.075009   100.09584  100.121772  100.153861  100.193331  100.241593  100.300254  100.371138  100.456292  100.557996  100.678773  100.821383  100.988831  101.184352  101.411404   101.67365  101.974938  102.319273  102.710789  103.153712  103.652327  104.210936  104.833819  105.525191 |  100.0 |   0.000011 | 105.525191 |
    | 08:30 | 3.717828  3.717828  3.717829  3.717829  3.717829   3.71783   3.71783  3.717831  3.717832  3.717834  3.717836   3.71784  3.717845  3.717851   3.71786  3.717872  3.717888  3.717909  3.717937  3.717975  3.718024  3.718088  3.718172  3.718279  3.718417  3.718593  3.718816  3.719097  3.719448  3.719885  3.720426  3.721091  3.721903   3.72289  3.724082  3.725513  3.727221  3.729249  3.731642   3.73445  3.737725  3.741526  3.745911  3.750943  3.756687  3.763207   3.77057  3.778842  3.788087              3.79837 | 124.878662  124.878667  124.878675  124.878686  124.878702  124.878724  124.878757  124.878803  124.878869   124.87896  124.879088  124.879266  124.879511  124.879846  124.880303  124.880922  124.881754  124.882867  124.884347  124.886303  124.888872  124.892227  124.896581  124.902199  124.909401  124.918582  124.930213  124.944861  124.963197  124.986012  125.014233  125.048933  125.091347  125.142886  125.205147  125.279924  125.369215   125.47523  125.600386  125.747312  125.918835  126.117976  126.347928  126.612038  126.913781  127.256732  127.644531  128.080847  128.569341  129.113624 |  52.914558   52.914559   52.914561   52.914563   52.914566   52.914571   52.914577   52.914586   52.914599   52.914617   52.914642   52.914676   52.914724    52.91479   52.914879      52.915   52.915163    52.91538   52.915669   52.916052   52.916554    52.91721   52.918061   52.919158   52.920566    52.92236   52.924633   52.927495   52.931078   52.935535   52.941049   52.947827   52.956111   52.966175    52.97833   52.992926   53.010349   53.031028    53.05543   53.084062   53.117469   53.156229    53.20095   53.252268   53.310839    53.37733   53.452418   53.536775   53.631063        53.735926 |  52.178283   52.178284   52.178286   52.178288   52.178291   52.178295   52.178301    52.17831   52.178323    52.17834   52.178365   52.178399   52.178446    52.17851   52.178598   52.178716   52.178876   52.179089   52.179373   52.179748    52.18024   52.180883   52.181717   52.182794   52.184174   52.185933   52.188162   52.190969   52.194482   52.198853   52.204259   52.210906   52.219029   52.228897   52.240817   52.255129   52.272214   52.292491   52.316419   52.344496   52.377254    52.41526   52.459113   52.509435   52.566868   52.632069   52.705698   52.788417   52.880874       52.9837 | 1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088386  1.088386  1.088386  1.088387  1.088388  1.088389   1.08839  1.088392  1.088394  1.088398  1.088402  1.088408  1.088415  1.088425  1.088438  1.088455  1.088476  1.088504  1.088538  1.088582  1.088636  1.088704  1.088788  1.088891  1.089017   1.08917  1.089355  1.089576  1.089841  1.090155  1.090526  1.090961  1.091468  1.092056  1.092735  1.093513  1.094401  1.095408  1.096545  1.097821  1.099246  1.100829 | 1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916  1.359159  1.359159  1.359159  1.359158  1.359157  1.359156  1.359155  1.359153   1.35915  1.359148  1.359144  1.359139  1.359134  1.359127  1.359119  1.359109  1.359096  1.359082  1.359064  1.359043  1.359019   1.35899  1.358957  1.358918  1.358874  1.358823  1.358765    1.3587  1.358626  1.358544  1.358452          1.358351 | 0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536159  0.536159  0.536159  0.536159  0.536159   0.53616   0.53616  0.536161  0.536162  0.536163  0.536165  0.536167   0.53617  0.536174  0.536179  0.536185  0.536194  0.536205  0.536219  0.536236  0.536258  0.536286  0.536321  0.536363  0.536416   0.53648  0.536557  0.536651  0.536764  0.536898  0.537057  0.537245  0.537466  0.537722   0.53802  0.538362  0.538755  0.539202  0.539708  0.540278  0.540916  0.541628      0.542416 | -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201869  -0.201869   -0.20187   -0.20187  -0.201871  -0.201872  -0.201874  -0.201876  -0.201879  -0.201883  -0.201888  -0.201895  -0.201903  -0.201914  -0.201929  -0.201946  -0.201969  -0.201998  -0.202033  -0.202077  -0.202131  -0.202198  -0.202279  -0.202377  -0.202494  -0.202635  -0.202803  -0.203001  -0.203234  -0.203506  -0.203822  -0.204186  -0.204605  -0.205083  -0.205625  -0.206237  -0.206924  -0.207691  -0.208542     -0.209482 | 0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571  0.665711  0.665711  0.665712  0.665712  0.665713  0.665714  0.665716  0.665718   0.66572  0.665723  0.665726  0.665731  0.665736  0.665743  0.665751   0.66576  0.665772  0.665786  0.665802  0.665821  0.665844   0.66587  0.665901  0.665936  0.665976  0.666021  0.666073  0.666131  0.666196  0.666269   0.66635      0.666438 | 100.000009 | 100.000011  100.000017  100.000025  100.000037  100.000054  100.000079  100.000114  100.000165  100.000236  100.000336  100.000475  100.000668  100.000935    100.0013  100.001797   100.00247  100.003376  100.004588  100.006199  100.008327  100.011124  100.014776  100.019515  100.025629  100.033469  100.043462  100.056123  100.072068  100.092028  100.116867  100.147593  100.185375  100.231563  100.287693  100.355512  100.436979  100.534282  100.649838  100.786301   100.94656   101.13373  101.351145  101.602344   101.89105  102.221146  102.596647  103.021667  103.500387  104.037016          104.635751 | 104.949387 |   0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207  0.720701  0.720701  0.720701  0.720702  0.720703  0.720704  0.720706  0.720708  0.720711  0.720716  0.720721  0.720728  0.720737  0.720748  0.720763  0.720782  0.720807  0.720837  0.720875  0.720922  0.720981  0.721052   0.72114  0.721247  0.721376   0.72153  0.721715  0.721934  0.722193  0.722496  0.722849   0.72326  0.723733  0.724275  0.724894  0.725596  0.726388  0.727278  0.728271       0.729374 | 2.591116  2.591116  2.591117  2.591117  2.591117  2.591117  2.591117  2.591118  2.591119   2.59112  2.591121  2.591124  2.591127  2.591131  2.591136  2.591143  2.591153  2.591167  2.591184  2.591208  2.591239  2.591279  2.591331  2.591398  2.591485  2.591595  2.591734   2.59191  2.592129  2.592403  2.592741  2.593156  2.593664  2.594281  2.595026  2.595921  2.596989  2.598256  2.599752  2.601507  2.603554  2.605928  2.608668  2.611811  2.615398  2.619469  2.624066  2.629228  2.634996         2.64141 | 100.000009  100.000013   100.00002   100.00003  100.000044  100.000064  100.000093  100.000135  100.000194  100.000277  100.000393  100.000555  100.000779  100.001087  100.001508  100.002079  100.002852  100.003888  100.005271  100.007105  100.009522   100.01269  100.016815  100.022155  100.029027  100.037816   100.04899   100.06311  100.080846  100.102991  100.130477  100.164387  100.205978  100.256688  100.318155  100.392229  100.480982  100.586714  100.711961  100.859495  101.032319  101.233662  101.466965  101.735864   102.04417   102.39584  102.794947  103.245651  103.752154  104.318672  104.949387 |  100.0 |   0.000009 | 104.949387 |
    | 09:00 | 3.717828  3.717828  3.717828  3.717829  3.717829  3.717829   3.71783   3.71783  3.717831  3.717833  3.717835  3.717838  3.717841  3.717847  3.717854  3.717864  3.717877  3.717894  3.717918  3.717949   3.71799  3.718044  3.718114  3.718204   3.71832  3.718469  3.718658  3.718896  3.719195  3.719569  3.720032  3.720603  3.721303  3.722157  3.723191  3.724436  3.725928  3.727704  3.729806  3.732282  3.735179  3.738552  3.742457  3.746953  3.752102  3.757966  3.764611  3.772102  3.780502             3.789876 |  124.87866  124.878664   124.87867  124.878679  124.878692   124.87871  124.878736  124.878774  124.878826  124.878901  124.879005   124.87915  124.879349  124.879624  124.879999  124.880508  124.881194  124.882114  124.883341  124.884966  124.887106  124.889908  124.893554  124.898271  124.904336  124.912088  124.921937  124.934376  124.949992  124.969481   124.99366  125.023479  125.060038  125.104601  125.158602  125.223663  125.301599  125.394427  125.504368   125.63385  125.785499  125.962139  126.166774  126.402574  126.672852  126.981041  127.330661  127.725289  128.168525  128.663948 |  52.914558   52.914559    52.91456   52.914562   52.914564   52.914568   52.914573    52.91458    52.91459   52.914605   52.914625   52.914654   52.914693   52.914746    52.91482   52.914919   52.915053   52.915233   52.915473    52.91579   52.916209   52.916756   52.917469   52.918391   52.919576   52.921091   52.923016   52.925446   52.928498   52.932306   52.937029   52.942855   52.949996   52.958699   52.969243   52.981945   52.997156   53.015267    53.03671   53.061953   53.091502     53.1259   53.165722   53.211572   53.264078   53.323886   53.391653   53.468042    53.55371        53.649305 |  52.178283   52.178284   52.178285   52.178286   52.178289   52.178292   52.178297   52.178305   52.178315   52.178329   52.178349   52.178377   52.178415   52.178468   52.178539   52.178637   52.178768   52.178945    52.17918   52.179491   52.179901   52.180438   52.181137   52.182041   52.183203   52.184689   52.186576    52.18896   52.191952   52.195686   52.200318    52.20603   52.213033   52.221567   52.231906   52.244361   52.259276   52.277036   52.298063   52.322815   52.351791   52.385521    52.42457   52.469529   52.521015   52.579662   52.646113   52.721018   52.805023     52.898761 | 1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088386  1.088386  1.088386  1.088387  1.088387  1.088388  1.088389  1.088391  1.088393  1.088395  1.088399  1.088404   1.08841  1.088418  1.088429  1.088443  1.088461  1.088484  1.088514  1.088551  1.088597  1.088655  1.088727  1.088815  1.088924  1.089056  1.089217   1.08941  1.089641  1.089916  1.090242  1.090625  1.091074  1.091596    1.0922  1.092896  1.093692  1.094598  1.095625  1.096781  1.098077  1.099521 | 1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916  1.359159  1.359159  1.359158  1.359158  1.359157  1.359156  1.359154  1.359152   1.35915  1.359147  1.359143  1.359138  1.359132  1.359125  1.359116  1.359105  1.359093  1.359077  1.359059  1.359038  1.359012  1.358983  1.358949  1.358909  1.358863  1.358811  1.358752  1.358686  1.358611  1.358527          1.358435 | 0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536159  0.536159  0.536159  0.536159  0.536159   0.53616   0.53616  0.536161  0.536162  0.536164  0.536165  0.536168  0.536171  0.536175  0.536181  0.536188  0.536197  0.536209  0.536224  0.536243  0.536266  0.536296  0.536332  0.536377  0.536432    0.5365  0.536581  0.536679  0.536796  0.536936  0.537101  0.537296  0.537523  0.537787  0.538092  0.538443  0.538845  0.539301  0.539817  0.540396  0.541044      0.541765 | -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201869  -0.201869   -0.20187   -0.20187  -0.201872  -0.201873  -0.201875  -0.201877   -0.20188  -0.201885   -0.20189  -0.201897  -0.201907  -0.201919  -0.201934  -0.201953  -0.201977  -0.202007  -0.202045  -0.202092  -0.202149  -0.202219  -0.202304  -0.202406  -0.202529  -0.202676   -0.20285  -0.203055  -0.203295  -0.203575    -0.2039  -0.204274  -0.204702   -0.20519  -0.205743  -0.206366  -0.207063  -0.207841     -0.208702 | 0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571  0.665711  0.665711  0.665712  0.665713  0.665713  0.665715  0.665716  0.665718  0.665721  0.665724  0.665727  0.665732  0.665738  0.665745  0.665753  0.665763  0.665775   0.66579  0.665807  0.665827   0.66585  0.665877  0.665908  0.665944  0.665985  0.666031  0.666084  0.666143  0.666209  0.666283      0.666364 | 100.000007 | 100.000009  100.000013   100.00002  100.000029  100.000043  100.000063  100.000092  100.000132   100.00019  100.000271  100.000384  100.000542  100.000759  100.001058  100.001466   100.00202  100.002767  100.003768  100.005103  100.006872  100.009201  100.012251   100.01622  100.021354  100.027955  100.036393  100.047114  100.060654  100.077653   100.09887  100.125193  100.157659  100.197468  100.245996  100.304811  100.375683  100.460597  100.561759  100.681604  100.822796  100.988223  101.180998  101.404439  101.662063  101.957558  102.294762  102.677636  103.110231  103.596653          104.141026 | 104.426552 |   0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207  0.720701  0.720701  0.720702  0.720702  0.720703  0.720705  0.720707  0.720709  0.720713  0.720717  0.720723  0.720731   0.72074  0.720753  0.720769  0.720789  0.720815  0.720847  0.720888  0.720938     0.721  0.721075  0.721168  0.721279  0.721414  0.721575  0.721767  0.721994  0.722262  0.722575  0.722939   0.72336  0.723845    0.7244  0.725032  0.725747  0.726553  0.727456       0.728463 | 2.591116  2.591116  2.591116  2.591117  2.591117  2.591117  2.591117  2.591118  2.591118  2.591119   2.59112  2.591122  2.591125  2.591128  2.591132  2.591138  2.591147  2.591158  2.591172  2.591192  2.591218  2.591251  2.591295  2.591351  2.591424  2.591517  2.591635  2.591784  2.591971  2.592204  2.592494  2.592851  2.593289  2.593823  2.594469  2.595248   2.59618   2.59729  2.598605  2.600152  2.601962   2.60407   2.60651  2.609319  2.612535  2.616197  2.620346  2.625022  2.630264        2.636112 | 100.000007  100.000011  100.000016  100.000024  100.000035  100.000051  100.000075  100.000108  100.000156  100.000223  100.000317  100.000449  100.000632  100.000883  100.001228  100.001698  100.002334  100.003189  100.004333  100.005855  100.007865  100.010507  100.013956  100.018434  100.024211   100.03162  100.041066  100.053037  100.068117  100.087001  100.110509  100.139598  100.175385  100.219151  100.272366  100.336694  100.414012  100.506413  100.616215  100.745969  100.898451  101.076664  101.283828  101.523364   101.79888  102.114148  102.473075  102.879676  103.338039  103.852289  104.426552 |  100.0 |   0.000007 | 104.426552 |
    | 09:30 | 3.717828  3.717828  3.717828  3.717829  3.717829  3.717829  3.717829   3.71783  3.717831  3.717832  3.717834  3.717836  3.717839  3.717843  3.717849  3.717857  3.717868  3.717883  3.717902  3.717928  3.717962  3.718007  3.718065  3.718141  3.718239  3.718364  3.718524  3.718726   3.71898  3.719299  3.719695  3.720185  3.720788  3.721525   3.72242  3.723502  3.724802  3.726355  3.728199  3.730377  3.732935  3.735923  3.739393  3.743402  3.748008  3.753272  3.759257  3.766025  3.773641             3.782168 | 124.878658  124.878662  124.878666  124.878673  124.878684  124.878698  124.878719   124.87875  124.878792  124.878853  124.878937  124.879055  124.879218  124.879443   124.87975  124.880168  124.880733  124.881493  124.882508  124.883857  124.885638  124.887975  124.891025  124.894981  124.900082  124.906619  124.914948  124.925497  124.938778  124.955402  124.976086  125.001672  125.033137  125.071606  125.118367  125.174879  125.242787  125.323927   125.42033  125.534231   125.66806  125.824446  126.006203  126.216318  126.457935  126.734334  127.048901  127.405104  127.806457  128.256488 |  52.914558   52.914558   52.914559   52.914561   52.914563   52.914565    52.91457   52.914575   52.914584   52.914596   52.914612   52.914635   52.914667   52.914711   52.914771   52.914853   52.914963   52.915112    52.91531   52.915574   52.915922   52.916379   52.916975   52.917748   52.918745   52.920022    52.92165   52.923711   52.926307   52.929555   52.933596   52.938595   52.944741   52.952255   52.961387   52.972421   52.985678   53.001513    53.02032   53.042533    53.06862   53.099088   53.134477   53.175359   53.222331   53.276014   53.337046   53.406071    53.48374        53.570696 |  52.178283   52.178283   52.178284   52.178285   52.178287    52.17829   52.178294     52.1783   52.178308    52.17832   52.178336   52.178359    52.17839   52.178433   52.178492   52.178572    52.17868   52.178826    52.17902   52.179279    52.17962   52.180068   52.180653   52.181411   52.182388   52.183641   52.185237   52.187258   52.189803   52.192988   52.196951   52.201853    52.20788   52.215248   52.224203   52.235023   52.248021   52.263549   52.281991   52.303772   52.329353    52.35923   52.393932   52.434019   52.480079    52.53272   52.592566   52.660251   52.736412     52.821679 | 1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088386  1.088386  1.088386  1.088387  1.088387  1.088388   1.08839  1.088391  1.088393  1.088397  1.088401  1.088406  1.088413  1.088422  1.088434  1.088449  1.088468  1.088493  1.088524  1.088564  1.088613  1.088675  1.088751  1.088844  1.088958  1.089097  1.089265  1.089466  1.089707  1.089993   1.09033  1.090726  1.091189  1.091726  1.092347  1.093059  1.093873  1.094798  1.095843  1.097019  1.098334 | 1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916  1.359159  1.359159  1.359158  1.359157  1.359156  1.359155  1.359153  1.359151  1.359149  1.359145  1.359141  1.359136   1.35913  1.359123  1.359113  1.359102  1.359089  1.359073  1.359054  1.359032  1.359006  1.358975   1.35894  1.358899  1.358853    1.3588  1.358739  1.358672  1.358596          1.358511 | 0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536159  0.536159  0.536159  0.536159  0.536159  0.536159   0.53616  0.536161  0.536161  0.536163  0.536164  0.536166  0.536169  0.536172  0.536177  0.536183  0.536191  0.536201  0.536213  0.536229  0.536249  0.536274  0.536306  0.536344  0.536392   0.53645   0.53652  0.536606  0.536708   0.53683  0.536975  0.537146  0.537347  0.537581  0.537853  0.538166  0.538526  0.538936  0.539401  0.539926  0.540515      0.541173 | -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201869  -0.201869  -0.201869   -0.20187  -0.201871  -0.201872  -0.201873  -0.201876  -0.201878  -0.201882  -0.201886  -0.201892    -0.2019   -0.20191  -0.201923  -0.201939   -0.20196  -0.201986  -0.202018  -0.202058  -0.202107  -0.202167  -0.202241  -0.202329  -0.202437  -0.202565  -0.202717  -0.202897   -0.20311  -0.203358  -0.203646  -0.203979  -0.204363  -0.204801  -0.205299  -0.205862  -0.206495  -0.207203     -0.207991 | 0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571  0.665711  0.665711  0.665711  0.665712  0.665713  0.665714  0.665715  0.665717  0.665719  0.665721  0.665725  0.665729  0.665734   0.66574  0.665747  0.665756  0.665766  0.665779  0.665794  0.665811  0.665832  0.665856  0.665883  0.665915  0.665952  0.665994  0.666041  0.666095  0.666155  0.666222      0.666296 | 100.000006 | 100.000007  100.000011  100.000016  100.000023  100.000034   100.00005  100.000073  100.000106  100.000153  100.000218   100.00031  100.000439  100.000616   100.00086  100.001195   100.00165  100.002265  100.003092  100.004197  100.005665  100.007604  100.010148  100.013467  100.017773  100.023325   100.03044  100.039506  100.050988  100.065446  100.083543  100.106061  100.133917  100.168176  100.210065  100.260988  100.322541  100.396518  100.484927  100.589993  100.714163  100.860109   101.03072  101.229101  101.458555  101.722572  102.024805  102.369048  102.759207  103.199265          103.693255 |  103.95269 |   0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207  0.720701  0.720701  0.720702  0.720703  0.720704  0.720706  0.720708   0.72071  0.720714  0.720719  0.720725  0.720733  0.720744  0.720758  0.720775  0.720797  0.720824  0.720859  0.720902  0.720955   0.72102  0.721099  0.721196  0.721313  0.721454  0.721621  0.721821  0.722056  0.722332  0.722655  0.723029  0.723462  0.723959  0.724526  0.725171  0.725899  0.726719       0.727635 | 2.591116  2.591116  2.591116  2.591116  2.591117  2.591117  2.591117  2.591117  2.591118  2.591119   2.59112  2.591121  2.591123  2.591126  2.591129  2.591134  2.591141   2.59115  2.591162  2.591179    2.5912  2.591228  2.591265  2.591312  2.591373  2.591451  2.591551  2.591678  2.591837  2.592036  2.592284   2.59259  2.592967  2.593428  2.593987  2.594664  2.595477  2.596447    2.5976  2.598961   2.60056  2.602427  2.604596    2.6071  2.609978  2.613266  2.617003  2.621229  2.625983        2.631303 | 100.000006  100.000008  100.000013  100.000019  100.000028  100.000041   100.00006  100.000087  100.000125   100.00018  100.000256  100.000363  100.000512  100.000717     100.001  100.001385  100.001908  100.002614  100.003559   100.00482  100.006491  100.008691  100.011572  100.015321  100.020172   100.02641  100.034385  100.044521  100.057325  100.073405  100.093482  100.118399  100.149144  100.186858  100.232854  100.288628  100.355872  100.436484  100.532579  100.646493  100.780785  100.938237  101.121848  101.334826  101.580573  101.862667  102.184837  102.550942  102.964935  103.430835   103.95269 |  100.0 |   0.000006 |  103.95269 |
    | 10:00 | 3.717828  3.717828  3.717828  3.717828  3.717829  3.717829  3.717829   3.71783   3.71783  3.717831  3.717833  3.717834  3.717837   3.71784  3.717845  3.717852  3.717861  3.717873  3.717889   3.71791  3.717939  3.717976  3.718025  3.718088   3.71817  3.718276  3.718411  3.718582  3.718798  3.719069  3.719408  3.719828  3.720346  3.720981  3.721755  3.722694  3.723825   3.72518  3.726795  3.728708  3.730962  3.733603  3.736681  3.740248   3.74436  3.749076  3.754454  3.760558  3.767448             3.775187 | 124.878657   124.87866  124.878663  124.878669  124.878677  124.878689  124.878706   124.87873  124.878765  124.878814  124.878882  124.878978  124.879111  124.879294  124.879546  124.879889  124.880354  124.880981  124.881821  124.882939  124.884419  124.886367  124.888915  124.892229  124.896513  124.902019  124.909053  124.917987  124.929268  124.943428  124.961097  124.983019  125.010057  125.043214   125.08364  125.132647  125.191718  125.262519  125.346904  125.446919  125.564807  125.703005  125.864137  126.051006  126.266583  126.513984  126.796452  127.117328  127.480027     127.888 |  52.914557   52.914558   52.914559    52.91456   52.914561   52.914564   52.914567   52.914572   52.914578   52.914588   52.914601    52.91462   52.914646   52.914682   52.914731   52.914798   52.914889   52.915012   52.915176   52.915394   52.915684   52.916064   52.916562    52.91721   52.918047   52.919123   52.920498   52.922244   52.924448   52.927215   52.930668   52.934951   52.940233    52.94671   52.954605   52.964176   52.975709   52.989529   53.005996   53.025506   53.048494    53.07543   53.106818   53.143197   53.185134   53.233221   53.288072   53.350313   53.420579        53.499506 |  52.178282   52.178283   52.178284   52.178285   52.178286   52.178288   52.178292   52.178296   52.178303   52.178312   52.178325   52.178344   52.178369   52.178404   52.178453   52.178518   52.178608   52.178728   52.178889   52.179103   52.179387    52.17976   52.180248   52.180883   52.181704   52.182759   52.184107   52.185819   52.187981   52.190694    52.19408   52.198279   52.203459    52.20981   52.217553   52.226937   52.238246   52.251798   52.267945   52.287077   52.309618   52.336031   52.366809   52.402482   52.443605   52.490758   52.544543   52.605575   52.674477     52.751872 | 1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088386  1.088386  1.088386  1.088386  1.088387  1.088388  1.088389   1.08839  1.088392  1.088394  1.088398  1.088402  1.088408  1.088416  1.088425  1.088438  1.088454  1.088475  1.088502  1.088535  1.088578   1.08863  1.088695  1.088775  1.088874  1.088994  1.089139  1.089315  1.089525  1.089775  1.090072  1.090421   1.09083  1.091306  1.091859  1.092495  1.093224  1.094056  1.094999  1.096063  1.097257 | 1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916  1.359159  1.359159  1.359159  1.359158  1.359157  1.359156  1.359155  1.359153  1.359151  1.359148  1.359144   1.35914  1.359135  1.359128   1.35912  1.359111  1.359099  1.359085  1.359069  1.359049  1.359026  1.358999  1.358968  1.358931   1.35889  1.358842  1.358788  1.358726  1.358657           1.35858 | 0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536159  0.536159  0.536159  0.536159  0.536159   0.53616   0.53616  0.536161  0.536162  0.536163  0.536165  0.536167   0.53617  0.536174  0.536179  0.536185  0.536194  0.536204  0.536218  0.536235  0.536256  0.536283  0.536316  0.536357  0.536407  0.536468  0.536542  0.536631  0.536738  0.536865  0.537015  0.537192  0.537399   0.53764   0.53792  0.538241  0.538609  0.539028  0.539502  0.540036      0.540635 | -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201869  -0.201869   -0.20187   -0.20187  -0.201871  -0.201872  -0.201874  -0.201876  -0.201879  -0.201883  -0.201888  -0.201895  -0.201903  -0.201914  -0.201928  -0.201945  -0.201967  -0.201995  -0.202029  -0.202071  -0.202123  -0.202186  -0.202263  -0.202356  -0.202468  -0.202601   -0.20276  -0.202946  -0.203166  -0.203421  -0.203718   -0.20406  -0.204452    -0.2049  -0.205408  -0.205981  -0.206625     -0.207343 | 0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571  0.665711  0.665711  0.665712  0.665712  0.665713  0.665714  0.665716  0.665717   0.66572  0.665722  0.665726   0.66573  0.665735  0.665741  0.665749  0.665758  0.665769  0.665782  0.665798  0.665816  0.665837  0.665862   0.66589  0.665923  0.665961  0.666003  0.666051  0.666106  0.666167      0.666235 | 100.000004 | 100.000006  100.000008  100.000013  100.000019  100.000027   100.00004  100.000059  100.000085  100.000123  100.000176  100.000251  100.000355    100.0005  100.000699  100.000973  100.001347  100.001853  100.002535  100.003449  100.004666  100.006277  100.008397  100.011171  100.014777   100.01944  100.025433   100.03309  100.042815  100.055094  100.070507  100.089743  100.113608  100.143046  100.179148  100.223169  100.276541  100.340884  100.418016  100.509966  100.618975  100.747503  100.898226  101.074032  101.278017  101.513466  101.783842  102.092761   102.44397  102.841319          103.288728 | 103.524001 |   0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207  0.720701  0.720701  0.720701  0.720702  0.720703  0.720704  0.720706  0.720709  0.720712  0.720716  0.720721  0.720728  0.720737  0.720748  0.720763  0.720781  0.720805  0.720834   0.72087  0.720916  0.720972  0.721041  0.721124  0.721226  0.721348  0.721494  0.721669  0.721876  0.722119  0.722404  0.722737  0.723122  0.723565  0.724074  0.724653  0.725311  0.726052       0.726885 | 2.591116  2.591116  2.591116  2.591116  2.591117  2.591117  2.591117  2.591117  2.591118  2.591118  2.591119   2.59112  2.591122  2.591124  2.591127  2.591131  2.591137  2.591144  2.591154  2.591168  2.591185  2.591209  2.591239  2.591279   2.59133  2.591396  2.591481  2.591588  2.591723  2.591892  2.592104  2.592367  2.592691  2.593088  2.593572  2.594158  2.594865  2.595713  2.596722  2.597918  2.599327  2.600978  2.602901   2.60513  2.607699  2.610645  2.614004  2.617815  2.622117        2.626947 | 100.000004  100.000007   100.00001  100.000015  100.000022  100.000033  100.000048  100.000069  100.000101  100.000144  100.000207  100.000293  100.000415  100.000582  100.000813  100.001129  100.001559   100.00214  100.002921  100.003965  100.005351  100.007182  100.009585  100.012721  100.016789  100.022035   100.02876   100.03733  100.048187  100.061861  100.078983  100.100295  100.126669  100.159119  100.198813  100.247092  100.305479  100.375688  100.459643  100.559477  100.677538  100.816397  100.978836   101.16785  101.386634  101.638566  101.927192  102.256203  102.629403  103.050686  103.524001 |  100.0 |   0.000004 | 103.524001 |
    | 10:30 | 3.717828  3.717828  3.717828  3.717828  3.717829  3.717829  3.717829  3.717829   3.71783  3.717831  3.717832  3.717833  3.717835  3.717838  3.717842  3.717847  3.717855  3.717865  3.717878  3.717896  3.717919   3.71795  3.717991  3.718044  3.718113  3.718202  3.718316   3.71846  3.718644  3.718875  3.719163  3.719523  3.719967  3.720514  3.721182  3.721995  3.722977  3.724158  3.725569  3.727246  3.729229   3.73156  3.734284  3.737453  3.741117  3.745332  3.750156  3.755648  3.761868             3.768878 | 124.878656  124.878658  124.878661  124.878666  124.878672  124.878682  124.878695  124.878715  124.878743  124.878782  124.878838  124.878916  124.879024  124.879174   124.87938  124.879661  124.880043   124.88056  124.881254   124.88218  124.883409   124.88503  124.887157  124.889929  124.893523  124.898155  124.904088  124.911645  124.921213  124.933257   124.94833  124.967085  124.990285  125.018819  125.053714  125.096144  125.147444   125.20912  125.282857  125.370525  125.474185  125.596087   125.73867  125.904553  126.096529  126.317546  126.570694  126.859177  127.186292  127.555397 |  52.914557   52.914558   52.914558   52.914559    52.91456   52.914562   52.914565   52.914569   52.914574   52.914582   52.914593   52.914608   52.914629   52.914658   52.914699   52.914754   52.914828   52.914929   52.915065   52.915246   52.915486   52.915803   52.916219    52.91676   52.917463   52.918368   52.919528   52.921004   52.922874   52.925228   52.928173   52.931837    52.93637   52.941944   52.948761   52.957047   52.967065   52.979106   52.993498   53.010604   53.030824   53.054592   53.082379   53.114688   53.152056   53.195043   53.244238   53.300244    53.36368        53.435169 |  52.178282   52.178282   52.178283   52.178284   52.178285   52.178287    52.17829   52.178293   52.178299   52.178306   52.178317   52.178332   52.178353   52.178381   52.178421   52.178475   52.178548   52.178647    52.17878   52.178957   52.179193   52.179504   52.179911   52.180443   52.181131   52.182019   52.183156   52.184604   52.186438   52.188745   52.191633   52.195227   52.199671   52.205138   52.211821   52.219947    52.22977   52.241577    52.25569   52.272464   52.292291   52.315598   52.342845   52.374527   52.411168   52.453321   52.501561    52.55648   52.618683     52.688784 | 1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088386  1.088386  1.088386  1.088387  1.088387  1.088388  1.088389  1.088391  1.088393  1.088396  1.088399  1.088404   1.08841  1.088419  1.088429  1.088443  1.088461  1.088483  1.088512  1.088547  1.088592  1.088648  1.088717  1.088802  1.088905  1.089031  1.089183  1.089366  1.089585  1.089845  1.090152  1.090513  1.090935  1.091426  1.091993  1.092645  1.093391   1.09424  1.095201  1.096284 | 1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916  1.359159  1.359159  1.359158  1.359158  1.359157  1.359156  1.359154  1.359152   1.35915  1.359147  1.359143  1.359139  1.359133  1.359126  1.359118  1.359108  1.359096  1.359081  1.359064  1.359044   1.35902  1.358992   1.35896  1.358923   1.35888  1.358831  1.358776  1.358713          1.358643 | 0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536159  0.536159  0.536159  0.536159  0.536159   0.53616   0.53616  0.536161  0.536162  0.536164  0.536165  0.536168  0.536171  0.536175  0.536181  0.536188  0.536197  0.536208  0.536223  0.536241  0.536264  0.536292  0.536327   0.53637  0.536423  0.536487  0.536564  0.536657  0.536768    0.5369  0.537056  0.537239  0.537453  0.537701  0.537988  0.538317  0.538693  0.539121  0.539604      0.540147 | -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201869  -0.201869  -0.201869   -0.20187  -0.201871  -0.201872  -0.201873  -0.201875  -0.201877  -0.201881  -0.201885   -0.20189  -0.201897  -0.201906  -0.201918  -0.201933  -0.201951  -0.201975  -0.202004   -0.20204  -0.202085  -0.202139  -0.202206  -0.202287  -0.202384    -0.2025  -0.202639  -0.202803  -0.202996  -0.203223  -0.203486  -0.203791  -0.204142  -0.204543  -0.205001  -0.205518  -0.206102     -0.206755 | 0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571  0.665711  0.665711  0.665712  0.665713  0.665713  0.665715  0.665716  0.665718   0.66572  0.665723  0.665727  0.665731  0.665737  0.665743  0.665751  0.665761  0.665772  0.665786  0.665802  0.665821  0.665842  0.665868  0.665897  0.665931  0.665969  0.666012  0.666062  0.666117      0.666179 | 100.000004 | 100.000004  100.000007   100.00001  100.000015  100.000022  100.000032  100.000047  100.000068  100.000099  100.000142  100.000202  100.000287  100.000405  100.000568  100.000792  100.001098  100.001514  100.002077  100.002832   100.00384  100.005177  100.006942  100.009256  100.012274  100.016186  100.021227  100.027685  100.035911  100.046326  100.059436  100.075844  100.096261  100.121518  100.152586  100.190581  100.236786  100.292657   100.35984  100.440175  100.535709  100.648699  100.781613  100.937132  101.118142  101.327724  101.569147  101.845843  102.161392  102.519493          102.923938 | 103.136892 |   0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207  0.720701  0.720701  0.720702  0.720703  0.720704  0.720705  0.720707   0.72071  0.720713  0.720717  0.720723   0.72073   0.72074  0.720752  0.720768  0.720788  0.720813  0.720844  0.720883  0.720931   0.72099  0.721062   0.72115  0.721256  0.721384  0.721537  0.721718  0.721932  0.722184  0.722478   0.72282  0.723215   0.72367   0.72419  0.724782  0.725452       0.726206 | 2.591116  2.591116  2.591116  2.591116  2.591116  2.591117  2.591117  2.591117  2.591117  2.591118  2.591118  2.591119  2.591121  2.591122  2.591125  2.591128  2.591133  2.591139  2.591147  2.591159  2.591173  2.591193  2.591218  2.591251  2.591294   2.59135  2.591421  2.591512  2.591626  2.591771  2.591951  2.592176  2.592454  2.592795  2.593213  2.593721  2.594335  2.595074  2.595956  2.597004  2.598244    2.5997  2.601403  2.603383  2.605673  2.608306  2.611319  2.614749  2.618634         2.62301 | 100.000004  100.000005  100.000008  100.000012  100.000018  100.000026  100.000038  100.000056  100.000081  100.000116  100.000166  100.000237  100.000335  100.000472  100.000661   100.00092  100.001272  100.001751  100.002395  100.003258  100.004407  100.005929  100.007932  100.010551  100.013959  100.018365  100.024028  100.031266   100.04046  100.052072  100.066653  100.084856  100.107448  100.135325  100.169529  100.211255  100.261868  100.322918  100.396143  100.483486  100.587098  100.709342  100.852791  101.020231   101.21465  101.439226  101.697314  101.992427  102.328212  102.708423  103.136892 |  100.0 |   0.000004 | 103.136892 |
    | 11:00 | 3.717828  3.717828  3.717828  3.717828  3.717828  3.717829  3.717829  3.717829   3.71783   3.71783  3.717831  3.717832  3.717834  3.717836  3.717839  3.717844   3.71785  3.717858  3.717869  3.717884  3.717903  3.717929  3.717963  3.718007  3.718065   3.71814  3.718236  3.718358  3.718513  3.718709  3.718955  3.719262  3.719643  3.720113  3.720689  3.721391  3.722243   3.72327  3.724502   3.72597  3.727711  3.729763  3.732171  3.734979  3.738238  3.741999  3.746316  3.751248  3.756852             3.763188 | 124.878655  124.878657  124.878659  124.878663  124.878668  124.878676  124.878687  124.878702  124.878725  124.878757  124.878802  124.878865  124.878953  124.879075  124.879244  124.879474  124.879788  124.880214  124.880786  124.881552  124.882572   124.88392  124.885692   124.88801  124.891021  124.894913  124.899911  124.906294  124.914399   124.92463  124.937471  124.953493   124.97337   124.99789  125.027964  125.064641   125.10912  125.162759  125.227083  125.303797  125.394786  125.502121   125.62806  125.775041  125.945678   126.14275  126.369184  126.628038   126.92248   127.25576 |  52.914557   52.914557   52.914558   52.914558   52.914559   52.914561   52.914563   52.914566   52.914571   52.914577   52.914586   52.914598   52.914615   52.914639   52.914672   52.914717   52.914778   52.914862   52.914973   52.915123   52.915322   52.915586   52.915932   52.916385   52.916974   52.917734   52.918711   52.919959   52.921543   52.923542   52.926051   52.929182   52.933065   52.937856   52.943731   52.950895   52.959581   52.970055   52.982612   52.997584   53.015337   53.036272   53.060824   53.089465   53.122695   53.161048   53.205082   53.255375   53.312526        53.377142 |  52.178282   52.178282   52.178283   52.178283   52.178284   52.178286   52.178288   52.178291   52.178295   52.178301    52.17831   52.178322   52.178339   52.178362   52.178395   52.178439   52.178499   52.178581    52.17869   52.178837   52.179032   52.179291   52.179631   52.180075   52.180652   52.181398   52.182355   52.183579   52.185132   52.187092   52.189553   52.192623   52.196431   52.201128   52.206889   52.213914   52.222432   52.232702   52.245016   52.259697   52.277105   52.297633   52.321709   52.349793   52.382378   52.419987   52.463165   52.512482   52.568523     52.631884 | 1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088386  1.088386  1.088386  1.088387  1.088387  1.088388   1.08839  1.088391  1.088394  1.088397  1.088401  1.088406  1.088413  1.088422  1.088433  1.088448  1.088467  1.088491  1.088522   1.08856  1.088607  1.088667  1.088739  1.088829  1.088938   1.08907  1.089229   1.08942  1.089647  1.089917  1.090235  1.090608  1.091043  1.091547  1.092129  1.092798   1.09356  1.094426  1.095405 | 1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916  1.359159  1.359159  1.359158  1.359157  1.359156  1.359155  1.359154  1.359152  1.359149  1.359146  1.359142  1.359137  1.359131  1.359124  1.359115  1.359105  1.359092  1.359077  1.359059  1.359038  1.359014  1.358985  1.358952  1.358914   1.35887   1.35882  1.358764            1.3587 | 0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536159  0.536159  0.536159  0.536159  0.536159   0.53616   0.53616  0.536161  0.536162  0.536163  0.536164  0.536166  0.536169  0.536172  0.536177  0.536183   0.53619    0.5362  0.536212  0.536228  0.536247  0.536271  0.536301  0.536338  0.536384  0.536439  0.536506  0.536587  0.536684    0.5368  0.536937  0.537098  0.537287  0.537507  0.537762  0.538057  0.538394  0.538778  0.539214      0.539706 | -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201869  -0.201869  -0.201869   -0.20187  -0.201871  -0.201872  -0.201874  -0.201876  -0.201878  -0.201882  -0.201886  -0.201892    -0.2019   -0.20191  -0.201922  -0.201938  -0.201958  -0.201983  -0.202014  -0.202052  -0.202099  -0.202157  -0.202227  -0.202311  -0.202413  -0.202534  -0.202678  -0.202848  -0.203048  -0.203281  -0.203552  -0.203865  -0.204224  -0.204635  -0.205102  -0.205629     -0.206222 | 0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571  0.665711  0.665711  0.665711  0.665712  0.665713  0.665714  0.665715  0.665717  0.665719  0.665721  0.665724  0.665728  0.665733  0.665738  0.665745  0.665754  0.665764  0.665776   0.66579  0.665806  0.665825  0.665848  0.665874  0.665904  0.665938  0.665977  0.666022  0.666072      0.666128 | 100.000003 | 100.000003  100.000005  100.000008  100.000012  100.000017  100.000026  100.000038  100.000055  100.000079  100.000114  100.000163  100.000232  100.000328  100.000461  100.000644  100.000895  100.001237    100.0017  100.002323  100.003157  100.004266  100.005733  100.007663  100.010185  100.013463  100.017698  100.023139  100.030087  100.038909  100.050046  100.064023  100.081464  100.103104  100.129798  100.162543   100.20248  100.250918  100.309339  100.379409  100.462992   100.56215  100.679155  100.816482  100.976813  101.163029  101.378201  101.625572  101.908547  102.230667          102.595581 | 102.787966 |   0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207  0.720701  0.720701  0.720701  0.720702  0.720703  0.720704  0.720706  0.720708  0.720711  0.720714  0.720719  0.720725  0.720733  0.720744  0.720757  0.720774  0.720795  0.720822  0.720855  0.720896  0.720947  0.721009  0.721085  0.721177  0.721288  0.721421   0.72158  0.721768   0.72199   0.72225  0.722553  0.722905  0.723311  0.723776  0.724308  0.724912       0.725594 | 2.591116  2.591116  2.591116  2.591116  2.591116  2.591117  2.591117  2.591117  2.591117  2.591117  2.591118  2.591119   2.59112  2.591121  2.591123  2.591126   2.59113  2.591135  2.591142  2.591151  2.591163  2.591179  2.591201  2.591228  2.591264  2.591311  2.591371  2.591447  2.591545  2.591667  2.591821  2.592013  2.592251  2.592545  2.592905  2.593344  2.593877  2.594519  2.595289  2.596206  2.597295  2.598578  2.600082  2.601838  2.603874  2.606224  2.608921  2.612002  2.615501        2.619458 | 100.000003  100.000004  100.000006  100.000009  100.000014  100.000021   100.00003  100.000045  100.000065  100.000093  100.000134  100.000191  100.000271  100.000383  100.000537  100.000748  100.001038  100.001431  100.001962  100.002675  100.003627   100.00489  100.006557  100.008743  100.011594   100.01529  100.020054  100.026158  100.033934  100.043782  100.056183  100.071708  100.091032  100.114947  100.144374   100.18038  100.224187  100.277185  100.340946  100.417234  100.508009  100.615437  100.741893  100.889955  101.062407  101.262228  101.492579  101.756791  102.058341  102.400832  102.787966 |  100.0 |   0.000003 | 102.787966 |
    | 11:30 | 3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717829  3.717829  3.717829   3.71783   3.71783  3.717831  3.717833  3.717835  3.717837  3.717841  3.717846  3.717853  3.717862  3.717874   3.71789  3.717911   3.71794  3.717977  3.718025  3.718088  3.718168  3.718271  3.718403  3.718569  3.718779   3.71904  3.719366  3.719769  3.720265  3.720871  3.721609  3.722501  3.723573  3.724856  3.726382  3.728187   3.73031  3.732795  3.735687  3.739036  3.742893  3.747313  3.752352             3.758067 | 124.878655  124.878656  124.878658  124.878661  124.878665  124.878671   124.87868  124.878692   124.87871  124.878736  124.878773  124.878824  124.878895  124.878995  124.879133  124.879321  124.879579  124.879929  124.880401  124.881034  124.881879  124.882998  124.884475  124.886409   124.88893  124.892195  124.896401  124.901787  124.908643  124.917323  124.928246  124.941914  124.958921  124.979959  125.005839  125.037495     125.076  125.122572  125.178593  125.245607  125.325336   125.41968  125.530719  125.660715  125.812104  125.987495  126.189648  126.421471  126.685989  126.986332 |  52.914557   52.914557   52.914557   52.914558   52.914559    52.91456   52.914562   52.914564   52.914568   52.914573    52.91458    52.91459   52.914604   52.914623    52.91465   52.914687   52.914738   52.914806   52.914898   52.915022   52.915187   52.915406   52.915694   52.916072   52.916565   52.917203   52.918025   52.919078   52.920418   52.922114   52.924248   52.926919   52.930242   52.934353   52.939409   52.945593   52.953113   52.962208   52.973146   52.986228   53.001788   53.020193   53.041848   53.067189   53.096684   53.130836   53.170171   53.215245   53.266629        53.324912 |  52.178282   52.178282   52.178282   52.178283   52.178284   52.178285   52.178287   52.178289   52.178292   52.178297   52.178304   52.178314   52.178328   52.178347   52.178373    52.17841   52.178459   52.178526   52.178616   52.178738     52.1789   52.179114   52.179397   52.179768   52.180251   52.180877   52.181683   52.182715   52.184029   52.185692   52.187785   52.190404   52.193663   52.197693   52.202651   52.208715   52.216089   52.225008   52.235733   52.248561   52.263819   52.281867   52.303101   52.327949   52.356872   52.390361   52.428933    52.47313   52.523517     52.580668 | 1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088386  1.088386  1.088386  1.088386  1.088387  1.088388  1.088389   1.08839  1.088392  1.088395  1.088398  1.088402  1.088408  1.088416  1.088425  1.088438  1.088454  1.088474    1.0885  1.088532  1.088573  1.088624  1.088686  1.088763  1.088857  1.088971   1.08911  1.089276  1.089475  1.089711  1.089991   1.09032  1.090705  1.091153  1.091671  1.092268  1.092952  1.093731  1.094614 | 1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916  1.359159  1.359159  1.359159  1.359158  1.359157  1.359156  1.359155  1.359153  1.359151  1.359148  1.359145  1.359141  1.359136  1.359129  1.359122  1.359113  1.359102  1.359088  1.359073  1.359054  1.359033  1.359007  1.358978  1.358944  1.358905   1.35886  1.358809          1.358751 | 0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536159  0.536159  0.536159  0.536159  0.536159   0.53616   0.53616  0.536161  0.536162  0.536163  0.536165  0.536167   0.53617  0.536174  0.536179  0.536185  0.536193  0.536204  0.536217  0.536233  0.536254   0.53628  0.536311   0.53635  0.536398  0.536456  0.536527  0.536611  0.536712  0.536832  0.536974  0.537141  0.537336  0.537563  0.537825  0.538126  0.538472  0.538864      0.539309 | -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201869  -0.201869   -0.20187   -0.20187  -0.201871  -0.201873  -0.201874  -0.201876  -0.201879  -0.201883  -0.201888  -0.201895  -0.201903  -0.201913  -0.201927  -0.201944  -0.201965  -0.201991  -0.202024  -0.202065  -0.202114  -0.202175  -0.202248  -0.202336  -0.202442  -0.202568  -0.202718  -0.202894    -0.2031   -0.20334  -0.203619   -0.20394  -0.204308  -0.204728  -0.205204     -0.205741 | 0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571  0.665711  0.665711  0.665712  0.665712  0.665713  0.665714  0.665716  0.665717  0.665719  0.665722  0.665725  0.665729  0.665734   0.66574  0.665747  0.665756  0.665767  0.665779  0.665793   0.66581   0.66583  0.665854   0.66588  0.665911  0.665946  0.665986  0.666031      0.666082 | 100.000002 | 100.000003  100.000004  100.000006  100.000009  100.000014   100.00002   100.00003  100.000044  100.000064  100.000092  100.000131  100.000187  100.000265  100.000373  100.000523  100.000729  100.001009   100.00139  100.001904  100.002593  100.003512  100.004731  100.006337  100.008443  100.011186  100.014741  100.019318   100.02518  100.032644  100.042091  100.053981   100.06886  100.087373  100.110277  100.138454  100.172921  100.214849  100.265569  100.326586   100.39959  100.486463  100.589284  100.710334  100.852097  101.017253  101.208677  101.429423  101.682716  101.971926          102.300554 | 102.474029 |   0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207  0.720701  0.720701  0.720702  0.720702  0.720703  0.720705  0.720706  0.720709  0.720712  0.720716  0.720721  0.720728  0.720736  0.720748  0.720762   0.72078  0.720802  0.720831  0.720866   0.72091  0.720963  0.721029  0.721108  0.721205  0.721321  0.721459  0.721624  0.721819  0.722049  0.722317  0.722629  0.722991  0.723407  0.723884  0.724427       0.725043 | 2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591117  2.591117  2.591117  2.591117  2.591118  2.591118  2.591119   2.59112  2.591122  2.591124  2.591127  2.591132  2.591137  2.591145  2.591155  2.591168  2.591186  2.591209  2.591239  2.591279  2.591329  2.591393  2.591476   2.59158   2.59171  2.591874  2.592078   2.59233   2.59264  2.593019   2.59348  2.594038  2.594708   2.59551  2.596464  2.597592  2.598919  2.600472   2.60228  2.604373  2.606783  2.609544  2.612691         2.61626 | 100.000002  100.000003  100.000005  100.000007  100.000011  100.000017  100.000024  100.000036  100.000052  100.000075  100.000108  100.000154  100.000219   100.00031  100.000435  100.000608  100.000846  100.001169  100.001606  100.002194  100.002982   100.00403  100.005416  100.007238  100.009621  100.012718   100.01672  100.021861  100.028429  100.036771  100.047303  100.060526  100.077031  100.097517  100.122798  100.153821  100.191677  100.237613  100.293044  100.359565  100.438959  100.533207  100.644486  100.775181  100.927874  101.105345  101.310562  101.546669  101.816968  102.124902  102.474029 |  100.0 |   0.000002 | 102.474029 |
    | 12:00 | 3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717829  3.717829  3.717829  3.717829   3.71783  3.717831  3.717832  3.717833  3.717836  3.717839  3.717843  3.717848  3.717856  3.717866  3.717879  3.717897   3.71792  3.717951  3.717992  3.718044  3.718112  3.718199   3.71831  3.718451  3.718629  3.718852   3.71913  3.719475  3.719901  3.720424  3.721062  3.721835  3.722767  3.723886  3.725221  3.726805  3.728675  3.730869  3.733432  3.736408  3.739847    3.7438  3.748321             3.753466 | 124.878654  124.878655  124.878657  124.878659  124.878662  124.878667  124.878674  124.878684  124.878699   124.87872  124.878749  124.878791  124.878849   124.87893  124.879042  124.879197  124.879408  124.879695  124.880084  124.880607  124.881306  124.882235  124.883463  124.885076  124.887183   124.88992  124.893455  124.897993  124.903787   124.91114   124.92042  124.932065  124.946594   124.96462  124.986858  125.014139  125.047419  125.087792  125.136502  125.194946  125.264691  125.347472  125.445202  125.559969  125.694038  125.849844  126.029984  126.237204  126.474384  126.744522 |  52.914557   52.914557   52.914557   52.914558   52.914558   52.914559   52.914561   52.914563   52.914566    52.91457   52.914575   52.914583   52.914595   52.914611   52.914633   52.914663   52.914704    52.91476   52.914836   52.914938   52.915075   52.915257   52.915497   52.915812   52.916224   52.916759   52.917449   52.918336   52.919469   52.920906   52.922719   52.924995   52.927834   52.931356   52.935701    52.94103   52.947531   52.955416   52.964928   52.976339   52.989953   53.006107   53.025171   53.047551   53.073682   53.104035   53.139106   53.179421   53.225527        53.277992 |  52.178282   52.178282   52.178282   52.178283   52.178283   52.178284   52.178286   52.178287    52.17829   52.178294     52.1783   52.178308   52.178319   52.178335   52.178356   52.178386   52.178426   52.178481   52.178556   52.178656    52.17879   52.178968   52.179203   52.179512   52.179916   52.180441   52.181118   52.181988   52.183098   52.184507   52.186286   52.188517   52.191301   52.194755   52.199015   52.204241   52.210616   52.218348   52.227675   52.238864   52.252213   52.268054   52.286748   52.308693   52.334317    52.36408    52.39847   52.438002   52.483213     52.534659 | 1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088386  1.088386  1.088386  1.088387  1.088387  1.088388  1.088389  1.088391  1.088393  1.088396  1.088399  1.088404   1.08841  1.088419  1.088429  1.088443   1.08846  1.088482  1.088509  1.088544  1.088587  1.088641  1.088707  1.088788  1.088886  1.089006  1.089151  1.089324  1.089531  1.089777  1.090066  1.090406  1.090803  1.091264  1.091797  1.092408  1.093108  1.093903 | 1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916  1.359159  1.359159  1.359158  1.359158  1.359157  1.359156  1.359154  1.359152   1.35915  1.359147  1.359144  1.359139  1.359134  1.359127  1.359119   1.35911  1.359098  1.359085  1.359068  1.359049  1.359027  1.359001   1.35897  1.358935  1.358895   1.35885          1.358798 | 0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536159  0.536159  0.536159  0.536159  0.536159  0.536159   0.53616   0.53616  0.536161  0.536162  0.536164  0.536166  0.536168  0.536171  0.536175  0.536181  0.536188  0.536196  0.536207  0.536221  0.536239  0.536261  0.536288  0.536322  0.536363  0.536413  0.536474  0.536548  0.536636  0.536741  0.536865  0.537012  0.537185  0.537386  0.537619  0.537888  0.538197   0.53855      0.538951 | -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201869  -0.201869  -0.201869   -0.20187  -0.201871  -0.201872  -0.201873  -0.201875  -0.201877  -0.201881  -0.201885   -0.20189  -0.201897  -0.201906  -0.201917  -0.201932   -0.20195  -0.201972     -0.202  -0.202035  -0.202078   -0.20213  -0.202193   -0.20227  -0.202362  -0.202473  -0.202604  -0.202758   -0.20294  -0.203153  -0.203401  -0.203687  -0.204016  -0.204393  -0.204822     -0.205307 | 0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571  0.665711  0.665711  0.665712  0.665712  0.665713  0.665715  0.665716  0.665718   0.66572  0.665723  0.665726  0.665731  0.665736  0.665742   0.66575  0.665759  0.665769  0.665782  0.665797  0.665815  0.665835  0.665859  0.665887  0.665918  0.665954  0.665995      0.666041 | 100.000002 | 100.000002  100.000003  100.000005  100.000007  100.000011  100.000016  100.000024  100.000035  100.000051  100.000074  100.000106  100.000151  100.000214  100.000302  100.000425  100.000593  100.000823  100.001136  100.001559  100.002128  100.002889    100.0039  100.005236  100.006992  100.009286  100.012265  100.016112  100.021051  100.027357  100.035361  100.045462  100.058138  100.073954  100.093578  100.117788   100.14749  100.183727  100.227691   100.28074    100.3444  100.420382  100.510585  100.617103  100.742227  100.888445  101.058435  101.255064   101.48137  101.740552           102.03595 | 102.192082 |   0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207  0.720701  0.720701  0.720702  0.720703  0.720704  0.720705  0.720707   0.72071  0.720713  0.720717  0.720723   0.72073   0.72074  0.720752  0.720767  0.720786   0.72081   0.72084  0.720878  0.720924   0.72098  0.721049  0.721133  0.721234  0.721355  0.721499   0.72167  0.721872  0.722109  0.722386  0.722707  0.723078  0.723505  0.723992       0.724547 | 2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591117  2.591117  2.591117  2.591117  2.591117  2.591118  2.591119   2.59112  2.591121  2.591123  2.591125  2.591129  2.591133   2.59114  2.591148  2.591159  2.591174  2.591193  2.591218  2.591251  2.591294  2.591348  2.591417  2.591506  2.591617  2.591756   2.59193  2.592146  2.592413  2.592739  2.593138  2.593621  2.594205  2.594904  2.595739  2.596729  2.597897  2.599269   2.60087   2.60273  2.604879  2.607349  2.610173        2.613387 | 100.000002  100.000003  100.000004  100.000006  100.000009  100.000013  100.000019  100.000028  100.000041   100.00006  100.000087  100.000124  100.000177  100.000251  100.000353  100.000494  100.000688  100.000953  100.001313  100.001798  100.002449  100.003317  100.004469  100.005986  100.007975  100.010567  100.013925  100.018251  100.023792  100.030848  100.039782  100.051029  100.065107   100.08263  100.104318  100.131008  100.163672  100.203424  100.251536  100.309447  100.378772  100.461316  100.559074  100.674237  100.809196  100.966534  101.149028  101.359633  101.601472  101.877819  102.192082 |  100.0 |   0.000002 | 102.192082 |
    | 12:30 | 3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717829  3.717829  3.717829   3.71783   3.71783  3.717831  3.717832  3.717834  3.717837   3.71784  3.717844  3.717851  3.717859   3.71787  3.717885  3.717904   3.71793  3.717964  3.718008  3.718064  3.718138  3.718231  3.718351  3.718501  3.718691  3.718929  3.719224   3.71959   3.72004   3.72059  3.721259  3.722069  3.723043  3.724209  3.725597   3.72724  3.729175  3.731441  3.734081  3.737141   3.74067  3.744719             3.749341 | 124.878654  124.878654  124.878656  124.878657   124.87866  124.878664   124.87867  124.878678  124.878689  124.878706   124.87873  124.878764  124.878811  124.878877  124.878968  124.879095  124.879268  124.879504  124.879824  124.880255  124.880833  124.881603  124.882623  124.883967  124.885727  124.888018  124.890985  124.894805  124.899694  124.905915  124.913789  124.923697  124.936093  124.951516  124.970596  124.994072  125.022793  125.057738  125.100023   125.15091  125.211819  125.284332  125.370199  125.471344  125.589862  125.728019  125.888246  126.073128  126.285394  126.527901 |  52.914557   52.914557   52.914557   52.914557   52.914558   52.914559    52.91456   52.914561   52.914564   52.914567   52.914572   52.914578   52.914587     52.9146   52.914618   52.914643   52.914677   52.914723   52.914785    52.91487   52.914983   52.915133   52.915333   52.915595   52.915939   52.916387   52.916967   52.917713   52.918669   52.919885   52.921423   52.923359   52.925782   52.928795   52.932524    52.93711   52.942721   52.949546   52.957805   52.967742   52.979633   52.993786   53.010541    53.03027   53.053378   53.080304   53.111513   53.147502   53.188792        53.235925 |  52.178282   52.178282   52.178282   52.178282   52.178283   52.178284   52.178285   52.178286   52.178288   52.178292   52.178296   52.178303   52.178312   52.178324   52.178342   52.178366   52.178399   52.178445   52.178506   52.178589   52.178699   52.178847   52.179042     52.1793   52.179637   52.180076   52.180645   52.181377   52.182314   52.183506   52.185015   52.186913   52.189289   52.192244     52.1959   52.200397   52.205899   52.212592    52.22069   52.230434   52.242094   52.255972   52.272402   52.291748   52.314408    52.34081   52.371413   52.406703   52.447191     52.493409 | 1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088386  1.088386  1.088386  1.088386  1.088387  1.088388  1.088389   1.08839  1.088392  1.088394  1.088397  1.088401  1.088406  1.088413  1.088422  1.088433  1.088448  1.088466  1.088489  1.088519  1.088556  1.088602  1.088658  1.088728  1.088813  1.088917  1.089043  1.089194  1.089374   1.08959  1.089844  1.090144  1.090495  1.090904  1.091378  1.091924   1.09255  1.093265 | 1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916  1.359159  1.359159  1.359158  1.359157  1.359156  1.359155  1.359154  1.359152  1.359149  1.359146  1.359142  1.359138  1.359132  1.359125  1.359117  1.359107  1.359095  1.359081  1.359064  1.359044  1.359021  1.358994  1.358963  1.358927  1.358886          1.358839 | 0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536159  0.536159  0.536159  0.536159  0.536159   0.53616   0.53616  0.536161  0.536162  0.536163  0.536164  0.536166  0.536169  0.536172  0.536177  0.536183   0.53619    0.5362  0.536211  0.536226  0.536245  0.536268  0.536297  0.536333  0.536376  0.536429  0.536493  0.536569  0.536661   0.53677    0.5369  0.537052   0.53723  0.537437  0.537676  0.537953  0.538269       0.53863 | -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201869  -0.201869  -0.201869   -0.20187  -0.201871  -0.201872  -0.201874  -0.201876  -0.201878  -0.201882  -0.201886  -0.201892    -0.2019  -0.201909  -0.201921  -0.201937  -0.201956   -0.20198   -0.20201  -0.202046  -0.202091  -0.202146  -0.202213  -0.202293  -0.202389  -0.202504   -0.20264    -0.2028  -0.202988  -0.203208  -0.203462  -0.203756  -0.204094  -0.204479     -0.204916 | 0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571  0.665711  0.665711  0.665711  0.665712  0.665713  0.665714  0.665715  0.665717  0.665718  0.665721  0.665724  0.665727  0.665732  0.665737  0.665744  0.665752  0.665761  0.665772  0.665786  0.665801  0.665819  0.665841  0.665865  0.665893  0.665925  0.665962      0.666004 | 100.000001 | 100.000002  100.000003  100.000004  100.000006  100.000009  100.000013  100.000019  100.000028  100.000041  100.000059  100.000085  100.000122  100.000173  100.000245  100.000344  100.000482   100.00067  100.000927  100.001275  100.001745  100.002374  100.003212  100.004322  100.005785    100.0077  100.010194  100.013424  100.017581  100.022902  100.029674  100.038245  100.049029  100.062523  100.079312  100.100084  100.125641  100.156912  100.194963   100.24101  100.296433  100.362781  100.441782  100.535353  100.645601  100.774824  100.925513  101.100344  101.302171  101.534017          101.799056 | 101.939318 |   0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207  0.720701  0.720701  0.720701  0.720702  0.720703  0.720704  0.720706  0.720708  0.720711  0.720714  0.720719  0.720725  0.720733  0.720743  0.720756  0.720772  0.720793  0.720819  0.720851   0.72089  0.720939  0.720998  0.721071  0.721158  0.721264   0.72139   0.72154  0.721717  0.721926  0.722171  0.722456  0.722786  0.723167  0.723604       0.724102 | 2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591117  2.591117  2.591117  2.591117  2.591118  2.591118  2.591119   2.59112  2.591122  2.591124  2.591126   2.59113  2.591135  2.591142  2.591152  2.591164   2.59118  2.591201  2.591228  2.591264   2.59131  2.591368  2.591443  2.591537  2.591656  2.591805  2.591989  2.592218  2.592499  2.592843  2.593262  2.593768  2.594377  2.595106  2.595974  2.597001   2.59821  2.599626  2.601276  2.603189  2.605394  2.607923         2.61081 | 100.000001  100.000002  100.000003  100.000005  100.000007   100.00001  100.000015  100.000023  100.000033  100.000048   100.00007    100.0001  100.000143  100.000203  100.000286  100.000401   100.00056  100.000777  100.001073  100.001473   100.00201  100.002729  100.003684  100.004946  100.006605  100.008771  100.011586  100.015222   100.01989  100.025851  100.033419  100.042973  100.054965  100.069933  100.088511   100.11144  100.139582   100.17393  100.215625  100.265959  100.326394  100.398568  100.484302  100.585605  100.704681  100.843925  101.005921  101.193438  101.409418  101.656963  101.939318 |  100.0 |   0.000001 | 101.939318 |
    | 13:00 | 3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717829  3.717829  3.717829  3.717829   3.71783  3.717831  3.717832  3.717833  3.717835  3.717838  3.717841  3.717846  3.717853  3.717862  3.717875  3.717891  3.717912   3.71794  3.717977  3.718025  3.718086  3.718165  3.718266  3.718394  3.718555  3.718758   3.71901  3.719323   3.71971  3.720184  3.720762  3.721464  3.722312  3.723328  3.724542  3.725984  3.727686  3.729686  3.732024  3.734742  3.737887  3.741506              3.74565 | 124.878653  124.878654  124.878655  124.878656  124.878658  124.878661  124.878666  124.878673  124.878682  124.878696  124.878715  124.878742   124.87878  124.878834  124.878908  124.879011  124.879153  124.879347   124.87961  124.879965  124.880443   124.88108  124.881927  124.883045  124.884513   124.88643  124.888917  124.892128  124.896248  124.901506  124.908178  124.916596  124.927158  124.940335  124.956684  124.976855  125.001605  125.031805  125.068456  125.112693  125.165799  125.229211  125.304527  125.393513  125.498099  125.620388  125.762644  125.927293  126.116909    126.3342 |  52.914557   52.914557   52.914557   52.914557   52.914558   52.914558   52.914559    52.91456   52.914562   52.914565   52.914569   52.914574   52.914581   52.914592   52.914606   52.914627   52.914654   52.914692   52.914744   52.914813   52.914906   52.915031   52.915196   52.915415   52.915702   52.916076   52.916563    52.91719   52.917995   52.919023   52.920327   52.921972   52.924036   52.926611   52.929805   52.933746   52.938582   52.944481    52.95164   52.960279   52.970648   52.983028   52.997727   53.015089   53.035487   53.059329    53.08705   53.119116   53.156021        53.198281 |  52.178282   52.178282   52.178282   52.178282   52.178283   52.178283   52.178284   52.178285   52.178287    52.17829   52.178293   52.178299   52.178306   52.178316    52.17833    52.17835   52.178377   52.178414   52.178465   52.178533   52.178624   52.178747   52.178909   52.179123   52.179405   52.179772   52.180249   52.180864   52.181654   52.182661    52.18394   52.185553   52.187577   52.190102   52.193234   52.197099    52.20184   52.207625   52.214645   52.223116   52.233284   52.245423   52.259837   52.276862   52.296864   52.320242   52.347425   52.378869   52.415057     52.456496 | 1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088386  1.088386  1.088386  1.088387  1.088387  1.088388  1.088389   1.08839  1.088392  1.088395  1.088398  1.088402  1.088408  1.088416  1.088425  1.088437  1.088453  1.088473  1.088498  1.088529  1.088568  1.088617  1.088677   1.08875   1.08884  1.088949   1.08908  1.089238  1.089426  1.089649  1.089913  1.090223  1.090585  1.091006  1.091493  1.092053  1.092694 | 1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916  1.359159  1.359159  1.359159  1.359158  1.359157  1.359156  1.359155  1.359153  1.359151  1.359148  1.359145  1.359141  1.359136   1.35913  1.359123  1.359114  1.359104  1.359092  1.359077  1.359059  1.359039  1.359015  1.358987  1.358955  1.358919          1.358877 | 0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536159  0.536159  0.536159  0.536159  0.536159   0.53616   0.53616  0.536161  0.536162  0.536163  0.536165  0.536167   0.53617  0.536174  0.536179  0.536185  0.536193  0.536203  0.536216  0.536232  0.536252  0.536276  0.536307  0.536344   0.53639  0.536445  0.536512  0.536592  0.536687  0.536801  0.536935  0.537092  0.537275  0.537489  0.537735  0.538018      0.538342 | -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201869  -0.201869   -0.20187   -0.20187  -0.201871  -0.201873  -0.201874  -0.201877  -0.201879  -0.201883  -0.201888  -0.201894  -0.201903  -0.201913  -0.201926  -0.201942  -0.201963  -0.201988   -0.20202  -0.202058  -0.202106  -0.202163  -0.202233  -0.202317  -0.202417  -0.202536  -0.202677  -0.202843  -0.203037  -0.203263  -0.203525  -0.203827  -0.204172     -0.204566 | 0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571  0.665711  0.665711  0.665712  0.665712  0.665713  0.665714  0.665715  0.665717  0.665719  0.665722  0.665725  0.665729  0.665733  0.665739  0.665746  0.665754  0.665764  0.665776  0.665789  0.665805  0.665824  0.665846  0.665871    0.6659  0.665933       0.66597 | 100.000001 | 100.000001  100.000002  100.000003  100.000005  100.000007   100.00001  100.000015  100.000022  100.000033  100.000047  100.000068  100.000098   100.00014  100.000198  100.000279  100.000391  100.000545  100.000756  100.001042  100.001429  100.001949  100.002643  100.003565  100.004782  100.006379  100.008465  100.011173  100.014667  100.019152  100.024875  100.032137    100.0413  100.052797  100.067141  100.084939  100.106898  100.133843  100.166725  100.206634  100.254809   100.31265  100.381727  100.463788  100.560763   100.67477  100.808114  100.963288  101.142962   101.34998          101.587344 | 101.713118 |   0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207  0.720701  0.720701  0.720702  0.720702  0.720703  0.720705  0.720706  0.720709  0.720712  0.720716  0.720721  0.720728  0.720736  0.720747  0.720761  0.720778    0.7208  0.720827  0.720861  0.720903  0.720954  0.721017  0.721093  0.721184  0.721294  0.721426  0.721581  0.721765  0.721981  0.722234  0.722527  0.722867  0.723257       0.723704 | 2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591117  2.591117  2.591117  2.591117  2.591118  2.591118  2.591119  2.591121  2.591122  2.591125  2.591128  2.591132  2.591138  2.591145  2.591155  2.591169  2.591186  2.591209  2.591239  2.591278  2.591327   2.59139   2.59147  2.591571  2.591697  2.591855  2.592051  2.592293  2.592589  2.592951   2.59339   2.59392  2.594555  2.595314  2.596215  2.597279   2.59853  2.599991   2.60169  2.603654  2.605916        2.608505 | 100.000001  100.000002  100.000002  100.000004  100.000006  100.000008  100.000012  100.000018  100.000027  100.000039  100.000056   100.00008  100.000115  100.000164  100.000231  100.000326  100.000455  100.000633  100.000876  100.001205  100.001648  100.002242  100.003034  100.004083  100.005464  100.007274   100.00963  100.012682   100.01661   100.02164  100.028043  100.036148  100.046349  100.059118   100.07501  100.094679  100.118889  100.148524  100.184602  100.228282  100.280883  100.343887  100.418951  100.507912  100.612794  100.735809  100.879356  101.046019  101.238556  101.459896  101.713118 |  100.0 |   0.000001 | 101.713118 |
    | 13:30 | 3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717829  3.717829  3.717829   3.71783   3.71783  3.717831  3.717832  3.717834  3.717836  3.717839  3.717843  3.717849  3.717856  3.717866   3.71788  3.717898  3.717921  3.717952  3.717992  3.718043   3.71811  3.718195  3.718303   3.71844  3.718612  3.718828  3.719096  3.719427  3.719835  3.720335  3.720942  3.721677  3.722563  3.723623  3.724885  3.726381  3.728143   3.73021   3.73262  3.735416  3.738644             3.742353 | 124.878653  124.878653  124.878654  124.878655  124.878657   124.87866  124.878663  124.878668  124.878676  124.878687  124.878702  124.878724  124.878755  124.878799   124.87886  124.878944   124.87906  124.879218  124.879434  124.879727  124.880121  124.880648   124.88135   124.88228  124.883503  124.885104  124.887187  124.889883  124.893352   124.89779  124.903436   124.91058  124.919567   124.93081  124.944798  124.962105  124.983401  125.009462   125.04118  125.079576  125.125804  125.181167   125.24712  125.325274  125.417408   125.52546  125.651536  125.797901  125.966972  126.161308 |  52.914557   52.914557   52.914557   52.914557   52.914557   52.914558   52.914559    52.91456   52.914561   52.914563   52.914566   52.914571   52.914577   52.914585   52.914597   52.914613   52.914636   52.914667   52.914709   52.914766   52.914843   52.914947   52.915084   52.915265   52.915505   52.915817   52.916225   52.916751   52.917429   52.918297     52.9194   52.920796   52.922552   52.924749   52.927483   52.930865   52.935025   52.940117   52.946312   52.953812   52.962839   52.973649   52.986523   53.001775    53.01975   53.040823     53.0654   53.093918   53.126841        53.164659 |  52.178281   52.178282   52.178282   52.178282   52.178282   52.178283   52.178283   52.178284   52.178286   52.178288   52.178291   52.178295   52.178301   52.178309   52.178321   52.178337   52.178359    52.17839   52.178431   52.178487   52.178563   52.178664   52.178798   52.178977   52.179211   52.179518   52.179917   52.180434   52.181098   52.181949   52.183031     52.1844   52.186122   52.188276   52.190957   52.194273   52.198353   52.203345   52.209421   52.216774   52.225627   52.236226   52.248851   52.263807   52.281432   52.302096   52.326195    52.35416   52.386443     52.423527 | 1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088386  1.088386  1.088386  1.088387  1.088387  1.088388  1.088389  1.088391  1.088393  1.088396  1.088399  1.088404   1.08841  1.088418  1.088429  1.088442  1.088459   1.08848  1.088507   1.08854  1.088582  1.088633  1.088696  1.088774  1.088868  1.088982  1.089119  1.089283  1.089479  1.089711  1.089984  1.090304  1.090678  1.091111   1.09161  1.092184 | 1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916  1.359159  1.359159  1.359158  1.359158  1.359157  1.359156  1.359154  1.359153   1.35915  1.359148  1.359144   1.35914  1.359135  1.359129  1.359121  1.359112  1.359101  1.359088  1.359073  1.359055  1.359034  1.359009  1.358981  1.358948           1.35891 | 0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536159  0.536159  0.536159  0.536159  0.536159  0.536159   0.53616  0.536161  0.536161  0.536162  0.536164  0.536166  0.536168  0.536171  0.536175  0.536181  0.536187  0.536196  0.536207   0.53622  0.536237  0.536258  0.536284  0.536317  0.536356  0.536404  0.536462  0.536532  0.536615  0.536714  0.536832  0.536971  0.537133  0.537322  0.537541  0.537794      0.538084 | -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201869  -0.201869  -0.201869   -0.20187  -0.201871  -0.201872  -0.201873  -0.201875  -0.201877  -0.201881  -0.201885   -0.20189  -0.201897  -0.201906  -0.201917   -0.20193  -0.201948   -0.20197  -0.201997   -0.20203  -0.202071  -0.202121  -0.202181  -0.202254  -0.202341  -0.202446  -0.202569  -0.202715  -0.202887  -0.203087   -0.20332  -0.203589  -0.203898     -0.204251 | 0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571  0.665711  0.665711  0.665712  0.665712  0.665713  0.665714  0.665716  0.665718   0.66572  0.665723  0.665726   0.66573  0.665735  0.665741  0.665748  0.665757  0.665767  0.665779  0.665793  0.665809  0.665829  0.665851  0.665877  0.665906       0.66594 | 100.000001 | 100.000001  100.000002  100.000002  100.000004  100.000006  100.000008  100.000012  100.000018  100.000026  100.000038  100.000055  100.000079  100.000113   100.00016  100.000226  100.000318  100.000444  100.000616  100.000851   100.00117  100.001599  100.002173  100.002937  100.003949   100.00528  100.007023   100.00929  100.012224     100.016   100.02083  100.026976  100.034751  100.044534  100.056772  100.071999   100.09084  100.114025  100.142398  100.176934  100.218743  100.269089  100.329391  100.401238  100.486395  100.586807    100.7046  100.842086  101.001755  101.186273          101.398469 | 101.511047 |   0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207  0.720701  0.720701  0.720702  0.720703  0.720704  0.720705  0.720707   0.72071  0.720713  0.720717  0.720723   0.72073  0.720739  0.720751  0.720766  0.720784  0.720808  0.720837  0.720873  0.720917  0.720971  0.721036  0.721116  0.721212  0.721326  0.721463  0.721624  0.721815  0.722038  0.722298    0.7226  0.722949       0.723349 | 2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591117  2.591117  2.591117  2.591117  2.591117  2.591118  2.591119   2.59112  2.591121  2.591123  2.591126  2.591129  2.591134   2.59114  2.591149   2.59116  2.591174  2.591194  2.591219  2.591251  2.591292  2.591346  2.591413  2.591499  2.591607  2.591741  2.591909  2.592116  2.592371  2.592683  2.593063  2.593523  2.594076  2.594739  2.595528  2.596463  2.597565  2.598857  2.600363   2.60211  2.604128        2.606445 | 100.000001  100.000001  100.000002  100.000003  100.000004  100.000007   100.00001  100.000014  100.000021  100.000031  100.000045  100.000065  100.000093  100.000132  100.000187  100.000264   100.00037  100.000515  100.000714  100.000985   100.00135  100.001841  100.002497  100.003367  100.004517  100.006026  100.007997  100.010555  100.013857  100.018096  100.023507  100.030374   100.03904  100.049918  100.063493  100.080343  100.101141  100.126671  100.157841  100.195689    100.2414  100.296311  100.361925  100.439918  100.532143  100.640634  100.767612  100.915478  101.086812  101.284365  101.511047 |  100.0 |   0.000001 | 101.511047 |
    | 14:00 | 3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717829  3.717829  3.717829   3.71783   3.71783  3.717831  3.717833  3.717834  3.717837   3.71784  3.717845  3.717851   3.71786  3.717871  3.717885  3.717905  3.717931  3.717964  3.718007  3.718063  3.718135  3.718226  3.718342  3.718489  3.718672  3.718901  3.719185  3.719536  3.719966  3.720491  3.721129  3.721898  3.722822  3.723926  3.725238  3.726788  3.728611  3.730744  3.733226  3.736101             3.739413 | 124.878653  124.878653  124.878654  124.878655  124.878656  124.878658  124.878661  124.878665  124.878671   124.87868  124.878692   124.87871  124.878735  124.878771   124.87882  124.878889  124.878983  124.879113   124.87929  124.879531  124.879856  124.880292  124.880873  124.881645  124.882663  124.883999  124.885742  124.888003   124.89092  124.894661  124.899434  124.905488  124.913125  124.922705  124.934657  124.949486  124.967784   124.99024  125.017648  125.050922    125.0911  125.139359  125.197017  125.265545  125.346569  125.441878  125.553419  125.683297  125.833776  126.007264 |  52.914557   52.914557   52.914557   52.914557   52.914557   52.914558   52.914558   52.914559    52.91456   52.914562   52.914564   52.914568   52.914573    52.91458   52.914589   52.914603   52.914621   52.914646   52.914681   52.914728   52.914792   52.914877   52.914991   52.915141    52.91534   52.915601   52.915942   52.916384   52.916954   52.917685   52.918618   52.919801   52.921293   52.923166   52.925501   52.928399   52.931974   52.936361   52.941716   52.948215   52.956062   52.965486   52.976743   52.990119   53.005931   53.024523   53.046274   53.071589   53.100905        53.134684 |  52.178281   52.178282   52.178282   52.178282   52.178282   52.178282   52.178283   52.178284   52.178285   52.178287   52.178289   52.178292   52.178297   52.178304   52.178314   52.178327   52.178345    52.17837   52.178404    52.17845   52.178512   52.178596   52.178707   52.178855    52.17905   52.179306    52.17964   52.180073   52.180632   52.181349   52.182264   52.183424   52.184888   52.186723   52.189014   52.191855   52.195361   52.199663   52.204913   52.211287   52.218981   52.228222    52.23926   52.252377   52.267881   52.286113   52.307441   52.332265   52.361012     52.394134 | 1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088386  1.088386  1.088386  1.088386  1.088387  1.088388  1.088389   1.08839  1.088392  1.088394  1.088397  1.088401  1.088406  1.088413  1.088422  1.088433  1.088447  1.088465  1.088487  1.088516  1.088551  1.088596   1.08865  1.088717  1.088798  1.088897  1.089016  1.089159  1.089331  1.089534  1.089774  1.090057  1.090387  1.090772  1.091217  1.091729 | 1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916  1.359159  1.359159  1.359158  1.359157  1.359156  1.359155  1.359154  1.359152   1.35915  1.359147  1.359143  1.359139  1.359133  1.359127  1.359119  1.359109  1.359098  1.359084  1.359069   1.35905  1.359028  1.359003  1.358974           1.35894 | 0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536159  0.536159  0.536159  0.536159  0.536159   0.53616   0.53616  0.536161  0.536162  0.536163  0.536164  0.536166  0.536169  0.536172  0.536177  0.536183   0.53619  0.536199   0.53621  0.536225  0.536243  0.536265  0.536293  0.536327  0.536368  0.536419  0.536479  0.536552  0.536639  0.536742  0.536864  0.537007  0.537175   0.53737  0.537595      0.537854 | -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201869  -0.201869   -0.20187   -0.20187  -0.201871  -0.201872  -0.201874  -0.201876  -0.201878  -0.201882  -0.201886  -0.201892  -0.201899  -0.201909  -0.201921  -0.201935  -0.201954  -0.201977  -0.202006  -0.202041  -0.202084  -0.202136  -0.202199  -0.202275  -0.202366  -0.202475  -0.202603  -0.202755  -0.202932  -0.203138  -0.203378  -0.203653      -0.20397 | 0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571  0.665711  0.665711  0.665711  0.665712  0.665713  0.665714  0.665715  0.665716  0.665718  0.665721  0.665723  0.665727  0.665731  0.665736  0.665743   0.66575  0.665759   0.66577  0.665782  0.665797  0.665814  0.665834  0.665857  0.665883      0.665913 | 100.000001 | 100.000001  100.000001  100.000002  100.000003  100.000004  100.000007   100.00001  100.000014  100.000021   100.00003  100.000044  100.000063  100.000091  100.000129  100.000183  100.000257   100.00036  100.000502  100.000695  100.000957   100.00131  100.001785  100.002418  100.003258  100.004366   100.00582  100.007717  100.010178  100.013352  100.017425  100.022619  100.029209  100.037521   100.04795   100.06096  100.077103  100.097022  100.121469  100.151311  100.187541  100.231294  100.283852  100.346656  100.421312  100.509602  100.613481  100.735085  100.876729  101.040901          101.230259 | 101.330846 |   0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207  0.720701  0.720701  0.720701  0.720702  0.720703  0.720704  0.720706  0.720708  0.720711  0.720714  0.720719  0.720725  0.720733  0.720743  0.720755  0.720771  0.720791  0.720816  0.720846  0.720884  0.720931  0.720988  0.721057   0.72114   0.72124  0.721359  0.721501  0.721668  0.721865  0.722096  0.722364  0.722674       0.723032 | 2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591117  2.591117  2.591117  2.591117  2.591118  2.591118  2.591119   2.59112  2.591122  2.591124  2.591127  2.591131  2.591136  2.591143  2.591152  2.591164   2.59118  2.591201  2.591228  2.591263  2.591308  2.591365  2.591438  2.591529  2.591644  2.591787  2.591965  2.592184  2.592453  2.592781   2.59318  2.593661  2.594239  2.594929  2.595749  2.596718  2.597858  2.599191  2.600742  2.602539        2.604608 | 100.000001  100.000001  100.000002  100.000002  100.000004  100.000005  100.000008  100.000012  100.000017  100.000025  100.000036  100.000052  100.000075  100.000107  100.000151  100.000214    100.0003  100.000419  100.000582  100.000804  100.001105   100.00151  100.002053  100.002774   100.00373  100.004988  100.006634  100.008776  100.011549  100.015117  100.019683  100.025494  100.032848  100.042102  100.053683  100.068097  100.085938  100.107901  100.134791  100.167536  100.207197  100.254979  100.312242  100.380508  100.461467  100.556988  100.669117  100.800079  100.952276  101.128286  101.330846 |  100.0 |   0.000001 | 101.330846 |
    | 14:30 | 3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717829  3.717829  3.717829  3.717829   3.71783  3.717831  3.717832  3.717833  3.717835  3.717838  3.717842  3.717847  3.717854  3.717863  3.717875  3.717892  3.717913  3.717941  3.717977  3.718024  3.718085  3.718162   3.71826  3.718384   3.71854  3.718736  3.718979   3.71928   3.71965  3.720103  3.720654  3.721322  3.722127   3.72309  3.724239    3.7256  3.727206   3.72909   3.73129  3.733845             3.736798 | 124.878653  124.878653  124.878653  124.878654  124.878655  124.878657  124.878659  124.878662  124.878667  124.878674  124.878684  124.878699  124.878719  124.878748  124.878788  124.878844  124.878921  124.879027  124.879173   124.87937  124.879638  124.879998  124.880479  124.881119  124.881966   124.88308  124.884536   124.88643   124.88888   124.89203  124.896059  124.901183  124.907665  124.915817  124.926016  124.938704  124.954404  124.973725  124.997375  125.026167  125.061033  125.103032  125.153358  125.213346  125.284483  125.368408  125.466918  125.581966  125.715659  125.870256 |  52.914556   52.914557   52.914557   52.914557   52.914557   52.914557   52.914558   52.914558   52.914559   52.914561   52.914563   52.914566   52.914569   52.914575   52.914583   52.914594   52.914609    52.91463   52.914658   52.914697   52.914749   52.914819   52.914913   52.915039   52.915204   52.915422   52.915706   52.916077   52.916555   52.917171   52.917958    52.91896   52.920226    52.92182   52.923813   52.926292    52.92936   52.933135   52.937755    52.94338    52.95019   52.958393   52.968219   52.979931   52.993815   53.010191   53.029407   53.051839   53.077895         53.10801 |  52.178281   52.178281   52.178282   52.178282   52.178282   52.178282   52.178283   52.178283   52.178284   52.178286   52.178288    52.17829   52.178294     52.1783   52.178307   52.178318   52.178333   52.178353   52.178381   52.178419    52.17847   52.178539   52.178631   52.178754   52.178916    52.17913   52.179409   52.179772   52.180241   52.180845   52.181617   52.182599   52.183841   52.185404   52.187358   52.189789   52.192797   52.196499    52.20103   52.206545   52.213223   52.221266   52.230902   52.242386   52.256001   52.272059   52.290901   52.312898   52.338448     52.367978 | 1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088386  1.088386  1.088386  1.088387  1.088387  1.088388  1.088389   1.08839  1.088392  1.088395  1.088398  1.088403  1.088408  1.088415  1.088425  1.088437  1.088452  1.088471  1.088496  1.088526  1.088564   1.08861  1.088668  1.088738  1.088823  1.088927  1.089052  1.089201  1.089379   1.08959  1.089839  1.090131  1.090472  1.090867  1.091325 | 1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916  1.359159  1.359159  1.359159  1.359158  1.359157  1.359156  1.359155  1.359153  1.359151  1.359149  1.359146  1.359142  1.359137  1.359132  1.359125  1.359116  1.359107  1.359095  1.359081  1.359064  1.359045  1.359023  1.358997          1.358966 | 0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536159  0.536159  0.536159  0.536159  0.536159  0.536159   0.53616   0.53616  0.536161  0.536162  0.536163  0.536165  0.536167   0.53617  0.536174  0.536179  0.536185  0.536192  0.536202  0.536214   0.53623  0.536249  0.536273  0.536302  0.536338  0.536381  0.536434  0.536497  0.536573  0.536663  0.536771  0.536897  0.537045  0.537218  0.537418       0.53765 | -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201869  -0.201869  -0.201869   -0.20187   -0.20187  -0.201871  -0.201873  -0.201874  -0.201877  -0.201879  -0.201883  -0.201888  -0.201894  -0.201902  -0.201912  -0.201925  -0.201941   -0.20196  -0.201985  -0.202015  -0.202052  -0.202097  -0.202152  -0.202218  -0.202298  -0.202393  -0.202505  -0.202638  -0.202795  -0.202978   -0.20319  -0.203436     -0.203719 | 0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571  0.665711  0.665711  0.665712  0.665712  0.665713  0.665714  0.665715  0.665717  0.665719  0.665721  0.665724  0.665728  0.665732  0.665738  0.665744  0.665752  0.665761  0.665772  0.665785    0.6658  0.665818  0.665838  0.665862      0.665889 | 100.000001 | 100.000001  100.000001  100.000002  100.000002  100.000003  100.000005  100.000008  100.000011  100.000017  100.000024  100.000035  100.000051  100.000073  100.000104  100.000148  100.000209  100.000293  100.000408  100.000567  100.000782  100.001073  100.001465  100.001989  100.002685  100.003607  100.004819  100.006404  100.008466  100.011132  100.014561  100.018946  100.024524  100.031578  100.040453  100.051554  100.065366  100.082457   100.10349  100.129238  100.160586  100.198551  100.244288  100.299099  100.364444  100.441947  100.533402  100.640777  100.766214   100.91203          101.080711 | 101.170423 |   0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207  0.720701  0.720701  0.720702  0.720702  0.720703  0.720705  0.720707  0.720709  0.720712  0.720716  0.720721  0.720727  0.720736  0.720746   0.72076  0.720777  0.720798  0.720824  0.720857  0.720897  0.720946  0.721005  0.721077  0.721164  0.721269  0.721393   0.72154  0.721713  0.721917  0.722155  0.722431       0.722749 | 2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591117  2.591117  2.591117  2.591117  2.591117  2.591118  2.591119  2.591119  2.591121  2.591122  2.591125  2.591128  2.591132  2.591138  2.591146  2.591156  2.591169  2.591187  2.591209  2.591239  2.591277  2.591325  2.591386  2.591464  2.591562  2.591684  2.591836  2.592024  2.592255  2.592539  2.592883  2.593301  2.593804  2.594406  2.595124  2.595975  2.596979  2.598157  2.599532  2.601129        2.602974 | 100.000001  100.000001  100.000001  100.000002  100.000003  100.000004  100.000006  100.000009  100.000014   100.00002  100.000029  100.000042   100.00006  100.000086  100.000122  100.000173  100.000243  100.000341  100.000474  100.000657  100.000904  100.001238  100.001686  100.002284  100.003077  100.004124  100.005498   100.00729  100.009615  100.012615  100.016464  100.021376  100.027607   100.03547  100.045336   100.05765  100.072934  100.091801  100.114965  100.143253  100.177613  100.219128  100.269023  100.328678  100.399633  100.483595  100.582444  100.698235  100.833199  100.989739  101.170423 |  100.0 |   0.000001 | 101.170423 |
    | 15:00 | 3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717829  3.717829  3.717829   3.71783   3.71783  3.717831  3.717832  3.717834  3.717836  3.717839  3.717844  3.717849  3.717857  3.717867  3.717881  3.717898  3.717922  3.717952  3.717991  3.718042  3.718107   3.71819  3.718296  3.718428  3.718595  3.718803   3.71906  3.719378  3.719769  3.720245  3.720824  3.721523  3.722363  3.723367   3.72456  3.725972  3.727634   3.72958  3.731847             3.734474 | 124.878652  124.878653  124.878653  124.878654  124.878655  124.878656  124.878658   124.87866  124.878664   124.87867  124.878678   124.87869  124.878706  124.878729  124.878762  124.878807   124.87887  124.878957  124.879076  124.879238  124.879459  124.879755  124.880153  124.880684  124.881387  124.882314   124.88353  124.885115   124.88717   124.88982  124.893217  124.897549  124.903043  124.909972  124.918663  124.929505  124.942957  124.959557  124.979934  125.004811  125.035021  125.071517  125.115373  125.167802  125.230155  125.303933  125.390787  125.492521  125.611093  125.748611 |  52.914556   52.914557   52.914557   52.914557   52.914557   52.914557   52.914557   52.914558   52.914559    52.91456   52.914561   52.914564   52.914567   52.914571   52.914578   52.914587   52.914599   52.914616   52.914639   52.914671   52.914714   52.914772    52.91485   52.914953   52.915091   52.915272    52.91551    52.91582   52.916221   52.916739   52.917403    52.91825   52.919323   52.920677   52.922376   52.924494   52.927123   52.930367   52.934348   52.939208   52.945109   52.952238   52.960802    52.97104   52.983212   52.997611   53.014557     53.0344   53.057517        53.084316 |  52.178281   52.178281   52.178282   52.178282   52.178282   52.178282   52.178282   52.178283   52.178284   52.178285   52.178286   52.178289   52.178292   52.178296   52.178302   52.178311   52.178323    52.17834   52.178363   52.178394   52.178436   52.178493   52.178569   52.178671   52.178806   52.178983   52.179216    52.17952   52.179914   52.180422   52.181073   52.181903   52.182956   52.184283   52.185949   52.188026   52.190604   52.193785   52.197688   52.202454   52.208241   52.215231   52.223629   52.233668   52.245604   52.259723    52.27634   52.295797   52.318466     52.344744 | 1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088386  1.088386  1.088386  1.088386  1.088387  1.088387  1.088388   1.08839  1.088391  1.088393  1.088396    1.0884  1.088404   1.08841  1.088418  1.088428  1.088441  1.088458  1.088478  1.088504  1.088536  1.088576  1.088625  1.088686   1.08876   1.08885  1.088958  1.089088  1.089244  1.089429  1.089648  1.089905  1.090207  1.090558  1.090965 | 1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916  1.359159  1.359159  1.359158  1.359158  1.359157  1.359156  1.359154  1.359153  1.359151  1.359148  1.359145  1.359141  1.359136   1.35913  1.359123  1.359114  1.359104  1.359091  1.359077   1.35906   1.35904  1.359017           1.35899 | 0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536159  0.536159  0.536159  0.536159  0.536159   0.53616   0.53616  0.536161  0.536161  0.536162  0.536164  0.536166  0.536168  0.536171  0.536175   0.53618  0.536187  0.536195  0.536206  0.536219  0.536235  0.536256  0.536281  0.536311  0.536349  0.536395   0.53645  0.536516  0.536595  0.536689    0.5368  0.536931  0.537083  0.537261      0.537468 | -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201869  -0.201869  -0.201869   -0.20187  -0.201871  -0.201872  -0.201873  -0.201875  -0.201877  -0.201881  -0.201885   -0.20189  -0.201897  -0.201905  -0.201916  -0.201929  -0.201946  -0.201967  -0.201993  -0.202025  -0.202064  -0.202111  -0.202169  -0.202238  -0.202321  -0.202419  -0.202536  -0.202674  -0.202836  -0.203024  -0.203243     -0.203496 | 0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571  0.665711  0.665711  0.665712  0.665712  0.665713  0.665714  0.665716  0.665717   0.66572  0.665722  0.665725  0.665729  0.665734  0.665739  0.665746  0.665754  0.665764  0.665775  0.665789  0.665804  0.665822  0.665843      0.665868 |      100.0 | 100.000001  100.000001  100.000001  100.000002  100.000003  100.000004  100.000006  100.000009  100.000013  100.000019  100.000028  100.000041  100.000059  100.000084   100.00012  100.000169  100.000237  100.000332  100.000462  100.000638  100.000878  100.001201  100.001634  100.002211  100.002977  100.003986  100.005309  100.007035  100.009271  100.012155  100.015853  100.020568  100.026548   100.03409   100.04355  100.055352  100.069995  100.088066  100.110249  100.137333  100.170227  100.209967  100.257728  100.314832  100.382756   100.46314  100.557792  100.668688  100.797978          100.947978 | 101.027852 |   0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207  0.720701  0.720701  0.720701  0.720702  0.720703  0.720704  0.720705  0.720707   0.72071  0.720713  0.720717  0.720723   0.72073  0.720739   0.72075  0.720765  0.720783  0.720805  0.720833  0.720867   0.72091  0.720961  0.721024  0.721099   0.72119  0.721299  0.721428   0.72158   0.72176   0.72197  0.722215       0.722498 | 2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591117  2.591117  2.591117  2.591117  2.591118  2.591118  2.591119   2.59112  2.591121  2.591123  2.591126  2.591129  2.591134  2.591141  2.591149   2.59116  2.591175  2.591194  2.591218   2.59125  2.591291  2.591343  2.591409  2.591492  2.591596  2.591726  2.591887  2.592086   2.59233  2.592628  2.592989  2.593427  2.593952  2.594579  2.595325  2.596208  2.597247  2.598463   2.59988        2.601522 |      100.0  100.000001  100.000001  100.000001  100.000002  100.000003  100.000005  100.000007  100.000011  100.000016  100.000023  100.000033  100.000048  100.000069  100.000099   100.00014  100.000197  100.000277  100.000386  100.000535  100.000739  100.001014  100.001384  100.001879  100.002537  100.003407  100.004552   100.00605  100.007998  100.010517  100.013758  100.017903  100.023177   100.02985  100.038245   100.04875  100.061825   100.07801  100.097936  100.122338  100.152061  100.188076  100.231484  100.283533  100.345619  100.419299  100.506297  100.608503  100.727981  100.866963  101.027852 |  100.0 |        0.0 | 101.027852 |
    | 15:30 | 3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717829  3.717829  3.717829  3.717829   3.71783  3.717831  3.717832  3.717833  3.717835  3.717837  3.717841  3.717845  3.717852   3.71786  3.717871  3.717886  3.717906  3.717931  3.717964  3.718007  3.718061  3.718132  3.718221  3.718333  3.718475  3.718652  3.718873  3.719146  3.719482  3.719893  3.720394     3.721  3.721731  3.722607  3.723652  3.724891  3.726354  3.728072   3.73008             3.732414 | 124.878652  124.878653  124.878653  124.878653  124.878654  124.878655  124.878656  124.878659  124.878662  124.878666  124.878673  124.878682  124.878695  124.878714  124.878741  124.878778  124.878829    124.8789  124.878997   124.87913  124.879311  124.879556  124.879885  124.880324  124.880908  124.881679  124.882693  124.884017  124.885739  124.887965  124.890827  124.894485  124.899136  124.905017  124.912413  124.921665  124.933176  124.947419   124.96495  124.986413  125.012551  125.044216  125.082375  125.128124  125.182691  125.247442  125.323891    125.4137  125.518679  125.640791 |  52.914556   52.914556   52.914557   52.914557   52.914557   52.914557   52.914557   52.914558   52.914558   52.914559    52.91456   52.914562   52.914565   52.914569   52.914574   52.914581   52.914591   52.914605   52.914624    52.91465   52.914685   52.914733   52.914797   52.914883   52.914997   52.915148   52.915346   52.915605   52.915942   52.916377   52.916936   52.917651    52.91856   52.919709   52.921154   52.922963   52.925212   52.927995    52.93142   52.935614    52.94072   52.946905   52.954358   52.963292   52.973946   52.986586   53.001506   53.019027   53.039501        53.063306 |  52.178281   52.178281   52.178281   52.178282   52.178282   52.178282   52.178282   52.178283   52.178283   52.178284   52.178285   52.178287    52.17829   52.178293   52.178298   52.178305   52.178315   52.178329   52.178347   52.178373   52.178408   52.178455   52.178518   52.178602   52.178714   52.178861   52.179056    52.17931    52.17964   52.180066   52.180615   52.181316   52.182207   52.183334   52.184751   52.186524    52.18873   52.191459   52.194818    52.19893   52.203937   52.210002    52.21731   52.226071   52.236518   52.248912   52.263542   52.280723   52.300799     52.324142 | 1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088386  1.088386  1.088386  1.088386  1.088387  1.088388  1.088389   1.08839  1.088392  1.088394  1.088397  1.088401  1.088406  1.088413  1.088421  1.088432  1.088446  1.088463  1.088485  1.088513  1.088547  1.088589  1.088641  1.088705  1.088783  1.088877   1.08899  1.089126  1.089288   1.08948  1.089707  1.089973  1.090284  1.090646 | 1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916  1.359159  1.359159  1.359158  1.359157  1.359157  1.359155  1.359154  1.359152   1.35915  1.359147  1.359144  1.359139  1.359134  1.359128   1.35912  1.359111  1.359101  1.359088  1.359073  1.359055  1.359035          1.359011 | 0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536159  0.536159  0.536159  0.536159  0.536159   0.53616   0.53616  0.536161  0.536162  0.536163  0.536164  0.536166  0.536169  0.536172  0.536177  0.536182  0.536189  0.536198  0.536209  0.536223  0.536241  0.536262  0.536289  0.536321  0.536361  0.536408  0.536466  0.536535  0.536617  0.536715   0.53683  0.536965  0.537123      0.537306 | -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201869  -0.201869   -0.20187   -0.20187  -0.201871  -0.201872  -0.201874  -0.201876  -0.201878  -0.201882  -0.201886  -0.201892  -0.201899  -0.201908   -0.20192  -0.201934  -0.201952  -0.201974  -0.202001  -0.202035  -0.202076  -0.202126  -0.202186  -0.202258  -0.202344  -0.202447  -0.202568  -0.202711  -0.202878  -0.203072     -0.203297 | 0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571  0.665711  0.665711  0.665711  0.665712  0.665713  0.665714  0.665715  0.665716  0.665718   0.66572  0.665723  0.665726   0.66573  0.665735  0.665741  0.665748  0.665757  0.665767  0.665778  0.665792  0.665808  0.665827      0.665848 |      100.0 |      100.0  100.000001  100.000001  100.000001  100.000002  100.000003  100.000005  100.000007  100.000011  100.000016  100.000023  100.000033  100.000047  100.000068  100.000096  100.000137  100.000193   100.00027  100.000376  100.000521  100.000718  100.000984  100.001342   100.00182  100.002455  100.003294  100.004398   100.00584  100.007714  100.010137  100.013251  100.017232  100.022295  100.028696  100.036747  100.046818  100.059348  100.074853  100.093937  100.117303  100.145761  100.180239  100.221792  100.271616   100.33105  100.401589  100.484888  100.582765  100.697207          100.830367 | 100.901359 |   0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207  0.720701  0.720701  0.720702  0.720702  0.720703  0.720704  0.720706  0.720708  0.720711  0.720714  0.720719  0.720725  0.720732  0.720742  0.720754   0.72077  0.720789  0.720813  0.720842  0.720878  0.720923  0.720977  0.721043  0.721122  0.721216  0.721329  0.721463  0.721621  0.721807  0.722024       0.722276 | 2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591117  2.591117  2.591117  2.591117  2.591118  2.591118  2.591119   2.59112  2.591122  2.591124  2.591127  2.591131  2.591136  2.591143  2.591153  2.591165  2.591181  2.591201  2.591228  2.591262  2.591306  2.591362  2.591432  2.591521  2.591632   2.59177   2.59194   2.59215  2.592407   2.59272    2.5931  2.593557  2.594104  2.594757  2.595532  2.596447  2.597521  2.598776        2.600234 |      100.0       100.0  100.000001  100.000001  100.000002  100.000003  100.000004  100.000006  100.000009  100.000013  100.000018  100.000027  100.000039  100.000056   100.00008  100.000113   100.00016  100.000224  100.000314  100.000436  100.000603  100.000829  100.001135  100.001544  100.002089  100.002812  100.003766  100.005016  100.006646  100.008759  100.011485   100.01498  100.019438  100.025093  100.032227  100.041178  100.052348  100.066213   100.08333  100.104349  100.130024   100.16122  100.198928  100.244269  100.298509  100.363064  100.439505   100.52957   100.63516  100.758344  100.901359 |  100.0 |        0.0 | 100.901359 |
    | 16:00 | 3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717829  3.717829  3.717829   3.71783   3.71783  3.717831  3.717832  3.717834  3.717836  3.717838  3.717842  3.717848  3.717854  3.717864  3.717876  3.717892  3.717913  3.717941  3.717977  3.718023  3.718082  3.718157  3.718253  3.718373  3.718525  3.718713  3.718947  3.719235   3.71959  3.720022  3.720548  3.721183  3.721946  3.722859  3.723945  3.725231  3.726746  3.728521             3.730591 | 124.878652  124.878652  124.878653  124.878653  124.878654  124.878654  124.878656  124.878657   124.87866  124.878663  124.878669  124.878676  124.878687  124.878702  124.878724  124.878753  124.878795  124.878853  124.878933  124.879042  124.879191  124.879392  124.879663  124.880026   124.88051  124.881151  124.881996  124.883102  124.884543  124.886411  124.888818  124.891903  124.895836  124.900822  124.907109  124.914994  124.924829  124.937034  124.952096  124.970587  124.993168    125.0206  125.053753  125.093611  125.141287  125.198024  125.265205  125.344354  125.437142  125.545386 |  52.914556   52.914556   52.914557   52.914557   52.914557   52.914557   52.914557   52.914557   52.914558   52.914559    52.91456   52.914561   52.914563   52.914566    52.91457   52.914576   52.914584   52.914596   52.914611   52.914633   52.914662   52.914701   52.914754   52.914825    52.91492   52.915045    52.91521   52.915426   52.915708   52.916073   52.916543   52.917146   52.917915   52.918889   52.920118   52.921659   52.923581   52.925966   52.928909   52.932522   52.936933   52.942292   52.948768   52.956553   52.965863    52.97694   52.990053   53.005498   53.023599        53.044708 |  52.178281   52.178281   52.178281   52.178282   52.178282   52.178282   52.178282   52.178282   52.178283   52.178283   52.178284   52.178286   52.178288   52.178291   52.178295   52.178301   52.178309    52.17832   52.178335   52.178356   52.178385   52.178423   52.178475   52.178545   52.178637    52.17876   52.178922   52.179134    52.17941   52.179768    52.18023   52.180821   52.181575    52.18253   52.183735   52.185246    52.18713   52.189469   52.192355   52.195898   52.200224   52.205479   52.211829   52.219462   52.228591   52.239453   52.252312   52.267457   52.285207     52.305905 | 1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088386  1.088386  1.088386  1.088387  1.088387  1.088388  1.088389  1.088391  1.088392  1.088395  1.088398  1.088403  1.088408  1.088415  1.088424  1.088436  1.088451   1.08847  1.088493  1.088522  1.088559  1.088603  1.088658  1.088725  1.088807  1.088905  1.089024  1.089165  1.089334  1.089533  1.089768  1.090043  1.090363 | 1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916  1.359159  1.359159  1.359159  1.359158  1.359157  1.359156  1.359155  1.359153  1.359151  1.359149  1.359146  1.359142  1.359138  1.359133  1.359126  1.359118  1.359109  1.359098  1.359085  1.359069  1.359051           1.35903 | 0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536159  0.536159  0.536159  0.536159  0.536159  0.536159   0.53616   0.53616  0.536161  0.536162  0.536163  0.536165  0.536167   0.53617  0.536174  0.536178  0.536184  0.536192  0.536201  0.536213  0.536228  0.536247  0.536269  0.536297  0.536331  0.536373  0.536423  0.536483  0.536555   0.53664  0.536742  0.536861     0.537      0.537163 | -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201869  -0.201869  -0.201869   -0.20187  -0.201871  -0.201871  -0.201873  -0.201874  -0.201877  -0.201879  -0.201883  -0.201888  -0.201894  -0.201902  -0.201911  -0.201924  -0.201939  -0.201958  -0.201981   -0.20201  -0.202046  -0.202089  -0.202141  -0.202204  -0.202279  -0.202369  -0.202475  -0.202601  -0.202748  -0.202921     -0.203121 | 0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571  0.665711  0.665711  0.665712  0.665712  0.665713  0.665714  0.665715  0.665717  0.665719  0.665721  0.665724  0.665727  0.665732  0.665737  0.665743   0.66575  0.665759  0.665769  0.665782  0.665796  0.665812      0.665832 |      100.0 |      100.0       100.0  100.000001  100.000001  100.000002  100.000003  100.000004  100.000006  100.000008  100.000012  100.000018  100.000026  100.000038  100.000055  100.000078   100.00011  100.000156  100.000219  100.000306  100.000424  100.000586  100.000805  100.001101  100.001496  100.002023   100.00272  100.003639  100.004843  100.006412  100.008445  100.011065  100.014423  100.018703   100.02413  100.030973  100.039556  100.050262  100.063547  100.079943  100.100073  100.124658  100.154525  100.190623  100.234028  100.285952  100.347753  100.420942  100.507187  100.608317          100.726327 | 100.789318 |   0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207  0.720701  0.720701  0.720702  0.720702  0.720703  0.720705  0.720707  0.720709  0.720712  0.720716  0.720721  0.720727  0.720735  0.720746  0.720759  0.720775  0.720795  0.720821  0.720852   0.72089  0.720937  0.720994  0.721062  0.721145  0.721244  0.721361    0.7215  0.721664  0.721855       0.722079 | 2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591117  2.591117  2.591117  2.591117  2.591117  2.591118  2.591119   2.59112  2.591121  2.591123  2.591125  2.591128  2.591133  2.591139  2.591146  2.591156   2.59117  2.591187  2.591209  2.591238  2.591275  2.591322  2.591382  2.591457  2.591552   2.59167  2.591816  2.591996  2.592218  2.592488  2.592817  2.593214  2.593691  2.594262  2.594941  2.595745  2.596691  2.597801        2.599095 |      100.0       100.0  100.000001  100.000001  100.000001  100.000002  100.000003  100.000005  100.000007   100.00001  100.000015  100.000021  100.000031  100.000045  100.000064  100.000091  100.000129  100.000182  100.000255  100.000355  100.000492  100.000678   100.00093  100.001267  100.001719  100.002319  100.003112  100.004154  100.005517  100.007288  100.009577  100.012521  100.016285  100.021072  100.027126  100.034743  100.044274  100.056135  100.070818  100.088899  100.111044  100.138028  100.170734  100.210172  100.257484  100.313953  100.381012  100.460248  100.553409  100.662408  100.789318 |  100.0 |        0.0 | 100.789318 |
    | 16:30 | 3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717829  3.717829  3.717829  3.717829   3.71783   3.71783  3.717831  3.717833  3.717834  3.717837   3.71784  3.717844   3.71785  3.717857  3.717868  3.717881  3.717899  3.717922  3.717952  3.717991   3.71804  3.718104  3.718185  3.718287  3.718416  3.718577  3.718777  3.719024  3.719329  3.719703  3.720157  3.720708  3.721372  3.722169  3.723119  3.724247   3.72558  3.727147             3.728979 | 124.878652  124.878652  124.878653  124.878653  124.878653  124.878654  124.878655  124.878656  124.878658  124.878661  124.878665  124.878671   124.87868  124.878692   124.87871  124.878734  124.878768  124.878815   124.87888   124.87897  124.879092  124.879257  124.879481  124.879781  124.880182  124.880714  124.881416  124.882339  124.883544   124.88511  124.887133  124.889732  124.893053  124.897274  124.902612  124.909323  124.917717  124.928159  124.941083  124.956991  124.976472  125.000202  125.028962  125.063635  125.105226  125.154862  125.213803  125.283442  125.365317  125.461107 |  52.914556   52.914556   52.914556   52.914557   52.914557   52.914557   52.914557   52.914557   52.914558   52.914558   52.914559    52.91456   52.914562   52.914564   52.914568   52.914572   52.914579   52.914588   52.914601   52.914618   52.914642   52.914675   52.914718   52.914777   52.914855   52.914959   52.915097   52.915277   52.915513   52.915819   52.916214   52.916722   52.917371   52.918196   52.919239   52.920551   52.922191   52.924232   52.926757   52.929865   52.933671   52.938308   52.943926   52.950698   52.958821   52.968513    52.98002   52.993612   53.009588        53.028273 |  52.178281   52.178281   52.178281   52.178281   52.178282   52.178282   52.178282   52.178282   52.178282   52.178283   52.178284   52.178285   52.178287   52.178289   52.178292   52.178297   52.178304   52.178313   52.178325   52.178342   52.178366   52.178397    52.17844   52.178498   52.178574   52.178676   52.178811   52.178988   52.179219   52.179519   52.179907   52.180405   52.181041    52.18185   52.182873   52.184159   52.185768   52.187769   52.190245   52.193293   52.197025   52.201571    52.20708   52.213721   52.221686    52.23119   52.242474   52.255802   52.271468      52.28979 | 1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088386  1.088386  1.088386  1.088386  1.088387  1.088388  1.088388   1.08839  1.088391  1.088393  1.088396    1.0884  1.088404   1.08841  1.088418  1.088428   1.08844  1.088456  1.088476  1.088501  1.088532  1.088571  1.088618  1.088676  1.088746  1.088832  1.088935  1.089058  1.089205   1.08938  1.089587   1.08983  1.090114 | 1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916  1.359159  1.359159  1.359158  1.359158  1.359157  1.359156  1.359155  1.359153  1.359151  1.359148  1.359145  1.359141  1.359137  1.359131  1.359124  1.359116  1.359106  1.359095  1.359081  1.359065          1.359046 | 0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536159  0.536159  0.536159  0.536159  0.536159   0.53616   0.53616  0.536161  0.536161  0.536162  0.536164  0.536166  0.536168  0.536171  0.536175   0.53618  0.536186  0.536195  0.536205  0.536217  0.536233  0.536253  0.536277  0.536306  0.536342  0.536385  0.536438    0.5365  0.536575  0.536664  0.536769  0.536892      0.537036 | -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201869  -0.201869  -0.201869   -0.20187  -0.201871  -0.201872  -0.201873  -0.201875  -0.201877  -0.201881  -0.201885   -0.20189  -0.201896  -0.201904  -0.201915  -0.201928  -0.201944  -0.201964  -0.201989   -0.20202  -0.202057  -0.202102  -0.202157  -0.202222  -0.202301  -0.202394  -0.202505  -0.202635  -0.202787     -0.202964 | 0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571  0.665711  0.665711  0.665712  0.665712  0.665713  0.665714  0.665716  0.665717  0.665719  0.665722  0.665725  0.665728  0.665733  0.665738  0.665745  0.665752  0.665761  0.665772  0.665785  0.665799      0.665816 |      100.0 |      100.0       100.0  100.000001  100.000001  100.000001  100.000002  100.000003  100.000005  100.000007   100.00001  100.000015  100.000021  100.000031  100.000044  100.000063  100.000089  100.000126  100.000178  100.000248  100.000346  100.000479  100.000659  100.000902  100.001229  100.001665  100.002244  100.003009  100.004013  100.005325  100.007029   100.00923  100.012059  100.015674  100.020269  100.026078  100.033383   100.04252  100.053887  100.067955  100.085272   100.10648  100.132316  100.163629  100.201385  100.246677  100.300738  100.364941  100.440813  100.530033          100.634442 | 100.690239 |   0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207  0.720701  0.720701  0.720701  0.720702  0.720703  0.720704  0.720705  0.720707   0.72071  0.720713  0.720717  0.720723  0.720729  0.720738  0.720749  0.720763  0.720781  0.720802  0.720829  0.720862  0.720902  0.720952  0.721011  0.721083  0.721169  0.721272  0.721394  0.721538  0.721707       0.721905 | 2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591117  2.591117  2.591117  2.591117  2.591118  2.591118  2.591119   2.59112  2.591122  2.591123  2.591126   2.59113  2.591135  2.591141  2.591149   2.59116  2.591175  2.591194  2.591218  2.591249  2.591289  2.591339  2.591403  2.591484  2.591584  2.591709  2.591864  2.592055  2.592288  2.592572  2.592917  2.593332   2.59383  2.594424   2.59513  2.595963  2.596942        2.598087 |      100.0       100.0       100.0  100.000001  100.000001  100.000002  100.000002  100.000004  100.000005  100.000008  100.000012  100.000017  100.000025  100.000036  100.000052  100.000074  100.000104  100.000147  100.000207  100.000289  100.000401  100.000554  100.000761   100.00104  100.001413  100.001911   100.00257  100.003438  100.004575  100.006058  100.007979  100.010455  100.013629  100.017676  100.022809  100.029282  100.037403  100.047537  100.060116  100.075646  100.094721  100.118027  100.146353  100.180605  100.221811  100.271131  100.329865  100.399462  100.481524   100.57781  100.690239 |  100.0 |        0.0 | 100.690239 |
    | 17:00 | 3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717829  3.717829  3.717829  3.717829   3.71783  3.717831  3.717832  3.717833  3.717835  3.717838  3.717841  3.717846  3.717852  3.717861  3.717872  3.717887  3.717906  3.717931  3.717964  3.718005  3.718059  3.718128  3.718214  3.718324  3.718461  3.718632  3.718844  3.719106  3.719427   3.71982  3.720297  3.720874  3.721568  3.722398  3.723387  3.724558  3.725938             3.727557 | 124.878652  124.878652  124.878652  124.878653  124.878653  124.878653  124.878654  124.878655  124.878657  124.878659  124.878663  124.878668  124.878674  124.878684  124.878698  124.878718  124.878746  124.878784  124.878837  124.878911  124.879011  124.879147  124.879331  124.879579   124.87991  124.880351  124.880935  124.881704  124.882711  124.884022   124.88572  124.887906  124.890708  124.894278  124.898804  124.904509  124.911663  124.920587   124.93166  124.945327  124.962109  124.982609   125.00752  125.037638  125.073865  125.117221   125.16885  125.230024  125.302151  125.386778 |  52.914556   52.914556   52.914556   52.914556   52.914557   52.914557   52.914557   52.914557   52.914557   52.914558   52.914558   52.914559   52.914561   52.914563   52.914565   52.914569   52.914575   52.914582   52.914593   52.914607   52.914626   52.914653   52.914689   52.914737   52.914802   52.914889   52.915003   52.915153    52.91535   52.915606   52.915938   52.916365   52.916913    52.91761   52.918495    52.91961   52.921008   52.922752   52.924915   52.927586   52.930865    52.93487   52.939737    52.94562   52.952696   52.961163   52.971244   52.983186   52.997263        53.013775 |  52.178281   52.178281   52.178281   52.178281   52.178281   52.178282   52.178282   52.178282   52.178282   52.178283   52.178283   52.178284   52.178286   52.178288    52.17829   52.178294   52.178299   52.178307   52.178317   52.178331    52.17835   52.178376   52.178411   52.178459   52.178522   52.178607   52.178719   52.178866   52.179059    52.17931   52.179636   52.180055   52.180592   52.181276   52.182143   52.183236   52.184607   52.186317   52.188439   52.191058   52.194273   52.198201   52.202973   52.208742   52.215681   52.223983   52.233868   52.245579   52.259382     52.275573 | 1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088386  1.088386  1.088386  1.088387  1.088387  1.088388  1.088389   1.08839  1.088392  1.088394  1.088397  1.088401  1.088406  1.088413  1.088421  1.088431  1.088445  1.088462  1.088483   1.08851  1.088543  1.088583  1.088633  1.088694  1.088768  1.088857  1.088965  1.089094  1.089247  1.089428  1.089642  1.089893 | 1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916  1.359159  1.359159  1.359158  1.359157  1.359157  1.359155  1.359154  1.359152   1.35915  1.359147  1.359144   1.35914  1.359135  1.359129  1.359122  1.359114  1.359103  1.359091  1.359077          1.359061 | 0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536159  0.536159  0.536159  0.536159  0.536159  0.536159   0.53616   0.53616  0.536161  0.536162  0.536163  0.536164  0.536166  0.536169  0.536172  0.536177  0.536182  0.536189  0.536197  0.536208  0.536222  0.536238  0.536259  0.536284  0.536315  0.536353  0.536398  0.536453  0.536519  0.536596  0.536689  0.536797      0.536925 | -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201869  -0.201869   -0.20187   -0.20187  -0.201871  -0.201872  -0.201874  -0.201876  -0.201878  -0.201882  -0.201886  -0.201892  -0.201899  -0.201907  -0.201918  -0.201932   -0.20195  -0.201971  -0.201997  -0.202029  -0.202068  -0.202116  -0.202173  -0.202241  -0.202323   -0.20242  -0.202535  -0.202669     -0.202826 | 0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571  0.665711  0.665711  0.665711  0.665712  0.665713  0.665714  0.665715  0.665716  0.665718   0.66572  0.665722  0.665726   0.66573  0.665734   0.66574  0.665747  0.665755  0.665764  0.665775  0.665788      0.665803 |      100.0 |      100.0       100.0       100.0  100.000001  100.000001  100.000002  100.000002  100.000004  100.000005  100.000008  100.000012  100.000017  100.000025  100.000035  100.000051  100.000072  100.000102  100.000144  100.000202  100.000281   100.00039  100.000538  100.000739  100.001008  100.001369   100.00185  100.002485  100.003322  100.004418  100.005845  100.007692  100.010073  100.013122  100.017008  100.021933  100.028143   100.03593  100.045644  100.057697  100.072575  100.090844  100.113161  100.140283  100.173076  100.212525  100.259741  100.315974  100.382613  100.461199          100.553422 | 100.602766 |   0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207  0.720701  0.720701  0.720702  0.720702  0.720703  0.720704  0.720706  0.720708  0.720711  0.720714  0.720719  0.720725  0.720732  0.720741  0.720753  0.720768  0.720787   0.72081  0.720838  0.720873  0.720915  0.720967  0.721029  0.721104  0.721194  0.721301  0.721427  0.721576       0.721751 | 2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591117  2.591117  2.591117  2.591117  2.591117  2.591118  2.591118  2.591119  2.591121  2.591122  2.591124  2.591127  2.591131  2.591137  2.591144  2.591153  2.591165  2.591181  2.591201  2.591227  2.591261  2.591303  2.591358  2.591426  2.591512  2.591619  2.591751  2.591915  2.592116  2.592362   2.59266  2.593021  2.593455  2.593974  2.594592  2.595324  2.596187        2.597199 |      100.0       100.0       100.0  100.000001  100.000001  100.000001  100.000002  100.000003  100.000004  100.000006  100.000009  100.000014   100.00002  100.000029  100.000042  100.000059  100.000084  100.000119  100.000168  100.000235  100.000327  100.000452  100.000622  100.000852  100.001161  100.001573   100.00212  100.002842  100.003791  100.005031  100.006641  100.008721  100.011395  100.014813  100.019158  100.024652  100.031564  100.040211  100.050972  100.064295  100.080702  100.100802    100.1253  100.155004  100.190838  100.233847   100.28521  100.346244  100.418412  100.503329  100.602766 |  100.0 |        0.0 | 100.602766 |
    | 17:30 | 3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717829  3.717829  3.717829   3.71783   3.71783  3.717831  3.717832  3.717834  3.717836  3.717839  3.717843  3.717848  3.717855  3.717864  3.717877  3.717893  3.717914  3.717941  3.717976  3.718021  3.718079  3.718153  3.718245  3.718362  3.718508   3.71869  3.718914  3.719191   3.71953  3.719943  3.720443  3.721046  3.721771  3.722635  3.723663  3.724877             3.726305 | 124.878652  124.878652  124.878652  124.878652  124.878653  124.878653  124.878654  124.878655  124.878656  124.878658   124.87866  124.878664   124.87867  124.878678  124.878689  124.878705  124.878728  124.878759  124.878802  124.878862  124.878944  124.879056  124.879207  124.879412  124.879686  124.880051  124.880536  124.881177  124.882016  124.883113  124.884536  124.886374  124.888735  124.891751  124.895583  124.900427  124.906516  124.914132  124.923608  124.935334  124.949771  124.967454  124.989001  125.015124  125.046632  125.084445  125.129598  125.183251  125.246688  125.321329 |  52.914556   52.914556   52.914556   52.914556   52.914557   52.914557   52.914557   52.914557   52.914557   52.914558   52.914558   52.914559    52.91456   52.914561   52.914564   52.914567   52.914571   52.914577   52.914586   52.914597   52.914613   52.914635   52.914665   52.914705   52.914758    52.91483   52.914925    52.91505   52.915214   52.915428   52.915706   52.916066   52.916527   52.917116   52.917865   52.918812   52.920002    52.92149   52.923342   52.925634   52.928455    52.93191   52.936119   52.941223   52.947377   52.954763    52.96358   52.974056   52.986439        53.001006 |  52.178281   52.178281   52.178281   52.178281   52.178281   52.178282   52.178282   52.178282   52.178282   52.178282   52.178283   52.178284   52.178285   52.178286   52.178288   52.178292   52.178296   52.178302    52.17831   52.178322   52.178337   52.178359   52.178388   52.178427   52.178479    52.17855   52.178642   52.178765   52.178926   52.179136   52.179409   52.179761   52.180214   52.180792   52.181526   52.182454   52.183621   52.185081   52.186896   52.189143    52.19191   52.195298   52.199426    52.20443   52.210465   52.217707   52.226353   52.236625   52.248768     52.263052 | 1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088386  1.088386  1.088386  1.088387  1.088387  1.088388  1.088389  1.088391  1.088393  1.088395  1.088398  1.088403  1.088408  1.088415  1.088424  1.088435   1.08845  1.088468   1.08849  1.088519  1.088553  1.088596  1.088649  1.088713  1.088791  1.088884  1.088996   1.08913   1.08929  1.089478  1.089699 | 1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916  1.359159  1.359159  1.359159  1.359158  1.359157  1.359156  1.359155  1.359154  1.359152  1.359149  1.359147  1.359143  1.359139  1.359134  1.359127   1.35912  1.359111  1.359101  1.359088          1.359074 | 0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536159  0.536159  0.536159  0.536159  0.536159  0.536159   0.53616   0.53616  0.536161  0.536162  0.536163  0.536165  0.536167   0.53617  0.536174  0.536178  0.536184  0.536191    0.5362  0.536212  0.536226  0.536244  0.536266  0.536293  0.536325  0.536365  0.536412  0.536469  0.536537  0.536618  0.536714      0.536826 | -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201869  -0.201869  -0.201869   -0.20187  -0.201871  -0.201872  -0.201873  -0.201874  -0.201877  -0.201879  -0.201883  -0.201888  -0.201894  -0.201901   -0.20191  -0.201922  -0.201937  -0.201955  -0.201978  -0.202006  -0.202039   -0.20208   -0.20213   -0.20219  -0.202261  -0.202346  -0.202447  -0.202565     -0.202704 | 0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571  0.665711  0.665711  0.665712  0.665712  0.665713  0.665714  0.665715  0.665716  0.665718  0.665721  0.665723  0.665727  0.665731  0.665736  0.665741  0.665748  0.665757  0.665766  0.665778      0.665791 |      100.0 |      100.0       100.0       100.0  100.000001  100.000001  100.000001  100.000002  100.000003  100.000004  100.000006  100.000009  100.000014   100.00002  100.000028  100.000041  100.000058  100.000083  100.000117  100.000164  100.000229  100.000318   100.00044  100.000605  100.000827  100.001125  100.001523  100.002051  100.002748  100.003662  100.004855  100.006404  100.008405  100.010974  100.014257  100.018428    100.0237  100.030328  100.038619  100.048932  100.061697  100.077413  100.096663  100.120121  100.148562  100.182869  100.224046  100.273221   100.33166  100.400767          100.482096 | 100.525661 |   0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207  0.720701  0.720701  0.720702  0.720703  0.720704  0.720705  0.720707  0.720709  0.720712  0.720716  0.720721  0.720727  0.720735  0.720745  0.720757  0.720773  0.720793  0.720817  0.720847  0.720884  0.720928  0.720982  0.721048  0.721126  0.721219   0.72133  0.721462       0.721616 | 2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591117  2.591117  2.591117  2.591117  2.591118  2.591118  2.591119   2.59112  2.591121  2.591123  2.591125  2.591129  2.591133  2.591139  2.591146  2.591157   2.59117  2.591187  2.591209  2.591237  2.591273  2.591319  2.591377   2.59145  2.591541  2.591655  2.591795  2.591968   2.59218  2.592438  2.592751  2.593129  2.593581  2.594122  2.594764  2.595523        2.596416 |      100.0       100.0       100.0       100.0  100.000001  100.000001  100.000002  100.000002  100.000003  100.000005  100.000008  100.000011  100.000016  100.000023  100.000033  100.000048  100.000068  100.000097  100.000136  100.000191  100.000266  100.000369  100.000509  100.000698  100.000953  100.001294  100.001747  100.002348  100.003138  100.004174  100.005522  100.007268  100.009518    100.0124  100.016075  100.020733  100.026607  100.033976  100.043171  100.054584  100.068677  100.085988  100.107145  100.132868  100.163984  100.201434  100.246282  100.299722  100.363089   100.43786  100.525661 |  100.0 |        0.0 | 100.525661 |
    | 18:00 | 3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717829  3.717829  3.717829  3.717829   3.71783   3.71783  3.717831  3.717833  3.717834  3.717837   3.71784  3.717844   3.71785  3.717858  3.717868  3.717882  3.717899  3.717922  3.717952   3.71799  3.718038    3.7181  3.718179  3.718278  3.718403  3.718558   3.71875  3.718988   3.71928  3.719637   3.72007  3.720594  3.721225   3.72198   3.72288  3.723946             3.725204 | 124.878652  124.878652  124.878652  124.878652  124.878653  124.878653  124.878653  124.878654  124.878655  124.878657  124.878659  124.878662  124.878666  124.878673  124.878682  124.878695  124.878713  124.878739  124.878774  124.878823   124.87889  124.878982  124.879106  124.879274  124.879501  124.879803  124.880206  124.880738  124.881438  124.882354  124.883546   124.88509  124.887076  124.889621  124.892862   124.89697  124.902147  124.908639  124.916736  124.926784  124.939187  124.954419   124.97303  124.995654  125.023017  125.055946  125.095376  125.142358  125.198063  125.263792 |  52.914556   52.914556   52.914556   52.914556   52.914556   52.914557   52.914557   52.914557   52.914557   52.914557   52.914558   52.914558   52.914559    52.91456   52.914562   52.914565   52.914568   52.914573    52.91458    52.91459   52.914603   52.914621   52.914645   52.914678   52.914722   52.914781    52.91486   52.914964   52.915101    52.91528   52.915513   52.915815   52.916203     52.9167   52.917334   52.918136   52.919148   52.920417   52.921999   52.923963   52.926386   52.929363   52.932999   52.937419   52.942764   52.949197   52.956897   52.966072   52.976947        52.989777 |  52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178282   52.178282   52.178282   52.178282   52.178283   52.178283   52.178284   52.178285   52.178287    52.17829   52.178293   52.178298   52.178305   52.178314   52.178327   52.178344   52.178368   52.178401   52.178444   52.178502   52.178579   52.178681   52.178815   52.178991   52.179219   52.179515   52.179896   52.180383   52.181005   52.181792   52.182784   52.184028    52.18558   52.187505   52.189882     52.1928   52.196366     52.2007   52.205942   52.212249     52.2198   52.228796   52.239461     52.252041 | 1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088386  1.088386  1.088386  1.088386  1.088387  1.088388  1.088388   1.08839  1.088391  1.088393  1.088396    1.0884  1.088404   1.08841  1.088418  1.088427  1.088439  1.088455  1.088474  1.088498  1.088528  1.088565   1.08861  1.088666  1.088733  1.088814  1.088912  1.089029  1.089168  1.089334  1.089529 | 1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916  1.359159  1.359159  1.359158  1.359158  1.359157  1.359156  1.359155  1.359153  1.359151  1.359149  1.359146  1.359142  1.359138  1.359132  1.359126  1.359118  1.359109  1.359098          1.359085 | 0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536159  0.536159  0.536159  0.536159  0.536159   0.53616   0.53616  0.536161  0.536161  0.536163  0.536164  0.536166  0.536168  0.536171  0.536175   0.53618  0.536186  0.536194  0.536204  0.536216  0.536231   0.53625  0.536273  0.536301  0.536335  0.536376  0.536426  0.536486  0.536556   0.53664      0.536739 | -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201869  -0.201869  -0.201869   -0.20187  -0.201871  -0.201872  -0.201873  -0.201875  -0.201877  -0.201881  -0.201884  -0.201889  -0.201896  -0.201904  -0.201914  -0.201926  -0.201942  -0.201961  -0.201985  -0.202014   -0.20205  -0.202093  -0.202145  -0.202207  -0.202282   -0.20237  -0.202474     -0.202597 | 0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571  0.665711  0.665711  0.665712  0.665712  0.665713  0.665714  0.665715  0.665717  0.665719  0.665721  0.665724  0.665728  0.665732  0.665737  0.665743   0.66575  0.665759  0.665769      0.665781 |      100.0 |      100.0       100.0       100.0       100.0  100.000001  100.000001  100.000002  100.000002  100.000003  100.000005  100.000007  100.000011  100.000016  100.000023  100.000033  100.000047  100.000067  100.000094  100.000133  100.000186  100.000259  100.000359  100.000494  100.000678  100.000924  100.001253  100.001691  100.002271  100.003033   100.00403  100.005327  100.007007  100.009169  100.011939  100.015467  100.019937  100.025573  100.032639  100.041452  100.052389  100.065891  100.082473  100.102733  100.127364  100.157156  100.193012   100.23595  100.287118  100.347796          100.419402 | 100.457804 |   0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207  0.720701  0.720701  0.720701  0.720702  0.720703  0.720704  0.720705  0.720707   0.72071  0.720713  0.720717  0.720722  0.720729  0.720738  0.720748  0.720762  0.720779  0.720799  0.720825  0.720857  0.720895  0.720942  0.720999  0.721067  0.721149  0.721246  0.721361       0.721497 | 2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591117  2.591117  2.591117  2.591117  2.591118  2.591118  2.591119   2.59112  2.591122  2.591124  2.591126   2.59113  2.591135  2.591141   2.59115  2.591161  2.591175  2.591193  2.591217  2.591248  2.591287  2.591336  2.591398  2.591476  2.591573  2.591693  2.591842  2.592024  2.592247  2.592518  2.592846   2.59324  2.593712  2.594275  2.594941        2.595728 |      100.0       100.0       100.0       100.0  100.000001  100.000001  100.000001  100.000002  100.000003  100.000004  100.000006  100.000009  100.000013  100.000019  100.000027  100.000039  100.000055  100.000078   100.00011  100.000155  100.000216  100.000301  100.000415  100.000571  100.000781  100.001063  100.001439  100.001938  100.002596  100.003459  100.004587  100.006051  100.007942   100.01037  100.013474  100.017418  100.022405  100.028677  100.036523  100.046287  100.058376  100.073266  100.091511  100.113755  100.140735  100.173296  100.212396  100.259116  100.314669  100.380401  100.457804 |  100.0 |        0.0 | 100.457804 |
    | 18:30 | 3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717829  3.717829  3.717829  3.717829   3.71783  3.717831  3.717832  3.717833  3.717835  3.717838  3.717841  3.717846  3.717853  3.717861  3.717872  3.717887  3.717906  3.717931  3.717963  3.718004  3.718056  3.718123  3.718207  3.718313  3.718446   3.71861  3.718814  3.719065  3.719373  3.719748  3.720202  3.720751  3.721409  3.722196  3.723131             3.724238 | 124.878652  124.878652  124.878652  124.878652  124.878652  124.878653  124.878653  124.878654  124.878654  124.878656  124.878657   124.87866  124.878664  124.878669  124.878676  124.878687  124.878701  124.878722  124.878751  124.878791  124.878846  124.878921  124.879023  124.879162  124.879348  124.879599  124.879932  124.880374  124.880957  124.881722  124.882719  124.884013  124.885684  124.887828  124.890567  124.894045  124.898442  124.903968  124.910879  124.919477  124.930118  124.943222  124.959274  124.978841  125.002569  125.031203  125.065583  125.106661    125.1555  125.213288 |  52.914556   52.914556   52.914556   52.914556   52.914556   52.914557   52.914557   52.914557   52.914557   52.914557   52.914557   52.914558   52.914559    52.91456   52.914561   52.914563   52.914566    52.91457   52.914576   52.914583   52.914594   52.914609   52.914629   52.914656   52.914692   52.914741   52.914807   52.914893   52.915007   52.915156   52.915351   52.915604   52.915931    52.91635   52.916885   52.917565   52.918424   52.919504   52.920855   52.922535   52.924614   52.927175   52.930311   52.934134    52.93877   52.944363   52.951079   52.959101   52.968638        52.979919 |  52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178282   52.178282   52.178282   52.178282   52.178282   52.178283   52.178284   52.178285   52.178286   52.178288   52.178291   52.178295     52.1783   52.178308   52.178318   52.178333   52.178352   52.178379   52.178415   52.178463   52.178527   52.178611   52.178723    52.17887   52.179061   52.179309   52.179629    52.18004   52.180565   52.181231   52.182074   52.183133   52.184457   52.186105   52.188144   52.190655    52.19373   52.197479   52.202025    52.20751   52.214094   52.221961   52.231312     52.242375 | 1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088386  1.088386  1.088386  1.088387  1.088387  1.088388  1.088389   1.08839  1.088392  1.088394  1.088397  1.088401  1.088406  1.088412   1.08842  1.088431  1.088444   1.08846  1.088481  1.088506  1.088538  1.088577  1.088625  1.088683  1.088753  1.088838   1.08894  1.089062  1.089207  1.089379 | 1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916  1.359159  1.359159  1.359158  1.359158  1.359157  1.359156  1.359154  1.359153   1.35915  1.359148  1.359145  1.359141  1.359136  1.359131  1.359124  1.359116  1.359106          1.359095 | 0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536159  0.536159  0.536159  0.536159  0.536159  0.536159   0.53616   0.53616  0.536161  0.536162  0.536163  0.536164  0.536166  0.536169  0.536172  0.536176  0.536182  0.536188  0.536197  0.536207   0.53622  0.536236  0.536256   0.53628   0.53631  0.536346  0.536389  0.536441  0.536503  0.536576      0.536663 | -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201869  -0.201869  -0.201869   -0.20187   -0.20187  -0.201871  -0.201872  -0.201874  -0.201876  -0.201878  -0.201882  -0.201886  -0.201891  -0.201898  -0.201907  -0.201917  -0.201931  -0.201947  -0.201968  -0.201993  -0.202024  -0.202061  -0.202106   -0.20216  -0.202225  -0.202303  -0.202394     -0.202502 | 0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571  0.665711  0.665711  0.665711  0.665712  0.665713  0.665713  0.665714  0.665716  0.665717   0.66572  0.665722  0.665725  0.665729  0.665733  0.665739  0.665745  0.665752  0.665761      0.665772 |      100.0 |      100.0       100.0       100.0       100.0  100.000001  100.000001  100.000001  100.000002  100.000003  100.000004  100.000006  100.000009  100.000013  100.000018  100.000026  100.000038  100.000054  100.000076  100.000108  100.000151  100.000211  100.000292  100.000404  100.000555  100.000758   100.00103  100.001393  100.001875  100.002509  100.003341  100.004427  100.005835  100.007653  100.009987  100.012968  100.016754  100.021539  100.027555  100.035077  100.044436   100.05602  100.070283  100.087758  100.109059  100.134893  100.166069  100.203506  100.248239  100.301433           100.36438 | 100.398177 |   0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207  0.720701  0.720701  0.720702  0.720702  0.720703  0.720704  0.720706  0.720708  0.720711  0.720714  0.720719  0.720724  0.720732  0.720741  0.720752  0.720766  0.720784  0.720806  0.720833  0.720867  0.720907  0.720956  0.721016  0.721087  0.721172  0.721273       0.721393 | 2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591117  2.591117  2.591117  2.591117  2.591117  2.591118  2.591119  2.591119  2.591121  2.591122  2.591125  2.591128  2.591132  2.591137  2.591144  2.591153  2.591165   2.59118  2.591201  2.591226  2.591259  2.591301  2.591353   2.59142  2.591502  2.591605  2.591733   2.59189  2.592082  2.592317  2.592601  2.592944  2.593355  2.593847  2.594432        2.595124 |      100.0       100.0       100.0       100.0       100.0  100.000001  100.000001  100.000001  100.000002  100.000003  100.000005  100.000007   100.00001  100.000015  100.000022  100.000031  100.000044  100.000063  100.000089  100.000126  100.000176  100.000245  100.000339  100.000467   100.00064  100.000873  100.001184  100.001598  100.002145  100.002865  100.003807  100.005033   100.00662  100.008664  100.011282  100.014618  100.018847  100.024178  100.030864  100.039208  100.049564  100.062354  100.078067  100.097274  100.120635  100.148904  100.182942  100.223726  100.272353  100.330049  100.398177 |  100.0 |        0.0 | 100.398177 |
    | 19:00 | 3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717829  3.717829  3.717829  3.717829   3.71783   3.71783  3.717831  3.717832  3.717834  3.717836  3.717839  3.717843  3.717848  3.717855  3.717865  3.717877  3.717893  3.717914  3.717941  3.717975  3.718019  3.718076  3.718147  3.718237   3.71835  3.718491  3.718666  3.718881  3.719146   3.71947  3.719864   3.72034  3.720913    3.7216  3.722419              3.72339 | 124.878652  124.878652  124.878652  124.878652  124.878652  124.878653  124.878653  124.878653  124.878654  124.878655  124.878656  124.878658  124.878661  124.878665  124.878671   124.87868  124.878692  124.878709  124.878732  124.878765  124.878809  124.878871  124.878955  124.879069  124.879223   124.87943  124.879706  124.880073  124.880558  124.881195  124.882029  124.883113  124.884516  124.886321  124.888632  124.891575  124.895303  124.900002  124.905893  124.913241  124.922359  124.933616  124.947443  124.964341  124.984889  125.009751  125.039683  125.075545  125.118299  125.169026 |  52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914557   52.914557   52.914557   52.914557   52.914557   52.914558   52.914558   52.914559    52.91456   52.914562   52.914564   52.914567   52.914572   52.914578   52.914587   52.914599   52.914616   52.914638   52.914668   52.914708   52.914762   52.914834   52.914929   52.915053   52.915216   52.915428   52.915702   52.916055   52.916507   52.917082   52.917811   52.918729    52.91988   52.921316   52.923098   52.925298      52.928   52.931301   52.935316   52.940173    52.94602   52.953024   52.961374        52.971279 |  52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178282   52.178282   52.178282   52.178282   52.178283   52.178283   52.178284   52.178285   52.178287   52.178289   52.178292   52.178297   52.178303   52.178311   52.178323   52.178339   52.178361   52.178391    52.17843   52.178483   52.178554   52.178647   52.178769   52.178928   52.179136   52.179405   52.179751   52.180194   52.180758   52.181472   52.182373   52.183502    52.18491   52.186657   52.188814   52.191463   52.194701   52.198638     52.2034   52.209134   52.216002    52.22419     52.233902 | 1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088386  1.088386  1.088386  1.088386  1.088387  1.088387  1.088388  1.088389  1.088391  1.088393  1.088395  1.088398  1.088402  1.088408  1.088415  1.088423  1.088435  1.088448  1.088466  1.088488  1.088515  1.088548  1.088589   1.08864  1.088701  1.088775  1.088863   1.08897  1.089097  1.089247 | 1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916  1.359159  1.359159  1.359159  1.359158  1.359157  1.359156  1.359155  1.359154  1.359152   1.35915  1.359147  1.359144   1.35914  1.359135  1.359129  1.359122  1.359113          1.359103 | 0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536159  0.536159  0.536159  0.536159  0.536159  0.536159   0.53616   0.53616  0.536161  0.536162  0.536163  0.536165  0.536167   0.53617  0.536173  0.536178  0.536183  0.536191  0.536199  0.536211  0.536224  0.536241  0.536262  0.536288  0.536319  0.536356  0.536402  0.536456   0.53652      0.536597 | -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201869  -0.201869  -0.201869   -0.20187  -0.201871  -0.201872  -0.201873  -0.201874  -0.201877  -0.201879  -0.201883  -0.201887  -0.201893    -0.2019   -0.20191  -0.201921  -0.201935  -0.201953  -0.201974  -0.202001  -0.202033  -0.202072  -0.202119  -0.202176  -0.202244  -0.202324     -0.202419 | 0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571  0.665711  0.665711  0.665711  0.665712  0.665713  0.665714  0.665715  0.665716  0.665718   0.66572  0.665723  0.665726   0.66573  0.665735   0.66574  0.665747  0.665754      0.665764 |      100.0 |      100.0       100.0       100.0       100.0       100.0  100.000001  100.000001  100.000001  100.000002  100.000003  100.000005  100.000007   100.00001  100.000015  100.000021   100.00003  100.000043  100.000062  100.000087  100.000123  100.000171  100.000238   100.00033  100.000454  100.000621  100.000846  100.001147  100.001546  100.002074  100.002768  100.003675  100.004855  100.006382  100.008347  100.010862  100.014065  100.018123  100.023238   100.02965  100.037648  100.047573  100.059827  100.074878  100.093274  100.115644  100.142712  100.175304  100.214354  100.260915          100.316165 | 100.345862 |   0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207  0.720701  0.720701  0.720701  0.720702  0.720703  0.720704  0.720705  0.720707  0.720709  0.720712  0.720716   0.72072  0.720726  0.720734  0.720744  0.720756  0.720771   0.72079  0.720814  0.720842  0.720877   0.72092  0.720971  0.721033  0.721108  0.721196       0.721301 | 2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591117  2.591117  2.591117  2.591117  2.591118  2.591118  2.591119   2.59112  2.591121  2.591123  2.591126  2.591129  2.591133  2.591139  2.591147  2.591157   2.59117  2.591187  2.591208  2.591236  2.591271  2.591316  2.591372  2.591443  2.591531   2.59164  2.591775   2.59194  2.592143  2.592389  2.592687  2.593045  2.593475  2.593987        2.594594 |      100.0       100.0       100.0       100.0       100.0  100.000001  100.000001  100.000001  100.000002  100.000003  100.000004  100.000006  100.000008  100.000012  100.000017  100.000025  100.000036  100.000051  100.000072  100.000102  100.000143  100.000199  100.000276  100.000382  100.000524  100.000716  100.000973  100.001316  100.001771   100.00237  100.003157  100.004182  100.005513  100.007232  100.009438  100.012256  100.015837  100.020363  100.026055  100.033175  100.042035  100.053006   100.06652  100.083083  100.103281  100.127789  100.157378  100.192927  100.235427  100.285991  100.345862 |  100.0 |        0.0 | 100.345862 |
    | 19:30 | 3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717829  3.717829  3.717829  3.717829   3.71783  3.717831  3.717832  3.717833  3.717835  3.717837   3.71784  3.717845  3.717851  3.717858  3.717868  3.717882  3.717899  3.717922  3.717951  3.717988  3.718036  3.718096  3.718173  3.718269  3.718389  3.718539  3.718724  3.718952  3.719231  3.719571  3.719984  3.720483  3.721082  3.721797             3.722649 | 124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878653  124.878653  124.878654  124.878654  124.878655  124.878657  124.878659  124.878663  124.878668  124.878674  124.878684  124.878698  124.878717  124.878743   124.87878   124.87883  124.878899  124.878993  124.879119   124.87929  124.879519  124.879823  124.880226  124.880757  124.881453   124.88236  124.883537  124.885055  124.887003   124.88949  124.892648  124.896638  124.901654  124.907926  124.915728  124.925386   124.93728  124.951853  124.969622  124.991178  125.017201  125.048462  125.085833  125.130293 |  52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914557   52.914557   52.914557   52.914557   52.914557   52.914557   52.914558   52.914558   52.914559   52.914561   52.914563   52.914565   52.914569   52.914574   52.914581   52.914591   52.914605   52.914623   52.914648   52.914681   52.914726   52.914785   52.914864   52.914968   52.915104   52.915281   52.915511   52.915808   52.916188   52.916675   52.917292   52.918072   52.919052   52.920277   52.921802    52.92369   52.926014   52.928861   52.932333   52.936545   52.941628   52.947735   52.955034        52.963716 |  52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178282   52.178282   52.178282   52.178282   52.178282   52.178283   52.178283   52.178284   52.178286   52.178287    52.17829   52.178294   52.178299   52.178306   52.178315   52.178329   52.178347   52.178371   52.178404   52.178447   52.178506   52.178583   52.178685   52.178818   52.178992   52.179217   52.179508   52.179882   52.180358   52.180964   52.181728   52.182689   52.183891   52.185386   52.187237   52.189516   52.192309   52.195713   52.199843   52.204828   52.210815   52.217973     52.226486 | 1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088386  1.088386  1.088386  1.088386  1.088387  1.088388  1.088389   1.08839  1.088391  1.088393  1.088396    1.0884  1.088404   1.08841  1.088417  1.088427  1.088439  1.088453  1.088472  1.088495  1.088524  1.088559  1.088603  1.088655  1.088719  1.088797   1.08889  1.089001  1.089133 | 1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916  1.359159  1.359159  1.359158  1.359158  1.359157  1.359156  1.359155  1.359153  1.359151  1.359149  1.359146  1.359143  1.359138  1.359133  1.359127   1.35912          1.359111 | 0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536159  0.536159  0.536159  0.536159  0.536159   0.53616   0.53616  0.536161  0.536161  0.536163  0.536164  0.536166  0.536168  0.536171  0.536175  0.536179  0.536185  0.536193  0.536203  0.536214  0.536229  0.536247  0.536269  0.536296  0.536328  0.536368  0.536415  0.536471      0.536538 | -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201869  -0.201869  -0.201869   -0.20187  -0.201871  -0.201872  -0.201873  -0.201875  -0.201877   -0.20188  -0.201884  -0.201889  -0.201895  -0.201903  -0.201913  -0.201925   -0.20194  -0.201959  -0.201981  -0.202009  -0.202043  -0.202084  -0.202133  -0.202193  -0.202263     -0.202346 | 0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571  0.665711  0.665711  0.665712  0.665712  0.665713  0.665714  0.665715  0.665717  0.665719  0.665721  0.665724  0.665727  0.665731  0.665736  0.665742  0.665748      0.665757 |      100.0 |      100.0       100.0       100.0       100.0       100.0  100.000001  100.000001  100.000001  100.000002  100.000003  100.000004  100.000006  100.000008  100.000012  100.000017  100.000024  100.000035   100.00005  100.000071  100.000099  100.000139  100.000194  100.000269  100.000371  100.000509  100.000695  100.000943  100.001274  100.001713  100.002291  100.003048  100.004036  100.005317  100.006969  100.009089  100.011796  100.015233  100.019577  100.025036  100.031862  100.040355  100.050868  100.063815   100.07968  100.099023  100.122491  100.150824  100.184862  100.225557          100.273977 | 100.300032 |   0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207  0.720701  0.720701  0.720701  0.720702  0.720703  0.720704  0.720705  0.720707   0.72071  0.720713  0.720717  0.720722  0.720729  0.720737  0.720747   0.72076  0.720776  0.720797  0.720821  0.720851  0.720888  0.720933  0.720987  0.721051  0.721129       0.721221 | 2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591117  2.591117  2.591117  2.591117  2.591117  2.591118  2.591118  2.591119   2.59112  2.591122  2.591124  2.591127   2.59113  2.591135  2.591141   2.59115  2.591161  2.591175  2.591193  2.591216  2.591246  2.591284  2.591332  2.591392  2.591467  2.591561  2.591676  2.591819  2.591993  2.592206  2.592464  2.592776   2.59315  2.593598         2.59413 |      100.0       100.0       100.0       100.0       100.0       100.0  100.000001  100.000001  100.000001  100.000002  100.000003  100.000004  100.000007   100.00001  100.000014   100.00002  100.000029  100.000041  100.000058  100.000082  100.000116  100.000162  100.000225  100.000312  100.000429  100.000587    100.0008  100.001084  100.001461  100.001959  100.002615  100.003472  100.004587   100.00603  100.007887  100.010265  100.013294  100.017133  100.021971   100.02804  100.035611  100.045009  100.056617   100.07088   100.08832  100.109537  100.135221  100.166161  100.203251  100.247499  100.300032 |  100.0 |        0.0 | 100.300032 |
    | 20:00 | 3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717829  3.717829  3.717829   3.71783   3.71783  3.717831  3.717832  3.717833  3.717835  3.717838  3.717842  3.717847  3.717853  3.717861  3.717873  3.717887  3.717906   3.71793  3.717962  3.718002  3.718053  3.718118    3.7182  3.718302   3.71843  3.718589  3.718785  3.719025  3.719319  3.719677  3.720109  3.720631  3.721255             3.722001 | 124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878653  124.878653  124.878653  124.878654  124.878655  124.878656  124.878658  124.878661  124.878664   124.87867  124.878678  124.878689  124.878704  124.878726  124.878756  124.878797  124.878853   124.87893  124.879035  124.879175  124.879364  124.879616  124.879951  124.880393  124.880973  124.881732  124.882718  124.883993  124.885633  124.887732  124.890404  124.893789  124.898054    124.9034  124.910069  124.918344  124.928561  124.941113  124.956457  124.975121  124.997712  125.024923   125.05754  125.096449 |  52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914557   52.914557   52.914557   52.914557   52.914557   52.914558   52.914558   52.914559    52.91456   52.914561   52.914564   52.914567   52.914571   52.914577   52.914585   52.914596   52.914611   52.914631   52.914659   52.914696   52.914745    52.91481   52.914897    52.91501   52.915158   52.915351     52.9156   52.915921   52.916331   52.916853   52.917515   52.918348   52.919393   52.920696   52.922313    52.92431   52.926763   52.929761   52.933408   52.937821   52.943137   52.949508        52.957107 |  52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178282   52.178282   52.178282   52.178282   52.178282   52.178283   52.178284   52.178285   52.178286   52.178288   52.178291   52.178295   52.178301   52.178309    52.17832   52.178335   52.178355   52.178382   52.178418   52.178466    52.17853   52.178615   52.178726   52.178872    52.17906   52.179305   52.179619   52.180021   52.180534   52.181182      52.182   52.183024   52.184302   52.185888   52.187846   52.190251   52.193191   52.196766   52.201094   52.206307   52.212554     52.220005 | 1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088386  1.088386  1.088386  1.088387  1.088387  1.088388  1.088389   1.08839  1.088392  1.088394  1.088397  1.088401  1.088406  1.088412   1.08842   1.08843  1.088443  1.088459  1.088478  1.088503  1.088533  1.088571  1.088616  1.088672  1.088739   1.08882  1.088917  1.089032 | 1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916  1.359159  1.359159  1.359158  1.359158  1.359157  1.359156  1.359154  1.359153  1.359151  1.359148  1.359145  1.359142  1.359137  1.359132  1.359125          1.359118 | 0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536159  0.536159  0.536159  0.536159  0.536159  0.536159   0.53616   0.53616  0.536161  0.536162  0.536163  0.536164  0.536166  0.536169  0.536172  0.536176  0.536181  0.536188  0.536196  0.536206  0.536218  0.536234  0.536253  0.536276  0.536304  0.536338  0.536379  0.536429      0.536487 | -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201869  -0.201869  -0.201869   -0.20187   -0.20187  -0.201871  -0.201872  -0.201874  -0.201876  -0.201878  -0.201882  -0.201886  -0.201891  -0.201897  -0.201906  -0.201916  -0.201929  -0.201945  -0.201965  -0.201989  -0.202018  -0.202053  -0.202096  -0.202148  -0.202209     -0.202283 | 0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571  0.665711  0.665711  0.665711  0.665712  0.665712  0.665713  0.665714  0.665716  0.665717  0.665719  0.665722  0.665724  0.665728  0.665732  0.665737  0.665743       0.66575 |      100.0 |      100.0       100.0       100.0       100.0       100.0       100.0  100.000001  100.000001  100.000001  100.000002  100.000003  100.000004  100.000006  100.000009  100.000014   100.00002  100.000028   100.00004  100.000057   100.00008  100.000113  100.000158  100.000219  100.000303  100.000416   100.00057  100.000775  100.001049  100.001414  100.001894  100.002526  100.003352  100.004425  100.005813  100.007598  100.009883  100.012791  100.016475  100.021117  100.026937  100.034195  100.043203  100.054324  100.067988  100.084692   100.10501  100.129605  100.159232  100.194747          100.237118 | 100.259943 |   0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207  0.720701  0.720701  0.720702  0.720702  0.720703  0.720704  0.720706  0.720708  0.720711  0.720714  0.720718  0.720724  0.720731   0.72074  0.720751  0.720765  0.720782  0.720803  0.720829  0.720861    0.7209  0.720946  0.721003   0.72107       0.721151 | 2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591117  2.591117  2.591117  2.591117  2.591117  2.591118  2.591119   2.59112  2.591121  2.591123  2.591125  2.591128  2.591132  2.591137  2.591144  2.591153  2.591165   2.59118    2.5912  2.591225  2.591257  2.591298  2.591349  2.591413  2.591493  2.591592  2.591714  2.591865  2.592048  2.592272  2.592543  2.592869  2.593259        2.593725 |      100.0       100.0       100.0       100.0       100.0       100.0       100.0  100.000001  100.000001  100.000002  100.000002  100.000004  100.000005  100.000008  100.000011  100.000016  100.000023  100.000033  100.000047  100.000067  100.000094  100.000132  100.000183  100.000254   100.00035  100.000481  100.000656  100.000891  100.001204  100.001618  100.002164   100.00288  100.003813  100.005023  100.006585  100.008589  100.011149    100.0144  100.018508  100.023674  100.030135  100.038177  100.048133    100.0604  100.075438   100.09378  100.116043  100.142933  100.175255  100.213917  100.259943 |  100.0 |        0.0 | 100.259943 |
    | 20:30 | 3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717829  3.717829  3.717829  3.717829   3.71783   3.71783  3.717831  3.717832  3.717834  3.717836  3.717839  3.717843  3.717849  3.717856  3.717865  3.717877  3.717893  3.717913   3.71794  3.717974  3.718017  3.718072  3.718141  3.718229  3.718338  3.718473  3.718641  3.718848  3.719102  3.719411  3.719786  3.720239  3.720784             3.721435 | 124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878653  124.878653  124.878653  124.878654  124.878655  124.878657  124.878659  124.878662  124.878666  124.878673  124.878682  124.878694  124.878712  124.878736   124.87877  124.878816  124.878879  124.878965  124.879081  124.879237  124.879445  124.879723   124.88009  124.880574  124.881207  124.882033  124.883102  124.884482  124.886252   124.88851  124.891378     124.895  124.899552  124.905244  124.912326  124.921092  124.931888  124.945121  124.961258  124.980842  125.004494  125.032919  125.066919 |  52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914557   52.914557   52.914557   52.914557   52.914557   52.914558   52.914558   52.914559    52.91456   52.914562   52.914565   52.914568   52.914573   52.914579   52.914588   52.914601   52.914618    52.91464   52.914671   52.914711   52.914766   52.914837   52.914932   52.915056   52.915217   52.915426   52.915696   52.916042   52.916483   52.917044   52.917751   52.918641   52.919753   52.921137    52.92285    52.92496   52.927546   52.930699   52.934525   52.939146   52.944699         52.95134 |  52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178282   52.178282   52.178282   52.178282   52.178283   52.178283   52.178284   52.178285   52.178287   52.178289   52.178293   52.178297   52.178304   52.178313   52.178325   52.178341   52.178363   52.178393   52.178433   52.178487   52.178557    52.17865   52.178771   52.178929   52.179134   52.179399   52.179738   52.180171    52.18072   52.181414   52.182287   52.183377   52.184734   52.186414   52.188483   52.191018    52.19411   52.197862   52.202393   52.207838      52.21435 | 1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088386  1.088386  1.088386  1.088386  1.088387  1.088387  1.088388  1.088389  1.088391  1.088393  1.088395  1.088398  1.088402  1.088408  1.088414  1.088423  1.088434  1.088447  1.088464  1.088485  1.088511  1.088543  1.088583  1.088631  1.088689  1.088759  1.088843  1.088944 | 1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916  1.359159  1.359159  1.359159  1.359158  1.359157  1.359156  1.359155  1.359154  1.359152   1.35915  1.359147  1.359144   1.35914  1.359136   1.35913          1.359123 | 0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536159  0.536159  0.536159  0.536159  0.536159   0.53616   0.53616   0.53616  0.536161  0.536162  0.536163  0.536165  0.536167   0.53617  0.536173  0.536178  0.536183   0.53619  0.536199  0.536209  0.536222  0.536239  0.536259  0.536283  0.536313  0.536348  0.536391      0.536443 | -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201869  -0.201869  -0.201869   -0.20187  -0.201871  -0.201872  -0.201873  -0.201874  -0.201877  -0.201879  -0.201883  -0.201887  -0.201893    -0.2019  -0.201909   -0.20192  -0.201933   -0.20195  -0.201971  -0.201996  -0.202027  -0.202064  -0.202109  -0.202163     -0.202227 | 0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571  0.665711  0.665711  0.665711  0.665712  0.665713  0.665714  0.665715  0.665716  0.665718   0.66572  0.665722  0.665725  0.665729  0.665733  0.665739      0.665745 |      100.0 |      100.0       100.0       100.0       100.0       100.0       100.0       100.0  100.000001  100.000001  100.000002  100.000002  100.000004  100.000005  100.000008  100.000011  100.000016  100.000023  100.000032  100.000046  100.000065  100.000092  100.000128  100.000178  100.000247  100.000341  100.000467  100.000636  100.000863  100.001165  100.001565  100.002092  100.002781  100.003679  100.004844  100.006345  100.008272   100.01073  100.013851  100.017793  100.022748  100.028944  100.036652  100.046193  100.057946   100.07235  100.089918  100.111238  100.136988  100.167938          100.204961 | 100.224927 |   0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207  0.720701  0.720701  0.720701  0.720702  0.720703  0.720704  0.720705  0.720707  0.720709  0.720712  0.720715   0.72072  0.720726  0.720734  0.720743  0.720755  0.720769  0.720788   0.72081  0.720837  0.720871  0.720911   0.72096  0.721019        0.72109 | 2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591117  2.591117  2.591117  2.591117  2.591118  2.591118  2.591119   2.59112  2.591121  2.591123  2.591126  2.591129  2.591133  2.591139  2.591147  2.591157   2.59117  2.591186  2.591207  2.591234  2.591269  2.591312  2.591367  2.591435   2.59152  2.591625  2.591754  2.591913  2.592106  2.592341  2.592624  2.592964        2.593371 |      100.0       100.0       100.0       100.0       100.0       100.0       100.0  100.000001  100.000001  100.000001  100.000002  100.000003  100.000004  100.000006  100.000009  100.000013  100.000019  100.000027  100.000038  100.000054  100.000076  100.000107  100.000149  100.000207  100.000286  100.000393  100.000538  100.000732  100.000991  100.001335   100.00179  100.002387  100.003167  100.004181  100.005493   100.00718   100.00934  100.012091  100.015575  100.019967  100.025474  100.032346  100.040876  100.051411   100.06436  100.080196  100.099467  100.122804   100.15093  100.184662  100.224927 |  100.0 |        0.0 | 100.224927 |
    | 21:00 | 3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717829  3.717829  3.717829  3.717829   3.71783  3.717831  3.717832  3.717833  3.717835  3.717837  3.717841  3.717845  3.717851  3.717859  3.717869  3.717882  3.717899  3.717921   3.71795  3.717986  3.718033  3.718092  3.718166  3.718259  3.718375  3.718519  3.718697  3.718915  3.719182  3.719507    3.7199  3.720374             3.720942 | 124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878653  124.878653  124.878654  124.878655  124.878656  124.878658   124.87866  124.878664  124.878669  124.878676  124.878686    124.8787   124.87872  124.878748  124.878785  124.878837  124.878908  124.879003  124.879132  124.879304  124.879534  124.879839  124.880242   124.88077   124.88146  124.882357  124.883516  124.885007  124.886913   124.88934  124.892413  124.896284  124.901137  124.907189    124.9147  124.923974  124.935371  124.949304  124.966258  124.986787  125.011525  125.041192 |  52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914557   52.914557   52.914557   52.914557   52.914557   52.914557   52.914558   52.914559    52.91456   52.914561   52.914563   52.914566    52.91457   52.914575   52.914582   52.914593   52.914606   52.914625    52.91465   52.914684   52.914729   52.914788   52.914867    52.91497   52.915105   52.915281   52.915507   52.915798   52.916171   52.916645   52.917246   52.918002   52.918951   52.920134   52.921601   52.923414   52.925641   52.928363   52.931676   52.935687   52.940519        52.946315 |  52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178282   52.178282   52.178282   52.178282   52.178282   52.178283   52.178284   52.178285   52.178286   52.178288   52.178291   52.178294     52.1783   52.178307   52.178317    52.17833   52.178349   52.178373   52.178406    52.17845   52.178509   52.178586   52.178687    52.17882   52.178991   52.179214   52.179499   52.179865    52.18033   52.180918    52.18166    52.18259    52.18375   52.185189   52.186967    52.18915    52.19182   52.195068   52.199001    52.20374     52.209423 | 1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088386  1.088386  1.088386  1.088386  1.088387  1.088388  1.088389   1.08839  1.088391  1.088393  1.088396  1.088399  1.088404   1.08841  1.088417  1.088426  1.088437  1.088452   1.08847  1.088492   1.08852  1.088554  1.088595  1.088645  1.088706   1.08878  1.088868 | 1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916  1.359159  1.359159  1.359158  1.359158  1.359157  1.359156  1.359155  1.359153  1.359152  1.359149  1.359147  1.359143  1.359139  1.359134          1.359129 | 0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536159  0.536159  0.536159  0.536159  0.536159  0.536159   0.53616   0.53616  0.536161  0.536162  0.536163  0.536164  0.536166  0.536168  0.536171  0.536174  0.536179  0.536185  0.536192  0.536201  0.536213  0.536227  0.536244  0.536265  0.536291  0.536322  0.536359      0.536404 | -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201869  -0.201869   -0.20187   -0.20187  -0.201871  -0.201872  -0.201873  -0.201875  -0.201877   -0.20188  -0.201884  -0.201889  -0.201895  -0.201902  -0.201912  -0.201923  -0.201938  -0.201956  -0.201978  -0.202004  -0.202036  -0.202075  -0.202122     -0.202178 | 0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571  0.665711  0.665711  0.665712  0.665712  0.665713  0.665714  0.665715  0.665717  0.665718   0.66572  0.665723  0.665726   0.66573  0.665735       0.66574 |      100.0 |      100.0       100.0       100.0       100.0       100.0       100.0       100.0  100.000001  100.000001  100.000001  100.000002  100.000003  100.000004  100.000006  100.000009  100.000013  100.000018  100.000026  100.000037  100.000053  100.000074  100.000104  100.000145  100.000202  100.000278  100.000382  100.000522   100.00071   100.00096  100.001292   100.00173  100.002306  100.003057  100.004033  100.005294  100.006916  100.008992  100.011633  100.014977  100.019191  100.024473  100.031061  100.039236  100.049332  100.061737  100.076905  100.095361  100.117711  100.144644          100.176946 | 100.194385 |   0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207  0.720701  0.720701  0.720701  0.720702  0.720703  0.720704  0.720705  0.720707   0.72071  0.720713  0.720717  0.720722  0.720728  0.720736  0.720746  0.720759  0.720774  0.720794  0.720817  0.720846  0.720881  0.720924  0.720975       0.721036 | 2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591117  2.591117  2.591117  2.591117  2.591117  2.591118  2.591118  2.591119   2.59112  2.591122  2.591124  2.591127   2.59113  2.591135  2.591142   2.59115  2.591161  2.591175  2.591192  2.591215  2.591244  2.591281  2.591328  2.591386  2.591458  2.591548  2.591659  2.591796  2.591963  2.592166  2.592412  2.592708        2.593063 |      100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0  100.000001  100.000001  100.000002  100.000002  100.000003  100.000005  100.000007   100.00001  100.000015  100.000021  100.000031  100.000044  100.000062  100.000087  100.000121  100.000169  100.000234  100.000322  100.000441  100.000601  100.000816  100.001101  100.001479  100.001976  100.002628  100.003476  100.004577  100.005996  100.007817  100.010142  100.013093  100.016823  100.021512  100.027376  100.034674  100.043711  100.054847    100.0685  100.085159  100.105385  100.129824  100.159213  100.194385 |  100.0 |        0.0 | 100.194385 |
    | 21:30 | 3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717829  3.717829  3.717829  3.717829   3.71783   3.71783  3.717831  3.717832  3.717834  3.717836  3.717838  3.717842  3.717847  3.717853  3.717862  3.717873  3.717887  3.717906   3.71793  3.717961     3.718   3.71805  3.718113  3.718192  3.718291  3.718414  3.718567  3.718755  3.718985  3.719266  3.719607  3.720018             3.720514 | 124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878653  124.878653  124.878653  124.878654  124.878655  124.878656  124.878658  124.878661  124.878665  124.878671   124.87868  124.878691  124.878707   124.87873  124.878761  124.878803  124.878861  124.878939  124.879045  124.879188  124.879378  124.879631  124.879966  124.880406  124.880983  124.881734  124.882707  124.883961  124.885568   124.88762  124.890223  124.893511  124.897644   124.90281  124.909237  124.917194  124.926996  124.939011  124.953668  124.971462   124.99296  125.018809 |  52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914557   52.914557   52.914557   52.914557   52.914557   52.914558   52.914558   52.914559    52.91456   52.914562   52.914564   52.914567   52.914572   52.914578   52.914586   52.914597   52.914613   52.914633   52.914661   52.914698   52.914748   52.914813   52.914899   52.915012   52.915159   52.915349   52.915594   52.915908   52.916309   52.916818   52.917461   52.918268   52.919278   52.920534   52.922089   52.924004   52.926352   52.929216   52.932693   52.936893        52.941942 |  52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178282   52.178282   52.178282   52.178282   52.178283   52.178283   52.178284   52.178285   52.178287   52.178289   52.178292   52.178296   52.178302    52.17831   52.178321   52.178336   52.178357   52.178384    52.17842   52.178469   52.178533   52.178618   52.178728   52.178872   52.179058   52.179299   52.179607       52.18   52.180499   52.181129   52.181921   52.182911   52.184143   52.185667   52.187546   52.189848   52.192656   52.196065   52.200184     52.205135 | 1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088386  1.088386  1.088386  1.088387  1.088387  1.088388  1.088389   1.08839  1.088392  1.088394  1.088397  1.088401  1.088406  1.088412  1.088419  1.088429  1.088441  1.088457  1.088476    1.0885  1.088529  1.088564  1.088608  1.088661  1.088725  1.088801 | 1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916  1.359159  1.359159  1.359158  1.359158  1.359157  1.359156  1.359155  1.359153  1.359151  1.359149  1.359146  1.359142  1.359138          1.359133 | 0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536159  0.536159  0.536159  0.536159  0.536159  0.536159   0.53616   0.53616  0.536161  0.536162  0.536163  0.536164  0.536166  0.536169  0.536172  0.536176  0.536181  0.536187  0.536195  0.536205  0.536217  0.536231   0.53625  0.536272  0.536299  0.536331       0.53637 | -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201869  -0.201869  -0.201869   -0.20187   -0.20187  -0.201871  -0.201872  -0.201874  -0.201876  -0.201878  -0.201881  -0.201885   -0.20189  -0.201897  -0.201905  -0.201915  -0.201927  -0.201943  -0.201961  -0.201984  -0.202012  -0.202046  -0.202087     -0.202136 | 0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571  0.665711  0.665711  0.665711  0.665712  0.665712  0.665713  0.665714  0.665715  0.665717  0.665719  0.665721  0.665724  0.665727  0.665731      0.665736 |      100.0 |      100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0  100.000001  100.000001  100.000002  100.000002  100.000003  100.000005  100.000007   100.00001  100.000015  100.000021   100.00003  100.000043   100.00006  100.000085  100.000118  100.000164  100.000227  100.000313  100.000428  100.000583   100.00079  100.001066   100.00143   100.00191  100.002537  100.003354  100.004413  100.005778  100.007528   100.00976  100.012594  100.016173  100.020671  100.026294   100.03329  100.041951  100.052621    100.0657  100.081655  100.101026  100.124431          100.152574 | 100.167785 |   0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207  0.720701  0.720701  0.720702  0.720702  0.720703  0.720704  0.720706  0.720708  0.720711  0.720714  0.720718  0.720724   0.72073  0.720739   0.72075  0.720763   0.72078    0.7208  0.720825  0.720855  0.720892  0.720937        0.72099 | 2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591117  2.591117  2.591117  2.591117  2.591118  2.591118  2.591119   2.59112  2.591121  2.591123  2.591125  2.591128  2.591132  2.591137  2.591144  2.591153  2.591165   2.59118  2.591199  2.591224  2.591255  2.591294  2.591344  2.591406  2.591483  2.591578  2.591696  2.591839  2.592015  2.592228  2.592486        2.592795 |      100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0  100.000001  100.000001  100.000001  100.000002  100.000003  100.000004  100.000006  100.000008  100.000012  100.000017  100.000025  100.000035   100.00005   100.00007  100.000098  100.000137   100.00019  100.000263  100.000361  100.000493  100.000671  100.000907  100.001221  100.001635  100.002178  100.002888   100.00381  100.005003  100.006536  100.008498  100.010996   100.01416  100.018146  100.023145  100.029382  100.037124  100.046688  100.058444  100.072824   100.09033  100.111536  100.137104  100.167785 |  100.0 |        0.0 | 100.167785 |
    | 22:00 | 3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717829  3.717829  3.717829  3.717829   3.71783   3.71783  3.717831  3.717833  3.717834  3.717837   3.71784  3.717844  3.717849  3.717856  3.717865  3.717877  3.717893  3.717913  3.717939  3.717972  3.718014  3.718068  3.718135   3.71822  3.718325  3.718456  3.718617  3.718815  3.719058  3.719353   3.71971             3.720141 | 124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878653  124.878653  124.878654  124.878654  124.878656  124.878657  124.878659  124.878663  124.878668  124.878674  124.878684  124.878697  124.878715   124.87874  124.878775  124.878822  124.878887  124.878974  124.879092  124.879249  124.879459  124.879737  124.880104  124.880585  124.881213  124.882029  124.883082  124.884437  124.886169  124.888372  124.891162  124.894676  124.899081  124.904575  124.911392  124.919812  124.930159  124.942814  124.958215  124.976871  124.999362 |  52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914557   52.914557   52.914557   52.914557   52.914557   52.914557   52.914558   52.914558   52.914559   52.914561   52.914563   52.914565   52.914569   52.914574    52.91458    52.91459   52.914602   52.914619   52.914642   52.914673   52.914714   52.914768    52.91484   52.914934   52.915057   52.915216   52.915422   52.915687   52.916025   52.916456   52.917001   52.917688   52.918549   52.919623   52.920955     52.9226   52.924622   52.927095   52.930104    52.93375        52.938143 |  52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178282   52.178282   52.178282   52.178282   52.178282   52.178283   52.178283   52.178284   52.178286   52.178287    52.17829   52.178293   52.178298   52.178305   52.178314   52.178326   52.178343   52.178366   52.178396   52.178436   52.178489    52.17856   52.178652   52.178772   52.178928    52.17913    52.17939   52.179722   52.180144   52.180679   52.181352   52.182196   52.183249   52.184556   52.186169   52.188152   52.190576   52.193527   52.197102      52.20141 | 1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088386  1.088386  1.088386  1.088386  1.088387  1.088387  1.088388  1.088389  1.088391  1.088393  1.088395  1.088398  1.088402  1.088407  1.088414  1.088422  1.088433  1.088446  1.088462  1.088482  1.088507  1.088538  1.088576  1.088621  1.088677  1.088744 | 1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916  1.359159  1.359159  1.359159  1.359158  1.359157  1.359157  1.359155  1.359154  1.359152   1.35915  1.359148  1.359145  1.359141          1.359137 | 0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536159  0.536159  0.536159  0.536159  0.536159   0.53616   0.53616   0.53616  0.536161  0.536162  0.536163  0.536165  0.536167   0.53617  0.536173  0.536177  0.536183  0.536189  0.536197  0.536208  0.536221  0.536236  0.536255  0.536279  0.536307      0.536341 | -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201869  -0.201869  -0.201869   -0.20187  -0.201871  -0.201872  -0.201873  -0.201874  -0.201877  -0.201879  -0.201883  -0.201887  -0.201892  -0.201899  -0.201908  -0.201918  -0.201931  -0.201948  -0.201967  -0.201992  -0.202021  -0.202056     -0.202099 | 0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571  0.665711  0.665711  0.665711  0.665712  0.665713  0.665714  0.665715  0.665716  0.665717  0.665719  0.665722  0.665725  0.665728      0.665732 |      100.0 |      100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0  100.000001  100.000001  100.000001  100.000002  100.000003  100.000004  100.000006  100.000008  100.000012  100.000017  100.000024  100.000034  100.000049  100.000069  100.000096  100.000134  100.000185  100.000256  100.000351  100.000479   100.00065  100.000878  100.001181   100.00158  100.002104  100.002788  100.003675  100.004822  100.006296  100.008181   100.01058  100.013616  100.017441  100.022236  100.028215  100.035636  100.044801  100.056064  100.069839  100.086606  100.106915          100.131401 | 100.144649 |   0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207  0.720701  0.720701  0.720701  0.720702  0.720703  0.720704  0.720705  0.720707  0.720709  0.720712  0.720715   0.72072  0.720726  0.720733  0.720742  0.720753  0.720768  0.720785  0.720806  0.720833  0.720865  0.720903        0.72095 | 2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591117  2.591117  2.591117  2.591117  2.591117  2.591118  2.591118  2.591119   2.59112  2.591122  2.591123  2.591126  2.591129  2.591134  2.591139  2.591147  2.591157  2.591169  2.591186  2.591206  2.591233  2.591266  2.591308  2.591361  2.591427  2.591509  2.591609  2.591733  2.591885   2.59207  2.592293        2.592562 |      100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0  100.000001  100.000001  100.000001  100.000002  100.000003  100.000005  100.000007   100.00001  100.000014   100.00002  100.000028   100.00004  100.000057   100.00008  100.000112  100.000155  100.000215  100.000295  100.000404  100.000551  100.000747  100.001007  100.001351  100.001804  100.002397  100.003169   100.00417   100.00546  100.007114  100.009226  100.011906  100.015292  100.019548  100.024871  100.031495  100.039698  100.049808  100.062205  100.077336  100.095712  100.117925  100.144649 |  100.0 |        0.0 | 100.144649 |
    | 22:30 | 3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717829  3.717829  3.717829   3.71783   3.71783  3.717831  3.717832  3.717833  3.717835  3.717838  3.717841  3.717845  3.717851  3.717859  3.717869  3.717882  3.717899  3.717921  3.717949  3.717984   3.71803  3.718087  3.718159  3.718249  3.718361  3.718499  3.718669  3.718879  3.719134  3.719443             3.719818 | 124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878653  124.878653  124.878653  124.878654  124.878655  124.878656  124.878658  124.878661  124.878664   124.87867  124.878678  124.878688  124.878703  124.878724  124.878752  124.878791  124.878844  124.878916  124.879013  124.879143  124.879316  124.879547  124.879852  124.880254  124.880778  124.881462  124.882346  124.883486  124.884946  124.886809  124.889174  124.892159   124.89591    124.9006  124.906434  124.913657  124.922556  124.933467  124.946781  124.962949   124.98249 |  52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914557   52.914557   52.914557   52.914557   52.914557   52.914558   52.914558   52.914559    52.91456   52.914561   52.914563   52.914566    52.91457   52.914576   52.914583   52.914594   52.914608   52.914627   52.914652   52.914686   52.914731   52.914791   52.914869   52.914972   52.915105   52.915278   52.915501   52.915787   52.916151   52.916613   52.917196   52.917929   52.918846   52.919986   52.921398   52.923137   52.925269    52.92787   52.931029        52.934847 |  52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178282   52.178282   52.178282   52.178282   52.178282   52.178283   52.178284   52.178285   52.178286   52.178288   52.178291   52.178295     52.1783   52.178308   52.178318   52.178332    52.17835   52.178375   52.178409   52.178453   52.178511   52.178588   52.178689    52.17882   52.178989   52.179208   52.179488   52.179845   52.180298    52.18087   52.181589   52.182487   52.183605    52.18499   52.186695   52.188786   52.191337   52.194434     52.198178 | 1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088386  1.088386  1.088386  1.088386  1.088387  1.088388  1.088389   1.08839  1.088391  1.088393  1.088396  1.088399  1.088404  1.088409  1.088416  1.088425  1.088436   1.08845  1.088468  1.088489  1.088516  1.088548  1.088588  1.088636  1.088694 | 1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916  1.359159  1.359159  1.359158  1.359158  1.359157  1.359156  1.359155  1.359154  1.359152   1.35915  1.359147  1.359144           1.35914 | 0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536159  0.536159  0.536159  0.536159  0.536159  0.536159   0.53616   0.53616  0.536161  0.536162  0.536163  0.536164  0.536166  0.536168  0.536171  0.536174  0.536179  0.536184  0.536191    0.5362  0.536211  0.536225  0.536241  0.536261  0.536286      0.536315 | -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201869  -0.201869   -0.20187   -0.20187  -0.201871  -0.201872  -0.201873  -0.201875  -0.201877   -0.20188  -0.201884  -0.201888  -0.201894  -0.201902  -0.201911  -0.201922  -0.201936  -0.201953  -0.201974  -0.201999   -0.20203     -0.202067 | 0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571  0.665711  0.665711  0.665712  0.665712  0.665713  0.665714  0.665715  0.665716  0.665718   0.66572  0.665723  0.665726      0.665729 |      100.0 |      100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0  100.000001  100.000001  100.000001  100.000002  100.000003  100.000004  100.000007  100.000009  100.000014   100.00002  100.000028  100.000039  100.000056  100.000078  100.000109  100.000151  100.000209  100.000287  100.000392  100.000534  100.000723  100.000974  100.001307  100.001743  100.002314  100.003058  100.004021  100.005261  100.006851  100.008879  100.011452  100.014701  100.018784  100.023888  100.030239  100.038101  100.047788  100.059665  100.074158  100.091758          100.113032 | 100.124554 |   0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207  0.720701  0.720701  0.720701  0.720702  0.720703  0.720704  0.720705  0.720707   0.72071  0.720713  0.720717  0.720721  0.720728  0.720735  0.720745  0.720757  0.720772  0.720791  0.720813  0.720841  0.720874       0.720915 | 2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591117  2.591117  2.591117  2.591117  2.591117  2.591118  2.591119  2.591119  2.591121  2.591122  2.591124  2.591127  2.591131  2.591135  2.591142   2.59115  2.591161  2.591174  2.591192  2.591214  2.591242  2.591278  2.591323  2.591379  2.591449  2.591536  2.591642  2.591773  2.591933  2.592126         2.59236 |      100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0  100.000001  100.000001  100.000001  100.000002  100.000002  100.000004  100.000005  100.000008  100.000011  100.000016  100.000023  100.000033  100.000046  100.000065  100.000091  100.000126  100.000175  100.000241  100.000331  100.000452  100.000614   100.00083  100.001116  100.001493  100.001988  100.002634  100.003473  100.004557   100.00595  100.007732  100.010001  100.012873  100.016492   100.02103  100.026691  100.033719  100.042401  100.053075  100.066135  100.082037  100.101309  100.124554 |  100.0 |        0.0 | 100.124554 |
    | 23:00 | 3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717829  3.717829  3.717829  3.717829   3.71783   3.71783  3.717831  3.717832  3.717834  3.717836  3.717839  3.717842  3.717847  3.717854  3.717862  3.717873  3.717887  3.717905  3.717929  3.717959  3.717998  3.718046  3.718107  3.718184  3.718279  3.718398  3.718545  3.718725  3.718945  3.719213             3.719538 | 124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878653  124.878653  124.878653  124.878654  124.878655  124.878657  124.878659  124.878662  124.878666  124.878673  124.878681  124.878693   124.87871  124.878733  124.878765  124.878808  124.878868  124.878947  124.879055  124.879199   124.87939  124.879644  124.879978  124.880416  124.880987  124.881729  124.882687  124.883918  124.885491  124.887492  124.890026  124.893216  124.897215  124.902201   124.90839  124.916034   124.92543  124.936923  124.950916  124.967871 |  52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914557   52.914557   52.914557   52.914557   52.914557   52.914557   52.914558   52.914558   52.914559    52.91456   52.914562   52.914564   52.914568   52.914572   52.914578   52.914587   52.914599   52.914614   52.914635   52.914663   52.914701    52.91475   52.914815   52.914901   52.915013   52.915158   52.915345   52.915586   52.915893   52.916284   52.916779   52.917403   52.918184   52.919159   52.920368   52.921862   52.923698   52.925944   52.928678        52.931991 |  52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178282   52.178282   52.178282   52.178282   52.178283   52.178283   52.178284   52.178285   52.178287   52.178289   52.178292   52.178297   52.178303   52.178311   52.178323   52.178338   52.178359   52.178386   52.178423   52.178471   52.178535   52.178619   52.178729   52.178871   52.179055    52.17929   52.179592   52.179976   52.180461   52.181072   52.181839   52.182794    52.18398   52.185445   52.187246   52.189448   52.192129     52.195377 | 1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088386  1.088386  1.088386  1.088386  1.088387  1.088387  1.088388  1.088389   1.08839  1.088392  1.088394  1.088397  1.088401  1.088405  1.088411  1.088419  1.088428   1.08844  1.088455  1.088473  1.088496  1.088524  1.088558    1.0886   1.08865 | 1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916  1.359159  1.359159  1.359158  1.359158  1.359157  1.359156  1.359155  1.359153  1.359151  1.359149  1.359146          1.359143 | 0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536159  0.536159  0.536159  0.536159  0.536159  0.536159   0.53616   0.53616  0.536161  0.536162  0.536163  0.536164  0.536166  0.536169  0.536172  0.536175   0.53618  0.536186  0.536194  0.536203  0.536215  0.536229  0.536246  0.536268      0.536293 | -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201869  -0.201869  -0.201869   -0.20187   -0.20187  -0.201871  -0.201872  -0.201874  -0.201876  -0.201878  -0.201881  -0.201885   -0.20189  -0.201896  -0.201904  -0.201914  -0.201926   -0.20194  -0.201958   -0.20198  -0.202007     -0.202039 | 0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571  0.665711  0.665711  0.665711  0.665712  0.665712  0.665713  0.665714  0.665715  0.665717  0.665719  0.665721  0.665723      0.665727 |      100.0 |      100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0  100.000001  100.000001  100.000001  100.000002  100.000002  100.000004  100.000005  100.000008  100.000011  100.000016  100.000022  100.000032  100.000045  100.000063  100.000088  100.000123   100.00017  100.000235  100.000322  100.000439  100.000595  100.000803  100.001079  100.001443   100.00192  100.002542  100.003349  100.004392  100.005731  100.007444  100.009622  100.012379  100.015852  100.020204  100.025632  100.032368  100.040689  100.050916  100.063427  100.078659          100.097117 | 100.107124 |   0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207  0.720701  0.720701  0.720702  0.720702  0.720703  0.720704  0.720706  0.720708  0.720711  0.720714  0.720718  0.720723   0.72073  0.720738  0.720748  0.720761  0.720777  0.720797   0.72082  0.720849       0.720885 | 2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591117  2.591117  2.591117  2.591117  2.591118  2.591118  2.591119   2.59112  2.591121  2.591123  2.591125  2.591128  2.591132  2.591137  2.591144  2.591153  2.591165  2.591179  2.591198  2.591222  2.591253  2.591291  2.591339  2.591398  2.591473  2.591564  2.591677  2.591814  2.591982        2.592185 |      100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0  100.000001  100.000001  100.000001  100.000002  100.000003  100.000004  100.000006  100.000009  100.000013  100.000018  100.000026  100.000037  100.000053  100.000074  100.000103  100.000143  100.000197  100.000271  100.000371  100.000505  100.000683  100.000921  100.001234  100.001647  100.002187  100.002889  100.003799  100.004972  100.006475  100.008392  100.010826    100.0139  100.017764  100.022596  100.028609  100.036056  100.045235  100.056493  100.070236  100.086933  100.107124 |  100.0 |        0.0 | 100.107124 |
    | 23:30 | 3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717828  3.717829  3.717829  3.717829  3.717829   3.71783  3.717831  3.717832  3.717833  3.717834  3.717837   3.71784  3.717844  3.717849  3.717856  3.717865  3.717877  3.717893  3.717912  3.717938   3.71797  3.718011  3.718063  3.718129   3.71821  3.718312  3.718437  3.718592  3.718782  3.719014             3.719295 | 124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878652  124.878653  124.878653  124.878653  124.878654  124.878655  124.878656  124.878658   124.87866  124.878664  124.878669  124.878676  124.878685  124.878699  124.878718  124.878744   124.87878  124.878828  124.878894  124.878983  124.879101   124.87926  124.879471  124.879749  124.880114  124.880592  124.881213  124.882017  124.883053   124.88438  124.886073   124.88822   124.89093  124.894335  124.898593  124.903889  124.910447  124.918527  124.928436   124.94053  124.955222 |  52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914556   52.914557   52.914557   52.914557   52.914557   52.914557   52.914557   52.914558   52.914559    52.91456   52.914561   52.914563   52.914566   52.914569   52.914574   52.914581   52.914591   52.914604   52.914621   52.914644   52.914675   52.914716   52.914771   52.914842   52.914935   52.915057   52.915214   52.915417   52.915676   52.916007   52.916426   52.916956   52.917622   52.918454   52.919489    52.92077   52.922349   52.924286   52.926649        52.929519 |  52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178281   52.178282   52.178282   52.178282   52.178282   52.178282   52.178283   52.178284   52.178284   52.178286   52.178288    52.17829   52.178294   52.178299   52.178306   52.178315   52.178328   52.178345   52.178367   52.178398   52.178438   52.178491   52.178561   52.178653   52.178772   52.178926   52.179125   52.179379   52.179703   52.180115   52.180634   52.181287   52.182103   52.183118   52.184374   52.185923   52.187822   52.190139     52.192954 | 1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088385  1.088386  1.088386  1.088386  1.088386  1.088387  1.088387  1.088388  1.088389  1.088391  1.088393  1.088395  1.088398  1.088402  1.088407  1.088413  1.088422  1.088432  1.088444   1.08846   1.08848  1.088504  1.088533  1.088569  1.088613 | 1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161  1.359161   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916   1.35916  1.359159  1.359159  1.359159  1.359158  1.359157  1.359157  1.359156  1.359154  1.359153  1.359151  1.359148          1.359145 | 0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536158  0.536159  0.536159  0.536159  0.536159  0.536159   0.53616   0.53616  0.536161  0.536161  0.536162  0.536163  0.536165  0.536167   0.53617  0.536173  0.536177  0.536182  0.536188  0.536196  0.536206  0.536219  0.536234  0.536252      0.536274 | -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201868  -0.201869  -0.201869  -0.201869   -0.20187  -0.201871  -0.201872  -0.201873  -0.201874  -0.201876  -0.201879  -0.201882  -0.201887  -0.201892  -0.201898  -0.201907  -0.201917   -0.20193  -0.201945  -0.201964  -0.201987     -0.202015 | 0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709  0.665709   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571   0.66571  0.665711  0.665711  0.665711  0.665712  0.665713  0.665713  0.665714  0.665716  0.665717  0.665719  0.665721      0.665724 |      100.0 |      100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0  100.000001  100.000001  100.000001  100.000002  100.000003  100.000004  100.000006  100.000009  100.000013  100.000018  100.000026  100.000036  100.000051  100.000072    100.0001  100.000139  100.000192  100.000263   100.00036  100.000489  100.000662  100.000891  100.001194  100.001591  100.002111  100.002788  100.003663   100.00479  100.006235  100.008077  100.010413  100.013364   100.01707  100.021704  100.027469  100.034607  100.043402  100.054188  100.067353          100.083346 | 100.092026 |   0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207    0.7207  0.720701  0.720701  0.720701  0.720702  0.720703  0.720704  0.720705  0.720707  0.720709  0.720712  0.720715  0.720719  0.720725  0.720732  0.720741  0.720752  0.720766  0.720782  0.720803  0.720828       0.720858 | 2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591116  2.591117  2.591117  2.591117  2.591117  2.591117  2.591118  2.591118  2.591119   2.59112  2.591122  2.591124  2.591126  2.591129  2.591134  2.591139  2.591147  2.591157  2.591169  2.591185  2.591205  2.591231  2.591263  2.591304  2.591355  2.591419  2.591497  2.591594  2.591713  2.591858        2.592034 |      100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0       100.0  100.000001  100.000001  100.000002  100.000002  100.000003  100.000005  100.000007   100.00001  100.000015  100.000021   100.00003  100.000043   100.00006  100.000084  100.000116  100.000161  100.000222  100.000304  100.000414  100.000562  100.000759   100.00102  100.001363  100.001814  100.002402  100.003165   100.00415  100.005417  100.007036  100.009096  100.011704   100.01499  100.019109  100.024248  100.030628   100.03851  100.048203  100.060064  100.074512  100.092026 |  100.0 |        0.0 | 100.092026 |

:cite:t:`ref-Todini2007` reports a peak discharge simulated for the channel outlet
of 643.72 m³/s, perfectly agreeing with our results:

>>> from hydpy import round_
>>> round_(numpy.max(fluxes.outflow.series))
643.742834

There is no indication of an error in the water balance:

>>> round_(model.check_waterbalance(conditions))
0.0

.. _musk_mct_long_segments:

Long segments
_____________

One advantage of the *MC* routing method is its relatively low sensitivity to changes
in spatial discretisation.  However, too long channel segments can cause unrealistic
discharge reductions at the start of the rising limb of flood waves.  We show this by
increasing the segment length to 25 km:

>>> nmbsegments(4)
>>> length(25.0)
>>> bottomwidth(15.0)
>>> sideslope(5.0)
>>> bottomslope(0.00025)
>>> stricklercoefficient(1.0/0.035)
>>> model.prepare_states(discharge=100.0)
>>> test.inits = ((states.discharge, 100.0),
...               (states.courantnumber, states.courantnumber[0]),
...               (states.reynoldsnumber, states.reynoldsnumber[0]))

.. integration-test::

    >>> conditions = test("musk_mct_long_segments", get_conditions="2000-01-01")
    |  date |                               referencewaterlevel |                                     wettedarea |                                     wettedperimeter |                                     surfacewidth |                               celerity |                               correctingfactor |                                  coefficient1 |                               coefficient2 |                               coefficient3 |     inflow |                                     referencedischarge |    outflow |                               courantnumber |                               reynoldsnumber |                                                  discharge | input1 |     input2 |     output |
    -------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
    | 00:00 | 3.717828  3.717828  3.717828             3.717828 | 124.878652  124.878652  124.878652  124.878652 |  52.914556   52.914556   52.914556        52.914556 |  52.178281   52.178281   52.178281     52.178281 | 1.088385  1.088385  1.088385  1.088385 | 1.359161  1.359161  1.359161          1.359161 | -0.581096  -0.581096  -0.581096     -0.581096 | 0.672256  0.672256  0.672256      0.672256 |  0.90884   0.90884   0.90884       0.90884 |      100.0 |      100.0       100.0       100.0               100.0 |      100.0 | 0.057656  0.057656  0.057656       0.057656 | 0.207289  0.207289  0.207289        0.207289 |      100.0       100.0       100.0       100.0       100.0 |  100.0 |        0.0 |      100.0 |
    | 00:30 | 3.717828  3.717828  3.717828             3.717828 | 124.878652  124.878652  124.878652  124.878652 |  52.914556   52.914556   52.914556        52.914556 |  52.178281   52.178281   52.178281     52.178281 | 1.088385  1.088385  1.088385  1.088385 | 1.359161  1.359161  1.359161          1.359161 | -0.581096  -0.581096  -0.581096     -0.581096 | 0.672256  0.672256  0.672256      0.672256 |  0.90884   0.90884   0.90884       0.90884 |      100.0 |      100.0       100.0       100.0               100.0 |      100.0 | 0.057656  0.057656  0.057656       0.057656 | 0.207289  0.207289  0.207289        0.207289 |      100.0       100.0       100.0       100.0       100.0 |  100.0 |        0.0 |      100.0 |
    | 01:00 | 3.717828  3.717828  3.717828             3.717828 | 124.878652  124.878652  124.878652  124.878652 |  52.914556   52.914556   52.914556        52.914556 |  52.178281   52.178281   52.178281     52.178281 | 1.088385  1.088385  1.088385  1.088385 | 1.359161  1.359161  1.359161          1.359161 | -0.581096  -0.581096  -0.581096     -0.581096 | 0.672256  0.672256  0.672256      0.672256 |  0.90884   0.90884   0.90884       0.90884 |      100.0 |      100.0       100.0       100.0               100.0 |      100.0 | 0.057656  0.057656  0.057656       0.057656 | 0.207289  0.207289  0.207289        0.207289 |      100.0       100.0       100.0       100.0       100.0 |  100.0 |        0.0 |      100.0 |
    | 01:30 | 3.717828  3.717828  3.717828             3.717828 | 124.878652  124.878652  124.878652  124.878652 |  52.914556   52.914556   52.914556        52.914556 |  52.178281   52.178281   52.178281     52.178281 | 1.088385  1.088385  1.088385  1.088385 | 1.359161  1.359161  1.359161          1.359161 | -0.581096  -0.581096  -0.581096     -0.581096 | 0.672256  0.672256  0.672256      0.672256 |  0.90884   0.90884   0.90884       0.90884 |      100.0 |      100.0       100.0       100.0               100.0 |      100.0 | 0.057656  0.057656  0.057656       0.057656 | 0.207289  0.207289  0.207289        0.207289 |      100.0       100.0       100.0       100.0       100.0 |  100.0 |        0.0 |      100.0 |
    | 02:00 | 3.717828  3.717828  3.717828             3.717828 | 124.878652  124.878652  124.878652  124.878652 |  52.914556   52.914556   52.914556        52.914556 |  52.178281   52.178281   52.178281     52.178281 | 1.088385  1.088385  1.088385  1.088385 | 1.359161  1.359161  1.359161          1.359161 | -0.581096  -0.581096  -0.581096     -0.581096 | 0.672256  0.672256  0.672256      0.672256 |  0.90884   0.90884   0.90884       0.90884 |      100.0 |      100.0       100.0       100.0               100.0 |      100.0 | 0.057656  0.057656  0.057656       0.057656 | 0.207289  0.207289  0.207289        0.207289 |      100.0       100.0       100.0       100.0       100.0 |  100.0 |        0.0 |      100.0 |
    | 02:30 | 3.717828  3.717828  3.717828             3.717828 | 124.878652  124.878652  124.878652  124.878652 |  52.914556   52.914556   52.914556        52.914556 |  52.178281   52.178281   52.178281     52.178281 | 1.088385  1.088385  1.088385  1.088385 | 1.359161  1.359161  1.359161          1.359161 | -0.581096  -0.581096  -0.581096     -0.581096 | 0.672256  0.672256  0.672256      0.672256 |  0.90884   0.90884   0.90884       0.90884 |      100.0 |      100.0       100.0       100.0               100.0 |      100.0 | 0.057656  0.057656  0.057656       0.057656 | 0.207289  0.207289  0.207289        0.207289 |      100.0       100.0       100.0       100.0       100.0 |  100.0 |        0.0 |      100.0 |
    | 03:00 | 3.717828  3.717828  3.717828             3.717828 | 124.878653  124.878651  124.878652  124.878652 |  52.914557   52.914556   52.914556        52.914556 |  52.178282   52.178281   52.178281     52.178281 | 1.088385  1.088385  1.088385  1.088385 | 1.359161  1.359161  1.359161          1.359161 | -0.581096  -0.581096  -0.581096     -0.581096 | 0.672256  0.672256  0.672256      0.672256 |  0.90884   0.90884   0.90884       0.90884 | 100.000003 | 100.000001   99.999999       100.0               100.0 |      100.0 | 0.057656  0.057656  0.057656       0.057656 | 0.207289  0.207289  0.207289        0.207289 | 100.000003   99.999999  100.000001       100.0       100.0 |  100.0 |   0.000003 |      100.0 |
    | 03:30 | 3.717828  3.717828  3.717828             3.717828 | 124.878663  124.878648  124.878654  124.878651 |  52.914558   52.914556   52.914557        52.914556 |  52.178283    52.17828   52.178282     52.178281 | 1.088385  1.088385  1.088385  1.088385 | 1.359161  1.359161  1.359161          1.359161 | -0.581096  -0.581096  -0.581096     -0.581096 | 0.672256  0.672256  0.672256      0.672256 |  0.90884   0.90884   0.90884       0.90884 | 100.000029 | 100.000012   99.999995  100.000002           99.999999 | 100.000001 | 0.057656  0.057656  0.057656       0.057656 | 0.207289  0.207289  0.207289        0.207289 | 100.000029   99.999989  100.000004   99.999998  100.000001 |  100.0 |   0.000029 | 100.000001 |
    | 04:00 | 3.717829  3.717828  3.717828             3.717828 | 124.878717  124.878627  124.878662  124.878648 |  52.914569   52.914551   52.914558        52.914556 |  52.178294   52.178276   52.178283     52.178281 | 1.088385  1.088385  1.088385  1.088385 | 1.359161  1.359161  1.359161          1.359161 | -0.581096  -0.581096  -0.581096     -0.581096 | 0.672256  0.672256  0.672256      0.672256 |  0.90884   0.90884   0.90884       0.90884 | 100.000175 | 100.000071   99.999972  100.000011           99.999996 | 100.000004 | 0.057656  0.057656  0.057656       0.057656 | 0.207289  0.207289  0.207289        0.207289 | 100.000175   99.999932  100.000026    99.99999  100.000004 |  100.0 |   0.000175 | 100.000004 |
    | 04:30 | 3.717834  3.717826  3.717829             3.717828 | 124.878958  124.878535  124.878696  124.878635 |  52.914616   52.914534   52.914565        52.914553 |   52.17834   52.178259    52.17829     52.178278 | 1.088386  1.088385  1.088385  1.088385 | 1.359161  1.359161  1.359161          1.359161 | -0.581096  -0.581096  -0.581096     -0.581096 | 0.672256  0.672256  0.672256      0.672256 | 0.908841   0.90884   0.90884       0.90884 | 100.000826 | 100.000333   99.999873  100.000048           99.999982 | 100.000017 | 0.057656  0.057656  0.057656       0.057656 |  0.20729  0.207289  0.207289        0.207289 | 100.000826   99.999686  100.000119   99.999955  100.000017 |  100.0 |   0.000826 | 100.000017 |
    | 05:00 | 3.717851   3.71782  3.717831             3.717827 |  124.87983  124.878213  124.878815  124.878592 |  52.914787   52.914471   52.914588        52.914545 |  52.178507   52.178197   52.178313      52.17827 | 1.088389  1.088384  1.088386  1.088385 |  1.35916  1.359161  1.359161          1.359161 | -0.581094  -0.581097  -0.581096     -0.581096 | 0.672256  0.672255  0.672256      0.672256 | 0.908842   0.90884  0.908841       0.90884 | 100.003194 | 100.001282   99.999523  100.000177           99.999934 | 100.000061 | 0.057656  0.057656  0.057656       0.057656 |  0.20729  0.207289  0.207289        0.207289 | 100.003194    99.99881  100.000442   99.999836  100.000061 |  100.0 |   0.003194 | 100.000061 |
    | 05:30 | 3.717902  3.717801  3.717838             3.717825 | 124.882516  124.877248   124.87916  124.878469 |  52.915312   52.914282   52.914656        52.914521 |  52.179022   52.178012   52.178379     52.178246 | 1.088397  1.088381  1.088387  1.088384 |  1.35916  1.359161   1.35916          1.359161 | -0.581091  -0.581098  -0.581095     -0.581096 | 0.672258  0.672255  0.672256      0.672256 | 0.908845  0.908839  0.908841       0.90884 | 100.010515 | 100.004206   99.998472  100.000553           99.999801 | 100.000179 | 0.057657  0.057656  0.057656       0.057656 | 0.207293  0.207288   0.20729        0.207289 | 100.010515   99.996177  100.001384   99.999501  100.000179 |  100.0 |   0.010515 | 100.000179 |
    | 06:00 | 3.718041  3.717753  3.717855             3.717819 | 124.889772  124.874721  124.880033   124.87817 |   52.91673   52.913788   52.914826        52.914462 |  52.180412   52.177528   52.178546     52.178189 | 1.088418  1.088373  1.088389  1.088384 | 1.359158  1.359161   1.35916          1.359161 |  -0.58108  -0.581102  -0.581094     -0.581097 | 0.672261  0.672254  0.672256      0.672256 | 0.908853  0.908836  0.908842       0.90884 | 100.030315 | 100.012103   99.995722  100.001503           99.999475 | 100.000456 | 0.057658  0.057655  0.057656       0.057656 |   0.2073  0.207286  0.207291        0.207289 | 100.030315   99.989278   100.00377   99.998684  100.000456 |  100.0 |   0.030315 | 100.000456 |
    | 06:30 | 3.718378   3.71764  3.717892             3.717807 | 124.907324  124.868822  124.881993  124.877528 |   52.92016   52.912635   52.915209        52.914337 |  52.183776   52.176397   52.178922     52.178066 |  1.08847  1.088356  1.088395  1.088382 | 1.359155  1.359163   1.35916          1.359161 | -0.581056   -0.58111  -0.581091     -0.581098 | 0.672266  0.672253  0.672257      0.672255 |  0.90887   0.90883  0.908844      0.908839 | 100.078179 | 100.031208   99.989301  100.003636           99.998776 |  100.00102 | 0.057661  0.057654  0.057657       0.057656 | 0.207317   0.20728  0.207293        0.207288 | 100.078179    99.97319  100.009114   99.996933   100.00102 |  100.0 |   0.078179 |  100.00102 |
    | 07:00 | 3.719119    3.7174  3.717968             3.717783 | 124.946002  124.856327  124.885961  124.876294 |  52.927718   52.910193   52.915985        52.914096 |  52.191187   52.174003   52.179682     52.177829 | 1.088585  1.088319  1.088407  1.088378 | 1.359147  1.359165  1.359159          1.359161 | -0.581001  -0.581127  -0.581086     -0.581099 | 0.672272  0.672251  0.672257      0.672255 | 0.908905   0.90882  0.908847      0.908838 | 100.183351 | 100.073312   99.975702  100.007955           99.997434 | 100.002028 | 0.057667  0.057652  0.057657       0.057656 | 0.207354  0.207268  0.207296        0.207287 | 100.183351   99.939247  100.019883   99.993591  100.002028 |  100.0 |   0.183351 | 100.002028 |
    | 07:30 | 3.720626  3.716934  3.718109             3.717742 | 125.024672  124.832018  124.893291  124.874154 |  52.943088   52.905441   52.917417        52.913677 |  52.206259   52.169343   52.181087     52.177419 | 1.088819  1.088246  1.088429  1.088372 | 1.359132   1.35917  1.359158          1.359161 | -0.580891  -0.581162  -0.581075     -0.581102 | 0.672278  0.672251  0.672257      0.672256 |  0.90897    0.9088  0.908852      0.908837 | 100.396214 | 100.158964   99.949246  100.015934           99.995105 | 100.003618 |  0.05768  0.057648  0.057658       0.057655 | 0.207429  0.207245  0.207303        0.207285 | 100.396214   99.873628  100.039612   99.987857  100.003618 |  100.0 |   0.396214 | 100.003618 |
    | 08:00 | 3.723484  3.716093  3.718347             3.717677 | 125.173935  124.788138   124.90575  124.870788 |  52.972237   52.896863   52.919852        52.913019 |  52.234842   52.160931   52.183474     52.176774 | 1.089262  1.088116  1.088466  1.088362 | 1.359102  1.359178  1.359155          1.359162 | -0.580681  -0.581223  -0.581058     -0.581107 | 0.672276  0.672254  0.672254      0.672257 | 0.909083   0.90877   0.90886      0.908835 | 100.797305 | 100.321521   99.901495  100.029495           99.991441 | 100.005804 | 0.057705  0.057641   0.05766       0.057655 | 0.207572  0.207202  0.207315        0.207282 | 100.797305   99.756253  100.072747   99.978988  100.005804 |  100.0 |   0.797305 | 100.005804 |
    | 08:30 | 3.728583  3.714673  3.718723             3.717585 | 125.440382  124.714086  124.925346   124.86599 |  53.024231   52.882383   52.923682        52.912082 |  52.285826   52.146733   52.187229     52.175855 | 1.090052  1.087896  1.088524  1.088347 |  1.35905  1.359193  1.359151          1.359163 | -0.580308  -0.581328   -0.58103     -0.581114 | 0.672249  0.672268  0.672249      0.672259 | 0.909266  0.908723  0.908871      0.908833 | 101.507032 | 100.611865   99.820924  100.050826           99.986219 | 100.008338 | 0.057749  0.057629  0.057664       0.057654 | 0.207827  0.207131  0.207334        0.207277 | 101.507032   99.560533  100.124003   99.966666  100.008338 |  100.0 |   1.507032 | 100.008338 |
    | 09:00 | 3.737195  3.712419  3.719271             3.717468 | 125.891047  124.596584  124.953927  124.859857 |  53.112058   52.859399   52.929267        52.910883 |  52.371948   52.124195   52.192706     52.174679 | 1.091386  1.087546  1.088609  1.088329 | 1.358962  1.359216  1.359146          1.359164 | -0.579677  -0.581493   -0.58099     -0.581122 | 0.672166    0.6723  0.672237      0.672263 | 0.909547  0.908658  0.908884      0.908832 | 102.694853 | 101.103426   99.693112  100.081939           99.979544 |  100.01054 | 0.057823  0.057609  0.057668       0.057653 | 0.208258  0.207019  0.207361        0.207271 | 102.694853   99.254644  100.197098   99.951541   100.01054 |  100.0 |   2.694853 |  100.01054 |
    | 09:30 | 3.751048  3.709042   3.72001             3.717337 | 126.617518  124.420603   124.99252  124.853035 |  53.253332   52.824957   52.936807        52.909549 |  52.510478   52.090422     52.2001     52.173372 | 1.093529  1.087022  1.088723  1.088309 | 1.358822  1.359251  1.359138          1.359166 | -0.578665  -0.581741  -0.580936     -0.581132 | 0.671976  0.672365  0.672215       0.67227 | 0.909953  0.908573  0.908896      0.908831 | 104.586356 | 101.897084   99.501767  100.123954           99.972119 | 100.011152 | 0.057943   0.05758  0.057675       0.057652 |  0.20895   0.20685  0.207398        0.207265 | 104.586356   98.805047  100.292783   99.935895  100.011152 |  100.0 |   4.586356 | 100.011152 |
    | 10:00 | 3.772364  3.704255  3.720927             3.717221 | 127.739105  124.171337  125.040369  124.846996 |  53.470714   52.776134   52.946154        52.908369 |  52.723639   52.042547   52.209265     52.172214 | 1.096822  1.086278  1.088866  1.088291 | 1.358607  1.359301  1.359129          1.359167 | -0.577111  -0.582093  -0.580869     -0.581141 | 0.671599  0.672482   0.67218      0.672279 | 0.910509  0.908471  0.908906      0.908834 | 107.466641 | 103.125447   99.230898  100.176052           99.965547 |  100.00827 | 0.058127  0.057538  0.057683       0.057651 | 0.210015   0.20661  0.207444        0.207259 | 107.466641   98.182852  100.406193   99.924325   100.00827 |  100.0 |   7.466641 |  100.00827 |
    | 10:30 | 3.803855  3.697836  3.721948              3.71717 | 129.404414  123.837487  125.093715    124.8443 |  53.791867   52.710674   52.956573        52.907842 |  53.038555   51.978358   52.219482     52.171697 | 1.101674  1.085281  1.089024  1.088283 | 1.358296  1.359368  1.359118          1.359167 | -0.574823  -0.582564  -0.580794     -0.581144 | 0.670921  0.672673  0.672129      0.672291 | 0.911232  0.908361  0.908907      0.908841 | 111.678559 | 104.956077   98.868406  100.234143           99.962614 |  99.999413 | 0.058397  0.057483  0.057692        0.05765 | 0.211587  0.206289  0.207495        0.207256 | 111.678559     97.3735  100.523695    99.92429   99.999413 |  100.0 |  11.678559 |  99.999413 |
    | 11:00 | 3.848658  3.689721  3.722918             3.717256 |  131.79072  123.416028  125.144382  124.848781 |  54.248766    52.62792   52.966467        52.908718 |  53.486582   51.897211   52.229184     52.172556 | 1.108552   1.08402  1.089174  1.088296 | 1.357863  1.359452  1.359108          1.359166 | -0.571585  -0.583162  -0.580723     -0.581138 | 0.669786  0.672964   0.67206      0.672304 | 0.912117  0.908259  0.908893      0.908855 | 117.614836 |  107.59328   98.411265  100.289324           99.967491 |  99.981791 |  0.05878  0.057412    0.0577       0.057651 |  0.21382  0.205883  0.207544        0.207261 | 117.614836   96.389821   100.61956   99.946338   99.981791 |  100.0 |  17.614836 |  99.981791 |
    | 11:30 | 3.910188  3.680124  3.723569             3.717577 | 135.100656  122.918433  125.178341  124.865567 |  54.876247    52.53005   52.973097        52.911999 |  54.101877   51.801242   52.235685     52.175774 | 1.117953  1.082527  1.089275  1.088346 | 1.357285  1.359553  1.359102          1.359163 |  -0.56717  -0.583869  -0.580675     -0.581114 | 0.667997  0.673377  0.671976      0.672315 | 0.913135  0.908193  0.908853      0.908876 | 125.703736 | 111.278182   97.872232  100.326312            99.98576 |  99.952811 | 0.059304  0.057329  0.057706       0.057654 | 0.216882  0.205402  0.207576        0.207277 | 125.703736   95.287979  100.652992  100.003781   99.952811 |  100.0 |  25.703736 |  99.952811 |
    | 12:00 | 3.991919  3.669682  3.723495             3.718253 |  139.55587  122.378063  125.174507  124.900845 |  55.709746   52.423561   52.972349        52.918894 |   54.91919   51.696821   52.234951     52.182534 | 1.130361  1.080901  1.089264  1.088451 | 1.356548  1.359663  1.359102          1.359156 |  -0.56136  -0.584639  -0.580681     -0.581065 | 0.665323   0.67393  0.671884      0.672319 | 0.914224    0.9082  0.908779      0.908906 | 136.388647 | 116.286724   97.287706  100.322132          100.024158 |  99.910852 | 0.059995  0.057238  0.057705        0.05766 |  0.22094  0.204879  0.207573        0.207311 | 136.388647    94.18499  100.566289  100.111595   99.910852 |  100.0 |  36.388647 |  99.910852 |
    | 12:30 | 4.097098  3.659613  3.722148             3.719411 | 145.387535  121.858017  125.104158  124.961248 |  56.782366   52.320873   52.958612        52.930697 |  55.970981   51.596127   52.221482     52.194108 |   1.1462  1.079331  1.089055   1.08863 | 1.355648   1.35977  1.359116          1.359144 | -0.553974  -0.585383  -0.580779      -0.58098 | 0.661509  0.674624    0.6718      0.672309 | 0.915286  0.908332   0.90866      0.908945 | 150.102726 | 122.925104      96.726  100.245506           100.08991 |  99.856289 | 0.060876  0.057151  0.057693        0.05767 | 0.226147  0.204374  0.207505        0.207368 | 150.102726   93.275474  100.285189  100.284286   99.856289 |  100.0 |  50.102726 |  99.856289 |
    | 13:00 | 4.228418  3.651861  3.718841             3.721164 | 152.823881  121.458358  124.931493  125.052768 |  58.121575   52.241821   52.924883        52.948576 |  57.284183    51.51861   52.188407      52.21164 | 1.165779  1.078121  1.088542  1.088902 | 1.354594  1.359853   1.35915          1.359126 | -0.544891  -0.585956  -0.581022     -0.580851 | 0.656307  0.675434   0.67175      0.672277 | 0.916196  0.908649  0.908488      0.908991 | 167.240349 | 131.521964   96.294888  100.057502          100.189557 |  99.792735 | 0.061964  0.057083  0.057665       0.057685 | 0.232626  0.203986   0.20734        0.207456 | 167.240349   92.844349   99.722725  100.532538   99.792735 |  100.0 |  67.240349 |  99.792735 |
    | 13:30 | 4.387894  3.649242  3.712774              3.72359 | 162.086462  121.323445  124.615066  125.179459 |  59.747911   52.215108   52.863015        52.973315 |  58.878937   51.492416    52.12774     52.235899 | 1.189277  1.077712  1.087601  1.089278 | 1.353409   1.35988  1.359213          1.359101 | -0.534058   -0.58615  -0.581467     -0.580674 | 0.649517    0.6763  0.671768      0.672211 | 0.916833  0.909216  0.908259      0.909037 | 188.127569 | 142.429764   96.149483   99.713198          100.327536 |  99.727995 | 0.063268   0.05706  0.057612       0.057706 | 0.240463  0.203854  0.207036        0.207578 | 188.127569   93.277116   98.785758  100.859457   99.727995 |  100.0 |  88.127569 |  99.727995 |
    | 14:00 |  4.57581  3.655275  3.703201             3.726652 | 173.327322  121.634305  124.116511  125.339472 |  61.664287   52.276638    52.76539        53.004546 |  60.758098   51.552751   52.032012     52.266523 | 1.216596  1.078654  1.086115  1.089753 | 1.352133  1.359815  1.359312           1.35907 | -0.521561  -0.585703   -0.58217     -0.580449 | 0.640958  0.677089  0.671909      0.672101 | 0.917024  0.910075  0.907984      0.909074 | 212.993962 | 155.953114   96.484681   99.171329          100.501875 |   99.67711 | 0.064783  0.057113  0.057529       0.057732 | 0.249657  0.204157  0.206557        0.207731 | 212.993962   95.029868   97.400118  101.250369    99.67711 |  100.0 | 112.993962 |   99.67711 |
    | 14:30 | 4.791858  3.674415   3.68942             3.730214 | 186.687365  122.622837  123.400398  125.525678 |  63.867552   52.471824   52.624849        53.040865 |  62.918577   51.744147     51.8942     52.302137 | 1.247541  1.081638  1.083973  1.090305 | 1.350807  1.359612  1.359456          1.359033 | -0.507531  -0.584289  -0.583184     -0.580188 |  0.63066  0.677635  0.672223      0.671939 | 0.916733  0.911275  0.907672      0.909092 | 241.948195 | 172.415486   97.552489   98.394304          100.704846 |  99.659742 | 0.066496   0.05728   0.05741       0.057763 | 0.260177  0.205116  0.205867        0.207909 | 241.948195   98.647581   95.515096  101.673842   99.659742 |  100.0 | 141.948195 |  99.659742 |
    | 15:00 | 5.034228  3.711521  3.671088             3.733928 | 202.230689  124.549766  122.450745      125.72 |  66.339255   52.850238   52.437897        53.078741 |  65.342282   52.115212   51.710878     52.339278 | 1.281704  1.087407   1.08112   1.09088 | 1.349476  1.359223  1.359649          1.358996 | -0.492198  -0.581558  -0.584535     -0.579917 | 0.618677   0.67766  0.672778      0.671722 | 0.915896  0.912787  0.907362      0.909072 |  274.95935 | 192.074476   99.642363   97.366248           100.91677 |   99.70227 | 0.068384  0.057601  0.057251       0.057795 | 0.271921  0.206974  0.204949        0.208094 |  274.95935  104.704416   93.142778  102.070179    99.70227 |  100.0 |  174.95935 |   99.70227 |
    | 15:30 |  5.29996  3.771591  3.648412             3.737208 | 219.947254  127.698353  121.280748  125.891743 |  69.049195   53.462831   52.206651        53.112194 |  67.999596   52.715909   51.484123     52.372081 | 1.318538  1.096702  1.077583  1.091388 |  1.34818  1.358613   1.35989          1.358962 | -0.475852  -0.577166  -0.586212     -0.579677 | 0.605203  0.676828  0.673634      0.671455 | 0.914523  0.914535  0.907108      0.908996 | 311.845554 | 215.111403  103.080918   96.103399          101.104163 |  99.836036 | 0.070417   0.05812  0.057053       0.057823 | 0.284732  0.209977  0.203813        0.208258 | 311.845554  113.768341   90.386002  102.347498   99.836036 |  100.0 | 211.845554 |  99.836036 |
    | 16:00 | 5.585172  3.859237  3.622409             3.739201 | 239.748304   132.35708  119.945391  125.996158 |  71.957801   54.356645   51.941473        53.132522 |  70.851719   53.592365   51.224094     52.392014 | 1.357401  1.110172  1.073518  1.091697 |  1.34695   1.35776  1.360171          1.358942 | -0.458816  -0.570823  -0.588141     -0.579531 | 0.590515  0.674746  0.674838      0.671157 |  0.91267  0.916369  0.906985      0.908844 | 352.270855 | 241.608497   108.22176   94.667148           101.21813 | 100.093373 | 0.072559  0.058871  0.056826       0.057841 | 0.298416  0.214347  0.202508        0.208358 | 352.270855  126.354752   87.465528  102.381498  100.093373 |  100.0 | 252.270855 | 100.093373 |
    | 16:30 | 5.885159  3.978264  3.595198             3.738769 |  261.45286  138.806893  118.555199  125.973499 |   75.01708   55.570493   51.663967        53.128111 |  73.851589   54.782642   50.951977     52.387689 | 1.397583  1.128295  1.069252   1.09163 |  1.34581  1.356667  1.360469          1.358947 | -0.441429  -0.562326  -0.590167     -0.579563 | 0.574938  0.671019  0.676407      0.670862 | 0.910417  0.918088  0.907094      0.908598 | 395.750601 | 271.510751  115.441037   93.177717          101.193385 | 100.502931 |  0.07477   0.05988  0.056588       0.057837 | 0.312742  0.220263  0.201142        0.208336 | 395.750601  142.873598   84.746077  102.017647  100.502931 |  100.0 | 295.750601 | 100.502931 |
    | 17:00 | 6.195328  4.130557  3.569896             3.734652 | 284.840396  147.265835  117.269206  125.757915 |  78.180201   57.123577   51.405934        53.086128 |  76.953284   56.305565   50.698956     52.346521 | 1.438433  1.151209  1.065276  1.090992 | 1.344771  1.355371  1.360751          1.358989 |  -0.42399  -0.551645  -0.592058     -0.579864 |  0.55885  0.665258  0.678287      0.670635 | 0.907894  0.919413  0.907522      0.908257 | 441.664962 | 304.679205  125.082912   91.805234          100.958102 | 101.075512 | 0.077015  0.061155  0.056366       0.057801 |  0.32749    0.2278   0.19987        0.208131 | 441.664962  163.617961   82.705477  101.098976  101.075512 |  100.0 | 341.664962 | 101.075512 |
    | 17:30 | 6.510569  4.316782  3.551098             3.725397 | 309.596078  157.924771  116.317956  125.273892 |  81.395037   59.022713   51.214236        52.991749 |   80.10569   58.167821    50.51098     52.253974 | 1.479279  1.178836  1.062315  1.089559 | 1.343839  1.353924  1.360963          1.359083 | -0.406791  -0.538862  -0.593467     -0.580541 |  0.54256  0.657298  0.680396      0.670541 | 0.905183  0.920178  0.908394      0.907814 | 489.279607 | 340.799058  137.502117   90.793327          100.430369 | 101.804743 | 0.079257  0.062689    0.0562       0.057721 | 0.342419  0.236973  0.198925        0.207668 | 489.279607  188.699206   81.966749   99.465713  101.804743 |  100.0 | 389.279607 | 101.804743 |
    | 18:00 | 6.826008  4.535517  3.544319             3.709716 | 335.362024  170.887314  115.975781  124.455729 |  84.611893   61.253377   51.145106        52.831833 |  83.260077   60.355168   50.443192     52.097165 | 1.519519  1.210771  1.061246  1.087126 | 1.343012  1.352396  1.361039          1.359245 |  -0.39008  -0.524216  -0.593976     -0.581692 | 0.526387  0.647063  0.682513      0.670681 | 0.902385  0.920214   0.90977      0.907292 |  537.77215 | 379.437277  152.991805    90.43006           99.539908 | 102.649918 | 0.081463   0.06446  0.056141       0.057586 | 0.357303   0.24769  0.198584        0.206883 |  537.77215  218.055518   83.218981   97.004169  102.649918 |  100.0 |  437.77215 | 102.649918 |
    | 18:30 | 7.137058  4.783577  3.555781              3.68669 | 361.743837  186.166726  116.554603  123.258774 |  87.783994   63.783109   51.261992         52.59701 |  86.370578   62.835774   50.557809     51.866902 | 1.558619  1.246364  1.063053  1.083548 | 1.342286  1.350855  1.360908          1.359485 | -0.374063  -0.508061  -0.593115     -0.583385 | 0.510598  0.634697  0.684323       0.67116 | 0.899584  0.919466  0.911659      0.906736 | 586.262607 | 420.045204  171.766371   91.044925           98.240787 | 103.527965 | 0.083604  0.066431  0.056242       0.057386 | 0.371934  0.259775  0.199161        0.205731 | 586.262607  251.445898   87.166367   93.683686  103.527965 |  100.0 | 486.262607 | 103.527965 |
    | 19:00 | 7.439528  5.056408   3.59184              3.65608 | 388.325803  203.682453  118.384179  121.675784 |  90.868597    66.56545   51.629726        52.284842 |   89.39528   65.564084   50.918401     51.560796 | 1.596121  1.284803  1.068725   1.07878 | 1.341652  1.349361  1.360503          1.359809 | -0.358903  -0.490815  -0.590416     -0.585645 | 0.495414  0.620517  0.685398       0.67208 | 0.896843  0.917977   0.91398       0.90621 | 633.846018 | 461.978876   193.93754   92.995149           96.529295 | 104.309247 | 0.085656  0.068555  0.056559        0.05712 |  0.38612  0.272993  0.200974        0.204197 | 633.846018  288.461092   94.462252   89.596621  104.309247 |  100.0 | 533.846018 | 104.309247 |
    | 19:30 | 7.729694   5.34855  3.658291             3.618653 | 414.686212  223.263194  121.789846  119.753048 |  93.827716   69.544723   52.307398        51.903166 |  92.296935   68.485501   51.582913     51.186531 | 1.631644  1.325207  1.079125   1.07293 | 1.341103  1.347959   1.35978          1.360212 | -0.344725  -0.472914  -0.585479      -0.58842 | 0.481007  0.604951   0.68522      0.673527 | 0.894209  0.915862  0.916538      0.905811 | 679.625318 | 504.525034  219.494726   96.652743            94.46069 | 104.818785 | 0.087598  0.070785  0.057139       0.056793 | 0.399694  0.287068  0.204309        0.202319 | 679.625318  328.544691   105.64603   84.992626  104.818785 |  100.0 | 579.625318 | 104.818785 |
    | 20:00 | 8.004332   5.65411  3.759772             3.576495 | 440.411628  244.656483  127.075988  117.604018 |   96.62849   72.660839   53.342298        51.473238 |  95.043319   71.541105   52.597716     50.764952 | 1.664877  1.366696  1.094877  1.066314 |  1.34063  1.346675  1.358729          1.360678 | -0.331621  -0.454774  -0.578026     -0.591564 | 0.467508  0.588473  0.683245      0.675551 | 0.891716  0.913277  0.919055      0.905658 | 722.742733 | 546.930209  248.293743  102.399101           92.162023 | 104.844225 | 0.089414   0.07307  0.058018       0.056424 | 0.412514  0.301714  0.209388        0.200202 | 722.742733  371.021466  121.092917   80.300104  104.844225 |  100.0 | 622.742733 | 104.844225 |
    | 20:30 | 8.260733  5.966973    3.8984             3.533072 | 465.109512  267.528399  134.463609  115.409065 |  99.243274   75.851419   54.756035        51.030405 |  97.607327   74.669726   53.983999     50.330719 | 1.695578  1.408424  1.116156  1.059471 | 1.340225  1.345524   1.35739           1.36117 | -0.319652  -0.436778  -0.568011     -0.594822 |  0.45501  0.571542  0.678931      0.678134 | 0.889384  0.910371  0.921131       0.90588 | 762.408181 | 588.430697  280.034809  110.566868           89.829016 |  104.16117 |  0.09109  0.075366  0.059204       0.056041 | 0.424461  0.316638  0.216297        0.198017 | 762.408181  415.130083  140.974098   76.105761   104.16117 |  100.0 | 662.408181 |  104.16117 |
    | 21:00 | 8.496687  6.281784  4.074749             3.493376 | 488.418803  291.530849  144.139111  113.419005 |  101.64955   79.061883   56.554445        50.625582 |  99.966875   77.817845   55.747486     49.933757 | 1.723565  1.449701  1.142846  1.053188 |  1.33988  1.344505  1.355832          1.361631 | -0.308854  -0.419222  -0.555534     -0.597818 | 0.443576  0.554601  0.672034      0.681188 | 0.887225  0.907322  0.922527      0.906618 | 797.923464 |  628.28149  314.340692   121.49651            87.72687 | 102.556077 | 0.092618  0.077633   0.06069        0.05569 | 0.435436  0.331589  0.225043        0.196017 | 797.923464  460.057558  165.274061   73.133666  102.556077 |  100.0 | 697.923464 | 102.556077 |
    | 21:30 | 8.710468  6.593308  4.286613             3.463618 | 510.018278  316.258206  156.174461  111.937508 | 103.829692   82.238817   58.715049         50.32211 | 102.104679   80.933084   57.866132     49.636178 | 1.748711  1.489894  1.174389  1.048462 | 1.339588  1.343613  1.354148          1.361983 | -0.299245  -0.402361  -0.540912     -0.600075 | 0.433243  0.537971  0.662422      0.684507 | 0.885244  0.904226  0.922988      0.907975 | 828.701418 | 665.782574   350.68953   135.44274           86.170115 |  99.868602 | 0.093989  0.079839  0.062442       0.055426 | 0.445366  0.346328   0.23549        0.194515 | 828.701418  504.973156  193.761421    72.17475   99.868602 |  100.0 | 728.701418 |  99.868602 |
    | 22:00 | 8.900795  6.897033  4.529813             3.450789 |  529.63265  341.300839  170.543249  111.301568 | 105.770652   85.336214   61.195214        50.191283 | 104.007947   83.970333   60.298134     49.507892 | 1.770934  1.528496  1.209945   1.04642 | 1.339344   1.34284  1.352433          1.362136 | -0.290824  -0.386384  -0.524593     -0.601051 | 0.424026  0.521939  0.650292      0.687761 |  0.88344  0.901185  0.922453      0.909992 |  854.27953 | 700.301394  388.487969  152.575413           85.504072 |  96.038156 | 0.095201  0.081954  0.064414       0.055312 | 0.454197  0.360648  0.247411        0.193868 |  854.27953  549.059907  226.032088   74.004435   96.038156 |  100.0 |  754.27953 |  96.038156 |
    | 22:30 | 9.066802  7.189125  4.798717             3.462058 | 547.036472  366.254479  187.119152  111.860085 | 107.463596   88.314979   63.937499          50.3062 | 105.668015   86.891251   62.987166     49.620577 | 1.790195   1.56511  1.248515  1.048214 | 1.339142  1.342173  1.350767          1.361999 |  -0.28358  -0.371425   -0.50709     -0.600193 | 0.415927  0.506709  0.636058      0.690489 | 0.881809  0.898265  0.920993      0.912608 | 874.327871 | 731.290705  427.090043  172.954491           86.089105 |   91.14655 | 0.096251  0.083959   0.06655       0.055412 | 0.461891  0.374378  0.260511        0.194437 | 874.327871  591.542379  261.536665   79.304348    91.14655 |  100.0 | 774.327871 |   91.14655 |
    | 23:00 | 9.207997  7.466423  5.086901             3.504007 | 562.056028  390.733671  205.686327  113.950433 | 108.903516   91.142869   66.876415        50.734001 | 107.079973   89.664226   65.869011     50.040071 |  1.80649  1.599432  1.289056  1.054873 | 1.338977    1.3416  1.349206          1.361501 |  -0.27749  -0.357574   -0.48892     -0.597012 | 0.408933  0.492427  0.620265      0.692115 | 0.880344   0.89551  0.918779      0.915632 | 888.651531 | 758.301449  465.825892   196.51638           88.287315 |  85.451249 | 0.097139  0.085837   0.06879       0.055785 |  0.46843  0.387379  0.274465        0.196554 | 888.651531  631.709843  299.617153   88.596376   85.451249 |  100.0 | 788.651531 |  85.451249 |
    | 23:30 | 9.324227  7.726403  5.387747             3.581976 | 574.569426  414.382579  225.955306  117.882397 | 110.088829   93.794161   69.944456         51.52913 | 108.242268   92.264032   68.877472     50.819759 | 1.819844  1.631243  1.330572  1.067176 | 1.338847   1.34111  1.347785           1.36061 | -0.272527  -0.344885  -0.470554     -0.591152 | 0.403021  0.479189  0.603495      0.692008 | 0.879038  0.892945  0.916022      0.918754 | 897.187983 | 780.990593  504.029431  223.069512           92.459425 |   79.40292 | 0.097867  0.087576   0.07108       0.056472 |  0.47381   0.39954  0.288951        0.200479 | 897.187983  668.934343  339.545148  102.199141    79.40292 |  100.0 | 797.187983 |   79.40292 |
    | 00:00 | 9.415632  7.967131  5.694736             3.698593 | 584.505141  436.882893  247.571158  123.876823 | 111.020986   96.249118   73.075144        52.718391 | 109.156323   94.671315   71.947364     51.985926 | 1.830309  1.660397  1.372156  1.085399 | 1.338747  1.340692  1.346518          1.359353 | -0.268653  -0.333379  -0.452405     -0.582506 | 0.398162   0.46705  0.586287       0.68951 | 0.877882  0.890582   0.91292      0.921507 |      900.0 | 799.124185  541.062935  252.285099           98.911569 |   73.61812 | 0.098437  0.089169  0.073371        0.05749 | 0.478038  0.410779  0.303655        0.206329 |      900.0  702.683574  380.558409  120.204024    73.61812 |  100.0 |      800.0 |   73.61812 |
    | 00:30 | 9.482616  8.187198  6.002264             3.854722 | 591.839295  457.959014  270.169838  132.115248 |  111.70409   98.493363   76.211324        54.310607 | 109.826162   96.871979   75.022642     53.547222 | 1.837958  1.686804  1.413086  1.109481 | 1.338676  1.340338  1.345404            1.3578 | -0.265832   -0.32306  -0.434783     -0.571146 | 0.394322  0.456033  0.569129      0.684267 | 0.876867  0.888424  0.909675      0.923577 | 897.264896 | 812.576082  576.337507  283.761114          107.953588 |  68.854546 | 0.098854  0.090611  0.075622       0.058832 | 0.481135  0.421036  0.318317        0.214124 | 897.264896  732.528757  421.893769  142.501224   68.854546 |  100.0 | 797.264896 |  68.854546 |
    | 01:00 | 9.525805  8.385655  6.305206             4.048225 |  596.59193  477.380872  293.356216  142.663993 | 112.144536  100.517237   79.300739        56.283955 | 110.258055    98.85655   78.052062     55.482248 | 1.842881  1.710427  1.452745  1.138858 | 1.338631   1.34004  1.344435          1.356056 |  -0.26402   -0.31391  -0.417937     -0.557392 | 0.391463   0.44614  0.552372      0.676056 | 0.875985  0.886467  0.906411      0.924628 | 889.260928 |  821.32296  609.328938  316.989555          119.813639 |  65.927469 | 0.099122  0.091901    0.0778       0.060468 | 0.483132  0.430273  0.332699         0.22373 | 889.260928  758.147887  462.815168  168.779253   65.927469 |  100.0 | 789.260928 |  65.927469 |
    | 01:30 | 9.546018  8.561956   6.59931             4.274905 | 598.822552  494.964743  316.744125  155.497653 | 112.350663  102.315156   82.300023        58.595651 | 110.460179  100.619555   80.993102     57.749052 | 1.845183  1.731263  1.490662   1.17266 |  1.33861  1.339789  1.343598          1.354236 | -0.263174  -0.305904  -0.402041      -0.54171 | 0.389543  0.437356  0.536298       0.66504 | 0.875228  0.884704  0.903233      0.924562 | 876.351729 | 825.436312  639.588887  351.413508          134.648494 |  65.629814 | 0.099247  0.093038  0.079881       0.062346 | 0.484066  0.438469  0.346611        0.234915 | 876.351729  779.324917  502.636265  198.573393   65.629814 |  100.0 | 776.351729 |  65.629814 |
    | 02:00 | 9.544231  8.715894  6.881123             4.528469 | 598.625179  510.572439  339.966124  170.462192 | 112.332439  103.885026   85.173962        61.181504 | 110.442309  102.158939   83.811231      60.28469 | 1.844979  1.749347  1.526488   1.20975 | 1.338611  1.339582  1.342879          1.352442 | -0.263249  -0.299003   -0.38721     -0.524682 |  0.38852  0.429651  0.521102        0.6516 | 0.874588  0.883124  0.900208      0.923394 | 858.969595 | 825.072141  666.751615  386.449009          152.477329 |    68.6319 | 0.099236  0.094024  0.081844       0.064404 | 0.483983  0.445618  0.359898        0.247345 | 858.969595  795.945527  540.737537  231.303215     68.6319 |  100.0 | 758.969595 |    68.6319 |
    | 02:30 | 9.521553  8.847553  7.147934             4.802635 | 596.123115  524.109262  362.683824  187.366055 | 112.101166   105.22769   87.894911        63.977462 | 110.215527   103.47553   86.479341     63.026352 | 1.842397  1.764732  1.559976  1.249072 | 1.338635  1.339411  1.342263          1.350744 | -0.264198  -0.293167  -0.373511      -0.50684 | 0.388349   0.42299   0.50691      0.636396 | 0.874057  0.881717  0.897377      0.921377 |   837.5984 | 820.459102  690.536734  421.510527          173.262774 |  75.458274 | 0.099095  0.094863  0.083678        0.06658 | 0.482935  0.451727  0.372444        0.260701 |   837.5984   807.99013  576.578143  266.312961   75.458274 |  100.0 |   737.5984 |  75.458274 |
    | 03:00 | 9.479195  8.957256  7.397697             5.090451 | 591.463655  535.521026  384.595039  205.920202 | 111.669204  106.346447      90.442        66.912615 | 109.791954  104.572561   88.976968     65.904507 | 1.837568  1.777497  1.590964   1.28955 | 1.338679  1.339274  1.341737          1.349189 | -0.265975   -0.28835  -0.360976       -0.4887 | 0.388987  0.417331  0.493794       0.62001 | 0.873629   0.88047  0.894761      0.918706 | 812.756821 | 811.885741  710.748529  456.033301          196.817835 |  86.408892 | 0.098832  0.095559  0.085374       0.068817 | 0.480977  0.456814  0.384159        0.274637 | 812.756821   815.52483  609.703051  302.909099   86.408892 |  100.0 | 712.756821 |  86.408892 |
    | 03:30 | 9.418451  9.045525  7.628943             5.385423 | 584.812859  544.790436  405.438012  225.795256 | 111.049731  107.246612    92.80026        69.920755 |  109.18451  105.455245   91.289431     68.854231 | 1.830632  1.787732  1.619358  1.330254 | 1.338744  1.339167  1.341289          1.347796 | -0.268534  -0.284503  -0.349609     -0.470694 | 0.390387   0.41263  0.481789      0.603055 | 0.873298  0.879373  0.892366      0.915629 | 784.982427 | 799.687452  727.272474  489.491469           222.85647 | 101.568214 | 0.098455  0.096117  0.086927       0.071063 | 0.478168  0.460905  0.394984        0.288839 | 784.982427  818.690957  639.746032  340.392999  101.568214 |  100.0 | 684.982427 | 101.568214 |
    | 04:00 | 9.340669   9.11304  7.840698             5.681494 | 576.350556  551.933078   424.99323  246.619275 |  110.25651  107.935136   94.959748        72.940097 | 108.406693  106.130399   93.406984     71.814939 | 1.821729   1.79554  1.645121  1.370378 | 1.338829  1.339087  1.340908           1.34657 | -0.271828  -0.281579  -0.339393     -0.453177 | 0.392506  0.408839  0.470898      0.586031 | 0.873057  0.878413  0.890191      0.912341 | 754.817086 | 784.233653  740.069587  521.411689          250.979681 | 120.805106 |  0.09797  0.096543  0.088335       0.073273 |  0.47457  0.464033  0.404878        0.303022 | 754.817086  817.693771  666.429205  378.087807  120.805106 |  100.0 | 654.817086 | 120.805106 |
    | 04:30 | 9.247238  9.160612  8.032401             5.973954 | 566.265664  556.993208  443.083353  268.049935 | 109.303698  108.420276    96.91474        75.922615 | 107.472384  106.606117   95.324011     74.739539 | 1.811004   1.80103  1.668253  1.409347 | 1.338933  1.339032  1.340585          1.345501 |  -0.27581  -0.279527  -0.330299     -0.436383 | 0.395298  0.405911  0.461104       0.56937 | 0.872903  0.877582  0.888226      0.909032 |  722.79402 | 765.915628  749.169203  551.382426           280.76932 | 143.831832 | 0.097385  0.096842  0.089598       0.075417 | 0.470247  0.466236  0.413823        0.316969 |  722.79402   812.79086  689.559816  415.359329  143.831832 |  100.0 |  622.79402 | 143.831832 |
    | 05:00 | 9.139567  9.189149  8.203832              6.25855 | 554.751903  560.039502  459.571733  289.725497 | 108.205658  108.711297   98.662994        78.824938 | 106.395668  106.891487   97.038315     77.585501 | 1.798602  1.804319  1.688791  1.446678 | 1.339056  1.338999  1.340313          1.344577 | -0.280433    -0.2783  -0.322287       -0.4205 | 0.398718  0.403797  0.452377      0.553336 | 0.872829  0.876868  0.886461      0.905793 | 689.426711 | 745.135325  754.660717  579.059321          311.725884 |  170.20398 | 0.096709  0.097021   0.09072       0.077467 | 0.465262  0.467558  0.421811        0.330487 | 689.426711  804.280661   709.02488  451.631237   170.20398 |  100.0 | 589.426711 |  170.20398 |
    | 05:30 | 9.019068  9.199635  8.355049             6.531967 | 542.003966  561.160899  474.359936  311.312464 | 106.976808  108.818232  100.205114        81.613254 | 105.190681  106.996346   98.550488     80.319669 | 1.784668  1.805527  1.706795  1.482029 | 1.339199  1.338987  1.340085          1.343782 | -0.285653  -0.277849  -0.315312     -0.405643 | 0.402724  0.402453  0.444678      0.538141 | 0.872833  0.876264  0.884883      0.902703 | 655.199767 | 722.295367  756.684775  604.167195          343.339972 | 199.381888 |  0.09595  0.097087  0.091703       0.079407 | 0.459679  0.468043  0.428849        0.343429 | 655.199767  792.491457    724.7843  486.395168  199.381888 |  100.0 | 555.199767 | 199.381888 |
    | 06:00 | 8.887147  9.193106  8.486337             6.791652 | 528.214162  560.462628  487.384614  332.507461 | 105.631477  108.751658  101.543994        84.261532 | 103.871475  106.931064   99.863368      82.91652 | 1.769345  1.804775  1.722343  1.515166 | 1.339361  1.338994  1.339895            1.3431 | -0.291424   -0.27813  -0.309324     -0.391878 | 0.407271   0.40183   0.43796      0.523914 |  0.87291   0.87576  0.883478      0.899808 | 620.561767 | 697.790409  755.424304  626.499299          375.105399 | 230.765266 | 0.095115  0.097046  0.092551       0.081224 | 0.453564  0.467741  0.434955        0.355684 | 620.561767  777.771123  736.862967  519.216443  230.765266 |  100.0 | 520.561767 | 230.765266 |
    | 06:30 |  8.74519  9.170637  8.598158             7.035717 | 513.569559  558.062493  498.613985  353.042308 | 104.183786  108.522516  102.684352        86.750515 | 102.451897  106.706372  100.981581     85.357168 | 1.752776  1.802186  1.735525  1.545941 | 1.339543   1.33902   1.33974          1.342515 | -0.297699  -0.279095  -0.304273     -0.379234 | 0.412316  0.401885  0.432175       0.51073 | 0.873057   0.87535  0.882233       0.89713 | 585.920022 | 671.999903  751.095708  645.914413          406.537274 | 263.727396 | 0.094211  0.096905   0.09327        0.08291 | 0.446978    0.4667  0.440151        0.367171 | 585.920022  760.477815  745.342307  549.736153  263.727396 |  100.0 | 485.920022 | 263.727396 |
    | 07:00 | 8.594549   9.13332  8.691115             7.262838 | 498.249603  554.087455  508.044139  372.686692 | 102.647547  108.141952  103.632331         89.06671 | 100.945491  106.333198  101.911151     87.628385 | 1.735101  1.797881  1.746443  1.574274 | 1.339744  1.339063  1.339614          1.342016 | -0.304435  -0.280703  -0.300108     -0.367711 | 0.417817  0.402575  0.427274      0.498618 |  0.87327  0.875027  0.881135      0.894678 |  551.63713 | 645.282291   743.94049  662.332342          437.186397 | 297.643833 | 0.093247   0.09667  0.093866       0.084461 | 0.439983  0.464972  0.444468        0.377836 |  551.63713  740.971724  750.351626  577.670353  297.643833 |  100.0 |  451.63713 | 297.643833 |
    | 07:30 | 8.436541  9.082254  8.765918             7.472165 | 482.424211  548.670548  515.695321  391.248755 | 101.036172  107.621186   104.39517        91.201434 |  99.365408  105.822545  102.659176     89.721653 | 1.716455  1.791982  1.755201  1.600138 | 1.339966  1.339123  1.339516          1.341589 | -0.311588   -0.28291  -0.296779     -0.357291 | 0.423733  0.403858  0.423206      0.487581 | 0.873546  0.874785  0.880172      0.892448 | 518.029154 | 617.970552  734.217475  675.728322          466.649613 | 331.915055 |  0.09223  0.096349  0.094343       0.085876 |  0.43264  0.462607   0.44794        0.387647 | 518.029154  719.607964  752.059583   602.80704  331.915055 |  100.0 | 418.029154 | 331.915055 |
    | 08:00 | 8.272434  9.018537  8.823355             7.663232 | 466.252304  541.948142  521.608314  408.574127 |  99.362602  106.971396  104.980921        93.149942 |  97.724337  105.185374  103.233552     91.632323 | 1.696972  1.784607   1.76191  1.623545 | 1.340208  1.339199  1.339442          1.341226 | -0.319112  -0.285676  -0.294235     -0.347943 | 0.430021  0.405693  0.419921      0.477599 | 0.873883  0.874618  0.879333      0.890432 | 485.365238 |  590.36904  722.195771  686.126733          494.576445 | 365.983315 | 0.091166  0.095947  0.094709       0.087156 | 0.425005  0.459655  0.450605        0.396587 | 485.365238  696.730612  750.665996    625.0015  365.983315 |  100.0 | 385.365238 | 365.983315 |
    | 08:30 | 8.103445  8.943252  8.864275             7.835887 | 449.880747  534.057546  525.840994  424.543892 |  97.639245  106.203631  105.398224        94.910677 |  96.034447  104.432518  103.642751     93.358866 | 1.676781  1.775871  1.766681  1.644538 |  1.34047  1.339291   1.33939          1.340917 | -0.326965  -0.288962   -0.29243     -0.339624 | 0.436642  0.408043  0.417373       0.46864 | 0.874278  0.874522  0.878607      0.888618 |  453.86844 | 562.751483   708.14853  693.594492          520.672515 | 399.344293 | 0.090064   0.09547  0.094969       0.088303 | 0.417134  0.456165  0.452503        0.404654 |  453.86844  672.667864  746.394175  644.170556  399.344293 |  100.0 |  353.86844 | 399.344293 |
    | 09:00 | 7.930733  8.857461  8.889563             7.990225 | 433.443657  525.134954  528.465118   439.07186 |  95.877928  105.328729  105.656113        96.484627 |  94.307333  103.574606  103.895632     94.902251 | 1.656007  1.765887  1.769627  1.663179 | 1.340753  1.339398  1.339358          1.340655 | -0.335104   -0.29273  -0.291318     -0.332288 | 0.443556  0.410869  0.415516      0.460661 | 0.874728  0.874492  0.877985      0.886992 | 423.717566 | 535.360001  692.347568   698.23439          544.700315 |  431.55428 | 0.088929  0.094926   0.09513       0.089321 | 0.409081  0.452187  0.453676        0.411856 | 423.717566  647.728272  739.483886  660.286121   431.55428 |  100.0 | 323.717566 |  431.55428 |
    | 09:30 | 7.755399    8.7622  8.900129             8.126541 | 417.062036  515.313719  529.563376  452.101437 |   94.08986  104.357255  105.763859         97.87478 |  92.553988  102.621998  104.001286     96.265408 |  1.63477  1.754766  1.770856  1.679548 | 1.341058  1.339521  1.339345          1.340434 | -0.343486  -0.296944  -0.290853     -0.325886 | 0.450725  0.414136  0.414303      0.453616 | 0.875232  0.874523   0.87746       0.88554 | 395.049792 | 508.405017  675.058813  700.178602          566.477913 | 462.233722 | 0.087769   0.09432  0.095197       0.090215 | 0.400895  0.447767  0.454166         0.41821 | 395.049792  622.197956  730.185039  673.368423  462.233722 |  100.0 | 295.049792 | 462.233722 |
    | 10:00 | 7.578477  8.658474  8.896891             8.245279 | 400.843683  504.722923  529.226723  463.602319 |    92.2856  103.299451  105.730843        99.085678 |  90.784766  101.584735  103.968911     97.452791 | 1.613185  1.742613   1.77048  1.693736 | 1.341383  1.339658  1.339349          1.340249 | -0.352071  -0.301567  -0.290995     -0.320367 | 0.458109  0.417809  0.413695      0.447456 | 0.875786  0.874612  0.877022      0.884248 | 367.963873 | 482.065892  656.538584  699.582521          585.876103 | 491.067842 | 0.086589  0.093657  0.095176        0.09099 | 0.392623  0.442952  0.454016        0.423741 | 367.963873  596.338725  718.752118  683.479171  491.067842 |  100.0 | 267.963873 | 491.067842 |
    | 10:30 | 7.400936  8.547251  8.880771             8.347001 |  384.88335  493.486234  527.551997  473.567171 |  90.475039  102.165196  105.566446       100.123045 |  89.009365  100.472507  103.807707     98.470013 | 1.591364  1.729531  1.768603   1.70584 |  1.34173  1.339809  1.339369          1.340097 | -0.360815  -0.306567  -0.291704     -0.315681 | 0.465672  0.421855   0.41365      0.442131 | 0.876388  0.874756  0.876666      0.883103 | 342.523746 | 456.492164  637.030628  696.619034          602.814464 | 517.805057 | 0.085396  0.092943  0.095074        0.09165 | 0.384311  0.437786  0.453268        0.428475 | 342.523746  570.386975  705.439383  690.714868  517.805057 |  100.0 | 242.523746 | 517.805057 |
    | 11:00 | 7.223682  8.429461  8.852679             8.432355 | 369.263099  481.721019  524.639826  482.008394 |  88.667386  100.963976  105.279967       100.993487 |  87.236816   99.294614  103.526791     99.323552 |  1.56941  1.715617   1.76533   1.71596 | 1.342098  1.339976  1.339404          1.339972 | -0.369681   -0.31191  -0.292941     -0.311779 | 0.473378  0.426244   0.41413      0.437591 | 0.877037  0.874951  0.876384      0.882092 | 318.762374 |  431.80523   616.76388  691.473321          617.256708 | 542.253764 | 0.084195  0.092184  0.094896       0.092203 |    0.376  0.432311  0.451965        0.432445 | 318.762374  544.553288  690.496802  695.200443  542.253764 |  100.0 | 218.762374 | 542.253764 |
    | 11:30 | 7.047547  8.305995  8.813513              8.50205 |  354.05281  469.537683  520.592797  488.955065 |  86.871161    99.70486  104.880554       101.704242 |  85.475471   98.059949  103.135134    100.020504 | 1.547424  1.700966  1.760762  1.724199 | 1.342488  1.340157  1.339454          1.339873 | -0.378627  -0.317565   -0.29467     -0.308612 |  0.48119  0.430943  0.415099      0.433788 |  0.87773  0.875194  0.876173      0.881204 |  296.68568 |  408.10037  595.950855  684.338206          629.205644 | 564.277999 | 0.082991  0.091384  0.094647       0.092652 | 0.367727  0.426567  0.450148        0.435685 |  296.68568  519.022592  674.166705  697.083297  564.277999 |  100.0 |  196.68568 | 564.277999 |
    | 12:00 | 6.873301  8.177698   8.76415             8.556839 |  339.31082  457.039218  515.513896  494.450088 |  85.094188   98.396486  104.377146       102.262982 |  83.733007   96.776983  102.641502    100.568393 |   1.5255  1.685669  1.754994  1.730661 | 1.342898  1.340353  1.339518          1.339797 | -0.387617  -0.323501  -0.296857     -0.306134 | 0.489076  0.435923  0.416522      0.430676 | 0.878464  0.875484  0.876025      0.880428 | 276.276442 | 385.448996  574.786611  675.410084          638.697992 | 583.792382 |  0.08179  0.090549  0.094332       0.093005 |  0.35953  0.420594  0.447858        0.438231 | 276.276442  493.954812  656.681098  696.527847  583.792382 |  100.0 | 176.276442 | 583.792382 |
    | 12:30 | 6.701642  8.045375  8.705441             8.597501 | 325.084653  444.320923  509.505179  498.547591 |  83.343606   97.047049  103.778432       102.677647 |  82.016419    95.45375  102.054415    100.975006 | 1.503727  1.669812  1.748122  1.735448 | 1.343328  1.340564  1.339595          1.339741 | -0.396612  -0.329689  -0.299469     -0.304303 | 0.497001  0.441157  0.418368      0.428211 | 0.879236  0.875817  0.875938      0.879755 | 257.498082 | 363.901014  553.448205  664.885412          645.799246 | 600.756667 | 0.080597  0.089684  0.093957       0.093266 |  0.35144  0.414428  0.445133         0.44012 | 257.498082  469.485886  638.259582  693.710608  600.756667 |  100.0 | 157.498082 | 600.756667 |
    | 13:00 | 6.533204  7.909785  8.638212             8.624827 | 311.411814   431.47028  502.666669  501.310568 |  81.625867   95.664297  103.092819       102.956318 |  80.332038   94.097851  101.382116    101.248266 | 1.482187  1.653478  1.740234  1.738661 | 1.343778  1.340789  1.339685          1.339703 | -0.405577  -0.336099  -0.302475     -0.303075 | 0.504934  0.446616  0.420604      0.426349 | 0.880046   0.87619  0.875907      0.879175 | 240.298231 | 343.487248  532.094567  652.957756          650.598734 | 615.170163 | 0.079416  0.088791  0.093527       0.093441 | 0.343488  0.408104  0.442011         0.44139 | 240.298231  445.729084  619.107829  688.815827  615.170163 |  100.0 | 140.298231 | 615.170163 |
    | 13:30 | 6.368553  7.771645  8.563254             8.639613 | 298.320616  418.566963  495.095452  502.808786 |   79.94675   94.255534  102.328403       103.107113 |  78.685528   92.716445  100.632545    101.396133 | 1.460959  1.636744  1.731416  1.740399 | 1.344247  1.341029  1.339787          1.339683 | -0.414477  -0.342705  -0.305845     -0.302412 | 0.512843  0.452276  0.423202      0.425051 | 0.880888  0.876603  0.875928      0.878682 | 224.612057 | 324.221832  510.866723  639.815349          653.204963 |  627.06621 | 0.078251  0.087877  0.093046       0.093536 | 0.335699  0.401654  0.438529        0.442076 | 224.612057  422.776529  599.416529  682.031672   627.06621 |  100.0 | 124.612057 |  627.06621 |
    | 14:00 | 6.208191  7.631626  8.481332             8.642652 |  285.83103  405.682945  486.884962  503.116926 |  78.311372   92.827617  101.492957         103.1381 |  77.081908   91.316258   99.813322    101.426518 | 1.440113  1.619686  1.721752  1.740756 | 1.344733  1.341284  1.339902          1.339679 | -0.423279  -0.349478  -0.309551     -0.302276 | 0.520701  0.458109  0.426133      0.424278 | 0.881762  0.877052  0.875997      0.878268 | 210.365288 | 306.104527  489.888294  625.639123          653.741329 | 636.506858 | 0.077107  0.086945  0.092519       0.093556 | 0.328099  0.395109  0.434722        0.442217 | 210.365288  400.700859  579.360777  673.546951  636.506858 |  100.0 | 110.365288 | 636.506858 |
    | 14:30 | 6.052556  7.490358  8.393174             8.634722 | 273.955531  392.882706  478.124416  502.312968 |  76.724205   91.386966  100.593912       103.057234 |  75.525563   89.903582   98.931736    101.347222 | 1.419714  1.602375  1.711318  1.739824 | 1.345236  1.341553  1.340028           1.33969 | -0.431953  -0.356394  -0.313566     -0.302631 | 0.528479  0.464093  0.429371      0.423996 | 0.882664  0.877535  0.876112      0.877928 |  197.47692 | 289.122932  469.266228  610.601192          652.342224 | 643.577847 | 0.075986  0.085998  0.091949       0.093505 | 0.320708  0.388499  0.430623        0.441849 |  197.47692  379.556975  559.099811  663.548351  643.577847 |  100.0 |   97.47692 | 643.577847 |
    | 15:00 | 5.902026  7.348429  8.299473             8.616588 | 262.699926  380.223502  468.898399  500.476789 |  75.189089   89.939568   99.638354       102.872303 |  74.020257   88.484293   97.994734    101.165882 | 1.399822  1.584878  1.700191  1.737693 | 1.345753  1.341837  1.340167          1.339714 | -0.440469  -0.363427  -0.317865     -0.303445 |  0.53615  0.470203  0.432891      0.424171 |  0.88359  0.878051   0.87627      0.877654 | 185.861607 | 273.254535  449.091688  594.863703          649.149557 | 648.383966 | 0.074893  0.085041  0.091342       0.093388 | 0.313544   0.38185  0.426264        0.441007 | 185.861607  359.383816  538.777055  652.218142  648.383966 |  100.0 |  85.861607 | 648.383966 |
    | 15:30 | 5.756916  7.206385  8.200893             8.588994 | 252.064158  367.755691  459.286578  497.688968 |  73.709255   88.490995   98.633022       102.590893 |  72.569161   87.063849   97.008925    100.889937 | 1.380489  1.567258   1.68844  1.734447 | 1.346283  1.342136  1.340317          1.339752 | -0.448799  -0.370553  -0.322423     -0.304685 |  0.54369  0.476417   0.43667      0.424769 | 0.884538  0.878598  0.876467      0.877442 | 175.431719 | 258.468613  429.441067   578.57805            644.3097 | 651.044809 | 0.073829  0.084077  0.090701       0.093211 | 0.306624  0.375188  0.421674        0.439725 | 175.431719  340.206113  518.520423   639.73232  651.044809 |  100.0 |  75.431719 | 651.044809 |
    | 16:00 | 5.617487  7.064731  8.098057             8.552659 | 242.043071  355.523097   449.36351  494.029777 |  72.287347   87.046404   97.584303       102.220351 |  71.174865   85.647311   95.980572    100.526591 | 1.361763  1.549576  1.676135  1.730168 | 1.346823  1.342449  1.340479          1.339802 | -0.456919  -0.377748  -0.327217     -0.306323 | 0.551076  0.482713  0.440683      0.425763 | 0.885504  0.879173  0.876703      0.877288 | 166.099094 | 244.727941  410.377081  561.884376           637.97084 | 651.690965 | 0.072799  0.083109  0.090029       0.092978 | 0.299961  0.368535  0.416883        0.438037 | 166.099094  322.036109  498.442811  626.259151  651.690965 |  100.0 |  66.099094 | 651.690965 |
    | 16:30 | 5.483941  6.923935   7.99156              8.50828 | 232.627131  343.563405  439.198556  489.578338 |   70.92544   85.610559   96.498241       101.767771 |  69.839406    84.23935     94.9156      100.0828 | 1.343684  1.531889   1.66334  1.724934 | 1.347372  1.342776  1.340652          1.339864 | -0.464806   -0.38499  -0.332224     -0.308329 | 0.558286  0.489071   0.44491      0.427123 | 0.886484  0.879776  0.876974      0.877188 | 157.776495 | 231.990332  391.949899  544.911334          630.280722 | 650.460645 | 0.071803   0.08214   0.08933       0.092692 | 0.293565  0.361913  0.411919        0.435975 | 157.776495  304.875194  478.642758  611.958062  650.460645 |  100.0 |  57.776495 | 650.460645 |
    | 17:00 | 5.356429  6.784426   7.88196             8.456525 | 223.803099  331.908569  428.855867  484.411931 |  69.625073   84.187841   95.380536        101.23997 |  68.564291    82.84426     93.8196     99.565248 | 1.326286  1.514249  1.650115  1.718819 | 1.347927  1.343117  1.340836          1.339937 |  -0.47244  -0.392256  -0.337424     -0.310678 | 0.565303  0.495471   0.44933      0.428824 | 0.887474  0.880404  0.877278      0.877137 | 150.378786 | 220.209985  374.198298  527.776065          621.384763 |  647.49672 | 0.070844  0.081174  0.088608       0.092359 | 0.287445  0.355343  0.406805        0.433569 | 150.378786  288.715454  459.205217  596.978846   647.49672 |  100.0 |  50.378786 |  647.49672 |
    | 17:30 | 5.235052  6.646597  7.769784             8.398034 | 215.554658  320.585234  418.394435  478.605434 |  68.387269   82.782258   94.236556       100.643484 |  67.350525   81.465972   92.697835     98.980345 | 1.309599  1.496707  1.636518  1.711894 | 1.348485  1.343472  1.341032          1.340021 | -0.479803  -0.399526  -0.342794     -0.313344 | 0.572108  0.501893  0.453922      0.430839 | 0.888471  0.881055  0.877613      0.877132 | 143.823868 | 209.338663  357.150794  510.584351          611.424492 | 642.944166 | 0.069924  0.080212  0.087865       0.091981 | 0.281607  0.348843  0.401567        0.430849 | 143.823868  273.541114  440.202417  581.461135  642.944166 |  100.0 |  43.823868 | 642.944166 |
    | 18:00 | 5.119864  6.510807  7.655525             8.333421 | 207.862978  309.615145  407.868197   472.23089 |   67.21257   81.397464   93.071344        99.984557 |  66.198637    80.10807   91.555251     98.334215 | 1.293643   1.47931  1.622605  1.704227 | 1.349044  1.343841  1.341239          1.340116 |  -0.48688  -0.406779  -0.348317     -0.316304 | 0.578686  0.508319  0.458669      0.433146 | 0.889471  0.881728  0.877978       0.87717 | 138.033385 | 199.326703  340.826755  493.430914          600.536311 | 636.947894 | 0.069043  0.079258  0.087104       0.091562 | 0.276055  0.342429  0.396227        0.427843 | 138.033385  259.329855   421.69477  565.534115  636.947894 |  100.0 |  38.033385 | 636.947894 |
    | 18:30 | 5.010871   6.37738  7.539648              8.26327 | 200.707237  299.015568  397.326186  465.357164 |  66.101063   80.036769   91.889625        99.269146 |  65.108715   78.773799   90.396481     97.632696 | 1.278436  1.462102  1.608427   1.69588 | 1.349601  1.344221  1.341458          1.340221 | -0.493659  -0.413996  -0.353972     -0.319535 | 0.585024  0.514731  0.463552      0.435722 | 0.890468  0.882421  0.878371      0.877248 | 132.933235 | 190.123869  325.237462  476.399831          588.850518 | 629.650923 | 0.068203  0.078314  0.086329       0.091107 |  0.27079  0.336117  0.390807        0.424578 | 132.933235  246.054029  403.731814  549.316427  629.650923 |  100.0 |  32.933235 | 629.650923 |
    | 19:00 | 4.908043  6.246608  7.422586             8.188134 | 194.065089  288.799687  386.812712  458.049703 |  65.052416   78.703153   90.695823         98.50291 |  64.080432   77.466081   89.225861      96.88134 | 1.263988  1.445123  1.594033  1.686916 | 1.350153  1.344614  1.341687          1.340337 |  -0.50013  -0.421158  -0.359742     -0.323016 |  0.59111  0.521113  0.468554      0.438546 | 0.891459  0.883131   0.87879      0.877364 | 128.453911 | 181.680059  310.387121  459.565033          576.490586 | 621.192879 | 0.067405  0.077382  0.085542       0.090618 | 0.265813  0.329921  0.385325         0.42108 | 128.453911  233.681736  386.353159  532.916244  621.192879 |  100.0 |  28.453911 | 621.192879 |
    | 19:30 | 4.811309  6.118753  7.304744             8.108542 | 187.913079  278.976988  376.367567  450.370369 |  64.065912   77.399282   89.494062        97.691225 |  63.113086    76.18753   88.047438     96.085417 | 1.250303  1.428411  1.579471  1.677392 | 1.350698  1.345018  1.341927          1.340462 | -0.506286  -0.428248   -0.36561     -0.326727 | 0.596937  0.527449  0.473658      0.441598 | 0.892441  0.883858  0.879233      0.877515 | 124.530701 | 173.945887  296.273804  442.990866          563.572639 | 611.708784 | 0.066648  0.076464  0.084745       0.090097 | 0.261121  0.323853  0.379802        0.417372 | 124.530701  222.177798  369.589423  516.431475  611.708784 |  100.0 |  24.530701 | 611.708784 |
    | 20:00 | 4.720563  5.994046  7.186498             8.024992 | 182.227013  269.553633  366.026256  442.377333 |  63.140485   76.127516   88.288189        96.839178 |  62.205629   74.940461   86.864982     95.249917 | 1.237382  1.412001  1.564783  1.667362 | 1.351233  1.345433  1.342179          1.340597 | -0.512123  -0.435248  -0.371558     -0.330648 | 0.602496  0.533723   0.47885      0.444859 | 0.893408    0.8846  0.879699      0.877699 | 121.103761 | 166.873145  282.890332  426.732699          550.205134 |   601.3281 | 0.065934  0.075562  0.083941        0.08955 |  0.25671  0.317925  0.374254        0.413478 | 121.103761   211.50461  353.463144  499.950082    601.3281 |  100.0 |  21.103761 |   601.3281 |
    | 20:30 | 4.635671   5.87269    7.0682             7.937956 | 176.982286  260.532805  355.820232  434.125041 |  62.274753   74.889924   87.081777        95.951582 |  61.356709   73.726902   85.681997     94.379557 | 1.225219  1.395926  1.550011  1.656879 | 1.351755  1.345858  1.342441          1.340741 |  -0.51764  -0.442143  -0.377571     -0.334761 | 0.607784  0.539921  0.484113      0.448312 | 0.894357  0.885354  0.880187      0.877914 | 118.118088 | 160.415166  270.225075  410.837555          536.488681 | 590.174011 |  0.06526  0.074679  0.083133       0.088977 | 0.252576  0.312146  0.368698        0.409418 | 118.118088  201.622886  337.989656   483.55048  590.174011 |  100.0 |  18.118088 | 590.174011 |
    | 21:00 |  4.55647  5.754861  6.950173             7.847879 | 172.154159  251.915041  345.777147   425.66421 |  61.467061   73.688297   85.878139        95.032976 |  60.564702    72.54861   84.501733      93.47879 | 1.213803  1.380214  1.535194   1.64599 | 1.352262  1.346291  1.342714          1.340895 | -0.522836  -0.448918  -0.383634      -0.33905 | 0.612797   0.54603  0.489434       0.45194 | 0.895284  0.886119  0.880695      0.878158 | 115.523427 | 154.527091  258.262698  395.344753           522.51601 | 578.362897 | 0.064628  0.073814  0.082321       0.088382 | 0.248712  0.306526  0.363148        0.405214 | 115.523427   192.49231  323.177925  467.302002  578.362897 |  100.0 |  15.523427 | 578.362897 |
    | 21:30 | 4.482775  5.640708   6.83272             7.755181 | 167.718009  243.698537  335.921094  417.041873 |  60.715519   72.524158   84.680343        94.087638 |  59.827754   71.407078   83.327198      92.55181 |  1.20312  1.364892  1.520369  1.634743 | 1.352753  1.346731  1.342997          1.341058 | -0.527714  -0.455559  -0.389731     -0.343497 | 0.617534  0.552037  0.494798      0.455726 | 0.896185  0.886893  0.881221      0.878429 | 113.274094 | 149.166072  246.984817  380.286551           508.37204 | 566.003986 | 0.064036  0.072971  0.081509       0.087768 | 0.245109  0.301072  0.357619        0.400885 | 113.274094  184.072085  309.031326  451.265411  566.003986 |  100.0 |  13.274094 | 566.003986 |
    | 22:00 | 4.414381  5.530355  6.716117             7.660256 |  163.64953  235.879434  326.272863   408.30146 |  60.018033   71.398773    83.49122        93.119591 |  59.143813   70.303547   82.161166     91.602561 | 1.193152  1.349983   1.50557  1.623182 | 1.353225  1.347178  1.343291          1.341231 | -0.532281  -0.462053  -0.395848     -0.348087 | 0.621997   0.55793  0.500193      0.459656 | 0.897058  0.887673  0.881765      0.878725 | 111.328777 | 144.291395  236.370603  365.688769          494.134047 | 553.199158 | 0.063483   0.07215  0.080698       0.087136 |  0.24176   0.29579  0.352123        0.396448 | 111.328777  176.321409  295.548376  435.493434  553.199158 |  100.0 |  11.328777 | 553.199158 |
    | 22:30 | 4.351067  5.423902  6.600619             7.563475 | 159.924917  228.452086   316.85018  399.482893 |   59.37235   70.313163   82.313375        92.132613 |  58.510669   69.239019   81.006195      90.63475 | 1.183877  1.335509   1.49083  1.611348 | 1.353677   1.34763  1.343595          1.341412 | -0.536542  -0.468388  -0.401972     -0.352804 | 0.626189  0.563699  0.505607      0.463714 |   0.8979  0.888459  0.882326      0.879046 | 109.650285 | 139.864564  226.397314  351.571411          479.871893 | 540.042868 | 0.062969  0.071352   0.07989       0.086489 | 0.238655  0.290685  0.346672        0.391921 | 109.650285  169.199867  282.723409  420.031329  540.042868 |  100.0 |   9.650285 | 540.042868 |
    | 23:00 | 4.292599  5.321427  6.486463             7.465185 | 156.521003  221.409301  307.667949  390.622713 |   58.77609   69.268116   81.149202        91.130248 |  57.925988   68.214266   79.864629      89.65185 | 1.175272  1.321487  1.476179   1.59928 | 1.354108  1.348085  1.343909          1.341602 | -0.540507  -0.474554  -0.408089     -0.357636 | 0.630114  0.569332  0.511026      0.467887 | 0.898708  0.889247  0.882901      0.879389 | 108.205287 | 135.849332  217.040767  337.949242          465.648323 | 526.622182 | 0.062491  0.070579  0.079086       0.085829 | 0.235783  0.285763  0.341278        0.387321 | 108.205287  162.667758  270.547191   404.91744  526.622182 |  100.0 |   8.205287 | 526.622182 |
    | 23:30 | 4.238735  5.222984  6.373862             7.365712 |  153.41538  214.742569   298.73848  381.754216 |  58.226782   68.264195   80.000889        90.115815 |  57.387347    67.22984   78.738616     88.657117 | 1.167308  1.307932  1.461646  1.587014 | 1.354517  1.348543  1.344232          1.341802 | -0.544185  -0.480541  -0.414188     -0.362567 | 0.633779  0.574822   0.51644      0.472162 |  0.89948  0.890036  0.883489      0.879754 | 106.964036 | 132.211695  208.275748  324.832351          451.519296 | 513.016893 | 0.062049  0.069832  0.078289       0.085158 | 0.233133  0.281026   0.33595         0.38266 | 106.964036  156.686363  259.007482  390.183759  513.016893 |  100.0 |   6.964036 | 513.016893 |
    | 00:00 | 4.189227  5.128609  6.263011             7.265359 | 150.586493  208.442264  290.071708  372.907602 |  57.721896   67.301751   78.870432        89.092416 |  56.892265   66.286087   77.630111     87.653591 | 1.159958  1.294859  1.447259  1.574587 | 1.354904  1.349001  1.344564           1.34201 | -0.547587   -0.48634  -0.420255     -0.367584 | 0.637189   0.58016  0.521837      0.476527 | 0.900215  0.890823   0.88409      0.880138 | 105.900084 | 128.919864  200.076377  312.226675          437.534351 | 499.299712 | 0.061641   0.06911  0.077499       0.084478 | 0.230694  0.276477  0.330698        0.377955 | 105.900084  151.218161  248.089545  375.856477  499.299712 |  100.0 |   5.900084 | 499.299712 |
    | 00:30 | 4.143823  5.038315  6.154089             7.164412 | 148.013711  202.497836  281.675409  364.110134 |  57.258872   66.380936   77.759642        88.062948 |  56.438234   65.383153   76.540892     86.644115 | 1.153191  1.282275  1.433041  1.562031 | 1.355267  1.349459  1.344905          1.342227 | -0.550726  -0.491945   -0.42628     -0.372676 | 0.640355  0.585338  0.527208      0.480969 | 0.900911  0.891607  0.884702      0.880541 |  104.99001 | 125.944213  192.416419  300.134488          423.736988 | 485.536517 | 0.061264  0.068415  0.076718       0.083791 | 0.228454  0.272117  0.325531        0.373217 |  104.99001   146.22699  237.776601  361.956514  485.536517 |  100.0 |    4.99001 | 485.536517 |
    | 01:00 | 4.102274    4.9521  6.047256             7.063134 | 145.677378  196.897978   273.55539  355.386298 |  56.835152    65.50171   76.670156        87.030114 |  56.022741   64.521001   75.472563     85.631338 | 1.146976  1.270191  1.419016  1.549376 | 1.355608  1.349914  1.345254          1.342453 | -0.553614  -0.497349  -0.432251      -0.37783 | 0.643285  0.590351  0.532542      0.485477 | 0.901568  0.892386  0.885325      0.880961 | 104.213141 | 123.257205   185.26955  288.554864          410.165069 | 471.786633 | 0.060919  0.067748  0.075948       0.083098 | 0.226402  0.267946  0.320455         0.36846 | 104.213141  141.678184  228.050232  348.500027  471.786633 |  100.0 |   4.213141 | 471.786633 |
    | 01:30 |  4.06433  4.869942  5.942657             6.961773 |  143.55885  191.630782  265.715684   346.75797 |  56.448197   64.663855   75.603444        85.996428 |  55.643301   63.699416   74.426566     84.617725 | 1.141281  1.258609  1.405207  1.536654 | 1.355925  1.350365  1.345611          1.342687 | -0.556266  -0.502547  -0.438158     -0.383035 |  0.64599  0.595192  0.537829      0.490041 | 0.902186  0.893157  0.885956      0.881398 | 103.551298 | 120.833311   178.60958  277.484089          396.851211 | 458.103159 | 0.060602  0.067108  0.075189       0.082401 | 0.224525  0.263966  0.315479        0.363694 | 103.551298  137.538654  218.890743  335.498901  458.103159 |  100.0 |   3.551298 | 458.103159 |
    | 02:00 | 4.029747  4.791802  5.840418             6.860557 | 141.640513  186.683883  258.158717  338.244574 |  56.095518   63.866987   74.560816        84.964229 |  55.297471   62.918023   73.404185     83.605571 | 1.136074  1.247533  1.391633   1.52389 | 1.356221  1.350811  1.345974          1.342929 | -0.558693  -0.507536  -0.443991      -0.38828 |  0.64848  0.599857  0.543062       0.49465 | 0.902765  0.893919  0.886594      0.881849 | 102.988541 | 118.648921   172.41065  266.916054          383.823187 | 444.533304 | 0.060313  0.066495  0.074442       0.081702 | 0.222813  0.260173  0.310608         0.35893 | 102.988541   133.77696  210.277481  322.961202  444.533304 |  100.0 |   2.988541 | 444.533304 |
    | 02:30 | 3.998287  4.717629  5.740656               6.7597 | 139.905791  182.044579  250.885477   329.86325 |  55.774685   63.110569   73.543431        83.935688 |  54.982868   62.176294   72.406557     82.597004 | 1.131324  1.236963  1.378313  1.511111 | 1.356494  1.351251  1.346345           1.34318 | -0.560911  -0.512313   -0.44974     -0.393554 | 0.650767  0.604343  0.548231      0.499295 | 0.903305  0.894669  0.887239      0.882315 |  102.51094 | 116.682248  166.647382  256.842607          371.104312 | 431.118745 | 0.060048   0.06591   0.07371       0.081002 | 0.221255  0.256567  0.305848        0.354178 |  102.51094   130.36334  202.189107  310.891607  431.118745 |  100.0 |    2.51094 | 431.118745 |
    | 03:00 | 3.969719  4.647356  5.643468               6.6594 | 138.339137  177.699949  243.895663  321.629011 |  55.483351   62.393921   72.552305        82.912817 |  54.697191   61.473563   71.434678     81.593996 |    1.127  1.226897  1.365264  1.498342 | 1.356747  1.351682  1.346721          1.343439 | -0.562933  -0.516877  -0.455398     -0.398847 | 0.652862  0.608648  0.553328      0.503965 | 0.903807  0.895406  0.887889      0.882794 | 102.106358 | 114.913226  161.295013  247.253868          358.713826 | 417.895987 | 0.059808  0.065353  0.072991       0.080302 | 0.219838  0.253145  0.301203        0.349447 | 102.106358  127.269725   194.60384  299.291807  417.895987 |  100.0 |   2.106358 | 417.895987 |
    | 03:30 | 3.943823  4.580904  5.548941             6.559837 | 136.926026  173.636956  237.187825  313.554896 |  55.219257   61.716236   71.588314        81.897478 |  54.438227   60.809038   70.489407     80.598374 |  1.12307  1.217332  1.352501  1.485605 |  1.35698  1.352104  1.347102          1.343706 | -0.564772  -0.521228  -0.460954      -0.40415 | 0.654777  0.612769  0.558345      0.508652 | 0.904272  0.896127  0.888543      0.883286 | 101.764251 | 113.323411  156.329491  238.138524          346.667254 | 404.896726 | 0.059589  0.064823  0.072289       0.079603 | 0.218552  0.249904   0.29668        0.344746 | 101.764251  124.469736  187.499666  288.160872  404.896726 |  100.0 |   1.764251 | 404.896726 |
    | 04:00 | 3.920386  4.518181  5.457147             6.461182 | 135.652922  169.842532  230.759494  305.652117 |  54.980249   61.076589   70.652201        80.891389 |   54.20386   60.181813   69.571473     79.611823 | 1.119506   1.20826  1.340039  1.472923 | 1.357193  1.352515  1.347487           1.34398 | -0.566442  -0.525365  -0.466402     -0.409453 | 0.656524  0.616705  0.563276      0.513348 | 0.904702  0.896832  0.889199      0.883788 |  101.47548 | 111.895882  151.727562  229.484085          334.976754 | 392.148208 | 0.059391  0.064321  0.071602       0.078908 | 0.217388  0.246841   0.29228        0.340082 |  101.47548  121.938659   180.85452  277.495594  392.148208 |  100.0 |    1.47548 | 392.148208 |
    | 04:30 | 3.899208  4.459088  5.368148              6.36359 | 134.507258  166.303665  224.607301  297.930207 |   54.76428   60.473956   69.744585        79.896135 |  53.992084   59.590883   68.681482     78.635896 |  1.11628  1.199673  1.327891  1.460316 | 1.357388  1.352915  1.347875          1.344262 | -0.567956  -0.529292  -0.471735     -0.414747 | 0.658114  0.620457  0.568115      0.518044 | 0.905097  0.897518  0.889857      0.884301 | 101.232151 | 110.615142  147.466825  221.277119          323.651449 | 379.673581 | 0.059211  0.063845  0.070932       0.078216 | 0.216335   0.24395  0.288008        0.335464 | 101.232151  119.653415  174.646433  267.290795  379.673581 |  100.0 |   1.232151 | 379.673581 |
    | 05:00 |   3.8801  4.403515  5.281992             6.267202 | 133.477395  163.007467  218.727085  290.397149 |  54.569414   59.907221   68.865962        78.913173 |  53.801003   59.035153   67.819921     77.672022 | 1.113363  1.191563  1.316067  1.447804 | 1.357566  1.353302  1.348266          1.344551 | -0.569325   -0.53301  -0.476945     -0.420025 | 0.659558  0.624027  0.572854      0.522732 | 0.905461  0.898185  0.890516      0.884823 | 101.027458 |  109.46702  143.525777  213.503454          312.697738 | 367.492232 | 0.059048  0.063395   0.07028       0.077529 | 0.215385  0.241228  0.283866        0.330897 | 101.027458  117.592512  168.853669  257.539611  367.492232 |  100.0 |   1.027458 | 367.492232 |
    | 05:30 | 3.862883  4.351346  5.198716             6.172151 | 132.552588  159.941233  213.113999   283.05951 |  54.393835   59.375193   68.016712        77.943837 |  53.628833   58.513457   66.987163     76.721511 |  1.11073  1.183918  1.304578  1.435404 | 1.357729  1.353676  1.348659          1.344847 | -0.570562  -0.536524  -0.482026     -0.425277 | 0.660867  0.627414  0.577488      0.527405 | 0.905794   0.89883  0.891173      0.885353 | 100.855551 | 108.438582  139.883841  206.148367          302.119582 | 355.620121 | 0.058902  0.062971  0.069647       0.076848 | 0.214528  0.238669  0.279857        0.326388 | 100.855551  115.735993   163.45483  248.233751  355.620121 |  100.0 |   0.855551 | 355.620121 |
    | 06:00 | 3.847391  4.302456  5.118347             6.078556 | 131.722936  157.092503  207.762594  275.922558 |   54.23584   58.876619   67.197104        76.989352 |  53.473907   58.024564   66.183471      75.78556 | 1.108358  1.176725  1.293432  1.423133 | 1.357876  1.354035  1.349052           1.34515 | -0.571677  -0.539837  -0.486974     -0.430495 | 0.662052  0.630622  0.582012      0.532056 | 0.906098  0.899453  0.891828      0.885891 | 100.711411 |  107.51804  136.521382  199.196742          291.918782 | 344.070084 |  0.05877  0.062572  0.069032       0.076174 | 0.213756  0.236267  0.275982        0.321943 | 100.711411  114.065373  158.428951  239.363723  344.070084 |  100.0 |   0.711411 | 344.070084 |
    | 06:30 | 3.833467   4.25672  5.040901             5.986526 | 130.979347  154.449099  202.666908   268.99038 |  54.093845   58.410192   66.407302        76.050829 |  53.334669   57.567195   65.409007     74.865263 | 1.106223  1.169971  1.282637  1.411008 | 1.358009   1.35438  1.349446          1.345459 | -0.572681  -0.542954  -0.491784     -0.435673 | 0.663123  0.633654  0.586421      0.536678 | 0.906375  0.900053   0.89248      0.886436 | 100.590745 | 106.694665  133.419714  192.633212          282.095224 | 332.852139 | 0.058651  0.062197  0.068435       0.075508 | 0.213063  0.234018  0.272242        0.317567 | 100.590745  112.563577  153.755573  230.919043  332.852139 |  100.0 |   0.590745 | 332.852139 |
    | 07:00 | 3.820968  4.214003  4.966383             5.896161 | 130.313486  151.999173  197.820546  262.265986 |  53.966378   57.974571   65.647368         75.12928 |  53.209677   57.140034    64.66383      73.96161 | 1.104304   1.16364  1.272197  1.399043 |  1.35813  1.354709  1.349837          1.345774 | -0.573585  -0.545882   -0.49645     -0.440803 | 0.664089  0.636515  0.590712      0.541265 | 0.906627  0.900628  0.893128      0.886986 | 100.489885 | 105.958707  130.561096  186.442284          272.647114 | 321.973758 | 0.058544  0.061845  0.067859        0.07485 |  0.21244  0.231915  0.268638        0.313264 | 100.489885  111.214868  149.414807   222.88842  321.973758 |  100.0 |   0.489885 | 321.973758 |
    | 07:30 |  3.80976  4.174174  4.894791              5.80755 |  129.71773  149.731233  193.216743  255.751408 |  53.852076   57.568386   64.917267        74.225618 |  53.097595   56.741736   63.947907     73.075496 | 1.102582  1.157717  1.262119  1.387253 | 1.358239  1.355024  1.350227          1.346095 | -0.574396  -0.548626   -0.50097     -0.445879 | 0.664959  0.639207   0.59488       0.54581 | 0.906855   0.90118  0.893769      0.887542 | 100.405708 | 105.301319  127.928719  180.608447          263.571188 | 311.440134 | 0.058448  0.061516  0.067302       0.074201 | 0.211881  0.229951   0.26517         0.30904 | 100.405708  110.004779   145.38738  215.259922  311.440134 |  100.0 |   0.405708 | 311.440134 |
    | 08:00 | 3.799719  4.137095  4.826112             5.720773 | 129.185124  147.634173  188.848434  249.447798 |  53.749686   57.190251   64.216874        73.340664 |  52.997193   56.370945   63.261115     72.207728 | 1.101037  1.152186  1.252403  1.375649 | 1.358338  1.355322  1.350614           1.34642 | -0.575124  -0.551193   -0.50534     -0.450893 | 0.665741  0.641736  0.598922      0.550308 | 0.907062  0.901706  0.894404      0.888102 |  100.33556 | 104.714485  125.506694   175.11626          254.862913 | 301.254422 | 0.058362  0.061209  0.066764       0.073563 |  0.21138  0.228122  0.261839        0.304898 |  100.33556  108.920034   141.65468  208.021117  301.254422 |  100.0 |    0.33556 | 301.254422 |
    | 08:30 | 3.790734   4.10263  4.760324             5.635902 | 128.709334  145.697297  184.708314  243.355505 |  53.658054   56.838778   63.545975        72.475151 |  52.907341   56.026297   62.603245     71.359023 | 1.099654  1.147029  1.243054  1.364245 | 1.358426  1.355605  1.350996          1.346751 | -0.575776   -0.55359  -0.509559      -0.45584 | 0.666444  0.644107  0.602835      0.554752 | 0.907249  0.902207   0.89503      0.888665 | 100.277186 | 104.190948  123.280027  169.950437          246.516656 | 291.417967 | 0.058284  0.060922  0.066247       0.072935 | 0.210932  0.226419  0.258643        0.300842 | 100.277186  107.948482  138.198777  201.159204  291.417967 |  100.0 |   0.277186 | 291.417967 |
    | 09:00 |   3.7827  4.070643  4.697401             5.553002 | 128.284602   143.91033  180.788885  237.474163 |  53.576123   56.512579   62.904277        71.629728 |  52.827001   55.706432   61.974008     70.530017 | 1.098416   1.14223  1.234071  1.353051 | 1.358505  1.355872  1.351374          1.347085 | -0.576359  -0.555824  -0.513623     -0.460714 | 0.667075  0.646326  0.606618      0.559138 | 0.907417  0.902682  0.895647      0.889231 | 100.228681 | 103.724149  121.234594  165.095912          238.525852 | 281.930514 | 0.058215  0.060655   0.06575       0.072319 | 0.210531  0.224838  0.255583        0.296874 | 100.228681  107.079021  135.002451  194.661128  281.930514 |  100.0 |   0.228681 | 281.930514 |
    | 09:30 | 3.775523  4.041001  4.637304             5.472127 | 127.905704  142.263433  177.082511  231.802758 |  53.502929   56.210281   62.291409        70.804962 |  52.755228   55.410005   61.373042     69.721268 | 1.097309   1.13777  1.225453  1.342077 | 1.358577  1.356124  1.351745          1.347423 | -0.576881  -0.557902  -0.517533      -0.46551 |  0.66764  0.648399  0.610268      0.563461 | 0.907569  0.903133  0.896254      0.889798 | 100.188431 | 103.308168  119.357116  160.537893          230.883147 | 272.790401 | 0.058154  0.060407  0.065273       0.071714 | 0.210173  0.223371  0.252655        0.292999 | 100.188431  106.301525  132.049202  188.513672  272.790401 |  100.0 |   0.188431 | 272.790401 |
    | 10:00 | 3.769116  4.013569  4.579991             5.393326 | 127.567908  140.747209  173.581462    226.3397 |   53.43759   55.930532   61.706929        70.001347 |  52.691158   55.135689   60.799912     68.933258 |  1.09632  1.133633    1.2172  1.331334 | 1.358641  1.356361   1.35211          1.347764 | -0.577348  -0.559833  -0.521288     -0.470221 | 0.668147  0.650331  0.613784      0.567716 | 0.907706  0.903559   0.89685      0.890367 | 100.155078 | 102.937665  117.635123  156.261914          223.580529 | 263.994736 | 0.058099  0.060177  0.064816       0.071122 | 0.209853  0.222012   0.24986        0.289217 | 100.155078  105.606779  129.323261  182.703548  263.994736 |  100.0 |   0.155078 | 263.994736 |
    | 10:30 | 3.763401  3.988218  4.525412              5.31664 | 127.266935  139.352708  170.277951   221.08288 |  53.379307   55.672007   61.150329        69.219299 |  52.634007   54.882184    60.25412     68.166396 | 1.095438  1.129801  1.209308   1.32083 | 1.358698  1.356583  1.352468          1.348107 | -0.577764  -0.561623  -0.524887     -0.474844 | 0.668599   0.65213  0.617166      0.571899 | 0.907829   0.90396  0.897433      0.890935 | 100.127477 | 102.607832   116.05693  152.253868          216.609446 | 255.539557 | 0.058049  0.059964  0.064379       0.070543 | 0.209567  0.220755  0.247194        0.285533 | 100.127477  104.986411  126.809588  177.217468  255.539557 |  100.0 |   0.127477 | 255.539557 |
    | 11:00 | 3.758306  3.964823  4.473511             5.242103 | 126.998921  138.071427  167.164176  216.029732 |  53.327353   55.433415   60.621041        68.459166 |  52.583062   54.648225   59.735111     67.421025 | 1.094651  1.126258  1.201773  1.310571 | 1.358749  1.356791  1.352817          1.348453 | -0.578136   -0.56328  -0.528331     -0.479373 | 0.669004  0.653801  0.620413      0.576005 | 0.907939  0.904337  0.898003      0.891503 | 100.104666 | 102.314343  114.611597  148.500043          209.960915 | 247.419981 | 0.058005  0.059766  0.063961       0.069977 | 0.209313  0.219595  0.244656        0.281946 | 100.104666  104.432824  124.493873  172.042211  247.419981 |  100.0 |   0.104666 | 247.419981 |
    | 11:30 | 3.753768  3.943258  4.424227             5.169742 | 126.760382  136.895307  164.232348  211.177281 |  53.281071   55.213503   60.118443        67.721229 |  52.537678   54.432583   59.242273     66.697418 |  1.09395  1.122985   1.19459  1.300566 | 1.358795  1.356985  1.353157          1.348799 | -0.578467  -0.564812  -0.531622     -0.483805 | 0.669365  0.655352  0.623526       0.58003 | 0.908037   0.90469  0.898559       0.89207 | 100.085838 | 102.053308  113.288898  144.987142          203.625613 | 239.630339 | 0.057966  0.059584  0.063563       0.069425 | 0.209086  0.218524  0.242243         0.27846 | 100.085838   103.93914  122.362528  167.164674  239.630339 |  100.0 |   0.085838 | 239.630339 |
    | 12:00 | 3.749727  3.923407  4.377495             5.099579 | 126.548182  135.816728  161.474728  206.522196 |  53.239865    55.01106   59.641863        67.005703 |  52.497273   54.234072   58.774948     65.995787 | 1.093325  1.119966  1.187754  1.290821 | 1.358835  1.357166  1.353487          1.349147 | -0.578762  -0.566227   -0.53476     -0.488136 | 0.669687  0.656788  0.626505      0.583971 | 0.908126  0.905022    0.8991      0.892634 | 100.070317 | 101.821236  112.079289  141.702306          197.593959 | 232.164289 | 0.057932  0.059416  0.063184       0.068887 | 0.208884  0.217538  0.239952        0.275076 | 100.070317  103.499137  120.402677  162.571924  232.164289 |  100.0 |   0.070317 | 232.164289 |
    | 12:30 | 3.746132  3.905155  4.333243             5.031628 |   126.3595  134.828499  158.883648  202.060826 |  53.203199   54.824922   59.190586        66.312739 |  52.461319   54.051549   58.332435     65.316281 | 1.092769  1.117186  1.181258  1.281341 | 1.358872  1.357333  1.353808          1.349494 | -0.579024   -0.56753  -0.537748     -0.492362 | 0.669974  0.658115  0.629351      0.587826 | 0.908205  0.905331  0.899625      0.893195 | 100.057539 | 101.614996  110.973872  138.633125          191.856195 | 225.014933 | 0.057901  0.059261  0.062823       0.068364 | 0.208704  0.216631   0.23778        0.271793 | 100.057539  103.107199  118.602143   158.25123  225.014933 |  100.0 |   0.057539 | 225.014933 |
    | 13:00 | 3.742934  3.888392  4.291399             4.965899 | 126.191801  133.923851   156.45154  197.789247 |   53.17059   54.653975    58.76386        65.642431 |  52.429343   53.883922   57.913995      64.65899 | 1.092274  1.114629  1.175095  1.272129 | 1.358904  1.357489  1.354118           1.34984 | -0.579258  -0.568731  -0.540589      -0.49648 |  0.67023  0.659342  0.632066      0.591589 | 0.908275  0.905619  0.900134      0.893752 | 100.047032 |  101.43178  109.964362  135.767644          186.402447 | 218.174909 | 0.057873  0.059119  0.062481       0.067855 | 0.208545  0.215797  0.235724        0.268614 | 100.047032  102.758258  116.949431  154.190108  218.174909 |  100.0 |   0.047032 | 218.174909 |
    | 13:30 | 3.740092  3.873015  4.251886             4.902395 | 126.042813  133.096426  154.170958  193.703298 |  53.141603   54.497153   58.360899        64.994814 |  52.400919   53.730145   57.518859     64.023948 | 1.091834   1.11228  1.169256  1.263191 | 1.358933  1.357633  1.354417          1.350185 | -0.579466  -0.569834  -0.543285     -0.500488 | 0.670458  0.660472  0.634651       0.59526 | 0.908338  0.905887  0.900626      0.894304 | 100.038402 | 101.269076  109.043055  133.094377          181.222781 | 211.636477 | 0.057848  0.058988  0.062157       0.067361 | 0.208403  0.215032   0.23378        0.265539 | 100.038402  102.447753  115.433709  150.376338  211.636477 |  100.0 |   0.038402 | 211.636477 |
    | 14:00 | 3.737566  3.858923  4.214623             4.841114 | 125.910498   132.34026  152.034593   189.79861 |  53.115846   54.353444   57.980892        64.369866 |  52.375662   53.589227   57.146232     63.411136 | 1.091443  1.110124  1.163732   1.25453 | 1.358959  1.357766  1.354705          1.350528 |  -0.57965  -0.570847  -0.545839     -0.504383 | 0.670661  0.661513  0.637109      0.598836 | 0.908394  0.906135  0.901101      0.894851 | 100.031322 | 101.124634  108.202797  130.602298           176.30726 | 205.391599 | 0.057827  0.058868   0.06185       0.066882 | 0.208276  0.214331  0.231945        0.262567 | 100.031322  102.171578  114.044792  146.797995  205.391599 |  100.0 |   0.031322 | 205.391599 |
    | 14:30 | 3.735323  3.846022   4.17953             4.782048 |  125.79303  131.649773  150.035298  186.070641 |  53.092969   54.221886   57.623009        63.767513 |  52.353229   53.460223   56.795299     62.820481 | 1.091096  1.108148  1.158515  1.246146 | 1.358981  1.357889  1.354981          1.350869 | -0.579815  -0.571776  -0.548256     -0.508162 | 0.670841   0.66247  0.639442      0.602314 | 0.908444  0.906366  0.901559      0.895392 | 100.025521 | 100.996444  107.436948  128.280849          171.645982 | 199.432007 | 0.057807  0.058758   0.06156       0.066418 | 0.208164  0.213688  0.230215        0.259699 | 100.025521  101.926047  112.773115  143.443466  199.432007 |  100.0 |   0.025521 | 199.432007 |
    | 15:00 | 3.733331  3.834224  4.146523             4.725186 | 125.688779  131.019754  148.166095  182.514705 |  53.072658   54.101571   57.286401        63.187631 |  52.333312   53.342245   56.465227      62.25186 | 1.090788  1.106339  1.153594  1.238042 | 1.359002  1.358002  1.355246          1.351206 |  -0.57996  -0.572627  -0.550539     -0.511825 | 0.671001   0.66335  0.641653      0.605693 | 0.908489  0.906579  0.901998      0.895926 | 100.020773 | 100.882712  106.739358   126.11993          167.229121 |  193.74926 |  0.05779  0.058657  0.061287        0.06597 | 0.208065    0.2131  0.228587        0.256935 | 100.020773  101.707849  111.609714  140.301464   193.74926 |  100.0 |   0.020773 |  193.74926 |
    | 15:30 | 3.731564  3.823445  4.115518              4.67051 | 125.596286  130.445341  146.420194  179.125995 |  53.054631   53.991643    56.97021        62.630048 |  52.315636   53.234451   56.155177     61.705104 | 1.090514  1.104685  1.148959  1.230219 |  1.35902  1.358106  1.355499          1.351539 |  -0.58009  -0.573406  -0.552692     -0.515369 | 0.671144  0.664156  0.643746       0.60897 | 0.908528  0.906776   0.90242      0.896453 | 100.016892 | 100.781833  106.104333    124.1099           163.04696 | 188.334803 | 0.057775  0.058565  0.061029       0.065537 | 0.207976  0.212563  0.227056        0.254273 | 100.016892  101.514017  110.546201  137.361045  188.334803 |  100.0 |   0.016892 | 188.334803 |
    | 16:00 | 3.729995  3.813605   4.08643                4.618 | 125.514248  129.922009  144.791001   175.89961 |  53.038636   53.891296   56.673573        62.094543 |  52.299952   53.136054     55.8643     61.179998 | 1.090271  1.103173    1.1446  1.222677 | 1.359036  1.358202   1.35574          1.351867 | -0.580204  -0.574117   -0.55472     -0.518795 |  0.67127  0.664895  0.645723      0.612146 | 0.908564  0.906958  0.902823      0.896971 | 100.013722 | 100.692378  105.526609  122.241561           159.08992 | 183.180008 | 0.057761  0.058481  0.060787       0.065119 | 0.207898  0.212073  0.225618        0.251714 | 100.013722  101.341891   109.57474  134.611611  183.180008 |  100.0 |   0.013722 | 183.180008 |
    | 16:30 | 3.728604  3.804631  4.059175             4.567628 | 125.441501  129.445555   143.27213  172.830575 |   53.02445   53.799776   56.395625        61.580853 |   52.28604   53.046311   55.591749     60.676284 | 1.090055  1.101793  1.140506  1.215415 |  1.35905  1.358289   1.35597           1.35219 | -0.580306  -0.574767  -0.556627     -0.522101 | 0.671382  0.665571  0.647588      0.615219 | 0.908595  0.907125  0.903208       0.89748 | 100.011136 | 100.613072  105.001323  120.506158          155.348584 | 178.276218 | 0.057749  0.058404  0.060559       0.064717 | 0.207828  0.211625   0.22427        0.249256 | 100.011136  101.189093   108.68802  132.042924  178.276218 |  100.0 |   0.011136 | 178.276218 |
    | 17:00 |  3.72737  3.796453  4.033668             4.519366 | 125.377011  129.012076  141.857408  169.913863 |   53.01187   53.716377   56.135504        61.088676 |  52.273705   52.964531    55.33668     60.193665 | 1.089864  1.100534  1.136665  1.208431 | 1.359063   1.35837  1.356187          1.352508 | -0.580397   -0.57536  -0.558418     -0.525287 | 0.671482  0.666189  0.649345      0.618188 | 0.908623  0.907279  0.903575      0.897979 | 100.009028 | 100.542779  104.523989  118.895364          151.813716 | 173.614785 | 0.057738  0.058334  0.060346        0.06433 | 0.207767  0.211217  0.223008        0.246898 | 100.009028  101.053496  107.879232  129.645107  173.614785 |  100.0 |   0.009028 | 173.614785 |
    | 17:30 | 3.726277  3.789007  4.009825              4.47318 | 125.319851  128.617958  140.540881  167.144413 |  53.000717   53.640438   55.892356        60.617667 |  52.262769   52.890067   55.098254     59.731803 | 1.089695  1.099388  1.133068  1.201725 | 1.359074  1.358443  1.356394          1.352819 | -0.580477  -0.575901  -0.560097     -0.528353 |  0.67157  0.666754  0.650998      0.621052 | 0.908648   0.90742  0.903925      0.898469 | 100.007313 | 100.480488  104.090471  117.401271          148.476279 | 169.187095 | 0.057729   0.05827  0.060145       0.063958 | 0.207712  0.210846  0.221826         0.24464 | 100.007313  100.933198  107.142046  127.408648  169.187095 |  100.0 |   0.007313 | 169.187095 |
    | 18:00 | 3.725308  3.782232  3.987564             4.429033 | 125.269201  128.259855   139.31682   164.51715 |  52.990833   53.571346   55.665338        60.167449 |  52.253077   52.822316   54.875645     59.290328 | 1.089545  1.098343  1.129703  1.195292 | 1.359084   1.35851  1.356589          1.353123 | -0.580548  -0.576393  -0.561669       -0.5313 | 0.671649  0.667269   0.65255      0.623813 |  0.90867  0.907549  0.904256      0.898947 | 100.005918 | 100.425298  103.696962  116.016377          145.327452 |   164.9846 | 0.057721  0.058211  0.059958       0.063602 | 0.207663  0.210507  0.220723        0.242478 | 100.005918  100.826502  106.470579  125.324401    164.9846 |  100.0 |   0.005918 |   164.9846 |
    | 18:30 | 3.724449  3.776072  3.966804             4.386884 | 125.224327  127.934679  138.179722  162.026997 |  52.982074    53.50853    55.45362        59.737611 |  52.244488    52.76072   54.668039     58.868837 | 1.089412  1.097394  1.126558   1.18913 | 1.359093  1.358571  1.356773           1.35342 | -0.580611  -0.576841   -0.56314     -0.534128 | 0.671718  0.667739  0.654006       0.62647 | 0.908689  0.907668  0.904571      0.899414 | 100.004785 | 100.376408   103.33996  114.733573          142.358637 |  160.99884 | 0.057713  0.058158  0.059783        0.06326 |  0.20762    0.2102  0.219693        0.240412 | 100.004785  100.731896   105.85938  123.383586   160.99884 |  100.0 |   0.004785 |  160.99884 |
    | 19:00 | 3.723688  3.770476  3.947464              4.34669 | 125.184578  127.639579  137.124313  159.668893 |  52.974315    53.45146    55.25639         59.32771 |  52.236879   52.704759   54.474639     58.466895 | 1.089294   1.09653  1.123623  1.183235 |   1.3591  1.358627  1.356947           1.35371 | -0.580666  -0.577249  -0.564513     -0.536839 |  0.67178  0.668167   0.65537      0.629023 | 0.908706  0.907776  0.904868       0.89987 | 100.003865 | 100.333107  103.016244  113.546135          139.561472 | 157.221461 | 0.057707   0.05811   0.05962       0.062933 | 0.207582  0.209921  0.218733         0.23844 | 100.003865  100.648029  105.303395  121.577787  157.221461 |  100.0 |   0.003865 | 157.221461 |
    | 19:30 | 3.723014  3.765395  3.929467             4.308404 | 125.149374  127.371928  136.145547  157.437808 |  52.967442   53.399646   55.072855        58.937276 |   52.23014   52.653951   54.294668     58.084044 | 1.089189  1.095746  1.120888  1.177602 | 1.359107  1.358678   1.35711          1.353991 | -0.580716  -0.577619  -0.565795     -0.539433 | 0.671835  0.668556  0.656646      0.631474 | 0.908722  0.907875  0.905149      0.900313 | 100.003119 | 100.294761   102.72286  112.447702          136.927835 | 153.644235 | 0.057701  0.058067  0.059467        0.06262 | 0.207549  0.209667  0.217839         0.23656 | 100.003119  100.573698  104.797955   119.89895  153.644235 |  100.0 |   0.003119 | 153.644235 |
    | 20:00 | 3.722417  3.760785  3.912737              4.27198 | 125.118201  127.129309  135.238608  155.328752 |  52.961355   53.352635   54.902244        58.565816 |  52.224171   52.607853    54.12737     57.719798 | 1.089097  1.095034  1.118342  1.172227 | 1.359113  1.358724  1.357264          1.354264 |  -0.58076  -0.577955  -0.566989     -0.541912 | 0.671883  0.668911  0.657839      0.633822 | 0.908735  0.907965  0.905413      0.900743 | 100.002515 | 100.260809  102.457095  111.432274          134.449853 | 150.259074 | 0.057696  0.058027  0.059326       0.062322 | 0.207519  0.209437  0.217008        0.234769 | 100.002515  100.507833  104.338742  118.339378  150.259074 |  100.0 |   0.002515 | 150.259074 |
    | 20:30 | 3.721889  3.756605  3.897201             4.237365 |   125.0906  126.909498  134.398904   153.33679 |  52.955965   53.310008    54.74381        58.212814 |  52.218885   52.566053   53.972012     57.373651 | 1.089015  1.094388  1.115973  1.167105 | 1.359119  1.358766  1.357407          1.354529 | -0.580798   -0.57826    -0.5681     -0.544279 | 0.671926  0.669232  0.658951       0.63607 | 0.908747  0.908048  0.905662      0.901161 | 100.002026 |  100.23075  102.216465  110.494186          132.119903 | 147.058042 | 0.057691  0.057991  0.059194       0.062038 | 0.207492  0.209228  0.216235        0.233065 | 100.002026  100.449481  103.921774  116.891726  147.058042 |  100.0 |   0.002026 | 147.058042 |
    | 21:00 | 3.721421  3.752817  3.882789             4.204508 | 125.066166  126.710457  133.622067  151.457052 |  52.951193   53.271379    54.59683        57.877736 |  52.214206   52.528175   53.827886     57.045079 | 1.088942  1.093803  1.113773   1.16223 | 1.359124  1.358804  1.357541          1.354784 | -0.580833  -0.578536  -0.569132     -0.546535 | 0.671964  0.669524  0.659988      0.638219 | 0.908758  0.908123  0.905896      0.901565 | 100.001631 | 100.204141  101.998693  109.628105          129.930615 | 144.033364 | 0.057687  0.057958  0.059071       0.061767 | 0.207469  0.209039  0.215519        0.231447 | 100.001631  100.397793  103.543381  115.548994  144.033364 |  100.0 |   0.001631 | 144.033364 |
    | 21:30 | 3.721006  3.749387  3.869431             4.173354 | 125.044538  126.530313  132.903947  149.684741 |  52.946968   53.236394   54.460609        57.560029 |  52.210064   52.493869    53.69431     56.733542 | 1.088878  1.093272  1.111732  1.157595 | 1.359128  1.358839  1.357667           1.35503 | -0.580863  -0.578787  -0.570091     -0.548683 | 0.671997  0.669789  0.660952      0.640271 | 0.908767  0.908191  0.906116      0.901956 | 100.001312 |  100.18059  101.801698  108.829011          127.874874 |  141.17744 | 0.057683  0.057929  0.058958       0.061509 | 0.207448  0.208867  0.214854        0.229911 | 100.001312  100.352017  103.200185  114.304524   141.17744 |  100.0 |   0.001312 |  141.17744 |
    | 22:00 |  3.72064  3.746282  3.857063             4.143849 | 125.025395  126.367352  132.240612  148.015148 |  52.943229   53.204726   54.334478        57.259132 |  52.206397   52.462816   53.570629     56.438488 | 1.088821  1.092792  1.109839  1.153195 | 1.359132   1.35887  1.357784          1.355268 |  -0.58089  -0.579013  -0.570981     -0.550725 | 0.672027   0.67003   0.66185      0.642226 | 0.908776  0.908254  0.906322      0.902333 | 100.001054 | 100.159746  101.623575  108.092184           125.94582 | 138.482845 |  0.05768  0.057902  0.058852       0.061265 |  0.20743  0.208712  0.214238        0.228455 | 100.001054  100.311481  102.889079  113.151987  138.482845 |  100.0 |   0.001054 | 138.482845 |
    | 22:30 | 3.720315  3.743472  3.845621             4.115935 | 125.008454  126.220005  131.628338  146.443652 |   52.93992   53.176076   54.217797         56.97447 |  52.203152   52.434722   53.456214     56.159354 | 1.088771  1.092358  1.108087  1.149022 | 1.359135  1.358899  1.357893          1.355495 | -0.580914  -0.579219  -0.571805     -0.552663 | 0.672054  0.670247  0.662683      0.644089 | 0.908783   0.90831  0.906514      0.902696 | 100.000846 | 100.141301  101.462585  107.413191          124.136845 | 135.942344 | 0.057677  0.057878  0.058754       0.061033 | 0.207414  0.208572  0.213668        0.227077 | 100.000846  100.275592  102.607209  112.085378  135.942344 |  100.0 |   0.000846 | 135.942344 |
    | 23:00 | 3.720028  3.740932  3.835047             4.089557 | 124.993463  126.086835  131.063606  144.965733 |  52.936991   53.150169   54.109954        56.705461 |   52.20028   52.409319   53.350465     55.895569 | 1.088726  1.091964  1.106465  1.145069 | 1.359138  1.358924  1.357994          1.355714 | -0.580935  -0.579404  -0.572567     -0.554502 | 0.672077  0.670445  0.663456      0.645859 |  0.90879  0.908361  0.906693      0.903045 | 100.000679 | 100.124979  101.317142  106.787873          122.441593 | 133.548889 | 0.057675  0.057856  0.058664       0.060813 | 0.207399  0.208445  0.213141        0.225773 | 100.000679  100.243821  102.351957  111.099008  133.548889 |  100.0 |   0.000679 | 133.548889 |
    | 23:30 | 3.719774  3.738636  3.825281             4.064656 | 124.980199  125.966532  130.543095  143.576977 |    52.9344   53.126755   54.010366         56.45152 |  52.197739   52.386359   53.252811     55.646559 | 1.088687  1.091609  1.104967   1.14133 | 1.359141  1.358948  1.358089          1.355923 | -0.580953  -0.579572  -0.573273     -0.556243 | 0.672097  0.670623  0.664172      0.647541 | 0.908796  0.908408  0.906861      0.903381 | 100.000544 | 100.110538  101.185795  106.212333          120.853958 |  131.29563 | 0.057673  0.057836  0.058581       0.060605 | 0.207387   0.20833  0.212655        0.224541 | 100.000544  100.215698  102.120922  110.187493   131.29563 |  100.0 |   0.000544 |  131.29563 |
    | 00:00 | 3.719549  3.736562  3.816271             4.041175 | 124.968463  125.857898  130.063677  142.273084 |  52.932107   53.105603   53.918479        56.212057 |  52.195491   52.365618   53.162708     55.411747 | 1.088652  1.091288  1.103583  1.137796 | 1.359143  1.358969  1.358176          1.356123 |  -0.58097  -0.579724  -0.573924      -0.55789 | 0.672116  0.670785  0.664835      0.649136 | 0.908801   0.90845  0.907017      0.903702 | 100.000436 | 100.097762  101.067227   105.68292          119.368079 | 129.175913 | 0.057671  0.057818  0.058503       0.060408 | 0.207375  0.208226  0.212206        0.223379 | 100.000436  100.190807  101.911905  109.345745  129.175913 |  100.0 |   0.000436 | 129.175913 |
    | 00:30 |  3.71935  3.734689  3.807964             4.019055 | 124.958081  125.759843  129.622407  141.049867 |  52.930078   53.086504   53.833765        55.986485 |  52.193502    52.34689    53.07964     55.190555 | 1.088621  1.090998  1.102306  1.134461 | 1.359145  1.358988  1.358257          1.356314 | -0.580984  -0.579861  -0.574526     -0.559446 | 0.672132  0.670931  0.665448      0.650647 | 0.908805  0.908488  0.907162       0.90401 | 100.000349 | 100.086459  100.960234  105.196221          117.978338 | 127.183285 | 0.057669  0.057802  0.058432       0.060223 | 0.207365  0.208133  0.211791        0.222284 | 100.000349  100.168779  101.722896   108.56896  127.183285 |  100.0 |   0.000349 | 127.183285 |
    | 01:00 | 3.719174  3.732999  3.800312             3.998241 | 124.948896  125.671371   129.21652  139.903263 |  52.928284   53.069265   53.755727        55.774216 |  52.191742   52.329986   53.003117     54.982409 | 1.088594  1.090736  1.101128  1.131317 | 1.359147  1.359005  1.358332          1.356495 | -0.580997  -0.579985  -0.575081     -0.560915 | 0.672146  0.671063  0.666015      0.652076 | 0.908809  0.908523  0.907297      0.904304 | 100.000279 | 100.076461  100.863722  104.749046          116.679354 | 125.311496 | 0.057668  0.057787  0.058367       0.060048 | 0.207357  0.208048   0.21141        0.221252 | 100.000279  100.149287  101.552054  107.852613  125.311496 |  100.0 |   0.000279 | 125.311496 |
    | 01:30 | 3.719019  3.731474  3.793268             3.978674 | 124.940771  125.591576  128.843421  138.829333 |  52.926696   53.053713   53.683893        55.574671 |  52.190185   52.314735   52.932678     54.786738 |  1.08857    1.0905  1.100044  1.128357 | 1.359148  1.359021  1.358401          1.356668 | -0.581009  -0.580096  -0.575592     -0.562299 | 0.672159  0.671182  0.666538      0.653427 | 0.908813  0.908554  0.907422      0.904585 | 100.000223 | 100.067616  100.776696  104.338416          115.465981 |   123.5545 | 0.057666  0.057774  0.058306       0.059883 | 0.207349  0.207972  0.211058        0.220282 | 100.000223  100.132039  101.397701  107.192441    123.5545 |  100.0 |   0.000223 |   123.5545 |
    | 02:00 | 3.718881  3.730098  3.786789             3.960298 | 124.933585  125.519634  128.500679  137.824266 |  52.925292   53.039687    53.61782        55.387273 |  52.188808   52.300982   52.867888     54.602979 | 1.088548  1.090287  1.099046  1.125571 |  1.35915  1.359035  1.358465          1.356832 | -0.581019  -0.580197  -0.576062     -0.563602 |  0.67217  0.671289  0.667021      0.654702 | 0.908816  0.908583  0.907538      0.904852 | 100.000178 | 100.059793   100.69825  103.961554            114.3333 | 121.906452 | 0.057665  0.057762  0.058251       0.059728 | 0.207342  0.207903  0.210735         0.21937 | 100.000178   100.11678  101.258303  106.584438  121.906452 |  100.0 |   0.000178 | 121.906452 |
    | 02:30 | 3.718759  3.728858  3.780834             3.943058 | 124.927229  125.454796  128.186019   136.88438 |   52.92405   53.027043   53.557089        55.211455 |   52.18759   52.288583   52.808336     54.430576 | 1.088529  1.090095  1.098128  1.122954 | 1.359151  1.359047  1.358524          1.356987 | -0.581028  -0.580288  -0.576495     -0.564827 |  0.67218  0.671386  0.667465      0.655904 | 0.908819  0.908608  0.907645      0.905106 | 100.000142 | 100.052874  100.627563   103.61587          113.276615 | 120.361715 | 0.057664  0.057751  0.058199       0.059583 | 0.207336  0.207841  0.210438        0.218514 | 100.000142   100.10328  101.132463  106.024842  120.361715 |  100.0 |   0.000142 | 120.361715 |
    | 03:00 | 3.718651  3.727741  3.775364             3.926898 | 124.921607   125.39638  127.897313  136.006126 |  52.922951   53.015648   53.501307        55.046662 |  52.186513    52.27741   52.753638     54.268983 | 1.088513  1.089922  1.097284  1.120497 | 1.359152  1.359059  1.358578          1.357134 | -0.581036   -0.58037  -0.576893     -0.565978 | 0.672189  0.671474  0.667874      0.657035 | 0.908821  0.908631  0.907744      0.905347 | 100.000113 | 100.046754  100.563889  103.298954          112.291449 | 118.914853 | 0.057663  0.057742  0.058152       0.059446 |  0.20733  0.207785  0.210165        0.217712 | 100.000113  100.091337  101.018905  105.510123  118.914853 |  100.0 |   0.000113 | 118.914853 |
    | 03:30 | 3.718556  3.726735  3.770343             3.911767 | 124.916635  125.343768  127.632579  135.186086 |  52.921979   53.005384   53.450105        54.892348 |   52.18556   52.267345    52.70343     54.117666 | 1.088498  1.089766   1.09651  1.118194 | 1.359153  1.359069  1.358628          1.357273 | -0.581043  -0.580443  -0.577258     -0.567058 | 0.672196  0.671553  0.668251        0.6581 | 0.908824  0.908652  0.907836      0.905576 |  100.00009 | 100.041342   100.50655  103.008566          111.373536 | 117.560634 | 0.057662  0.057733  0.058109       0.059317 | 0.207326  0.207735  0.209914         0.21696 |  100.00009  100.080773  100.916471  105.036974  117.560634 |  100.0 |    0.00009 | 117.560634 |
    | 04:00 | 3.718472  3.725828  3.765738              3.89761 | 124.912237  125.296399  127.389963  134.420975 |   52.92112   52.996141   53.403139         54.74798 |  52.184718   52.258281   52.657376     53.976101 | 1.088485  1.089625  1.095799  1.116036 | 1.359154  1.359078  1.358674          1.357403 | -0.581049   -0.58051  -0.577594      -0.56807 | 0.672203  0.671624  0.668597        0.6591 | 0.908826  0.908671   0.90792      0.905793 | 100.000072 | 100.036556  100.454932   102.74262          110.518814 | 116.294026 | 0.057662  0.057725  0.058069       0.059197 | 0.207321  0.207689  0.209684        0.216256 | 100.000072  100.071428  100.824104  104.602302  116.294026 |  100.0 |   0.000072 | 116.294026 |
    | 04:30 | 3.718397  3.725012  3.761516             3.884378 | 124.908349  125.253763  127.167744  133.707642 |   52.92036    52.98782   53.360085         54.61304 |  52.183972   52.250122   52.615158     53.843782 | 1.088473  1.089499  1.095147  1.114016 | 1.359155  1.359087  1.358717          1.357527 | -0.581054  -0.580569  -0.577902     -0.569018 | 0.672209  0.671688  0.668915      0.660039 | 0.908827  0.908688  0.907998      0.905998 | 100.000057 | 100.032323  100.408478  102.499184          109.723425 | 115.110196 | 0.057661  0.057718  0.058033       0.059085 | 0.207318  0.207649  0.209473        0.215598 | 100.000057  100.063163  100.740845  104.203215  115.110196 |  100.0 |   0.000057 | 115.110196 |
    | 05:00 | 3.718331  3.724278  3.757648             3.872021 |  124.90491  125.215401  126.964318  133.043067 |  52.919688   52.980332   53.320642        54.487025 |  52.183313    52.24278   52.576481     53.720213 | 1.088463  1.089385  1.094549  1.112128 | 1.359155  1.359094  1.358756          1.357643 | -0.581059  -0.580623  -0.578184     -0.569905 | 0.672215  0.671746  0.669207      0.660919 | 0.908829  0.908703   0.90807      0.906192 | 100.000046 |  100.02858  100.366683  102.276462          108.983701 | 114.004509 |  0.05766  0.057712     0.058        0.05898 | 0.207314  0.207612   0.20928        0.214983 | 100.000046  100.055853  100.665822  103.837011  114.004509 |  100.0 |   0.000046 | 114.004509 |
    | 05:30 | 3.718273  3.723617  3.754107             3.860492 | 124.901869  125.180892  126.778195   132.42436 |  52.919094   52.973595   53.284528        54.369446 |  52.182731   52.236174   52.541069     53.604918 | 1.088454  1.089283  1.094002  1.110364 | 1.359156  1.359101  1.358791          1.357752 | -0.581063  -0.580672  -0.578442     -0.570734 | 0.672219  0.671798  0.669475      0.661743 |  0.90883  0.908717  0.908136      0.906374 | 100.000036 |  100.02527  100.329092  102.072791          108.296161 | 112.972522 |  0.05766  0.057706  0.057969       0.058881 | 0.207312  0.207579  0.209103        0.214409 | 100.000036  100.049388  100.598243  103.501174  112.972522 |  100.0 |   0.000036 | 112.972522 |
    | 06:00 | 3.718222  3.723023  3.750866             3.849743 |  124.89918   125.14986  126.607992  131.848762 |  52.918568   52.967537   53.251482        54.259832 |  52.182215   52.230233   52.508664     53.497432 | 1.088446  1.089191  1.093501  1.108718 | 1.359157  1.359107  1.358824          1.357854 | -0.581067  -0.580715  -0.578678     -0.571508 | 0.672224  0.671845  0.669721      0.662514 | 0.908831   0.90873  0.908197      0.906546 | 100.000029 | 100.022343   100.29529   101.88663          107.657508 | 112.009987 | 0.057659  0.057701  0.057941        0.05879 | 0.207309  0.207549  0.208941        0.213874 | 100.000029   100.04367  100.537389  103.193355  112.009987 |  100.0 |   0.000029 | 112.009987 |
    | 06:30 | 3.718176  3.722489  3.747903             3.839731 | 124.896802  125.121962  126.452426  131.313641 |  52.918104   52.962089    53.22126        54.157728 |   52.18176   52.224891   52.479029     53.397311 | 1.088439  1.089108  1.093043  1.107184 | 1.359157  1.359113  1.358854          1.357949 | -0.581071  -0.580754  -0.578895     -0.572229 | 0.672227  0.671887  0.669945      0.663235 | 0.908832  0.908741  0.908253      0.906708 | 100.000023 | 100.019755  100.264905  101.716553          107.064617 | 111.112838 | 0.057659  0.057696  0.057916       0.058704 | 0.207307  0.207522  0.208793        0.213375 | 100.000023  100.038614  100.482609  102.911372  111.112838 |  100.0 |   0.000023 | 111.112838 |
    | 07:00 | 3.718136  3.722009  3.745194             3.830413 | 124.894699  125.096888  126.310309   130.81649 |  52.917693   52.957193   53.193636        54.062697 |  52.181357    52.22009   52.451942     53.304126 | 1.088433  1.089033  1.092624  1.105754 | 1.359157  1.359118  1.358881          1.358039 | -0.581074   -0.58079  -0.579093     -0.572902 | 0.672231  0.671925  0.670151      0.663908 | 0.908833  0.908751  0.908304       0.90686 | 100.000018 | 100.017466  100.237597  101.561242          106.514532 |   110.2772 | 0.057659  0.057692  0.057892       0.058624 | 0.207305  0.207498  0.208658        0.212911 | 100.000018  100.034143   100.43331  102.653193    110.2772 |  100.0 |   0.000018 |   110.2772 |
    | 07:30 |   3.7181  3.721577   3.74272             3.821746 |  124.89284  125.074358  126.180541  130.354925 |  52.917329   52.952793     53.1684        53.974319 |     52.181   52.215775   52.427195     53.217464 | 1.088427  1.088966  1.092241  1.104424 | 1.359158  1.359122  1.358906          1.358123 | -0.581076  -0.580821  -0.579274     -0.573528 | 0.672234  0.671959   0.67034      0.664536 | 0.908834   0.90876  0.908351      0.907002 | 100.000014 | 100.015443  100.213062  101.419479          106.004458 | 109.499371 | 0.057658  0.057688  0.057871        0.05855 | 0.207303  0.207477  0.208534        0.212479 | 100.000014  100.030188  100.388958  102.416932  109.499371 |  100.0 |   0.000014 | 109.499371 |
    | 08:00 | 3.718069   3.72119   3.74046             3.813693 | 124.891196  125.054119  126.062105   129.92668 |  52.917008    52.94884   53.145357        53.892193 |  52.180685   52.211899     52.4046     53.136932 | 1.088422  1.088906  1.091891  1.103186 | 1.359158  1.359126  1.358929          1.358201 | -0.581078   -0.58085  -0.579439     -0.574111 | 0.672236  0.671989  0.670512      0.665121 | 0.908835  0.908768  0.908395      0.907136 | 100.000011 | 100.013653  100.191023  101.290137          105.531756 | 108.775831 | 0.057658  0.057685  0.057852       0.058481 | 0.207301  0.207457  0.208421        0.212077 | 100.000011  100.026692  100.349068  102.200839  108.775831 |  100.0 |   0.000011 | 108.775831 |
    | 08:30 | 3.718041  3.720842  3.738398             3.806215 | 124.889743  125.035943   125.95406  129.529607 |  52.916724   52.945289   53.124327        53.815933 |  52.180407   52.208417   52.383979     53.062153 | 1.088418  1.088852  1.091572  1.102037 | 1.359158   1.35913   1.35895          1.358274 |  -0.58108  -0.580875   -0.57959     -0.574653 | 0.672238  0.672017  0.670669      0.665666 | 0.908835  0.908776  0.908434      0.907261 | 100.000009 | 100.012071  100.171231  101.172181          105.093937 | 108.103227 | 0.057658  0.057682  0.057834       0.058417 |   0.2073   0.20744  0.208318        0.211704 | 100.000009    100.0236  100.313201   102.00329  108.103227 |  100.0 |   0.000009 | 108.103227 |
    | 09:00 | 3.718016  3.720529  3.736517             3.799277 | 124.888458  125.019623   125.85554  129.161671 |  52.916473   52.942101   53.105144        53.745173 |  52.180161   52.205291   52.365168     52.992768 | 1.088414  1.088804  1.091281  1.100969 | 1.359159  1.359133  1.358969          1.358342 | -0.581082  -0.580898  -0.579727     -0.575156 |  0.67224  0.672041  0.670813      0.666173 | 0.908836  0.908782   0.90847      0.907378 | 100.000007 | 100.010673  100.153461  101.064653          104.688652 | 107.478376 | 0.057658  0.057679  0.057818       0.058358 | 0.207299  0.207424  0.208224        0.211358 | 100.000007  100.020866  100.280959  101.822781  107.478376 |  100.0 |   0.000007 | 107.478376 |
    | 09:30 | 3.717994  3.720249  3.734802             3.792843 | 124.887321  125.004973  125.765744  128.820947 |  52.916251    52.93924   53.087653        53.679563 |  52.179943   52.202485   52.348017     52.928432 | 1.088411   1.08876  1.091015  1.099978 | 1.359159  1.359136  1.358987          1.358405 | -0.581084  -0.580919  -0.579853     -0.575622 | 0.672242  0.672064  0.670944      0.666645 | 0.908837  0.908788  0.908503      0.907487 | 100.000006 | 100.009436  100.137511  100.966671          104.313692 | 106.898254 | 0.057657  0.057677  0.057803       0.058303 | 0.207298   0.20741  0.208138        0.211037 | 100.000006  100.018449  100.251985  101.657923  106.898254 |  100.0 |   0.000006 | 106.898254 |
    | 10:00 | 3.717975  3.719997  3.733239             3.786882 | 124.886317  124.991826  125.683934  128.505615 |  52.916054   52.936671   53.071713        53.618772 |   52.17975   52.199967   52.332387     52.868822 | 1.088408  1.088721  1.090773   1.09906 | 1.359159  1.359138  1.359003          1.358464 | -0.581085  -0.580937  -0.579967     -0.576055 | 0.672244  0.672084  0.671064      0.667082 | 0.908837  0.908794  0.908534      0.907589 | 100.000004 | 100.008342  100.123196  100.877425          103.966977 | 106.359998 | 0.057657  0.057675  0.057789       0.058251 | 0.207297  0.207398   0.20806         0.21074 | 100.000004  100.016311  100.225954  101.507427  106.359998 |  100.0 |   0.000004 | 106.359998 |
    | 10:30 | 3.717958  3.719771  3.731815             3.781363 | 124.885429  124.980029  125.609431  128.213959 |  52.915881   52.934366   53.057193        53.562484 |   52.17958   52.197707   52.318148     52.813627 | 1.088405  1.088686  1.090553  1.098209 | 1.359159  1.359141  1.359017          1.358519 | -0.581087  -0.580954  -0.580071     -0.576457 | 0.672245  0.672101  0.671173      0.667489 | 0.908837  0.908799  0.908561      0.907684 | 100.000004 | 100.007376  100.110353  100.796167          103.646551 | 105.860892 | 0.057657  0.057673  0.057777       0.058204 | 0.207296  0.207386  0.207989        0.210464 | 100.000004  100.014421  100.202572  101.370104  105.860892 |  100.0 |   0.000004 | 105.860892 |
    | 11:00 | 3.717943  3.719568  3.730518             3.776255 | 124.884643  124.969446   125.54161  127.944359 |  52.915727   52.932299   53.043972        53.510401 |   52.17943   52.195679   52.305183     52.762555 | 1.088403  1.088655  1.090352  1.097422 | 1.359159  1.359143   1.35903          1.358569 | -0.581088  -0.580968  -0.580166     -0.576828 | 0.672246  0.672117  0.671273      0.667866 | 0.908838  0.908803  0.908587      0.907773 | 100.000003 | 100.006521  100.098832  100.722211          103.350579 | 105.398371 | 0.057657  0.057671  0.057766        0.05816 | 0.207295  0.207376  0.207924        0.210209 | 100.000003   100.01275  100.181576  101.244859  105.398371 |  100.0 |   0.000003 | 105.398371 |
    | 11:30 |  3.71793  3.719386  3.729338             3.771533 | 124.883949  124.959955  125.479897  127.695291 |  52.915592   52.930444   53.031938        53.462239 |  52.179296   52.193861   52.293383     52.715328 | 1.088401  1.088627  1.090169  1.096693 |  1.35916  1.359145  1.359043          1.358616 | -0.581089  -0.580982  -0.580253     -0.577172 | 0.672247  0.672132  0.671364      0.668215 | 0.908838  0.908807   0.90861      0.907855 | 100.000002 | 100.005765    100.0885  100.654926          103.077336 | 104.970009 | 0.057657  0.057669  0.057756       0.058119 | 0.207294  0.207367  0.207865        0.209973 | 100.000002  100.011273  100.162725  101.130677  104.970009 |  100.0 |   0.000002 | 104.970009 |
    | 12:00 | 3.717918  3.719223  3.728265             3.767169 | 124.883335  124.951444  125.423762  127.465323 |  52.915472   52.928781   53.020989        53.417732 |  52.179179    52.19223   52.282648     52.671686 | 1.088399  1.088601  1.090003   1.09602 |  1.35916  1.359146  1.359053           1.35866 | -0.581089  -0.580994  -0.580331      -0.57749 | 0.672248  0.672145  0.671446      0.668538 | 0.908838   0.90881  0.908631      0.907932 | 100.000002 | 100.005097  100.079235  100.593734          102.825207 | 104.573516 | 0.057657  0.057668  0.057746       0.058082 | 0.207294  0.207359  0.207811        0.209755 | 100.000002  100.009966  100.145806  101.026625  104.573516 |  100.0 |   0.000002 | 104.573516 |
    | 12:30 | 3.717907  3.719077  3.727288             3.763138 | 124.882792  124.943814  125.372721  127.253107 |  52.915366   52.927291   53.011033        53.376628 |  52.179075   52.190768   52.272884      52.63138 | 1.088397  1.088579  1.089852  1.095397 |  1.35916  1.359148  1.359063            1.3587 |  -0.58109  -0.581004  -0.580403     -0.577783 | 0.672249  0.672156  0.671522      0.668838 | 0.908839  0.908813   0.90865      0.908003 | 100.000001 | 100.004506  100.070929  100.538103          102.592679 | 104.206735 | 0.057657  0.057667  0.057738       0.058047 | 0.207293  0.207352  0.207762        0.209554 | 100.000001  100.008811  100.130622  100.931845  104.206735 |  100.0 |   0.000001 | 104.206735 |
    | 13:00 | 3.717898  3.718946  3.726401             3.759418 | 124.882312  124.936975  125.326329   127.05738 |  52.915272   52.925954   53.001981         53.33869 |  52.178983   52.189458   52.264008     52.594178 | 1.088396  1.088558  1.089714  1.094823 |  1.35916  1.359149  1.359073          1.358738 | -0.581091  -0.581014  -0.580468     -0.578055 |  0.67225  0.672167   0.67159      0.669115 | 0.908839  0.908816  0.908668      0.908069 | 100.000001 | 100.003984  100.063484  100.487546          102.378335 | 103.867632 | 0.057657  0.057666   0.05773       0.058015 | 0.207293  0.207345  0.207718        0.209368 | 100.000001   100.00779     100.117  100.845546  103.867632 |  100.0 |   0.000001 | 103.867632 |
    | 13:30 |  3.71789  3.718828  3.725594             3.755986 | 124.881888  124.930846  125.284177  126.876956 |  52.915189   52.924756   52.993756        53.303694 |  52.178901   52.188283   52.255943     52.559862 | 1.088395   1.08854  1.089589  1.094293 |  1.35916   1.35915  1.359081          1.358772 | -0.581092  -0.581023  -0.580527     -0.578305 | 0.672251  0.672176  0.671652       0.66937 | 0.908839  0.908819  0.908684       0.90813 | 100.000001 | 100.003522  100.056811  100.441615           102.18085 | 103.554294 | 0.057657  0.057665  0.057723       0.057985 | 0.207292  0.207339  0.207678        0.209197 | 100.000001  100.006887   100.10478  100.766998  103.554294 |  100.0 |   0.000001 | 103.554294 |
    | 14:00 | 3.717883  3.718723  3.724862             3.752823 | 124.881513  124.925354  125.245892  126.710727 |  52.915116   52.923683   52.986284        53.271431 |   52.17883   52.187231   52.248616     52.528226 | 1.088394  1.088524  1.089476  1.093804 |  1.35916  1.359151  1.359088          1.358804 | -0.581092   -0.58103   -0.58058     -0.578536 | 0.672251  0.672184  0.671709      0.669607 | 0.908839  0.908821  0.908698      0.908187 | 100.000001 | 100.003114  100.050833  100.399901          101.998986 | 103.264924 | 0.057656  0.057664  0.057717       0.057958 | 0.207292  0.207334  0.207641        0.209039 | 100.000001  100.006089  100.093822  100.695534  103.264924 |  100.0 |   0.000001 | 103.264924 |
    | 14:30 | 3.717877  3.718629  3.724196             3.749908 | 124.881181  124.920434   125.21113  126.557653 |  52.915051   52.922722   52.979498        53.241705 |  52.178766   52.186288   52.241962     52.499077 | 1.088393  1.088509  1.089372  1.093353 |  1.35916  1.359152  1.359095          1.358834 | -0.581093  -0.581037  -0.580629     -0.578748 | 0.672252  0.672192   0.67176      0.669825 | 0.908839  0.908823  0.908711      0.908239 | 100.000001 | 100.002753  100.045478  100.362031          101.831587 | 102.997832 | 0.057656  0.057663  0.057711       0.057933 | 0.207292  0.207329  0.207608        0.208893 | 100.000001  100.005383  100.083996   100.63054  102.997832 |  100.0 |   0.000001 | 102.997832 |
    | 15:00 | 3.717871  3.718544  3.723592             3.747223 | 124.880888  124.916027  125.179579  126.416765 |  52.914993   52.921861   52.973339         53.21433 |   52.17871   52.185444   52.235922     52.472234 | 1.088392  1.088496  1.089279  1.092938 |  1.35916  1.359153  1.359101          1.358861 | -0.581093  -0.581044  -0.580673     -0.578945 | 0.672252  0.672199  0.671807      0.670026 | 0.908839  0.908825  0.908723      0.908288 |      100.0 | 100.002434  100.040681  100.327661          101.677575 | 102.751434 | 0.057656  0.057662  0.057706        0.05791 | 0.207291  0.207325  0.207578        0.208759 |      100.0  100.004759  100.075188  100.571452  102.751434 |  100.0 |        0.0 | 102.751434 |
    | 15:30 | 3.717866  3.718469  3.723044             3.744753 | 124.880629   124.91208   125.15095  126.287156 |  52.914943   52.921089   52.967749        53.189135 |   52.17866   52.184687   52.230441     52.447527 | 1.088391  1.088484  1.089194  1.092556 |  1.35916  1.359154  1.359107          1.358886 | -0.581093  -0.581049  -0.580714     -0.579125 | 0.672253  0.672205  0.671849      0.670211 | 0.908839  0.908826  0.908734      0.908333 |      100.0 | 100.002152  100.036385  100.296477          101.535945 | 102.524246 | 0.057656  0.057662  0.057701       0.057889 | 0.207291  0.207321   0.20755        0.208636 |      100.0  100.004208  100.067293  100.517751  102.524246 |  100.0 |        0.0 | 102.524246 |
    | 16:00 | 3.717862  3.718401  3.722547              3.74248 |   124.8804  124.908546  125.124982  126.167981 |  52.914898   52.920399   52.962679        53.165957 |  52.178616    52.18401   52.225469       52.4248 |  1.08839  1.088474  1.089117  1.092204 |  1.35916  1.359155  1.359112          1.358909 | -0.581094  -0.581054   -0.58075     -0.579291 | 0.672253   0.67221  0.671888      0.670382 |  0.90884  0.908828  0.908744      0.908374 |      100.0 | 100.001902  100.032538  100.268194          101.405759 | 102.314879 | 0.057656  0.057661  0.057697       0.057869 | 0.207291  0.207318  0.207525        0.208522 |      100.0   100.00372  100.060218  100.468965  102.314879 |  100.0 |        0.0 | 102.314879 |
    | 16:30 | 3.717858   3.71834  3.722096              3.74039 | 124.880197  124.905383  125.101434  126.058451 |  52.914858   52.919781    52.95808        53.144646 |  52.178577   52.183404    52.22096     52.403903 |  1.08839  1.088464  1.089047  1.091881 |  1.35916  1.359155  1.359117           1.35893 | -0.581094  -0.581059  -0.580783     -0.579444 | 0.672253  0.672215  0.671923      0.670539 |  0.90884  0.908829  0.908753      0.908413 |      100.0 | 100.001682  100.029094  100.242548          101.286147 |  102.12203 | 0.057656   0.05766  0.057693       0.057851 | 0.207291  0.207315  0.207503        0.208418 |      100.0  100.003289  100.053879  100.424659   102.12203 |  100.0 |        0.0 |  102.12203 |
    | 17:00 | 3.717854  3.718286  3.721687              3.73847 | 124.880018  124.902551  125.080088  125.957833 |  52.914823   52.919227   52.953912        53.125062 |  52.178543   52.182861   52.216872     52.384699 | 1.088389  1.088456  1.088983  1.091583 |  1.35916  1.359156  1.359121          1.358949 | -0.581094  -0.581062  -0.580813     -0.579584 | 0.672254  0.672219  0.671955      0.670684 |  0.90884   0.90883  0.908762      0.908448 |      100.0 | 100.001487  100.026012  100.219302          101.176298 | 101.944485 | 0.057656   0.05766  0.057689       0.057834 | 0.207291  0.207312  0.207482        0.208322 |      100.0  100.002908  100.048201  100.384435  101.944485 |  100.0 |        0.0 | 101.944485 |
    | 17:30 | 3.717851  3.718238  3.721317             3.736706 |  124.87986  124.900016  125.060743  125.865442 |  52.914792   52.918732   52.950133        53.107072 |  52.178513   52.182376   52.213168     52.367059 | 1.088389  1.088449  1.088926   1.09131 |  1.35916  1.359156  1.359125          1.358967 | -0.581094  -0.581066   -0.58084     -0.579713 | 0.672254  0.672223  0.671983      0.670817 |  0.90884  0.908831  0.908769      0.908481 |      100.0 | 100.001314  100.023253  100.198236          101.075458 |  101.78111 | 0.057656   0.05766  0.057686       0.057819 |  0.20729   0.20731  0.207464        0.208233 |      100.0  100.002571  100.043115  100.347929   101.78111 |  100.0 |        0.0 |  101.78111 |
    | 18:00 | 3.717849  3.718194  3.720981             3.735086 |  124.87972  124.897748  125.043217  125.780645 |  52.914765   52.918288    52.94671        53.090556 |  52.178486   52.181941   52.209811     52.350863 | 1.088388  1.088442  1.088874   1.09106 |  1.35916  1.359157  1.359128          1.358984 | -0.581095  -0.581069  -0.580865     -0.579832 | 0.672254  0.672227   0.67201      0.670939 |  0.90884  0.908832  0.908776      0.908511 |      100.0 | 100.001162  100.020784  100.179152          100.982928 | 101.630844 | 0.057656  0.057659  0.057683       0.057805 |  0.20729  0.207308  0.207447        0.208152 |      100.0  100.002273  100.038561  100.314808  101.630844 |  100.0 |        0.0 | 101.630844 |
    | 18:30 | 3.717846  3.718155  3.720677               3.7336 | 124.879596  124.895719  125.027344   125.70285 |  52.914741   52.917892    52.94361        53.075399 |  52.178462   52.181552    52.20677     52.336001 | 1.088388  1.088436  1.088827  1.090829 |  1.35916  1.359157  1.359131          1.358999 | -0.581095  -0.581072  -0.580887     -0.579941 | 0.672254   0.67223  0.672033      0.671051 |  0.90884  0.908833  0.908782      0.908538 |      100.0 | 100.001027  100.018576  100.161868          100.898058 |   101.4927 | 0.057656  0.057659  0.057681       0.057792 |  0.20729  0.207306  0.207432        0.208078 |      100.0  100.002009  100.034483  100.284768    101.4927 |  100.0 |        0.0 |   101.4927 |
    | 19:00 | 3.717844   3.71812  3.720402             3.732237 | 124.879487  124.893903  125.012971   125.63151 |  52.914719   52.917537   52.940802        53.061497 |  52.178441   52.181204   52.204017     52.322368 | 1.088388   1.08843  1.088784  1.090618 |  1.35916  1.359158  1.359134          1.359013 | -0.581095  -0.581075  -0.580907      -0.58004 | 0.672254  0.672232  0.672055      0.671155 |  0.90884  0.908834  0.908788      0.908564 |      100.0 | 100.000908  100.016599  100.146218          100.820245 |  101.36576 | 0.057656  0.057659  0.057678       0.057781 |  0.20729  0.207304  0.207418         0.20801 |      100.0  100.001776  100.030832   100.25753   101.36576 |  100.0 |        0.0 |  101.36576 |
    | 19:30 | 3.717842  3.718089  3.720152             3.730987 |  124.87939  124.892279   124.99996  125.566117 |  52.914701    52.91722    52.93826        53.048749 |  52.178423   52.180893   52.201525     52.309868 | 1.088387  1.088426  1.088746  1.090425 |  1.35916  1.359158  1.359137          1.359026 | -0.581095  -0.581077  -0.580926     -0.580132 | 0.672255  0.672235  0.672074      0.671249 |  0.90884  0.908835  0.908793      0.908587 |      100.0 | 100.000803  100.014832  100.132053          100.748932 | 101.249166 | 0.057656  0.057658  0.057676        0.05777 |  0.20729  0.207302  0.207406        0.207947 |      100.0   100.00157  100.027565  100.232841  101.249166 |  100.0 |        0.0 | 101.249166 |
    | 20:00 | 3.717841  3.718061  3.719927             3.729841 | 124.879304  124.890827  124.988186  125.506199 |  52.914684   52.916936    52.93596        53.037067 |  52.178406   52.180615   52.199269     52.298413 | 1.088387  1.088421  1.088711  1.090247 |  1.35916  1.359158  1.359139          1.359037 | -0.581095  -0.581079  -0.580942     -0.580216 | 0.672255  0.672237  0.672092      0.671336 |  0.90884  0.908835  0.908797      0.908609 |      100.0 |  100.00071  100.013251  100.119234          100.683601 | 101.142123 | 0.057656  0.057658  0.057674        0.05776 |  0.20729  0.207301  0.207394         0.20789 |      100.0  100.001388   100.02464  100.210468  101.142123 |  100.0 |        0.0 | 101.142123 |
    | 20:30 | 3.717839  3.718037  3.719723             3.728792 | 124.879229  124.889528  124.977534   125.45132 |  52.914669   52.916682   52.933879        53.026365 |  52.178392   52.180366   52.197229     52.287918 | 1.088387  1.088417  1.088679  1.090085 |  1.35916  1.359158  1.359141          1.359048 | -0.581095  -0.581081  -0.580957     -0.580293 | 0.672255  0.672239  0.672107      0.671416 |  0.90884  0.908836  0.908801      0.908628 |      100.0 | 100.000628  100.011838  100.107637          100.623774 | 101.043891 | 0.057656  0.057658  0.057672       0.057751 |  0.20729    0.2073  0.207384        0.207838 |      100.0  100.001227  100.022023    100.1902  101.043891 |  100.0 |        0.0 | 101.043891 |
    | 21:00 | 3.717838  3.718014  3.719538             3.727831 | 124.879162  124.888367  124.967898  125.401076 |  52.914656   52.916455   52.931996        53.016564 |  52.178379   52.180143   52.195383     52.278308 | 1.088387  1.088414   1.08865  1.089936 |  1.35916  1.359159  1.359143          1.359058 | -0.581095  -0.581082  -0.580971     -0.580363 | 0.672255  0.672241  0.672122      0.671489 |  0.90884  0.908836  0.908805      0.908647 |      100.0 | 100.000555  100.010574  100.097147          100.569007 | 100.953784 | 0.057656  0.057658  0.057671       0.057742 |  0.20729  0.207299  0.207375         0.20779 |      100.0  100.001085  100.019682  100.171845  100.953784 |  100.0 |        0.0 | 100.953784 |
    | 21:30 | 3.717837  3.717994  3.719371             3.726951 | 124.879103  124.887329  124.959185  125.355093 |  52.914644   52.916252   52.930294        53.007594 |  52.178368   52.179944   52.193713     52.269512 | 1.088386  1.088411  1.088624  1.089799 |  1.35916  1.359159  1.359145          1.359067 | -0.581095  -0.581084  -0.580983     -0.580427 | 0.672255  0.672242  0.672135      0.671556 |  0.90884  0.908837  0.908808      0.908663 |      100.0 | 100.000491  100.009444  100.087662          100.518891 | 100.871163 | 0.057656  0.057657  0.057669       0.057735 |  0.20729  0.207298  0.207366        0.207746 |      100.0  100.000959  100.017588  100.155226  100.871163 |  100.0 |        0.0 | 100.871163 |
    | 22:00 | 3.717836  3.717977   3.71922             3.726146 |  124.87905  124.886401  124.951308  125.313026 |  52.914634   52.916071   52.928755        52.999385 |  52.178358   52.179766   52.192204     52.261463 | 1.088386  1.088408  1.088601  1.089675 |  1.35916  1.359159  1.359146          1.359075 | -0.581096  -0.581085  -0.580994     -0.580486 | 0.672255  0.672244  0.672147      0.671617 |  0.90884  0.908837  0.908812      0.908679 |      100.0 | 100.000434  100.008434  100.079087          100.473049 | 100.795438 | 0.057656  0.057657  0.057668       0.057728 |  0.20729  0.207297  0.207359        0.207705 |      100.0  100.000848  100.015715  100.140183  100.795438 |  100.0 |        0.0 | 100.795438 |
    | 22:30 | 3.717835  3.717961  3.719084              3.72541 | 124.879004  124.885571  124.944189  125.274555 |  52.914625   52.915909   52.927364        52.991878 |  52.178349   52.179607    52.19084     52.254101 | 1.088386  1.088406   1.08858  1.089561 |  1.35916  1.359159  1.359148          1.359083 | -0.581096  -0.581086  -0.581004      -0.58054 | 0.672255  0.672245  0.672157      0.671673 |  0.90884  0.908837  0.908814      0.908693 |      100.0 | 100.000383  100.007531  100.071336           100.43113 | 100.726061 | 0.057656  0.057657  0.057667       0.057722 |  0.20729  0.207296  0.207352        0.207669 |      100.0   100.00075  100.014039  100.126571  100.726061 |  100.0 |        0.0 | 100.726061 |
    | 23:00 | 3.717834  3.717947  3.718961             3.724737 | 124.878963   124.88483  124.937756  125.239386 |  52.914617   52.915764   52.926107        52.985014 |  52.178341   52.179465   52.189607      52.24737 | 1.088386  1.088403  1.088561  1.089456 | 1.359161  1.359159  1.359149           1.35909 | -0.581096  -0.581087  -0.581013      -0.58059 | 0.672255  0.672246  0.672167      0.671725 |  0.90884  0.908838  0.908817      0.908705 |      100.0 | 100.000339  100.006724  100.064333          100.392813 | 100.662525 | 0.057656  0.057657  0.057666       0.057716 |  0.20729  0.207295  0.207346        0.207635 |      100.0  100.000663  100.012541  100.114256  100.662525 |  100.0 |        0.0 | 100.662525 |
    | 23:30 | 3.717833  3.717934  3.718849             3.724122 | 124.878927  124.884167  124.931944  125.207246 |   52.91461   52.915634   52.924971         52.97874 |  52.178334   52.179338   52.188494     52.241219 | 1.088386  1.088401  1.088543  1.089361 | 1.359161  1.359159   1.35915          1.359096 | -0.581096  -0.581088  -0.581021     -0.580635 | 0.672255  0.672247  0.672176      0.671772 |  0.90884  0.908838  0.908819      0.908717 |      100.0 |   100.0003  100.006003  100.058006            100.3578 | 100.604361 | 0.057656  0.057657  0.057665        0.05771 |  0.20729  0.207295   0.20734        0.207604 |      100.0  100.000586  100.011202  100.103118  100.604361 |  100.0 |        0.0 | 100.604361 |

The coarse spatial resolution also leads to a higher peak flow, compensating for the
initial discharge reduction:

>>> round_(numpy.max(fluxes.outflow.series))
651.690965
>>> round_(model.check_waterbalance(conditions))
0.0

.. _musk_mct_no_baseflow:

No baseflow
___________

The accidental discharge reductions discussed in the :ref:`musk_mct_long_segments`
example may become more problematic when leading to negative discharge values.
|musk_mct| or, more specifically, method |Calc_Discharge_V2| avoids this by simply
setting negative discharge values simulated for any segment endpoint to zero.  We
demonstrate this by removing the base flow from the synthetic flood wave:

>>> nmbsegments(4)
>>> length(10.0)
>>> bottomwidth(15.0)
>>> sideslope(5.0)
>>> bottomslope(0.00025)
>>> stricklercoefficient(1.0/0.035)
>>> model.prepare_states(discharge=0.0)
>>> test.inits = ((states.discharge, 0.0),
...               (states.courantnumber, states.courantnumber[0]),
...               (states.reynoldsnumber, states.reynoldsnumber[0]))
>>> nodes.input1.sequences.sim.series = 0.0

.. integration-test::

    >>> conditions = test("musk_mct_no_baseflow", get_conditions="2000-01-01")
    |  date |                               referencewaterlevel |                                     wettedarea |                                     wettedperimeter |                                     surfacewidth |                               celerity |                               correctingfactor |                                  coefficient1 |                               coefficient2 |                               coefficient3 |     inflow |                                     referencedischarge |    outflow |                               courantnumber |                               reynoldsnumber |                                                  discharge | input1 |     input2 |     output |
    -------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
    | 00:00 |      0.0       0.0       0.0                  0.0 |        0.0         0.0         0.0         0.0 |       15.0        15.0        15.0             15.0 |       15.0        15.0        15.0          15.0 |      0.0       0.0       0.0       0.0 |      1.0       1.0       1.0               1.0 |      -1.0       -1.0       -1.0          -1.0 |      1.0       1.0       1.0           1.0 |      1.0       1.0       1.0           1.0 |        0.0 |        0.0         0.0         0.0                 0.0 |        0.0 |      0.0       0.0       0.0            0.0 |      0.0       0.0       0.0             0.0 |        0.0         0.0         0.0         0.0         0.0 |    0.0 |        0.0 |        0.0 |
    | 00:30 |      0.0       0.0       0.0                  0.0 |        0.0         0.0         0.0         0.0 |       15.0        15.0        15.0             15.0 |       15.0        15.0        15.0          15.0 |      0.0       0.0       0.0       0.0 |      0.0       1.0       1.0               1.0 |      -1.0       -1.0       -1.0          -1.0 |      1.0       1.0       1.0           1.0 |      1.0       1.0       1.0           1.0 |        0.0 |        0.0         0.0         0.0                 0.0 |        0.0 |      0.0       0.0       0.0            0.0 |      0.0       0.0       0.0             0.0 |        0.0         0.0         0.0         0.0         0.0 |    0.0 |        0.0 |        0.0 |
    | 01:00 |      0.0       0.0       0.0                  0.0 |        0.0         0.0         0.0         0.0 |       15.0        15.0        15.0             15.0 |       15.0        15.0        15.0          15.0 |      0.0       0.0       0.0       0.0 |      0.0       1.0       1.0               1.0 |      -1.0       -1.0       -1.0          -1.0 |      1.0       1.0       1.0           1.0 |      1.0       1.0       1.0           1.0 |        0.0 |        0.0         0.0         0.0                 0.0 |        0.0 |      0.0       0.0       0.0            0.0 |      0.0       0.0       0.0             0.0 |        0.0         0.0         0.0         0.0         0.0 |    0.0 |        0.0 |        0.0 |
    | 01:30 |      0.0       0.0       0.0                  0.0 |        0.0         0.0         0.0         0.0 |       15.0        15.0        15.0             15.0 |       15.0        15.0        15.0          15.0 |      0.0       0.0       0.0       0.0 |      0.0       1.0       1.0               1.0 |      -1.0       -1.0       -1.0          -1.0 |      1.0       1.0       1.0           1.0 |      1.0       1.0       1.0           1.0 |        0.0 |        0.0         0.0         0.0                 0.0 |        0.0 |      0.0       0.0       0.0            0.0 |      0.0       0.0       0.0             0.0 |        0.0         0.0         0.0         0.0         0.0 |    0.0 |        0.0 |        0.0 |
    | 02:00 |      0.0       0.0       0.0                  0.0 |        0.0         0.0         0.0         0.0 |       15.0        15.0        15.0             15.0 |       15.0        15.0        15.0          15.0 |      0.0       0.0       0.0       0.0 |      0.0       1.0       1.0               1.0 |      -1.0       -1.0       -1.0          -1.0 |      1.0       1.0       1.0           1.0 |      1.0       1.0       1.0           1.0 |        0.0 |        0.0         0.0         0.0                 0.0 |        0.0 |      0.0       0.0       0.0            0.0 |      0.0       0.0       0.0             0.0 |        0.0         0.0         0.0         0.0         0.0 |    0.0 |        0.0 |        0.0 |
    | 02:30 |      0.0       0.0       0.0                  0.0 |        0.0         0.0         0.0         0.0 |       15.0        15.0        15.0             15.0 |       15.0        15.0        15.0          15.0 |      0.0       0.0       0.0       0.0 |      0.0       1.0       1.0               1.0 |      -1.0       -1.0       -1.0          -1.0 |      1.0       1.0       1.0           1.0 |      1.0       1.0       1.0           1.0 |        0.0 |        0.0         0.0         0.0                 0.0 |        0.0 |      0.0       0.0       0.0            0.0 |      0.0       0.0       0.0             0.0 |        0.0         0.0         0.0         0.0         0.0 |    0.0 |        0.0 |        0.0 |
    | 03:00 |      0.0       0.0       0.0                  0.0 |        0.0         0.0         0.0         0.0 |       15.0        15.0        15.0             15.0 |       15.0        15.0        15.0          15.0 |      0.0       0.0       0.0       0.0 |      0.0       1.0       1.0               1.0 |      -1.0       -1.0       -1.0          -1.0 |      1.0       1.0       1.0           1.0 |      1.0       1.0       1.0           1.0 |   0.000003 |   0.000002         0.0         0.0                 0.0 |        0.0 |      0.0       0.0       0.0            0.0 |      0.0       0.0       0.0             0.0 |   0.000003         0.0         0.0         0.0         0.0 |    0.0 |   0.000003 |        0.0 |
    | 03:30 |      0.0       0.0       0.0                  0.0 |        0.0         0.0         0.0         0.0 |       15.0        15.0        15.0             15.0 |       15.0        15.0        15.0          15.0 |      0.0       0.0       0.0       0.0 |      0.0       1.0       1.0               1.0 |      -1.0       -1.0       -1.0          -1.0 |      1.0       1.0       1.0           1.0 |      1.0       1.0       1.0           1.0 |   0.000029 |   0.000014         0.0         0.0                 0.0 |        0.0 |      0.0       0.0       0.0            0.0 |      0.0       0.0       0.0             0.0 |   0.000029         0.0         0.0         0.0         0.0 |    0.0 |   0.000029 |        0.0 |
    | 04:00 |      0.0       0.0       0.0                  0.0 |        0.0         0.0         0.0         0.0 |       15.0        15.0        15.0             15.0 |       15.0        15.0        15.0          15.0 |      0.0       0.0       0.0       0.0 |      0.0       1.0       1.0               1.0 |      -1.0       -1.0       -1.0          -1.0 |      1.0       1.0       1.0           1.0 |      1.0       1.0       1.0           1.0 |   0.000175 |   0.000088         0.0         0.0                 0.0 |        0.0 |      0.0       0.0       0.0            0.0 |      0.0       0.0       0.0             0.0 |   0.000175         0.0         0.0         0.0         0.0 |    0.0 |   0.000175 |        0.0 |
    | 04:30 |      0.0       0.0       0.0                  0.0 |        0.0         0.0         0.0         0.0 |       15.0        15.0        15.0             15.0 |       15.0        15.0        15.0          15.0 |      0.0       0.0       0.0       0.0 |      0.0       1.0       1.0               1.0 |      -1.0       -1.0       -1.0          -1.0 |      1.0       1.0       1.0           1.0 |      1.0       1.0       1.0           1.0 |   0.000826 |   0.000413         0.0         0.0                 0.0 |        0.0 |      0.0       0.0       0.0            0.0 |      0.0       0.0       0.0             0.0 |   0.000826         0.0         0.0         0.0         0.0 |    0.0 |   0.000826 |        0.0 |
    | 05:00 |      0.0       0.0       0.0                  0.0 |        0.0         0.0         0.0         0.0 |       15.0        15.0        15.0             15.0 |       15.0        15.0        15.0          15.0 |      0.0       0.0       0.0       0.0 |      0.0       1.0       1.0               1.0 |      -1.0       -1.0       -1.0          -1.0 |      1.0       1.0       1.0           1.0 |      1.0       1.0       1.0           1.0 |   0.003194 |   0.001597         0.0         0.0                 0.0 |        0.0 |      0.0       0.0       0.0            0.0 |      0.0       0.0       0.0             0.0 |   0.003194         0.0         0.0         0.0         0.0 |    0.0 |   0.003194 |        0.0 |
    | 05:30 |      0.0       0.0       0.0                  0.0 |        0.0         0.0         0.0         0.0 |       15.0        15.0        15.0             15.0 |       15.0        15.0        15.0          15.0 |      0.0       0.0       0.0       0.0 |      0.0       1.0       1.0               1.0 |      -1.0       -1.0       -1.0          -1.0 |      1.0       1.0       1.0           1.0 |      1.0       1.0       1.0           1.0 |   0.010515 |   0.005258         0.0         0.0                 0.0 |        0.0 |      0.0       0.0       0.0            0.0 |      0.0       0.0       0.0             0.0 |   0.010515         0.0         0.0         0.0         0.0 |    0.0 |   0.010515 |        0.0 |
    | 06:00 |    0.009       0.0       0.0                  0.0 |   0.135405         0.0         0.0         0.0 |  15.091782        15.0        15.0             15.0 |      15.09        15.0        15.0          15.0 | 0.032431       0.0       0.0       0.0 | 0.289707       1.0       1.0               1.0 | -0.881619       -1.0       -1.0          -1.0 | 0.940809       1.0       1.0           1.0 | 0.940809       1.0       1.0           1.0 |   0.030315 |   0.015158         0.0         0.0                 0.0 |        0.0 |  0.02015       0.0       0.0            0.0 | 0.042765       0.0       0.0             0.0 |   0.030315         0.0         0.0         0.0         0.0 |    0.0 |   0.030315 |        0.0 |
    | 06:30 |  0.05994       0.0       0.0                  0.0 |   0.917064         0.0         0.0         0.0 |   15.61127        15.0        15.0             15.0 |    15.5994        15.0        15.0          15.0 |  0.11203       0.0       0.0       0.0 | 2.628283       1.0       1.0               1.0 | -0.978089       -1.0       -1.0          -1.0 | 0.368083       1.0       1.0           1.0 | 0.385116       1.0       1.0           1.0 |   0.078179 |    0.03909         0.0         0.0                 0.0 |        0.0 | 0.007672       0.0       0.0            0.0 | 0.003404       0.0       0.0             0.0 |   0.078179         0.0         0.0         0.0         0.0 |    0.0 |   0.078179 |        0.0 |
    | 07:00 | 0.081548       0.0       0.0                  0.0 |   1.256467         0.0         0.0         0.0 |  15.831628        15.0        15.0             15.0 |  15.815478        15.0        15.0          15.0 | 0.136203       0.0       0.0       0.0 | 1.746644       1.0       1.0               1.0 | -0.952262       -1.0       -1.0          -1.0 | 1.793394       1.0       1.0           1.0 |  1.77815       1.0       1.0           1.0 |   0.183351 |   0.097979         0.0         0.0                 0.0 |        0.0 | 0.014036       0.0       0.0            0.0 | 0.010416       0.0       0.0             0.0 |   0.183351         0.0         0.0         0.0         0.0 |    0.0 |   0.183351 |        0.0 |
    | 07:30 | 0.117131       0.0       0.0                  0.0 |   1.825567         0.0         0.0         0.0 |  16.194509        15.0        15.0             15.0 |  16.171312        15.0        15.0          15.0 | 0.170703       0.0       0.0       0.0 | 1.573043       1.0       1.0               1.0 | -0.927187       -1.0       -1.0          -1.0 | 1.345811       1.0       1.0           1.0 | 1.336102       1.0       1.0           1.0 |   0.396214 |   0.198107         0.0         0.0                 0.0 |        0.0 | 0.019533       0.0       0.0            0.0 | 0.018249       0.0       0.0             0.0 |   0.396214         0.0         0.0         0.0         0.0 |    0.0 |   0.396214 |        0.0 |
    | 08:00 | 0.176719       0.0       0.0                  0.0 |   2.806925         0.0         0.0         0.0 |  16.802182        15.0        15.0             15.0 |  16.767185        15.0        15.0          15.0 | 0.219099       0.0       0.0       0.0 | 1.542686       1.0       1.0               1.0 | -0.898071       -1.0       -1.0          -1.0 |  1.24366       1.0       1.0           1.0 | 1.240469       1.0       1.0           1.0 |   0.797305 |   0.398652         0.0         0.0                 0.0 |        0.0 | 0.025564       0.0       0.0            0.0 | 0.028137       0.0       0.0             0.0 |   0.797305         0.0         0.0         0.0         0.0 |    0.0 |   0.797305 |        0.0 |
    | 08:30 | 0.261729       0.0       0.0                  0.0 |   4.268453         0.0         0.0         0.0 |  17.669127        15.0        15.0             15.0 |  17.617294        15.0        15.0          15.0 | 0.275712       0.0       0.0       0.0 | 1.561829       1.0       1.0               1.0 | -0.866531       -1.0       -1.0          -1.0 | 1.157032       1.0       1.0           1.0 |    1.163       1.0       1.0           1.0 |   1.507032 |   0.753516         0.0         0.0                 0.0 |        0.0 | 0.031776       0.0       0.0            0.0 | 0.039731       0.0       0.0             0.0 |   1.507032         0.0         0.0         0.0         0.0 |    0.0 |   1.507032 |        0.0 |
    | 09:00 | 0.367729       0.0       0.0                  0.0 |   6.192066         0.0         0.0         0.0 |  18.750119        15.0        15.0             15.0 |  18.677294        15.0        15.0          15.0 |  0.33378       0.0       0.0       0.0 | 1.533879       1.0       1.0               1.0 | -0.825596       -1.0       -1.0          -1.0 |  1.11623       1.0       1.0           1.0 | 1.134131       1.0       1.0           1.0 |   2.694853 |   1.347427         0.0         0.0                 0.0 |        0.0 | 0.039169       0.0       0.0            0.0 | 0.056364       0.0       0.0             0.0 |   2.694853         0.0         0.0         0.0         0.0 |    0.0 |   2.694853 |        0.0 |
    | 09:30 |  0.50018       0.0       0.0                  0.0 |   8.753594         0.0         0.0         0.0 |  20.100852        15.0        15.0             15.0 |  20.001797        15.0        15.0          15.0 |  0.39416       0.0       0.0       0.0 |   1.5046       1.0       1.0               1.0 | -0.778596       -1.0       -1.0          -1.0 | 1.052196       1.0       1.0           1.0 | 1.089014       1.0       1.0           1.0 |   4.586356 |   2.293178         0.0         0.0                 0.0 |        0.0 | 0.047155       0.0       0.0            0.0 | 0.077328       0.0       0.0             0.0 |   4.586356         0.0         0.0         0.0         0.0 |    0.0 |   4.586356 |        0.0 |
    | 10:00 | 0.664409       0.0       0.0                  0.0 |  12.173324         0.0         0.0         0.0 |  21.775665        15.0        15.0             15.0 |  21.644086        15.0        15.0          15.0 | 0.457114       0.0       0.0       0.0 | 1.490521       1.0       1.0               1.0 | -0.729406       -1.0       -1.0          -1.0 | 0.981739       1.0       1.0           1.0 | 1.042826       1.0       1.0           1.0 |   7.466641 |   3.733321         0.0         0.0                 0.0 |        0.0 | 0.055202       0.0       0.0            0.0 | 0.101264       0.0       0.0             0.0 |   7.466641         0.0         0.0         0.0         0.0 |    0.0 |   7.466641 |        0.0 |
    | 10:30 | 0.855024       0.0       0.0                  0.0 |   16.48068         0.0         0.0         0.0 |  23.719564        15.0        15.0             15.0 |  23.550236        15.0        15.0          15.0 | 0.519564       0.0       0.0       0.0 | 1.465443       1.0       1.0               1.0 |  -0.67481       -1.0       -1.0          -1.0 | 0.923507       1.0       1.0           1.0 |  1.01269       1.0       1.0           1.0 |  11.678559 |   5.843125         0.0         0.0                 0.0 |        0.0 | 0.063818       0.0       0.0            0.0 | 0.130347       0.0       0.0             0.0 |  11.678559         0.0         0.0         0.0         0.0 |    0.0 |  11.678559 |        0.0 |
    | 11:00 | 1.090327       0.0       0.0                  0.0 |   22.29897         0.0         0.0         0.0 |  26.119197        15.0        15.0             15.0 |   25.90327        15.0        15.0          15.0 | 0.586491       0.0       0.0       0.0 | 1.441898       1.0       1.0               1.0 | -0.614416       -1.0       -1.0          -1.0 | 0.864456       1.0       1.0           1.0 | 0.987678       1.0       1.0           1.0 |  17.614836 |   9.070094         0.0         0.0                 0.0 |        0.0 | 0.073215       0.0       0.0            0.0 | 0.165623       0.0       0.0             0.0 |  17.614836         0.0         0.0         0.0         0.0 |    0.0 |  17.614836 |        0.0 |
    | 11:30 | 1.352656       0.0       0.0                  0.0 |  29.438234         0.0         0.0         0.0 |   28.79444        15.0        15.0             15.0 |  28.526561        15.0        15.0          15.0 | 0.652397       0.0       0.0       0.0 | 1.422267       1.0       1.0               1.0 | -0.554452       -1.0       -1.0          -1.0 | 0.795502       1.0       1.0           1.0 | 0.957494       1.0       1.0           1.0 |  25.703736 |  13.503379         0.0         0.0                 0.0 |        0.0 | 0.082566       0.0       0.0            0.0 | 0.204061       0.0       0.0             0.0 |  25.703736         0.0         0.0         0.0         0.0 |    0.0 |  25.703736 |        0.0 |
    | 12:00 | 1.641741  0.174336       0.0                  0.0 |  38.102679    2.767003         0.0         0.0 |  31.742538   16.777884        15.0             15.0 |  31.417409   16.743359        15.0          15.0 | 0.717867  0.217331       0.0       0.0 | 1.405257  1.540559       1.0               1.0 | -0.495202  -0.898892       -1.0          -1.0 | 0.731428  0.949446       1.0           1.0 | 0.933737  0.949446       1.0           1.0 |  36.388647 |  19.464518     0.39035         0.0                 0.0 |        0.0 | 0.091952  0.025393       0.0            0.0 |  0.24566  0.027853       0.0             0.0 |  36.388647    0.780699         0.0         0.0         0.0 |    0.0 |  36.388647 |        0.0 |
    | 12:30 | 1.972289  0.413655       0.0                  0.0 |  49.033972    7.060384         0.0         0.0 |  35.113485   19.218474        15.0             15.0 |  34.722895   19.136554        15.0          15.0 | 0.786342  0.355966       0.0       0.0 | 1.392611   1.51632       1.0               1.0 | -0.435855  -0.807595       -1.0          -1.0 | 0.671575  1.500288       1.0           1.0 | 0.915525  1.507687       1.0           1.0 |  50.102726 |  27.687197     1.65747         0.0                 0.0 |        0.0 | 0.101638  0.042256       0.0            0.0 | 0.291261  0.064186       0.0             0.0 |  50.102726    3.314939         0.0         0.0         0.0 |    0.0 |  50.102726 |        0.0 |
    | 13:00 | 2.343179   0.70859       0.0                  0.0 |  62.600128   13.139343         0.0         0.0 |  38.895832   22.226226        15.0             15.0 |  38.431791   22.085897        15.0          15.0 | 0.857387  0.472436       0.0       0.0 | 1.381947   1.48407       1.0               1.0 | -0.376615  -0.716227       -1.0          -1.0 | 0.612877  1.138114       1.0           1.0 | 0.899696  1.189151       1.0           1.0 |  67.240349 |  38.838357    4.182751         0.0                 0.0 |        0.0 | 0.111676  0.057301       0.0            0.0 | 0.341163  0.108046       0.0             0.0 |  67.240349    8.365501         0.0         0.0         0.0 |    0.0 |  67.240349 |        0.0 |
    | 13:30 | 2.745362  1.036978       0.0                  0.0 |   78.86551   20.931292         0.0         0.0 |  42.997313   25.575144        15.0             15.0 |  42.453624   25.369782        15.0          15.0 | 0.929354  0.572082       0.0       0.0 | 1.372924   1.44646       1.0               1.0 | -0.319204  -0.627438       -1.0          -1.0 | 0.554512  0.959668       1.0           1.0 | 0.884821  1.062272       1.0           1.0 |  88.127569 |  53.385336    8.278436         0.0                 0.0 |        0.0 | 0.121845  0.071191       0.0            0.0 | 0.394221  0.157734       0.0             0.0 |  88.127569   16.556871         0.0         0.0         0.0 |    0.0 |  88.127569 |        0.0 |
    | 14:00 | 3.171291  1.409512       0.0                  0.0 |  97.854818   31.076307         0.0         0.0 |  47.340954   29.374261        15.0             15.0 |  46.712914   29.095122        15.0          15.0 | 1.001217  0.665782       0.0       0.0 | 1.365821  1.418423       1.0               1.0 | -0.264921  -0.542208       -1.0          -1.0 | 0.498356   0.83594       1.0           1.0 | 0.871462  0.994339       1.0           1.0 | 112.993962 |  71.732614   14.586662         0.0                 0.0 |        0.0 | 0.131949  0.084489       0.0            0.0 | 0.449177  0.212353       0.0             0.0 | 112.993962   28.413164         0.0         0.0         0.0 |    0.0 | 112.993962 |        0.0 |
    | 14:30 | 3.612903  1.819674     0.112                  0.0 | 119.458887   43.851179     1.74272         0.0 |  51.844526   33.557107    16.14218             15.0 |   51.12903   33.196741       16.12          15.0 | 1.072029  0.755435  0.166047       0.0 |  1.36027  1.398002  1.672364               1.0 | -0.214375  -0.462396  -0.935483          -1.0 | 0.445703  0.734145  0.967741           1.0 | 0.859865  0.949412  0.967741           1.0 | 141.948195 |  94.145537   23.695757    0.173033                 0.0 |        0.0 | 0.141858  0.097266  0.017872            0.0 |  0.50508  0.270352  0.015462             0.0 | 141.948195   44.363108    0.346065         0.0         0.0 |    0.0 | 141.948195 |        0.0 |
    | 15:00 | 4.063244  2.241431  0.435121                  0.0 | 143.498406   58.741517    7.473465         0.0 |  56.437118   37.858197    19.43738             15.0 |  55.632438   37.414307   19.351209          15.0 | 1.141117  0.838403  0.365846       0.0 | 1.355932  1.384316  1.534324               1.0 | -0.167771  -0.392049  -0.804175          -1.0 | 0.397033  0.645079  2.171579           1.0 | 0.849974   0.91513  2.161136           1.0 |  174.95935 | 120.764531   35.576459    1.781984                 0.0 |        0.0 | 0.151483  0.109016  0.042919            0.0 | 0.561181  0.327715  0.065621             0.0 |  174.95935   64.712461    3.563967         0.0         0.0 |    0.0 |  174.95935 |        0.0 |
    | 15:30 | 4.516153  2.705859  0.836251                  0.0 | 169.720457   77.196253   16.040348         0.0 |    61.0559   42.594456   23.528122             15.0 |  60.161525   42.058591   23.362512          15.0 | 1.207965   0.92248  0.513798       0.0 | 1.352527  1.373695  1.467627               1.0 | -0.125102  -0.324556  -0.679939          -1.0 | 0.352414   0.57373  1.205274           1.0 | 0.841597  0.894924  1.261268           1.0 | 211.845554 | 151.580265   51.839751    5.615526                 0.0 |        0.0 | 0.160761  0.120876  0.063016            0.0 | 0.616855  0.389064  0.127504             0.0 | 211.845554   89.617514   11.231052         0.0         0.0 |    0.0 | 211.845554 |        0.0 |
    | 16:00 |  4.96616  3.195329   1.26977                  0.0 | 197.806124    98.98055    27.10814         0.0 |  65.645093   47.586085   27.949168             15.0 |    64.6616   46.953285   27.697704          15.0 | 1.272166   1.00516   0.63236       0.0 | 1.349837  1.365476  1.427507               1.0 | -0.086242  -0.262029  -0.572511          -1.0 | 0.311727  0.506201  0.930727           1.0 | 0.834526  0.877218  1.059043           1.0 | 252.270855 | 186.424116   72.862002   12.008413                 0.0 |        0.0 | 0.169643  0.132502  0.079737            0.0 | 0.671569  0.452247  0.192114             0.0 | 252.270855  119.069824   24.016826         0.0         0.0 |    0.0 | 252.270855 |        0.0 |
    | 16:30 |  5.40825   3.69563  1.755582                  0.0 | 227.369571  123.722846   41.744086         0.0 |  70.153542   52.688177   32.903498             15.0 |  69.082497   51.956298   32.555825          15.0 | 1.333373  1.084939  0.742108       0.0 | 1.347697  1.359386  1.400489               1.0 | -0.051015  -0.205466  -0.473972          -1.0 |  0.27477  0.444537  0.782508           1.0 | 0.828561  0.862434  0.980647           1.0 | 295.750601 | 224.952861   98.744333    22.11984                 0.0 |        0.0 | 0.178087   0.14366   0.09538            0.0 | 0.724836  0.515449  0.261497             0.0 | 295.750601  152.885352    42.23111         0.0         0.0 |    0.0 | 295.750601 |        0.0 |
    | 17:00 | 5.838418  4.195833  2.251273             0.336393 | 258.011914  150.962575   59.110244    5.611691 |  74.540417    57.78927   37.958569        18.430546 |  73.384183   56.958331   37.512729     18.363927 | 1.391367  1.160941  0.840255  0.317717 | 1.345981  1.354849  1.384443          1.556531 | -0.019186  -0.154875   -0.39073       -0.8396 | 0.241327  0.389464  0.664152        0.9198 | 0.823543  0.850451   0.92876        0.9198 | 341.664962 | 266.711889   129.35651   35.875573            1.145452 |        0.0 | 0.186069  0.154238  0.109247       0.036741 | 0.776282  0.577551  0.328847        0.050451 | 341.664962  190.725253   65.920136    2.290903         0.0 |    0.0 | 341.664962 |        0.0 |
    | 17:30 | 6.252656  4.687739  2.764274             0.778475 | 289.268407  180.190596   79.670146   14.707231 |  78.764834    62.80575    43.19017        22.938914 |  77.526564   61.877394   42.642736     22.784745 |  1.44591  1.232688  0.932631  0.495569 | 1.344595  1.351431  1.372561          1.468628 |  0.009451  -0.109899  -0.316662     -0.694274 | 0.211131   0.34067   0.57518      1.381236 |  0.81931  0.840802  0.898885      1.419635 | 389.279607 | 311.064706  164.358206    54.13448            4.962761 |        0.0 | 0.193563  0.164184  0.122307       0.060739 | 0.825519  0.637782  0.396686        0.119708 | 389.279607  232.078222   94.895136    9.925522         0.0 |    0.0 | 389.279607 |        0.0 |
    | 18:00 | 6.647776  5.165305  3.301022             1.270525 | 320.681281  210.881447  103.999054   27.129047 |  82.794281   67.675981    48.66395        27.956865 |  81.477761   66.653049   48.010219     27.705251 | 1.496858  1.299952  1.022372  0.632545 | 1.343469   1.34882  1.364026          1.427454 |   0.03512  -0.070036  -0.249517     -0.572343 | 0.183969  0.297577  0.500069      0.971534 | 0.815746  0.833031  0.878251      1.093296 |  437.77215 | 357.294572  203.241134   77.949913           12.021649 |        0.0 | 0.200551  0.173479  0.134915       0.079763 | 0.872246  0.695617  0.465704        0.192224 |  437.77215  276.307048  128.760432   24.043298         0.0 |    0.0 |  437.77215 |        0.0 |
    | 18:30 | 7.020895  5.623705  3.841494              1.80357 | 351.778282  242.485842  131.407799   43.317862 |  86.599364    72.35076   54.175707        33.392874 |  85.208953   71.237047   53.414942     33.035697 | 1.544082  1.362601  1.107454  0.752107 | 1.342549  1.346798  1.357928          1.398352 |  0.058028  -0.034791  -0.190151     -0.465153 | 0.159615  0.259548  0.433309      0.789176 | 0.812741  0.826741  0.861677      0.989171 | 486.262607 | 404.584454  245.331067  107.169228           23.298609 |        0.0 |  0.20702  0.182112  0.146798       0.096813 | 0.916186  0.750646   0.53366        0.268232 | 486.262607  322.657946  166.940964   44.766531         0.0 |    0.0 | 486.262607 |        0.0 |
    | 19:00 | 7.369609  6.059708  4.372091             2.352905 | 382.099796  274.495917  161.157238   62.974393 |  90.155557   76.797138    59.58675        38.995019 |  88.696087   75.597079   58.720906     38.529052 | 1.587496  1.420655  1.186962  0.859184 | 1.341794  1.345212  1.353524          1.381705 |   0.07837  -0.003657   -0.13826     -0.375152 | 0.137867  0.226013  0.375225      0.658663 |  0.81021  0.821639  0.848724      0.931194 | 533.846018 | 452.067853  289.890323  141.325608           39.159274 |   2.559518 | 0.212961  0.190095   0.15785       0.111929 | 0.957108  0.802619  0.599219        0.342456 | 533.846018  370.297461  208.735961   71.775124    2.559518 |    0.0 | 533.846018 |   2.559518 |
    | 19:30 | 7.691933  6.470252  4.883694             2.903087 | 411.208151  306.374572  192.507754    85.68588 |  93.442632   80.983881   64.804103        44.605796 |  91.919329   79.702518   63.836941     44.030871 | 1.627045  1.474092  1.260552  0.956428 | 1.341172  1.343954  1.350287          1.370048 |  0.096327   0.023775  -0.093123     -0.298395 | 0.118539   0.19643  0.325033       0.56081 | 0.808077  0.817463  0.838644      0.896838 | 579.625318 |   498.8578  336.041342  179.714371           59.817178 |  11.380359 | 0.218367   0.19743  0.168038       0.125658 | 0.994823  0.851278  0.661582        0.414706 | 579.625318  418.343747  253.317517  104.450167   11.380359 |    0.0 | 579.625318 |  11.380359 |
    | 20:00 | 7.986285   6.85342  5.370554             3.469455 | 438.698052  337.648115  224.772536    112.2274 |   96.44445   84.891443   69.769115        50.381634 |  94.862854   83.534198   68.705536     49.694547 | 1.662704  1.522987   1.32822   1.04939 |  1.34066  1.342947  1.347864          1.361907 |  0.112066   0.047898  -0.053915     -0.230221 |  0.10146  0.170379  0.281713      0.482687 | 0.806281  0.814042  0.830774      0.875175 | 622.742733 | 544.079006  382.914515  221.496687            86.47455 |  27.691348 | 0.223238  0.204131  0.177377       0.138695 | 1.029181  0.896484   0.72031        0.487029 | 622.742733  465.899685  299.803876  141.973404   27.691348 |    0.0 | 622.742733 |  27.691348 |
    | 20:30 | 8.251465  7.207565  5.828711             4.030095 | 464.205329  367.858405  257.300041   141.65976 |   99.14876   88.503025   74.441425        56.099067 |  97.514648   87.075646   73.287112     55.300951 | 1.694474  1.567405  1.390074  1.136127 | 1.340239  1.342133  1.346017          1.356214 |   0.12574    0.06904   -0.01988      -0.17105 | 0.086476   0.14747  0.244265      0.414839 | 0.804769  0.811222  0.824574      0.858325 | 662.408181 | 586.898004  429.601988  265.721814          118.671107 |  51.294495 | 0.227575  0.210212  0.185892       0.150789 | 1.060072  0.938108  0.775124         0.55708 | 662.408181  512.084824  347.268002  183.395665   51.294495 |    0.0 | 662.408181 |  51.294495 |
    | 21:00 | 8.486629  7.531557  6.256062             4.571035 | 487.413777  396.595074  289.532504  173.037347 | 101.546972   91.807108   78.799566        61.615596 |  99.866288   90.315566   77.560622     60.710353 | 1.722377  1.607434  1.446354  1.215907 | 1.339894  1.341473  1.344585          1.352166 |  0.137486    0.08749   0.009679     -0.120185 | 0.073445  0.127382  0.211856      0.356949 | 0.803498  0.808892  0.819658      0.845486 | 697.923464 | 626.549863  475.224193  311.446543          155.600226 |  81.463924 | 0.231382  0.215687  0.193624       0.161861 | 1.087421  0.976069  0.825922        0.623558 | 697.923464  556.064612  394.783036  227.713435   81.463924 |    0.0 | 697.923464 |  81.463924 |
    | 21:30 | 8.691271  7.824662  6.651091             5.083877 | 508.060025  423.496615   320.95144  205.487169 | 103.633921   94.796209   82.828088        66.845574 |  101.91271   93.246621   81.510912     65.838768 | 1.746461  1.643177  1.497281  1.288634 | 1.339614  1.340936  1.343461          1.349225 |  0.147432   0.103499   0.035329     -0.076593 | 0.062239  0.109841  0.183754      0.307768 | 0.802431  0.806962   0.81572      0.835707 | 728.701418 | 662.360433  518.951155  357.698918          196.259235 | 117.184815 | 0.234667  0.220571  0.200609       0.171917 | 1.111188  1.010325  0.872637        0.685795 | 728.701418  597.075741  441.450454   273.88928  117.184815 |    0.0 | 728.701418 | 117.184815 |
    | 22:00 | 8.865196  8.086494  7.012986             5.563758 | 525.936465  448.254362  351.104657   238.23342 | 105.407617   97.466384   86.518705        71.739426 | 103.651962   95.864943    85.12986     70.637585 | 1.766789  1.674749  1.543089  1.354506 | 1.339388  1.340497  1.342568           1.34704 |  0.155692   0.117288   0.057555     -0.039246 | 0.052743  0.094607  0.159382      0.265943 | 0.801536  0.805359  0.812542      0.828199 |  754.27953 | 693.763484  560.026015  403.544533          239.553825 | 157.296946 | 0.237438  0.224883  0.206884       0.180998 | 1.131365  1.040861  0.915255        0.743474 |  754.27953  634.446782  486.426712  320.902301  157.296946 |    0.0 |  754.27953 | 157.296946 |
    | 22:30 | 9.008495  8.316974  7.341442              6.00853 | 540.892351  470.614888  379.605492  270.640128 | 106.868985   99.816825   89.868313        76.275225 | 105.084951    98.16974   88.414421     75.085302 | 1.783443  1.702272  1.584014  1.413913 | 1.339211  1.340141  1.341852          1.345384 |  0.162368   0.129047   0.076763       -0.0072 |  0.04485  0.081476  0.138265      0.230284 |  0.80079  0.804027  0.809963      0.822385 | 774.327871 | 720.312429  597.783845  448.112349          284.425464 | 200.609184 | 0.239708  0.228639  0.212484       0.189168 | 1.147974  1.067695  0.953806        0.796534 | 774.327871  667.613648  528.945362  367.778224  200.609184 |    0.0 | 774.327871 | 200.609184 |
    | 23:00 | 9.121518  8.516295  7.636528             6.416852 | 552.833214  490.380778   406.13072  302.132734 | 108.021596  101.849504   92.877611        80.439308 | 106.215179  100.162945    91.36528     79.168521 | 1.796519   1.72588  1.620285  1.467204 | 1.339077  1.339853  1.341275          1.344107 |  0.167553   0.138944   0.093295      0.020303 | 0.038466  0.070267  0.120009       0.19977 | 0.800169  0.802919  0.807857      0.817819 | 788.651531 | 741.686635  631.665316  490.613558          329.802953 | 245.930574 |  0.24149   0.23186  0.217443       0.196485 | 1.161065  1.090868  0.988346        0.844963 | 788.651531   696.12974  568.334215  413.613254  245.930574 |    0.0 | 788.651531 | 245.930574 |
    | 23:30 | 9.204848   8.68489  7.898607              6.78871 | 561.718812  507.409929  430.419109   332.26358 | 108.871395  103.568848   95.550307        84.231531 | 107.048476  101.848901   93.986074     82.887101 | 1.806127  1.745713  1.652127  1.514793 |  1.33898  1.339622  1.340808          1.343107 |  0.171332   0.147126   0.107441      0.043918 | 0.033503  0.060825  0.104287      0.173622 | 0.799657  0.801998  0.806129      0.814205 | 797.187983 | 757.692619  661.225244  530.357066          374.735822 | 292.146753 | 0.242799  0.234565  0.221794       0.203009 | 1.170712  1.110447  1.018954        0.888862 | 797.187983  719.670909  604.027394  457.592933  292.146753 |    0.0 | 797.187983 | 292.146753 |
    | 00:00 | 9.259274  8.823406  8.128283             7.124405 |  567.55985  521.613564  452.269152  360.651796 | 109.426434  104.981439   97.892546        87.654959 | 107.592737  103.234061   96.282828     86.244049 | 1.812388  1.761916  1.679757  1.557039 | 1.338919  1.339441  1.340431          1.342316 |   0.17378   0.153723   0.119445       0.06417 | 0.029884  0.053014  0.090825      0.151183 | 0.799238  0.801234  0.804708      0.811315 |      800.0 |  768.26063  686.136262  566.759817          418.343422 | 338.224051 | 0.243652  0.236774  0.225566       0.208794 |  1.17701  1.126519  1.045729        0.928346 |      800.0  738.035642  635.572452  499.005608  338.224051 |    0.0 |      800.0 | 338.224051 |
    | 00:30 | 9.285766  8.932671  8.326352             7.424667 | 570.413695  532.953148  471.535998  386.998447 |   109.6966   106.09573   99.912466        90.717049 | 107.857656  104.326713   98.263523     89.246675 | 1.815431  1.774641  1.703387   1.59429 | 1.338889  1.339304  1.340127          1.341683 |  0.174966    0.15885   0.129518      0.081493 | 0.027537   0.04671  0.079392       0.13194 |   0.7989  0.800602  0.803536      0.808985 | 797.264896 | 773.437273   706.18808  599.352292          459.860961 | 383.239129 | 0.244066  0.238508  0.228791        0.21389 | 1.180076  1.139187  1.068786        0.963556 | 797.264896  751.141006  662.632884  537.250875  383.239129 |    0.0 | 797.264896 | 383.239129 |
    | 01:00 | 9.285447  9.013672  8.493775             7.690485 | 570.379363  541.436449  488.127688  411.075059 | 109.693354  106.921774  101.619848        93.427865 | 107.854473  105.136716   99.937749     91.904849 | 1.815394  1.784043  1.723222  1.626869 | 1.338889  1.339205  1.339884          1.341175 |  0.174952   0.162607   0.137838      0.096248 | 0.026397  0.041809  0.069791      0.115475 | 0.798635  0.800083  0.802568      0.807094 | 789.260928 | 773.374945  721.282922  627.779681          498.641083 | 426.387788 | 0.244061   0.23979  0.231498       0.218343 | 1.180039  1.148574  1.088251        0.994653 | 789.260928  759.015041   684.98655   571.84348  426.387788 |    0.0 | 789.260928 | 426.387788 |
    | 01:30 | 9.259572  9.067525  8.631642             7.923021 | 567.591989  547.112954  502.000833  432.716665 | 109.429481  107.470977   103.02582        95.799282 | 107.595724  105.675253  101.316418     94.230214 | 1.812422  1.790279  1.739462  1.655076 | 1.338918   1.33914  1.339694          1.340767 |  0.173793   0.165084    0.14456      0.108734 | 0.026404  0.038214  0.061855      0.101443 | 0.798432  0.799661   0.80177      0.805552 | 776.351729 | 768.318876  731.427823  651.799314          534.156281 | 466.990071 | 0.243656   0.24064  0.233713       0.222197 | 1.177045  1.154812  1.104266        1.021802 | 776.351729  761.786351  702.520611  602.413217  466.990071 |    0.0 | 776.351729 | 466.990071 |
    | 02:00 | 9.209501   9.09546  8.741151             8.123567 | 562.217021  550.068838  513.155867  451.815223 | 108.918847  107.755854  104.142599        97.844455 | 107.095006  105.954598   102.41151     96.235671 | 1.806663  1.793509  1.752304  1.679192 | 1.338975  1.339107  1.339548          1.340438 |  0.171542   0.166364   0.149819      0.119202 | 0.027504   0.03584  0.055439      0.089561 | 0.798287  0.799321  0.801112      0.804289 | 758.969595 | 758.592577  736.724488  671.274979           565.99741 | 504.490785 | 0.242872   0.24108  0.235464       0.225489 |  1.17125  1.158048  1.116977         1.04518 | 758.969595  759.671646  715.223656  628.701415  504.490785 |    0.0 | 758.969595 | 504.490785 |
    | 02:30 | 9.136676   9.09879  8.823585             8.293501 | 554.444329  550.421706  521.632041  468.313266 | 108.176173  107.789812  104.983265        99.577442 | 106.366755  105.987896   103.23585     97.935006 | 1.798269  1.793893  1.761937   1.69948 | 1.339059  1.339103  1.339441          1.340176 |  0.168243   0.166516   0.153732      0.127865 | 0.029646   0.03461  0.050416      0.079593 | 0.798195  0.799055  0.800573      0.803252 |   737.5984 | 744.582446  737.357436  686.168795          593.868962 | 538.456373 | 0.241728  0.241132  0.236777       0.228258 |  1.16282  1.158433  1.126539        1.064964 |   737.5984  752.961973  723.175704  650.554663  538.456373 |    0.0 |   737.5984 | 538.456373 |
    | 03:00 | 9.042603  9.078896  8.880292             8.434261 | 544.482412  548.315256  527.502282  482.197751 |  107.21682  107.586941  105.561562       101.012927 | 105.426032  105.788965  103.802917     99.342614 | 1.787394  1.791594  1.768547  1.716185 |  1.33917  1.339127  1.339369          1.339969 |   0.16394   0.165606     0.1564        0.1349 | 0.032784  0.034457  0.046675      0.071338 | 0.798152  0.798853  0.800135        0.8024 | 712.756821 | 726.722207  733.581096  696.531276          617.581842 | 568.568932 | 0.240247  0.240819  0.237678       0.230538 | 1.151926  1.156129  1.133115        1.081334 | 712.756821   742.00832  726.536729  667.916378  568.568932 |    0.0 | 712.756821 | 568.568932 |
    | 03:30 | 8.928833  9.037211  8.912666              8.54733 |  532.55283  543.914027  530.868059  493.494155 | 106.056591  107.161826  105.891715          102.166 | 104.288334  105.372105  104.126659    100.473295 | 1.774195   1.78677  1.772315   1.72954 | 1.339309  1.339177  1.339329          1.339809 |  0.158671   0.163692   0.157916      0.140463 | 0.036875  0.035317   0.04412       0.06463 | 0.798155  0.798709  0.799783        0.8017 | 684.982427 | 705.477773  725.706455  702.490198          637.044297 | 594.618081 | 0.238448  0.240161  0.238191       0.232359 | 1.138743  1.151301  1.136868        1.094474 | 684.982427  727.207191  725.534346   680.81683  594.618081 |    0.0 | 684.982427 | 594.618081 |
    | 04:00 | 8.796941  8.975193  8.922134             8.634208 | 518.885007  537.398356  531.854371  502.260825 | 104.711551   106.52937  105.988269       103.051987 | 102.969414  104.751931  104.221339    101.342077 | 1.758827   1.77958  1.773416  1.739764 | 1.339475  1.339252  1.339317           1.33969 |  0.152471   0.160826   0.158358      0.144684 |  0.04188  0.037133  0.042666      0.059322 | 0.798202  0.798617  0.799506      0.801127 | 654.817086 |  681.33299   714.08779  704.238816          652.251594 | 616.491303 | 0.236353  0.239182  0.238342       0.233754 | 1.123449  1.144115  1.137966        1.104564 | 654.817086  708.986651  720.451181  689.362192  616.491303 |    0.0 | 654.817086 | 616.491303 |
    | 04:30 | 8.648512  8.894322  8.910139             8.696408 | 503.711457  528.959628   530.60494  508.583631 | 103.197861  105.704641  105.865942       103.686303 | 101.485118  103.943218  104.101387    101.964075 | 1.741444  1.770181  1.772021  1.747063 | 1.339671  1.339352  1.339332          1.339607 |  0.145375   0.157058   0.157798      0.147679 | 0.047761  0.039853  0.042235       0.05529 | 0.798294  0.798574  0.799296       0.80066 |  622.79402 | 654.776659  699.109917  702.023924          663.274959 | 634.163297 | 0.233983  0.237901  0.238151       0.234749 | 1.106225  1.134742  1.136575        1.111784 |  622.79402  687.793266  711.612401  693.723091  634.163297 |    0.0 |  622.79402 | 634.163297 |
    | 05:00 | 8.485124  8.796076  8.878128             8.735438 | 487.263534  518.795912  527.277695  512.570955 | 101.531629  104.702727  105.539495       104.084338 |  99.851243  102.960761  103.781279     102.35438 |   1.7222  1.758725  1.768295  1.751635 | 1.339896  1.339476  1.339372          1.339555 |  0.137412    0.15243   0.156299      0.149546 | 0.054483   0.04343  0.042758      0.052427 | 0.798429  0.798576  0.799143      0.800285 | 589.426711 |  626.29106  681.176306  696.134151          670.250238 | 647.684806 | 0.231358  0.236339  0.237644       0.235372 | 1.087246  1.123349  1.132864        1.116314 | 589.426711  664.080216  699.373788  694.123118  647.684806 |    0.0 | 589.426711 | 647.684806 |
    | 05:30 |  8.30834  8.681925  8.827542             8.752795 | 469.767709  507.108026  522.040608  514.349046 |   99.72878  103.538614  105.023617       104.261347 |  98.083404  101.819254  103.275419    102.527952 | 1.701245  1.745365  1.762399  1.753666 | 1.340154  1.339626  1.339436          1.339533 |  0.128612   0.146984   0.153918      0.150374 | 0.062013   0.04782  0.044176      0.050637 | 0.798607  0.798622  0.799044      0.799988 | 555.199767 | 596.342165  660.698304   686.88882           673.36664 | 657.171323 | 0.228499  0.234517   0.23684       0.235649 | 1.066691  1.110103  1.126998        1.118328 | 555.199767  638.296819  684.110637  690.827627  657.171323 |    0.0 | 555.199767 | 657.171323 |
    | 06:00 | 8.119692  8.553318  8.759804             8.749954 | 451.442412  494.096056  515.067906  514.057764 |   97.80494  102.227075  104.332825        104.23237 |  96.196924  100.533184  102.598042    102.499538 | 1.678728  1.730246  1.754486  1.753334 | 1.340444  1.339801  1.339523          1.339536 |  0.119003   0.140755   0.150708      0.150239 | 0.070318  0.052981  0.046431      0.049838 |  0.79883  0.798709  0.798993       0.79976 | 520.561767 | 565.371586  638.085636  674.627631          672.855888 | 662.792006 | 0.225426  0.232456  0.235761       0.235604 | 1.044729   1.09517  1.119141        1.117998 | 520.561767  610.879566  666.207712  684.133126  662.792006 |    0.0 | 520.561767 | 662.792006 |
    | 06:30 | 7.920673  8.411673  8.676313              8.72836 | 432.495432  479.956351  506.536699  511.846728 |  95.775336  100.782574  103.481375       104.012155 |  94.206734   99.116734  101.763127    102.283599 | 1.654793   1.71351  1.744707  1.750806 |  1.34077  1.340001  1.339633          1.339565 |   0.10861    0.13378   0.146714      0.149208 | 0.079368  0.058876  0.049474      0.049956 | 0.799098  0.798838  0.798986      0.799591 | 485.920022 | 533.790269  613.738284  659.701316          668.981989 | 664.759057 | 0.222158  0.230173  0.234428       0.235259 | 1.021529  1.078709  1.109452        1.115492 | 485.920022  582.244723  646.050378  674.357457  664.759057 |    0.0 | 485.920022 | 664.759057 |
    | 07:00 | 7.712729  8.258371  8.578433             8.689425 | 413.121854  464.878991  496.624021  507.871874 |  93.654709   99.219186  102.483191       103.615092 |  92.127287   97.583707  100.784326    101.894247 | 1.629579  1.695297  1.733204  1.746244 | 1.341135  1.340229  1.339766          1.339616 |  0.097459   0.126089   0.141979      0.147344 | 0.089134  0.065468  0.053257      0.050925 | 0.799414  0.799006  0.799022      0.799477 |  451.63713 | 501.973849   588.03976  642.463406          662.031824 | 663.317774 | 0.218713  0.227688  0.232859       0.234637 | 0.997252  1.060876  1.098087        1.110974 |  451.63713  552.782481  624.016986  661.830928  663.317774 |    0.0 |  451.63713 | 663.317774 |
    | 07:30 |  7.49725  8.094746   8.46749             8.634519 | 393.502502  449.045708  485.504293   502.29242 |  91.457244   97.550531  101.351794       103.055166 |  89.972496   95.947455   99.674901    101.345194 | 1.603222  1.675738  1.720116    1.7398 |  1.34154  1.340484  1.339921           1.33969 |  0.085574   0.117715   0.136543      0.144699 | 0.099587  0.072725  0.057739      0.052687 | 0.799777  0.799216  0.799098      0.799411 | 418.029154 | 470.259558  561.351753  623.263141          652.306651 | 658.737406 | 0.215111  0.225018  0.231074       0.233759 | 0.972055  1.041823  1.085197          1.1046 | 418.029154  522.852578  600.472538  646.888463  658.737406 |    0.0 | 418.029154 | 658.737406 |
    | 08:00 | 7.275567  7.922083  8.344767             8.564971 | 373.802862  432.628244  473.347179  495.268228 |  89.196514   95.789712  100.100259       102.345911 |  87.755668    94.22083   98.447669    100.649712 | 1.575854  1.654963  1.705574  1.731619 | 1.341989  1.340768  1.340099          1.339785 |  0.072981   0.108685   0.130442      0.141324 | 0.110699  0.080613   0.06288      0.055188 | 0.800191  0.799466  0.799213       0.79939 | 385.365238 | 438.944532  534.010076  602.439537          640.114597 | 651.302927 | 0.211368  0.222181   0.22909       0.232643 | 0.946084  1.021694  1.070928        1.096523 | 385.365238  492.781296  575.763586   629.86284  651.302927 |    0.0 | 385.365238 | 651.302927 |
    | 08:30 | 7.048946  7.741614  8.211497             8.482059 | 354.172426  415.787113  460.315838  486.957549 |  86.885431   93.949278   98.741164       101.500373 |  85.489464   92.416136   97.114967     99.820594 | 1.547599  1.633094  1.689706  1.721838 | 1.342485  1.341083  1.340301          1.339901 |  0.059701   0.099026   0.123708      0.137261 | 0.122442  0.089103  0.068643      0.058381 | 0.800656  0.799757  0.799366       0.79941 |  353.86844 | 408.285338  506.321825  580.316575          625.764172 | 641.307747 | 0.207502  0.219194  0.226925       0.231309 | 0.919482  1.000627   1.05542         1.08689 |  353.86844  462.859685  550.214315  611.078986  641.307747 |    0.0 |  353.86844 | 641.307747 |
    | 09:00 | 6.818587   7.55451  8.068861             8.387013 | 334.744419  398.670724  446.565525  477.515123 |  84.536213   92.041184   97.286562       100.531085 |  83.185866   90.545096   95.688612     98.870129 | 1.518579  1.610249  1.672633  1.710588 | 1.343033  1.341429  1.340526          1.340037 |   0.04576   0.088766   0.116374      0.132553 | 0.134788  0.098164  0.074995      0.062222 | 0.801176   0.80009  0.799556       0.79947 | 323.717566 | 378.498544  478.563608  557.199455           609.55877 | 629.047408 | 0.203528  0.216072  0.224594       0.229774 | 0.892382  0.978755  1.038807        1.075841 | 323.717566  433.342869  524.123708  590.849323  629.047408 |    0.0 | 323.717566 | 629.047408 |
    | 09:30 | 6.585615  7.361883  7.917988             8.281008 | 315.635848  381.414839  432.242524  467.090626 |  82.160359   90.076769   95.747954        99.450047 |   80.85615   88.618829   94.179884     97.810084 | 1.488909  1.586541  1.654468  1.697993 | 1.343636   1.34181  1.340775          1.340195 |  0.031181    0.07793   0.108468      0.127234 | 0.147711   0.10777  0.081903      0.066671 | 0.801752  0.800467  0.799783      0.799567 | 295.049792 | 349.762141  450.980734  533.371827          591.792153 | 614.814224 | 0.199461   0.21283  0.222114       0.228056 | 0.864908  0.956202  1.021216        1.063511 | 295.049792  404.450279  497.763696   569.47008  614.814224 |    0.0 | 295.049792 | 614.814224 |
    | 10:00 | 6.351087  7.164783   7.75995             8.165169 | 296.947842  364.142315  417.483377  455.827413 |  79.768634   88.066736   94.136274        98.268707 |  78.510871   86.647829   92.599501     96.651685 | 1.458697  1.562077  1.635323  1.684171 | 1.344299  1.342226   1.34105          1.340372 |  0.015989   0.066541   0.100018      0.121341 | 0.161182  0.117892  0.089338      0.071692 | 0.802386  0.800887  0.800047        0.7997 | 267.963873 |  322.21764  423.787212    509.0939           572.74483 | 598.892831 | 0.195318  0.209483  0.219498       0.226169 | 0.837179  0.933087  1.002769        1.050026 | 267.963873  376.366639  471.378165  547.218521  598.892831 |    0.0 | 267.963873 | 598.892831 |
    | 10:30 | 6.115986  6.964196  7.595761             8.040561 | 278.766204    346.9631  402.414382  443.861554 |  77.371063   86.021146   92.461872        96.997958 |  76.159859   84.641964   90.957615     95.405613 | 1.428048  1.536958  1.615301  1.669234 | 1.345028  1.342681  1.341351          1.340571 |  0.000207   0.054626    0.09105      0.114903 | 0.175175  0.128505   0.09727      0.077249 |  0.80308  0.801353  0.800349      0.799868 | 242.523746 | 295.972664  397.166413  484.601335           552.68128 | 581.556604 |  0.19111  0.206045  0.216762        0.22413 | 0.809303  0.909519   0.98358        1.035509 | 242.523746  349.243566  445.182705   524.35098  581.556604 |    0.0 | 242.523746 | 581.556604 |
    | 11:00 | 5.881222  6.761046   7.42638               7.9082 | 261.162165   329.97443  387.151313  431.321114 |  74.976928   83.949414   90.734515         95.64813 |  73.812217   82.610463   89.263801     94.081998 |  1.39706  1.511282  1.594501  1.653286 | 1.345828  1.343177  1.341679          1.340792 | -0.016139   0.042205    0.08159       0.10795 | 0.189662  0.139582  0.105674      0.083312 | 0.803837  0.801864  0.800688       0.80007 | 218.762374 | 271.103896  371.272285  460.104808          531.847925 | 563.064852 | 0.186852  0.202528  0.213919       0.221952 | 0.781382  0.885603  0.963758        1.020074 | 218.762374  323.201628  419.364983  501.101623  563.064852 |    0.0 | 218.762374 | 563.064852 |
    | 11:30 | 5.647633  6.556193  7.252706             7.769042 | 244.193265   313.26121  371.799347  418.325658 |  72.594778    81.86031   88.963383        94.228989 |  71.476327   80.561928   87.527064     92.690416 | 1.365824  1.485138  1.573016  1.636428 | 1.346704  1.343716  1.342037          1.341034 | -0.033025   0.029304    0.07166      0.100509 | 0.204615  0.151097  0.114522      0.089852 | 0.804656  0.802423  0.801065      0.800307 |  196.68568 | 247.660232   346.23098  435.790129          510.471778 | 543.660735 | 0.182556  0.198944   0.21098       0.219649 | 0.753506  0.861434  0.943403        1.003831 |  196.68568  298.332736  394.085611  477.681836  543.660735 |    0.0 |  196.68568 | 543.660735 |
    | 12:00 | 5.415986  6.350433  7.075583              7.62399 | 227.904337  296.896524  356.453159  404.985955 |   70.23244   79.761968   87.157076        92.749746 |  69.159864   78.504334   85.755835     91.239899 | 1.334429  1.458612  1.550935  1.618753 | 1.347665  1.344301  1.342425          1.341298 | -0.050423   0.015946   0.061284      0.092606 | 0.220005  0.163027  0.123791       0.09684 | 0.805541   0.80303  0.801481      0.800577 | 176.276442 | 225.666027  322.142784  411.818837          488.759667 | 523.569808 | 0.178232  0.195306  0.207958       0.217234 | 0.725763  0.837102  0.922612        0.986881 | 176.276442  274.702751  369.479423  454.280135  523.569808 |    0.0 | 176.276442 | 523.569808 |
    | 12:30 | 5.186981  6.144505  6.895798             7.473894 | 212.328567  280.942257  341.197155  391.403841 |  67.897034   77.661898   85.323621         91.21906 |  66.869809   76.445047   83.957984     89.738937 | 1.302954  1.431786  1.528341  1.600351 | 1.348716  1.344936  1.342844          1.341585 | -0.068307   0.002152   0.050484      0.084266 | 0.235803  0.175347  0.133455       0.10425 | 0.806493  0.803687  0.801935       0.80088 | 157.498082 | 205.124339  299.084253  388.329153          466.897961 |  502.99912 | 0.173893  0.191624  0.204865       0.214718 | 0.698228  0.812689  0.901472        0.969321 | 157.498082  252.354197  345.657036  431.062532   502.99912 |    0.0 | 157.498082 |  502.99912 |
    | 13:00 | 4.961247  5.939083  6.714083             7.319549 | 197.488577  265.449795  326.105815  377.672234 |  65.594993   75.567002   83.470483        89.645048 |  64.612472   74.390832   82.140832     88.195491 | 1.271476  1.404734  1.505311  1.581304 | 1.349865  1.345623  1.343297          1.341897 | -0.086649  -0.012054   0.039282       0.07551 | 0.251982  0.188032  0.143493      0.112058 | 0.807512  0.804393   0.80243      0.801218 | 140.298231 |  186.02008  277.110474  365.437224          445.052707 | 482.136793 | 0.169547  0.187907   0.20171       0.212114 | 0.670973  0.788272  0.880067        0.951241 | 140.298231  231.309013  322.706634  408.173245  482.136793 |    0.0 | 140.298231 | 482.136793 |
    | 13:30 | 4.739351  5.734788  6.531116               7.1617 | 183.397523  250.460777  311.244132  363.875265 |  63.332089    73.48359   81.604577          88.0353 |  62.393513   72.347879   80.311161     86.617004 | 1.240064  1.377527  1.481919  1.561693 | 1.351121  1.346367  1.343785          1.342233 | -0.105422   -0.02665   0.027698      0.066361 |  0.26851  0.201058  0.153881       0.12024 | 0.808601  0.805151  0.802964      0.801589 | 124.612057 | 168.323022  256.257361  343.238569          423.370115 | 461.151993 | 0.165205  0.184166  0.198503       0.209431 | 0.644059  0.763918  0.858472        0.932725 | 124.612057  211.571253  300.695872  385.735689  461.151993 |    0.0 | 124.612057 | 461.151993 |
    | 14:00 | 4.521797  5.532181  6.347523             7.001041 | 170.060183  236.007883  296.668116  350.088503 |   61.11346   71.417403   79.732291        86.396891 |  60.217968   70.321815   78.475234     85.010411 | 1.208784  1.350231  1.458235   1.54159 | 1.352491  1.347171   1.34431          1.342595 | -0.124595  -0.041612   0.015754       0.05684 | 0.285358  0.214402  0.164599      0.128774 | 0.809759   0.80596  0.803539      0.801994 | 110.365288 | 151.990606  236.543949  321.809664          401.977299 | 440.195238 | 0.160874  0.180409  0.195254       0.206679 | 0.617544  0.739692  0.836757        0.913851 | 110.365288  193.129693  279.673861  363.853677  440.195238 |    0.0 | 110.365288 | 440.195238 |
    | 14:30 | 4.309028  5.331773  6.163881             6.838216 | 157.474012  222.115629  282.425346  336.379265 |  58.943632   69.373632   77.859497        84.736398 |  58.090277   68.317733   76.638808     83.382164 | 1.177694  1.322907  1.434322  1.521064 | 1.353986  1.348039  1.344874          1.342984 | -0.144139  -0.056918   0.003468      0.046966 | 0.302494  0.228041  0.175624       0.13764 | 0.810987  0.806822  0.804155      0.802433 |   97.47692 | 136.970524  217.974612  301.209606          380.983239 |  419.39897 | 0.156564  0.176644  0.191972       0.203868 | 0.591476   0.71565  0.814989        0.894694 |   97.47692  175.960302  259.673157  342.612768   419.39897 |    0.0 |   97.47692 |  419.39897 |
    | 15:00 | 4.101431  5.134021  5.980716             6.673824 | 145.630152  208.801168   268.55555  322.806985 |  56.826554   67.356946   75.991574        83.059916 |  56.014311   66.340209   74.807159     81.738239 | 1.146849  1.295611  1.410241  1.500182 | 1.355615  1.348975  1.345479          1.343401 | -0.164022  -0.072545  -0.009138       0.03676 | 0.319889  0.241949  0.186936      0.146815 | 0.812285  0.807737  0.804812      0.802906 |  85.861607 | 123.203056  200.541191   281.48181          360.479883 | 398.878331 |  0.15228  0.172879  0.188664       0.201007 |   0.5659  0.691844  0.793226        0.875319 |  85.861607  160.028534  240.711722  322.081705  398.878331 |    0.0 |  85.861607 | 398.878331 |
    | 15:30 |  3.89934  4.939333   5.79851             6.508416 | 134.514387  196.075069  255.091212  309.423622 |  54.765626   65.371514   74.133427        81.373079 |  53.993405   64.393333   72.985096     80.084158 |   1.1163  1.268395  1.386046  1.479003 | 1.357386  1.349983  1.346128          1.343848 | -0.184213   -0.08847  -0.022046      0.026239 | 0.337509  0.256105  0.198515      0.156281 | 0.813654  0.808704  0.805512      0.803414 |  75.431719 | 110.623148   184.22499  262.655699          340.543357 | 378.732102 |  0.14803  0.169121  0.185338       0.198103 | 0.540855   0.66832  0.771522        0.855789 |  75.431719  145.291441  222.794818   302.31391  378.732102 |    0.0 |  75.431719 | 378.732102 |
    | 16:00 | 3.703038  4.748073  5.617699             6.342501 | 124.108033  183.942071  242.058178  296.274113 |  52.763728   63.421032   72.289511        79.681073 |  52.030382   62.480728   71.176988      78.42501 |  1.08609  1.241308  1.361791  1.457584 | 1.359312  1.351069  1.346823          1.344325 | -0.204679   -0.10467  -0.035236      0.015422 | 0.355324  0.270485  0.210341      0.166018 | 0.815093  0.809725  0.806253      0.803957 |  66.099094 |  99.162232  168.998621  244.748346          321.235236 | 359.043749 |  0.14382  0.165377     0.182       0.195165 | 0.516374  0.645119  0.749927        0.836163 |  66.099094  131.699571  205.916804  283.348979  359.043749 |    0.0 |  66.099094 | 359.043749 |
    | 16:30 | 3.512759  4.560558  5.438678             6.176547 | 114.388768   172.40181  229.476254  283.396834 |  50.823254   61.508747   70.463849        77.988663 |  50.127591   60.605579   69.386779     76.765465 | 1.056259  1.214393  1.337523  1.435979 | 1.361402  1.352236  1.347566          1.344834 | -0.225385  -0.121121  -0.048689      0.004327 | 0.373299  0.285066  0.222395      0.176009 |   0.8166  0.810799  0.807037      0.804535 |  57.776495 |  88.749805  154.827699   227.76606          302.603839 |  339.88254 | 0.139655  0.161651  0.178659       0.192199 | 0.492486  0.622277  0.728485        0.816492 |  57.776495  119.198674  190.062829  265.214165   339.88254 |    0.0 |  57.776495 |  339.88254 |
    | 17:00 | 3.328693  4.377066  5.261802              6.01098 | 105.331374  161.449499  217.359802  270.824077 |   48.94614   59.637486   68.660058        76.300205 |  48.286929   58.770656   67.618016     75.109797 | 1.026845  1.187691  1.313287  1.414236 | 1.363667   1.35349   1.34836          1.345376 | -0.246296  -0.137801  -0.062386      -0.00703 | 0.391401  0.299825  0.234658      0.186235 | 0.818176  0.811927  0.807863      0.805147 |  50.378786 |  79.314757   141.67236  211.705875          284.685526 | 321.304702 |  0.13554   0.15795  0.175318       0.189213 | 0.469214  0.599827  0.707237        0.796825 |  50.378786  107.731194  175.210387  247.925814  321.304702 |    0.0 |  50.378786 | 321.304702 |
    | 17:30 | 3.150988  4.197834  5.087387              5.84619 |  96.908424  151.076567  205.718306  258.582534 |  47.133894   57.809676   66.881366        74.619673 |  46.509875   56.978341   65.873865       73.4619 | 0.997877  1.161238  1.289123  1.392401 | 1.366118  1.354836  1.349208          1.345954 | -0.267378  -0.154685  -0.076309     -0.018631 | 0.409596  0.314739   0.24711      0.196679 |  0.81982  0.813107  0.808732      0.805795 |  43.823868 |  70.786484  129.488629  196.556954           267.50596 | 303.354596 |  0.13148  0.154279  0.171984       0.186212 | 0.446581  0.577795  0.686218        0.777207 |  43.823868   97.237567  161.330746  231.490735  303.354596 |    0.0 |  43.823868 | 303.354596 |
    | 18:00 | 2.979752  4.023065  4.915714             5.682531 |  89.090886  141.271256  194.556917  246.693765 |  45.387626   56.027378   65.130641        72.950674 |  44.797519   55.230654   64.157138     71.825311 | 0.969384  1.135067  1.265069  1.370517 | 1.368765  1.356278  1.350112          1.346568 | -0.288593  -0.171751   -0.09044     -0.030461 |  0.42785  0.329785  0.259734      0.207325 | 0.821529  0.814339  0.809643      0.806478 |  38.033385 |  63.095799  118.229608  182.301873           251.08134 | 286.065864 | 0.127479  0.150642  0.168662       0.183201 | 0.424602  0.556206   0.66546        0.757677 |  38.033385   87.657332  148.390238    215.9075  286.065864 |    0.0 |  38.033385 | 286.065864 |
    | 18:30 | 2.815059  3.852927  4.747031             5.520323 |  81.848659  132.019157  183.876972  235.174665 |  43.708079   54.292304   63.410406        71.296468 |  43.150587   53.529274   62.470309     70.203229 |  0.94139  1.109206  1.241159  1.348623 | 1.371617  1.357823  1.351075           1.34722 | -0.309902  -0.188974  -0.104759     -0.042503 | 0.446126   0.34494  0.272511      0.218156 | 0.823302  0.815624  0.810596      0.807196 |  32.933235 |  56.175643  107.846523  168.917805          235.419559 | 269.462568 |  0.12354  0.147042  0.165356       0.180187 | 0.403291  0.535081  0.644993        0.738272 |  32.933235   78.930063  136.351411   201.16765  269.462568 |    0.0 |  32.933235 | 269.462568 |
    | 19:00 | 2.656947  3.687557  4.581554             5.359853 |  75.151046  123.303717  173.676482  224.037913 |   42.09565   52.605846   61.722864         69.65999 |  41.569471   51.875566   60.815538      68.59853 | 0.913914  1.083683  1.217425  1.326755 | 1.374686  1.359474    1.3521          1.347913 | -0.331268  -0.206331   -0.11925     -0.054742 | 0.464391  0.360182  0.285423      0.229157 | 0.825137  0.816959  0.811591      0.807949 |  28.453911 |   49.96165   98.289616  156.377579          220.521309 | 253.560268 | 0.119667  0.143484  0.162071       0.177175 | 0.382661  0.514436   0.62484        0.719024 |  28.453911   70.996141  125.174047  187.256811  253.560268 |    0.0 |  28.453911 | 253.560268 |
    | 19:30 | 2.505425   3.52706  4.419469             5.201379 |  68.967154  115.106678   163.95058    213.2924 |  40.550424   50.969099    60.06992        68.043866 |  40.054252   50.270603   59.194692      67.01379 | 0.886972  1.058521  1.193895  1.304946 | 1.377978  1.361237   1.35319          1.348647 | -0.352649  -0.223797  -0.133893     -0.067162 | 0.482606  0.375489  0.298454      0.240311 | 0.827032  0.818344  0.812628      0.808738 |  24.530701 |  44.392548     89.5089  144.650644            206.3811 | 238.367054 | 0.115862  0.139971  0.158811       0.174167 | 0.362718  0.494287  0.605024        0.699963 |  24.530701   63.797382  114.816054  174.155716  238.367054 |    0.0 |  24.530701 | 238.367054 |
    | 20:00 | 2.360473  3.371519  4.260936              5.04513 |  63.266248  107.408485  154.691942  202.943641 |  39.072193   49.382882   58.453196        66.450434 |  38.604727    48.71519   57.609364     65.451301 |  0.86058   1.03374  1.170595  1.283228 | 1.381503  1.363117  1.354348          1.349424 | -0.374007   -0.24135  -0.148671     -0.079748 | 0.500736  0.390838  0.311586      0.251605 | 0.828985  0.819778  0.813705       0.80956 |  21.103761 |  39.410449   81.454793  133.703904          192.988209 | 223.884516 | 0.112127  0.136506  0.155578        0.17117 | 0.343469  0.474644  0.585564        0.681115 |  21.103761   57.277526  105.234238  161.841128  223.884516 |    0.0 |  21.103761 | 223.884516 |
    | 20:30 | 2.221289    3.2212  4.106073             4.891309 |  57.989957  100.198663  145.890245  192.994163 |  37.652791   47.849927   56.873888        64.881761 |   37.21289   47.212003   56.060725     63.913091 | 0.834603  1.009393  1.147545  1.261627 | 1.384358  1.365117  1.355576          1.350247 | -0.394922  -0.258939  -0.163569     -0.092484 | 0.518854  0.406214  0.324801      0.263023 | 0.831171  0.821274  0.814822      0.810418 |  18.118088 |  34.961025    74.08875  123.501401          180.327549 | 210.108646 | 0.108519  0.133095  0.152377       0.168186 | 0.325254  0.455544  0.566473        0.662503 |  18.118088    51.40198   96.383045  150.286681  210.108646 |    0.0 |  18.118088 | 210.108646 |
    | 21:00 | 2.089694  3.075701  3.955051             4.740091 |  53.179529   93.435206  137.537911  183.443666 |  36.310785   46.366121   55.333765        63.339631 |  35.896944   45.757012   54.550511     62.400908 | 0.809405  0.985423  1.124775   1.24017 | 1.388382  1.367245  1.356878          1.351117 | -0.416074  -0.276592  -0.178561     -0.105358 | 0.536279  0.421552  0.338086       0.27455 | 0.833063  0.822786  0.815982      0.811309 |  15.523427 |  31.002854   67.342144  114.011133          168.380204 | 197.030653 | 0.104937  0.129733   0.14921       0.165219 | 0.307418  0.436938  0.547774        0.644149 |  15.523427   46.078547   88.226249  139.463185  197.030653 |    0.0 |  15.523427 | 197.030653 |
    | 21:30 | 1.964064  2.935251  3.807915             4.591637 |   48.74868   87.107274  129.619796  174.290194 |  35.029596   44.933808   53.833263        61.825691 |  34.640635   44.352514   53.079148     60.916368 | 0.784704   0.96188  1.102298   1.21888 |  1.39266  1.369504  1.358257          1.352035 | -0.437137   -0.29426  -0.193636     -0.118354 | 0.553877  0.436895  0.351417      0.286174 | 0.835122  0.824359  0.817178      0.812234 |  13.274094 |  27.467791   61.180316  105.193382          157.125167 | 184.637741 | 0.101422  0.126424   0.14608       0.162273 | 0.290233  0.418861  0.529473         0.62607 |  13.274094   41.276686   80.715496   129.34107  184.637741 |    0.0 |  13.274094 | 184.637741 |
    | 22:00 | 1.844679  2.799852  3.664723             4.446081 |  44.684378   81.193642  122.121797  165.529401 |  33.812105   43.553001   52.372984        60.341308 |  33.446787   42.998521   51.647226     59.460811 | 0.760576  0.938774  1.080128  1.197778 | 1.397184  1.371898  1.359715          1.353005 | -0.458007  -0.311917  -0.208777     -0.131457 | 0.571321  0.452193  0.364783       0.29788 |  0.83728  0.825976  0.818412      0.813193 |  11.328777 |  24.324552   55.559854   97.010869          146.538658 | 172.913914 | 0.097985  0.123172  0.142988       0.159349 |  0.27375  0.401314  0.511576        0.608283 |  11.328777   36.955247   73.807106  119.888701  172.913914 |    0.0 |  11.328777 | 172.913914 |
    | 22:30 | 1.731424  2.669471  3.525526             4.303535 |  40.960491   75.672462  115.029549  157.155081 |  32.657125   42.223374    50.95345        58.887616 |  32.314235   41.694715   50.255258     58.035348 | 0.737026  0.916113  1.058279  1.176884 | 1.401953   1.37443  1.361255          1.354027 | -0.478638  -0.329541  -0.223967     -0.144655 | 0.588496  0.467423  0.378169      0.309655 | 0.839484  0.827634  0.819683      0.814184 |   9.650285 |  21.533498   50.438739   89.427253          136.595004 | 161.840493 | 0.094628  0.119977  0.139937       0.156451 | 0.257967  0.384301  0.494093        0.590801 |   9.650285   33.071295   67.460677  111.073294  161.840493 |    0.0 |   9.650285 | 161.840493 |
    | 23:00 | 1.624152  2.544072  3.390358             4.164091 |   37.55164   70.522575  108.327992  149.159623 |  31.563169   40.944541      49.575         57.46556 |  31.241524   40.440716   48.903577     56.640908 | 0.714052  0.893904  1.036763  1.156215 |  1.40696  1.377103   1.36288          1.355104 | -0.498988  -0.347108   -0.23919     -0.157933 | 0.605364  0.482566  0.391559      0.321485 | 0.841731  0.829333   0.82099      0.815207 |   8.205287 |  19.057989   45.777559   82.406707          127.267374 | 151.396604 | 0.091353  0.116841  0.136929       0.153581 | 0.242881  0.367821  0.477029        0.573635 |   8.205287   29.584714   61.637355   102.86187  151.396604 |    0.0 |   8.205287 | 151.396604 |
    | 23:30 | 1.522705  2.423601  3.259236             4.027825 |  34.433742   65.723212  102.001635  141.534255 |   30.52861   39.715974   48.237816        56.075918 |  30.227055   39.236007    47.59236     55.278252 | 0.691652  0.872153  1.015592  1.135784 | 1.412198   1.37992  1.364593          1.356238 | -0.519014  -0.364592  -0.254428     -0.171278 | 0.621889  0.497601  0.404938      0.333359 | 0.844016   0.83107  0.822332      0.816261 |   6.964036 |  16.864616   41.539137   75.914252          118.528202 |  141.55969 | 0.088159  0.113766  0.133964       0.150741 | 0.228485  0.351873  0.460388        0.556795 |   6.964036   26.458314   56.299798   95.221676   141.55969 |    0.0 |   6.964036 |  141.55969 |
    | 00:00 | 1.426911  2.307229  3.132404             3.894778 |  31.584049   61.224961   96.045852  134.268144 |  29.551697   38.529211   46.944382        54.719097 |  29.269113   38.072289   46.324044     53.947779 | 0.669824  0.850719  0.994813  1.115604 | 1.417657  1.381992  1.366392          1.357429 | -0.538676  -0.381596  -0.269637     -0.184679 |  0.63804   0.51261  0.418299      0.345263 | 0.846337  0.833013  0.823724      0.817345 |   5.900084 |  14.923036   37.688543    69.92703          110.348308 | 132.306032 | 0.085048  0.110803  0.131051       0.147933 | 0.214771  0.336797  0.444203        0.540287 |   5.900084   23.657758    51.43354   88.118195  132.306032 |    0.0 |   5.900084 | 132.306032 |
    | 00:30 | 1.336588  2.196468  3.009323             3.765075 |  28.981165   57.069368   90.419965  127.355093 |  28.630579   37.399663   45.689192        53.396385 |  28.365883   36.964677   45.093229     52.650754 | 0.648563    0.8299  0.974348  1.095697 | 1.423324   1.38512  1.368286          1.358681 | -0.557935  -0.398862  -0.284863     -0.198116 | 0.653786  0.526923  0.431576       0.35719 | 0.848692  0.834624  0.825111      0.818463 |    4.99001 |  13.205779   34.193322   64.387467          102.704445 | 123.610408 |  0.08202  0.107848  0.128177       0.145159 | 0.201731  0.321886  0.428409        0.524128 |    4.99001   21.151446   46.956938   81.528529  123.610408 |    0.0 |    4.99001 | 123.610408 |
    | 01:00 | 1.251546  2.090672  2.890346             3.638678 |  26.605033   53.214622   85.125698  120.780064 |  27.763317   36.320753   44.475863        52.107381 |  27.515462   35.906718   43.903462     51.386781 | 0.627864  0.809595  0.954263  1.076062 | 1.429182  1.388388  1.370274          1.359994 | -0.576753  -0.415934  -0.300046     -0.211583 | 0.669097  0.541543  0.444842      0.369118 | 0.851078  0.836496  0.826558      0.819606 |   4.213141 |  11.688049    31.03043   59.281788           95.564334 | 115.449111 | 0.079077  0.104961  0.125353       0.142421 | 0.189353  0.307534  0.413054        0.508312 |   4.213141   18.910389   42.868236   75.414052  115.449111 |    0.0 |   4.213141 | 115.449111 |
    | 01:30 | 1.171588   1.98914  2.775305             3.515624 |  24.436914   49.620485   80.141153  114.532429 |  26.947901   35.285326   43.302666        50.852472 |  26.715881   34.891399   42.753047     50.156242 | 0.607724  0.789688  0.934539  1.056713 | 1.435213  1.391807  1.372357           1.36137 | -0.595095  -0.432882  -0.315188     -0.225067 | 0.683949  0.555893  0.458025      0.381043 | 0.853491  0.838323  0.828024      0.820779 |   3.551298 |  10.347523   28.153842   54.574017           88.901539 | 107.796655 | 0.076219  0.102129  0.122575       0.139718 | 0.177625   0.29366   0.39812        0.492846 |   3.551298    16.90807   39.130401   69.745936  107.796655 |    0.0 |   3.551298 | 107.796655 |
    | 02:00 | 1.096511  1.892105  2.664215             3.395913 |  22.459355    46.28188   75.453419  108.599804 |  26.182265    34.29576   42.169765         49.63165 |  25.965113   33.921049   41.642147     48.959127 | 0.588137  0.770242   0.91519  1.037653 | 1.441395  1.395369  1.374538          1.362811 |  -0.61293  -0.449632  -0.330265     -0.238555 |  0.69832  0.570103  0.471134      0.392952 |  0.85593  0.840224  0.829523      0.821978 |   2.988541 |   9.164157   25.547565   50.238163           82.688625 | 100.627714 | 0.073446   0.09936  0.119847       0.137053 | 0.166533    0.2803  0.383612        0.477732 |   2.988541     15.1203   35.719054   64.494759  100.627714 |    0.0 |   2.988541 | 100.627714 |
    | 02:30 | 1.026109  1.799467  2.557021             3.279544 |  20.656144   43.182423   71.047095  102.970215 |  25.464304   33.351038   41.076599         48.44492 |  25.261094   32.994673   40.570209     47.795442 | 0.569097  0.751257  0.896217  1.018891 | 1.447705  1.399068  1.376816          1.364318 |  -0.63023  -0.466162  -0.345265     -0.252037 |  0.71219  0.584088  0.484146      0.404836 | 0.858391  0.842153  0.831049      0.823203 |    2.51094 |   8.119997   23.187635   46.247029            76.89953 |  93.917099 | 0.070759  0.096654  0.117168       0.134426 | 0.156062  0.267452  0.369528        0.462971 |    2.51094   13.525068   32.607611   59.633329   93.917099 |    0.0 |    2.51094 |  93.917099 |
    | 03:00 | 0.960173  1.711109  2.453669             3.166511 |  19.012261   40.306106   66.907492   97.631604 |  24.791884   32.449957   40.022612        47.292198 |  24.601732    32.11109    39.53669     46.665106 | 0.550601  0.732727  0.877622  1.000431 | 1.454116  1.402902  1.379192          1.365893 |  -0.64697  -0.482449   -0.36017       -0.2655 | 0.725545  0.597828  0.497051      0.416685 | 0.860871  0.844106  0.832602      0.824454 |   2.106358 |   7.198992   21.051643   42.575251           71.509018 |  87.640117 | 0.068157  0.094013  0.114539       0.131839 | 0.146194  0.255106  0.355865        0.448564 |   2.106358   12.102391   29.771165   55.135775   87.640117 |    0.0 |   2.106358 |  87.640117 |
    | 03:30 | 0.898492  1.626907  2.354098             3.056795 |  17.513827   37.637732   63.020344   92.571903 |   24.16286   31.591259    39.00718        46.173315 |  23.984923   31.269068   38.540977      45.56795 | 0.532643  0.714651  0.859404  0.982278 | 1.460604  1.406864  1.381668          1.367537 |  -0.66313  -0.498474  -0.374968     -0.278934 | 0.738373  0.611307  0.509835      0.428487 | 0.863368  0.846082  0.834179       0.82573 |   1.764251 |   6.386826    19.11899   39.198953            66.49276 |  81.772748 | 0.065641  0.091435  0.111961       0.129291 | 0.136911  0.243256  0.342619        0.434512 |   1.764251   10.834167   27.186579   50.977397   81.772748 |    0.0 |   1.764251 |  81.772748 |
    | 04:00 | 0.840856  1.546734  2.257518             2.950617 |  16.148043   35.162926     59.3447   87.789938 |  23.575086   30.773649   38.022255        45.090503 |  23.408564   30.467335   37.575178     44.506165 | 0.515219  0.697021  0.841428  0.964476 |  1.46714  1.410949  1.383386          1.369247 | -0.678692  -0.514217  -0.389285     -0.292296 | 0.750667  0.624509   0.52259      0.440244 | 0.865878  0.848079  0.835947      0.827047 |    1.47548 |   5.670752   17.370788   36.095704           61.837828 |  76.289206 | 0.063211  0.088922  0.109483       0.126789 | 0.128192  0.231893  0.330106        0.420844 |    1.47548    9.704034   24.832411   47.154922   76.289206 |    0.0 |    1.47548 |  76.289206 |
    | 04:30 | 0.787056  1.470459  2.165701             2.847454 |  14.903129   32.868128    55.93683   83.251798 |  23.026429   29.995796   37.085906        44.038451 |  22.870561   29.704588   36.657013     43.474544 | 0.498325  0.679834  0.824039  0.946945 | 1.473696  1.415149  1.386069          1.371031 | -0.693647  -0.529661  -0.403782     -0.305637 | 0.762425  0.637421  0.534695      0.451889 | 0.868398  0.850093  0.837415      0.828348 |   1.232151 |   5.039439   15.789754   33.255283            57.50042 |   71.17562 | 0.060866  0.086471  0.107013       0.124323 | 0.120018  0.221007   0.31771        0.407497 |   1.232151    8.697229   22.688836   43.604649    71.17562 |    0.0 |   1.232151 |   71.17562 |
    | 05:00 | 0.736885   1.39795  2.077018             2.747509 |  13.768268   30.740569   52.725275   78.956664 |   22.51478   29.256348   36.181507        43.019204 |  22.368848   28.979499   35.770176      42.47509 | 0.481956  0.663082  0.806943  0.929727 | 1.480245  1.419456  1.388856          1.372887 | -0.707985   -0.54479  -0.418193     -0.318917 | 0.773646   0.65003   0.54698      0.463494 | 0.870925  0.852124  0.839003      0.829691 |   1.027458 |   4.482835   14.360101   30.634033           53.469901 |  66.404742 | 0.058607  0.084085  0.104582       0.121897 | 0.112364   0.21059  0.305663        0.394498 |   1.027458    7.800455   20.737557   40.322469   66.404742 |    0.0 |   1.027458 |  66.404742 |
    | 05:30 | 0.690139  1.329072   1.99181             2.650752 |  12.733541   28.768253   49.713673   74.893714 |  22.038063   28.553932   35.312553        42.032473 |  21.901388   28.290724   34.918096     41.507521 | 0.466107  0.646761  0.790218  0.912824 | 1.486761   1.42386  1.391739          1.374814 | -0.721703  -0.559587  -0.432442     -0.332124 | 0.784336  0.662325  0.559182      0.475018 | 0.873455  0.854169  0.840664      0.831055 |   0.855551 |   3.992031   13.067429   28.226991            49.72658 |  61.953676 | 0.056431  0.081762  0.102202       0.119513 |  0.10521   0.20063  0.294015        0.381848 |   0.855551    7.001761   18.961716   37.293892   61.953676 |    0.0 |   0.855551 |  61.953676 |
    | 06:00 | 0.644317  1.264259   1.90985             2.557159 |  11.740487   26.955641    46.88537    71.05269 |  21.570775   27.892963    34.47672        41.078006 |  21.443175   27.642591   34.098496     40.571589 | 0.449948  0.631004  0.773831  0.896242 | 1.484233  1.428309  1.394721          1.376813 | -0.733129   -0.57391   -0.44654     -0.345245 | 0.797072   0.67441  0.571193      0.486449 |  0.87882  0.856371  0.842323       0.83244 |   0.711411 |   3.559147   11.908569    26.01334           46.252039 |  57.802457 | 0.054567  0.079521  0.099869       0.117172 | 0.099415    0.1912   0.28274        0.369546 |   0.711411    6.313664    17.33685   34.502687   57.802457 |    0.0 |   0.711411 |  57.802457 |
    | 06:30 |  0.60536  1.202242  1.831402             2.466588 |  10.912707   25.260555   44.241189   67.419107 |  21.173487   27.260509   33.676707        40.154362 |  21.053602   27.022418   33.314018     39.665881 | 0.435667  0.615528   0.75785  0.879963 | 1.490265  1.432876  1.397784          1.378885 | -0.745236  -0.588001  -0.460415     -0.358283 | 0.803763  0.685836  0.583071      0.497772 | 0.879242  0.858281   0.84405      0.833835 |   0.590745 |    3.19024   10.851313   23.986688           43.024852 |  53.934206 | 0.052622  0.077324  0.097592       0.114871 | 0.093355  0.182122  0.271882        0.357577 |   0.590745    5.682799   15.868535   31.924513   53.934206 |    0.0 |   0.590745 |  53.934206 |
    | 07:00 | 0.567731  1.143428  1.756227             2.379034 |  10.127549   23.688565   41.765067    63.98451 |  20.789739   26.660726    32.91007        39.261478 |  20.677306   26.434283   32.562269     38.790336 | 0.421347  0.600458  0.742243  0.863995 | 1.496689   1.43751  1.400931           1.38103 | -0.757264  -0.601728  -0.474079     -0.371221 |  0.81164  0.697128  0.594709      0.508995 |  0.88057  0.860348  0.845768      0.835253 |   0.489885 |   2.851101    9.894888   22.128009           40.029776 |  50.330062 | 0.050673  0.075187  0.095368       0.112611 |  0.08746  0.173464  0.261412        0.345944 |   0.489885    5.112604   14.537711   29.545879   50.330062 |    0.0 |   0.489885 |  50.330062 |
    | 07:30 |  0.53265  1.087764  1.684165             2.293746 |    9.40834   22.232621   39.444546    60.71254 |   20.43199   26.093063   32.175184         38.39171 |  20.326505   25.877643   31.841654     37.937459 | 0.407482  0.585808  0.726989  0.848208 | 1.502962   1.44219  1.404162          1.382399 | -0.768716  -0.615063  -0.487534     -0.383689 | 0.820358  0.708098  0.606125      0.520207 | 0.883019  0.862446  0.847488      0.836853 |   0.405708 |   2.550785    9.030744   20.421985           37.251793 |  46.993203 | 0.048802  0.073115  0.093193       0.110444 | 0.081962  0.165227  0.251314        0.334968 |   0.405708    4.604531   13.326141   27.354499   46.993203 |    0.0 |   0.405708 |  46.993203 |
    | 08:00 | 0.500073  1.035031  1.615139             2.212158 |   8.751466   20.881899    37.27044   57.650597 |  20.099767   25.555282   31.471246        37.559677 |  20.000733   25.350305   31.151385     37.121583 | 0.394116  0.571549  0.712089  0.832875 | 1.509078  1.446914  1.407471            1.3847 | -0.779574   -0.62802  -0.500766     -0.396402 | 0.828689  0.718686  0.617329      0.530767 | 0.885534  0.864519  0.849222      0.838114 |    0.33556 |    2.28556    8.248603   18.856433           34.675917 |  43.861247 | 0.047009  0.071102  0.091068       0.108267 | 0.076855  0.157384  0.241584        0.323986 |    0.33556    4.152079   12.222326   25.336122   43.861247 |    0.0 |    0.33556 |  43.861247 |
    | 08:30 | 0.469837  0.985092  1.549052             2.133719 |   8.151284     19.6284   35.233591   54.769564 |  19.791414   25.046002   30.797293        36.759748 |  19.698368   24.850915    30.49052     36.337189 | 0.381242  0.557672  0.697537  0.817908 | 1.515023  1.451671  1.410854          1.387067 | -0.789855  -0.640594  -0.513767      -0.40894 | 0.836562  0.728926  0.628311      0.541648 | 0.888034   0.86659  0.850965      0.839612 |   0.277186 |   2.051197    7.540417   17.419762           32.295824 |  40.949084 | 0.045295  0.069149  0.088993        0.10614 | 0.072114  0.149922  0.232214        0.313367 |   0.277186    3.748965    11.21674   23.476792   40.949084 |    0.0 |   0.277186 |  40.949084 |
    | 09:00 | 0.441779  0.937824  1.485806             2.057724 |   7.602519   18.464923   33.325199   52.036987 |  19.505275   24.563963   30.152312        35.984746 |  19.417785   24.378238   29.858064     35.577236 | 0.368853  0.544171  0.683326  0.803182 | 1.520785  1.456447  1.414305          1.389508 | -0.799577  -0.652781  -0.526528     -0.421393 | 0.843993  0.738819  0.639063      0.552305 | 0.890509  0.868661  0.852716      0.841044 |   0.228681 |   1.843924    6.899037   16.101185           30.079115 |  38.235295 | 0.043657  0.067253  0.086968       0.104046 | 0.067715  0.142829  0.223195        0.303025 |   0.228681    3.389583   10.300702   21.763636   38.235295 |    0.0 |   0.228681 |  38.235295 |
    | 09:30 | 0.415744  0.893106  1.425303             1.984455 |    7.10038   17.384771   31.536989   49.457117 |  19.239776   24.107925   29.535296        35.237545 |  19.157443   23.931056    29.25303     34.844545 |  0.35694  0.531042  0.669451  0.788759 | 1.526355  1.461232  1.417821          1.392015 | -0.808763  -0.664577  -0.539042       -0.4337 | 0.851002  0.748367   0.64958      0.562885 | 0.892956  0.870731  0.854473      0.842533 |   0.188431 |   1.660433    6.317991   14.890795           28.023941 |  35.710767 | 0.042093  0.065416   0.08499       0.101994 | 0.063635  0.136091  0.214519        0.292999 |   0.188431    3.068961    9.466226   20.184872   35.710767 |    0.0 |   0.188431 |  35.710767 |
    | 10:00 | 0.391588  0.850817  1.367445             1.913854 |   6.640527     16.3817   29.861201   47.021997 |  18.993431   23.676664   28.945256        34.517558 |  18.915881   23.508169   28.674449      34.13854 | 0.345493  0.518278  0.655905  0.774639 | 1.531726  1.466012  1.421394          1.394587 | -0.817435  -0.675981  -0.551302     -0.445851 |  0.85761  0.757569  0.659857      0.573314 | 0.895371  0.872795  0.856234      0.844035 |   0.155078 |   1.497824    5.791412   13.779512           26.118888 |  33.362736 |   0.0406  0.063635  0.083061       0.099983 | 0.059851  0.129696  0.206178        0.283286 |   0.155078    2.782694    8.705976   18.729711   33.362736 |    0.0 |   0.155078 |  33.362736 |
    | 10:30 | 0.369173   0.81084  1.312136              1.84585 |   6.219033   15.449899   28.290558   44.723563 |  18.764837   23.268974   28.381219        33.824051 |  18.691727   23.108396   28.121365     33.458501 | 0.334501  0.505873  0.642683  0.760816 | 1.536893  1.470777  1.425021          1.397221 | -0.825617  -0.686994  -0.563303     -0.457839 | 0.863839  0.766429   0.66989      0.583582 | 0.897752  0.874852  0.857999      0.845545 |   0.127477 |   1.353557    5.313987   12.759008           24.352931 |  31.178996 | 0.039177  0.061911   0.08118       0.098014 | 0.056344   0.12363  0.198163         0.27388 |   0.127477    2.526883    8.013226   17.388252   31.178996 |    0.0 |   0.127477 |  31.178996 |
    | 11:00 | 0.348369  0.773058  1.259284              1.78037 |   5.832341   14.583972   26.818236   42.554121 |  18.552682    22.88368   27.842225        33.156278 |  18.483691   22.730584   27.592838     32.803695 | 0.323953  0.493821  0.629777  0.747288 | 1.541854  1.475515  1.428693          1.399915 |  -0.83333  -0.697618  -0.575038     -0.469659 | 0.869711   0.77495  0.679677      0.593685 | 0.900093    0.8769  0.859766      0.847063 |   0.104666 |   1.225409    4.880916   11.821649           22.715799 |  29.148079 | 0.037819  0.060242  0.079345       0.096086 | 0.053092  0.117879  0.190465        0.264774 |   0.104666    2.298078    7.381826   16.151399   29.148079 |    0.0 |   0.104666 |  29.148079 |
    | 11:30 | 0.329057  0.737361  1.208794             1.717339 |   5.477239   13.778914   25.437833   40.506361 |  18.355731   22.519633   27.327331        32.513493 |  18.290565   22.373607   27.087943     32.173393 | 0.313836  0.482114  0.617183  0.734048 | 1.546609  1.480217  1.432406          1.402665 | -0.840601  -0.707857  -0.586505     -0.481304 | 0.875245  0.783138  0.689214      0.603618 | 0.902394  0.878936  0.861534      0.848588 |   0.085838 |   1.111434    4.487862   10.960435           21.197944 |  27.259242 | 0.036525  0.058627  0.077557       0.094198 | 0.050076  0.112432  0.183076        0.255964 |   0.085838    2.093224    6.806162   15.010787   27.259242 |    0.0 |   0.085838 |  27.259242 |
    | 12:00 | 0.311122  0.703637  1.160577             1.656686 |   5.150817   13.030088   24.143346   38.573348 |  18.172836   22.175721   26.835608        31.894953 |  18.111221   22.036374   26.605769     31.566865 | 0.304138  0.470747  0.604894  0.721093 | 1.551157  1.484872  1.436153          1.405471 | -0.847451  -0.717714  -0.597698     -0.492769 | 0.880461  0.790999    0.6985      0.613376 | 0.904651  0.880958    0.8633      0.850119 |   0.070317 |   1.009929     4.13091   10.168951           19.790499 |  25.502447 | 0.035293  0.057065  0.075814       0.092351 |  0.04728  0.107273  0.175987        0.247441 |   0.070317    1.909625    6.281112   13.958718   25.502447 |    0.0 |   0.070317 |  25.502447 |
    | 12:30 | 0.294461  0.671783  1.114542             1.598339 |   4.850449   12.333205   22.929149   36.748509 |  18.002923   21.850868   26.366143        31.299919 |  17.944609   21.717829    26.14542     30.983386 | 0.294846  0.459712  0.592905  0.708417 | 1.555503  1.489471  1.439927          1.408328 | -0.853904  -0.727198  -0.608615     -0.504048 | 0.885379  0.798539  0.707534      0.622957 | 0.906863  0.882963  0.865065      0.851655 |   0.057539 |   0.919403    3.806533    9.441317           18.485239 |  23.868329 | 0.034119  0.055555  0.074117       0.090544 | 0.044685  0.102389  0.169189        0.239201 |   0.057539    1.744892    5.802018   12.988108   23.868329 |    0.0 |   0.057539 |  23.868329 |
    | 13:00 | 0.278975  0.643004  1.070794             1.542163 |   4.573758   11.712328   21.794908    35.02379 |  17.844997   21.557379   25.919998        30.727042 |  17.789749   21.430039    25.70794     30.421634 | 0.285946  0.449474  0.581261  0.696003 | 1.559649  1.493827  1.443704          1.411237 | -0.859983   -0.73592  -0.619207      -0.51515 | 0.890017  0.806528  0.716366      0.632347 | 0.909027  0.885786  0.866889      0.853182 |   0.047032 |   0.838552    3.524099    8.775027           17.273258 |  22.349085 | 0.033001   0.05416  0.072471       0.088774 | 0.042278  0.097967  0.162702        0.231228 |   0.047032    1.596918    5.371445   12.089587   22.349085 |    0.0 |   0.047032 |  22.349085 |
    | 13:30 | 0.264573  0.612208  1.029504             1.488016 |   4.318597   11.057111    20.74196   33.391213 |   17.69813    21.24332   25.498925         30.17485 |  17.645735   21.122079   25.295043     29.880165 | 0.277424  0.438215  0.570032  0.683828 |   1.5636  1.488526  1.447449            1.4142 | -0.865709  -0.742926  -0.629404     -0.526085 | 0.894393  0.815306  0.724991       0.64155 | 0.911142  0.890012  0.868799      0.854698 |   0.038402 |   0.766234    3.255165    8.168559           16.146125 |  20.937376 | 0.031937  0.052991  0.070887       0.087038 | 0.040042  0.094504  0.156555        0.223505 |   0.038402    1.463839    4.995103   11.253731   20.937376 |    0.0 |   0.038402 |  20.937376 |
    | 14:00 | 0.251173  0.585671  0.989302             1.436281 |   4.083027   10.500117   19.733127   31.858721 |  17.561468   20.972695   25.088942        29.647246 |  17.511726    20.85671   24.893022     29.362807 | 0.269267  0.428242  0.558857   0.67199 | 1.567363  1.492968  1.451286          1.417183 | -0.871104  -0.751349  -0.639531     -0.536756 | 0.898522   0.81778  0.732901       0.65066 | 0.913206  0.888618  0.870239      0.856313 |   0.031322 |   0.701449    3.011847    7.598769           15.106546 |  19.617586 | 0.030923  0.051631  0.069314       0.085351 | 0.037964  0.090346  0.150547        0.216092 |   0.031322    1.344006    4.626016   10.495866   19.617586 |    0.0 |   0.031322 |  19.617586 |
    | 14:30 | 0.238694  0.560194  0.951204             1.386529 |   3.865292    9.971998   18.792001   30.410259 |  17.434216   20.712881   24.700414        29.139881 |  17.386945   20.601941   24.512038     28.865294 |  0.26146  0.418412  0.548031  0.660405 | 1.570942  1.497368  1.455095          1.420208 | -0.876188  -0.759556  -0.649311     -0.547232 | 0.902421  0.823876  0.741045      0.659472 | 0.915218  0.890238  0.872084      0.857847 |   0.025521 |    0.64332     2.78649    7.077612           14.140941 |  18.388881 | 0.029958  0.050298  0.067793       0.083701 | 0.036033  0.086353  0.144834         0.20893 |   0.025521    1.235964    4.286766    9.797918   18.388881 |    0.0 |   0.025521 |  18.388881 |
    | 15:00 | 0.227067  0.536111  0.914871             1.338718 |   3.663809    9.478738   17.908008   29.041603 |  17.315642    20.46728    24.32989          28.6523 |  17.270674   20.361109   24.148709     28.387181 | 0.253989  0.408874  0.537478  0.649072 | 1.574346  1.501677  1.458897          1.423268 |  -0.88098   -0.76743  -0.658811     -0.557507 | 0.906103  0.830045   0.74881      0.668102 | 0.917179  0.892139  0.873819      0.859387 |   0.020773 |   0.591082    2.580853    6.597557           13.244237 |   17.24703 | 0.029039   0.04901  0.066314       0.082088 | 0.034236  0.082576  0.139368        0.202016 |   0.020773    1.138424    3.976635    9.151725    17.24703 |    0.0 |   0.020773 |   17.24703 |
    | 15:30 | 0.216225   0.51337  0.880186             1.292799 |   3.477142    9.018304   17.076418   27.748632 |  17.205072   20.235372   23.976167        28.184015 |  17.162251   20.133705   23.801856      27.92799 |  0.24684  0.399632   0.52718  0.637993 |  1.57758  1.505886  1.462692          1.426357 | -0.885499  -0.774973  -0.668044     -0.567572 | 0.909584  0.835967  0.756314      0.676545 | 0.919086  0.894037  0.875526      0.860931 |   0.016892 |   0.544059    2.393277     6.15464           12.411631 |  16.185591 | 0.028164  0.047768  0.064875       0.080512 | 0.032563  0.079009  0.134134        0.195346 |   0.016892    1.050247    3.693031    8.553048   16.185591 |    0.0 |   0.016892 |  16.185591 |
    | 16:00 | 0.206107  0.491894  0.847077             1.248702 |   3.303997    8.588214   16.293844   26.526812 |  17.101882   20.016357   23.638521        27.734311 |  17.061065   19.918943   23.470767     27.487019 | 0.239999  0.390679  0.517132   0.62716 | 1.580653  1.509993  1.466475          1.429472 | -0.889762  -0.782197   -0.67701     -0.577427 | 0.912874  0.841626  0.763582      0.684793 | 0.920941  0.895908  0.877222      0.862472 |   0.013722 |   0.501663    2.222023      5.7458           11.638248 |   15.19821 |  0.02733  0.046571  0.063475       0.078972 | 0.031004  0.075639  0.129124        0.188915 |   0.013722    0.970427    3.433466    7.998365    15.19821 |    0.0 |   0.013722 |   15.19821 |
    | 16:30 | 0.196656  0.471606  0.815479             1.206355 |   3.143201    8.186147   15.557212     25.3718 |  17.005502   19.809454   23.316285         27.30246 |  16.966556   19.716058   23.154789     27.063555 | 0.233452  0.382008  0.507331  0.616568 | 1.583571  1.513995  1.470238          1.432608 | -0.893785  -0.789113  -0.685712     -0.587072 | 0.915988   0.84703  0.770619      0.692844 | 0.922743  0.897748   0.87891      0.864008 |   0.011136 |   0.463374    2.065512     5.36828           10.919545 |  14.279104 | 0.026536  0.045417  0.062112       0.077469 |  0.02955  0.072455   0.12433        0.182714 |   0.011136    0.898071    3.195688    7.484414   14.279104 |    0.0 |   0.011136 |  14.279104 |
    | 17:00 | 0.187821  0.452433  0.785328             1.165691 |   2.993694    7.809966   14.863626   24.279553 |  16.915403   19.613926   23.008808        26.887766 |  16.878207   19.524326   22.853282     26.656914 | 0.227186   0.37361  0.497771   0.60621 | 1.586343  1.517892  1.473977          1.435763 | -0.897584  -0.795733  -0.694152     -0.596506 | 0.918935  0.852191  0.777428      0.700699 | 0.924492  0.899555  0.880586      0.865538 |   0.009028 |   0.428738    1.922325    5.019542           10.251353 |  13.423044 | 0.025778  0.044305  0.060787          0.076 | 0.028193  0.069447  0.119744        0.176736 |   0.009028    0.832389    2.977666    7.008138   13.423044 |    0.0 |   0.009028 |  13.423044 |
    | 17:30 | 0.179554  0.434307  0.756562             1.126645 |   2.854514    7.457713   14.210355   23.246308 |  16.831102   19.429077   22.715446        26.489565 |  16.795543   19.343067   22.565618     26.266445 | 0.221188  0.365477  0.488451  0.596082 | 1.588974  1.521682  1.477686          1.438933 | -0.901173  -0.802069  -0.702332     -0.605729 | 0.921727  0.857119  0.784013      0.708361 | 0.926189   0.90133  0.882251      0.867063 |   0.007313 |   0.397354     1.79119    4.697248            9.629854 |  12.625287 | 0.025056  0.043232  0.059499       0.074566 | 0.026926  0.066603  0.115359        0.170975 |   0.007313    0.772681    2.777573    6.566683   12.625287 |    0.0 |   0.007313 |  12.625287 |
    | 18:00 | 0.171813  0.417164  0.729118             1.089153 |   2.724792     7.12759   13.594828   22.268556 |  16.752155   19.254255    22.43557        26.107221 |  16.718129   19.171641   22.291177     25.891526 | 0.215446  0.357601  0.479364  0.586178 | 1.591473  1.525365  1.481361          1.442113 | -0.904567  -0.808134  -0.710256     -0.614742 | 0.924373  0.861825  0.790376       0.71583 | 0.927835  0.903072  0.883901      0.868581 |   0.005918 |   0.368869    1.670965    4.399246             9.05154 |  11.881509 | 0.024368  0.042199  0.058247       0.073165 |  0.02574  0.063914  0.111168        0.165423 |   0.005918    0.718327    2.593759     6.15739   11.881509 |    0.0 |   0.005918 |  11.881509 |
    | 18:30 | 0.164556  0.400944  0.702936             1.053156 |   2.603741    6.817948   13.014628    21.34302 |  16.678153   19.088847   22.168565        25.740123 |  16.645564   19.009444   22.029357     25.531557 | 0.209947  0.349974  0.470507  0.576494 | 1.593847  1.528941  1.484997          1.445301 | -0.907778  -0.813939  -0.717928     -0.623545 | 0.926882  0.866319  0.796521      0.723107 |  0.92943   0.90478  0.885536      0.870093 |   0.004785 |   0.342973    1.560625    4.123559            8.513187 |  11.187755 |  0.02371  0.041202  0.057031       0.071797 |  0.02463  0.061371  0.107162        0.160074 |   0.004785    0.668777    2.424738    5.777789   11.187755 |    0.0 |   0.004785 |  11.187755 |
    | 19:00 | 0.157748  0.385591  0.677957             1.018596 |   2.490645    6.527272   12.467489   20.466628 |  16.608722   18.932275   21.913835        25.387681 |  16.577482   18.855912   21.779573      25.18596 | 0.204678  0.342588  0.461874  0.567024 | 1.596102  1.532411  1.488589          1.448492 | -0.910817  -0.819495  -0.725353      -0.63214 | 0.929263  0.870612  0.802455      0.730195 | 0.930974  0.906454  0.887156      0.871597 |   0.003865 |   0.319392    1.459248     3.86837            8.011824 |  10.540394 | 0.023083  0.040241   0.05585       0.070462 |  0.02359  0.058965  0.103333        0.154923 |   0.003865     0.62354    2.269169    5.425591   10.540394 |    0.0 |   0.003865 |  10.540394 |
    | 19:30 | 0.151354  0.371052  0.654126             0.985418 |   2.384856     6.25417   11.951288   19.636503 |  16.543518   18.783998     21.6708        25.049327 |  16.513544   18.710516   21.541257     24.854176 |  0.19963  0.335437  0.453462  0.557764 | 1.598244  1.535775  1.492136          1.451682 | -0.913696  -0.824814  -0.732536     -0.640526 | 0.931524  0.874712  0.808181      0.737095 |  0.93247  0.908093  0.888757      0.873092 |   0.003119 |   0.297883    1.366006    3.632014            7.544713 |   9.936082 | 0.022483  0.039315  0.054702       0.069159 | 0.022615  0.056688  0.099675        0.149962 |   0.003119    0.582183    2.125843    5.098676    9.936082 |    0.0 |   0.003119 |   9.936082 |
    | 20:00 | 0.145344  0.357276  0.630236             0.954591 |   2.285787    5.997364   11.439521   18.875076 |  16.482225   18.643511   21.427168        24.734952 |  16.453442   18.572756   21.302357     24.545906 | 0.194791  0.328511  0.444846  0.549003 |  1.60028  1.539036  1.485571          1.454757 | -0.916426  -0.829905   -0.73727     -0.648437 | 0.933672   0.87863  0.817403      0.744144 | 0.933919  0.909698  0.894387      0.874973 |   0.002515 |   0.278233    1.280153      3.4255            7.123163 |   9.357298 |  0.02191  0.038421    0.0539       0.067929 | 0.021699  0.054531  0.097331        0.145341 |   0.002515    0.544319    1.993666    4.827992    9.357298 |    0.0 |   0.002515 |   9.357298 |
    | 20:30 | 0.139689  0.344217  0.608934             0.923695 |   2.192902    5.755679    10.98802   18.121489 |  16.424555   18.510337   21.209936        24.419878 |  16.396891   18.442169   21.089343     24.236951 | 0.190151  0.321805  0.436999  0.540062 | 1.602216  1.542194  1.489015          1.457975 | -0.919015   -0.83478  -0.743946     -0.656494 | 0.935714  0.882374  0.817494      0.750028 |  0.93532  0.911269  0.891728      0.875893 |   0.002026 |   0.260253    1.201019    3.224785            6.712554 |   8.834899 | 0.021362   0.03756  0.052827       0.066676 | 0.020839  0.052489  0.093998        0.140694 |   0.002026    0.509603    1.871652    4.542655    8.834899 |    0.0 |   0.002026 |   8.834899 |
    | 21:00 | 0.134363  0.331832  0.588085             0.894419 |   2.105716    5.528035   10.550493   17.416216 |  16.370242   18.384032   20.997313        24.121322 |  16.343633   18.318316   20.880849     23.944192 | 0.185699   0.31531   0.42916  0.531433 | 1.604057  1.545251  1.492488          1.461133 | -0.921474  -0.839449  -0.750561     -0.664241 | 0.937657  0.885952  0.822272      0.756578 | 0.936676  0.912805  0.892887      0.877532 |   0.001631 |   0.243775       1.128     3.03376            6.334508 |   8.350207 | 0.020838  0.036729  0.051758       0.065468 |  0.02003  0.050553  0.090733        0.136281 |   0.001631     0.47773    1.758908    4.274918    8.350207 |    0.0 |   0.001631 |   8.350207 |
    | 21:30 | 0.129343  0.320079   0.56816              0.86634 |   2.023787    5.313437   10.136432   16.747822 |  16.319041   18.264178   20.794119        23.834968 |  16.293426   18.200789   20.681601     23.663399 | 0.181426  0.309019  0.421513  0.523005 | 1.605809   1.54821  1.495903          1.464272 | -0.923809  -0.843921  -0.756959      -0.67178 | 0.939507  0.889373  0.827306      0.762742 | 0.937988  0.914307  0.894409      0.878997 |   0.001312 |    0.22865     1.06055     2.85623            5.981942 |   7.895747 | 0.020337  0.035928   0.05072       0.064292 | 0.019267  0.048718   0.08761        0.132038 |   0.001312    0.448426    1.654627    4.026194    7.895747 |    0.0 |   0.001312 |   7.895747 |
    | 22:00 | 0.124605  0.308921  0.549154             0.839345 |   1.946711     5.11097    9.745161   16.112684 |  16.270729   18.150385   20.600294        23.559677 |  16.246052   18.089207    20.49154     23.393454 | 0.177324  0.302926   0.41407  0.514754 | 1.607477  1.551072  1.499253          1.467398 |  -0.92603  -0.848206  -0.763135     -0.679131 | 0.941268  0.892646   0.83219      0.768699 | 0.939258  0.915775  0.895941       0.88042 |   0.001054 |   0.214745    0.998178    2.691461            5.652222 |   7.469089 | 0.019856  0.035154  0.049713       0.063143 | 0.018549  0.046976   0.08463        0.127949 |   0.001054    0.421446    1.558079     3.79517    7.469089 |    0.0 |   0.001054 |   7.469089 |
    | 22:30 | 0.120131  0.298321  0.531022             0.813392 |   1.874123    4.919791    9.375245   15.508918 |  16.225101   18.042289    20.41538        23.295006 |   16.20131   17.983209   20.310217     23.133922 | 0.173383  0.297024  0.406826  0.506676 | 1.609066   1.55384  1.502536          1.470509 | -0.928144  -0.852312  -0.769093     -0.686297 | 0.942947  0.895776  0.836893      0.774492 | 0.940486   0.91721  0.897451      0.881832 |   0.000846 |   0.201944    0.940441    2.538435            5.343723 |   7.068569 | 0.019396  0.034408  0.048737        0.06202 | 0.017871  0.045324  0.081786         0.12401 |   0.000846    0.396572    1.468603    3.580431    7.068569 |    0.0 |   0.000846 |   7.068569 |
    | 23:00 |  0.10702  0.290519  0.512793             0.788798 |   1.662563    4.779799     9.00668   14.942988 |  16.091392   17.962728   20.229484        23.044196 |  16.070198   17.905194   20.127931     22.887983 | 0.161435  0.292605  0.399394  0.498882 | 1.411566  1.556153  1.505928          1.473552 | -0.920575   -0.85539  -0.775152     -0.693178 | 1.020769  0.902575  0.840667       0.78028 | 1.017662  0.922498  0.898133      0.883411 |   0.000679 |   0.190141    0.898752    2.388703            5.059061 |   6.688011 | 0.020586  0.033846  0.047739        0.06094 | 0.020769  0.044095  0.078925         0.12027 |   0.000679    0.403815    1.367299    3.390443    6.688011 |    0.0 |   0.000679 |   6.688011 |
    | 23:30 |  0.10595  0.280079  0.496451             0.764784 |   1.645371    4.593398    8.679089   14.396222 |  16.080478   17.856252    20.06283        22.799292 |  16.059496   17.800785   19.964513     22.647835 | 0.160432  0.286589  0.392598  0.491135 | 1.407215  1.558742  1.509043          1.476627 |  -0.92083  -0.859467  -0.780642     -0.699986 | 0.957221  0.899787  0.846127      0.785508 | 0.957572  0.918422  0.900602      0.884595 |   0.000544 |   0.187583    0.844538    2.257982            4.788273 |   6.335284 | 0.020521  0.033095  0.046829       0.059869 | 0.020695  0.042482  0.076361        0.116611 |   0.000544    0.386831    1.286637    3.205945    6.335284 |    0.0 |   0.000544 |   6.335284 |
    | 00:00 |  0.10489  0.269761  0.481278             0.741422 |   1.628361     4.41027    8.377305    13.86986 |  16.069673   17.751033   19.908088        22.561049 |  16.048901    17.69761   19.812776     22.414219 | 0.159434  0.280522  0.386169  0.483462 | 1.486019  1.561531  1.512017          1.479716 | -0.927356  -0.863558  -0.785793     -0.706695 |  0.90674  0.901878  0.850661      0.790697 | 0.907056  0.918972  0.902433      0.885826 |   0.000436 |   0.174706    0.792285    2.139562            4.531648 |   6.010872 | 0.019312  0.032336  0.045972       0.058811 | 0.018379   0.04088  0.073979        0.113045 |   0.000436    0.350967    1.228177     3.02255    6.010872 |    0.0 |   0.000436 |   6.010872 |
    | 00:30 | 0.103841   0.25989   0.46688               0.7189 |   1.611533    4.236068    8.093088    13.36758 |  16.058977   17.650371   19.761262        22.331368 |  16.038412   17.598902   19.668801     22.188997 | 0.158442  0.274599  0.379957   0.47593 | 1.648842  1.564365  1.514903          1.482786 | -0.937834  -0.867521  -0.790728     -0.713243 | 0.868612  0.904595  0.854658       0.79589 | 0.866992  0.920185  0.903909       0.88717 |   0.000349 |   0.154857    0.743574     2.02985            4.290594 |   5.708487 | 0.017297  0.031596  0.045146       0.057775 | 0.014784  0.039342  0.071718        0.109602 |   0.000349    0.304337    1.183614    2.845865    5.708487 |    0.0 |   0.000349 |   5.708487 |
    | 01:00 | 0.102803  0.254901  0.451397             0.697975 |   1.594884    4.148393    7.789754   12.905461 |  16.048387   17.599494   19.603366        22.117972 |  16.028028   17.549013   19.513972     21.979746 | 0.157455   0.27156  0.373151  0.468807 | 1.563529  1.565848  1.518046          1.485714 | -0.933472  -0.869541  -0.796081     -0.719399 | 1.015683  0.916394  0.856741      0.801361 | 1.010591  0.930702  0.903513      0.888966 |   0.000279 |   0.160613    0.719442    1.914798            4.072229 |   5.413331 | 0.018127  0.031217  0.044246       0.056798 | 0.016282  0.038565   0.06929          0.1064 |   0.000279    0.307654    1.112967    2.699316    5.413331 |    0.0 |   0.000279 |   5.413331 |
    | 01:30 | 0.101775  0.246375  0.437691             0.677574 |   1.578412    3.999122    7.523227   12.459146 |  16.037903   17.512538   19.463587        21.909928 |  16.017748   17.463746   19.376907     21.775742 | 0.156474   0.26629   0.36701  0.461741 | 1.562185  1.568167  1.520909          1.488658 | -0.933898  -0.872986   -0.80087     -0.725471 | 0.963524  0.910224  0.861813      0.805986 | 0.959975    0.9237  0.906088      0.890116 |   0.000223 |   0.158099     0.67909    1.815428            3.864483 |    5.13406 | 0.018029  0.030566  0.043436       0.055831 | 0.016152  0.037248  0.067139        0.103273 |   0.000223      0.2954    1.050201    2.563913     5.13406 |    0.0 |   0.000223 |    5.13406 |
    | 02:00 | 0.100757  0.237788  0.425099             0.657679 |   1.562115     3.84953    7.280022   12.027886 |  16.027524   17.424968   19.335172        21.707033 |   16.00757   17.377877   19.250986     21.576787 | 0.155498  0.260884  0.361269  0.454728 | 1.638346  1.570612  1.523593          1.491615 | -0.938686  -0.876492  -0.805308     -0.731463 | 0.920244  0.911635  0.865945      0.810525 | 0.916794  0.923901  0.907993       0.89126 |   0.000178 |   0.148263     0.63942    1.726214            3.666773 |   4.876172 | 0.017084  0.029899  0.042681       0.054874 | 0.014542   0.03592  0.065163        0.100219 |   0.000178    0.270859    1.002172    2.430373    4.876172 |    0.0 |   0.000178 |   4.876172 |
    | 02:30 | 0.099749  0.229551  0.413126             0.639863 |   1.545992     3.70674    7.050252   11.645069 |  16.017249   17.340974   19.213073        21.525348 |  15.997494   17.295514   19.131258      21.39863 | 0.154528  0.255602  0.355718  0.448341 |  1.79208   1.57308    1.5262           1.49436 | -0.946363  -0.879893  -0.809565     -0.736893 | 0.886393  0.913933  0.869364      0.815921 | 0.881899  0.925005  0.909354      0.893431 |   0.000142 |   0.133308    0.602289    1.643234            3.493777 |   4.644212 | 0.015521  0.029247  0.041953       0.054004 | 0.012036  0.034643  0.063285        0.097478 |   0.000142    0.238894    0.964361    2.300608    4.644212 |    0.0 |   0.000142 |   4.644212 |
    | 03:00 | 0.080797  0.225736  0.400007             0.619744 |   1.244597     3.64082    6.800131    11.21657 |  15.823972   17.302063   19.079286        21.320172 |  15.807971   17.257358   19.000069     21.197438 | 0.135411  0.253121  0.349528  0.441001 | 1.455681  1.574437  1.529107          1.487264 | -0.938724    -0.8815  -0.814269      -0.74055 | 1.049385  0.925793  0.870675      0.822719 | 1.042097  0.935838   0.90863      0.897504 |   0.000113 |   0.115776    0.585331    1.554393            3.325917 |   4.436324 | 0.016744  0.028938  0.041145       0.053373 | 0.014862  0.034044  0.061227        0.095689 |   0.000113    0.248993    0.904164    2.193814    4.436324 |    0.0 |   0.000113 |   4.436324 |
    | 03:30 | 0.079989  0.218795  0.388394             0.602744 |   1.231828    3.521276    6.580164   10.857669 |  15.815732   17.231277    18.96086         21.14681 |  15.799891   17.187947   18.883943     21.027443 | 0.134556  0.248551  0.343949  0.434689 | 1.408157  1.576368   1.53173          1.490036 | -0.936245  -0.884377  -0.818469     -0.745901 | 0.996346  0.919327  0.875254      0.822512 | 0.992603  0.928762  0.911129      0.895198 |    0.00009 |   0.117707     0.55521    1.477573             3.16751 |   4.213555 |   0.0172  0.028381  0.040419       0.052511 | 0.015727  0.032978  0.059407        0.093029 |    0.00009     0.24718    0.850059    2.094474    4.213555 |    0.0 |    0.00009 |   4.213555 |
    | 04:00 | 0.079189  0.211733   0.37774             0.585701 |   1.219193    3.400156    6.379531   10.500743 |  15.807574   17.159265   18.852204        20.973002 |  15.791892   17.117334   18.777396      20.85701 | 0.133705  0.243824  0.338746  0.428253 | 1.428568    1.5784  1.534171          1.492894 | -0.938024   -0.88733  -0.822352     -0.751323 | 0.950528  0.920277  0.878951      0.826159 | 0.947733  0.928776  0.912977      0.895934 |   0.000072 |   0.114109     0.52524    1.408605            3.012255 |   4.007298 | 0.016847  0.027806  0.039744       0.051635 | 0.015132  0.031892  0.057739        0.090359 |   0.000072    0.234279    0.809106    1.993997    4.007298 |    0.0 |   0.000072 |   4.007298 |
    | 04:30 | 0.078397  0.204929  0.367607             0.569215 |    1.20669     3.28392    6.189785   10.158253 |  15.799499   17.089877   18.748874        20.804877 |  15.783973   17.049293   18.676073      20.69215 |  0.13286  0.239192  0.333719  0.421922 | 1.513985   1.58051  1.536542          1.495727 | -0.943376   -0.89021  -0.826075     -0.756621 | 0.912631  0.922129   0.88194      0.830241 | 0.909507  0.929697  0.914261      0.897132 |   0.000057 |   0.105893    0.496983     1.34435             2.86549 |   3.817939 | 0.015796  0.027241  0.039094       0.050775 | 0.013341  0.030843  0.056151        0.087775 |   0.000057     0.21309    0.778564    1.893464    3.817939 |    0.0 |   0.000057 |   3.817939 |
    | 05:00 | 0.077613  0.202216  0.356322             0.554033 |   1.194319    3.237689    5.979666    9.845264 |  15.791504   17.062202    18.63379        20.650053 |  15.776133   17.022155   18.563225     20.540333 |  0.13202  0.237322  0.328026  0.415995 | 1.369329  1.581639  1.539232          1.498386 | -0.935165  -0.891392  -0.830254     -0.761543 | 1.065639  0.934258  0.882628      0.834785 | 1.060421  0.941012  0.913261      0.898931 |   0.000046 |   0.115147     0.48581    1.274329            2.733331 |   3.634239 | 0.017354  0.027009   0.03836       0.049973 |  0.01615  0.030413  0.054384        0.085395 |   0.000046    0.225983    0.730279    1.810092    3.634239 |    0.0 |   0.000046 |   3.634239 |
    | 05:30 | 0.076837  0.192603  0.347985              0.53853 |   1.182078    3.074521    5.825243    9.528017 |  15.783589   16.964171   18.548765        20.491947 |  15.768372   16.926028    18.47985     20.385297 | 0.131184  0.230596  0.323754  0.409843 | 1.586712  1.584626  1.541247          1.501178 | -0.947814  -0.895503  -0.833364     -0.766621 | 0.836166  0.916023  0.889082      0.837859 | 0.834154  0.922282   0.91804      0.899396 |   0.000036 |    0.09773    0.447406    1.223651            2.601284 |   3.468251 | 0.014882  0.026194  0.037811       0.049143 |  0.01191  0.028935   0.05308        0.082962 |   0.000036    0.188509     0.71172    1.717893    3.468251 |    0.0 |   0.000036 |   3.468251 |
    | 06:00 | 0.076069  0.189886   0.33816             0.524191 |   1.169965    3.028579    5.644168    9.236745 |  15.775753   16.936468   18.448573         20.34572 |  15.760688   16.898863   18.381604     20.241909 | 0.130353  0.228664  0.318645  0.404059 | 1.555746  1.585218  1.543642          1.503804 | -0.946756  -0.896648  -0.837052     -0.771357 | 0.989397  0.937453  0.888846      0.842175 | 0.983535  0.942607  0.916411      0.901134 |   0.000029 |   0.098029    0.436866    1.165094            2.481832 |    3.30973 | 0.015082  0.025965  0.037156       0.048364 | 0.012268  0.028527  0.051545        0.080713 |   0.000029    0.185413    0.681339    1.636588     3.30973 |    0.0 |   0.000029 |    3.30973 |
    | 06:30 | 0.075308  0.184494  0.329163             0.510391 |   1.157979    2.937594    5.479191    8.958366 |  15.767995   16.881473    18.35682         20.20499 |  15.753081   16.844936   18.291633     20.103913 | 0.129527  0.224788  0.313893  0.398403 | 1.585924  1.586922  1.545881          1.506385 | -0.948574  -0.898981  -0.840455     -0.775956 | 0.952365  0.930005  0.892165      0.845774 |  0.94702  0.934784  0.918213      0.902323 |   0.000023 |   0.094576    0.416112    1.112555            2.369276 |   3.157765 | 0.014701  0.025497  0.036549       0.047606 |  0.01169  0.027699  0.050139        0.078548 |   0.000023    0.175596    0.651483    1.563061    3.157765 |    0.0 |   0.000023 |   3.157765 |
    | 07:00 |    0.061  0.182592  0.319341             0.497585 |   0.933599    2.905584    5.300005    8.701727 |  15.622076   16.862083    18.25665         20.07439 |  15.609996   16.825923   18.193408     19.975849 | 0.113291  0.223408   0.30862  0.393074 | 1.201139  1.588173  1.548408          1.508825 | -0.935061  -0.899866  -0.844205     -0.780259 | 1.120711  0.941279  0.892833      0.849671 | 1.113983  0.945435  0.917434      0.903932 |   0.000018 |   0.088057    0.408729    1.056367            2.266942 |   3.008384 | 0.016978  0.025321  0.035877       0.046893 | 0.016582  0.027385  0.048602        0.076539 |   0.000018     0.19562    0.605188    1.504768    3.008384 |    0.0 |   0.000018 |   3.008384 |
    | 07:30 |  0.07381  0.174088  0.312443              0.48429 |   1.134382    2.762861    5.174751    8.437034 |  15.752712   16.775361   18.186307        19.938808 |  15.738095   16.740885   18.124432       19.8429 |  0.12789  0.217147  0.304863  0.387454 | 1.521467  1.590447  1.550131          1.511424 | -0.946311  -0.903541  -0.846844     -0.784767 | 0.867613  0.921865  0.899572      0.852082 | 0.866926   0.92568  0.922761      0.904147 |   0.000014 |   0.095353    0.377219    1.017713            2.162838 |   2.875603 |  0.01513  0.024576    0.0354       0.046143 | 0.012455  0.026098  0.047528        0.074451 |   0.000014     0.16959    0.587314    1.435587    2.875603 |    0.0 |   0.000014 |   2.875603 |
    | 08:00 | 0.059786  0.172324  0.303889             0.472074 |   0.914657    2.733337    5.020077    8.195376 |  15.609697   16.757366   18.099071        19.814228 |  15.597857   16.723239    18.03889     19.720739 | 0.111846  0.215829  0.300139  0.382211 | 1.254852  1.590954  1.552351          1.513847 | -0.939978  -0.904311  -0.850146     -0.788941 | 1.031288  0.944635  0.898407      0.856015 | 1.025785  0.947515  0.920465      0.905891 |   0.000011 |   0.081524    0.370805    0.970604            2.069139 |   2.750247 | 0.016044  0.024419  0.034802       0.045446 | 0.014896   0.02583  0.046193        0.072534 |   0.000011    0.173967    0.559369     1.37351    2.750247 |    0.0 |   0.000011 |   2.750247 |
    | 08:30 | 0.059188  0.167925  0.296111             0.460297 |   0.905334    2.659867    4.880069    7.963823 |    15.6036   16.712504   18.019749        19.694128 |  15.591878   16.679249   17.961108     19.602971 |  0.11113  0.212514  0.295779   0.37708 | 1.230848   1.59233  1.554397          1.516226 | -0.938769  -0.906249  -0.853169     -0.792997 | 0.983089  0.936351  0.901535      0.859154 | 0.980835  0.938996  0.922311      0.906996 |   0.000009 |    0.08174    0.354988    0.928605            1.980572 |   2.630994 | 0.016252  0.024023  0.034251       0.044765 | 0.015331  0.025158  0.044981        0.070686 |   0.000009    0.170636    0.533501    1.315927    2.630994 |    0.0 |   0.000009 |   2.630994 |
    | 09:00 | 0.058596   0.16332  0.289051             0.448726 |   0.896107    2.583164    4.753516    7.737667 |  15.597564   16.665542   17.947753        19.576126 |   15.58596   16.633198   17.890509     19.487261 | 0.110418  0.208998  0.291766  0.371963 | 1.260853  1.593849  1.556273          1.518611 | -0.941315  -0.908294  -0.855929     -0.797013 | 0.942356  0.936402  0.904464      0.861985 | 0.940622   0.93853  0.924083       0.90786 |   0.000007 |   0.078476    0.338724    0.891178            1.895234 |   2.520233 | 0.015763  0.023603  0.033746       0.044088 | 0.014466  0.024454  0.043881        0.068869 |   0.000007    0.160505    0.514705    1.258008    2.520233 |    0.0 |   0.000007 |   2.520233 |
    | 09:30 |  0.05801  0.158852  0.282353             0.437487 |   0.886976    2.508951    4.633917    7.519276 |  15.591588    16.61998    17.87945        19.461508 |    15.5801   16.588521   17.823533     19.374868 |  0.10971   0.20554   0.28791  0.366918 | 1.343356  1.595428   1.55808          1.520969 | -0.946815  -0.910297  -0.858564     -0.800945 | 0.908947  0.937619  0.906652      0.864905 | 0.906593  0.939215  0.925219       0.90886 |   0.000006 |   0.072438     0.32323    0.856279            1.813947 |   2.417413 |   0.0147   0.02319  0.033261       0.043423 | 0.012619  0.023768  0.042839        0.067105 |   0.000006    0.145514     0.50145    1.200064    2.417413 |    0.0 |   0.000006 |   2.417413 |
    | 10:00 |  0.05743  0.157649  0.274613             0.427213 |    0.87794       2.489    4.496256    7.320754 |  15.585672    16.60771   17.800514        19.356737 |  15.574299   16.576489    17.74613     19.272132 | 0.109007  0.204601  0.283391   0.36224 | 1.217351  1.595951  1.560203          1.523144 | -0.939237  -0.910847  -0.861628     -0.804561 | 1.065336  0.950196  0.906193       0.86844 |  1.06091  0.951296  0.923719      0.910571 |   0.000004 |   0.078615     0.31909    0.816686            1.741051 |   2.314997 | 0.016118  0.023076  0.032695       0.042808 | 0.015215  0.023581  0.041634        0.065494 |   0.000004    0.154379    0.474678    1.153936    2.314997 |    0.0 |   0.000004 |   2.314997 |
    | 10:30 | 0.056856  0.156072  0.269744             0.416305 |   0.868997    2.462879    4.409968    7.111119 |  15.579816   16.591633   17.750859        19.245492 |  15.568556   16.560724   17.697439     19.163047 | 0.108308  0.203365  0.280512  0.357201 | 1.405515  1.614867  1.561557          1.525523 | -0.950891  -0.912998  -0.863567     -0.808435 | 0.840197  0.939111  0.913282      0.870053 | 0.838681  0.940059  0.929757      0.910445 |   0.000004 |   0.066964    0.310158    0.792189            1.665068 |   2.224056 | 0.013871  0.022668  0.032334       0.042147 | 0.011302  0.022811  0.040876        0.063782 |   0.000004    0.129475    0.472994    1.097932    2.224056 |    0.0 |   0.000004 |   2.224056 |
    | 11:00 | 0.056287  0.149624  0.262331             0.406751 |   0.860147    2.356294    4.279056    6.928495 |  15.574018    16.52587   17.675263        19.148062 |   15.56287   16.496239   17.623312     19.067509 | 0.107613  0.198246  0.276075  0.352724 | 1.369865  1.598495  1.563603          1.527606 | -0.949467  -0.914454   -0.86653     -0.811846 | 0.996237  0.942556  0.909467      0.874022 | 0.991132  0.942826  0.925137      0.912677 |   0.000003 |   0.067571    0.292229    0.755525             1.59979 |   2.132624 |  0.01414  0.022324  0.031781       0.041562 | 0.011781  0.022361  0.039725        0.062285 |   0.000003    0.128328    0.450639    1.055418    2.132624 |    0.0 |   0.000003 |   2.132624 |
    | 11:30 | 0.055724  0.145706  0.256504             0.397242 |   0.851389    2.291746    4.176527    6.747627 |  15.568277   16.485919   17.615835        19.051085 |  15.557242   16.457063   17.565038     18.972415 | 0.106923  0.195085   0.27254  0.348208 | 1.383714  1.599795  1.565239          1.529726 | -0.950569  -0.916233  -0.868875     -0.815266 |  0.96159  0.942038  0.914186      0.876227 | 0.957063  0.942108  0.928826      0.913311 |   0.000002 |   0.065789    0.279465    0.727217            1.535946 |   2.044125 | 0.013909   0.02195  0.031342       0.040973 | 0.011433  0.021764  0.038821        0.060794 |   0.000002    0.122819     0.43291    1.016123    2.044125 |    0.0 |   0.000002 |   2.044125 |
    | 12:00 | 0.045137  0.144879  0.249704             0.388458 |   0.687235    2.278138    4.057322    6.581364 |  15.460305   16.477484   17.546492        18.961508 |  15.451366   16.448792   17.497041     18.884578 | 0.093372  0.194413  0.268359   0.34398 | 1.035547  1.600728  1.567243          1.531708 | -0.936445  -0.916659  -0.871636     -0.818445 | 1.132588  0.954646  0.913403      0.879241 | 1.126993  0.954292  0.927169        0.9148 |   0.000002 |   0.061966    0.276686    0.694735            1.477995 |   1.956525 |  0.01623  0.021861  0.030821       0.040423 |  0.01659  0.021621  0.037763        0.059418 |   0.000002    0.138417     0.40349    0.985842    1.956525 |    0.0 |   0.000002 |   1.956525 |
    | 12:30 | 0.054615   0.14343   0.24604             0.378829 |   0.834143    2.254318    3.993285    6.399983 |  15.556969   16.462709   17.509129        18.863308 |  15.546153   16.434304   17.460404     18.788285 | 0.105554   0.19323  0.266082  0.339282 | 1.295776  1.606809  1.568239          1.533943 | -0.946576  -0.917733  -0.873119     -0.821958 | 0.878988  0.949656   0.92158      0.880184 | 0.879622  0.949199  0.934463      0.914269 |   0.000001 |   0.067949    0.271098    0.677537            1.415565 |   1.882662 | 0.014663  0.021646   0.03054       0.039813 | 0.012783  0.021252  0.037197        0.057907 |   0.000001    0.121755    0.402703    0.941474    1.882662 |    0.0 |   0.000001 |   1.882662 |
    | 13:00 | 0.044238  0.138264  0.238918             0.370834 |   0.673361    2.169542    3.869171    6.250096 |  15.451145    16.41002    17.43649        18.781778 |  15.442384   16.382638   17.389175     18.708338 | 0.092168  0.188967  0.261601  0.335328 | 1.052599  1.602626  1.570336          1.535786 | -0.938917  -0.919665  -0.876035     -0.824887 | 1.044042  0.941477  0.914861      0.884432 | 1.040124  0.940734  0.927123      0.917028 |   0.000001 |   0.058961    0.255813    0.644563            1.364666 |   1.807639 | 0.015761  0.021224  0.029986       0.039302 | 0.015742  0.020624  0.036092        0.056657 |   0.000001     0.12664    0.376999    0.911015    1.807639 |    0.0 |   0.000001 |   1.807639 |
    | 13:30 | 0.043796  0.134767  0.233947             0.362654 |    0.66653    2.112324    3.782867    6.097391 |  15.446633   16.374364   17.385804        18.698355 |   15.43796   16.347675   17.339474     18.626535 | 0.091572  0.186039  0.258433  0.331233 | 1.018318  1.603619  1.571758          1.537717 | -0.936407  -0.921265  -0.878074     -0.827905 | 0.994341  0.945731  0.921167      0.886008 | 0.994303  0.944598  0.932485      0.917304 |   0.000001 |   0.059937    0.245055    0.621991            1.313411 |    1.73611 | 0.016186  0.020882  0.029596       0.038773 | 0.016654  0.020098  0.035324        0.055376 |   0.000001    0.125919    0.359875     0.88079     1.73611 |    0.0 |   0.000001 |    1.73611 |
    | 14:00 | 0.043358  0.131475  0.229241             0.354501 |    0.65977    2.058552    3.701372    5.945861 |  15.442167   16.340786   17.337809         18.61521 |   15.43358   16.314749    17.29241     18.545005 | 0.090979   0.18325  0.255401  0.327098 | 1.029476  1.604768  1.573127          1.539675 | -0.937906  -0.922795  -0.880016     -0.830933 | 0.951803  0.947045  0.922852      0.887898 | 0.952693  0.945562  0.933485      0.917879 |   0.000001 |   0.058307    0.235068    0.600926            1.263174 |   1.671512 | 0.015907  0.020554  0.029223        0.03824 | 0.016134  0.019598  0.034597        0.054099 |   0.000001    0.119962    0.348834    0.847336    1.671512 |    0.0 |   0.000001 |   1.671512 |
    | 14:30 | 0.042924  0.128319   0.22468               0.3466 |   0.653079    2.007107    3.622613    5.799658 |  15.437745   16.308598     17.2913        18.534641 |  15.429244   16.283186   17.246805        18.466 |  0.09039  0.180545  0.252431  0.323039 | 1.083669  1.606005  1.574485          1.541598 |  -0.94282  -0.924277  -0.881909     -0.833885 | 0.916649  0.948114  0.924217      0.890092 | 0.917065  0.946302  0.934204      0.918779 |   0.000001 |   0.054474    0.225637    0.580798            1.215308 |   1.612649 | 0.015014  0.020235  0.028859       0.037719 | 0.014417  0.019116  0.033892        0.052862 |   0.000001    0.110013    0.342158    0.812231    1.612649 |    0.0 |   0.000001 |   1.612649 |
    | 15:00 | 0.042495  0.127982  0.219094             0.339496 |   0.646457    2.001621    3.526416    5.668719 |  15.433368   16.305162   17.234326        18.462189 |  15.424952   16.279816   17.190937     18.394955 | 0.089805  0.180255  0.248749  0.319345 | 0.973427  1.606184  1.576185          1.543324 | -0.933679  -0.924439  -0.884242      -0.83655 | 1.070003  0.961639  0.922711      0.893026 | 1.068727  0.959489  0.932047      0.920489 |        0.0 |    0.05964    0.224633    0.556529            1.172971 |   1.553062 | 0.016606  0.020201  0.028407       0.037246 | 0.017692  0.019063  0.033028        0.051753 |        0.0    0.117575    0.325399    0.785018    1.553062 |    0.0 |        0.0 |   1.553062 |
    | 15:30 |  0.04207  0.126702  0.215988             0.331714 |   0.639903    1.980794     3.47307    5.525878 |  15.429034    16.29211   17.202651        18.382831 |  15.420702   16.267018   17.159877     18.317138 | 0.089223  0.179148  0.246681  0.315247 | 1.115168  1.623077   1.57713           1.54528 | -0.946001  -0.926121  -0.885544     -0.839494 | 0.842912  0.948262  0.930057      0.893642 | 0.844744  0.946108  0.938692      0.919952 |        0.0 |   0.051198    0.218631    0.543228            1.127315 |   1.500959 | 0.014402  0.019868  0.028154       0.036721 | 0.013347  0.018489  0.032548        0.050535 |        0.0     0.09932    0.327372    0.749628    1.500959 |    0.0 |        0.0 |   1.500959 |
    | 16:00 |  0.04165  0.122302  0.210593             0.325033 |   0.633417    1.909323    3.380647    5.403725 |  15.424744   16.247243   17.147639        18.314698 |  15.416495   16.223022   17.105933     18.250329 | 0.088645  0.175304  0.243053  0.311686 | 1.081898  1.608028  1.578757          1.546935 | -0.944033  -0.927099  -0.887812     -0.842028 | 0.996465  0.953008  0.924982      0.897071 | 0.994366  0.950384  0.933147      0.922201 |        0.0 |   0.051899     0.20815    0.520458            1.088776 |   1.447569 | 0.014748  0.019623  0.027711       0.036268 | 0.014041  0.018206  0.031716        0.049492 |        0.0    0.098761    0.314221    0.723357    1.447569 |    0.0 |        0.0 |   1.447569 |
    | 16:30 | 0.041233   0.11949   0.20652             0.318313 |   0.626997    1.863738    3.311052    5.281315 |  15.420496   16.218563   17.106099        18.246171 |   15.41233   16.194899     17.0652     18.183132 |  0.08807  0.172813  0.240282  0.308063 | 1.090733  1.609032  1.580005          1.548636 | -0.945124  -0.928431  -0.889534     -0.844593 |  0.95911   0.95127  0.929519      0.898532 | 0.957754  0.948578  0.936994      0.922616 |        0.0 |   0.050626    0.200169    0.503533            1.050587 |   1.394914 | 0.014534  0.019332  0.027374       0.035807 | 0.013678   0.01778  0.031088        0.048443 |        0.0    0.094589    0.304192    0.699267    1.394914 |    0.0 |        0.0 |   1.394914 |
    | 17:00 | 0.033399  0.118295  0.199547             0.312979 |   0.506559    1.844394    3.192294    5.184468 |  15.340602   16.206377   17.034984        18.191774 |  15.333988    16.18295   16.995466     18.129792 | 0.076815  0.171747  0.235471  0.305156 | 0.816299  1.689982  1.582698          1.549996 |  -0.92907  -0.933759  -0.892552     -0.846639 | 1.125054  0.916306  0.922315      0.902258 | 1.123129  0.913466  0.929193      0.925353 |        0.0 |   0.047668    0.187439    0.474944            1.020695 |   1.334578 | 0.016938  0.018293   0.02678       0.035438 | 0.019831  0.015962  0.029994        0.047611 |        0.0    0.106235    0.265343    0.693482    1.334578 |    0.0 |        0.0 |   1.334578 |
    | 17:30 | 0.039077  0.117112  0.198144             0.305867 |   0.593783    1.825257    3.168469    5.055787 |  15.398504   16.194313   17.020683        18.119249 |  15.390766   16.171121   16.981442     18.058675 | 0.085059  0.170686  0.234494  0.301238 | 1.140598  1.772739  1.582598          1.551849 | -0.950676  -0.938552  -0.893098     -0.849383 | 0.770712  0.920462   0.93965      0.900628 | 0.775183  0.916182  0.945709      0.922826 |        0.0 |   0.044281    0.175743    0.469472            0.981406 |   1.287284 | 0.013423  0.017331  0.026671       0.034941 | 0.011862  0.014367  0.029798        0.046501 |        0.0    0.082352    0.263597    0.669744    1.287284 |    0.0 |        0.0 |   1.287284 |
    | 18:00 | 0.031652  0.115941  0.197192              0.29799 |   0.479789    1.806326     3.15231    4.913842 |  15.322788    16.18237   17.010976        18.038915 |   15.31652   16.159409   16.971924     17.979901 | 0.074175  0.169632  0.233828  0.296838 | 0.859349  1.561823  1.583499           1.55408 | -0.937039  -0.927005  -0.893564     -0.852462 | 1.122748  1.090083  0.940606      0.900856 | 1.119248  1.083639  0.946509      0.921929 |        0.0 |   0.041413    0.196187    0.465487            0.938571 |   1.259254 | 0.015537   0.01955   0.02658       0.034381 | 0.016967   0.01833  0.029629        0.045263 |        0.0    0.092172     0.28997    0.622752    1.259254 |    0.0 |        0.0 |   1.259254 |
    | 18:30 | 0.031335  0.106526  0.191601             0.292662 |   0.474942    1.654637    3.057564    4.818189 |  15.319561   16.086361   16.953951         17.98458 |  15.313355   16.065265   16.916007     17.926622 | 0.073691  0.160973  0.229885  0.293825 |  0.77477  1.444386  1.584664          1.555369 | -0.927177  -0.923432  -0.895908     -0.854523 | 1.060285  0.988034  0.928443       0.90711 | 1.063323  0.985626  0.934123      0.927071 |        0.0 |   0.045173    0.184405    0.443556            0.910206 |   1.224403 |  0.01712  0.020061  0.026112       0.034004 | 0.020667  0.019747  0.028791         0.04444 |        0.0    0.098009    0.286367    0.594389    1.224403 |    0.0 |        0.0 |   1.224403 |
    | 19:00 | 0.037916  0.105461  0.185388             0.288279 |   0.575927    1.637529     2.95266    4.739704 |  15.386668   16.075498   16.890592        17.939878 |  15.379159   16.054612   16.853878     17.882787 | 0.083413  0.159972  0.225435  0.291324 | 1.027574  1.603648  1.587069          1.556482 | -0.944017  -0.934572  -0.898633     -0.856232 | 0.826623  0.866077   0.92704      0.909961 | 0.832508  0.865535   0.93202      0.929155 |        0.0 |   0.046751    0.163352     0.41941            0.887123 |   1.174238 | 0.014611  0.017956  0.025568        0.03369 | 0.014186  0.015865  0.027822        0.043761 |        0.0    0.081593    0.256489    0.588967    1.174238 |    0.0 |        0.0 |   1.174238 |
    | 19:30 | 0.030712  0.104407  0.182538              0.28279 |   0.465395    1.620603    2.904664    4.641707 |  15.313201   16.064743   16.861526        17.883907 |  15.307119   16.044066   16.825376     17.827903 | 0.072732  0.158977  0.223369  0.288163 | 0.878943  1.660042  1.587756           1.55796 | -0.940547  -0.938173   -0.89985     -0.858391 | 0.989512  0.932287   0.93869      0.908994 | 0.988671  0.928395  0.942931      0.927489 |        0.0 |   0.038511      0.1552    0.408634            0.858539 |    1.12531 | 0.014895  0.017238  0.025323       0.033293 | 0.015742  0.014662  0.027392        0.042907 |        0.0    0.080669     0.23851    0.581495     1.12531 |    0.0 |        0.0 |    1.12531 |
    | 20:00 | 0.035933  0.103363  0.184975             0.274382 |    0.54545    1.603857    2.945697    4.492158 |  15.366446   16.054095   16.886378        17.798159 |  15.359329   16.033626   16.849746      17.74382 | 0.080555  0.157988  0.225136  0.283255 | 0.971199  1.435518  1.588164          1.560505 | -0.941767  -0.924508    -0.8989     -0.861751 | 0.972344  1.108678  0.954739      0.904744 | 0.973993   1.10298  0.958698      0.922308 |        0.0 |   0.045242    0.176514    0.417579            0.815393 |    1.10414 |  0.01493   0.01981  0.025517       0.032673 |  0.01506  0.019417  0.027725        0.041585 |        0.0    0.078571    0.279868    0.533619     1.10414 |    0.0 |        0.0 |    1.10414 |
    | 20:30 | 0.029106  0.102329  0.175724             0.271379 |   0.440821     1.58729     2.79026    4.438924 |  15.296821   16.043554   16.792044        17.767537 |  15.291057   16.023289   16.757243     17.713794 | 0.070227  0.157003  0.218363  0.281482 | 0.839277  1.553269  1.590272          1.561076 | -0.939051  -0.933081  -0.902854     -0.862912 | 0.977959  0.888053  0.919545      0.917042 | 0.978214  0.887355  0.923615      0.933535 |        0.0 |   0.036886    0.160442    0.383135            0.800395 |   1.064707 | 0.015062  0.018194  0.024716       0.032456 | 0.016371  0.016424  0.026337        0.041132 |        0.0    0.076859    0.246402    0.527745    1.064707 |    0.0 |        0.0 |   1.064707 |
    | 21:00 | 0.028815  0.101306  0.172725             0.267062 |   0.436371    1.570899    2.740052    4.362539 |  15.293853   16.033119   16.761461        17.723508 |  15.288146   16.013056   16.727254     17.670619 | 0.069768  0.156025  0.216129  0.278914 | 0.710245  1.588408  1.591253          1.562287 | -0.922495   -0.93567  -0.904172     -0.864636 | 1.126976  0.942191   0.94024      0.915089 | 1.129931   0.93886  0.943292      0.931106 |        0.0 |   0.042865    0.154305    0.372163            0.778842 |   1.017483 | 0.017682  0.017681  0.024448       0.032135 | 0.022633  0.015553  0.025877         0.04046 |        0.0    0.086846    0.222494    0.528322    1.017483 |    0.0 |        0.0 |   1.017483 |
    | 21:30 | 0.028526  0.100293  0.173662             0.260325 |   0.431966    1.554682    2.755722    4.243721 |  15.290914   16.022788   16.771012        17.654805 |  15.285265   16.002926    16.73662      17.60325 | 0.069312  0.155052  0.216829  0.274863 | 0.771738  1.517764  1.590663          1.564225 | -0.932094  -0.931886  -0.903735     -0.867344 | 0.878883  1.006736  0.953938      0.911319 |  0.88763   1.00246  0.956668      0.926619 |        0.0 |   0.038796    0.158823    0.375642            0.745699 |   0.988873 | 0.016166  0.018388  0.024536       0.031629 |  0.01898  0.016869   0.02603        0.039411 |        0.0    0.077087    0.238636    0.502011    0.988873 |    0.0 |        0.0 |   0.988873 |
    | 22:00 | 0.028241   0.09929  0.171925             0.257213 |   0.427606    1.538637    2.726672    4.188994 |  15.288005    16.01256   16.753302        17.623072 |  15.282412   15.992897   16.719254     17.572134 | 0.068859  0.154084   0.21553  0.272972 | 0.718281  1.666616  1.632046          1.565034 | -0.925022  -0.940812  -0.907919     -0.868588 | 1.024489  0.879555  0.922822      0.920173 |  1.03027  0.876886  0.925582      0.934606 |        0.0 |   0.040993    0.142252    0.360088            0.730642 |   0.950838 | 0.017256  0.016642  0.023771       0.031396 | 0.021693  0.013855  0.024491        0.038931 |        0.0     0.07942    0.202339    0.501163    0.950838 |    0.0 |        0.0 |   0.950838 |
    | 22:30 | 0.027959  0.098297  0.167627             0.251261 |   0.423291    1.522763    2.654902    4.084571 |  15.285125   16.002434   16.709469        17.562367 |  15.279588   15.982968   16.676272     17.512607 | 0.068408  0.153122  0.212288  0.269321 | 0.703922  1.544786  1.592618          1.566789 | -0.923343  -0.934588  -0.906397     -0.871003 | 0.970543  1.039953  0.961409      0.915008 | 0.979195  1.034173  0.962795      0.928903 |        0.0 |   0.041136    0.150939    0.353885            0.702113 |    0.92512 | 0.017493  0.017842  0.023993       0.030941 | 0.022363   0.01597  0.025107        0.038004 |        0.0    0.077768    0.219166    0.478396     0.92512 |    0.0 |        0.0 |    0.92512 |
    | 23:00 | 0.027679  0.097314  0.164965             0.246181 |   0.419019    1.507057    2.610549    3.995745 |  15.282274    15.99241   16.682324        17.510566 |  15.276792   15.973138   16.649654     17.461813 |  0.06796  0.152165   0.21026  0.266169 | 0.723732  1.488094  1.593442          1.568306 | -0.927063   -0.93153  -0.907573     -0.873074 | 0.926491  0.998159  0.943134      0.918149 |  0.93556  0.994428  0.945236      0.931212 |        0.0 |   0.039347    0.154104     0.34447            0.678149 |   0.905954 | 0.016903  0.018406  0.023752       0.030549 | 0.020946  0.017043  0.024701        0.037214 |        0.0    0.072757    0.227794     0.45216    0.905954 |    0.0 |        0.0 |   0.905954 |
    | 23:30 |  0.02242  0.096341  0.163316             0.243154 |   0.338816    1.491517    2.583097    3.942924 |  15.228642   15.982486   16.665501        17.479691 |  15.224202   15.963407   16.633158     17.431537 | 0.059205  0.151214  0.208995  0.264274 | 0.555633     1.652  1.671426          1.569053 | -0.908371  -0.941426  -0.914347     -0.874295 | 1.078356   0.87012  0.906163      0.923833 | 1.087112  0.867751  0.907885       0.93623 |        0.0 |   0.036102    0.136524     0.32299            0.664104 |   0.875626 |  0.01918  0.016476  0.022507       0.030317 | 0.028835  0.013694  0.022236        0.036751 |        0.0    0.079095    0.186514     0.44639    0.875626 |    0.0 |        0.0 |   0.875626 |

Unfortunately, this simple approach for preventing negative discharges introduces an
(in this case, relatively small) water balance error:

>>> round_(model.check_waterbalance(conditions))
-0.000149

.. _musk_mct_little_baseflow:

Little baseflow
_______________

Here, we repeat the :ref:`musk_mct_no_baseflow` example but assume a small baseflow
of 5 m³/s.  The trimming of negative discharge values again works and, unfortunately,
again introduces some error into the water balance.  Besides that, we see a kind of
fluctuation progressing from the first to the last segment endpoint with rising
amplitude.  As a result, the channel outflow first rises to around 8 m³/s and then
drops to zero before it finally starts to reflect the expected flood wave shape:

>>> model.prepare_states(discharge=5.0)
>>> test.inits = ((states.discharge, 5.0),
...               (states.courantnumber, states.courantnumber[0]),
...               (states.reynoldsnumber, states.reynoldsnumber[0]))
>>> nodes.input1.sequences.sim.series = 5.0

.. integration-test::

    >>> conditions = test("musk_mct_little_baseflow", get_conditions="2000-01-01")
    |  date |                               referencewaterlevel |                                     wettedarea |                                     wettedperimeter |                                     surfacewidth |                               celerity |                               correctingfactor |                                  coefficient1 |                               coefficient2 |                               coefficient3 |     inflow |                                     referencedischarge |    outflow |                               courantnumber |                               reynoldsnumber |                                                  discharge | input1 |     input2 |     output |
    -------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
    | 00:00 | 0.783626  0.783626  0.783626             0.783626 |  14.824738   14.824738   14.824738   14.824738 |  22.991448   22.991448   22.991448        22.991448 |   22.83626    22.83626    22.83626      22.83626 | 0.497225  0.497225  0.497225  0.497225 | 1.474247  1.474247  1.474247          1.474247 | -0.694649  -0.694649  -0.694649     -0.694649 | 0.797528  0.797528  0.797528      0.797528 | 0.897118  0.897118  0.897118      0.897118 |        5.0 |        5.0         5.0         5.0                 5.0 |        5.0 | 0.060709  0.060709  0.060709       0.060709 | 0.119476  0.119476  0.119476        0.119476 |        5.0         5.0         5.0         5.0         5.0 |    5.0 |        0.0 |        5.0 |
    | 00:30 | 0.783626  0.783626  0.783626             0.783626 |  14.824738   14.824738   14.824738   14.824738 |  22.991448   22.991448   22.991448        22.991448 |   22.83626    22.83626    22.83626      22.83626 | 0.497225  0.497225  0.497225  0.497225 | 1.474247  1.474247  1.474247          1.474247 | -0.694649  -0.694649  -0.694649     -0.694649 |  0.79753   0.79753   0.79753       0.79753 | 0.897119  0.897119  0.897119      0.897119 |        5.0 |        5.0         5.0         5.0                 5.0 |        5.0 | 0.060709  0.060709  0.060709       0.060709 | 0.119476  0.119476  0.119476        0.119476 |        5.0         5.0         5.0         5.0         5.0 |    5.0 |        0.0 |        5.0 |
    | 01:00 | 0.783626  0.783626  0.783626             0.783626 |  14.824738   14.824738   14.824738   14.824738 |  22.991448   22.991448   22.991448        22.991448 |   22.83626    22.83626    22.83626      22.83626 | 0.497225  0.497225  0.497225  0.497225 | 1.474247  1.474247  1.474247          1.474247 | -0.694649  -0.694649  -0.694649     -0.694649 |  0.79753   0.79753   0.79753       0.79753 | 0.897119  0.897119  0.897119      0.897119 |        5.0 |        5.0         5.0         5.0                 5.0 |        5.0 | 0.060709  0.060709  0.060709       0.060709 | 0.119476  0.119476  0.119476        0.119476 |        5.0         5.0         5.0         5.0         5.0 |    5.0 |        0.0 |        5.0 |
    | 01:30 | 0.783626  0.783626  0.783626             0.783626 |  14.824738   14.824738   14.824738   14.824738 |  22.991448   22.991448   22.991448        22.991448 |   22.83626    22.83626    22.83626      22.83626 | 0.497225  0.497225  0.497225  0.497225 | 1.474247  1.474247  1.474247          1.474247 | -0.694649  -0.694649  -0.694649     -0.694649 |  0.79753   0.79753   0.79753       0.79753 | 0.897119  0.897119  0.897119      0.897119 |        5.0 |        5.0         5.0         5.0                 5.0 |        5.0 | 0.060709  0.060709  0.060709       0.060709 | 0.119476  0.119476  0.119476        0.119476 |        5.0         5.0         5.0         5.0         5.0 |    5.0 |        0.0 |        5.0 |
    | 02:00 | 0.783626  0.783626  0.783626             0.783626 |  14.824738   14.824738   14.824738   14.824738 |  22.991448   22.991448   22.991448        22.991448 |   22.83626    22.83626    22.83626      22.83626 | 0.497225  0.497225  0.497225  0.497225 | 1.474247  1.474247  1.474247          1.474247 | -0.694649  -0.694649  -0.694649     -0.694649 |  0.79753   0.79753   0.79753       0.79753 | 0.897119  0.897119  0.897119      0.897119 |        5.0 |        5.0         5.0         5.0                 5.0 |        5.0 | 0.060709  0.060709  0.060709       0.060709 | 0.119476  0.119476  0.119476        0.119476 |        5.0         5.0         5.0         5.0         5.0 |    5.0 |        0.0 |        5.0 |
    | 02:30 | 0.783626  0.783626  0.783626             0.783626 |  14.824738   14.824738   14.824738   14.824738 |  22.991448   22.991448   22.991448        22.991448 |   22.83626    22.83626    22.83626      22.83626 | 0.497225  0.497225  0.497225  0.497225 | 1.474247  1.474247  1.474247          1.474247 | -0.694649  -0.694649  -0.694649     -0.694649 |  0.79753   0.79753   0.79753       0.79753 | 0.897119  0.897119  0.897119      0.897119 |        5.0 |        5.0         5.0         5.0                 5.0 |        5.0 | 0.060709  0.060709  0.060709       0.060709 | 0.119476  0.119476  0.119476        0.119476 |        5.0         5.0         5.0         5.0         5.0 |    5.0 |        0.0 |        5.0 |
    | 03:00 | 0.783626  0.783626  0.783626             0.783626 |  14.824741   14.824736   14.824738   14.824737 |  22.991449   22.991448   22.991448        22.991448 |  22.836261   22.836259    22.83626     22.836259 | 0.497226  0.497225  0.497225  0.497225 | 1.474247  1.474247  1.474247          1.474247 | -0.694649  -0.694649  -0.694649     -0.694649 |  0.79753   0.79753   0.79753       0.79753 | 0.897119  0.897119  0.897119      0.897119 |   5.000003 |   5.000001    4.999999         5.0                 5.0 |        5.0 | 0.060709  0.060709  0.060709       0.060709 | 0.119476  0.119476  0.119476        0.119476 |   5.000003    4.999998    5.000001         5.0         5.0 |    5.0 |   0.000003 |        5.0 |
    | 03:30 | 0.783627  0.783625  0.783626             0.783626 |  14.824761   14.824727   14.824742   14.824735 |  22.991459   22.991443    22.99145        22.991447 |   22.83627   22.836255   22.836262     22.836259 | 0.497226  0.497225  0.497226  0.497225 | 1.474247  1.474248  1.474247          1.474247 | -0.694649  -0.694649  -0.694649     -0.694649 |  0.79753   0.79753   0.79753       0.79753 | 0.897119  0.897119  0.897119      0.897119 |   5.000029 |   5.000012    4.999995    5.000002            4.999999 |   5.000001 | 0.060709  0.060709  0.060709       0.060709 | 0.119476  0.119476  0.119476        0.119476 |   5.000029    4.999987    5.000006    4.999997    5.000001 |    5.0 |   0.000029 |   5.000001 |
    | 04:00 | 0.783632  0.783623  0.783627             0.783625 |   14.82488   14.824674   14.824766   14.824725 |  22.991512    22.99142   22.991461        22.991443 |  22.836322   22.836232   22.836272     22.836254 | 0.497227  0.497225  0.497226  0.497225 | 1.474247  1.474248  1.474247          1.474248 | -0.694647   -0.69465  -0.694649     -0.694649 | 0.797532  0.797529   0.79753       0.79753 | 0.897122  0.897118  0.897119      0.897119 |   5.000175 |   5.000071    4.999968    5.000014            4.999994 |   5.000007 |  0.06071  0.060709  0.060709       0.060709 | 0.119477  0.119476  0.119476        0.119476 |   5.000175    4.999922    5.000035    4.999985    5.000007 |    5.0 |   0.000175 |   5.000007 |
    | 04:30 | 0.783655  0.783613  0.783632             0.783624 |  14.825403   14.824447   14.824864   14.824682 |  22.991745   22.991318   22.991505        22.991423 |  22.836551   22.836132   22.836315     22.836235 | 0.497235  0.497221  0.497227  0.497225 | 1.474244  1.474249  1.474247          1.474248 | -0.694641  -0.694652  -0.694647      -0.69465 | 0.797539  0.797526  0.797532      0.797529 |  0.89713  0.897114  0.897121      0.897118 |   5.000826 |   5.000332    4.999855    5.000063            4.999972 |    5.00003 | 0.060711  0.060709   0.06071       0.060709 | 0.119481  0.119474  0.119477        0.119476 |   5.000826    4.999639    5.000158    4.999931     5.00003 |    5.0 |   0.000826 |    5.00003 |
    | 05:00 | 0.783738  0.783578  0.783646             0.783617 |  14.827287   14.823646   14.825204   14.824539 |  22.992586   22.990961   22.991656        22.991359 |  22.837376   22.835782   22.836464     22.836173 | 0.497261   0.49721  0.497232  0.497223 | 1.474233  1.474254  1.474245          1.474249 | -0.694617  -0.694662  -0.694643     -0.694651 | 0.797561  0.797517  0.797535      0.797528 | 0.897159  0.897102  0.897126      0.897116 |   5.003194 |   5.001271    4.999456    5.000233            4.999901 |   5.000106 | 0.060714  0.060707   0.06071       0.060709 | 0.119493  0.119469  0.119479        0.119475 |   5.003194    4.998631    5.000586     4.99975    5.000106 |    5.0 |   0.003194 |   5.000106 |
    | 05:30 | 0.783991  0.783473   0.78369             0.783599 |  14.833071   14.821249   14.826192   14.824133 |  22.995169    22.98989   22.992098        22.991178 |  22.839909   22.834732   22.836897     22.835995 | 0.497343  0.497176  0.497246  0.497217 | 1.474199  1.474268  1.474239          1.474251 | -0.694545  -0.694692  -0.694631     -0.694656 | 0.797619  0.797493  0.797545      0.797524 | 0.897242  0.897068   0.89714       0.89711 |   5.010515 |   5.004154    4.998261    5.000725            4.999699 |   5.000317 | 0.060726  0.060703  0.060712       0.060708 | 0.119532  0.119453  0.119486        0.119472 |   5.010515    4.995595     5.00184    4.999235    5.000317 |    5.0 |   0.010515 |   5.000317 |
    | 06:00 | 0.784672  0.783199  0.783799             0.783556 |  14.848642   14.814987   14.828689   14.823145 |  23.002121   22.987093   22.993213        22.990737 |  22.846725   22.831989    22.83799     22.835562 | 0.497561  0.497088  0.497281  0.497203 | 1.474108  1.474304  1.474224          1.474257 | -0.694351   -0.69477    -0.6946     -0.694669 | 0.797756   0.79744  0.797565      0.797516 | 0.897445  0.896988  0.897171      0.897098 |   5.030315 |   5.011917    4.995141    5.001969            4.999206 |   5.000812 | 0.060756   0.06069  0.060717       0.060706 | 0.119637  0.119411  0.119503        0.119465 |   5.030315    4.987629    5.005022    4.997974    5.000812 |    5.0 |   0.030315 |   5.000812 |
    | 06:30 | 0.786315  0.782559  0.784044             0.783463 |  14.886188   14.800386    14.83429   14.821023 |  23.018874   22.980571   22.995713        22.989789 |  22.863153   22.825593   22.840442     22.834633 | 0.498088  0.496883   0.49736  0.497173 | 1.473892  1.474389  1.474192          1.474269 | -0.693885  -0.694952   -0.69453     -0.694695 | 0.798035  0.797336  0.797602      0.797504 | 0.897892  0.896818  0.897234      0.897076 |   5.078179 |   5.030646     4.98787     5.00476            4.998149 |   5.001823 | 0.060829  0.060662  0.060728       0.060702 | 0.119888  0.119313   0.11954        0.119451 |   5.078179    4.969046    5.012164    4.995266    5.001823 |    5.0 |   0.078179 |   5.001823 |
    | 07:00 | 0.789919  0.781206   0.78454             0.783285 |  14.968635   14.769515   14.845605   14.816954 |   23.05562   22.966774   23.000765        22.987972 |  22.899185   22.812065   22.845395     22.832851 |  0.49924  0.496448  0.497519  0.497116 | 1.473421  1.474568  1.474127          1.474292 | -0.692865  -0.695336   -0.69439     -0.694746 | 0.798541  0.797157  0.797659      0.797486 | 0.898783  0.896491  0.897346      0.897038 |   5.183351 |   5.071834    4.972508    5.010399            4.996123 |    5.00363 |  0.06099  0.060601   0.06075       0.060694 | 0.120439  0.119106  0.119616        0.119424 |   5.183351    4.929869    5.026551    4.990098     5.00363 |    5.0 |   0.183351 |    5.00363 |
    | 07:30 | 0.797199   0.77858   0.78545             0.782978 |  15.135615   14.709637   14.866415   14.809939 |  23.129866   22.939991   23.010053        22.984839 |  22.971989   22.785801   22.854503     22.829778 |  0.50156  0.495603  0.497811  0.497017 | 1.472483  1.474914  1.474008          1.474332 | -0.690812  -0.696083  -0.694131     -0.694833 |  0.79935  0.796881  0.797735      0.797468 |   0.9004  0.895917  0.897528      0.896983 |   5.396214 |   5.155522    4.942755    5.020771            4.992631 |   5.006454 | 0.061312  0.060484  0.060791        0.06068 | 0.121552  0.118704  0.119755        0.119377 |   5.396214    4.854398    5.052824    4.981281    5.006454 |    5.0 |   0.396214 |   5.006454 |
    | 08:00 | 0.810875  0.773865  0.786982             0.782502 |  15.450728   14.602316   14.901439   14.799075 |   23.26934   22.891908   23.025676        22.979985 |  23.108755   22.738652   22.869822     22.825019 | 0.505884  0.494081  0.498301  0.496865 |  1.47076  1.475538   1.47381          1.474394 |  -0.68698  -0.697426  -0.693698     -0.694968 |  0.80045   0.79651  0.797808      0.797461 | 0.903069  0.894989  0.897788      0.896915 |   5.797305 |   5.314453    4.889561     5.03824            4.987226 |   5.010225 | 0.061913  0.060273  0.060859       0.060659 | 0.123637  0.117982  0.119989        0.119304 |   5.797305    4.720622    5.096515    4.967876    5.010225 |    5.0 |   0.797305 |   5.010225 |
    | 08:30 | 0.835996  0.766277  0.789219             0.781905 |  16.034394   14.430053   14.952624   14.785442 |  23.525523   22.814521   23.048489        22.973893 |  23.359963   22.662768   22.892192     22.819045 | 0.513719  0.491621  0.499017  0.496673 | 1.467713  1.476545  1.473523          1.474471 | -0.680026  -0.699592  -0.693066     -0.695137 | 0.802032  0.796224  0.797776      0.797513 | 0.907554  0.893752  0.898051      0.896881 |   6.507032 |   5.612252    4.804539    5.063791            4.980448 |   5.013521 | 0.063002  0.059932  0.060958       0.060633 | 0.127455  0.116821  0.120331        0.119213 |   6.507032    4.508893    5.159305    4.951545    5.013521 |    5.0 |   1.507032 |   5.013521 |
    | 09:00 | 0.874828  0.753842  0.792652             0.781064 |  16.949051   14.149021   15.031259   14.766268 |  23.921535   22.687711   23.083493        22.965323 |  23.748285   22.538421   22.926517     22.810642 | 0.525569  0.487559  0.500113  0.496403 | 1.463244  1.478219  1.473085          1.474579 | -0.669469  -0.703162  -0.692099     -0.695375 | 0.801387  0.795597  0.797802      0.797537 | 0.911807  0.891579  0.898517      0.896792 |   7.694853 |   6.087771    4.666751    5.103116             4.97092 |   5.016701 | 0.064653  0.059369   0.06111       0.060595 | 0.133334  0.114917  0.120854        0.119085 |   7.694853    4.174428    5.251891    4.930323    5.016701 |    5.0 |   2.694853 |   5.016701 |
    | 09:30 | 0.937948  0.737254  0.796443             0.780521 |  18.467957   13.776526   15.118248   14.753869 |  24.565232   22.518545   23.122155        22.959779 |  24.379482   22.372539   22.964428     22.805205 | 0.544208  0.482078   0.50132  0.496228 | 1.456492  1.480471  1.472612          1.474644 | -0.652767  -0.707954  -0.691034     -0.695527 | 0.800615  0.796262  0.797177      0.797862 | 0.918699  0.889926  0.898483      0.896993 |   9.586356 |   6.900418    4.485984    5.146689            4.964777 |   5.012715 | 0.067256  0.058613  0.061277       0.060571 | 0.142836  0.112379  0.121431        0.119003 |   9.586356    3.738002    5.351398    4.918498    5.012715 |    5.0 |   4.586356 |   5.012715 |
    | 10:00 | 1.030132  0.717149  0.800408              0.78047 |  20.757836   13.328749   15.209381   14.752724 |  25.505325   22.313514    23.16259        22.959267 |  25.301318    22.17149   23.004078     22.804703 | 0.570204  0.475339  0.502578  0.496212 | 1.447593  1.483275  1.472125          1.474643 | -0.629319  -0.713823  -0.689925     -0.695539 | 0.793912  0.797991  0.796392      0.798207 | 0.923733  0.888677  0.898336      0.897278 |  12.466641 |   8.176469    4.271406    5.192429            4.964234 |   4.997456 | 0.070902  0.057684  0.061451       0.060569 | 0.156605  0.109298  0.122033        0.118996 |  12.466641    3.218142    5.441369    4.926139    4.997456 |    5.0 |   7.466641 |   4.997456 |
    | 10:30 | 1.158827  0.696765  0.801312              0.78239 |  24.096814   12.878892   15.230188   14.796523 |  26.817766   22.105641   23.171813        22.978845 |  26.588273   21.967654   23.013122     22.823901 | 0.604443  0.468391  0.502865  0.496829 | 1.433889  1.486092  1.472051          1.474379 | -0.597194  -0.719812  -0.689683     -0.694991 | 0.781396   0.80208  0.794153      0.799114 | 0.927887  0.889384   0.89658      0.898288 |  16.678559 |  10.157799    4.059213    5.202761            4.986056 |   4.955091 | 0.075877  0.056733   0.06149       0.060655 | 0.176319  0.106185  0.122164        0.119292 |  16.678559     2.76712    5.428868    4.993754    4.955091 |    5.0 |  11.678559 |   4.955091 |
    | 11:00 | 1.333998  0.680053  0.797012             0.786881 |  28.907718   12.513154   15.131319   14.899129 |  28.604162   21.935206   23.127959        23.024645 |  28.339978   21.800529   22.970119     22.868812 | 0.647942  0.462606    0.5015  0.498269 | 1.423946  1.488342  1.472654          1.473787 | -0.558684  -0.724742  -0.690908     -0.693716 | 0.756764  0.808375  0.791677      0.799833 | 0.925759  0.892487  0.893952      0.899475 |  22.614836 |  13.153964    3.889333    5.152848            5.037209 |   4.881759 | 0.081906  0.055948  0.061298       0.060856 | 0.201227  0.103646  0.121499         0.11998 |  22.614836    2.548864    5.234796    5.145323    4.881759 |    5.0 |  17.614836 |   4.881759 |
    | 11:30 | 1.555271   0.68429  0.781374             0.795732 |  35.423418   12.605623   14.773343   15.101933 |  30.860719    21.97842   22.968485        23.114909 |  30.552715   21.842904   22.813743     22.957323 | 0.698919  0.464081  0.496502  0.501094 | 1.410805  1.479583  1.474761          1.472648 | -0.512655  -0.721311  -0.695352     -0.691218 | 0.725177  0.827083  0.787578      0.800742 | 0.921683  0.909937  0.888479      0.901379 |  30.703736 |  17.548914    3.953834    4.973685             5.13869 |   4.770368 | 0.089173  0.056458    0.0606       0.061248 | 0.233006  0.105447  0.119096        0.121331 |  30.703736    3.008585    4.701335    5.425238    4.770368 |    5.0 |  25.703736 |   4.770368 |
    | 12:00 | 1.824281  0.725479  0.743074             0.811957 |  44.004228   13.513782   13.906896   15.475732 |  33.604091   22.398462   22.577894        23.280372 |  33.242812   22.254789   22.430736     23.119572 | 0.756385  0.478144  0.484009  0.506225 | 1.398195  1.481918   1.47738          1.470614 | -0.461753  -0.711333  -0.705623     -0.686676 |  0.68331  0.837087  0.781334      0.801891 | 0.912897  0.923328  0.878425      0.904412 |  41.388647 |  23.805063    4.360253    4.556085            5.327159 |    4.60281 | 0.097375  0.058077   0.05897       0.061961 | 0.270845  0.110603  0.113622        0.123804 |  41.388647     4.61536    3.576267    5.915484     4.60281 |    5.0 |  36.388647 |    4.60281 |
    | 12:30 | 2.137883  0.816723  0.683024             0.832408 |  54.920948   15.586025   12.577972   15.950635 |   36.80221   23.328972   21.965507        23.488929 |  36.378826   23.167229   21.830242      23.32408 | 0.818709  0.507721   0.46364  0.512609 | 1.386898  1.469971  1.485228          1.468143 | -0.408246  -0.685332  -0.723145     -0.681014 | 0.635063  0.854682  0.776087      0.799816 | 0.901634  0.949444   0.86582      0.905263 |  55.102726 |  32.420731    5.383335    3.926439            5.569233 |   4.451439 | 0.106257  0.062171   0.05619       0.062848 |  0.31395  0.124539  0.104478         0.12691 |  55.102726    7.950315    1.891521    6.529392    4.451439 |    5.0 |  50.102726 |   4.451439 |
    | 13:00 |  2.49076  0.976001  0.611603             0.852212 |  68.380823   19.402896   11.044335   16.414496 |  40.400867   24.953293   21.237151        23.690887 |  39.907599   24.760007    21.11603     23.522115 | 0.884331  0.555104  0.437991  0.518705 | 1.378314  1.452373  1.498145          1.465841 | -0.354767  -0.642849  -0.745448     -0.675598 | 0.583324  0.852276  0.777867        0.7947 | 0.889146  0.965656  0.856802       0.90349 |  72.240349 |  43.873366    7.415877    3.228872            5.808462 |   4.437126 | 0.115489  0.068797  0.052624       0.063695 |  0.36078    0.1486  0.093214        0.129908 |  72.240349   13.583255         0.0    7.065749    4.437126 |    5.0 |  67.240349 |   4.437126 |
    | 13:30 | 2.875422  1.220564  0.616576             0.837185 |  84.471588    25.75733    11.14948   16.062162 |  44.323666   27.447354   21.287871        23.537643 |   43.75422   27.205635   21.165765     23.371847 | 0.951721  0.620143  0.439833  0.514087 | 1.370531  1.431248    1.4888          1.468015 | -0.301982  -0.583701  -0.741797     -0.679835 | 0.531751  0.826047  0.844329      0.776169 | 0.877405  0.969324  0.915771      0.886243 |  93.127569 |  58.658585   11.160349    3.293864            5.624835 |   5.017644 | 0.124995  0.077992  0.053177       0.063034 | 0.411124  0.184872  0.095062        0.127559 |  93.127569   22.209071         0.0    6.470609    5.017644 |    5.0 |  88.127569 |   5.017644 |
    | 14:00 | 3.284974  1.584747  0.576197             0.826145 | 103.229895   36.328308   10.302974   15.804741 |  48.500296   31.161308   20.876081        23.425054 |  47.849743   30.847466   20.761972     23.261445 | 1.019772  0.705433  0.424617  0.510663 | 1.364242  1.408797  1.496952          1.463391 | -0.251397  -0.506612  -0.754943     -0.681067 | 0.480812   0.77752  0.807214       0.78353 | 0.866244  0.963613  0.877791      0.891618 | 117.993962 |  77.164428   18.190837    2.922482            5.515202 |   5.675387 |  0.13455  0.090132  0.051058       0.062813 | 0.463664   0.23735  0.088581        0.126908 | 117.993962   34.352028         0.0    5.679841    5.675387 |    5.0 | 112.993962 |   5.675387 |
    | 14:30 | 3.712098   1.95602  0.596296             0.781846 | 124.579808   48.470374   10.722295   14.784103 |  52.856116   34.947569   21.081055        22.973295 |  52.120976   34.560201   20.962965     22.818459 | 1.087496  0.783099  0.432267  0.496654 | 1.359216  1.393001  1.501187          1.470121 | -0.203713  -0.438543  -0.750366     -0.693879 | 0.432185  0.688634    0.8551       0.76738 | 0.856214  0.926395  0.921773      0.872489 | 146.948195 |  99.675129   27.248446    3.087484            4.994542 |   6.469725 | 0.144016   0.10119  0.051831        0.06081 | 0.517509  0.289106  0.090787        0.119912 | 146.948195   50.472631    1.521595    4.093774    6.469725 |    5.0 | 141.948195 |   6.469725 |
    | 15:00 | 4.149975  2.374733  0.715259             0.661393 | 148.361066   63.817775   13.286853   12.108106 |  57.321603   39.217619   22.294235        21.744915 |  56.499746   38.747329   22.152586     21.613934 | 1.154109  0.863205  0.474699  0.456046 | 1.355215  1.380385  1.477985          1.490561 | -0.159296  -0.371452  -0.712854     -0.730211 | 0.386536  0.619441  0.918047      0.737174 | 0.847403  0.906118  0.992474      0.829785 |  179.95935 |  126.34513   39.907565    4.267471            3.704552 |   7.681888 | 0.153289  0.112561  0.057812       0.055072 | 0.571896  0.345748  0.109829        0.100856 |  179.95935   70.904597    6.305892    0.964678    7.681888 |    5.0 |  174.95935 |   7.681888 |
    | 15:30 | 4.592338  2.831363  0.991152             0.668339 | 174.332898   82.553543    19.77919   12.258476 |   61.83284   43.874354   25.107806         21.81575 |  60.923378   43.313634    24.91152     21.683393 | 1.218981  0.944189  0.559376  0.458501 | 1.352029  1.371316  1.450865          1.485885 |  -0.11829  -0.307748  -0.638978     -0.727166 | 0.344165  0.552066  0.932551      0.831087 | 0.839767  0.887832  1.034891      0.910839 | 216.845554 | 157.177442   56.840379    7.625801            3.782609 |   7.798696 | 0.162287  0.123935  0.069398       0.055543 | 0.626157  0.405411  0.150874        0.102423 | 216.845554   95.828348   15.251601         0.0    7.798696 |    5.0 | 211.845554 |   7.798696 |
    | 16:00 | 5.033507  3.309537  1.410195             0.643575 | 202.183561  104.408211   31.096166   11.724579 |    66.3319   48.750783   29.381221        21.563208 |  65.335069   48.095366   29.101947     21.435755 | 1.281603   1.02375  0.665941   0.44968 | 1.349483  1.363915  1.418262          1.485892 | -0.080697  -0.248523  -0.542014     -0.733752 | 0.305154  0.488982  0.862485      0.810338 | 0.833205  0.872093  1.015494      0.890053 | 257.270855 | 192.013678   78.368446   14.601122            3.548248 |   6.941256 | 0.170946  0.135107  0.084519       0.054474 | 0.679712  0.466785  0.212487        0.099093 | 257.270855  125.255044   29.030359         0.0    6.941256 |    5.0 | 252.270855 |   6.941256 |
    | 16:30 | 5.468218  3.797807  1.892858             0.599074 | 231.530316  129.083816   46.307419   10.780561 |  70.765101   53.730188   34.303437        21.109381 |  69.682181   52.978074   33.928578     20.990741 | 1.341546  1.100743  0.770395  0.433312 | 1.347439  1.358353  1.394987          1.494228 |  -0.04644  -0.194687  -0.449325     -0.747932 | 0.269453     0.431  0.743245      0.800088 | 0.827592  0.858796  0.961383      0.874822 | 300.750601 | 230.517753  104.603238   25.573714             3.12626 |   6.051365 | 0.179213  0.145863  0.099407       0.052198 | 0.732029  0.528215  0.280546        0.092011 | 300.750601  159.015613   47.957718    0.028073    6.051365 |    5.0 | 295.750601 |   6.051365 |
    | 17:00 | 5.892261  4.286565  2.364244             0.683867 |  261.97758  156.171645   63.411908   12.596381 |  75.089503   58.714552   39.110652        21.974105 |  73.922605   57.865645    38.64244     21.838673 | 1.398526  1.174381  0.861274  0.463934 | 1.345787  1.354152  1.381184          1.487669 | -0.015357  -0.146249  -0.373318     -0.723571 | 0.236958  0.378843  0.634889      0.889858 | 0.822822  0.847883  0.915776      0.963652 | 346.664962 | 272.243835   135.43909   39.542209            3.928215 |    3.28918 | 0.187054  0.156104  0.112244       0.056133 | 0.782697  0.588717  0.344083        0.104248 | 346.664962  196.783382    72.12504    3.547969     3.28918 |    5.0 | 341.664962 |    3.28918 |
    | 17:30 | 6.301381  4.768179  2.875948             0.914832 | 293.057752  185.200321   84.494586   17.907064 |  79.261733   63.626072   44.329026        24.329491 |  78.013813   62.681787   43.759476     24.148318 | 1.452248  1.244172  0.951811  0.537466 | 1.344447  1.350948   1.37052          1.457598 |    0.0127  -0.102939  -0.301913     -0.658397 | 0.207485  0.332277  0.556902       0.93327 | 0.818763  0.838977  0.893061      1.027617 | 394.279607 | 316.555817  170.562513   58.680541            6.602951 |        0.0 | 0.194433  0.165773  0.125008       0.066372 | 0.831293  0.647564  0.411193        0.139612 | 394.279607  238.054014  101.392974   12.723335         0.0 |    5.0 | 389.279607 |        0.0 |
    | 18:00 | 6.692249  5.236786  3.406576             1.383007 | 324.314749  215.671409  109.122444   30.308661 |   83.24782   68.404945   49.740396        29.103964 |  81.922494   67.367857   49.065761     28.830075 |  1.50253  1.309838  1.039361  0.659577 | 1.343353  1.348477  1.362673          1.419903 |  0.037916  -0.064356  -0.237335      -0.54776 | 0.180881  0.290867  0.485012      0.903512 | 0.815325  0.831718  0.873917      1.046317 |  442.77215 |  362.74371  209.491616     83.2317           14.079051 |        0.0 | 0.201329  0.174842  0.137293       0.083614 | 0.877492  0.704227  0.479085        0.208576 |  442.77215  282.197221  135.411208   28.158102         0.0 |    5.0 |  437.77215 |        0.0 |
    | 18:30 | 7.061824  5.687518  3.938666              1.92468 |  355.27411  247.052046  136.645456   47.392159 |  87.016752   73.001526   55.166673         34.62796 |  85.618235   71.875176   54.386663     34.246798 | 1.549212  1.371187  1.122287  0.776819 | 1.342455  1.346548  1.357023          1.394972 |  0.060467  -0.030096  -0.180214     -0.444364 | 0.156959  0.254106  0.421149      0.757567 | 0.812412  0.825782  0.858535      0.973939 | 491.262607 | 409.991311  251.572654  113.008719           26.391296 |        0.0 | 0.207723  0.183294  0.148864       0.100237 | 0.920996  0.758274  0.545743        0.284456 | 491.262607  328.462634  173.642713   49.910151         0.0 |    5.0 | 486.262607 |        0.0 |
    | 19:00 | 7.407578  6.117045   4.46026             2.455528 | 385.474711  278.846839  166.373492   66.981017 |  90.542767   77.381859   60.485905        40.041572 |  89.075778   76.170446   59.602599     39.555282 | 1.592183  1.428187  1.199844  0.877959 | 1.341717  1.345024  1.352905          1.379042 |  0.080526    0.00028  -0.130163      -0.35984 |  0.13555  0.221533  0.365474      0.634123 | 0.809947  0.820925  0.846469      0.920518 | 538.846018 | 457.433469  296.087928  147.550836           42.643066 |    3.70053 | 0.213602  0.191129  0.159636       0.114596 | 0.961556   0.80943   0.61002        0.356166 | 538.846018  376.019282   215.41802   77.669725     3.70053 |    5.0 | 533.846018 |    3.70053 |
    | 19:30 |  7.72743  6.522112  4.963186             3.005976 | 414.477358  310.521421  197.613875   90.269119 |  93.804637   81.512755   65.614766        45.655065 |  92.274304   80.221122   64.631861     45.059764 | 1.631368  1.480763  1.271748  0.973787 | 1.341108  1.343809  1.349853          1.368334 |  0.098258    0.02712  -0.086488      -0.28528 |  0.11649  0.192683  0.317184      0.544827 | 0.807865  0.816923  0.837019      0.891896 | 584.625318 | 504.184055  342.168065   186.17959           64.240797 |  13.991659 | 0.218958  0.198345  0.169585       0.128099 |  0.99897  0.857408  0.671209        0.427982 | 584.625318  423.986816  259.931219  110.857195   13.991659 |    5.0 | 579.625318 |  13.991659 |
    | 20:00 | 8.019719  6.900698  5.441995             3.568239 |  441.87529  341.608602  229.706449  117.185247 |  96.785412   85.373584   70.497674        51.389043 |  95.197194   84.006976   69.419946     50.682393 | 1.666728  1.528959  1.337975  1.065015 | 1.340605  1.342832  1.347551          1.360764 |  0.113816   0.050782  -0.048435     -0.219252 | 0.099625    0.1672  0.275332      0.469391 | 0.806107  0.813626  0.829585      0.871504 | 627.742733 | 549.368245  388.958162   228.07416           91.716208 |  31.627235 | 0.223788  0.204949  0.178721       0.140879 | 1.033079  0.902049  0.728884        0.499471 | 627.742733  471.468912  306.319654  148.696296   31.627235 |    5.0 | 622.742733 |  31.627235 |
    | 20:30 | 8.283179  7.251001  5.893306             4.122335 | 467.302977  371.650136  262.054854  146.803262 |  99.472185   88.945996   75.100163        57.039735 |  97.831792   87.510015   73.933058     56.223351 | 1.698251  1.572805  1.398664  1.149979 | 1.340191   1.34204  1.345784          1.355439 |  0.127344   0.071561  -0.015283      -0.16198 | 0.084811   0.14473  0.239027      0.403961 | 0.804624  0.810897  0.823701      0.855647 | 667.408181 | 592.152777  435.555452  272.351847          124.550557 |  56.305306 | 0.228091  0.210951  0.187073       0.152715 | 1.063763  0.943203  0.782822        0.568485 | 667.408181  517.585439  353.668395  190.294979   56.305306 |    5.0 | 662.408181 |  56.305306 |
    | 21:00 | 8.516916  7.571776  6.314571             4.655812 | 490.442981  400.235619  294.087638  178.220105 | 101.855837   92.217269   79.396247        62.480152 | 100.169155   90.717762   78.145715      61.55812 | 1.725953  1.612365  1.453961  1.228111 | 1.339851  1.341396  1.344408          1.351628 |  0.138974   0.089724   0.013575      -0.11269 | 0.071917  0.124987  0.207477      0.348096 | 0.803375  0.808633  0.818997      0.843543 | 702.923464 |  631.77275  481.085151  318.052266          161.933723 |  87.291547 |  0.23187  0.216361  0.194668       0.163551 | 1.090941  0.980775  0.832855        0.633895 | 702.923464  561.501849  401.059612  234.673229   87.291547 |    5.0 | 697.923464 |  87.291547 |
    | 21:30 |  8.72038  7.862188  6.704376             5.161248 | 511.030831  427.002844  325.308936  210.611155 | 103.930775   95.178903   83.371489        67.634613 | 102.203799   93.621882   82.043761     66.612484 | 1.749872  1.647723  1.504075  1.299389 | 1.339575   1.34087  1.343322           1.34884 |  0.148827   0.105504   0.038675      -0.07036 | 0.060821  0.107718  0.180049      0.300533 | 0.802326  0.806751  0.815214       0.83429 | 733.701418 | 667.553934  524.720523  364.238172          202.889828 | 123.594233 | 0.235132  0.221192   0.20154       0.173401 | 1.114567  1.014705  0.878922        0.695128 | 733.701418  602.454634  447.601008  280.830413  123.594233 |    5.0 | 728.701418 | 123.594233 |
    | 22:00 | 8.893344  8.121768  7.061807             5.634203 |  528.85795  451.642074  355.272727  243.234231 | 105.694665   97.826105   87.016587        72.457818 | 103.933435   96.217678   85.618074     71.342026 | 1.770067  1.678977   1.54921  1.364016 | 1.339353  1.340441  1.342456          1.346757 |  0.157012    0.11911   0.060466     -0.034015 | 0.051413  0.092701  0.156204       0.25997 | 0.801445  0.805185  0.812148      0.827151 |  759.27953 | 698.929955  565.706784  409.989046          246.351312 | 164.089596 | 0.237885   0.22546  0.207722       0.182307 | 1.134629  1.044971  0.920993        0.751901 |  759.27953  639.772052  492.453177  327.769756  164.089596 |    5.0 |  754.27953 | 164.089596 |
    | 22:30 |  9.03587  8.350367  7.386465             6.072745 | 543.772743  473.898641  383.596272  275.482311 | 107.148151  100.157368   90.327455        76.930087 | 105.358696   98.503669   88.864647     75.727447 | 1.786614  1.706239  1.589578   1.42237 | 1.339178  1.340091   1.34176          1.345169 |   0.16363   0.130722   0.079328     -0.002759 |  0.04359  0.079742  0.135501      0.225294 | 0.800709  0.803881   0.80965      0.821598 | 779.327871 | 725.454046   603.38024  454.444951          291.292432 | 207.626939 |  0.24014  0.229181  0.213245        0.19033 | 1.151146  1.071579  0.959082        0.804167 | 779.327871  672.889639  534.852291  374.535729  207.626939 |    5.0 | 774.327871 | 207.626939 |
    | 23:00 | 9.148286  8.548123  7.678326              6.47552 | 555.680001  493.573849   409.95834  306.794583 |  108.29458  102.174089   93.303868        81.037604 | 106.482863  100.481227    91.78326     79.755198 | 1.799608  1.729634  1.625387   1.47477 | 1.339045  1.339808  1.341198           1.34394 |  0.168771   0.140501   0.095584      0.024116 |  0.03726  0.068672  0.117573      0.195546 | 0.800097  0.802795  0.807604      0.817216 | 793.651531 | 746.805375  637.182207  496.825205          336.660536 | 253.049301 | 0.241911  0.232372  0.218141       0.197523 | 1.164165  1.094566  0.993232        0.851901 | 793.651531   701.36039  574.127692   420.23856  253.049301 |    5.0 | 788.651531 | 253.049301 |
    | 23:30 | 9.231159  8.715423  7.937674             6.842542 | 564.538845  510.524352  434.098427  336.740044 | 109.139718  103.880226   95.948706        84.780511 | 107.311588  102.154232   94.376737     83.425421 | 1.809155  1.749291  1.656845  1.521611 |  1.33895  1.339582  1.340742          1.342974 |  0.172517    0.14859   0.109508      0.047231 |  0.03234  0.059341  0.102112      0.170001 | 0.799592  0.801891  0.805923      0.813737 | 802.187983 | 762.790263  666.667743   536.44449          381.532056 | 299.275198 | 0.243211  0.235053  0.222438       0.203943 | 1.173757  1.113991  1.023512        0.895203 | 802.187983  724.859756  609.714278  464.074034  299.275198 |    5.0 | 797.187983 | 299.275198 |
    | 00:00 | 9.285262  8.852876   8.16504             7.174039 | 570.359426  524.660199  455.814945  364.944793 | 109.691468  105.281974   98.267392        88.161133 | 107.852624  103.528759   96.650395     86.740393 | 1.815373  1.765353  1.684156  1.563231 |  1.33889  1.339404  1.340373          1.342206 |  0.174943   0.155112   0.121334      0.067083 | 0.028752  0.051618   0.08886      0.148038 | 0.799178   0.80114  0.804536      0.810944 |      805.0 | 773.338763  691.509452  572.723731           425.04113 | 345.295704 | 0.244058  0.237243  0.226167       0.209641 | 1.180017  1.129936   1.05001        0.934173 |      805.0  743.185836  641.159856  505.337799  345.295704 |    5.0 |      800.0 | 345.295704 |
    | 00:30 | 9.311555  8.961279  8.361161             7.470676 | 573.198636  535.941785  474.962478  391.115174 | 109.959604  106.387472  100.267446        91.186249 | 108.115553  104.612789    98.61161     89.706764 | 1.818391  1.777965  1.707521  1.599955 |  1.33886  1.339269  1.340075          1.341592 |  0.176117   0.160181   0.131262      0.084085 | 0.026426  0.045385  0.077596      0.129176 | 0.798845  0.800519   0.80339      0.808686 | 802.264896 | 778.497299  711.496843  605.196039          466.435929 | 390.207716 | 0.244469  0.238961  0.229356       0.214664 | 1.183059  1.142503  1.072835        0.968944 | 802.264896  756.255331  668.127835  543.434504  390.207716 |    5.0 | 797.264896 | 390.207716 |
    | 01:00 | 9.311152  9.041593  8.526946             7.733373 | 573.155082   544.37591  491.448226  415.025896 | 109.955496  107.206518  101.958128        93.865241 | 108.111524   105.41593   100.26946     92.333731 | 1.818345  1.787277  1.727137  1.632091 | 1.338861  1.339171  1.339838          1.341098 |  0.176099   0.163893   0.139466       0.09858 | 0.025297  0.040539  0.068132      0.113015 | 0.798583  0.800008  0.802443       0.80685 | 794.260928 | 778.418099  726.531837  633.508246          505.078941 |  433.22275 | 0.244463  0.240231  0.232032       0.219057 | 1.183013  1.151809  1.092106        0.999664 | 794.260928  764.095946   690.39577  577.882263   433.22275 |    5.0 | 789.260928 |  433.22275 |
    | 01:30 | 9.285301  9.094916  8.663443             7.963231 | 570.363548  550.011195  505.227821  436.513702 | 109.691858  107.750305  103.350125         96.20934 | 107.853006  105.949157  101.634425      94.63231 | 1.815377  1.793446  1.743197  1.659927 | 1.338889  1.339108  1.339651          1.340699 |  0.174945   0.166339   0.146094      0.110855 | 0.025308  0.036986  0.060306      0.099229 | 0.798383  0.799592   0.80166      0.805349 | 781.351729 |  773.34624  736.621109  657.418605          540.449923 | 473.672767 | 0.244059  0.241071  0.234222       0.222859 | 1.180022  1.157985  1.107958        1.026492 | 781.351729  766.835978  707.850388  608.312995  473.672767 |    5.0 | 776.351729 | 473.672767 |
    | 02:00 | 9.235354  9.122458  8.771813             8.161483 | 564.989181   552.93305  516.300679  455.471268 | 109.182506  108.031181  104.455288        98.231122 | 107.353545  106.224578  102.718127      96.61483 | 1.809638  1.796627   1.75589   1.68373 | 1.338946  1.339076  1.339508          1.340379 |  0.172706   0.167596   0.151279      0.121151 | 0.026404  0.034643  0.053977      0.087547 |  0.79824  0.799258  0.801015      0.804118 | 763.969595 | 763.605094   741.86598  676.791324          572.144955 | 511.011422 | 0.243277  0.241505  0.235953       0.226109 | 1.174243  1.161174  1.120534        1.049596 | 763.969595   764.69186   720.47978  634.469259  511.011422 |    5.0 | 758.969595 | 511.011422 |
    | 02:30 | 9.162754   9.12552  8.853309             8.329458 | 557.221597  553.258405  524.705042  471.841183 | 108.442121  108.062412  105.286391        99.944134 | 106.627538  106.255203  103.533091     98.294576 | 1.801277  1.796981  1.765404  1.703756 | 1.339029  1.339072  1.339403          1.340122 |  0.169427   0.167735   0.155133      0.129674 | 0.028535  0.033435  0.049025       0.07774 | 0.798149  0.798996  0.800486      0.803106 |   742.5984 | 749.580932  742.450587  691.588594          599.872173 |  544.81157 | 0.242138  0.241553  0.237249       0.228842 | 1.165839  1.161529  1.129986        1.069147 |   742.5984   757.95439  728.363435  656.198234   544.81157 |    5.0 |   737.5984 |  544.81157 |
    | 03:00 | 9.069003  9.105474  8.909254             8.468551 | 547.269076  551.130365  530.512878  485.610039 | 107.486042  107.857976  105.856923       101.362613 | 105.690026  106.054737  104.092543     99.685509 | 1.790449  1.794666  1.771918  1.720241 | 1.339139  1.339096  1.339333           1.33992 |  0.165152   0.166821   0.157757      0.136595 | 0.031655  0.033294   0.04534      0.069618 | 0.798107  0.798797  0.800056      0.802272 | 717.756821 | 731.707367  738.628988  701.860769          623.444923 | 574.759825 | 0.240663  0.241237  0.238137       0.231091 | 1.154984  1.159207  1.136473         1.08532 | 717.756821  746.974334  731.660807  673.443481  574.759825 |    5.0 | 712.756821 | 574.759825 |
    | 03:30 | 8.955649  9.063739  8.941023             8.580205 | 535.352997   546.71295   533.82476  496.802694 |  106.33006  107.432368  106.180897       102.501268 | 104.556492  105.637394  104.410226    100.802053 | 1.777311  1.789841  1.775612  1.733413 | 1.339276  1.339145  1.339294          1.339764 |  0.159919   0.164911   0.159239      0.142065 | 0.035721  0.034158  0.042828      0.063017 |  0.79811  0.798655  0.799711      0.801587 | 689.982427 | 710.450211  730.711818  707.735315          642.772995 | 600.648881 | 0.238872   0.24058  0.238641       0.232888 | 1.141851  1.154374  1.140155        1.098292 | 689.982427  732.147996  730.599008  686.235124  600.648881 |    5.0 | 684.982427 | 600.648881 |
    | 04:00 | 8.824269  9.001772  8.950023             8.665891 | 521.702626   540.18604  534.764868  505.476733 | 104.990237  106.800419   106.27268       103.375098 | 103.242687  105.017717  104.500227    101.658913 | 1.762017  1.782663  1.776657  1.743484 |  1.33944  1.339219  1.339283          1.339648 |  0.153764   0.162057   0.159657      0.146212 | 0.040693  0.035971  0.041405      0.057797 | 0.798158  0.798565  0.799439      0.801026 | 659.817086 | 686.293223  719.053028  709.405091          657.852536 | 622.368204 | 0.236788  0.239602  0.238783       0.234261 | 1.126619  1.147195  1.141198        1.108242 | 659.817086  713.903266  725.460187  694.678989  622.368204 |    5.0 | 654.817086 | 622.368204 |
    | 04:30 | 8.676447  8.921042  8.937683             8.727093 | 506.550351  531.740623  533.476178  511.717201 | 103.482743  105.977138  106.146845       103.999239 | 101.764468  104.210424  104.376834    102.270934 | 1.744722  1.773289  1.775224  1.750658 | 1.339633  1.339318  1.339298          1.339566 |   0.14672   0.158307   0.159083      0.149148 | 0.046534  0.038682  0.040997      0.053836 | 0.798249  0.798523  0.799232       0.80057 |  627.79402 | 659.725122  704.037138  707.116452          668.755176 | 639.893675 |  0.23443  0.238324  0.238588       0.235239 | 1.109467   1.13784  1.139768        1.115345 |  627.79402  692.686549  716.569074  698.945247  639.893675 |    5.0 |  622.79402 | 639.893675 |
    | 05:00 | 8.513765  8.823027   8.90544             8.765297 | 490.127419   521.57442   530.11595  515.631642 | 101.823705  104.977572  105.818029       104.388844 | 100.137647  103.230269  104.054404    102.652973 | 1.725581  1.761872  1.771474  1.755128 | 1.339856  1.339442  1.339338          1.339516 |   0.13882   0.153705   0.157578      0.150969 | 0.053208  0.042242  0.041535      0.051029 | 0.798383  0.798526  0.799083      0.800202 | 594.426711 | 631.228119  686.067344  701.157571          675.616829 | 653.276628 | 0.231819  0.236768  0.238077       0.235849 | 1.090575  1.126475  1.136031        1.119778 | 594.426711  668.950888  704.281049  699.256968  653.276628 |    5.0 | 589.426711 | 653.276628 |
    | 05:30 | 8.337786  8.709191  8.854724             8.781979 | 472.660213  509.887941  524.851556  517.345485 | 100.029072  103.816673  105.300822       104.558967 |  98.377865  102.091914  103.547241    102.819792 | 1.704745  1.748561  1.765568  1.757079 |  1.34011   1.33959  1.339402          1.339495 |  0.130092   0.148292   0.155199      0.151762 | 0.060682  0.046609  0.042959      0.049283 |  0.79856  0.798572  0.798986      0.799911 | 560.199767 | 601.268121  665.554747  691.847321          678.626545 | 662.632736 | 0.228977  0.234953  0.237272       0.236115 | 1.070116  1.113268  1.130151        1.121714 | 560.199767  643.145475  688.971046  695.878963  662.632736 |    5.0 | 555.199767 | 662.632736 |
    | 06:00 | 8.150048  8.580982   8.78695             8.778597 | 454.367097  496.881006  517.856687  516.997812 |  98.114504  102.509191  104.609658       104.524478 |  96.500476  100.809822  102.869499    102.785973 | 1.682362  1.733504  1.757659  1.756683 | 1.340396  1.339763  1.339488          1.339499 |  0.120565   0.142103   0.151998      0.151601 | 0.068922  0.051741  0.045213      0.048518 | 0.798781  0.798659  0.798936      0.799687 | 525.561767 | 570.286687  642.908854  679.524967          678.015734 | 668.131046 | 0.225922    0.2329  0.236194       0.236061 | 1.048265  1.098383   1.12229        1.121321 | 525.561767   615.70669  671.023501  689.107199  668.131046 |    5.0 | 520.561767 | 668.131046 |
    | 06:30 | 7.952045  8.439816  8.703509             8.756583 | 435.455805  482.749726  509.307977  514.737513 |  96.095269  101.069575  103.758724       104.299979 |  94.520453   99.398162   102.03509    102.565834 | 1.658578  1.716843  1.747895  1.754109 | 1.340718  1.339961  1.339597          1.339528 |  0.110266   0.135175   0.148019      0.150554 | 0.077898  0.057599  0.048248       0.04866 | 0.799047  0.798786  0.798931      0.799523 | 490.920022 |  538.69471  618.529448  664.540836          674.048004 | 669.983457 | 0.222675  0.230627  0.234863        0.23571 | 1.025188   1.08198  1.112609        1.118767 | 490.920022  587.050698  650.823484  679.258995  669.983457 |    5.0 | 485.920022 | 669.983457 |
    | 07:00 |  7.74523  8.287073  8.605761             8.717337 | 416.121383  467.683974  499.382038  510.719851 |  93.986157   99.511892  102.761888        103.89974 |  92.452299   97.870728  101.057611    102.173367 | 1.633534  1.698715   1.73642  1.749516 | 1.341076  1.340185  1.339729          1.339579 |  0.099222   0.127541   0.143307      0.148681 | 0.087579  0.064149  0.052018      0.049645 |  0.79936  0.798954  0.798967      0.799411 |  456.63713 | 506.867783  592.799865   647.24808          667.009782 | 668.434842 | 0.219254  0.228154  0.233298       0.235083 | 1.001049  1.064216  1.101261        1.114213 |  456.63713  557.567603  628.749088  666.664164  668.434842 |    5.0 |  451.63713 | 668.434842 |
    | 07:30 | 7.530998  8.124087  8.495028             8.662219 | 396.544626  451.865298  488.252914  505.103431 |  91.801412   97.849761  101.632626       103.337643 |   90.30998   96.240875   99.950279    101.622186 | 1.607366  1.679255  1.723369  1.743053 | 1.341474  1.340437  1.339882          1.339652 |  0.087459   0.119229     0.1379      0.146036 | 0.097936  0.071355   0.05648      0.051416 |  0.79972  0.799162  0.799043      0.799347 | 423.029154 | 475.143097  566.081638  627.995593          657.201854 | 663.753958 | 0.215678  0.225498  0.231518       0.234202 | 0.976004  1.045241  1.088397        1.107816 | 423.029154  527.617062   605.16508  651.657171  663.753958 |    5.0 | 418.029154 | 663.753958 |
    | 08:00 | 7.310687  7.952146  8.372588             8.592547 | 376.891005  435.465279  476.089936  498.047555 |  89.554669   96.096291  100.383976       102.627133 |  88.106868   94.521456   98.725877    100.925473 | 1.580207   1.65859  1.708877  1.734865 | 1.341915  1.340717  1.340058          1.339747 |  0.075001   0.110271   0.131833      0.142666 |  0.10894  0.079186  0.061596      0.053919 | 0.800131   0.79941  0.799157      0.799328 | 390.365238 | 443.817751  538.710442  607.122074          644.931877 |  656.22525 | 0.211964  0.222676  0.229541       0.233086 | 0.950202    1.0252  1.074164        1.099726 | 390.365238  497.525287  580.417807  634.570371   656.22525 |    5.0 | 385.365238 |  656.22525 |
    | 08:30 | 7.085569  7.772478  8.239671             8.509595 | 357.310005  418.644284  463.055965   489.70997 |  87.258913   94.264039   99.028487       101.781183 |  85.855693   92.724785   97.396711    100.095951 | 1.552185  1.636845  1.693068  1.725089 | 1.342402  1.341027  1.340257          1.339862 |  0.061876   0.100694   0.125141      0.138615 | 0.120561  0.087611  0.067328      0.057108 | 0.800592  0.799699  0.799309      0.799349 |  358.86844 | 413.148282  510.993245  584.951223          630.507902 | 646.141597 | 0.208129  0.219706  0.227383       0.231752 | 0.923785  1.004232    1.0587         1.09009 |  358.86844  467.583262  554.831271  615.728309  646.141597 |    5.0 |  353.86844 | 646.141597 |
    | 09:00 | 6.856852   7.58626  8.097458             8.414585 | 337.934851  401.550593  449.305992  480.244962 |  84.926441   92.364974   97.578192       100.812265 |  83.568517   90.862599   95.974579     99.145848 | 1.523421  1.614138  1.676063  1.713855 | 1.342938  1.341369   1.34048          1.339997 |  0.048108   0.090525   0.117855      0.133924 | 0.132772    0.0966  0.073642      0.060939 | 0.801108  0.800029  0.799498      0.799409 | 328.717566 | 383.351228  483.206548  561.787983          614.232891 | 633.798018 | 0.204191  0.216603  0.225062        0.23022 | 0.896888  0.982469  1.042139        1.079047 | 328.717566  438.046056  528.704295  595.443056  633.798018 |    5.0 | 323.717566 | 633.798018 |
    | 09:30 |  6.62567  7.394603  7.947074             8.308688 |  318.88256  384.319854  434.986065  469.801798 |  82.568841   90.410455   96.044574        99.732324 |  81.256699   88.946035   94.470743      98.08688 | 1.494034  1.590582  1.657979  1.701287 | 1.343528  1.341743  1.340726          1.340153 |  0.033723   0.079791   0.110004       0.12863 | 0.145542  0.106125  0.080507      0.065373 | 0.801679  0.800402  0.799724      0.799506 | 300.049792 | 354.604556  455.595563  537.915775          596.400201 | 619.486329 | 0.200164  0.213383  0.222593       0.228505 | 0.869637  0.960036  1.024609        1.066731 | 300.049792  409.133047  502.308667  574.010533  619.486329 |    5.0 | 295.049792 | 619.486329 |
    | 10:00 | 6.393089  7.198561  7.789591             8.193023 | 300.254276  367.074819  420.232515  458.523467 |  80.196972   88.411206   94.438554        98.552768 |  78.930891    86.98561   92.895911     96.930229 | 1.464134  1.566285  1.638923    1.6875 | 1.344176  1.342153  1.340997          1.340329 |  0.018749   0.068515   0.101616      0.122766 | 0.158843  0.116157  0.087893      0.070372 | 0.802307  0.800819  0.799986      0.799639 | 272.963873 | 327.049752  428.374211  513.594608          577.289966 | 603.490698 | 0.196064  0.210059   0.21999       0.226624 | 0.842151  0.937051   1.00623        1.053269 | 272.963873  381.028915  475.888148  551.707724  603.490698 |    5.0 | 267.963873 | 603.490698 |
    | 10:30 | 6.160102  6.999122  7.626023             8.068654 | 282.135824  349.925341   405.17145  446.545711 |  77.820962   86.377315   92.770477         97.28445 |  76.601021   84.991216   91.260227     95.686541 | 1.433828  1.541348  1.619002  1.672608 | 1.344886    1.3426  1.341294          1.340526 |  0.003212   0.056725   0.092718      0.116363 | 0.172646  0.126669   0.09577      0.075904 | 0.802995   0.80128  0.800285      0.799806 | 247.523746 |  300.79442  401.725789  489.059962          557.166324 | 586.084066 | 0.191904  0.206646  0.217268       0.224591 | 0.814541  0.913626  0.987118        1.038783 | 247.523746  353.885236  449.658219  528.790716  586.084066 |    5.0 | 242.523746 | 586.084066 |
    | 11:00 |  5.92763  6.797211  7.457326             7.936592 |  264.59846  332.968572  389.918474  433.996299 |  75.450205   84.318225   91.050105         95.93767 |  74.276303   82.972112   89.573263     94.365915 | 1.403217  1.515871  1.598312  1.656714 | 1.345663  1.343086  1.341618          1.340743 |  -0.01286   0.044443   0.083335      0.109451 |  0.18692  0.137635  0.104112      0.081936 | 0.803743  0.801787  0.800621      0.800007 | 223.762374 | 275.915225  375.804179  464.522355          536.275388 | 567.525343 | 0.187699  0.203157   0.21444        0.22242 | 0.786908  0.889864  0.967381        1.023386 | 223.762374  327.822542   423.80645  505.493455  567.525343 |    5.0 | 218.762374 | 567.525343 |
    | 11:30 | 5.696524  6.593693  7.284402              7.79779 | 247.699755  316.289338  374.578613  420.994536 |  73.093369   82.242739   89.286618        94.522171 |  71.965235   80.936931   87.844022     92.977904 | 1.372396  1.489943  1.576949  1.639918 | 1.346514  1.343614   1.34197          1.340983 | -0.029437   0.031695    0.07349      0.102057 | 0.201636  0.149028  0.112892      0.088439 | 0.804555   0.80234  0.800995      0.800242 |  201.68568 | 252.461047  350.735476  440.167457          514.843892 | 548.057327 |  0.18346  0.199603  0.211518       0.220126 | 0.759349  0.865862   0.94712        1.007188 |  201.68568  302.932715  398.493371  482.027133  548.057327 |    5.0 |  196.68568 | 548.057327 |
    | 12:00 | 5.467562  6.389368  7.108094             7.653152 | 231.484577  299.960652  359.246402  407.650951 |  70.758407   80.159026   87.488619        93.047142 |  69.675617   78.893682   86.080939     91.531519 | 1.341456  1.463652  1.555001  1.622315 | 1.347442  1.344187  1.342352          1.341244 | -0.046491   0.018505   0.063208      0.094207 | 0.216762  0.160824  0.122085      0.095385 | 0.805431  0.802941  0.801407       0.80051 | 181.276442 |  230.45623  326.619915  416.156685          493.078413 | 527.905244 |   0.1792  0.195998  0.208515       0.217721 | 0.731949  0.841711  0.926431         0.99029 | 181.276442  279.281587  373.853743  458.580097  527.905244 |    5.0 | 176.276442 | 527.905244 |
    | 12:30 | 5.241456  6.184977  6.929189             7.503523 |  215.98615  284.044335  344.006122  394.067168 |  68.452574   78.074634   85.664138        91.521225 |  67.414561   76.849767   84.291888     90.035234 | 1.310482  1.437081  1.532551  1.603993 | 1.348456  1.344807  1.342764          1.341527 | -0.063988   0.004897   0.052511      0.085925 | 0.232267  0.172996  0.131666      0.102749 | 0.806373  0.803591  0.801857      0.800811 | 162.498082 | 209.903819  303.534012  392.628154          471.165098 | 507.275852 | 0.174931  0.192351  0.205441       0.215216 | 0.704789  0.817492  0.905401        0.972789 | 162.498082  256.911663  349.998122  435.318209  507.275852 |    5.0 | 157.498082 | 507.275852 |
    | 13:00 | 5.018852  5.981199   6.74842               7.3497 | 201.227132  268.591711  328.932138  380.335917 |  66.182445   75.996503   83.820648        89.952524 |  65.188516   74.811993   82.484197     88.496996 | 1.279553  1.410305  1.509678  1.585035 |  1.34956  1.345478  1.343209          1.341835 | -0.081898  -0.009104   0.041421      0.077235 | 0.248118  0.185519  0.141613      0.110505 | 0.807382  0.804289  0.802346      0.801146 | 145.298231 | 190.788719  281.532817  369.697921          449.269798 | 486.357006 | 0.170663  0.188673  0.202308       0.212624 |  0.67794  0.793283  0.884114        0.954774 | 145.298231  235.844864  327.014639  412.385555  486.357006 |    5.0 | 140.298231 | 486.357006 |
    | 13:30 | 4.800329  5.778659  6.566466             7.192424 | 187.220724  253.644365  314.089342  366.541154 |  63.953942   73.930987   81.965073        88.348618 |   63.00329   72.786587   80.664657     86.924238 | 1.248744  1.383396  1.486455  1.565521 | 1.350762  1.346202  1.343688          1.342166 | -0.100187  -0.023475   0.029961      0.068157 | 0.264282  0.198368  0.151902      0.118629 | 0.808459  0.805038  0.802875      0.801514 | 129.612057 | 173.080693  260.652219  347.461425          427.538546 | 465.317641 | 0.166405  0.184973  0.199125       0.209954 | 0.651468  0.769154  0.862647        0.936331 | 129.612057  216.085231  304.970906  389.905439  465.317641 |    5.0 | 124.612057 | 465.317641 |
    | 14:00 | 4.586407  5.577922  6.383954             7.032389 | 173.971746  239.234925  299.533649  352.758296 |  61.772357    71.88387   80.103812        86.716576 |  60.864069   70.779224   78.839539     85.323888 | 1.218125  1.356421  1.462952  1.545524 | 1.352069  1.346983  1.344203          1.342523 | -0.118818   -0.03819    0.01815      0.058714 | 0.280726  0.211518  0.162511        0.1271 | 0.809604  0.805838  0.803444      0.801915 | 115.365288 |  156.73718  240.911229  325.995074          406.098305 | 444.308066 | 0.162168  0.181261  0.195902       0.207218 | 0.625432  0.745168   0.84107        0.917536 | 115.365288  197.621536  283.915998  367.981576  444.308066 |    5.0 | 110.365288 | 444.308066 |
    | 14:30 | 4.377545  5.379504  6.201462             6.870239 | 161.477696  225.387862   285.31255  339.054521 |  59.642378    69.86039   78.242747        85.062967 |  58.775453   68.795038   77.014616     83.702392 | 1.187762  1.329444  1.439234  1.525113 | 1.353487  1.347825  1.344755          1.342906 | -0.137756  -0.053225   0.006011      0.048926 | 0.297414  0.224944  0.173419      0.135896 | 0.810819   0.80669  0.804053       0.80235 |  102.47692 | 141.705872  222.314206   305.35791           385.05792 | 423.460541 |  0.15796  0.177545  0.192646       0.204423 | 0.599886  0.721384  0.819448        0.898464 |  102.47692  180.429743  263.882445   346.69944  423.460541 |    5.0 |   97.47692 | 423.460541 |
    | 15:00 | 4.174148  5.183865  6.019517             6.706572 | 149.729758  212.120283  271.465697  325.489136 |  57.568121   67.865262   76.387272        83.393885 |  56.741476   66.838654   75.195172     82.065722 | 1.157713  1.302523  1.415362  1.504355 | 1.355023  1.348731  1.345347          1.343316 | -0.156962  -0.068555  -0.006437      0.038813 | 0.314311  0.238621  0.184604      0.144996 | 0.812103  0.807592  0.804704      0.802819 |  90.861607 | 127.927047  204.852985  285.593302          364.509222 | 402.890046 |  0.15379  0.173833  0.189368       0.201579 | 0.574876  0.697853  0.797841        0.879181 |  90.861607  164.475313  244.888186  326.127704  402.890046 |    5.0 |  85.861607 | 402.890046 |
    | 15:30 | 3.976564  4.991421  5.838603              6.54194 | 138.713758   199.44272  258.025502  312.113988 |  55.553153   65.902704   74.542306        81.714959 |  54.765639   64.914208   73.386034     80.419399 | 1.128037  1.275711  1.391391  1.483309 | 1.356686  1.349704  1.345981          1.343755 | -0.176396  -0.084155  -0.019173      0.028392 | 0.331381  0.252525  0.196045       0.15438 | 0.813456  0.808547  0.805395      0.803322 |  80.431719 | 115.335657  188.508865  266.730635          344.528237 | 382.695222 | 0.149664  0.170132  0.186073       0.198694 | 0.550444  0.674623  0.776303         0.85975 |  80.431719  149.715304  226.938471   306.31973  382.695222 |    5.0 |  75.431719 | 382.695222 |
    | 16:00 | 3.785093  4.802537  5.659158             6.376851 | 128.411057  187.359878  245.017745  298.973911 |  53.600529   63.976462   72.712318        80.031376 |  52.850933   63.025372   71.591584      78.76851 | 1.098785  1.249058  1.367375  1.462033 | 1.358481  1.350749  1.346659          1.344223 | -0.196017  -0.099998  -0.032177      0.017684 | 0.348585  0.266629  0.207723       0.16403 | 0.814878  0.809554  0.806128      0.803859 |  71.099094 | 103.863141  173.254465  248.786954          325.176452 |  362.95941 |  0.14559  0.166449  0.182769       0.195775 | 0.526627  0.651736  0.754884        0.840229 |  71.099094  136.100284  210.027652  287.315065   362.95941 |    5.0 |  66.099094 |  362.95941 |
    | 16:30 | 3.599988  4.617538  5.481578             6.211773 | 118.799369  175.871363  232.462174  286.107188 |  51.712814   62.089834   70.901349          78.3479 |  50.999876   61.175381   69.815782     77.117727 | 1.070004  1.222611  1.343363   1.44058 | 1.360415   1.35187  1.347383          1.344723 | -0.215782   -0.11606  -0.045428      0.006706 | 0.365886  0.280911  0.219616      0.173925 | 0.816368  0.810613  0.806902       0.80443 |  62.776495 |  93.439005  159.055408  231.768548          306.502114 | 343.751774 | 0.141575  0.162789  0.179463       0.192831 | 0.503457  0.629229  0.733629        0.820671 |  62.776495  123.576022  194.140876  269.140921  343.751774 |    5.0 |  57.776495 | 343.751774 |
    | 17:00 | 3.421453  4.436706  5.306219             6.047132 | 109.853515  164.972362  220.373096  273.546029 |  49.892115   60.245696    69.11303        76.668891 |  49.214533   59.367055   68.062192     75.471323 |  1.04174  1.196411  1.319398     1.419 | 1.362494   1.35307  1.348155          1.345255 | -0.235645  -0.132314  -0.058907     -0.004525 | 0.383244  0.295344  0.231705      0.184049 | 0.817926  0.811723  0.807717      0.805036 |  55.378786 |  83.992153  145.871851  215.672437          288.541519 | 325.128449 | 0.137625   0.15916  0.176161       0.189867 | 0.480963  0.607135  0.712579        0.801123 |  55.378786  112.084993  179.255644  251.813612  325.128449 |    5.0 |  50.378786 | 325.128449 |
    | 17:30 | 3.249655  4.260283    5.1334              5.88332 | 101.546103  154.654286  208.759949  261.317049 |  48.140106    58.44653   67.350609        74.998324 |  47.496548   57.602827   66.333996     73.833197 | 1.014033  1.170498  1.295524  1.397338 | 1.364723  1.354353  1.348978           1.34582 |  -0.25556  -0.148733  -0.072595     -0.015991 |  0.40062  0.309903   0.24397      0.194384 |  0.81955  0.812885  0.808574      0.805677 |  48.823868 |     75.452  133.659839  200.487777          271.320281 | 307.133714 | 0.133746  0.155565  0.172867        0.18689 | 0.459169  0.585483  0.691769        0.781631 |  48.823868  101.567672  165.343234  235.339924  307.133714 |    5.0 |  43.823868 | 307.133714 |
    | 18:00 | 3.084717  4.088477  4.963402             5.720689 |   93.84813  144.905359  197.627842  249.441742 |   46.45806   56.694445    65.61697        73.339809 |  45.847166   55.884767   64.634022     72.206889 |  0.98692  1.144907  1.271779  1.375638 | 1.367107  1.355723  1.349854          1.346421 |  -0.27548  -0.165291  -0.086471     -0.027675 | 0.417972  0.324565  0.256393      0.204912 |  0.82124  0.814097  0.809472      0.806352 |  43.033385 |  67.749376  122.372508  186.197146          254.854554 | 289.801148 | 0.129943   0.15201  0.169589       0.183906 | 0.438095  0.564299  0.671233        0.762235 |  43.033385    91.96364  152.369995  219.718413  289.801148 |    5.0 |  38.033385 | 289.801148 |
    | 18:30 | 2.926726   3.92146  4.796476              5.55956 |  86.729523   135.71117  186.978077  237.936942 |  44.846868   54.991207   63.914654        71.696611 |  44.267262   54.214605   62.964764     70.595601 | 0.960437   1.11967  1.248197  1.353938 | 1.369648  1.357183  1.350785          1.347058 | -0.295353  -0.181961  -0.100516     -0.039561 |  0.43526  0.339304  0.268954      0.215617 | 0.822994  0.815359  0.810412      0.807061 |  37.933235 |  60.817251   111.96112  172.777721          239.152199 | 273.154754 | 0.126221  0.148499   0.16633       0.180919 | 0.417759  0.543604     0.651         0.74297 |  37.933235   83.212525  140.298498  204.940608  273.154754 |    5.0 |  32.933235 | 273.154754 |
    | 19:00 | 2.775736  3.759376   4.63284             5.400221 |  80.159604   127.05516   176.80863  226.815272 |  43.307067   53.338259   62.245883        70.071668 |  42.757363   52.593756     61.3284     69.002213 | 0.934614  1.094816  1.224812  1.332276 | 1.372348  1.358737  1.351773          1.347734 |  -0.31513  -0.198715   -0.11471     -0.051632 | 0.452442  0.354095  0.281634      0.226483 |  0.82481  0.816671  0.811391      0.807805 |  33.453911 |  54.591284  102.375962  160.202344          224.213881 | 257.210045 | 0.122586  0.145037  0.163094       0.177936 | 0.398177  0.523416  0.631094        0.723871 |  33.453911    75.25477  129.088552  190.992133  257.210045 |    5.0 |  28.453911 | 257.210045 |
    | 19:30 | 2.631767  3.602334  4.472681             5.242931 |  74.107496  118.919081   167.11461  216.085577 |  41.838863   51.736747   60.612579        68.467613 |  41.317671   51.023344   59.726813     67.429308 |  0.90948  1.070372  1.201652  1.310686 | 1.375208  1.360389  1.352822          1.348449 | -0.334757  -0.215526  -0.129035     -0.063872 | 0.469476  0.368915  0.294416      0.237493 | 0.826686   0.81803  0.812411      0.808583 |  29.530701 |   49.01024     93.5671  148.440479          210.034092 | 241.975074 | 0.119041  0.141626  0.159886       0.174959 | 0.379358  0.503751  0.611538        0.704966 |  29.530701   68.032259  118.698099  177.853719  241.975074 |    5.0 |  24.530701 | 241.975074 |
    | 20:00 | 2.494808  3.450422  4.316162             5.087918 |  68.542469  111.283392  157.888674  205.753329 |  40.442153   50.187539   59.016384        66.886788 |  39.948084    49.50422   58.161615     65.879182 | 0.885061  1.046361  1.178745  1.289197 | 1.378223  1.362141  1.353934          1.349205 | -0.354183  -0.232366  -0.143471     -0.076266 | 0.486319  0.383737  0.307281      0.248633 | 0.828621  0.819435  0.813471      0.809395 |  26.103761 |  44.016266   85.485008  137.459055          196.602096 | 227.451397 | 0.115592  0.138271  0.156709       0.171994 | 0.361314  0.484623  0.592352        0.686282 |  26.103761   61.488813  109.083982  165.502138  227.451397 |    5.0 |  21.103761 | 227.451397 |
    | 20:30 | 2.364822  3.303699  4.163416             4.935387 |  63.434229  104.127619  149.121413  195.821013 |  39.116543   48.691251    57.45868        65.331267 |  38.648216    48.03699   56.634162     64.353867 | 0.861381  1.022805  1.156114   1.26784 | 1.381391  1.363996   1.35511          1.350005 | -0.373352  -0.249207  -0.157998     -0.088799 | 0.502931  0.398537  0.320211      0.259887 | 0.830612  0.820887  0.814569       0.81024 |  23.118088 |  39.555076    78.08109  127.223206          183.902793 | 213.634985 | 0.112241  0.134975  0.153567       0.169045 | 0.344049  0.466041  0.573552        0.667842 |  23.118088    55.57056  100.202603  153.911023  213.634985 |    5.0 |  18.118088 | 213.634985 |
    | 21:00 | 2.240984  3.162427  4.014539             4.785515 |  58.724794    97.44114  140.800677  186.288504 |  37.853639   47.250557   55.940421         63.80287 |  37.409837   46.624273   55.145386     62.855152 | 0.838319   0.99976   1.13378  1.246639 |   1.3838  1.365952  1.356353          1.350849 | -0.391835  -0.265993  -0.172599     -0.101453 | 0.519379  0.413299  0.333188       0.27124 | 0.832834  0.822397  0.815705      0.811118 |  20.523427 |   35.57603   71.318554  117.695756           171.91752 | 200.517046 | 0.109046  0.131745  0.150463       0.166114 | 0.327906  0.448043   0.55515        0.649669 |  20.523427   50.246316   92.008376  143.051638  200.517046 |    5.0 |  15.523427 | 200.517046 |
    | 21:30 | 2.125112  3.026166  3.869688             4.638455 |    54.4572   91.180914  132.917743  177.153134 |  36.671979   45.860964   54.463229        62.303143 |  36.251124   45.261665   53.696879     61.384548 | 0.816252  0.977167  1.111771  1.225619 | 1.387264  1.368018  1.357664          1.351738 | -0.410297  -0.282751  -0.187249     -0.114216 | 0.534981  0.427953  0.346198      0.282678 | 0.834764  0.823917  0.816881      0.812029 |  18.274094 |   32.04206   65.129983  108.844383          160.624428 | 188.084776 |  0.10591  0.128573  0.147399       0.163206 | 0.312231  0.430576  0.537167        0.631778 |  18.274094   45.425651   84.466172   132.89296  188.084776 |    5.0 |  13.274094 | 188.084776 |
    | 22:00 | 2.015534  2.895139  3.728886             4.494347 |  50.544897   85.336242  125.456243  168.410993 |  35.554495   44.524742   53.027325        60.833528 |   35.15534   43.951392    52.28886     59.943472 | 0.794905  0.955078  1.090099  1.204801 | 1.390861   1.37019  1.359047          1.352676 | -0.428406  -0.299425  -0.201934     -0.127071 | 0.550596  0.442543  0.359215      0.294188 | 0.836856  0.825495   0.81809      0.812972 |  16.328777 |  28.887435   59.482797  100.629187           150.00032 | 176.322044 | 0.102874  0.125467  0.144379       0.160322 | 0.297289  0.413675  0.519606        0.614188 |  16.328777    41.08104   77.528472  123.404356  176.322044 |    5.0 |  11.328777 | 176.322044 |
    | 22:30 | 1.912522  2.769335  3.592174             4.353307 |   46.97653   79.886126   118.40118   160.05604 |  34.503973   43.241791   51.633131        59.395199 |  34.125219   42.693355    50.92174     58.533074 |  0.77437  0.933507  1.068778  1.184206 | 1.394565   1.37247  1.360502          1.353662 | -0.446047  -0.315986  -0.216635     -0.140004 | 0.565903  0.457016  0.372228      0.305754 | 0.839046  0.827112  0.819335      0.813947 |  14.650285 |  26.084989   54.335823    93.01315          140.019755 | 165.210171 |  0.09995   0.12243  0.141404       0.157467 | 0.283131  0.397343  0.502475        0.596912 |  14.650285   37.174668   71.152738  114.553568  165.210171 |    5.0 |   9.650285 | 165.210171 |
    | 23:00 | 1.815936  2.648711  3.459588             4.215429 |  43.727163    74.80902  111.737571   152.08062 |  33.518988    42.01166   50.281015        57.989105 |  33.159362   41.487111   49.595881     57.154286 | 0.754664  0.912465  1.047821  1.163852 |  1.39836  1.374856  1.362031          1.354698 |  -0.46317  -0.332407  -0.231333     -0.153001 | 0.580785  0.471347  0.385221      0.317362 |  0.84128  0.828765  0.820613      0.814953 |  13.205287 |  23.598568   49.649298   85.960565          130.655867 | 154.728365 | 0.097142  0.119463  0.138475       0.154642 | 0.269753  0.381581   0.48578        0.579962 |  13.205287   33.666679   65.300084  106.307552  154.728365 |    5.0 |   8.205287 | 154.728365 |
    | 23:30 | 1.725604  2.533218  3.331148             4.080787 |   40.77261   70.084256   105.44995  144.475915 |  32.597778   40.833861   48.971176        56.616025 |  32.256041   40.332185   48.311479      55.80787 | 0.735797  0.891962  1.027241  1.143753 | 1.402225  1.377346  1.363636          1.355787 | -0.479728  -0.348658  -0.246012     -0.166046 | 0.595206  0.485515  0.398177      0.329001 | 0.843552  0.830454  0.821924      0.815988 |  11.964036 |  21.394824   45.386166   79.436508          121.881056 | 144.854137 | 0.094452  0.116567  0.135596       0.151849 | 0.257147  0.366389  0.469525        0.563348 |  11.964036   30.519981   59.933372   98.633424  144.854137 |    5.0 |   6.964036 | 144.854137 |
    | 00:00 | 1.641337  2.422797  3.206859             3.949443 |  38.089984   65.691685   99.522597   137.23217 |  31.738417   39.707779   47.703671        55.276578 |  31.413368   39.227971   47.068588     54.494435 | 0.717779  0.872006  1.007048  1.123924 | 1.406138  1.379939  1.365318          1.356929 |  -0.49568  -0.364711  -0.260651     -0.179127 | 0.609137  0.499496  0.411081      0.340657 | 0.845857  0.832176  0.823267      0.817053 |  10.900084 |  19.443467   41.511642    73.40711          113.667361 | 135.563746 | 0.091883  0.113745  0.132767       0.149091 | 0.245302  0.351767  0.453715        0.547078 |  10.900084   27.700336   55.017106   91.498826  135.563746 |    5.0 |   5.900084 | 135.563746 |
    | 00:30 | 1.562931   2.31661  3.086962             3.821423 |  35.657721   61.582547   93.951087  130.337688 |  30.938828   38.624876   46.480956        53.971016 |  30.629307   38.166097   45.869617     53.214225 | 0.700617  0.852463  0.987293  1.104374 | 1.410075  1.381745  1.367073          1.358126 | -0.510987  -0.380164  -0.275203     -0.192232 | 0.622551  0.513371  0.423926      0.352317 |  0.84819  0.834099  0.824656      0.818145 |    9.99001 |  17.717084   37.993174   67.850978          105.985451 | 126.832692 | 0.089436   0.11105  0.129995       0.146369 | 0.234203  0.338053  0.438382        0.531156 |    9.99001   25.176256   50.539146   84.869751  126.832692 |    5.0 |    4.99001 | 126.832692 |
    | 01:00 | 1.490169  2.216157   2.97089             3.696843 |  33.455568   57.799129   88.694308  123.785868 |  30.196807    37.60046   45.297257        52.700546 |  29.901695   37.161574   44.708905     51.968427 | 0.684316  0.833632  0.967893  1.085127 | 1.414012  1.384549  1.368911          1.359377 | -0.525616  -0.395752  -0.289716     -0.205339 | 0.635426  0.526472  0.436642      0.363973 | 0.850545  0.835684  0.826035      0.819269 |   9.213141 |  16.190941   34.800669   62.711626           98.812442 | 118.635188 | 0.087112  0.108377   0.12727       0.143686 | 0.223834  0.324542  0.423459        0.515598 |   9.213141   22.918875   46.419136   78.724515  118.635188 |    5.0 |   4.213141 | 118.635188 |
    | 01:30 | 1.422826  2.120802  2.859004             3.575649 |  31.464561   54.301045   83.754568  117.561079 |  29.510036   36.628024   44.156232        51.464611 |  29.228261   36.208023   43.590037     50.756493 | 0.668877  0.815421  0.948919  1.066181 | 1.417923  1.387436  1.370825          1.360686 | -0.539541  -0.411016  -0.304125     -0.218442 | 0.647741  0.539796  0.449304      0.375602 | 0.852915  0.837524  0.827471      0.820416 |   8.551298 |   14.84278   31.913723    57.97702           92.116346 | 110.947442 | 0.084911  0.105789  0.124601       0.141041 | 0.214177  0.311629  0.408994        0.500399 |   8.551298   20.901801   42.657689   73.025282  110.947442 |    5.0 |   3.551298 | 110.947442 |
    | 02:00 | 1.360664  2.029822  2.751118             3.457871 |   29.66699   51.048222   79.110003  111.652441 |  28.876104   35.700206   43.056004        50.263507 |  28.606639   35.298222   42.511176     49.578713 | 0.654299  0.797717  0.930352  1.047548 | 1.421784  1.390409  1.372817          1.362052 |  -0.55274  -0.426026  -0.318431     -0.231527 | 0.659483  0.552767  0.461837      0.387201 | 0.855291  0.839312  0.828921      0.821589 |   7.988541 |  13.652625   29.287795   53.612517           85.871367 | 103.744023 | 0.082835  0.103271  0.121985       0.138437 |  0.20521  0.299228  0.394969        0.485563 |   7.988541   19.100975   39.219454   67.744403  103.744023 |    5.0 |   2.988541 | 103.744023 |
    | 02:30 | 1.303439  1.943454  2.647245             3.343498 |  28.046351   48.036866     74.7482  106.047353 |  28.292522   34.819416   41.996707        49.097121 |   28.03439   34.434537   41.472449     48.434978 | 0.640578  0.780586  0.912207  1.029232 |  1.42557  1.393454  1.374886          1.363476 | -0.565199  -0.440703   -0.33261     -0.244581 | 0.670639  0.565517  0.474252      0.398757 | 0.857666  0.841167  0.830401      0.822786 |    7.51094 |  12.602587   26.909334   49.593801           80.050773 |  96.999881 | 0.080883  0.100833  0.119426       0.135875 |  0.19691  0.287379  0.381388        0.471092 |    7.51094   17.494513   36.082162   62.853613   96.999881 |    5.0 |    2.51094 |  96.999881 |
    | 03:00 | 1.250902  1.861594  2.547325             3.232521 |  26.587302   45.251582   70.654198  100.733764 |  27.756745   33.984612    40.97772        47.965373 |  27.509018   33.615943    40.47325     47.325208 | 0.627705  0.764037  0.894486   1.01124 | 1.429257  1.396562  1.377031           1.36496 | -0.576909  -0.455021  -0.346644     -0.257592 | 0.681204   0.57796  0.486524      0.410259 | 0.860029  0.843043  0.831902      0.824007 |   7.106358 |   11.67668   24.756416   45.895246           74.629354 |  90.690254 | 0.079053  0.098475  0.116924       0.133354 | 0.189251  0.276076   0.36825        0.456986 |   7.106358   16.062556    33.22114   58.326996   90.690254 |    5.0 |   2.106358 |  90.690254 |
    | 03:30 | 1.202799  1.784124  2.451299             3.124925 |  25.275609   42.677339   66.813826   95.699638 |   27.26619   33.194562   39.998445        46.868103 |  27.027989   32.841236   39.512992     46.249246 | 0.615669   0.74807  0.877192  0.993578 | 1.432826  1.399721  1.379249          1.366503 |  -0.58787  -0.468957  -0.360518     -0.270549 | 0.691173  0.590077  0.498642      0.421698 | 0.862371  0.844935  0.833425       0.82525 |   6.764251 |  10.860648    22.80855   42.493066           69.582816 |   84.79101 | 0.077344    0.0962  0.114479       0.130877 | 0.182205  0.265311   0.35555        0.443246 |   6.764251   14.787118   30.613325   54.140007    84.79101 |    5.0 |   1.764251 |   84.79101 |
    | 04:00 | 1.158875  1.710916  2.359103             3.020686 |  24.098087   40.299907   63.213384   90.933013 |  26.818255   32.447988   39.058226        45.805075 |  26.588752    32.10916   38.591031     45.206861 | 0.604456  0.732686  0.860327   0.97625 | 1.436257  1.402921  1.381539          1.368105 | -0.598085   -0.48249  -0.374213     -0.283437 | 0.700549  0.601851  0.510592      0.433062 | 0.864681  0.846841  0.834968      0.826515 |    6.47548 |  10.141796   21.046937   39.364941           64.887837 |  79.278807 | 0.075754  0.094006  0.112092       0.128444 | 0.175744  0.255075  0.343286        0.429871 |    6.47548   13.651941   28.237351   50.269294   79.278807 |    5.0 |    1.47548 |  79.278807 |
    | 04:30 | 1.118876  1.641842  2.269943             2.920026 |  23.042545   38.105868   59.812351   86.433139 |  26.410336   31.743573   38.148967        44.778537 |  26.188755   31.418424    37.69943     44.200258 | 0.594045  0.717889  0.843758  0.959302 | 1.439534   1.40615  1.383043          1.369763 | -0.607567  -0.495599  -0.387358     -0.296214 | 0.709335  0.613271  0.522464      0.444352 | 0.866947  0.848758  0.836695      0.827818 |   6.232151 |   9.508843   19.454377   36.489959           60.532727 |   74.12852 |  0.07428  0.091896  0.109813       0.126061 | 0.169837   0.24536  0.331776        0.416893 |   6.232151   12.642348   26.073473   46.713316    74.12852 |    5.0 |   1.232151 |   74.12852 |
    | 05:00 | 1.082546  1.576771  2.185596             2.822405 |  22.097732   36.082585   56.668106   82.165918 |  26.039851   31.079968   37.288798        43.782995 |  25.825465   30.767706   36.855965     43.224048 | 0.584414  0.703676  0.827833  0.942652 | 1.442646  1.409394  1.385475          1.371484 | -0.616331  -0.508267  -0.400607     -0.308932 | 0.717539  0.624322   0.53364      0.455499 | 0.869159  0.850682  0.838119      0.829101 |   6.027458 |    8.95177    18.01516   33.859645           56.474475 |  69.327385 | 0.072918   0.08987  0.107551       0.123718 | 0.164452  0.236155  0.320401        0.404245 |   6.027458   11.745114    24.10348    43.40911   69.327385 |    5.0 |   1.027458 |  69.327385 |
    | 05:30 | 1.049638  1.515566  2.104423             2.728025 |  21.253277   34.218183    53.70933   78.130984 |  25.704252   30.455798   36.460989        42.820507 |  25.496383   30.155657   36.044231     42.280252 | 0.575538  0.690049  0.812259  0.926342 | 1.445581  1.412641  1.387978          1.373264 |   -0.6244  -0.520478  -0.413698     -0.321549 | 0.725173  0.634995  0.544945      0.466578 | 0.871306  0.852611  0.839656      0.830422 |   5.855551 |   8.461695   16.714953   31.431245           52.703636 |  64.848953 | 0.071664  0.087927  0.105338        0.12142 | 0.159559  0.227449  0.309392        0.391956 |   5.855551   10.948333   22.310608   40.353949   64.848953 |    5.0 |   0.855551 |  64.848953 |
    | 06:00 | 1.019906  1.458091  2.026771             2.636855 |  20.499638   32.501511   50.940554   74.317867 |  25.401044   29.869669   35.669085        41.890755 |  25.199063    29.58091   35.267706     41.368555 | 0.567386  0.677006  0.797117  0.910377 | 1.448333  1.415876  1.390536          1.375102 | -0.631799  -0.532217  -0.426549     -0.334051 | 0.732251  0.645281  0.556119      0.477546 | 0.873378  0.854539  0.841261      0.831761 |   5.711411 |   8.030751   15.540702   29.201374           49.201634 |  60.671164 | 0.070515  0.086068  0.103184       0.119168 | 0.155126  0.219231  0.298801        0.380026 |   5.711411     10.2413   20.679446   37.534748   60.671164 |    5.0 |   0.711411 |  60.671164 |
    | 06:30 | 0.993112  1.404208   1.95257             2.548835 |  19.828047   30.922117   48.351187   70.715329 |  25.127799   29.320167   34.912382         40.99312 |  24.931124   29.042079   34.525697     40.488351 | 0.559926  0.664545   0.78241  0.894756 |   1.4509  1.419088  1.393146          1.376997 | -0.638558  -0.543473  -0.439145     -0.346429 |  0.73879  0.655171  0.567069       0.48839 | 0.875365  0.856465  0.842878      0.833114 |   5.590745 |   7.651972   14.480526   27.154704            45.94993 |  56.775401 | 0.069465  0.084292  0.101091       0.116962 |  0.15112  0.211487  0.288623        0.368449 |   5.590745    9.614397   19.195852   34.934134   56.775401 |    5.0 |   0.590745 |  56.775401 |
    | 07:00 | 0.969027  1.353777  1.881733             2.463913 |  19.230461   29.470209   45.930595   67.313044 |  24.882171   28.805868   34.189988        40.127084 |  24.690266   28.537768   33.817331     39.639133 | 0.553125  0.652663  0.768138  0.879479 |  1.45328  1.422261    1.3958          1.378949 |  -0.64471  -0.554238  -0.451472     -0.358673 |  0.74481   0.66466   0.57778      0.499103 | 0.877261  0.858383  0.844505      0.834483 |   5.489885 |     7.3192   13.523626   25.276568           42.931545 |  53.144147 | 0.068509    0.0826  0.099058       0.114802 | 0.147511  0.204204  0.278854        0.357223 |   5.489885    9.058991   17.846865   32.535686   53.144147 |    5.0 |   0.489885 |  53.144147 |
    | 07:30 | 0.947428  1.306657  1.814172             2.382037 |  18.699516    28.13662   43.668669   64.101043 |  24.661906    28.32534   33.500993        39.292103 |  24.474279   28.066571   33.141717     38.820367 | 0.546945  0.641357  0.754299  0.864547 | 1.455475  1.425383  1.398493          1.380954 | -0.650288  -0.564503  -0.463517     -0.370771 | 0.750331  0.673745  0.588244      0.509677 | 0.879057  0.860289   0.84614      0.835864 |   5.405708 |   7.026987   12.660192   23.553386           40.130492 |  49.760685 | 0.067641  0.080992  0.097086       0.112689 | 0.144268  0.197369  0.269485        0.346344 |   5.405708    8.567336    16.62062   30.324098   49.760685 |    5.0 |   0.405708 |  49.760685 |
    | 08:00 | 0.928105  1.262708  1.749795             2.302422 |  18.228476   26.912778   41.555841   61.042068 |  24.464854   27.877146   32.844479         38.48019 |  24.281053    27.62708   32.497951     38.024221 | 0.541349  0.630622  0.740893  0.849825 | 1.457489  1.428442  1.401218          1.382166 |  -0.65533  -0.574267  -0.475269     -0.382357 | 0.755379  0.682422  0.598451      0.520205 |  0.88075  0.862179   0.84778      0.837421 |    5.33556 |   6.770521   11.881322    21.97262           37.531718 |  46.630352 | 0.066857  0.079466  0.095175       0.110673 | 0.141362  0.190967   0.26051        0.336131 |    5.33556    8.132488   15.506265   28.285135   46.630352 |    5.0 |    0.33556 |  46.630352 |
    | 08:30 | 0.910859  1.221789  1.688513             2.226506 |  17.811195   25.790664   39.583068    58.18425 |  24.288971   27.459847   32.219519        37.705999 |  24.108586   27.217885   31.885128     37.265064 | 0.536298   0.62045  0.727918  0.835589 | 1.459329  1.431425  1.403969          1.384288 | -0.659871  -0.583527  -0.486716     -0.394144 | 0.759976  0.690692  0.608395      0.530053 | 0.882336  0.864047  0.849423      0.838635 |   5.277186 |   6.545549   11.178942   20.522705           35.121382 |  43.690879 | 0.066149  0.078021  0.093325       0.108652 | 0.138763  0.184983  0.251922         0.32592 |   5.277186    7.748222   14.493889   26.405579   43.690879 |    5.0 |   0.277186 |  43.690879 |
    | 09:00 | 0.895499  1.183758  1.630234             2.153745 |  17.442075   24.762785   37.741815   55.499279 |  24.132333    27.07201   31.625187         36.96398 |  23.954989    26.83758   31.302337     36.537455 | 0.531754  0.610835  0.715373  0.821751 | 1.461002  1.434323  1.406737          1.386453 | -0.663949  -0.592285  -0.497847     -0.405711 | 0.764149  0.698554  0.618068      0.540192 | 0.883814  0.865889  0.851068      0.840081 |   5.228681 |   6.348314   10.545732   19.192985           32.894453 |   40.95781 | 0.065514  0.076657  0.091536       0.106686 | 0.136446    0.1794  0.243714        0.316081 |   5.228681    7.408962    13.57444   24.673166    40.95781 |    5.0 |   0.228681 |   40.95781 |
    | 09:30 | 0.881849  1.148477  1.574867             2.083432 |  17.116024   23.822139   36.024033   52.954907 |  23.993131   26.712208   31.060554        36.246916 |   23.81849   26.484765   30.748669     35.834315 | 0.527679  0.601768  0.703256  0.808189 | 1.462516  1.437124  1.409515          1.388669 |   -0.6676  -0.600544  -0.508653     -0.417147 | 0.767924  0.706012  0.627463      0.550077 | 0.885182    0.8677  0.852713      0.841458 |   5.188431 |   6.175499    9.975061   17.973649           30.819159 |  38.410115 | 0.064944  0.075372  0.089808       0.104758 | 0.134384  0.174203  0.235878        0.306528 |   5.188431    7.109716    12.73966   23.076536   38.410115 |    5.0 |   0.188431 |  38.410115 |
    | 10:00 | 0.869743  1.115805  1.522322             2.015853 |  16.828421   22.962185   34.422139   50.556108 |  23.869678   26.379025   30.524695        35.557746 |  23.697435   26.158052   30.223216     35.158529 | 0.524035  0.593238  0.691566  0.794968 | 1.463881  1.439821  1.412295          1.390926 | -0.670859  -0.608312  -0.519126     -0.428391 | 0.771329  0.713069  0.636577      0.559856 | 0.886443  0.869475  0.854356      0.842887 |   5.155078 |   6.024174    9.460922   16.855674           28.894774 |   36.03947 | 0.064436  0.074164  0.088142       0.102877 | 0.132553  0.169376  0.228405          0.2973 |   5.155078    6.846015   11.982018   21.605174    36.03947 |    5.0 |   0.155078 |   36.03947 |
    | 10:30 | 0.859029  1.085607  1.472507             1.950957 |  16.575082   22.176814   32.928995   48.295505 |  23.760408   26.071061   30.016686        34.895931 |  23.590287   25.856068   29.725072     34.509565 | 0.520785  0.585232  0.680301  0.782088 | 1.465108  1.442406  1.415069           1.39322 | -0.673761  -0.615597  -0.529257      -0.43943 | 0.774389  0.719732  0.645403      0.569453 | 0.887599  0.871208  0.855994      0.844324 |   5.127477 |   5.891751    8.997875    15.83076           27.110806 |  33.833918 | 0.063982  0.073032  0.086536       0.101043 | 0.130931    0.1649  0.221289        0.288395 |   5.127477    6.613861   11.294648   20.249358   33.833918 |    5.0 |   0.127477 |  33.833918 |
    | 11:00 | 0.849563  1.057746  1.425334             1.888675 |  16.352226   21.460319   31.537884   46.165598 |  23.663874   25.786933   29.535608        34.260783 |  23.495628   25.577458   29.253336     33.886752 | 0.517894  0.577739  0.669458  0.769547 | 1.466206  1.444873  1.417829          1.395548 | -0.676338   -0.62241   -0.53904     -0.450258 | 0.777132  0.726006  0.653937      0.578859 | 0.888654  0.872894  0.857625      0.845764 |   5.104666 |   5.775946    8.580998   14.891286           25.457095 |  31.782086 |  0.06358  0.071974  0.084991       0.099257 | 0.129497   0.16076  0.214521        0.279807 |   5.104666     6.40968   10.671289   19.000107   31.782086 |    5.0 |   0.104666 |  31.782086 |
    | 11:30 | 0.841215  1.032089  1.380711             1.828941 |  16.156442   20.807374    30.24249   44.159245 |  23.578745   25.525284   29.080549        33.651613 |  23.412151    25.32089   28.807114     33.289411 |  0.51533  0.570742  0.659036  0.757344 | 1.467185  1.447219  1.420567          1.397906 | -0.678621  -0.628763  -0.548469     -0.460867 | 0.779583    0.7319  0.662178      0.588069 | 0.889612   0.87453  0.859248      0.847208 |   5.085838 |   5.674743    8.205834   14.030247           23.924177 |  29.873331 | 0.063223  0.070987  0.083506       0.097519 | 0.128231  0.156939  0.208092        0.271531 |   5.085838    6.230278   10.106234   17.849132   29.873331 |    5.0 |   0.085838 |  29.873331 |
    | 12:00 | 0.833866  1.008506  1.338552             1.771687 |  15.984659   20.213009   29.036874   42.269668 |  23.503801   25.284782   28.650601         33.06773 |  23.338663   25.085059   28.385515     32.716867 |  0.51306  0.564225  0.649032  0.745477 | 1.468056  1.449439  1.423274          1.400289 | -0.680639  -0.634672  -0.557541      -0.47125 | 0.781767  0.737424  0.670122      0.597078 | 0.890477   0.87611  0.860859      0.848653 |   5.070317 |   5.586362    7.868348   13.241214           22.503258 |  28.097714 | 0.062907  0.070069  0.082082       0.095827 | 0.127117  0.153418  0.201993        0.263561 |   5.070317    6.072805    9.594282   16.788785   28.097714 |    5.0 |   0.070317 |  28.097714 |
    | 12:30 | 0.827407  0.986869  1.298766             1.716845 |  15.834122   19.672585   27.915449   40.490445 |  23.437931   25.064128   28.244864        32.508448 |  23.274072   24.868689   27.987658     32.168446 | 0.511056  0.558172  0.639444  0.733943 | 1.468829  1.451533  1.425942          1.402693 | -0.682419  -0.640154  -0.566251     -0.481399 | 0.783708  0.742587  0.677767      0.605881 | 0.891257   0.87763  0.862456      0.850099 |   5.057539 |   5.509234    7.564892   12.518285           21.186174 |  26.445957 | 0.062628  0.069217  0.080718       0.094183 | 0.126136  0.150181  0.196216        0.255892 |   5.057539    5.934715    9.130687   15.812016   26.445957 |    5.0 |   0.057539 |  26.445957 |
    | 13:00 | 0.821739  0.967054  1.261267             1.664348 |  15.702359   19.181779   26.872967   38.815495 |  23.380126   24.862055   27.862446        31.973087 |   23.21739   24.670541   27.612666     31.643481 |  0.50929  0.552564  0.630266  0.722742 | 1.469513    1.4535  1.428565          1.405113 | -0.683987  -0.645224  -0.574598     -0.491309 | 0.785429  0.747401  0.685114      0.614475 | 0.891956  0.879088  0.864037      0.851543 |   5.047032 |   5.441973    7.292162   11.856042            19.96535 |  24.909409 | 0.062383  0.068429  0.079414       0.092586 | 0.125275  0.147211  0.190751        0.248518 |   5.047032    5.813744    8.711124   14.912329   24.909409 |    5.0 |   0.047032 |  24.909409 |
    | 13:30 | 0.816772  0.948941  1.225967             1.614131 |  15.587162    18.73656   25.904487   37.239069 |  23.329473   24.677337   27.502462        31.460974 |   23.16772    24.48941   27.259672     31.141313 | 0.507736   0.54738  0.621497   0.71187 | 1.470116   1.45534  1.431134          1.407545 | -0.685364  -0.649902  -0.582581     -0.500973 | 0.786952  0.751879  0.692162      0.622854 | 0.892582   0.88048  0.865599      0.852986 |   5.038402 |    5.38336     7.04717   11.249513           18.833763 |  23.480013 | 0.062167  0.067701  0.078168       0.091036 |  0.12452  0.144492   0.18559        0.241434 |   5.038402    5.707871    8.331642   14.083739   23.480013 |    5.0 |   0.038402 |  23.480013 |
    | 14:00 | 0.812426  0.932413  1.192782             1.566129 |   15.48656   18.333176   25.005365   35.755728 |  23.285148   24.508789   27.164034        30.971442 |  23.124256   24.324135   26.927816     30.661288 | 0.506372  0.542602  0.613131  0.701325 | 1.470647  1.457055  1.433643          1.409982 | -0.686572  -0.654208  -0.590201     -0.510387 | 0.788295  0.756034  0.698912      0.631017 | 0.893139  0.881804  0.867139      0.854425 |   5.031322 |   5.332319    6.827209   10.694135           17.784905 |  22.150264 | 0.061977  0.067031  0.076981       0.089532 |  0.12386  0.142007  0.180722        0.234632 |   5.031322    5.615302    7.988639   13.320734   22.150264 |    5.0 |   0.031322 |  22.150264 |
    | 14:30 | 0.808627   0.91736  1.161625             1.520276 |  15.398796   17.968135   24.171229   34.360337 |  23.246411   24.355268   26.846294        30.503835 |  23.086271   24.173595   26.616247     30.202761 | 0.505176  0.538208  0.605164  0.691107 | 1.471114  1.458649  1.436085          1.412421 |  -0.68763  -0.658161  -0.597458     -0.519544 | 0.789478  0.759878  0.705365       0.63896 | 0.893633  0.883058  0.868654      0.855859 |   5.025521 |   5.287904    6.629832   10.185719           16.812743 |  20.913182 | 0.061812  0.066416  0.075852       0.088075 | 0.123282   0.13974  0.176137        0.228109 |   5.025521     5.53444    7.678825   12.618241   20.913182 |    5.0 |   0.025521 |  20.913182 |
    | 15:00 | 0.805312  0.903672  1.132412              1.47651 |  15.322309   17.638192   23.397964   33.048046 |    23.2126   24.215681   26.548381        30.057502 |  23.053117   24.036718    26.32412     29.765096 |  0.50413  0.534178   0.59759  0.681213 | 1.471523  1.460124  1.438455          1.414856 | -0.688556  -0.661781  -0.604356     -0.528442 | 0.790519  0.763428  0.711524       0.64668 | 0.894072  0.884241  0.870142      0.857287 |   5.020773 |   5.249282    6.452824    9.720423           15.911694 |  19.762275 | 0.061666  0.065852  0.074779       0.086665 | 0.122778  0.137676  0.171827        0.221857 |   5.020773    5.463872    7.399193   11.971587   19.762275 |    5.0 |   0.020773 |  19.762275 |
    | 15:30 | 0.802421  0.891248   1.10506             1.434767 |  15.255716   17.340332   22.681693   31.814277 |  23.183122   24.088981   26.269447        29.631806 |  23.024212   23.912479   26.050602     29.347667 | 0.503216  0.530489  0.590403  0.671641 | 1.471882  1.461486  1.440747          1.417281 | -0.689364  -0.665088  -0.610899     -0.537075 | 0.791431  0.766696  0.717392      0.654176 | 0.894459  0.885354  0.871601      0.858707 |   5.016892 |   5.215721    6.294179    9.294715           15.076586 |  18.691512 |  0.06154  0.065336  0.073762       0.085301 | 0.122338  0.135801  0.167781        0.215872 |   5.016892    5.402341    7.146999   11.376468   18.691512 |    5.0 |   0.016892 |  18.691512 |
    | 16:00 | 0.799904   0.87999  1.079487             1.394986 |   15.19779   17.071761   22.018763   30.654707 |  23.157452   23.974172   26.008649        29.226117 |  22.999039     23.7999   25.794869     28.949855 | 0.502419  0.527121  0.583595  0.662388 | 1.472195  1.462739  1.442957          1.419691 | -0.690068  -0.668104  -0.617092      -0.54544 | 0.792231  0.769699  0.722972      0.661445 |   0.8948  0.886396  0.873026      0.860118 |   5.013722 |   5.186577    6.152082    8.905355           14.302632 |  17.695287 | 0.061429  0.064866    0.0728       0.083983 | 0.121955    0.1341  0.163988        0.210147 |   5.013722    5.348739    6.919733   10.828923   17.695287 |    5.0 |   0.013722 |  17.695287 |
    | 16:30 | 0.797714  0.869806  1.055611             1.357105 |  15.147449   16.829891   21.405732   29.565258 |  23.135119   23.870311    25.76516        28.839814 |   22.97714   23.698055   25.556108     28.571055 | 0.501724  0.524053   0.57716  0.653454 | 1.472468  1.463887  1.445081          1.422082 | -0.690682  -0.670847  -0.622941     -0.553535 |  0.79293  0.772451   0.72827      0.668487 |   0.8951  0.887368  0.874416      0.861518 |   5.011136 |   5.161287    6.024891    8.549367           13.585402 |  16.768399 | 0.061333  0.064438  0.071891       0.082711 | 0.121622   0.13256  0.160439        0.204676 |   5.011136    5.302083    6.715103     10.3253   16.768399 |    5.0 |   0.011136 |  16.768399 |
    | 17:00 | 0.795811  0.860607  1.033352             1.321066 |  15.103735   16.612327   20.839351   28.542081 |  23.115709   23.776504    25.53816        28.472288 |  22.958108    23.60607   25.333516     28.210665 | 0.501119  0.521265  0.571088  0.644837 | 1.472707  1.464938  1.447117          1.424446 | -0.691216  -0.673336  -0.628454     -0.561356 |  0.79354  0.774967  0.733289        0.6753 | 0.895363  0.888271  0.875768      0.862907 |   5.009028 |   5.139355    5.911119    8.224015           12.920795 |  15.906017 | 0.061249  0.064049  0.071035       0.081485 | 0.121332  0.131168  0.157124        0.199453 |   5.009028    5.261507    6.531011    9.862235   15.906017 |    5.0 |   0.009028 |  15.906017 |
    | 17:30 | 0.794158  0.852312   1.01263              1.28681 |  15.065807   16.416858   20.316557   27.581542 |  23.098855   23.691911   25.326844        28.122936 |  22.941581    23.52312   25.126304     27.868097 | 0.500593  0.518736  0.565371  0.636533 | 1.472914  1.465896  1.449061           1.42678 | -0.691681  -0.675592  -0.633639     -0.568903 | 0.794073  0.777263  0.738036      0.681884 | 0.895594  0.889108  0.877079      0.864282 |   5.007313 |   5.120347    5.809422    7.926786           12.305013 |  15.103658 | 0.061176  0.063696  0.070229       0.080304 | 0.121081  0.129912  0.154031        0.194472 |   5.007313    5.226247    6.365543    9.436624   15.103658 |    5.0 |   0.007313 |  15.103658 |
    | 18:00 | 0.792724  0.844843   0.99337             1.254277 |  15.032924   16.241447   19.834459    26.68021 |  23.084234   23.615743   25.130422        27.791166 |  22.927243   23.448431   24.933696      27.54277 | 0.500136  0.516446  0.559999   0.62854 | 1.473094  1.466768  1.450912          1.429079 | -0.692084  -0.677631  -0.638505     -0.576175 | 0.794537  0.779352  0.742517      0.688238 | 0.895796  0.889881  0.878348      0.865642 |   5.005918 |   5.103884    5.718582    7.655371           11.734542 |  14.357164 | 0.061112  0.063378  0.069473       0.079168 | 0.120862   0.12878  0.151151        0.189727 |   5.005918     5.19563    6.216948    9.045605   14.357164 |    5.0 |   0.005918 |  14.357164 |
    | 18:30 | 0.791481  0.838128  0.975493             1.223411 |  15.004436   16.084215   19.390327   25.834843 |  23.071559   23.547263   24.948115        27.476395 |  22.914814   23.381281    24.75493     27.234112 | 0.499739  0.514378   0.55496  0.620857 | 1.473251  1.467559  1.452669          1.431337 | -0.692433  -0.679471  -0.643063     -0.583169 |  0.79494   0.78125  0.746738      0.694364 | 0.895972  0.890593  0.879572      0.866986 |   5.004785 |   5.089633    5.637501    7.407645           11.206127 |  13.662677 | 0.061058   0.06309  0.068765       0.078077 | 0.120673  0.127762  0.148474        0.185212 |   5.004785    5.169064    6.083629    8.686537   13.662677 |    5.0 |   0.004785 |  13.662677 |
    | 19:00 | 0.790405  0.832099  0.958926             1.194156 |  14.979772    15.94344   18.981586   25.042376 |  23.060579   23.485783   24.779165        27.178047 |  22.904048   23.320995    24.58926     26.941557 | 0.499396  0.512513  0.550245   0.61348 | 1.473387  1.468275  1.454332          1.433549 | -0.692736  -0.681128  -0.647322     -0.589888 | 0.795291  0.782971  0.750708      0.700262 | 0.896125  0.891245  0.880749      0.868311 |   5.003865 |   5.077304    5.565182    7.181655           10.716755 |  13.016617 |  0.06101   0.06283  0.068103        0.07703 | 0.120509  0.126847  0.145989         0.18092 |   5.003865     5.14603    5.964124    8.356976   13.016617 |    5.0 |   0.003865 |  13.016617 |
    | 19:30 | 0.789473  0.826695  0.943596             1.166455 |  14.958432   15.817539   18.605805   24.299911 |  23.051076   23.430664   24.622828        26.895553 |   22.89473   23.266946   24.435959      26.66455 | 0.499098  0.510834   0.54584  0.606407 | 1.473505  1.468922  1.455902          1.435713 | -0.692999  -0.682618  -0.651295     -0.596331 | 0.795595  0.784527  0.754433      0.705932 | 0.896259  0.891842  0.881879      0.869616 |   5.003119 |   5.066643    5.500728    6.975605           10.263632 |  12.415664 | 0.060969  0.062597  0.067485       0.076027 | 0.120367  0.126027  0.143686        0.176846 |   5.003119    5.126072    5.857103    8.054668   12.415664 |    5.0 |   0.003119 |  12.415664 |
    | 20:00 | 0.788667  0.821855  0.929432             1.140254 |  14.939981   15.705063   18.260692   23.604705 |  23.042856   23.381314   24.478381        26.628354 |  22.886669   23.218554   24.294317     26.402539 |  0.49884  0.509326  0.541735  0.599633 | 1.473607  1.469505  1.457379          1.437823 | -0.693226  -0.683957  -0.654992       -0.6025 | 0.795859  0.785932  0.757921      0.711376 | 0.896375  0.892386  0.882959      0.870899 |   5.002515 |   5.057431    5.443326    6.787839             9.84417 |  11.856737 | 0.060933  0.062387  0.066909       0.075068 | 0.120244  0.125292  0.141556        0.172983 |   5.002515     5.10879    5.761351    7.777523   11.856737 |    5.0 |   0.002515 |  11.856737 |
    | 20:30 |  0.78797  0.817528  0.916364             1.115498 |  14.924036   15.604688   17.944081   22.954158 |   23.03575   23.337186   24.345119        26.375896 |  22.879701   23.175283   24.163643     26.154984 | 0.498617  0.507973  0.537916  0.593157 | 1.473695   1.47003  1.458765          1.439875 | -0.693422  -0.685156  -0.658426     -0.608396 | 0.796088  0.787198  0.761183      0.716597 | 0.896476  0.892882  0.883989       0.87216 |   5.002026 |   5.049474    5.392243    6.616837             9.45597 |  11.336976 | 0.060902  0.062199  0.066375       0.074151 | 0.120138  0.124634  0.139588        0.169324 |   5.002026    5.093835    5.675762    7.523611   11.336976 |    5.0 |   0.002026 |  11.336976 |
    | 21:00 | 0.787368  0.813664  0.904327             1.092135 |  14.910265   15.515205   17.653936   22.345809 |  23.029611   23.297777   24.222359        26.137631 |  22.873681    23.13664   24.043267     25.921346 | 0.498425  0.506761  0.534371  0.586973 | 1.473771  1.470501  1.460063          1.441867 | -0.693592  -0.686229   -0.66161     -0.614024 | 0.796286  0.788336  0.764225      0.721596 | 0.896564  0.893331  0.884968      0.873395 |   5.001631 |   5.042604    5.346818    6.461198            9.096807 |  10.853727 | 0.060875  0.062031  0.065879       0.073277 | 0.120046  0.124047  0.137773        0.165863 |   5.001631    5.080902    5.599327    7.291141   10.853727 |    5.0 |   0.001631 |  10.853727 |
    | 21:30 | 0.786848  0.810217  0.893254              1.07011 |  14.898377   15.435514   17.388334   21.777319 |   23.02431   23.262625   24.109443         25.91302 |  22.868483    23.10217   23.932544     25.701097 | 0.498258  0.505677  0.531087  0.581077 | 1.473837  1.470923  1.461273          1.443795 | -0.693739  -0.687189  -0.664556     -0.619385 | 0.796458  0.789359  0.767057      0.726375 |  0.89664  0.893738  0.885896      0.874603 |   5.001312 |   5.036676    5.306456    6.319633            8.764614 |  10.404524 | 0.060852  0.061881  0.065419       0.072444 | 0.119967  0.123523  0.136102        0.162593 |   5.001312    5.069723    5.531132    7.078455   10.404524 |    5.0 |   0.001312 |  10.404524 |
    | 22:00 |   0.7864  0.807146  0.883085             1.049371 |   14.88812   15.364613   17.145467    21.24647 |  23.019736   23.231306   24.005734        25.701529 |  22.863998    23.07146   23.830849     25.493713 | 0.498115  0.504709  0.528049  0.575465 | 1.473894  1.471301  1.462401          1.445657 | -0.693865  -0.688045  -0.667276     -0.624484 | 0.796606  0.790276   0.76969      0.730939 | 0.896706  0.894105  0.886774      0.875784 |   5.001054 |   5.031563    5.270618    6.190954            8.457474 |   9.987076 | 0.060832  0.061747  0.064995       0.071652 | 0.119899  0.123056  0.134567        0.159508 |   5.001054    5.060065    5.470343    6.884015    9.987076 |    5.0 |   0.001054 |   9.987076 |
    | 22:30 | 0.786013  0.804413  0.873758             1.029868 |  14.879274   15.301592   16.923636   20.751152 |   23.01579   23.203433   23.910619        25.502631 |  22.860129   23.044128    23.73758     25.298677 | 0.497991  0.503846  0.525246  0.570131 | 1.473944  1.471638  1.463447           1.44745 | -0.693975  -0.688808  -0.669783     -0.629326 | 0.796734  0.791097  0.772131       0.73529 | 0.896763  0.894436    0.8876      0.876935 |   5.000846 |   5.027155    5.238821    6.074068            8.173606 |   9.599254 | 0.060815  0.061627  0.064604         0.0709 |  0.11984  0.122641  0.133157        0.156602 |   5.000846    5.051727    5.416205    6.706393    9.599254 |    5.0 |   0.000846 |   9.599254 |
    | 23:00 | 0.785679  0.801983  0.865216             1.011548 |  14.871649   15.245629   16.721243    20.28936 |  23.012388   23.178654   23.823511        25.315803 |  22.856793    23.01983   23.652164     25.115477 | 0.497884  0.503077  0.522664  0.565071 | 1.473986   1.47194  1.464417          1.449173 | -0.694069  -0.689487  -0.672089     -0.633915 | 0.796845   0.79183   0.77439      0.739432 | 0.896812  0.894734  0.888378      0.878054 |   5.000679 |   5.023356     5.21063    5.967971            7.911355 |   9.239075 | 0.060801   0.06152  0.064244       0.070187 | 0.119789  0.122271  0.131865        0.153867 |   5.000679    5.044531    5.368034    6.544267    9.239075 |    5.0 |   0.000679 |   9.239075 |
    | 23:30 | 0.785392  0.799825  0.857405             0.994361 |  14.865078   15.195978    16.53679   19.859186 |  23.009456   23.156648   23.743849        25.140533 |  22.853918   22.998251   23.574049     24.943611 | 0.497792  0.502394   0.52029  0.560276 | 1.474022  1.472208  1.465313          1.450824 |  -0.69415  -0.690091  -0.674208     -0.638257 |  0.79694  0.792485  0.776478      0.743368 | 0.896854  0.895001  0.889106      0.879141 |   5.000544 |   5.020083    5.185656    5.871734            7.669182 |   8.904695 | 0.060788  0.061425  0.063913       0.069512 | 0.119745  0.121943  0.130682        0.151298 |   5.000544    5.038323    5.325209    6.396407    8.904695 |    5.0 |   0.000544 |   8.904695 |
    | 00:00 | 0.785144  0.797911  0.850271             0.978258 |  14.859419   15.151966   16.368871   19.458813 |  23.006931   23.137124   23.671098        24.976313 |  22.851441   22.979106   23.502711      24.78258 | 0.497712  0.501786  0.518111  0.555743 | 1.474054  1.472446   1.46614          1.452403 |  -0.69422  -0.690628   -0.67615     -0.642358 | 0.797022  0.793069  0.778403      0.747104 | 0.896891  0.895241  0.889787      0.880193 |   5.000436 |   5.017264    5.163546    5.784507            7.445656 |   8.594396 | 0.060777  0.061341  0.063609       0.068875 | 0.119708  0.121651  0.129602        0.148887 |   5.000436    5.032971    5.287171    6.261674    8.594396 |    5.0 |   0.000436 |   8.594396 |
    | 00:30 | 0.784931  0.796214  0.843765             0.963189 |  14.854546   15.112986   16.216173    19.08651 |  23.004756   23.119818   23.604749        24.822643 |  22.849309   22.962136    23.43765     24.631894 | 0.497644  0.501247  0.516115  0.551462 | 1.474081  1.472659    1.4669          1.453909 |  -0.69428  -0.691104  -0.677927     -0.646225 | 0.797093  0.793588  0.780174      0.750643 | 0.896923  0.895455  0.890421      0.881211 |   5.000349 |   5.014837    5.143986    5.705506            7.239443 |   8.306575 | 0.060767  0.061266  0.063331       0.068273 | 0.119675  0.121393  0.128615        0.146627 |   5.000349    5.028358    5.253411     6.13901    8.306575 |    5.0 |   0.000349 |   8.306575 |
    | 01:00 | 0.784747  0.794711  0.837839             0.949107 |  14.850352    15.07849   16.077467   18.740628 |  23.002884   23.104492    23.54432        24.679031 |  22.847473   22.947109   23.378395     24.491071 | 0.497585  0.500769  0.514289  0.547428 | 1.474105  1.472847  1.467598          1.455342 | -0.694332  -0.691526  -0.679551     -0.649866 | 0.797154   0.79405  0.781802      0.753991 |  0.89695  0.895646  0.891011      0.882192 |   5.000279 |   5.012749    5.126694    5.634009            7.049304 |   8.039736 | 0.060759    0.0612  0.063077       0.067707 | 0.119647  0.121164  0.127717        0.144513 |   5.000279    5.024384    5.223475    6.027433    8.039736 |    5.0 |   0.000279 |   8.039736 |
    | 01:30 | 0.784589  0.793381   0.83245             0.935964 |  14.846743   15.047987   15.951607   18.419597 |  23.001273   23.090933   23.489354        24.544995 |  22.845894   22.933812   23.324497     24.359638 | 0.497535  0.500345  0.512621  0.543633 | 1.474125  1.473014  1.468238          1.456702 | -0.694377    -0.6919  -0.681033     -0.653286 | 0.797207  0.794461  0.783295      0.757153 | 0.896974  0.895817  0.891558      0.883137 |   5.000223 |   5.010952    5.111417    5.569353            6.874085 |   7.792482 | 0.060752  0.061141  0.062845       0.067175 | 0.119623  0.120962  0.126899        0.142537 |   5.000223    5.020961     5.19695    5.926033    7.792482 |    5.0 |   0.000223 |   7.792482 |
    | 02:00 | 0.784453  0.792206  0.827553             0.923713 |  14.843638   15.021036   15.837526   18.121918 |  22.999887   23.078945   23.439423        24.420059 |  22.844535   22.922057   23.275535     24.237128 | 0.497491   0.49997  0.511101  0.540068 | 1.474142  1.473161  1.468823           1.45799 | -0.694415   -0.69223  -0.682383     -0.656494 | 0.797252  0.794824  0.784662      0.760135 | 0.896994  0.895969  0.892063      0.884043 |   5.000178 |   5.009407     5.09793     5.51093            6.712707 |   7.563508 | 0.060746  0.061089  0.062634       0.066675 | 0.119602  0.120783  0.126156        0.140693 |   5.000178    5.018015    5.173467    5.833964    7.563508 |    5.0 |   0.000178 |   7.563508 |
    | 02:30 | 0.784337  0.791167  0.823111             0.912309 |  14.840969   14.997241   15.734229   17.846165 |  22.998695   23.068356   23.394121        24.303759 |  22.843366   22.911674   23.231113     24.123087 | 0.497454  0.499639  0.509718  0.536725 | 1.474157  1.473292  1.469357          1.459207 | -0.694448  -0.692522   -0.68361     -0.659499 | 0.797292  0.795147  0.785912      0.762941 | 0.897012  0.896104   0.89253      0.884912 |   5.000142 |   5.008078     5.08603     5.45818             6.56417 |   7.351593 | 0.060741  0.061044  0.062442       0.066208 | 0.119584  0.120625  0.125482        0.138976 |   5.000142    5.015479    5.152692    5.750446    7.351593 |    5.0 |   0.000142 |   7.351593 |
    | 03:00 | 0.784236  0.790251  0.819086             0.901707 |  14.838674   14.976247   15.640796   17.590978 |  22.997671    23.05901   23.353069        24.195641 |  22.842361   22.902509   23.190858     24.017068 | 0.497421  0.499347   0.50846  0.533596 |  1.47417  1.473408  1.469844          1.460354 | -0.694476   -0.69278  -0.684725     -0.662307 | 0.797325  0.795433  0.787052      0.765578 | 0.897027  0.896224   0.89296      0.885741 |   5.000113 |   5.006935    5.075538    5.410591            6.427537 |   7.155592 | 0.060736  0.061003  0.062267        0.06577 | 0.119569  0.120485   0.12487        0.137377 |   5.000113    5.013297    5.134326    5.674754    7.155592 |    5.0 |   0.000113 |   7.155592 |
    | 03:30 |  0.78415  0.789443  0.815442             0.891864 |  14.836701   14.957737   15.556369   17.355064 |   22.99679   23.050767   23.315914        24.095262 |  22.841498   22.894426   23.154425     23.918639 | 0.497394  0.499088  0.507319  0.530673 | 1.474181   1.47351  1.470287          1.461432 | -0.694501  -0.693008  -0.685736     -0.664929 | 0.797354  0.795685  0.788091      0.768052 |  0.89704   0.89633  0.893355      0.886532 |    5.00009 |   5.005954    5.066293     5.36769            6.301937 |   6.974435 | 0.060733  0.060967  0.062109       0.065361 | 0.119556  0.120362  0.124317        0.135892 |    5.00009     5.01142    5.118103    5.606219    6.974435 |    5.0 |    0.00009 |   6.974435 |
    | 04:00 | 0.784076   0.78873  0.812149             0.882737 |  14.835006   14.941429   15.480158   17.137189 |  22.996033   23.043501   23.282324        24.002191 |  22.840756   22.887302   23.121487     23.827374 |  0.49737  0.498861  0.506285  0.527945 |  1.47419    1.4736  1.470689          1.462445 | -0.694522  -0.693208  -0.686652     -0.667371 | 0.797379  0.795908  0.789036      0.770369 | 0.897051  0.896424  0.893717      0.887284 |   5.000072 |    5.00511    5.058151    5.329046            6.186558 |   6.807116 | 0.060729  0.060936  0.061965        0.06498 | 0.119545  0.120254  0.123816        0.134513 |   5.000072    5.009807    5.103782     5.54422    6.807116 |    5.0 |   0.000072 |   6.807116 |
    | 04:30 | 0.784012  0.788103  0.809174             0.874287 |   14.83355    14.92707    15.41143   16.936182 |  22.995383   23.037102   23.251991        23.916008 |  22.840118   22.881027   23.091743     23.742865 | 0.497349   0.49866  0.505349  0.525406 | 1.474198  1.473679  1.471054          1.463393 |  -0.69454  -0.693385   -0.68748     -0.669642 |   0.7974  0.796105  0.789894      0.772536 | 0.897061  0.896508  0.894048      0.887998 |   5.000057 |   5.004385    5.050985    5.294265            6.080642 |   6.652695 | 0.060726  0.060908  0.061835       0.064626 | 0.119535  0.120158  0.123364        0.133236 |   5.000057     5.00842    5.091148    5.488185    6.652695 |    5.0 |   0.000057 |   6.652695 |
    | 05:00 | 0.783957   0.78755  0.806491             0.866471 |    14.8323   14.914435   15.349512   16.750932 |  22.994825    23.03147    23.22463        23.836308 |  22.839571   22.875504   23.064914     23.664713 | 0.497332  0.498483  0.504503  0.523045 | 1.474205  1.473749  1.471384          1.464279 | -0.694555  -0.693541  -0.688228     -0.671751 | 0.797419  0.796279  0.790673      0.774559 | 0.897069  0.896581  0.894351      0.888673 |   5.000046 |   5.003763    5.044682    5.262985            5.983482 |   6.510285 | 0.060724  0.060883  0.061718       0.064297 | 0.119527  0.120074  0.122956        0.132054 |   5.000046    5.007228    5.080012    5.437588    6.510285 |    5.0 |   0.000046 |   6.510285 |
    | 05:30 |  0.78391  0.787065  0.804074             0.859253 |  14.831226   14.903323    15.29378   16.580379 |  22.994345   23.026516   23.199976        23.762698 |    22.8391   22.870646   23.040738     23.592532 | 0.497317  0.498328  0.503739  0.520853 | 1.474211  1.473811  1.471682          1.465105 | -0.694569  -0.693678  -0.688903     -0.673707 | 0.797434  0.796432  0.791378      0.776444 | 0.897076  0.896647  0.894626      0.889311 |   5.000036 |   5.003228    5.039141    5.234876             5.89442 |   6.379059 | 0.060722  0.060862  0.061612       0.063991 | 0.119519      0.12  0.122589        0.130961 |   5.000036    5.006203      5.0702    5.391942    6.379059 |    5.0 |   0.000036 |   6.379059 |
    | 06:00 |  0.78387  0.786638  0.801898             0.852595 |  14.830303   14.893557   15.243664   16.423524 |  22.993934   23.022161   23.177784          23.6948 |  22.838697   22.866376   23.018977     23.525953 | 0.497304  0.498191   0.50305  0.518822 | 1.474216  1.473865  1.471952          1.465874 |  -0.69458  -0.693799  -0.689512     -0.675518 | 0.797448  0.796567  0.792016      0.778198 | 0.897082  0.896704  0.894877      0.889911 |   5.000029 |   5.002769    5.034272    5.209635            5.812841 |   6.258236 |  0.06072  0.060843  0.061516       0.063708 | 0.119513  0.119935  0.122258        0.129953 |   5.000029    5.005324    5.061562    5.350799    6.258236 |    5.0 |   0.000029 |   6.258236 |
    | 06:30 | 0.783835  0.786262  0.799941             0.846462 |  14.829512   14.884979   15.198638   16.279415 |   22.99358   23.018335   23.157827         23.63225 |   22.83835   22.862624   22.999408     23.464618 | 0.497293  0.498071   0.50243  0.516943 | 1.474221  1.473912  1.472195          1.466588 |  -0.69459  -0.693904  -0.690059     -0.677191 |  0.79746  0.796686  0.792592      0.779827 | 0.897087  0.896755  0.895105      0.890476 |   5.000023 |   5.002376    5.029996    5.186987            5.738174 |   6.147086 | 0.060719  0.060826   0.06143       0.063446 | 0.119508  0.119878   0.12196        0.129024 |   5.000023    5.004568    5.053962    5.313748    6.147086 |    5.0 |   0.000023 |   6.147086 |
    | 07:00 | 0.783805  0.785933  0.798183             0.840818 |  14.828832   14.877448   15.158221   16.147156 |  22.993277   23.014975     23.1399          23.5747 |  22.838053    22.85933   22.981828     23.408185 | 0.497283  0.497965  0.501872  0.515208 | 1.474225  1.473954  1.472414           1.46725 | -0.694598  -0.693997  -0.690552     -0.678735 |  0.79747  0.796791  0.793112      0.781338 | 0.897092    0.8968  0.895312      0.891006 |   5.000018 |   5.002037    5.026243    5.166681            5.669885 |    6.04492 | 0.060717  0.060812  0.061353       0.063205 | 0.119503  0.119828  0.121692        0.128168 |   5.000018    5.003919    5.047278    5.280412     6.04492 |    5.0 |   0.000018 |    6.04492 |
    | 07:30 |  0.78378  0.785644  0.796605             0.835633 |  14.828249   14.870839   15.121973   16.025897 |  22.993016   23.012027   23.123809        23.521813 |  22.837797   22.856438    22.96605     23.356326 | 0.497275  0.497873  0.501371  0.513607 | 1.474228  1.473991  1.472611          1.467863 | -0.694605  -0.694079  -0.690995     -0.680159 | 0.797478  0.796883  0.793581      0.782737 | 0.897096  0.896839  0.895499      0.891501 |   5.000014 |   5.001747    5.022951    5.148489            5.607479 |   5.951096 | 0.060716  0.060799  0.061284       0.062982 |   0.1195  0.119784  0.121452        0.127382 |   5.000014    5.003362    5.041403    5.250443    5.951096 |    5.0 |   0.000014 |   5.951096 |
    | 08:00 | 0.783758   0.78539   0.79519             0.830873 |  14.827749   14.865043   15.089491   15.914839 |  22.992793   23.009441   23.109381        23.473273 |  22.837578   22.853903   22.951902     23.308728 | 0.497268  0.497791  0.500921  0.512132 | 1.474231  1.474023  1.472788          1.468429 | -0.694612  -0.694151  -0.691392     -0.681468 | 0.797485  0.796963  0.794002      0.784031 | 0.897099  0.896874  0.895668      0.891964 |   5.000011 |   5.001498    5.020064    5.132202            5.550494 |   5.865007 | 0.060715  0.060788  0.061221       0.062777 | 0.119496  0.119745  0.121237         0.12666 |   5.000011    5.002884    5.036243    5.223524    5.865007 |    5.0 |   0.000011 |   5.865007 |
    | 08:30 | 0.783739  0.785168  0.793923             0.826509 |   14.82732   14.859963   15.060407   15.813228 |  22.992601   23.007173   23.096455        23.428774 |   22.83739   22.851679   22.939227     23.265093 | 0.497262   0.49772  0.500518  0.510777 | 1.474233  1.474051  1.472947          1.468951 | -0.694617  -0.694213  -0.691748     -0.682672 | 0.797492  0.797034  0.794382      0.785225 | 0.897102  0.896904  0.895821      0.892395 |   5.000009 |   5.001285    5.017534    5.117632            5.498499 |   5.786087 | 0.060714  0.060778  0.061165       0.062589 | 0.119493  0.119711  0.121044        0.125997 |   5.000009    5.002473    5.031713    5.199365    5.786087 |    5.0 |   0.000009 |   5.786087 |
    | 09:00 | 0.783723  0.784973  0.792788             0.822514 |  14.826951   14.855511   15.034387   15.720353 |  22.992437   23.005187   23.084884        23.388029 |  22.837229   22.849731   22.927881     23.225139 | 0.497257  0.497658  0.500156  0.509531 | 1.474235  1.474076  1.473089          1.469432 | -0.694621  -0.694268  -0.692067     -0.683776 | 0.797497  0.797096  0.794722      0.786326 | 0.897104  0.896931  0.895958      0.892796 |   5.000007 |   5.001101    5.015317    5.104606            5.451096 |   5.713803 | 0.060714  0.060769  0.061115       0.062416 | 0.119491  0.119681  0.120871        0.125391 |   5.000007    5.002121    5.027738    5.177701    5.713803 |    5.0 |   0.000007 |   5.713803 |
    | 09:30 | 0.783709  0.784803  0.791773              0.81886 |  14.826636   14.851612   15.011126    15.63555 |  22.992296   23.003447   23.074536        23.350763 |  22.837091   22.848025   22.917734     23.188596 | 0.497252  0.497603  0.499832   0.50839 | 1.474237  1.474098  1.473217          1.469874 | -0.694625  -0.694316  -0.692352     -0.684788 | 0.797502  0.797151  0.795028      0.787339 | 0.897106  0.896954  0.896082      0.893169 |   5.000006 |   5.000944    5.013376     5.09297            5.407914 |   5.647658 | 0.060713  0.060762   0.06107       0.062257 | 0.119489  0.119655  0.120717        0.124836 |   5.000006    5.001819    5.024252     5.15829    5.647658 |    5.0 |   0.000006 |   5.647658 |
    | 10:00 | 0.783697  0.784653  0.790867             0.815521 |  14.826365     14.8482   14.990348   15.558194 |  22.992175   23.001923   23.065288        23.316718 |  22.836972   22.846531   22.908666     23.155213 | 0.497248  0.497555  0.499543  0.507344 | 1.474238  1.474117  1.473331          1.470279 | -0.694629  -0.694359  -0.692607     -0.685715 | 0.797506  0.797199  0.795302       0.78827 | 0.897108  0.896975  0.896194      0.893513 |   5.000004 |   5.000809    5.011677    5.082582            5.368608 |   5.587184 | 0.060713  0.060755   0.06103       0.062112 | 0.119487  0.119633  0.120579        0.124329 |   5.000004    5.001559    5.021196     5.14091    5.587184 |    5.0 |   0.000004 |   5.587184 |
    | 10:30 | 0.783687  0.784522  0.790057             0.812475 |  14.826132   14.845213   14.971801   15.487699 |  22.992071    23.00059    23.05703         23.28565 |   22.83687   22.845224   22.900568     23.124748 | 0.497245  0.497513  0.499285  0.506387 |  1.47424  1.474133  1.473433          1.470651 | -0.694632  -0.694396  -0.692835     -0.686561 | 0.797509  0.797241  0.795547      0.789124 |  0.89711  0.896993  0.896294      0.893832 |   5.000004 |   5.000694     5.01019    5.073314             5.33286 |   5.531946 | 0.060712  0.060749  0.060994       0.061979 | 0.119485  0.119613  0.120456        0.123866 |   5.000004    5.001337    5.018518    5.125362    5.531946 |    5.0 |   0.000004 |   5.531946 |
    | 11:00 | 0.783678  0.784408  0.789334             0.809698 |  14.825933   14.842601   14.955257   15.423519 |  22.991982   22.999424   23.049662        23.257329 |  22.836783   22.844081   22.893343     23.096978 | 0.497242  0.497476  0.499054  0.505514 | 1.474241  1.474148  1.473524          1.470991 | -0.694634  -0.694428  -0.693038     -0.687334 | 0.797512  0.797277  0.795767      0.789908 | 0.897111  0.897009  0.896384      0.894126 |   5.000003 |   5.000595     5.00889    5.065052            5.300372 |   5.481536 | 0.060712  0.060744  0.060962       0.061858 | 0.119484  0.119595  0.120346        0.123444 |   5.000003    5.001146    5.016173    5.111463    5.481536 |    5.0 |   0.000003 |   5.481536 |
    | 11:30 | 0.783671  0.784308   0.78869             0.807169 |  14.825762   14.840318    14.94051   15.365143 |  22.991906   22.998405   23.043092        23.231541 |  22.836708   22.843081     22.8869      23.07169 |  0.49724  0.497444  0.498848  0.504717 | 1.474242  1.474161  1.473605          1.471302 | -0.694636  -0.694456   -0.69322     -0.688039 | 0.797515  0.797309  0.795963      0.790624 | 0.897112  0.897023  0.896465      0.894397 |   5.000002 |    5.00051    5.007753     5.05769            5.270872 |   5.435573 | 0.060711   0.06074  0.060934       0.061747 | 0.119483   0.11958  0.120248        0.123059 |   5.000002    5.000983    5.014119    5.099047    5.435573 |    5.0 |   0.000002 |   5.435573 |
    | 12:00 | 0.783664  0.784221  0.788116             0.804869 |  14.825616   14.838322   14.927375   15.312097 |   22.99184   22.997514   23.037238        23.208082 |  22.836644   22.842207    22.88116     23.048686 | 0.497238  0.497416  0.498664   0.50399 | 1.474243  1.474172  1.473678          1.471586 | -0.694638  -0.694481  -0.693382     -0.688682 | 0.797517  0.797337  0.796138       0.79128 | 0.897113  0.897035  0.896537      0.894647 |   5.000002 |   5.000437     5.00676    5.051136            5.244105 |   5.393703 | 0.060711  0.060736  0.060909       0.061647 | 0.119482  0.119567   0.12016        0.122709 |   5.000002    5.000842    5.012323    5.087963    5.393703 |    5.0 |   0.000002 |   5.393703 |
    | 12:30 | 0.783659  0.784144  0.787605             0.802778 |   14.82549   14.836578   14.915683   15.263939 |  22.991784   22.996735   23.032026        23.186764 |  22.836589   22.841444    22.87605     23.027783 | 0.497236  0.497392  0.498501  0.503329 | 1.474243  1.474181  1.473743          1.471844 |  -0.69464  -0.694502  -0.693526     -0.689266 | 0.797519  0.797362  0.796295      0.791878 | 0.897114  0.897046  0.896601      0.894876 |   5.000001 |   5.000374    5.005892    5.045303            5.219837 |   5.355596 | 0.060711  0.060732  0.060886       0.061555 | 0.119481  0.119555  0.120082        0.122391 |   5.000001    5.000722    5.010751    5.078077    5.355596 |    5.0 |   0.000001 |   5.355596 |
    | 13:00 | 0.783654  0.784078   0.78715             0.800881 |  14.825383   14.835055   14.905281   15.220256 |  22.991736   22.996055   23.027389        23.167411 |  22.836542   22.840777   22.871502     23.008805 | 0.497235   0.49737  0.498355  0.502728 | 1.474244   1.47419    1.4738          1.472079 | -0.694641  -0.694521  -0.693654     -0.689797 |  0.79752  0.797383  0.796434      0.792424 | 0.897115  0.897055  0.896659      0.895086 |   5.000001 |   5.000321    5.005134    5.040116            5.197853 |   5.320943 | 0.060711  0.060729  0.060866       0.061472 |  0.11948  0.119545  0.120013        0.122103 |   5.000001    5.000619    5.009377    5.069263    5.320943 |    5.0 |   0.000001 |   5.320943 |
    | 13:30 |  0.78365   0.78402  0.786746             0.799159 |   14.82529   14.833726   14.896034   15.180669 |  22.991695   22.995462   23.023265        23.149858 |  22.836502   22.840195   22.867459     22.991594 | 0.497233  0.497352  0.498226  0.502182 | 1.474244  1.474197  1.473851          1.472293 | -0.694642  -0.694538  -0.693768     -0.690279 | 0.797522  0.797402  0.796558       0.79292 | 0.897115  0.897063  0.896711      0.895278 |   5.000001 |   5.000275    5.004472    5.035505            5.177952 |   5.289459 |  0.06071  0.060727  0.060848       0.061396 |  0.11948  0.119536  0.119951        0.121841 |   5.000001     5.00053    5.008176    5.061412    5.289459 |    5.0 |   0.000001 |   5.289459 |
    | 14:00 | 0.783647  0.783969  0.786387               0.7976 |  14.825211   14.832566   14.887818   15.144824 |   22.99166   22.994944   23.019601        23.133954 |  22.836467   22.839687   22.863866     22.975998 | 0.497232  0.497335  0.498111  0.501687 | 1.474245  1.474204  1.473897          1.472488 | -0.694643  -0.694552  -0.693869     -0.690716 | 0.797523  0.797419  0.796669      0.793372 | 0.897116   0.89707  0.896757      0.895453 |   5.000001 |   5.000236    5.003895     5.03141            5.159952 |    5.26088 |  0.06071  0.060725  0.060832       0.061327 | 0.119479  0.119528  0.119897        0.121604 |   5.000001    5.000454    5.007127    5.054423     5.26088 |    5.0 |   0.000001 |    5.26088 |
    | 14:30 | 0.783644  0.783924  0.786067             0.796188 |  14.825143   14.831553   14.880522   15.112395 |  22.991629   22.994492   23.016347        23.119556 |  22.836437   22.839244   22.860675     22.961879 | 0.497231  0.497321  0.498008  0.501239 | 1.474245  1.474209  1.473937          1.472664 | -0.694644  -0.694565  -0.693959     -0.691112 | 0.797524  0.797433  0.796768      0.793783 | 0.897116  0.897077  0.896798      0.895614 |   5.000001 |   5.000202    5.003391    5.027775            5.143682 |   5.234959 |  0.06071  0.060723  0.060818       0.061265 | 0.119479  0.119522  0.119848        0.121389 |   5.000001    5.000389    5.006211    5.048204    5.234959 |    5.0 |   0.000001 |   5.234959 |
    | 15:00 | 0.783641  0.783886  0.785784             0.794911 |  14.825085    14.83067   14.874048   15.083082 |  22.991603   22.994097   23.013458        23.106533 |  22.836412   22.838857   22.857842      22.94911 |  0.49723  0.497309  0.497918  0.500832 | 1.474246  1.474214  1.473973          1.472824 | -0.694645  -0.694575  -0.694039      -0.69147 | 0.797525  0.797445  0.796855      0.794156 | 0.897117  0.897082  0.896834       0.89576 |        5.0 |   5.000173    5.002952    5.024549            5.128988 |   5.211469 |  0.06071  0.060721  0.060805       0.061209 | 0.119478  0.119516  0.119805        0.121194 |        5.0    5.000334    5.005411    5.042675    5.211469 |    5.0 |        0.0 |   5.211469 |
    | 15:30 | 0.783639  0.783852  0.785533             0.793757 |  14.825036     14.8299   14.868305   15.056606 |  22.991581   22.993754   23.010896        23.094765 |   22.83639    22.83852    22.85533      22.93757 |  0.49723  0.497298  0.497837  0.500465 | 1.474246  1.474219  1.474005          1.472968 | -0.694645  -0.694585   -0.69411     -0.691794 | 0.797525  0.797456  0.796933      0.794494 | 0.897117  0.897087  0.896867      0.895893 |        5.0 |   5.000148    5.002569    5.021688            5.115726 |   5.190199 |  0.06071  0.060719  0.060794       0.061158 | 0.119478  0.119511  0.119767        0.121019 |        5.0    5.000286    5.004712    5.037762    5.190199 |    5.0 |        0.0 |   5.190199 |
    | 16:00 | 0.783637  0.783823   0.78531             0.792715 |  14.824993   14.829229   14.863215   15.032712 |  22.991562   22.993454   23.008625        23.084139 |  22.836371   22.838227   22.853103     22.927151 | 0.497229  0.497289  0.497766  0.500133 | 1.474246  1.474222  1.474033          1.473099 | -0.694646  -0.694593  -0.694173     -0.692087 | 0.797526  0.797466  0.797002        0.7948 | 0.897117  0.897091  0.896896      0.896014 |        5.0 |   5.000127    5.002235    5.019153            5.103766 |   5.170956 |  0.06071  0.060718  0.060784       0.061112 | 0.119478  0.119506  0.119733         0.12086 |        5.0    5.000245    5.004103    5.033399    5.170956 |    5.0 |        0.0 |   5.170956 |
    | 16:30 | 0.783636  0.783797  0.785113             0.791775 |  14.824956   14.828645   14.858704   15.011165 |  22.991546   22.993193   23.006612        23.074553 |  22.836355    22.83797   22.851129     22.917751 | 0.497229   0.49728  0.497702  0.499833 | 1.474246  1.474226  1.474058          1.473217 | -0.694646  -0.694601  -0.694229     -0.692352 | 0.797527  0.797474  0.797063      0.795078 | 0.897118  0.897095  0.896921      0.896124 |        5.0 |   5.000109    5.001944    5.016907            5.092987 |    5.15356 |  0.06071  0.060717  0.060775        0.06107 | 0.119478  0.119502  0.119703        0.120717 |        5.0     5.00021    5.003572    5.029527     5.15356 |    5.0 |        0.0 |    5.15356 |
    | 17:00 | 0.783634  0.783775  0.784938             0.790928 |  14.824925   14.828135    14.85471   14.991749 |  22.991532   22.992965   23.004829        23.065912 |  22.836342   22.837747   22.849381     22.909277 | 0.497228  0.497273  0.497646  0.499563 | 1.474246  1.474229  1.474081          1.473324 | -0.694647  -0.694607  -0.694278      -0.69259 | 0.797527  0.797481  0.797118      0.795328 | 0.897118  0.897098  0.896944      0.896224 |        5.0 |   5.000093    5.001691    5.014918             5.08328 |   5.137846 |  0.06071  0.060716  0.060768       0.061033 | 0.119477  0.119499  0.119676        0.120588 |        5.0     5.00018    5.003108    5.026091    5.137846 |    5.0 |        0.0 |   5.137846 |
    | 17:30 | 0.783633  0.783755  0.784783             0.790164 |  14.824898   14.827692   14.851175   14.974266 |   22.99152   22.992767   23.003252        23.058128 |   22.83633   22.837553   22.847834     22.901645 | 0.497228  0.497267  0.497597  0.499319 | 1.474247  1.474231    1.4741           1.47342 | -0.694647  -0.694612  -0.694322     -0.692805 | 0.797527  0.797488  0.797166      0.795555 | 0.897118    0.8971  0.896965      0.896314 |        5.0 |    5.00008     5.00147    5.013158            5.074545 |   5.123664 |  0.06071  0.060715  0.060761       0.060999 | 0.119477  0.119496  0.119652        0.120472 |        5.0    5.000154    5.002705    5.023046    5.123664 |    5.0 |        0.0 |   5.123664 |
    | 18:00 | 0.783632  0.783738  0.784646             0.789478 |  14.824875   14.827306   14.848048   14.958536 |   22.99151   22.992595   23.001856        23.051122 |   22.83632   22.837384   22.846465     22.894775 | 0.497227  0.497262  0.497553  0.499099 | 1.474247  1.474233  1.474118          1.473506 | -0.694647  -0.694617  -0.694361     -0.692998 | 0.797528  0.797493  0.797209      0.795759 | 0.897118  0.897103  0.896982      0.896396 |        5.0 |   5.000068    5.001278    5.011601            5.066688 |   5.110872 |  0.06071  0.060714  0.060755       0.060969 | 0.119477  0.119493  0.119632        0.120367 |        5.0    5.000132    5.002353    5.020347    5.110872 |    5.0 |        0.0 |   5.110872 |
    | 18:30 | 0.783631  0.783724  0.784525              0.78886 |  14.824856   14.826969   14.845283   14.944393 |  22.991501   22.992445   23.000621        23.044822 |  22.836311   22.837237   22.845255     22.888597 | 0.497227  0.497257  0.497514  0.498902 | 1.474247  1.474235  1.474133          1.473584 | -0.694647  -0.694621  -0.694395     -0.693172 | 0.797528  0.797498  0.797246      0.795943 | 0.897118  0.897105  0.896998       0.89647 |        5.0 |   5.000059     5.00111    5.010224            5.059627 |   5.099344 |  0.06071  0.060714  0.060749       0.060941 | 0.119477  0.119491  0.119613        0.120273 |        5.0    5.000113    5.002046    5.017956    5.099344 |    5.0 |        0.0 |   5.099344 |
    | 19:00 |  0.78363  0.783711  0.784418             0.788304 |  14.824839   14.826677   14.842839   14.931686 |  22.991493   22.992314    22.99953        23.039159 |  22.836304   22.837109   22.844185     22.883044 | 0.497227  0.497253   0.49748  0.498724 | 1.474247  1.474237  1.474147          1.473654 | -0.694648  -0.694625  -0.694425     -0.693329 | 0.797528  0.797502   0.79728      0.796109 | 0.897118  0.897107  0.897013      0.896537 |        5.0 |    5.00005    5.000965    5.009008            5.053285 |   5.088962 |  0.06071  0.060713  0.060745       0.060917 | 0.119477  0.119489  0.119597        0.120189 |        5.0    5.000097    5.001779    5.015841    5.088962 |    5.0 |        0.0 |   5.088962 |
    | 19:30 |  0.78363    0.7837  0.784324             0.787806 |  14.824824   14.826422   14.840679   14.920276 |  22.991487     22.9922   22.998566        23.034074 |  22.836297   22.836997   22.843239     22.878058 | 0.497227  0.497249  0.497449  0.498565 | 1.474247  1.474238  1.474159          1.473717 | -0.694648  -0.694628  -0.694452     -0.693469 | 0.797529  0.797506  0.797309      0.796258 | 0.897118  0.897108  0.897025      0.896597 |        5.0 |   5.000043    5.000838    5.007933            5.047593 |    5.07962 |  0.06071  0.060713   0.06074       0.060895 | 0.119477  0.119487  0.119583        0.120113 |        5.0    5.000083    5.001546    5.013968     5.07962 |    5.0 |        0.0 |    5.07962 |
    | 20:00 | 0.783629   0.78369   0.78424             0.787358 |  14.824812   14.826201   14.838773   14.910038 |  22.991481   22.992102   22.997715         23.02951 |  22.836292     22.8369   22.842405     22.873582 | 0.497227  0.497246  0.497423  0.498422 | 1.474247  1.474239  1.474169          1.473774 | -0.694648  -0.694631  -0.694475     -0.693595 | 0.797529  0.797509  0.797336      0.796393 | 0.897118   0.89711  0.897036      0.896651 |        5.0 |   5.000037    5.000728    5.006984            5.042487 |   5.071218 | 0.060709  0.060712  0.060737       0.060875 | 0.119477  0.119486   0.11957        0.120045 |        5.0    5.000071    5.001344    5.012313    5.071218 |    5.0 |        0.0 |   5.071218 |
    | 20:30 | 0.783629  0.783682  0.784167             0.786957 |  14.824801   14.826008    14.83709   14.900858 |  22.991476   22.992016   22.996964        23.025416 |  22.836287   22.836816   22.841668     22.869568 | 0.497226  0.497243  0.497399  0.498293 | 1.474247   1.47424  1.474179          1.473825 | -0.694648  -0.694633  -0.694496     -0.693709 | 0.797529  0.797512  0.797359      0.796513 | 0.897119  0.897111  0.897046        0.8967 |        5.0 |   5.000032    5.000632    5.006146            5.037909 |   5.063667 | 0.060709  0.060712  0.060733       0.060857 | 0.119477  0.119485  0.119559        0.119984 |        5.0    5.000061    5.001168    5.010849    5.063667 |    5.0 |        0.0 |   5.063667 |
    | 21:00 | 0.783628  0.783674  0.784102             0.786597 |  14.824792   14.825841   14.835605   14.892631 |  22.991472   22.991941   22.996301        23.021748 |  22.836283   22.836743   22.841018     22.865971 | 0.497226  0.497241  0.497378  0.498178 | 1.474247  1.474241  1.474187           1.47387 | -0.694648  -0.694635  -0.694514      -0.69381 | 0.797529  0.797514  0.797379      0.796622 | 0.897119  0.897112  0.897055      0.896744 |        5.0 |   5.000027    5.000549    5.005407            5.033808 |   5.056886 | 0.060709  0.060711   0.06073       0.060841 | 0.119477  0.119484  0.119549        0.119929 |        5.0    5.000052    5.001015    5.009556    5.056886 |    5.0 |        0.0 |   5.056886 |
    | 21:30 | 0.783628  0.783668  0.784044             0.786275 |  14.824784   14.825695   14.834295   14.885263 |  22.991469   22.991876   22.995716        23.018461 |   22.83628   22.836679   22.840445     22.862748 | 0.497226  0.497239   0.49736  0.498075 | 1.474247  1.474242  1.474194          1.473911 | -0.694648  -0.694637  -0.694531     -0.693901 | 0.797529  0.797516  0.797397      0.796719 | 0.897119  0.897113  0.897062      0.896784 |        5.0 |   5.000023    5.000476    5.004756            5.030136 |   5.050801 | 0.060709  0.060711  0.060728       0.060827 | 0.119476  0.119483   0.11954         0.11988 |        5.0    5.000045    5.000881    5.008414    5.050801 |    5.0 |        0.0 |   5.050801 |
    | 22:00 | 0.783628  0.783662  0.783994             0.785986 |  14.824778   14.825569   14.833141   14.878669 |  22.991466   22.991819     22.9952         23.01552 |  22.836277   22.836624   22.839939     22.859864 | 0.497226  0.497237  0.497344  0.497982 | 1.474247  1.474243  1.474201          1.473948 | -0.694648  -0.694639  -0.694545     -0.693982 | 0.797529  0.797518  0.797413      0.796806 | 0.897119  0.897114  0.897069       0.89682 |        5.0 |    5.00002    5.000414    5.004181             5.02685 |   5.045342 | 0.060709  0.060711  0.060726       0.060814 | 0.119476  0.119482  0.119532        0.119836 |        5.0    5.000038    5.000765    5.007406    5.045342 |    5.0 |        0.0 |   5.045342 |
    | 22:30 | 0.783627  0.783658  0.783949             0.785728 |  14.824772   14.825459   14.832123    14.87277 |  22.991463    22.99177   22.994746        23.012888 |  22.836275   22.836575   22.839494     22.857283 | 0.497226  0.497236  0.497329    0.4979 | 1.474247  1.474243  1.474206           1.47398 | -0.694648   -0.69464  -0.694557     -0.694055 | 0.797529   0.79752  0.797427      0.796885 | 0.897119  0.897114  0.897075      0.896852 |        5.0 |   5.000017    5.000359    5.003675            5.023912 |    5.04045 | 0.060709  0.060711  0.060724       0.060803 | 0.119476  0.119481  0.119525        0.119796 |        5.0    5.000033    5.000664    5.006517     5.04045 |    5.0 |        0.0 |    5.04045 |
    | 23:00 | 0.783627  0.783653   0.78391             0.785498 |  14.824767   14.825363   14.831227   14.867496 |  22.991461   22.991728   22.994346        23.010535 |  22.836272   22.836534   22.839101     22.854976 | 0.497226  0.497234  0.497317  0.497826 | 1.474247  1.474244  1.474211           1.47401 | -0.694648  -0.694641  -0.694569      -0.69412 | 0.797529  0.797521   0.79744      0.796955 | 0.897119  0.897115   0.89708       0.89688 |        5.0 |   5.000015    5.000311    5.003229            5.021285 |   5.036068 | 0.060709  0.060711  0.060722       0.060792 | 0.119476   0.11948  0.119519        0.119761 |        5.0    5.000028    5.000577    5.005732    5.036068 |    5.0 |        0.0 |   5.036068 |
    | 23:30 | 0.783627   0.78365  0.783876             0.785291 |  14.824763   14.825281   14.830438   14.862784 |  22.991459   22.991691   22.993994        23.008432 |  22.836271   22.836497   22.838756     22.852914 | 0.497226  0.497233  0.497306   0.49776 | 1.474247  1.474244  1.474216          1.474036 | -0.694649  -0.694642  -0.694578     -0.694179 | 0.797529  0.797522  0.797451      0.797017 | 0.897119  0.897116  0.897085      0.896906 |        5.0 |   5.000012     5.00027    5.002836            5.018938 |   5.032144 | 0.060709   0.06071   0.06072       0.060783 | 0.119476   0.11948  0.119514         0.11973 |        5.0    5.000024    5.000501     5.00504    5.032144 |    5.0 |        0.0 |   5.032144 |

>>> round_(model.check_waterbalance(conditions))
-0.000044
"""
# import...
# ...from standard library
from typing import *

# ...from site-packages
import numpy

# ...from HydPy
from hydpy.exe.modelimports import *
from hydpy.core import masktools
from hydpy.core import modeltools
from hydpy.core.typingtools import *

# ...from musk
from hydpy.models.musk import musk_masks
from hydpy.models.musk import musk_model
from hydpy.models.musk import musk_solver


class NmbRuns(musk_solver.NmbRuns):
    """The number of (repeated) runs of the |RunModel.RUN_METHODS| of the |musk_mct|
    model per simulation step [-]."""

    INIT = 2


class Model(modeltools.SegmentModel):
    """The Muskingum-Cunge-Todini version of *HydPy-Musk* (|musk_mct|)."""

    SOLVERPARAMETERS = (
        NmbRuns,
        musk_solver.ToleranceWaterLevel,
        musk_solver.ToleranceDischarge,
    )
    INLET_METHODS = (
        musk_model.Pick_Inflow_V1,
        musk_model.Update_Discharge_V1,
    )
    RECEIVER_METHODS = ()
    RUN_METHODS = (
        musk_model.Calc_ReferenceDischarge_V1,
        musk_model.Calc_ReferenceWaterLevel_V1,
        musk_model.Calc_WettedArea_V1,
        musk_model.Calc_WettedPerimeter_V1,
        musk_model.Calc_SurfaceWidth_V1,
        musk_model.Calc_Celerity_V1,
        musk_model.Calc_CorrectingFactor_V1,
        musk_model.Calc_CourantNumber_V1,
        musk_model.Calc_ReynoldsNumber_V1,
        musk_model.Calc_Coefficient1_Coefficient2_Coefficient3_V1,
        musk_model.Calc_Discharge_V2,
    )
    ADD_METHODS = (
        musk_model.Return_WettedArea_V1,
        musk_model.Return_WettedPerimeter_V1,
        musk_model.Return_SurfaceWidth_V1,
        musk_model.Return_Discharge_V1,
        musk_model.Return_ReferenceDischargeError_V1,
        musk_model.Return_Celerity_V1,
    )
    OUTLET_METHODS = (
        musk_model.Calc_Outflow_V1,
        musk_model.Pass_Outflow_V1,
    )
    SENDER_METHODS = ()
    SUBMODELS = (musk_model.PegasusReferenceWaterLevel,)

    def prepare_states(self, discharge: float, tolerance: float = 1e-6) -> None:
        """Determine the Courant and cell Reynolds numbers of all channel segments for
        the given discharge under the assumption of stationarity.

        We prepare a channel consisting of three segments.  The first and the last
        segment are equal to those defined in the :ref:`musk_mct_base_example`:

        >>> from hydpy.models.musk_mct import *
        >>> simulationstep("30m")
        >>> parameterstep()
        >>> catchmentarea(25000.0)
        >>> nmbsegments(3)
        >>> length(2.0)
        >>> bottomwidth(15.0, 0.0, 15.0)
        >>> sideslope(5.0, 1.0, 5.0)
        >>> bottomslope(0.00025, 0.01, 0.00025)
        >>> stricklercoefficient(1.0/0.035, 20.0, 1.0/0.035)

        We apply |Model.prepare_states| for a constant discharge of 100 m³/s:

        >>> model.prepare_states(discharge=100.0)
        >>> states.courantnumber
        courantnumber(0.7207, 2.846072, 0.7207)
        >>> states.reynoldsnumber
        reynoldsnumber(2.591116, 0.079081, 2.591116)

        The smoothness of the discharge simulated for the first timesteps of the
        :ref:`musk_mct_base_example` indicates that the Courant number and the cell
        Reynolds number are consistent with the given discharge.

        For zero discharge, both the Courant and the cell Reynolds number are zero
        (even within a triangle profile):

        >>> model.prepare_states(discharge=0.0)
        >>> states.courantnumber
        courantnumber(0.0, 0.0, 0.0)
        >>> states.reynoldsnumber
        reynoldsnumber(0.0, 0.0, 0.0)
        """
        self.parameters.update()
        states = self.sequences.states
        states.courantnumber.new, states.courantnumber.old = 0.5, 0.5
        states.reynoldsnumber.new, states.reynoldsnumber.old = 0.5, 0.5
        states.discharge.old = discharge
        states.discharge.new[0], states.discharge.new[1:] = discharge, numpy.inf
        while max(numpy.abs(discharge - states.discharge.new)) > tolerance:
            self.run()
            self.new2old()

    def check_waterbalance(
        self,
        initial_conditions: Dict[str, Dict[str, ArrayFloat]],
    ) -> float:
        r"""Determine the water balance error of the previous simulation run in %.

        Method |Model.check_waterbalance| calculates the balance error as follows:

        :math:`100 \cdot \frac{
        Seconds \cdot \sum_{t=t0}^{t1} \Big( Inflow_t - Outflow_t \Big) +
        1000 \cdot \sum_{s=1}^{NmbSegments} \Big( Length^s \cdot
        \big( WettedArea_{t0}^s - WettedArea_{t1}^s \big) \Big)
        }{
        Seconds \cdot \sum_{t=t0}^{t1} \Big( Inflow_t \Big) + 1000 \cdot
        \sum_{s=1}^{NmbSegments} \Big(Length^s \cdot  WettedArea_{t0}^s \Big)
        }`

        :cite:t:`ref-Todini2007` reports the relative volume error to be less than
        0.00 % in all performed experiments.  However, we are not sure
        |Model.check_waterbalance| calculates this error in the same way.  Please
        inform us if you encounter larger errors that are not due to preventing
        negative discharge (see :ref:`musk_mct_no_baseflow`).

        You can pick the required initial conditions before starting the simulation run
        via property |Sequences.conditions|.
        """

        seconds = self.parameters.derived.seconds.value
        inflow = seconds * sum(self.sequences.fluxes.inflow.series)
        outflow = seconds * sum(self.sequences.fluxes.outflow.series)

        self.idx_run = 1
        length = 1000.0 * self.parameters.control.length.values
        self.run_segments(self.calc_referencedischarge_v1)
        self.run_segments(self.calc_referencewaterlevel_v1)
        self.run_segments(self.calc_wettedarea_v1)
        volume1 = numpy.dot(length, self.sequences.factors.wettedarea.values)
        self.sequences.states.discharge = initial_conditions["states"]["discharge"]
        self.run_segments(self.calc_referencedischarge_v1)
        self.run_segments(self.calc_referencewaterlevel_v1)
        self.run_segments(self.calc_wettedarea_v1)
        volume0 = numpy.dot(length, self.sequences.factors.wettedarea.values)

        return (inflow - outflow + volume0 - volume1) / (inflow + volume0) / 100.0


class Masks(masktools.Masks):
    """Masks applicable to |musk_mct|."""

    CLASSES = musk_masks.Masks.CLASSES


tester = Tester()
cythonizer = Cythonizer()
